/*
 * Irq wake driver for Cashew Platform.
 *
 * Copyright (C) 2016 Imagination Technologies Ltd.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 */

#include <linux/interrupt.h>
#include <linux/irqchip/mips-gic.h>
#include <linux/irq.h>
#include <linux/of_platform.h>

#define NUM_SYSWAKES	3
#define EVENT_INVAL	(-1)

struct syswake {
	int irq;
	int event;
};

struct wake_mapping {
	u32 gic_sirq;	/* gic shared irq */
	u32 event;
	u32 length;
};

static void __iomem *event_sel_reg;
static struct syswake syswakes[NUM_SYSWAKES];
static struct wake_mapping *wake_map;
static int wake_map_sz;

bool cashew_irqwake_is_set(void)
{
	int wake_num;

	for (wake_num = 0; wake_num < NUM_SYSWAKES; wake_num++)
		if (syswakes[wake_num].event != EVENT_INVAL)
			return true;

	return false;
}

static int gic_sirq_to_event(u32 gic_sirq)
{
	int i;

	for (i = 0; i < wake_map_sz; i++)
		if (gic_sirq >= wake_map[i].gic_sirq &&
		    gic_sirq < wake_map[i].gic_sirq + wake_map[i].length)
			return wake_map[i].event +
			       gic_sirq - wake_map[i].gic_sirq;

	/* Not found */
	return EVENT_INVAL;
}

static int cashew_irq_set_wake(struct irq_data *data, unsigned int on)
{
	const unsigned int gic_sirq = GIC_HWIRQ_TO_SHARED(data->hwirq);
	const unsigned int event = gic_sirq_to_event(gic_sirq);
	int wake_num, ret;
	u32 reg;

	if (event == EVENT_INVAL || event > 0x1F)
		return -EINVAL;

	/* Find whether a syswake has been configured for this event */
	for (wake_num = 0; wake_num < NUM_SYSWAKES; wake_num++)
		if (syswakes[wake_num].event == event)
			break;

	if (on) {
		if (wake_num == NUM_SYSWAKES) {
			/* Find a free syswake */
			for (wake_num = 0; wake_num < NUM_SYSWAKES; wake_num++)
				if (syswakes[wake_num].event == EVENT_INVAL)
					break;
		}

		if (wake_num == NUM_SYSWAKES) {
			/* All syswakes in use */
			return -EBUSY;
		}

		syswakes[wake_num].event = event;

		/* Set CR_ALWAYS_ON_EVENT_SEL register */
		reg = ioread32(event_sel_reg);
		reg &= ~(0x1F << 8 * wake_num);
		reg |= event << 8 * wake_num;
		iowrite32(reg, event_sel_reg);
	} else {
		if (wake_num == NUM_SYSWAKES) {
			/* syswake not found */
			return -EINVAL;
		}

		/* Mark the syswake as free */
		syswakes[wake_num].event = EVENT_INVAL;
	}

	/* Pass on the request to the syswake irq controller */
	ret = irq_set_irq_wake(syswakes[wake_num].irq, on);

	if (ret)
		syswakes[wake_num].event = EVENT_INVAL;

	return ret;
}

static int cashew_irqwake_probe(struct platform_device *pdev)
{
	const struct device_node *np = pdev->dev.of_node;
	struct resource *res_regs;
	struct device_node *gic_np;
	struct irq_domain *gic_domain;
	struct irq_chip *chip;
	int ret, i, j, length;
	unsigned int virq;

	/* Get resources */
	res_regs = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!res_regs) {
		dev_err(&pdev->dev, "no register resource\n");
		return -EINVAL;
	}
	event_sel_reg = devm_ioremap_nocache(&pdev->dev, res_regs->start,
					     res_regs->end - res_regs->start);
	if (!event_sel_reg)
		return -ENOMEM;

	/* Get syswake irqs */
	for (i = 0; i < NUM_SYSWAKES; i++) {
		syswakes[i].irq = platform_get_irq(pdev, i);
		if (syswakes[i].irq < 0) {
			dev_err(&pdev->dev, "cannot find IRQ #%u\n", i);
			return syswakes[i].irq;
		}
		syswakes[i].event = EVENT_INVAL; /* Mark as unsued */
	}

	/* Check if too many were provided */
	if (platform_get_irq(pdev, i) >= 0) {
		dev_err(&pdev->dev, "surplus IRQs detected\n");
		return -EINVAL;
	}

	/* Get gic domain */
	gic_np = of_parse_phandle(np, "gic", 0);
	if (!gic_np) {
		dev_err(&pdev->dev,
			"missing 'gic' node\n");
		return -EINVAL;
	}

	gic_domain = irq_find_host(gic_np);
	if (!gic_domain) {
		dev_err(&pdev->dev, "unable to obtain GIC domain\n");
		return -EINVAL;
	}

	/* Get irq wake map */
	if (!of_find_property(np, "irq-wake-map", &length)) {
		dev_err(&pdev->dev, "missing 'irq-wake-map' property\n");
		return -EINVAL;
	}

	if (length % sizeof(struct wake_mapping)) {
		dev_err(&pdev->dev, "invalid 'irq-wake-map' property\n");
		return -EINVAL;
	}

	wake_map = devm_kzalloc(&pdev->dev, length, GFP_KERNEL);
	if (!wake_map)
		return -ENOMEM;

	ret = of_property_read_u32_array(np, "irq-wake-map",
					 &wake_map->gic_sirq,
					 length / sizeof(u32));
	if (ret) {
		dev_err(&pdev->dev, "cannot read 'irq-wake-map' property\n");
		return ret;
	}
	wake_map_sz = length / sizeof(struct wake_mapping);

	/* Override gic irq_set_wake functions */
	for (i = 0; i < wake_map_sz; i++) {
		for (j = 0; j < wake_map[i].length; j++) {
			virq = irq_find_mapping(gic_domain,
						wake_map[i].gic_sirq + j);
			chip = irq_get_chip(virq);
			chip->irq_set_wake = cashew_irq_set_wake;
		}
	}

	return 0;
}

static int cashew_irqwake_remove(struct platform_device *pdev)
{
	return 0;
}

static const struct of_device_id cashew_irqwake_match[] = {
	{ .compatible = "img,cashew-irq-wake" },
	{}
};

static struct platform_driver cashew_irqwake_driver = {
	.probe		= cashew_irqwake_probe,
	.remove		= cashew_irqwake_remove,
	.driver		= {
		.name	= "cashew-irq-wake",
		.of_match_table	= of_match_ptr(cashew_irqwake_match),
	}
};
module_platform_driver(cashew_irqwake_driver);
