/*
 * This file is subject to the terms and conditions of the GNU General Public
 * License. See the file "COPYING" in the main directory of this archive for
 * more details.
 *
 * Copyright (C) 2016 Imagination Technologies Ltd.
 */

#include <linux/init.h>
#include <linux/pm.h>
#include <linux/slab.h>
#include <linux/suspend.h>

#include <asm/cacheflush.h>
#include <asm/tlbflush.h>

#define EXC_VECTOR_SIZE		0x400

extern bool cashew_irqwake_is_set(void);
extern void cashew_sleep(bool irqwake_set);

/**
 * struct cashew_suspend_state - State to save across suspend to RAM.
 * @ebase:	CP0 EBase register.
 * @excvec:	Memory for saving exception vector.
 *
 * This structure contains state that must be saved across suspend to RAM, but
 * can be done so by C code and allocated dynamically.
 */
struct cashew_suspend_state {
	u32 ebase;
	void *excvec;
};

/* Suspend state storage */
struct cashew_suspend_state *cashew_suspend_state;

struct cashew_suspend_state *cashew_alloc_suspend_state(void)
{
	struct cashew_suspend_state *s;

	s = kzalloc(sizeof(*s), GFP_KERNEL);
	if (!s)
		return NULL;

	/* Allocate memory to save exception vector */
	s->excvec = kmalloc(EXC_VECTOR_SIZE, GFP_KERNEL);
	if (!s->excvec) {
		kfree(s);
		return NULL;
	}

	return s;
}

void cashew_free_suspend_state(struct cashew_suspend_state *s)
{
	kfree(s->excvec);
	kfree(s);
}

void cashew_save_suspend_state(struct cashew_suspend_state *s)
{
	void *ebase_addr;

	/* Save exception vector */
	s->ebase = read_c0_ebase();
	ebase_addr = (void *)(long)(s32)(s->ebase & 0xfffff000);
	memcpy(s->excvec, ebase_addr, EXC_VECTOR_SIZE);
}

void cashew_restore_suspend_state(const struct cashew_suspend_state *s)
{
	void *ebase_addr;

	/* Restore CP0 state */
	write_c0_ebase(s->ebase);

	/* Restore exception vector */
	ebase_addr = (void *)(long)(s32)(s->ebase & 0xfffff000);
	memcpy(ebase_addr, s->excvec, EXC_VECTOR_SIZE);

	/* flush exception vector from icache so it gets used */
	local_flush_icache_range((unsigned long)ebase_addr,
				 (unsigned long)ebase_addr + EXC_VECTOR_SIZE);

	/* flush any pre-existing TLB entries */
	local_flush_tlb_all();
}

static int cashew_pm_enter(suspend_state_t state)
{
	bool irqwake_set = cashew_irqwake_is_set();

#ifdef CONFIG_SMP
	/* Disable Core0 coherence */
	write_gcr_cl_coherence(0);
#endif /* CONFIG_SMP */

	/* Save important CPU state */
	cashew_save_suspend_state(cashew_suspend_state);

	/*
	 * Perform the sleep.
	 * After wake, this will return.
	 */
	cashew_sleep(irqwake_set);

	/* Restore important CPU state */
	cashew_restore_suspend_state(cashew_suspend_state);

#ifdef CONFIG_SMP
	/* Make core 0 coherent with everything */
	write_gcr_cl_coherence(0xff);
#endif /* CONFIG_SMP */

	return 0;
}

static int cashew_pm_begin(suspend_state_t state)
{
	cashew_suspend_state = cashew_alloc_suspend_state();
	if (!cashew_suspend_state)
		return -ENOMEM;

	return 0;
}

static void cashew_pm_end(void)
{
	cashew_free_suspend_state(cashew_suspend_state);
}

static const struct platform_suspend_ops cashew_pm_ops = {
	.valid		= suspend_valid_only_mem,
	.begin		= cashew_pm_begin,
	.enter		= cashew_pm_enter,
	.end		= cashew_pm_end,
};

/*
 * Initialize suspend interface
 */
static int __init cashew_pm_init(void)
{
	suspend_set_ops(&cashew_pm_ops);

	return 0;
}

late_initcall(cashew_pm_init);
