/*
 * This file is subject to the terms and conditions of the GNU General Public
 * License. See the file "COPYING" in the main directory of this archive for
 * more details.
 *
 * Copyright (C) 2016 Imagination Technologies Ltd.
 */

#include <linux/clk.h>
#include <linux/delay.h>
#include <linux/of.h>
#include <linux/of_address.h>
#include <linux/platform_device.h>
#include <linux/pm_domain.h>
#include <linux/slab.h>

/* Registers */
#define SOC_POWER		0x0
#define SOC_RESETN		0x4
#define SOC_DELAY		0xc

#define SOC_DELAY_SHIFT		0
#define FUNC_DELAY_SHIFT	4
#define CLK_DELAY_SHIFT		8
#define CLKRST_DELAY_SHIFT	12
#define ISO_DELAY_SHIFT		16
#define PWR_DELAY_SHIFT		20
#define RST_ISO_DELAY_SHIFT	24
#define SYS_CLK_DIV_SHIFT	28

#define REPAIR_TIMEOUTMS	30

/*
 * struct pdc_intc_priv
 * @genpd:	generic_pm_domain struct.
 * @base:	Base of the SOC_POWER PDC register.
 * @index:	The index of the power domain.
 * @soc_power:	Cached value of the SOC_POWER PDC register, required as reads
		of this register during a power transition may not get the
		last value written.
 * @lock:	Lock to protect writes to the SOC_POWER PDC register and cached
		value.
 * @on_timeout:	The timeout for power-on transitions in microseconds
 * @off_timeout:The timeout for power-off transitions in microseconds
 */
struct pdc_pm_domain {
	struct generic_pm_domain genpd;
	void __iomem *base;
	void __iomem *repair_resetn;
	unsigned int index;
	unsigned int *soc_power;
	spinlock_t *lock;
	unsigned int on_timeout;
	unsigned int off_timeout;
	struct clk *clk;
};

static inline struct pdc_pm_domain *to_pdc_pd(struct generic_pm_domain *d)
{
	return container_of(d, struct pdc_pm_domain, genpd);
}

static int pdc_pd_power_off(struct generic_pm_domain *genpd)
{
	struct pdc_pm_domain *pdc_pd = to_pdc_pd(genpd);
	unsigned int poll_count, soc_power;

	spin_lock(pdc_pd->lock);
	*pdc_pd->soc_power &= ~BIT(pdc_pd->index);
	iowrite32(*pdc_pd->soc_power, pdc_pd->base + SOC_POWER);
	spin_unlock(pdc_pd->lock);

	/* Poll for soc_power to take correct value */
	for (poll_count = pdc_pd->off_timeout; poll_count; poll_count--) {
		soc_power = ioread32(pdc_pd->base + SOC_POWER);
		if (!(soc_power & BIT(pdc_pd->index)))
			break;
		udelay(1);
	}
	if (!poll_count)
		return -EIO;

	return 0;
}

static int pdc_pd_power_on(struct generic_pm_domain *genpd)
{
	struct pdc_pm_domain *pdc_pd = to_pdc_pd(genpd);
	unsigned int poll_count, resetn;
	int ret = 0;

	/* hold clock for duration of power-on */
	if (pdc_pd->clk) {

		ret = clk_prepare_enable(pdc_pd->clk);
		if (ret)
			return ret;
	}

	spin_lock(pdc_pd->lock);
	*pdc_pd->soc_power |= BIT(pdc_pd->index);
	iowrite32(*pdc_pd->soc_power, pdc_pd->base + SOC_POWER);
	spin_unlock(pdc_pd->lock);

	/* Poll for soc_resetn to take correct value */
	for (poll_count = pdc_pd->on_timeout; poll_count; poll_count--) {
		resetn = ioread32(pdc_pd->base + SOC_RESETN);
		if (resetn & BIT(pdc_pd->index))
			break;
		udelay(1);
	}
	if (!poll_count) {
		ret = -EIO;
		goto out;
	}

	/* For domains with memory repair, also poll repair_resetn */
	if (pdc_pd->index < 13) {
		for (poll_count = REPAIR_TIMEOUTMS; poll_count; poll_count--) {
			resetn = ioread32(pdc_pd->repair_resetn);
			if (resetn & BIT(pdc_pd->index + 2))
				break;
			usleep_range(1000, 2000);
		}
		if (!poll_count) {
			ret = -EIO;
			goto out;
		}
	}

out:
	if (pdc_pd->clk)
		clk_disable_unprepare(pdc_pd->clk);

	return ret;
}

static bool pdc_pd_is_off(struct generic_pm_domain *genpd)
{
	struct pdc_pm_domain *pdc_pd = to_pdc_pd(genpd);

	return !(*pdc_pd->soc_power & BIT(pdc_pd->index));
}

static bool pdc_pd_active_wakeup(struct device *dev)
{
	return true;
}

static int pdc_pd_probe(struct platform_device *pdev)
{
	void __iomem *pdc_base, *repair_resetn;
	unsigned int *soc_power;
	spinlock_t *lock;
	struct device_node *pmd;
	struct device_node *domain;
	struct pdc_pm_domain *pdc_pd;
	u32 index;
	struct generic_pm_domain *genpd;
	bool is_off;
	struct clk *sys_clk;
	unsigned long clk_rate;
	u32 soc_delay, func_delay, clk_delay, clkrst_delay;
	u32 iso_delay, pwr_delay, rst_iso_delay, val, reg;
	unsigned int on_timeout, off_timeout;

	if (!pdev->dev.of_node)
		return -ENOENT;

	/* Ioremap the registers */
	pdc_base = of_iomap(pdev->dev.of_node, 0);
	if (!pdc_base)
		return -ENOMEM;

	repair_resetn = of_iomap(pdev->dev.of_node, 1);
	if (!repair_resetn)
		return -ENOMEM;

	sys_clk = devm_clk_get(&pdev->dev, "sys");
	if (IS_ERR(sys_clk)) {
		dev_err(&pdev->dev, "could not get sys clock resource\n");
		return PTR_ERR(sys_clk);
	}
	clk_rate = clk_get_rate(sys_clk);

	if (of_property_read_u32(pdev->dev.of_node, "soc_delay",
				 &soc_delay))
		soc_delay = 3;
	if (of_property_read_u32(pdev->dev.of_node, "func_delay",
				 &func_delay))
		func_delay = 3;
	if (of_property_read_u32(pdev->dev.of_node, "clk_delay",
				 &clk_delay))
		clk_delay = 3;
	if (of_property_read_u32(pdev->dev.of_node, "clkrst_delay",
				 &clkrst_delay))
		clkrst_delay = 6;
	if (of_property_read_u32(pdev->dev.of_node, "iso_delay",
				 &iso_delay))
		iso_delay = 6;
	if (of_property_read_u32(pdev->dev.of_node, "pwr_delay",
				 &pwr_delay))
		pwr_delay = 3;
	if (of_property_read_u32(pdev->dev.of_node, "rst_iso_delay",
				 &rst_iso_delay))
		rst_iso_delay = 6;

	/* Set clock divider to 1 to select sys_clk/2 as the timer source */
	reg = 1 << SYS_CLK_DIV_SHIFT;
	clk_rate /= 2;

	/* Convert delays in microseconds to log2(sys_clk_cycles) rounded up */
	val = fls(soc_delay * (clk_rate / 1000000));
	if (val > 0xf) {
		dev_err(&pdev->dev, "soc_delay property out of range\n");
		return -EINVAL;
	}
	reg |= val << SOC_DELAY_SHIFT;
	val = fls(func_delay * (clk_rate / 1000000));
	if (val > 0xf) {
		dev_err(&pdev->dev, "func_delay property out of range\n");
		return -EINVAL;
	}
	reg |= val << FUNC_DELAY_SHIFT;
	val = fls(clk_delay * (clk_rate / 1000000));
	if (val > 0xf) {
		dev_err(&pdev->dev, "clk_delay property out of range\n");
		return -EINVAL;
	}
	reg |= val << CLK_DELAY_SHIFT;
	val = fls(clkrst_delay * (clk_rate / 1000000));
	if (val > 0xf) {
		dev_err(&pdev->dev, "clkrst_delay property out of range\n");
		return -EINVAL;
	}
	reg |= val << CLKRST_DELAY_SHIFT;
	val = fls(iso_delay * (clk_rate / 1000000));
	if (val > 0xf) {
		dev_err(&pdev->dev, "iso_delay property out of range\n");
		return -EINVAL;
	}
	reg |= val << ISO_DELAY_SHIFT;
	val = fls(pwr_delay * (clk_rate / 1000000));
	if (val > 0xf) {
		dev_err(&pdev->dev, "pwr_delay property out of range\n");
		return -EINVAL;
	}
	reg |= val << PWR_DELAY_SHIFT;
	val = fls(rst_iso_delay * (clk_rate / 1000000));
	if (val > 0xf) {
		dev_err(&pdev->dev, "rst_iso_delay property out of range\n");
		return -EINVAL;
	}
	reg |= val << RST_ISO_DELAY_SHIFT;
	iowrite32(reg, pdc_base + SOC_DELAY);

	/* Set timeouts to 8x the expected transition time */
	on_timeout = max3(func_delay, clkrst_delay, iso_delay) + rst_iso_delay;
	on_timeout <<= 3;
	off_timeout = rst_iso_delay +
		      max3(func_delay, clkrst_delay, iso_delay) + pwr_delay;
	off_timeout <<= 3;

	/* Allocate and initialise shared data */
	soc_power = devm_kzalloc(&pdev->dev, sizeof(*soc_power), GFP_KERNEL);
	if (!soc_power)
		return -ENOMEM;
	*soc_power = ioread32(pdc_base + SOC_POWER);

	lock = devm_kzalloc(&pdev->dev, sizeof(*lock), GFP_KERNEL);
	if (!lock)
		return -ENOMEM;
	spin_lock_init(lock);

	/* Add domains */
	pmd = of_get_child_by_name(pdev->dev.of_node, "pm-domains");
	if (!pmd) {
		dev_err(&pdev->dev, "No pm-domains node found\n");
		return -EINVAL;
	}

	for_each_child_of_node(pmd, domain) {
		const char *clk_name = 0;
		struct clk *clk = 0;

		if (of_property_read_u32(domain, "index", &index)) {
			dev_err(&pdev->dev,
				"No index found for pm-domain %s\n",
				domain->name);
			of_node_put(domain);
			return -EINVAL;
		}
		if (index < 2 || index > 14) {
			dev_err(&pdev->dev,
				"Index %d out of range for pm-domain %s\n",
				index, domain->name);
			of_node_put(domain);
			return -EINVAL;
		}

		if (of_property_read_string(domain, "clk", &clk_name) == 0) {

			clk = devm_clk_get(&pdev->dev, clk_name);
			if (IS_ERR(clk)) {
				dev_err(&pdev->dev, "could not get %s clock resource for pm-domain %s\n",
					clk_name, domain->name);
				of_node_put(domain);
				return PTR_ERR(clk);
			}
		}

		pdc_pd = devm_kzalloc(&pdev->dev, sizeof(*pdc_pd), GFP_KERNEL);
		if (!pdc_pd) {
			of_node_put(domain);
			return -ENOMEM;
		}

		pdc_pd->genpd.name = domain->name;
		pdc_pd->base = pdc_base;
		pdc_pd->repair_resetn = repair_resetn;
		pdc_pd->index = index;
		pdc_pd->soc_power = soc_power;
		pdc_pd->lock = lock;
		pdc_pd->on_timeout = on_timeout;
		pdc_pd->off_timeout = off_timeout;
		pdc_pd->clk = clk;

		genpd = &pdc_pd->genpd;
		is_off = pdc_pd_is_off(genpd);
		pm_genpd_init(genpd, NULL, is_off);

		genpd->dev_ops.active_wakeup	= pdc_pd_active_wakeup;
		genpd->power_off		= pdc_pd_power_off;
		genpd->power_on			= pdc_pd_power_on;
		of_genpd_add_provider_simple(domain, genpd);
	}
	of_node_put(pmd);

	return 0;
}

static const struct of_device_id pdc_pd_match[] = {
	{ .compatible = "img,pdc-pd" },
	{}
};

static struct platform_driver pdc_pd_driver = {
	.driver = {
		.name		= "pdc-pd",
		.of_match_table	= pdc_pd_match,
	},
	.probe = pdc_pd_probe,
};

static int __init pdc_pd_init(void)
{
	return platform_driver_register(&pdc_pd_driver);
}
subsys_initcall(pdc_pd_init);
