/*
 * Imagination Technologies Random Number Generator (RNG)
 *
 * Copyright (C) 2016 Imagination Technologies Ltd.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 */

#include <linux/clk.h>
#include <linux/err.h>
#include <linux/hw_random.h>
#include <linux/io.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/of.h>
#include <linux/platform_device.h>
#include <linux/pm_runtime.h>

#define IMG_RNG_SEED_127_96		0x00

#define IMG_RNG_SEED_95_64		0x04

#define IMG_RNG_SEED_63_32		0x08

#define IMG_RNG_SEED_31_0		0x0C

#define IMG_RNG_CONTROL			0x10
#define IMG_RNG_CONTROL_SNOW3G_INIT	BIT(0)
#define IMG_RNG_CONTROL_ENABLE		BIT(1)
#define IMG_RNG_CONTROL_RING_DISABLE	BIT(2)
#define IMG_RNG_CONTROL_RESET_ERROR	BIT(3)
#define IMG_RNG_CONTROL_ENT_DISABLE	BIT(4)
#define IMG_RNG_CONTROL_ENT_RESET	BIT(5)
#define IMG_RNG_CONTROL_MODE		BIT(6)

#define IMG_RNG_RAND_NUM_STAT_0		0x14
#define IMG_RNG_RAND_VAL_MASK		0x1
#define IMG_RNG_RAND_VAL_WAIT_MS	50

#define IMG_RNG_RAND_NUM_STAT_1		0x18

#define IMG_RNG_RAND_ERROR		0x1C

#define IMG_RNG_RAND_READ		0x20
#define IMG_RNG_RAND_READ_ENABLE	1
#define IMG_RNG_RAND_READ_DISABLE	0

#define IMG_RNG_SEED_VAL		0x0

#define IMG_RNG_AUTOSUSPEND_DELAY_MS	100

struct img_rng {
	struct device *dev;
	struct hwrng rng;
	void __iomem *regs;
	struct clk *clk;
};

static u32 img_rng_readl(struct img_rng *irng, u32 read_from)
{
	return readl(irng->regs + read_from);
}

static void img_rng_writel(struct img_rng *irng, u32 val, u32 write_to)
{
	writel(val, irng->regs + write_to);
}

static void img_rng_config(struct img_rng *irng)
{
	/* load 128-bit seed */
	img_rng_writel(irng, IMG_RNG_SEED_VAL, IMG_RNG_SEED_31_0);
	img_rng_writel(irng, IMG_RNG_SEED_VAL, IMG_RNG_SEED_63_32);
	img_rng_writel(irng, IMG_RNG_SEED_VAL, IMG_RNG_SEED_95_64);
	img_rng_writel(irng, IMG_RNG_SEED_VAL, IMG_RNG_SEED_127_96);

	/* Enable RNG */
	img_rng_writel(irng, IMG_RNG_CONTROL_SNOW3G_INIT, IMG_RNG_CONTROL);
}

static int img_rng_init(struct hwrng *rng)
{
	struct img_rng *irng = container_of(rng, struct img_rng, rng);

	pm_runtime_get_sync(irng->dev);
	img_rng_config(irng);
	pm_runtime_mark_last_busy(irng->dev);
	pm_runtime_put_autosuspend(irng->dev);

	return 0;
}

static int img_rng_read(struct hwrng *rng, void *data, size_t max, bool wait)
{
	struct img_rng *irng = container_of(rng, struct img_rng, rng);
	u32 no_of_reads = DIV_ROUND_UP(max, sizeof(u32));
	u32 last_chunk_mask = 0xFFFFFFFF >> ((4 - (max % sizeof(u32))) * 8);
	u32 bytes_remaining = max;
	u32 read_count = 0;
	u32 reg_val;
	u32 rand_val;
	unsigned long timeout = jiffies +
				msecs_to_jiffies(IMG_RNG_RAND_VAL_WAIT_MS);

	pm_runtime_get_sync(irng->dev);

	/* Enable RNG and start number generation */
	reg_val = img_rng_readl(irng, IMG_RNG_CONTROL) | IMG_RNG_CONTROL_ENABLE;
	img_rng_writel(irng, reg_val, IMG_RNG_CONTROL);
	while (no_of_reads) {
		/* Clear previous rand_val read */
		rand_val = 0;
		/* wait for a valid random value */
		while (time_before(jiffies, timeout) && (!rand_val)) {
			rand_val = img_rng_readl(irng, IMG_RNG_RAND_NUM_STAT_0)
				& IMG_RNG_RAND_VAL_MASK;
		}
		if (rand_val) {
			/* enable RAND_READ to indicate output is being read */
			img_rng_writel(irng, IMG_RNG_RAND_READ_ENABLE,
					IMG_RNG_RAND_READ);
			/* read random number output */
			reg_val = img_rng_readl(irng, IMG_RNG_RAND_NUM_STAT_1);
			/* mask read value if this is the last read */
			if (no_of_reads == 1)
				reg_val &= last_chunk_mask;
			memcpy(data + read_count, &reg_val,
						sizeof(u32));
			/* disable RAND_READ to indicate reading is done */
			img_rng_writel(irng, IMG_RNG_RAND_READ_DISABLE,
					IMG_RNG_RAND_READ);
			no_of_reads--;
			read_count += 4;
			bytes_remaining -= (bytes_remaining % 4) ?
					bytes_remaining : 4;
		} else {
			dev_err(irng->dev,
				"Waiting for valid number Timed out\n");
			no_of_reads = 0;
		}
	}
	/* Disable RNG */
	reg_val = img_rng_readl(irng, IMG_RNG_CONTROL)
				& !((u32)IMG_RNG_CONTROL_ENABLE);
	img_rng_writel(irng, reg_val, IMG_RNG_CONTROL);

	pm_runtime_mark_last_busy(irng->dev);
	pm_runtime_put_sync_autosuspend(irng->dev);

	return read_count;
}

static int img_rng_runtime_suspend(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);
	struct img_rng *irng = platform_get_drvdata(pdev);

	clk_disable_unprepare(irng->clk);

	return 0;
}

static int img_rng_runtime_resume(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);
	struct img_rng *irng = platform_get_drvdata(pdev);
	int error;

	error = clk_prepare_enable(irng->clk);
	if (error)
		dev_err(dev, "Clock enable failed\n");

	return error;
}

static int img_rng_probe(struct platform_device *pdev)
{
	struct img_rng *irng;
	struct resource *res;
	unsigned int error;

	irng = devm_kzalloc(&pdev->dev, sizeof(irng), GFP_KERNEL);
	if (!irng)
		return -ENOMEM;

	irng->dev = &pdev->dev;
	irng->rng.name = "img-rng";

	irng->clk = devm_clk_get(&pdev->dev, "rng_clk");

	if (IS_ERR(irng->clk))
		return PTR_ERR(irng->clk);

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	irng->regs = devm_ioremap_resource(irng->dev, res);
	if (IS_ERR(irng->regs))
		return PTR_ERR(irng->regs);

	irng->rng.init = img_rng_init;
	irng->rng.read = img_rng_read;
	platform_set_drvdata(pdev, irng);

	pm_runtime_set_autosuspend_delay(&pdev->dev,
					 IMG_RNG_AUTOSUSPEND_DELAY_MS);
	pm_runtime_use_autosuspend(&pdev->dev);
	pm_runtime_enable(&pdev->dev);
	if (!pm_runtime_enabled(&pdev->dev)) {
		error = img_rng_runtime_resume(&pdev->dev);
		if (error)
			return error;
	}

	error = devm_hwrng_register(&pdev->dev, &irng->rng);
	if (error) {
		if (!pm_runtime_enabled(&pdev->dev))
			img_rng_runtime_suspend(&pdev->dev);
		pm_runtime_disable(&pdev->dev);
		pm_runtime_dont_use_autosuspend(&pdev->dev);
	}

	return error;
}

static int img_rng_remove(struct platform_device *pdev)
{
	struct img_rng *irng = platform_get_drvdata(pdev);

	devm_hwrng_unregister(&pdev->dev, &irng->rng);

	pm_runtime_disable(&pdev->dev);
	pm_runtime_dont_use_autosuspend(irng->dev);
	if (!pm_runtime_status_suspended(&pdev->dev))
		img_rng_runtime_suspend(&pdev->dev);

	return 0;
}

#ifdef CONFIG_PM_SLEEP
static int img_rng_suspend(struct device *dev)
{
	return pm_runtime_force_suspend(dev);
}

static int img_rng_resume(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);
	struct img_rng *irng = platform_get_drvdata(pdev);
	int ret;

	ret = pm_runtime_force_resume(dev);
	if (ret)
		return ret;

	img_rng_config(irng);

	return 0;
}
#endif /* CONFIG_PM_SLEEP */

static const struct dev_pm_ops img_rng_pm_ops = {
	SET_SYSTEM_SLEEP_PM_OPS(img_rng_suspend, img_rng_resume)
	SET_RUNTIME_PM_OPS(img_rng_runtime_suspend,
			   img_rng_runtime_resume, NULL)
};

static const struct of_device_id img_rng_of_match[] = {
	{ .compatible = "img,img-rng", },
	{ },
};
MODULE_DEVICE_TABLE(of, img_rng_of_match);

static struct platform_driver img_rng_driver = {
	.driver = {
		.name = "img,img-rng",
		.pm = &img_rng_pm_ops,
		.of_match_table = of_match_ptr(img_rng_of_match),
	},
	.probe = img_rng_probe,
	.remove = img_rng_remove,
};
module_platform_driver(img_rng_driver);

MODULE_DESCRIPTION("IMG Random Number Generator (RNG) driver");
MODULE_AUTHOR("Seraj Mohammed <seraj.mohammed@imgtec.com>");
MODULE_LICENSE("GPL v2");
