/*
 * Cashew SoC clock controllers
 * Copyright (C) 2015 Imagination Technologies, Ltd.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 */

#include <linux/clk-provider.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/io.h>
#include <linux/kernel.h>
#include <linux/of.h>
#include <linux/slab.h>
#include <dt-bindings/clock/cashew-clk.h>
#include "clk.h"
#include "clock_connections.h"

#ifdef CONFIG_PM_SLEEP
#include <linux/syscore_ops.h>
static void __iomem *clk_regs_base;
static const u32 ctrls_idx[] = {
	PLL0_CONTROL_A_OFFSET,
	PLL0_CONTROL_B_OFFSET,
	PLL1_CONTROL_A_OFFSET,
	PLL1_CONTROL_B_OFFSET,
	PLL2_CONTROL_A_OFFSET,
	PLL2_CONTROL_B_OFFSET,
	PLL3_CONTROL_A_OFFSET,
	PLL3_CONTROL_B_OFFSET,
	PLL4_CONTROL_A_OFFSET,
	PLL4_CONTROL_B_OFFSET,
	PLL5_CONTROL_A_OFFSET,
	PLL5_CONTROL_B_OFFSET,
	PLL6_CONTROL_A_OFFSET,
	PLL6_CONTROL_B_OFFSET,
	PLL7_CONTROL_A_OFFSET,
	PLL7_CONTROL_B_OFFSET,
};

static const u32 status_idx[] = {
	PLL0_STATUS_OFFSET,
	PLL1_STATUS_OFFSET,
	PLL2_STATUS_OFFSET,
	PLL3_STATUS_OFFSET,
	PLL4_STATUS_OFFSET,
	PLL5_STATUS_OFFSET,
	PLL6_STATUS_OFFSET,
	PLL7_STATUS_OFFSET,
};

static u32 clk_regs_saved[PLL7_STATUS_OFFSET / 4 + 1];

static int cashew_clk_suspend(void)
{
	int idx;

	/* Save the register values */
	for (idx = 0; idx <= PLL7_STATUS_OFFSET; idx += 4)
		clk_regs_saved[idx / 4] = readl(clk_regs_base + idx);

	return 0;
}

static void cashew_clk_resume(void)
{
	int idx, i;
	int count;

	/* Restore the register values */
	for (idx = 0; idx <= CLOCK_NOC_CIP3_OFFSET; idx += 4)
		writel(clk_regs_saved[idx / 4], clk_regs_base + idx);

	for (i = 0; i < ARRAY_SIZE(ctrls_idx); i++) {
		idx = ctrls_idx[i];
		writel(clk_regs_saved[idx / 4],
		       clk_regs_base + idx);
	}

	/* Wait for PLLs that were locked prior to suspend to re-lock */
	for (i = 0; i < ARRAY_SIZE(status_idx); i++) {
		idx = status_idx[i];
		if (clk_regs_saved[idx / 4] & BIT(0)) {
			for (count = 0; count < 10; count++) {
				if (readl(clk_regs_base + idx) & BIT(0))
					break;
				udelay(5);
			}
			if (count == 10)
				pr_err("PLL%d did not lock\n", i);
		}
	}
}
static struct syscore_ops cashew_clk_syscore_ops = {
	.suspend = cashew_clk_suspend,
	.resume = cashew_clk_resume,
};

static void cashew_clk_sleep_init(void __iomem *reg_base)
{
	clk_regs_base = reg_base;
	register_syscore_ops(&cashew_clk_syscore_ops);
}

#else /* CONFIG_PM_SLEEP */
static void cashew_clk_sleep_init(void __iomem *reg_base) {}
#endif

static bool pll_mux(struct cashew_clk_provider *p,
		    struct cashew_pll *pll, struct cashew_mux *mux)
{
	if (!strcmp(pll->name, mux->parents[0]) ||
	    !strcmp(pll->name, mux->parents[1])) {
		cashew_clk_pmx(p->clk_data.clks[pll->id],
					   p->base + mux->reg, mux->shift);
		return true;
	   }
	return false;
}

static void add_cs_to_pll(struct cashew_clk_provider *p,
			  struct cashew_primary_clock *pr)
{
	int i, j;

	for (i = 0; i < pr->no_of_plls; i++)
		for (j = 0; j < pr->no_of_muxes; j++)
			if (pll_mux(p, &pr->prim_plls[i], &pr->prim_muxes[j]))
				break;
}

static void cashew_clk_register_primary(struct cashew_clk_provider *p,
					struct cashew_primary_clock *primary)
{
	if (!p || !primary)
		return;
	/* TODO more input check */
	cashew_clk_register_pll(p, primary->prim_plls, primary->no_of_plls);
	cashew_clk_register_mux(p, primary->prim_muxes, primary->no_of_muxes);
	cashew_clk_register_gate(p, primary->prim_gates, primary->no_of_gates,
				 CLK_SET_RATE_PARENT);
	add_cs_to_pll(p, primary);
}

static void cashew_clk_register_devices(struct cashew_clk_provider *p,
					struct cashew_device_clock *dev_clks,
					 unsigned int no_of_devices)
{
	int i;

	for (i = 0; i < no_of_devices; i++) {
		if (dev_clks[i].no_of_muxes > 0)
			cashew_clk_register_mux(p, dev_clks[i].dev_muxes,
						dev_clks[i].no_of_muxes);
		if (dev_clks[i].dev_div.name)
			cashew_clk_register_div(p, &dev_clks[i].dev_div, 1);
		if (dev_clks[i].dev_ddiv.name)
			cashew_clk_register_ddiv(p, &dev_clks[i].dev_ddiv, 1);
/* cashew_clk_register_fixed_factor(p, cashew_ffs,  ARRAY_SIZE(cashew_ffs)); */
		if (dev_clks[i].dev_gate.name)
			cashew_clk_register_gate(p, &dev_clks[i].dev_gate, 1,
						 CLK_SET_RATE_PARENT);
	}
}

void enable_driverless_clks(struct cashew_clk_provider *p)
{
	size_t sz;
	u32 *forced_clk_ids;
	int error;
	struct device_node *np = p->node;
	error = of_property_read_u32(np, "fe-size", &sz);
	if (error)
		return;
	forced_clk_ids = (u32 *)kmalloc (sz * sizeof(u32), GFP_KERNEL);
	error = of_property_read_u32_array(np, "forced-enable-clks",
					   forced_clk_ids, sz);
	if (error) {
		pr_err("clk_provider has no forced enabled clks.");
		return;
	}
	cashew_clk_force_enable(p, forced_clk_ids, sz);
	kfree(forced_clk_ids);
}

#if 0
static void set_pll_clock_frequency(struct cashew_clk_provider *p,
				    u32 pll_clk_id,
				     char *property_name)
{
	struct clk *pllclk;
	u32 rate = 0;
	u32 ret = 0;

	pllclk = p->clk_data.clks[pll_clk_id];
	ret = of_property_read_u32(p->node, property_name, (u32 *)&rate);
	if (ret == 0 && rate != 0) {
		pllclk = p->clk_data.clks[pll_clk_id];
		ret = clk_set_rate(pllclk, rate);
	}
}
#endif

static void __init cashew_clk_core_init(struct device_node *np)
{
	struct cashew_clk_provider *p;

	p = cashew_clk_alloc_provider(np, CLK_NR_CORE_CLKS);
	if (!p)
		return;

	cashew_clk_sleep_init(p->base);

	cashew_clk_register_primary(p,  &system_primary);
	cashew_clk_register_devices(p, cashew_sys_devices,
				    ARRAY_SIZE(cashew_sys_devices));
	cashew_clk_register_provider(p);
	enable_driverless_clks(p);
}

CLK_OF_DECLARE(cashew_clk_core, "img,cashew-clk-core", cashew_clk_core_init);

static void __init cashew_clk_av_init(struct device_node *np)
{
	struct cashew_clk_provider *p;

	p = cashew_clk_alloc_provider(np, CLK_NR_AV_CLKS);
	if (!p)
		return;

	cashew_clk_register_primary(p,  &av_primary);
	cashew_clk_register_cv(p, prim_av_cvs, 2);
	cashew_clk_register_devices(p, cashew_av_devices,
				    ARRAY_SIZE(cashew_av_devices));
	cashew_clk_register_provider(p);
	enable_driverless_clks(p);
}

CLK_OF_DECLARE(cashew_clk_av, "img,cashew-clk-av", cashew_clk_av_init);

static void __init cashew_clk_afe_init(struct device_node *np)
{
	struct cashew_clk_provider *p;

	p = cashew_clk_alloc_provider(np, CLK_NR_AFE_CLKS);
	if (!p)
		return;

	cashew_clk_register_primary(p,  &afe_primary);
	cashew_clk_register_devices(p, cashew_afe_devices,
				    ARRAY_SIZE(cashew_afe_devices));

	cashew_clk_register_provider(p);
	enable_driverless_clks(p);
}

CLK_OF_DECLARE(cashew_clk_afe, "img,cashew-clk-afe", cashew_clk_afe_init);

static void __init cashew_clk_aon_init(struct device_node *np)
{
	struct cashew_clk_provider *p;
	struct cashew_fixed_factor ff = {MEMREPAIR_FIXED_DIV, 16,
					 "memrepair_fixed_div", "xtal1"};
	p = cashew_clk_alloc_provider(np, CLK_NR_AON_CLKS);
	if (!p)
		return;
	cashew_clk_register_fixed_factor(p, &ff, 1);

	cashew_clk_register_devices(p, cashew_aon_devices,
				    ARRAY_SIZE(cashew_aon_devices));
	cashew_clk_register_provider(p);
	enable_driverless_clks(p);
}

CLK_OF_DECLARE(cashew_clk_aon, "img,cashew-clk-aon", cashew_clk_aon_init);

/* sys clocks */
static void __init cashew_cr_perip_init(struct device_node *np)
{
	struct cashew_clk_provider *p;

	p = cashew_clk_alloc_provider(np, CLK_NR_PERIP_SYS_CLKS);
	if (!p)
		return;

	cashew_clk_register_gate(p, cashew_perip_sys_gates,
				 ARRAY_SIZE(cashew_perip_sys_gates), 0);

	cashew_clk_register_provider(p);
	enable_driverless_clks(p);
}

CLK_OF_DECLARE(cashew_cr_perip, "img,cashew-cr-perip",
	       cashew_cr_perip_init);
static void __init cashew_clk_connections_init(struct device_node *np)
{
	struct cashew_clk_provider *p;

	p = cashew_clk_alloc_provider(np, 2);
	if (!p)
		return;
	cashew_clk_register_provider(p);
}

CLK_OF_DECLARE(cashew_clk_connections, "img,cashew-clk-connections",
	       cashew_clk_connections_init);
