/*
 * double-divider clk-cashew
 *
 * Copyright (C) 2015 Imagination Technologies Ltd.
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 */

#include <linux/clk-provider.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/io.h>
#include <linux/err.h>
#include <linux/string.h>
#include <linux/log2.h>

#include "clk.h"
/*
 * DOC: basic adjustable divider clock that cannot gate
 *
 * Traits of this clock:
 * prepare - clk_prepare only ensures that parents are prepared
 * enable - clk_enable only ensures that parents are enabled
 * rate - rate is adjustable.  clk->rate = parent->rate / divisor
 * parent - fixed parent.  No clk_set_parent support
 */

#define to_cashew_double_divider(_hw) \
	container_of(_hw, struct cashew_double_divider, hw)
#define BIT(nr)         (1UL << (nr))
#define div_mask(width)	((1 << (width)) - 1)
unsigned long pdp_pll_rates[] = {594000000, 495000000, 371250000, 402000000};
static struct clk *get_pll(struct clk_hw *hw);

static bool is_clk_a_pll(struct clk *clock)
{
	const char *name = __clk_get_name(clock);

	if (strlen(name) == strlen("pllX") && strstr(name, "pll"))
		return true;

	return false;
}

static struct clk *get_pll(struct clk_hw *hw)
{
	struct clk *pll;

	pll = clk_get_parent(hw->clk);
	while (pll && !is_clk_a_pll(pll))
		pll = clk_get_parent(pll);

	return pll;
}

static bool set_pll_if_pdp(struct clk_hw *hw, unsigned long rate,
			   unsigned long *prate)
{
	struct clk *pll;
	int i;

	if (!strstr(__clk_get_name(hw->clk), "pixel0") &&
	    !strstr(__clk_get_name(hw->clk), "pixel1"))
		return true;

	pll = get_pll(hw);
	if (!pll)
		return false;

	if (clk_get_rate(pll) % rate == 0)
		return true;
	for (i = 0; i < sizeof(pdp_pll_rates) / sizeof(unsigned long); i++) {
		if (pdp_pll_rates[i] % rate == 0) {
			clk_set_rate(pll, pdp_pll_rates[i]);
			*prate = clk_get_rate(clk_get_parent(hw->clk));
			return true;
		}
	}
	return false;
}

struct clk *clk_register_double_divider(struct device *dev, const char *name,
					const char *parent_name,
					unsigned long flags,
					void __iomem *reg, u8 shift1,
					u8 shift2, u8 width1,
					u8 width2,
					u8 cashew_double_divider_flags,
					spinlock_t *lock);

static bool is_div1_valid(u32 div1, unsigned long parent_rate,
			  struct cashew_double_divider *divider)
{
	return ((parent_rate / div1) <= divider->max_div_pre_rate);
}

static bool _is_valid_div(struct cashew_double_divider *divider,
			  unsigned int div)
{
	return true;
}

static unsigned int _get_maxdiv(u32 width)
{
	return div_mask(width) + 1;
}

static unsigned long
cashew_double_divider_recalc_rate(struct clk_hw *hw, unsigned long parent_rate)
{
	struct cashew_double_divider *divider = to_cashew_double_divider(hw);
	u32 div;
	u32 val1;
	u32 val2;
/* set_pll_if_pdp (hw, rate); */
/* printk ("\nrecalculating the rate for %s",__clk_get_name(get_pll(hw)) ); */
	val1 = clk_readl(divider->reg) >> divider->shift1;
	val1 &= div_mask(divider->width1);

	val2 = clk_readl(divider->reg) >> divider->shift2;
	val2 &= div_mask(divider->width2);

	if (!is_div1_valid((val1 + 1), parent_rate, divider)) {
		WARN(true,
		     "%s: the first divider rate constraint not met.\n",
		     __clk_get_name(hw->clk));
		return parent_rate;
	}

	div = (val1 + 1) * (val2 + 1);
	return parent_rate / div;
}

/*
 * The reverse of DIV_ROUND_UP: The maximum number which
 * divided by m is r
 */
#define MULT_ROUND_UP(r, m) ((r) * (m) + (m) - 1)

static bool _is_best_div(struct cashew_double_divider *divider,
			 unsigned long rate, unsigned long now,
			 unsigned long best)
{
	return abs(rate - now) < abs(rate - best);
	if (divider->flags & CASHEW_DOUBLE_DIVIDER_ROUND_CLOSEST)
		return abs(rate - now) < abs(rate - best);

	return now <= rate && now > best;
}

static bool _get_optimised_values(unsigned long rate,
				  unsigned long  parent_rate,
				  u32 *val1, u32 *val2,
				  struct cashew_double_divider *divider)
{
	u32 div1, div2;
	u32 diff, min_diff = ~0;
	unsigned long new_rate, rate1;
	u32 max_div1 = 1 << divider->width1;
	u32 max_div2 = 1 << divider->width2;
	bool ret;

	*val1 = 0;
	ret = false;
	if (DIV_ROUND_CLOSEST(parent_rate, rate) >= (max_div1 * max_div2)) {
		*val1 = div_mask(divider->width1);
		*val2 = div_mask(divider->width2);
		return is_div1_valid(*val1, parent_rate, divider);
	}
	for (div1 = max_div1; div1 > 0; div1--) {
		if (is_div1_valid(div1, parent_rate, divider)) {
			rate1 = DIV_ROUND_CLOSEST(parent_rate, div1);
			div2  = DIV_ROUND_CLOSEST(rate1, rate);
			new_rate = DIV_ROUND_CLOSEST(rate1, div2);
			diff = abs(new_rate - rate);
			if (diff < min_diff && div2 > 0 && div2 < max_div2) {
				min_diff = diff;
				*val1 = div1 - 1;
		*val2 = div2 - 1;
				ret = true;
			}
		}
	}

	return ret;
}

/*
inputs:
	rate: desired rate for the double-divider
	*best_parent_rate: The current parent rate
	hw   : pointer to clk_hw of the double-divider
outputs:
	*best_parent_rate: the best possible acheivable parent's rate
	if change in parent rate is allowed.
returns:
	The best possible valid divide value for the whole double-divider
*/
static int cashew_double_divider_bestdiv(struct clk_hw *hw, unsigned long rate,
					 unsigned long *best_parent_rate)
{
	struct cashew_double_divider *divider = to_cashew_double_divider(hw);
	int bestdiv = 0;
	unsigned long parent_rate, best_rate, now, maxdiv;
	unsigned long parent_rate_saved = *best_parent_rate;
	u32 maxdiv1, maxdiv2;
	u32 val1, val2;
	u32 div1, div2;

	if (!rate)
		rate = 1;

	maxdiv1 = _get_maxdiv(divider->width1);
	maxdiv2 = _get_maxdiv(divider->width2);
	maxdiv  = maxdiv1 * maxdiv2;
	if (!(__clk_get_flags(hw->clk) & CLK_SET_RATE_PARENT)) {
		parent_rate = *best_parent_rate;
		bestdiv = (bool)_get_optimised_values(rate, parent_rate,
						&val1, &val2, divider);
	/* Do not change the parent rate
	 *  if the divider constraints have not been met 
	 */
		bestdiv = (!bestdiv) ? 1 : ((val1 + 1) * (val2 + 1));
		bestdiv = bestdiv > maxdiv ? maxdiv : bestdiv;
		return bestdiv;
	}
	/* FIXME. As CLK_SET_RATE_PARENT is always 0,
	 * we never get to this part of the
	 * code and therefore it has not been tested
	 */
	
	/*
	 * rate * maxdiv munst be less than ULONG_MAX
	 */
	/* maxdiv = min((ULONG_MAX / rate) + 1, maxdiv); */
	best_rate = parent_rate_saved / maxdiv;

	for (div1 = maxdiv1; div1 > 0; div1--) {
		for (div2 = 1; div2 < maxdiv2; div2++) {
			if (div1 * div2 > maxdiv)
				continue;
			/*
			 * It's the ideal case if the requested rate can be
			 * divided from parent clock without needing to change
			 * parent rate, so return the divider immediately.
			 */
			if (rate * div1 * div2 == parent_rate_saved) {
				*best_parent_rate = parent_rate_saved;
				return div1 * div2;
			}
			parent_rate =
			__clk_round_rate(__clk_get_parent(hw->clk),
					 rate * div1 * div2);

			if (is_div1_valid(div1, parent_rate, divider)) {
				now = MULT_ROUND_UP(parent_rate, div1 * div2);
				if (_is_best_div(divider, rate, now,
						 best_rate)) {
					bestdiv = (div1 * div2);
					best_rate = now;
					*best_parent_rate = parent_rate;
					if (*best_parent_rate ==
					    parent_rate_saved)
						return bestdiv;
				}
			}
		}
	}

	if (!bestdiv) {
		/*TODO: see how parent_rate = 1 will make sense. */
		bestdiv = maxdiv;
		*best_parent_rate = __clk_round_rate(__clk_get_parent(hw->clk),
						     1);
	}
	return bestdiv;
}

static long cashew_double_divider_round_rate(struct clk_hw *hw,
					     unsigned long rate,
					     unsigned long *prate)
{
	int div;

	set_pll_if_pdp(hw, rate, prate);
	div = cashew_double_divider_bestdiv(hw, rate, prate);

	return *prate / div;
}

static int cashew_double_divider_set_rate(struct clk_hw *hw,
					  unsigned long rate,
					  unsigned long parent_rate)
{
	struct cashew_double_divider *divider = to_cashew_double_divider(hw);
	unsigned int div;
	unsigned long flags = 0;
	u32 val, val1, val2;

	set_pll_if_pdp(hw, rate, &parent_rate);
	div = parent_rate / rate;

	if (!_is_valid_div(divider, div))
		return -EINVAL;

	if (!_get_optimised_values(rate, parent_rate, &val1, &val2, divider))
		return -EINVAL;

	if (divider->lock)
		spin_lock_irqsave(divider->lock, flags);
	val  = clk_readl(divider->reg);
	val &= ~(div_mask(divider->width1) << divider->shift1);
	val |= val1 << divider->shift1;

	val &= ~(div_mask(divider->width2) << divider->shift2);
	val |= val2 << divider->shift2;
/*
	val = clk_readl(divider->reg);
	val &= ~(div_mask(divider) << divider->shift);
	val |= value << divider->shift;
*/
	clk_writel(val, divider->reg);

	if (divider->lock)
		spin_unlock_irqrestore(divider->lock, flags);

	return 0;
}

const struct clk_ops cashew_double_divider_ops = {
	.recalc_rate = cashew_double_divider_recalc_rate,
	.round_rate = cashew_double_divider_round_rate,
	.set_rate = cashew_double_divider_set_rate,
};

static struct clk *_register_double_divider(struct device *dev,
					    const char *name,
					    const char *parent_name,
					    unsigned long flags,
					    void __iomem *reg, u8 shift1,
					    u8 shift2, u8 width1, u8 width2,
					    u8 cashew_double_divider_flags,
					    spinlock_t *lock)
{
	struct cashew_double_divider *div;
	struct clk *clk;
	struct clk_init_data init;

	/* allocate the divider */
	div = kzalloc(sizeof(*div), GFP_KERNEL);
	if (!div) {
		pr_err("%s: could not allocate divider clk\n", __func__);
		return ERR_PTR(-ENOMEM);
	}

	init.name = name;
	init.ops = &cashew_double_divider_ops;
	init.flags = flags | CLK_IS_BASIC;
	init.parent_names = (parent_name ? &parent_name : NULL);
	init.num_parents = (parent_name ? 1 : 0);

	/* struct cashew_double_divider assignments */
	div->reg = reg;
	div->shift1 = shift1;
	div->width1 = width1;
	div->shift2 = shift2;
	div->width2 = width2;

	div->flags = cashew_double_divider_flags;
	div->lock = lock;
	div->hw.init = &init;
	/* TODO, add this to the definition of the double divider */
	div->max_div_pre_rate = 3800000000UL;
	/* register the clock */
	clk = clk_register(dev, &div->hw);

	if (IS_ERR(clk))
		kfree(div);

	return clk;
}

/**
 * clk_register_double_divider - register a double divider clock with the clock framework
 * @dev: device registering this clock
 * @name: name of this clock
 * @parent_name: name of clock's parent
 * @flags: framework-specific flags
 * @reg: register address to adjust divider
 * @shift1: number of bits to shift the bitfield for the first divider
 * @shift2: number of bits to shift the bitfield for the second divider
 * @width1: width of the bitfield of the first divider
 * @width2: width of the bitfield of the second divider
 * @cashew_double_divider_flags: divider-specific flags for this clock
 * @lock: shared register lock for this clock
 */
struct clk *clk_register_double_divider(struct device *dev, const char *name,
					const char *parent_name,
					unsigned long flags,
					void __iomem *reg, u8 shift1,
					u8 shift2, u8 width1, u8 width2,
					u8 cashew_double_divider_flags,
					spinlock_t *lock)
{
	return _register_double_divider(dev, name, parent_name, flags, reg,
					shift1, shift2,	width1, width2,
					cashew_double_divider_flags, lock);
}

void cashew_clk_register_ddiv(struct cashew_clk_provider *p,
			      struct cashew_double_divider *ddiv,
			      unsigned int num)
{
	struct clk *clk;
	unsigned int i;

	for (i = 0; i < num; i++) {
		clk =
		clk_register_double_divider(NULL, ddiv[i].name,
					    ddiv[i].parent,
					    0,
					    (void __iomem *)
					    (p->base + ddiv[i].reg_index),
					    ddiv[i].shift1,
					    ddiv[i].shift2,
					    ddiv[i].width1,
					    ddiv[i].width2,
					    ddiv[i].flags, NULL);

		p->clk_data.clks[ddiv[i].id] = clk;
	}
}
