/*
 * Copyright (C) 2015 Imagination Technologies Ltd.
 * Copyright (C) 2014 Google, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 */

#define pr_fmt(fmt) "%s: " fmt, __func__

#include <linux/clk-provider.h>
#include <linux/delay.h>
#include <linux/io.h>
#include <linux/kernel.h>
#include <linux/printk.h>
#include <linux/slab.h>

#include "clk.h"

#define PLL_CTRLA			0x0
#define PLL_CTRLA_REFDIV_SHIFT		1
#define PLL_CTRLA_REFDIV_MASK		0x3f
#define PLL_CTRLA_FBDIV_SHIFT		7
#define PLL_CTRLA_FBDIV_MASK		0xfff
#define PLL_CTRLA_POSTDIV1_SHIFT	19
#define PLL_CTRLA_POSTDIV1_MASK	0x7
#define PLL_CTRLA_POSTDIV2_SHIFT	22
#define PLL_CTRLA_POSTDIV2_MASK	0x7
#define PLL_CTRLA_BYPASS		BIT(28)
#define PLL_CTRLA_PD		BIT(25)
#define PLL_CTRLA_DACPD		BIT(26)
#define PLL_CTRLA_DSMPD		BIT(27)
#define PLL_CTRLA_UPDATE		BIT(0)

#define PLL_CTRLB			0x4
#define PLL_CTRLB_FRAC_SHIFT	0
#define PLL_CTRLB_FRAC_MASK	0xffffff
#define PLL_STATUS			0x8
#define PLL_STATUS_LOCK			BIT(0)

/* Fractional PLL operating modes */
#define PLL_MODE_INT			1
#define PLL_MODE_FRAC			0

#define MIN_PFD_INT			1000000UL
#define MIN_PFD_FRAC			10000000UL
#define MIN_VCO_FRAC_FRAC		800000000UL
#define MAX_VCO_FRAC_FRAC		3200000000UL
#define MIN_OUTPUT_FRAC			16000000UL
#define MAX_OUTPUT_FRAC			3200000000UL
#define FBDIV_FRAC_MIN			20
#define FBDIV_FRAC_MAX			320
struct pll_mux {
	void __iomem *adr;
	int shift;
};

struct cashew_clk_pll {
	struct clk_hw hw;
	void __iomem *base;
	struct cashew_pll_rate_table *rates;
	unsigned int nr_rates;
	struct pll_mux pmx;
	u32 type;
};

static inline u32 pll_readl(struct cashew_clk_pll *pll, u32 reg)
{
	return readl(pll->base + reg);
}

static inline void pll_writel(struct cashew_clk_pll *pll, u32 val, u32 reg)
{
	writel(val, pll->base + reg);
}

static inline int pll_lock(struct cashew_clk_pll *pll)
{
	int count = 0;

	while (!(pll_readl(pll, PLL_STATUS) & PLL_STATUS_LOCK) && count < 5) {
		count++;
		udelay(10);
	}
	if (count == 5) {
		pr_err("%s did not lock\n", __clk_get_name(pll->hw.clk));
		return -EINVAL;
	}
	return 0;
}

static inline u64 do_div_round_closest(u64 dividend, u64 divisor)
{
	dividend += divisor / 2;
	return div64_u64(dividend, divisor);
}

static inline struct cashew_clk_pll *to_cashew_pll(struct clk_hw *hw)
{
	return container_of(hw, struct cashew_clk_pll, hw);
}

static inline u32 pll_frac_get_mode(struct clk_hw *hw)
{
	struct cashew_clk_pll *pll = to_cashew_pll(hw);
	u32 val;

	val = pll_readl(pll, PLL_CTRLA) & PLL_CTRLA_DSMPD;
	return val ? PLL_MODE_INT : PLL_MODE_FRAC;
}

static inline void pll_frac_set_mode(struct clk_hw *hw, u32 mode)
{
	struct cashew_clk_pll *pll = to_cashew_pll(hw);
	u32 val;

	val = pll_readl(pll, PLL_CTRLA);
	if (mode == PLL_MODE_INT)
		val |= PLL_CTRLA_DSMPD | PLL_CTRLA_DACPD;
	else
		val &= ~(PLL_CTRLA_DSMPD | PLL_CTRLA_DACPD);
	pll_writel(pll, val, PLL_CTRLA);
}

static int pll_get_params_var(unsigned long fref,
			      unsigned long fout,
			      struct cashew_pll_rate_table *tbl)
{
	u64 fbdiv;
	u64 vco;
	u32 postdiv1, postdiv2 = 1;
	u32 fin;
	u32 refdiv;
	u32 max_post_div = (PLL_CTRLA_POSTDIV1_MASK) *
			   (PLL_CTRLA_POSTDIV2_MASK);
	u64 vco_min = (u64) fout * (u64) max_post_div;

	if (vco_min < (u64) MIN_VCO_FRAC_FRAC) {
		pr_err("%s: fout %lu low vco_min = %llu\n",
		       __func__, fout, vco_min);
		return -EINVAL;
	}
	/* Set postdiv1 and postdiv2 so that vco > MIN_VCO_FRAC_FRAC */
	postdiv1 = 1 + MIN_VCO_FRAC_FRAC / (fout * postdiv2);
	/* Make sure postdiv1 is not bigger than it is allowed */
	while (postdiv1 > PLL_CTRLA_POSTDIV1_MASK) {
		postdiv2++;
		if (postdiv2 > PLL_CTRLA_POSTDIV2_MASK) {
			pr_err("%s: postdiv2 high fout = %lu\n",
			       __func__, fout);
			return -EINVAL;
		}
		postdiv1 = 1 + MIN_VCO_FRAC_FRAC / (fout * postdiv2);
	}
	pr_info("postdiv1 = %u postdiv2 = %u\n", postdiv1, postdiv2);
	vco = fout * postdiv1 * postdiv2;
	if (vco > MAX_VCO_FRAC_FRAC) {
		pr_err("%s: vco = %lu is too high\n",
		       __func__, (unsigned long)vco);
		return -EINVAL;
	}

	/* if fref is bigger than 40MHz, use a bigger refdiv */
	refdiv = (1 + fref / 40000000);
	vco <<= 24;

	fin = fref / refdiv;
	fbdiv = div64_u64(vco, fin);

	tbl->fout = fout;
	tbl->fbdiv = (unsigned long)(fbdiv >> 24);
	if (tbl->fbdiv > PLL_CTRLA_FBDIV_MASK + 1) {
		pr_err("%s: fbdiv = %llu too large\n", __func__, tbl->fbdiv);
		return -EINVAL;
	}
	tbl->frac = (unsigned long)(fbdiv & 0xFFFFFF);
	tbl->fout_min = fin * tbl->fbdiv;
	if (tbl->frac)
		tbl->fout_max = fin * (1 + tbl->fbdiv);
	else
		tbl->fout_max = fout;

	tbl->fref = fref;
	tbl->postdiv1 = postdiv1;
	tbl->postdiv2 = postdiv2;
	tbl->refdiv = refdiv;
	return 0;
}

static struct cashew_pll_rate_table *
pll_get_params(struct cashew_clk_pll *pll, unsigned long fref,
	       unsigned long fout)
{
	unsigned int i, best = ~0;
	unsigned long err, best_err = ~0;

	for (i = 0; i < pll->nr_rates; i++) {
		err = abs(pll->rates[i].fout - fout);
		if (pll->rates[i].fref == fref && err < best_err) {
			best = i;
			best_err = err;
		}
	}
	if (best == ~0)
		return NULL;

	return &pll->rates[best];
}

static long pll_round_rate(struct clk_hw *hw, unsigned long rate,
			   unsigned long *parent_rate)
{
	struct cashew_clk_pll *pll = to_cashew_pll(hw);
	unsigned int i, best = ~0;
	unsigned long err, best_err = ~0;

	if (pll->type == PLL_GF40LP_FRAC_AUDIO)
		return rate;
	for (i = 0; i < pll->nr_rates; i++) {
		err = abs(pll->rates[i].fout - rate);
		if (pll->rates[i].fref == *parent_rate && err < best_err) {
			best = i;
			best_err = err;
		}
	}
	/*TODO: check that best is not ~0 */
	/* Make sure fout_{min,max} parameters have sane values */
	if (!pll->rates[best].fout_min)
		pll->rates[best].fout_min = pll->rates[best].fout;
	if (!pll->rates[best].fout_max)
		pll->rates[best].fout_max = pll->rates[best].fout;

	/*
	 * If the chosen rate is within the maximum allowed PLL adjustment
	 * then we accept it.
	 * Otherwise, just return the closest nominal table rate.
	 */
	if (rate <= pll->rates[best].fout_max &&
	    rate >= pll->rates[best].fout_min)
		return rate;
	return pll->rates[best].fout;
}
static bool set_pll_mx_parent(struct clk_hw *pll_hw, bool to_pll)
{
	struct cashew_clk_pll *pll = to_cashew_pll(pll_hw);
	u32 val, temp;

    if (!strcmp(__clk_get_name(pll_hw->clk), "pll3")) {
        printk("set_pll_mx_parent() called for PLL3 - operation inhibited.\n");
        return 0;
    }

	val = readl(pll->pmx.adr);
	temp = val & BIT(pll->pmx.shift);
	/* val is zero if the mux parent is xtal */
	if ((to_pll && temp) || (!to_pll && !temp)) {
		return false;
	} else {
	/* we need to change the mux parent */
		if (to_pll)
			val |= BIT(pll->pmx.shift);
		else
			val &= ~BIT(pll->pmx.shift);
		writel(val, pll->pmx.adr);
		return true;
	}
}

static int pll_gf40lp_frac_enable(struct clk_hw *hw)
{
	struct cashew_clk_pll *pll = to_cashew_pll(hw);
	u32 val;
	bool mux_swapped;

	/* Make sure that the mux parent is xtal */
	mux_swapped = set_pll_mx_parent(hw, false);
	/*
	 * De-assert the PD. Power up the PLL.
	 * Asserting PD may create a glitch but de-asserting it will not.
	 */
	val = pll_readl(pll, PLL_CTRLA);
	val &= ~(PLL_CTRLA_PD);

	/* Propagate REFDIV and FBDIV to PLL */
	val |= PLL_CTRLA_UPDATE;
	pll_writel(pll, val, PLL_CTRLA);
#ifndef __TARGET_IS_SIMULATOR__
	if (pll_lock(pll))
		return -EINVAL;
#endif
	/* Switching in and out of bypass may create a glitch */
	val &= ~PLL_CTRLA_BYPASS;
	pll_writel(pll, val, PLL_CTRLA);

	/* Set the mux parent to pll */
	if (mux_swapped)
		mux_swapped = set_pll_mx_parent(hw, true);

	return 0;
}

static void pll_gf40lp_frac_disable(struct clk_hw *hw)
{
	struct cashew_clk_pll *pll = to_cashew_pll(hw);
	u32 val;
	bool mux_swapped;

	/* Make sure that the mux parent is xtal */
	mux_swapped = set_pll_mx_parent(hw, false);
	val = pll_readl(pll, PLL_CTRLA);
	val |= PLL_CTRLA_BYPASS;
	val |= PLL_CTRLA_PD;
	pll_writel(pll, val, PLL_CTRLA);
	if (mux_swapped)
		mux_swapped = set_pll_mx_parent(hw, true);
}

static int pll_gf40lp_frac_is_enabled(struct clk_hw *hw)
{
	struct cashew_clk_pll *pll = to_cashew_pll(hw);

	return !(pll_readl(pll, PLL_CTRLA) & PLL_CTRLA_PD);
}

static int pll_gf40lp_frac_set_rate(struct clk_hw *hw, unsigned long rate,
				    unsigned long parent_rate)
{
/*
 *	we have these situations
 *	1. pll is disabled and we want to change the rate and enable it later.
 *	2. pll is enabled but the change is small.
 *	3. pll is enabled and the change is large
 */
	struct cashew_clk_pll *pll = to_cashew_pll(hw);
	struct cashew_pll_rate_table *params, p;
	int enabled = pll_gf40lp_frac_is_enabled(hw);
	u64 val, vco, old_postdiv1, old_postdiv2, frac;
	const char *name = __clk_get_name(hw->clk);
	bool big_change = true;
	unsigned long current_rate = clk_get_rate(hw->clk);
	int diff = abs(rate - current_rate);
	u32 diff_percent = div64_u64((u64)diff * 100, (u64)current_rate);

	if (diff_percent < 2)
		big_change = false;

	if (rate < MIN_OUTPUT_FRAC || rate > MAX_OUTPUT_FRAC)
		return -EINVAL;
	if (pll->type == PLL_GF40LP_FRAC_AUDIO) {
		if (pll_get_params_var(parent_rate, rate, &p))
			return -EINVAL;
		params = &p;
	} else {
		params = pll_get_params(pll, parent_rate, rate);
	}
	if (!params || !params->refdiv)
		return -EINVAL;

	vco = params->fref;
	vco *= (params->fbdiv << 24) + params->frac;
	vco = div64_u64(vco, params->refdiv << 24);
	if (vco < MIN_VCO_FRAC_FRAC || vco > MAX_VCO_FRAC_FRAC)
		pr_warn("%s: VCO %llu is out of range %lu..%lu\n", name, vco,
			MIN_VCO_FRAC_FRAC, MAX_VCO_FRAC_FRAC);

	/* Calculate the frac parameter */
	frac = rate * params->refdiv * params->postdiv1 * params->postdiv2;
	frac -= (params->fbdiv * parent_rate);
	frac = do_div_round_closest(frac << 24, parent_rate);

	val = div64_u64(params->fref, params->refdiv);
	if (val < (frac ? MIN_PFD_FRAC : MIN_PFD_INT))
		pr_warn("%s: PFD %llu is too low (min %lu)\n",
			name, val, frac ? MIN_PFD_FRAC : MIN_PFD_INT);
	if (val > vco / 16)
		pr_warn("%s: PFD %llu is too high (max %llu)\n",
			name, val, vco / 16);
	val = pll_readl(pll, PLL_CTRLA);
	val &= ~((PLL_CTRLA_REFDIV_MASK << PLL_CTRLA_REFDIV_SHIFT) |
		 (PLL_CTRLA_FBDIV_MASK << PLL_CTRLA_FBDIV_SHIFT));
	val |= (params->refdiv << PLL_CTRLA_REFDIV_SHIFT) |
		(params->fbdiv << PLL_CTRLA_FBDIV_SHIFT);

	old_postdiv1 = (val >> PLL_CTRLA_POSTDIV1_SHIFT) &
		       PLL_CTRLA_POSTDIV1_MASK;
	old_postdiv2 = (val >> PLL_CTRLA_POSTDIV2_SHIFT) &
		       PLL_CTRLA_POSTDIV2_MASK;
	if (enabled &&
	    (params->postdiv1 != old_postdiv1 ||
	     params->postdiv2 != old_postdiv2))
		pr_warn("%s: changing postdiv while PLL is enabled\n", name);

	if (params->postdiv2 > params->postdiv1)
		pr_warn("%s: postdiv2 should not exceed postdiv1\n", name);

	val &= ~((PLL_CTRLA_POSTDIV2_MASK << PLL_CTRLA_POSTDIV2_SHIFT) |
		 (PLL_CTRLA_POSTDIV1_MASK << PLL_CTRLA_POSTDIV1_SHIFT));
	val |=	(params->postdiv1 << PLL_CTRLA_POSTDIV1_SHIFT) |
		(params->postdiv2 << PLL_CTRLA_POSTDIV2_SHIFT);

	if (enabled && big_change) {
		/* PLL is enabled and the change is big */
		pll_gf40lp_frac_disable(hw);
	}
	pll_writel(pll, val, PLL_CTRLA);

	val = pll_readl(pll, PLL_CTRLB);
	val &= ~(PLL_CTRLB_FRAC_MASK << PLL_CTRLB_FRAC_SHIFT);
	val |= (frac << PLL_CTRLB_FRAC_SHIFT);
	pll_writel(pll, val, PLL_CTRLB);

	/* set operating mode */
	if (frac)
		pll_frac_set_mode(hw, PLL_MODE_FRAC);
	else
		pll_frac_set_mode(hw, PLL_MODE_INT);

	if (enabled && big_change) {
		/*PLL is enabled and the change is big */
		pll_gf40lp_frac_enable(hw);
	}
#ifndef __TARGET_IS_SIMULATOR__
	if (enabled && !big_change)
		return pll_lock(pll);
#endif
	return 0;
}

static unsigned long pll_gf40lp_frac_recalc_rate(struct clk_hw *hw,
						 unsigned long parent_rate)
{
	struct cashew_clk_pll *pll = to_cashew_pll(hw);
	u64 val, prediv, fbdiv, frac, postdiv1, postdiv2, rate;

	val = pll_readl(pll, PLL_CTRLA);
	prediv = (val >> PLL_CTRLA_REFDIV_SHIFT) & PLL_CTRLA_REFDIV_MASK;
	fbdiv = (val >> PLL_CTRLA_FBDIV_SHIFT) & PLL_CTRLA_FBDIV_MASK;

	postdiv1 = (val >> PLL_CTRLA_POSTDIV1_SHIFT) &
		PLL_CTRLA_POSTDIV1_MASK;
	postdiv2 = (val >> PLL_CTRLA_POSTDIV2_SHIFT) &
		PLL_CTRLA_POSTDIV2_MASK;

	val = pll_readl(pll, PLL_CTRLB);
	frac = (val >> PLL_CTRLB_FRAC_SHIFT) & PLL_CTRLB_FRAC_MASK;

	rate = parent_rate;
	if (pll_frac_get_mode(hw) == PLL_MODE_FRAC)
		rate *= (fbdiv << 24) + frac;
	else
		rate *= (fbdiv << 24);
	rate = do_div_round_closest(rate, (prediv * postdiv1 * postdiv2) << 24);

	return rate;
}

static struct clk_ops pll_gf40lp_frac_ops = {
	.enable = pll_gf40lp_frac_enable,
	.disable = pll_gf40lp_frac_disable,
	.is_enabled = pll_gf40lp_frac_is_enabled,
	.recalc_rate = pll_gf40lp_frac_recalc_rate,
	.round_rate = pll_round_rate,
	.set_rate = pll_gf40lp_frac_set_rate,
};

static struct clk_ops pll_gf40lp_frac_fixed_ops = {
	.enable = pll_gf40lp_frac_enable,
	.disable = pll_gf40lp_frac_disable,
	.is_enabled = pll_gf40lp_frac_is_enabled,
	.recalc_rate = pll_gf40lp_frac_recalc_rate,
};

static struct clk *pll_register(const char *name, const char *parent_name,
				unsigned long flags, void __iomem *base,
				enum cashew_pll_type type,
				struct cashew_pll_rate_table *rates,
				unsigned int nr_rates)
{
	struct cashew_clk_pll *pll;
	struct clk_init_data init;
	struct clk *clk;

	pll = kzalloc(sizeof(*pll), GFP_KERNEL);
	if (!pll)
		return ERR_PTR(-ENOMEM);

	init.name = name;
	init.flags = flags | CLK_GET_RATE_NOCACHE;
	init.parent_names = &parent_name;
	init.num_parents = 1;
	if (rates)
		init.ops = &pll_gf40lp_frac_ops;
	else
		init.ops = &pll_gf40lp_frac_fixed_ops;
	pll->hw.init = &init;
	pll->base = base;
	pll->rates = rates;
	pll->nr_rates = nr_rates;
	pll->type = type;
	clk = clk_register(NULL, &pll->hw);
	if (IS_ERR(clk))
		kfree(pll);

	return clk;
}

void cashew_clk_pmx(struct clk *pll_clk, void __iomem *reg, u32 shift)
{
	struct clk_hw *hw = __clk_get_hw(pll_clk);
	struct cashew_clk_pll *pll = to_cashew_pll(hw);

	pll->pmx.adr = reg;
	pll->pmx.shift = shift;
}

void cashew_clk_register_pll(struct cashew_clk_provider *p,
			     struct cashew_pll *pll,
			     unsigned int num)
{
	struct clk *clk;
	unsigned int i;

	for (i = 0; i < num; i++) {
		clk = pll_register(pll[i].name, pll[i].parent,
				   0, p->pll_base + pll[i].reg_base,
				   pll[i].type, pll[i].rates,
				   pll[i].nr_rates);
		p->clk_data.clks[pll[i].id] = clk;
	}
}
