/*
 * Copyright (C) 2016 Imagination Technologies Ltd.
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 */

#include <linux/kernel.h>
#include <linux/clk-provider.h>
#include <linux/slab.h>

#include "clk.h"

struct cashew_clk_cv {
	struct clk_hw hw;
	bool is_enabled;
	unsigned long rate;
	unsigned long max_rate;
	unsigned long min_rate;
};

static inline struct cashew_clk_cv *cashew_clk_cv(struct clk_hw *hw)
{
	return container_of(hw, struct cashew_clk_cv, hw);
}

static int cv_enable(struct clk_hw *hw)
{
	struct cashew_clk_cv *cv = cashew_clk_cv(hw);

	cv->is_enabled = true;

	return 0;
}

static void cv_disable(struct clk_hw *hw)
{
	struct cashew_clk_cv *cv = cashew_clk_cv(hw);

	cv->is_enabled = false;
}

static int cv_is_enabled(struct clk_hw *hw)
{
	struct cashew_clk_cv *cv = cashew_clk_cv(hw);

	return cv->is_enabled;
}

static int cv_set_rate(struct clk_hw *hw, unsigned long rate,
		       unsigned long parent_rate)
{
	struct cashew_clk_cv *cv = cashew_clk_cv(hw);

	if (rate > cv->max_rate || rate < cv->min_rate)
		pr_warn("%s: rate %lu is out of range %lu..%lu\n",
			__clk_get_name(hw->clk), rate,
			cv->min_rate, cv->max_rate);

	cv->rate = rate;
	return 0;
}

static long cv_round_rate(struct clk_hw *hw, unsigned long rate,
			  unsigned long *parent_rate)
{
	struct cashew_clk_cv *cv = cashew_clk_cv(hw);

	if (rate > cv->max_rate || rate < cv->min_rate)
		pr_warn("%s: rate %lu is out of range %lu..%lu\n",
			__clk_get_name(hw->clk), rate,
			cv->min_rate, cv->max_rate);

	return rate;
}

static unsigned long cv_recalc_rate(struct clk_hw *hw,
				    unsigned long parent_rate)
{
	struct cashew_clk_cv *cv = cashew_clk_cv(hw);

	return cv->rate ? cv->rate : parent_rate;
}

static struct clk_ops cv_ops = {
	.enable = cv_enable,
	.disable = cv_disable,
	.is_enabled = cv_is_enabled,
	.set_rate = cv_set_rate,
	.round_rate = cv_round_rate,
	.recalc_rate = cv_recalc_rate,
};

static struct clk *cv_register(struct cashew_cv *input)
{
	struct cashew_clk_cv *cv;
	struct clk_init_data init;
	struct clk *clk;

	cv = kzalloc(sizeof(*cv), GFP_KERNEL);
	if (!cv)
		return ERR_PTR(-ENOMEM);

	init.name = input->name;
	init.flags = CLK_GET_RATE_NOCACHE;
	init.parent_names = &input->parent;
	init.num_parents = 1;
	init.ops = &cv_ops;
	cv->hw.init = &init;
	cv->max_rate = input->max_rate;
	cv->min_rate = input->min_rate;
	clk = clk_register(NULL, &cv->hw);
	if (IS_ERR(clk))
		kfree(cv);

	return clk;
}

void cashew_clk_register_cv(struct cashew_clk_provider *p,
			    struct cashew_cv *cv, unsigned int num)
{
	struct clk *clk;
	unsigned int i;

	for (i = 0; i < num; i++) {
		clk = cv_register(&cv[i]);
		p->clk_data.clks[cv[i].id] = clk;
	}
}
