/*
 * Copyright (C) 2015 Imagination Technologies Ltd.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 */

#ifndef __CASHEW_CLK_H
#define __CASHEW_CLK_H

#include <linux/clk-provider.h>
struct cashew_double_divider {
	struct clk_hw	hw;
/* This is the actual register value which is the provider->base + reg_index */
	void __iomem	*reg;
	unsigned int id;
/* This is the index to the provider base address. */
	unsigned long reg_index;
	u8		shift1;
	u8		shift2;
	u8		width1;
	u8		width2;
	u8		flags;
	const char *name;
	const char *parent;
/* maximum rate for the parent of the divider */
	unsigned long max_div_parent;
/* maximum rate for the output of pre-divider */
	unsigned long max_div_pre_rate;
/* private lock */
	spinlock_t	*lock;
};

/* TODO: Think about the max_div_pre_rate */
#define DDIV(_id, _name, _pname, _reg_idx, _width1, _width2, _shift1, _shift2)\
	{							\
		.id		= _id,				\
		.reg_index	= _reg_idx,			\
		.width1		= _width1,			\
		.width2		= _width2,			\
		.shift1		= _shift1,			\
		.shift2		= _shift2,			\
		.flags		= 0,				\
		.name		= _name,			\
		.parent		= _pname,			\
		.max_div_pre_rate = 3800000000ul,		\
	}
#define CASHEW_DOUBLE_DIVIDER_ONE_BASED		BIT(0)
#define CASHEW_DOUBLE_DIVIDER_ROUND_CLOSEST	BIT(4)
struct cashew_gate {
	unsigned int id;
	unsigned long reg;
	unsigned int shift;
	const char *name;
	const char *parent;
};

#define GATE(_id, _name, _pname, _reg, _shift)	\
	{					\
		.id	= _id,			\
		.reg	= _reg,			\
		.shift	= _shift,		\
		.name	= _name,		\
		.parent = _pname,		\
	}

struct cashew_mux {
	unsigned int id;
	unsigned long reg;
	unsigned int shift;
	unsigned int num_parents;
	const char *name;
	const char **parents;
};

#define PNAME(x) static const char *x[] __initconst

#define MUX(_id, _name, _pnames, _reg, _shift)			\
	{							\
		.id		= _id,				\
		.reg		= _reg,				\
		.shift		= _shift,			\
		.name		= _name,			\
		.parents	= _pnames,			\
		.num_parents	= ARRAY_SIZE(_pnames)		\
	}

struct cashew_div {
	unsigned int id;
	unsigned long reg;
	unsigned int width;
	unsigned int shift;
	unsigned int div_flags;
	unsigned int flags;
	const char *name;
	const char *parent;
};

struct cashew_cv {
	unsigned int id;
	const char *name;
	const char *parent;
	unsigned long max_rate;
	unsigned long min_rate;
};

#define DIVC(_id, _name, _pname, _reg, _shift, _width, _flags)	\
	{							\
		.id		= _id,				\
		.reg		= _reg,				\
		.width		= _width,			\
		.shift		= _shift,			\
		.flags	= _flags,				\
		.name		= _name,			\
		.parent		= _pname,			\
	}
#define DIV(_id, _name, _pname, _reg, _width)			\
	{							\
		.id		= _id,				\
		.reg		= _reg,				\
		.width		= _width,			\
		.div_flags	= 0,				\
		.name		= _name,			\
		.parent		= _pname,			\
	}
#define DIV_F(_id, _name, _pname, _reg, _width, _div_flags)	\
	{							\
		.id		= _id,				\
		.reg		= _reg,				\
		.width		= _width,			\
		.div_flags	= _div_flags,			\
		.name		= _name,			\
		.parent		= _pname,			\
	}

#define DIV_FC(_id, _name, _pname, _reg, _width, _flags)	\
	{							\
		.id		= _id,				\
		.reg		= _reg,				\
		.width		= _width,			\
		.flags		= _flags,			\
		.name		= _name,			\
		.parent		= _pname,			\
	}

struct cashew_fixed_factor {
	unsigned int id;
	unsigned int div;
	const char *name;
	const char *parent;
};

#define FIXED_FACTOR(_id, _name, _pname, _div)			\
	{							\
		.id		= _id,				\
		.div		= _div,				\
		.name		= _name,			\
		.parent		= _pname,			\
	}

/*
 * in order to avoid u32 multiplication overflow, declare all
 * members of this structure as u64
 */
struct cashew_pll_rate_table {
	unsigned long long fref;
	unsigned long long fout;
	unsigned long long fout_min;
	unsigned long long fout_max;
	unsigned long long refdiv;
	unsigned long long fbdiv;
	unsigned long long postdiv1;
	unsigned long long postdiv2;
	unsigned long long frac;
};

enum cashew_pll_type {
	PLL_GF40LP_LAINT,
	PLL_GF40LP_FRAC,
	PLL_GF40LP_FRAC_AUDIO
};

struct cashew_pll {
	unsigned int id;
	unsigned long reg_base;
	enum cashew_pll_type type;
	struct cashew_pll_rate_table *rates;
	unsigned int nr_rates;
	const char *name;
	const char *parent;
};

#define PLL(_id, _name, _pname, _type, _reg, _rates)		\
	{							\
		.id		= _id,				\
		.reg_base	= _reg,				\
		.type		= _type,			\
		.rates		= _rates,			\
		.nr_rates	= ARRAY_SIZE(_rates),		\
		.name		= _name,			\
		.parent		= _pname,			\
	}

#define PLL_FIXED(_id, _name, _pname, _type, _reg)		\
	{							\
		.id		= _id,				\
		.reg_base	= _reg,				\
		.type		= _type,			\
		.rates		= NULL,				\
		.nr_rates	= 0,				\
		.name		= _name,			\
		.parent		= _pname,			\
	}

struct cashew_clk_provider {
	struct device_node *node;
	void __iomem *base;
	void __iomem *pll_base;
	struct clk_onecell_data clk_data;
};

struct cashew_device_clock {
	unsigned int		index;
	struct cashew_gate	dev_gate;
	struct cashew_double_divider dev_ddiv;
	struct cashew_mux	dev_muxes[4];
	unsigned int no_of_muxes;
	struct cashew_div	dev_div;
};

struct cashew_primary_clock {
	unsigned int no_of_gates;
	unsigned int no_of_muxes;
	unsigned int no_of_plls;
	struct cashew_gate	prim_gates[8];
	struct cashew_mux	prim_muxes[10];
	struct cashew_pll	prim_plls[4];
	struct cashew_cv	prim_cvs[2];
};

struct cashew_clk_provider *
cashew_clk_alloc_provider(struct device_node *node, unsigned int num_clks);
void cashew_clk_register_provider(struct cashew_clk_provider *p);
void cashew_clk_register_gate(struct cashew_clk_provider *p,
			      struct cashew_gate *gate,
			      unsigned int num, u32 flags);
void cashew_clk_register_mux(struct cashew_clk_provider *p,
			     struct cashew_mux *mux,
			     unsigned int num);
void cashew_clk_register_div(struct cashew_clk_provider *p,
			     struct cashew_div *div,
			     unsigned int num);
void cashew_clk_register_ddiv(struct cashew_clk_provider *p,
			      struct cashew_double_divider *ddiv,
			      unsigned int num);
void cashew_clk_register_fixed_factor(struct cashew_clk_provider *p,
				      struct cashew_fixed_factor *ff,
				      unsigned int num);
void cashew_clk_register_pll(struct cashew_clk_provider *p,
			     struct cashew_pll *pll,
			     unsigned int num);
void cashew_clk_register_cv(struct cashew_clk_provider *p,
			    struct cashew_cv *cv, unsigned int num);
void cashew_clk_force_enable(struct cashew_clk_provider *p,
			     unsigned int *clk_ids, unsigned int num);
void cashew_clk_pmx(struct clk *pll_clk, void __iomem *reg, u32 shift);
#endif
