/*
 * Imagination Technologies timer clocksource
 *
 * Copyright (C) 2016 Imagination Technologies
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License. See the file "COPYING" in the main directory of this archive
 * for more details.
 */

#include <linux/clk.h>
#include <linux/clocksource.h>
#include <linux/err.h>
#include <linux/init.h>
#include <linux/spinlock.h>
#include <linux/of.h>
#include <linux/of_address.h>
#include <linux/sched_clock.h>

/* Top level reg */
#define CR_TIMER_CTRL_CFG		0x00
#define TIMER_ME_GLOBAL			BIT(0)
#define CR_TIMER_REV			0x10

/* Timer specific registers */
#define TIMER_CFG			0x20
#define TIMER_ME_LOCAL			BIT(0)
#define TIMER_RELOAD_VALUE		0x24
#define TIMER_CURRENT_VALUE		0x28
#define TIMER_CURRENT_OVERFLOW_VALUE	0x2C
#define TIMER_IRQ_STATUS		0x30
#define TIMER_IRQ_CLEAR			0x34
#define TIMER_IRQ_MASK			0x38

/* Timer specific configuration Values */
#define RELOAD_VALUE			0xffffffff

static bool enabled;
static void __iomem *timer_base;
static struct clk *sys_clk, *core_clk;
static DEFINE_RAW_SPINLOCK(lock);

static inline u32 timer_readl(u32 offset, u32 timer_id)
{
	return readl(timer_base + 0x20 * timer_id + offset);
}

static inline void timer_writel(u32 value, u32 offset, u32 timer_id)
{
	writel(value, timer_base + 0x20 * timer_id + offset);
}

static cycle_t timer_clocksource_read_cycles(struct clocksource *cs)
{
	u32 counter, overflw;
	unsigned long flags;

	raw_spin_lock_irqsave(&lock, flags);
	overflw = timer_readl(TIMER_CURRENT_OVERFLOW_VALUE, 0);
	counter = timer_readl(TIMER_CURRENT_VALUE, 0);
	raw_spin_unlock_irqrestore(&lock, flags);

	return ~(cycle_t)counter;
}

static u64 notrace timer_read_sched_clock(void)
{
	return timer_clocksource_read_cycles(NULL);
}

static void timer_clksrc_enable(int timeridx)
{
	u32 val;

	/* Disable timer local before loading reload value */
	val = timer_readl(TIMER_CFG, timeridx);
	val &= ~TIMER_ME_LOCAL;
	timer_writel(val, TIMER_CFG, timeridx);

	timer_writel(RELOAD_VALUE, TIMER_RELOAD_VALUE, timeridx);

	val = timer_readl(TIMER_CFG, timeridx);
	val |= TIMER_ME_LOCAL;
	timer_writel(val, TIMER_CFG, timeridx);
}

static void timer_clksrc_disable(int timeridx)
{
	u32 val;

	/* Disable timer local */
	val = timer_readl(TIMER_CFG, timeridx);
	val &= ~TIMER_ME_LOCAL;
	timer_writel(val, TIMER_CFG, timeridx);
}

static int timer_clocksource_enable(struct clocksource *cs)
{
	timer_clksrc_enable(0);
	enabled = true;
	return 0;
}

static void timer_clocksource_disable(struct clocksource *cs)
{
	timer_clksrc_disable(0);
	enabled = false;
}

static void timer_clocksource_suspend(struct clocksource *cs)
{
	if (enabled)
		timer_clksrc_disable(0);

	clk_disable(sys_clk);
	clk_disable(core_clk);
}

static void timer_clocksource_resume(struct clocksource *cs)
{
	clk_enable(sys_clk);
	clk_enable(core_clk);

	/* Enable timer block */
	writel(TIMER_ME_GLOBAL, timer_base);

	if (enabled)
		timer_clksrc_enable(0);
}

/* Desirable clock source for IMG timer */
static struct clocksource timer_clocksource = {
	.name		= "img_timer",
	.rating		= 300,
	.enable		= timer_clocksource_enable,
	.disable	= timer_clocksource_disable,
	.read		= timer_clocksource_read_cycles,
	.suspend	= timer_clocksource_suspend,
	.resume		= timer_clocksource_resume,
	.mask		= CLOCKSOURCE_MASK(32),
	.flags		= CLOCK_SOURCE_IS_CONTINUOUS,
};

static void __init timer_clksrc_of_init(struct device_node *node)
{
	unsigned long rate;
	int ret;

	enabled = false;

	timer_base = of_iomap(node, 0);
	if (!timer_base) {
		pr_err("cannot iomap\n");
		return;
	}

	sys_clk = of_clk_get_by_name(node, "sys");
	if (IS_ERR(sys_clk)) {
		pr_err("clock get failed (%lu)\n", PTR_ERR(sys_clk));
		return;
	}

	core_clk = of_clk_get_by_name(node, "core");
	if (IS_ERR(core_clk)) {
		pr_err("clock get failed (%lu)\n", PTR_ERR(core_clk));
		return;
	}

	ret = clk_prepare_enable(sys_clk);
	if (ret < 0) {
		pr_err("failed to enable clock (%d)\n", ret);
		return;
	}

	ret = clk_prepare_enable(core_clk);
	if (ret < 0) {
		pr_err("failed to enable clock (%d)\n", ret);
		clk_disable_unprepare(sys_clk);
		return;
	}

	rate = clk_get_rate(core_clk);

	/* Disable irq's for clocksource usage */
	timer_writel(0, TIMER_IRQ_MASK, 0);
	timer_writel(0, TIMER_IRQ_MASK, 1);
	timer_writel(0, TIMER_IRQ_MASK, 2);
	timer_writel(0, TIMER_IRQ_MASK, 3);

	/* Enable timer block */
	writel(TIMER_ME_GLOBAL, timer_base);

	sched_clock_register(timer_read_sched_clock, 32, rate);
	clocksource_register_hz(&timer_clocksource, rate);
}

CLOCKSOURCE_OF_DECLARE(img_timer, "img,timer", timer_clksrc_of_init);
