/*
 * Cashew PDC GPIO handling.
 *
 * Copyright (C) 2016 Imagination Technologies Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/bitops.h>
#include <linux/gpio.h>
#include <linux/io.h>
#include <linux/of_irq.h>
#include <linux/platform_device.h>
#include <linux/slab.h>

#define GPIO_PDC_NGPIO		5

/**
 * struct cashew_pdc_gpio - GPIO bank private data
 * @chip:	Generic GPIO chip for GPIO bank
 * @reg:	Base of first SOC_SYS_WAKE register
 * @irq:	IRQ numbers
 *
 * This is the main private data for the PDC GPIO driver. It encapsulates a
 * gpio_chip, and the callbacks for the gpio_chip can access the private data
 * with the to_pdc() macro below.
 */
struct cashew_pdc_gpio {
	struct gpio_chip chip;
	void __iomem *reg;
	int irq[GPIO_PDC_NGPIO];
};
#define to_pdc(c)	container_of(c, struct cashew_pdc_gpio, chip)

/* Generic GPIO interface */

static int cashew_pdc_gpio_direction_input(struct gpio_chip *chip,
					   unsigned int offset)
{
	/* Syswake pins can only be used as inputs */
	return 0;
}

static int cashew_pdc_gpio_direction_output(struct gpio_chip *chip,
					    unsigned int offset,
					    int output_value)
{
	/* Syswake pins can only be used as inputs */
	return -EINVAL;
}

static int cashew_pdc_gpio_get(struct gpio_chip *chip, unsigned int offset)
{
	struct cashew_pdc_gpio *priv = to_pdc(chip);

	return readl(priv->reg + 8 * offset) & BIT(0);
}

static void cashew_pdc_gpio_set(struct gpio_chip *chip, unsigned int offset,
				int output_value)
{
	/* Syswake pins can only be used as inputs */
}

static int cashew_pdc_gpio_to_irq(struct gpio_chip *chip, unsigned int offset)
{
	struct cashew_pdc_gpio *priv = to_pdc(chip);

	if (!priv->irq[offset])
		return -EINVAL;

	return priv->irq[offset];
}

static int cashew_pdc_gpio_probe(struct platform_device *pdev)
{
	struct device_node *np = pdev->dev.of_node;
	struct resource *res_regs;
	struct cashew_pdc_gpio *priv;
	unsigned int i;

	if (!np) {
		dev_err(&pdev->dev, "must be instantiated via devicetree\n");
		return -ENOENT;
	}

	res_regs = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!res_regs) {
		dev_err(&pdev->dev, "cannot find registers resource\n");
		return -ENOENT;
	}

	priv = devm_kzalloc(&pdev->dev, sizeof(*priv), GFP_KERNEL);
	if (!priv)
		return -ENOMEM;

	/* Ioremap the registers */
	priv->reg = devm_ioremap(&pdev->dev, res_regs->start,
				 resource_size(res_regs));
	if (!priv->reg) {
		dev_err(&pdev->dev, "unable to ioremap registers\n");
		return -ENOMEM;
	}

	/* Set up GPIO chip */
	priv->chip.label		= "cashew-pdc-gpio";
	priv->chip.dev			= &pdev->dev;
	priv->chip.direction_input	= cashew_pdc_gpio_direction_input;
	priv->chip.direction_output	= cashew_pdc_gpio_direction_output;
	priv->chip.get			= cashew_pdc_gpio_get;
	priv->chip.set			= cashew_pdc_gpio_set;
	priv->chip.to_irq		= cashew_pdc_gpio_to_irq;
	priv->chip.of_node		= np;

	/* GPIO numbering */
	priv->chip.base			= -1;
	priv->chip.ngpio		= GPIO_PDC_NGPIO;

	/* Map the syswake irqs */
	for (i = 0; i < GPIO_PDC_NGPIO; ++i)
		priv->irq[i] = irq_of_parse_and_map(np, i);

	/* Add the GPIO bank */
	gpiochip_add(&priv->chip);

	return 0;
}

static const struct of_device_id cashew_pdc_gpio_of_match[] = {
	{ .compatible = "img,cashew-pdc-gpio" },
	{ },
};

static struct platform_driver cashew_pdc_gpio_driver = {
	.driver = {
		.name		= "cashew-pdc-gpio",
		.of_match_table	= cashew_pdc_gpio_of_match,
	},
	.probe		= cashew_pdc_gpio_probe,
};

static int __init cashew_pdc_gpio_init(void)
{
	return platform_driver_register(&cashew_pdc_gpio_driver);
}
subsys_initcall(cashew_pdc_gpio_init);
