/* Dummy HDMI module */

#include <linux/clk.h>
#include <linux/component.h>
#include <linux/iopoll.h>
#include <linux/math64.h>
#include <linux/module.h>
#include <linux/of_device.h>
#include <linux/regmap.h>
#include <linux/mfd/syscon.h>
#include <drm/drm_atomic_helper.h>
#include <drm/drm_crtc.h>
#include <drm/drm_crtc_helper.h>
#include <drm/drm_of.h>
#include <drm/drm_panel.h>
#include <drm/drmP.h>

#define DRIVER_NAME    "dummy-hdmi"

#define XRES_DEF 1920
#define YRES_DEF 1080

struct dumb_hdmi {
	struct drm_encoder encoder;
	struct drm_connector connector;
	struct device *dev;
	struct platform_device *pdev;
};

static inline struct dumb_hdmi *con_to_hdmi(struct drm_connector *con)
{
	return container_of(con, struct dumb_hdmi, connector);
}

static inline struct dumb_hdmi *encoder_to_hdmi(struct drm_encoder *encoder)
{
	return container_of(encoder, struct dumb_hdmi, encoder);
}

static void dumb_hdmi_encoder_mode_set(struct drm_encoder *encoder,
					struct drm_display_mode *mode,
					struct drm_display_mode *adjusted_mode)
{
}

static void dumb_hdmi_encoder_disable(struct drm_encoder *encoder)
{
}


static bool dumb_hdmi_encoder_mode_fixup(struct drm_encoder *encoder,
					const struct drm_display_mode *mode,
					struct drm_display_mode *adjusted_mode)
{
	return true;
}

static void dumb_hdmi_encoder_commit(struct drm_encoder *encoder)
{
}

static struct drm_encoder_helper_funcs
dumb_hdmi_encoder_helper_funcs = {
	.mode_fixup = dumb_hdmi_encoder_mode_fixup,
	.commit = dumb_hdmi_encoder_commit,
	.mode_set = dumb_hdmi_encoder_mode_set,
	.disable = dumb_hdmi_encoder_disable,
};

static struct drm_encoder_funcs dumb_hdmi_encoder_funcs = {
	.destroy = drm_encoder_cleanup,
};
static int dumb_hdmi_connector_get_modes(struct drm_connector *connector)
{
	struct drm_device *dev = connector->dev;
	int num_modes = 0;

	num_modes = drm_add_modes_noedid(connector,
					 dev->mode_config.max_width,
					 dev->mode_config.max_height);
	drm_set_preferred_mode(connector, XRES_DEF, YRES_DEF);

	DRM_DEBUG_DRIVER("[CONNECTOR:%d:%s] found %d modes\n",
			 connector->base.id,
			 connector->name,
			 num_modes);

	return num_modes;
}

static enum drm_mode_status dumb_hdmi_mode_valid(
					struct drm_connector *connector,
					struct drm_display_mode *mode)
{
	enum drm_mode_status mode_status = MODE_OK;
	return mode_status;
}

static struct drm_encoder *dumb_hdmi_connector_best_encoder(
					struct drm_connector *connector)
{
	struct dumb_hdmi *hdmi = con_to_hdmi(connector);

	return &hdmi->encoder;
}

static struct drm_connector_helper_funcs dumb_hdmi_connector_helper_funcs = {
	.get_modes = dumb_hdmi_connector_get_modes,
	.mode_valid = dumb_hdmi_mode_valid,
	.best_encoder = dumb_hdmi_connector_best_encoder,
};

static enum drm_connector_status
dumb_hdmi_detect(struct drm_connector *connector, bool force)
{
	return connector_status_connected;
}

static void dumb_hdmi_drm_connector_destroy(struct drm_connector *connector)
{
	drm_connector_unregister(connector);
	drm_connector_cleanup(connector);
}

static struct drm_connector_funcs dumb_hdmi_atomic_connector_funcs = {
	.dpms = drm_atomic_helper_connector_dpms,
	.fill_modes = drm_helper_probe_single_connector_modes,
	.detect = dumb_hdmi_detect,
	.destroy = dumb_hdmi_drm_connector_destroy,
	.reset = drm_atomic_helper_connector_reset,
	.atomic_duplicate_state = drm_atomic_helper_connector_duplicate_state,
	.atomic_destroy_state = drm_atomic_helper_connector_destroy_state,
};


static int dumb_hdmi_register(struct drm_device *drm,
				      struct dumb_hdmi *hdmi)
{
	struct drm_encoder *encoder = &hdmi->encoder;
	struct drm_connector *connector = &hdmi->connector;
	struct device *dev = hdmi->dev;
	int ret;

	encoder->possible_crtcs = 1;
	drm_encoder_helper_add(&hdmi->encoder,
			       &dumb_hdmi_encoder_helper_funcs);
	ret = drm_encoder_init(drm, &hdmi->encoder, &dumb_hdmi_encoder_funcs,
			 DRM_MODE_ENCODER_TMDS);
	if (ret) {
		dev_err(dev, "Failed to initialize encoder with drm\n");
		return ret;
	}

	drm_connector_helper_add(connector,
			&dumb_hdmi_connector_helper_funcs);

	drm_connector_init(drm, &hdmi->connector,
			   &dumb_hdmi_atomic_connector_funcs,
			   DRM_MODE_CONNECTOR_HDMIA);

	ret = drm_connector_register(connector);
	if (ret) {
		return ret;
	}

	drm_mode_connector_attach_encoder(connector, encoder);
	return 0;
}

static const struct of_device_id dumb_hdmi_dt_ids[] = {
	{
		.compatible = "dummy-hdmi",
	},
	{},
};

static int dumb_hdmi_bind(struct device *dev, struct device *master,
			     void *data)
{
	struct dumb_hdmi *hdmi;
	int ret;

	hdmi = devm_kzalloc(dev, sizeof(*hdmi), GFP_KERNEL);
	if (!hdmi)
		return -ENOMEM;

	hdmi->dev = dev;

	ret = dumb_hdmi_register(data, hdmi);
	if (ret) {
		dev_err(dev, "Failed to register hdmi: %d\n", ret);
	}

	dev_set_drvdata(dev, hdmi);

	return ret;
}

static void dumb_hdmi_unbind(struct device *dev, struct device *master, void *data)
{
}

static const struct component_ops dumb_hdmi_ops = {
	.bind	= dumb_hdmi_bind,
	.unbind	= dumb_hdmi_unbind,
};

static int dumb_hdmi_probe(struct platform_device *pdev)
{
	return component_add(&pdev->dev, &dumb_hdmi_ops);
}

static int dumb_hdmi_remove(struct platform_device *pdev)
{
	component_del(&pdev->dev, &dumb_hdmi_ops);
	return 0;
}

struct platform_driver dumb_hdmi_driver = {
	.probe		= dumb_hdmi_probe,
	.remove		= dumb_hdmi_remove,
	.driver		= {
		.of_match_table = dumb_hdmi_dt_ids,
		.name 		= DRIVER_NAME,
	},
};
module_platform_driver(dumb_hdmi_driver);

MODULE_DESCRIPTION("Imagination Technologies Dummy HDMI Driver");
MODULE_AUTHOR("Will Thomas <will.thomas@imgtec.com>");
MODULE_LICENSE("GPL");
MODULE_ALIAS("platform:" DRIVER_NAME);
