/*
 * Copyright (c) 2016, Imagination Technologies plc
 * Based on dw_hdmi-rockchip.c
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <linux/component.h>
#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/mfd/syscon.h>
#include <linux/regmap.h>
#include <drm/drm_of.h>
#include <drm/drmP.h>
#include <drm/drm_crtc_helper.h>
#include <drm/drm_edid.h>
#include <drm/drm_encoder_slave.h>
#include <drm/bridge/dw_hdmi.h>

struct cashew_hdmi {
	struct device *dev;
	struct regmap *regmap;
	struct drm_encoder encoder;
};

#define to_cashew_hdmi(x)	container_of(x, struct cashew_hdmi, x)

static const u32 valid_freqs[] = {
	/* FIXME: 594MHz doesn't work, was disabled */
	25175000, 27000000, 29400000, 54000000,
	59400000, 72000000, 74250000, 82500000,
	90000000, 99000000, 108000000, 118800000,
	148500000, 165000000, 185625000, 198000000,
	297000000, 371250000, 495000000, /* 594000000 */
	~0UL, };

/* TODO add modes to this... */
static const struct dw_hdmi_mpll_config cashew_mpll_cfg[] = {
	{
		27000000, {
			{ 0x00b3, 0x0000},
			{ 0x2153, 0x0000},
			{ 0x40f3, 0x0000}
		},
	},	{
		148500000, {
			{ 0x0051, 0x0002},
			{ 0x214c, 0x0003},
			{ 0x4064, 0x0003}
		},
	},	{
		297000000, {
			{ 0x0040, 0x0003},
			{ 0x0000, 0x0000},
			{ 0x0000, 0x0000},
		},
	},	{
		594000000, {
			{ 0x1a40, 0x0003},
			{ 0x0000, 0x0000},
			{ 0x0000, 0x0000},
		},
	},

};

/*
 * The following information extracted from "Table B-1 MPLL Generic
 * Configuration Settings of the phy databook" and covers all
 * frequencies of CEA-861-F spec.
 * TODO add modes to this....
 */
static const struct dw_hdmi_curr_ctrl cashew_cur_ctr[] = {
	/*      pixelclk    bpp8    bpp10   bpp12 */
	{
		40000000,  { 0x0018, 0x0018, 0x0018 },
	}, {
		65000000,  { 0x0028, 0x0028, 0x0028 },
	}, {
		66000000,  { 0x0038, 0x0038, 0x0038 },
	}, {
		74250000,  { 0x0028, 0x0038, 0x0038 },
	}, {
		83500000,  { 0x0028, 0x0038, 0x0038 },
	}, {
		146250000, { 0x0038, 0x0038, 0x0038 },
	}, {
		148500000, { 0x0038, 0x0038, 0x0038 },
	}, {
		297000000, { 0x0038, 0x0038, 0x0038 },
	}, {
		594000000, { 0x0018, 0x0018, 0x0018 },
	}, {
		~0UL,      { 0x0000, 0x0000, 0x0000},
	}
};

static const struct dw_hdmi_phy_config cashew_phy_config[] = {
	/* This table covers all frequencies.
	 * The calculation is based on Apendix A of
	 * dwc_hdmi20_tx_ns_6gbps_tsmc_28hpm18 Databook
	 * and Synopsys driver (table in phy_303.c)
	 * We are only concerend with color_depth 8 and 10
	 * Note that in some frequencies the vlev is different for
	 * color depth of 8 and 10, and the value for color_depth 10
	 * has been selected.
	 */
	/*pixelclk   symbol   term   vlev*/
	{ 74250000,  0x8009, 0x0004, 0x0272},
	{ 148500000, 0x802b, 0x0004, 0x028d},
	{ 297000000, 0x8039, 0x0005, 0x028d},
	{ 594000000, 0x8039, 0x0004, 0x014a},
	{ ~0UL,	     0x0000, 0x0000, 0x0000}
};

static enum drm_mode_status
dw_hdmi_cashew_mode_valid(struct drm_connector *connector,
			  struct drm_display_mode *mode)
{
	int pclk = mode->clock * 1000;
	int i;

	for (i = 0; valid_freqs[i] != (~0UL); i++) {
		if (pclk == valid_freqs[i]) {
			return MODE_OK;
		}
	}

	return MODE_BAD;
}

static struct drm_encoder_funcs dw_hdmi_cashew_encoder_funcs = {
	.destroy = drm_encoder_cleanup,
};

static void dw_hdmi_cashew_encoder_disable(struct drm_encoder *encoder)
{
}

static bool
dw_hdmi_cashew_encoder_mode_fixup(struct drm_encoder *encoder,
				  const struct drm_display_mode *mode,
				  struct drm_display_mode *adj_mode)
{
	return true;
}

static void dw_hdmi_cashew_encoder_mode_set(struct drm_encoder *encoder,
					    struct drm_display_mode *mode,
					    struct drm_display_mode *adj_mode)
{
}

static void dw_hdmi_cashew_encoder_commit(struct drm_encoder *encoder)
{
}

static void dw_hdmi_cashew_encoder_prepare(struct drm_encoder *encoder)
{
}

static struct drm_encoder_helper_funcs dw_hdmi_cashew_encoder_helper_funcs = {
	.mode_fixup = dw_hdmi_cashew_encoder_mode_fixup,
	.mode_set   = dw_hdmi_cashew_encoder_mode_set,
	.prepare    = dw_hdmi_cashew_encoder_prepare,
	.commit     = dw_hdmi_cashew_encoder_commit,
	.disable    = dw_hdmi_cashew_encoder_disable,
};

static const struct dw_hdmi_plat_data cashew_hdmi_drv_data = {
	.mode_valid = dw_hdmi_cashew_mode_valid,
	.mpll_cfg   = cashew_mpll_cfg,
	.cur_ctr    = cashew_cur_ctr,
	.phy_config = cashew_phy_config,
	.dev_type   = CASHEW_HDMI,
};

static const struct of_device_id dw_hdmi_cashew_dt_ids[] = {
	{ .compatible = "img,cashew-dw-hdmi",
	  .data = &cashew_hdmi_drv_data
	},
	{},
};
MODULE_DEVICE_TABLE(of, dw_hdmi_cashew_dt_ids);

int dw_hdmi_cashew_bind(struct device *dev, struct device *master,
			void *data)
{
	struct platform_device *pdev = to_platform_device(dev);
	const struct dw_hdmi_plat_data *plat_data;
	const struct of_device_id *match;
	struct drm_device *drm = data;
	struct drm_encoder *encoder;
	struct cashew_hdmi *hdmi;
	struct resource *iores;
	int irq;

	if (!pdev->dev.of_node) {
		dev_err(dev, "Could not find dts node\n");
		return -ENODEV;
	}

	hdmi = devm_kzalloc(&pdev->dev, sizeof(*hdmi), GFP_KERNEL);
	if (!hdmi)
		return -ENOMEM;

	match = of_match_node(dw_hdmi_cashew_dt_ids, pdev->dev.of_node);
	plat_data = match->data;
	hdmi->dev = &pdev->dev;
	encoder = &hdmi->encoder;

	irq = platform_get_irq(pdev, 0);
	if (irq < 0) {
		dev_err(dev, "Could not get irq\n");
		return irq;
	}

	iores = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!iores) {
		dev_err(dev, "Could not get io resource\n");
		return -ENXIO;
	}

	platform_set_drvdata(pdev, hdmi);
	encoder->possible_crtcs = 1;
	drm_encoder_helper_add(encoder, &dw_hdmi_cashew_encoder_helper_funcs);
	drm_encoder_init(drm, encoder, &dw_hdmi_cashew_encoder_funcs,
			 DRM_MODE_ENCODER_TMDS);

	return dw_hdmi_bind(dev, master, data, encoder, iores, irq, plat_data);
}

static void dw_hdmi_cashew_unbind(struct device *dev, struct device *master,
				  void *data)
{
	return dw_hdmi_unbind(dev, master, data);
}

static const struct component_ops dw_hdmi_cashew_ops = {
	.bind	= dw_hdmi_cashew_bind,
	.unbind	= dw_hdmi_cashew_unbind,
};

static int dw_hdmi_cashew_probe(struct platform_device *pdev)
{
	return component_add(&pdev->dev, &dw_hdmi_cashew_ops);
}

static int dw_hdmi_cashew_remove(struct platform_device *pdev)
{
	component_del(&pdev->dev, &dw_hdmi_cashew_ops);

	return 0;
}

static int dw_hdmi_cashew_suspend(struct device *dev)
{
	return dw_hdmi_suspend(dev);
}

static int dw_hdmi_cashew_resume(struct device *dev)
{
	return dw_hdmi_resume(dev);
}

static UNIVERSAL_DEV_PM_OPS(dw_hdmi_cashew_pm,
			    dw_hdmi_cashew_suspend,
			    dw_hdmi_cashew_resume,
			    NULL);

static struct platform_driver dw_hdmi_cashew_pltfm_driver = {
	.probe  = dw_hdmi_cashew_probe,
	.remove = dw_hdmi_cashew_remove,
	.driver = {
		.name = "dwhdmi-cashew",
		.of_match_table = dw_hdmi_cashew_dt_ids,
		.pm = &dw_hdmi_cashew_pm,
	},
};

module_platform_driver(dw_hdmi_cashew_pltfm_driver);

MODULE_AUTHOR("Will Thomas <will.thomas@imgtec.com>");
MODULE_DESCRIPTION("Cashew DW-HDMI Driver Extension");
MODULE_LICENSE("GPL");
MODULE_ALIAS("platform:dwhdmi-cashew");
