#include <linux/fence.h>
#include <linux/reservation.h>
#include <linux/of_irq.h>
#include <linux/pm_runtime.h>

#include <drm/drmP.h>
#include <drm/drm_atomic_helper.h>
#include <drm/drm_crtc.h>
#include <drm/drm_crtc_helper.h>
#include <drm/drm_modes.h>
#include <drm/drm_plane_helper.h>

#include "pdp_crtc.h"
#include "pdp_drv.h"
#include "pdp_plane.h"
#include "pdp_plane_funcs.h"
#include "pdp_tiler.h"

#include "pdpfb2_regs.h"

static void pdp_crtc_finish_page_flip(struct pdp_crtc *pcrtc);

static irqreturn_t pdp_irq_handler(int irq, void *data)
{
	struct pdp_private *priv = data;
	struct drm_device *dev = priv->dev;
	u32 status;

	status = pdp_read(dev, PDP_INTSTAT_OFFSET);
	pdp_write(dev, PDP_INTCLR_OFFSET, status);

	if (status & PDP_INTSTAT_INTS_VBLNK0_MASK) {
		pdp_tiler_fix_faults(priv);
		drm_handle_vblank(dev, 0);
		pdp_crtc_finish_page_flip(priv->crtc);
		DRM_DEBUG_DRIVER("vblank irq handeled\n");
	} else {
		DRM_DEBUG_DRIVER("ignored non vblank irq\n");
	}
	return IRQ_HANDLED;
}

static void pdp_crtc_sync_gen(struct pdp_crtc *pcrtc, bool enable)
{
	struct pdp_private *priv = pcrtc->priv;
	u32 reg;

	reg = pdp_read(priv->dev, PDP_SYNCCTRL_OFFSET);
	SET_FIELD(reg, PDP_SYNCCTRL_PDP_RST, 1);
	pdp_write(priv->dev, PDP_SYNCCTRL_OFFSET, reg);
	SET_FIELD(reg, PDP_SYNCCTRL_PDP_RST, 0);
	pdp_write(priv->dev, PDP_SYNCCTRL_OFFSET, reg);

	DRM_DEBUG_DRIVER("Setting output sync active %s\n",
			 enable ? "on" : "off");
	SET_FIELD(reg, PDP_SYNCCTRL_SYNCACTIVE, enable);
	pdp_write(priv->dev, PDP_SYNCCTRL_OFFSET, reg);
}

/*
struct csc_coeffs {
	int ru, ry, rv;
	int gy, gu, gv;
	int by, bu, bv;
} csc_config = {
	.ry = 298,	.rv = 459,	.ru = 0,
	.gy = 289,	.gv = -137,	.gu = -55,
	.by = 289,	.bv = 0,	.bu = 541,
};

static void pdp_crtc_set_csc_coeff(struct pdp_crtc *pcrtc)
{
	struct pdp_private *priv = pcrtc->priv;
	u32 reg;

	DRM_DEBUG_DRIVER("Setting CSC config\n");
	pdp_write(priv->dev, PDP_CSCCOEFF0_OFFSET,
		  PLACE_FIELD(PDP_CSCCOEFF0_CSCCOEFFRU, csc_config.ru)
		  | PLACE_FIELD(PDP_CSCCOEFF0_CSCCOEFFRY, csc_config.ry));
	pdp_write(priv->dev, PDP_CSCCOEFF1_OFFSET,
		  PLACE_FIELD(PDP_CSCCOEFF1_CSCCOEFFGY, csc_config.gy)
		  | PLACE_FIELD(PDP_CSCCOEFF1_CSCCOEFFRV, csc_config.rv));
	pdp_write(priv->dev, PDP_CSCCOEFF2_OFFSET,
		  PLACE_FIELD(PDP_CSCCOEFF2_CSCCOEFFGV, csc_config.gv)
		  | PLACE_FIELD(PDP_CSCCOEFF2_CSCCOEFFGU, csc_config.gu));
	pdp_write(priv->dev, PDP_CSCCOEFF3_OFFSET,
		  PLACE_FIELD(PDP_CSCCOEFF3_CSCCOEFFBU, csc_config.bu)
		  | PLACE_FIELD(PDP_CSCCOEFF3_CSCCOEFFBY, csc_config.by));
	pdp_write(priv->dev, PDP_CSCCOEFF4_OFFSET,
		  PLACE_FIELD(PDP_CSCCOEFF4_CSCCOEFFBV, csc_config.bv));
} */

void pdp_crtc_set_background(struct pdp_crtc *pcrtc,
			     u16 a, u16 r, u16 g, u16 b)
{
	struct pdp_private *priv =  pcrtc->priv;
	u32 reg;

	DRM_DEBUG_DRIVER("Setting background colour %x %x %x %x\n", a, r, g, b);
	reg = pdp_read(priv->dev, PDP_BGNDCOL_AR_OFFSET);
	SET_FIELD(reg, PDP_BGNDCOL_AR_BGNDCOL_A, a);
	SET_FIELD(reg, PDP_BGNDCOL_AR_BGNDCOL_R, r);
	pdp_write(priv->dev, PDP_BGNDCOL_AR_OFFSET, reg);

	reg = pdp_read(priv->dev, PDP_BGNDCOL_GB_OFFSET);
	SET_FIELD(reg, PDP_BGNDCOL_GB_BGNDCOL_G, g);
	SET_FIELD(reg, PDP_BGNDCOL_GB_BGNDCOL_B, b);
	pdp_write(priv->dev, PDP_BGNDCOL_GB_OFFSET, reg);

	pdp_plane_update_flush_regs(priv->dev);
}

static void pdp_crtc_set_gamma(struct drm_crtc *crtc, u16 *r, u16 *g, u16 *b,
			       uint32_t start, uint32_t size)
{
	struct pdp_crtc *pcrtc = to_pdp_crtc(crtc);
	struct pdp_private *priv = pcrtc->priv;
	struct drm_device *dev = priv->dev;
	u32 reg;
	int i;

	/*
	 * This should use field masks etc but Non europa PDP's have a
	 * very different layout so this code wouldnt have been portable
	 * anyways.
	 */
	for (i = 0; i < size; i++) {
		/* write to full gamma R register */
		pdp_write(dev, PDP_GAMMA0_R_OFFSET + (i * 2 * 4), r[i]);

		/* write G and B regsiter */
		reg = ((u32)g[i] << 16) | b[i];
		pdp_write(dev, PDP_GAMMA0_GB_OFFSET + (i * 2 * 4), reg);
	}
}

static void pdp_crtc_set_display_timings(struct pdp_crtc *pcrtc)
{
	const struct drm_display_mode *mode = &pcrtc->crtc.state->adjusted_mode;
	struct pdp_private *priv = pcrtc->priv;
	struct drm_device *dev = priv->dev;

	unsigned long mode_clock = mode->clock * 1000;
	uint32_t ht = mode->htotal;
	uint32_t hbps = mode->hsync_end - mode->hsync_start;
	uint32_t has = (mode->htotal - mode->hsync_start);
	uint32_t hlbs = has;
	uint32_t hfps = (hlbs + mode->hdisplay);
	uint32_t hrbs = hfps;
	uint32_t vt = mode->vtotal;
	uint32_t vbps = mode->vsync_end - mode->vsync_start;
	uint32_t vas = (mode->vtotal - mode->vsync_start);
	uint32_t vtbs = vas;
	uint32_t vfps = (vtbs + mode->vdisplay);
	uint32_t vbbs = vfps;

	u32 reg;

	DRM_DEBUG_DRIVER("Setting CRTC mode to %dx%d\n",
			 mode->hdisplay, mode->vdisplay);

	DRM_DEBUG_DRIVER("Setting px clk %lu\n", mode_clock);
	clk_set_rate(priv->pixel_clk, mode_clock);

	reg = pdp_read(dev, PDP_VSYNC1_OFFSET);
	SET_FIELD(reg, PDP_VSYNC1_VBPS, vbps);
	SET_FIELD(reg, PDP_VSYNC1_VT, vt);
	pdp_write(dev, PDP_VSYNC1_OFFSET, reg);

	reg = pdp_read(dev, PDP_VSYNC2_OFFSET);
	SET_FIELD(reg, PDP_VSYNC2_VAS, vas);
	SET_FIELD(reg, PDP_VSYNC2_VTBS, vtbs);
	pdp_write(dev, PDP_VSYNC2_OFFSET, reg);

	reg = pdp_read(dev, PDP_VSYNC3_OFFSET);
	SET_FIELD(reg, PDP_VSYNC3_VFPS, vfps);
	SET_FIELD(reg, PDP_VSYNC3_VBBS, vbbs);
	pdp_write(dev, PDP_VSYNC3_OFFSET, reg);

	reg = pdp_read(dev, PDP_VEVENT_OFFSET);
	SET_FIELD(reg, PDP_VEVENT_VFETCH, vbps);
	SET_FIELD(reg, PDP_VEVENT_VEVENT, vfps);
	pdp_write(dev, PDP_VEVENT_OFFSET, reg);

	reg = pdp_read(dev, PDP_VDECTRL_OFFSET);
	SET_FIELD(reg, PDP_VDECTRL_VDES, vtbs);
	SET_FIELD(reg, PDP_VDECTRL_VDEF, vfps);
	pdp_write(dev, PDP_VDECTRL_OFFSET, reg);

	reg = pdp_read(dev, PDP_HSYNC1_OFFSET);
	SET_FIELD(reg, PDP_HSYNC1_HBPS, hbps);
	SET_FIELD(reg, PDP_HSYNC1_HT, ht);
	pdp_write(dev, PDP_HSYNC1_OFFSET, reg);

	reg = pdp_read(dev, PDP_HSYNC2_OFFSET);
	SET_FIELD(reg, PDP_HSYNC2_HAS, has);
	SET_FIELD(reg, PDP_HSYNC2_HLBS, hlbs);
	pdp_write(dev, PDP_HSYNC2_OFFSET, reg);

	reg = pdp_read(dev, PDP_HSYNC3_OFFSET);
	SET_FIELD(reg, PDP_HSYNC3_HFPS, hfps);
	SET_FIELD(reg, PDP_HSYNC3_HRBS, hrbs);
	pdp_write(dev, PDP_HSYNC3_OFFSET, reg);

	reg = pdp_read(dev, PDP_HDECTRL_OFFSET);
	SET_FIELD(reg, PDP_HDECTRL_HDES, hlbs);
	SET_FIELD(reg, PDP_HDECTRL_HDEF, hfps);
	pdp_write(dev, PDP_HDECTRL_OFFSET, reg);

	/* Setting polarities of sync */
	reg = pdp_read(dev, PDP_SYNCCTRL_OFFSET);
	if (mode->flags & DRM_MODE_FLAG_NHSYNC)
		SET_FIELD(reg, PDP_SYNCCTRL_HSPOL, 1);
	else
		SET_FIELD(reg, PDP_SYNCCTRL_HSPOL, 0);

	if (mode->flags & DRM_MODE_FLAG_NVSYNC)
		SET_FIELD(reg, PDP_SYNCCTRL_VSPOL, 1);
	else
		SET_FIELD(reg, PDP_SYNCCTRL_VSPOL, 0);
	/* Set blanking polatity */
	SET_FIELD(reg, PDP_SYNCCTRL_BLNKPOL, 1);

	pdp_write(dev, PDP_SYNCCTRL_OFFSET, reg);
}

void pdp_crtc_cancel_page_flip(struct pdp_crtc *pcrtc, struct drm_file *file)
{
	struct drm_pending_vblank_event *event;
	struct drm_device *dev = pcrtc->crtc.dev;
	unsigned long flags;

	spin_lock_irqsave(&dev->event_lock, flags);
	event = pcrtc->event;
	if (event && event->base.file_priv == file) {
		pcrtc->event = NULL;
		event->base.destroy(&event->base);
		drm_crtc_vblank_put(&pcrtc->crtc);
	}
	spin_unlock_irqrestore(&dev->event_lock, flags);
}

static void pdp_crtc_finish_page_flip(struct pdp_crtc *pcrtc)
{
	struct drm_pending_vblank_event *event;
	struct drm_device *dev = pcrtc->crtc.dev;
	unsigned long flags;

	spin_lock_irqsave(&dev->event_lock, flags);
	event = pcrtc->event;
	pcrtc->event = NULL;
	spin_unlock_irqrestore(&dev->event_lock, flags);

	if (!event)
		return;

	spin_lock_irqsave(&dev->event_lock, flags);
	drm_send_vblank_event(dev, 0, event);
	wake_up(&pcrtc->flip_wait);
	spin_unlock_irqrestore(&dev->event_lock, flags);

	drm_crtc_vblank_put(&pcrtc->crtc);
}

static bool pdp_crtc_page_flip_pending(struct pdp_crtc *pcrtc)
{
	struct drm_device *dev = pcrtc->crtc.dev;
	unsigned long flags;
	bool pending;

	spin_lock_irqsave(&dev->event_lock, flags);
	pending = pcrtc->event;
	spin_unlock_irqrestore(&dev->event_lock, flags);

	return pending;
}

static void pdp_crtc_wait_page_flip(struct pdp_crtc *pcrtc)
{
	if (wait_event_timeout(pcrtc->flip_wait,
			       !pdp_crtc_page_flip_pending(pcrtc),
			       msecs_to_jiffies(50)))
		return;
	DRM_INFO("page flip timeout\n");

	pdp_crtc_finish_page_flip(pcrtc);
}

static void pdp_crtc_start(struct pdp_crtc *pcrtc)
{
	struct drm_crtc *crtc = &pcrtc->crtc;

	if (pcrtc->running)
		return;

	pdp_crtc_set_display_timings(pcrtc);
	pdp_configure_dithering(pcrtc->priv);
	drm_crtc_vblank_on(crtc);
	pdp_crtc_sync_gen(pcrtc, true);

	pcrtc->running = true;
}

static void pdp_crtc_stop(struct pdp_crtc *pcrtc)
{
	struct drm_crtc *crtc = &pcrtc->crtc;

	if (!pcrtc->running)
		return;

	pdp_crtc_wait_page_flip(pcrtc);
	pdp_crtc_sync_gen(pcrtc, false);
	drm_crtc_vblank_off(crtc);

	pcrtc->running = false;
}

static int pdp_crtc_clock_enable(struct pdp_crtc *pcrtc)
{
	struct pdp_private *priv = pcrtc->priv;
	int ret;

	ret = clk_enable(priv->pixel_clk);
	if (ret < 0)
		DRM_DEBUG("Could not enable pixel clock\n");

	return ret;
}

static void pdp_crtc_clock_disable(struct pdp_crtc *pcrtc)
{
	struct pdp_private *priv = pcrtc->priv;

	clk_disable(priv->pixel_clk);
}

/************************************************************************/
/*    CRTC helper functions*/

static bool pdp_crtc_helper_mode_fixup(struct drm_crtc *crtc,
				       const struct drm_display_mode *mode,
				       struct drm_display_mode *adjusted_mode)
{
	return true;
}

static void pdp_crtc_helper_atomic_begin(struct drm_crtc *crtc)
{
	struct drm_pending_vblank_event *event = crtc->state->event;
	struct pdp_crtc *pcrtc = to_pdp_crtc(crtc);
	struct drm_device *dev = crtc->dev;
	unsigned long flags;

	if (event) {
		WARN_ON(drm_crtc_vblank_get(crtc) < 0);

		spin_lock_irqsave(&dev->event_lock, flags);
		pcrtc->event = event;
		event = NULL;
		spin_unlock_irqrestore(&dev->event_lock, flags);
	}
}

static void pdp_crtc_helper_atomic_flush(struct drm_crtc *crtc)
{
}

static void pdp_crtc_helper_enable(struct drm_crtc *crtc)
{
	struct pdp_crtc *pcrtc = to_pdp_crtc(crtc);
	int ret;

	if (pcrtc->enabled)
		return;

	ret = pm_runtime_get_sync(crtc->dev->dev);
	if (ret < 0) {
		DRM_ERROR("Failed to enable power domain: %d\n", ret);
		return;
	}

	pdp_crtc_clock_enable(pcrtc);
	pdp_crtc_start(pcrtc);

	pcrtc->enabled = true;
}

static void pdp_crtc_helper_disable(struct drm_crtc *crtc)
{
	struct pdp_crtc *pcrtc = to_pdp_crtc(crtc);

	if (!pcrtc->enabled)
		return;

	pdp_crtc_stop(pcrtc);
	pdp_crtc_clock_disable(pcrtc);
	pm_runtime_put(crtc->dev->dev);

	pcrtc->enabled = false;
}

static const struct drm_crtc_helper_funcs pdp_crtc_helper_funcs = {
	.mode_fixup = pdp_crtc_helper_mode_fixup,
	.disable = pdp_crtc_helper_disable,
	.enable = pdp_crtc_helper_enable,
	.atomic_flush = pdp_crtc_helper_atomic_flush,
	.atomic_begin = pdp_crtc_helper_atomic_begin,
};

static const struct drm_crtc_funcs pdp_crtc_funcs = {
	.atomic_destroy_state = drm_atomic_helper_crtc_destroy_state,
	.atomic_duplicate_state = drm_atomic_helper_crtc_duplicate_state,
	.destroy = drm_crtc_cleanup,
	.gamma_set = pdp_crtc_set_gamma,
	.page_flip = drm_atomic_helper_page_flip,
	.reset = drm_atomic_helper_crtc_reset,
	.set_config = drm_atomic_helper_set_config,
};

int pdp_crtc_create(struct drm_device *dev, uint32_t number)
{
	struct platform_device *pdev = to_platform_device(dev->dev);
	struct pdp_private *priv = dev->dev_private;
	struct pdp_crtc *pcrtc;
	int ret;

	pcrtc = devm_kzalloc(dev->dev, sizeof(*pcrtc), GFP_KERNEL);
	if (pcrtc < 0)
		return PTR_ERR(pcrtc);
	pdp_planes_init(dev);
	ret = drm_crtc_init_with_planes(dev, &pcrtc->crtc,
					&priv->planes[0].plane,
					&priv->planes[3].plane,
					&pdp_crtc_funcs);

	if (ret < 0) {
		dev_err(dev->dev, "cant init crtc\n");
		return ret;
	}

	ret = drm_mode_crtc_set_gamma_size(&pcrtc->crtc, 33);
	if (ret < 0) {
		dev_err(dev->dev, "Could not allocate CRCT gamma table\n");
		return ret;
	}

	init_waitqueue_head(&pcrtc->flip_wait);

	pcrtc->priv = priv;
	pcrtc->enabled = false;

	drm_crtc_helper_add(&pcrtc->crtc, &pdp_crtc_helper_funcs);
	drm_crtc_vblank_reset(&pcrtc->crtc);

	priv->irq = irq_of_parse_and_map(pdev->dev.of_node, 0);
	if (priv->irq < 0) {
		dev_err(dev->dev, "no IRQ for CRTC\n");
		return priv->irq;
	}
	ret = devm_request_irq(dev->dev, priv->irq, pdp_irq_handler, 0, "pdp",
			       priv);
	if (ret < 0) {
		dev_err(dev->dev, "cannot register irq %d for CRTC\n",
			priv->irq);
		return ret;
	}

	DRM_DEBUG_DRIVER("registered irqs and crtc\n");

	priv->crtc = pcrtc;

	return 0;
}

void pdp_crtc_enable_vblank(struct pdp_crtc *pcrtc, bool enable)
{
	struct drm_crtc *crtc = &pcrtc->crtc;
	struct drm_device *dev = crtc->dev;
	int reg;

	DRM_DEBUG_DRIVER("Setting vblank: %s\n", enable ? "yes" : "no");

	reg = pdp_read(dev, PDP_INTENAB_OFFSET);
	SET_FIELD(reg, PDP_INTENAB_INTEN_VBLNK0, enable ? 1 : 0);
	pdp_write(dev, PDP_INTENAB_OFFSET, reg);
}
