#include <linux/module.h>
#include <linux/clk.h>
#include <linux/component.h>
#include <linux/kernel.h>
#include <linux/of_platform.h>
#include <linux/of_graph.h>
#include <linux/pm_runtime.h>
#include <linux/sizes.h>

#include <linux/img-iommu.h>
#include <linux/iommu.h>

#include <drm/drmP.h>
#include <drm/pdp_drm.h>
#include <drm/drm_fb_cma_helper.h>
#include <drm/drm_gem.h>
#include <drm/drm_gem_cma_helper.h>
#include <drm/drm_atomic_helper.h>
#include <drm/drm_of.h>

#include "pdp_crtc.h"
#include "pdp_drv.h"
#include "pdp_modeset.h"
#include "pdp_fb.h"
#include "pdpfb2_regs.h"
#include "pdp_plane_funcs.h"

#define DRIVER_NAME	"pdp-drm"
#define DRIVER_DESC	"Img. Tec. Cashew soc DRM"
#define DRIVER_DATE	"20151010"
#define DRIVER_MAJOR	1
#define DRIVER_MINOR	0

u32 pdp_read(struct drm_device *dev, unsigned int reg)
{
	struct pdp_private *priv = dev->dev_private;

	return ioread32(priv->reg + reg);
}

void pdp_write(struct drm_device *dev, unsigned int reg_offs,
	       unsigned int data)
{
	struct pdp_private *priv = dev->dev_private;

	iowrite32(data, priv->reg + reg_offs);
}

static int pdp_load_mem(struct drm_device *dev)
{
	struct platform_device *pdev = dev->platformdev;
	struct pdp_private *priv = dev->dev_private;

	struct resource *res;

	res = platform_get_resource_byname(pdev, IORESOURCE_MEM,
					   "pdp_reg");
	if (!res) {
		dev_err(dev->dev, "no pdp register ioresource\n");
		return -ENODEV;
	}
	priv->reg = devm_ioremap_resource(&pdev->dev, res);
	if (!priv->reg)
		return -ENOMEM;

	res = platform_get_resource_byname(pdev, IORESOURCE_MEM,
					   "cma");
	return pdp_tiler_init(priv, res);
}

static int pdp_load_clk(struct drm_device *dev)
{
	struct pdp_private *priv = dev->dev_private;
	int ret = 0;

	priv->pdp_clk = devm_clk_get(dev->dev, "core");
	if (IS_ERR(priv->pdp_clk)) {
		DRM_ERROR("could not get core clock resource\n");
		return -ENOENT;
	}

	ret = clk_prepare_enable(priv->pdp_clk);
	if (ret < 0) {
		DRM_ERROR("Failed to enable pdp core clock: %d\n", ret);
		return ret;
	}

	priv->pixel_clk = devm_clk_get(dev->dev, "pixel");
	if (IS_ERR(priv->pixel_clk)) {
		DRM_ERROR("could not get pixel clock resource\n");
		clk_disable_unprepare(priv->pdp_clk);
		return -ENOENT;
	}

	ret = clk_prepare(priv->pixel_clk);
	if (ret < 0) {
		DRM_ERROR("Failed to prepare pdp pixel clock: %d\n", ret);
		clk_disable_unprepare(priv->pdp_clk);
		return ret;
	}

	return ret;
}

static int pdp_load(struct drm_device *dev, unsigned long flags)
{
	struct platform_device *pdev = dev->platformdev;
	struct pdp_private *priv;
	int err = 0;

	priv = kzalloc(sizeof(*priv), GFP_KERNEL);
	if (!priv)
		return -ENOMEM;

	platform_set_drvdata(pdev, priv);

	init_waitqueue_head(&priv->commit.wait);

	dev->dev_private = priv;
	priv->dev = dev;
#ifdef __PDP_TARGET_IS_SIMULATOR__
	priv->auto_kick = true;
#else
	priv->auto_kick = false;
#endif

	err = pdp_load_mem(dev);
	if (err < 0) {
		DRM_ERROR("cannot load memio\n");
		goto err_free_priv;
	}

	err = pdp_load_clk(dev);
	if (err)
		goto err_free_mem;

	err = drm_vblank_init(priv->dev, 1);
	if (err) {
		DRM_ERROR("failed to complete vblank init (err=%d)\n", err);
		goto err_free_clk;
	}
	dev->irq_enabled = true;

	err = pdp_modeset_init(priv);
	if (err) {
		DRM_ERROR("modeset initialisation failed (err=%d)\n", err);
		goto err_free_vblank;
	}

	err = component_bind_all(dev->dev, dev);
	if (err)
		goto err_free_vblank;

	drm_kms_helper_poll_init(dev);

	drm_mode_config_reset(dev);

	pm_runtime_enable(dev->dev);

#ifdef CONFIG_DRM_IMG_PDP_FBDEV
	priv->fbdev = drm_fbdev_cma_init(dev, 32, 1, dev->mode_config.num_connector);
#endif
	return 0;

err_free_vblank:
	drm_vblank_cleanup(dev);
err_free_clk:
	clk_disable_unprepare(priv->pdp_clk);
	clk_unprepare(priv->pixel_clk);
err_free_mem:
err_free_priv:
	drm_mode_config_cleanup(dev);
	dev->dev_private = NULL;
	return err;
}

static int pdp_unload(struct drm_device *dev)
{
	struct pdp_private *priv = dev->dev_private;

	if (priv->fbdev)
		drm_fbdev_cma_fini(priv->fbdev);

	drm_vblank_cleanup(priv->dev);
	drm_kms_helper_poll_fini(dev);
	component_unbind_all(dev->dev, dev);
	pm_runtime_disable(dev->dev);
	drm_mode_config_cleanup(dev);
	dev->irq_enabled = 0;
	dev->dev_private = NULL;

	kfree(priv);

	return 0;
}

static void pdp_preclose(struct drm_device *dev, struct drm_file *file)
{
	struct pdp_private *priv = dev->dev_private;

	pdp_crtc_cancel_page_flip(priv->crtc, file);
}

static void pdp_lastclose(struct drm_device *dev)
{
	struct pdp_private *priv = dev->dev_private;

	drm_fbdev_cma_restore_mode(priv->fbdev);
}

static int pdp_enable_vblank(struct drm_device *dev, int crtc)
{
	struct pdp_private *priv = dev->dev_private;

	pdp_crtc_enable_vblank(priv->crtc, true);

	DRM_DEBUG_DRIVER("vblank interrupts enabled for crtc %d\n", crtc);

	return 0;
}

static void pdp_disable_vblank(struct drm_device *dev, int crtc)
{
	struct pdp_private *priv = dev->dev_private;

	pdp_crtc_enable_vblank(priv->crtc, false);

	DRM_DEBUG_DRIVER("vblank interrupts disabled for crtc %d\n", crtc);
}

static int pdp_ioctl_get_sig(struct drm_device *dev, void *data,
			     struct drm_file *file)
{
	struct pdp_private *priv = dev->dev_private;
	u32 r, gb, sig;

	r = pdp_read(priv->dev, PDP_SIGNAT_R_OFFSET);
	gb = pdp_read(priv->dev, PDP_SIGNAT_GB_OFFSET);

	DRM_DEBUG_DRIVER("ioctl red: %x gb: %x\n", r, gb);

	sig = GET_FIELD(r, PDP_SIGNAT_R_SIGNATURE_R) << 20
		| GET_FIELD(gb, PDP_SIGNAT_GB_SIGNATURE_G) << 10
		| GET_FIELD(gb, PDP_SIGNAT_GB_SIGNATURE_B);
	*(u32 *)data = sig;

	return 0;
}

static int pdp_kick_simulator_ioctl(struct drm_device *dev, void *data,
				    struct drm_file *file)
{
#ifdef __PDP_TARGET_IS_SIMULATOR__
	struct pdp_private *priv = dev->dev_private;

	if (priv->auto_kick) {
		DRM_ERROR("Cannot kick simulator if auto-kick enabled.\n");
		return -EINVAL;
	}

	pdp_kick_simulator(priv);

#endif
	return 0;
}

static int pdp_auto_kick_ioctl(struct drm_device *dev, void *data,
			       struct drm_file *file)
{
#ifdef __PDP_TARGET_IS_SIMULATOR__
	struct pdp_private *priv = dev->dev_private;

	priv->auto_kick = !!(*(u32 *)data);

	DRM_DEBUG_DRIVER("%s pdp-drm auto kick simulator\n",
			 priv->auto_kick ? "Enabled" : "Disabled");

#endif
	return 0;
}

static int pdp_set_bkgnd_ioctl(struct drm_device *dev, void *data,
			       struct drm_file *file)
{
	struct pdp_private *priv = dev->dev_private;
	struct pdp_crtc *pcrtc = priv->crtc;
	struct drm_pdp_bkgnd *background = data;

	pdp_crtc_set_background(pcrtc, background->alpha, background->red,
				background->green, background->blue);

	return 0;
}

static const struct drm_ioctl_desc pdp_ioctls[] = {
	DRM_IOCTL_DEF_DRV(PDP_GET_SIGNATURE, pdp_ioctl_get_sig, DRM_UNLOCKED),
	DRM_IOCTL_DEF_DRV(PDP_GEM_CREATE, pdp_gem_create_ioctl, DRM_AUTH | DRM_UNLOCKED),
	DRM_IOCTL_DEF_DRV(PDP_GEM_MMAP, pdp_gem_mmap_ioctl, DRM_AUTH | DRM_UNLOCKED),
	DRM_IOCTL_DEF_DRV(PDP_KICK_SIMULATOR, pdp_kick_simulator_ioctl, DRM_UNLOCKED),
	DRM_IOCTL_DEF_DRV(PDP_AUTO_KICK, pdp_auto_kick_ioctl, DRM_UNLOCKED),
	DRM_IOCTL_DEF_DRV(PDP_SET_BKGND, pdp_set_bkgnd_ioctl, DRM_UNLOCKED),
};

static const struct file_operations pdp_driver_fops = {
	.owner		= THIS_MODULE,
	.open		= drm_open,
	.release	= drm_release,
	.unlocked_ioctl	= drm_ioctl,
	.mmap		= drm_gem_mmap,
	.poll		= drm_poll,
	.read		= drm_read,
	.llseek		= noop_llseek,
#ifdef CONFIG_COMPAT
	.compat_ioctl	= drm_compat_ioctl,
#endif
};

static struct drm_driver pdp_drm_driver = {
	.driver_features	= DRIVER_GEM |
				  DRIVER_MODESET |
				  DRIVER_PRIME | DRIVER_ATOMIC,
	/* Generics */
	.load			= pdp_load,
	.unload			= pdp_unload,
	.preclose		= pdp_preclose,
	.lastclose		= pdp_lastclose,
	.fops			= &pdp_driver_fops,
	.name			= DRIVER_NAME,
	.desc			= DRIVER_DESC,
	.date			= DRIVER_DATE,
	.major			= DRIVER_MAJOR,
	.minor			= DRIVER_MINOR,
	.patchlevel		= 0,

	/* VBlank ops */
	.get_vblank_counter	= drm_vblank_count,
	.enable_vblank		= pdp_enable_vblank,
	.disable_vblank		= pdp_disable_vblank,

	/* PRIME ops */
	.prime_handle_to_fd	= drm_gem_prime_handle_to_fd,
	.prime_fd_to_handle	= drm_gem_prime_fd_to_handle,
	.gem_prime_export	= pdp_gem_prime_export,
	.gem_prime_import	= pdp_gem_prime_import,

	/* GEM operations*/
	.gem_vm_ops		= &pdp_gem_vm_ops,
	.gem_free_object	= pdp_gem_free_object,

	/* IOCTLs */
	.ioctls			= pdp_ioctls,
	.num_ioctls		= ARRAY_SIZE(pdp_ioctls),

	/* FB ops*/
	.dumb_create		= pdp_gem_dumb_create,
	.dumb_map_offset	= pdp_gem_dumb_map_offset,
	.dumb_destroy		= pdp_gem_dumb_destroy,
};

static int compare_of(struct device *dev, void *data)
{
	return dev->of_node == data;
}

static int pdp_bind(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);
	struct drm_device *drm;
	int ret;

	drm = drm_dev_alloc(&pdp_drm_driver, dev);
	if (!drm)
		return -ENOMEM;

	drm->platformdev = pdev;

	ret = drm_dev_set_unique(drm, "%s", pdev->name);
	if (ret)
		goto err_free;

	ret = drm_dev_register(drm, 0);
	if (ret)
		goto err_free;

	return 0;

err_free:
	drm_dev_unref(drm);
	return ret;
}

static void pdp_unbind(struct device *dev)
{
	dev_set_drvdata(dev, NULL);
}

static const struct component_master_ops pdp_ops = {
	.bind = pdp_bind,
	.unbind = pdp_unbind,
};

static void pdp_add_endpoints(struct device *dev,
	struct component_match **match, struct device_node *port)
{
	struct device_node *ep, *remote;

	for_each_child_of_node(port, ep) {
		remote = of_graph_get_remote_port_parent(ep);
		if (!remote ) {
			of_node_put(remote);
			continue;
		} else if (!of_device_is_available(remote->parent)) {
			dev_warn(dev, "parent device of %s is not available\n",
				 remote->full_name);
			of_node_put(remote);
			continue;
		}

		component_match_add(dev, match, compare_of, remote);
		of_node_put(remote);
	}
}

#ifdef CONFIG_PM_SLEEP
static int pdp_pm_suspend(struct device *dev)
{
	struct pdp_private *priv = dev_get_drvdata(dev);
	struct drm_device *drm = priv->dev;

	drm_kms_helper_poll_disable(drm);
	priv->suspend_state = drm_atomic_helper_suspend(drm);
	if (IS_ERR(priv->suspend_state)) {
		drm_kms_helper_poll_enable(drm);
		return PTR_ERR(priv->suspend_state);
	}

	return 0;
}

static int pdp_pm_resume(struct device *dev)
{
	struct pdp_private *priv = dev_get_drvdata(dev);
	struct drm_device *drm = priv->dev;

	drm_atomic_helper_resume(drm, priv->suspend_state);
	drm_kms_helper_poll_enable(drm);

	return 0;
}
#endif

static SIMPLE_DEV_PM_OPS(pdp_pm_ops, pdp_pm_suspend, pdp_pm_resume);

static int pdp_probe(struct platform_device *pdev)
{
	struct device *dev = &pdev->dev;
	struct device_node *node = dev->of_node;
	struct component_match *match = NULL;

	if (!node)
		return -ENODEV;
	/* Ideally should create crtc here so will be found by encoder init
	 * Only have 1 crtc so can get away with this for now
	 */

	pdp_add_endpoints(dev, &match, node);
	of_node_put(node);

	return component_master_add_with_match(&pdev->dev, &pdp_ops, match);
}

static int pdp_remove(struct platform_device *pdev)
{
	component_master_del(&pdev->dev, &pdp_ops);
	return 0;
}

static const struct of_device_id pdp_of_table[] = {
	{ .compatible = "img,pdp-cashew", .data = 0 },
	{ }
};

MODULE_DEVICE_TABLE(of, pdp_of_table);

static struct platform_driver pdp_platform_driver = {
	.probe		= pdp_probe,
	.remove		= pdp_remove,
	.driver		= {
		.name	= "pdp-drm",
		.of_match_table = pdp_of_table,
		.pm = &pdp_pm_ops,
	},
};
module_platform_driver(pdp_platform_driver);

MODULE_AUTHOR("Imagination Technologies Ltd. <gpl-support@imgtec.com>");
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("Dual MIT/GPL");
