#include <drm/drmP.h>
#include <drm/drm_crtc.h>
#include <drm/drm_fb_helper.h>
#include <drm/drm_gem_cma_helper.h>
#include <drm/drm_fb_cma_helper.h>
#include <linux/img-iommu.h>

#include "pdp_fb.h"
#include "pdp_modeset.h"

void pdp_fill_fb_struct(struct drm_framebuffer *fb,
			struct drm_mode_fb_cmd2 *mode_cmd)
{
	int i;

	fb->width = mode_cmd->width;
	fb->height = mode_cmd->height;
	for (i = 0; i < 4; i++) {
		fb->pitches[i] = mode_cmd->pitches[i];
		fb->offsets[i] = mode_cmd->offsets[i];
		fb->modifier[i] = mode_cmd->modifier[i];
	}	/* below, use driver bpp's, kernel helpers dont have some*/
	fb->bits_per_pixel = pdp_format_info(mode_cmd->pixel_format)->bpp;
	fb->depth = 0;

	fb->pixel_format = mode_cmd->pixel_format;
	fb->flags = mode_cmd->flags;
}

static void pdp_fb_destroy(struct drm_framebuffer *fb)
{
	struct pdp_fb_cma *fb_cma = drm_fb_to_pdp_fb(fb);
	int i;

	drm_framebuffer_cleanup(fb);
	for (i = 0; i < 4; i++) {
		if (fb_cma->obj[i])
			drm_gem_object_unreference_unlocked(
					&fb_cma->obj[i]->obj);
	}

	kfree(fb_cma);
}

static int
pdp_fb_cma_create_handle(struct drm_framebuffer *fb,
			 struct drm_file *file_priv,
			 unsigned int *handle)
{
	struct pdp_fb_cma *fb_cma = drm_fb_to_pdp_fb(fb);

	return drm_gem_handle_create(file_priv,
		&fb_cma->obj[0]->obj, handle);
}

static struct drm_framebuffer_funcs pdp_fb_cma_funcs = {
	.destroy = pdp_fb_destroy,
	.create_handle = pdp_fb_cma_create_handle,
};

static struct
pdp_fb_cma *pdp_fb_cma_alloc(struct drm_device *dev,
			     struct drm_mode_fb_cmd2 *mode_cmd,
			     struct pdp_gem_object **obj,
			     unsigned int num_planes)
{
	struct pdp_fb_cma *fb_cma;
	int ret, i;

	fb_cma = kzalloc(sizeof(*fb_cma), GFP_KERNEL);
	if (!fb_cma)
		return ERR_PTR(-ENOMEM);

	pdp_fill_fb_struct(&fb_cma->fb, mode_cmd);

	for (i = 0; i < num_planes; i++)
		fb_cma->obj[i] = obj[i];

	ret = drm_framebuffer_init(dev, &fb_cma->fb, &pdp_fb_cma_funcs);
	if (ret) {
		dev_err(dev->dev, "Failed to initialize fb: %d\n", ret);
		kfree(fb_cma);
		return ERR_PTR(ret);
	}

	return fb_cma;
}

struct drm_framebuffer *
pdp_fb_create(struct drm_device *dev, struct drm_file *file_priv,
	      struct drm_mode_fb_cmd2 *mode_cmd)
{
	struct pdp_private *priv = dev->dev_private;
	struct pdp_fb_cma *fb_cma;
	struct pdp_gem_object *objs[4];
	struct drm_gem_object *obj;
	unsigned int hsub, vsub;
	int ret;
	int i;

	if (!pdp_format_info(mode_cmd->pixel_format)) {
		DRM_DEBUG_DRIVER("Invalid pixel format %s\n",
				 drm_get_format_name(mode_cmd->pixel_format));
		return ERR_PTR(-EINVAL);
	}

	hsub = drm_format_horz_chroma_subsampling(mode_cmd->pixel_format);
	vsub = drm_format_vert_chroma_subsampling(mode_cmd->pixel_format);

	for (i = 0; i < drm_format_num_planes(mode_cmd->pixel_format); i++) {
		unsigned int width = mode_cmd->width / (i ? hsub : 1);
		unsigned int height = mode_cmd->height / (i ? hsub : 1);
		unsigned int min_size;

		obj = drm_gem_object_lookup(dev, file_priv,
					    mode_cmd->handles[i]);
		if (!obj) {
			dev_err(dev->dev, "Failed to lookup GEM object\n");
			ret = -ENXIO;
			goto err_gem_object_unreference;
		}

		objs[i] = to_pdp_gem_object(obj);

		if (objs[i]->mm) {
			DRM_DEBUG_DRIVER("handle %u already mapped\n",
					 mode_cmd->handles[i]);
			continue;
		}

		if (pdp_format_info(mode_cmd->pixel_format)->bpp == 10) {
			min_size = (height - 1) * mode_cmd->pitches[i]
				 + width * 4 / 3
				 + mode_cmd->offsets[i];
		} else {
			min_size = (height - 1) * mode_cmd->pitches[i]
				 + width *
			 drm_format_plane_cpp(mode_cmd->pixel_format, i)
				 + mode_cmd->offsets[i];
		}

		if (obj->size < min_size) {
			drm_gem_object_unreference_unlocked(obj);
			ret = -EINVAL;
			goto err_gem_object_unreference;
		}

		switch (mode_cmd->modifier[i]) {
		case IMG_FORMAT_MOD_256x16_TILED:
			objs[i]->tiled = true;
			objs[i]->tile_stride = mode_cmd->pitches[i];
			pdp_tiler_set_tile_mode(priv, TILE_256x16);
			break;
		case IMG_FORMAT_MOD_512x8_TILED:
			objs[i]->tiled = true;
			objs[i]->tile_stride = mode_cmd->pitches[i];
			pdp_tiler_set_tile_mode(priv, TILE_512x8);
			break;
		case DRM_FORMAT_MOD_NONE:
			break;
		default:
			DRM_DEBUG_DRIVER("Unsupported fb modifier %llx\n",
					 mode_cmd->modifier[i]);
		}
		pdp_tiler_iommu_map(priv, objs[i]);
	}
	fb_cma = pdp_fb_cma_alloc(dev, mode_cmd, objs, i);
	if (IS_ERR(fb_cma)) {
		ret = PTR_ERR(fb_cma);
		goto err_gem_object_unreference;
	}

	return &fb_cma->fb;

err_gem_object_unreference:
	for (i--; i >= 0; i--) {
		pdp_tiler_iommu_unmap(priv, objs[i]);
		drm_gem_object_unreference_unlocked(&objs[i]->obj);
	}
	return ERR_PTR(ret);
}
