#include <linux/dma-buf.h>
#include <linux/dma-mapping.h>
#include <linux/sizes.h>
#include <linux/gfp.h>
#include <linux/pagemap.h>

#include "pdp_drv.h"
#include "pdp_gem.h"

#include <drm/pdp_drm.h>

static int pdp_gem_vm_fault(struct vm_area_struct *vma, struct vm_fault *vmf)
{
	struct pdp_gem_object *pobj = to_pdp_gem_object(vma->vm_private_data);
	pgoff_t pgoff;
	int err;
	unsigned long pfn;

	pgoff = ((unsigned long)vmf->virtual_address - vma->vm_start);
	pgoff >>= PAGE_SHIFT;

	if (!pobj->pages)
		pfn = (pobj->paddr >> PAGE_SHIFT) + pgoff;
	else
		pfn = page_to_pfn(pobj->pages[pgoff]);

	err = vm_insert_pfn(vma, (unsigned long)vmf->virtual_address, pfn);

	switch (err) {
	case -EAGAIN:
	case -0:
	case -ERESTARTSYS:
	case -EINTR:
	case -EBUSY:
		return VM_FAULT_NOPAGE;
	case -ENOMEM:
		return VM_FAULT_OOM;
	default:
		return VM_FAULT_SIGBUS;
	}
}

const struct vm_operations_struct pdp_gem_vm_ops = {
	.fault = pdp_gem_vm_fault,
	.open = drm_gem_vm_open,
	.close = drm_gem_vm_close,
};

static void pdp_gem_free(struct drm_device *dev, struct pdp_gem_object *pobj)
{
	if (pobj->pages) {
		dma_unmap_sg(dev->dev, pobj->sgt->sgl, pobj->sgt->nents,
			     DMA_TO_DEVICE);
		drm_gem_put_pages(&pobj->obj, pobj->pages, true, true);
		sg_free_table(pobj->sgt);
		kfree(pobj->sgt);
	}
}

void pdp_gem_free_object(struct drm_gem_object *obj)
{
	struct pdp_private *priv = obj->dev->dev_private;
	struct pdp_gem_object *pobj = to_pdp_gem_object(obj);

	if (pobj->size) {
		pdp_tiler_iommu_unmap(priv, pobj);
		drm_mm_remove_node(pobj->mm);
		kfree(pobj->mm);
	}

	if (obj->import_attach) {
		dma_buf_unmap_attachment(obj->import_attach, pobj->sgt, 0);
		drm_prime_gem_destroy(obj, NULL);
	} else {
		pdp_gem_free(obj->dev, pobj);
	}

	drm_gem_object_release(obj);
	kfree(pobj);
}

int pdp_gem_get_pages(struct drm_device *dev, struct pdp_gem_object *pobj)
{
	int npages = pobj->obj.size >> PAGE_SHIFT;

	pobj->pages = drm_gem_get_pages(&pobj->obj);
	if (IS_ERR(pobj->pages))
		return PTR_ERR(pobj->pages);

	pobj->sgt = drm_prime_pages_to_sg(pobj->pages, npages);
	if (IS_ERR(pobj->sgt))
		goto put_pages;

	dma_map_sg(dev->dev, pobj->sgt->sgl, pobj->sgt->nents,
		   DMA_TO_DEVICE);

	return 0;

put_pages:
	drm_gem_put_pages(&pobj->obj, pobj->pages, false, false);
	return PTR_ERR(pobj->sgt);
}

static struct pdp_gem_object *
pdp_gem_alloc_object(struct drm_device *dev, size_t size)
{
	struct pdp_gem_object *pobj;
	int err;

	pobj = kzalloc(sizeof(*pobj), GFP_KERNEL);
	if (!pobj)
		return ERR_PTR(-ENOMEM);

	size = roundup(size, PAGE_SIZE);
	size = PAGE_ALIGN(size);

	err = drm_gem_object_init(dev, &pobj->obj, size);
	if (err < 0)
		goto free;

	err = drm_gem_create_mmap_offset(&pobj->obj);
	if (err < 0)
		goto release;

	reservation_object_init(&pobj->resv);

	DRM_DEBUG_DRIVER("alloc private obj %p size %zu\n", pobj, size);

	return pobj;
release:
	drm_gem_object_release(&pobj->obj);
free:
	kfree(pobj);
	return ERR_PTR(err);
}

static int pdp_gem_alloc(struct drm_device *dev, struct pdp_gem_object *pobj)
{
	struct pdp_private *priv = dev->dev_private;
	int err;

	if (!priv->carveout) {
		err = pdp_gem_get_pages(dev, pobj);
		if (err < 0)
			return err;
	}

	return 0;
}

struct pdp_gem_object *pdp_gem_create(struct drm_device *dev, size_t size,
				      bool tiled, unsigned int pitch)
{
	struct pdp_gem_object *pobj;
	int err;

	pobj = pdp_gem_alloc_object(dev, size);
	if (IS_ERR(pobj))
		return pobj;

	pobj->tiled = tiled;
	if (tiled)
		pobj->tile_stride = pitch;

	err = pdp_gem_alloc(dev, pobj);
	if (err < 0)
		goto release;

	return pobj;
release:
	drm_gem_object_release(&pobj->obj);
	kfree(pobj);
	return ERR_PTR(err);
}

int pdp_gem_dumb_create(struct drm_file *file, struct drm_device *dev,
			struct drm_mode_create_dumb *args)
{
	struct pdp_gem_object *pobj;
	u32 pitch;
	size_t size;
	unsigned int align = 64;
	unsigned int min_pitch;
	int tiled = false;
	int ret;

	DRM_DEBUG_DRIVER("dumb create req: width:%u height:%u flags:%u\n",
			 args->width, args->height, args->flags);

	if (args->bpp == 10)
		min_pitch = DIV_ROUND_UP(args->width, 3) * 4;
	else
		min_pitch = DIV_ROUND_UP(args->width * args->bpp, 8);

	if (args->width != 1) {
		/* If we are using a tiled format then we need a stride
		 * that is equal to a power of two
		 */
		if (tiled)
			pitch = BIT(fls(min_pitch));
		else
			pitch = roundup(min_pitch, align);
	} else {
		pitch = min_pitch;
	}

	size = pitch * args->height;

	pobj = pdp_gem_create(dev, size, tiled, pitch);
	if (IS_ERR(pobj))
		return PTR_ERR(pobj);

	ret = drm_gem_handle_create(file, &pobj->obj, &args->handle);
	if (ret) {
		pdp_gem_free_object(&pobj->obj);
		return ret;
	}

	drm_gem_object_unreference_unlocked(&pobj->obj);

	args->pitch = pitch;
	args->size = size;
	DRM_DEBUG_DRIVER("obj %p size %zu handle %x\n",
			 pobj, size, args->handle);
	return 0;
}

int pdp_gem_dumb_map_offset(struct drm_file *file, struct drm_device *dev,
			    uint32_t handle, uint64_t *offset)
{
	struct drm_gem_object *obj;
	struct pdp_gem_object *pobj;

	mutex_lock(&dev->struct_mutex);
	obj = drm_gem_object_lookup(dev, file, handle);
	if (!obj) {
		dev_err(dev->dev, "failed to lookup GEM object\n");
		mutex_unlock(&dev->struct_mutex);
		return -EINVAL;
	}

	pobj = to_pdp_gem_object(obj);

	*offset = drm_vma_node_offset_addr(&pobj->obj.vma_node);

	drm_gem_object_unreference(obj);

	mutex_unlock(&dev->struct_mutex);

	return 0;
}

int pdp_gem_dumb_destroy(struct drm_file *file, struct drm_device *dev,
			 u32 handle)
{
	return drm_gem_handle_delete(file, handle);
}

int pdp_gem_create_ioctl(struct drm_device *dev, void *data,
			 struct drm_file *file)
{
	struct drm_pdp_gem_create *args = data;
	struct pdp_gem_object *pobj;
	int ret;

	if (!args->size)
		return -EINVAL;

	pobj = pdp_gem_create(dev, args->size, false, 0);
	if (IS_ERR(pobj))
		return PTR_ERR(pobj);

	ret = drm_gem_handle_create(file, &pobj->obj, &args->handle);
	if (ret) {
		pdp_gem_free_object(&pobj->obj);
		return ret;
	}

	DRM_DEBUG_DRIVER("obj %p size %zu handle %#x\n", pobj, pobj->obj.size,
			 args->handle);

	drm_gem_object_unreference_unlocked(&pobj->obj);

	return 0;
}

int pdp_gem_mmap_ioctl(struct drm_device *dev, void *data,
		       struct drm_file *file)
{
	struct drm_pdp_gem_mmap *args = data;

	return pdp_gem_dumb_map_offset(file, dev, args->handle, &args->offset);
}

struct sg_table *
pdp_gem_prime_map_dma_buf(struct dma_buf_attachment *attach,
			  enum dma_data_direction dir)
{
	struct pdp_gem_object *pobj = to_pdp_gem_object(attach->dmabuf->priv);
	struct sg_table *sgt;
	unsigned int npages = pobj->obj.size >> PAGE_SHIFT;

	if (pobj->pages) {
		sgt = drm_prime_pages_to_sg(pobj->pages, npages);
		if (IS_ERR(sgt))
			return NULL;

		if (dma_map_sg(attach->dev, sgt->sgl, sgt->nents, dir) == 0)
			goto free;

	} else {
		sgt = kmalloc(sizeof(*sgt), GFP_KERNEL);
		if (!sgt)
			return NULL;

		if (sg_alloc_table(sgt, 1, GFP_KERNEL))
			goto free;

		sg_dma_address(sgt->sgl) = pobj->paddr;
		sg_dma_len(sgt->sgl) = pobj->obj.size;
	}

	return sgt;

free:
	sg_free_table(sgt);
	kfree(sgt);
	return NULL;
}

static void
pdp_gem_prime_unmap_dma_buf(struct dma_buf_attachment *attach,
			    struct sg_table *sgt,
			    enum dma_data_direction dir)
{
	sg_free_table(sgt);
	kfree(sgt);
}

static int pdp_gem_dmabuf_mmap(struct dma_buf *buf, struct vm_area_struct *vma)
{
	struct drm_gem_object *obj = buf->priv;
	int err;

	mutex_lock(&obj->dev->struct_mutex);
	err = drm_gem_mmap_obj(obj, obj->size, vma);
	mutex_unlock(&obj->dev->struct_mutex);

	return err;
}

static void *pdp_gem_dmabuf_no_kmap(struct dma_buf *buf, unsigned long n)
{
	return NULL;
}

static const struct dma_buf_ops pdp_gem_prime_dmabuf_ops = {
	.map_dma_buf	= pdp_gem_prime_map_dma_buf,
	.unmap_dma_buf	= pdp_gem_prime_unmap_dma_buf,
	.release	= drm_gem_dmabuf_release,
	.kmap_atomic	= pdp_gem_dmabuf_no_kmap,
	.kmap		= pdp_gem_dmabuf_no_kmap,
	.mmap		= pdp_gem_dmabuf_mmap,
};

static struct pdp_gem_object *pdp_gem_import(struct drm_device *dev,
					     struct dma_buf *buf)
{
	struct pdp_private *priv = dev->dev_private;
	struct dma_buf_attachment *attach;
	struct pdp_gem_object *pobj;
	int err;

	pobj = pdp_gem_alloc_object(dev, buf->size);
	if (IS_ERR(pobj))
		return pobj;

	attach = dma_buf_attach(buf, dev->dev);
	if (IS_ERR(attach)) {
		err = PTR_ERR(attach);
		goto free;
	}

	get_dma_buf(buf);

	pobj->sgt = dma_buf_map_attachment(attach, DMA_TO_DEVICE);
	if (!pobj->sgt) {
		err = -ENOMEM;
		goto detach;
	}

	if (IS_ERR(pobj->sgt)) {
		err = PTR_ERR(pobj->sgt);
		goto detach;
	}

	err = pdp_tiler_iommu_map(priv, pobj);
	if (err < 0)
		goto detach;

	pobj->obj.import_attach = attach;

	return pobj;

detach:
	if (!IS_ERR_OR_NULL(pobj->sgt))
		dma_buf_unmap_attachment(attach, pobj->sgt, DMA_TO_DEVICE);

	dma_buf_detach(buf, attach);
	dma_buf_put(buf);
free:
	drm_gem_object_release(&pobj->obj);
	kfree(pobj);
	return ERR_PTR(err);
}

struct dma_buf *
pdp_gem_prime_export(struct drm_device *dev, struct drm_gem_object *obj,
		     int flags)
{
	struct pdp_gem_object *pobj = to_pdp_gem_object(obj);

	DEFINE_DMA_BUF_EXPORT_INFO(exp_info);

	exp_info.ops = &pdp_gem_prime_dmabuf_ops;
	exp_info.size = obj->size;
	exp_info.flags = flags;
	exp_info.resv = &pobj->resv;
	exp_info.priv = obj;

	return dma_buf_export(&exp_info);
}

struct drm_gem_object *
pdp_gem_prime_import(struct drm_device *dev, struct dma_buf *buf)
{
	struct pdp_gem_object *pobj;

	if (buf->ops == &pdp_gem_prime_dmabuf_ops) {
		struct drm_gem_object *obj = buf->priv;

		if (obj->dev == dev) {
			drm_gem_object_reference(obj);
			return obj;
		}
	}

	pobj = pdp_gem_import(dev, buf);
	if (IS_ERR(pobj))
		return ERR_CAST(pobj);

	return &pobj->obj;
}
