#include <linux/clk.h>
#include <linux/component.h>
#include <linux/iopoll.h>
#include <linux/math64.h>
#include <linux/module.h>
#include <linux/of_device.h>
#include <linux/regmap.h>
#include <linux/mfd/syscon.h>
#include <drm/drm_atomic_helper.h>
#include <drm/drm_crtc.h>
#include <drm/drm_crtc_helper.h>
#include <drm/drm_of.h>
#include <drm/drm_panel.h>
#include <drm/drmP.h>

struct pdp_lvds {
	struct device *dev;
	struct platform_device *pdev;

	struct drm_encoder encoder;
	struct drm_connector connector;

	void __iomem *en;

	struct drm_panel *panel;
};

static inline struct pdp_lvds *connector_to_lvds(struct drm_connector *con)
{
	return container_of(con, struct pdp_lvds, connector);
}

static inline struct pdp_lvds *encoder_to_lvds(struct drm_encoder *encoder)
{
	return container_of(encoder, struct pdp_lvds, encoder);
}

static void pdp_lvds_encoder_mode_set(struct drm_encoder *encoder,
				      struct drm_display_mode *mode,
				      struct drm_display_mode *adjusted_mode)
{
}

static bool pdp_lvds_encoder_mode_fixup(struct drm_encoder *encoder,
					const struct drm_display_mode *mode,
					struct drm_display_mode *adjusted_mode)
{
	return true;
}

static void pdp_lvds_encoder_enable(struct drm_encoder *encoder)
{
	struct pdp_lvds *lvds = encoder_to_lvds(encoder);

	writel(1, lvds->en);

	if (lvds->panel) {
		drm_panel_prepare(lvds->panel);
		drm_panel_enable(lvds->panel);
	}
}

static void pdp_lvds_encoder_disable(struct drm_encoder *encoder)
{
	struct pdp_lvds *lvds = encoder_to_lvds(encoder);

	if (lvds->panel) {
		drm_panel_disable(lvds->panel);
		drm_panel_unprepare(lvds->panel);
	}

	writel(0, lvds->en);
}

static struct drm_encoder_helper_funcs
pdp_lvds_encoder_helper_funcs = {
	.mode_fixup = pdp_lvds_encoder_mode_fixup,
	.mode_set = pdp_lvds_encoder_mode_set,
	.enable = pdp_lvds_encoder_enable,
	.disable = pdp_lvds_encoder_disable,
};

static struct drm_encoder_funcs pdp_lvds_encoder_funcs = {
	.destroy = drm_encoder_cleanup,
};

static int pdp_lvds_connector_get_modes(struct drm_connector *connector)
{
	struct pdp_lvds *lvds = connector_to_lvds(connector);
	int num_modes = 0;

	WARN_ON(!lvds->panel);

	if (lvds->panel->funcs && lvds->panel->funcs->get_modes)
		num_modes = lvds->panel->funcs->get_modes(lvds->panel);

	return num_modes;
}

static struct drm_encoder *
pdp_lvds_connector_best_encoder(struct drm_connector *connector)
{
	struct pdp_lvds *lvds = connector_to_lvds(connector);

	return &lvds->encoder;
}

static struct drm_connector_helper_funcs pdp_lvds_connector_helper_funcs = {
	.get_modes = pdp_lvds_connector_get_modes,
	.best_encoder = pdp_lvds_connector_best_encoder,
};

static enum drm_connector_status
pdp_lvds_detect(struct drm_connector *connector, bool force)
{
	return connector_status_connected;
}

static void pdp_lvds_drm_connector_destroy(struct drm_connector *connector)
{
	drm_connector_unregister(connector);
	drm_connector_cleanup(connector);
}

static struct drm_connector_funcs pdp_lvds_atomic_connector_funcs = {
	.dpms = drm_atomic_helper_connector_dpms,
	.fill_modes = drm_helper_probe_single_connector_modes,
	.detect = pdp_lvds_detect,
	.destroy = pdp_lvds_drm_connector_destroy,
	.reset = drm_atomic_helper_connector_reset,
	.atomic_duplicate_state = drm_atomic_helper_connector_duplicate_state,
	.atomic_destroy_state = drm_atomic_helper_connector_destroy_state,
};

static int pdp_lvds_register(struct drm_device *drm, struct pdp_lvds *lvds)
{
	struct drm_encoder *encoder = &lvds->encoder;
	struct drm_connector *connector = &lvds->connector;
	struct device *dev = lvds->dev;
	int ret;

	encoder->possible_crtcs = 1;
	drm_encoder_helper_add(&lvds->encoder,
			       &pdp_lvds_encoder_helper_funcs);
	ret = drm_encoder_init(drm, &lvds->encoder, &pdp_lvds_encoder_funcs,
			       DRM_MODE_ENCODER_TMDS);
	if (ret) {
		dev_err(dev, "Failed to initialize encoder with drm\n");
		return ret;
	}

	drm_connector_helper_add(connector,
				 &pdp_lvds_connector_helper_funcs);

	drm_connector_init(drm, &lvds->connector,
			   &pdp_lvds_atomic_connector_funcs,
			   DRM_MODE_CONNECTOR_HDMIA);

	drm_panel_attach(lvds->panel, &lvds->connector);

	ret = drm_connector_register(connector);
	if (ret)
		return ret;

	drm_mode_connector_attach_encoder(connector, encoder);

	return 0;
}

static int pdp_lvds_bind(struct device *dev, struct device *master,
			 void *data)
{
	struct platform_device *pdev = to_platform_device(dev);
	struct pdp_lvds *lvds;
	struct device_node *np = dev->of_node;
	struct device_node *panel;
	struct resource *res;
	int ret;

	lvds = devm_kzalloc(dev, sizeof(*lvds), GFP_KERNEL);
	if (!lvds)
		return -ENOMEM;

	lvds->dev = dev;

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!res) {
		dev_err(dev, "Failed to get enable register resource.\n");
		return -ENODEV;
	}
	lvds->en = devm_ioremap_resource(dev, res);
	if (IS_ERR(lvds->en))
		return PTR_ERR(lvds->en);

	panel = of_parse_phandle(np, "panel", 0);
	if (panel) {
		lvds->panel = of_drm_find_panel(panel);
		if (!lvds->panel)
			return -EPROBE_DEFER;
		of_node_put(panel);
	} else {
		return -ENOENT;
	}
	ret = pdp_lvds_register(data, lvds);
	if (ret) {
		dev_err(dev, "Failed to register lvds: %d\n", ret);
		return ret;
	}

	dev_set_drvdata(dev, lvds);

	return ret;
}

static void pdp_lvds_unbind(struct device *dev, struct device *master,
			    void *data)
{
}

static const struct component_ops pdp_lvds_ops = {
	.bind	= pdp_lvds_bind,
	.unbind	= pdp_lvds_unbind,
};

static int pdp_lvds_probe(struct platform_device *pdev)
{
	return component_add(&pdev->dev, &pdp_lvds_ops);
}

static int pdp_lvds_remove(struct platform_device *pdev)
{
	component_del(&pdev->dev, &pdp_lvds_ops);
	return 0;
}

static const struct of_device_id pdp_lvds_dt_ids[] = {
	{.compatible = "img,cashew-lvds",},
	{},
};

struct platform_driver pdp_lvds_driver = {
	.probe		= pdp_lvds_probe,
	.remove		= pdp_lvds_remove,
	.driver		= {
		.of_match_table = pdp_lvds_dt_ids,
		.name		= "cashew-lvds",
	},
};
module_platform_driver(pdp_lvds_driver);

MODULE_DESCRIPTION("Cashew LVDS driver");
MODULE_AUTHOR("Will Thomas <will.thomas@imgtec.com>");
MODULE_LICENSE("GPL");
MODULE_ALIAS("platform:cashew-lvds");
