#include <linux/wait.h>
#include <linux/reservation.h>

#include <drm/drmP.h>
#include <drm/drm_crtc.h>
#include <drm/drm_crtc_helper.h>
#include <drm/drm_atomic_helper.h>
#include <drm/drm_fb_cma_helper.h>
#include <drm/drm_gem_cma_helper.h>
#include <drm/drm_gem.h>
#include <drm/drm_atomic.h>

#include <drm/drm_fourcc.h>

#include "pdp_crtc.h"
#include "pdp_drv.h"
#include "pdp_modeset.h"
#include "pdp_plane_funcs.h"
#include "pdpfb2_regs.h"
#include "pdp_fb.h"

/* bpp field represents total bpp for single plane formats or
 * bpp of Y plane for multi planar formats
 */
static const struct pdp_format_info pdp_format_infos[] = {
	{
		.fourcc = DRM_FORMAT_RGB332,
		.bpp = 8,
		.planes = 1,
		.hwv = PDP_SURF_PIXFMT_RGB332,
	},
	{
		.fourcc = DRM_FORMAT_ARGB4444,
		.hwv = PDP_SURF_PIXFMT_ARGB4444,
		.bpp = 16,
		.planes = 1,
	},
	{
		.fourcc = DRM_FORMAT_ARGB1555,
		.hwv = PDP_SURF_PIXFMT_ARGB1555,
		.bpp = 16,
		.planes = 1,
	},
	{
		.fourcc = DRM_FORMAT_RGB888,
		.hwv = PDP_SURF_PIXFMT_RGB888,
		.bpp = 24,
		.planes = 1,
	},
	{
		.fourcc = DRM_FORMAT_RGB565,
		.hwv = PDP_SURF_PIXFMT_RGB565,
		.bpp = 16,
		.planes = 1,
	},
	{
		.fourcc = DRM_FORMAT_ARGB8888,
		.hwv = PDP_SURF_PIXFMT_ARGB8888,
		.bpp = 32,
		.planes = 1,
	},
	{
		.fourcc = DRM_FORMAT_AYUV,
		.hwv = PDP_SURF_PIXFMT_AYUV8888,
		.bpp = 32,
		.planes = 1,
	},
	{
		.fourcc = DRM_FORMAT_YUV420,
		.hwv = 0x9,
		.bpp = 8,
		.planes = 3,
	},
	{
		.fourcc = DRM_FORMAT_NV12,
		.hwv = 0xa,
		.bpp = 8,
		.planes = 2,
	},
	{
		.fourcc = DRM_FORMAT_NV21,
		.hwv = 0xb,
		.bpp = 8,
		.planes = 2,
	},
	{
		.fourcc = DRM_FORMAT_UYVY,
		.hwv = 0xc,
		.bpp = 32,
		.planes = 1,
	},
	{
		.fourcc = DRM_FORMAT_VYUY,
		.hwv = 0xd,
		.bpp = 32,
		.planes = 1,
	},
	{
		.fourcc = DRM_FORMAT_YUYV,
		.hwv = 0xe,
		.bpp = 32,
		.planes = 1,
	},
	{
		.fourcc = DRM_FORMAT_YVYU,
		.hwv = 0xf,
		.bpp = 32,
		.planes = 1,
	},
	{
		.fourcc = DRM_FORMAT_XYUV101010,
		.hwv = 0x15,
		.bpp = 32,
		.planes = 1,
	},
	{
		.fourcc = DRM_FORMAT_XRGB8888,
		.hwv = PDP_SURF_PIXFMT_ARGB8888,
		.bpp = 32,
		.planes = 1,
	},
	{
		.fourcc = DRM_FORMAT_XRGB2101010,
		.hwv = PDP_SURF_PIXFMT_RGB101010,
		.bpp = 32,
		.planes = 1,
	},
	{
		.fourcc = DRM_FORMAT_NV15,
	       .hwv = 0x18,
	       .bpp = 10,
	       .planes = 2,
	},
	{
		.fourcc = DRM_FORMAT_NV51,
	       .hwv = 0x19,
	       .bpp = 10,
	       .planes = 2,
	},
	{
		.fourcc = DRM_FORMAT_NV20,
	       .hwv = 0x1a,
	       .bpp = 10,
	       .planes = 2,
	},
	{
		.fourcc = DRM_FORMAT_NV02,
	       .hwv = 0x1b,
	       .bpp = 10,
	       .planes = 2,
	},
};

const struct pdp_format_info *pdp_format_info(u32 fourcc)
{
	unsigned int i;

	for (i = 0; i < ARRAY_SIZE(pdp_format_infos); ++i) {
		if (pdp_format_infos[i].fourcc == fourcc)
			return &pdp_format_infos[i];
	}
	return NULL;
}

/* --------------------------------------------------------------------------
 *	Atomic
 */

struct pdp_commit {
	struct work_struct work;
	struct drm_device *dev;
	struct drm_atomic_state *state;

	atomic_t waiting_fences;
	struct fence_cb cb;
};

static void pdp_atomic_complete(struct pdp_commit *commit)
{
	struct drm_device *dev =  commit->dev;
	struct pdp_private *priv = dev->dev_private;
	struct drm_atomic_state *old_state = commit->state;

	drm_atomic_helper_commit_planes(dev, old_state);
	drm_atomic_helper_commit_modeset_disables(dev, old_state);
	drm_atomic_helper_commit_modeset_enables(dev, old_state);

	drm_atomic_helper_cleanup_planes(dev, old_state);

	drm_atomic_state_free(old_state);

	spin_lock(&priv->commit.wait.lock);
	priv->commit.pending &= ~1;
	wake_up_all_locked(&priv->commit.wait);
	spin_unlock(&priv->commit.wait.lock);

	if (priv->auto_kick)
		pdp_kick_simulator(priv);

	kfree(commit);
}

static void pdp_atomic_work(struct work_struct *work)
{
	struct pdp_commit *commit = container_of(work, struct pdp_commit, work);

	DRM_DEBUG("work queue - completing commit\n");
	pdp_atomic_complete(commit);
}

static void pdp_fence_callback(struct fence *fence, struct fence_cb *cb)
{
	struct pdp_commit *commit = container_of(cb, struct pdp_commit, cb);

	if (atomic_dec_and_test(&commit->waiting_fences)) {
		DRM_DEBUG("All fences signaled, scheduling commit work\n");
		schedule_work(&commit->work);
	}
}

static void pdp_add_fence_callbacks(struct drm_atomic_state *state,
				    struct pdp_commit *commit)
{
	struct drm_plane *plane;
	struct drm_plane_state *plane_state;
	int i;
	int ret;

	atomic_inc(&commit->waiting_fences);

	for_each_plane_in_state(state, plane, plane_state, i) {
		struct drm_framebuffer *fb = plane->state->fb;
		struct pdp_gem_object *pobj;
		struct drm_gem_object *obj;
		struct reservation_object *resv;
		struct fence *fence;

		if (!fb)
			continue;

		pobj = drm_fb_obj(fb, 0);
		if (!pobj)
			continue;
		obj = &pobj->obj;
		if (!obj || !obj->dma_buf || !obj->dma_buf->resv)
			continue;

		resv = obj->dma_buf->resv;
		fence = fence_get(reservation_object_get_excl(resv));
		if (fence) {
			ret = fence_add_callback(fence, &commit->cb,
						 pdp_fence_callback);
			fence_put(fence);
			if (ret == -ENOENT) {
				DRM_INFO("Fence already signaled\n");
				continue;
			}
			atomic_inc(&commit->waiting_fences);
		}
	}

	/* If no planes had fences then send commit here */
	if (atomic_dec_and_test(&commit->waiting_fences)) {
		DRM_DEBUG("No wait fences, scheduling commit immediately\n");
		schedule_work(&commit->work);
	} else {
		DRM_DEBUG("Waiting for unsignaled fence...");
	}
}

static int pdp_atomic_commit(struct drm_device *dev,
			     struct drm_atomic_state *state, bool async)
{
	struct pdp_private *priv = dev->dev_private;
	struct pdp_commit *commit;
	int ret;

	ret = drm_atomic_helper_prepare_planes(dev, state);
	if (ret)
		return ret;

	commit = kzalloc(sizeof(*commit), GFP_KERNEL);
	if (!commit)
		return -ENOMEM;

	INIT_WORK(&commit->work, pdp_atomic_work);
	commit->dev = dev;
	commit->state = state;

	spin_lock(&priv->commit.wait.lock);
	ret = wait_event_interruptible_locked(priv->commit.wait,
					      !(priv->commit.pending));

	if (ret == 0)
		priv->commit.pending |= 1;
	spin_unlock(&priv->commit.wait.lock);

	if (ret) {
		kfree(commit);
		return ret;
	}

	drm_atomic_helper_swap_state(dev, state);

	if (async)
		pdp_add_fence_callbacks(state, commit);
	else
		pdp_atomic_complete(commit);

	return 0;
}

static void pdp_output_poll_changed(struct drm_device *dev)
{
	struct pdp_private *priv = dev->dev_private;

	drm_fbdev_cma_hotplug_event(priv->fbdev);
}

/* -----------------------------------------------------------------------
 *	Init
 */
static const struct drm_mode_config_funcs pdp_mode_config_funcs = {
	.fb_create = pdp_fb_create,
	.atomic_commit = pdp_atomic_commit,
	.atomic_check = drm_atomic_helper_check,
	.output_poll_changed = pdp_output_poll_changed,
};

void pdp_mode_config_init(struct drm_device *dev)
{
	dev->mode_config.min_width = 0;
	dev->mode_config.min_height = 0;

	dev->mode_config.max_width = 4096;
	dev->mode_config.max_height = 4096;

	dev->mode_config.fb_base = 0;
	dev->mode_config.allow_fb_modifiers = true;
	dev->mode_config.async_page_flip = false;
	dev->mode_config.funcs = &pdp_mode_config_funcs;
}

static int pdp_properties_init(struct pdp_private *priv)
{
	/* True range 0..1023, high bit enables Global plane alpha */
	priv->props.alpha = drm_property_create_range(priv->dev, 0, "alpha",
						      0, 4095);
	if (!priv->props.alpha)
		return -ENOMEM;

	priv->props.zpos = drm_property_create_range(priv->dev, 0, "zpos",
						     0, 7);
	if (!priv->props.zpos)
		return -ENOMEM;

	priv->props.porter_duff_mode = drm_property_create_enum(priv->dev, 0,
				"porter_duff_mode",
				pdp_porter_duff_mode_enum_list,
				ARRAY_SIZE(pdp_porter_duff_mode_enum_list));
	priv->props.porter_duff_type = drm_property_create_enum(priv->dev, 0,
				"porter_duff_type",
				pdp_porter_duff_type_enum_list,
				ARRAY_SIZE(pdp_porter_duff_type_enum_list));

	priv->props.colourkey_r_property = drm_property_create_range(priv->dev, 0,
				"colourkey_r", 0, 0x3FF);
	priv->props.colourkey_g_property = drm_property_create_range(priv->dev, 0,
				"colourkey_g", 0, 0x3FF);
	priv->props.colourkey_b_property = drm_property_create_range(priv->dev, 0,
				"colourkey_b", 0, 0x3FF);
	priv->props.colourkey_r_mask_property = drm_property_create_range(priv->dev, 0,
				"colourkey_r_mask", 0, 0x3FF);
	priv->props.colourkey_g_mask_property = drm_property_create_range(priv->dev, 0,
				"colourkey_g_mask", 0, 0x3FF);
	priv->props.colourkey_b_mask_property = drm_property_create_range(priv->dev, 0,
				"colourkey_b_mask", 0, 0x3FF);
	priv->props.colourkey_mode_property = drm_property_create_enum(priv->dev, 0,
				"colourkey_mode", pdp_colourkey_mode_enum_list,
				ARRAY_SIZE(pdp_colourkey_mode_enum_list));
	priv->props.colourkey_enable_property = drm_property_create_range(priv->dev, 0,
				"colourkey_enable", 0, 1);
	priv->props.gamma_enable_property = drm_property_create_range(priv->dev, 0,
				"gamma_enable", 0, 1);

	return 0;
}

int pdp_modeset_init(struct pdp_private *priv)
{
	struct drm_device *dev = priv->dev;
	int err;

	drm_mode_config_init(dev);
	pdp_mode_config_init(dev);

	err = pdp_properties_init(priv);
	if (err < 0)
		return err;

	err = pdp_crtc_create(dev, 0);
	if (err < 0) {
		DRM_ERROR("failed to create a CRTC\n");
		goto err_config_cleanup;
	}

	return 0;

err_config_cleanup:
	drm_mode_config_cleanup(dev);

	return err;
}

void pdp_modeset_cleanup(struct pdp_private *priv)
{
	drm_mode_config_cleanup(priv->dev);

	DRM_DEBUG_DRIVER("cleaned up\n");
}
