#include <drm/drmP.h>
#include <drm/drm_atomic_helper.h>
#include <drm/drm_crtc.h>
#include <drm/drm_crtc_helper.h>
#include <drm/drm_fb_cma_helper.h>
#include <drm/drm_gem_cma_helper.h>
#include <drm/drm_plane_helper.h>

#include "pdp_drv.h"
#include "pdp_modeset.h"
#include "pdp_plane.h"
#include "pdp_plane_funcs.h"
#include "pdpfb2_regs.h"

const u32 grph_formats[] = {
	DRM_FORMAT_RGB332,
	DRM_FORMAT_ARGB4444,
	DRM_FORMAT_ARGB1555,
	DRM_FORMAT_RGB888,
	DRM_FORMAT_RGB565,
	DRM_FORMAT_ARGB8888,
	DRM_FORMAT_AYUV,
	DRM_FORMAT_XYUV101010,
	DRM_FORMAT_XRGB8888,
};

const u32 vid_formats[] = {
	DRM_FORMAT_RGB888,
	DRM_FORMAT_ARGB8888,
	DRM_FORMAT_AYUV,
	DRM_FORMAT_YUV420,
	DRM_FORMAT_NV21,
	DRM_FORMAT_NV12,
	DRM_FORMAT_YUYV,
	DRM_FORMAT_YVYU,
	DRM_FORMAT_VYUY,
	DRM_FORMAT_UYVY,
	DRM_FORMAT_XYUV101010,
	DRM_FORMAT_XRGB8888,
	DRM_FORMAT_XRGB2101010,
	DRM_FORMAT_NV15,
	DRM_FORMAT_NV51,
	DRM_FORMAT_NV20,
	DRM_FORMAT_NV02,
};

static struct pdp_plane_conf _stream_template[] = {
{
	.format = PDP_FORMAT_GRPH,
	.type = DRM_PLANE_TYPE_PRIMARY,
	.caps = {
		.has_scaler = 0,
		.has_doubler = 1,
		.has_decimator = 1,
		.has_cosited_chroma_control = 0,
		.has_csc = 1,
	},
	.regs = {
		.surf = PDP_GRPH1SURF_OFFSET,
		.blend2_r = PDP_GRPH1BLND2_R_OFFSET,
		.blend2_gb = PDP_GRPH1BLND2_GB_OFFSET,
		.ctrl = PDP_GRPH1CTRL_OFFSET,
		.uctrl = PDPFB_NO_REG,
		.stride = PDP_GRPH1STRIDE_OFFSET,
		.size = PDP_GRPH1SIZE_OFFSET,
		.posn = PDP_GRPH1POSN_OFFSET,
		.galpha = PDP_GRPH1GALPHA_OFFSET,
		.ckey_r = PDP_GRPH1CKEY_R_OFFSET,
		.ckey_gb = PDP_GRPH1CKEY_GB_OFFSET,
		.interleave_ctrl = PDP_GRPH1INTERLEAVE_CTRL_OFFSET,
		.base_addr = PDP_GRPH1BASEADDR_OFFSET,
		.u_base_addr = PDPFB_NO_REG,
		.v_base_addr = PDPFB_NO_REG,
		.memctrl = PDP_GRPH1_MEMCTRL_OFFSET,
		.memthresh = PDP_GRPH1_MEM_THRESH_OFFSET,
		.skip_ctrl = PDPFB_NO_REG,
		.postskip_ctrl = PDPFB_NO_REG,
		.decimate_ctrl = PDP_GRPH1DECIMATE_CTRL_OFFSET,
		.scale_ctrl = PDPFB_NO_REG,
		.vsinit = PDPFB_NO_REG,
		.first_vscoeff = PDPFB_NO_REG,
		.last_vscoeff = PDPFB_NO_REG,
		.hsinit = PDPFB_NO_REG,
		.first_hscoeff = PDPFB_NO_REG,
		.last_hscoeff = PDPFB_NO_REG,
		.scale_size = PDPFB_NO_REG,
		.lumakey_alpha_trans_opaque = PDPFB_NO_REG,
		.lumakey_luma_max_min = PDPFB_NO_REG,
		.lumakey_c_rg = PDPFB_NO_REG,
		.lumakey_c_b = PDPFB_NO_REG,
	}
},
{
	.format = PDP_FORMAT_GRPH,
	.type = DRM_PLANE_TYPE_OVERLAY,
	.caps = {
		.has_scaler = 0,
		.has_doubler = 1,
		.has_decimator = 1,
		.has_cosited_chroma_control = 0,
		.has_csc = 1,
	},
	.regs = {
		.surf = PDP_GRPH2SURF_OFFSET,
		.blend2_r = PDP_GRPH2BLND2_R_OFFSET,
		.blend2_gb = PDP_GRPH2BLND2_GB_OFFSET,
		.ctrl = PDP_GRPH2CTRL_OFFSET,
		.uctrl = PDPFB_NO_REG,
		.stride = PDP_GRPH2STRIDE_OFFSET,
		.size = PDP_GRPH2SIZE_OFFSET,
		.posn = PDP_GRPH2POSN_OFFSET,
		.galpha = PDP_GRPH2GALPHA_OFFSET,
		.ckey_r = PDP_GRPH2CKEY_R_OFFSET,
		.ckey_gb = PDP_GRPH2CKEY_GB_OFFSET,
		.interleave_ctrl = PDP_GRPH2INTERLEAVE_CTRL_OFFSET,
		.base_addr = PDP_GRPH2BASEADDR_OFFSET,
		.u_base_addr = PDPFB_NO_REG,
		.v_base_addr = PDPFB_NO_REG,
		.memctrl = PDP_GRPH2_MEMCTRL_OFFSET,
		.memthresh = PDP_GRPH2_MEM_THRESH_OFFSET,
		.skip_ctrl = PDPFB_NO_REG,
		.postskip_ctrl = PDPFB_NO_REG,
		.decimate_ctrl = PDP_GRPH2DECIMATE_CTRL_OFFSET,
		.scale_ctrl = PDPFB_NO_REG,
		.vsinit = PDPFB_NO_REG,
		.first_vscoeff = PDPFB_NO_REG,
		.last_vscoeff = PDPFB_NO_REG,
		.hsinit = PDPFB_NO_REG,
		.first_hscoeff = PDPFB_NO_REG,
		.last_hscoeff = PDPFB_NO_REG,
		.scale_size = PDPFB_NO_REG,
		.lumakey_alpha_trans_opaque = PDPFB_NO_REG,
		.lumakey_luma_max_min = PDPFB_NO_REG,
		.lumakey_c_rg = PDPFB_NO_REG,
		.lumakey_c_b = PDPFB_NO_REG,
	}
},
{
	.format = PDP_FORMAT_GRPH,
	.type = DRM_PLANE_TYPE_OVERLAY,
	.caps = {
		.has_scaler = 0,
		.has_doubler = 1,
		.has_decimator = 1,
		.has_cosited_chroma_control = 0,
		.has_csc = 1,
	},
	.regs = {
		.surf = PDP_GRPH3SURF_OFFSET,
		.blend2_r = PDP_GRPH3BLND2_R_OFFSET,
		.blend2_gb = PDP_GRPH3BLND2_GB_OFFSET,
		.ctrl = PDP_GRPH3CTRL_OFFSET,
		.uctrl = PDPFB_NO_REG,
		.stride = PDP_GRPH3STRIDE_OFFSET,
		.size = PDP_GRPH3SIZE_OFFSET,
		.posn = PDP_GRPH3POSN_OFFSET,
		.galpha = PDP_GRPH3GALPHA_OFFSET,
		.ckey_r = PDP_GRPH3CKEY_R_OFFSET,
		.ckey_gb = PDP_GRPH3CKEY_GB_OFFSET,
		.interleave_ctrl = PDP_GRPH3INTERLEAVE_CTRL_OFFSET,
		.base_addr = PDP_GRPH3BASEADDR_OFFSET,
		.u_base_addr = PDPFB_NO_REG,
		.v_base_addr = PDPFB_NO_REG,
		.memctrl = PDP_GRPH3_MEMCTRL_OFFSET,
		.memthresh = PDP_GRPH3_MEM_THRESH_OFFSET,
		.skip_ctrl = PDPFB_NO_REG,
		.postskip_ctrl = PDPFB_NO_REG,
		.decimate_ctrl = PDP_GRPH3DECIMATE_CTRL_OFFSET,
		.scale_ctrl = PDPFB_NO_REG,
		.vsinit = PDPFB_NO_REG,
		.first_vscoeff = PDPFB_NO_REG,
		.last_vscoeff = PDPFB_NO_REG,
		.hsinit = PDPFB_NO_REG,
		.first_hscoeff = PDPFB_NO_REG,
		.last_hscoeff = PDPFB_NO_REG,
		.scale_size = PDPFB_NO_REG,
		.lumakey_alpha_trans_opaque = PDPFB_NO_REG,
		.lumakey_luma_max_min = PDPFB_NO_REG,
		.lumakey_c_rg = PDPFB_NO_REG,
		.lumakey_c_b = PDPFB_NO_REG,
	}
},
{
	.format = PDP_FORMAT_GRPH,
	.type = DRM_PLANE_TYPE_CURSOR,
	.caps = {
		.has_scaler = 0,
		.has_doubler = 1,
		.has_decimator = 1,
		.has_cosited_chroma_control = 0,
		.has_csc = 1,
	},
	.regs = {
		.surf = PDP_GRPH4SURF_OFFSET,
		.blend2_r = PDP_GRPH4BLND2_R_OFFSET,
		.blend2_gb = PDP_GRPH4BLND2_GB_OFFSET,
		.ctrl = PDP_GRPH4CTRL_OFFSET,
		.uctrl = PDPFB_NO_REG,
		.stride = PDP_GRPH4STRIDE_OFFSET,
		.size = PDP_GRPH4SIZE_OFFSET,
		.posn = PDP_GRPH4POSN_OFFSET,
		.galpha = PDP_GRPH4GALPHA_OFFSET,
		.ckey_r = PDP_GRPH4CKEY_R_OFFSET,
		.ckey_gb = PDP_GRPH4CKEY_GB_OFFSET,
		.interleave_ctrl = PDP_GRPH4INTERLEAVE_CTRL_OFFSET,
		.base_addr = PDP_GRPH4BASEADDR_OFFSET,
		.u_base_addr = PDPFB_NO_REG,
		.v_base_addr = PDPFB_NO_REG,
		.memctrl = PDP_GRPH4_MEMCTRL_OFFSET,
		.memthresh = PDP_GRPH4_MEM_THRESH_OFFSET,
		.skip_ctrl = PDPFB_NO_REG,
		.postskip_ctrl = PDPFB_NO_REG,
		.decimate_ctrl = PDP_GRPH4DECIMATE_CTRL_OFFSET,
		.scale_ctrl = PDPFB_NO_REG,
		.vsinit = PDPFB_NO_REG,
		.first_vscoeff = PDPFB_NO_REG,
		.last_vscoeff = PDPFB_NO_REG,
		.hsinit = PDPFB_NO_REG,
		.first_hscoeff = PDPFB_NO_REG,
		.last_hscoeff = PDPFB_NO_REG,
		.scale_size = PDPFB_NO_REG,
		.lumakey_alpha_trans_opaque = PDPFB_NO_REG,
		.lumakey_luma_max_min = PDPFB_NO_REG,
		.lumakey_c_rg = PDPFB_NO_REG,
		.lumakey_c_b = PDPFB_NO_REG,
	}
},
{
	.format = PDP_FORMAT_VIDEO,
	.type = DRM_PLANE_TYPE_OVERLAY,
	.caps = {
		.has_scaler = 1,
		.vs_taps = 4,
		.hs_taps = 8,
		.vs_interpolation_points = 16,
		.hs_interpolation_points = 16,
		.max_scaler_input_width = 2560,
		.has_doubler = 0,
		.has_decimator = 1,
		.has_cosited_chroma_control = 1,
		.has_csc = 1,
	},
	.regs = {
		.surf = PDP_VID1SURF_OFFSET,
		.blend2_r = PDP_VID1BLND2_R_OFFSET,
		.blend2_gb = PDP_VID1BLND2_GB_OFFSET,
		.ctrl = PDP_VID1CTRL_OFFSET,
		.uctrl = PDP_VID1UCTRL_OFFSET,
		.stride = PDP_VID1STRIDE_OFFSET,
		.size = PDP_VID1SIZE_OFFSET,
		.posn = PDP_VID1POSN_OFFSET,
		.galpha = PDP_VID1GALPHA_OFFSET,
		.ckey_r = PDP_VID1CKEY_R_OFFSET,
		.ckey_gb = PDP_VID1CKEY_GB_OFFSET,
		.interleave_ctrl = PDP_VID1INTERLEAVE_CTRL_OFFSET,
		.base_addr = PDP_VID1BASEADDR_OFFSET,
		.u_base_addr = PDP_VID1UBASEADDR_OFFSET,
		.v_base_addr = PDP_VID1VBASEADDR_OFFSET,
		.memctrl = PDP_VID1_MEMCTRL_OFFSET,
		.memthresh = PDP_VID1_MEM_THRESH_OFFSET,
		.skip_ctrl = PDP_VID1SKIPCTRL_OFFSET,
		.postskip_ctrl = PDP_VID1POSTSKIPCTRL_OFFSET,
		.decimate_ctrl = PDP_VID1DECIMATE_CTRL_OFFSET,
		.scale_ctrl = PDP_VID1SCALECTRL_OFFSET,
		.vsinit = PDP_VID1VSINIT_OFFSET,
		.first_vscoeff = PDP_VID1VCOEFF0_OFFSET,
		.last_vscoeff = PDP_VID1VCOEFF8_OFFSET,
		.hsinit = PDP_VID1HSINIT_OFFSET,
		.first_hscoeff = PDP_VID1HCOEFF0_OFFSET,
		.last_hscoeff = PDP_VID1HCOEFF16_OFFSET,
		.scale_size = PDP_VID1SCALESIZE_OFFSET,
		.lumakey_alpha_trans_opaque =
				PDP_VID1LUMAKEY_ALPHA_TRANS_OPAQUE_OFFSET,
		.lumakey_luma_max_min = PDP_VID1LUMAKEY_LUMA_MAX_MIN_OFFSET,
		.lumakey_c_rg = PDP_VID1LUMAKEY_C_RG_OFFSET,
		.lumakey_c_b = PDP_VID1LUMAKEY_C_B_OFFSET,
	}
},
{
	.format = PDP_FORMAT_VIDEO,
	.type = DRM_PLANE_TYPE_OVERLAY,
	.caps = {
		.has_scaler = 1,
		.vs_taps = 4,
		.hs_taps = 8,
		.vs_interpolation_points = 16,
		.hs_interpolation_points = 16,
		.max_scaler_input_width = 2560,
		.has_doubler = 0,
		.has_decimator = 1,
		.has_cosited_chroma_control = 1,
		.has_csc = 1,
	},
	.regs = {
		.surf = PDP_VID2SURF_OFFSET,
		.blend2_r = PDP_VID2BLND2_R_OFFSET,
		.blend2_gb = PDP_VID2BLND2_GB_OFFSET,
		.ctrl = PDP_VID2CTRL_OFFSET,
		.uctrl = PDP_VID2UCTRL_OFFSET,
		.stride = PDP_VID2STRIDE_OFFSET,
		.size = PDP_VID2SIZE_OFFSET,
		.posn = PDP_VID2POSN_OFFSET,
		.galpha = PDP_VID2GALPHA_OFFSET,
		.ckey_r = PDP_VID2CKEY_R_OFFSET,
		.ckey_gb = PDP_VID2CKEY_GB_OFFSET,
		.interleave_ctrl = PDP_VID2INTERLEAVE_CTRL_OFFSET,
		.base_addr = PDP_VID2BASEADDR_OFFSET,
		.u_base_addr = PDP_VID2UBASEADDR_OFFSET,
		.v_base_addr = PDP_VID2VBASEADDR_OFFSET,
		.memctrl = PDP_VID2_MEMCTRL_OFFSET,
		.memthresh = PDP_VID2_MEM_THRESH_OFFSET,
		.skip_ctrl = PDP_VID2SKIPCTRL_OFFSET,
		.postskip_ctrl = PDP_VID2POSTSKIPCTRL_OFFSET,
		.decimate_ctrl = PDP_VID2DECIMATE_CTRL_OFFSET,
		.scale_ctrl = PDP_VID2SCALECTRL_OFFSET,
		.vsinit = PDP_VID2VSINIT_OFFSET,
		.first_vscoeff = PDP_VID2VCOEFF0_OFFSET,
		.last_vscoeff = PDP_VID2VCOEFF8_OFFSET,
		.hsinit = PDP_VID2HSINIT_OFFSET,
		.first_hscoeff = PDP_VID2HCOEFF0_OFFSET,
		.last_hscoeff = PDP_VID2HCOEFF16_OFFSET,
		.scale_size = PDP_VID2SCALESIZE_OFFSET,
		.lumakey_alpha_trans_opaque =
				PDP_VID2LUMAKEY_ALPHA_TRANS_OPAQUE_OFFSET,
		.lumakey_luma_max_min = PDP_VID2LUMAKEY_LUMA_MAX_MIN_OFFSET,
		.lumakey_c_rg = PDP_VID2LUMAKEY_C_RG_OFFSET,
		.lumakey_c_b = PDP_VID2LUMAKEY_C_B_OFFSET,
	}
},
{
	.format = PDP_FORMAT_VIDEO,
	.type = DRM_PLANE_TYPE_OVERLAY,
	.caps = {
		.has_scaler = 1,
		.vs_taps = 4,
		.hs_taps = 8,
		.vs_interpolation_points = 16,
		.hs_interpolation_points = 16,
		.max_scaler_input_width = 2560,
		.has_doubler = 0,
		.has_decimator = 1,
		.has_cosited_chroma_control = 1,
		.has_csc = 1,
	},
	.regs = {
		.surf = PDP_VID3SURF_OFFSET,
		.blend2_r = PDP_VID3BLND2_R_OFFSET,
		.blend2_gb = PDP_VID3BLND2_GB_OFFSET,
		.ctrl = PDP_VID3CTRL_OFFSET,
		.uctrl = PDP_VID3UCTRL_OFFSET,
		.stride = PDP_VID3STRIDE_OFFSET,
		.size = PDP_VID3SIZE_OFFSET,
		.posn = PDP_VID3POSN_OFFSET,
		.galpha = PDP_VID3GALPHA_OFFSET,
		.ckey_r = PDP_VID3CKEY_R_OFFSET,
		.ckey_gb = PDP_VID3CKEY_GB_OFFSET,
		.interleave_ctrl = PDP_VID3INTERLEAVE_CTRL_OFFSET,
		.base_addr = PDP_VID3BASEADDR_OFFSET,
		.u_base_addr = PDP_VID3UBASEADDR_OFFSET,
		.v_base_addr = PDP_VID3VBASEADDR_OFFSET,
		.memctrl = PDP_VID3_MEMCTRL_OFFSET,
		.memthresh = PDP_VID3_MEM_THRESH_OFFSET,
		.skip_ctrl = PDP_VID3SKIPCTRL_OFFSET,
		.postskip_ctrl = PDP_VID3POSTSKIPCTRL_OFFSET,
		.decimate_ctrl = PDP_VID3DECIMATE_CTRL_OFFSET,
		.scale_ctrl = PDP_VID3SCALECTRL_OFFSET,
		.vsinit = PDP_VID3VSINIT_OFFSET,
		.first_vscoeff = PDP_VID3VCOEFF0_OFFSET,
		.last_vscoeff = PDP_VID3VCOEFF8_OFFSET,
		.hsinit = PDP_VID3HSINIT_OFFSET,
		.first_hscoeff = PDP_VID3HCOEFF0_OFFSET,
		.last_hscoeff = PDP_VID3HCOEFF16_OFFSET,
		.scale_size = PDP_VID3SCALESIZE_OFFSET,
		.lumakey_alpha_trans_opaque =
				PDP_VID3LUMAKEY_ALPHA_TRANS_OPAQUE_OFFSET,
		.lumakey_luma_max_min = PDP_VID3LUMAKEY_LUMA_MAX_MIN_OFFSET,
		.lumakey_c_rg = PDP_VID3LUMAKEY_C_RG_OFFSET,
		.lumakey_c_b = PDP_VID3LUMAKEY_C_B_OFFSET,
	}
},
{
	.format = PDP_FORMAT_VIDEO,
	.type = DRM_PLANE_TYPE_OVERLAY,
	.caps = {
		.has_scaler = 1,
		.vs_taps = 4,
		.hs_taps = 8,
		.vs_interpolation_points = 16,
		.hs_interpolation_points = 16,
		.max_scaler_input_width = 2560,
		.has_doubler = 0,
		.has_decimator = 1,
		.has_cosited_chroma_control = 1,
		.has_csc = 1,
	},
	.regs = {
		.surf = PDP_VID4SURF_OFFSET,
		.blend2_r = PDP_VID4BLND2_R_OFFSET,
		.blend2_gb = PDP_VID4BLND2_GB_OFFSET,
		.ctrl = PDP_VID4CTRL_OFFSET,
		.uctrl = PDP_VID4UCTRL_OFFSET,
		.stride = PDP_VID4STRIDE_OFFSET,
		.size = PDP_VID4SIZE_OFFSET,
		.posn = PDP_VID4POSN_OFFSET,
		.galpha = PDP_VID4GALPHA_OFFSET,
		.ckey_r = PDP_VID4CKEY_R_OFFSET,
		.ckey_gb = PDP_VID4CKEY_GB_OFFSET,
		.interleave_ctrl = PDP_VID4INTERLEAVE_CTRL_OFFSET,
		.base_addr = PDP_VID4BASEADDR_OFFSET,
		.u_base_addr = PDP_VID4UBASEADDR_OFFSET,
		.v_base_addr = PDP_VID4VBASEADDR_OFFSET,
		.memctrl = PDP_VID4_MEMCTRL_OFFSET,
		.memthresh = PDP_VID4_MEM_THRESH_OFFSET,
		.skip_ctrl = PDP_VID4SKIPCTRL_OFFSET,
		.postskip_ctrl = PDP_VID4POSTSKIPCTRL_OFFSET,
		.decimate_ctrl = PDP_VID4DECIMATE_CTRL_OFFSET,
		.scale_ctrl = PDP_VID4SCALECTRL_OFFSET,
		.vsinit = PDP_VID4VSINIT_OFFSET,
		.first_vscoeff = PDP_VID4VCOEFF0_OFFSET,
		.last_vscoeff = PDP_VID4VCOEFF8_OFFSET,
		.hsinit = PDP_VID4HSINIT_OFFSET,
		.first_hscoeff = PDP_VID4HCOEFF0_OFFSET,
		.last_hscoeff = PDP_VID4HCOEFF16_OFFSET,
		.scale_size = PDP_VID4SCALESIZE_OFFSET,
		.lumakey_alpha_trans_opaque =
				PDP_VID4LUMAKEY_ALPHA_TRANS_OPAQUE_OFFSET,
		.lumakey_luma_max_min = PDP_VID4LUMAKEY_LUMA_MAX_MIN_OFFSET,
		.lumakey_c_rg = PDP_VID4LUMAKEY_C_RG_OFFSET,
		.lumakey_c_b = PDP_VID4LUMAKEY_C_B_OFFSET,
	}
},
};

/*
 * Check requested plane state from userspace. Must adhere to restrictions
 * for stride, 0.125 to 8x horizontal scaling and 0.125-32x vertical scaling
 * For planes that do not support scaling, only line doubling/decicimation and
 * pixel doubling/halving it fixes up the crtc width/height to possible values.
 * Note. Planes to be scaled must have an even number of displayed pixels in
 * width and height dimensions.
 */
static int pdp_plane_atomic_check(struct drm_plane *plane,
				  struct drm_plane_state *state)
{
	struct pdp_plane *pplane = to_pdp_plane(plane);
	struct pdp_plane_conf *conf = pplane->conf;
	int max_n, max_geomh, geomh_thresh;
	u32 src_w = state->src_w >> 16;
	u32 src_h = state->src_h >> 16;
	u32 read_or_skip_n;

	DRM_DEBUG_DRIVER("atomic checking plane %i\n", pplane->pos);

	if (conf->caps.has_scaler) {
		/* Clamp min/max scaling to 1/8 or 8x scaling */
		if (state->crtc_w != src_w) {
			if (state->crtc_w > (src_w * 8)) {
				state->crtc_w = src_w * 8;
			} else if ((state->crtc_w * 8) < src_w) {
				state->crtc_w = (src_w + 7) / 8;
				state->crtc_w &= ~0x1;
			} else {
				state->crtc_w &= ~0x1;
			}
			/* HW Workaround for plane ypos 0 and down scaling */
			if ((state->crtc_y == 0) && (src_w > state->crtc_w))
				state->crtc_y = 1;
		}
		/* Clamp min/max scaling to 1/8 or 32x scaling */
		if (state->crtc_h != src_h) {
			if (state->crtc_h > (src_h * 8)) {
				state->crtc_h = src_h * 8;
			} else if (state->crtc_h * 31 < src_h) {
				state->crtc_h =
					(src_h + 30) / 31;
				state->crtc_h &= ~0x1;
			} else {
				state->crtc_h &= ~0x1;
			};
		}
	} else {
		/* Clamp fixup crtc_w/h to closest supported by line doubling
		 * or decimation
		 */
		if (state->crtc_w != src_w) {
			if (((u32)(state->crtc_w * 2) >= src_w * 3)
			    && conf->caps.has_doubler)
				state->crtc_w = src_w * 2;
			else if (((u32)(state->crtc_w) * 3 < src_w * 2)
				 && conf->caps.has_decimator)
				state->crtc_w = src_w / 2;
			else
				state->crtc_w = src_w;
		}
		if (state->crtc_h != src_h) {
			max_n = MAX_FIELD_VALUE(PDP_GRPH1DECIMATE_CTRL_GRPH1DECIMATE_LINE_DISCARD_COUNT);
			max_geomh = ((src_h + 1) * max_n) / (max_n + 1);
			geomh_thresh = ((src_h - max_geomh) / 2) + max_geomh;

			if ((u32)(state->crtc_h * 2) >= src_h * 3)
				state->crtc_h = src_h * 2;
			else if (state->crtc_h <= geomh_thresh
				 && conf->caps.has_decimator) {
				read_or_skip_n = 0;

				if ((state->crtc_h * 2) > src_h) {
					read_or_skip_n = (src_h + state->crtc_h) / 2;
					read_or_skip_n /= (src_h - state->crtc_h);
					if (read_or_skip_n > max_n)
						read_or_skip_n = max_n;
					state->crtc_h =	(src_h + 1) * read_or_skip_n / (read_or_skip_n + 1);
				} else {
					read_or_skip_n = ((2 * src_h) - state->crtc_h) / (2*(state->crtc_h));
					if (read_or_skip_n > max_n)
						read_or_skip_n = max_n;
					state->crtc_h =	(src_h + read_or_skip_n) / (read_or_skip_n + 1);
				}
			} else {
				state->crtc_h = src_h;
			}
		}
	}

	return 0;
}

static void pdp_plane_atomic_update(struct drm_plane *plane,
				    struct drm_plane_state *old_state)
{
	struct pdp_plane *pplane = to_pdp_plane(plane);
	struct pdp_private *priv = pplane->dev;

	/* for now just update everything can refine later by
	comparing plane->state to old_state */
	pplane->stale_coeffs = true;

	pdp_plane_update_props(pplane);

	if (!plane->state->fb) {
		/* updating plane property wont call kicl_sim through crtc*/
		pdp_plane_update_flush_regs(priv->dev);
#ifdef __PDP_TARGET_IS_SIMULATOR__
		pdp_kick_simulator(priv);
#endif
		return;
	}

	pdp_update_plane_buffer(pplane);
	pdp_update_stride(pplane);
	pdp_plane_update_mem_thresh(pplane);
	pdp_plane_update_format(pplane);

	pdp_plane_update_vscaler(pplane);
	pdp_plane_update_hscaler(pplane);
	pdp_configure_plane(pplane);
	pdp_plane_update_flush_regs(priv->dev);
}

static void pdp_plane_atomic_disable(struct drm_plane *plane,
				     struct drm_plane_state *old_state)
{
	struct pdp_plane *pplane = to_pdp_plane(plane);

	pdp_disable_plane(pplane);
}

static const struct drm_plane_helper_funcs pdp_plane_helper_funcs = {
	.atomic_disable = pdp_plane_atomic_disable,
	.atomic_check = pdp_plane_atomic_check,
	.atomic_update = pdp_plane_atomic_update,
};

static void pdp_plane_reset(struct drm_plane *plane)
{
	struct pdp_plane_state *state;

	if (plane->state && plane->state->fb)
		drm_framebuffer_unreference(plane->state->fb);

	kfree(plane->state);
	plane->state = NULL;

	state = kzalloc(sizeof(*state), GFP_KERNEL);
	if (!state)
		return;

	state->zpos = to_pdp_plane(plane)->pos;
	state->alpha = 0;
	state->porter_duff_mode = PDP_BLENDMODE_F_OVER_B;
	state->porter_duff_type = PDP_BLENDTYPE_NORMAL;
	state->colourkey_r = 0;
	state->colourkey_g = 0;
	state->colourkey_b = 0;
	state->colourkey_r_mask = 0x3FF;
	state->colourkey_g_mask = 0x3FF;
	state->colourkey_b_mask = 0x3FF;
	state->colourkey_mode = PDP_COLOURKEY_MODE_PREVIOUS;
	state->colourkey_enable = 0;

	plane->state = &state->state;
	plane->state->plane = plane;
}

static struct
drm_plane_state *pdp_plane_atomic_duplicate_state(struct drm_plane *plane)
{
	struct pdp_plane_state *state;
	struct pdp_plane_state *copy;

	if (WARN_ON(!plane->state))
		return NULL;

	state = to_pdp_plane_state(plane->state);
	copy = kmemdup(state, sizeof(*state), GFP_KERNEL);
	if (!copy)
		return NULL;
	if (copy->state.fb)
		drm_framebuffer_reference(copy->state.fb);
	return &copy->state;
}

static void pdp_plane_atomic_destroy_state(struct drm_plane *plane,
					   struct drm_plane_state *state)
{
	if (state->fb)
		drm_framebuffer_unreference(state->fb);

	kfree(to_pdp_plane_state(state));
}

static int pdp_plane_atomic_set_property(struct drm_plane *plane,
					 struct drm_plane_state *state,
					 struct drm_property *property,
					 uint64_t val)
{
	struct pdp_plane_state *pstate = to_pdp_plane_state(state);
	struct pdp_plane *pplane = to_pdp_plane(plane);
	struct pdp_private *priv = pplane->dev;

	if (property == priv->props.alpha) {
		pstate->alpha = val;
	} else if (property == priv->props.zpos) {
		pstate->zpos = val;
		DRM_DEBUG_DRIVER("Ensure that no planes have the same zpos\n");
	} else if (property == priv->props.porter_duff_mode) {
		pstate->porter_duff_mode = val;
	} else if (property == priv->props.porter_duff_type) {
		pstate->porter_duff_type = val;
	} else if (property == priv->props.colourkey_r_property) {
		pstate->colourkey_r = val;
	} else if (property == priv->props.colourkey_g_property) {
		pstate->colourkey_g = val;
	} else if (property == priv->props.colourkey_b_property) {
		pstate->colourkey_b = val;
	} else if (property == priv->props.colourkey_r_mask_property) {
		pstate->colourkey_r_mask = val;
	} else if (property == priv->props.colourkey_g_mask_property) {
		pstate->colourkey_g_mask = val;
	} else if (property == priv->props.colourkey_b_mask_property) {
		pstate->colourkey_b_mask = val;
	} else if (property == priv->props.colourkey_mode_property) {
		pstate->colourkey_mode = val;
	} else if (property == priv->props.colourkey_enable_property) {
		pstate->colourkey_enable = val;
	} else if (property == priv->props.gamma_enable_property) {
		pstate->gamma_enable = val;
	} else {
	       return -EINVAL;
	}

	return 0;
}

static int pdp_plane_atomic_get_property(struct drm_plane *plane,
					 const struct drm_plane_state *state,
					 struct drm_property *property,
					 uint64_t *val)
{
	const struct pdp_plane_state *pstate =
		container_of(state, const struct pdp_plane_state, state);
	struct pdp_private *priv = to_pdp_plane(plane)->dev;

	if (property == priv->props.alpha)
		*val = pstate->alpha;
	else if (property == priv->props.zpos)
		*val = pstate->zpos;
	else if (property == priv->props.porter_duff_mode)
		*val = pstate->porter_duff_mode;
	else if (property == priv->props.porter_duff_type)
		*val = pstate->porter_duff_type;
	else if (property == priv->props.colourkey_r_property)
		*val = pstate->colourkey_r;
	else if (property == priv->props.colourkey_g_property)
		*val = pstate->colourkey_g;
	else if (property == priv->props.colourkey_b_property)
		*val = pstate->colourkey_b;
	else if (property == priv->props.colourkey_r_mask_property)
		*val = pstate->colourkey_r_mask;
	else if (property == priv->props.colourkey_g_mask_property)
		*val = pstate->colourkey_g_mask;
	else if (property == priv->props.colourkey_b_mask_property)
		*val = pstate->colourkey_b_mask;
	else if (property == priv->props.colourkey_mode_property)
		*val = pstate->colourkey_mode;
	else if (property == priv->props.colourkey_enable_property)
		*val = pstate->colourkey_enable;
	else if (property == priv->props.gamma_enable_property)
		*val = pstate->gamma_enable;
	else
		return -EINVAL;

	return 0;
}

static const struct drm_plane_funcs pdp_plane_funcs = {
	.update_plane = drm_atomic_helper_update_plane,
	.disable_plane = drm_atomic_helper_disable_plane,
	.reset = pdp_plane_reset,
	.set_property = drm_atomic_helper_plane_set_property,
	.destroy = drm_plane_cleanup,
	.atomic_duplicate_state = pdp_plane_atomic_duplicate_state,
	.atomic_destroy_state = pdp_plane_atomic_destroy_state,
	.atomic_set_property = pdp_plane_atomic_set_property,
	.atomic_get_property = pdp_plane_atomic_get_property,
};

int pdp_planes_init(struct drm_device *dev)
{
	struct pdp_private *priv;
	struct pdp_plane *plane;
	int ret;
	int i;

	priv = dev->dev_private;

	for (i = 0; i < DRM_PDP_NUM_PLANES; i++) {
		enum drm_plane_type type;

		priv->planes[i].conf = &_stream_template[i];
		type = priv->planes[i].conf->type;
		plane = &priv->planes[i];
		if (plane->conf->format == PDP_FORMAT_GRPH)
			ret = drm_universal_plane_init(dev, &plane->plane,
						       0x1, &pdp_plane_funcs,
						       grph_formats,
						       ARRAY_SIZE(grph_formats),
						       type);
		else
			ret = drm_universal_plane_init(dev, &plane->plane,
						       0x1, &pdp_plane_funcs,
						       vid_formats,
						       ARRAY_SIZE(vid_formats),
						       type);

		if (ret < 0) {
			DRM_INFO("Failed plane init\n");
			return ret;
		}
		drm_plane_helper_add(&plane->plane, &pdp_plane_helper_funcs);
		plane->dev = priv;
		plane->pos = i;

		drm_object_attach_property(&plane->plane.base,
					   priv->props.alpha, 0);
		drm_object_attach_property(&plane->plane.base,
					   priv->props.zpos, plane->pos);
		drm_object_attach_property(&plane->plane.base,
					   priv->props.porter_duff_type,
					   PDP_BLENDTYPE_NORMAL);
		drm_object_attach_property(&plane->plane.base,
					   priv->props.porter_duff_mode,
					   PDP_BLENDMODE_F_OVER_B);
		drm_object_attach_property(&plane->plane.base,
			priv->props.colourkey_r_property, 0);
		drm_object_attach_property(&plane->plane.base,
			priv->props.colourkey_g_property, 0);
		drm_object_attach_property(&plane->plane.base,
			priv->props.colourkey_b_property, 0);
		drm_object_attach_property(&plane->plane.base,
			priv->props.colourkey_r_mask_property, 0x3FF);
		drm_object_attach_property(&plane->plane.base,
			priv->props.colourkey_g_mask_property, 0x3FF);
		drm_object_attach_property(&plane->plane.base,
			priv->props.colourkey_b_mask_property, 0x3FF);
		drm_object_attach_property(&plane->plane.base,
					   priv->props.colourkey_mode_property,
					   PDP_COLOURKEY_MODE_PREVIOUS);
		drm_object_attach_property(&plane->plane.base,
					   priv->props.colourkey_enable_property,
					   0);
		drm_object_attach_property(&plane->plane.base,
					   priv->props.gamma_enable_property,
					   0);
	}
	return 0;
}
