#include <linux/lcm.h>

#include <drm/drm_atomic_helper.h>
#include <drm/drm_crtc.h>
#include <drm/drm_plane_helper.h>
#include <drm/drmP.h>
#include <drm/drm_fb_cma_helper.h>
#include <drm/drm_gem_cma_helper.h>

#include "pdp_plane_funcs.h"
#include "pdp_crtc.h"
#include "pdp_drv.h"
#include "pdp_modeset.h"
#include "pdp_plane.h"
#include "pdpfb2_regs.h"
#include "pdp_fb.h"

enum pdp_display_mode {
	PDP_FULLY_POWERED_DOWN,
	PDP_PARTIALLY_POWERED_DOWN,
	PDP_ENABLED,
};

static inline int fix(int p, int a)
{
	return a << p;
}

static inline int fix_div(int p, int n, int d)
{
	return (n << p) / d;
}

static inline int fix_divl(int p, int n, int d)
{
	return (int)(((u64)n << p) / d);
}

static inline int fix_mul(int p, int a, int b)
{
	return (a * b) >> p;
}

static inline int fix_mull(int p, int a, int b)
{
	return (int)(((u64)a * (u64)b) >> p);
}

static inline int fix_sin(int p, int x)
{
	/* tailor expansion */
	int pow;
	int sum;

	/* in [-pi,pi] */
	int pi2 = fix_divl(p, 6283185, 1000000);
	int pi = pi2 / 2;

	x = x % pi2;
	if (x > pi)
		x -= pi2;
	else if (x < -pi)
		x += pi2;

	sum = x;

	/* x represents x^2 from now on */
	x = fix_mull(p, x, x);

	/* -x^3/3! */
	pow = fix_mull(p, sum, x);
	sum -= pow / 6;
	/* +x^5/5! */
	pow = fix_mull(p, pow, x);
	sum += pow / 120;
	/* -x^7/7! */
	pow = fix_mull(p, pow, x);
	sum -= pow / 5040;
	/* +x^9/9! */
	pow = fix_mull(p, pow, x);
	sum += pow / 362880;

	return sum;
}

static inline int fix_cos(int p, int x)
{
	/* tailor expansion */
	int pow;
	int sum = fix(p, 1);

	/* in [-pi,pi] */
	int pi2 = fix_divl(p, 6283185, 1000000);
	int pi = pi2 / 2;

	x = x % pi2;
	if (x > pi)
		x -= pi2;
	else if (x < -pi)
		x += pi2;

	/* x represents x^2 from now on */
	x = fix_mull(p, x, x);

	/* -x^2/2! */
	pow = x;
	sum -= pow / 2;
	/* +x^4/4! */
	pow = fix_mull(p, pow, x);
	sum += pow / 24;
	/* -x^6/6! */
	pow = fix_mull(p, pow, x);
	sum -= pow / 720;
	/* +x^8/8! */
	pow = fix_mull(p, pow, x);
	sum += pow / 40320;
	/* -x^10/10! */
	pow = fix_mull(p, pow, x);
	sum -= pow / 3628800;

	return sum;
}

static void
pdp_calc_scale_coeffs(u32 res, u32 size, int T, int I, int *coeffs, int count)
{
	int midpoint = T * I / 2;
#define SCALE_PRECISION 8
#define RESULT_PRECISION 6
#define DIV(a, b)       fix_div(SCALE_PRECISION, a, b)
#define DIVL(a, b)      fix_divl(SCALE_PRECISION, a, b)
#define MUL(a, b)       fix_mul(SCALE_PRECISION, a, b)
#define SIN(a)	  fix_sin(SCALE_PRECISION, a)
#define COS(a)	  fix_cos(SCALE_PRECISION, a)
#define FIX(a)	  fix(SCALE_PRECISION, a)
#define RESULT(a)       ((a) >> (SCALE_PRECISION - RESULT_PRECISION))
	int fpi = DIVL(3141593, 1000000);
	int fS = min(FIX(1), /* scale factor clamped to 1 */
		DIV(size, res));
	int fA = DIV(54, 100);
	int c, i;

	int fpiS = MUL(fpi, fS);
	int f2piSoT = fpiS * 2 / T;
	int f1mA = FIX(1) - fA;
	int fT = FIX(T);
	int ftotal = 0;

	/* Calculate coefficients */
	for (c = 0; c <= midpoint; ++c) {
		/* t + i/I */
		int ftpioI = FIX(c) / I;
		int fx = MUL(fpiS, fT / 2 - ftpioI);
		int fco = fA - MUL(f1mA, COS(MUL(f2piSoT, ftpioI)));

		if (fx > DIV(1, 10))
			fco = MUL(DIV(SIN(fx), fx), fco);
		ftotal += fco;
		coeffs[c] = fco;
	}

	/*
	 * Interpolation points I/2+1 to I-1 have same values as 1 to I/2-1,
	 * but in opposite order, so we only need to normalise the first half
	 * of the interpolation points taking mirroring into account.
	 */
	for (i = 0; i <= I / 2; ++i) {
		int sum, err, dir = 0;

		/*
		 * Normalise coefficients in each interpolation point and
		 * convert into result format.
		 */
		sum = 0;
		for (c = i; c <= midpoint; c += I)
			sum += coeffs[c];
		for (; c < midpoint * 2; c += I)
			sum += coeffs[midpoint * 2 - c];
		/* Careful not to modify a coefficient twice */
		for (c = i; c <= midpoint; c += I)
			coeffs[c] = RESULT(DIV(coeffs[c], sum));
		if (i & (I / 2 - 1))
			for (c = I - i; c <= midpoint; c += I)
				coeffs[c] = RESULT(DIV(coeffs[c], sum));

		/*
		 * Find fixed point error from normalisation.
		 */
		sum = 0;
		for (c = i; c <= midpoint; c += I)
			sum += coeffs[c];
		for (; c < midpoint * 2; c += I)
			sum += coeffs[midpoint * 2 - c];
		err = sum - (1 << RESULT_PRECISION);
		if (err > 0)
			dir = -1;
		else if (err < 0)
			dir = 1;
		else
			continue;
		/*
		 * Distribute the error over the tap coefficients, preferring
		 * to change central coefficients.
		 */
		if (i == 0) {
			/*
			 * Special case: contains midpoint. Work out from
			 * center until no error or an odd value (which can be
			 * fixed by adjusting midpoint again).
			 */
			while (err > 1 || err < -1) {
				coeffs[midpoint] += dir;
				err += dir;
				for (c = midpoint - I;
					 (err > 1 || err < -1) && c >= I;
					 c -= I) {
					coeffs[c] += dir;
					err += dir * 2;
				}
			}
			if (err & 1) {
				coeffs[midpoint] += dir;
				err += dir;
			}
		} else if (i == I / 2) {
			/*
			 * Special case: tap coeffs mirror. This also means err
			 * is always even with this interpolation point.
			 */
			int last = i + I * (T / 2 - 1);

			while (err) {
				for (c = last; err && c >= 0; c -= I) {
					coeffs[c] += dir;
					err += dir * 2;
				}
			}
		} else {
			/*
			 * No tap coefficient mirrors another. Work out from
			 * center adjusting values.
			 */
			int last = midpoint - i;
			int offset = i * 2 - I;

			while (err) {
				for (c = last; err && c >= 0; c -= I) {
					coeffs[c] += dir;
					err += dir;
					if (!err)
						break;
					coeffs[c + offset] += dir;
					err += dir;
				}
			}
		}
	}

	/* reflect about midpoint in spare coeffs */
	if (count > midpoint * 2)
		count = midpoint * 2;
	for (c = midpoint + 1; c < count; ++c)
		coeffs[c] = coeffs[2 * midpoint - c];
#undef SCALE_PRECISION
#undef RESULT_PRECISION
#undef DIV
#undef DIVL
#undef MUL
#undef SIN
#undef COS
#undef FIX
#undef RESULT
}

static void pdp_calc_pass_thru_coeffs(int t, int ip, int *p, int c)
{
	const int m = t * ip / 2;
	int i;

	for (i = 0; i < c; ++c)
		p[i] = 0;

	for (i = m; i < m + ip / 2; ++i)
		p[i] = 0x40;

	p[m + ip / 2] = 0x20;
	p[m - ip / 2] = 0x20;

	for (i = m - ip + 1; i < m; ++i)
		p[i] = 0x40;
}

void pdp_plane_update_hscaler(struct pdp_plane *pplane)
{
	struct drm_plane *plane = &pplane->plane;
	struct drm_plane_state *state = plane->state;
	struct pdp_plane_state *pstate =
				to_pdp_plane_state(pplane->plane.state);
	struct pdp_private *priv = pplane->dev;
	int coeffs[1 + PDP_MAX_TAPS * PDP_MAX_INTERPOLATION_POINTS];
	u32 scale_width, hpitch, hinitial, hpostclip;
	u32 reg;
	int i;
	u32 en;

	DRM_DEBUG_DRIVER(
		"updating hscaler, plane source: %dx%d target: %dx%d\n",
			 state->src_w >> 16, state->src_h >> 16,
			 state->crtc_w, state->crtc_h);

	if (pplane->conf->caps.has_scaler) {
		DRM_DEBUG_DRIVER("Using scaler for horizontal scaling\n");
		en = !((state->src_w >> 16) == state->crtc_w);

		if (pstate->hskip) {
			DRM_DEBUG_DRIVER(
				"forcing scaler on and adjusting plane size\n");
			en = 1;
			reg = pdp_read(priv->dev, pplane->conf->regs.size);
			SET_FIELD(reg, PDP_GRPH1SIZE_GRPH1WIDTH,
				  (state->src_w >> 16) + pstate->hskip - 1);
			pdp_write(priv->dev, pplane->conf->regs.size, reg);
		}

		reg = pdp_read(priv->dev, pplane->conf->regs.scale_ctrl);
		SET_FIELD(reg, PDP_VID1SCALECTRL_VID1HSCALEBP, !en);
		pdp_write(priv->dev, pplane->conf->regs.scale_ctrl, reg);

		if (!en) {
			reg = pdp_read(priv->dev,
				       pplane->conf->regs.scale_size);
			SET_FIELD(reg, PDP_VID1SCALESIZE_VID1SCALEWIDTH,
				  state->crtc_w - 1);
			pdp_write(priv->dev, pplane->conf->regs.scale_size,
				  reg);
			return;
		}

		reg = pdp_read(priv->dev, pplane->conf->regs.scale_ctrl);
		SET_FIELD(reg, PDP_VID1SCALECTRL_VID1HORDER, 0x7);
		pdp_write(priv->dev, pplane->conf->regs.scale_ctrl, reg);

		hpitch =
		   (fix_div(12, state->src_w >> 16, state->crtc_w) + 1) >> 1;
		hinitial = fix_div(11, pplane->conf->caps.hs_taps, 2);
		hpostclip = 0;
		scale_width = state->crtc_w;

		if (pstate->hskip < 19) {
			hinitial += pstate->hskip << 11;
		} else {
			u32 hpostclip_to_add =
					((pstate->hskip - 19) << 26) / hpitch;
			hpostclip_to_add >>= 15;
			hpostclip += hpostclip_to_add;
			if (hpostclip > MAX_FIELD_VALUE(PDP_VID1POSTSKIPCTRL_VID1HPOSTCLIP)) {
				hpostclip = MAX_FIELD_VALUE(PDP_VID1POSTSKIPCTRL_VID1HPOSTCLIP);
				DRM_DEBUG_DRIVER(
		"insufficient post scale clip px for requested dimensions\n");
			}
			hinitial += 19 << 11;
		}
		reg = pdp_read(priv->dev, pplane->conf->regs.hsinit);
		SET_FIELD(reg, PDP_VID1HSINIT_VID1HINITIAL, hinitial);
		SET_FIELD(reg, PDP_VID1HSINIT_VID1HPITCH, hpitch);
		pdp_write(priv->dev, pplane->conf->regs.hsinit, reg);

		reg = pdp_read(priv->dev, pplane->conf->regs.postskip_ctrl);
		SET_FIELD(reg, PDP_VID1POSTSKIPCTRL_VID1HPOSTCLIP, hpostclip);
		pdp_write(priv->dev, pplane->conf->regs.postskip_ctrl, reg);

		reg = pdp_read(priv->dev, pplane->conf->regs.scale_size);
		SET_FIELD(reg, PDP_VID1SCALESIZE_VID1SCALEWIDTH,
			  scale_width + hpostclip - 1);
		pdp_write(priv->dev, pplane->conf->regs.scale_size, reg);

		if (pplane->stale_coeffs) {
			if ((state->src_w >> 16) == state->crtc_w) {
				pdp_calc_pass_thru_coeffs(
						pplane->conf->caps.hs_taps,
				pplane->conf->caps.hs_interpolation_points,
						coeffs, ARRAY_SIZE(coeffs));
			} else {
				pdp_calc_scale_coeffs(state->src_w >> 16,
						      state->crtc_w,
						pplane->conf->caps.hs_taps,
				pplane->conf->caps.hs_interpolation_points,
						coeffs,
						ARRAY_SIZE(coeffs));
			}
			/* Now to update the regs */
			for (i = 0;
			     i < pplane->conf->caps.hs_interpolation_points;
			     ++i) {
				reg = ((0xFF & coeffs[i << 2])
				| ((0xFF & coeffs[(i << 2) + 1]) << 8)
				| ((0xFF & coeffs[(i << 2) + 2]) << 16)
				| ((0xFF & coeffs[(i << 2) + 3]) << 24));
				pdp_write(priv->dev,
					  pplane->conf->regs.first_hscoeff
					  +(i << 2), reg);
			}
			pdp_write(priv->dev, pplane->conf->regs.last_hscoeff,
							0xff & coeffs[i << 2]);
		}
	} else {
	       DRM_DEBUG_DRIVER("Using pixel doubling/halving\n");
		/* plane doesnt have a scaler */
		if (pplane->conf->caps.has_doubler) {
			reg = pdp_read(priv->dev, pplane->conf->regs.blend2_r);
			SET_FIELD(reg, PDP_GRPH1BLND2_R_GRPH1PIXDBL,
					state->crtc_w > (state->src_w >> 16));
			pdp_write(priv->dev, pplane->conf->regs.blend2_r, reg);
		}
		if (pplane->conf->caps.has_decimator) {
			reg = pdp_read(priv->dev,
				       pplane->conf->regs.decimate_ctrl);
			SET_FIELD(reg,
			PDP_GRPH1DECIMATE_CTRL_GRPH1DECIMATE_PIXEL_HALVE,
				  state->crtc_w < (state->src_w >> 16));
			if (state->crtc_w < (state->src_w >> 16))
				SET_FIELD(reg,
				PDP_GRPH1DECIMATE_CTRL_GRPH1DECIMATE_EN, 1);
			pdp_write(priv->dev, pplane->conf->regs.decimate_ctrl,
				  reg);
		}
	}
}
#define MAX_LINE_DISCARD_COUNT \
	(MAX_FIELD_VALUE(\
	PDP_GRPH1DECIMATE_CTRL_GRPH1DECIMATE_LINE_DISCARD_COUNT))

static void
pdp_configure_decimator(struct pdp_plane *pplane, u32 src_h, u32 crtc_h)
{
	struct pdp_private *priv = pplane->dev;
	u32 read_or_skip = 0;
	u32 reg = pdp_read(priv->dev, pplane->conf->regs.decimate_ctrl);

	if (crtc_h < src_h) {
		/* output display plane is smaller than source */
		if ((crtc_h * 2) > src_h) {
			SET_FIELD(reg,
		PDP_GRPH1DECIMATE_CTRL_GRPH1DECIMATE_LINE_DISCARD_MODE, 1);
			read_or_skip = crtc_h / (src_h - crtc_h);
		} else {
			SET_FIELD(reg,
		PDP_GRPH1DECIMATE_CTRL_GRPH1DECIMATE_LINE_DISCARD_MODE, 0);
			read_or_skip = (src_h - 1) / crtc_h;
		}

		if (read_or_skip > MAX_LINE_DISCARD_COUNT)
			read_or_skip = MAX_LINE_DISCARD_COUNT;

		SET_FIELD(reg, PDP_GRPH1DECIMATE_CTRL_GRPH1DECIMATE_EN, 1);
	}
	SET_FIELD(reg, PDP_GRPH1DECIMATE_CTRL_GRPH1DECIMATE_LINE_DISCARD_COUNT,
		  read_or_skip);
	pdp_write(priv->dev, pplane->conf->regs.decimate_ctrl, reg);
}

static void pdp_calculate_vinit(bool plane_interlaced, bool crtc_interlaced,
				u32 taps, u32 vpitch, u32 *vinit0,
				u32 *vinit1)
{
	*vinit0 = taps << 10;
	*vinit1 = *vinit0;

	if (plane_interlaced) {
		if (crtc_interlaced) {
			if (vpitch > (1 << 11))
				*vinit1 += (vpitch - (1 << 11)) >> 1;
			else if (vpitch < (1 << 11))
				*vinit0 += ((1 << 11) - vpitch) >> 1;
		} else {
			*vinit0 += 1 << 10;
		}
	} else if (crtc_interlaced) {
		*vinit0 += vpitch >> 1;
	}
}

void pdp_plane_update_vscaler(struct pdp_plane *pplane)
{
	struct pdp_private *priv = pplane->dev;
	struct drm_plane *plane = &pplane->plane;
	struct drm_plane_state *state = plane->state;
	int vscoeffs[1 + PDP_MAX_TAPS * PDP_MAX_INTERPOLATION_POINTS];
	u32 reg;

	DRM_DEBUG_DRIVER("updating vscaler: %dx%d target: %dx%d\n",
			 state->src_w >> 16, state->src_h >> 16,
			 state->crtc_w, state->crtc_h);

	if (pplane->conf->caps.has_scaler) {
		u32 i, vpitch, hs_before_vs, vinit0, vinit1;
		bool vs_en = !(state->crtc_h == (state->src_h >> 16));

		vpitch = fix_div(11, (state->src_h >> 16) - 2, state->crtc_h);
		pdp_calculate_vinit(0, 0, pplane->conf->caps.vs_taps, vpitch,
				    &vinit0, &vinit1);
		reg = pdp_read(priv->dev, pplane->conf->regs.scale_ctrl);
		SET_FIELD(reg, PDP_VID1SCALECTRL_VID1VSCALEBP, !vs_en);
		pdp_write(priv->dev, pplane->conf->regs.scale_ctrl, reg);

		reg = pdp_read(priv->dev, pplane->conf->regs.scale_size);
		SET_FIELD(reg, PDP_VID1SCALESIZE_VID1SCALEHEIGHT,
			  state->crtc_h - 1);
		pdp_write(priv->dev, pplane->conf->regs.scale_size, reg);

		reg = pdp_read(priv->dev, pplane->conf->regs.postskip_ctrl);
		SET_FIELD(reg, PDP_VID1POSTSKIPCTRL_VID1VPOSTCLIP, 0);
		pdp_write(priv->dev, pplane->conf->regs.postskip_ctrl, reg);

		if (!vs_en)
			return;
		hs_before_vs = ((state->src_h >> 16) > 2048);
					/*TODO move to dts 2048*/
		reg = pdp_read(priv->dev, pplane->conf->regs.scale_ctrl);
		SET_FIELD(reg, PDP_VID1SCALECTRL_VID1HSBEFOREVS, hs_before_vs);
		SET_FIELD(reg, PDP_VID1SCALECTRL_VID1VSURUNCTRL, 1);
		SET_FIELD(reg, PDP_VID1SCALECTRL_VID1VORDER, 3);
		SET_FIELD(reg, PDP_VID1SCALECTRL_VID1VPITCH, vpitch);
		pdp_write(priv->dev, pplane->conf->regs.scale_ctrl, reg);

		reg = pdp_read(priv->dev, pplane->conf->regs.vsinit);
		SET_FIELD(reg, PDP_VID1VSINIT_VID1VINITIAL0, vinit0);
		SET_FIELD(reg, PDP_VID1VSINIT_VID1VINITIAL1, vinit1);
		pdp_write(priv->dev, pplane->conf->regs.vsinit, reg);

		pdp_calc_scale_coeffs(state->src_h >> 16,
				      state->crtc_h,
				      pplane->conf->caps.vs_taps,
				pplane->conf->caps.vs_interpolation_points,
				      vscoeffs, ARRAY_SIZE(vscoeffs));
		for (i = 0;
		     i < (pplane->conf->caps.vs_interpolation_points / 2);
		     ++i) {
			reg = ((0xFF & vscoeffs[i << 2])
				| ((0xFF & vscoeffs[(i << 2) + 1]) << 8)
				| ((0xFF & vscoeffs[(i << 2) + 2]) << 16)
				| ((0xFF & vscoeffs[(i << 2) + 3]) << 24));
			pdp_write(priv->dev,
				  pplane->conf->regs.first_vscoeff + (i << 2),
				  reg);
		}
		pdp_write(priv->dev,
			  pplane->conf->regs.last_vscoeff,
			  0xff & vscoeffs[i << 2]);
	} else {
		/* no scaler */
		if (pplane->conf->caps.has_doubler) {
			reg = pdp_read(priv->dev,
				       pplane->conf->regs.blend2_r);
			SET_FIELD(reg, PDP_GRPH1BLND2_R_GRPH1LINDBL,
				  (state->crtc_h > (state->src_h >> 16)));
			pdp_write(priv->dev, pplane->conf->regs.blend2_r,
				  reg);
		}
		if (pplane->conf->caps.has_decimator)
			pdp_configure_decimator(pplane,
						state->src_h >> 16,
						state->crtc_h);
	}
}

bool zpos_valid(struct drm_device *dev)
{
	int i;
	u8 used = 0;
	u32 zpos;

	for (i = 0; i < 8; i++) {
		zpos = pdp_read(dev, PDP_GRPH1CTRL_OFFSET + (4 * i)) >> 24;
		used |= BIT(zpos & 0x7);
	}

	return used == 0xff;
}

void pdp_plane_update_flush_regs(struct drm_device *dev)
{
	u32 reg;

	if (!zpos_valid(dev)) {
		DRM_DEBUG_DRIVER("Delayed updating registers - bad zpos\n");
		return;
	}

	reg = pdp_read(dev, PDP_REGISTER_UPDATE_CTRL_OFFSET);

	DRM_DEBUG_DRIVER("Flush Register update\n");
	SET_FIELD(reg, PDP_REGISTER_UPDATE_CTRL_REGISTERS_VALID, 1);
	SET_FIELD(reg, PDP_REGISTER_UPDATE_CTRL_USE_VBLANK, 1);
	pdp_write(dev, PDP_REGISTER_UPDATE_CTRL_OFFSET, reg);
}

void pdp_kick_simulator(struct pdp_private *priv)
{
	DRM_INFO("Kicking simulator\n");
	pdp_write(priv->dev, 0x7a4, 0x11111111);
}

void pdp_update_plane_buffer(struct pdp_plane *pplane)
{
	struct drm_plane *plane = &pplane->plane;
	struct drm_framebuffer *fb = pplane->plane.state->fb;
	struct pdp_fb_cma *pdpfb = drm_fb_to_pdp_fb(fb);
	struct pdp_gem_object *gem = drm_fb_obj(fb, 0);
	struct drm_plane_state *state = plane->state;
	struct pdp_plane_state *pstate =
				to_pdp_plane_state(pplane->plane.state);
	/* state ->src_x/y are in Q16 format */
	u32 xoffset = state->src_x >> 16;
	u32 yo = state->src_y >> 16;
	u32 base_addr;
	u32 reg = 0;
	u32 uctrl;

	u32 ybo = 0;
	u32 ubo = 0;
	u32 vbo = 0;
	u32 xob = 0;
	u32 bpp = pdp_format_info(fb->pixel_format)->bpp;

	switch (fb->pixel_format) {
	case DRM_FORMAT_YUV420:
		xob = xoffset & ~0x3F;
		ybo = yo * fb->pitches[0] + xob;
		ubo = (yo / 2) * fb->pitches[1] + xob / 2;
		vbo = (yo / 2) * fb->pitches[1] + xob / 2;
		break;
	case DRM_FORMAT_NV12:
	case DRM_FORMAT_NV21:
		xob = xoffset & ~0x1F;
		ybo = yo * fb->pitches[0] +  xob;
		ubo = (yo / 2) * fb->pitches[1] + xob;
		vbo = ubo;
		break;
	case DRM_FORMAT_NV15:
	case DRM_FORMAT_NV51:
		xob = (xoffset / 24) * 24;
		ybo = yo * fb->pitches[0] + xob * 4 / 3;
		ubo = (yo / 2) * fb->pitches[1] + xob * 4 / 3;
		vbo = ubo;
		break;
	case DRM_FORMAT_NV20:
	case DRM_FORMAT_NV02:
		xob = (xoffset / 24) * 24;
		ybo = yo * fb->pitches[0] + xob * 4 / 3;
		ubo = yo * fb->pitches[1] + xob * 4 / 3;
		vbo = ubo;
		break;
	default:
		{
			u32 ybxo = xoffset * bpp / 8;
			u32 align = lcm(1 << PDP_YADDR_ALIGN, bpp/8);
			u32 ybxoa = (ybxo / align) * align;

			ybo = yo * fb->pitches[0] + ybxoa;
			xob = ybxoa * 8 / bpp;
		}
	}

	pstate->hskip = xoffset - xob;

	base_addr = gem->paddr + fb->offsets[0] + ybo;
	DRM_DEBUG_DRIVER("Writing plane %i base buffer: %x yo: %x\n",
			 pplane->pos, base_addr, ybo);

	SET_FIELD(reg, PDP_GRPH1BASEADDR_GRPH1BASEADDR, base_addr);
	pdp_write(plane->dev, pplane->conf->regs.base_addr, base_addr);

	if (pdp_format_info(fb->pixel_format)->planes > 1) {
		DRM_DEBUG_DRIVER("uoffset: %x %x voffset: %x %x\n",
				 fb->offsets[1], ubo, fb->offsets[2], vbo);
		gem = pdpfb->obj[1];
		pdp_write(plane->dev, pplane->conf->regs.u_base_addr,
			  gem->paddr + ubo + fb->offsets[1]);

		/* If format is bi planar, write plane 3hw equal to plane 2 */
		if (pdp_format_info(fb->pixel_format)->planes == 3) {
			gem = pdpfb->obj[2];
			pdp_write(plane->dev, pplane->conf->regs.v_base_addr,
				  gem->paddr + vbo + fb->offsets[2]);
		} else {
			pdp_write(plane->dev, pplane->conf->regs.v_base_addr,
				  gem->paddr + vbo + fb->offsets[1]);
		}

		if (fb->pitches[0] == fb->pitches[1])
			reg = PDP_UADDR_UV_STRIDE_EQUAL_TO_Y_STRIDE;
		else if (fb->pitches[0] < fb->pitches[1])
			reg = PDP_UADDR_UV_STRIDE_EQUAL_TO_DOUBLE_Y_STRIDE;
		else
			reg = PDP_UADDR_UV_STRIDE_EQUAL_TO_HALF_Y_STRIDE;

		DRM_DEBUG_DRIVER("Plane pitches %x %x %x\n",
				 fb->pitches[0], fb->pitches[1],
				 fb->pitches[2]);
		uctrl = pdp_read(plane->dev, pplane->conf->regs.uctrl);
		SET_FIELD(uctrl, PDP_VID1UCTRL_VID1UVHALFSTR, reg);
		pdp_write(plane->dev, pplane->conf->regs.uctrl, uctrl);
	}
}

/* TODO this isnt a plane function, should be with crtc/bgnd col??*/
void pdp_configure_dithering(struct pdp_private *priv)
{
	pdp_write(priv->dev, PDP_DITHERMATRIX2_OFFSET, 0x0000006C);
	pdp_write(priv->dev, PDP_DITHERMATRIX4_0_OFFSET, 0x7B48F3C0);
	pdp_write(priv->dev, PDP_DITHERMATRIX4_1_OFFSET, 0x596AC1E2);
	pdp_write(priv->dev, PDP_DITHERMATRIX8_0_OFFSET, 0x03F0CC00);
	pdp_write(priv->dev, PDP_DITHERMATRIX8_1_OFFSET, 0x1083F3F3);
	pdp_write(priv->dev, PDP_DITHERMATRIX8_2_OFFSET, 0x2F4E372C);
	pdp_write(priv->dev, PDP_DITHERMATRIX8_3_OFFSET, 0x3413821F);
	pdp_write(priv->dev, PDP_DITHERMATRIX8_4_OFFSET, 0x28DC7ECB);
	pdp_write(priv->dev, PDP_DITHERMATRIX8_5_OFFSET, 0x1BAD4918);
	pdp_write(priv->dev, PDP_DITHERMATRIX8_6_OFFSET, 0x0EC825E7);
	pdp_write(priv->dev, PDP_DITHERMATRIX8_7_OFFSET, 0x3D37107E);
	pdp_write(priv->dev, PDP_DITHERMATRIX8_8_OFFSET, 0x217AE4A2);
	pdp_write(priv->dev, PDP_DITHERMATRIX8_9_OFFSET, 0x3A29DB51);
	pdp_write(priv->dev, PDP_DITHERMATRIX8_10_OFFSET, 0x05E49D86);
	pdp_write(priv->dev, PDP_DITHERMATRIX8_11_OFFSET, 0x1699AAB5);
	pdp_write(priv->dev, PDP_DITHERMATRIX8_12_OFFSET, 0x00565669);
}

void pdp_plane_update_blender(struct pdp_plane *pplane)
{
	struct pdp_private *priv = pplane->dev;
	struct pdp_plane_state
			*pstate = to_pdp_plane_state(pplane->plane.state);
	unsigned int reg_off = 0x668 + (pstate->zpos * 4);
	unsigned int mode = pstate->porter_duff_mode;
	unsigned int type = pstate->porter_duff_type;
	u32 reg;

	reg = pdp_read(priv->dev, reg_off);
	SET_FIELD(reg, PDP_PORTER_BLND0_BLND0PORTERMODE, mode);
	SET_FIELD(reg, PDP_PORTER_BLND0_BLND0BLENDTYPE, type);
	pdp_write(priv->dev, reg_off, reg);

	DRM_DEBUG_DRIVER("Setting blend mode '%s' '%s'for plane %u\n",
			 pdp_porter_duff_mode_enum_list[mode].name,
			 pdp_porter_duff_type_enum_list[type].name,
			 pplane->pos);
}

void pdp_plane_update_mem_thresh(struct pdp_plane *pplane)
{
	struct pdp_private *priv = pplane->dev;
	int burst_length = 8;
	u32 reg;

	DRM_DEBUG_DRIVER("Writing memory thresholds for plane %d\n",
			 pplane->pos);
	reg = pdp_read(priv->dev, pplane->conf->regs.memthresh);
	SET_FIELD(reg, PDP_GRPH1_MEM_THRESH_GRPH1_THRESHOLD,
		  256 - burst_length);
	SET_FIELD(reg, PDP_GRPH1_MEM_THRESH_GRPH1_YTHRESHOLD,
		  256 - burst_length);
	SET_FIELD(reg, PDP_GRPH1_MEM_THRESH_GRPH1_UVTHRESHOLD,
		  128 - burst_length);
	pdp_write(priv->dev, pplane->conf->regs.memthresh, reg);

	reg = pdp_read(priv->dev, pplane->conf->regs.memctrl);
	SET_FIELD(reg, PDP_GRPH1_MEMCTRL_GRPH1_BURSTLEN, burst_length);
	SET_FIELD(reg, PDP_GRPH1_MEMCTRL_GRPH1_LOCAL_GLOBAL_MEMCTRL, 1);
	pdp_write(priv->dev, pplane->conf->regs.memctrl, reg);
}

void pdp_update_stride(struct pdp_plane *pplane)
{
	struct pdp_private *priv = pplane->dev;
	struct drm_plane *plane = &pplane->plane;
	struct drm_framebuffer *fb = plane->state->fb;
	u32 reg, stride;

	stride = fb->pitches[0] >> 5;

	DRM_DEBUG_DRIVER("Writing Plane %d stride %i(%i)\n", pplane->pos,
			 stride, fb->pitches[0]);
	reg = pdp_read(priv->dev, pplane->conf->regs.stride);
	SET_FIELD(reg, PDP_GRPH1STRIDE_GRPH1STRIDE, stride - 1);
	pdp_write(plane->dev, pplane->conf->regs.stride, reg);
}

void pdp_plane_update_format(struct pdp_plane *pplane)
{
	struct pdp_private *priv = pplane->dev;
	struct drm_plane_state *state = pplane->plane.state;
	u32 reg, format;
	const struct pdp_format_info *pformat;

	pformat = pdp_format_info(state->fb->pixel_format);

	format = pformat->hwv;

	DRM_DEBUG_DRIVER("Writing format for plane %i\n", pplane->pos);
	reg = pdp_read(priv->dev, pplane->conf->regs.surf);
	switch (state->fb->pixel_format) {
	case DRM_FORMAT_NV21:
	case DRM_FORMAT_NV12:
	case DRM_FORMAT_NV15:
	case DRM_FORMAT_NV51:
	case DRM_FORMAT_NV20:
	case DRM_FORMAT_NV02:
	case DRM_FORMAT_YUV420:
	case DRM_FORMAT_XYUV101010:
	case DRM_FORMAT_YUYV:
	case DRM_FORMAT_YVYU:
	case DRM_FORMAT_UYVY:
	case DRM_FORMAT_VYUY:
	case DRM_FORMAT_AYUV:
		SET_FIELD(reg, PDP_GRPH1SURF_GRPH1USECSC, 1);
		break;
	default:
		SET_FIELD(reg, PDP_GRPH1SURF_GRPH1USECSC, 0);
	}
	SET_FIELD(reg, PDP_GRPH1SURF_GRPH1PIXFMT, format);
	pdp_write(priv->dev, pplane->conf->regs.surf, reg);
}

void pdp_plane_update_props(struct pdp_plane *pplane)
{
	struct pdp_private *priv = pplane->dev;
	struct pdp_plane_state
	*pstate = to_pdp_plane_state(pplane->plane.state);
	u32 reg;
	unsigned int alpha_mode = pstate->alpha >> 10;

	reg = pdp_read(priv->dev, pplane->conf->regs.ctrl);
	DRM_DEBUG_DRIVER("Write zpos for plane %i: %i\n", pplane->pos,
			 pstate->zpos);
	SET_FIELD(reg, PDP_GRPH1CTRL_GRPH1BLENDPOS, pstate->zpos);

	SET_FIELD(reg, PDP_GRPH1CTRL_GRPH1BLEND, alpha_mode);
	pdp_write(priv->dev, pplane->conf->regs.ctrl, reg);

	DRM_DEBUG_DRIVER("Setting gaplha for plane: %u\n", pstate->alpha);
	pdp_write(priv->dev, pplane->conf->regs.galpha, pstate->alpha);

	DRM_DEBUG_DRIVER("Setting Colourkey properties.");
	reg = pdp_read(priv->dev, pplane->conf->regs.ctrl);
	SET_FIELD(reg, PDP_GRPH1CTRL_GRPH1CKEYEN, pstate->colourkey_enable);
	SET_FIELD(reg, PDP_GRPH1CTRL_GRPH1CKEYSRC, pstate->colourkey_mode);
	pdp_write(priv->dev, pplane->conf->regs.ctrl, reg);

	reg = pdp_read(priv->dev, pplane->conf->regs.ckey_r);
	SET_FIELD(reg, PDP_GRPH1CKEY_R_GRPH1CKEY_R, pstate->colourkey_r);
	pdp_write(priv->dev, pplane->conf->regs.ckey_r, reg);

	reg = pdp_read(priv->dev, pplane->conf->regs.ckey_gb);
	SET_FIELD(reg, PDP_GRPH1CKEY_GB_GRPH1CKEY_G, pstate->colourkey_g);
	SET_FIELD(reg, PDP_GRPH1CKEY_GB_GRPH1CKEY_B, pstate->colourkey_b);
	pdp_write(priv->dev, pplane->conf->regs.ckey_gb, reg);

	reg = pdp_read(priv->dev, pplane->conf->regs.blend2_r);
	SET_FIELD(reg, PDP_GRPH1BLND2_R_GRPH1CKEYMASK_R, pstate->colourkey_r_mask);
	pdp_write(priv->dev, pplane->conf->regs.blend2_r, reg);

	reg = pdp_read(priv->dev, pplane->conf->regs.blend2_gb);
	SET_FIELD(reg, PDP_GRPH1BLND2_GB_GRPH1CKEYMASK_G, pstate->colourkey_g_mask);
	SET_FIELD(reg, PDP_GRPH1BLND2_GB_GRPH1CKEYMASK_B, pstate->colourkey_b_mask);
	pdp_write(priv->dev, pplane->conf->regs.blend2_gb, reg);

	reg = pdp_read(priv->dev, pplane->conf->regs.surf);
	SET_FIELD(reg, PDP_GRPH1SURF_GRPH1USEGAMMA, pstate->gamma_enable);
	pdp_write(priv->dev, pplane->conf->regs.surf, reg);
}

void pdp_disable_plane(struct pdp_plane *pplane)
{
	struct pdp_private *priv = pplane->dev;
	u32 reg = pdp_read(priv->dev, pplane->conf->regs.ctrl);

	pdp_plane_update_props(pplane);

	DRM_DEBUG_DRIVER("Disabling plane %u\n", pplane->pos);
	SET_FIELD(reg, PDP_GRPH1CTRL_GRPH1STREN, 0);
	pdp_write(priv->dev, pplane->conf->regs.ctrl, reg);
}

void pdp_configure_plane(struct pdp_plane *pplane)
{
	struct pdp_private *priv = pplane->dev;
	struct pdp_plane_regs *regs = &pplane->conf->regs;
	struct drm_plane_state *state = pplane->plane.state;
	u32 reg;

	DRM_DEBUG_DRIVER("Write position (%d,%d) for plane %i\n",
			 state->crtc_x, state->crtc_y, pplane->pos);
	reg = pdp_read(priv->dev, regs->posn);
	SET_FIELD(reg, PDP_GRPH1POSN_GRPH1XSTART, state->crtc_x);
	SET_FIELD(reg, PDP_GRPH1POSN_GRPH1YSTART, state->crtc_y);
	pdp_write(priv->dev, regs->posn, reg);

	DRM_DEBUG_DRIVER("Write source w.h (%d,%d)\n", state->src_w >> 16,
			 state->src_h >> 16);
	reg = pdp_read(priv->dev, regs->size);
	SET_FIELD(reg, PDP_GRPH1SIZE_GRPH1WIDTH, (state->src_w >> 16) - 1);
	SET_FIELD(reg, PDP_GRPH1SIZE_GRPH1HEIGHT, (state->src_h >> 16) - 1);
	pdp_write(priv->dev, regs->size, reg);

	DRM_DEBUG_DRIVER("Write Enable bit for plane %i\n", pplane->pos);
	reg = pdp_read(priv->dev, regs->ctrl);
	SET_FIELD(reg, PDP_GRPH1CTRL_GRPH1STREN, 1);
	pdp_write(priv->dev, regs->ctrl, reg);

	/*move this to atomic helper in pdp_planes.c?*/
	DRM_DEBUG_DRIVER("Updating blender config for plane %i\n",
			 pplane->pos);
	pdp_plane_update_blender(pplane);
}

