/* PDP Memory management
 *
 * Copyright (C) 2016 Imagination Technologies.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 *
 * The PDP has its own iommu in a single master-slave relationship
 * The iommu allows the PDP to scanout tiled buffers from the GPU and
 * other video encoding/decoding blocks.
 *
 * The PDP Memory manager manages the domains provided by the iommu.
 * The iommu supports 4 domains that can have different tiling properties
 *
 * The linear virtual address space presented to the PDP by the iommu is
 * managed by the generic drm_mm manager. This can operate on the complete
 * 32-bit virtual address space.  The 32-bit address space is split into
 * 4 1GB blocks that relate to an iommu_domain configuration. In most cases
 * there will not be over 1GB of buffers allocated at once so this suffices.
 *
 * A list is kept of the currently used domains. When a domain's last buffer
 * is freed the domain is unallocated allowing it to be re-used with a
 * different configuration.
 */

/* TODO
 * Use another list to hold unallocated domains would save some clutter
 * with the bitmap and re-getting the geometry. This would allow better
 * static configuration possibly from dts.
 *
 */

#include <linux/sizes.h>
#include <linux/iommu.h>
#include <linux/img-iommu.h>
#include "pdp_drv.h"
#include "pdp_tiler.h"

int pdp_tiler_init(struct pdp_private *priv, struct resource *cma_region)
{
	struct pdp_tiler *tiler = &priv->tiler;

	memset(tiler, 0, sizeof(*tiler));

	if (cma_region) {
		DRM_DEBUG_DRIVER("using carveout memory %pr\n", cma_region);
		drm_mm_init(&tiler->vram, cma_region->start,
			    resource_size(cma_region));
		tiler->domain_size = resource_size(cma_region) / 4;
		tiler->domain_base = cma_region->start;
		priv->carveout = true;
	} else {
		DRM_DEBUG_DRIVER("using page buffer allocation\n");
		drm_mm_init(&tiler->vram, 0, SZ_2G * 2 - 1);
		tiler->domain_size = SZ_1G;
		priv->carveout = false;
	}

	INIT_LIST_HEAD(&tiler->domains.list);

	return 0;
}

static struct pdp_tiler_domain *pdp_mmu_get_domain(struct pdp_private *priv)
{
	struct iommu_domain *domain;
	struct pdp_tiler_domain *pdp_domain;
	struct device *dev = priv->dev->dev;
	int err;

	pdp_domain = devm_kzalloc(dev, sizeof(*pdp_domain), GFP_KERNEL);
	if (!pdp_domain)
		return pdp_domain;

	domain = iommu_domain_alloc(&platform_bus_type);
	if (!domain) {
		devm_kfree(dev, pdp_domain);
		return ERR_PTR(-ENODEV);
	}

	err = iommu_attach_device(domain, priv->dev->dev);
	if (err) {
		iommu_domain_free(domain);
		devm_kfree(dev, pdp_domain);
		return ERR_PTR(err);
	}

	pdp_domain->domain = domain;

	return pdp_domain;
}

static void pdp_tiler_domain_remove(struct pdp_private *priv,
				    struct pdp_tiler_domain *domain)
{
	priv->tiler.domain_bitmap &= ~BIT(domain->id);
	list_del(&domain->list);
}

static void pdp_tiler_domain_remove_buf(struct pdp_private *priv,
					struct pdp_tiler_domain *domain)
{
	domain->nbuffs--;

	if (domain->nbuffs == 0) {
		iommu_detach_device(domain->domain, priv->dev->dev);
		iommu_domain_free(domain->domain);
		pdp_tiler_domain_remove(priv, domain);
	}
}

static bool pdp_tiler_domain_equal(struct pdp_tiler_domain *domain,
				   struct img_iommu_tile_cfg *conf)
{
	return (domain->config.enable == conf->enable) &&
	       (domain->config.stride == conf->stride);
}

static void pdp_tiler_domain_add(struct pdp_private *priv,
				 struct pdp_tiler_domain *domain)
{
	struct pdp_tiler *tiler = &priv->tiler;

	list_add(&domain->list, &tiler->domains.list);
}

static struct pdp_tiler_domain *
pdp_tiler_domain_for_conf(struct pdp_private *priv,
			  struct img_iommu_tile_cfg *conf, bool force)
{
	struct pdp_tiler *tiler = &priv->tiler;
	struct list_head *pos;
	struct pdp_tiler_domain *domain;
	struct iommu_domain_geometry geom;
	size_t size = tiler->domain_size;

	DRM_DEBUG_DRIVER("Allocating IOMMU domain prop:%u %u\n",
			 conf->enable, conf->stride);

	/* Check to see if we have a valid region already */
	if (!force)
		list_for_each(pos, &tiler->domains.list) {
			domain = list_entry(pos, struct pdp_tiler_domain, list);

			if (pdp_tiler_domain_equal(domain, conf))
				return domain;
		}
	/* Here we dont have a valid region so make one */
	domain = pdp_mmu_get_domain(priv);
	if (IS_ERR(domain))
		return domain;
	/* add domain to list */
	pdp_tiler_domain_add(priv, domain);

	/* Setup domain parameters */
	domain->id = ffz(priv->tiler.domain_bitmap);

	tiler->domain_bitmap |= BIT(domain->id);

	geom.aperture_start = tiler->domain_base + (domain->id) * size;
	geom.aperture_end = tiler->domain_base + ((domain->id + 1) * size) - 1;

	iommu_domain_set_attr(domain->domain, DOMAIN_ATTR_GEOMETRY, &geom);
	iommu_domain_set_attr(domain->domain, DOMAIN_ATTR_TILED, conf);
	img_iommu_set_bypass(domain->domain, false);

	domain->config.enable = conf->enable;
	domain->config.stride = conf->stride;

	return domain;
}

void pdp_tiler_set_tile_mode(struct pdp_private *priv, int mode)
{
	struct pdp_tiler *tiler = &priv->tiler;
	struct list_head *pos;
	struct pdp_tiler_domain *domain;

	DRM_DEBUG_DRIVER("Set tiling scheme %s\n",
			 mode == TILE_256x16 ? "256x16"  : "512x8");

	list_for_each(pos, &tiler->domains.list) {
		domain = list_entry(pos, struct pdp_tiler_domain, list);
		if (domain) {
			img_iommu_set_tiling_scheme(domain->domain, mode);
			tiler->tile_mode = mode;
			break;
		}
	}
}

int pdp_tiler_iommu_unmap(struct pdp_private *priv,
			  struct pdp_gem_object *pobj)
{
	iommu_unmap(pobj->domain->domain, pobj->paddr, pobj->size);
	pdp_tiler_domain_remove_buf(priv, pobj->domain);

	return 0;
}

int pdp_tiler_iommu_map(struct pdp_private *priv, struct pdp_gem_object *pobj)
{
	struct pdp_tiler *tiler = &priv->tiler;
	ssize_t err;
	struct pdp_tiler_domain *domain;
	struct img_iommu_tile_cfg tile;

	if (pobj->mm)
		return -EBUSY;

	pobj->mm = kzalloc(sizeof(*pobj->mm), GFP_KERNEL);
	if (!pobj->mm)
		return -ENOMEM;

	/* grab domain to map into */
	tile.enable = pobj->tiled;
	tile.stride = pobj->tile_stride;

	domain = pdp_tiler_domain_for_conf(priv, &tile, false);
	if (!domain)
		return PTR_ERR(domain);

	/* grab a suitable virtual addr from vram allocator */
	mutex_lock(&priv->dev->struct_mutex);
	err = drm_mm_insert_node_in_range_generic(&tiler->vram, pobj->mm,
			pobj->obj.size, PAGE_SIZE, 0,
			domain->domain->geometry.aperture_start,
			domain->domain->geometry.aperture_end, 0, 0);
	if (err < 0) {
		/* There was not enough memory in first domain, try another.
		 * If this fails the system truly is out of memory
		 */
		domain = pdp_tiler_domain_for_conf(priv, &tile, true);
		if (!domain)
			return PTR_ERR(domain);
		err = drm_mm_insert_node_in_range_generic(&tiler->vram,
			pobj->mm, pobj->obj.size, PAGE_SIZE, 0,
			domain->domain->geometry.aperture_start,
			domain->domain->geometry.aperture_end, 0, 0);
		if (err < 0) {
			dev_err(priv->dev->dev, "out of IO virtual memory: %zd\n",
				err);
			goto free;
		}
	}
	domain->nbuffs++;
	pobj->domain = domain;
	pobj->paddr = pobj->mm->start;

	if (!priv->carveout) {
		err = iommu_map_sg(pobj->domain->domain, pobj->paddr,
				   pobj->sgt->sgl, pobj->sgt->nents,
				   IOMMU_READ);
		if (err < 0) {
			dev_err(priv->dev->dev, "failed to map buffer: %zd\n",
				err);
			goto remove;
		}
		pobj->size = err;
	} else {
		img_iommu_set_bypass(pobj->domain->domain, true);
	}

	mutex_unlock(&priv->dev->struct_mutex);

	return 0;
remove:
	drm_mm_remove_node(pobj->mm);
free:
	pdp_tiler_domain_remove_buf(priv, pobj->domain);
	mutex_unlock(&priv->dev->struct_mutex);
	kfree(pobj->mm);
	return err;
}

void pdp_tiler_fix_faults(struct pdp_private *priv)
{
	struct pdp_tiler *tiler = &priv->tiler;
	struct pdp_tiler_domain *domain;
	struct list_head *pos;

	list_for_each(pos, &tiler->domains.list) {
		domain = list_entry(pos, struct pdp_tiler_domain, list);
		if (img_iommu_has_pagefault(domain->domain)) {
			img_iommu_clear_pagefault(domain->domain);
			return;
		}
	}
}
