/*
 * Copyright (C) 2016 Imagination Technologies
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/backlight.h>
#include <linux/gpio/consumer.h>
#include <linux/module.h>
#include <linux/of.h>
#include <linux/regulator/consumer.h>

#include <drm/drmP.h>
#include <drm/drm_crtc.h>
#include <drm/drm_mipi_dsi.h>
#include <drm/drm_panel.h>

#include <video/mipi_display.h>

struct sharp_ls_panel {
	struct drm_panel base;
	struct mipi_dsi_device *dsi;

	struct backlight_device *backlight;
	struct regulator *supply;
	struct gpio_desc *reset_gpio;

	bool prepared;
	bool enabled;

	const struct drm_display_mode *mode;
};

static inline struct sharp_ls_panel *to_sharp_ls_panel(struct drm_panel *panel)
{
	return container_of(panel, struct sharp_ls_panel, base);
}

static void
sharp_ls_wait_frames(struct sharp_ls_panel *sharp, unsigned int frames)
{
	unsigned int refresh = drm_mode_vrefresh(sharp->mode);

	if (WARN_ON(frames > refresh))
		return;

	msleep(1000 / (refresh / frames));
}

static int sharp_ls_panel_disable(struct drm_panel *panel)
{
	struct sharp_ls_panel *sharp = to_sharp_ls_panel(panel);

	if (!sharp->enabled)
		return 0;

	if (sharp->backlight) {
		sharp->backlight->props.power = FB_BLANK_POWERDOWN;
		backlight_update_status(sharp->backlight);
	};

	sharp->enabled = false;

	return 0;
}

static int sharp_ls_panel_unprepare(struct drm_panel *panel)
{
	struct sharp_ls_panel *sharp = to_sharp_ls_panel(panel);
	int err;

	if (!sharp->prepared)
		return 0;

	err = mipi_dsi_dcs_set_display_off(sharp->dsi);
	if (err < 0)
		dev_err(panel->dev, "failed to set display off: %d\n", err);

	sharp_ls_wait_frames(sharp, 1);

	err = mipi_dsi_dcs_enter_sleep_mode(sharp->dsi);
	if (err < 0)
		dev_err(panel->dev, "failed to enter sleep mode: %d\n", err);

	sharp_ls_wait_frames(sharp, 4);

	gpiod_set_value(sharp->reset_gpio, 0);
	regulator_disable(sharp->supply);

	sharp->prepared = false;

	return 0;
}

/*
 * Power ON Sequence:
 * Reset Low
 * Regulator ON
 * Assert Reset high
 * Wait 10ms
 * Sleep out	0x11
 * Wait 6 frames
 * Display on
 * Backlight on
 */
static int sharp_ls_panel_prepare(struct drm_panel *panel)
{
	struct sharp_ls_panel *sharp = to_sharp_ls_panel(panel);
	u8 format = MIPI_DCS_PIXEL_FMT_24BIT;
	int err;

	if (sharp->prepared)
		return 0;

	gpiod_set_value(sharp->reset_gpio, 0);

	err = regulator_enable(sharp->supply);
	if (err < 0)
		return err;

	msleep(150);

	gpiod_set_value(sharp->reset_gpio, 1);

	msleep(20);

	err = mipi_dsi_dcs_exit_sleep_mode(sharp->dsi);
	if (err < 0) {
		dev_err(panel->dev, "failed to exit sleep mode %d\n", err);
		goto poweroff;
	}

	err = mipi_dsi_dcs_set_pixel_format(sharp->dsi, format);
	if (err < 0) {
		dev_err(panel->dev, "failed to set pixel format: %d\n", err);
		goto poweroff;
	}

	sharp->prepared = true;

	return 0;

poweroff:
	regulator_disable(sharp->supply);
	return err;
}

static int sharp_ls_panel_enable(struct drm_panel *panel)
{
	int err;
	struct sharp_ls_panel *sharp = to_sharp_ls_panel(panel);

	if (sharp->enabled)
		return 0;

	sharp_ls_wait_frames(sharp, 6);

	err = mipi_dsi_dcs_set_display_on(sharp->dsi);
	if (err < 0) {
		dev_err(panel->dev, "failed to set display on %d\n", err);
		return err;
	}

	if (sharp->backlight) {
		sharp->backlight->props.power = FB_BLANK_UNBLANK;
		backlight_update_status(sharp->backlight);
	}

	sharp->enabled = true;

	return 0;
}

static const struct drm_display_mode panel_modes[] = {
	{
		.clock = 148500,
		.hdisplay = 1080,
		.vdisplay = 1920,
		.vrefresh = 60,

		.hsync_start = 1250 - 50 - 10, /* sync width 10 */
		.hsync_end = 1250 - 50, /* back porch 50 */
		.htotal = 1250,
		.vsync_start = 1980 - 4 - 2, /* sync width 2 */
		.vsync_end = 1980 - 4, /* back porch 4 */
		.vtotal = 1980,
	},
	{
		.clock = 148500,
		.hdisplay = 1080,
		.vdisplay = 1920,
		.vrefresh = 50,

		.hsync_start = 1320 - 40 - 10, /* sync width 10 */
		.hsync_end = 1320 - 40, /* back porch 40 */
		.htotal = 1320,
		.vsync_start = 2250 - 4 - 2, /* sync width 2 */
		.vsync_end = 2250 - 4, /* back porch 4 */
		.vtotal = 2250,
	},
};


static int sharp_ls_panel_get_modes(struct drm_panel *panel)
{
	int i, mode_count = 0;
	struct drm_display_mode *mode;

	for (i = 0; i < ARRAY_SIZE(panel_modes); i++) {
		const struct drm_display_mode *m = &panel_modes[i];

		mode = drm_mode_duplicate(panel->drm, m);
		if (!mode) {
			dev_err(panel->drm->dev, "failed to add mode %ux%u@%u\n",
				m->hdisplay, m->vdisplay, m->vrefresh);
			continue;
		}

		mode_count++;

		drm_mode_set_name(mode);

		drm_mode_probed_add(panel->connector, mode);
	}

	panel->connector->display_info.width_mm = 62;
	panel->connector->display_info.height_mm = 110;

	return mode_count;
}

static const struct drm_panel_funcs sharp_ls_panel_funcs = {
	.disable = sharp_ls_panel_disable,
	.unprepare = sharp_ls_panel_unprepare,
	.prepare = sharp_ls_panel_prepare,
	.enable = sharp_ls_panel_enable,
	.get_modes = sharp_ls_panel_get_modes,
};

static const struct of_device_id sharp_ls_of_match[] = {
	{ .compatible = "sharp,ls050t1sx06", },
	{ }
};
MODULE_DEVICE_TABLE(of, sharp_ls_of_match);

static int sharp_ls_panel_add(struct sharp_ls_panel *sharp)
{
	struct device *dev = &sharp->dsi->dev;
	struct device_node *np;
	int err;

	sharp->mode = &panel_modes[0];

	sharp->supply = devm_regulator_get(dev, "power");
	if (IS_ERR(sharp->supply)) {
		dev_err(dev, "cannot get power regulator %ld\n",
			PTR_ERR(sharp->supply));
		return PTR_ERR(sharp->supply);
	}

	sharp->reset_gpio = devm_gpiod_get(dev, "reset", GPIOD_OUT_LOW);
	if (IS_ERR(sharp->reset_gpio)) {
		dev_err(dev, "cannot get reset-gpios %ld\n",
			PTR_ERR(sharp->reset_gpio));
		return PTR_ERR(sharp->reset_gpio);
	}

	gpiod_set_value(sharp->reset_gpio, 0);

	np = of_parse_phandle(dev->of_node, "backlight", 0);
	if (np) {
		sharp->backlight = of_find_backlight_by_node(np);
		of_node_put(np);

		if (!sharp->backlight)
			return -EPROBE_DEFER;
	}

	drm_panel_init(&sharp->base);
	sharp->base.funcs = &sharp_ls_panel_funcs;
	sharp->base.dev = &sharp->dsi->dev;

	err = drm_panel_add(&sharp->base);
	if (err < 0)
		goto put_backlight;

	return 0;

put_backlight:
	if (sharp->backlight)
		put_device(&sharp->backlight->dev);

	return err;
}

static void sharp_ls_panel_del(struct sharp_ls_panel *sharp)
{
	if (sharp->base.dev)
		drm_panel_remove(&sharp->base);

	if (sharp->backlight)
		put_device(&sharp->backlight->dev);
}

static int sharp_ls_panel_probe(struct mipi_dsi_device *dsi)
{
	struct sharp_ls_panel *sharp;
	int err;

	dsi->lanes = 4;
	dsi->format = MIPI_DSI_FMT_RGB888;
	dsi->mode_flags = MIPI_DSI_MODE_VIDEO |
			  MIPI_DSI_MODE_VIDEO_BURST |
			  MIPI_DSI_MODE_VIDEO_SYNC_PULSE;

	sharp = devm_kzalloc(&dsi->dev, sizeof(*sharp), GFP_KERNEL);
	if (!sharp)
		return -ENOMEM;

	mipi_dsi_set_drvdata(dsi, sharp);

	sharp->dsi = dsi;

	err = sharp_ls_panel_add(sharp);
	if (err < 0)
		return err;

	return mipi_dsi_attach(dsi);
}

static int sharp_ls_panel_remove(struct mipi_dsi_device *dsi)
{
	struct sharp_ls_panel *sharp = mipi_dsi_get_drvdata(dsi);
	int err;

	err = sharp_ls_panel_disable(&sharp->base);
	if (err < 0)
		dev_err(&dsi->dev, "failed to disable panel: %d\n", err);

	err = mipi_dsi_detach(dsi);
	if (err < 0)
		dev_err(&dsi->dev, "failed to detach from DSI host: %d\n", err);

	drm_panel_detach(&sharp->base);
	sharp_ls_panel_del(sharp);

	return 0;
}

static void sharp_ls_panel_shutdown(struct mipi_dsi_device *dsi)
{
	struct sharp_ls_panel *sharp = mipi_dsi_get_drvdata(dsi);

	sharp_ls_panel_disable(&sharp->base);
}

static struct mipi_dsi_driver sharp_ls_panel_driver = {
	.driver = {
		.name = "panel-sharp-ls050t1sx06",
		.of_match_table = sharp_ls_of_match,
	},
	.probe = sharp_ls_panel_probe,
	.remove = sharp_ls_panel_remove,
	.shutdown = sharp_ls_panel_shutdown,
};
module_mipi_dsi_driver(sharp_ls_panel_driver);

MODULE_AUTHOR("Will Thomas <will.thomas@igmtec.com>");
MODULE_DESCRIPTION("Sharp LS050T1SX06 panel driver");
MODULE_LICENSE("GPL v2");
