/*
 * Imagination Technologies Temperature Sensor Driver
 *
 * Copyright (C) 2016 Imagination Technologies Ltd.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 */

#include <linux/clk.h>
#include <linux/device.h>
#include <linux/err.h>
#include <linux/hwmon.h>
#include <linux/hwmon-sysfs.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/of.h>
#include <linux/platform_device.h>
#include <linux/pm_runtime.h>

#define CR_SNR_RESET			0x00
#define CR_SNR_RESET_REQUEST		BIT(0)

#define	CR_SNR_CONTROL			0x04
#define CR_SNR_CONTROL_READ		BIT(0)
#define CR_SNR_CONTROL_SLEEP		BIT(1)

#define CR_SNR_MAXVAL			0x08
#define CR_SNR_MAXVAL_MASK		0xFFF

#define CR_SNR_MINVAL			0x0C
#define CR_SNR_MINVAL_MASK		0xFFF

#define CR_SNR_RESPONSE			0x10
#define CR_SNR_RESPONSE_READ_MASK	0xFFF
#define CR_SNR_RESPONSE_VALID		BIT(16)
#define CR_SNR_RESPONSE_TARGET		BIT(28)
#define CR_SNR_RESPONSE_LOWALARM	BIT(29)
#define CR_SNR_RESPONSE_LOWALARM_SHIFT	29
#define CR_SNR_RESPONSE_HIGHALARM	BIT(30)
#define CR_SNR_RESPONSE_HIGHALARM_SHIFT	30

#define CR_SNR_EVENT_ENAB		0x14
#define CR_SNR_EVENT_RNAB_UPDATE	BIT(0)
#define CR_SNR_EVENT_ENAB_LOWALARM	BIT(1)
#define CR_SNR_EVENT_ENAB_HIGHALARM	BIT(2)

#define CR_SNR_TRIM			0x18

#define RESPONSE_TIMEOUT_MSEC		5000
#define REG_TO_CELSIUS(reg)		((((int)(reg) * 859) - 1090000) / 10000)
#define CELSIUS_TO_REG(temp)		((((temp) + 109) / 352) * 4096)
#define MAX_SAMPLE_VAL			4095
#define MAX_DEGREE_CELSIUS		REG_TO_CELSIUS(MAX_SAMPLE_VAL)
#define MIN_DEGREE_CELSIUS		REG_TO_CELSIUS(0)

#define IMG_TEMP_PM_TIMEOUT		5000 /* ms */

struct img_temp {
	void __iomem *regs;
	struct device *dev;
	struct device *hwmon_dev;
	struct clk *sys_clk;
	struct clk *snr_clk;
};

static inline u32 img_temp_readl(struct img_temp *hwmon, u32 reg)
{
	return readl(hwmon->regs + reg);
}

static inline void img_temp_writel(struct img_temp *hwmon, u32 val,
					u32 reg)
{
	writel(val, hwmon->regs + reg);
}

static inline void img_temp_reset(struct img_temp *hwmon)
{
	pm_runtime_get_sync(hwmon->dev);

	img_temp_writel(hwmon, CR_SNR_RESET_REQUEST, CR_SNR_RESET);
	img_temp_writel(hwmon, 0, CR_SNR_RESET);

	pm_runtime_mark_last_busy(hwmon->dev);
	pm_runtime_put_autosuspend(hwmon->dev);
}

static ssize_t img_temp_show_name(struct device *dev,
			 struct device_attribute *devattr, char *buf)
{
	return sprintf(buf, "img_temp_sensor\n");
}

static ssize_t img_temp_show_temp(struct device *dev,
			 struct device_attribute *devattr, char *buf)
{
	struct img_temp *hwmon = dev_get_drvdata(dev);
	unsigned long timeout;
	int temp;
	u32 sample_val;
	u32 valid;
	int ret;

	ret = pm_runtime_get_sync(dev);
	if (ret < 0)
		return ret;

	/* Request sample and set device back to sleep after reading */
	img_temp_writel(hwmon, CR_SNR_CONTROL_SLEEP | CR_SNR_CONTROL_READ,
			CR_SNR_CONTROL);

	timeout = jiffies + msecs_to_jiffies(RESPONSE_TIMEOUT_MSEC);

	/* Wait for valid sample */
	do {
		valid = CR_SNR_RESPONSE_VALID
			& img_temp_readl(hwmon, CR_SNR_RESPONSE);
		cpu_relax();
	} while ((!valid) && time_before(jiffies, timeout));

	if (!time_before(jiffies, timeout)) {
		dev_err(dev, "Timed out waiting for valid reading.\n");
		return -ETIMEDOUT;
	}

	sample_val = CR_SNR_RESPONSE_READ_MASK &
			img_temp_readl(hwmon, CR_SNR_RESPONSE);

	pm_runtime_mark_last_busy(dev);
	pm_runtime_put_autosuspend(dev);

	/* Convert sample to millidegree Celsius */
	temp = REG_TO_CELSIUS(sample_val);
	return sprintf(buf, "%d\n", temp);
}

/* Sysfs attributes */
static DEVICE_ATTR(name, S_IRUGO, img_temp_show_name, NULL);
static SENSOR_DEVICE_ATTR(temp1_input, S_IRUGO,
				img_temp_show_temp, NULL, 0);
static struct attribute *img_temp_attrs[] = {
	&dev_attr_name.attr,
	&sensor_dev_attr_temp1_input.dev_attr.attr,
	NULL,
};

static const struct attribute_group img_temp_attr_group = {
	.name = "img_temp",
	.attrs = img_temp_attrs,
};

static int img_temp_runtime_suspend(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);
	struct img_temp *hwmon = platform_get_drvdata(pdev);

	clk_disable_unprepare(hwmon->snr_clk);
	clk_disable_unprepare(hwmon->sys_clk);

	return 0;
}

static int img_temp_runtime_resume(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);
	struct img_temp *hwmon = platform_get_drvdata(pdev);
	int ret;

	ret = clk_prepare_enable(hwmon->sys_clk);
	if (ret)
		return ret;

	ret = clk_prepare_enable(hwmon->snr_clk);
	if (ret) {
		clk_disable_unprepare(hwmon->sys_clk);
		return ret;
	}

	return 0;
}

#ifdef CONFIG_PM_SLEEP
static int img_temp_suspend(struct device *dev)
{
	return pm_runtime_force_suspend(dev);
}

static int img_temp_resume(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);
	struct img_temp *hwmon = platform_get_drvdata(pdev);
	int ret;

	ret = pm_runtime_force_resume(dev);
	if (ret)
		return ret;

	img_temp_reset(hwmon);

	return 0;
}
#endif /* CONFIG_PM_SLEEP */

static int img_temp_probe(struct platform_device *pdev)
{
	int ret;
	struct img_temp *hwmon;
	struct resource *mem;

	hwmon = devm_kzalloc(&pdev->dev, sizeof(*hwmon), GFP_KERNEL);
	if (!hwmon)
		return -ENOMEM;
	hwmon->dev = &pdev->dev;
	mem = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	hwmon->regs = devm_ioremap_resource(&pdev->dev, mem);
	if (IS_ERR(hwmon->regs))
		return PTR_ERR(hwmon->regs);

	hwmon->sys_clk = devm_clk_get(&pdev->dev, "sys_clk");
	if (IS_ERR(hwmon->sys_clk)) {
		dev_err(&pdev->dev, "Failed to get sys_clock.\n");
		return PTR_ERR(hwmon->sys_clk);
	}
	hwmon->snr_clk = devm_clk_get(&pdev->dev, "snr_clk");
	if (IS_ERR(hwmon->snr_clk)) {
		dev_err(&pdev->dev, "Failed to get snr_clock.\n");
		return PTR_ERR(hwmon->snr_clk);
	}

	platform_set_drvdata(pdev, hwmon);

	pm_runtime_set_autosuspend_delay(&pdev->dev, IMG_TEMP_PM_TIMEOUT);
	pm_runtime_use_autosuspend(&pdev->dev);
	pm_runtime_enable(&pdev->dev);
	if (!pm_runtime_enabled(&pdev->dev)) {
		ret = img_temp_runtime_resume(&pdev->dev);
		if (ret)
			return ret;
	}

	ret = sysfs_create_group(&pdev->dev.kobj, &img_temp_attr_group);
	if (ret) {
		dev_err(&pdev->dev, "Failed to create sysfs group: %d.\n", ret);
		goto rpm_disable;
	}

	hwmon->hwmon_dev = hwmon_device_register(&pdev->dev);
	if (IS_ERR(hwmon->hwmon_dev)) {
		dev_err(&pdev->dev, "Failed to register device.\n");
		ret = PTR_ERR(hwmon->hwmon_dev);
		goto remove_sysfs;
	}
	img_temp_reset(hwmon);

	return 0;

remove_sysfs:
	sysfs_remove_group(&pdev->dev.kobj, &img_temp_attr_group);
rpm_disable:
	if (!pm_runtime_enabled(&pdev->dev))
		img_temp_runtime_suspend(&pdev->dev);
	pm_runtime_disable(&pdev->dev);
	pm_runtime_dont_use_autosuspend(&pdev->dev);

	return ret;
}

static int img_temp_remove(struct platform_device *pdev)
{
	struct img_temp *hwmon = platform_get_drvdata(pdev);

	hwmon_device_unregister(hwmon->hwmon_dev);
	sysfs_remove_group(&pdev->dev.kobj, &img_temp_attr_group);
	pm_runtime_disable(&pdev->dev);
	if (!pm_runtime_status_suspended(&pdev->dev))
		img_temp_runtime_suspend(&pdev->dev);

	return 0;
}

static const struct dev_pm_ops img_temp_pm_ops = {
	SET_RUNTIME_PM_OPS(img_temp_runtime_suspend,
			   img_temp_runtime_resume,
			   NULL)
	SET_SYSTEM_SLEEP_PM_OPS(img_temp_suspend, img_temp_resume)
};

static const struct of_device_id img_temp_of_match[] = {
	{ .compatible = "img,temp", },
	{ },
};
MODULE_DEVICE_TABLE(of, img_temp_of_match);

static struct platform_driver img_temp_driver = {
	.probe	= img_temp_probe,
	.remove = img_temp_remove,
	.driver = {
		.name = "img-temp",
		.pm = &img_temp_pm_ops,
		.of_match_table = of_match_ptr(img_temp_of_match),
	},
};
module_platform_driver(img_temp_driver);

MODULE_DESCRIPTION("IMG Temperature Sensor Driver");
MODULE_AUTHOR("Seraj Mohammed <seraj.mohammed@imgtec.com>");
MODULE_LICENSE("GPL V2");
