/*
 * Imagination Voltage Monitor driver.
 *
 * Copyright (C) 2016 Imagination Technologies Ltd.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 */

#include <linux/clk.h>
#include <linux/device.h>
#include <linux/err.h>
#include <linux/hwmon.h>
#include <linux/hwmon-sysfs.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/of.h>
#include <linux/platform_device.h>
#include <linux/pm_runtime.h>
#include <linux/spinlock.h>

#define CR_SNR_RESET			0x00
#define CR_SNR_RESET_REQUEST		BIT(0)

#define	CR_SNR_CONTROL			0x04
#define CR_SNR_CONTROL_READ		BIT(0)
#define CR_SNR_CONTROL_SLEEP		BIT(1)

#define CR_SNR_MAXVAL			0x08
#define CR_SNR_MAXVAL_MASK		0xFF

#define CR_SNR_MINVAL			0x0C
#define CR_SNR_MINVAL_MASK		0xFF

#define CR_SNR_RESPONSE			0x10
#define CR_SNR_RESPONSE_READ_MASK	0xFF
#define CR_SNR_RESPONSE_VALID		BIT(16)
#define CR_SNR_RESPONSE_TARGET		BIT(28)
#define CR_SNR_RESPONSE_LOWALARM	BIT(29)
#define CR_SNR_RESPONSE_LOWALARM_SHIFT	29
#define CR_SNR_RESPONSE_HIGHALARM	BIT(30)
#define CR_SNR_RESPONSE_HIGHALARM_SHIFT	30

#define CR_SNR_EVENT_ENAB		0x14
#define CR_SNR_EVENT_RNAB_UPDATE	BIT(0)
#define CR_SNR_EVENT_ENAB_LOWALARM	BIT(1)
#define CR_SNR_EVENT_ENAB_HIGHALARM	BIT(2)

#define CR_SNR_TRIM			0x18

#define CR_SNR_SEL_VIN			0x1C
#define CR_SNR_SEL_VIN_MASK		0xF

#define CR_SNR_REV			0x3C
#define CR_SNR_MAINT_REV_MASK		0xFF
#define CR_SNR_MAINT_REV_SHIFT		0
#define CR_SNR_MINOR_REV_MASK		0xFF00
#define CR_SNR_MINOR_REV_SHIFT		8
#define CR_SNR_MAJOR_REV_MASK		0xFF0000
#define CR_SNR_MAJOR_REV_SHIFT		16
#define CR_SNR_DESIGNER_REV_MASK	0xFF000000
#define CR_SNR_DESIGNER_REV_SHIFT	24

#define NUM_OF_MONITOR_POINTS		9
#define RESPONSE_TIMEOUT_MSEC		5000
#define MAX_SAMPLE_VAL			255
#define MAX_VOLTAGE			REG_TO_MV(MAX_SAMPLE_VAL)
#define MIN_VOLTAGE			REG_TO_MV(0)
#define UPDATE_AVERAGE(source)		\
	source.avg = ((source.max + source.min) / 2)
/*
* The equation retrieved from the data sheet is : Voltage = (NX) / 256
* where N is the sampled value and X is a technology dependant coefficient.
* for the MR74070 Voltage monitor, X = 1.2, Therefore the equation for
* retrieving the voltage in volts is V = ((N*1.2) / 256) * 1000
* the result is also multiplied by 1000 to convert to millivolt unit as required
* by the Sysfs API. to avoid performing fpu operations, the equation was
* converted to:
* mV = ((reg * (1.2/256 * 10,000,000)) / 10,000,000) * 1,000
* when minimised:
* mV = (reg*46875)/10,000
*/
#define REG_TO_MV(reg)			(((reg) * 46875) / 10000)
/*
* Similar to the previous equation, the equation used to retrieved the register
* value from a mV value is reg = (mV / 1000) * 256/1.2
* which is minimised to
* reg = (((mV) * 213) / 1000)
*/
#define MV_TO_REG(mV)			(((mV) * 213) / 1000)

#define IMG_VMON_PM_TIMEOUT		5000 /* ms */

/*order of labels depends on mapping*/
static const char * const monitor_labels[] = {
	"ALWAYS_ON_DOMAIN",
	"CPU_0",
	"CPU_1",
	"GPU",
	"VXD",
	"VXE",
	"ISP",
	"RPU",
	"SWITCHED_TOP",
};

struct img_vmon {
	void __iomem *regs;
	struct device *dev;
	struct device *hwmon_dev;
	struct clk *sys_clk;
	struct clk *snr_clk;
	spinlock_t lock;
};

static inline u32 img_vmon_readl(struct img_vmon *hwmon, u32 reg)
{
	return readl(hwmon->regs + reg);
}

static inline void img_vmon_writel(struct img_vmon *hwmon, u32 val,
					u32 reg)
{
	writel(val, hwmon->regs + reg);
}

static inline void img_vmon_reset(struct img_vmon *hwmon)
{
	pm_runtime_get_sync(hwmon->dev);

	img_vmon_writel(hwmon, CR_SNR_RESET_REQUEST, CR_SNR_RESET);
	img_vmon_writel(hwmon, 0, CR_SNR_RESET);

	pm_runtime_mark_last_busy(hwmon->dev);
	pm_runtime_put_autosuspend(hwmon->dev);
}

static ssize_t img_vmon_show_name(struct device *dev,
			 struct device_attribute *devattr, char *buf)
{
	return sprintf(buf, "img_voltage_monitor\n");
}

static ssize_t img_vmon_show_version(struct device *dev,
			 struct device_attribute *devattr, char *buf)
{
	struct img_vmon *hwmon = dev_get_drvdata(dev);
	u32 reg_val;
	int ret;

	ret = pm_runtime_get_sync(dev);
	if (ret < 0)
		return ret;

	reg_val = img_vmon_readl(hwmon, CR_SNR_REV);

	pm_runtime_mark_last_busy(dev);
	pm_runtime_put_autosuspend(dev);

	return sprintf(buf, "%u.%u.%u.%u\n",
	(reg_val & CR_SNR_MAINT_REV_MASK) >> CR_SNR_MAINT_REV_SHIFT,
	(reg_val & CR_SNR_MINOR_REV_MASK) >> CR_SNR_MINOR_REV_SHIFT,
	(reg_val & CR_SNR_MAJOR_REV_MASK) >> CR_SNR_MAJOR_REV_SHIFT,
	(reg_val & CR_SNR_DESIGNER_REV_MASK) >> CR_SNR_DESIGNER_REV_SHIFT);

}

static ssize_t img_vmon_show_label(struct device *dev,
			 struct device_attribute *devattr, char *buf)
{
	return sprintf(buf, "%s\n",
		monitor_labels[to_sensor_dev_attr(devattr)->index]);
}


static ssize_t img_vmon_show_in(struct device *dev,
			 struct device_attribute *devattr, char *buf)
{
	struct img_vmon *hwmon = dev_get_drvdata(dev);
	unsigned long timeout;
	long voltage;
	u32 sample_val;
	u32 valid;
	u32 reg_val;
	u32 source_index = to_sensor_dev_attr(devattr)->index;
	u32 sel_vin = source_index == 8 ? 0 : source_index + 8;
	int ret;

	ret = pm_runtime_get_sync(dev);
	if (ret < 0)
		return ret;

	spin_lock(&hwmon->lock);
	/* change voltage monitor source */
	img_vmon_writel(hwmon, sel_vin & CR_SNR_SEL_VIN_MASK,
				CR_SNR_SEL_VIN);
	/* Request sample and set device back to sleep after reading */
	img_vmon_writel(hwmon, CR_SNR_CONTROL_SLEEP | CR_SNR_CONTROL_READ,
			CR_SNR_CONTROL);

	timeout = jiffies + msecs_to_jiffies(RESPONSE_TIMEOUT_MSEC);

	/* Wait for valid sample */
	do {
		valid = CR_SNR_RESPONSE_VALID
			& img_vmon_readl(hwmon, CR_SNR_RESPONSE);
		cpu_relax();
	} while ((!valid) && time_before(jiffies, timeout));

	if (!time_before(jiffies, timeout)) {
		dev_err(dev, "Timed out waiting for valid reading.\n");
		return -ETIMEDOUT;
	}

	reg_val = img_vmon_readl(hwmon, CR_SNR_RESPONSE);
	spin_unlock(&hwmon->lock);
	sample_val = reg_val & CR_SNR_RESPONSE_READ_MASK;

	pm_runtime_mark_last_busy(dev);
	pm_runtime_put_autosuspend(dev);

	/* Convert sample to millivolts */
	voltage = REG_TO_MV(sample_val);
	return sprintf(buf, "%ld\n", voltage);
}

/* Sysfs attributes */
static DEVICE_ATTR(name, S_IRUGO, img_vmon_show_name, NULL);
static DEVICE_ATTR(vrm, S_IRUGO, img_vmon_show_version, NULL);

#define REGISTER_MONITOR_SOURCE(x)					\
static SENSOR_DEVICE_ATTR(in##x##_label, S_IRUGO,			\
				img_vmon_show_label, NULL, x);		\
static SENSOR_DEVICE_ATTR(in##x##_input, S_IRUGO, img_vmon_show_in, NULL, x)

REGISTER_MONITOR_SOURCE(0);
REGISTER_MONITOR_SOURCE(1);
REGISTER_MONITOR_SOURCE(2);
REGISTER_MONITOR_SOURCE(3);
REGISTER_MONITOR_SOURCE(4);
REGISTER_MONITOR_SOURCE(5);
REGISTER_MONITOR_SOURCE(6);
REGISTER_MONITOR_SOURCE(7);
REGISTER_MONITOR_SOURCE(8);

#define RETRIEVE_ATTRS_SET(x) \
&sensor_dev_attr_in##x##_label.dev_attr.attr,\
&sensor_dev_attr_in##x##_input.dev_attr.attr

static struct attribute *img_vmon_attrs[] = {
	&dev_attr_name.attr,
	&dev_attr_vrm.attr,
	RETRIEVE_ATTRS_SET(0),
	RETRIEVE_ATTRS_SET(1),
	RETRIEVE_ATTRS_SET(2),
	RETRIEVE_ATTRS_SET(3),
	RETRIEVE_ATTRS_SET(4),
	RETRIEVE_ATTRS_SET(5),
	RETRIEVE_ATTRS_SET(6),
	RETRIEVE_ATTRS_SET(7),
	RETRIEVE_ATTRS_SET(8),
	NULL
};

static const struct attribute_group img_vmon_attr_group = {
	.name = "img_vmon",
	.attrs = img_vmon_attrs,
};

static int img_vmon_runtime_suspend(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);
	struct img_vmon *hwmon = platform_get_drvdata(pdev);

	clk_disable_unprepare(hwmon->snr_clk);
	clk_disable_unprepare(hwmon->sys_clk);

	return 0;
}

static int img_vmon_runtime_resume(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);
	struct img_vmon *hwmon = platform_get_drvdata(pdev);
	u32 ret;

	ret = clk_prepare_enable(hwmon->sys_clk);
	if (ret)
		return ret;

	ret = clk_prepare_enable(hwmon->snr_clk);
	if (ret) {
		clk_disable_unprepare(hwmon->sys_clk);
		return ret;
	}

	return 0;
}

#ifdef CONFIG_PM_SLEEP
static int img_vmon_suspend(struct device *dev)
{
	return pm_runtime_force_suspend(dev);
}

static int img_vmon_resume(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);
	struct img_vmon *hwmon = platform_get_drvdata(pdev);
	int ret;

	ret = pm_runtime_force_resume(dev);
	if (ret)
		return ret;

	img_vmon_reset(hwmon);

	return 0;
}
#endif /* CONFIG_PM_SLEEP */

static int img_vmon_probe(struct platform_device *pdev)
{
	int ret;
	struct img_vmon *hwmon;
	struct resource *mem;

	hwmon = devm_kzalloc(&pdev->dev, sizeof(*hwmon), GFP_KERNEL);
	if (!hwmon)
		return -ENOMEM;
	hwmon->dev = &pdev->dev;
	mem = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	hwmon->regs = devm_ioremap_resource(&pdev->dev, mem);
	if (IS_ERR(hwmon->regs))
		return PTR_ERR(hwmon->regs);

	hwmon->sys_clk = devm_clk_get(&pdev->dev, "sys_clk");
	if (IS_ERR(hwmon->sys_clk)) {
		dev_err(&pdev->dev, "Failed to get sys_clock.\n");
		return PTR_ERR(hwmon->sys_clk);
	}
	hwmon->snr_clk = devm_clk_get(&pdev->dev, "snr_clk");
	if (IS_ERR(hwmon->snr_clk)) {
		dev_err(&pdev->dev, "Failed to get snr_clock.\n");
		return PTR_ERR(hwmon->snr_clk);
	}

	platform_set_drvdata(pdev, hwmon);

	pm_runtime_set_autosuspend_delay(&pdev->dev, IMG_VMON_PM_TIMEOUT);
	pm_runtime_use_autosuspend(&pdev->dev);
	pm_runtime_enable(&pdev->dev);
	if (!pm_runtime_enabled(&pdev->dev)) {
		ret = img_vmon_runtime_resume(&pdev->dev);
		if (ret)
			return ret;
	}

	spin_lock_init(&hwmon->lock);
	ret = sysfs_create_group(&pdev->dev.kobj, &img_vmon_attr_group);
	if (ret) {
		dev_err(&pdev->dev, "Failed to create sysfs group: %u.\n", ret);
		goto rpm_disable;
	}

	hwmon->hwmon_dev = hwmon_device_register(&pdev->dev);
	if (IS_ERR(hwmon->hwmon_dev)) {
		dev_err(&pdev->dev, "Failed to register device.\n");
		ret = PTR_ERR(hwmon->hwmon_dev);
		goto remove_sysfs;
	}
	img_vmon_reset(hwmon);

	return 0;

remove_sysfs:
	sysfs_remove_group(&pdev->dev.kobj, &img_vmon_attr_group);
rpm_disable:
	if (!pm_runtime_enabled(&pdev->dev))
		img_vmon_runtime_suspend(&pdev->dev);
	pm_runtime_disable(&pdev->dev);
	pm_runtime_dont_use_autosuspend(&pdev->dev);

	return ret;
}

static int img_vmon_remove(struct platform_device *pdev)
{
	struct img_vmon *hwmon = platform_get_drvdata(pdev);

	hwmon_device_unregister(hwmon->hwmon_dev);
	sysfs_remove_group(&pdev->dev.kobj, &img_vmon_attr_group);
	pm_runtime_disable(&pdev->dev);
	if (!pm_runtime_status_suspended(&pdev->dev))
		img_vmon_runtime_suspend(&pdev->dev);

	return 0;
}

static const struct dev_pm_ops img_vmon_pm_ops = {
	SET_RUNTIME_PM_OPS(img_vmon_runtime_suspend,
			   img_vmon_runtime_resume,
			   NULL)
	SET_SYSTEM_SLEEP_PM_OPS(img_vmon_suspend, img_vmon_resume)
};

static const struct of_device_id img_vmon_of_match[] = {
	{ .compatible = "img,vmon", },
	{ },
};
MODULE_DEVICE_TABLE(of, img_vmon_of_match);

static struct platform_driver img_vmon_driver = {
	.probe	= img_vmon_probe,
	.remove = img_vmon_remove,
	.driver = {
		.name = "img-vmon",
		.pm = &img_vmon_pm_ops,
		.of_match_table = of_match_ptr(img_vmon_of_match),
	},
};
module_platform_driver(img_vmon_driver);

MODULE_DESCRIPTION("IMG Voltage Monitor Driver");
MODULE_AUTHOR("Seraj Mohammed <seraj.mohammed@imgtec.com>");
MODULE_LICENSE("GPL V2");
