/*
 * Copyright (C) 2016 Imagination Technologies Ltd.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 *
 */

#include <linux/clk.h>
#include <linux/delay.h>
#include <linux/err.h>
#include <linux/kernel.h>
#include <linux/mfd/syscon.h>
#include <linux/module.h>
#include <linux/of.h>
#include <linux/of_device.h>
#include <linux/platform_device.h>
#include <linux/regmap.h>
#include <linux/regulator/consumer.h>
#include <linux/slab.h>

#include <linux/iio/buffer.h>
#include <linux/iio/iio.h>
#include <linux/iio/sysfs.h>
#include <linux/iio/trigger.h>
#include <linux/iio/trigger_consumer.h>
#include <linux/iio/triggered_buffer.h>

/* Registers */
#define DWC12B5M_ADC_CTRL		0x84
#define DWC12B5M_ADC_CTRL_SOC		BIT(31)
#define DWC12B5M_ADC_CTRL_RPU_CONV	BIT(13)
#define DWC12B5M_ADC_CTRL_SINGLE_CONV	BIT(12)
#define DWC12B5M_ADC_CTRL_SELREF	BIT(9)
#define DWC12B5M_ADC_CTRL_SELDIFF	BIT(8)
#define DWC12B5M_ADC_CTRL_SELBG		BIT(7)
#define DWC12B5M_ADC_CTRL_ENADC		BIT(0)

#define DWC12B5M_ADC_STATUS		0x88
#define DWC12B5M_ADC_STATUS_DATA_MASK	GENMASK(11, 0)
#define DWC12B5M_ADC_STATUS_EOC		BIT(31)

#define DWC12B5M_ADC_EOC_CLEAR		0x98
#define DWC12B5M_ADC_EOC_CLEAR_BIT	BIT(31)

#define DWC12B5M_ADC_CH_MASK		GENMASK(6, 4)

/* Top level register */
#define DWC12B5M_ADC_SOFT_RESET		0x44
#define DWC12B5M_ADC_SOFT_RESET_BIT	BIT(0)

#define DWC12B5M_ADC_INVALID_SAMPLED	0xffff
#define DWC12B5M_ADC_MAX_POLL_COUNT	20

/* TODO: Confirm the number of channels */
#define DWC12B5M_ADC_NUM_CHANNELS	8

struct dwc12b5m_adc_device {
	struct clk *adc_clk;
	struct regmap *reg_base;
	struct regmap *reset_base;
	struct regulator *reg;
	u16 *buf;

	bool shared;
	struct mutex lock;
	unsigned int start_delay_ns;
	unsigned int eoc_delay_ns;
};

static inline void dwc12b5m_reset_adc(struct dwc12b5m_adc_device *adc_dev)
{
	regmap_write(adc_dev->reset_base,
		     DWC12B5M_ADC_SOFT_RESET,
		     DWC12B5M_ADC_SOFT_RESET_BIT);
	udelay(10);
	regmap_write(adc_dev->reset_base, DWC12B5M_ADC_SOFT_RESET, 0);
}

static inline void dwc12b5m_adc_write_reg(struct dwc12b5m_adc_device *adc_dev,
					  u32 reg, u32 val)
{
	 regmap_write(adc_dev->reg_base, reg, val);
}

static inline u32 dwc12b5m_adc_read_reg(struct dwc12b5m_adc_device *adc_dev,
					u32 reg)
{
	u32 val;

	regmap_read(adc_dev->reg_base, reg, &val);

	return val;
}

static void dwc12b5m_adc_power_up(struct dwc12b5m_adc_device *adc_dev)
{
	u32 val;

	dwc12b5m_adc_write_reg(adc_dev, DWC12B5M_ADC_CTRL,
			       DWC12B5M_ADC_CTRL_ENADC);
	dwc12b5m_reset_adc(adc_dev);
	val = dwc12b5m_adc_read_reg(adc_dev, DWC12B5M_ADC_CTRL);
	val |= DWC12B5M_ADC_CTRL_SOC;
	dwc12b5m_adc_write_reg(adc_dev, DWC12B5M_ADC_CTRL, val);
}

static void dwc12b5m_adc_power_down(struct dwc12b5m_adc_device *adc_dev)
{
	u32 val;

	val = dwc12b5m_adc_read_reg(adc_dev, DWC12B5M_ADC_CTRL);
	val &= ~(DWC12B5M_ADC_CTRL_ENADC | DWC12B5M_ADC_CTRL_SELBG);
	dwc12b5m_adc_write_reg(adc_dev, DWC12B5M_ADC_CTRL, val);
}

static void __maybe_unused
dwc12b5m_adc_standby(struct dwc12b5m_adc_device *adc_dev)
{
	u32 val;

	val = dwc12b5m_adc_read_reg(adc_dev, DWC12B5M_ADC_CTRL);
	val &= ~DWC12B5M_ADC_CTRL_ENADC;
	val |= DWC12B5M_ADC_CTRL_SELBG;
	dwc12b5m_adc_write_reg(adc_dev, DWC12B5M_ADC_CTRL, val);
}

static void __maybe_unused
dwc12b5m_adc_wakeup(struct dwc12b5m_adc_device *adc_dev)
{
	u32 val;

	val = dwc12b5m_adc_read_reg(adc_dev, DWC12B5M_ADC_CTRL);
	val |= DWC12B5M_ADC_CTRL_ENADC | DWC12B5M_ADC_CTRL_SELBG;
	dwc12b5m_adc_write_reg(adc_dev, DWC12B5M_ADC_CTRL, val);
}

static void dwc12b5m_adc_start(struct dwc12b5m_adc_device *adc_dev,
			       unsigned int channel)
{
	u32 val;

	/* Channel selection and mode of operation */
	val = (channel & DWC12B5M_ADC_CH_MASK) | DWC12B5M_ADC_CTRL_SINGLE_CONV;
	dwc12b5m_adc_write_reg(adc_dev, DWC12B5M_ADC_CTRL, val);

	udelay(1);
	val = dwc12b5m_adc_read_reg(adc_dev, DWC12B5M_ADC_CTRL);
	val |= DWC12B5M_ADC_CTRL_SOC;
	dwc12b5m_adc_write_reg(adc_dev, DWC12B5M_ADC_CTRL, val);
}

static u16 dwc12b5m_adc_poll_done(struct iio_dev *indio_dev,
				  unsigned int channel,
				  unsigned int delay)
{
	struct dwc12b5m_adc_device *adc_dev = iio_priv(indio_dev);
	unsigned int poll_count = 0;
	u16 ret;

	while (!(dwc12b5m_adc_read_reg(adc_dev, DWC12B5M_ADC_STATUS) &
			DWC12B5M_ADC_STATUS_EOC)) {
		ndelay(delay);
		if (++poll_count == DWC12B5M_ADC_MAX_POLL_COUNT)
			return DWC12B5M_ADC_INVALID_SAMPLED;
	}

	/* Read the 12 bit output register */
	ret = dwc12b5m_adc_read_reg(adc_dev, DWC12B5M_ADC_STATUS) &
				    DWC12B5M_ADC_STATUS_DATA_MASK;

	/* Clear the EOC event in register CR_LOW_RATE_ADC_STATUS */
	dwc12b5m_adc_write_reg(adc_dev, DWC12B5M_ADC_EOC_CLEAR,
			       DWC12B5M_ADC_EOC_CLEAR_BIT);

	return ret;
}

static irqreturn_t dwc12b5m_adc_trigger_h(int irq, void *p)
{
	struct dwc12b5m_adc_device *adc_dev;
	struct iio_poll_func *pf = p;
	struct iio_dev *indio_dev;
	unsigned int delay_ns;
	unsigned int channel;
	unsigned int scan_idx;
	bool sample_invalid;
	u16 *data;
	int i;

	indio_dev = pf->indio_dev;
	adc_dev = iio_priv(indio_dev);
	data = adc_dev->buf;

	mutex_lock(&adc_dev->lock);

	/* Calculate delay step for eoc and sampled data */
	delay_ns = adc_dev->eoc_delay_ns / DWC12B5M_ADC_MAX_POLL_COUNT;

	i = 0;
	sample_invalid = false;
	for_each_set_bit(scan_idx, indio_dev->active_scan_mask,
			 indio_dev->masklength) {
		channel = indio_dev->channels[scan_idx].channel;
		dwc12b5m_adc_start(adc_dev, channel);

		data[i] = dwc12b5m_adc_poll_done(indio_dev, channel, delay_ns);
		if (data[i] == DWC12B5M_ADC_INVALID_SAMPLED) {
			dev_warn(&indio_dev->dev,
				 "invalid sample on channel %d\n", channel);
			sample_invalid = true;
			goto done;
		}
		i++;
	}

done:
	mutex_unlock(&adc_dev->lock);

	if (!sample_invalid)
		iio_push_to_buffers_with_timestamp(indio_dev, data,
						   iio_get_time_ns());
	iio_trigger_notify_done(indio_dev->trig);

	return IRQ_HANDLED;
}

static u16 dwc12b5m_adc_read_raw_voltage(struct iio_dev *indio_dev,
					 struct iio_chan_spec const *chan)
{
	struct dwc12b5m_adc_device *adc_dev = iio_priv(indio_dev);
	unsigned int delay_ns;
	u16 val;

	/* Calculate delay step for eoc and sampled data */
	delay_ns = adc_dev->eoc_delay_ns / DWC12B5M_ADC_MAX_POLL_COUNT;

	dwc12b5m_adc_start(adc_dev, chan->channel);

	val = dwc12b5m_adc_poll_done(indio_dev, chan->channel, delay_ns);

	return val;
}

static int dwc12b5m_adc_read_raw(struct iio_dev *indio_dev,
				 struct iio_chan_spec const *chan,
				 int *val, int *val2, long mask)
{
	struct dwc12b5m_adc_device *adc_dev = iio_priv(indio_dev);
	int ret;

	switch (mask) {
	case IIO_CHAN_INFO_RAW:
		if (iio_buffer_enabled(indio_dev))
			return -EBUSY;
		mutex_lock(&adc_dev->lock);
		*val = dwc12b5m_adc_read_raw_voltage(indio_dev, chan);
		mutex_unlock(&adc_dev->lock);

		if (*val == DWC12B5M_ADC_INVALID_SAMPLED)
			return -EIO;
		return IIO_VAL_INT;

	case IIO_CHAN_INFO_SCALE:
		ret = regulator_get_voltage(adc_dev->reg);
		if (ret < 0)
			return ret;

		*val = ret / 1000;
		*val2 = chan->scan_type.realbits;
		return IIO_VAL_FRACTIONAL_LOG2;

	default:
		return -EINVAL;
	}
}

static int dwc12b5m_update_scan_mode(struct iio_dev *indio_dev,
				     const unsigned long *scan_mask)
{
	struct dwc12b5m_adc_device *adc_dev = iio_priv(indio_dev);

	kfree(adc_dev->buf);
	adc_dev->buf = kmalloc(indio_dev->scan_bytes, GFP_KERNEL);
	if (!adc_dev->buf)
		return -ENOMEM;

	return 0;
}

static const struct iio_info dwc12b5m_adc_info = {
	.driver_module = THIS_MODULE,
	.read_raw = &dwc12b5m_adc_read_raw,
	.update_scan_mode = &dwc12b5m_update_scan_mode,
};

static int dwc12b5m_adc_probe(struct platform_device *pdev)
{
	struct device_node *node = pdev->dev.of_node;
	struct dwc12b5m_adc_device *adc_dev;
	unsigned long adc_clk_rate;
	struct iio_dev *indio_dev;
	unsigned long channel_map;
	int ret;

	indio_dev = devm_iio_device_alloc(&pdev->dev, sizeof(*adc_dev));
	if (!indio_dev)
		return -ENOMEM;

	adc_dev = iio_priv(indio_dev);

	channel_map = GENMASK(DWC12B5M_ADC_NUM_CHANNELS - 1, 0);
	if (!of_property_read_u32(node, "adc-reserved-channels", &ret)) {
		adc_dev->shared = true;
		channel_map &= ~ret;
	}

	adc_dev->reg = devm_regulator_get(&pdev->dev, "vref");
	if (IS_ERR(adc_dev->reg))
		return PTR_ERR(adc_dev->reg);

	ret = regulator_enable(adc_dev->reg);
	if (ret)
		return ret;

	indio_dev->dev.parent = &pdev->dev;
	indio_dev->name = dev_name(&pdev->dev);
	indio_dev->info = &dwc12b5m_adc_info;
	indio_dev->modes = INDIO_DIRECT_MODE;

	adc_dev->reg_base = syscon_regmap_lookup_by_phandle(node,
						"img,cashew-cr-rpu");
	if (IS_ERR(adc_dev->reg_base)) {
		ret = PTR_ERR(adc_dev->reg_base);
		goto err_disable_reg;
	}

	adc_dev->reset_base = syscon_regmap_lookup_by_phandle(node,
						"img,cashew-cr-top");
	if (IS_ERR(adc_dev->reset_base)) {
		ret = PTR_ERR(adc_dev->reset_base);
		goto err_disable_reg;
	}

	adc_dev->adc_clk = devm_clk_get(&pdev->dev, "adc");
	if (IS_ERR(adc_dev->adc_clk)) {
		dev_err(&pdev->dev, "failed to get the clock\n");
		ret = PTR_ERR(adc_dev->adc_clk);
		goto err_disable_reg;
	}

	ret = clk_prepare_enable(adc_dev->adc_clk);
	if (ret) {
		dev_err(&pdev->dev, "failed to enable the clock\n");
		goto err_disable_reg;
	}

	adc_clk_rate = clk_get_rate(adc_dev->adc_clk);
	if (!adc_clk_rate) {
		ret = -EINVAL;
		dev_err(&pdev->dev, "null clock rate!\n");
		goto err_disable_clk;
	}

	adc_dev->eoc_delay_ns = NSEC_PER_SEC / adc_clk_rate;

	/*
	 * There is only one register to power-up/power-down the AUX ADC.
	 * If the ADC is shared among multiple CPUs, always power it up here.
	 * If the ADC is used only by the MIPS, power-up/power-down at runtime.
	 */
	dwc12b5m_adc_power_up(adc_dev);

	mutex_init(&adc_dev->lock);

	ret = iio_triggered_buffer_setup(indio_dev, NULL,
					 &dwc12b5m_adc_trigger_h, NULL);
	if (ret < 0)
		goto err_disable_clk;

	ret = iio_device_register(indio_dev);
	if (ret < 0)
		goto err_cleanup_buffer;

	platform_set_drvdata(pdev, indio_dev);

	return 0;

err_cleanup_buffer:
	iio_triggered_buffer_cleanup(indio_dev);
err_disable_clk:
	clk_disable_unprepare(adc_dev->adc_clk);
err_disable_reg:
	regulator_disable(adc_dev->reg);
	return ret;
}

static int dwc12b5m_adc_remove(struct platform_device *pdev)
{
	struct iio_dev *indio_dev = platform_get_drvdata(pdev);
	struct dwc12b5m_adc_device *adc_dev = iio_priv(indio_dev);

	dwc12b5m_adc_power_down(adc_dev);
	iio_device_unregister(indio_dev);
	iio_triggered_buffer_cleanup(indio_dev);
	clk_disable_unprepare(adc_dev->adc_clk);
	regulator_disable(adc_dev->reg);

	return 0;
}

#ifdef CONFIG_PM_SLEEP
static int dwc12b5m_adc_suspend(struct device *dev)
{
	struct iio_dev *indio_dev = dev_get_drvdata(dev);
	struct dwc12b5m_adc_device *adc_dev = iio_priv(indio_dev);

	dwc12b5m_adc_power_down(adc_dev);
	clk_disable_unprepare(adc_dev->adc_clk);
	regulator_disable(adc_dev->reg);

	return 0;
}

static int dwc12b5m_adc_resume(struct device *dev)
{
	struct iio_dev *indio_dev = dev_get_drvdata(dev);
	struct dwc12b5m_adc_device *adc_dev = iio_priv(indio_dev);
	int ret;

	ret = regulator_enable(adc_dev->reg);
	if (ret)
		return ret;
	ret = clk_prepare_enable(adc_dev->adc_clk);
	if (ret) {
		dev_err(dev, "failed to enable the clock\n");
		regulator_disable(adc_dev->reg);
		return ret;
	}
	dwc12b5m_adc_power_up(adc_dev);

	return 0;
}
#endif /* CONFIG_PM_SLEEP */

static const struct dev_pm_ops dwc12b5m_adc_pm_ops = {
	SET_SYSTEM_SLEEP_PM_OPS(dwc12b5m_adc_suspend, dwc12b5m_adc_resume)
};

static const struct of_device_id dwc12b5m_adc_dt_ids[] = {
	{ .compatible = "snps,dwc12b5m-adc", },
	{ }
};
MODULE_DEVICE_TABLE(of, dwc12b5m_adc_dt_ids);

static struct platform_driver dwc12b5m_adc_driver = {
	.driver = {
		.name   = "dwc12b5m-adc",
		.pm	= &dwc12b5m_adc_pm_ops,
		.of_match_table = dwc12b5m_adc_dt_ids,
	},
	.probe	= dwc12b5m_adc_probe,
	.remove	= dwc12b5m_adc_remove,
};
module_platform_driver(dwc12b5m_adc_driver);

MODULE_AUTHOR("Sifan Naeem <sifan.naeem@imgtec.com>");
MODULE_DESCRIPTION("Synopsys 12b5m ADC driver");
MODULE_LICENSE("GPL v2");
