/*
 * Copyright (C) 2016 Imagination Technologies.
 * Author Will Thomas <will.thomas@imgtec.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <asm/cacheflush.h>
#include <asm/pgtable.h>
#include <linux/compiler.h>
#include <linux/delay.h>
#include <linux/device.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/io.h>
#include <linux/iommu.h>
#include <linux/mm.h>
#include <linux/module.h>
#include <linux/of.h>
#include <linux/of_platform.h>
#include <linux/platform_device.h>
#include <linux/pm_runtime.h>
#include <linux/sizes.h>
#include <linux/slab.h>
#include <linux/spinlock.h>

#include <linux/img-iommu.h>
#include "img_iommu.h"

static struct img_iommu_domain *to_img_domain(struct iommu_domain *domain)
{
	return container_of(domain, struct img_iommu_domain, domain);
}

/*
 * The Imagination Technologies iommu uses a 2-level page table.
 * The first level is the directory table(DT).
 * The DT consists of 4-byte Directory Table Entries(DTEs), each pointing
 * to a "Page Table"
 * Depending on the system page size there are different amounts of DTEs in
 * the DT. Eg with a 4K page system has 1024 DTEs and with a 16K page size
 * each DT has 4096 DTEs Later revisions of the hardware enable software
 * configuration of the page size for better optimisation.
 *
 * The second level is the Page Tables (PT)
 * Each PD consists of 4-byte Page Table Entries(PTEs), each pointing
 * to a page of physical memory. Like the DT, the number of PTEs per PT is
 * depenant on the page size.
 */
/*
 * Each DTE has a PT address and a valid bit:
 *
 *  31: 8 - PT address
 *   7: 4 - Reserved
 *      3 - Cache Snoop state
 *   2: 1 - Reserved (Used for PTE)
 *      0 - 1 if PT for this DTE is valid
 */

#define IMG_DTE_PT_ADDRESS_MASK	0xffffff00
#define IMG_DTE_PT_VALID	BIT(0)

/* Get the PT address for a given DTE */
static inline phys_addr_t img_dte_pt_address(u32 dte)
{
	return (phys_addr_t)dte & IMG_DTE_PT_ADDRESS_MASK;
}

/* Check whether a given DTE is valid */
static inline bool img_dte_is_pt_valid(u32 dte)
{
	return dte & IMG_DTE_PT_VALID;
}

/* Generate a DTE for a given PT, marks valid bit valid */
static inline u32 img_mk_dte(u32 *pt)
{
	phys_addr_t pt_phys = virt_to_phys(pt);

	return (pt_phys & IMG_DTE_PT_ADDRESS_MASK) | IMG_DTE_PT_VALID;
}

static inline u32 img_mk_dte_invalid(u32 dte)
{
	return dte & ~IMG_DTE_PT_VALID;
}

/*
 * Each PTE has a page address, some flags and a valid bit:
 *
 * 31: 8 - Page Address
 *  7: 4 - Reserved
 *     3 - Cache Snoop state
 *     2 - Readable
 *     1 - Writeable
 *     0 - 1 if this PTE is valid
 */

#define IMG_PTE_PAGE_ADDRESS_MASK	0xffffff00
#define IMG_PTE_PAGE_FLAGS_MASK		0x0000000f
#define IMG_PTE_PAGE_READABLE		BIT(2)
#define IMG_PTE_PAGE_WRITABLE		BIT(1)
#define IMG_PTE_PAGE_VALID		BIT(0)

/* Get the physical address of a page in PTE */
static inline phys_addr_t img_pte_page_address(u32 pte)
{
	return (phys_addr_t)pte & IMG_PTE_PAGE_ADDRESS_MASK;
}

/* Check whether a given PTE is valid */
static inline bool img_pte_is_page_valid(u32 pte)
{
	return pte & IMG_PTE_PAGE_VALID;
}

/* Generate a PTE given a page and R/W protection */
static u32 img_mk_pte(phys_addr_t page, int prot)
{
	u32 flags = 0;

	flags |= (prot & IOMMU_READ) ? IMG_PTE_PAGE_READABLE : 0;
	flags |= (prot & IOMMU_WRITE) ? IMG_PTE_PAGE_WRITABLE : 0;
	page &= IMG_PTE_PAGE_ADDRESS_MASK;
	return page | flags | IMG_PTE_PAGE_VALID;
}

static u32 img_mk_pte_invalid(u32 pte)
{
	return pte & ~IMG_PTE_PAGE_VALID;
}

/*
 * IMG IOMMU virtual Address (iova) format
 *
 * 31:22 - DTE index	- index of DTE in DT
 * 21:12 - PTE index	- index of PTE in PT @ DTE.pt_address
 * 11: 0 - Page offset	- offset into page at PTE.page_address
 */
#define IMG_IOVA_DTE_MASK	0xffc00000
#define IMG_IOVA_DTE_SHIFT	22
#define IMG_IOVA_PTE_MASK	0x003ff000
#define IMG_IOVA_PTE_SHIFT	12
#define IMG_IOVA_PAGE_MASK	0x00000fff
#define IMG_IOVA_PAGE_SHIFT	0

static u32 img_iova_dte_index(dma_addr_t iova)
{
	return (u32)(iova & IMG_IOVA_DTE_MASK) >> IMG_IOVA_DTE_SHIFT;
}

static u32 img_iova_pte_index(dma_addr_t iova)
{
	return (u32)(iova & IMG_IOVA_PTE_MASK) >> IMG_IOVA_PTE_SHIFT;
}

static u32 img_iova_page_offset(dma_addr_t iova)
{
	return (u32)(iova & IMG_IOVA_PAGE_MASK) >> IMG_IOVA_PAGE_SHIFT;
}

static u32 img_iommu_read(struct img_iommu *iommu, u32 offset)
{
	return readl(iommu->base + offset);
}

static u32 __maybe_unused
img_iommu_read_domain(struct img_iommu *iommu, u32 offset, u32 domain)
{
	return readl(iommu->base + offset + (domain * 4));
}

static void img_iommu_write(struct img_iommu *iommu, u32 offset, u32 value)
{
	writel(value, iommu->base + offset);
}

static void img_iommu_write_domain(struct img_iommu *iommu, u32 offset,
				   u32 domain, u32 value)
{
	writel(value, iommu->base + offset + (domain * 4));
}

static inline void img_table_flush(struct img_iommu_domain *domain,
				   dma_addr_t dma, unsigned int count)
{
	size_t size = count * sizeof(u32);

	dma_sync_single_for_device(domain->iommu->dev, dma,
				   size, DMA_TO_DEVICE);
}

/* Invalidate cache used after software update to dt */
static void
img_iommu_domain_invalidate_cache(struct img_iommu_domain *img_domain)
{
	struct img_iommu *iommu = img_domain->iommu;

	img_iommu_write(iommu, IMG_MMU_CTRL1, MMU_CTRL1_INVALIDATE_CACHE0);
}

/* Flush cache used after software update to a valid pt */
static void img_iommu_domain_flush_cache(struct img_iommu_domain *img_domain)
{
	struct img_iommu *iommu = img_domain->iommu;

	img_iommu_write(iommu, IMG_MMU_CTRL1, MMU_CTRL1_FLUSH_CACHE0);
}

void img_iommu_set_bypass(struct iommu_domain *domain, bool bypass)
{
	struct img_iommu_domain *img_domain = to_img_domain(domain);
	struct img_iommu *iommu = img_domain->iommu;

	img_iommu_write(iommu, IMG_MMU_ADDR_CTRL, bypass);
}
EXPORT_SYMBOL_GPL(img_iommu_set_bypass);

void img_iommu_set_tiling_scheme(struct iommu_domain *domain, int tiling)
{
	struct img_iommu_domain *img_domain = to_img_domain(domain);
	struct img_iommu *iommu = img_domain->iommu;

	u32 reg;

	reg = img_iommu_read(iommu, IMG_MMU_CTRL0);

	reg &= ~(IMG_TILE_MASK); /* reset tiling scheme bit*/
	reg |= tiling & IMG_TILE_MASK;

	img_iommu_write(iommu, IMG_MMU_CTRL0, reg);

	iommu->tile_mode = 0x1 & tiling;
}
EXPORT_SYMBOL_GPL(img_iommu_set_tiling_scheme);

void img_iommu_clear_pagefault(struct iommu_domain *domain)
{
	struct img_iommu_domain *img_domain = to_img_domain(domain);
	struct img_iommu *iommu = img_domain->iommu;

	img_iommu_write(iommu, IMG_MMU_CTRL1, MMU_CTRL1_RESET);
}
EXPORT_SYMBOL_GPL(img_iommu_clear_pagefault);

bool img_iommu_has_pagefault(struct iommu_domain *domain)
{
	struct img_iommu_domain *img_domain = to_img_domain(domain);
	struct img_iommu *iommu = img_domain->iommu;
	u32 stat0, stat1;

	stat0 = img_iommu_read(iommu, IMG_MMU_STATUS0);
	stat1 = img_iommu_read(iommu, IMG_MMU_STATUS1);

	if (stat0 || stat1) {
		dev_err(iommu->dev, "page fault at iova: %#x (%#x, %#x)\n",
			stat0 & MMU_STATUS0_ADDR_MASK, stat0, stat1);
		return true;
	}
	return false;
}
EXPORT_SYMBOL_GPL(img_iommu_has_pagefault);

static struct iommu_domain *img_iommu_domain_alloc(unsigned iommu_domain_type)
{
	struct img_iommu_domain *img_domain;

	img_domain = kzalloc(sizeof(*img_domain), GFP_KERNEL);
	if (!img_domain)
		return NULL;

	img_domain->iommu = NULL;
	img_domain->index = -1;

	img_domain->domain.geometry.aperture_start = 0;
	img_domain->domain.geometry.aperture_end = 0;
	img_domain->domain.geometry.force_aperture = true;

	return &img_domain->domain;
}

static void img_iommu_domain_free(struct iommu_domain *domain)
{
	struct img_iommu_domain *img_domain = to_img_domain(domain);
	int i;
	u32 dte_start, dte_stop;

	dte_start = img_iova_dte_index(domain->geometry.aperture_start);
	dte_stop = img_iova_dte_index(domain->geometry.aperture_end);

	spin_lock(&img_domain->iommu->dt_lock);

	for (i = dte_start; i < dte_stop; i++) {
		u32 dte = img_domain->iommu->dt_vaddr[i];

		if (img_dte_is_pt_valid(dte)) {
			phys_addr_t pt_phys = img_dte_pt_address(dte);
			u32 *pt = phys_to_virt(pt_phys);

			dma_unmap_single(img_domain->iommu->dev, pt_phys,
					 SPAGE_SIZE, DMA_TO_DEVICE);
			free_page((unsigned long)pt);
			img_domain->iommu->dt_vaddr[i] =
						img_mk_dte_invalid(dte);
		}
	}

	spin_unlock(&img_domain->iommu->dt_lock);

	img_iommu_domain_invalidate_cache(img_domain);
	kfree(img_domain);
}

static struct img_iommu *img_iommu_from_dev(struct device *dev)
{
	struct iommu_group *group;
	struct device *iommu_dev;
	struct img_iommu *img_iommu;

	group = iommu_group_get(dev);
	if (!group)
		return NULL;
	iommu_dev = iommu_group_get_iommudata(group);
	img_iommu = dev_get_drvdata(iommu_dev);
	iommu_group_put(group);

	return img_iommu;
}

static int img_iommu_attach_device(struct iommu_domain *domain,
				   struct device *dev)
{
	struct img_iommu *iommu;
	struct img_iommu_domain *img_domain = to_img_domain(domain);
	u32 domain_id;
	int ret;

	iommu = img_iommu_from_dev(dev);
	if (!iommu)
		return -ENODEV;

	if (img_domain->index >= 0) {
		dev_err(dev, "domain already attached to %u\n",
			img_domain->index);
		return -EINVAL;
	}

	/* if the bitmap is same as mask then no free domains */
	if (iommu->domain_bitmap == IMG_IOMMU_NUM_DOMAINS_MASK) {
		dev_err(dev, "could not attach device, no free domains\n");
		return -EINVAL;
	}

	ret = pm_runtime_get_sync(dev);
	if (ret < 0) {
		dev_err(dev, "could not enable power domain: %d\n", ret);
		return ret;
	}

	domain_id = ffz(iommu->domain_bitmap);

	img_domain->iommu = iommu;
	iommu->domain_bitmap |= BIT(domain_id);
	iommu->domain[domain_id] = domain;
	img_domain->index = domain_id;

	img_iommu_write_domain(iommu, IMG_MMU_DIR_BASE_ADDR, domain_id,
			       (u32)iommu->dt_bus);

	img_iommu_write(iommu, IMG_MMU_CTRL1, MMU_CTRL1_INVALIDATE_CACHE0);
	dev_dbg(dev, "Attached to iommu domain %u\n", domain_id);

	return 0;
}

static void img_iommu_detach_device(struct iommu_domain *domain,
				    struct device *dev)
{
	struct img_iommu *iommu;
	struct img_iommu_domain *img_domain = to_img_domain(domain);

	iommu = img_iommu_from_dev(dev);
	if (!iommu)
		return;

	img_iommu_write_domain(iommu, IMG_MMU_DIR_BASE_ADDR,
			       img_domain->index, 0);
	img_iommu_write_domain(iommu, IMG_MMU_TILE_CFG, img_domain->index, 0);
	img_iommu_write_domain(iommu, IMG_MMU_TILE_MIN, img_domain->index, 0);
	img_iommu_write_domain(iommu, IMG_MMU_TILE_MAX, img_domain->index, 0);

	iommu->domain_bitmap &= ~BIT(img_domain->index);
	iommu->domain[img_domain->index] = NULL;

	pm_runtime_put(dev);

	dev_dbg(dev, "Detached from iommu domain %u\n", img_domain->index);
}

static u32 *img_dte_get_page_table(struct img_iommu_domain *img_domain,
				   dma_addr_t iova)
{
	struct device *dev = img_domain->iommu->dev;
	u32 *pt, *dte_addr;
	u32 dte;
	phys_addr_t pt_phys;
	dma_addr_t pt_dma;

	assert_spin_locked(&img_domain->iommu->dt_lock);

	dte_addr = &img_domain->iommu->dt_vaddr[img_iova_dte_index(iova)];
	dte = *dte_addr;
	if (img_dte_is_pt_valid(dte))
		goto done;

	pt = (u32 *)get_zeroed_page(GFP_ATOMIC | GFP_DMA);
	if (!pt)
		return ERR_PTR(-ENOMEM);

	pt_dma = dma_map_single(dev, pt, SPAGE_SIZE, DMA_TO_DEVICE);
	if (dma_mapping_error(dev, pt_dma)) {
		free_page((unsigned long)pt);
		return ERR_PTR(-ENOMEM);
	}

	dte = img_mk_dte(pt);
	*dte_addr = dte;
	img_iommu_domain_invalidate_cache(img_domain);
	img_table_flush(img_domain, pt_dma, NUM_PT_ENTRIES);

done:
	pt_phys = img_dte_pt_address(dte);
	return (u32 *)phys_to_virt(pt_phys);
}

static size_t img_iommu_unmap_iova(struct img_iommu_domain *img_domain,
				   u32 *pte_addr, dma_addr_t pte_dma,
				   dma_addr_t iova, size_t size)
{
	unsigned int pte_count;
	unsigned int pte_total = size / SPAGE_SIZE;

	assert_spin_locked(&img_domain->iommu->dt_lock);

	for (pte_count = 0; pte_count < pte_total; pte_count++) {
		u32 pte = pte_addr[pte_count];

		if (!img_pte_is_page_valid(pte))
			break;

		pte_addr[pte_count] = img_mk_pte_invalid(pte);
	}

	img_table_flush(img_domain, pte_dma, pte_count);

	return pte_count * SPAGE_SIZE;
}

static int img_iommu_map_iova(struct img_iommu_domain *img_domain,
			      u32 *pte_addr, dma_addr_t pte_dma,
			      dma_addr_t iova, phys_addr_t paddr,
			      size_t size, int prot)
{
	unsigned int pte_count;
	unsigned int pte_total = size / SPAGE_SIZE;
	phys_addr_t page_phys;

	assert_spin_locked(&img_domain->iommu->dt_lock);

	for (pte_count = 0; pte_count < pte_total; pte_count++) {
		if (img_pte_is_page_valid(pte_addr[pte_count]))
			goto unwind;

		pte_addr[pte_count] = img_mk_pte(paddr, prot);
		page_phys = img_pte_page_address(pte_addr[pte_count]);
		paddr += SPAGE_SIZE;
	}

	img_table_flush(img_domain, pte_dma, pte_total);

	return 0;
unwind:
	img_iommu_unmap_iova(img_domain, pte_addr, pte_dma, iova,
			     pte_count * SPAGE_SIZE);

	iova += pte_count * SPAGE_SIZE;
	page_phys = img_pte_page_address(pte_addr[pte_count]);
	pr_err("iova: %pad already mapped to %pa cannot remap to phys: %pa prot %#x\n",
	       &iova, &page_phys, &paddr, prot);

	return -EADDRINUSE;
}

static int img_iommu_map(struct iommu_domain *domain, unsigned long _iova,
			 phys_addr_t paddr, size_t size, int prot)
{
	struct img_iommu_domain *img_domain = to_img_domain(domain);
	struct img_iommu *img_iommu = img_domain->iommu;
	unsigned long flags;
	dma_addr_t iova = (dma_addr_t)_iova;
	u32 *pt, *pte_addr;
	int ret;
	u32 dte_index, pte_index;
	dma_addr_t pte_dma;

	spin_lock_irqsave(&img_domain->iommu->dt_lock, flags);

	pt = img_dte_get_page_table(img_domain, iova);
	if (IS_ERR(pt)) {
		spin_unlock_irqrestore(&img_domain->iommu->dt_lock, flags);
		return PTR_ERR(pt);
	}
	dte_index = img_iommu->dt_vaddr[img_iova_dte_index(iova)];
	pte_index = img_iova_pte_index(iova);
	pte_addr = &pt[pte_index];
	pte_dma = img_dte_pt_address(dte_index) + pte_index * sizeof(u32);
	ret = img_iommu_map_iova(img_domain, pte_addr, pte_dma, iova, paddr,
				 size, prot);
	spin_unlock_irqrestore(&img_domain->iommu->dt_lock, flags);

	return ret;
}

static size_t img_iommu_unmap(struct iommu_domain *domain, unsigned long _iova,
			      size_t size)
{
	struct img_iommu_domain *img_domain = to_img_domain(domain);
	unsigned long flags;
	dma_addr_t pte_dma, iova = (dma_addr_t)_iova;
	phys_addr_t pt_phys;
	u32 dte;
	u32 *pte_addr;
	size_t unmap_size;

	spin_lock_irqsave(&img_domain->iommu->dt_lock, flags);
	dte = img_domain->iommu->dt_vaddr[img_iova_dte_index(iova)];
	if (!img_dte_is_pt_valid(dte)) {
		spin_unlock_irqrestore(&img_domain->iommu->dt_lock, flags);
		return 0;
	}

	pt_phys = img_dte_pt_address(dte);
	pte_addr = (u32 *)phys_to_virt(pt_phys) + img_iova_pte_index(iova);
	pte_dma = pt_phys + img_iova_pte_index(iova) * sizeof(u32);
	unmap_size = img_iommu_unmap_iova(img_domain, pte_addr, pte_dma,
					  iova, size);

	spin_unlock_irqrestore(&img_domain->iommu->dt_lock, flags);

	img_iommu_domain_flush_cache(to_img_domain(domain));

	return unmap_size;
}

static phys_addr_t img_iommu_iova_to_phys(struct iommu_domain *domain,
					  dma_addr_t iova)
{
	struct img_iommu_domain *img_domain = to_img_domain(domain);
	unsigned long flags;
	phys_addr_t pt_phys, phys = 0;
	u32 dte, pte;
	u32 *pt;

	spin_lock_irqsave(&img_domain->iommu->dt_lock, flags);
	/* Check if Directory table entry for iova exists */
	dte = img_domain->iommu->dt_vaddr[img_iova_dte_index(iova)];
	if (!img_dte_is_pt_valid(dte))
		goto out;

	pt_phys = img_dte_pt_address(dte);
	pt = (u32 *)phys_to_virt(pt_phys);
	/* Check if page table entry iova exists in its pagetable*/
	pte = pt[img_iova_pte_index(iova)];
	if (!img_pte_is_page_valid(pte))
		goto out;

	phys = img_pte_page_address(pte) + img_iova_page_offset(iova);
out:
	spin_unlock_irqrestore(&img_domain->iommu->dt_lock, flags);

	return phys;
}

static bool img_iommu_is_dev_iommu_master(struct device *dev)
{
	struct device_node *np = dev->of_node;
	int ret;

	ret = of_count_phandle_with_args(np, "iommus", "#iommu-cells");
	return (ret > 0);
}

static int img_iommu_group_set_iommudata(struct iommu_group *group,
					 struct device *dev)
{
	struct device_node *np = dev->of_node;
	struct platform_device *pd;
	struct of_phandle_args args;
	int ret;

	ret = of_parse_phandle_with_args(np, "iommus", "#iommu-cells", 0,
					 &args);
	if (ret) {
		dev_err(dev, "of_parse_phandle_with_args(%s) => %d",
			np->full_name, ret);
		return ret;
	}
	if (args.args_count != 0) {
		dev_err(dev, "incorrect number of iommu params found for %s %u\n",
			args.np->full_name, args.args_count);
		return -EINVAL;
	}
	pd = of_find_device_by_node(args.np);
	of_node_put(np);
	if (!pd) {
		dev_err(dev, "iommu %s not found\n", args.np->full_name);
		return -EPROBE_DEFER;
	}

	iommu_group_set_iommudata(group, &pd->dev, NULL);

	return 0;
}

static int img_iommu_add_device(struct device *dev)
{
	struct iommu_group *group;
	int ret;

	if (!img_iommu_is_dev_iommu_master(dev))
		return -ENODEV;

	group = iommu_group_get(dev);
	if (!group) {
		group = iommu_group_alloc();
		if (IS_ERR(group)) {
			dev_err(dev, "Failed to allocate IOMMU group\n");
			return PTR_ERR(group);
		}
	}

	ret = iommu_group_add_device(group, dev);
	if (ret)
		goto err_put_group;

	ret = img_iommu_group_set_iommudata(group, dev);
	if (ret)
		goto err_remove_device;

	iommu_group_put(group);

	return 0;

err_remove_device:
	iommu_group_remove_device(dev);
err_put_group:
	iommu_group_put(group);
	return ret;
}

static void img_iommu_remove_device(struct device *dev)
{
	if (!img_iommu_is_dev_iommu_master(dev))
		return;

	iommu_group_remove_device(dev);
}

static int img_iommu_domain_set_attr(struct iommu_domain *domain,
				     enum iommu_attr attr, void *data)
{
	struct img_iommu_domain *img_domain = to_img_domain(domain);
	struct img_iommu *iommu = img_domain->iommu;
	u32 reg;

	if (WARN_ON(!img_domain->iommu))
		return -ENODEV;

	switch (attr) {
	case DOMAIN_ATTR_TILED:
		memcpy(&img_domain->tile_cfg, data,
		       sizeof(struct img_iommu_tile_cfg));
		pr_debug("Setting domain tile cfg:%u %u\n",
			 img_domain->tile_cfg.enable,
			 img_domain->tile_cfg.stride);

		reg = fls(img_domain->tile_cfg.stride) - 10 - iommu->tile_mode;
		reg |= (img_domain->tile_cfg.enable ? BIT(3) : 0);
		img_iommu_write_domain(iommu, IMG_MMU_TILE_CFG,
				       img_domain->index, reg);
		break;
	case DOMAIN_ATTR_GEOMETRY:
		memcpy(&domain->geometry, data,
		       sizeof(struct iommu_domain_geometry));
		pr_debug("Setting domain %u aperture %pad:%pad\n",
			 img_domain->index,
			 &domain->geometry.aperture_start,
			 &domain->geometry.aperture_end);
		img_iommu_write_domain(iommu, IMG_MMU_TILE_MIN,
				       img_domain->index,
				       domain->geometry.aperture_start);
		img_iommu_write_domain(iommu, IMG_MMU_TILE_MAX,
				       img_domain->index,
				       domain->geometry.aperture_end);
		break;
	default:
		return -EINVAL;
	}

	return 0;
}

static int img_iommu_domain_get_attr(struct iommu_domain *domain,
				     enum iommu_attr attr, void *data)
{
	return 0;
}

size_t img_iommu_map_sg(struct iommu_domain *domain, unsigned long iova,
			struct scatterlist *sg, unsigned int nents, int prot)
{
	size_t sz = default_iommu_map_sg(domain, iova, sg, nents, prot);

	img_iommu_domain_flush_cache(to_img_domain(domain));
	img_iommu_domain_invalidate_cache(to_img_domain(domain));

	return sz;
}

static const struct iommu_ops img_iommu_ops = {
	.domain_alloc = img_iommu_domain_alloc,
	.domain_free = img_iommu_domain_free,
	.attach_dev = img_iommu_attach_device,
	.detach_dev = img_iommu_detach_device,
	.map = img_iommu_map,
	.unmap = img_iommu_unmap,
	.add_device = img_iommu_add_device,
	.remove_device = img_iommu_remove_device,
	.domain_get_attr = img_iommu_domain_get_attr,
	.domain_set_attr = img_iommu_domain_set_attr,
	.iova_to_phys = img_iommu_iova_to_phys,
	.map_sg = img_iommu_map_sg,
	.pgsize_bitmap = IMG_IOMMU_PAGES,
};

static int img_iommu_probe(struct platform_device *pdev)
{
	struct device *dev = &pdev->dev;
	struct img_iommu *iommu;
	struct resource *res;
	dma_addr_t bus_addr;

	iommu = devm_kzalloc(dev, sizeof(*iommu), GFP_KERNEL);
	if (!iommu)
		return -ENOMEM;

	platform_set_drvdata(pdev, iommu);
	iommu->dev = dev;

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	iommu->base = devm_ioremap_resource(&pdev->dev, res);
	if (IS_ERR(iommu->base))
		return PTR_ERR(iommu->base);

	spin_lock_init(&iommu->dt_lock);

	iommu->dt_vaddr = dma_zalloc_coherent(dev, SPAGE_SIZE, &bus_addr,
					      GFP_DMA);
	iommu->dt_bus = bus_addr;
	if (!iommu->dt_vaddr)
		return PTR_ERR(iommu->dt_vaddr);

	pm_runtime_enable(dev);

	return 0;
}

static int img_iommu_remove(struct platform_device *pdev)
{
	pm_runtime_disable(&pdev->dev);

	return 0;
}

#ifdef CONFIG_PM_SLEEP
static int img_iommu_suspend_late(struct device *dev)
{
	struct img_iommu *iommu = dev_get_drvdata(dev);
	struct img_iommu_suspend_reg *reg = &iommu->suspend_reg;
	int i;

	if (pm_runtime_suspended(dev))
		return 0;

	reg->ctrl0 = img_iommu_read(iommu, IMG_MMU_CTRL0);
	reg->ctrl1 = img_iommu_read(iommu, IMG_MMU_CTRL1);
	reg->addr_ctrl = img_iommu_read(iommu, IMG_MMU_ADDR_CTRL);
	for (i = 0; i < 4; i++) {
		reg->dir_base_addr[i] =
			img_iommu_read_domain(iommu, IMG_MMU_DIR_BASE_ADDR, i);
		reg->tile_cfg[i] =
			img_iommu_read_domain(iommu, IMG_MMU_TILE_CFG, i);
		reg->tile_min[i] =
			img_iommu_read_domain(iommu, IMG_MMU_TILE_MIN, i);
		reg->tile_max[i] =
			img_iommu_read_domain(iommu, IMG_MMU_TILE_MAX, i);
	}

	return 0;
}

static int img_iommu_resume_early(struct device *dev)
{
	struct img_iommu *iommu = dev_get_drvdata(dev);
	struct img_iommu_suspend_reg *reg = &iommu->suspend_reg;
	int i;

	if (pm_runtime_suspended(dev))
		return 0;

	img_iommu_write(iommu, IMG_MMU_CTRL0, reg->ctrl0);
	img_iommu_write(iommu, IMG_MMU_CTRL1, reg->ctrl1);
	img_iommu_write(iommu, IMG_MMU_ADDR_CTRL, reg->addr_ctrl);
	for (i = 0; i < 4; i++) {
		img_iommu_write_domain(iommu, IMG_MMU_DIR_BASE_ADDR, i,
				       reg->dir_base_addr[i]);
		img_iommu_write_domain(iommu, IMG_MMU_TILE_CFG, i,
				       reg->tile_cfg[i]);
		img_iommu_write_domain(iommu, IMG_MMU_TILE_MIN, i,
				       reg->tile_min[i]);
		img_iommu_write_domain(iommu, IMG_MMU_TILE_MAX, i,
				       reg->tile_max[i]);
	}

	return 0;
}
#endif /* CONFIG_PM_SLEEP */

static const struct dev_pm_ops img_iommu_pm_ops = {
	SET_LATE_SYSTEM_SLEEP_PM_OPS(img_iommu_suspend_late,
				     img_iommu_resume_early)
};

static const struct of_device_id img_iommu_dt_ids[] = {
	{ .compatible = "img,cashew-iommu" },
	{ },
};
MODULE_DEVICE_TABLE(of, img_iommu_dt_ids);

static struct platform_driver img_iommu_driver = {
	.probe = img_iommu_probe,
	.remove = img_iommu_remove,
	.driver = {
		   .name = "img-iommu",
		   .of_match_table = img_iommu_dt_ids,
		   .pm = &img_iommu_pm_ops,
	},
};

static int __init img_iommu_init(void)
{
	struct device_node *np;
	int ret;

	np = of_find_matching_node(NULL, img_iommu_dt_ids);
	if (!np)
		return 0;

	of_node_put(np);

	ret = bus_set_iommu(&platform_bus_type, &img_iommu_ops);
	if (ret)
		return ret;

	return platform_driver_register(&img_iommu_driver);
}

static void __exit img_iommu_exit(void)
{
	platform_driver_unregister(&img_iommu_driver);
}

subsys_initcall(img_iommu_init);
module_exit(img_iommu_exit);

MODULE_DESCRIPTION("IOMMU API for Imagination Technologies");
MODULE_AUTHOR("Will Thomas <will.thomas@imgtec.com>");
MODULE_ALIAS("platform:img-iommu");
MODULE_LICENSE("GPL v2");
