/*
 * Copyright 2016 Imagination Technologies.
 *
 * This program is free software; you may redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/kmod.h>
#include <linux/mutex.h>
#include <linux/types.h>
#include <linux/videodev2.h>
#include <linux/v4l2-dv-timings.h>
#include <media/v4l2-device.h>
#include <media/v4l2-dev.h>
#include <media/v4l2-ioctl.h>
#include <media/v4l2-dv-timings.h>
#include <media/v4l2-ctrls.h>
#include <video/img_hdmi_rx.h>
#include "hdmi_drv.h"


static int hdmi_rx_dev_s_ctrl(struct v4l2_ctrl *ctrl)
{
	return 0;
}

static const struct v4l2_ctrl_ops hdev_ctrl_ops = {
	.s_ctrl = hdmi_rx_dev_s_ctrl,
};

static const struct v4l2_ctrl_config hdmi_ctrls[] = {

	{/* 0 */
		.ops = &hdev_ctrl_ops,
		.id = V4L2_CID_DV_RX_COLOR_DEPTH,
		.name = "color depth",
		.type = V4L2_CTRL_TYPE_INTEGER,
		.def = 24,
		.min = 24,
		.max = 48,
		.step = 6,
	},
	{/* 1 */
		.ops = &hdev_ctrl_ops,
		.id = V4L2_CID_DV_RX_PIXEL_ENCODING,
		.name = "color/pixel encoding",
		.type = V4L2_CTRL_TYPE_INTEGER,
		.def = PE_YCB444,
		.min = PE_RGB,
		.max = PE_INVALID - 1,
		.step = 1,
	},
	{/* 2 */
		.ops = &hdev_ctrl_ops,
		.id = V4L2_CID_DV_RX_COLOR_SPACE,
		.name = "color_space, colorimetry",
		.type = V4L2_CTRL_TYPE_INTEGER,
		.def = V4L2_COLORSPACE_REC709,
		.min = V4L2_COLORSPACE_DEFAULT,
		.max = V4L2_COLORSPACE_REC709,
		.step = 1,
	},
	{/* 3 */
		.ops = &hdev_ctrl_ops,
		.id = V4L2_CID_DV_RX_TMDS_RATE,
		.name = "TMDS clk rate",
		.type = V4L2_CTRL_TYPE_INTEGER,
		.def = 130000000,
		.min = 0,
		.max = 600000000,
		.step = 1,
	},
#if (LINUX_VERSION_CODE <= KERNEL_VERSION(4, 6, 0))
	{/* 4 */
		.ops = &hdev_ctrl_ops,
		.id = V4L2_CID_DV_RX_IT_CONTENT_TYPE,
		.name = "it content type",
		.type = V4L2_CTRL_TYPE_INTEGER,
		.def = V4L2_DV_IT_CONTENT_TYPE_NO_ITC,
		.min = V4L2_DV_IT_CONTENT_TYPE_GRAPHICS,
		.max = V4L2_DV_IT_CONTENT_TYPE_NO_ITC,
		.step = 1,
	},
#else
	/* 4 */
	{
	},
#endif
	{/* 5 */
		.ops = &hdev_ctrl_ops,
		.id = V4L2_CID_DV_RX_VIDEO_PARAMS,
		.name = "all video parameters",
		.type = V4L2_CTRL_TYPE_INTEGER,
		.def = 0,
		.min = 0,
		.max = 0xFFFFFFFF,
		.step = 1,
		.dims = { 15 },
	},
	{/* 6 */
		.ops = &hdev_ctrl_ops,
		.id = V4L2_CID_DV_RX_AUD_SAMPLE_FREQ,
		.name = "Audio sample frequency",
		.type = V4L2_CTRL_TYPE_INTEGER,
		.def = 48000,
		.min = 32000,
		.max = 192000,
		.step = 1,
	},
	{/* 7 */
		.ops = &hdev_ctrl_ops,
		.id = V4L2_CID_DV_RX_AUD_SAMPLE_SIZE,
		.name = "Audio sample size",
		.type = V4L2_CTRL_TYPE_INTEGER,
		.def = 20,
		.min = 16,
		.max = 28,
		.step = 4,
	},
};

int hdmi_rx_ctrl_register_ctrls(struct hdmi_device *hdev)
{
	struct v4l2_ctrl_handler *hdl = &hdev->ctrls.ctrl_handler;

	dev_info(hdev->dev, "v4l2 controls registeration started.\n");
	hdev->ctrls.rx_rgb_range_ctrl =
		v4l2_ctrl_new_std_menu(hdl, &hdev_ctrl_ops,
				       V4L2_CID_DV_RX_RGB_RANGE,
				       V4L2_DV_RGB_RANGE_FULL,
				       0, V4L2_DV_RGB_RANGE_AUTO);
	hdev->ctrls.color_depth_ctrl =
		v4l2_ctrl_new_custom(hdl, &hdmi_ctrls[0], NULL);
	hdev->ctrls.pixel_encoding_ctrl =
		v4l2_ctrl_new_custom(hdl, &hdmi_ctrls[1], NULL);
	hdev->ctrls.color_space_ctrl =
		v4l2_ctrl_new_custom(hdl, &hdmi_ctrls[2], NULL);
	hdev->ctrls.tmds_clk_ctrl =
		v4l2_ctrl_new_custom(hdl, &hdmi_ctrls[3], NULL);
#if (LINUX_VERSION_CODE <= KERNEL_VERSION(4, 6, 0))
	hdev->ctrls.it_content_type_ctrl =
		v4l2_ctrl_new_custom(hdl, &hdmi_ctrls[4], NULL);
#else
	hdev->ctrls.it_content_type_ctrl =
		v4l2_ctrl_new_std_menu(hdl, &hdev_ctrl_ops,
				  V4L2_CID_DV_RX_IT_CONTENT_TYPE,
				  V4L2_DV_IT_CONTENT_TYPE_NO_ITC,
				  0, V4L2_DV_IT_CONTENT_TYPE_NO_ITC);
#endif
	hdev->ctrls.video_params_ctrl =
		v4l2_ctrl_new_custom(hdl, &hdmi_ctrls[5], NULL);
	hdev->ctrls.aud_sample_freq_ctrl =
		v4l2_ctrl_new_custom(hdl, &hdmi_ctrls[6], NULL);
	hdev->ctrls.aud_sample_size_ctrl =
		v4l2_ctrl_new_custom(hdl, &hdmi_ctrls[7], NULL);
	if (hdl->error)
		return(hdl->error);
	dev_info(hdev->dev, "v4l2 controls registered\n");
	hdev->v4l2_dev.ctrl_handler = hdl;
	return 0;
}
