/*
 * Copyright 2016 Imagination Technologies.
 *
 * This program is free software; you may redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
#include "hdmi_drv.h"
#include <video/img_hdmi_rx.h>
#include "linux/of_platform.h"
#include "linux/clk-provider.h"
#include <linux/reset.h>
#include <linux/math64.h>

MODULE_AUTHOR("Sadegh Abbasi <sadegh.abbasi@imgtec.com>");
MODULE_DESCRIPTION("IMG HDMI_RX");
MODULE_LICENSE("GPL v2");

static void interrupts_cfg(struct hdmi_device *hdev,
			   bool enable, bool clear_status);
static void packet_fifo_rst(struct hdmi_device *hdev);
static void dbg_init(struct hdmi_device *hdev);
static int video_handler(struct hdmi_device *hdev);
static int hdmi_rx_ctrl_config(struct hdmi_device *hdev, u32 port);

#define TMDS_TOLERANCE  (20000)
#define LONG_WAIT_FOR_TMDS_IN_US 80000
#define SHORT_WAIT_FOR_TMDS_IN_US 200
enum {
	CLOSED = 0,
	OPENED,
	CONFIGURED
};

enum {
	PM_SUS_STREAMING,
	PM_SUS_NOT_STREAMING,
	PM_NOT_STREAMING,
};
/* A timings may have two vics for two aspect ratios */
struct vic_timings {
	int vic1;
	int vic2;
	struct v4l2_dv_timings dv_timings;
} cea_vic[] = {
	{1, 1, V4L2_DV_BT_CEA_640X480P59_94},
	{2, 3, V4L2_DV_BT_CEA_720X480P59_94},
	{17, 18, V4L2_DV_BT_CEA_720X576P50},
	{60, 65, V4L2_DV_BT_CEA_1280X720P24},
	{61, 66, V4L2_DV_BT_CEA_1280X720P25},
	{62, 67, V4L2_DV_BT_CEA_1280X720P30},
	{19, 68, V4L2_DV_BT_CEA_1280X720P50},
	{4, 69, V4L2_DV_BT_CEA_1280X720P60},
	{32, 72, V4L2_DV_BT_CEA_1920X1080P24},
	{33, 73, V4L2_DV_BT_CEA_1920X1080P25},
	{34, 74, V4L2_DV_BT_CEA_1920X1080P30},
	{20, 20, V4L2_DV_BT_CEA_1920X1080I50},
	{31, 75, V4L2_DV_BT_CEA_1920X1080P50},
	{5, 5, V4L2_DV_BT_CEA_1920X1080I60},
	{16, 76, V4L2_DV_BT_CEA_1920X1080P60},
	{93, 103, V4L2_DV_BT_CEA_3840X2160P24},
	{94, 104, V4L2_DV_BT_CEA_3840X2160P25},
	{95, 105, V4L2_DV_BT_CEA_3840X2160P30},
	{96, 106, V4L2_DV_BT_CEA_3840X2160P50},
	{97, 107, V4L2_DV_BT_CEA_3840X2160P60},
	{98, 98, V4L2_DV_BT_CEA_4096X2160P24},
	{99, 99, V4L2_DV_BT_CEA_4096X2160P25},
	{100, 100, V4L2_DV_BT_CEA_4096X2160P30},
	{101, 101, V4L2_DV_BT_CEA_4096X2160P50},
	{102, 102, V4L2_DV_BT_CEA_4096X2160P60},
};

/*
 * Get bit field
 * @param[in] data raw data
 * @param[in] mask bit field mask
 * @return bit field value
 */
static inline u32 get_field(u32 data, u32 mask)
{
	if (unlikely(!mask))
		return 0;
	return (data & mask) >> __ffs(mask);
}

static inline u32 hdmi_read(struct hdmi_device *hdev, u32 reg_id)
{
	return readl(hdev->regs + reg_id);
}

static inline void hdmi_write(struct hdmi_device *hdev, u32 reg_id, u32 value)
{
	writel(value, hdev->regs + reg_id);
}

static inline u32 hdmi_read_mask(struct hdmi_device *hdev, u32 reg_id, u32 mask)
{
	u32 val = hdmi_read(hdev, reg_id);

	return get_field(val, mask);
}

static inline void hdmi_write_mask(struct hdmi_device *hdev, u32 reg_id,
				   u32 mask, u32 value)
{
	u32 old;

	if (unlikely(!mask)) {
		dev_err(hdev->dev, "mask is zero");
		return;
	}
	value <<= __ffs(mask);
	old = readl(hdev->regs + reg_id);
	value = (value & mask) | (old & ~mask);
	hdmi_write(hdev, reg_id, value);
}

static inline struct hdmi_device *get_hdev(struct v4l2_device *v4l2_dev)
{
	return container_of(v4l2_dev, struct hdmi_device, v4l2_dev);
}

static inline u32 mult_divide_64(u64 mul1, u64 mul2, u64 div)
{
	return (u32)div64_u64(mul1 * mul2, div);
}

static u32 round_sample_rate(u32 rate)
{
	u32 s_rates[] = {22050, 24000, 32000, 44100, 48000, 64000,
			 88200, 96000, 176400, 192000};
	u32 i, best = 0;
	u32 err, best_err = ~0;
	u32 size;

	size = ARRAY_SIZE(s_rates) - 1;
	/* if rate is too high or too low (5%) */
	if (rate < s_rates[0] - s_rates[0] / 20 ||
	    rate > s_rates[size] + s_rates[size] / 20)
		return -EINVAL;

	for (i = 0; i < ARRAY_SIZE(s_rates); i++) {
		err = abs(s_rates[i] - rate);
		if (err < best_err) {
			best_err = err;
			best = s_rates[i];
		}
	}
	if (!best_err || (best_err && best / best_err > 20))
		return best;
	else
		return -EINVAL;
}

static ssize_t debug_counters_show(struct device *dev,
				   struct device_attribute *attr, char *buf)
{
	struct v4l2_device *v4l2_dev = dev_get_drvdata(dev);
	struct hdmi_device *hdev = get_hdev(v4l2_dev);

	return scnprintf(buf, PAGE_SIZE,
			 "audio_fifo_rst %04d, audio_fifo_under %04d, audio_fifo_over: %04d, packet_fifo_rst %04d, irq_handling %04d, irq_packet_decoder %04d, irq_audio_clock %04d, irq_audio_fifo %04d, irq_video_mode %04d, irq_hdmi %04d\n",
			 hdev->dbg.debug_audio_fifo_rst,
			 hdev->dbg.debug_irq_audio_fifo_under,
			 hdev->dbg.debug_irq_audio_fifo_over,
			 hdev->dbg.debug_packet_fifo_rst,
			 hdev->dbg.debug_irq_handling,
			 hdev->dbg.debug_irq_packet_decoder,
			 hdev->dbg.debug_irq_audio_clock,
			 hdev->dbg.debug_irq_audio_fifo,
			 hdev->dbg.debug_irq_video_mode,
			 hdev->dbg.debug_irq_hdmi);
}
DEVICE_ATTR_RO(debug_counters);

static struct device_attribute *hdmi_rx_attrs[] = {
	&dev_attr_debug_counters,
	NULL
};

static int get_timings(struct hdmi_device *hdev,
		       struct v4l2_dv_timings *timings)
{
	int n;
	int vic;

	vic = hdev->video_params.video_mode;
	for (n = 0; n < ARRAY_SIZE(cea_vic); n++) {
		if (vic == cea_vic[n].vic1 ||
		    vic == cea_vic[n].vic2) {
				*timings = cea_vic[n].dv_timings;
				break;
		}
	}
	if (n == ARRAY_SIZE(cea_vic)) {
		dev_err(hdev->dev, "vic = %d not supported", vic);
		return -EINVAL;
	} else {
		return 0;
	}
}

int hdmi_clk_set_pll_rate(struct hdmi_device *hdev, unsigned long req_rate)
{
	unsigned long rate;
	int ret = 0;
	struct clk *clock = hdev->res.pll_clk;
	bool disabled = false;

	if (__clk_is_enabled(clock)) {
		clk_disable_unprepare(clock);
		disabled = true;
	}
	/* If you do not change parent, everything OK */
	clk_set_parent(hdev->res.tmds_mux_clk, hdev->res.tmds_clk);
	clk_set_rate(hdev->res.tmds_clk, hdev->tmds_clk_rate_from_aud);
	dev_info(hdev->dev, "%s: tmds_clk rate = %lu expected %lu", __func__,
		 clk_get_rate(hdev->res.tmds_clk),
		 (unsigned long)hdev->tmds_clk_rate_from_aud);

	ret = clk_set_rate(clock, req_rate);
	if (ret) {
		pr_err("failed to set %s to %lu", __clk_get_name(clock),
		       req_rate);
		clk_set_parent(hdev->res.tmds_mux_clk, hdev->res.xtal_mux_clk);
		goto end;
	}
	rate = clk_get_rate(clock);
	if (rate != req_rate) {
		pr_info("%s: req_rate of %lu not achieved. rate = %lu",
			__clk_get_name(clock), req_rate, rate);
	}
end:
	if (disabled)
		clk_prepare_enable(clock);

	return ret;
}

void hdmi_rx_scdc_setpowerprovided(struct hdmi_device *hdev, uint8_t enable)
{
	hdmi_write_mask(hdev, RA_SCDC_CONFIG, BIT(0), enable);
}

void hdmi_rx_scdc_sethpdlow(struct hdmi_device *hdev, uint8_t enable)
{
	hdmi_write_mask(hdev, RA_SCDC_CONFIG, BIT(1), enable);
}

void hdmi_rx_scdc_scrambling(struct hdmi_device *hdev, uint8_t enable)
{
	/*
	 * enable: 0, scrambling under SCDC control.
	 *	   1: auto, enable when detected scrambled data
	 *	   2: auto, enable when detected scrambled data ??
	 *	   3: forces descrambling
	 */
	if (enable < 4)
		hdmi_write_mask(hdev, RA_HDMI20_CONTROL, SCRAMBEN_SEL, enable);
	else
		dev_err(hdev->dev, "couldn't set scrambl-en");
}

void hdmi_rx_scdc_initconfig(struct hdmi_device *hdev)
{
	/*enable on hdmi20_control register the SCDC communication*/
	hdmi_write(hdev, RA_HDMI20_CONTROL, 0x00000010);
	/*SCDC i2c config register, spike suppression and timeout enable*/
	hdmi_write(hdev, RA_SCDC_I2CCONFIG, 0x101C0000);
	/*
	 * SCDC config, enable +5v provided by source
	 * and hpd voltage low more than 100ms
	 */
	hdmi_write(hdev, RA_SCDC_CONFIG, 0x00000001);
	/* SDCD manufacturer specific register, Manufacturer_OUI, version 1 */
	hdmi_write(hdev, RA_SCDC_WRDATA0, ((0xABCD << 8) | 0x01));
	/* SCDC Channel lock config register, default values, period 27Mhz */
	/* hdmi_write(hdev, RA_CHLOCK_CONFIG, 0x6978); */

	hdmi_rx_scdc_scrambling(hdev, 2);
	hdmi_rx_scdc_setpowerprovided(hdev, 1);
}


void hdmi_rx_scdc_setsinkversion(struct hdmi_device *hdev, uint8_t version)
{
	hdmi_write(hdev, RA_SCDC_WRDATA0, (version & 0xFF));
}

void hdmi_rx_scdc_setmanoui(struct hdmi_device *hdev, uint32_t man_oui)
{
	hdmi_write(hdev, RA_SCDC_WRDATA0, ((man_oui<<8) & 0xFFFFFF00));
}

/* This function prints out the read-only registers of scdc_reg0-3 */
void __maybe_unused hdmi_rx_scdc_getstatusregisters(struct hdmi_device *hdev)
{
	uint32_t scdc_reg0 = 0;
	uint32_t scdc_reg1 = 0;
	uint32_t scdc_reg2 = 0;
	uint32_t scdc_reg3 = 0;

	scdc_reg0 = hdmi_read(hdev, RA_SCDC_REGS0);
	scdc_reg1 = hdmi_read(hdev, RA_SCDC_REGS1);
	scdc_reg2 = hdmi_read(hdev, RA_SCDC_REGS2);
	scdc_reg3 = hdmi_read(hdev, RA_SCDC_REGS3);
	if (scdc_reg0) {
		dev_info(hdev->dev, "Rx controller Register 0x%x:",
			 (u32)RA_SCDC_REGS0);
		dev_info(hdev->dev, "SCDC Status, Scrambling_Status: 0x%x",
			 ((scdc_reg0 & 0x01000000) >> 24));
		dev_info(hdev->dev, "SCDC TMDS, TMDSBitClockRatio: 0x%x",
			 ((scdc_reg0 & 0x00020000) >> 17));
		dev_info(hdev->dev, "SCDC TMDS, Scrambling_Enable: 0x%x",
			 ((scdc_reg0 & 0x00010000) >> 16));
		dev_info(hdev->dev, "SCDC Update Flags, RR_Test: 0x%x",
			 ((scdc_reg0 & 0x00000400) >> 10));
		dev_info(hdev->dev, "SCDC Update Flags, CED_Update: 0x%x",
			 ((scdc_reg0 & 0x00000200) >> 9));
		dev_info(hdev->dev, "SCDC Update Flags, Status_Update: 0x%x",
			 ((scdc_reg0 & 0x00000100) >> 8));
		dev_info(hdev->dev, "SCDC Source Version/Revision: 0x%x",
			 (scdc_reg0 & 0x00000FF));
	}
	if (scdc_reg1) {
		dev_info(hdev->dev, "Rx Controller Reg 0x%x:",
			 (u32)RA_SCDC_REGS1);
		dev_info(hdev->dev, "SCDC CED,Channel 0 Char Err Valid: 0x%x",
			 ((scdc_reg1 & 0x80000000) >> 31));
		dev_info(hdev->dev, "SCDC CED, Channel 0 Char Error Cnt: 0x%x",
			 ((scdc_reg1 & 0x7FFF0000) >> 16));
		dev_info(hdev->dev, "SCDC Status Flags, Ch2_Locked: 0x%x",
			 ((scdc_reg1 & 0x00000800) >> 11));
		dev_info(hdev->dev, "SCDC Status Flags, Ch1_Locked: 0x%x",
			 ((scdc_reg1 & 0x00000400) >> 10));
		dev_info(hdev->dev, "SCDC Status Flags, Ch0_Locked: 0x%x",
			 ((scdc_reg1 & 0x00000200) >> 9));
		dev_info(hdev->dev, "SCDC Status Flags, Clock_Detected: 0x%x",
			 ((scdc_reg1 & 0x00000100) >> 8));
		dev_info(hdev->dev, "SCDC Configuration, RR_Enable: 0x%x",
			 (scdc_reg1 & 0x00000001));
	}
	if (scdc_reg2) {
		dev_info(hdev->dev, "Rx Controller Reg 0x%x:",
			 (u32)RA_SCDC_REGS2);
		dev_info(hdev->dev, "SCDC CED, Channel 2 Char Err Valid: 0x%x",
			 ((scdc_reg2 & 0x80000000) >> 31));
		dev_info(hdev->dev, "SCDC CED, Channel 2 Char Err Cnt: 0x%x",
			 ((scdc_reg2 & 0x7FFF0000) >> 16));
		dev_info(hdev->dev, "SCDC CED, Channel 1 Char Err Valid: 0x%x",
			 ((scdc_reg2 & 0x00008000) >> 15));
		dev_info(hdev->dev, "SCDC CED, Channel 1 Char Error Cnt: 0x%x",
			 (scdc_reg2 & 0x00007FFF));
	}
	if (scdc_reg3) {
		dev_info(hdev->dev, "Rx Controller 0x%x:", (u32)RA_SCDC_REGS3);
		dev_info(hdev->dev, "SCDC Test Config, TestReadRequest: 0x%x",
			 ((scdc_reg3 & 0x0000080) >> 7));
		dev_info(hdev->dev, "SCDC Test Config, TestREadReqDelay: 0x%x",
			 (scdc_reg3 & 0x000007F));
	}
}


/* Get error count from previously enabled counters */
error_cnt_t hdmi_rx_scdc_errorcnt(struct hdmi_device *hdev,
				  uint8_t verbose, uint8_t check_valid)
{
	uint32_t value;
	error_cnt_t error_cnt;

	value = hdmi_read(hdev, RA_SCDC_REGS1);
	error_cnt.errorCh0 = ((value >> 16) & 0x7fff);
	value = 0;
	value = hdmi_read(hdev, RA_SCDC_REGS2);
	error_cnt.errorCh1 = ((value) & 0x7fff);
	error_cnt.errorCh2 = ((value >> 16) & 0x7fff);
	if (verbose) {
		dev_info(hdev->dev, "Ch0: 0x%x", error_cnt.errorCh0);
		dev_info(hdev->dev, "Ch1: 0x%x", error_cnt.errorCh1);
		dev_info(hdev->dev, "Ch2: 0x%x", error_cnt.errorCh2);
	}
	if (check_valid == 1) {
		if ((hdmi_read(hdev, RA_HDMI_PLL_LCK_STS) & 0x0001UL) == 0)
			dev_info(hdev->dev, "TMDS Valid = 0");
		else
			dev_info(hdev->dev, "TMDS Valid = 1");
	}
	return error_cnt;
}

/* Enable or disable all error counters in SCDC structure */
void hdmi_rx_scdc_errorcntconfig(struct hdmi_device *hdev, uint8_t enable,
				 uint8_t verbose)
{
	unsigned long md_clk_rate_khz;
	(enable) ?
	((verbose) ? dev_info(hdev->dev, " SCDC enabled error counters") : 0) :
	((verbose) ? dev_info(hdev->dev, " SCDC disabled error counters") : 0);

	/* Enable SCDC and all error counters. */
	(enable) ? hdmi_write(hdev, RA_HDMI20_CONTROL, 0x10 || 0x1f00) :
			  hdmi_write(hdev, RA_HDMI20_CONTROL, 0x0000);
	/* Confirm that the source has been detected and is providing +5V */
	(enable) ? hdmi_write(hdev, RA_SCDC_CONFIG, 0x0001) :
			  hdmi_write(hdev, RA_SCDC_CONFIG, 0x0000);
	/* hdmi_swreset, resets ihdmi_clk domain, Sec 1.5 databook table1-4 */
	(enable) ? hdmi_write(hdev, RA_DMI_SW_RST, 0x0004) : 0;
	md_clk_rate_khz = clk_get_rate(hdev->res.mode_detect_clk)/1000;
	/* Number of imodetclk cycles in 1 ms for SCDC and channel lock use */
	if (md_clk_rate_khz > 0)
		hdmi_write(hdev, RA_CHLOCK_CONFIG, md_clk_rate_khz);
	else
		dev_err(hdev->dev, "md_clk_rate is 0");
}

void hdmi_rx_scdc_errorcntconfig_enablespecificcounter(struct hdmi_device *hdev,
						       uint8_t enable,
						       uint8_t counters,
						       uint8_t verbose)
{
	unsigned long md_clk_rate_khz;
	(enable) ? hdmi_write(hdev, RA_HDMI20_CONTROL,
			     (0x0010 | ((0x1f & counters) << 8))) :
		   hdmi_write(hdev, RA_HDMI20_CONTROL, 0x0000);
	(enable) ? hdmi_write(hdev, RA_SCDC_CONFIG, 0x0001) :
		   /* Video Data Period check enable */
		   hdmi_write(hdev, RA_SCDC_CONFIG, 0x0000);
	(enable) ? hdmi_write(hdev, RA_DMI_SW_RST, 0x0004) : 0;
	md_clk_rate_khz = clk_get_rate(hdev->res.mode_detect_clk)/1000;
	/* Number of imodetclk cycles in 1 ms for SCDC and channel lock use */
	if (md_clk_rate_khz > 0 && md_clk_rate_khz < 0xffff)
		hdmi_write(hdev, RA_CHLOCK_CONFIG, (u32)md_clk_rate_khz);
	else
		dev_err(hdev->dev, " md_clk_rate invalid = %x",
			(u32)md_clk_rate_khz);
}

/* Configure packet decoder after reset */
static void packet_cfg(struct hdmi_device *hdev)
{
	/* Enable packet storage and error detection */
	hdmi_write(hdev, RA_PDEC_CTRL,
		   PFIFO_STORE_FILTER_EN | PD_FIFO_WE | PDEC_BCH_EN);
	/* Mute audio and/or video after Set_AVMUTE command from source */
	hdmi_write(hdev, RA_PDEC_ASP_CTRL, AUTO_VMUTE|AUTO_SPFLAT_MUTE);
	/* clear the packet fifo */
	packet_fifo_rst(hdev);
}

/* enable/disable the storage of a packet type */
static int hdmi_rx_ctrl_store(struct hdmi_device *hdev,
		       enum hdmi_rx_ctrl_packet_t type, bool enable)
{
	if (hdev == 0 || type >= hdmi_rx_ctrl_packet_cnt)
		return -EINVAL;
	/* is not configured? */
	if (hdev->status != CONFIGURED)
		return -EPERM;
	hdmi_write_mask(hdev, RA_PDEC_CTRL,
			PFIFO_STORE_PACKET << type, enable ? 1 : 0);
	return 0;
}

/* Enable/disable interrupts after HPD change. */
static void interrupts_hpd(struct hdmi_device *hdev, bool enable)
{
	if (enable) {
		/*
		 * Enable the following interrupts of Packet Detector:
		 * DVI Mode Detected, Audio InfoFrame Checksum field value
		 * changed, AVI checksum field value changed,
		 * New packet in fifo, and Packet fifo overflow
		 */
		hdmi_write(hdev, RA_PDEC_IEN_SET, DVIDET | AIF_CKS_CHG |
			   AVI_CKS_CHG | ACR_N_CHG  | AUD_TYPE_CHG);
		hdmi_write(hdev, RA_AUD_FIFO_IEN_SET,
			      AFIF_OVERFL | AFIF_UNDERFL);

		/* Almost all interrupts in Video Mode enabled */
		hdmi_write(hdev, RA_MD_IEN_SET, VIDEO_MODE);
		hdmi_write(hdev, RA_HDMI_IEN_SET,
			   AKSV_RCV | DCM_CURRENT_MODE_CHG | CLK_CHANGE |
			   SCDCTMDSCFGCHANGE | SCDCSCSTATUSCHANGE |
			   SCDCCFGCHANGE);
	} else {
		/* Disable the above interrupts */
		hdmi_write(hdev, RA_PDEC_IEN_CLR, DVIDET | AIF_CKS_CHG |
			   AVI_CKS_CHG | PD_FIFO_NEW_ENTRY | PD_FIFO_OVERFL |
			   GCP_AV_MUTE_CHG | ACR_N_CHG  | AUD_TYPE_CHG);
		hdmi_write(hdev, RA_AUD_FIFO_IEN_CLR,
			   AFIF_OVERFL|AFIF_UNDERFL);
		hdmi_write(hdev, RA_MD_IEN_CLR, VIDEO_MODE);
		hdmi_write(hdev, RA_HDMI_IEN_CLR,
			   AKSV_RCV | DCM_CURRENT_MODE_CHG);
		/* Clear the status of the above interrupts */
		hdmi_write(hdev, RA_PDEC_ICLR, DVIDET | AIF_CKS_CHG |
			   AVI_CKS_CHG | PD_FIFO_NEW_ENTRY | PD_FIFO_OVERFL |
			  GCP_AV_MUTE_CHG | ACR_N_CHG  | AUD_TYPE_CHG);
		hdmi_write(hdev, RA_AUD_FIFO_ICLR, AFIF_OVERFL | AFIF_UNDERFL);
		hdmi_write(hdev, RA_MD_ICLR, VIDEO_MODE);
		hdmi_write(hdev, RA_HDMI_ICLR,
			   AKSV_RCV | DCM_CURRENT_MODE_CHG | CLK_CHANGE |
			   SCDCTMDSCFGCHANGE | SCDCSCSTATUSCHANGE |
			   SCDCCFGCHANGE);
	}
}

/* Reset packet decoder FIFO */
static void packet_fifo_rst(struct hdmi_device *hdev)
{
	hdmi_write_mask(hdev, RA_PDEC_CTRL, PD_FIFO_CLR, 1);
	usleep_range(10, 20);
	hdmi_write_mask(hdev, RA_PDEC_CTRL, PD_FIFO_CLR,  0);
	hdmi_write_mask(hdev, RA_PDEC_CTRL, PD_FIFO_FILL_INFO_CLR, 1);
	hdev->dbg.debug_packet_fifo_rst++;
}

/********** End of Packet-related functions ***************************/
static inline u32 calc_evaltime(struct hdmi_device *hdev)
{
	u32 ref_clk_khz;
	/*
	 * for mode_detect_clk at 158MHz, evaltime is 4095.
	 * It has a linear relationship
	 */
	ref_clk_khz = clk_get_rate(hdev->res.mode_detect_clk)/1000;
	return (ref_clk_khz * 4095) / 158000;
}
static unsigned long get_tmds_clk_rate(struct hdmi_device *hdev)
{
	u32 result;
	u32 evaltime;
	u32 ref_clk_rate_hz;
	u32 rate;

	ref_clk_rate_hz = clk_get_rate(hdev->res.mode_detect_clk);
	evaltime = hdmi_read_mask(hdev, RA_HDMI_CKM_EVLTM, EVAL_TIME);
	if (!evaltime) {
		dev_err(hdev->dev, "evaltime is zero");
		return 0;
	}
	dev_info(hdev->dev, "evaltime = %u", evaltime);
	result = hdmi_read(hdev, RA_HDMI_CKM_RESULT);
	rate = mult_divide_64((u64)ref_clk_rate_hz,
			      (u64)get_field(result, CLKRATE), (u64)evaltime);
	return rate;
}


/* Configure legal TMDS clock range */
static void control_clk_range(struct hdmi_device *hdev, unsigned long min,
			      unsigned long max)
{
	u32 evaltime = 0;
	unsigned long ref_clk_hz;
	u32 data;

	ref_clk_hz = clk_get_rate(hdev->res.mode_detect_clk);
	if (!ref_clk_hz) {
		dev_err(hdev->dev, "%s: ref_clk is zero", __func__);
		WARN_ON(!ref_clk_hz);
		return;
	}
	evaltime = calc_evaltime(hdev);
	data = mult_divide_64((u64)min, (u64)evaltime, (u64)ref_clk_hz);
	hdmi_write_mask(hdev, RA_HDMI_CKM_F, MINFREQ_CKM, data);
	data = mult_divide_64((u64)max, (u64)evaltime, (u64)ref_clk_hz);
	hdmi_write_mask(hdev, RA_HDMI_CKM_F, MAXFREQ_CKM, data);
}

/* Reset audio FIFO */
static void audio_fifo_rst(struct hdmi_device *hdev)
{
	hdmi_write_mask(hdev, RA_AUD_FIFO_CTRL, AFIF_INIT, 1);
	/* usleep_range(15, 20); */
	hdmi_write_mask(hdev, RA_AUD_FIFO_CTRL, AFIF_INIT, 0);
	/* Don't ignore subpackets that have no audio samples! */
	hdmi_write_mask(hdev, RA_AUD_FIFO_CTRL, AFIF_SUBPACKETS, 0);
	hdev->dbg.debug_audio_fifo_rst++;
}

/* Read packet AVI parameters */
static int packet_get_avi(struct hdmi_device *hdev,
			  struct hdmi_rx_ctrl_video *params)
{
	int error = 0;
	int color_space;

	if (!hdev || !params)
		return -EINVAL;
	params->video_format =
		hdmi_read_mask(hdev, RA_PDEC_AVI_PB, VIDEO_FORMAT);
	params->active_valid =
		hdmi_read_mask(hdev, RA_PDEC_AVI_PB, ACT_INFO_PRESENT);
	params->bar_valid =
		hdmi_read_mask(hdev, RA_PDEC_AVI_PB, BAR_INFO_VALID);
	params->scan_info = hdmi_read_mask(hdev, RA_PDEC_AVI_PB, SCAN_INFO);
	color_space = hdmi_read_mask(hdev, RA_PDEC_AVI_PB, COLORIMETRY);
	if (color_space == C_SMPTE170_ITU601)
		params->colorimetry = V4L2_COLORSPACE_SMPTE170M;
	else if (color_space == C_ITU709)
		params->colorimetry = V4L2_COLORSPACE_REC709;
	else
		params->colorimetry = V4L2_COLORSPACE_DEFAULT;
	params->picture_ratio = hdmi_read_mask(hdev, RA_PDEC_AVI_PB,
					       PIC_ASPECT_RATIO);
	params->active_ratio = hdmi_read_mask(hdev, RA_PDEC_AVI_PB,
					      ACT_ASPECT_RATIO);
	params->it_content = hdmi_read_mask(hdev, RA_PDEC_AVI_PB, IT_CONTENT);
	params->ext_colorimetry =
		hdmi_read_mask(hdev, RA_PDEC_AVI_PB, EXT_COLORIMETRY);
	params->rgb_quant_range =
		hdmi_read_mask(hdev, RA_PDEC_AVI_PB, RGB_QUANT_RANGE);
	params->n_uniform_scale =
		hdmi_read_mask(hdev, RA_PDEC_AVI_PB, NON_UNIF_SCALE);
	params->video_mode =
		hdmi_read_mask(hdev, RA_PDEC_AVI_PB, VID_IDENT_CODE);
	params->pixel_repetition =
		hdmi_read_mask(hdev, RA_PDEC_AVI_HB, PIX_REP_FACTOR);
	if (params->it_content)
		params->it_content_type =
			hdmi_read_mask(hdev, RA_PDEC_AVI_HB, IT_CONTENT_TYPE);
	else
		params->it_content_type = V4L2_DV_IT_CONTENT_TYPE_NO_ITC;
	/** @note HW does not support AVI YQ1-0, YCC quantization range */
	/** @note HW does not support AVI CN1-0, IT content type */
	params->bar_end_top = hdmi_read_mask(hdev, RA_PDEC_AVI_TBB,
					     LIN_END_TOP_BAR);
	params->bar_start_bottom = hdmi_read_mask(hdev, RA_PDEC_AVI_TBB,
						  LIN_ST_BOT_BAR);
	params->bar_end_left = hdmi_read_mask(hdev, RA_PDEC_AVI_LRB,
					      PIX_END_LEF_BAR);
	params->bar_start_right = hdmi_read_mask(hdev, RA_PDEC_AVI_LRB,
						 PIX_ST_RIG_BAR);
	return error;
}

static void packet_get_aif(struct hdmi_device *hdev,
			   struct hdmi_rx_ctrl_audio *params)
{
	u32 smpl_freqs[] = {0, 32000, 44100, 48000, 88200, 96000,
			    176400, 192000};
	u32 idx;

	params->coding_type = hdmi_read_mask(hdev, RA_PDEC_AIF_PB0,
					     CODING_TYPE);
	params->channel_count = hdmi_read_mask(hdev, RA_PDEC_AIF_PB0,
					       CHANNEL_COUNT);
	idx = hdmi_read_mask(hdev, RA_PDEC_AIF_PB0, SAMPLE_FREQ);
	params->sample_frequency = smpl_freqs[idx];
	params->sample_size = hdmi_read_mask(hdev, RA_PDEC_AIF_PB0,
					     SAMPLE_SIZE);
	params->coding_extension = hdmi_read_mask(hdev, RA_PDEC_AIF_PB0,
						  AIF_DATA_BYTE_3);
	params->channel_allocation = hdmi_read_mask(hdev, RA_PDEC_AIF_PB0,
						    CH_SPEAK_ALLOC);
	params->down_mix_inhibit = hdmi_read_mask(hdev, RA_PDEC_AIF_PB1,
						  DWNMIX_INHIBIT);
	params->level_shift_value = hdmi_read_mask(hdev, RA_PDEC_AIF_PB1,
						   LEVEL_SHIFT_VAL);
}

/*
 * Read input video parameters
 * @note If EGAIN is returned,
 * @note clear interrupts before call it again
 * @note or let interrupt handler run in between
 */

int hdmi_rx_ctrl_get_video(struct hdmi_device *hdev,
			   struct hdmi_rx_ctrl_video *params)
{
	int error = 0;
	const unsigned factor = 100;
	unsigned divider = 0;
	u64 tmp = 0;
	u64 mult;
	unsigned long ref_clk_khz;

	if (!hdev || !params)
		return -EINVAL;
	/* DVI mode */
	params->dvi = hdmi_read_mask(hdev, RA_PDEC_STS, DVIDET) != 0;
	/* AVI parameters */
	error |= packet_get_avi(hdev, params);
	if (error)
		goto exit;
	ref_clk_khz = clk_get_rate(hdev->res.mode_detect_clk)/1000;

	/* pixel clock */
	params->pixel_clk = hdev->tmds_clk_rate;
	/* image parameters */
	params->interlaced = hdmi_read_mask(hdev, RA_MD_STS, ILACE) != 0;
	params->voffset = hdmi_read_mask(hdev, RA_MD_VOL, VOFS_LIN);
	params->vactive = hdmi_read_mask(hdev, RA_MD_VAL, VACT_LIN);
	params->vtotal = hdmi_read_mask(hdev, RA_MD_VTL, VTOT_LIN);
	params->vtot_clk = hdmi_read_mask(hdev, RA_MD_VTC, VTOT_CLK);
	params->vs_clk = hdmi_read_mask(hdev, RA_MD_VSC, VS_CLK);

	/* refresh rate */
	tmp = (u64)hdmi_read_mask(hdev, RA_MD_VTC, VTOT_CLK);
	if (params->interlaced)	{
		params->voffset <<= 1;
		params->vactive <<= 1;
		params->vtotal <<= 1;
		params->vtot_clk <<= 1;
		tmp <<= 1;
	}
	mult = (u64)ref_clk_khz * (u64)100000;
	if (tmp != 0)
		mult = div64_u64(mult, tmp);
	params->refresh_rate = (tmp == 0) ? 0 : (u32)mult;
	if (params->video_format == 3) /* YCC420 */
		params->refresh_rate = (params->refresh_rate) * 2;
	dev_info(hdev->dev, "ref_clk_khz:%lu, MD_VTC:%u, refresh_rate:%lu",
		 ref_clk_khz, (u32)tmp, params->refresh_rate);

	params->hoffset = hdmi_read_mask(hdev, RA_MD_HT1, HOFS_PIX);
	params->hactive = hdmi_read_mask(hdev, RA_MD_HACT_PX, HACT_PIX);
	params->htotal = hdmi_read_mask(hdev, RA_MD_HT1, HTOT_PIX);
	params->htot32_clk = hdmi_read_mask(hdev, RA_MD_HT0, HTOT32_CLK);
	params->hs_clk = hdmi_read_mask(hdev, RA_MD_HT0, HS_CLK);
	/* deep color mode */
	tmp = hdmi_read_mask(hdev, RA_HDMI_STS, DCM_CURRENT_MODE);

	switch (tmp) {
	case DCM_CURRENT_MODE_48b:
		dev_info(hdev->dev, "clr_depth is DCM_CURRENT_MODE_48b\n");
		params->deep_color_mode = 48;
		/* does not seem that we need floating point here.*/
		divider = 2 * factor;
		break;
	case DCM_CURRENT_MODE_36b:
		dev_info(hdev->dev, "clr_depth is DCM_CURRENT_MODE_36b\n");
		params->deep_color_mode = 36;
		divider = 3 * factor/2;
		break;
	case DCM_CURRENT_MODE_30b:
		dev_info(hdev->dev, "clr_depth is DCM_CURRENT_MODE_30b\n");
		params->deep_color_mode = 30;
		divider = 5 * factor/4;
		break;
	case DCM_CURRENT_MODE_24b:
		dev_info(hdev->dev, "clr_depth is DCM_CURRENT_MODE_24b\n");
		params->deep_color_mode = 24;
		divider = factor;
		break;
	default:
		dev_info(hdev->dev,
			 "clr_depth not known, set to DCM_CURRENT_MODE_24b\n");
		params->deep_color_mode = 24;
		/* divider = 1.00 * factor; */
		divider = factor;
		break;
	}
	/* YCC420 */
	if (params->video_format == PE_YCB420) {
		params->pixel_clk = params->pixel_clk * 2;
		params->hoffset = params->hoffset * 2;
		params->hactive	= params->hactive * 2;
		params->htotal = params->htotal * 2;
	}
	params->pixel_clk = mult_divide_64((u64)params->pixel_clk,
					   (u64)factor, (u64)divider);
	params->hoffset = (params->hoffset * factor) / divider;
	params->hactive	= (params->hactive * factor) / divider;
	params->htotal = (params->htotal  * factor) / divider;
	if (params->pixel_clk < PIXEL_CLK_MIN ||
	    params->pixel_clk > PIXEL_CLK_MAX ||
	    params->hactive < HRESOLUTION_MIN ||
	    params->hactive > HRESOLUTION_MAX ||
	    params->htotal < (params->hactive + params->hoffset) ||
	    params->vactive < VRESOLUTION_MIN ||
	    params->vactive > VRESOLUTION_MAX ||
	    params->vtotal < (params->vactive + params->voffset) ||
	    params->refresh_rate < REFRESH_RATE_MIN ||
	    params->refresh_rate > REFRESH_RATE_MAX)
		error = -ERANGE;
exit:
	return error;
}

static void set_video_ctrls(struct hdmi_device *hdev)
{
	/*
	 * No conversion needed. Table 11 of CEA-861-D
	 * matches enum v4l2_dv_rgb_range.
	 */
	v4l2_ctrl_s_ctrl(hdev->ctrls.rx_rgb_range_ctrl,
			 hdev->video_params.rgb_quant_range);
	dev_info(hdev->dev, " hdev->video_params.rgb_quant_range = %d",
		 hdev->video_params.rgb_quant_range);
	/*
	 * No conversion needed. Table 8-6 of HDMI-Spec1.4
	 * matches enum v4l2_dv_it_content_type
	 */
	v4l2_ctrl_s_ctrl(hdev->ctrls.it_content_type_ctrl,
			 hdev->video_params.it_content_type);
	v4l2_ctrl_s_ctrl(hdev->ctrls.tmds_clk_ctrl,
			 hdev->tmds_clk_rate);
	v4l2_ctrl_s_ctrl(hdev->ctrls.color_space_ctrl,
			 hdev->video_params.colorimetry);
	v4l2_ctrl_s_ctrl(hdev->ctrls.color_depth_ctrl,
			 hdev->video_params.deep_color_mode);
	v4l2_ctrl_s_ctrl(hdev->ctrls.pixel_encoding_ctrl,
			 hdev->video_params.video_format);
	dev_info(hdev->dev,
		 "Vid: %ux%u%c@%lu.%02luHz: %s, DCM %ub, VIC:%u, pix_clk:%lu ",
		 hdev->video_params.hactive,
		 hdev->video_params.vactive,
		 hdev->video_params.interlaced ? 'i' : 'p',
		 hdev->video_params.refresh_rate / 100,
		 hdev->video_params.refresh_rate % 100,
		 hdev->video_params.dvi ? "DVI" : "HDMI",
		 hdev->video_params.deep_color_mode,
		 hdev->video_params.video_mode, hdev->video_params.pixel_clk);
}

static void set_audio_ctrls(struct hdmi_device *hdev)
{
	v4l2_ctrl_s_ctrl(hdev->ctrls.aud_sample_freq_ctrl,
			 hdev->audio_params.sample_frequency);
	v4l2_ctrl_s_ctrl(hdev->ctrls.aud_sample_size_ctrl,
			 hdev->audio_params.sample_size);
}

u32 aud_rate_from_tmds(struct hdmi_device *hdev)
{
	u32 n, cts;
	u32 aud_rate;
	u32 sample_rate;
	u64 mult;
	/* Fout = 128 x Fs = (Ftmds x N)/CTS */
	n = hdmi_read_mask(hdev, RA_PDEC_ACR_N, N_DECODED);
	cts = hdmi_read_mask(hdev, RA_PDEC_ACR_CTS, CTS_DECODED);
	dev_info(hdev->dev, "N= %d, CTS = %d", n, cts);
	hdev->tmds_clk_rate = get_tmds_clk_rate(hdev);
	aud_rate = mult_divide_64((u64)hdev->tmds_clk_rate,
				  (u64)n, (u64)cts);
	sample_rate = round_sample_rate(aud_rate / 128);
	mult = (u64)cts * (u64)128 * (u64)sample_rate;
	hdev->tmds_clk_rate_from_aud = (unsigned long)div64_u64(mult, (u64)n);

	return aud_rate;
}

static void dump_audio_params(struct hdmi_device *hdev)
{
	struct hdmi_rx_ctrl_audio *params = &hdev->audio_params;

	dev_info(hdev->dev, "audio:\n sample_rate = %d\n channels=%d\n",
		 params->sample_frequency, params->channel_count);

}

static int audio_handler(struct hdmi_device *hdev)
{
	struct hdmi_rx_ctrl_audio *params = &hdev->audio_params;
	char * const aud_fmts[] = {"Refer to Stream Header", "L-PCM",
					 "AC-3", "MPEG-1", "MP3", "MPEG2",
					 "DTS", "ATRAC", "One Bit Audio",
					 "Enhanced AC-3", "DTS-HD",
					 "MAT", "DST", "WMA Pro", "AAC LC"};
	u32 iaud_rate;
	unsigned long pll_rate;
	/* Clear the audio FIFO */
	audio_fifo_rst(hdev);
	packet_get_aif(hdev, params);
	if (params->sample_frequency) {
		iaud_rate = round_sample_rate(128 * params->sample_frequency);
	} else {
		iaud_rate = aud_rate_from_tmds(hdev);
		params->sample_frequency = round_sample_rate(iaud_rate / 128);
	}
	switch (params->sample_frequency) {
	case 32000:
	case 24000:
	case 48000:
	case 64000:
	case 96000:
	case 192000:
		pll_rate = 921600000;
		break;
	case 22050:
	case 44100:
	case 88200:
	case 176400:
		pll_rate = 846720000;
		break;
	default:
		dev_err(hdev->dev, "sample_rate (%d) is invalid",
			(int)params->sample_frequency);
		return -EINVAL;
	}
	hdmi_clk_set_pll_rate(hdev, pll_rate);
	hdmi_write_mask(hdev, RA_AUD_FIFO_CTRL, AFIF_SUBPACKETS,
			params->channel_count >= 7 ? 0 : 1);

	dump_audio_params(hdev);

	/* print out both results and see the difference */
	dev_info(hdev->dev, "audio_rate read from: timing-> %d, tmds(hw)-> %d\n",
		 128 * params->sample_frequency,
		 aud_rate_from_tmds(hdev));
	dev_info(hdev->dev, "audio coding type: %s\n",
		 aud_fmts[params->coding_type]);
	/* We need to change the iaudio input clock */
	return clk_set_rate(hdev->res.audio_clk,
			    128 * params->sample_frequency);
}

static inline bool tmds_is_valid(struct hdmi_device *hdev)
{
	return hdmi_read_mask(hdev, RA_HDMI_PLL_LCK_STS, PLL_LOCKED);
}

static bool wait_for_tmds_valid(struct hdmi_device *hdev, u32 time_in_us)
{
	int cnt;

	for (cnt = 0; cnt < 1 + time_in_us / 20; cnt++) {
		if (tmds_is_valid(hdev)) {
			dev_info(hdev->dev, "tmdsvalid iterations: %d", cnt);
			return true;
		}
		usleep_range(20, 25);
	}
	dev_err(hdev->dev, "wait_for_tmds_valid failed");

	return false;
}

static void hdmi_rx_ctrl_reset_hdmi(struct hdmi_device *hdev, bool hard_reset)
{
	/* TODO: Assert ipon_rst_n asynchronous reset
	* (ipon_rst_n is an input pin for the HDMI system and must be
	* controlled by your system). This reset activates all other
	* clock domain resets simultaneously. It is in
	* CR_TOP_SOFT_RESET(0xF00A01A4). This is needed once
	* at power up. Page 57 of user guide.
	*/
	if (!hard_reset) {
		/* Reset all configuration bits and all functional units. */
		hdmi_write(hdev, RA_DMI_DISABLE_IF, 0);
		usleep_range(500, 600);
		hdmi_write(hdev, RA_DMI_DISABLE_IF, 0x100FF);
		hdmi_write(hdev, RA_DMI_SW_RST, 0xFE);
	} else if (!IS_ERR(hdev->rst)) {
		dev_info(hdev->dev, "toggling top level reset.\n");
		reset_control_assert(hdev->rst);
		/* Wait for 500us. See "Note" in Section 1.5 Resets */
		usleep_range(500, 600);
		reset_control_deassert(hdev->rst);
	} else {
		dev_err(hdev->dev, "top level reset not accessible.\n");
	}
}

static int set_phy_color_depth(struct hdmi_device *hdev, int val)
{
	struct v4l2_control ctrl;

	ctrl.id = V4L2_CID_DV_RX_PHY_COLOR_DEPTH;
	ctrl.value = val;
	return v4l2_subdev_s_ctrl(hdev->phy_sd, &ctrl);
}

static int reset_phy_and_core(struct hdmi_device *hdev)
{
	int ret;

	interrupts_cfg(hdev, false, true);
	/* Reset phy */
	ret = set_phy_color_depth(hdev, 24);
	ret = v4l2_subdev_call(hdev->phy_sd, video, s_stream, 0);
	v4l2_ctrl_s_ctrl(hdev->ctrls.color_depth_ctrl, 24);

	if (ret) {
		dev_err(hdev->dev, "RX PHY close failed");
		return ret;
	}
	/* Reset controller */
	dbg_init(hdev);

	/* Set registers to default. Internal reset is enough. */
	hdmi_rx_ctrl_reset_hdmi(hdev, 0);

	return 0;
}

/*
 *	TODO: This will need to be moved to the phy driver but I would need
 *	time to work out how to shoehorn it through the framework so it is here
 *	for now. It is unsafe to leave it here though.
 */
static void dw_hdmi_rx_toggle_phy_mhl(struct hdmi_device *hdev, int status)
{
	/* This is safe because hdev is zeroed when allocated */

	dev_info(hdev->dev, "Changed phy MHL for modes %s 3.4Gbps\n",
		status ? "over" : "under");
	hdmi_write(hdev, 0x2C8, 0xe);
	hdmi_write(hdev, 0x2CC, 0x08 | status << 8);
	hdmi_write(hdev, 0x2D4, 0x01);
	if (status)
		wait_for_tmds_valid(hdev, LONG_WAIT_FOR_TMDS_IN_US);
}

static int config_phy_and_core(struct hdmi_device *hdev)
{
	int i, ret;

	hdmi_write(hdev, RA_HDMI_SETUP_CTRL, 0x0);
	usleep_range(10, 20);

	/* Change the status to "OPENED" and start configuration */
	hdev->status = OPENED;
	ret = hdmi_rx_ctrl_config(hdev, 0);
	if (ret) {
		hdev->status = CLOSED;
		return ret;
	}

	/* Store all packets. We may change it later */
	for (i = 0; i < hdmi_rx_ctrl_packet_cnt; i++)
		ret |= hdmi_rx_ctrl_store(hdev, i, true);

	if (ret) {
		hdev->status = CLOSED;
		return ret;
	}
	ret = v4l2_subdev_call(hdev->phy_sd, video, s_stream, 1);
	if (ret)
		return ret;

	dw_hdmi_rx_toggle_phy_mhl(hdev, hdev->mhl);

	return ret;
}

static int clock_handler(struct hdmi_device *hdev)
{
	int ret = 0;
	unsigned long tmdsclk;
	struct hdmi_rx_ctrl_video v;

	/*
	 * TMDS clock has changed
	 * if 0 do nothing
	 * if PLL can lock all is good
	 * if PLL cant lock need to reset phy and bring back up alternate mode
	 */

	tmdsclk = get_tmds_clk_rate(hdev);
	dev_info(hdev->dev, "new tmds_clk =  %lu", tmdsclk);
	if (tmdsclk == 0) {
		hdev->tmds_clk_rate = 0;
		return 0;
	}
	if (tmdsclk < TMDS_CLK_MIN || tmdsclk > TMDS_CLK_MAX)
		dev_err(hdev->dev, "tmds clk rate out of range");
	interrupts_hpd(hdev, true);
	ret = wait_for_tmds_valid(hdev, SHORT_WAIT_FOR_TMDS_IN_US);
	if (!ret) {
		dev_info(hdev->dev, "clk changed but cant lock\n");
		ret = v4l2_subdev_call(hdev->phy_sd, video, s_stream, 0);
		ret = v4l2_subdev_call(hdev->phy_sd, video, s_stream, 1);
		hdev->mhl = 1 - hdev->mhl;
		dw_hdmi_rx_toggle_phy_mhl(hdev, hdev->mhl);
		video_handler(hdev);
	}
	/* wait for valid tmds */
	if (wait_for_tmds_valid(hdev, SHORT_WAIT_FOR_TMDS_IN_US)) {
		dev_info(hdev->dev, "getting video after clock change");
		ret = hdmi_rx_ctrl_get_video(hdev, &v);
		tmdsclk = get_tmds_clk_rate(hdev);
		if (abs(hdev->tmds_clk_rate - tmdsclk) > TMDS_TOLERANCE) {
			if (tmdsclk < TMDS_CLK_MIN || tmdsclk > TMDS_CLK_MAX) {
				dev_err(hdev->dev,
					"tmds rate %lu invalid. was:%lu\n",
					tmdsclk, hdev->tmds_clk_rate);
				return -EINVAL;
			} else {
				hdev->tmds_clk_rate = tmdsclk;
				ret = audio_handler(hdev);
				if (ret)
					return ret;
				/* Set ctrl values */
				set_audio_ctrls(hdev);
				ret = set_phy_color_depth(hdev,
							  v.deep_color_mode);
			}
		}
	}
	return 0;
}

int hdmi_rx_ctrl_get_packet(struct hdmi_device *hdev, u8 *data, u32 size)
{
	int ret = 0;
	const unsigned max = 32;
	u32 tmp = 0;
	u32 i = 0;

	if (hdmi_read_mask(hdev, RA_PDEC_STS, PD_FIFO_NEW_ENTRY) == 0) {
		/* re-align */
		hdmi_write_mask(hdev, RA_PDEC_CTRL, PD_FIFO_SKIP, 1);
		hdmi_write_mask(hdev, RA_PDEC_CTRL, PD_FIFO_SKIP, 0);
	}
	if (hdmi_read_mask(hdev, RA_PDEC_STS, PD_FIFO_NEW_ENTRY) == 0) {
		ret = -ENODATA;
		goto exit;
	}
	for (i = 0; i < max; i++) {
		if ((i % 4) == 0)
			tmp = hdmi_read(hdev, RA_PDEC_FIFO_DATA);
		if (hdmi_read_mask(hdev, RA_PDEC_STS, PD_FIFO_UNDERFL) != 0) {
			ret = -ENODATA;
			goto exit;
		}
		if (i < size)
			data[i] = (u8)tmp;
		tmp >>= 8;
	}
exit:
	return ret;
}


static int packet_handler(struct hdmi_device *hdev)
{
	/* Read packets and possibly just print out the type */
	int ret;
	u8 data[32] = {0};
	u32 i;

	ret = hdmi_rx_ctrl_get_packet(hdev, data, sizeof(data));
	if (ret)
		return  ret;
	/*
	 * Stop storing this type of packets. Only one packet of each type is
	 * stored and printed out. Remove this to print all received packets.
	 */
	ret = hdmi_rx_ctrl_store(hdev, data[0], false);
	for (i = 0; i < sizeof(data); i += 8)
		dev_info(hdev->dev, " %02X %02X %02X %02X %02X %02X %02X %02X",
			 data[i], data[i + 1], data[i + 2], data[i + 3],
			 data[i + 4], data[i + 5], data[i + 6], data[i + 7]);

	return 0;
}

void inform_user(struct hdmi_device *hdev, int i, u8 *ev_data)
{
	struct v4l2_event event;
	int ret = 0;

	memset(&event, 0, sizeof(event));
	if (ev_data)
		memcpy(event.u.data, ev_data, sizeof(event.u.data));
	switch (i) {
	case EVENT_CABLE_CONNECTED:
	case EVENT_CABLE_DISCONNECTED:
		dev_info(hdev->dev, "HDMI cable state changed\n");
		if (pm_runtime_get_sync(hdev->dev) < 0) {
			dev_err(hdev->dev, "pm_runtime_get_sync failed\n");
		} else {
			if (i == EVENT_CABLE_CONNECTED) {
				hdmi_rx_scdc_setpowerprovided(hdev, 1);
				event.type = VIDEO_IN_CABLE_CONNECT;
			} else {
				hdmi_rx_scdc_setpowerprovided(hdev, 0);
				event.type = VIDEO_IN_CABLE_DISCONNECT;
			}
			pm_runtime_put(hdev->dev);
		}
		break;
	case EVENT_PACKET_FIFO_ERROR:
		packet_fifo_rst(hdev);
		break;
	case EVENT_CLOCK_CHANGE:
		ret = clock_handler(hdev);
		event.type = VIDEO_IN_CLOCK_CHANGED;
		if (!ret)
			v4l2_ctrl_s_ctrl(hdev->ctrls.tmds_clk_ctrl,
					 hdev->tmds_clk_rate);
		break;
	case EVENT_VIDEO_STATUS:
		event.type = VIDEO_IN_VIDEO_CHANGED;
		ret = video_handler(hdev);
		if (!ret)
			set_video_ctrls(hdev);
		break;
	case EVENT_AUDIO_STATUS:
		event.type = VIDEO_IN_AUDIO_CHANGED;
		ret = audio_handler(hdev);
		/* Set ctrl values */
		if (!ret)
			set_audio_ctrls(hdev);
		break;
	case EVENT_PACKET_RECEPTION:
		ret = packet_handler(hdev);
		break;
	case EVENT_AKSV_RECEPTION:
/* event.type = VIDEO_IN_ASK_RCVD; */
		break;
	default:
		dev_err(hdev->dev, "invalid event");
	}
	if (ret)
		dev_err(hdev->dev, "wq: error happened. event = %d\n", i);
	else if (event.type)
		v4l2_event_queue(&hdev->vdev, &event);
}

void wq_handler(struct work_struct *work)
{
	struct hdmi_device *hdev;
	int i;

	if (!work) {
		pr_err("%s: work_struct is null\n", __func__);
		return;
	}
	hdev = container_of(work, struct hdmi_device, wks);
	for (i = 0; i < EVENT_INVALID; i++) {
		if (hdev->events[i]) {
			hdev->events[i] = 0;
			inform_user(hdev, i, NULL);
		}
	}
}

/* Interrupt handler */
static irqreturn_t hdmi_irq_handler(int irq, void *dev_data)
{
	struct hdmi_device *hdev = dev_data;
	int error = 0;
	u32 intr_hdmi, intr_md, intr_aud, intr_pdec, intr_aud_fifo;

	intr_hdmi = hdmi_read(hdev, RA_HDMI_ISTS) &
			      hdmi_read(hdev, RA_HDMI_IEN);
	intr_md = hdmi_read(hdev, RA_MD_ISTS) &
			    hdmi_read(hdev, RA_MD_IEN);
	intr_pdec = hdmi_read(hdev, RA_PDEC_ISTS) &
			      hdmi_read(hdev, RA_PDEC_IEN);
	intr_aud = hdmi_read(hdev, RA_AUD_CEC_ISTS) &
		 hdmi_read(hdev, RA_AUD_CEC_IEN);
	intr_aud_fifo = hdmi_read(hdev, RA_AUD_FIFO_ISTS) &
		 hdmi_read(hdev, RA_AUD_FIFO_IEN);

	if (intr_hdmi) {
		hdmi_write(hdev, RA_HDMI_ICLR, intr_hdmi);
		/* Check if clock rate has changed */
		if (get_field(intr_hdmi, CLK_CHANGE)) {
			/* handle outside interrupt context: may sleep */
			hdev->events[EVENT_CLOCK_CHANGE] = true;
		}
		/* Has the color depth changed? */
		if (get_field(intr_hdmi, DCM_CURRENT_MODE_CHG)) {
			/*
			 * soft reset video block if deep color mode
			 * is more than the current deep colore mode ??
			 */
			if (hdmi_read_mask(hdev, RA_PDEC_GCP_AVMUTE,
					   PKTDEC_GCP_CD) >
			    hdmi_read_mask(hdev, RA_HDMI_STS,
					   DCM_CURRENT_MODE))
				/* WA - 24-bit DCM fall back */
				hdmi_write_mask(hdev, RA_DMI_SW_RST,
						VID_SWRESET, 0x1);
			hdev->events[EVENT_VIDEO_STATUS] = true;
		}
		/* Start of HDCP? KSV Key Selection Vector */
		if (get_field(intr_hdmi, AKSV_RCV)) {
			hdev->events[EVENT_AKSV_RECEPTION] = true;
			dev_info(hdev->dev, "aksv event 0x%x", intr_hdmi);
		}
		hdev->dbg.debug_irq_hdmi++;
	}
	/* Check Video Mode Interrupt Status */
	if (intr_md) {
		/* dev_info(hdev->dev, "Video Mode: 0x%x", intr_md); */
		hdmi_write(hdev, RA_MD_ICLR, intr_md);
		if (get_field(intr_md, VIDEO_MODE) != 0)
			hdev->events[EVENT_VIDEO_STATUS] = true;
		hdev->dbg.debug_irq_video_mode++;
	}
	/* Check Packet Decoder Interrupt Flags */
	/* dev_info(hdev->dev, "RX INTERRUPT: 0x%x", intr_pdec); */
	if (intr_pdec) {
		hdmi_write(hdev, RA_PDEC_ICLR, intr_pdec);
		/* DVI mode detected or change in AVI checksum */
		if (get_field(intr_pdec, DVIDET | AVI_CKS_CHG))
			hdev->events[EVENT_VIDEO_STATUS] = true;
		/* AIF (Audio InfoFrame)Packet received */
		if (get_field(intr_pdec, AIF_CKS_CHG | GCP_AV_MUTE_CHG |
			      ACR_N_CHG  | AUD_TYPE_CHG))
			hdev->events[EVENT_AUDIO_STATUS] = true;
		/*
		 * New packet in packet FIFO. If we have chosen to store some
		 * types of packets in the external FIFO, this interrupt
		 * signals the addition of one packet to the FIFO.
		 */
		if (get_field(intr_pdec, PD_FIFO_NEW_ENTRY)) {
			hdev->events[EVENT_PACKET_RECEPTION] = true;
			/* dev_info(hdev->dev, "packet reception event 0x%x", */
			/* intr_pdec); */
		}
		/* packet fifo overflow or underflow */
		if (get_field(intr_pdec, PD_FIFO_OVERFL | PD_FIFO_UNDERFL))
			hdev->events[EVENT_PACKET_FIFO_ERROR] = true;
		hdev->dbg.debug_irq_packet_decoder++;
	}

	if (intr_aud) {
		hdmi_write(hdev, RA_AUD_CEC_ICLR, intr_aud);
		hdev->dbg.debug_irq_audio_clock++;
	}

	if (intr_aud_fifo) {
		if (get_field(intr_aud_fifo, AFIF_OVERFL | AFIF_UNDERFL) != 0) {
			audio_fifo_rst(hdev);

			if (get_field(intr_aud_fifo, AFIF_UNDERFL))
				++hdev->dbg.debug_irq_audio_fifo_under;
			if (get_field(intr_aud_fifo, AFIF_OVERFL))
				++hdev->dbg.debug_irq_audio_fifo_over;
		}
		hdmi_write(hdev, RA_AUD_FIFO_ICLR, intr_aud_fifo);
		hdev->dbg.debug_irq_audio_fifo++;
	}
	hdev->dbg.debug_irq_handling++;
	queue_work(hdev->wq, &hdev->wks);
	return error;
}

/* Add as many registers as needed to this list */
static void __maybe_unused hdmi_dumpregs(struct hdmi_device *hdev,
					 char *prefix)
{
#define DUMPREG(reg_id) \
	dev_dbg(hdev->dev, "%s:" #reg_id " = %08x\n", prefix, \
		readl(hdev->regs + reg_id))

	dev_dbg(hdev->dev, "%s: ---- CONTROL REGISTERS ----\n", prefix);
	DUMPREG(RA_HDMI_SETUP_CTRL);
	DUMPREG(RA_HDMI_OVR_CTRL);
	DUMPREG(RA_HDMI_TIMERA_CTRL);
	DUMPREG(RA_HDMI_RES_OVR);
	DUMPREG(RA_HDMI_RES_STS);
	DUMPREG(RA_HDMI_PLL_CTRL);
	DUMPREG(RA_HDMI_PLL_FRQSET1);
	DUMPREG(RA_HDMI_PLL_FRQSET2);

	dev_dbg(hdev->dev, "%s: ---- CORE REGISTERS ----\n", prefix);

	dev_dbg(hdev->dev, "%s: ---- CORE SYNC REGISTERS ----\n", prefix);
#undef DUMPREG
}

static int hdmi_rx_streamon(struct file *file, void *_fh, u32 val)
{
	struct hdmi_device *hdev = video_drvdata(file);

	/*
	 * Enable clk_chg and pll_lck_chg interrupts
	 * HPD workaround. clk_chg interrupt means HPD.
	 */
	hdmi_write(hdev, RA_HDMI_IEN_SET, CLK_CHANGE | SCDCTMDSCFGCHANGE |
		   SCDCSCSTATUSCHANGE | SCDCCFGCHANGE);
	return 0;
}

static int hdmi_rx_streamoff(struct file *file, void *_fh, u32 val)
{
	struct hdmi_device *hdev = video_drvdata(file);

	dev_info(hdev->dev, "sourcing audio clock from xtal");
	clk_set_parent(hdev->res.tmds_mux_clk, hdev->res.xtal_mux_clk);
	return 0;
}


/* Enable/disable interrupts */
static void interrupts_cfg(struct hdmi_device *hdev, bool enable,
			   bool clear_status)
{
	if (enable) {
		/* set enable */
		hdmi_write(hdev, RA_PDEC_IEN_SET, ~0);
		hdmi_write(hdev, RA_AUD_CEC_IEN_SET, ~0);
		hdmi_write(hdev, RA_AUD_FIFO_IEN_SET, ~0);
		hdmi_write(hdev, RA_MD_IEN_SET, ~0);
		hdmi_write(hdev, RA_HDMI_IEN_SET, ~0);
	} else {
		/* clear enable */
		hdmi_write(hdev, RA_PDEC_IEN_CLR, ~0);
		hdmi_write(hdev, RA_AUD_CEC_IEN_CLR, ~0);
		hdmi_write(hdev, RA_AUD_FIFO_IEN_CLR, ~0);
		hdmi_write(hdev, RA_MD_IEN_CLR, ~0);
		hdmi_write(hdev, RA_HDMI_IEN_CLR, ~0);
	}
	if (clear_status) {
		/* clear status */
		hdmi_write(hdev, RA_PDEC_ICLR, ~0);
		hdmi_write(hdev, RA_AUD_CEC_ICLR, ~0);
		hdmi_write(hdev, RA_AUD_FIFO_ICLR, ~0);
		hdmi_write(hdev, RA_MD_ICLR, ~0);
		hdmi_write(hdev, RA_HDMI_ICLR, ~0);
	}

}

/* Configure control after reset (default settings) */
static void control_cfg(struct hdmi_device *hdev, unsigned port)
{
	unsigned evaltime = 0;

	/* Need to know if this works for our configuration. */
	evaltime = calc_evaltime(hdev);

	hdmi_write(hdev, RA_HDMI_OVR_CTRL, ~0);	/* enable all */
	hdmi_write_mask(hdev, RA_HDMI_SYNC_CTRL, VS_POL_ADJ_MODE,
			VS_POL_ADJ_AUTO);
	hdmi_write_mask(hdev, RA_HDMI_SYNC_CTRL, HS_POL_ADJ_MODE,
			HS_POL_ADJ_AUTO);
	hdmi_write_mask(hdev, RA_HDMI_CKM_EVLTM, EVAL_TIME, evaltime);
	/*
	 * trigger interrupt if the number of tmds pulses duringh evlatime
	 * (~tmds_rate) has changed more than 6.25% (1/16)
	 * |f_clk_curr - f_clk_last| >= f_clk_last/(2^(clk_hyst+2))
	 */
	hdmi_write_mask(hdev, RA_HDMI_CKM_EVLTM, CLK_HYST, 2);
	control_clk_range(hdev, TMDS_CLK_MIN, TMDS_CLK_MAX);
	/* bit field shared between phy and controller */
	hdmi_write_mask(hdev, RA_HDMI_PCB_CTRL, INPUT_SELECT, port);
	hdmi_write_mask(hdev, RA_SNPS_PHYG3_CTRL, MSK(2, 2), port);
}

/* Configure audio after reset (default settings) */
static void audio_cfg(struct hdmi_device *hdev)
{
	/* UNmute all audio channels */
	hdmi_write_mask(hdev, RA_AUD_MUTE_CTRL, AUD_MUTE_SEL,
			AUD_MUTE_FORCE_UN);
	/* Disable I2S and SPDIF outputs */
	hdmi_write_mask(hdev, RA_AUD_SAO_CTRL, SPDIF_DISABLE, 0x0F);
	hdmi_write_mask(hdev, RA_AUD_SAO_CTRL, I2S_DISABLE, 0x0F);
	/* Enable Parallel Audio out */
	hdmi_write_mask(hdev, RA_AUD_PAO_CTRL, PAO_DISABLE, 0x01);
	usleep_range(TMOUT, TMOUT * 2);
	hdmi_write_mask(hdev, RA_AUD_PAO_CTRL, PAO_DISABLE, 0x0);

	/*
	 * Automatic PAO output format. Manual format selection is possible. eg
	 * set the PAO audio format to L-PCM. Also select automatic audio rate.
	 */
	hdmi_write_mask(hdev, RA_AUD_PAO_CTRL, AUDIO_FMT_SEL, 0x0);
	hdmi_write_mask(hdev, RA_AUD_PAO_CTRL, PAO_RATE, 0x3);
	/* Clear the audio FIFO */
	audio_fifo_rst(hdev);
}

/* Enable/disable hot plug detect output */
static void hdmi_rx_ctrl_hpd(struct hdmi_device *hdev, bool ready)
{
	hdmi_write_mask(hdev, RA_HDMI_SETUP_CTRL, HOT_PLUG_DETECT_INPUT_A, 0x1);
	hdmi_write_mask(hdev, RA_HDMI_SETUP_CTRL, HOT_PLUG_DETECT, ready);
}

void cea_cfg(struct hdmi_device *hdev)
{
	/* Reset cea video output interface */
	hdmi_write_mask(hdev, RA_CEAVID_CONFIG, CEAVID_RST, 0x1);
	mdelay(33);
	hdmi_write_mask(hdev, RA_CEAVID_CONFIG, CEAVID_RST, 0x0);
	mdelay(33);
}

 /* Configure hdmi_rx_core device. See user guide, "Configure Controller" */
static int hdmi_rx_ctrl_config(struct hdmi_device *hdev, u32 port)
{
	if (!hdev || port > 3)
		return -EINVAL;
	if (hdev->status == CLOSED)
		return -EPERM;
	/* Disable HPD output to tell tx that the rx is not ready. */
	hdmi_rx_ctrl_hpd(hdev, false);
	/* disable and clear all interrupts */
	interrupts_cfg(hdev, false, true);
	/* change the status to "not configured" */
	hdev->status = OPENED;
	/* The initial state is cable not connected. tmds_rate = 0 */
	hdev->tmds_clk_rate = 0;

	/* Power on for Clocks, Polarity and Port */
	control_cfg(hdev, port);

	/*
	 * Enable HPD (Hot Plug Detect) output to indicat to the tx that the rx
	 * is ready. Wait for tmdsvalid and aud_pll_lock. P62 of user guide
	 */
	hdmi_rx_ctrl_hpd(hdev, true);
	/* Audio output enable and audio fifo reset */
	audio_cfg(hdev);
	/* Packet filter and reset, mute audio and video if needed. */
	packet_cfg(hdev);
	/* Reset CEA video output */
	cea_cfg(hdev);

	/* Set hdmi to dvi  and dvi to hdmi mode switching hysteresis */
	hdmi_write(hdev, RA_HDMI_MODE_RECOVER, 0x2100);
	/* Set verical lines interrupt thresholds to maximum */
	hdmi_write(hdev, RA_MD_VTH, 0xfff);


	hdev->status = CONFIGURED;

	/* Filter (discard) all packets with errors */
	hdmi_write(hdev, RA_PDEC_ERR_FILTER, 0x1FF);
	hdmi_write(hdev, RA_HDMI_ERROR_PROTECT, 0x50C18);
	/* Configure SCDC (Status and Data Channel) */
	hdmi_rx_scdc_initconfig(hdev);
	return 0;
}
static void dbg_init(struct hdmi_device *hdev)
{
	hdev->dbg.debug_audio_fifo_rst = 0;
	hdev->dbg.debug_packet_fifo_rst = 0;
	hdev->dbg.debug_irq_handling = 0;
	hdev->dbg.debug_irq_packet_decoder = 0;
	hdev->dbg.debug_irq_audio_clock = 0;
	hdev->dbg.debug_irq_audio_fifo = 0;
	hdev->dbg.debug_irq_video_mode = 0;
	hdev->dbg.debug_irq_hdmi = 0;
	hdev->dbg.debug_irq_audio_fifo_under = 0;
	hdev->dbg.debug_irq_audio_fifo_over = 0;
}

int set_tmds_pll(unsigned long tmds_clk_rate)
{
	/* TODO: Set the relevant PLL when tmds clk changes.
	 * This is an important action and needs to be done
	 * after careful study of the clocks.
	 * 1.	This might be the responsibility of the application
	 *	as it differs from one clock arrangement from another.
	 *	We may only inform the application here by triggering
	 *	an event. Note that the caller of this function is
	 *	video handler() that is called after getting interrupt.
	 * 2.	We may also do it in this driver. If dts provides a handle
	 *	to a pll, it would mean that the tmds clk is the ref clk
	 *	to that pll, which will be re-configured by this driver.
	 *	If dts does not provide the handle to a pll, then we can just
	 *	call clk_set_rat() to inform the kernel that the clk rate has
	 *	and expect the interested parties to have registered a
	 *	notification on the clk rate change.
	 *	This approach seems more reasonable.
	 */
	return 0;
}

static int hdmi_rx_querycap(struct file *file, void *priv,
			    struct v4l2_capability *cap)
{
	strlcpy(cap->driver, KBUILD_MODNAME, sizeof(cap->driver));
	strlcpy(cap->card, HDMI_RX_DRIVER_NAME, sizeof(cap->card));
	cap->device_caps = V4L2_CAP_VIDEO_CAPTURE |
			   V4L2_CAP_STREAMING;
	cap->capabilities = cap->device_caps | V4L2_CAP_DEVICE_CAPS;

	return 0;
}

static int hdmi_rx_fh_open(struct file *filp)
{
	int ret = 0;
	struct hdmi_device *hdev = video_drvdata(filp);

	/* check for exclusive access */
	ret = mutex_trylock(&hdev->access_lock);
	if (!ret) {
		dev_err(hdev->dev, "Already in use\n");
		return -EPERM;
	}

	ret = v4l2_fh_open(filp);
	if (ret) {
		dev_err(hdev->dev, "v4l2_fh_open failed.\n");
		goto mutex_tryload_cleanup;
	}
	/* enable the clocks */
	ret = pm_runtime_get_sync(hdev->dev);
	if (ret < 0)
		return ret;
	/* It makes sense to configure hdmi controller and phy here.
	 * First call hdmi_rx_phy_open() to open/configre the phy for
	 * the first time. Then configure the controller.
	 */
	ret = config_phy_and_core(hdev);
	if (ret)
		return ret;
	dev_info(hdev->dev, "\nhdmi_open done.");
	hdev->status = OPENED;
	return 0;

mutex_tryload_cleanup:
	mutex_unlock(&hdev->access_lock);
	return ret;
}
static int hdmi_rx_fop_release(struct file *filp)
{
#define CLEAR(x) memset(&(x), 0, sizeof(x))
	int ret = 0;
	struct hdmi_device *hdev = video_drvdata(filp);

	v4l2_fh_release(filp);
	hdmi_rx_streamoff(filp, NULL, 0);
	hdev->status = CLOSED;
	reset_phy_and_core(hdev);
	ret = cancel_work_sync(&hdev->wks);
	if (ret == 0)
		flush_workqueue(hdev->wq);
	else
		dev_err(hdev->dev, "cancel_work_sync error = %d", ret);
	/* disable hdcp */
	hdmi_write(hdev, RA_DMI_HDCP_CTRL, 0x0);
	pm_runtime_put(hdev->dev);
	CLEAR(hdev->video_params);
	CLEAR(hdev->audio_params);
	CLEAR(hdev->events);

	/* release exclusive access */
	mutex_unlock(&hdev->access_lock);

	return ret;
}
static int hdmi_rx_s_dv_timings(struct file *file, void *_fh,
				struct v4l2_dv_timings *timings)
{
/* It is possible to set the timings manually through registers CEAVID_CONFIG
 * to CEAVID_VCONFIG_HI (0x400-0x414). We will not support this for the
 * time being
 */
	return -EINVAL;
}

static int get_video_mode(struct hdmi_device *hdev)
{
#ifdef __HDMI_RX_TARGET_IS_SIMULATOR__
	hdev->video_params.video_mode = 69;
	return 0;
#endif
	/*FIXME: This is a workaround before the edid flash memory is fixed */
	if (!hdev->video_params.video_mode) {
		if (hdev->video_params.hactive == 640 &&
		    hdev->video_params.vactive == 480)
			hdev->video_params.video_mode = 1;
		else if (hdev->video_params.hactive == 1920 &&
			 hdev->video_params.vactive == 1080)
			hdev->video_params.video_mode = 76;
		else if (hdev->video_params.hactive == 3840 &&
			 hdev->video_params.vactive == 2160)
			hdev->video_params.video_mode = 95;
		else if (hdev->video_params.hactive == 4096 &&
			 hdev->video_params.vactive == 2160)
			hdev->video_params.video_mode = 98;
		else {
			dev_err(hdev->dev, "input timing not supported.");
			return -EINVAL;
		}
	}
	return 0;
}
static int hdmi_rx_g_dv_timings(struct file *flip, void *_fh,
				struct v4l2_dv_timings *timings)
{
	struct hdmi_device *hdev;

	hdev = video_drvdata(flip);
	return get_timings(hdev, timings);
}

static int hdmi_rx_subscribe_event(struct v4l2_fh *fh,
				  const struct v4l2_event_subscription *sub)
{
	if (!sub || !fh)
		return -EINVAL;
	if (sub->type < V4L2_EVENT_VIDEO_IN_CLASS ||
		sub->type >= V4L2_EVENT_VIDEO_IN_CLASS + VIDEO_IN_INVALID_EVENT)
		return -EINVAL;

	return v4l2_event_subscribe(fh, sub, 3, NULL);
}
/*
 * The set of file operations. Note that all these ops are standard core
 * helper functions.
 */
static const struct v4l2_file_operations pdev_fops = {
	.owner = THIS_MODULE,
	.open = hdmi_rx_fh_open,
	.release = hdmi_rx_fop_release,
	.poll = v4l2_ctrl_poll,
	.unlocked_ioctl = video_ioctl2,
/* .close = vpp_dev_close, */
};
static const struct v4l2_ioctl_ops pdev_ioctl_ops = {
	.vidioc_streamon = hdmi_rx_streamon,
	.vidioc_streamoff = hdmi_rx_streamoff,


	.vidioc_querycap = hdmi_rx_querycap,
	.vidioc_s_dv_timings = hdmi_rx_s_dv_timings,
	.vidioc_g_dv_timings = hdmi_rx_g_dv_timings,
	.vidioc_subscribe_event = hdmi_rx_subscribe_event,
	.vidioc_unsubscribe_event = v4l2_event_unsubscribe,

	.vidioc_log_status = v4l2_ctrl_log_status,
};

static void hdmi_resource_clear_clocks(struct hdmi_resources *res)
{
	res->cec_clk = ERR_PTR(-EINVAL);
	res->config_clk = ERR_PTR(-EINVAL);
	res->mode_detect_clk = ERR_PTR(-EINVAL);
	res->audio_clk = ERR_PTR(-EINVAL);
}

static void hdmi_resources_cleanup(struct hdmi_device *hdev)
{
	struct hdmi_resources *res = &hdev->res;

	dev_dbg(hdev->dev, "HDMI resource cleanup\n");
	/* put clocks, power */
	if (!IS_ERR(res->audio_clk))
		clk_put(res->audio_clk);
	if (!IS_ERR(res->mode_detect_clk))
		clk_put(res->mode_detect_clk);
	if (!IS_ERR(res->config_clk))
		clk_put(res->config_clk);
	if (!IS_ERR(res->cec_clk))
		clk_put(res->cec_clk);
	memset(res, 0, sizeof(*res));
	if (!IS_ERR(hdev->regs))
		iounmap(hdev->regs);
	if (hdev->irq != IRQ_INVALID) {
		free_irq(hdev->irq, (void *)hdev);
		hdev->irq = IRQ_INVALID;
	}
	hdmi_resource_clear_clocks(res);
}

static int hdmi_rx_runtime_resume(struct device *dev)
{
	struct v4l2_device *v4l2_dev = dev_get_drvdata(dev);
	struct hdmi_device *hdev = get_hdev(v4l2_dev);
	int ret;

	dev_info(hdev->dev, "enabling the clocks...");
	ret = clk_prepare_enable(hdev->res.mode_detect_clk);
	ret |= clk_prepare_enable(hdev->res.audio_clk);
	ret |= clk_prepare_enable(hdev->res.config_clk);
	ret |= clk_prepare_enable(hdev->res.cec_clk);
	ret |= clk_prepare_enable(hdev->res.esm_clk);
	/* TODO: disable the clocks that we have enabled here? */
	if (ret)
		dev_err(hdev->dev, "couldn't enable clks\n");

	dev_info(hdev->dev, "enabled the colcks successfully.");
	dev_info(hdev->dev, "mode:%lu audio:%lu config:%lu cec:%lu esm:%lu\n",
		 clk_get_rate(hdev->res.mode_detect_clk),
		 clk_get_rate(hdev->res.audio_clk),
		 clk_get_rate(hdev->res.config_clk),
		 clk_get_rate(hdev->res.cec_clk),
		 clk_get_rate(hdev->res.esm_clk));
	return ret;
}
static int hdmi_rx_runtime_suspend(struct device *dev)
{
	struct v4l2_device *v4l2_dev = dev_get_drvdata(dev);
	struct hdmi_device *hdev = get_hdev(v4l2_dev);

	dev_info(hdev->dev, "disabling the clocks...");
	clk_disable_unprepare(hdev->res.mode_detect_clk);
	clk_disable_unprepare(hdev->res.audio_clk);
	clk_disable_unprepare(hdev->res.config_clk);
	clk_disable_unprepare(hdev->res.cec_clk);
	clk_disable_unprepare(hdev->res.esm_clk);
	dev_info(hdev->dev, "clocks disabled...");
	return 0;
}

static int hdmi_resources_init(struct hdmi_device *hdev)
{
	struct device *dev = hdev->dev;
	struct hdmi_resources *res = &hdev->res;
	unsigned long tmds_rate;

	dev_dbg(dev, "HDMI resource init\n");

	memset(res, 0, sizeof(*res));
	hdmi_resource_clear_clocks(res);

	res->tmds_clk = devm_clk_get(dev, "tmds");
	if (IS_ERR(res->tmds_clk))  {
		dev_err(dev, "failed to get clock 'tmds'\n");
		goto fail;
	}
	tmds_rate = clk_get_rate(res->tmds_clk);
	dev_info(dev, "hdmi_rx_tmds_clk rate = %lu\n", tmds_rate);
	/* get clocks, power */
	res->cec_clk = devm_clk_get(dev, "cec");
	if (IS_ERR(res->cec_clk)) {
		dev_err(dev, "failed to get clock 'cec'\n");
		goto fail;
	}
	/* pll clk for registers, also used for hdmi-phy */
	res->pll_clk = devm_clk_get(dev, "pll");
	if (IS_ERR(res->pll_clk)) {
		dev_err(dev, "failed to get clock 'pll'\n");
		goto fail;
	}
	/* pll clk for registers, also used for hdmi-phy */
	res->tmds_mux_clk = devm_clk_get(dev, "tmds_mux");
	if (IS_ERR(res->tmds_mux_clk)) {
		dev_err(dev, "failed to get clock 'tmds_mux'\n");
		goto fail;
	}
	res->xtal_mux_clk = devm_clk_get(dev, "xtal_mux");
	if (IS_ERR(res->xtal_mux_clk)) {
		dev_err(dev, "failed to get clock 'xtal_mux'\n");
		goto fail;
	}
	/* config clk for registers, also used for hdmi-phy */
	res->config_clk = devm_clk_get(dev, "cfg");
	if (IS_ERR(res->config_clk)) {
		dev_err(dev, "failed to get clock 'config'\n");
		goto fail;
	}
	res->mode_detect_clk = devm_clk_get(dev, "mode_detect");
	if (IS_ERR(res->mode_detect_clk)) {
		dev_err(dev, "failed to get clock 'mode_detect'\n");
		goto fail;
	}
	/* TODO: Check the mode_detect rate,  MD_CLK_MIN and MD_CLK_MAX */
	res->audio_clk = devm_clk_get(dev, "audio");
	if (IS_ERR(res->audio_clk)) {
		dev_err(dev, "failed to get clock 'audio'\n");
		goto fail;
	}
	return 0;
fail:
	dev_err(dev, "HDMI resource init - failed\n");
	hdmi_resources_cleanup(hdev);
	return -ENODEV;
}
static int find_reg_callback(struct device *dev, void *p)
{
	struct v4l2_subdev **sd = p;

	*sd = dev_get_drvdata(dev);
	/* non-zero value stops iteration */
	return 1;
}
static struct v4l2_subdev *find_and_register_subdev(
	struct hdmi_device *hdev, char *module_name)
{
	struct device_driver *drv;
	struct v4l2_subdev *sd = NULL;
	int ret;

	drv = driver_find(module_name, &platform_bus_type);
	if (!drv) {
		pr_info("\nmodule %s is missing\n", module_name);
		return NULL;
	}
	pr_info("\nmodule %s found\n", module_name);
	/* driver refcnt is increased, it is safe to iterate over devices */
	ret = driver_for_each_device(drv, NULL, &sd, find_reg_callback);
	/* ret = 0 means that find_reg_callback was never called */
	if (sd == NULL) {
		pr_info("module %s provides no subdev!\n", module_name);
		goto done;
	}
	/* v4l2_device_register_subdev detects if sd is NULL */
	ret = v4l2_device_register_subdev(&hdev->v4l2_dev, sd);
	if (ret) {
		pr_err("failed to register subdev %s\n", sd->name);
		sd = NULL;
	} else
		pr_info("registered subdev %s\n", sd->name);

done:
	return sd;
}

static int video_handler(struct hdmi_device *hdev)
{
	/* SA: Something in the input video has changed.
	 * Here we need to decide which parts need to be done by the driver
	 * and which parts need to be done by the application.
	 */
	int error = 0;
	struct hdmi_rx_ctrl_video v = {0};
	struct hdmi_rx_ctrl_video *video_params;

	if (!hdev)
		return -EINVAL;
	video_params = &hdev->video_params;
	dev_info(hdev->dev, "video handler called. waiting for tmdsvalid.");
	/* Step 3, page 65 */
	/* wait for the video mode is stable */
	if (!wait_for_tmds_valid(hdev, SHORT_WAIT_FOR_TMDS_IN_US))
		return -EINVAL;
	/* Don't bother waiting for audio here for now. */

	/* Step 4, Clear packet fifo */
	packet_fifo_rst(hdev);
	/* Step 5, Reset audio clock and fifo */
	audio_cfg(hdev);
	/* Step 6 Reset the CEA video interface */
	cea_cfg(hdev);

	error = hdmi_rx_ctrl_get_video(hdev, &v);
	dev_info(hdev->dev, "status = %d", error);
	if (!error) {
		dev_info(hdev->dev,
			 "Video: %ux%u%c@%lu.%02luHz: %s, DCM %ub, VIC:%u, pix_clk:%lu",
			 v.hactive, v.vactive, v.interlaced ? 'i' : 'p',
			 v.refresh_rate / 100, v.refresh_rate % 100,
			 v.dvi ? "DVI" : "HDMI", v.deep_color_mode,
			 v.video_mode, v.pixel_clk);
		video_params->dvi = v.dvi;
		video_params->pixel_clk = v.pixel_clk;
		video_params->refresh_rate = v.refresh_rate;
		video_params->interlaced = v.interlaced;
		video_params->video_mode = v.video_mode;
		video_params->htotal = v.htotal;
		video_params->vtotal = v.vtotal;
		video_params->hactive = v.hactive;
		video_params->vactive  = v.vactive;
		video_params->pixel_repetition = v.pixel_repetition;
		if (video_params->deep_color_mode != v.deep_color_mode) {
			if (v.deep_color_mode == 24 ||
			    v.deep_color_mode == 30 ||
			    v.deep_color_mode == 36 ||
			    v.deep_color_mode == 48) {
				dev_info(hdev->dev, "new deep color = %d",
					 v.deep_color_mode);
				video_params->deep_color_mode =
					v.deep_color_mode;
				error = set_phy_color_depth(hdev,
							    v.deep_color_mode);
			}
		}
		video_params->rgb_quant_range = v.rgb_quant_range;
		video_params->video_format = v.video_format;
		video_params->colorimetry = v.colorimetry;
		video_params->it_content_type = v.it_content_type;

		/* Let the application know that tmds_clk_rate has changed. */
		error |= set_tmds_pll(hdev->tmds_clk_rate);
	} else {
		/* Something wrong. Reconfigure the phy. */
		dev_info(hdev->dev, "%s failed to get video parameters",
			 __func__);
		error = v4l2_subdev_call(hdev->phy_sd, video, s_stream, 0);
		error = v4l2_subdev_call(hdev->phy_sd, video, s_stream, 1);

		hdev->mhl = 1 - hdev->mhl;
		dw_hdmi_rx_toggle_phy_mhl(hdev, hdev->mhl);
	}
	/* clear the packet fifo */
	/* Video changed: disable audio, reset audio fifo and enable audio */
	/* make sure that video_mode is not 0 */
	if (!hdev->video_params.video_mode)
		error = get_video_mode(hdev);
	return error;
}

static irqreturn_t source_power_handler(int irq, void *dev_data)
{
	struct hdmi_device *hdev = dev_data;
	int level;

	level = gpiod_get_value(hdev->gpiodesc);
	if (level)
		hdev->events[EVENT_CABLE_CONNECTED] = true;
	else
		hdev->events[EVENT_CABLE_DISCONNECTED] = true;

	queue_work(hdev->wq, &hdev->wks);

	return 0;
}

static int hdmi_probe(struct platform_device *pdev)
{
	struct device *dev = &pdev->dev;
	struct resource *res;
	struct hdmi_device *hdev = NULL;
	int ret;
	struct video_device *vdev;
	struct device_driver *drv;
	int irq_gpio_p_detect;
	struct v4l2_ctrl_handler *hdl;
	int i;

	dev_info(dev, "probe start\n");
	drv = driver_find("img,hdmiphy", &platform_bus_type);
	if (!drv) {
		dev_info(dev, "\nimg,hdmiphy not registered yet\n");
		return -EPROBE_DEFER;
	}
	hdev = devm_kzalloc(&pdev->dev, sizeof(*hdev), GFP_KERNEL);
	if (!hdev) {
		ret = -ENOMEM;
		goto fail;
	}
	hdev->dev = dev;
	hdev->status = CLOSED;
	hdev->irq = IRQ_INVALID;

	ret = hdmi_resources_init(hdev);
	if (ret)
		goto fail_free;

	/* mapping HDMI registers */
	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (res == NULL) {
		dev_err(dev, "get memory resource failed.\n");
		ret = -ENXIO;
		goto fail_init;
	}
	hdev->regs = devm_ioremap(&pdev->dev, res->start,
				  resource_size(res));
	if (hdev->regs == NULL) {
		dev_err(dev, "register mapping failed.\n");
		ret = -ENXIO;
		goto fail_init;
	}
	res = platform_get_resource(pdev, IORESOURCE_IRQ, 0);
	if (res == NULL) {
		dev_err(dev, "get interrupt resource failed.\n");
		ret = -ENXIO;
		goto fail_init;
	}
	dev_info(dev, "got the interrupt resource.\n");
	ret = devm_request_irq(&pdev->dev, res->start, hdmi_irq_handler, 0,
			       "hdmi", hdev);
	if (ret) {
		dev_err(dev, "request interrupt failed.\n");
		goto fail_init;
	}
	hdev->irq = res->start;
	mutex_init(&hdev->lock);

	/* setting v4l2 name to prevent WARN_ON in v4l2_device_register */
	strlcpy(hdev->v4l2_dev.name, dev_name(dev),
		sizeof(hdev->v4l2_dev.name));
	ret = v4l2_device_register(&pdev->dev, &hdev->v4l2_dev);
	if (ret) {
		dev_err(dev, "could not register v4l2 device.\n");
		goto fail_init;
	} else
		dev_info(dev, "hdmi-rx registered\n");

	/* initialise the access control */
	mutex_init(&hdev->access_lock);

	dbg_init(hdev);
	/* Add the controls */
	hdl = &hdev->ctrls.ctrl_handler;
	/* 10 is a hint to the number of controls. */
	v4l2_ctrl_handler_init(hdl, 10);
	ret = hdmi_rx_ctrl_register_ctrls(hdev);
	if (ret)
		goto fail_hdl;

	/* Initialize the video_device structure */
	vdev = &hdev->vdev;
	strlcpy(vdev->name, KBUILD_MODNAME, sizeof(vdev->name));
	/*
	 * The release callback must be non-NULL.
	 */
	vdev->release = video_device_release_empty;
	vdev->fops = &pdev_fops;
	vdev->ioctl_ops = &pdev_ioctl_ops;
	/*
	 * The main serialization lock. All ioctls are serialized by this
	 * lock. Exception: if q->lock is set, then the streaming ioctls
	 * are serialized by that separate lock.
	 */
	vdev->lock = &hdev->lock;
	vdev->v4l2_dev = &hdev->v4l2_dev;
	video_set_drvdata(vdev, hdev);

	ret = video_register_device(vdev, VFL_TYPE_GRABBER, -1);
	if (ret)
		goto fail_v4l2_dev;

	hdev->phy_sd = find_and_register_subdev(hdev, "img,hdmiphy");
	if (hdev->phy_sd)
		dev_info(dev, "\n found the subdev");
	else {
		ret = -EPROBE_DEFER;
		goto fail_vdev;
	}
	hdev->cur_conf_dirty = 5;
	v4l2_set_subdev_hostdata(hdev->phy_sd, hdev->res.config_clk);

	pm_runtime_enable(dev);
	if (!pm_runtime_enabled(dev)) {
		ret = hdmi_rx_runtime_resume(dev);
		if (ret)
			goto pm_rt_disable;
	}
	ret = pm_runtime_get_sync(dev);
	if (ret < 0)
		goto pm_rt_suspend;


	/* reset and disable interrupts. At power up you need to
	 * assert ipon_rst_n asynchronous reset(P57 of user guide) */
	hdev->rst = devm_reset_control_get(hdev->dev, "rst");
	interrupts_cfg(hdev, false, true);
	/* workqueue. Remove WQ_HIGHPRI if not needed */
	hdev->wq = alloc_workqueue(HDMI_RX_DRIVER_NAME, WQ_MEM_RECLAIM |
				   WQ_HIGHPRI | WQ_FREEZABLE, 0);
	if (!hdev->wq) {
		ret = -ENOMEM;
		goto fail_gpio_irq;
	}
	INIT_WORK(&hdev->wks, wq_handler);
	/* disable hdcp */
	hdmi_write(hdev, RA_DMI_HDCP_CTRL, 0x0);
	hdev->gpiodesc = gpiod_get(hdev->dev, "5v_detect", GPIOD_IN);
	if (!hdev->gpiodesc) {
		dev_err(hdev->dev, "gpiodesc is null.");
	} else {
		irq_gpio_p_detect = gpiod_to_irq(hdev->gpiodesc);
		/*interrutp must come at both rising/falling edges of gpio */
		if (irq_gpio_p_detect < 0)
			goto fail_gpio_irq;
		ret = devm_request_irq(hdev->dev, irq_gpio_p_detect,
				       source_power_handler,
				       IRQF_TRIGGER_FALLING | IRQF_TRIGGER_RISING,
				       "hdmi_rx_source", hdev);
		if (ret) {
			dev_err(hdev->dev, "gpio irq request failed");
			goto fail_gpio_irq;
		}
	}
	hdmi_dumpregs(hdev, "probe");
	pm_runtime_put(dev);
	hdev->suspended_closed = true;
	/* hdmi_rx_ctrl_hdcp_config (hdev, hdcp); */

	/* add the attributes */
	for (i = 0; hdmi_rx_attrs[i]; ++i) {
		ret = device_create_file(dev, hdmi_rx_attrs[i]);
		if (ret)
			dev_err(hdev->dev, "Failed to add %s attr\n",
				hdmi_rx_attrs[i]->attr.name);
	}

	return 0;
fail_gpio_irq:
	if (hdev->gpiodesc)
		gpiod_put(hdev->gpiodesc);
	pm_runtime_put(dev);
pm_rt_suspend:
	if (!pm_runtime_enabled(dev))
		hdmi_rx_runtime_suspend(dev);
pm_rt_disable:
	pm_runtime_disable(dev);
fail_vdev:
	video_unregister_device(&hdev->vdev);
fail_hdl:
	v4l2_ctrl_handler_free(&hdev->ctrls.ctrl_handler);
fail_v4l2_dev:
	v4l2_device_unregister(&hdev->v4l2_dev);
fail_init:
	hdmi_resources_cleanup(hdev);
fail_free:
	devm_kfree(&pdev->dev, hdev);
fail:
	dev_err(dev, "probe failed\n");
	return ret;

}

static int hdmi_remove(struct platform_device *ofdev)
{
	struct v4l2_device *v4l2_dev = dev_get_drvdata(&ofdev->dev);
	struct hdmi_device *hdev = get_hdev(v4l2_dev);
	int i;

	/* remove the attributes */
	for (i = 0; hdmi_rx_attrs[i]; ++i)
		device_remove_file(hdev->dev, hdmi_rx_attrs[i]);

	/* TODO:shut down all current operations */

	flush_workqueue(hdev->wq);
	destroy_workqueue(hdev->wq);

	hdmi_resources_cleanup(hdev);
	gpiod_put(hdev->gpiodesc);
	v4l2_ctrl_handler_free(&hdev->ctrls.ctrl_handler);

	/* unregister phy sub-driver */
	v4l2_device_unregister_subdev(hdev->phy_sd);
	v4l2_device_unregister(v4l2_dev);
	video_unregister_device(&hdev->vdev);
	/* release all resources including clks, irq, iomap, ... */
	dev_info(hdev->dev, "remove successful\n");

	return 0;
}
#ifdef CONFIG_PM_SLEEP
static int hdmi_rx_suspend(struct device *dev)
{
	int ret = 0;
	struct hdmi_device *hdev = container_of(dev_get_drvdata(dev),
						struct hdmi_device, v4l2_dev);
	struct clk *clock = hdev->res.pll_clk;

	if (pm_runtime_status_suspended(hdev->dev)) {
		dev_info(dev, "%s: already suspended. return.\n", __func__);
		hdev->suspended_closed = true;
		return 0;
	}

	hdev->suspended_closed = false;
	/*
	 * hdmi is using audio pll. The pll source is tmds_clk.
	 * If other drivers that use audio pll resume earlier, they will enable
	 * the pll while the input to the pll is disconnected. The pll fails
	 * to lock in this situation. Change the source from tmds_clk to xtal.
	 */
	if (clk_get_parent(hdev->res.tmds_mux_clk) != hdev->res.xtal_mux_clk) {
		unsigned long rate = clk_get_rate(clock);

		clk_set_parent(hdev->res.tmds_mux_clk, hdev->res.xtal_mux_clk);
		ret = clk_set_rate(clock, rate);
	}

	ret = pm_runtime_force_suspend(hdev->dev);
	if (ret) {
		dev_err(hdev->dev, "%s: force_suspend failed.\n", __func__);
		return ret;
	}
	return ret;
}

static int hdmi_rx_resume(struct device *dev)
{
	int ret = 0;
	int cnt;
	struct hdmi_device *hdev = container_of(dev_get_drvdata(dev),
						struct hdmi_device, v4l2_dev);
	if (hdev->suspended_closed) {
		if (hdmi_rx_runtime_resume(hdev->dev)) {
			dev_err(hdev->dev, "%s: pm_runtime_get_sync failed\n",
				__func__);
		} else {
			interrupts_cfg(hdev, false, true);
			hdmi_write(hdev, RA_DMI_HDCP_CTRL, 0x0);
			hdmi_rx_runtime_suspend(hdev->dev);
		}

		return 0;
	}

	/* enable the clocks */
	ret = pm_runtime_force_resume(hdev->dev);
	if (ret < 0) {
		dev_info(hdev->dev, "%s: force_resume failed\n", __func__);
		return ret;
	}
	/* disable hdcp */
	hdmi_write(hdev, RA_DMI_HDCP_CTRL, 0x0);
	ret = config_phy_and_core(hdev);
	if (ret)
		return ret;
	/* get the latest status of the cable and pass it to the user */
	if (gpiod_get_value(hdev->gpiodesc))
		hdev->events[EVENT_CABLE_CONNECTED] = true;
	else
		hdev->events[EVENT_CABLE_DISCONNECTED] = true;
	queue_work(hdev->wq, &hdev->wks);

	hdev->status = OPENED;
	hdmi_write(hdev, RA_HDMI_IEN_SET, CLK_CHANGE | SCDCTMDSCFGCHANGE |
		   SCDCSCSTATUSCHANGE | SCDCCFGCHANGE);
	for (cnt = 0; cnt < TMOUT * 20; cnt++) {
		if (tmds_is_valid(hdev))
			break;
		usleep_range(10, 20);
	}
	if (cnt == TMOUT * 20)
		return -ETIME;

	return 0;
}
#endif /* CONFIG_PM_SLEEP */

static const struct of_device_id hdmi_driver_types[] = {
	{
		.compatible = HDMI_RX_DRIVER_NAME,
	}, {
		/* end node */
	}
};
static const struct dev_pm_ops hdmi_rx_pm_ops = {
	SET_SYSTEM_SLEEP_PM_OPS(hdmi_rx_suspend, hdmi_rx_resume)
	SET_RUNTIME_PM_OPS(hdmi_rx_runtime_suspend,
			   hdmi_rx_runtime_resume, NULL)
};

static struct platform_driver img_hdmi_driver = {
	.probe = hdmi_probe,
	.remove = hdmi_remove,
	.driver = {
		.name = HDMI_RX_DRIVER_NAME,
		.owner = THIS_MODULE,
		.pm = &hdmi_rx_pm_ops,
		.of_match_table = hdmi_driver_types
	}
};
/* Add this to enable autoloading as device tree is instantiated. */
MODULE_DEVICE_TABLE(of, hdmi_driver_types);

module_platform_driver(img_hdmi_driver);

