/*
 * Copyright 2016 Imagination Technologies.
 *
 * This program is free software; you may redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
#ifndef HDMI_DRV_H
#define HDMI_DRV_H
#include <linux/bug.h>
#include <linux/clk.h>
#include <linux/delay.h>
#include <linux/gpio/consumer.h>
#include <linux/interrupt.h>
#include <linux/io.h>
#include <linux/irq.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/pm_runtime.h>
#include <linux/slab.h>
#include <linux/v4l2-dv-timings.h>
#include <linux/workqueue.h>
#include <media/v4l2-common.h>
#include <media/v4l2-ctrls.h>
#include <media/v4l2-dev.h>
#include <media/v4l2-device.h>
#include <media/v4l2-event.h>
#include <media/v4l2-ioctl.h>
#include <media/v4l2-subdev.h>
#include <video/img_vpp.h>
#include "hdmi-rx-regs.h"

#define HDMI_RX_DRIVER_NAME "img,hdmi-rx"
#define IRQ_INVALID 0xFFFFFFFF
/** Mode detection clock minimum [Hz] */
#define MD_CLK_MIN				(10000000UL)
/** Mode detection clock maximum [Hz] */
#define MD_CLK_MAX				(50000000UL)
/** TMDS clock minimum [Hz] */
#define TMDS_CLK_MIN			(25000000UL)
/** TMDS clock maximum [Hz] */
#define TMDS_CLK_MAX			(600000000UL)
/** Configuration clock minimum [Hz] */
#define CFG_CLK_MIN				(10000000UL)
/** Configuration clock maximum [Hz] */
#define CFG_CLK_MAX				(160000000UL)

/** TMDS clock delta [kHz] */
#define TMDS_CLK_DELTA			(125)
/** Pixel clock minimum [kHz] */
#define PIXEL_CLK_MIN			0 /* TMDS_CLK_MIN */
/** Pixel clock maximum [kHz] */
#define PIXEL_CLK_MAX			TMDS_CLK_MAX
/** Horizontal resolution minimum */
#define HRESOLUTION_MIN			(320)
/** Horizontal resolution maximum */
#define HRESOLUTION_MAX			(4096)
/** Vertical resolution minimum */
#define VRESOLUTION_MIN			(240)
/** Vertical resolution maximum */
#define VRESOLUTION_MAX			(4455)
/** Refresh rate minimum [1/100 Hz] */
#define REFRESH_RATE_MIN		(100)
/** Refresh rate maximum [1/100 Hz] */
#define REFRESH_RATE_MAX		(25000)
#define V_MODE_STABLE	(50000)
#define TMOUT (50)

/**
 * @short HDMI RX controller packet reception
 */
enum hdmi_rx_ctrl_packet_t {
	/** Audio clock regeneration packet */
	hdmi_rx_ctrl_packet_ACR = 0,
	/** General control packet */
	hdmi_rx_ctrl_packet_GCP,
	/** Audio content protection packet */
	hdmi_rx_ctrl_packet_ACP,
	/** International standard recording code 1 packet */
	hdmi_rx_ctrl_packet_ISRC1,
	/** International standard recording code 2 packet */
	hdmi_rx_ctrl_packet_ISRC2,
	/** Gamut metadata packet */
	hdmi_rx_ctrl_packet_GAMUT,
	/** Vendor specific info frame packet */
	hdmi_rx_ctrl_packet_VSI,
	/** Auxiliary video info frame packet */
	hdmi_rx_ctrl_packet_AVI,
	/** Source product descriptor info frame packet */
	hdmi_rx_ctrl_packet_SPD,
	/** Audio info frame packet */
	hdmi_rx_ctrl_packet_AIF,
	/** MPEG source info frame packet */
	hdmi_rx_ctrl_packet_MPEGS,
	/** Sentinel value */
	hdmi_rx_ctrl_packet_cnt
};

/**
 * @short HDMI RX controller video parameters
 *
 * For Auxiliary Video InfoFrame (AVI) details see HDMI 1.4a section 8.2.2
 */
struct hdmi_rx_ctrl_video {
	/** DVI detection status: DVI (true) or HDMI (false) */
	bool dvi;
	/** Deep color mode: 24, 30, 36 or 48 [bits per pixel] */
	u32 deep_color_mode;

	/** Pixel clock frequency [Hz] */
	/* pixel_clk is related to tmds clock depending of color depth
	 * and is PREPCLK signal output of the PHY going to pixel_clk input
	 * of the controller.
	 */
	 /* See Figure 2-4 in page 59 of databook. */
	unsigned long pixel_clk; /* Pixel clock in Hz. */
	/** Refresh rate [0.01Hz] */
	unsigned long refresh_rate;
	/** Interlaced */
	bool interlaced;
	/** Vertical offset */
	u32 voffset;
	/** Vertical active in lines */
	/*
	 * Height of the active video frame in lines.
	 * For interlaced formats the height of the active video in each field
	 * is height/2.
	 */
	u32 vactive;
	/** Vertical total in lines */
	u32 vtotal;
	/** Vertical total in time(imodetec cycles)*/
	u32 vtot_clk;
	/** Vertical sync in time(imodetec cycles)*/
	u32 vs_clk;
	/** Horizontal offset */
	u32 hoffset;		/* hsync + hbackporch */
	/** Horizontal active */
	/* The same field can be used to show hsync,
	 we can then find the backporch from hoffset */
	u32 hactive;	/* Width of the active video in pixels. */
	/** Horizontal total in pixels*/
	/* hfrontporch = htotal - hoffset: Horizontal front porch in pixels */
	u32 htotal;
	/** Horizontal total in time(imodetec cycles) */
	u32 htot32_clk;
	/** Horizontal sync in time(imodetec cycles) */
	u32 hs_clk;
	/** AVI Y1-0, video format */
	u32 video_format;
	/** AVI A0, active format information present */
	u32 active_valid;
	/** AVI B1-0, bar valid information */
	u32 bar_valid;
	/** AVI S1-0, scan information */
	u32 scan_info;
	/** AVI C1-0, colorimetry information */
	u32 colorimetry;
	/** AVI M1-0, picture aspect ratio */
	u32 picture_ratio;
	/** AVI R3-0, active format aspect ratio */
	u32 active_ratio;
	/** AVI ITC, IT content present or not */
	u32 it_content;
	/** AVI CN1-0, IT content type */
	u32 it_content_type;
	/** AVI EC2-0, extended colorimetry */
	u32 ext_colorimetry;
	/** AVI Q1-0, RGB quantization range */
	u32 rgb_quant_range;
	/** AVI SC1-0, non-uniform scaling information */
	u32 n_uniform_scale;
	/** AVI VIC7-0, video mode identification code */
	u32 video_mode;
	/** AVI PR3-0, pixel repetition factor */
	u32 pixel_repetition;
	/** AVI, line number of end of top bar */
	u32 bar_end_top;
	/** AVI, line number of start of bottom bar. */
	u32 bar_start_bottom;
	/** AVI, pixel number of end of left bar */
	u32 bar_end_left;
	/** AVI, pixel number of start of right bar */
	u32 bar_start_right;
};
struct hdmi_rx_ctrl_audio {
	/** AIF CT3-0, coding type */
	u32 coding_type;
	/** AIF CC2-0, channel count */
	u32 channel_count;
	/** AIF SF2-0, sample freq converted to 32000->192000 (PDEC_AIF_PB0) */
	u32 sample_frequency;
	/** AIF SS1-0, sample size */
	u32 sample_size;
	/** AIF CTX, coding extension */
	u32 coding_extension;
	/** AIF CA7-0, channel/speaker allocation */
	u32 channel_allocation;
	/** AIF DM_INH, down-mix inhibit */
	u32 down_mix_inhibit;
	/** AIF LSV3-0, level shift value */
	u32 level_shift_value;
};
struct hdmi_pulse {
	u32 beg;
	u32 end;
};

struct hdmi_timings {
	struct hdmi_pulse hact;
	u32 hsyn_pol; /* 0 - high, 1 - low */
	struct hdmi_pulse hsyn;
	u32 interlaced;
	struct hdmi_pulse vact[2];
	u32 vsyn_pol; /* 0 - high, 1 - low */
	u32 vsyn_off;
	struct hdmi_pulse vsyn[2];
};

struct hdmi_resources {
	struct clk *tmds_clk;
	struct clk *mode_detect_clk;
	struct clk *audio_clk;
	struct clk *config_clk;
	struct clk *cec_clk;
	struct clk *esm_clk;
	struct clk *pll_clk;
	struct clk *tmds_mux_clk;
	struct clk *xtal_mux_clk;
	int regul_count;
};

struct debug_info {
	/** Debug status, audio FIFO reset count */
	u32 debug_audio_fifo_rst;
	/** Debug status, packet FIFO reset count */
	u32 debug_packet_fifo_rst;
	/** Debug status, IRQ handling count */
	u32 debug_irq_handling;
	/** Debug status, IRQ packet decoder count */
	u32 debug_irq_packet_decoder;
	/** Debug status, IRQ audio clock count */
	u32 debug_irq_audio_clock;
	/** Debug status, IRQ audio FIFO count */
	u32 debug_irq_audio_fifo;
	/** Debug status, IRQ video mode count */
	u32 debug_irq_video_mode;
	/** Debug status, IRQ HDMI count */
	u32 debug_irq_hdmi;
	/** Debug status, audio fifo underflow count */
	u32 debug_irq_audio_fifo_under;
	/** Debug status, audio fifo overflow count */
	u32 debug_irq_audio_fifo_over;
};
/**
 * @short HDMI RX controller events
 */
enum event_t {
	EVENT_CABLE_CONNECTED = 0,
	EVENT_CABLE_DISCONNECTED,
	/** TMDS clock change event */
	EVENT_CLOCK_CHANGE,
	/** Video status change event */
	EVENT_VIDEO_STATUS,
	/** Audio status change event */
	EVENT_AUDIO_STATUS,
	/** Packet reception event */
	EVENT_PACKET_RECEPTION,
	/** AKSV, Key Selection Vector, reception event */
	EVENT_AKSV_RECEPTION,
	/** TMDSVALID changed to valid or invalid */
	EVENT_PLL_LOCKED,
	/**  Packet fifo overflow or underflow */
	EVENT_PACKET_FIFO_ERROR,
	/** Sentinel value */
	EVENT_INVALID
};

struct hdmi_rx_ctrls {
	struct v4l2_ctrl_handler ctrl_handler;
	struct v4l2_ctrl *tmds_clk_ctrl;
	/* AUTO, LIMITED, FULL */
	struct v4l2_ctrl *rx_rgb_range_ctrl;
	/* CINEMA, PHOTO, GAME, GRAPHICS */
	struct v4l2_ctrl *it_content_type_ctrl;
	/*
	 * V4L2_COLORSPACE_REC709, V4L2_COLORSPACE_SMPTE170M,
	 * V4L2_COLORSPACE_DEFAULT
	 */
	struct v4l2_ctrl *color_space_ctrl;
	struct v4l2_ctrl *color_depth_ctrl;
	struct v4l2_ctrl *pixel_encoding_ctrl; /* video_format */
	/* We may use an array to pass all video paramaters but not for now */
	struct v4l2_ctrl *video_params_ctrl;
	struct v4l2_ctrl *aud_sample_freq_ctrl;
	struct v4l2_ctrl *aud_sample_size_ctrl;
};

struct hdmi_device {
	/** base address of HDMI registers */
	void __iomem *regs;
	/** HDMI interrupt */
	u32 irq;
	/** pointer to device parent */
	struct device *dev;
	/** V4L2 device structure */
	struct v4l2_device v4l2_dev;
	/** subdev of HDMIPHY interface */
	struct v4l2_subdev *phy_sd;
	struct debug_info dbg;
	struct mutex access_lock;
	/** configuration of current graphic mode */
	const struct hdmi_timings *cur_conf;
	/** flag indicating that timings are dirty */
	int cur_conf_dirty;
	/** current timings */
	struct v4l2_dv_timings cur_timings;
	/** other resources */
	struct hdmi_resources res;
	struct video_device vdev;
	struct mutex lock;
	/** TDMS clock frequency [Hz] */
	unsigned long tmds_clk_rate;
	unsigned long tmds_clk_rate_from_aud;
	int status;
	bool suspended_closed;
	struct hdmi_rx_ctrl_video video_params;
	struct hdmi_rx_ctrl_audio audio_params;
	struct gpio_desc *gpiodesc;
	struct hdmi_rx_ctrls ctrls;
	struct reset_control *rst;
	bool events[EVENT_INVALID];
	struct workqueue_struct *wq;
	struct work_struct wks;
	bool mhl;
};

/* HDMI 2.0 features - error counting */
typedef struct {
	uint32_t errorCh0;
	uint32_t errorCh1;
	uint32_t errorCh2;
	uint8_t validCnt;
} error_cnt_t;

/* error counting and equalizer settings */
typedef struct {
	uint32_t error_ch_count;
	uint8_t ch_bestEqSetting;
} eqResult_t;
extern int hdmi_rx_ctrl_register_ctrls(struct hdmi_device *hdev);
/* HDMI 2.0 features - SCDC config*/
void hdmi_rx_scdc_initConfig(struct hdmi_device *hdev);
/* HDMI 2.0 features - Scrambling enable*/
void hdmi_rx_scdc_scrambling(struct hdmi_device *hdev, uint8_t enable);

error_cnt_t hdmi_rx_scdc_errorCnt(struct hdmi_device *hdev, uint8_t verbose, uint8_t check_valid);

void hdmi_rx_scdc_errorCntConfig(struct hdmi_device *hdev, uint8_t enable, uint8_t verbose);

void hdmi_rx_scdc_errorCntConfig_EnableSpecificCounter(struct hdmi_device *hdev, uint8_t enable, uint8_t counters, uint8_t verbose);
#endif
