/*
 * Copyright 2016 Imagination Technologies. All rights reserved.
 *
 * This program is free software; you may redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <linux/module.h>
#include <linux/slab.h>
#include <linux/clk.h>
#include <linux/io.h>
#include <linux/interrupt.h>
#include <linux/irq.h>
#include <linux/err.h>
#include <linux/platform_device.h>
#include <media/v4l2-subdev.h>
#include <media/v4l2-common.h>
#include <media/v4l2-dev.h>
#include <media/v4l2-device.h>
#include <media/v4l2-dv-timings.h>
#include <media/v4l2-ctrls.h>
#include <linux/regmap.h>
#include <linux/mfd/syscon.h>
#include <linux/delay.h>
#include <video/img_hdmi_rx.h>

#define IMG_HDMI_PHY_NAME "img,hdmiphy"
enum {
	STATE_CLOSED = 0,
	STATE_OPENED,
	STATE_CONFIGURED,
};
#define MSK(m, n) (((1 << (m)) - 1) << (n))
/** Register address: SNPS PHY GEN3 control - starting version 1.30a */
#define RA_SNPS_PHYG3_CTRL	(0x2C0UL)
#define PHYRESET BIT(0)
#define PHYPDDQ BIT(1)
#define PORTSELECT  MSK(2, 2)
#define CFGCLKFREQ MSK(2, 4)
#define PHYSVSRETMODEZ BIT(6)

/* I2C registers from the controller. starting version 1.30a */
/** Register address:  I2C Master: slave address */
#define RA_I2CM_PHYG3_SLAVE		(0x2C4UL)
/** Register address: I2C Master: register address*/
#define RA_I2CM_PHYG3_ADDRESS		(0x2C8UL)
/** Register address: I2C Master: data to write to slave*/
#define RA_I2CM_PHYG3_DATAO		(0x2CCUL)
/** Register address: I2C Master: data read from slave*/
#define RA_I2CM_PHYG3_DATAI		(0x2D0UL)
/** Register address: I2C Master: operation RD/WR  */
#define RA_I2CM_PHYG3_OPERATION		(0x2D4UL)
/** Register address: I2C Master: SS/HS mode  */
#define RA_I2CM_PHYG3_MODE			(0x2D8UL)
#define I2CMP_FASTMODE BIT(0)
/** Register address: I2C Master: soft reset */
#define RA_I2CM_PHYG3_SOFTRST		(0x2DCUL)
#define I2CMP_SOFTRST BIT(0)
/** Register address: I2C Master: ss mode counters */
#define RA_I2CM_PHYG3_SS_CNTS		(0x2E0UL)
/** Register address:I2C Master:  hs mode counters  */
#define RA_I2CM_PHYG3_FS_HCNT		(0x2E4UL)
#define I2CMP_FS_SCL_HCNT MSK(16, 16)
#define I2CMP_FS_SCL_LCNT MSK(16, 0)
/*
 * HDMI Interrupt registers
 */
/** Register address: HDMI interrupt clear enable */
#define RA_HDMI_IEN_CLR		(0xFD8UL)
/** Register address: HDMI interrupt set enable */
#define RA_HDMI_IEN_SET		(0xFDCUL)
/** Register address: HDMI interrupt status */
#define RA_HDMI_ISTS		(0xFE0UL)
/** Register address: HDMI interrupt enable */
#define RA_HDMI_IEN		(0xFE4UL)
/** Register address: HDMI interrupt clear status */
#define RA_HDMI_ICLR		(0xFE8UL)
/** Register address: HDMI interrupt set status */
#define RA_HDMI_ISET		(0xFECUL)

#define RA_SCDC_REGS0		(0x820UL)
/*
 * PHY registers within the controller
 */
/** HPD and PHY target state control */
#define GEN2_REG_HDMI_SETUP_CTRL		(0x000UL)
/** PHY override configurations */
#define GEN2_REG_HDMI_OVR_CTRL		(0x004UL)
/** PHY register to store the number of CFG_CLK cycles
 * that occur within a 4-us period. */
#define GEN2_REG_HDMI_TIMER_CTRL		(0x008UL)
/** PHY configurations override */
#define GEN2_REG_HDMI_RES_OVR		(0x010UL)
/** PHY PLL general configurations */
#define GEN2_REG_HDMI_PLL_CTRL		(0x018UL)
/** PLL sub ranges 1 */
#define GEN2_REG_HDMI_PLL_FRQSET1	(0x01CUL)
/** PLL sub ranges 2 */
#define GEN2_REG_HDMI_PLL_FRQSET2	(0x020UL)
/** PLL coefficients (Integral & Proportional Charge Pumps) 1 */
#define GEN2_REG_HDMI_PLL_PAR1		(0x024UL)
/** PLL coefficients (Integral & Proportional Charge Pumps) 2 */
#define GEN2_REG_HDMI_PLL_PAR2		(0x028UL)
/** PLL coefficients (Oscillator gains and CCO Low Pass filter) */
#define GEN2_REG_HDMI_PLL_PAR3		(0x02CUL)
/** hysteresis of clock input and active edge selection of PLL clock
 * to clock out data to DWC_hdmi_rx */
#define GEN2_REG_HDMI_CLK_CTRL		(0x034UL)
/** clock inversions and input selection */
#define GEN2_REG_HDMI_PCB_CTRL		(0x038UL)
/** Equalizer offset register */
#define GEN2_REG_HDMI_EQOFF_CTRL		(0x04CUL)
/** Equalizer gain control */
#define GEN2_REG_HDMI_EQGAIN_CTRL	(0x050UL)
/**  HDMI frequency change detection time period */
#define GEN2_REG_HDMI_CKM_EVLTM		(0x094UL)
/** default values according to PHY macro */
/*HDMIRX_PHY_TSMC_65GP_2_5V*/ /* E201 */
#define E201_EQ_RSV_DEFAULT			(0x11UL)
#define E201_INV_ALIGNCLOCK_DEFAULT	(0x1UL)


/*HDMIRX_PHY_TSMC_40LP_2_5V*/ /* E202 */
#define E202_EQ_RSV_DEFAULT			(0x31UL)
#define E202_INV_ALIGNCLOCK_DEFAULT	(0x3UL)

/*HDMIRX_PHY_TSMC_65LP_2_5V*/ /* E203 */
#define E203_EQ_RSV_DEFAULT			(0x1CUL)
#define E203_INV_ALIGNCLOCK_DEFAULT	(0x1UL)

/* HDMIRX_PHY_TSMC_55LP_2_5V*/ /* E206 */
#define E206_EQ_RSV_DEFAULT			(0x11UL)
#define E206_INV_ALIGNCLOCK_DEFAULT	(0x1UL)

/* HDMIRX_PHY_TSMC_55GP_2_5V*/ /* E207 */
#define E207_EQ_RSV_DEFAULT			(0x1CUL)
#define E207_INV_ALIGNCLOCK_DEFAULT	(0x1UL)

/* PLL configuration values */
/* HDMIRX_PHY_TSMC_55GP_2_5V*/ /* E207 */
#define E207_HDMI_PLL_PAR1_DEFAULT		(0xFCFBFF69UL)
#define E207_HDMI_PLL_PAR2_DEFAULT		(0x6F6DAFBFUL)
#define E207_HDMI_PLL_PAR3_DEFAULT		(0x7EDCA998UL)
/*all other SNPS PHYs - E201, E202, E203 and E206 */
#define E201_E206_HDMI_PLL_PAR1_DEFAULT		(0xF8F8FBC6UL)
#define E201_E206_HDMI_PLL_PAR2_DEFAULT		(0x6A6AABBBUL)
#define E201_E206_HDMI_PLL_PAR3_DEFAULT		(0x76DCA998UL)

/* common default values for PHYs */
#define GEN2_HDMI_PLL_FRQSET1_DEFAULT	(0x72523121UL)
#define GEN2_HDMI_PLL_FRQSET2_DEFAULT	(0x40D48A93UL)
#define GEN2_HDMI_SETUP_CTRL_DEFAULT		(0x03000C1BUL)
#define GEN2_HDMI_CLK_CTRL_DEFAULT		(0x10UL)

/** register address: PHY Gen3 clock measurement unit configuration */
#define PHY_CMU_CONFIG			(0x02UL)
#define LOCK_THRES MSK(6, 10)
#define TIMEBASE_OVR_EN BIT(9)
#define TIMEBASE_OVR MSK(9, 0)
/** register address: PHY Gen3 system configuration */
#define PHY_SYSTEM_CONFIG		(0x03UL)
/** register address: PHY Gen3 main FSM Override 2 */
#define PHY_MAINFSM_OVR2	(0x08UL)
/** register address: PHY Gen3 main FSM status 1 */
#define PHY_MAINFSM_STATUS1	(0x09UL)

#define PHY_CDR_CTRL_CNT	(0x0E)

/** register address: PHY Gen3 DAC control */
#define PHY_DAC_CTRL		(0x12UL)
/** register address: PHY Gen3 Clock Channel control */
#define PHY_CLK_CH_CTRL		(0x20UL)
/** register address: PHY Gen3 Analog Test Bus Selection */
#define PHY_SUP_ATB_SELECT	(0x21UL)
/** register address: PHY Gen3 Analog Test Bus Selection and
 * Slicer/Driver Voltage Threshold */
#define PHY_REC_VOLTAGE_LEVEL	(0x22UL)
/** MPLL Transfer Function Parameters 1 */
#define PHY_MPLL_PARAMETERS1	(0x26UL)
/** MPLL Transfer Function Parameters 2 */
#define PHY_MPLL_PARAMETERS2	(0x27UL)
/** MPLL Transfer Function Parameters 3 */
#define PHY_MPLL_PARAMETERS3	(0x28UL)
/** MPLL Transfer Function Parameters 4 */
#define PHY_MPLL_PARAMETERS4	(0x29UL)
/** MPLL Transfer Function Parameters 5 */
#define PHY_MPLL_PARAMETERS5	(0x2AUL)
/** MPLL Transfer Function Parameters 6 */
#define PHY_MPLL_PARAMETERS6	(0x2BUL)
/** MPLL Transfer Function Parameters 7 */
#define PHY_MPLL_PARAMETERS7	(0x2CUL)
/** MPLL Transfer Function Parameters 8 */
#define PHY_MPLL_PARAMETERS8	(0x2DUL)
/** MPLL Transfer Function Parameters 9 */
#define PHY_MPLL_PARAMETERS9	(0x2EUL)
/** MPLL Transfer Function Parameters 10 */
#define PHY_MPLL_PARAMETERS10	(0xC0UL)
/** MPLL Transfer Function Parameters 11 */
#define PHY_MPLL_PARAMETERS11	(0xC1UL)
/** MPLL Transfer Function Parameters 12 */
#define PHY_MPLL_PARAMETERS12	(0xC2UL)
/** MPLL Transfer Function Parameters 13 */
#define PHY_MPLL_PARAMETERS13	(0xC3UL)
/** MPLL Transfer Function Parameters 14 */
#define PHY_MPLL_PARAMETERS14	(0xC4UL)
/** MPLL Transfer Function Parameters 15 */
#define PHY_MPLL_PARAMETERS15	(0xC5UL)
/** MPLL Transfer Function Parameters 16 */
#define PHY_MPLL_PARAMETERS16	(0xC6UL)
/** MPLL Transfer Function Parameters 17 */
#define PHY_MPLL_PARAMETERS17	(0xC7UL)
/** MPLL Transfer Function Parameters 18 */
#define PHY_MPLL_PARAMETERS18	(0xC8UL)
/** MPLL Transfer Function Parameters 14 */
#define PHY_MPLL_PARAMETERS19	(0xC9UL)
/** MPLL Transfer Function Parameters 15 */
#define PHY_MPLL_PARAMETERS20	(0xCAUL)
/** MPLL Transfer Function Parameters 16 */
#define PHY_MPLL_PARAMETERS21	(0xCBUL)

/** register address: PHY Gen3 Channel 0 Analog Test Bus Control 1 */
#define PHY_CH0_ATB_CTRL1		(0x3AUL)
/** register address: PHY Gen3 Channel 1 Analog Test Bus Control 1 */
#define PHY_CH1_ATB_CTRL1		(0x5AUL)
/** register address: PHY Gen3 Channel 2 Analog Test Bus Control 1 */
#define PHY_CH2_ATB_CTRL1		(0x7AUL)
/** register overload protections **/
#define PHY_OVL_PROT_CTRL      (0x0DUL)
/** PHY Gen3 Clock measurement lock threshold - default 8*/
#define GEN3_LOCK_THRES	(0x08UL)
/* active write wait to the zcal done */
#define MIN_WRITE_WAIT					(50000)
/** I2C operation max polling wait */
#ifdef __HDMI_RX_TARGET_IS_SIMULATOR__
	#define MAX_IIC_WAIT			(1)
#else
	#define MAX_IIC_WAIT			(8000)
#endif

/** Register address: input port configuration */
#define GEN2_TCHIP_RA_INPUT_PORT			(0x1FUL)
/** Register address: latch enable flag */
#define GEN2_TCHIP_RA_LATCH_EN				(0xFFUL)

/* #else // w/out TCHIP */
/** PLL range when less than 40500kHz */
#define GEN2_PLL_LESS_THAN_40500kHz	(40500UL)
/** PLL range when less than 74500kHz */
#define GEN2_PLL_LESS_THAN_74500kHz	(74500UL)
/** PLL range when less than 111375kHz */
#define GEN2_PLL_LESS_THAN_111375kHz	(111375UL)
/** PLL range when less than 148500kHz */
#define GEN2_PLL_LESS_THAN_148500kHz	(148500UL)
/** PLL range when less than 222750kHz */
#define GEN2_PLL_LESS_THAN_222750kHz	(222750UL)
/** PLL range when less than 297000kHz */
#define GEN2_PLL_LESS_THAN_297000kHz	(297000UL)
/** PLL range when more than 297000kHz */
#define GEN2_PLL_MORE_THAN_297000kHz	(340000UL)

/* In Cashew the connection is point to point so any value is OK */
#define PHY_GEN3_I2C_SLAVE_ADDR			(0x69UL)
#define PHY_GEN3_GLUE_I2C_SLAVE_ADDR	(0x48UL)


/** Configuration clock minimum [Hz] */
#define CFG_CLK_MIN				(10000000UL)
/** Configuration clock maximum [Hz] */
#define CFG_CLK_MAX				(160000000UL)
/** TMDS clock minimum [Hz] */
#define TMDS_CLK_MIN			(25000000UL)
/** TMDS clock maximum [Hz] */
#define TMDS_CLK_MAX			(600000000UL)
/* ***************************************************************************
 * Data Manipulation and Access
 * ***************************************************************************/
/**
 * Find first (least significant) bit set
 * @param[in] data word to search
 * @return bit position or 32 if none is set
 */
static unsigned first_bit_set(u32 data)
{
	if (data)
		return __ffs(data);
	pr_err("%s: input is zero", __func__);
	return 32;
}

/*
 * Get bit field
 * @param[in] data raw data
 * @param[in] mask bit field mask
 * @return bit field value
 */
static u32 get(u32 data, u32 mask)
{
	return (data & mask) >> first_bit_set(mask);
}

/*
 * Set bit field
 * @param[in] data raw data
 * @param[in] mask bit field mask
 * @param[in] value new value
 * @return new raw data
 */
static u32 set(u32 data, u32 mask, u32 value)
{
	return ((value << first_bit_set(mask)) & mask) | (data & ~mask);
}

struct hdmiphy_conf {
	unsigned long pixclk;
	const u8 *data;
};

struct hdmiphy_ctx {
	/** pointer to device parent */
	struct device *dev;
	/** subdev generated by HDMI PHY device */
	struct v4l2_subdev sd;
	/** V4L2 device structure */
	struct v4l2_device v4l2_dev;
	struct v4l2_ctrl_handler ctrl_handler;
	struct v4l2_ctrl *color_depth_ctrl;
	struct regmap *hdmi_ctrl_regs;
	int version;
	int generation;
	int status;
	/* Config clock frequency [kHz], valid range 10MHz to 160MHz */
	u32 cfg_clk_rate;
	/** Peaking configuration */
	u32 peaking;
	/** PLL configuration */
	u32 pll_cfg;
};

/*
 * Read bit field from device
 * @param[in,out] ctx context information
 * @param[in] addr bus address
 * @param[in] mask bitfield mask
 * @return bit field value.
 * eg. If mask is 0x00F0 and the register value is 0xE49D, returns 0x9
 */
static u32 __maybe_unused ctrl_read_mask(struct hdmiphy_ctx *ctx,
					 u32 addr, u32 mask)
{
	u32 val;

	WARN_ON(regmap_read(ctx->hdmi_ctrl_regs, addr, &val));
	return get(val, mask);
}

/* Write to controller registers not to phy registers via I2C */
static inline void ctrl_write_mask(struct hdmiphy_ctx *ctx, u32 addr,
				   u32 mask, u32 value)
{
	regmap_update_bits(ctx->hdmi_ctrl_regs, addr, mask,
			   value << first_bit_set(mask));
}
/* Read from "reg" index of the hdmi controller. */
static inline u32 ctrl_read(struct hdmiphy_ctx *ctx, u32 reg)
{
	u32 val;

	WARN_ON(regmap_read(ctx->hdmi_ctrl_regs, reg, &val));
	return val;
}

/* Write to the "reg" index of the hdmi controller. */
static inline void ctrl_write(struct hdmiphy_ctx *ctx, u32 reg, u32 val)
{
	WARN_ON(regmap_write(ctx->hdmi_ctrl_regs, reg, val));
}

/* Wait for i2c interrupt signalling the read/write op completion. */
static int wait_for_i2c_int(struct hdmiphy_ctx *ctx, u8 slave_address)
{
	int i;
	u32 ists;
	int error = -EIO;

	for (i = 0; i < MAX_IIC_WAIT; i++) {
		usleep_range(10, 20);
		/* read i2c interrupt status - whether done or error */
		ists = ctrl_read_mask(ctx, RA_HDMI_ISTS, MSK(3, 28));
		if (ists == 0) {
			continue;
		} else if (ists == 1) {
			error = 0;
			break;
		} else {
			WARN(1, "i2c failed. interrupt_sts = 0x%X\n", ists);
			break;
		}
	}
	/* clear i2c interrupts */
	ctrl_write(ctx, RA_HDMI_ICLR, set(ctrl_read(ctx, RA_HDMI_ISTS),
		   MSK(3, 28), 0x7));

#ifdef __HDMI_RX_TARGET_IS_SIMULATOR__
	return 0;
#else
	return error;
#endif
}

/**
 * PHY GEN 3 read function through internal I2C Master controller
 * @param[in,out] ctx context information
 * @param [in] slave_address of peripheral/phy
 * @param [in] reg_address to read from
 * @return 16-bit read data or error
 */
u16 hdmi_rx_phy_i2c_read(struct hdmiphy_ctx *ctx, u8 slave_address,
			 u8 reg_address)
{
	int error = -EIO;

	ctrl_write(ctx, RA_I2CM_PHYG3_SLAVE, slave_address);
	ctrl_write(ctx, RA_I2CM_PHYG3_ADDRESS, reg_address);
	ctrl_write(ctx, RA_I2CM_PHYG3_OPERATION, 0x02); /* read op */
	error = wait_for_i2c_int(ctx, slave_address);
	if (error)
		return error;
	return (u16)ctrl_read(ctx, RA_I2CM_PHYG3_DATAI);
}

/*
 * PHY GEN 3 write function through internal I2C Master controller
 * @param[in,out] ctx context information
 * @param [in] slave_address of peripheral/phy
 * @param [in] reg_address to write to
 * @param [in] data 16-bit data to be written to peripheral
 * @return error
 */
int hdmi_rx_phy_i2c_write(struct hdmiphy_ctx *ctx, u8 slave_address,
			  u8 reg_address, u16 data)
{
	u32 ists;
	int error = -EINVAL;

	ctrl_write(ctx, RA_HDMI_ICLR, set(ctrl_read(ctx, RA_HDMI_ISTS),
		   MSK(3, 28), 0x7));
	/* Make sure that i2c interrupt is cleared */
	ists = ctrl_read(ctx, RA_HDMI_ISTS);
	if ((ists & MSK(3, 28)) > 0)
		dev_err(ctx->dev, "i2c interrupt status is not clear.\n");
	ctrl_write(ctx, RA_I2CM_PHYG3_SLAVE, slave_address);
	ctrl_write(ctx, RA_I2CM_PHYG3_ADDRESS, reg_address);
	ctrl_write(ctx, RA_I2CM_PHYG3_DATAO, data);
	ctrl_write(ctx, RA_I2CM_PHYG3_OPERATION, 0x01); /* write op */
	error = wait_for_i2c_int(ctx, slave_address);

	if (error)
		dev_info(ctx->dev, "...I2C write: failed...");

	return error;
}

static inline struct hdmiphy_ctx *sd_to_ctx(struct v4l2_subdev *sd)
{
	return container_of(sd, struct hdmiphy_ctx, sd);
}

static u32 __maybe_unused phy_wrapper_read(struct hdmiphy_ctx *ctx, u8 addr)
{
	return hdmi_rx_phy_i2c_read(ctx, PHY_GEN3_I2C_SLAVE_ADDR, addr);
}

static void phy_wrapper_write(struct hdmiphy_ctx *ctx, u8 reg_address,
			      u16 data)
{
	hdmi_rx_phy_i2c_write(ctx, PHY_GEN3_I2C_SLAVE_ADDR, reg_address, data);
}

/*
 * PDDQ powers down all analog blocks and resets digital circuitry's data path
 * If digital circuitry's configuration is also need a reset, use PHY_RESET.
 * To go to configuration mode, write 1'b0 to PHY_RESET. It does not power up
 * the phy. To make it ready for power up write 1'b0 to PDDQ. PHY will power
 * up when tmds data starts coming. When it is up, the output tmdsvalid signal
 * will rise. (see figure 11-1 of databook.)
 */
static void phy_wrapper_pddq(struct hdmiphy_ctx *ctx, int enable)
{
	ctrl_write_mask(ctx, RA_SNPS_PHYG3_CTRL, PHYPDDQ, enable);
}

static void phy_wrapper_svsretmode(struct hdmiphy_ctx *ctx, int enable)
{
	ctrl_write_mask(ctx, RA_SNPS_PHYG3_CTRL, PHYSVSRETMODEZ,
			enable);
}

static void phy_wrapper_phy_reset(struct hdmiphy_ctx *ctx, bool enable)
{
	ctrl_write_mask(ctx, RA_SNPS_PHYG3_CTRL, PHYRESET,
			enable ? 1 : 0);

}
/*s_power: puts subdevice in power saving mode (on == 0) or normal operation
 mode (on == 1).
 */
static int hdmiphy_s_power(struct v4l2_subdev *sd, int on)
{

	struct hdmiphy_ctx *ctx = sd_to_ctx(sd);
	struct clk *config = (struct clk *)v4l2_get_subdev_hostdata(sd);

	dev_info(ctx->dev, "\nstarted %u cfg=%lu\n",
		 ctx->cfg_clk_rate, clk_get_rate(config));

	/* to be implemented */
	return 0;
}

static void go_to_power_down_mode(struct hdmiphy_ctx *ctx)
{
	/*
	 * In power-down mode all analog blocks powered down and the digital
	 * logic is in a well-defined state. Set PHY_RESET to 1'b0 to enter the
	 * configuration mode. Sec 6.1 of Phy databook
	 */
	phy_wrapper_phy_reset(ctx, true);
	phy_wrapper_pddq(ctx, 1);
	phy_wrapper_svsretmode(ctx, 0);
}

static void go_to_config_mode(struct hdmiphy_ctx *ctx)
{
	/* Any configuration programmed on the HDMI 2.0 Rx PHY
	 * must be done in PHY Configuration mode. Sec 6.1 of Phy databook */
	phy_wrapper_svsretmode(ctx, 1);
	phy_wrapper_phy_reset(ctx, false);
	phy_wrapper_pddq(ctx, 1);
}

static void go_to_active_mode(struct hdmiphy_ctx *ctx)
{
	/* After configuration go to active mode. Sec 6.1 of Phy databook */
	phy_wrapper_phy_reset(ctx, false);
	phy_wrapper_svsretmode(ctx, 1);
	phy_wrapper_pddq(ctx, 0);
}

static int config_dcm(struct hdmiphy_ctx *ctx, int clr_depth)
{

	int dcm = clr_depth == 48 ? 3 : clr_depth / 6 - 4;


#ifdef __HDMI_RX_TARGET_IS_HARDWARE__
	if (ctx->status == STATE_CLOSED || dcm < 0 || dcm > 3 || clr_depth % 6)
		return -EINVAL;
#endif
	dev_info(ctx->dev, "\n%s: color_depth is %d dcm = %d\n",
			 __func__, clr_depth, dcm);
	ctx->status = STATE_OPENED;
	go_to_config_mode(ctx);
	/* force port select:Disable Equalization:calibration at lowest rate */
	phy_wrapper_write(ctx, PHY_SYSTEM_CONFIG, (dcm << 5));
	go_to_active_mode(ctx);
	ctx->status = STATE_CONFIGURED;

	return 0;
}

static int hdmi_dev_s_ctrl(struct v4l2_ctrl *ctrl)
{
	struct hdmiphy_ctx *ctx = container_of(ctrl->handler,
					    struct hdmiphy_ctx, ctrl_handler);

	switch (ctrl->id) {
	case V4L2_CID_DV_RX_PHY_COLOR_DEPTH:
		if (config_dcm(ctx, ctrl->val))
			return -EINVAL;
		break;
	default:
		dev_err(ctx->dev, "\nunknown ctrl id 0x%x", (u32)ctrl->id);
		return -EINVAL;
	}

	return 0;
}
static const struct v4l2_ctrl_ops ctx_ctrl_ops = {
	.s_ctrl = hdmi_dev_s_ctrl,
};

static const struct v4l2_ctrl_config h_ctrls[] = {
	{/* 0 */
		.ops = &ctx_ctrl_ops,
		.id = V4L2_CID_DV_RX_PHY_COLOR_DEPTH,
		.name = "phy color depth",
		.type = V4L2_CTRL_TYPE_INTEGER,
		.def = 24,
		.min = 24,
		.max = 48,
		.step = 6,
	},
};

static int config_i2c_master(struct hdmiphy_ctx *ctx, unsigned long irate)
{
	u64 min_scl_htime;
	u64 min_scl_ltime;
	u64 rate = (u64)irate;
	#define A_BILLION 1000000000ULL

	if (rate < CFG_CLK_MIN || rate >  CFG_CLK_MAX) {
		dev_err(ctx->dev, "cfg rate out of rang: %llu Hz", rate);
		return -EINVAL;
	}
	min_scl_htime = (u32)ctrl_read_mask(ctx, RA_I2CM_PHYG3_FS_HCNT,
					    I2CMP_FS_SCL_HCNT);
	min_scl_ltime = (u32)ctrl_read_mask(ctx, RA_I2CM_PHYG3_FS_HCNT,
					    I2CMP_FS_SCL_LCNT);

	/* Reset I2C master. Auto clear */
	ctrl_write_mask(ctx, RA_I2CM_PHYG3_SOFTRST, I2CMP_SOFTRST, 1);
	/* Set up the fast mode. See Appendix B of the core data book */
	min_scl_htime = div64_u64((rate * 600 + A_BILLION - 1), A_BILLION);
	min_scl_ltime = div64_u64((rate * 1300 + A_BILLION - 1), A_BILLION);
	if (min_scl_htime >= 6 && min_scl_ltime >= 8) {
		ctrl_write_mask(ctx, RA_I2CM_PHYG3_FS_HCNT, I2CMP_FS_SCL_HCNT,
				(u32)min_scl_htime);
		ctrl_write_mask(ctx, RA_I2CM_PHYG3_FS_HCNT, I2CMP_FS_SCL_LCNT,
				(u32)min_scl_ltime);
		/* use the fast mode. 400 kbis/sec */
		ctrl_write_mask(ctx, RA_I2CM_PHYG3_MODE, I2CMP_FASTMODE, 1);
	} else {
		/* fast mode not possible, set up the slow mode */
		min_scl_htime = div64_u64(rate * 4000 + A_BILLION - 1,
					  (u64)A_BILLION);
		min_scl_ltime = div64_u64(rate * 4700 + A_BILLION - 1,
					       (u64)A_BILLION);
		if (min_scl_htime >= 6 && min_scl_ltime >= 8) {
			ctrl_write_mask(ctx, RA_I2CM_PHYG3_SS_CNTS,
					I2CMP_FS_SCL_HCNT, (u32)min_scl_htime);
			ctrl_write_mask(ctx, RA_I2CM_PHYG3_SS_CNTS,
					I2CMP_FS_SCL_LCNT, (u32)min_scl_ltime);
			/* use the slow mode. 100 kbis/sec */
			ctrl_write_mask(ctx, RA_I2CM_PHYG3_MODE,
					I2CMP_FASTMODE, 0);
		} else {
			dev_err(ctx->dev, "min_scl_time out of rang");
			return -EINVAL;
		}
	}
	return 0;
}

static void __maybe_unused dump_regs(struct hdmiphy_ctx *ctx)
{
	u32 i;
	u32 reg;

	for (i = 0; i < 0x9A; i++) {
		reg =  phy_wrapper_read(ctx, i);
		dev_info(ctx->dev, "\t0x%08x : 0x%08X", i, reg);
	}
	for (i = 0xB0; i < 0xB7; i++) {
		reg =  phy_wrapper_read(ctx, i);
		dev_info(ctx->dev, "\t0x%08x : 0x%08X", i, reg);
	}
	for (i = 0xC0; i < 0xCD; i++) {
		reg =  phy_wrapper_read(ctx, i);
		dev_info(ctx->dev, "\t0x%08x : 0x%08X", i, reg);
	}
}
/**
 * Open device validating context information
 * @param[in,out] ctx context information
 * @return error code
 */
static int hdmi_rx_phy_open(struct v4l2_subdev *sd)
{
	int error = 0;
	struct clk *cfg;
	struct hdmiphy_ctx *ctx = sd_to_ctx(sd);
	/* (void)val; */
	u16 reg = 0;

	if (!sd) {
		dev_err(ctx->dev, "\nPHY OPEN: input is null.");
		return -EINVAL;
	}
	ctx = sd_to_ctx(sd);
	cfg = (struct clk *)v4l2_get_subdev_hostdata(sd);
	if (cfg)
		ctx->cfg_clk_rate = clk_get_rate(cfg);
	if (!cfg || ctx->cfg_clk_rate < CFG_CLK_MIN ||
	    ctx->cfg_clk_rate >  CFG_CLK_MAX) {
		dev_err(ctx->dev, "\n clk not valid");
		return -EINVAL;
	}

	pr_info("\nRX Phy: %d", ctx->version);
	/* This core configuration needed befor phy configuration. */
	error = config_i2c_master(ctx, ctx->cfg_clk_rate);
	if (error) {
		dev_err(ctx->dev, "could not configure i2c master.");
		return error;
	}
#ifdef __HDMI_RX_TARGET_IS_HARDWARE__
	if (ctx->status != STATE_CLOSED)
		pr_info("\nPHY already open.");
		/*return -EPERM;*/
#endif
	ctx->status = STATE_OPENED;
	/* Assuming that the cfg clk is 50MHz */
	ctrl_write_mask(ctx, RA_SNPS_PHYG3_CTRL,
			CFGCLKFREQ, 0x2);


	go_to_power_down_mode(ctx);
	/* switch to Configuration Mode. The same as power-down but internal
	 * register bank is active
	 */
	/* Any configuration programmed on the HDMI 2.0 Rx PHY
	 * must be done in PHY Configuration mode */
	go_to_config_mode(ctx);
	/* configure the clock measurement unit, which is used
	 * to measure the tmds clock based on the input cfg clock.
	 * CMU_Threshold- time_base[8:0] = int(CFG_CLK_FREQ in MHz * 4)
	 * bit 9 enables the override and LOCK_THRESH represents the
	 * the variation in the TMDS clock which triggers the CMU reaction.
	 * color depth field clrdep[1:0] remains 2'b00: 8bit mode. It will
	 * change when the input color depth is known. We will receive an
	 * interrupt signalling the "video change" if the clr_dep is incorrect.
	 */
	/* first set the time_base which is 4 usec */
	reg = DIV_ROUND_CLOSEST(ctx->cfg_clk_rate, 1000000) * 4;
	/* Set the override enable to make the values of this register valid */
	reg |= TIMEBASE_OVR_EN;
	/* lock _thershod. Defines the accepted variation in TMDS clk */
	reg |= GEN3_LOCK_THRES << __ffs(LOCK_THRES);
	phy_wrapper_write(ctx, PHY_CMU_CONFIG, reg);

	if (ctx->version == 401) {
		/* configure the MPLL parameters (proportional and integral
		 *  charge pumps) values obtained from characterization */
		phy_wrapper_write(ctx, PHY_MPLL_PARAMETERS2, 0x1C94);
		phy_wrapper_write(ctx, PHY_MPLL_PARAMETERS3, 0x3713);
		phy_wrapper_write(ctx, PHY_MPLL_PARAMETERS4, 0x24DA);
		phy_wrapper_write(ctx, PHY_MPLL_PARAMETERS5, 0x5492);
		phy_wrapper_write(ctx, PHY_MPLL_PARAMETERS6, 0x4B0D);
		phy_wrapper_write(ctx, PHY_MPLL_PARAMETERS7, 0x4760);
		phy_wrapper_write(ctx, PHY_MPLL_PARAMETERS8, 0x008C);
		phy_wrapper_write(ctx, PHY_MPLL_PARAMETERS9, 0x0002);
		phy_wrapper_write(ctx, PHY_MPLL_PARAMETERS10, 0x2D20);
		phy_wrapper_write(ctx, PHY_MPLL_PARAMETERS11, 0x2E31);
		phy_wrapper_write(ctx, PHY_MPLL_PARAMETERS12, 0x4B64);
		phy_wrapper_write(ctx, PHY_MPLL_PARAMETERS13, 0x2493);
		phy_wrapper_write(ctx, PHY_MPLL_PARAMETERS14, 0x676D);
		phy_wrapper_write(ctx, PHY_MPLL_PARAMETERS15, 0x23E0);
		phy_wrapper_write(ctx, PHY_MPLL_PARAMETERS16, 0x001B);
		phy_wrapper_write(ctx, PHY_MPLL_PARAMETERS17, 0x2218);
		phy_wrapper_write(ctx, PHY_MPLL_PARAMETERS18, 0x1B25);
		phy_wrapper_write(ctx, PHY_MPLL_PARAMETERS19, 0x2492);
		phy_wrapper_write(ctx, PHY_MPLL_PARAMETERS20, 0x48EA);
		phy_wrapper_write(ctx, PHY_MPLL_PARAMETERS21, 0x0011);
	}

	ctx->status = STATE_CONFIGURED;

	go_to_active_mode(ctx);
	dev_info(ctx->dev, "hdmi_rx_phy_open finished...");

	return error;
}

static int hdmi_rx_phy_close(struct v4l2_subdev *sd)
{

	struct hdmiphy_ctx *ctx = sd_to_ctx(sd);

	ctx->status = STATE_CLOSED;
	go_to_power_down_mode(ctx);
	return 0;
}



static int hdmiphy_s_stream(struct v4l2_subdev *sd, int enable)
{
	if (enable)
		return hdmi_rx_phy_open(sd);
	else
		return hdmi_rx_phy_close(sd);
}

static const struct v4l2_subdev_core_ops hdmiphy_core_ops = {
	.reset = NULL,
	.s_power =  hdmiphy_s_power,
};

static const struct v4l2_subdev_video_ops hdmiphy_video_ops = {
#if 0
	.s_dv_timings = hdmiphy_s_dv_timings,
#endif
	.s_stream =  hdmiphy_s_stream,
};

static const struct v4l2_subdev_pad_ops hdmiphy_pad_ops = {
#if 0
	.dv_timings_cap = hdmiphy_dv_timings_cap,
#endif
};

static const struct v4l2_subdev_ops hdmiphy_ops = {
	.core = &hdmiphy_core_ops,
	.video = &hdmiphy_video_ops,
	.pad = &hdmiphy_pad_ops,
};

static int hdmi_phy_probe(struct platform_device *pdev)
{
	struct hdmiphy_ctx *ctx = NULL;
	struct device *dev = &pdev->dev;

	pr_info("\n hdmiphy_probe started.\n");
	ctx = kzalloc(sizeof(*ctx), GFP_KERNEL);
	if (!ctx)
		return -ENOMEM;
	ctx->dev = dev;
	/* setting v4l2 name to prevent WARN_ON in v4l2_device_register */
	strlcpy(ctx->v4l2_dev.name, dev_name(dev),
		sizeof(ctx->v4l2_dev.name));
	v4l2_subdev_init(&ctx->sd, &hdmiphy_ops);
	ctx->sd.owner = THIS_MODULE;
	strlcpy(ctx->sd.name, IMG_HDMI_PHY_NAME, sizeof(ctx->sd.name));

	/* storing subdev for call that have only access to struct device */
	dev_set_drvdata(dev, &ctx->sd);
	/* No regmap for Phy, registers are accessed via core regs (i2c)*/
	ctx->hdmi_ctrl_regs =
		syscon_regmap_lookup_by_compatible("img,hdmi-rx");
	if (ctx->hdmi_ctrl_regs) {
		dev_info(dev, "got the core regmap\n");
	} else {
		dev_err(dev, "core regmap not found");
		goto no_regmap;
	}
	v4l2_ctrl_handler_init(&ctx->ctrl_handler, 5);
	ctx->sd.ctrl_handler = &ctx->ctrl_handler;
	ctx->color_depth_ctrl =
		v4l2_ctrl_new_custom(&ctx->ctrl_handler, &h_ctrls[0], NULL);
	ctx->version = 401;
	ctx->generation = 3;
	ctx->status = STATE_CLOSED;
	dev_info(dev, "\n%s probe successful\n", __func__);
	return 0;
no_regmap:
	kfree(ctx);
	return -EINVAL;
}

static int hdmi_phy_remove(struct platform_device *pdev)
{
	/* TODO: needs completion. */
	struct v4l2_device *v4l2_dev = dev_get_drvdata(&pdev->dev);
	struct hdmiphy_ctx *ctx = container_of(v4l2_dev, struct hdmiphy_ctx,
					       v4l2_dev);

	dev_info(ctx->dev, "remove successful\n");
	kfree(ctx);
	return 0;
}
static const struct of_device_id hdmiphy_id[] = {
			{ .compatible = IMG_HDMI_PHY_NAME },
		{ }
	};
#if 0
static struct platform_device_id hdmiphy_id[] = {
	{
		.name		= "img,hdmi-phy-syn",
	}, {
		.name		= IMG_HDMI_PHY_NAME,
	}, {
		.name		= "hdmi_rx_phy",
	}, {
		/* end node */
	}
};
#endif
MODULE_DEVICE_TABLE(of, hdmiphy_id);
static struct platform_driver hdmiphy_driver = {
	.probe = hdmi_phy_probe,
	.remove = hdmi_phy_remove,
	.driver = {
		.name = IMG_HDMI_PHY_NAME,
		.owner	= THIS_MODULE,
		.of_match_table = hdmiphy_id

	}
};
module_platform_driver(hdmiphy_driver);
