/*
 * This is a V4L2 VPP Driver.
 *
 * Copyright 2015 Imagination Technologies. All rights reserved.
 *
 * This program is free software; you may redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
#include <linux/clk.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/kmod.h>
#include <linux/mutex.h>
#include <linux/interrupt.h>
#include <linux/videodev2.h>
#include <linux/v4l2-dv-timings.h>
#include <media/v4l2-device.h>
#include <media/v4l2-dev.h>
#include <media/v4l2-ioctl.h>
#include <media/v4l2-dv-timings.h>
#include <media/v4l2-ctrls.h>
#include <media/v4l2-event.h>
#include <media/videobuf2-dma-sg.h>
#include <media/videobuf2-dma-contig.h>
#include <media/videobuf2-vmalloc.h>
#include <media/videobuf2-core.h>
#include <linux/clk-provider.h>
#include <linux/pm_runtime.h>
#define VPP 1
#define VPP_DRIVER_NAME "img,vpp-cashew"

#include <linux/of_address.h>
#include <linux/platform_device.h>
#include "vpp_api.h"
#include "vpp_mmu.h"
#include "vpp_ctrl.h"
#include "csc2/csc2.h"
#include "csc2/fixedpointmaths.h"
#include "csc2/csc2_data.h"
#include "validation.h"
#include "vpp_buf.h"
#define	IMG_FALSE		0
#define	IMG_TRUE		1
MODULE_DESCRIPTION("IMG VPP");
MODULE_AUTHOR("Sadegh Abbasi");
MODULE_LICENSE("GPL v2");
#ifdef VPP_DEBUG
static unsigned int module_reg_base;
#endif
/* Output format descriptions */
struct v4l2_fmtdesc vpp_allowed_formats[] = {
	{	.index = 0,
		.type  = V4L2_BUF_TYPE_VIDEO_CAPTURE_MPLANE,
		.description = "444 YUV101010",
		.pixelformat = V4L2_PIX_FMT_YUV30
	}, {	.index = 1,
		.type  = V4L2_BUF_TYPE_VIDEO_CAPTURE_MPLANE,
		.description = "422 UYVY10101010",
		.pixelformat = V4L2_PIX_FMT_UYVY20
	}, {	.index = 2,
		.type  = V4L2_BUF_TYPE_VIDEO_CAPTURE_MPLANE,
		.description = "PL12Y10-422PL12UV10",
		.pixelformat = V4L2_PIX_FMT_PLA422
	}, {	.index = 3,
		.type  = V4L2_BUF_TYPE_VIDEO_CAPTURE_MPLANE,
		.description = "PL12Y10-420PL12UV10",
		.pixelformat = V4L2_PIX_FMT_PLA420
	}, {	.index = 4,
		.type  = V4L2_BUF_TYPE_VIDEO_CAPTURE_MPLANE,
		.description = "PL12Y8-422PL12UV8",
		.pixelformat = V4L2_PIX_FMT_PL8422
	}, {	.index = 5,
		.type  = V4L2_BUF_TYPE_VIDEO_CAPTURE_MPLANE,
		.description = "PL12Y8-420PL12UV8",
		.pixelformat = V4L2_PIX_FMT_PL8420
	},
};

	static const struct of_device_id of_ids[] = {
		{ .compatible = VPP_DRIVER_NAME },
		{ }
	};
	MODULE_DEVICE_TABLE(of, of_ids);

/* VPP supports all the standards and the ranges for width and height */
static const struct v4l2_dv_timings_cap idev_timings_cap = {
	.type = V4L2_DV_BT_656_1120,
	/* keep this initialization for compatibility with GCC < 4.4.6 */
	.reserved = { 0 },

	V4L2_INIT_BT_TIMINGS(
		640, 4096,		/* min/max width */
		480, 2304,		/* min/max height */
		13500000, 594000000,	/* min/max pixelclock*/
		V4L2_DV_BT_STD_CEA861,	/* Supported standards */
		/* capabilities */
		V4L2_DV_BT_CAP_INTERLACED | V4L2_DV_BT_CAP_PROGRESSIVE
	)
};

static void vpp_api_turn_off(struct vpp_dev *idev)
{
	vpp_io_soft_reset(idev);
	vpp_io_turn_off_hw(idev);
}

/*
 * Supported SDTV standards. This does the same job as idev_timings_cap, but
 * for standard TV formats.
 */
#define SKEL_TVNORMS V4L2_STD_ALL

/*
 * Interrupt handler: typically interrupts happen after a new frame has been
 * captured. It is the job of the handler to remove the new frame from the
 * internal list and give it back to the vb2 framework, updating the sequence
 * counter, field and timestamp at the same time.
 */
static irqreturn_t vpp_dev_irq(int irq, void *dev_id)
{
	struct vpp_dev *idev = (struct vpp_dev *)dev_id;
	u32 intr_status;
	u32 intr_enable;
	bool kick_again = false;
	struct timeval time;
	struct idev_buffer *cur_buf = idev->current_buf;
	static int counter;
	int state;

	state = cur_buf->vb.state;
	trace_printk("interrupt started\n");
	/* In VPP we are only concerned with DIF Frame Ready interrupt
	 * and two overrun interrupts. We will disable all other interrupts.
	 * The application will be informed in
	 * vb2_buffer_done() function by a call to wake_up().
	 */
	PDEBUG("\nvpp_dev_irq:  got an interrupt irq = %d status=0x%X\n",
	       (u32)irq, (u32)vpp_io_read_irq_status(idev));
	/* if not streaming ignore interrupts */
	if (vb2_is_streaming(&idev->queue) == 0) {
		PDEBUG("\nvpp_dev_irq: not streaming yet, ignoring");
		vpp_io_clear_interrupt(idev, 0xFFFFFFFF);
		return IRQ_HANDLED;
	}

	intr_enable = vpp_io_read_irq_enable(idev);
	vpp_io_disable_interrupt(idev, ~0u);
	intr_status = vpp_io_read_irq_status(idev);
	intr_status &= intr_enable;

	/* Is it DIFFRAMERDY? */
	if (intr_status & VPP_DIF_IRQ_STATUS_DIFFRAMERDY_MASK) {
		PDEBUG("\nvpp_dev_irq: VPP_DIF_IRQ_STATUS_DIFFRAMERDY_MASK");
		intr_enable |= VPP_DIF_IRQ_ENABLE_DIFOVERRUNIE_CHROMA_MASK |
			       VPP_DIF_IRQ_ENABLE_DIFOVERRUNIE_Y_MASK;
		#ifdef __VPP_TARGET_IS_SIMULATOR__
			cur_buf->vb.v4l2_buf.reserved =
				vpp_io_read_signatures
					(idev, &cur_buf->vb.v4l2_buf.reserved2);
		#endif
		/* vb2_buffer_done() will inform the application that is waiting
		 * through select(), that the buffer is ready. Application then
		 * gets it by VIDIOC_DQBUF ioctl handled by video_buffer2.
		 * When the application finished with the buffer it returns the
		 * buffer back to the vb2 queue by a call to VIDIOC_QBUF.
		 * video_buffer2 will add the buffer to its own queue and also
		 * call the driver's buffer_queue() to give the driver
		 * a chance to return the buffer to its own queue.
		 */
			v4l2_get_timestamp(&cur_buf->vb.v4l2_buf.timestamp);
			/* Inform the application that the buffer is ready
			 * and it is time for it to get it with VIDIOC_DQBUF.
			 */
			idev->state_count[cur_buf->vb.state]++;
			trace_printk("state = %d\n", cur_buf->vb.state);
			if (cur_buf->vb.state == VB2_BUF_STATE_ACTIVE)
				vb2_buffer_done(&cur_buf->vb,
						idev->buffer_error ?
						VB2_BUF_STATE_ERROR :
						VB2_BUF_STATE_DONE);
			else if (counter++ < 2) {
				pr_err("%u.%06u:buf state not active: %d\n",
				       (u32) cur_buf->vb.v4l2_buf.timestamp.tv_sec,
				       (u32) cur_buf->vb.v4l2_buf.timestamp.tv_usec,
				       cur_buf->vb.state);
				}
			/* Schedule the task of writing the next buffer
			 * addresses to vpp registers and setting up the mmu
			 */
			if (!idev->buffer_error && idev->pm_state == PM_RUNNING &&
			    state == VB2_BUF_STATE_ACTIVE) {
				queue_work(idev->set_next_buffer_workqueue,
					   &idev->set_next_buffer_work);
				kick_again = true;
			}
	}
	/* Is it FIFO overrun? Data is coming but either not written to memory
	 * at all or is written slower than required.
	 */
	if ((intr_status & (VPP_DIF_IRQ_STATUS_DIFOVERRUN_CHROMA_MASK |
	    VPP_DIF_IRQ_STATUS_DIFOVERRUN_Y_MASK))) {
		intr_enable &= ~(VPP_DIF_IRQ_ENABLE_DIFOVERRUNIE_CHROMA_MASK |
				 VPP_DIF_IRQ_ENABLE_DIFOVERRUNIE_Y_MASK);
		v4l2_get_timestamp(&time);
		dev_info(idev->dev, "time=%u.%06u: FIFO overrun detected\n",
			 (u32)time.tv_sec, (u32)time.tv_usec);
	} else {
		PDEBUG("\nvpp_dev_irq: interrupt status is 0x%X",
		       vpp_io_read_irq_status(idev));
	}
	/* clear all interrupts */
	vpp_io_clear_interrupt(idev, 0xFFFFFFFF);

	intr_status = vpp_io_read_irq_status(idev);
	PDEBUG("\nvpp_dev_irq: returning from irq, status=0x%X", intr_status);
	/*enable interrupt */
	vpp_io_enable_interrupt(idev, intr_enable);
	/*invalidate cache*/
	vpp_io_mmu_invalidate_cache(idev);
#ifdef __VPP_TARGET_IS_SIMULATOR__
	if (kick_again && v4l2_ctrl_g_ctrl(idev->vp_kick_ctrl) == VPP_KICK_AUTO)
		vpp_io_kick_simulator(idev);
#endif
	trace_printk("interrupt finished\n");
	return IRQ_HANDLED;
}

static int vpp_dev_querycap(struct file *file, void *priv,
			    struct v4l2_capability *cap)
{
	strlcpy(cap->driver, KBUILD_MODNAME, sizeof(cap->driver));
	cap->device_caps = V4L2_CAP_VIDEO_CAPTURE | V4L2_CAP_READWRITE |
			   V4L2_CAP_STREAMING;
	cap->capabilities = cap->device_caps | V4L2_CAP_DEVICE_CAPS;
	return 0;
}

/*
 * Helper function to check and correct struct v4l2_pix_format. It's used
 * not only in VIDIOC_TRY/S_FMT, but also elsewhere if changes to the SDTV
 * standard, HDTV timings or the video input would require updating the
 * current format.
 */
static void vpp_dev_fill_pix_format(struct vpp_dev *idev,
				    struct v4l2_pix_format *pix)
{
	pix->pixelformat = V4L2_PIX_FMT_YUYV;
	if (idev->input == 0) {
		/* S-Video input */
		pix->width = 720;
		pix->height = (idev->std & V4L2_STD_525_60) ? 480 : 576;
		pix->field = V4L2_FIELD_INTERLACED;
		pix->colorspace = V4L2_COLORSPACE_SMPTE170M;
	} else {
		/* HDMI input */
		pix->width = idev->timings.bt.width;
		pix->height = idev->timings.bt.height;
		if (idev->timings.bt.interlaced) {
			pix->field = V4L2_FIELD_ALTERNATE;
			pix->height /= 2;
		} else {
			pix->field = V4L2_FIELD_NONE;
		}
		pix->colorspace = V4L2_COLORSPACE_REC709;
	}

	/*
	 * The YUYV format is four bytes for every two pixels, so bytesperline
	 * is width * 2.
	 */
	pix->bytesperline = pix->width * 2;
	pix->sizeimage = pix->bytesperline * pix->height;
	pix->priv = 0;
	PDEBUG("\n vpp_dev_fill_pix_format: pix->sizeimage= %d ",
	       pix->sizeimage);
}

static int vpp_dev_s_fmt_vid_cap_mplane(struct file *file, void *priv,
					struct v4l2_format *f)
{
	struct vpp_dev *idev = video_drvdata(file);

	PDEBUG("\n vpp_dev_s_fmt_vid_cap_mplane: Checking vb2 busy\n");
	/*
	 * It is not allowed to change the format while buffers for use with
	 * streaming have already been allocated.
	 */
	if (vb2_is_busy(&idev->queue))
		return -EBUSY;

	idev->format.fmt.pix_mp = f->fmt.pix_mp;
	idev->format.type = f->type;
	PDEBUG("%s: v4l2 format type is %d\n", __func__, idev->format.type);
	PDEBUG("\n vpp_dev_s_fmt_vid_cap_mplane: v4l2 format field is %d\n",
	       idev->format.fmt.pix_mp.field);

	return 0;
}

static int vpp_dev_g_fmt_vid_cap_mplane(struct file *file, void *priv,
					struct v4l2_format *f)
{
	struct vpp_dev *idev = video_drvdata(file);

	f->fmt.pix_mp = idev->format.fmt.pix_mp;
	
	return 0;
}

static int vpp_dev_enum_fmt_vid_cap_mplane(struct file *file, void *priv,
					   struct v4l2_fmtdesc *f)
{
	/* The application will enumerate the supported video formats by
	 * using this ioctl with index value that is increamented each time
	 * until -EINVAL is returned. The driver fills the structure with one
	 * of the supported formats for each index value.
	 * Note that for vpp driver, format means the vidoe format of the
	 * output. Input formats are called modes.
	 */
	if (f->index >= ARRAY_SIZE(vpp_allowed_formats))
		return -EINVAL;
	strlcpy(f->description, vpp_allowed_formats[f->index].description,
		sizeof(vpp_allowed_formats[f->index].description));
	f->pixelformat = vpp_allowed_formats[f->index].pixelformat;
	f->flags = 0;
	return 0;
}


static int vpp_dev_try_fmt_vid_cap(struct file *file, void *priv,
				   struct v4l2_format *f)
{
	/* We do not provide this function */
	struct vpp_dev *idev = video_drvdata(file);
	struct v4l2_pix_format *pix = &f->fmt.pix;

	/*
	 * Due to historical reasons providing try_fmt with an unsupported
	 * pixelformat will return -EINVAL for video receivers. Webcam drivers,
	 * however, will silently correct the pixelformat. Some video capture
	 * applications rely on this behavior...
	 */
	if (pix->pixelformat != V4L2_PIX_FMT_YUYV)
		return -EINVAL;
	vpp_dev_fill_pix_format(idev, pix);
	return 0;
}

static int vpp_dev_s_fmt_vid_cap(struct file *file, void *priv,
				 struct v4l2_format *f)
{
	struct vpp_dev *idev = video_drvdata(file);
	int ret = 0;

	ret = vpp_dev_try_fmt_vid_cap(file, priv, f);
	if (ret)
		return ret;

	/*
	 * It is not allowed to change the format while buffers for use with
	 * streaming have already been allocated.
	 */
	if (vb2_is_busy(&idev->queue))
		return -EBUSY;

	idev->format.fmt.pix = f->fmt.pix;
	return 0;
}

static int vpp_dev_g_fmt_vid_cap(struct file *file, void *priv,
				 struct v4l2_format *f)
{
	struct vpp_dev *idev = video_drvdata(file);

	f->fmt.pix = idev->format.fmt.pix;
	return 0;
}

static int vpp_dev_enum_fmt_vid_cap(struct file *file, void *priv,
				    struct v4l2_fmtdesc *f)
{
	if (f->index != 0)
		return -EINVAL;

	strlcpy(f->description, "4:2:2, packed, YUYV", sizeof(f->description));
	f->pixelformat = V4L2_PIX_FMT_YUYV;
	f->flags = 0;
	return 0;
}

static int vpp_dev_s_std(struct file *file, void *priv, v4l2_std_id std)
{
	struct vpp_dev *idev = video_drvdata(file);

	/* S_STD is not supported on the HDMI input */
	if (idev->input)
		return -ENODATA;

	/*
	 * No change, so just return. Some applications call S_STD again after
	 * the buffers for streaming have been set up, so we have to allow for
	 * this behavior.
	 */
	if (std == idev->std)
		return 0;

	/*
	 * Changing the standard implies a format change, which is not allowed
	 * while buffers for use with streaming have already been allocated.
	 */
	if (vb2_is_busy(&idev->queue))
		return -EBUSY;

	/* TODO: handle changing std */

	idev->std = std;

	/* Update the internal format */
	vpp_dev_fill_pix_format(idev, &idev->format.fmt.pix);
	return 0;
}

static int vpp_dev_g_std(struct file *file, void *priv, v4l2_std_id *std)
{
	struct vpp_dev *idev = video_drvdata(file);

	/* G_STD is not supported on the HDMI input */
	if (idev->input)
		return -ENODATA;

	*std = idev->std;
	return 0;
}

/*
 * Query the current standard as seen by the hardware. This function shall
 * never actually change the standard, it just detects and reports.
 * The framework will initially set *std to tvnorms (i.e. the set of
 * supported standards by this input), and this function should just AND
 * this value. If there is no signal, then *std should be set to 0.
 */
static int vpp_dev_querystd(struct file *file, void *priv, v4l2_std_id *std)
{
	struct vpp_dev *idev = video_drvdata(file);

	/* QUERY_STD is not supported on the HDMI input */
	if (idev->input)
		return -ENODATA;

	return 0;
}

static int vpp_dev_s_dv_timings(struct file *file, void *_fh,
				struct v4l2_dv_timings *timings)
{
	struct vpp_dev *idev = video_drvdata(file);
	int ret = 0;
	/* S_DV_TIMINGS is not supported on the S-Video input */
	if (idev->input == 0) {
		dev_err(idev->dev, "%s: set timeings failed. input is S-Video", __func__);
		return -ENODATA;
	}

	/* Quick sanity check */
	if (!v4l2_valid_dv_timings(timings, &idev_timings_cap, NULL, NULL)) {
		pr_err("\nvpp_api.c: timings not valid. Sanity check failed");
		return -EINVAL;
	}

	/* Check if the timings are part of the CEA-861 timings. */
	if (!v4l2_find_dv_timings_cap(timings, &idev_timings_cap,
				      0, NULL, NULL)){
		pr_err("\nvpp_api.c: timings is not part of CEA-861");
		return -EINVAL;
	}

	/* Return 0 if the new timings are the same as the current timings. */
	if (v4l2_match_dv_timings(timings, &idev->timings, 0)) {
		pr_err("\nvpp_api.c: timings is the same as the current one.");
		return 0;
	}

	/*
	 * Changing the timings implies a format change, which is not allowed
	 * while buffers for use with streaming have already been allocated.
	 */
	if (vb2_is_busy(&idev->queue))
		return -EBUSY;

	/* Save timings */
	idev->timings = *timings;
	/* vpp_io_set_timing_registers(idev, timings); */
	ret = vpp_io_set_timing_registers(idev, &idev->timings);
	if (ret)
		pr_err("\n vpp_dev_s_dv_timings: failed to set timings registers");
	return ret;
}

static int vpp_dev_g_dv_timings(struct file *file, void *_fh,
				struct v4l2_dv_timings *timings)
{
	struct vpp_dev *idev = video_drvdata(file);

	/* G_DV_TIMINGS is not supported on the S-Video input */
	if (idev->input == 0)
		return -ENODATA;

	*timings = idev->timings;
	return 0;
}

static int vpp_dev_enum_dv_timings(struct file *file, void *_fh,
				   struct v4l2_enum_dv_timings *timings)
{
	struct vpp_dev *idev = video_drvdata(file);

	/* ENUM_DV_TIMINGS is not supported on the S-Video input */
	if (idev->input == 0)
		return -ENODATA;

	return v4l2_enum_dv_timings_cap(timings, &idev_timings_cap,
					NULL, NULL);
}

/*
 * Query the current timings as seen by the hardware. This function shall
 * never actually change the timings, it just detects and reports.
 * If no signal is detected, then return -ENOLINK. If the hardware cannot
 * lock to the signal, then return -ENOLCK. If the signal is out of range
 * of the capabilities of the system (e.g., it is possible that the receiver
 * can lock but that the DMA engine it is connected to cannot handle
 * pixelclocks above a certain frequency), then -ERANGE is returned.
 */
static int vpp_dev_query_dv_timings(struct file *file, void *_fh,
				    struct v4l2_dv_timings *timings)
{
	struct vpp_dev *idev = video_drvdata(file);

	/* QUERY_DV_TIMINGS is not supported on the S-Video input */
	if (idev->input == 0)
		return -ENODATA;

	return 0;
}

static int vpp_dev_dv_timings_cap(struct file *file, void *fh,
				  struct v4l2_dv_timings_cap *cap)
{
	struct vpp_dev *idev = video_drvdata(file);

	/* DV_TIMINGS_CAP is not supported on the S-Video input */
	if (idev->input == 0)
		return -ENODATA;
	*cap = idev_timings_cap;

	return 0;
}

static int vpp_dev_enum_input(struct file *file, void *priv,
			      struct v4l2_input *i)
{
	if (i->index > 1)
		return -EINVAL;

	i->type = V4L2_INPUT_TYPE_CAMERA;
	if (i->index == 0) {
		i->std = SKEL_TVNORMS;
		strlcpy(i->name, "S-Video", sizeof(i->name));
		i->capabilities = V4L2_IN_CAP_STD;
	} else {
		i->std = 0;
		strlcpy(i->name, "HDMI", sizeof(i->name));
		i->capabilities = V4L2_IN_CAP_DV_TIMINGS;
	}
	return 0;
}

static int vpp_dev_s_input(struct file *file, void *priv, unsigned int i)
{
	struct vpp_dev *idev = video_drvdata(file);

	if (i > 1)
		return -EINVAL;

	/*
	 * Changing the input implies a format change, which is not allowed
	 * while buffers for use with streaming have already been allocated.
	 */
	if (vb2_is_busy(&idev->queue)) {
		pr_err("Can't change the input. streaming is on.\n");
		return -EBUSY;
	}
	idev->input = i;
	/*
	 * Update tvnorms. The tvnorms value is used by the core to implement
	 * VIDIOC_ENUMSTD so it has to be correct. If tvnorms == 0, then
	 * ENUMSTD will return -ENODATA.
	 */
	idev->vdev.tvnorms = i ? 0 : SKEL_TVNORMS;

	/* Update the internal format */
	vpp_dev_fill_pix_format(idev, &idev->format.fmt.pix);
	return 0;
}

static int vpp_dev_g_input(struct file *file, void *priv, unsigned int *i)
{
	struct vpp_dev *idev = video_drvdata(file);

	*i = idev->input;
	return 0;
}

/*
 * The set of all supported ioctls. Note that all the streaming ioctls
 * use the vb2 helper functions that take care of all the locking and
 * that also do ownership tracking (i.e. only the filehandle that requested
 * the buffers can call the streaming ioctls, all other filehandles will
 * receive -EBUSY if they attempt to call the same streaming ioctls).
 *
 * The last three ioctls also use standard helper functions: these implement
 * standard behavior for drivers with controls.
 */
static const struct v4l2_ioctl_ops idev_ioctl_ops = {
	.vidioc_querycap = vpp_dev_querycap,
	/* cap means V4L2_BUF_TYPE_VIDEO_CAPTURE type. */
	.vidioc_s_fmt_vid_cap_mplane = vpp_dev_s_fmt_vid_cap_mplane,
	.vidioc_g_fmt_vid_cap_mplane = vpp_dev_g_fmt_vid_cap_mplane,
	.vidioc_enum_fmt_vid_cap_mplane = vpp_dev_enum_fmt_vid_cap_mplane,
	.vidioc_try_fmt_vid_cap = vpp_dev_try_fmt_vid_cap,
	.vidioc_s_fmt_vid_cap = vpp_dev_s_fmt_vid_cap,
	.vidioc_g_fmt_vid_cap = vpp_dev_g_fmt_vid_cap,
	.vidioc_enum_fmt_vid_cap = vpp_dev_enum_fmt_vid_cap,

	.vidioc_g_std = vpp_dev_g_std,
	.vidioc_s_std = vpp_dev_s_std,
	.vidioc_querystd = vpp_dev_querystd,

	.vidioc_s_dv_timings = vpp_dev_s_dv_timings,
	.vidioc_g_dv_timings = vpp_dev_g_dv_timings,
	.vidioc_enum_dv_timings = vpp_dev_enum_dv_timings,
	.vidioc_query_dv_timings = vpp_dev_query_dv_timings,
	.vidioc_dv_timings_cap = vpp_dev_dv_timings_cap,

	.vidioc_enum_input = vpp_dev_enum_input,
	.vidioc_g_input = vpp_dev_g_input,
	.vidioc_s_input = vpp_dev_s_input,

	.vidioc_reqbufs = vb2_ioctl_reqbufs,
	.vidioc_create_bufs = vb2_ioctl_create_bufs,
	.vidioc_querybuf = vb2_ioctl_querybuf,
	.vidioc_qbuf = vb2_ioctl_qbuf,
	.vidioc_dqbuf = vb2_ioctl_dqbuf,
	.vidioc_expbuf = vb2_ioctl_expbuf,
	.vidioc_streamon = vb2_ioctl_streamon,
	.vidioc_streamoff = vb2_ioctl_streamoff,

	.vidioc_log_status = v4l2_ctrl_log_status,
	.vidioc_subscribe_event = v4l2_ctrl_subscribe_event,
	.vidioc_unsubscribe_event = v4l2_event_unsubscribe,
};

static int  vpp_fh_open(struct file *filp)
{
	struct vpp_dev *idev = video_drvdata(filp);
	int ret;

	ret = pm_runtime_get_sync(idev->dev);
	if (ret < 0)
		return ret;

	ret = v4l2_fh_open(filp);
	if (ret) {
		dev_err(idev->dev, "%s v4l2_fh_open failed.\n", __func__);
		pm_runtime_put(idev->dev);
	}
	idev->pm_state = PM_RUNNING;
	return ret;
}

int vpp_fop_release(struct file *filp)
{
	struct vpp_dev *idev = video_drvdata(filp);
	int ret;

	dev_info(idev->dev, "%s: started\n", __func__);

	ret = vb2_fop_release(filp);
	if (ret)
		dev_err(idev->dev, "vb2_fop_release failed.\n");
	vpp_ctrl_release_ctrls(idev);

	pm_runtime_put(idev->dev);

	return ret;
}

static const struct v4l2_file_operations idev_fops = {
	.owner = THIS_MODULE,
	.open = vpp_fh_open,
	.release = vpp_fop_release,
	.unlocked_ioctl = video_ioctl2,
	.read = vb2_fop_read,
	.mmap = vb2_fop_mmap,
	.poll = vb2_fop_poll,
};

static int vpp_runtime_suspend(struct device *dev)
{
	struct v4l2_device *v4l2_dev = dev_get_drvdata(dev);
	struct vpp_dev *idev __maybe_unused = container_of(v4l2_dev,
							   struct vpp_dev,
							   v4l2_dev);

#ifndef __VPP_TARGET_IS_EMULATOR__
	dev_info(idev->dev, " %s: disabling the clock\n", __func__);
	clk_disable_unprepare(idev->clk_core);
#endif
	return 0;
}

static int vpp_runtime_resume(struct device *dev)
{
	struct v4l2_device *v4l2_dev = dev_get_drvdata(dev);
	struct vpp_dev *idev __maybe_unused = container_of(v4l2_dev,
							   struct vpp_dev,
							   v4l2_dev);
	int ret = 0;

#ifndef __VPP_TARGET_IS_EMULATOR__
	ret = clk_prepare_enable(idev->clk_core);
	if (ret)
		dev_err(idev->dev, "couldn't enable the vpp core clk\n");
#endif
	return ret;
}

/*
 * The initial setup of this device instance. Note that the initial state of
 * the driver should be complete. So the initial format, standard, timings
 * and video input should all be initialized to some reasonable value.
 */

static int vpp_dev_probe(struct platform_device *ofdev)
{
	struct vpp_dev *idev;
	struct video_device *vdev;
	struct v4l2_ctrl_handler *hdl;
	struct vb2_queue *q;
	int ret;
	struct resource *res;
	struct resource *irq_res;

	idev = devm_kzalloc(&ofdev->dev, sizeof(struct vpp_dev), GFP_KERNEL);
	if (!idev)
		return -ENOMEM;

	res = platform_get_resource(ofdev, IORESOURCE_MEM, 0);
	if (!res) {
		dev_err(&ofdev->dev, "missing 'reg' property in device tree\n");
		ret = -EINVAL;
		goto _free_alloc;
	}
	idev->base = ioremap_nocache(res->start, res->end - res->start);
	if (!idev->base) {
		ret = -ENOMEM;
		goto _free_alloc;
	}
	PDEBUG("register_base=0x%X size=0x%X\n", (unsigned int)idev->base,
	       (u32)(res->end - res->start + 1));

	res = platform_get_resource(ofdev, IORESOURCE_MEM, 1);
	if (!res) {
		dev_err(&ofdev->dev, "missing 'reg' property in device tree\n");
		ret = -EINVAL;
		goto _iounmap;
	}
	idev->mmu_base = ioremap_nocache(res->start, res->end - res->start);
	if (!idev->mmu_base) {
		ret = -ENOMEM;
		goto _iounmap;
	}
	PDEBUG("register_base=0x%X size=0x%X\n", (u32)idev->mmu_base,
	       (u32)(res->end - res->start + 1));
	res = platform_get_resource(ofdev, IORESOURCE_MEM, 2);
	if (!res)  {
		dev_err(&ofdev->dev, "missing 'reg' property in device tree\n");
		ret = -EINVAL;
		goto _iounmap;
	}
	idev->vpp_sel_base = ioremap_nocache(res->start, res->end - res->start);
	if (!idev->vpp_sel_base) {
		ret = -ENOMEM;
		goto _iounmap;
	}
	PDEBUG("register_base=0x%X size=0x%X\n", (u32)idev->vpp_sel_base,
	       (u32)(res->end - res->start + 1));

	ret = of_property_read_u32(ofdev->dev.of_node,
				   "img,vpp_lane_swap_flag",
				   (u32 *)&idev->sys_values.vpp_swap_lanes);
	if (ret)
		goto _iounmap;
	ret = of_property_read_u32(ofdev->dev.of_node,
				   "img,vpp_mem_burst_size",
				   (u32 *)&idev->sys_values.vpp_mem_burst_size);
	if (ret) {
		pr_err("\nvpp_dev_probe: can't get vpp_mem_burst_size.");
		goto _iounmap;
	}
	ret = of_property_read_u32(ofdev->dev.of_node,
				   "img,vpp_mem_burst_length", (u32 *)
				   &idev->sys_values.vpp_mem_burst_length);
	if (ret) {
		pr_err("\nvpp_dev_probe: Can't get vpp_mem_burst_length");
		goto _iounmap;
	}
	idev->sys_values.vpp_mem_burst_length = 31;
#ifndef __VPP_TARGET_IS_EMULATOR__
	idev->clk_core = devm_clk_get(&ofdev->dev, "vpp_core");
	if (IS_ERR(idev->clk_core)) {
		pr_err("\n vpp_probe: unable to get the clock");
		ret = PTR_ERR(idev->clk_core);
		goto _iounmap;
	}
	ret = of_property_read_u32(ofdev->dev.of_node, "clock-frequency",
				   (u32 *)&idev->clk_core_rate);
	if (ret)
		idev->clk_core_rate = 670000000;

	if (clk_get_rate(idev->clk_core) != idev->clk_core_rate) {
		ret = clk_set_rate(idev->clk_core, idev->clk_core_rate);
		if (ret) {
			dev_err(idev->dev, "vpp core clk rate not set\n");
			goto _iounmap;
		}

		pr_info("vpp_core rate is %lu\n", clk_get_rate(idev->clk_core));

	}
#endif

	idev->use_dma_contig = USE_DMA_CONTIG;
	idev->dev   = &ofdev->dev;
	irq_res = platform_get_resource(ofdev, IORESOURCE_IRQ, 0);
	if (!irq_res) {
		pr_err("\nget interrupt resource failed.\n");
		ret = -ENXIO;
		goto _iounmap;
	} else {
		pr_info("get interrupt resource ok. irq= 0x%X\n",
			irq_res->start);
	}
	ret = request_irq(irq_res->start, vpp_dev_irq, IRQF_SHARED,
			  "img_vpp", idev);
	if (ret) {
		pr_err("\nrequest interrupt failed.\n");
		ret = -ENXIO;
		goto _iounmap;
	}

	/*TODO: do we need idev->input?*/
	idev->input = 1; /* HDMI not S_Video */
	ret = v4l2_device_register(&ofdev->dev, &idev->v4l2_dev);
	if (ret)
		goto _free_irq;
	else
		pr_err("v4l2 device registration succeeded...\n");

	mutex_init(&idev->lock);
	mutex_init(&idev->vpp_workqueue_mutex);
	idev->streaming_is_being_stopped = false;
	init_completion(&idev->stop_streaming_notifier);

	hdl = &idev->ctrl_handler;
	/* 15 is a hint to the number of controls. */
	v4l2_ctrl_handler_init(hdl, 15);
	ret = vpp_ctrl_register_ctrls(idev);
	if (ret)
		goto _free_hdl;

	/* Initialize the vb2 queue */
	q = &idev->queue;
	q->type = V4L2_BUF_TYPE_VIDEO_CAPTURE_MPLANE;
	/* q->mem_ops = &vb2_vmalloc_memops; */
	q->io_modes = VB2_MMAP  | VB2_USERPTR | VB2_DMABUF;

	q->drv_priv = idev;
	q->buf_struct_size = sizeof(struct idev_buffer);

	vpp_buf_set_qops(idev);

	idev->set_next_buffer_workqueue =
		alloc_workqueue(VPP_DRIVER_NAME,
				WQ_MEM_RECLAIM | WQ_HIGHPRI, 0);

	if (!idev->set_next_buffer_workqueue) {
		ret = -ENOMEM;
		goto _free_hdl;
	}
	INIT_WORK(&idev->set_next_buffer_work, vpp_set_next_buffer);
	q->timestamp_flags = V4L2_BUF_FLAG_TIMESTAMP_MONOTONIC;

	/* q->mem_ops = &vb2_vmalloc_memops; */
	if (idev->use_dma_contig)
		q->mem_ops = &vb2_dma_contig_memops;
	else
		q->mem_ops = &vb2_dma_sg_memops;

	memset(idev->state_count, 0, sizeof(idev->state_count));
	/* The serialization lock for the streaming ioctls. */
	q->lock = &idev->lock;

	/*
	 * This makes sure that the memory is allocated in the first 4GB of
	 * physical address. We do not need it for the time being
	 * as the size of physical memory is well below 4GBytes but it will
	 * be needed for a 64-bit architecture.
	 * See http://lwn.net/Articles/152337/ for details and note that in
	 * 32-bit artitechture, "GFP_NORMAL (=0) is DMA able enough".
	 * using __GFP_HIGHMEM and GFP_DMA32 resulted in allocating memory
	 * only in GFP_NORMAL not in Highmem.
	 */
	if (idev->use_dma_contig)
		q->gfp_flags =  GFP_DMA32 | __GFP_HIGHMEM;
	else
		q->gfp_flags =  __GFP_HIGHMEM; /* GFP_DMA32 | __GFP_HIGHMEM; */
	ret = vb2_queue_init(q);
	if (ret)
		goto _free_workqueue;

	if (!idev->use_dma_contig) {
		idev->alloc_ctx = vb2_dma_sg_init_ctx(&ofdev->dev);
	} else {
		idev->alloc_ctx = vb2_dma_contig_init_ctx(&ofdev->dev);
		if (IS_ERR(idev->alloc_ctx)) {
			pr_err("Can't allocate buffer context");
			goto _release_vb2_queue;
		}
	}
	INIT_LIST_HEAD(&idev->buf_list);
	spin_lock_init(&idev->qlock);

	/* Initialize the video_device structure */
	vdev = &idev->vdev;
	strlcpy(vdev->name, KBUILD_MODNAME, sizeof(vdev->name));
	/*
	 * There is nothing to clean up, so release is set to an empty release
	 * function. The release callback must be non-NULL.
	 */
	vdev->release = video_device_release_empty;
	vdev->fops = &idev_fops,
	vdev->ioctl_ops = &idev_ioctl_ops,
	/*
	 * The main serialization lock. All ioctls are serialized by this
	 * lock. Exception: if q->lock is set, then the streaming ioctls
	 * are serialized by that separate lock.
	 */
	vdev->lock = &idev->lock;
	vdev->queue = q;
	vdev->v4l2_dev = &idev->v4l2_dev;
	/* Supported SDTV standards, if any */
	vdev->tvnorms = SKEL_TVNORMS;
	video_set_drvdata(vdev, idev);

	pm_runtime_enable(&ofdev->dev);
	if (!pm_runtime_enabled(&ofdev->dev)) {
		ret = vpp_runtime_resume(&ofdev->dev);
		if (ret)
			goto _pm_rt_disable;
	}
	ret = pm_runtime_get_sync(&ofdev->dev);
	if (ret < 0)
		goto _pm_rt_suspend;

	ret = vpp_mmu_probe(idev);
	if (ret)
		goto _pm_rt_put;

	ret = video_register_device(vdev, VFL_TYPE_GRABBER, -1);
	if (ret)
		goto _pm_rt_put;

	idev->about_to_start_streaming = false;

	/* soft_reset and then turn off the vpp*/
	ret = vpp_io_soft_reset(idev);
	if (!ret)
		ret = vpp_io_turn_off_hw(idev);
	else
		goto _unregister_device;
	vpp_io_disable_interrupt(idev, ~0u);
	vpp_io_clear_interrupt(idev, ~0u);
	PDEBUG(" %s registered.\n", video_device_node_name(&idev->vdev));
	PDEBUG(" %s registered.\n", video_device_node_name(vdev));
	idev->irq_message[0] = '\0';
	dev_info(idev->dev, "vpp_core rate is %lu\n",
		 clk_get_rate(idev->clk_core));

	/* show_free_areas(0); */
	/* show_mem(0); */
	pm_runtime_put(&ofdev->dev);
	idev->pm_state = PM_SUSPENDED_CLOSED;

	return 0;
_unregister_device:
	video_unregister_device(&idev->vdev);
_pm_rt_put:
	pm_runtime_put(&ofdev->dev);
_pm_rt_suspend:
	if (!pm_runtime_enabled(&ofdev->dev))
		vpp_runtime_suspend(&ofdev->dev);
_pm_rt_disable:
	pm_runtime_disable(&ofdev->dev);
	if (idev->use_dma_contig)
		vb2_dma_contig_cleanup_ctx(idev->alloc_ctx);
_release_vb2_queue:
	vb2_queue_release(&idev->queue);
_free_workqueue:
	flush_workqueue(idev->set_next_buffer_workqueue);
	destroy_workqueue(idev->set_next_buffer_workqueue);
_free_hdl:
	v4l2_ctrl_handler_free(&idev->ctrl_handler);
	v4l2_device_unregister(&idev->v4l2_dev);
_free_irq:
	if (irq_res)
		free_irq(irq_res->start, (void *)idev);
_iounmap:
	if (idev->base)
		iounmap(idev->base);
	if (idev->mmu_base)
		iounmap(idev->mmu_base);
_free_alloc:
	devm_kfree(&ofdev->dev, idev);
	return ret;
}

static int vpp_dev_remove(struct platform_device *ofdev)
{
	struct resource *res;
	struct v4l2_device *v4l2_dev = dev_get_drvdata(&ofdev->dev);
	struct vpp_dev *idev = container_of(v4l2_dev, struct vpp_dev, v4l2_dev);

	pr_err("\nvpp_dev_remove: started.\n");
	flush_workqueue(idev->set_next_buffer_workqueue);
	destroy_workqueue(idev->set_next_buffer_workqueue);
	/*Return all remaining buffers in the queue. Just in case.*/
	return_all_buffers(idev, VB2_BUF_STATE_ERROR);
	vb2_queue_release(&idev->queue);
	vpp_mmu_remove(idev);
	if (idev->base)
		iounmap(idev->base);
	v4l2_info(&idev->v4l2_dev, "unregistering %s. success\n",
		  video_device_node_name(&idev->vdev));
	PDEBUG("unregistering %s\n", video_device_node_name(&idev->vdev));
	res = platform_get_resource(ofdev, IORESOURCE_IRQ, 0);
	if (!res) {
		pr_err("\nget interrupt resource failed.\n");
		return -ENXIO;
	}
	free_irq(res->start, (void *)idev);

	/* irq_dispose_mapping(idev->irq); */
	v4l2_ctrl_handler_free(&idev->ctrl_handler);
	v4l2_device_unregister(&idev->v4l2_dev);
	video_unregister_device(&idev->vdev);

	pm_runtime_disable(&ofdev->dev);
	if (!pm_runtime_status_suspended(&ofdev->dev))
		vpp_runtime_suspend(&ofdev->dev);

	return 0;
}

/*
 *	To suspend:
 *	1. Wait for the current frame to finish
 *	2. Return the buffer (current_buffer) to the application.
 *	3. Do not set up the workqueue
 *	4. Power-off the device. It stops using the programmed next_buffer.
 *	5. mmu_free both current_buffer and next_buffer
 *	6. Return the next_buffer to the queue. Application will return the
 *	   current_buffer to the queue after it has finished with it.
 *	7. Stop the clocks
 */
#ifdef CONFIG_PM_SLEEP
static int vpp_suspend(struct device *dev)
{
	int ret = 0;
	struct vpp_dev *idev = container_of(dev_get_drvdata(dev),
					    struct vpp_dev, v4l2_dev);

	if (pm_runtime_suspended(dev)) {
		idev->pm_state = PM_SUSPENDED_CLOSED;
		return 0;
	}
	vpp_io_disable_interrupt(idev, ~0u);
	ret = vpp_buf_suspend(idev);
	if (ret) {
		dev_err(dev, "vpp_buf_suspend failed.\n");
		return ret;
	}
	ret = pm_runtime_force_suspend(dev);
	if (ret) {
		dev_err(dev, "error: force_suspend failed.\n");
		return ret;
	}
	idev->pm_state = PM_SUSPENDED_ACTIVE;

	return ret;
}

static int vpp_resume(struct device *dev)
{
	int ret = 0;
	struct vpp_dev *idev = container_of(dev_get_drvdata(dev),
					    struct vpp_dev, v4l2_dev);

	if (idev->pm_state != PM_SUSPENDED_ACTIVE &&
	    idev->pm_state != PM_SUSPENDED_CLOSED) {
		dev_err(idev->dev, "error: status must be SUSPENDED\n");
		return -EINVAL;
	}

	if (idev->pm_state == PM_SUSPENDED_CLOSED)
		return 0;

	ret = pm_runtime_force_resume(dev);
	if (ret)
		return ret;

	/*
	 * Soft_reset and turn_off the device to make
	 * registers ready for programming.
	 */
	vpp_api_turn_off(idev);
	/* The device will be turned on in start_streaming function */
	ret = vpp_buf_resume(idev);

	return ret;
}
#endif /* CONFIG_PM_SLEEP */

static const struct dev_pm_ops vpp_pm_ops = {
	SET_SYSTEM_SLEEP_PM_OPS(vpp_suspend, vpp_resume)
	SET_RUNTIME_PM_OPS(vpp_runtime_suspend, vpp_runtime_resume, NULL)
};

static struct platform_driver vpp_dev_driver = {
		.probe = vpp_dev_probe,
		.remove = vpp_dev_remove,
		.driver = {
			.name = VPP_DRIVER_NAME,
			.owner = THIS_MODULE,
			.pm = &vpp_pm_ops,
			.of_match_table = of_ids
		}
};

module_platform_driver(vpp_dev_driver);
