#ifndef _VPP_API_H
#define _VPP_API_H
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/kmod.h>
#include <linux/mutex.h>
#include <linux/mm.h>
#include <linux/interrupt.h>
#include <linux/videodev2.h>
#include <linux/v4l2-dv-timings.h>
#include <media/v4l2-device.h>
#include <media/v4l2-dev.h>
#include <media/v4l2-ioctl.h>
#include <media/v4l2-dv-timings.h>
#include <media/v4l2-ctrls.h>
#include <media/v4l2-event.h>
/* #include <media/videobuf2-dma-contig.h> */
#include <media/videobuf2-vmalloc.h>
#include <media/videobuf2-core.h>
#include <linux/of_address.h>
#include <linux/platform_device.h>
#include <linux/workqueue.h>
#include <linux/delay.h>
#include <video/img_vpp.h>
#include "csc2/csc2.h"
#include <linux/completion.h>
#define VPP 1
#define MIN_VPP_BUFFERS 3 /* This can't be more than 32! */
#define MAX_VPP_BUFFERS	32
#define IRQ_MESSAGE_LEN 750
#define MAX_PLANES_PER_BUFFER 4
#define USE_DMA_CONTIG 0   /* select 1 for no-mmu and 0 for mmu */
#define MAX_FRAME_PERIOD_IN_MS 50

enum {
	PM_RUNNING = 0,
	PM_READY_TO_SUSPEND,
	PM_SUSPENDED_ACTIVE,
	PM_SUSPENDED_CLOSED,
	PM_WAITING_FOR_BUFFERS,
};
/*TODO
1. Add selection of DMA-CONTIG to menuconfig selection for vpp
2. Allocate/reserve memory in vpp_api.c in case of DMA-CONTIG
4. Reserve the memory for the vpp_device in dts.
3. Change the number of MIN_VPP_BUFFERS to 4 or more.
*/
/**
 * struct vpp_addr -
 * @y_even:	 luminance plane address (even in interlaced case)
 * @y_odd:	 luminance odd plane address if interlaced
 * @uv_even: Chroma plane address (even in interlaced case)
 * @uv_odd:	 Chroma odd plane address if interlaced
 */
struct vpp_addr {
	__u32 y_even;
	__u32 y_odd;
	__u32 uv_even;
	__u32 uv_odd;
};
/**
 * struct vpp_fmt - the driver's internal color format data
 * @mbus_code: Media Bus pixel code, -1 if not applicable
 * @name: format description
 * @pixelformat: the fourcc code for this format, 0 if not applicable
 * @yorder: Y/C order
 * @corder: Chrominance order control
 * @num_planes: number of physically non-contiguous data planes
 * @nr_comp: number of physically contiguous data planes
 * @depth: per plane driver's private 'number of bits per pixel'
 * @flags: flags indicating which operation mode format applies to
 */
struct vpp_fmt {
/* enum v4l2_mbus_pixelcode mbus_code; */
	u32	mbus_code;
	char	*name;
	u32	pixelformat;
	u32	color;
	u32	yorder;
	u32	corder;
	u16	num_planes;
	u16	num_comp;
	u8	depth[MAX_PLANES_PER_BUFFER];
	u32	flags;
};
/**
 * struct vpp_target_frame - target frame properties
 * @f_width:	SRC : SRCIMG_WIDTH, DST : OUTPUTDMA_WHOLE_IMG_WIDTH
 * @f_height:	SRC : SRCIMG_HEIGHT, DST : OUTPUTDMA_WHOLE_IMG_HEIGHT
 * @crop:	cropped(source)/scaled(destination) size
 * @plane_image_sizes:	image size in bytes (w x h x bpp)
 * @addr:	image frame buffer physical addresses
 * @fmt:	G-Scaler color format pointer
 * @colorspace: value indicating v4l2_colorspace
 * @alpha:	frame's alpha value
 */
struct vpp_target_frame {
	u32 f_width;
	u32 f_height;
	unsigned long plane_image_sizes[MAX_PLANES_PER_BUFFER];
	struct vpp_addr	addr;
	const struct vpp_fmt *fmt;
	u32 colorspace;
	u8 alpha;
};
/**
 * struct vpp_source_frame - target frame properties
 * @f_width:	SRC : SRCIMG_WIDTH, DST : OUTPUTDMA_WHOLE_IMG_WIDTH
 * @f_height:	SRC : SRCIMG_HEIGHT, DST : OUTPUTDMA_WHOLE_IMG_HEIGHT
 * @crop:	cropped(source)/scaled(destination) size
 * @payload:	image size in bytes (w x h x bpp)
 * @addr:	image frame buffer physical addresses
 * @fmt:	G-Scaler color format pointer
 * @colorspace: value indicating v4l2_colorspace
 * @alpha:	frame's alpha value
 */
struct vpp_source_frame {
	u32 f_width;
	u32 f_height;
	struct v4l2_rect crop;
	unsigned long payload[MAX_PLANES_PER_BUFFER];
	struct vpp_addr	addr;
	const struct vpp_fmt *fmt;
	u32 colorspace;
	u8 alpha;
};

struct vpp_scaler {
	/* horizontal scaler */
	int hscoeffs[(VPP_H_TAPS * VPP_H_INTERPOLATION_POINTS) + 1];
	int hs_old_output_size;
	int hs_old_input_size;
	bool hs_en;
	/* vertical scaler */
	int vscoeffs[(VPP_V_TAPS * VPP_V_INTERPOLATION_POINTS) + 1];
	int vs_old_output_size;
	int vs_old_input_size;
	bool vs_en;

};

struct	vpp_csc_data {
	/* input data */
	CSC_eColourSpace		eInputColourSpace;
	CSC_eColourSpace		eOutputColourSpace;
	CSC_eRange				eInputRange;
	CSC_eRange				eOutputRange;
	CSC_sHSBCSettings		sHSBCSettings;
	/*output matrix and offsets */
	CSC_sConfiguration		sResultantConfigurationData;
};
struct system_level_values {
	bool	vpp_swap_lanes;
	u32		vpp_mem_burst_size;
	u32		vpp_mem_burst_length;
};
/**
 * struct vpp_dev - All internal data for one instance of device
 * @pdev: PCI device  // ofdev, platform device
 * @v4l2_dev: top-level v4l2 device struct
 * @vdev: video node structure
 * @ctrl_handler: control handler structure
 * @lock: ioctl serialization mutex
 * @std: current SDTV standard
 * @timings: current HDTV timings
 * @format: current pix format
 * @input: current video input (0 = SDTV, 1 = HDTV)
 * @queue: vb2 video capture queue
 * @alloc_ctx: vb2 contiguous DMA context
 * @qlock: spinlock controlling access to buf_list and sequence
 * @buf_list: list of buffers queued for DMA
 * @sequence: frame sequence counter
 */


struct vpp_dev {
	struct device		*dev;

	struct v4l2_device v4l2_dev;
	struct video_device vdev;
	struct v4l2_ctrl_handler ctrl_handler;
	struct mutex lock;
	struct mutex vpp_workqueue_mutex;
	bool	streaming_is_being_stopped;
	struct	completion stop_streaming_notifier;
	v4l2_std_id std;
	struct v4l2_dv_timings timings;
	struct v4l2_format format;

	/* S_VIDEO: 0 or HDMI: 1 we set it to 1. Will remove later. */
	unsigned input;

	struct vb2_queue queue;
	struct vb2_alloc_ctx *alloc_ctx;
	/* workqueue spinlock */
	spinlock_t wqlock;
	spinlock_t qlock;
	struct list_head buf_list;
	struct idev_buffer *current_buf;
	struct idev_buffer *next_buf;
	unsigned field;
	unsigned sequence;
	/* __SA Now adding vpp structures */
	void __iomem *base;
	void __iomem *mmu_base;
	void __iomem *vpp_sel_base;
	struct vpp_source_frame source_frame;
	struct vpp_target_frame target_frame;
	struct clk *pdp_clk;
	struct clk *pixel_clk;

	struct	vpp_csc_data csc_data;
	struct	vpp_scaler  v;
	/* true for custom coefficients. false for defaults */
	bool	vpp_user_defined_csc_matrix;
	bool	vpp_user_defined_hscb;
	bool	vpp_user_defined_hscale_coeff;
	bool	vpp_user_defined_vscale_coeff;
	struct {
	/* CSC cluster */
	/* If user defines csc_mult, then they are not allowed to set HSCB
	 * and their csc_in_offset and csc_out_offset will also be used.
	 * Otherwise, the input and output and also the HSCB are used to create
	 * the matrix and offsets.
	 */
		struct v4l2_ctrl *csc_mult;
		struct v4l2_ctrl *csc_in_offset;
		struct v4l2_ctrl *csc_out_offset;
	};

	struct v4l2_ctrl *csc_brightness_ctrl;
	struct v4l2_ctrl *csc_hue_ctrl;
	struct v4l2_ctrl *csc_saturation_ctrl;
	struct v4l2_ctrl *csc_contrast_ctrl;
	struct v4l2_ctrl *csc_input_colorspcae_ctrl;
	struct v4l2_ctrl *csc_input_colorspcae_range_ctrl;
	struct v4l2_ctrl *csc_input_mode_ctrl;
	struct v4l2_ctrl *input_pol_ctrl;
	struct v4l2_ctrl *hscaler_coeffs;
	struct v4l2_ctrl *vscaler_coeffs;
	struct v4l2_ctrl *swap_internal_uv_lanes_ctrl;
	struct v4l2_ctrl *tiling_ctrl;
#ifdef __VPP_TARGET_IS_SIMULATOR__
	struct v4l2_ctrl *vp_kick_ctrl;
#endif
	struct workqueue_struct *set_next_buffer_workqueue;
	struct work_struct set_next_buffer_work;
	u32 *pdir; /* kernel virtual base address of the page directories */
	dma_addr_t pdir_bus; /* bus base address of the page directories */
	u32 bus_index;
	struct system_level_values sys_values;
	bool use_dma_contig;
	struct clk *clk_core;
	unsigned int clk_core_rate;
	bool about_to_start_streaming;
	char irq_message[IRQ_MESSAGE_LEN];
	bool buffer_error;
	int pm_state;
	bool use_mmu;
	u32 state_count[VB2_BUF_STATE_ERROR + 1];

};


#endif /*_VPP_API_H */
