/*
 * This is a V4L2 VPP Driver.
 *
 * Copyright 2015 Imagination Technologies. All rights reserved.
 *
 * This program is free software; you may redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
#include <linux/dma-mapping.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/mutex.h>
#include <linux/list.h>
#include <linux/spinlock.h>
#include <linux/videodev2.h>
#include "vpp_api.h"
#include "vpp_scale.h"
#include <video/img_vpp.h>
#include "vpp_buf.h"
#include "vpp_ctrl.h"
#include "vpp_mmu.h"
#include <media/videobuf2-dma-contig.h>
#define VPP 1

static int queue_setup(struct vb2_queue *vq, const struct v4l2_format *fmt,
		       unsigned int *nbuffers, unsigned int *nplanes,
		       unsigned int sizes[], void *alloc_ctxs[]);
static int buffer_init (struct vb2_buffer *vb);
static int buffer_prepare(struct vb2_buffer *vb);
static void buffer_finish(struct vb2_buffer *vb);
static void buffer_cleanup(struct vb2_buffer *vb);
static void buffer_queue(struct vb2_buffer *vb);
static int start_streaming(struct vb2_queue *vq, unsigned int count);
static void stop_streaming(struct vb2_queue *vq);
static CSC_eColourSpace  convert_v4l2_to_csc (enum v4l2_colorspace v4l2_cs);
static int set_csc(struct vpp_dev *idev);

static struct vb2_ops idev_qops = {
	.queue_setup		= queue_setup,
	.buf_init		= buffer_init,
	.buf_prepare		= buffer_prepare,
	.buf_queue		= buffer_queue,
	.buf_finish		= buffer_finish,
	.buf_cleanup		= buffer_cleanup,
	.start_streaming	= start_streaming,
	.stop_streaming		= stop_streaming,
	.wait_prepare		= vb2_ops_wait_prepare,
	.wait_finish		= vb2_ops_wait_finish,
};

void vpp_buf_set_qops(struct vpp_dev *idev)
{
	idev->queue.ops = &idev_qops;
}
#if (LINUX_VERSION_CODE <= KERNEL_VERSION(3, 16, 0))
static inline bool vb2_fileio_is_active(struct vb2_queue *q)
{
         return q->fileio;
}
#endif
static inline struct idev_buffer *to_idev_buffer(struct vb2_buffer *vb2)
{
	return container_of(vb2, struct idev_buffer, vb);
}

static void __maybe_unused print_buf_addresses(struct vb2_buffer *p_vb)
{
	int i;
	unsigned int page_addr;

	for ( i=0; i< p_vb->num_planes; i++) {
		int count =0;
		unsigned int addr, buffer = (unsigned int)vb2_plane_vaddr(p_vb,i);
		unsigned int size = p_vb->v4l2_planes[i].length;
		PDEBUG("\n print_buf_addresses: length[%d]= %d start[%d]=0x%X , address=0x%X\n",i,p_vb->v4l2_planes[i].length,i,p_vb->v4l2_planes[i].m.mem_offset, (unsigned int ) vb2_plane_vaddr(p_vb,i));
		if (is_vmalloc_addr(vb2_plane_vaddr(p_vb,i))){
			PDEBUG ("\n address is vmalloc, buffer = 0x%X buffer & PAGE_MASK = 0x%X buffer & 0xFF800000= 0x%x ",buffer,buffer & PAGE_MASK,(buffer & 0xFF800000) >>22 );

			for (addr = buffer & PAGE_MASK; addr < buffer + size; addr += PAGE_SIZE) {
				struct page *pg = vmalloc_to_page((void *)addr);
				page_addr =  page_to_phys(pg);
				if (count++ <3){
					PDEBUG ("\nprint_buf_addresses: virtual addr=0x%X physical=0x%X ", addr,page_addr  );
					PDEBUG (" *addr= 0x%08X", *(unsigned int *)addr);
				}
			}

		}
		else{
			u32 physic;
			physic = (u32)vb2_dma_contig_plane_dma_addr(p_vb, i);
			PDEBUG ("\nprint_buf_addresses: address is not vmalloc: virtual addr=0x%08X physical=0x%08X *buffer=0x%08X", buffer,physic,*(u32 *)buffer );
		}
	}
}
/* Delete the first buffer from the queue
 * Store a pointer to this buffer in either idev->current_buffer or idev->next_buffer
 * Pass the buffer to mmu_setup to store the buffer physical addresses to mmu registers.
 * Write the virtual addresses to vpp registers.
 */
int set_buffer ( struct vpp_dev *idev, int cur_or_next)
{
	struct vb2_buffer *vb2_buffer_frame = NULL;
	unsigned long flags;
	struct idev_buffer *buf;
	int ret = 0;

	spin_lock_irqsave(&idev->qlock, flags);

	if (list_empty(&idev->buf_list)) {
		dev_info(idev->dev, "\n\n\n error: no buffer in the queue.\n\n\n");
		ret = -EINVAL;
		spin_unlock_irqrestore(&idev->qlock, flags);
		goto __end;
	}
		buf = list_first_entry(&idev->buf_list,
				       struct idev_buffer, list);
		if ((u32)buf->list.next != (u32)LIST_POISON1) {
			list_del(&buf->list);
		} else {
			WARN(1, "re-deleting a buffer from the list.");
			ret = -EINVAL;
			spin_unlock_irqrestore(&idev->qlock, flags);
			goto __end;
		}
	spin_unlock_irqrestore(&idev->qlock, flags);
	vb2_buffer_frame = &buf->vb;
	if ( cur_or_next == CURRENT_BUFFER){
		idev->current_buf = buf;
		PDEBUG("set_buffer: got the current buffer: 0x%X", (u32)vb2_buffer_frame );
	}
	else if ( cur_or_next == NEXT_BUFFER){
		idev->next_buf = buf;
		PDEBUG("set_buffer: got the next buffer: 0x%X", (u32)vb2_buffer_frame);
	}
		if (vb2_buffer_frame) {
			if (idev->use_mmu)
				ret = vpp_mmu_setup(buf, idev);
			else
				ret = vpp_mmu_setup_no_mmu(buf, idev);
		} else {
			ret = -EINVAL;
		}
__end:
		if (ret !=0){
			printk ("\n--------------------\n set_buffer: Could not set a new buffer.returning error.\n--------------------\n");
			idev->buffer_error = true;
			ret = -EINVAL;
		}
		else{
			vb2_buffer_frame->v4l2_buf.sequence = idev->sequence;
			idev->sequence++;
		}
	return ret;
}
#ifdef __VPP_TARGET_IS_SIMULATOR__
#define DELAY 1000
static void wait_for_kick(struct vpp_dev *idev)
{
	int counter = 0;
	int vpp_kick = v4l2_ctrl_g_ctrl(idev->vp_kick_ctrl);

	if (vpp_kick != VPP_KICK_AUTO) {
		while (vpp_kick != VPP_KICK_ON) {
			msleep(DELAY);
			counter++;
			vpp_kick = v4l2_ctrl_g_ctrl(idev->vp_kick_ctrl);
			WARN_ONCE(counter * DELAY > 50, "waiting too long");
		}
		dev_info(idev->dev, "waited for kick = %d for %d ms\n",
			 vpp_kick, counter * DELAY);
		v4l2_ctrl_s_ctrl(idev->vp_kick_ctrl, VPP_KICK_OFF);
		vpp_io_kick_simulator(idev);
	} else {
		dev_err(idev->dev, "shouldn't be here!\n");
	}
}
#endif

void vpp_set_next_buffer(struct work_struct *work)
{
	struct vpp_dev *idev;
	int ret;

	idev = container_of ( work, struct vpp_dev , set_next_buffer_work);
	if (idev == NULL){
		printk("\nvpp_set_next_buffer: PANIC: invalid device data structure. \
		No way to get a new buffer and set the device ready for the next frame.");
		return;
	}
	if (idev->buffer_error == true)
		return;
	mutex_lock(&idev->vpp_workqueue_mutex);
	PDEBUG("\n vpp_dev_irq: : got the  mutex");
	if (idev->streaming_is_being_stopped) {
		complete(&idev->stop_streaming_notifier);
		mutex_unlock(&idev->vpp_workqueue_mutex);
		PDEBUG("\n%s: mutex released %dth buf.",
		       __func__, cur_buf->vb.v4l2_buf.sequence);
		return;
	}
	mutex_unlock(&idev->vpp_workqueue_mutex);
#ifdef __VPP_TARGET_IS_SIMULATOR__
	if ( v4l2_ctrl_g_ctrl(idev->vp_kick_ctrl) != VPP_KICK_AUTO)
		wait_for_kick(idev);
#endif
	/*
	 * The buffer that we just used in the last frame is pointed to by idev->current_buf.
	 * Use this pointer to free the page tables and unmap all the mappings.
	 */
	if (idev->use_mmu) {
		ret = vpp_mmu_free(idev, idev->current_buf);
		if (ret)
			dev_err(idev->dev, " Failed to free the mmu.\n");
	}

	PDEBUG("\nvpp_set_next_buffer: freed mmu successfully");
	idev->current_buf = idev->next_buf;
	ret = set_buffer (idev, NEXT_BUFFER);
	if ( ret !=0 ){
		printk ("\nvpp_set_next_buffer: set_buffer returned error");
	}
	else {
		PDEBUG ("\nvpp_set_next_buffer:  finished successfully");
	}
	PDEBUG ("\nvpp_set_next_buffer:  finished");
	return;
}

static int buf_is_in_q (struct vpp_dev *idev, struct idev_buffer *in_idev_buf)
{
	struct list_head *ptr;
	struct idev_buffer *idev_buf;

	list_for_each (ptr, &idev->buf_list) {
		idev_buf = list_entry(ptr, struct idev_buffer, list);
		if (idev_buf == in_idev_buf)
			return -EINVAL;
	}
	return 0;
}

static inline void cfg_use_mmu_flag(struct vpp_dev *idev)
{
	struct list_head *ptr;
	struct idev_buffer *idev_buf;
	int i;

	/*
	 * Do not use mmu if mem_op:
	 * 1. is vb2_dma_contig_memops.
	 * 2. is vb2_dma_sg_memops but all buffers have only one segment.
	 */
	idev->use_mmu = false;
	if (idev->use_dma_contig)
		return;

	list_for_each(ptr, &idev->buf_list) {
		idev_buf = list_entry(ptr, struct idev_buffer, list);
		if (idev_buf && !idev_buf->ndesc[0]) {
			/* mem_ops is vb2_vmalloc_memops*/
			idev->use_mmu = true;
		} else if (idev_buf) {
			/* memop is vb2_dma_sg_memops. Check the segments. */
			for (i = 0; i < idev_buf->vb.num_planes; i++)
				if (idev_buf->ndesc[i] > 1)
					idev->use_mmu = true;
		}
	}
}

int get_number_of_buffers(struct vpp_dev *idev)
{
	int number_of_buffers = 0;
	struct list_head *ptr;
	struct idev_buffer *idev_buf;
	list_for_each(ptr, &idev->buf_list) {
		idev_buf = list_entry(ptr, struct idev_buffer, list);

		number_of_buffers++;
	}
	return number_of_buffers;
}

static int queue_setup(struct vb2_queue *vq, const struct v4l2_format *fmt,
		       unsigned int *nbuffers, unsigned int *nplanes,
		       unsigned int sizes[], void *alloc_ctxs[])
{
	struct vpp_dev *idev = vb2_get_drv_priv(vq);
	int i;

	/* TODO: The driver needs to set bytesperline for each plane based on
	 * the idev->format.fmt.pix_mp.field. Note that the driver can also
	 * change the field element of the input format.
	 */
	switch (idev->queue.type) {
	case V4L2_BUF_TYPE_VIDEO_CAPTURE_MPLANE:
		if (fmt){
			dev_err(idev->dev, "\nfmt is supppsed to be NULL");
			return -EINVAL;
		}
		fmt = &idev->format;
		if (idev->format.type != V4L2_BUF_TYPE_VIDEO_CAPTURE_MPLANE) {
			dev_err(idev->dev, "\n formats do not match.");
			return -EINVAL;
		}
		idev->field = idev->format.fmt.pix_mp.field;
		if (idev->field == V4L2_FIELD_ANY  ||
		    idev->field == V4L2_FIELD_TOP ||
		    idev->field == V4L2_FIELD_BOTTOM) {
			dev_err(idev->dev,
				"\n V4L2_FIELD_ANY/BOTTOM/TOP not correct.\n");
				return -EINVAL;
			}
		if (vq->num_buffers + *nbuffers > MAX_VPP_BUFFERS)
			*nbuffers = MAX_VPP_BUFFERS - vq->num_buffers;
		*nplanes = idev->format.fmt.pix_mp.num_planes;
		PDEBUG ("\n queue_setup: *nbuffers=%d *nplanes= %d\n",
			*nbuffers, *nplanes);
		for (i = 0; i < *nplanes; i++) {
			sizes[i] = idev->format.fmt.pix_mp.plane_fmt[i].sizeimage;
			PDEBUG ("\n size[%d]= %d\n", i, sizes[i]);
			alloc_ctxs[i] = idev->alloc_ctx;
		}
		break;
	case V4L2_BUF_TYPE_VIDEO_CAPTURE:
		if (fmt == NULL){
			fmt = &idev->format;
		}
		idev->field = idev->format.fmt.pix.field;
		if (idev->field == V4L2_FIELD_ALTERNATE) {
		/*
		* You cannot use read() with FIELD_ALTERNATE since the field
		* information (TOP/BOTTOM) cannot be passed back to the user.
		*/
		if (vb2_fileio_is_active(vq))
			return -EINVAL;
		idev->field = V4L2_FIELD_TOP;
		}

		if (vq->num_buffers + *nbuffers < MAX_VPP_BUFFERS)
			*nbuffers = MAX_VPP_BUFFERS - vq->num_buffers;

		if (fmt && fmt->fmt.pix.sizeimage < idev->format.fmt.pix.sizeimage)
			return -EINVAL;
		*nplanes = 1;
		sizes[0] = fmt ? fmt->fmt.pix.sizeimage : idev->format.fmt.pix.sizeimage;
		alloc_ctxs[0] = idev->alloc_ctx;
		break;
	default:
		dev_err(idev->dev, "\nqueue type invalid.");
		return -EINVAL;
		break;
	}
	PDEBUG("\n queue_setup: bufreq handled successfully.\n");
	return 0;
}

static int buffer_init (struct vb2_buffer *vb)
 {
	u32 i;
	u32 *buffer;
	u32 physic;
	struct sg_table *sg_t;
	struct idev_buffer * idev_b = to_idev_buffer(vb);
	struct vpp_dev *idev = vb2_get_drv_priv(vb->vb2_queue);

	for (i=0; i < vb->num_planes; i++){
		sg_t = vb2_dma_sg_plane_desc(vb, i);
		if (sg_t) {
			int max_desc, len;
			PDEBUG("\n%s: buffer has a sg_list",__func__);
			max_desc = idev->format.fmt.pix_mp.plane_fmt[i].sizeimage /
					VPP_PAGE_SIZE + 1;
			len = max_desc * sizeof(struct idev_dma_desc);
			idev_b->dma_desc[i] = kmalloc(len, GFP_KERNEL);
			PDEBUG("%s: i=%d alloc=%p\n", __func__,
			       i, idev_b->dma_desc[i]);
			if (idev_b->dma_desc[i] == NULL) {
				dev_err(idev->dev, "Unable to get DMA descriptor array\n");
				return -ENOMEM;
			}
			/*set the virtual address of the first seg to NULL */
			idev_b->dma_desc[i][0].virt = NULL;
		}
		else {
			buffer = (unsigned int *)vb2_plane_vaddr(vb,i);
			if (is_vmalloc_addr(buffer)) {
				struct page *pg = vmalloc_to_page((void *)buffer);
				physic =  page_to_phys(pg);
			}
			else {
				physic = (u32)vb2_dma_contig_plane_dma_addr(vb, i);
			}
			PDEBUG ("\nbuffer_init: virtual addr=0x%08X physical=0x%08X *buffer=0x%08X", (u32) buffer,physic,*(u32 *)buffer );
		}
	}
	PDEBUG ("\nvpp_api: buffer_init called");
	return 0;
 }

static int buffer_prepare(struct vb2_buffer *vb)
{
	struct vpp_dev *idev = vb2_get_drv_priv(vb->vb2_queue);
	struct idev_buffer *idev_b = to_idev_buffer(vb);
	unsigned long size = idev->format.fmt.pix.sizeimage;
	int i, j;
	struct sg_table *sg_t;
	int len = 0;

	if (vb2_plane_size(vb, 0) < size) {
		printk("\nbuffer too small (%lu < %lu)\n",
			 vb2_plane_size(vb, 0), size);
		return -EINVAL;
	}
	for (i=0; i < vb->num_planes; i++) {
		sg_t = vb2_dma_sg_plane_desc(vb, i);
		if (sg_t) {
			struct idev_dma_desc *desc = idev_b->dma_desc[i];
			struct scatterlist *sg;
			idev_b->ndesc[i] = 0;
			for_each_sg(sg_t->sgl, sg, sg_t->nents, j) {
				desc->dma_addr = sg_dma_address(sg);
				desc->seg_len = sg_dma_len(sg);
				len += desc->seg_len;
				PDEBUG("desc: plane=%d dmaadr=0x%X sglen=%d  ",
				       i, desc->dma_addr, desc->seg_len);
				desc++;
				idev_b->ndesc[i]++;
			}
			PDEBUG("\nplane=%d ndesc=%d len = %d",
			       i, idev_b->ndesc[i], len);
		}
	}

	vb2_set_plane_payload(vb, 0, size);
	return 0;
}

static void buffer_finish(struct vb2_buffer *vb)
{
	PDEBUG ("\n%s: num_buf = %d\n",__func__,
		atomic_read(&vb->vb2_queue->owned_by_drv_count));
}

static void buffer_cleanup(struct vb2_buffer *vb)
{
	struct idev_buffer *idev_b = to_idev_buffer(vb);
	struct vpp_dev *idev = vb2_get_drv_priv(vb->vb2_queue);
	u32 i;

	dev_info(idev->dev, "%s: Buffer=%p\n", __func__, vb);

	for (i = 0; i < vb->num_planes; i++) {
		dev_info(idev->dev, "%s: i=%d alloc=%p\n", __func__,
			 i, idev_b->dma_desc[i]);
		kfree(idev_b->dma_desc[i]);
	}
}

static void start_again(struct vpp_dev *idev)
{

	if (set_buffer(idev, CURRENT_BUFFER))
		dev_err(idev->dev, "can't set the current buffer\n");

	vpp_io_enable_memory(idev);
	vpp_io_turn_on_hw(idev);
	vpp_io_enable_double_buffering(idev);
	idev->streaming_is_being_stopped = false;

	set_buffer(idev, NEXT_BUFFER);

	dev_info(idev->dev, "streaming is on\n");
	vpp_io_enable_interrupt(idev,
				VPP_DIF_IRQ_STATUS_DIFFRAMERDY_MASK |
				VPP_DIF_IRQ_ENABLE_DIFOVERRUNIE_Y_MASK |
				VPP_DIF_IRQ_ENABLE_DIFOVERRUNIE_CHROMA_MASK);
}

static void buffer_queue(struct vb2_buffer *vb)
{
	struct vpp_dev *idev = vb2_get_drv_priv(vb->vb2_queue);
	struct idev_buffer *buf = to_idev_buffer(vb);
	unsigned long flags;
	int i;
	int number_of_buffers;

	PDEBUG ("\n buffer_queue: vb2->num_planes= %d\n", vb->num_planes);
	PDEBUG("num_buf=%d\n", atomic_read(&vb->vb2_queue->owned_by_drv_count));
	if (vb->num_planes != idev->format.fmt.pix_mp.num_planes)
		return;

	for (i = 0; i < VPP_MAX_NO_OF_PLANES_PER_FRAME; /*vb->num_planes;*/ i++) {
		if (vb->v4l2_planes[i].length < idev->format.fmt.pix_mp.plane_fmt[i].sizeimage) {
			printk("\n buffer_queue: plane sizes do not match, vb->v4l2_planes[%d].length=%d   idev->format.fmt.pix_mp.plane_fmt[%d].sizeimage = %d",i,vb->v4l2_planes[i].length ,i ,idev->format.fmt.pix_mp.plane_fmt[i].sizeimage);
			return;
		}
	}
	spin_lock_irqsave(&idev->qlock, flags);
	if (buf_is_in_q(idev, buf) == 0)
		list_add_tail(&buf->list, &idev->buf_list);
	spin_unlock_irqrestore(&idev->qlock, flags);
	PDEBUG ("\nbuffer_queue: buffer added to the internal queue\n");
	if (idev->pm_state == PM_WAITING_FOR_BUFFERS) {
		spin_lock_irqsave(&idev->qlock, flags);
		number_of_buffers = get_number_of_buffers(idev);
		spin_unlock_irqrestore(&idev->qlock, flags);
		if (number_of_buffers >= MIN_VPP_BUFFERS) {
			start_again(idev);
			idev->pm_state = PM_RUNNING;
		}
	}
}

void  return_all_buffers(struct vpp_dev *idev, enum vb2_buffer_state state)
{
	unsigned long flags;
	int ret;
	int i;
	struct vb2_queue *vq = &idev->queue;

	spin_lock_irqsave(&idev->qlock, flags);

	while (!list_empty(&idev->buf_list)) {
		struct idev_buffer *buf;
		buf = list_first_entry(&idev->buf_list,
				       struct idev_buffer, list);
		if ((u32)buf->list.next != (u32)LIST_POISON1) {
			list_del(&buf->list);
		} else {
			WARN(1, "re-deleting a buffer from the list.");
			break;
		}
		if (idev->use_mmu){
			ret = vpp_mmu_free (idev, buf);
			if (ret){
				printk("\nreturn_all_buffers. PANIC. TODO: we need to do something about this.");
			}
		}
		if (buf->vb.state == VB2_BUF_STATE_ACTIVE)
			vb2_buffer_done(&buf->vb, state);
		else
			dev_info(idev->dev, "Buffer state is not  VB2_BUF_STATE_ACTIVE");
	}
	/* release the current and next buffers */
	if (idev->current_buf) {
		if (idev->use_mmu)
			ret = vpp_mmu_free(idev, idev->current_buf);
		dev_info(idev->dev, "return_all_buffers. releasing current buffer\n");
		if (idev->current_buf->vb.state == VB2_BUF_STATE_ACTIVE)
			vb2_buffer_done(&idev->current_buf->vb, state);
		else
			dev_info(idev->dev, "Buffer state not VB2_BUF_STATE_ACTIVE");
	}
	if (idev->next_buf) {
		if (idev->use_mmu) {
			PDEBUG("mmu free: next buffer\n");
			ret = vpp_mmu_free(idev, idev->next_buf);
		}
		if (idev->next_buf->vb.state == VB2_BUF_STATE_ACTIVE) {
			vb2_buffer_done(&idev->next_buf->vb, state);
			dev_info(idev->dev, "\nreturn_all_buffers. next buffer released\n");
		} else {
			dev_info(idev->dev, "next buffer state not VB2_BUF_STATE_ACTIVE");
		}
	}

	for (i = 0; i < vq->num_buffers; ++i)
		if (vq->bufs[i]->state == VB2_BUF_STATE_ACTIVE)
			vb2_buffer_done(vq->bufs[i], VB2_BUF_STATE_ERROR);

	spin_unlock_irqrestore(&idev->qlock, flags);
}
static u32  convert_input_to_csc (struct vpp_dev *idev)
{
	u32 ret;
	s32 input_mode = v4l2_ctrl_g_ctrl(idev->csc_input_mode_ctrl);
	s32 colorspace = v4l2_ctrl_g_ctrl(idev->csc_input_colorspcae_ctrl);
	if (input_mode == VPP_B1BMODE_MODE_1_444_DVI_RGB_3X8 ||
		input_mode == VPP_B1BMODE_MODE_1_444_DVI_RGB_3X10 )
		ret = CSC_COLOURSPACE_RGB;
	/* The input format is not RGB */
	else if (colorspace == V4L2_COLORSPACE_REC709)
		ret = CSC_COLOURSPACE_YCC_BT709;
	else if (colorspace == V4L2_COLORSPACE_SMPTE170M)
		ret = CSC_COLOURSPACE_YCC_BT601;
	else {
		dev_err(idev->dev, "Could not determine the input colorspace\n");
		ret = -EINVAL;
	}
	return ret;
}

static int convert_range(enum v4l2_quantization q)
{
	if (q == V4L2_QUANTIZATION_LIM_RANGE)
		return CSC_RANGE_16_235;
	else if (q == V4L2_QUANTIZATION_FULL_RANGE ||
		 q == V4L2_QUANTIZATION_DEFAULT)
		return CSC_RANGE_0_255;
	else {
		pr_err("\n invalid input quantization range.");
		return -EINVAL;
	}
}
static CSC_eColourSpace  convert_v4l2_to_csc(enum v4l2_colorspace v4l2_cs)
{
	CSC_eColourSpace ret;
	switch (v4l2_cs)
	{
		/* We only support BT601 and BT709 */
		case	V4L2_COLORSPACE_SMPTE170M:
			ret = CSC_COLOURSPACE_YCC_BT601;
			break;
		case	V4L2_COLORSPACE_REC709:
			ret = CSC_COLOURSPACE_YCC_BT709;
			break;
		default:
			ret = -EINVAL;
			break;
	}
	return ret;
}
static int set_csc(struct vpp_dev *idev)
{
	CSC_sHSBCSettings sHSBCSettings;
	CSC_sConfiguration cfg = idev->csc_data.sResultantConfigurationData;
	CSC_eColourSpace in_colour_space;
	CSC_eColourSpace out_colour_space = 0;
	int ret = 0;
	CSC_eRange input_range;
	CSC_eRange output_range;
	u32 input_primary = CSC_COLOUR_PRIMARY_BT709;
	u32 output_primary = CSC_COLOUR_PRIMARY_BT709;
	bool swap = true;

	if (idev->vpp_user_defined_csc_matrix == true) {
		if (idev->vpp_user_defined_hscb == true) {
			dev_err(idev->dev, "hscb change not allowed\n");
			return -EINVAL;
		} else {
		/* Set the csc registers using the user-provided values. */
			ret = vpp_io_set_csc_regs(cfg, idev, false);
		}
	} else {
		sHSBCSettings.i32Hue = v4l2_ctrl_g_ctrl(idev->csc_hue_ctrl);
		sHSBCSettings.i32Contrast = v4l2_ctrl_g_ctrl(idev->csc_contrast_ctrl);
		sHSBCSettings.i32Saturation = v4l2_ctrl_g_ctrl(idev->csc_saturation_ctrl);
		sHSBCSettings.i32Brightness = v4l2_ctrl_g_ctrl(idev->csc_brightness_ctrl);
		/* Set input colorspace if it has not been set by the application */
		vpp_ctrl_set_csc_defaults(idev);
		input_range = v4l2_ctrl_g_ctrl(idev->csc_input_colorspcae_range_ctrl);
		input_range = convert_range(input_range);
		in_colour_space	= convert_input_to_csc (idev);
		if (idev->format.type == V4L2_BUF_TYPE_VIDEO_CAPTURE_MPLANE) {
			out_colour_space = convert_v4l2_to_csc (idev->format.fmt.pix_mp.colorspace);
			output_range = convert_range(idev->format.fmt.pix_mp.quantization);
		} else if (idev->format.type == V4L2_BUF_TYPE_VIDEO_CAPTURE) {
			out_colour_space = convert_v4l2_to_csc(idev->format.fmt.pix.colorspace);
			output_range = convert_range(idev->format.fmt.pix.quantization);
		} else {
			printk("\n set_csc: format.type not recognized");
			return -EINVAL;
		}
		if (input_range == -EINVAL || output_range == -EINVAL)
			return -EINVAL;
		PDEBUG("\n sat=0x%X, hue= 0x%X bright=0x%X con =0x%X",
		       sHSBCSettings.i32Saturation, sHSBCSettings.i32Hue,
		       sHSBCSettings.i32Brightness, sHSBCSettings.i32Contrast);
		if (in_colour_space == CSC_COLOURSPACE_YCC_BT601)
			input_primary = CSC_COLOUR_PRIMARY_BT601;
		if (out_colour_space == CSC_COLOURSPACE_YCC_BT709)
			output_primary = CSC_COLOUR_PRIMARY_BT709;
		ret = CSC_GenerateMatrix(in_colour_space, out_colour_space,
					 input_range, output_range,
					 input_primary, output_primary,
					 &sHSBCSettings, &cfg);
		if (ret == 0) {
			if (((in_colour_space == CSC_COLOURSPACE_RGB) &&
			    (out_colour_space == CSC_COLOURSPACE_RGB)) ||
			    ((in_colour_space != CSC_COLOURSPACE_RGB) &&
			    (out_colour_space != CSC_COLOURSPACE_RGB)))
				swap = false;
			ret = vpp_io_set_csc_regs(cfg, idev, swap);
		}
		else {
			printk("\n set_csc: CSC_GenerateMatrix failed");
			return ret;
		}
	}

	PDEBUG("\n set_csc: sat=0x%X, hue= 0x%X bright=0x%X con =0x%X",
	       sHSBCSettings.i32Saturation, sHSBCSettings.i32Hue,
	       sHSBCSettings.i32Brightness, sHSBCSettings.i32Contrast);

	return 0;
}

static int prepare_for_streaming(struct vpp_dev *idev)
{
	unsigned long flags;
	int ret = 0;

	vpp_io_pixel_clock(idev, true);
	spin_lock_irqsave(&idev->qlock, flags);
	cfg_use_mmu_flag(idev);
	if (idev->use_mmu)
		vpp_mmu_enable(idev);
	else
		vpp_mmu_disable(idev);
	spin_unlock_irqrestore(&idev->qlock, flags);
	if (vpp_io_is_double_buffer_enabled(idev)) {
		dev_err(idev->dev, "double buffering already enabled.\n");
		return -EINVAL;
	}
	ret = vpp_io_mmu_set_tiling(idev);
	if (ret)
		return ret;
	set_csc(idev);
	ret = vpp_io_set_timing_registers(idev, &idev->timings);
	if (ret) {
		dev_err(idev->dev, "failed to set timings\n");
		return ret;
	}

	ret |= vpp_io_set_output_format(idev);
	ret |= vpp_set_scalers(idev);
	ret |= vpp_io_set_input_mode(idev);
	ret |= vpp_io_set_input_itu_or_dvi(idev);
	ret |= vpp_io_set_input_polarity(idev);
	if (ret)
		return ret;
	ret = vpp_io_set_burst_size(idev);
	if (ret)
		return ret;
	ret = vpp_io_set_input_swap(idev);
	if (ret)
		return ret;
	idev->about_to_start_streaming = true;
	ret = v4l2_ctrl_handler_setup(&idev->ctrl_handler);
	if (ret) {
		dev_err(idev->dev, "v4l2_ctrl_handler_setup failed\n");
		return ret;
	}
	idev->about_to_start_streaming = false;
	idev->buffer_error = false;
	ret = vpp_ctrl_grab_controls(idev);

	return ret;
}

static int start_streaming(struct vb2_queue *vq, unsigned int count)
{
	struct vpp_dev *idev = vb2_get_drv_priv(vq);
	struct idev_buffer *idev_buf;
	int ret;
	int number_of_buffers = get_number_of_buffers(idev);

	dev_info(idev->dev, "number_of_buffers = %d\n", number_of_buffers);
	if (number_of_buffers < MIN_VPP_BUFFERS) {
		dev_err(idev->dev, "alloc buffs (%d) < min buffs (%d)\n",
			number_of_buffers, MIN_VPP_BUFFERS);
		ret = -EINVAL;
		goto Error;
	}
	ret = prepare_for_streaming(idev);
	if (ret)
		goto Error;	/* do not allow any changes in controls */

	if (idev->pm_state == PM_RUNNING)
		idev->sequence = 0;
	dev_info(idev->dev, "%s setting buffer\n", __func__);
	ret = set_buffer ( idev, CURRENT_BUFFER);
	if (ret)
		goto Error;
	idev_buf = idev->current_buf;
	printk("\nlocking ctrl_handler");
	mutex_lock(idev->ctrl_handler.lock);
	printk("\nctrl_handler locked");
	vpp_io_mmu_invalidate_cache(idev);
	vpp_io_enable_memory(idev);
	printk("\ncalling vpp_io_turn_on_hw");
	ret = vpp_io_turn_on_hw(idev);

	printk("\nunlocking ctrl_handler");
	mutex_unlock(idev->ctrl_handler.lock);
	printk("\nctrl_handler unlocked");

	/* New values of the dbl-buf regs are used when difframerdy = 1 */
	vpp_io_enable_double_buffering(idev);
	idev->streaming_is_being_stopped = false;
	if(!ret) {
#ifdef __VPP_TARGET_IS_SIMULATOR__
		if (v4l2_ctrl_g_ctrl(idev->vp_kick_ctrl) == VPP_KICK_AUTO)
			vpp_io_kick_simulator(idev);
		else
			wait_for_kick(idev);
#endif
		ret = set_buffer(idev, NEXT_BUFFER);
		if (ret)
			goto Error;
	}
	dev_info(idev->dev, "\n streaming is on. returning 0\n");
	vpp_io_enable_interrupt(idev,
				VPP_DIF_IRQ_STATUS_DIFFRAMERDY_MASK |
				VPP_DIF_IRQ_ENABLE_DIFOVERRUNIE_Y_MASK |
				VPP_DIF_IRQ_ENABLE_DIFOVERRUNIE_CHROMA_MASK);
	return 0;
Error:
	return_all_buffers(idev, VB2_BUF_STATE_QUEUED);
	vpp_io_pixel_clock(idev, false);
	printk("\n start streaming failed. returning ret =%d",ret);
	return ret;

}

/*
 * Stop the DMA engine. Any remaining buffers in the DMA queue are dequeued
 * and passed on to the vb2 framework marked as STATE_ERROR.
 */
static void stop_streaming(struct vb2_queue *vq)
{
	struct vpp_dev *idev = vb2_get_drv_priv(vq);
	int ret = 0;

	if (idev == NULL){
		dev_info(idev->dev, "%s: idev is NULL", __func__);
		return;
	}
	if (idev->streaming_is_being_stopped) {
		dev_info(idev->dev, "shouldn't be here");
		return;
	}
	printk("\n%s waiting for mutex", __func__);
	mutex_lock(&idev->vpp_workqueue_mutex);
	printk("\n%s got the  mutex", __func__);
	idev->streaming_is_being_stopped = true;
	mutex_unlock(&idev->vpp_workqueue_mutex);
	printk("\n%s released the  mutex", __func__);
	//wait for the completion of the next interrupt. timeout = 0.25 seconds.
	#ifdef __VPP_TARGET_IS_SIMULATOR__
		ret = (int)wait_for_completion_interruptible_timeout (&idev->stop_streaming_notifier, 25000000);
	#else
		ret = (int)wait_for_completion_interruptible_timeout (&idev->stop_streaming_notifier, HZ);
	#endif
	if (ret <= 0) {
		printk("\n%s timeout happened", __func__);
	} else if (ret > 0) {
		printk ("\n%s time= %d",  __func__, ret);
	}
	/* Double buffering of registers is disabled.
	 * Any write to registers will take effect immediately.
	 */
	vpp_io_disable_double_buffering(idev);
	vpp_io_disable_memory(idev);

	ret = vpp_io_soft_reset(idev);
	PDEBUG("\n vpp_dev_irq: reset the device");
	if (ret == 0)
		ret = vpp_io_turn_off_hw(idev);
	PDEBUG("\n turned off the device");
	/* Release all active buffers */
	return_all_buffers(idev, VB2_BUF_STATE_ERROR);

	vpp_io_disable_interrupt(idev, ~0u);
	printk("\n%s disabled interrupts", __func__);
	flush_work(&idev->set_next_buffer_work);
	printk("\n%s flushed work", __func__);
	/*remove all requested workqueue works */
	printk ("\n%s: calling cancel_work_sync", __func__);
	ret = cancel_work_sync(&idev->set_next_buffer_work);
	if (ret == 0)
		flush_workqueue(idev->set_next_buffer_workqueue);
	else
		dev_err(idev->dev, "cancel_work_sync error = %d", ret);
	/*reset scaler coeffs if set by the user. User needs to set them again.
	 * otherwise they are calculated by the driver.
	*/
	if (idev->vpp_user_defined_hscale_coeff ==true){
		idev->vpp_user_defined_hscale_coeff =false;
		idev->v.hs_old_output_size = 0;
		idev->v.hs_old_input_size = 0;
		idev->v.hs_en = false;
	}
	idev->vpp_user_defined_csc_matrix = false;
	idev->vpp_user_defined_hscb = false;

	if (idev->vpp_user_defined_vscale_coeff ==true){
		idev->vpp_user_defined_vscale_coeff =false;
		idev->v.vs_old_output_size = 0;
		idev->v.vs_old_input_size = 0;
		idev->v.vs_en = false;
	}

	PDEBUG("\n%s: ungrabbing controls", __func__);
	ret = vpp_ctrl_ungrab_controls(idev);
	if (ret)
		dev_err(idev->dev, " ungrab_controls failed\n");

	printk("\n%s ungrabbed controls", __func__);
	/* cleare the page table directory */
	memset(idev->pdir, 0, VPP_PAGE_DIR_BYTES);
	idev->irq_message[0] = '\0';
	vpp_io_pixel_clock(idev, false);
	printk("\n%s: returning with ret = %d", __func__, ret);
	return;
}

int vpp_buf_resume(struct vpp_dev *idev)
{
	int ret;
	int pm_state = idev->pm_state;

	idev->pm_state = PM_WAITING_FOR_BUFFERS;
	/* Set MMU directory base address */
	vpp_mmu_write(idev, VPP_MMU_DIR_BASE_ADDR_OFFSET, idev->pdir_bus);

	ret = prepare_for_streaming(idev);
	if (ret) {
		return_all_buffers(idev, VB2_BUF_STATE_QUEUED);
		vpp_io_pixel_clock(idev, false);
		dev_err(idev->dev, " %s failed. ret =%d\n", __func__, ret);
		idev->pm_state = pm_state;
		return ret;
	} else {
		mutex_lock(idev->ctrl_handler.lock);
		/*reset the mmu cache */
		vpp_io_mmu_invalidate_cache(idev);
		mutex_unlock(idev->ctrl_handler.lock);
		return 0;
	}
}

int vpp_buf_suspend(struct vpp_dev *idev)
{
	int ret = 0;

	vb2_discard_done(&idev->queue);
	if (idev->use_mmu) {
		ret = vpp_mmu_free(idev, idev->current_buf);
		if (ret) {
			dev_err(idev->dev, "unable to free mmu cur_buf\n");
			return ret;
		}
		ret = vpp_mmu_free(idev, idev->next_buf);
		if (ret) {
			dev_err(idev->dev, "unable to free mmu next_buf\n");
			return ret;
		}
	}

	return_all_buffers(idev, VB2_BUF_STATE_ERROR);

	return ret;
}
