/*
 * This is a V4L2 VPP Driver.
 *
 * Copyright 2015 Imagination Technologies. All rights reserved.
 *
 * This program is free software; you may redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/kmod.h>
#include <linux/mutex.h>

#include <linux/videodev2.h>
#include <linux/v4l2-dv-timings.h>
#include <media/v4l2-device.h>
#include <media/v4l2-dev.h>
#include <media/v4l2-ioctl.h>
#include <media/v4l2-dv-timings.h>
#include <media/v4l2-ctrls.h>
#include "vpp_api.h"
#include "vpp_ctrl.h"
#include "vpp_mmu.h"
#include "vpp_io.h"
#include "validation.h"
#include "csc2/csc2.h"
#include "csc2/csc2_data.h"
#include "csc2/fixedpointmaths.h"

static int vpp_dev_s_ctrl(struct v4l2_ctrl *ctrl)
{
	/* We will write to the registers in this function but will make
	 * sure that it does not happen while streaming is on. Note that apart
	 * from checking the vb2_streaming flag, we also use v4l2_ctrl_grab()
	 * to deactivate the controls while streaming is on.
	 */
	int i;
	int ret;
	struct vpp_dev *idev = container_of(ctrl->handler,
					    struct vpp_dev, ctrl_handler);
	int busy = vb2_is_streaming(&idev->queue);
	if (idev->pm_state == PM_WAITING_FOR_BUFFERS)
		busy = 0;
	if (idev->about_to_start_streaming)
		PDEBUG("\n vpp_dev_s_ctrl called by v4l2_ctrl_handler_setup()");
	if (!idev->about_to_start_streaming) {
		if (ctrl->id == V4L2_CID_BRIGHTNESS ||
		    ctrl->id == V4L2_CID_CONTRAST ||
		    ctrl->id == V4L2_CID_SATURATION ||
		    ctrl->id == V4L2_CID_HUE) {
			if (idev->vpp_user_defined_csc_matrix)
				pr_info("\ncsc matrix is ignored.");
			idev->vpp_user_defined_hscb = true;
			idev->vpp_user_defined_csc_matrix = false;
		}
	}
	switch (ctrl->id) {
	case V4L2_CID_BRIGHTNESS:
		if (VPP_DATA_VALID(ctrl->val, CSC_MIN_BRIGHTNESS,
				   CSC_MAX_BRIGHTNESS))
			idev->csc_data.sHSBCSettings.i32Brightness = ctrl->val;
		else
			return -EINVAL;
		PDEBUG("\nvpp_dev_s_ctrl: brightness is %d\n", ctrl->val);
		break;
	case V4L2_CID_CONTRAST:
		if (VPP_DATA_VALID(ctrl->val, CSC_MIN_CONTRAST,
				   CSC_MAX_CONTRAST))
			idev->csc_data.sHSBCSettings.i32Contrast = ctrl->val;
		else
			return -EINVAL;
		PDEBUG("\nvpp_dev_s_ctrl: contrast is %d\n", ctrl->val);
		break;
	case V4L2_CID_SATURATION:
		if (VPP_DATA_VALID(ctrl->val, CSC_MIN_SATURATION,
				   CSC_MAX_SATURATION))
			idev->csc_data.sHSBCSettings.i32Saturation = ctrl->val;
		else
			return -EINVAL;
		PDEBUG("\nvpp_dev_s_ctrl: saturation is %d\n", ctrl->val);
		break;
	case V4L2_CID_HUE:
		if (VPP_DATA_VALID(ctrl->val, CSC_MIN_HUE, CSC_MAX_HUE))
			idev->csc_data.sHSBCSettings.i32Hue = ctrl->val;
		else
			return -EINVAL;
		PDEBUG("\nvpp_dev_s_ctrl: hue is %d\n", ctrl->val);
		break;
	case IMG_VPP_CID_INPUT_COLORSPACE:
		PDEBUG("\nvpp_dev_s_ctrl: set color space is %d\n", ctrl->val);

		switch (ctrl->val) {
		case V4L2_COLORSPACE_SMPTE170M:/* ITU-R 601 */
		case V4L2_COLORSPACE_REC709:
		case V4L2_COLORSPACE_DEFAULT:
			PDEBUG("\n colour space is:%d\n", ctrl->val);
			break;
		default:
			pr_err("\n colour space invalid: %d\n", ctrl->val);
			return -EINVAL;
		}
		break;
	case IMG_VPP_CID_INPUT_MODE:
		PDEBUG("\nvpp_dev_s_ctrl input mode is:%d = %d\n",
		       ctrl->val, idev->csc_input_mode_ctrl->val);
		ret = vpp_io_set_input_mode(idev);
		ret = vpp_io_set_input_itu_or_dvi(idev);
		if (ret)
			return ret;
		break;
	case IMG_VPP_CID_INPUT_POLARITIES:
		if (busy) {
			pr_err("\nvpp_dev_s_ctrl: vb2 is streaming\n");
			return -EBUSY;
		}
		PDEBUG("\nvpp_dev_s_ctrl input polarity is:%d\n", ctrl->val);
		ret = vpp_io_set_input_polarity(idev);
		if (ret)
			return ret;
		break;
	case IMG_VPP_CID_SWAP_INTERNAL_UV_LANES:
		/* This is different from swapping external input lanes.
		 * This is an internal change in lanes and usually not needed.
		 */
		if (busy) {
			pr_err("\nvpp_dev_s_ctrl: vb2 is streaming\n");
			return -EBUSY;
		}
		break;
	case IMG_VPP_CID_HSCALER_COEFFS:
	case IMG_VPP_CID_VSCALER_COEFFS:
	case IMG_VPP_CID_INPUT_COLORSPACE_RANGE:
		break;
	case IMG_VPP_CID_CSC_MULT:
	case IMG_VPP_CID_CSC_IN_OFFSET:
	case IMG_VPP_CID_CSC_OUT_OFFSET:
	ctrl = idev->csc_mult;

	PDEBUG("\n csc_mult :ctrl->has_changed=%d ctrl->is_new=%d ",
	       ctrl->has_changed, ctrl->is_new);
	if (ctrl->is_new) {
		/* Check that the number of elements is 9 */
		if (ctrl->elems != ctrl->dims[0] * ctrl->dims[1]) {
			pr_err("vpp_dev_s_ctrl: invalid size of array");
			return -EINVAL;
		}
		if (!idev->about_to_start_streaming) {
			PDEBUG("\n vpp_dev_s_ctrl called by user ioctl.");
			if (idev->vpp_user_defined_hscb)
				pr_info("\n hsbc setting is ignored");
			idev->vpp_user_defined_csc_matrix = true;
			idev->vpp_user_defined_hscb = false;
		}
		for (i = 0; i < ctrl->elems; ++i) {
			PDEBUG(" 0x%X ",
			       ctrl->p_new.p_s32[i] & VPP_CSC_COEFFS_MASK);
		/* Change the fractional bits from
		 * VPP_CSC_MATRIX_FRACTIONAL_BITS to CSC_FRACTIONAL_BITS
		 * and store the values
		 */
			idev->csc_data.sResultantConfigurationData.sCoefficients.ai32Coefficients[i/3][i % 3] =
			ctrl->p_new.p_s32[i] << (CSC_FRACTIONAL_BITS - VPP_CSC_MATRIX_FRACTIONAL_BITS);
		}
	}
	PDEBUG("\n");

	ctrl = idev->csc_in_offset;
	PDEBUG("\n csc_in_offset :ctrl->has_changed=%d ctrl->is_new=%d ",
	       ctrl->has_changed, ctrl->is_new);
	ctrl = idev->csc_out_offset;
	PDEBUG("\n csc_out_offset :ctrl->has_changed=%d ctrl->is_new=%d ",
	       ctrl->has_changed, ctrl->is_new);
		break;
	case IMG_VPP_CID_TILING_METHOD:
		break;
#ifdef __VPP_TARGET_IS_SIMULATOR__
	case IMG_VPP_CID_VP_KICK:
		dev_info(idev->dev, "vpp_kick set to %d  ", ctrl->val);
		break;
#endif
	default:
		pr_err("\nvpp_dev_s_ctrl: unknown ctrl id 0x%x", (u32)ctrl->id);
		return -EINVAL;
	}
	return 0;
}

static const struct v4l2_ctrl_ops idev_ctrl_ops = {
	.s_ctrl = vpp_dev_s_ctrl,
};

static void vpp_ctrl_type_init(const struct v4l2_ctrl *ctrl,
			       union v4l2_ctrl_ptr ptr)
{
	unsigned int i;

	PDEBUG("\nvpp_ctrl_type_init started ctrl->elems= %d ctrl->id=0x%X",
	       ctrl->elems, ctrl->id);
	switch (ctrl->id) {
	case IMG_VPP_CID_CSC_MULT:

		/*
		 * Initialize the diagonal to 1.0 and all other elements to
		 * 0.0.
		 */
		for (i = 0; i < ctrl->elems; ++i)
			ptr.p_s32[i] = (i % 4) ? 0 : 256;
		break;
	}
}

static const struct v4l2_ctrl_type_ops vpp_ctrl_type_ops = {
	.equal = v4l2_ctrl_type_std_equal,
	.init = vpp_ctrl_type_init,
	.log = v4l2_ctrl_type_std_log,
	.validate = v4l2_ctrl_type_std_validate,
};

static const struct v4l2_ctrl_config vpp_ctrls[] = {
	{   /* 0 */
		.ops = &idev_ctrl_ops,
		.type_ops = &vpp_ctrl_type_ops,
		.id = IMG_VPP_CID_CSC_MULT,
		.name = "CSC Multiplier",
		.type = V4L2_CTRL_TYPE_S32,
		.def = 0,
		.min = -1 * CSC_MATRIX_COEFFS_MIN,
		.max = CSC_MATRIX_COEFFS_MAX,
		.step = 1,
		.dims = { 3, 3 },
		.flags = V4L2_CTRL_FLAG_EXECUTE_ON_WRITE,
	}, {/* 1 */
		.ops = &idev_ctrl_ops,
		.id = IMG_VPP_CID_CSC_IN_OFFSET,
		.name = "CSC Input Offset",
		.type = V4L2_CTRL_TYPE_S32,
		.def = 0,
		.min = -4096,
		.max = 4095,
		.step = 1,
		.dims = { 3 },
		.flags = V4L2_CTRL_FLAG_EXECUTE_ON_WRITE,
	},
	{/* 2 */
		.ops = &idev_ctrl_ops,
		.id = IMG_VPP_CID_CSC_OUT_OFFSET,
		.name = "CSC Output Offset",
		.type = V4L2_CTRL_TYPE_S32,
		.def = 0,
		.min = -4096,
		.max = 4095,
		.step = 1,
		.dims = { 3 },
		.flags = V4L2_CTRL_FLAG_EXECUTE_ON_WRITE,
	},
	{/* 3 */
		.ops = &idev_ctrl_ops,
		.type_ops = &vpp_ctrl_type_ops,
		.id = IMG_VPP_CID_HSCALER_COEFFS,
		.name = "Scaler H Coefficients",
		.type = V4L2_CTRL_TYPE_S16,
		.def = 0,
		.min = -128,
		.max = 127,
		.step = 1,
		.dims = { 17, 4 },
	},
	{/* 4 */
		.ops = &idev_ctrl_ops,
		.id = IMG_VPP_CID_INPUT_COLORSPACE,
		.name = "Input Color Space",
		.type = V4L2_CTRL_TYPE_INTEGER,
		.def = V4L2_COLORSPACE_DEFAULT,
		.min = 0,
		.max = 20,
		.step = 1,
	},
	{/* 5 */
		.ops = &idev_ctrl_ops,
		.id = IMG_VPP_CID_INPUT_MODE,
		.name = "Input Mode",
		.type = V4L2_CTRL_TYPE_INTEGER,
		.def = 1,
		.min = 1,
		.max = VPP_NO_OF_B1BMODE_MODE,
		.step = 1,
	},
	{/* 6 */
		.ops = &idev_ctrl_ops,
		.id = IMG_VPP_CID_INPUT_COLORSPACE_RANGE,
		.name = "Input Color Space Range",
		.type = V4L2_CTRL_TYPE_INTEGER,
		.def = V4L2_QUANTIZATION_DEFAULT,
		.min = 0,
		.max = CSC_NO_OF_RANGES,
		.step = 1,
	},
	{/* 7 */
		.ops = &idev_ctrl_ops,
		.id = IMG_VPP_CID_INPUT_POLARITIES,
		.name = "Input data enable and field Polarity",
		.type = V4L2_CTRL_TYPE_INTEGER,
		.def = VPP_POLARITY_FIELD_HIGH_DENABLE_LOW_VYSNC_HIGH_HSYNC_HIGH,
		.min = VPP_POLARITY_FIELD_HIGH_DENABLE_LOW_VYSNC_HIGH_HSYNC_HIGH,
		.max = VPP_POLARITY_INVALID - 1,
		.step = 1,
	},
	{/* 8, we may remove this later */
		.ops = &idev_ctrl_ops,
		.id = IMG_VPP_CID_SWAP_INTERNAL_UV_LANES,
		.name = "Control for swapping internal U and V after CSC",
		.type = V4L2_CTRL_TYPE_INTEGER,
		.def = 0,
		.min = 0,
		.max = 1,
		.step = 1,
	},
	{/* 9 */
		.ops = &idev_ctrl_ops,
		.type_ops = &vpp_ctrl_type_ops,
		.id = IMG_VPP_CID_VSCALER_COEFFS,
		.name = "Scaler V Coefficients",
		.type = V4L2_CTRL_TYPE_S16,
		.def = 0,
		.min = -128,
		.max = 127,
		.step = 1,
		.dims = {9, 4 },
	},
	{/* 10 */
		.ops = &idev_ctrl_ops,
		.id = IMG_VPP_CID_TILING_METHOD,
		.name = "Tiling method",
		.type = V4L2_CTRL_TYPE_INTEGER,
		.def = VPP_TILING_NONE,
		.min = VPP_TILING_NONE,
		.max = VPP_TILING_INVALID - 1,
		.step = 1,
	},
#ifdef __VPP_TARGET_IS_SIMULATOR__
	{/* 11 */
		.ops = &idev_ctrl_ops,
		.id = IMG_VPP_CID_VP_KICK,
		.name = "vp kick",
		.type = V4L2_CTRL_TYPE_INTEGER,
		.def = VPP_KICK_AUTO,
		.min = VPP_KICK_AUTO,
		.max = VPP_KICK_INVALID - 1,
		.step = 1,
	},
#endif
};

int vpp_ctrl_register_ctrls(struct vpp_dev *idev)
{
	struct v4l2_ctrl_handler *hdl = &idev->ctrl_handler;
	int ret = 0;

	idev->csc_brightness_ctrl =
		v4l2_ctrl_new_std(hdl, &idev_ctrl_ops, V4L2_CID_BRIGHTNESS,
				  CSC_MIN_BRIGHTNESS, CSC_MAX_BRIGHTNESS, 1,
				  VPP_DEFAULT_BRIGHTNESS);
	idev->csc_contrast_ctrl =
		v4l2_ctrl_new_std(hdl, &idev_ctrl_ops, V4L2_CID_CONTRAST,
				  CSC_MIN_CONTRAST, CSC_MAX_CONTRAST, 1,
				  VPP_DEFAULT_CONTRAST);
	idev->csc_saturation_ctrl =
		v4l2_ctrl_new_std(hdl, &idev_ctrl_ops, V4L2_CID_SATURATION,
				  CSC_MIN_SATURATION, CSC_MAX_SATURATION, 1,
				  VPP_DEFAULT_SATURATION);
	idev->csc_hue_ctrl =
		v4l2_ctrl_new_std(hdl, &idev_ctrl_ops, V4L2_CID_HUE,
				  CSC_MIN_HUE, CSC_MAX_HUE, 1, VPP_DEFAULT_HUE);
	PDEBUG("\n registering mul contol\n");
	idev->vpp_user_defined_csc_matrix = false;
	idev->vpp_user_defined_hscb = false;
	idev->csc_mult = v4l2_ctrl_new_custom(hdl, &vpp_ctrls[0], NULL);
	PDEBUG("\n registering in_offset contol\n");
	idev->csc_in_offset = v4l2_ctrl_new_custom(hdl,  &vpp_ctrls[1], NULL);
	PDEBUG("\n registering out_offset contol\n");
	idev->csc_out_offset = v4l2_ctrl_new_custom(hdl, &vpp_ctrls[2], NULL);
	PDEBUG("\n registering hscaler contol\n");
	idev->hscaler_coeffs = v4l2_ctrl_new_custom(hdl, &vpp_ctrls[3], NULL);
	PDEBUG("\n registering vscaler contol\n");
	idev->vscaler_coeffs = v4l2_ctrl_new_custom(hdl, &vpp_ctrls[9], NULL);
	idev->csc_input_colorspcae_ctrl =
		v4l2_ctrl_new_custom(&idev->ctrl_handler, &vpp_ctrls[4], NULL);
	if (!idev->csc_input_colorspcae_ctrl)
		pr_err("\n csc_input_colorspcae_ctrl registration failed.\n");
	idev->csc_input_mode_ctrl =
		v4l2_ctrl_new_custom(hdl, &vpp_ctrls[5], NULL);

	idev->csc_input_colorspcae_range_ctrl =
		v4l2_ctrl_new_custom(hdl, &vpp_ctrls[6], NULL);

	idev->input_pol_ctrl =
		v4l2_ctrl_new_custom(hdl, &vpp_ctrls[7], NULL);

	idev->swap_internal_uv_lanes_ctrl =
		v4l2_ctrl_new_custom(hdl, &vpp_ctrls[8], NULL);
	idev->tiling_ctrl =
		v4l2_ctrl_new_custom(hdl, &vpp_ctrls[10], NULL);
#ifdef __VPP_TARGET_IS_SIMULATOR__
	idev->vp_kick_ctrl =
		v4l2_ctrl_new_custom(hdl, &vpp_ctrls[11], NULL);
#endif
	if (hdl->error) {
		ret = hdl->error;
		return ret;
	}

	v4l2_ctrl_cluster(3, &idev->csc_mult);
	idev->v4l2_dev.ctrl_handler = hdl;
	return 0;
}

int vpp_ctrl_grab_controls(struct vpp_dev *idev)
{
	int ret = 0;

	v4l2_ctrl_grab(idev->csc_mult, true);
	v4l2_ctrl_grab(idev->csc_in_offset, true);
	v4l2_ctrl_grab(idev->csc_out_offset, true);
	v4l2_ctrl_grab(idev->csc_input_colorspcae_ctrl, true);
	v4l2_ctrl_grab(idev->csc_input_colorspcae_range_ctrl, true);
	v4l2_ctrl_grab(idev->csc_input_mode_ctrl, true);
	v4l2_ctrl_grab(idev->input_pol_ctrl, true);
	v4l2_ctrl_grab(idev->vscaler_coeffs, true);
	v4l2_ctrl_grab(idev->swap_internal_uv_lanes_ctrl, true);
	v4l2_ctrl_grab(idev->tiling_ctrl, true);
	return ret;
}

int vpp_ctrl_ungrab_controls(struct vpp_dev *idev)
{
	int ret = 0;

	v4l2_ctrl_grab(idev->csc_mult, false);
	v4l2_ctrl_grab(idev->csc_in_offset, false);
	v4l2_ctrl_grab(idev->csc_out_offset, false);
	v4l2_ctrl_grab(idev->csc_input_colorspcae_ctrl, false);
	v4l2_ctrl_grab(idev->csc_input_colorspcae_range_ctrl, false);
	v4l2_ctrl_grab(idev->csc_input_mode_ctrl, false);
	v4l2_ctrl_grab(idev->input_pol_ctrl, false);
	v4l2_ctrl_grab(idev->vscaler_coeffs, false);
	v4l2_ctrl_grab(idev->swap_internal_uv_lanes_ctrl, false);
	v4l2_ctrl_grab(idev->tiling_ctrl, false);

	return ret;
}

void vpp_ctrl_release_ctrls(struct vpp_dev *idev)
{
	vpp_ctrl_ungrab_controls(idev);
}

void vpp_ctrl_set_csc_defaults(struct vpp_dev *idev)
{
	s32 input_mode = v4l2_ctrl_g_ctrl(idev->csc_input_mode_ctrl);

	if (v4l2_ctrl_g_ctrl(idev->csc_input_colorspcae_ctrl) ==
			     V4L2_COLORSPACE_DEFAULT) {
		/* The input color space has not been set by the application.
		 * Find the default color space for this timing.
		 */
		 if (idev->timings.bt.height <= 576 &&
		     idev->timings.bt.height >= 240)
			v4l2_ctrl_s_ctrl(idev->csc_input_colorspcae_ctrl,
					 V4L2_COLORSPACE_SMPTE170M);
		 else if (idev->timings.bt.height > 576)
			v4l2_ctrl_s_ctrl(idev->csc_input_colorspcae_ctrl,
					 V4L2_COLORSPACE_REC709);
		 else
			dev_err(idev->dev, "couldn't find the colorspace.\n");
	}
	if (v4l2_ctrl_g_ctrl(idev->csc_input_colorspcae_range_ctrl) ==
			     V4L2_QUANTIZATION_DEFAULT) {
		if(input_mode != VPP_B1BMODE_MODE_1_444_DVI_RGB_3X8 &&
		   input_mode != VPP_B1BMODE_MODE_1_444_DVI_RGB_3X10)
		   /* default for yuv is V4L2_QUANTIZATION_LIM_RANGE */
			v4l2_ctrl_s_ctrl(idev->csc_input_colorspcae_range_ctrl,
					 V4L2_QUANTIZATION_LIM_RANGE);
		else
			v4l2_ctrl_s_ctrl(idev->csc_input_colorspcae_range_ctrl,
					 V4L2_QUANTIZATION_FULL_RANGE);
	}
}
