#ifndef _VPP_IO_H
#define _VPP_IO_H
#include "vpp_api.h"
#include <video/img_vpp.h>
#include "vpp_regs.h"
#include "vpp_mmu_regs.h"


/*
 * Bitfield operations
 * For each argument field, the following preprocessor macros must exist
 * field##_MASK - the number of bits in the bit field
 * field##_SHIFT - offset from the first bit
 */
#define PLACE_FIELD(field, val) \
	(((u32)(val) << (field##_SHIFT)) & (field##_MASK))

#define ADJ_FIELD(x, field, val) \
	(((x) & ~(field##_MASK)) \
	| PLACE_FIELD(field, val))

#define SET_FIELD(x, field, val) \
	(x) = ADJ_FIELD(x, field, val)

#define GET_FIELD(x, field) \
	(((x) & (field##_MASK)) >> (field##_SHIFT))

/* Keeps most significant bits */
#define MOVE_FIELD(x, o1, l1, o2, l2) \
	(((x) >> ((o1) + (l1) - (l2))) << (o2))

#define MAX_FIELD_VALUE(field) \
	((field##_MASK) >> (field##_SHIFT))



/* Info about video plane addresses */
#define VPP_YADDR_BITS		(VPP_DIF_FRAME_BASE_ADDR_EVEN_DIFFRBASE_EVEN_LENGTH)
/* TODO: This is related to the memory burst size which is a system level value.
 * Can (must) be extracted from memory burst size? discuss with Ed and Dan.
 */
#define VPP_YADDR_ALIGN		5
#define VPP_UADDR_BITS		(VPP_DIF_FRAME_CHROMA_BASE_ADDR_EVEN_DIFFRCBASE_EVEN_LENGTH)
#define VPP_UADDR_ALIGN		5
#define VPP_VADDR_BITS		(VPP_DIF_FRAME_CHROMA_BASE_ADDR_EVEN_DIFFRCBASE_EVEN_LENGTH)
#define VPP_VADDR_ALIGN		5

#define VPP_YSTRIDE_BITS	(VPP_DIF_FRAME_STRIDE_DIFFRSTRIDE_LENGTH)
#define VPP_YSTRIDE_ALIGN	5


/* Maximum 6 bytes per pixel for RGB161616 */
#define VPP_MAX_IMAGE_BYTES (VPP_MAX_OUTPUT_WIDTH * VPP_MAX_OUTPUT_HEIGHT * 6)

/* Round up */
//#define VPP_MAX_IMAGE_PAGES ((VPP_MAX_IMAGE_BYTES+PAGE_SIZE-1)/PAGE_SIZE)

#define VPP_YADDR_MAX		(((1 << VPP_YADDR_BITS) - 1) << VPP_YADDR_ALIGN)
#define VPP_UADDR_MAX		(((1 << VPP_UADDR_BITS) - 1) << VPP_UADDR_ALIGN)
#define VPP_VADDR_MAX		(((1 << VPP_VADDR_BITS) - 1) << VPP_VADDR_ALIGN)
#define VPP_YSTRIDE_MAX		((1 << VPP_YSTRIDE_BITS) << VPP_YSTRIDE_ALIGN)
#define VPP_YADDR_ALIGNMASK	((1 << VPP_YADDR_ALIGN) - 1)
#define VPP_UADDR_ALIGNMASK	((1 << VPP_UADDR_ALIGN) - 1)
#define VPP_VADDR_ALIGNMASK	((1 << VPP_VADDR_ALIGN) - 1)
#define VPP_YSTRIDE_ALIGNMASK	((1 << VPP_YSTRIDE_ALIGN) - 1)

#define VPP_CSC_COEFFS_MASK 	VPP_DIF_AA_V_INACTIVE_DIFAAVACTIVEN_LSBMASK


//mmu
#define VPP_PAGE_SIZE       4096
#define VPP_PAGES_PER_PTBL	1024  // Number of page addresses in each page table
#define VPP_PAGE_MASK       (~(VPP_PAGE_SIZE-1))
/*
 * The total size of the page directories required  in bytes
 */
#define VPP_PAGE_DIR_BYTES 4096
#define VPP_PAGE_DIR_ENTERIES VPP_PAGE_DIR_BYTES /4
//#define CONFIG_FB_VPP_MMU 1

/* Round up */
#define VPP_MAX_IMAGE_PAGES  ((VPP_MAX_IMAGE_BYTES + VPP_PAGE_SIZE - 1) / VPP_PAGE_SIZE)

/* The number of pointers per page table */
#define VPP_PTRS_PER_PTBL	(VPP_PAGE_SIZE / 4)

/*
 * The number of pointers per page directory, sized to allow the maximum
 * supported image size.
 * 1 is added to allow for buffers not starting on a page boundary
 */
#define VPP_PTRS_PER_PDIR	((VPP_MAX_IMAGE_PAGES / VPP_PTRS_PER_PTBL) + 1)



/*
 * This is the maximum number of page tables that we will need to map an image.
 * Added 4 as we may have 4 buffers per frame and page tables may overlap.
*/
#define MAX_PAGE_TABLES_PER_FRAME	((VPP_MAX_IMAGE_PAGES / VPP_PAGES_PER_PTBL) + 1) + 4
struct buf_offsets_strides{
	/* This is for storing the offsets in case of using only one buffer for all planes */
	bool is_interleaved;  /*true: Interlaced; false:Progressive */
	u32  y_offset;
	u32  uv_offset;
	u32  y_odd_offset;
	u32  uv_odd_offset;
};
#ifdef __VPP_TARGET_IS_SIMULATOR__
void vpp_io_kick_simulator(struct vpp_dev *idev);
#endif

int vpp_io_write_all_buffer_base_addresses(struct vpp_dev *idev, u32 dev_mem_start_addr, struct buf_offsets_strides *bofs);
int vpp_io_frame_buffers_base_addr_write(struct vpp_dev *idev, enum VPP_BUFFERS buffer, u32 buf_dev_address, enum v4l2_field field); 
void vpp_mmu_write(struct vpp_dev *idev, unsigned int reg_offs, unsigned int data);
unsigned int vpp_mmu_read(struct vpp_dev *idev, unsigned int reg_offs);
int vpp_io_mmu_invalidate_cache(struct vpp_dev *idev);
void vpp_write(struct vpp_dev *idev, unsigned int reg_offs, unsigned int data);
unsigned int vpp_read(struct vpp_dev *idev, unsigned int reg_offs);
void vpp_flush_regs(struct vpp_dev *idev);
int vpp_io_set_timing_registers(struct vpp_dev *idev, struct v4l2_dv_timings *timings);
int vpp_io_set_input_polarity(struct vpp_dev *idev);
int vpp_io_set_input_mode(struct vpp_dev *idev);
int vpp_io_set_input_swap(struct vpp_dev *idev);
int vpp_io_set_input_itu_or_dvi(struct vpp_dev *idev);
int vpp_io_set_output_format(struct vpp_dev *idev);
int vpp_io_enable_interrupt(struct vpp_dev *idev, u32 mask);
int vpp_io_disable_interrupt(struct vpp_dev *idev, u32 mask);
u32 vpp_io_read_irq_status(struct vpp_dev *idev);
u32 vpp_io_read_irq_enable(struct vpp_dev *idev);
void vpp_io_clear_interrupt(struct vpp_dev *idev, u32 mask);
int vpp_io_set_csc_regs(CSC_sConfiguration sResultantConfigurationData,
			struct vpp_dev *idev, bool swap);
int vpp_io_set_burst_size(struct vpp_dev *idev);
int vpp_io_set_h_scaler_coeffs(struct vpp_dev *idev);
int vpp_io_set_v_scaler_coeffs(struct vpp_dev *idev);
int vpp_io_set_hpitch_hinit(struct vpp_dev *idev, u32 hpitch, u32 hinitial);
int vpp_io_set_vinit(struct vpp_dev *idev, u32 vinit0, u32 vinit1);
int vpp_io_set_vpitch_vorder(struct vpp_dev *idev, u32 vpitch, u32 vorder);
int vpp_io_set_hscaled_size(struct vpp_dev *idev, u32 output_width_minus_one);
int vpp_io_set_hscale_bypass (struct vpp_dev *idev, bool hs_bp);
int vpp_io_set_vscale_bypass (struct vpp_dev *idev, bool vs_bp);
int vpp_io_set_vscaled_size(struct vpp_dev *idev, u32 output_height_minus_one);
void vpp_io_set_hs_before_vs(struct vpp_dev *idev, bool hs_before_vs);
bool vpp_io_is_double_buffer_enabled(struct vpp_dev *idev);
void vpp_io_enable_double_buffering(struct vpp_dev *idev);
void vpp_io_disable_double_buffering(struct vpp_dev *idev);
int vpp_io_soft_reset(struct vpp_dev *idev);
int vpp_io_turn_on_hw(struct vpp_dev *idev);
int vpp_io_turn_off_hw(struct vpp_dev *idev);
int vpp_io_enable_memory(struct vpp_dev *idev);
int vpp_io_disable_memory(struct vpp_dev *idev);
u32 vpp_io_read_signatures(struct vpp_dev *idev, u32 *chroma_sig);
void vpp_io_pixel_clock(struct vpp_dev *idev, bool enable);
int vpp_io_mmu_set_tiling(struct vpp_dev *idev);
#endif /*_VPP_IO_H*/
