#include <linux/dma-mapping.h>
#include <linux/vmalloc.h>
#include <asm/page.h>
#include <asm/io.h>
#include "vpp_mmu_regs.h"
#include "vpp_buf.h"
#include "vpp_mmu.h"
#include "vpp_io.h"
#include <media/videobuf2-dma-contig.h>
static int get_offsets(struct v4l2_format *format, struct buf_offsets_strides *bofs);
static unsigned long get_line_len(int width, int pixels_per_block, int bits_per_block, int linelen_align);

static unsigned long get_line_len(int width, int pixels_per_block, int bits_per_block, int linelen_align)
{
	const int linelen_align_in_bits = 8 << linelen_align;
	u_long length;

	/* round up to line length alignment */
	length = (width * bits_per_block) / pixels_per_block;
	length = (length + (linelen_align_in_bits - 1)) & -linelen_align_in_bits;
	length >>= 3;
	return length;
}
static int get_offsets(struct v4l2_format *format, struct buf_offsets_strides *bofs)
{
	u32  pixel_format;
	int width;
	int height;
	unsigned long stride;
	u32 buffer_size;
	u32 req_buf_size;
	/* lines must be at least 256-bit (32-byte) aligned */
	const int align = VPP_YSTRIDE_ALIGN;

	if ((format->type != V4L2_BUF_TYPE_VIDEO_CAPTURE_MPLANE) && (format->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)){
		return -EINVAL;
	}
	pixel_format= (format->type == V4L2_BUF_TYPE_VIDEO_CAPTURE_MPLANE)?
							format->fmt.pix_mp.pixelformat : format->fmt.pix.pixelformat;
	width 	= (format->type == V4L2_BUF_TYPE_VIDEO_CAPTURE_MPLANE)?
							format->fmt.pix_mp.width : format->fmt.pix.width;
	height = (format->type == V4L2_BUF_TYPE_VIDEO_CAPTURE_MPLANE)?
							format->fmt.pix_mp.height : format->fmt.pix.height;
	stride = (format->type == V4L2_BUF_TYPE_VIDEO_CAPTURE_MPLANE) ?
				format->fmt.pix_mp.plane_fmt[VPP_BUF_ALL].bytesperline :
				format->fmt.pix.bytesperline;
	buffer_size = (format->type == V4L2_BUF_TYPE_VIDEO_CAPTURE_MPLANE)?
				format->fmt.pix_mp.plane_fmt[VPP_BUF_ALL].sizeimage : format->fmt.pix.sizeimage;

	switch (pixel_format) {
		case V4L2_PIX_FMT_PL8420:
		case V4L2_PIX_FMT_PL8422:
			if (!stride)
				stride = get_line_len(width, 1, 8, align);
			bofs->y_offset = 0;
			bofs->uv_offset = stride * height;
			if (stride < get_line_len(width, 1, 8, align))
				return -EINVAL;
			break;
		case V4L2_PIX_FMT_PLA420:
		case V4L2_PIX_FMT_PLA422:
			if (!stride)
				stride = get_line_len(width, 3, 32, align);
			bofs->y_offset = 0;
			bofs->uv_offset = stride * height;
			if (stride < get_line_len(width, 3, 32, align))
				return -EINVAL;
			break;
		default:
			return -EINVAL;
	}
	if (bofs->is_interleaved){
		/* TODO: Needs to be corrected according to idev->format.fmt.pix_mp.field. */
		height /= 2;
		bofs->y_odd_offset = bofs->uv_offset + stride * height;
		bofs->uv_odd_offset = bofs->y_odd_offset + stride * height;
		req_buf_size = bofs->uv_odd_offset + stride * height;
	}
	else {
		if (pixel_format == V4L2_PIX_FMT_PL8420 ||
			pixel_format == V4L2_PIX_FMT_PLA420)
			height /= 2;
		req_buf_size = bofs->uv_offset + stride * height;
	}

	/* Check Buffer size is sufficient for both fields */
	if (req_buf_size > buffer_size) {
		printk("\n%s: sizeimage is too small.", __func__);
		return -EINVAL;
	}
	PDEBUG("\n%s: req_buf_size = %d  buffer_size = %d",
		__func__, req_buf_size , buffer_size);
	return 0;
}

void vpp_mmu_remove(struct vpp_dev *idev)
{
	if (idev->mmu_base){
		iounmap(idev->mmu_base);
	}
	return;
}

#if 0
static void print_pagetable_data(u32 *pagetable)
{
	int i;
	u32 *page_address;
	for (i=0;i< VPP_PAGES_PER_PTBL; i++){
		if ((pagetable[i] & 0x1) && ( i%8 == 0) ){
			page_address = (u32 *)((pagetable[i] |0x80000000) & (~1));
			printk("\nprint_pagetable_data: page_address[%d] = 0x%08X, data=0x%08X  0x%08X",
		i, (u32)page_address, (u32)page_address[0], (u32)page_address[VPP_PAGE_SIZE - 1]);
		}
	}
	return;
}
#endif

int vpp_mmu_free(struct vpp_dev *idev, struct idev_buffer *idev_buf)
{
	u32 ii;
	u32 dir_index;

	if (!idev_buf || !idev) {
		pr_warn("%s: inputs are not empty.\n", __func__);
		return -EINVAL;
	}
	BUG_ON(!idev->use_mmu);
	dir_index = idev_buf->dir_index_start;

	//Go through all the buffers in the queue and free/unmap the page tables.

	PDEBUG("vpp_mmu_free is called\n");
	for (ii= 0; ii< MAX_PAGE_TABLES_PER_FRAME; ii++)
	{
		if (idev_buf->page_tables[ii]){
		/*	PDEBUG("vpp_mmu_free: freeing the page tables\n"); */
			/* print_pagetable_data(idev_buf->page_tables[ii]); */
			if (idev_buf->ptbl_bus[ii])
				dma_unmap_single(idev->dev,idev_buf->ptbl_bus[ii] , VPP_PAGE_SIZE, DMA_TO_DEVICE);
			kfree (idev_buf->page_tables[ii]);
			idev_buf->page_tables[ii] = NULL;
			/*Invalidate the directory page entry. But before that make sure that it has been valid (integrity check)*/
			if ((idev->pdir[dir_index + ii] & 0x1) != 0x1){
				printk("\nvpp_mmu_free: invalid directory page entry dir_index=0x%x ", dir_index);
				return -EINVAL;
			}
			else{
				idev->pdir[dir_index + ii] = 0x0;
			}
		}
	}
	PDEBUG("vpp_mmu_free finished\n");
	return 0;
}

int vpp_mmu_setup_no_mmu(struct idev_buffer *idev_buf, struct vpp_dev *idev)
{
	struct vb2_buffer *p_vb;
	u32 physic;
	int frame_size_in_bytes = 0;
	u32 	plane_no,size;
	u32 error;
	p_vb = &(idev_buf->vb);
	if (p_vb == NULL){
		printk ("\vpp_mmu_setup_no_mmu: p_vb is NULL ");
		return -EINVAL;
	}
	PDEBUG("\nmmu_disable\n");
	BUG_ON(idev->use_mmu);
	for ( plane_no =0; plane_no< VPP_MAX_NO_OF_PLANES_PER_FRAME;/*p_vb->num_planes; */plane_no++){
		size = p_vb->v4l2_planes[plane_no].length;
		//TODO: When allocating memory, make sure the size for unused planes is zero.
		if (size == 0)
			continue;
		if(idev->format.fmt.pix_mp.plane_fmt[plane_no].sizeimage !=0 ){
			//Get the physical address of the buffer
			if (USE_DMA_CONTIG) {
				physic = (u32)vb2_dma_contig_plane_dma_addr(p_vb, plane_no);
			} else {
				/* Make sure that we only have one segment */
				if (idev_buf->ndesc[plane_no] == 1) {
					physic = (u32)idev_buf->dma_desc[plane_no][0].dma_addr;
				} else {
					dev_err(idev->dev, "no_mmu: dma segments must be 1");
					WARN_ON(1);
					return -EINVAL;
				}
			}
			frame_size_in_bytes  += size;
		
			if (unlikely(frame_size_in_bytes > VPP_MAX_IMAGE_BYTES)){
				printk("\n vpp_mmu_setup_no_mmu: frame_size_in_bytes is too large 0x%X. returning error",frame_size_in_bytes);
				error = -EINVAL;
				return error;
			}
			
			if (p_vb->num_planes == 1){
				u32 field = idev->format.fmt.pix_mp.field;
				u32 pixelformat= idev->format.fmt.pix_mp.pixelformat;
				struct buf_offsets_strides bofs;
				memset(&bofs, 0, sizeof(bofs));
				/* If interlaced or non_planar, then we have more than one address to deal with.  */
				if ( (field !=V4L2_FIELD_NONE) || ((pixelformat != V4L2_PIX_FMT_YUV30)&&(pixelformat != V4L2_PIX_FMT_UYVY20)) )
				{
				/* V4L2_FIELD_NONE means format is progressive */
					bofs.is_interleaved = (field != V4L2_FIELD_NONE);
					error = get_offsets(&(idev->format), &bofs);
					if (error)
						return error;
					error =  vpp_io_write_all_buffer_base_addresses(idev, physic, &bofs);
					if (error)
						return error;
					}
					else{
						error = vpp_io_frame_buffers_base_addr_write (idev,(enum VPP_BUFFERS) plane_no, physic,idev->format.fmt.pix_mp.field );
					}
			}
			else{
				error = vpp_io_frame_buffers_base_addr_write (idev,(enum VPP_BUFFERS) plane_no, physic,idev->format.fmt.pix_mp.field );
			}
			if (error){
				error = -EFAULT;
				return error;
			}
		}
		else{
			printk("\vpp_mmu_setup_no_mmu:plane_fmt[%d].sizeimage =0",  plane_no);
			error = -EINVAL;
			return error;
		}
		
	}
	return 0;
}
int add_table(struct vpp_dev *idev, struct idev_buffer *idev_buf,
	      int pgtbl_idx, int dir_index)
{
	int error = 0;
	idev_buf->ptbl_bus[pgtbl_idx] =
		dma_map_single(idev->dev,
		idev_buf->page_tables[pgtbl_idx], VPP_PAGE_SIZE,
		DMA_TO_DEVICE);
	if (unlikely(!idev_buf->ptbl_bus[pgtbl_idx])) {
		error = -EFAULT;
		printk ("\n Failed to get the physical address");
		return error;
	}
	idev->pdir[dir_index]= idev_buf->ptbl_bus[pgtbl_idx];
	idev->pdir[dir_index] |= 0x1;
	PDEBUG("\n%s: &idev->pdir[%d]=0x%X idev->pdir[%d]=0x%X ",
		 __func__, dir_index, (u32)&idev->pdir[dir_index],
		 dir_index, idev->pdir[dir_index]);
	idev_buf->page_tables[pgtbl_idx + 1] =
		(u32 *)kzalloc(VPP_PAGE_SIZE, GFP_KERNEL);
	return error;
}

int vpp_mmu_setup(struct idev_buffer *idev_buf, struct vpp_dev *idev)
{
	struct vb2_buffer *p_vb;
	int frame_size_in_bytes = 0;
	int plane_no, j, ii, error;
	int page_count=0;
	u32 addr_to_map;
	u32 len_to_map;
	bool dmabuf = false;
	u32	dir_index;
	int	pgtbl_idx = 0;

	if (!idev_buf || !idev) {
		printk("\nvpp_mmu_setup: error inputs are empty");
		return -EINVAL;
	}
	PDEBUG("\n---mmu_enable---\n");
	BUG_ON(!idev->use_mmu);
	/*
	 * idev->bus_index is either 0 or 0x80000000. Meaning that the device
	 * memory address for this buffer starts at either 0x0 or 0x80000000,
	 * and the dir_index would be either 0 or 512 (mmu_page_size is 4kytes)
	 *
	 */
	dir_index = idev->bus_index >> (LOG2_MMU_PAGESIZE * 2 - 2);
	idev_buf->dir_index_start = dir_index;
	p_vb = &(idev_buf->vb);
	if (!p_vb) {
		printk ("\vpp_mmu_setup: p_vb is NULL ");
		return -EINVAL;
	}
	for (ii=0; ii< MAX_PAGE_TABLES_PER_FRAME; ii++){
		idev_buf->page_tables[ii] = NULL;
		idev_buf->ptbl_bus[ii] = 0x0;
	}

	PDEBUG ("\n vpp_mmu_setup: buffer's memory type is %s  %d", (p_vb->v4l2_buf.memory == V4L2_MEMORY_MMAP)? "V4L2_MEMORY_MMAP":"V4L2_MEMORY_USERPTR", (int)p_vb->v4l2_buf.memory);
	
	/*****************
	 * For each plane;
	 ****************/
	
	for ( plane_no =0; plane_no< VPP_MAX_NO_OF_PLANES_PER_FRAME;/*p_vb->num_planes; */plane_no++) {
		
		u32 	plane_dev_mem_start_addr = dir_index << ( (LOG2_MMU_PAGESIZE * 2) -2);
		u32 	addr;
		u32 	cpu_v_addr;
		u32 	size = p_vb->v4l2_planes[plane_no].length;
		//TODO: When allocating memory, make sure the size for unused planes is zero.
		if (size == 0)
			continue;
		frame_size_in_bytes  += size;
		
		if (unlikely(frame_size_in_bytes > VPP_MAX_IMAGE_BYTES)){
			printk("\n vpp_mmu_setup: frame_size_in_bytes is too large 0x%X. returning error",frame_size_in_bytes);
			error = -EINVAL;
			goto virt_err;
		}
		if (idev_buf->ndesc[plane_no] != 0)
				dmabuf = true;
		if (!dmabuf) {
			/* Get the VA of the plane buffer from cpu (not the device) point of view */
			cpu_v_addr= (u32)vb2_plane_vaddr(p_vb,plane_no);
			if (!is_vmalloc_addr( (u32 *)cpu_v_addr) ){
				for( ii =0; ii< VPP_MAX_NO_OF_PLANES_PER_FRAME;/*p_vb->num_planes; */ii++)
					printk ("\vpp_mmu_setup: buffer has not been created by vmalloc. size = %d , plane_no=%d", size, ii);
				return -EFAULT;
			}
			/* cpu_v_addr addresses must be correctly aligned */
			if (cpu_v_addr & VPP_YADDR_ALIGNMASK){
					printk("\n vpp_mmu_setup :cpu_v_addr is not aligned correctly 0x%X . returning error",cpu_v_addr);
					error = -EFAULT;
					goto virt_err;
			}
			//check that the cpu_v_addr is page-alligned.
			if (cpu_v_addr & ~VPP_PAGE_MASK) {
			/*
			* If cpu_v_addr address is not on a page boundary, start mapping from
			* the start of the page
			*/
			addr_to_map = cpu_v_addr - (cpu_v_addr & ~VPP_PAGE_MASK);
			len_to_map = size + (cpu_v_addr & ~VPP_PAGE_MASK);
			/* Change the plane start address accordingly */
			plane_dev_mem_start_addr |= cpu_v_addr & ~VPP_PAGE_MASK;
			}
			else {
				addr_to_map = cpu_v_addr;
				len_to_map = size;
			}
		}
		if(idev->format.fmt.pix_mp.plane_fmt[plane_no].sizeimage !=0 ){
			//TODO: Check the number of planes. If it is 1, then make sure that the field is progressive and the format is not planar, otherwise we need to calculate the
			//start addresses based on the field and format. We need to call a different function here.
			if (p_vb->num_planes == 1){
					u32 field = idev->format.fmt.pix_mp.field;
					u32 pixelformat= idev->format.fmt.pix_mp.pixelformat;
					struct buf_offsets_strides bofs;
					memset(&bofs, 0, sizeof(bofs));
					/* If interlaced or non_planar, then we have more than one address to deal with.  */
					if ( (field !=V4L2_FIELD_NONE) || ((pixelformat != V4L2_PIX_FMT_YUV30)&&(pixelformat != V4L2_PIX_FMT_UYVY20)) )
					{
						/* V4L2_FIELD_NONE means format is progressive */
						bofs.is_interleaved = (field != V4L2_FIELD_NONE);
						error = get_offsets(&(idev->format), &bofs);
						if (error)
							goto virt_err;
						error =  vpp_io_write_all_buffer_base_addresses(idev, plane_dev_mem_start_addr, &bofs);
						if (error)
							goto virt_err;
					}
					else{
						error = vpp_io_frame_buffers_base_addr_write (idev,(enum VPP_BUFFERS) plane_no, plane_dev_mem_start_addr,idev->format.fmt.pix_mp.field );
					}
			}
			else{
				error = vpp_io_frame_buffers_base_addr_write (idev,(enum VPP_BUFFERS) plane_no, plane_dev_mem_start_addr,idev->format.fmt.pix_mp.field );
			}
			if (error){
				error = -EFAULT;
				goto virt_err;
			}
		}
		else{
			printk("\nvpp_mmu_setup:plane_fmt[%d].sizeimage =0",  plane_no);
			error = -EINVAL;
			goto virt_err;
		}
	
			/* Fill the page table and allocate new page tables if needed and fill them */
		page_count = 0;
		idev_buf->page_tables[pgtbl_idx] = (u32 *)kzalloc(VPP_PAGE_SIZE, GFP_KERNEL);
		if (!dmabuf) {
			for (addr = addr_to_map; addr < addr_to_map + size; addr += PAGE_SIZE) {
				struct page *pg = vmalloc_to_page((void *)addr);
				u32 page_physic; 
				if (page_count == VPP_PAGES_PER_PTBL){
					if (add_table(idev, idev_buf, pgtbl_idx, dir_index))
						goto virt_err;
					page_count = 0;
					pgtbl_idx++;
					dir_index ++;
				}
				/*Get the physical address of the page and put it in the page table*/
				page_physic =  page_to_phys(pg);
				for (j=0; j< PAGE_SIZE / VPP_PAGE_SIZE; j++){
					idev_buf->page_tables[pgtbl_idx][page_count] = (u32 )(page_physic + (j * VPP_PAGE_SIZE)) ;
					idev_buf->page_tables[pgtbl_idx][page_count] |= 0x1;
					page_count ++;
				}
			}
		}
		else {
			PDEBUG("\n %s: ndesc= %d  ", __func__, idev_buf->ndesc[plane_no]);
			for (ii = 0; ii < idev_buf->ndesc[plane_no]; ii++){
				for (addr = idev_buf->dma_desc[plane_no][ii].dma_addr;
					addr < idev_buf->dma_desc[plane_no][ii].dma_addr + idev_buf->dma_desc[plane_no][ii].seg_len;
					addr += VPP_PAGE_SIZE) {
					idev_buf->page_tables[pgtbl_idx][page_count] = (u32) addr;
					idev_buf->page_tables[pgtbl_idx][page_count] |= 0x1;
					page_count ++;
					if (page_count == VPP_PAGES_PER_PTBL){
						if (add_table(idev, idev_buf, pgtbl_idx, dir_index))
							goto virt_err;
						page_count = 0;
						pgtbl_idx++;
						dir_index ++;
					}
				}
			}
		}

			/*Finished one plane. finalise the current page table */
		PDEBUG("\n vpp_mmu_setup: mark the rest of the pages as unused. TableEntry= %d pgtbl_idx=%d  ", page_count,pgtbl_idx+1);
		for (; page_count < VPP_PAGES_PER_PTBL; page_count++){
			idev_buf->page_tables[pgtbl_idx][page_count] = 0x0;
		}
		idev_buf->ptbl_bus[pgtbl_idx] = dma_map_single(idev->dev, idev_buf->page_tables[pgtbl_idx], VPP_PAGE_SIZE,
									DMA_TO_DEVICE);
		if (unlikely(!idev_buf->ptbl_bus[pgtbl_idx])) {
			error = -EFAULT;
			printk ("\n Failed to get the physical address");
			goto virt_err;
		}
		idev->pdir[dir_index]= idev_buf->ptbl_bus[pgtbl_idx];
		idev->pdir[dir_index] |= 0x1;
		page_count = 0;
		PDEBUG ("\n vpp_mmu_setup:	 Start of the page tables directory2 : &idev->pdir[%d]=0x%X idev->pdir[%d]=0x%X ",dir_index,(u32)&idev->pdir[dir_index],dir_index,idev->pdir[dir_index]);
		pgtbl_idx++;
		dir_index ++;
	}
	/*Finished all planes. */
	idev->bus_index ^=0x80000000;  //0 <->0x80000000
	PDEBUG ("\n vpp_mmu_setup: finished successfully");
	
	return 0;
virt_err:
	/*plane_no indicates the last plane we were working on when the error happened.*/
	for (ii= 0; ii<= pgtbl_idx; ii++)
	{
		if (idev_buf->page_tables[ii]){
			printk("\nvpp_mmu_setup: ERROR freeing the page tables");
			if (idev_buf->ptbl_bus[ii])
				dma_unmap_single(idev->dev,idev_buf->ptbl_bus[ii] , VPP_PAGE_SIZE, DMA_TO_DEVICE);
			kfree (idev_buf->page_tables[ii]);
			idev->pdir[dir_index - ii] = 0x0;
		}
	}
	return error;
}
void vpp_mmu_enable(struct vpp_dev *idev)
{
	u32 reg;
	reg = vpp_mmu_read(idev, VPP_MMU_ADDRESS_CONTROL_OFFSET);
	SET_FIELD(reg, VPP_MMU_ADDRESS_CONTROL_MMU_BYPASS, 0);
	vpp_mmu_write(idev, VPP_MMU_ADDRESS_CONTROL_OFFSET, reg);
}
void vpp_mmu_disable(struct vpp_dev *idev)
{
	u32 reg;
	reg = vpp_mmu_read(idev, VPP_MMU_ADDRESS_CONTROL_OFFSET);
	SET_FIELD(reg, VPP_MMU_ADDRESS_CONTROL_MMU_BYPASS, 1);
	vpp_mmu_write(idev, VPP_MMU_ADDRESS_CONTROL_OFFSET, reg);
}
int vpp_mmu_probe( struct vpp_dev *idev)
{
		/* idev->bus_index will be either 0x0 or 0x80000000
			This is the start of buffers (current and next) in terms of device address.
		*/
		idev->bus_index = 0x0;
#if (PAGE_SIZE % VPP_PAGE_SIZE != 0)
#error PAGE_SIZE must be a multiple of VPP_PAGE_SIZE
#endif

		idev->pdir = dma_zalloc_coherent(idev->dev, VPP_PAGE_DIR_BYTES,
										&idev->pdir_bus, GFP_KERNEL);

		if (!idev->pdir) {
			dev_err(idev->dev, "cannot allocate mmu page directory\n");
			printk ("\nvpp_mmu_probe:cannot allocate mmu page directory\n");
			return -ENOMEM;
			
		}
		else{
		PDEBUG ("\n vpp_mmu_probe: mmu page directory allocated. virtual= 0x%X physical= 0x%X\n",(u32)idev->pdir,(u32) idev->pdir_bus );
		}
		if (idev->pdir_bus & ~VPP_PAGE_MASK) {
			dev_err(idev->dev,
					"mmu page directory bus address is not page aligned\n");
			return -EFAULT;
		}

		/* Set MMU directory base address */
		vpp_mmu_write(idev, VPP_MMU_DIR_BASE_ADDR_OFFSET, idev->pdir_bus);

	return 0;

}






