#include "vpp_io.h"
#include "vpp_scale.h"
#include "linux/math64.h"
#include <video/img_vpp.h>

static void calc_scale_coeffs(u32 res, u32 size, int T, int I, int *coeffs, int count);
static int scale_coeffs_stale(u32 input_width, u32 old_input_width,
										u32 output_width, u32 old_output_width);
/* Scaler Code */

static inline int fix(int p, int a)
{
	return a << p;
}

static inline int fix_div(int p, int n, int d)
{
	return (n << p) / d;
}

static inline int fix_divl(int p, int n, int d)
{
	//return (int)(((u64)n << p) / d);
	return (int)( div64_u64(((u64)n << p) , (u64) d) );
}

static inline int fix_mul(int p, int a, int b)
{
	return (a * b) >> p;
}

static inline int fix_mull(int p, int a, int b)
{
	return (int)(((u64)a * (u64)b) >> p);
}

static inline int fix_sin(int p, int x)
{
	/* tailor expansion */
	int pow;
	int sum;

	/* in [-pi,pi] */
	int pi2 = fix_divl(p, 6283185, 1000000);
	int pi = pi2 / 2;
	x = x % pi2;
	if (x > pi)
		x -= pi2;
	else if (x < -pi)
		x += pi2;

	sum = x;

	/* x represents x^2 from now on */
	x = fix_mull(p, x, x);

	/* -x^3/3! */
	pow = fix_mull(p, sum, x);
	sum -= pow / 6;
	/* +x^5/5! */
	pow = fix_mull(p, pow, x);
	sum += pow / 120;
	/* -x^7/7! */
	pow = fix_mull(p, pow, x);
	sum -= pow / 5040;
	/* +x^9/9! */
	pow = fix_mull(p, pow, x);
	sum += pow / 362880;

	return sum;
}

static inline int fix_cos(int p, int x)
{
	/* tailor expansion */
	int pow;
	int sum = fix(p, 1);

	/* in [-pi,pi] */
	int pi2 = fix_divl(p, 6283185, 1000000);
	int pi = pi2 / 2;
	x = x % pi2;
	if (x > pi)
		x -= pi2;
	else if (x < -pi)
		x += pi2;

	/* x represents x^2 from now on */
	x = fix_mull(p, x, x);

	/* -x^2/2! */
	pow = x;
	sum -= pow / 2;
	/* +x^4/4! */
	pow = fix_mull(p, pow, x);
	sum += pow / 24;
	/* -x^6/6! */
	pow = fix_mull(p, pow, x);
	sum -= pow / 720;
	/* +x^8/8! */
	pow = fix_mull(p, pow, x);
	sum += pow / 40320;
	/* -x^10/10! */
	pow = fix_mull(p, pow, x);
	sum -= pow / 3628800;

	return sum;
}

static int scale_coeffs_stale(u32 input_width, u32 old_input_width,
										u32 output_width, u32 old_output_width)
{
	/* scale factor clamped to max of 1, meaning that scaler coeffs will be the same for all scale factors of >= 1. */
	if (output_width >= input_width && old_output_width >= old_input_width)
		return 0;
	/* has scale factor changed? */
	return output_width * old_input_width != old_output_width * output_width;
}

static void calc_scale_coeffs(u32 res, //input width or height
										u32 size,//output width or height
										int T,	/* taps */
										int I,	/* interpolation points */
										int *coeffs, /*array for the output*/
										int count	/*size of the array*/
										)
{
	int midpoint = T * I / 2;
#define SCALE_PRECISION 10
#define RESULT_PRECISION 6
#define DIV(a, b)	fix_div(SCALE_PRECISION, a, b)
#define DIVL(a, b)	fix_divl(SCALE_PRECISION, a, b)
#define MUL(a, b)	fix_mul(SCALE_PRECISION, a, b)
#define SIN(a)		fix_sin(SCALE_PRECISION, a)
#define COS(a)		fix_cos(SCALE_PRECISION, a)
#define FIX(a)		fix(SCALE_PRECISION, a)
#define RESULT(a)	((a) >> (SCALE_PRECISION - RESULT_PRECISION))
#define CHECK_SCALE_COEFFS
	int fpi = DIVL(3141593, 1000000);
	int fS = min(FIX(1), /* scale factor clamped to 1 */
		DIV(size, res));
	int fA = DIV(54, 100);
	int c, i;

	int fpiS = MUL(fpi, fS);
	int f2piSoT = fpiS * 2 / T;
	int f1mA = FIX(1) - fA;
	int fT = FIX(T);
	int ftotal = 0;

	/* Calculate coefficients */
	for (c = 0; c <= midpoint; ++c) {
		/* t + i/I */
		int ftpioI = FIX(c) / I;
		int fx = MUL(fpiS, fT / 2 - ftpioI);
		int fco = fA - MUL(f1mA, COS(MUL(f2piSoT, ftpioI)));
		if (fx > DIV(1, 10))
			fco = MUL(DIV(SIN(fx), fx), fco);
		ftotal += fco;
		coeffs[c] = fco;
	}

	/*
	 * Interpolation points I/2+1 to I-1 have same values as 1 to I/2-1,
	 * but in opposite order, so we only need to normalise the first half
	 * of the interpolation points taking mirroring into account.
	 */
	for (i = 0; i <= I / 2; ++i) {
		int sum, err, dir = 0;

		/*
		 * Normalise coefficients in each interpolation point and
		 * convert into result format.
		 */
		sum = 0;
		for (c = i; c <= midpoint; c += I)
			sum += coeffs[c];
		for (; c < midpoint * 2; c += I)
			sum += coeffs[midpoint * 2 - c];
		/* Careful not to modify a coefficient twice */
		for (c = i; c <= midpoint; c += I)
			coeffs[c] = RESULT(DIV(coeffs[c], sum));
		if (i & (I / 2 - 1))
			for (c = I - i; c <= midpoint; c += I)
				coeffs[c] = RESULT(DIV(coeffs[c], sum));

		/*
		 * Find fixed point error from normalisation.
		 */
		sum = 0;
		for (c = i; c <= midpoint; c += I)
			sum += coeffs[c];
		for (; c < midpoint * 2; c += I)
			sum += coeffs[midpoint * 2 - c];
		err = sum - (1 << RESULT_PRECISION);
		if (err > 0)
			dir = -1;
		else if (err < 0)
			dir = 1;
		else
			continue;

#ifdef DEBUG_SCALE_COEFFS
		printk(KERN_DEBUG "pdp: i=%d, err=%d\n", i, err);
#define SCALE_DEBUG(C, N) \
		printk(KERN_DEBUG "pdp:   adj*%d [%d] to 0x%x (err=%d)\n", \
				(N), (C), coeffs[(C)], err)
#else
#define SCALE_DEBUG(C, N) do {} while (0)
#endif

		/*
		 * Distribute the error over the tap coefficients, preferring
		 * to change central coefficients.
		 */
		if (i == 0) {
			/*
			 * Special case: contains midpoint. Work out from
			 * center until no error or an odd value (which can be
			 * fixed by adjusting midpoint again).
			 */
			while (err > 1 || err < -1) {
				coeffs[midpoint] += dir;
				err += dir;
				SCALE_DEBUG(midpoint, 1);
				for (c = midpoint - I;
					 (err > 1 || err < -1) && c >= I;
					 c -= I) {
					coeffs[c] += dir;
					err += dir * 2;
					SCALE_DEBUG(c, 2);
				}
			}
			if (err & 1) {
				coeffs[midpoint] += dir;
				err += dir;
				SCALE_DEBUG(midpoint, 1);
			}
		} else if (i == I / 2) {
			/*
			 * Special case: tap coeffs mirror. This also means err
			 * is always even with this interpolation point.
			 */
			int last = i + I * (T / 2 - 1);
#ifdef CHECK_SCALE_COEFFS
			WARN(err & 1, "err (=%d) should be even for i=%d\n",
					err, i);
#endif
			while (err) {
				for (c = last; err && c >= 0; c -= I) {
					coeffs[c] += dir;
					err += dir * 2;
					SCALE_DEBUG(c, 2);
				}
			}
		} else {
			/*
			 * No tap coefficient mirrors another. Work out from
			 * center adjusting values.
			 */
			int last = midpoint - i;
			int offset = i * 2 - I;
			while (err) {
				for (c = last; err && c >= 0; c -= I) {
					coeffs[c] += dir;
					err += dir;
					SCALE_DEBUG(c, 1);
					if (!err)
						break;
					coeffs[c+offset] += dir;
					err += dir;
					SCALE_DEBUG(c+offset, 1);
				}
			}
		}
#ifdef CHECK_SCALE_COEFFS
		sum = 0;
		for (c = i; c <= midpoint; c += I)
			sum += coeffs[c];
		for (; c < midpoint * 2; c += I)
			sum += coeffs[midpoint * 2 - c];
		err = sum - (1 << RESULT_PRECISION);
		WARN_ONCE(err, "Scale coefficients not normalised"
				   " (i=%d, err=%d)\n", i, err);
#endif
	}

	/* reflect about midpoint in spare coeffs */
	if (count > midpoint * 2)
		count = midpoint * 2;
	for (c = midpoint + 1; c < count; ++c)
		coeffs[c] = coeffs[2 * midpoint - c];
#undef SCALE_PRECISION
#undef RESULT_PRECISION
#undef DIV
#undef DIVL
#undef MUL
#undef SIN
#undef COS
#undef FIX
#undef RESULT
#undef CHECK_SCALE_COEFFS
}
static int configure_vscale(struct vpp_dev *idev)
{
	u32 vpitch;
	u32 vinit0;
	u32 vinit1;
	u32 input_height = idev->timings.bt.height;//stream->info.var.xres;
	u32 output_height = idev->format.fmt.pix_mp.height;
	idev->v.vs_en = (output_height != input_height); // Output height is supposed to be smaller than input_height.
	vpp_io_set_vscale_bypass (idev, !(idev->v.vs_en));
	if (!idev->v.vs_en)
			return 0;
	if ( ((MAX_V_SCALE_FACTOR * input_height) < output_height ) ||
		((MAX_V_SCALE_FACTOR * output_height) < input_height ) ){
		/*scale factor out of range */
		printk ("\n configure_vscale: scale factor out of range");
		return -EINVAL;
	}
	
	vpp_io_set_vscaled_size(idev, output_height -1 ); 
	
	vpitch = (fix_div(PITCH_FRACTION +1 , input_height, output_height) + 1) >> 1;
	printk ("\nconfigure_vscale: vpitch = 0x%X", (u32) vpitch);
	vpp_io_set_vpitch_vorder (idev, vpitch, 0x3);
	
	vinit0= vinit1= (VPP_V_TAPS >> 1) << VINIT_FRACTION;
	vpp_io_set_vinit(idev, vinit0, vinit1);
	
	calc_scale_coeffs(	input_height,
							output_height,
							VPP_V_TAPS,
							VPP_V_INTERPOLATION_POINTS,
							idev->v.vscoeffs,
							ARRAY_SIZE(idev->v.vscoeffs));
	
	vpp_io_set_v_scaler_coeffs( idev); 
	return 0;
	
//	hinitial = fix_div(11, stream->caps.hs_taps, 2);
}
static int configure_hscale(struct vpp_dev *idev)
{
	u32 hpitch, hinitial;
	u32 input_width = idev->timings.bt.width;//stream->info.var.xres;
	u32 output_width = idev->format.fmt.pix_mp.width;
	idev->v.hs_en = (output_width != input_width); // Output width is supposed to be smaller than input_width.
	vpp_io_set_hscale_bypass (idev, !(idev->v.hs_en));
	if (!idev->v.hs_en)
			return 0;
	vpp_io_set_hscaled_size(idev, output_width -1); 
	if ( ((MAX_H_SCALE_FACTOR * input_width) < output_width ) ||
		((MAX_H_SCALE_FACTOR * output_width) < input_width ) ){
		/*scale factor out of range */
		printk ("\n configure_hscale: scale factor out of range");
		return -EINVAL;
	}
	hpitch = (fix_div(PITCH_FRACTION +1 , input_width, output_width) + 1) >> 1;
	
	hinitial = (VPP_H_TAPS >> 1) << HINIT_FRACTION;
	
	PDEBUG ("\nconfigure_hscale: hpitch = 0x%X", (u32) hpitch);
	
	vpp_io_set_hpitch_hinit(idev, hpitch, hinitial);
	/* Let's do horizontal scaling first */
	vpp_io_set_hs_before_vs(idev, true);
	/* do not calculate coefficients again if not needed */
	if ((idev->v.hs_old_input_size == 0) ||
				scale_coeffs_stale(input_width,
										idev->v.hs_old_input_size,
										output_width,
										idev->v.hs_old_output_size)){
		calc_scale_coeffs(	input_width,
							output_width,
							VPP_H_TAPS,
							VPP_H_INTERPOLATION_POINTS,
							idev->v.hscoeffs,
							ARRAY_SIZE(idev->v.hscoeffs));
	}
	idev->v.hs_old_input_size = input_width;
	idev->v.hs_old_output_size = output_width;
	vpp_io_set_h_scaler_coeffs( idev); 
	return 0;

}

#if 0
static int pdpfb_str_configure_hscale(struct pdpfb_priv *priv,
									struct pdpfb_stream *stream)
{
	u32 reg;
	u32 input_width = stream->info.var.xres;
	u32 output_width = stream->geom.w ? stream->geom.w : stream->info.var.xres;
	u32 i;
	u32 scale_width;
	u32 hpitch, hinitial;

	stream->hs_en = (output_width != input_width);
		reg = pdpfb_read(priv, stream->regs.scale_ctrl);
		SET_FIELD(reg, PDP_VID1SCALECTRL_VID1HSCALEBP, !stream->hs_en);
		pdpfb_write(priv, stream->regs.scale_ctrl, reg);

		scale_width = (stream->hs_en && stream->clip_geom.w) ?
						stream->clip_geom.w + stream->clip_geom.x : output_width;
		reg = pdpfb_read(priv, stream->regs.scale_size);
		SET_FIELD(reg, PDP_VID1SCALESIZE_VID1SCALEWIDTH, scale_width - 1);
		pdpfb_write(priv, stream->regs.scale_size, reg);
		if (!stream->hs_en)
			return 0;
	
		hpitch = (fix_div(12, input_width, output_width) + 1) >> 1;

		reg = pdpfb_read(priv, stream->regs.scale_ctrl);
		SET_FIELD(reg, PDP_VID1SCALECTRL_VID1HORDER, stream->bilinear_en ? 2 : 7);
		pdpfb_write(priv, stream->regs.scale_ctrl, reg);

		hpitch = (fix_div(12, input_width, output_width) + 1) >> 1;
		if (stream->hdecimation)
			hpitch /= 2;
		hinitial = fix_div(11, stream->caps.hs_taps, 2);

		/*
		   For the right image of a split scaling operation, add the fractional
		   part of (hpitch * scale_width) and the number of taps to hinitial.
		*/
		if (stream->split_scale == PDP_SPLIT_SCALE_RIGHT_PLANE)
			hinitial += ((hpitch * scale_width) & 0x3FF) + (stream->caps.hs_taps << 11);

		reg = pdpfb_read(priv, stream->regs.hsinit);
		SET_FIELD(reg, PDP_VID1HSINIT_VID1HINITIAL, hinitial);
		SET_FIELD(reg, PDP_VID1HSINIT_VID1HPITCH, hpitch);
		pdpfb_write(priv, stream->regs.hsinit, reg);

		if (stream->bilinear_en) {
			if (stream->caps.hs_interpolation_points != 16) {
				dev_err(priv->device, "unsupported number of interpolation points");
				return -EINVAL;
			}

			/* Number of coefficients = 2 * I */
			/* Number of coeff registers to write = I / 2 */
			for (i = 0; i < stream->caps.hs_interpolation_points / 2; ++i) {
				
				reg =  (0xFF & (bf_coeffs_i16[i << 2]>>0))
					| ((0xFF & (bf_coeffs_i16[(i << 2) + 1]>>0)) << 8)
					| ((0xFF & (bf_coeffs_i16[(i << 2) + 2])>>0) << 16)
					| ((0xFF & (bf_coeffs_i16[(i << 2) + 3])>>0) << 24);
				pdpfb_write(priv, stream->regs.first_hscoeff + (i << 2), reg);
			}
		}
		else {
			/*
			   Don't calculate coefficients unless they haven't been
			   calculated before, or something has changed.
			*/
			if (!stream->hs_oldres ||
				pdpfb_str_scale_coeffs_stale(stream->info.var.xres,
										stream->hs_oldres,
										stream->geom.w,
										stream->hs_oldsize)) {
				stream->hs_oldres = stream->info.var.xres;
				stream->hs_oldsize = stream->geom.w;
				pdpfb_str_calc_scale_coeffs(stream->info.var.xres,
											stream->geom.w,
											stream->caps.hs_taps,
											stream->caps.hs_interpolation_points,
											stream->hscoeffs,
											ARRAY_SIZE(stream->hscoeffs));
			}

			/* Number of coefficients = (T/2)*I + 1 */
			/* Number of coeff registers to write = ((T/2)*I)/4 + 1 */
			/* T = 8, so registers to write = I + 1 */
			for (i = 0; i < stream->caps.hs_interpolation_points; ++i) {
				reg =  (0xFF & stream->hscoeffs[i << 2])
					| ((0xFF & stream->hscoeffs[(i << 2) + 1]) << 8)
					| ((0xFF & stream->hscoeffs[(i << 2) + 2]) << 16)
					| ((0xFF & stream->hscoeffs[(i << 2) + 3]) << 24);
				pdpfb_write(priv, stream->regs.first_hscoeff + (i << 2), reg);
			}
			pdpfb_write(priv, stream->regs.last_hscoeff,
						0xFF & stream->hscoeffs[i << 2]);
		}
	
	return 0;
}

#endif




int vpp_set_scalers(struct vpp_dev *idev)
{
	int ret = 0;
	if (idev == NULL)
		return -EINVAL;
	ret = configure_hscale(idev);
	if (ret){
		return ret;
	}
	ret = configure_vscale(idev);
	
	
	return ret;
}






