/**
 * Imagination Technologies Ring Oscillator driver
 *
 * Copyright (C) 2017 Imagination Technologies Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/errno.h>
#include <linux/init.h>
#include <linux/kobject.h>
#include <linux/module.h>
#include <linux/of_device.h>
#include <linux/pm_runtime.h>
#include <linux/sysfs.h>
#include <linux/rwsem.h>
#include <linux/atomic.h>
#include <linux/delay.h>

/* Macros */
#define RING0_CTRL_CFG          0x00    /* Ring0 control reg */
#define RING1_CTRL_CFG          0x0c    /* Ring1 control reg */
#define RING0_FREQ_COUNT_CFG    0x08    /* Freq Count offset for Ring0 */
#define RING1_FREQ_COUNT_CFG    0x14
#define RING0_STATUS_CFG	0x04    /* Freq Count offset for Ring0 */
#define RING1_STATUS_CFG	0x10
#define NUM_RING_OSC            0x02

/* Bit positions */
#define OSC_EN			BIT(0)
#define FREQ_START		BIT(1)   /* Start bit position */
#define OSC_FREQ_DONE		BIT(0)

#define FREQ_PERIOD_SHIFT               0x02
#define DEFAULT_FREQ_PERIOD             (0xff << FREQ_PERIOD_SHIFT)
#define RING_OSC_FREQ_DONE		BIT(0)

/* helper macros */
#define IS_VALID(r)		(((r) == 0) || ((r) == 1))
#define GET_CTRL_REG(r)		((r) ? RING1_CTRL_CFG : RING0_CTRL_CFG)
#define GET_FREQ_COUNT_CFG(r)	((r) ? RING1_FREQ_COUNT_CFG : \
					RING0_FREQ_COUNT_CFG)
#define GET_STATUS_CFG(r)		((r) ? RING1_STATUS_CFG : \
						RING0_STATUS_CFG)

/* Device structure */
struct ring_osc_driver_data {
	struct platform_device *pdev;
	struct device *dev;
	void __iomem *base;
	struct kobject **ring_kobj;
	struct rw_semaphore rosc_lock;
	unsigned long flags;
};

static const struct of_device_id img_ring_osc_of_match[] = {
	{ .compatible = "img,img-ring", },
	{ }
};
MODULE_DEVICE_TABLE(of, img_ring_osc_of_match);

#define RING_ATTR(_name, _mode, _show, _store) \
	struct kobj_attribute dev_attr_ ## _name = { \
		.attr	= { .name = __stringify(_name), .mode = _mode }, \
		.show	= _show, \
		.store	= _store, \
	}

/* Function Prototypes */
static u32 img_ring_osc_readl(struct ring_osc_driver_data *, u32);
static void img_ring_osc_writel(struct ring_osc_driver_data *, u32, u32);
static ssize_t img_ring_osc_period_write(struct kobject *,
					 struct kobj_attribute *,
					 const char *, size_t);
static ssize_t img_ring_osc_freq_count_read(struct kobject *,
					    struct kobj_attribute *, char *);
static ssize_t img_ring_osc_ring_module_export(struct device *,
					       struct device_attribute *,
					       const char *, size_t);
static void img_ring_osc_config(struct ring_osc_driver_data *);
static int img_ring_osc_probe(struct platform_device *);
static int img_ring_osc_remove(struct platform_device *);
static void img_ring_osc_reset(struct ring_osc_driver_data *, u32);
static u32 osc_status_read_helper(struct ring_osc_driver_data *, u32);
static void img_ring_osc_reset(struct ring_osc_driver_data *priv, u32 ring_num)
{
	u32 offset, val;

	offset  = GET_CTRL_REG(ring_num);

	down_write(&priv->rosc_lock);

	val = img_ring_osc_readl(priv, offset);
	val &= ~FREQ_START; /* clear the freq start bit */
	img_ring_osc_writel(priv, val, offset);

	up_write(&priv->rosc_lock);
}

/* Function definitions */
static void img_ring_osc_start_new_count(struct ring_osc_driver_data *priv,
					 u32 ring_num)
{
	u32 offset, val;

	offset  = GET_CTRL_REG(ring_num);

	down_write(&priv->rosc_lock);

	val = img_ring_osc_readl(priv, offset);
	val |= FREQ_START; /* clear the freq start bit */
	img_ring_osc_writel(priv, val, offset);

	up_write(&priv->rosc_lock);
}

static u32 osc_status_read_helper(struct ring_osc_driver_data *priv,
				  u32 ring_num)
{
	u32 val, offset;

	offset  = GET_STATUS_CFG(ring_num);

	down_read(&priv->rosc_lock);

	val = img_ring_osc_readl(priv, offset);
	val &= OSC_FREQ_DONE;

	up_read(&priv->rosc_lock);

	pr_info(" FREQ_DONE val = %d\n", val);

	return val;
}

static u32 img_ring_osc_count_read(struct ring_osc_driver_data *priv,
				   u32 ring_num)
{
	u32 offset, val;

	if (!IS_VALID(ring_num))
		return -EINVAL;

	while (!osc_status_read_helper(priv, ring_num)) {
		pr_info("### Waiting ###\n");
		mdelay(1000);
	}

	offset  = GET_FREQ_COUNT_CFG(ring_num);

	down_read(&priv->rosc_lock);

	val = img_ring_osc_readl(priv, offset);

	up_read(&priv->rosc_lock);

	pr_info(" count read val = %d\n", val);

	return val;
}

static void img_ring_osc_change_freq_period(struct ring_osc_driver_data *priv,
					    u32 period, u32 ring_num)
{
	u32 val, offset;

	pr_info(" new period val = %d\n", period);
	offset = GET_CTRL_REG(ring_num);

	down_write(&priv->rosc_lock);

	val = img_ring_osc_readl(priv, offset);
	val |= period << FREQ_PERIOD_SHIFT;
	img_ring_osc_writel(priv, val, offset);

	val = img_ring_osc_readl(priv, offset);
	val |= FREQ_START;
	img_ring_osc_writel(priv, val, offset);

	up_write(&priv->rosc_lock);
}

static ssize_t img_ring_osc_period_write(struct kobject *kobj,
					 struct kobj_attribute *attr,
					 const char *buf, size_t size)
{
	int ret;
	u32 ring_num, period;
	unsigned char kobj_name[2];
	struct platform_device *pdev;
	struct ring_osc_driver_data *priv;

	pdev = to_platform_device(kobj_to_dev(kobj->parent));
	priv = platform_get_drvdata(pdev);

	kobj_name[0] = *(kobj->name + 4);
	kobj_name[1] = '\0';
	ret = kstrtou32(kobj_name, 10, &ring_num);
	if (ret) {
		dev_err(&pdev->dev, "could not parse ring number string\n");
		return -EINVAL;
	}

	ret = kstrtouint(buf, 10, &period);
	if (ret) {
		dev_err(&pdev->dev,
			"could not parse period attr value\n");
		return -EINVAL;
	}

	if (!IS_VALID(ring_num))
		return -EINVAL;

	/*
	 * CR_RAND_NUM_GEN_RING_OSC_FREQ_START[1]
	 * value must be returned to 0 before a subsequent
	 * count can be started.
	 * So before a new write reset this value.
	 */
	img_ring_osc_reset(priv, ring_num);
	img_ring_osc_change_freq_period(priv, period, ring_num);

	return size;
}

static ssize_t img_ring_osc_freq_count_read(struct kobject *kobj,
					    struct kobj_attribute *attr,
					    char *buf)
{
	int ret;
	u32 ring_num, val;
	unsigned char kobj_name[2];
	struct platform_device *pdev;
	struct ring_osc_driver_data *priv;

	pdev = to_platform_device(kobj_to_dev(kobj->parent));
	priv = platform_get_drvdata(pdev);

	kobj_name[0] = *(kobj->name + 4);
	kobj_name[1] = '\0';
	ret = kstrtou32(kobj_name, 10, &ring_num);
	if (ret) {
		dev_err(&pdev->dev, "could not parse ring number string\n");
		return -EINVAL;
	}

	img_ring_osc_reset(priv, ring_num);
	img_ring_osc_start_new_count(priv, ring_num);
	val = img_ring_osc_count_read(priv, ring_num);

	return sprintf(buf, "%d\n", val);
}

static RING_ATTR(period, S_IWUSR, NULL, img_ring_osc_period_write);
static RING_ATTR(count, S_IRUGO, img_ring_osc_freq_count_read, NULL);

static struct attribute *ring_sysfs_attrs[] = {
	&dev_attr_period.attr,
	&dev_attr_count.attr,
	NULL,
};

static struct attribute_group ring_osc_attr_group = {
	.name	= NULL,            /* put in device directory */
	.attrs	= ring_sysfs_attrs,
};

static ssize_t img_ring_osc_ring_module_export(struct device *dev,
					       struct device_attribute *attr,
					       const char *buf, size_t size)
{
	int ret;
	u32 ring_num;
	unsigned char kobj_name[6];
	struct ring_osc_driver_data *priv;
	struct platform_device *pdev;

	pdev = to_platform_device(dev);
	priv = platform_get_drvdata(pdev);

	ret = kstrtou32(buf, 10, &ring_num);
	if (ret) {
		dev_err(&pdev->dev, "could not parse module number string\n");
		return ret;
	}

	if (ring_num == 0) {
		if (!test_bit(0, &priv->flags))
			set_bit(0, &priv->flags);
		else
			return -EBUSY;
	} else if (ring_num == 1) {
		if (!test_bit(1, &priv->flags))
			set_bit(1, &priv->flags);
		else
			return -EBUSY;
	} else {
		return -EINVAL;
	}

	sprintf(kobj_name, "ring%d", ring_num);
	priv->ring_kobj[ring_num] = kobject_create_and_add(kobj_name,
							   &pdev->dev.kobj);

	ret = sysfs_create_group(priv->ring_kobj[ring_num],
				 &ring_osc_attr_group);
	if (ret) {
		kobject_put(priv->ring_kobj[ring_num]);
		dev_err(&pdev->dev, "unable to register device attributes\n");
		return ret;
	}
	return size;
}

static DEVICE_ATTR(export, S_IWUSR, NULL, img_ring_osc_ring_module_export);

static struct attribute *ring_sysfs_entries[] = {
	&dev_attr_export.attr,
	NULL
};

static struct attribute_group ring_attribute_group = {
	.name	= NULL,              /* put in device directory */
	.attrs	= ring_sysfs_entries,
};

static u32 img_ring_osc_readl(struct ring_osc_driver_data *rosc_dev,
			      u32 offset)
{
	return readl(rosc_dev->base + offset);
}

static void img_ring_osc_writel(struct ring_osc_driver_data *rosc_dev,
				u32 val, u32 offset)
{
	writel(val, rosc_dev->base + offset);
}

static void img_ring_osc_config(struct ring_osc_driver_data *priv)
{
	u32 val, offset;

	val = DEFAULT_FREQ_PERIOD;

	down_write(&priv->rosc_lock);
	offset = RING0_CTRL_CFG;
	val |= OSC_EN;
	img_ring_osc_writel(priv, val, offset);
	val |= FREQ_START;
	img_ring_osc_writel(priv, val, offset);

	offset = RING1_CTRL_CFG;
	val |= OSC_EN;
	img_ring_osc_writel(priv, val, offset);
	val |= FREQ_START;
	img_ring_osc_writel(priv, val, offset);

	up_write(&priv->rosc_lock);
}

static int img_ring_osc_probe(struct platform_device *pdev)
{
	struct ring_osc_driver_data *priv;
	struct resource *res;
	int ret;
	const struct of_device_id *of_dev_id;

	priv = devm_kzalloc(&pdev->dev, sizeof(*priv), GFP_KERNEL);
	if (!priv)
		return -ENOMEM;

	priv->dev = &pdev->dev;

	priv->ring_kobj = devm_kcalloc(&pdev->dev, NUM_RING_OSC,
				       sizeof(struct kobject), GFP_KERNEL);

	if (!priv->ring_kobj)
		return -ENOMEM;

	init_rwsem(&priv->rosc_lock);

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	priv->base = devm_ioremap_resource(priv->dev, res);

	if (IS_ERR(priv->base))
		return PTR_ERR(priv->base);

	of_dev_id = of_match_device(img_ring_osc_of_match, &pdev->dev);
	if (!of_dev_id)
		return -ENODEV;

	img_ring_osc_config(priv);

	platform_set_drvdata(pdev, priv);

	ret = sysfs_create_group(&priv->dev->kobj, &ring_attribute_group);
	if (ret) {
		dev_err(priv->dev,
			"Error %d creating sysfs device attributes\n",
			ret);
		platform_set_drvdata(pdev, NULL);
		return ret;
	}

	return 0;
}

static int img_ring_osc_remove(struct platform_device *pdev)
{
	u32 i;
	struct ring_osc_driver_data *priv;

	priv = platform_get_drvdata(pdev);

	for (i = 0; i < NUM_RING_OSC; i++) {
		if (priv->ring_kobj[i]) {
			sysfs_remove_group(priv->ring_kobj[i],
					   &ring_osc_attr_group);
			kobject_del(priv->ring_kobj[i]);
		}
	}

	clear_bit(0, &priv->flags);
	clear_bit(1, &priv->flags);

	return 0;
}

static struct platform_driver img_ring_osc_driver = {
	.driver				= {
		.name			= "img-ring",
		.owner			= THIS_MODULE,
		.of_match_table		= img_ring_osc_of_match,
	},
	.probe				= img_ring_osc_probe,
	.remove				= img_ring_osc_remove,
};
module_platform_driver(img_ring_osc_driver);

MODULE_AUTHOR("Vivek Bhadra <vivek.bhadra@imgtec.com>");
MODULE_DESCRIPTION("Imagination Technologies Ring Oscillator driver");
MODULE_LICENSE("GPL v2");
