/*
 * Imagination Technologies Secure Digital Host Controller Interface.
 *
 * Copyright (c) 2016 Imagination Technologies Ltd.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 *
 * Based on drivers/mmc/host/sdhci-msm.c
 *
 */
#include <linux/delay.h>
#include <linux/jiffies.h>
#include <linux/mmc/host.h>
#include <linux/mmc/slot-gpio.h>
#include <linux/module.h>
#include <linux/of_device.h>
#include <linux/pm_runtime.h>
#include <linux/slab.h>

#include "sdhci.h"
#include "sdhci-pltfm.h"

/* Tuning specific registers. */
#define TX_TUNE					0xE0
#define RX_TUNE					0xE4
#define TUNE_TAP_POINT_MASK			0x3F
#define TUNE_AUTO_INC_MASK			0x40
#define TUNE_INC_MASK				0x380
#define TUNE_DEFVAL				0x1
#define TUNE_INC_DEFVAL				0x80

#define SDHCI_DRV_TYPE_MSB_MASK			0x100
#define SDHCI_CTRL_DRV_TYPE_MSB			0x1

/* AHB burst size register. */
#define AHB_BURST_SIZE				0xEC
#define AHB_BURST_SIZE_MASK			0x7F
#define AHB_BURST_SIZE_DEFVAL			0xF

/* AXI burst size register. */
#define AXI_BURST_SIZE				0xF0
#define AXI_BURST_SIZE_MASK			0x7FF
#define AXI_BURST_SIZE_DEFVAL			0x200

/* SDH revision id register. */
#define SDH_REV_ID				0xF4
#define SDH_MAJOR_REV_MASK			0xFF
#define SDH_MINOR_REV_MASK			0xFF00
#define SDH_MAINTENENCE_REV_MASK		0xFF0000
#define SDH_MINOR_REV_SHIFT			8
#define SDH_MAINTENENCE_REV_SHIFT		16

#define TUNING_LOOP_COUNTER			40

struct sdhci_img_host {
	struct platform_device *pdev;
	struct clk *base_clk;
	struct clk *lp_clk;
	struct mmc_host *mmc;
	struct sdhci_pltfm_data sdhci_img_data;
};

static const unsigned char tuning_blk_pattern_4bit[] = {
	0xff, 0x0f, 0xff, 0x00, 0xff, 0xcc, 0xc3, 0xcc,
	0xc3, 0x3c, 0xcc, 0xff, 0xfe, 0xff, 0xfe, 0xef,
	0xff, 0xdf, 0xff, 0xdd, 0xff, 0xfb, 0xff, 0xfb,
	0xbf, 0xff, 0x7f, 0xff, 0x77, 0xf7, 0xbd, 0xef,
	0xff, 0xf0, 0xff, 0xf0, 0x0f, 0xfc, 0xcc, 0x3c,
	0xcc, 0x33, 0xcc, 0xcf, 0xff, 0xef, 0xff, 0xee,
	0xff, 0xfd, 0xff, 0xfd, 0xdf, 0xff, 0xbf, 0xff,
	0xbb, 0xff, 0xf7, 0xff, 0xf7, 0x7f, 0x7b, 0xde
};

static unsigned int sdhci_img_get_max_clk(struct sdhci_host *host)
{
	struct sdhci_pltfm_host *pltfm_host = sdhci_priv(host);
	struct sdhci_img_host *priv = pltfm_host->priv;

	return clk_get_rate(priv->base_clk);
}

static int set_img_rx_tune(struct sdhci_host *host, u32 opcode)
{
	unsigned char phase = 1, *data_buf;
	int tuned_phase[40], tuned_fail_cnt = 0, tuned_pass_cnt = 0;
	int start = 0, end = 0, val, ret = 0;
	int size = sizeof(tuning_blk_pattern_4bit);
	int curr_passcnt, prev_passcnt = 0;
	const unsigned char *tuning_block_pattern = tuning_blk_pattern_4bit;
	unsigned long flags = 0, timeout;
	struct mmc_host *mmc = host->mmc;

	data_buf = kmalloc(size, GFP_KERNEL);
	if (!data_buf)
		return -ENOMEM;

	spin_lock_irqsave(&host->lock, flags);
	/*
	 * As per the Host Controller spec v3.00, tuning command
	 * generates Buffer Read Ready interrupt, so enable that.
	 *
	 * Note: The spec clearly says that when tuning sequence
	 * is being performed, the controller does not generate
	 * interrupts other than Buffer Read Ready interrupt. But
	 * to make sure we don't hit a controller bug, we _only_
	 * enable Buffer Read Ready interrupt here.
	 */
	sdhci_writel(host, SDHCI_INT_DATA_AVAIL, SDHCI_INT_ENABLE);
	sdhci_writel(host, SDHCI_INT_DATA_AVAIL, SDHCI_SIGNAL_ENABLE);

	/*
	 * Issue CMD19 for SD and CMD21 for MMC repeatedly till the number of
	 * loops reaches 40 times or a timeout of 150ms occurs.
	 */
	timeout = jiffies + msecs_to_jiffies(150);

	do {
		struct mmc_command cmd = { 0 };
		struct mmc_data data = { 0 };
		struct mmc_request mrq = {
			.cmd = &cmd,
			.data = &data
		};
		struct scatterlist sg;

		sdhci_writew(host, phase, RX_TUNE);

		if (time_after(jiffies, timeout)) {
			dev_info(mmc_dev(mmc), "%s:  Timeout of 150ms "
				 "completed, tuning count  = %d\n",
				 mmc_hostname(mmc), phase);
			ret = -ETIMEDOUT;
			break;
		}

		/* Issue command 19 and read the data from the card. */
		cmd.opcode = opcode;
		cmd.flags = MMC_RSP_R1 | MMC_CMD_ADTC;

		data.blksz = size;
		data.blocks = 1;
		data.flags = MMC_DATA_READ;
		data.timeout_ns = NSEC_PER_SEC;	/* 1 second */

		data.sg = &sg;
		data.sg_len = 1;
		sg_init_one(&sg, data_buf, size);
		memset(data_buf, 0, size);
		mmc_wait_for_req(mmc, &mrq);

		spin_unlock_irqrestore(&host->lock, flags);
		/* Wait for Buffer Read Ready interrupt. */
		wait_event_interruptible_timeout(host->buf_ready_int,
						 host->tuning_done == 1,
						 msecs_to_jiffies(50));
		spin_lock_irqsave(&host->lock, flags);

		if (!host->tuning_done) {
			dev_dbg(mmc_dev(mmc),
				"%s: Timeout for Read Ready interrupt during tuning.",
				mmc_hostname(mmc));
			ret = -EIO;
		}

		host->tuning_done = 0;

		if (!cmd.error && !data.error &&
		    memcmp(data_buf, tuning_block_pattern, size)) {
			/* Tuning is successful at this tuning point. */
			tuned_phase[tuned_pass_cnt] = phase;
			if (tuned_fail_cnt) {
				start = phase;
				tuned_fail_cnt = 0;
			}
			if ((start + tuned_pass_cnt) == TUNING_LOOP_COUNTER)
				end = TUNING_LOOP_COUNTER;

			dev_dbg(mmc_dev(mmc),
				"%s: Found good Sampling point = %d\n",
				mmc_hostname(mmc), phase);
			tuned_pass_cnt++;
		} else {
			curr_passcnt = tuned_pass_cnt;
			if (curr_passcnt > prev_passcnt) {
				start = phase - tuned_pass_cnt;
				end = tuned_phase[tuned_pass_cnt - 1];
				prev_passcnt = curr_passcnt;
				tuned_pass_cnt = 0;
			}
			tuned_fail_cnt++;
		}
	} while (++phase <= TUNING_LOOP_COUNTER);

	spin_unlock_irqrestore(&host->lock, flags);

	/* If all the tuning cases are passed. */
	if (tuned_pass_cnt == TUNING_LOOP_COUNTER) {
		start = 1;
		end = TUNING_LOOP_COUNTER;
	}

	if (start <= 0 && end <= 0) {
		ret = -EINVAL;
	} else {
		/* Set the median value in RX_TUNE register. */
		val = sdhci_readw(host, RX_TUNE) & ~TUNE_TAP_POINT_MASK;
		val |= (start + end) / 2;
		sdhci_writew(host, val, RX_TUNE);
	}
	return ret;
}

void sdhci_img_set_uhs_signaling(struct sdhci_host *host, unsigned int uhs)
{
	unsigned short ctrl_2 = sdhci_readw(host, SDHCI_HOST_CONTROL2);

	/* Driver strength in Pads 3bits. */
	ctrl_2 &= ~SDHCI_DRV_TYPE_MSB_MASK;
	ctrl_2 |= SDHCI_CTRL_DRV_TYPE_MSB;
	sdhci_writew(host, ctrl_2, SDHCI_HOST_CONTROL2);

	/* Select Bus Speed Mode for host. */
	ctrl_2 &= ~SDHCI_CTRL_UHS_MASK;
	if (uhs == MMC_TIMING_MMC_HS200 || uhs == MMC_TIMING_UHS_SDR104)
		ctrl_2 |= SDHCI_CTRL_UHS_SDR104;
	else if (uhs == MMC_TIMING_UHS_SDR12)
		ctrl_2 |= SDHCI_CTRL_UHS_SDR12;
	else if (uhs == MMC_TIMING_UHS_SDR25)
		ctrl_2 |= SDHCI_CTRL_UHS_SDR25;
	else if (uhs == MMC_TIMING_UHS_SDR50)
		ctrl_2 |= SDHCI_CTRL_UHS_SDR50;
	else if (uhs == MMC_TIMING_UHS_DDR50)
		ctrl_2 |= SDHCI_CTRL_UHS_DDR50;

	sdhci_writew(host, ctrl_2, SDHCI_HOST_CONTROL2);
}

static int sdhci_img_execute_tuning(struct sdhci_host *host, u32 opcode)
{
	int ret = 0;
	int val = 0;
	struct mmc_ios ios = host->mmc->ios;
	struct mmc_host *mmc = host->mmc;

	/*
	 * Tuning is required for SDR104, HS200 and HS400 cards and
	 * if clock frequency is greater than 100MHz in these modes.
	 */
	if (host->clock <= 100 * 1000 * 1000 ||
	    !((ios.timing == MMC_TIMING_MMC_HS200) ||
	      (ios.timing == MMC_TIMING_UHS_SDR104)))
		return 0;

	ret = set_img_rx_tune(host, opcode);
	if (ret) {
		dev_dbg(mmc_dev(mmc), "%s: Tuning procedure failed.\n",
			mmc_hostname(mmc));

		/* Program Default values in RX tune register. */
		val = sdhci_readw(host, RX_TUNE);
		val &= ~(TUNE_TAP_POINT_MASK | TUNE_AUTO_INC_MASK |
			 TUNE_INC_MASK);
		val |= TUNE_INC_DEFVAL | TUNE_DEFVAL;

		sdhci_writew(host, val, RX_TUNE);
	}
	return ret;
}

static struct sdhci_ops sdhci_img_ops = {
	.platform_execute_tuning = sdhci_img_execute_tuning,
	.set_uhs_signaling = sdhci_img_set_uhs_signaling,
	.get_max_clock = sdhci_img_get_max_clk,

	/* SDHCI default functions. */
	.reset = sdhci_reset,
	.set_clock = sdhci_set_clock,
	.set_bus_width = sdhci_set_bus_width
};

static void sdhci_img_configure_defaults(struct sdhci_host *host)
{
	int val;

	/* Program the default values in TX tune register. */
	val = sdhci_readw(host, TX_TUNE);
	val &= ~(TUNE_TAP_POINT_MASK | TUNE_AUTO_INC_MASK | TUNE_INC_MASK);
	val |= (TUNE_INC_DEFVAL | TUNE_DEFVAL);
	sdhci_writew(host, val, TX_TUNE);

	/* Program the default values in RX tune register. */
	val = sdhci_readw(host, RX_TUNE);
	val &= ~(TUNE_TAP_POINT_MASK | TUNE_AUTO_INC_MASK | TUNE_INC_MASK);

	val |= (TUNE_INC_DEFVAL | TUNE_DEFVAL);
	sdhci_writew(host, val, RX_TUNE);

	/* Program the default value in AHB_BURST_SIZE register. */
	val = sdhci_readw(host, AHB_BURST_SIZE) & ~AHB_BURST_SIZE_MASK;
	val |= AHB_BURST_SIZE_DEFVAL;
	sdhci_writew(host, val, AHB_BURST_SIZE);

	/* Program the default value in AXI_BURST_SIZE register. */
	val = sdhci_readl(host, AXI_BURST_SIZE) & ~AXI_BURST_SIZE_MASK;
	val |= AXI_BURST_SIZE_DEFVAL;
	sdhci_writel(host, val, AXI_BURST_SIZE);
}

static int sdhci_img_probe(struct platform_device *pdev)
{
	int ret = 0;
	u32 host_revision;
	struct sdhci_host *host;
	struct sdhci_pltfm_host *pltfm_host;
	struct sdhci_img_host *img_host;

	img_host = devm_kzalloc(&pdev->dev, sizeof(*img_host), GFP_KERNEL);
	if (!img_host)
		return -ENOMEM;

	img_host->sdhci_img_data.ops = &sdhci_img_ops;
	host = sdhci_pltfm_init(pdev, &img_host->sdhci_img_data, 0);
	if (IS_ERR(host))
		return PTR_ERR(host);

	pltfm_host = sdhci_priv(host);

	pltfm_host->priv = img_host;

	img_host->mmc = host->mmc;
	img_host->pdev = pdev;

	ret = mmc_of_parse(host->mmc);
	if (ret)
		goto pltfm_free;

	sdhci_get_of_property(pdev);

	img_host->base_clk = devm_clk_get(&pdev->dev, "base");
	ret = IS_ERR(img_host->base_clk);
	if (ret) {
		dev_err(&pdev->dev, "base clock not found.\n");
		goto pltfm_free;
	}

	img_host->lp_clk = devm_clk_get(&pdev->dev, "lp");
	ret = IS_ERR(img_host->lp_clk);
	if (ret) {
		dev_err(&pdev->dev, "lp clock not found.\n");
		goto pltfm_free;
	}

	ret = clk_prepare_enable(img_host->base_clk);
	if (ret) {
		dev_err(&pdev->dev, "Unable to enable base clock.\n");
		goto pltfm_free;
	}

	ret = clk_prepare_enable(img_host->lp_clk);
	if (ret) {
		dev_err(&pdev->dev, "Unable to enable lp clock.\n");
		goto base_clk_disable;
	}

	sdhci_img_configure_defaults(host);

	/* Host controller quirks. */
	host->quirks = SDHCI_QUIRK_CAP_CLOCK_BASE_BROKEN;
	host->quirks |= SDHCI_QUIRK_SINGLE_POWER_WRITE;
	host->quirks |= SDHCI_QUIRK_DELAY_AFTER_POWER;

	/* Enable 8-bits support. */
	host->mmc->caps |= MMC_CAP_8_BIT_DATA;

	host_revision = sdhci_readl(host, SDH_REV_ID);
	dev_dbg(&pdev->dev,
		"Major rev: 0x%x, Minor rev: 0x%x, Maintenance: 0x%x.\n",
		host_revision & SDH_MAJOR_REV_MASK,
		(host_revision & SDH_MINOR_REV_MASK) >> SDH_MINOR_REV_SHIFT,
		(host_revision & SDH_MAINTENENCE_REV_MASK
		 >> SDH_MAINTENENCE_REV_SHIFT));

	pm_runtime_get_noresume(&pdev->dev);
	pm_runtime_set_active(&pdev->dev);
	pm_runtime_enable(&pdev->dev);
	pm_runtime_set_autosuspend_delay(&pdev->dev, 50);
	pm_runtime_use_autosuspend(&pdev->dev);

	ret = sdhci_add_host(host);
	if (ret) {
		dev_err(&pdev->dev, "Failed to add SDHCI host (%u).\n", ret);
		goto pm_runtime_disable;
	}

	/*
	 * When calling sdhci_runtime_suspend_host(), the sdhci layer makes
	 * the assumption that all the clocks of the controller are disabled.
	 * It means we can't get irq from it when it is runtime suspended.
	 * For that reason, it is not planned to wake-up on a card detect irq
	 * from the controller.
	 * If we want to use runtime PM and to be able to wake-up on card
	 * insertion, we have to use a GPIO for the card detection or we can
	 * use polling. Be aware that using polling will resume/suspend the
	 * controller between each attempt.
	 * Disable SDHCI_QUIRK_BROKEN_CARD_DETECTION to be sure nobody tries
	 * to enable polling via device tree with broken-cd property.
	 */
	if (!(host->mmc->caps & MMC_CAP_NONREMOVABLE) &&
	    mmc_gpio_get_cd(host->mmc) < 0) {
		host->mmc->caps |= MMC_CAP_NEEDS_POLL;
		host->quirks &= ~SDHCI_QUIRK_BROKEN_CARD_DETECTION;
	}

	pm_runtime_put_autosuspend(&pdev->dev);

	return 0;

pm_runtime_disable:
	pm_runtime_dont_use_autosuspend(&pdev->dev);
	pm_runtime_disable(&pdev->dev);
	pm_runtime_set_suspended(&pdev->dev);
	pm_runtime_put_noidle(&pdev->dev);
	clk_disable_unprepare(img_host->lp_clk);
base_clk_disable:
	clk_disable_unprepare(img_host->base_clk);
pltfm_free:
	sdhci_pltfm_free(pdev);

	return ret;
}

static int sdhci_img_remove(struct platform_device *pdev)
{
	struct sdhci_host *host = platform_get_drvdata(pdev);
	struct sdhci_pltfm_host *pltfm_host = sdhci_priv(host);
	struct sdhci_img_host *img_host = pltfm_host->priv;
	int dead = (readl_relaxed(host->ioaddr + SDHCI_INT_STATUS) ==
		    0xffffffff);

	pm_runtime_get_sync(&pdev->dev);
	pm_runtime_disable(&pdev->dev);
	pm_runtime_put_noidle(&pdev->dev);

	sdhci_remove_host(host, dead);
	sdhci_pltfm_free(pdev);

	clk_disable_unprepare(img_host->base_clk);
	clk_disable_unprepare(img_host->lp_clk);

	return 0;
}

#ifdef CONFIG_PM
static int sdhci_img_runtime_suspend(struct device *dev)
{
	struct sdhci_host *host = dev_get_drvdata(dev);
	struct sdhci_pltfm_host *pltfm_host = sdhci_priv(host);
	struct sdhci_img_host *img_host = pltfm_host->priv;
	int ret;

	ret = sdhci_runtime_suspend_host(host);
	if (ret)
		return ret;

	clk_disable_unprepare(img_host->lp_clk);
	clk_disable_unprepare(img_host->base_clk);

	return 0;
}

static int sdhci_img_runtime_resume(struct device *dev)
{
	struct sdhci_host *host = dev_get_drvdata(dev);
	struct sdhci_pltfm_host *pltfm_host = sdhci_priv(host);
	struct sdhci_img_host *img_host = pltfm_host->priv;
	int ret;

	ret = clk_prepare_enable(img_host->base_clk);
	if (ret) {
		dev_err(dev, "Unable to enable base clock.\n");
		return ret;
	}

	ret = clk_prepare_enable(img_host->lp_clk);
	if (ret) {
		dev_err(dev, "Unable to enable lp clock.\n");
		goto base_clk_disable;
	}

	sdhci_img_configure_defaults(host);

	ret = sdhci_runtime_resume_host(host);
	if (ret)
		goto lp_clk_disable;

	return 0;

lp_clk_disable:
	clk_disable_unprepare(img_host->lp_clk);
base_clk_disable:
	clk_disable_unprepare(img_host->base_clk);
	return ret;
}
#endif

static const struct dev_pm_ops sdhci_img_pm_ops = {
	SET_SYSTEM_SLEEP_PM_OPS(pm_runtime_force_suspend,
				pm_runtime_force_resume)
	SET_RUNTIME_PM_OPS(sdhci_img_runtime_suspend,
			   sdhci_img_runtime_resume,
			   NULL)
};

static const struct of_device_id sdhci_img_dt_match[] = {
	{ .compatible = "img,sdhci" },
	{ }
};

MODULE_DEVICE_TABLE(of, sdhci_img_dt_match);

static struct platform_driver sdhci_img_driver = {
	.driver = {
		.name = "img,sdhci",
		.owner = THIS_MODULE,
		.pm = &sdhci_img_pm_ops,
		.of_match_table = sdhci_img_dt_match,
	},
	.probe = sdhci_img_probe,
	.remove = sdhci_img_remove,
};

module_platform_driver(sdhci_img_driver);

MODULE_DESCRIPTION("Imagination Technologies SDHCI driver");
MODULE_AUTHOR("Krzysztof Antonowicz <krzysztof.antonowicz@imgtec.com>");
MODULE_LICENSE("GPL");
