/*
 * Fake flash driver used for Imagination Technologies VP (Virtual Platform)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
*/
#include <linux/module.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/device.h>
#include <linux/mutex.h>
#include <linux/err.h>
#include <linux/of.h>
#include <linux/of_device.h>
#include <linux/mtd/mtd.h>
#include <linux/platform_device.h>
#include <linux/io.h>

#include <asm/delay.h>

enum FAKE_FLASH_CONTROLLER_REGS {
    FS_CONTROL = 0x00000000,
    FS_STATUS = 0x00000004,
    FS_LENGTH = 0x00000008,
    FS_TRANSFER_OFFSET = 0x0000000C,
    FS_TRANSFER_LENGTH = 0x000000014,
    FS_TRANSFER_SOURCE_DESTINATION = 0x000000018,
};

/* Defines used for accessing FS_CONTROL register. */
#define FS_CONTROL_START_TRANSFER_MASK      (0x00000001)
#define FS_CONTROL_START_TRANSFER_OFFSET    (0)
#define FS_CONTROL_TRANSFER_WRITE_MASK      (0x00000002)
#define FS_CONTROL_TRANSFER_WRITE_OFFSET    (1)

/* Defines used for accessing FS_STATUS register. */
#define FS_STATUS_FS_READY_MASK             (0x00000001)
#define FS_STATUS_FS_READY_OFFSET           (0)
#define FS_STATUS_TRANSFER_FINISHED_MASK    (0x00000002)
#define FS_STATUS_TRANSFER_FINISHED_OFFSET  (1)
#define FS_STATUS_TRANSFER_ERROR_MASK       (0x00000004)
#define FS_STATUS_TRANSFER_ERROR_OFFSET     (2)

#define SET_REG_FLD(reg_base_addr, reg, field, val) \
    do { iowrite32((ioread32(reg_base_addr + reg) & ~reg##_##field##_MASK) | \
            (reg##_##field##_MASK & (val << reg##_##field##_OFFSET)), \
            reg_base_addr + reg); } while(0);

#define SET_REG(reg_base_addr, reg, val) \
    do { iowrite32(val, reg_base_addr + reg); } while(0);

#define GET_REG(reg_base_addr, reg) \
    ioread32(reg_base_addr + reg)

#define GET_REG_FLD(reg_base_addr, reg, field) \
    ((ioread32(reg_base_addr + reg) & reg##_##field##_MASK) >> \
     reg##_##field##_OFFSET)

#define REGISTER_READ_DELAY_US (10)
static struct mutex lock;

static const struct of_device_id fake_flash_vp_dt_ids[] = {
	{ .compatible = "fake_flash_vp_system", },
	{ .compatible = "fake_flash_vp_data", },
	{ .compatible = "fake_flash_vp_cache", },
	{ /* sentinel */ }
};

static int fake_erase(struct mtd_info *mtd, struct erase_info *instr) {
    unsigned err = -EIO;
    pr_err("Erasing not implemented in the driver!\n");
    return err;
}

static unsigned long fake_get_unmapped_area(struct mtd_info *mtd,
                       unsigned long len, unsigned long offset,
                       unsigned long flags) {
    pr_err("Getting unmapped area failed!\n");
    return -ENOSYS;
}

#if 0
static int fake_point(struct mtd_info *mtd, loff_t from, size_t len,
		size_t *retlen, void **virt, resource_size_t *phys) {
    pr_err("Pointing not implemented in the driver!\n");
	return -EIO;
}

static int fake_unpoint(struct mtd_info *mtd, loff_t from, size_t len) {
    pr_err("Un-pointing not implemented in the driver!\n");
	return -EIO;
}
#endif

static int fake_read(struct mtd_info *mtd, loff_t offset, size_t len,
        size_t *retlen, u_char *buf) {
    unsigned err = 0;

    mutex_lock(&lock);
    if (GET_REG_FLD(mtd->priv, FS_STATUS, FS_READY)) {
        SET_REG(mtd->priv, FS_TRANSFER_SOURCE_DESTINATION, virt_to_phys(buf));
        SET_REG(mtd->priv, FS_TRANSFER_LENGTH, len);
        SET_REG(mtd->priv, FS_TRANSFER_OFFSET, offset);
        SET_REG_FLD(mtd->priv, FS_CONTROL, TRANSFER_WRITE, 0);
        SET_REG_FLD(mtd->priv, FS_STATUS, TRANSFER_FINISHED, 0);
        SET_REG_FLD(mtd->priv, FS_STATUS, TRANSFER_ERROR, 0);
        SET_REG_FLD(mtd->priv, FS_CONTROL, START_TRANSFER, 1);

        // TODO: This may be endless!
        while (!GET_REG_FLD(mtd->priv, FS_STATUS, TRANSFER_FINISHED))
            udelay(REGISTER_READ_DELAY_US);

        if (GET_REG_FLD(mtd->priv, FS_STATUS, TRANSFER_ERROR)) {
            pr_err("Read transfer error!\n");
            err = -EIO;
        }
    } else {
        pr_err("FS not ready for reading!\n");
        err = -EIO;
    }
    mutex_unlock(&lock);
    return err;
}

static int fake_write(struct mtd_info *mtd, loff_t offset, size_t len,
        size_t *retlen, const u_char *buf) {
    unsigned err = 0;

    mutex_lock(&lock);
    if (GET_REG_FLD(mtd->priv, FS_STATUS, FS_READY)) {
        SET_REG(mtd->priv, FS_TRANSFER_SOURCE_DESTINATION, virt_to_phys(buf));
        SET_REG(mtd->priv, FS_TRANSFER_LENGTH, len);
        SET_REG(mtd->priv, FS_TRANSFER_OFFSET, offset);
        SET_REG_FLD(mtd->priv, FS_CONTROL, TRANSFER_WRITE, 1);
        SET_REG_FLD(mtd->priv, FS_STATUS, TRANSFER_FINISHED, 0);
        SET_REG_FLD(mtd->priv, FS_STATUS, TRANSFER_ERROR, 0);
        SET_REG_FLD(mtd->priv, FS_CONTROL, START_TRANSFER, 1);

        while (!GET_REG_FLD(mtd->priv, FS_STATUS, TRANSFER_FINISHED))
            udelay(REGISTER_READ_DELAY_US);

        if (GET_REG_FLD(mtd->priv, FS_STATUS, TRANSFER_ERROR)) {
            pr_err("Write transfer error!\n");
            err = -EIO;
        }
    } else {
        pr_err("FS not ready for writing!\n");
        err = -EIO;
    }
    mutex_unlock(&lock);

    return err;
}

static int fake_flash_vp_probe(struct platform_device *pdev) {
    const struct of_device_id *of_id = of_match_device(
            fake_flash_vp_dt_ids, &pdev->dev);
    void __iomem *regs_vaddr = NULL;
    struct resource *regs_res = NULL;
    struct mtd_info *mtd_info = NULL;
    unsigned fs_length = 0;
    unsigned err = 0;

    if (of_id) {
        /* Use of_id->data here. */
        pr_debug("%s DEBUG: Device matched up with the driver.\n",
                pdev->name);

        regs_res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
        if (!regs_res) {
            pr_err("%s ERROR: Cannot find registry range for the MTD device.\n",
                    pdev->name);

            err = -ENODEV;
            goto ERROR_GET_RESOURCE;
        }
        regs_vaddr = devm_ioremap(&pdev->dev, regs_res->start,
                resource_size(regs_res));

        if (!regs_vaddr) {
            pr_err("%s ERROR: Cannot map IO region for the MTD device.\n",
                    pdev->name);
            err = -ENOMEM;
            goto ERROR_GET_RESOURCE;
        }

        fs_length = GET_REG(regs_vaddr, FS_LENGTH);
        pr_debug("%s DEBUG: FS_LENGTH=%d\n", pdev->name, fs_length);

        if (!fs_length) {
            pr_err("%s ERROR: Filesystem size cannot be 0.\n", pdev->name);
            err = -ENODEV;
            goto ERROR_FS_LENGTH;
        }

        /* Create MTDRAM device. */
        mtd_info = devm_kzalloc(&pdev->dev, sizeof(*mtd_info), GFP_KERNEL);
        if (!mtd_info) {
            pr_err("%s ERROR: Cannot allocate memory for MTD device.\n",
                    pdev->name);
            err = -ENOMEM;
            goto ERROR_GET_RESOURCE;
        }

        /* Setup the MTD structure */
        mtd_info->name = of_id->compatible;
        mtd_info->type = MTD_RAM;
        mtd_info->flags = MTD_CAP_RAM;
        mtd_info->size = fs_length;
        mtd_info->writesize = 1;
        mtd_info->writebufsize = 64;
        mtd_info->erasesize = 0x20000;

        mtd_info->owner = THIS_MODULE;
        mtd_info->_erase = fake_erase;
        mtd_info->_point = NULL;
        mtd_info->_unpoint = NULL;
        mtd_info->_get_unmapped_area = fake_get_unmapped_area;
        mtd_info->_read = fake_read;
        mtd_info->_write = fake_write;

        mtd_info->priv = regs_vaddr;

        if (mtd_device_register(mtd_info, NULL, 0)) {
            pr_err("%s ERROR: Cannot register MTD device.\n", pdev->name);
            err = -EIO;
            goto ERROR_MTD_DEVICE_REGISTER;
        }

        platform_set_drvdata(pdev, mtd_info);
        mutex_init(&lock);
    } else pr_err("%s ERROR: Cannot match up the device!\n", pdev->name);

    return err;

ERROR_MTD_DEVICE_REGISTER:
    devm_kfree(&pdev->dev, mtd_info);
ERROR_FS_LENGTH:
    devm_iounmap(&pdev->dev, regs_vaddr);
ERROR_GET_RESOURCE:
	return err;
}

static int fake_flash_vp_remove(struct platform_device *pdev) {
    struct mtd_info *mtd_info = platform_get_drvdata(pdev);
    if (mtd_info) {
        mtd_device_unregister(mtd_info);
        devm_iounmap(&pdev->dev, mtd_info->priv);
        devm_kfree(&pdev->dev, mtd_info);
    }
	return 0;
}

static struct platform_driver fake_flash_vp_driver = {
	.probe	= fake_flash_vp_probe,
	.remove = fake_flash_vp_remove,
	.driver = {
		.name = "fake_flash_vp",
		.owner = THIS_MODULE,
		.of_match_table = of_match_ptr(fake_flash_vp_dt_ids),
	},
};

module_platform_driver(fake_flash_vp_driver);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Krzysztof Antonowicz");
MODULE_DESCRIPTION("MTD fake flash driver for VP");
