/*
 * Imagination Technologies Generic Nand Controller Driver
 *
 * Copyright (c) 2015 Imagination Technologies
 *
 * Based on drivers/mtd/nand/omap2.c
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/clk.h>
#include <linux/delay.h>
#include <linux/dmaengine.h>
#include <linux/dma-mapping.h>
#include <linux/interrupt.h>
#include <linux/jiffies.h>
#include <linux/module.h>
#include <linux/mtd/nand.h>
#include <linux/of_device.h>
#include <linux/of_mtd.h>
#include <linux/mtd/partitions.h>
#include <linux/slab.h>

/* Registers */
#define REG_NFLASH_CONTROL_3					0x44

#define REG_MODE_SELECT						0x0400
#define REG_MODE_SELECT_FLD_SDR_TIMING_MODE_SHIFT		0x3
#define REG_MODE_SELECT_FLD_DDR_TIMING_MODE_SHIFT		0x7
#define REG_MODE_SELECT_FLD_DDR2_TIMING_MODE_SHIFT		0xB

#define REG_CTRL_REG0						0x0404
#define REG_CTRL_REG0_FLD_WR_PROTECT_SHIFT			0x3
#define REG_CTRL_REG0_FLD_BUS_WIDTH_SHIFT			0x4
#define REG_CTRL_REG0_FLD_COL_ADDR_CYC_SHIFT			0x5
#define REG_CTRL_REG0_FLD_ROW_ADDR_CYC_SHIFT			0x9
#define REG_CTRL_REG0_FLD_PAGE_SIZE_SHIFT			0xD
#define REG_CTRL_REG0_FLD_TX_MODE_SHIFT				0x10
#define REG_CTRL_REG0_FLD_TX_MODE_VAL_WRITE_TO_FLASH		0x1
#define REG_CTRL_REG0_FLD_CHIP_SELECT_SHIFT			0x13

#define REG_BCH_CTRL_REG0					0x0408
#define REG_BCH_CTRL_REG0_FLD_ECC_ON_OFF_MASK			0x1
#define REG_BCH_CTRL_REG0_FLD_ECC_ALGO_SHIFT			0x1
#define REG_BCH_CTRL_REG0_BCH_T_SHIFT				0x2
#define REG_BCH_CTRL_REG0_BCH_N_SHIFT				0x6

#define REG_BCH_CTRL_REG1					0x040C

#define REG_CODEWORD_SIZE_CNT					0x041C
#define REG_CODEWORD_SIZE_CNT_FLD_CNT_SHIFT			0x10

#define REG_CMD_REG						0x0410
#define REG_CMD_REG_FLD_COMMAND_VAL_READ_STATUS			BIT(10)

#define REG_CMD_BUF						0x0420
#define REG_CMD_BUF_FLD_CMD_DATA_VAL_CMD_TYPE_SHIFT		0x8

#define REG_DATA_BUF						0x0424

#define REG_INT_EN						0x0448
#define REG_INT_EN_FLD_CMD_BUF_FULL				BIT(1)
#define REG_INT_EN_FLD_DATA_BUF_EMPTY				BIT(2)
#define REG_INT_EN_FLD_UNCORR_ERR_INT				BIT(5)
#define REG_INT_EN_FLD_DATA_BUF_WR_RDY_INT			BIT(6)
#define REG_INT_EN_FLD_DATA_BUF_RD_RDY_INT			BIT(7)
#define REG_INT_EN_FLD_WRRD_DONE_INT				BIT(8)
#define REG_INT_EN_FLD_TRANS_COMP				BIT(10)
#define REG_INT_EN_FLD_CMD_BUF_OVERFLOW				BIT(11)
#define REG_INT_EN_FLD_CMD_BUF_UNDERFLOW			BIT(12)
#define REG_INT_EN_FLD_DATA_BUF_OVERFLOW			BIT(13)
#define REG_INT_EN_FLD_DATA_BUF_UNDERFLOW			BIT(14)
#define REG_INT_EN_FLD_DMA_TRANS_DONE				BIT(15)
#define REG_INT_EN_FLD_DMA_BOUNDARY_CROSS_INT			BIT(16)
#define REG_INT_EN_FLD_SLAVE_ERROR				BIT(17)
#define REG_INT_EN_VAL_DISABLE_ALL_INTS				0x00000

/* User timing parameters registers */
#define REG_SDR_USER_TIMING_PARAMS0				0x042C
#define REG_SDR_USER_TIMING_PARAMS0_FLD_HOLD_TIME_SHIFT		0x6
#define REG_SDR_USER_TIMING_PARAMS0_FLD_WH_TIME_SHIFT		0xC
#define REG_SDR_USER_TIMING_PARAMS0_FLD_WRP_TIME_SHIFT		0x12
#define REG_SDR_USER_TIMING_PARAMS0_FLD_INTCMD_TIME_SHIFT	0x18

#define REG_SDR_USER_TIMING_PARAMS1				0x0430
#define REG_SDR_USER_TIMING_PARAMS1_FLD_WB_TIME_SHIFT		0x6
#define REG_SDR_USER_TIMING_PARAMS1_FLD_RD_DLY_SHIFT		0xC
#define NAND_SDR_EXTN_WB_TIME_SHIFT				0xE

#define REG_DDR_USER_TIMING_PARAMS0				0x0434
#define REG_DDR_USER_TIMING_PARAMS0_FLD_HOLD_TIME_SHIFT		0x6
#define REG_DDR_USER_TIMING_PARAMS0_FLD_WH_TIME_SHIFT		0xC
#define REG_DDR_USER_TIMING_PARAMS0_FLD_WRP_TIME_SHIFT		0x12
#define REG_DDR_USER_TIMING_PARAMS0_FLD_INTCMD_TIME_SHIFT	0x18

#define REG_DDR_USER_TIMING_PARAMS1				0x0438
#define REG_DDR_USER_TIMING_PARAMS1_FLD_WB_TIME_SHIFT		0x6
#define REG_DDR_USER_TIMING_PARAMS1_FLD_CLKDIV_SHIFT		0xC
#define REG_DDR_USER_TIMING_PARAMS1_FLD_CE_TIME_SHIFT		0x12
#define REG_DDR_USER_TIMING_PARAMS1_FLD_DQS_TIME_SHIFT		0x18

#define REG_DDR_USER_TIMING_PARAMS2				0x0458

#define REG_DDR2_USER_TIMING_PARAMS0				0x043C
#define REG_DDR2_USER_TIMING_PARAMS0_FLD_HOLD_TIME_SHIFT	0x6
#define REG_DDR2_USER_TIMING_PARAMS0_FLD_WH_TIME_SHIFT		0xC
#define REG_DDR2_USER_TIMING_PARAMS0_FLD_WRP_TIME_SHIFT		0x12
#define REG_DDR2_USER_TIMING_PARAMS0_FLD_INTCMD_TIME_SHIFT	0x18

#define REG_DDR2_USER_TIMING_PARAMS1				0x0440
#define REG_DDR2_USER_TIMING_PARAMS1_FLD_WB_TIME_SHIFT		0x6
#define REG_DDR2_USER_TIMING_PARAMS1_FLD_CLKDIV_SHIFT		0xC
#define REG_DDR2_USER_TIMING_PARAMS1_FLD_CE_TIME_SHIFT		0x12
#define REG_DDR2_USER_TIMING_PARAMS1_FLD_DQS_TIME_SHIFT		0x18

#define REG_DDR2_USER_TIMING_PARAMS2				0x045C

#define REG_RDY_BUSY_STATUS					0x044C

#define FLASH_STATUS_RDY					BIT(6)

#define REG_INT_STATUS						0x0450
#define REG_INT_STATUS_VAL_CLR_ALL_NAND_INT			0x3FFFF

#define REG_FLASH_STATUS					0x0454

#define REG_RESET						0x0460
#define REG_RESET_FLD_SOFT_RESET				BIT(0)
#define REG_RESET_FLD_CMD_FIFO_RESET				BIT(1)
#define REG_RESET_FLD_DATA_FIFO_RESET				BIT(2)
#define REG_RESET_FLD_DDRRX_FIFO_RESET				BIT(3)
#define REG_RESET_FLD_CMD_ENGINE_RESET				BIT(4)
#define REG_RESET_FLD_TIME_ENGINE_RESET				BIT(5)

#define REG_BCH_STATUS						0x0474
#define REG_BCH_STATUS_FLD_ERR_CNT_MASK				0x3F
#define REG_BCH_STATUS_FLD_ACC_ERR_CNT_SHIFT			0x6
#define REG_BCH_STATUS_FLD_ACC_ERR_CNT_MASK			0xFFF

#define REG_DMA_ADDR						0x047C

#define REG_DMA_CTRL						0x0480
#define REG_DMA_CTRL_FLD_DMA_EN					BIT(0)
#define REG_DMA_CTRL_FLD_DATA_DIR_SHIFT				0x1
#define REG_DMA_CTRL_FLD_BUF_SIZE_SHIFT				0x2
#define REG_DMA_CTRL_FLD_BUF_SIZE_MASK				0x7

#define REG_RESET_STATUS					0x048C
#define REG_RESET_STATUS_FLD_TIMING_ENGINE_RESET_DONE		BIT(0)

/* Other Macro definitions */
#define DRIVER_NAME						"img-nand"
#define IMG_MIN_NAND_PAGE_SUPPORT				2048

/* Location where the ECC data to be stored */
#define IMG_ECC_OFFSET_IN_OOB					16

/* Use 10s timeout for very long transfers, it should suffice. */
#define IMG_NAND_TIMEOUT					10000
#define IMG_BCH_ALGO_POLYNOMIAL					15
#define IMG_HAMMING_CW_SIZE					512
#define IMG_MIN_BCH_CW_SIZE					512
#define IMG_MAX_BCH_CW_SIZE					1024
#define IMG_MAX_NAND_CHIPS					8
#define IMG_NAND_BUS_WIDTH_16					16
#define IMG_MAX_ERR_CORR_BCH					40
#define IMG_MIN_ERR_CORR_BCH					4
#define IMG_DFL_ERR_CORR_HAMMING				1
#define IMG_NAND_ECC_ON						1
#define IMG_NAND_ECC_OFF					0
#define IMG_NAND_CLK_RATIO_200MHZ				1
#define IMG_NAND_PAGE_PROG_WAIT_CYCLE				16
#define IMG_NAND_DMA_WRITE					1
#define IMG_NAND_DMA_READ					0
#define	IMG_NAND_BYTE_MASK					0xFF
#define	IMG_NAND_WORD_MASK					0xFFFF
#define	IMG_NAND_BYTE_SHIFT					8
#define	IMG_NAND_WORD_SHIFT					16

#define IMG_NAND_MIN_DMA_BUF_SIZE				4096
#define IMG_NAND_ALL_INT_SET					0x3FFFF
/* Nand No Operation Command value */
#define IMG_NAND_NO_OP_CMD_VAL					0x00
/* Nand Wait for Ready Command value */
#define IMG_NAND_WAIT_FOR_RDY_CMD_VAL				0x00
#define IMG_NAND_MIN_DATA_BUF_RD_CNT				0x4
#define IMG_NAND_MIN_STATUS_RD_CNT				0x1
#define IMG_NAND_MIN_CW_SIZE					0x4
#define IMG_NAND_MIN_CW_CNT					0x1

#define PAGE_WRITTEN_MAGIC_NUMBER				(0xABBADABA)
#define PAGE_WRITTEN_MAGIC_NUMBER_SIZE				(4)

/* oob info generated runtime depending on ecc algorithm and layout selected */
static struct nand_ecclayout img_oobinfo;

enum img_mode {
	IMG_NAND_SDR_MODE,
	IMG_NAND_DDR_MODE,
	IMG_NAND_DDR2_MODE
};

enum img_timing_mode {
	IMG_NAND_TMODE_0,
	IMG_NAND_TMODE_1,
	IMG_NAND_TMODE_2,
	IMG_NAND_TMODE_3,
	IMG_NAND_TMODE_4,
	IMG_NAND_TMODE_5,
	IMG_NAND_TMODE_6,
	IMG_NAND_TMODE_7,
	IMG_NAND_TMODE_USER = 0xF
};

enum img_cmd_type {
	NO_OPERATION,
	COMMAND_WITHOUT_WAIT,
	NORMAL_COMMAND,
	ADDR_TYPE,
	WAIT_CYCLE_CNT,
	WAIT_FOR_READY,
	RD_BYTE_CNT_TYPE,
	WR_BYTE_CNT_TYPE,
	STATUS_RD_CMD,
	SPARE_RD_TYPE,
	SPARE_WR_TYPE,
	STATUS_WR_CMD,
};

enum img_ecc_algo {
	IMG_NAND_ECC_ALGO_HAMMING,
	IMG_NAND_ECC_ALGO_BCH
};

struct img_nand_sdr_timings {
	unsigned int tSETUP;
	unsigned int tHOLD;
	unsigned int tWH;
	unsigned int tWRP;
	unsigned int tINTCMD;
	unsigned int tRR;
	unsigned int tWB;
	unsigned int tRDDLY;
};

struct img_nand_ddr_timings {
	unsigned int tSETUP;
	unsigned int tHOLD;
	unsigned int tWH;
	unsigned int tWRP;
	unsigned int tINTCMD;
	unsigned int tRR;
	unsigned int tWB;
	unsigned int tCLKDiv;
	unsigned int tCE;
	unsigned int tDQSin;
	unsigned int tDQSout;
};

struct img_nand_ctrl {
	bool			dev_ready;
	bool			wp_ctrl;
	unsigned char		chip_sel;	/* select a bank */
	unsigned char		bus_width;	/* 8 bit or 16 bit */
	unsigned char		col_addr_cycles;
	unsigned char		row_addr_cycles;
};

struct img_nand_eccinfo {
	unsigned char		ecc_opt;	/* ECC ON/OFF */
	enum img_ecc_algo	ecc_algo;	/* BCH or Hamming */
	unsigned char		err_corr;	/* number of errors to be
						 * corrected
						 */
};

struct img_nand_dmactrl {
	u8			*dma_buf_virt;
	dma_addr_t		dma_buf_phys;
	bool			dma_en;
	bool			has_dma;
	bool			dma_done;
	unsigned int		dma_start_addr;
	unsigned int		dma_update_addr;
	unsigned char		dma_buf_size;
};

struct img_nand_host {
	int			prev_cmd;
	int			curr_cmd;
	int			curr_column;
	int			curr_page_addr;
	bool			cmd_post_proc_needed;
	bool			flash_bbt;
	bool			uncorrectable_err;
	bool			irq_initialized;
	bool			use_user_timings;
	unsigned char		mode;		/* SDR, DDR or DDR2 */
	unsigned char		tmode;		/* select a timing mode */
	unsigned char		oob_access;	/* OOB area access indication */
	unsigned char		raw_access;	/* raw data access indication */
	unsigned int		cw_size;	/* Codeword size */
	unsigned int		bch_cw_size;
	bool			bch_byte_align_8bitmode;
	unsigned int		bytes_xfered;
	void __iomem		*io_base;
	struct clk		*flash_clk;
	struct mtd_info		mtd;
	struct nand_chip	chip;
	struct completion	comp_nand;
	struct img_nand_ctrl	ctrl;		/* control_reg0 fields */
	struct img_nand_dmactrl	dma;
	struct img_nand_eccinfo	ecc_data;	/* ecc data sub-structure */
};

/* User timing settings for SDR mode. */
static const struct img_nand_sdr_timings user_sdr_timings = {
	.tSETUP = 4,
	.tHOLD = 4,
	.tWH = 2,
	.tWRP = 2,
	.tINTCMD = 16,
	.tRR = 8,
	.tWB = 20,
	.tRDDLY = 3
};

/* User timing settings for DDR mode. */
static const struct img_nand_ddr_timings user_ddr_timings = {
	.tSETUP = 1,
	.tHOLD = 8,
	.tWH = 4,
	.tWRP = 1,
	.tINTCMD = 16,
	.tRR = 12,
	.tWB = 0,
	.tCLKDiv = 4,
	.tCE = 16,
	.tDQSin = 4,
	.tDQSout = 4
};

/* User timing settings for DDR2 mode. */
static const struct img_nand_ddr_timings user_ddr2_timings = {
	.tSETUP = 2,
	.tHOLD = 2,
	.tWH = 2,
	.tWRP = 1,
	.tINTCMD = 16,
	.tRR = 8,
	.tWB = 0,
	.tCLKDiv = 2,
	.tCE = 15,
	.tDQSin = 30,
	.tDQSout = 30
};

static u32 img_nand_read_reg(struct img_nand_host *host, int reg)
{
	return readl(host->io_base + reg);
}

static __maybe_unused u16 img_nand_read_word(struct img_nand_host *host,
					     int reg)
{
	return readw(host->io_base + reg);
}

static void img_nand_write_reg(struct img_nand_host *host, int reg, u32 val)
{
	writel(val, host->io_base + reg);
}

static void img_nand_write_word(struct img_nand_host *host, int reg, u16 val)
{
	writew(val, host->io_base + reg);
}

static void img_nand_configure_sdr_user_timings(struct img_nand_host *host)
{
	const struct img_nand_sdr_timings *user_timings = &user_sdr_timings;

	u32 val = (user_timings->tSETUP |
		   (user_timings->tHOLD <<
		    REG_SDR_USER_TIMING_PARAMS0_FLD_HOLD_TIME_SHIFT) |
		   (user_timings->tWH <<
		    REG_SDR_USER_TIMING_PARAMS0_FLD_WH_TIME_SHIFT) |
		   (user_timings->tWRP <<
		    REG_SDR_USER_TIMING_PARAMS0_FLD_WRP_TIME_SHIFT) |
		   (user_timings->tINTCMD <<
		    REG_SDR_USER_TIMING_PARAMS0_FLD_INTCMD_TIME_SHIFT));
	img_nand_write_reg(host, REG_SDR_USER_TIMING_PARAMS0, val);

	val = (user_timings->tRR |
	       (user_timings->tWB <<
		REG_SDR_USER_TIMING_PARAMS1_FLD_WB_TIME_SHIFT) |
	       (user_timings->tRDDLY <<
		REG_SDR_USER_TIMING_PARAMS1_FLD_RD_DLY_SHIFT) |
	       ((user_timings->tWB >>
		 REG_SDR_USER_TIMING_PARAMS1_FLD_WB_TIME_SHIFT) <<
		NAND_SDR_EXTN_WB_TIME_SHIFT));
	img_nand_write_reg(host, REG_SDR_USER_TIMING_PARAMS1, val);
}

static void img_nand_configure_ddr_user_timings(struct img_nand_host *host)
{
	const struct img_nand_ddr_timings *user_timings = &user_ddr_timings;

	u32 val = user_timings->tSETUP |
		(user_timings->tHOLD <<
		 REG_DDR_USER_TIMING_PARAMS0_FLD_HOLD_TIME_SHIFT) |
		(user_timings->tWH <<
		 REG_DDR_USER_TIMING_PARAMS0_FLD_WH_TIME_SHIFT) |
		(user_timings->tWRP <<
		 REG_DDR_USER_TIMING_PARAMS0_FLD_WRP_TIME_SHIFT) |
		(user_timings->tINTCMD <<
		 REG_DDR_USER_TIMING_PARAMS0_FLD_INTCMD_TIME_SHIFT);
	img_nand_write_reg(host, REG_DDR_USER_TIMING_PARAMS0, val);

	val = user_timings->tRR |
		(user_timings->tWB <<
		REG_DDR_USER_TIMING_PARAMS1_FLD_WB_TIME_SHIFT) |
	       (user_timings->tCLKDiv <<
		REG_DDR_USER_TIMING_PARAMS1_FLD_CLKDIV_SHIFT) |
	       (user_timings->tCE <<
		REG_DDR_USER_TIMING_PARAMS1_FLD_CE_TIME_SHIFT) |
	       (user_timings->tDQSin <<
		REG_DDR_USER_TIMING_PARAMS1_FLD_DQS_TIME_SHIFT);
	img_nand_write_reg(host, REG_DDR_USER_TIMING_PARAMS1, val);

	val = user_timings->tDQSout;
	img_nand_write_reg(host, REG_DDR_USER_TIMING_PARAMS2, val);

}

static void img_nand_configure_ddr2_user_timings(struct img_nand_host *host)
{
	const struct img_nand_ddr_timings *user_timings = &user_ddr2_timings;

	u32 val = user_timings->tSETUP |
		(user_timings->tHOLD <<
		 REG_DDR2_USER_TIMING_PARAMS0_FLD_HOLD_TIME_SHIFT) |
		(user_timings->tWH <<
		 REG_DDR2_USER_TIMING_PARAMS0_FLD_WH_TIME_SHIFT) |
		(user_timings->tWRP <<
		 REG_DDR2_USER_TIMING_PARAMS0_FLD_WRP_TIME_SHIFT) |
		(user_timings->tINTCMD <<
		 REG_DDR2_USER_TIMING_PARAMS0_FLD_INTCMD_TIME_SHIFT);
	img_nand_write_reg(host, REG_DDR2_USER_TIMING_PARAMS0, val);

	val = user_timings->tRR |
		(user_timings->tWB <<
		REG_DDR2_USER_TIMING_PARAMS1_FLD_WB_TIME_SHIFT) |
	       (user_timings->tCLKDiv <<
		REG_DDR2_USER_TIMING_PARAMS1_FLD_CLKDIV_SHIFT) |
	       (user_timings->tCE <<
		REG_DDR2_USER_TIMING_PARAMS1_FLD_CE_TIME_SHIFT) |
	       (user_timings->tDQSin <<
		REG_DDR2_USER_TIMING_PARAMS1_FLD_DQS_TIME_SHIFT);
	img_nand_write_reg(host, REG_DDR2_USER_TIMING_PARAMS1, val);

	val = user_timings->tDQSout;
	img_nand_write_reg(host, REG_DDR2_USER_TIMING_PARAMS2, val);
}

static int img_nand_configure_timing_params(struct img_nand_host *host)
{
	u32 ret;
	u32 tmode_shift = 0;
	struct mtd_info *mtd = &host->mtd;
	u8 feature[ONFI_SUBFEATURE_PARAM_LEN] = {};

	/* Choose the slowest mode and the slowest timing mode. */
	img_nand_write_reg(host, REG_MODE_SELECT, 0);

	feature[0] = host->tmode | (host->mode << 4);
	feature[1] = feature[2] = feature[3] = 0;
	ret = host->chip.onfi_set_features(&host->mtd, &host->chip,
					   ONFI_FEATURE_ADDR_TIMING_MODE,
					   feature);

	if (ret == -EINVAL) {
		/* Some devices don't support set/get feature commands.
		 *
		 * Set default timing mode. */
		host->mode = IMG_NAND_SDR_MODE;
		host->tmode = IMG_NAND_TMODE_0;
		dev_info(&mtd->dev,
			 "Set features command not supported.\n");
	} else if (ret < 0) {
		dev_err(&mtd->dev,
			"Failed to set mode %d and tmode %d.\n",
			host->mode, host->tmode);
		goto err_host_mode;
	}

	switch (host->mode) {
	case IMG_NAND_SDR_MODE:
		tmode_shift = REG_MODE_SELECT_FLD_SDR_TIMING_MODE_SHIFT;
		if (host->use_user_timings)
			img_nand_configure_sdr_user_timings(host);
		break;
	case IMG_NAND_DDR_MODE:
		tmode_shift = REG_MODE_SELECT_FLD_DDR_TIMING_MODE_SHIFT;
		if (host->use_user_timings)
			img_nand_configure_ddr_user_timings(host);
		break;
	case IMG_NAND_DDR2_MODE:
		tmode_shift = REG_MODE_SELECT_FLD_DDR2_TIMING_MODE_SHIFT;
		if (host->use_user_timings)
			img_nand_configure_ddr2_user_timings(host);
		break;
	default:
		dev_err(&mtd->dev, "Invalid host mode: %d\n", host->mode);
		ret = -EINVAL;
		goto err_host_mode;
	}

	/* Choose mode and timing mode. */
	img_nand_write_reg(host, REG_MODE_SELECT,
			   host->mode | host->tmode << tmode_shift);

	/* Check whether the timing feature has been set in the memory. */
	ret = host->chip.onfi_get_features(&host->mtd, &host->chip,
					   ONFI_FEATURE_ADDR_TIMING_MODE,
					   feature);
	if (ret == -EINVAL) {
		dev_info(&mtd->dev, "Get features command not supported.\n");
		ret = 0;
	} else if ((feature[0] & 0x3F) != (host->tmode | (host->mode << 4))) {
		dev_err(&mtd->dev, "Failed to set timing feature.\n");
		ret = -EIO;
		goto err_host_mode;
	}
err_host_mode:
	return ret;
}

static int img_nand_wait_interrupt(struct img_nand_host *host)
{
	struct mtd_info *mtd = &host->mtd;

	if (host->irq_initialized) {
		if (!wait_for_completion_timeout(&host->comp_nand,
				msecs_to_jiffies(IMG_NAND_TIMEOUT))) {
			dev_err(&mtd->dev,
				"Interrupt time out: mask = %#x, stat = %#x\n",
				img_nand_read_reg(host, REG_INT_EN),
				img_nand_read_reg(host, REG_INT_STATUS));
			return -ETIMEDOUT;
		}
	} else {
		/* IRQs not initialized yet. */
		unsigned req_irq = img_nand_read_reg(host, REG_INT_EN);

		while (!(img_nand_read_reg(host, REG_INT_STATUS) & req_irq))
			;

		/* Clear the requested IRQs. */
		img_nand_write_reg(host, REG_INT_STATUS, req_irq);
	}

	return 0;
}

static void wait_for_cmd_buf_not_full(struct img_nand_host *host)
{
	while (img_nand_read_reg(host, REG_INT_STATUS) &
	       REG_INT_EN_FLD_CMD_BUF_FULL) {
		img_nand_write_reg(host, REG_INT_STATUS,
				   REG_INT_EN_FLD_CMD_BUF_FULL);
		ndelay(10);
	}
}

static void put_to_cmd_fifo(struct img_nand_host *host, unsigned int val)
{
	wait_for_cmd_buf_not_full(host);

	if (host->curr_cmd == NAND_CMD_STATUS)
		img_nand_write_reg(host, REG_CMD_REG,
				   REG_CMD_REG_FLD_COMMAND_VAL_READ_STATUS);
	else
		img_nand_write_reg(host, REG_CMD_BUF, val);
}

static int img_nand_read_parameter_page(struct mtd_info *mtd, unsigned *buf,
					unsigned int len)
{
	int ret;
	unsigned int count = 0;
	struct nand_chip *chip = mtd->priv;
	struct img_nand_host *host = chip->priv;
	unsigned val = 4 | (STATUS_RD_CMD <<
			    REG_CMD_BUF_FLD_CMD_DATA_VAL_CMD_TYPE_SHIFT);
	img_nand_write_reg(host, REG_INT_EN, REG_INT_EN_FLD_WRRD_DONE_INT);

	while (len > 0) {
		img_nand_write_reg(host, REG_INT_STATUS,
				   REG_INT_EN_FLD_WRRD_DONE_INT);
		put_to_cmd_fifo(host, val);
		ret = img_nand_wait_interrupt(host);
		if (ret < 0)
			return ret;

		*(buf + count) = img_nand_read_reg(host, REG_FLASH_STATUS);
		count++;
		len -= 4;
	}
	return 0;
}

static int img_uncorrectable_error_handle(struct img_nand_host *host)
{
	u32 val;
	unsigned long tim, limit;

	/* Reset Timing engine */
	img_nand_write_reg(host, REG_RESET, REG_RESET_FLD_TIME_ENGINE_RESET);
	/* poll for timing engine reset complete */
	tim = 0;
	limit = (loops_per_jiffy * msecs_to_jiffies(IMG_NAND_TIMEOUT));
	do {
		cpu_relax();
		val = img_nand_read_reg(host, REG_RESET_STATUS) &
			REG_RESET_STATUS_FLD_TIMING_ENGINE_RESET_DONE;
	} while (!val && (tim++ < limit));

	if (tim > limit) {
		dev_err(&host->mtd.dev, "Failed to timing reset.\n");
		return -ETIMEDOUT;
	}

	/* Reset Data FIFO */
	if (host->mode == IMG_NAND_SDR_MODE)
		img_nand_write_reg(host, REG_RESET,
				   REG_RESET_FLD_DATA_FIFO_RESET);
	else
		img_nand_write_reg(host, REG_RESET,
				   REG_RESET_FLD_DDRRX_FIFO_RESET);
	return 0;
}

static int img_nand_data_read(struct mtd_info *mtd, u8 *buf, unsigned int len)
{
	int ret;
	u32 interrupt, i;
	unsigned int count = 0;
	struct nand_chip *chip = mtd->priv;
	struct img_nand_host *host = chip->priv;
	unsigned int bytes_to_read = host->oob_access ? len : host->cw_size;
	bool need_reset = false;

	interrupt = REG_INT_EN_FLD_DATA_BUF_RD_RDY_INT;

	if (host->ecc_data.ecc_opt && len == mtd->writesize) {
		host->uncorrectable_err = false;
		interrupt |= REG_INT_EN_FLD_UNCORR_ERR_INT;
	}

	while (len > 0) {
		img_nand_write_reg(host, REG_INT_EN, interrupt);
		ret = img_nand_wait_interrupt(host);
		if (ret < 0)
			return ret;

		if (host->ecc_data.ecc_opt && host->uncorrectable_err &&
		    !need_reset) {
			need_reset = true;
		}

		for (i = 0; i < bytes_to_read; i = i + 4) {
#ifdef EUROPA_BUB_WORKAROUND
			/* This is a workaround. buf_rdrdy doesn't guarantee
			 * a whole codeword can be read out from DATA
			 * buffer! */
			while (img_nand_read_reg(host, REG_INT_STATUS) &
			       REG_INT_EN_FLD_DATA_BUF_EMPTY)
				img_nand_write_reg(host, REG_INT_STATUS,
					REG_INT_EN_FLD_DATA_BUF_EMPTY);
#endif
			/* Read data from Device via Data Buffer reg */
			*(buf + count) = img_nand_read_reg(host, REG_DATA_BUF);
			count++;
		}
		len -= bytes_to_read;
		if (host->ecc_data.ecc_opt && !host->oob_access &&
		    !host->raw_access) {
			int stat = chip->ecc.correct(mtd, NULL, NULL, NULL);
			if (stat < 0)
				mtd->ecc_stats.failed++;
			else
				mtd->ecc_stats.corrected += stat;
		}
	}

	if (need_reset) {
		/* Invoke Uncorrectable error handler function */
		if (img_uncorrectable_error_handle(host) < 0) {
			dev_err(&mtd->dev,
			"Timeout: poll for timing engine reset done\n");
			return -EIO;
		}
	}

	return 0;
}

static int img_nand_data_write(struct img_nand_host *host, const u8* buf,
			       unsigned int len)
{
	int i, ret;
	unsigned int count = 0;
	unsigned cw_size = host->oob_access ? len : host->cw_size;

	while (len > 0) {
		/* Enable Buffer_Wr_Ready interrupt */
		img_nand_write_reg(host, REG_INT_EN,
				   REG_INT_EN_FLD_DATA_BUF_WR_RDY_INT);

		ret = img_nand_wait_interrupt(host);
		if (ret < 0)
			return ret;

		for (i = 0; i < cw_size; i = i + 4) {
			/* Write data to Device via Data Buffer reg */
			img_nand_write_reg(host, REG_DATA_BUF, *(buf + count));
			count++;
		}

#ifdef EUROPA_BUB_WORKAROUND
		/* This is a workaround. Wait for data buffer empty IRQ,
		 * otherwise data buffer overflow IRQ occurs.*/
		img_nand_write_reg(host, REG_INT_STATUS,
				   REG_INT_EN_FLD_DATA_BUF_EMPTY);

		while (!(img_nand_read_reg(host, REG_INT_STATUS) &
			 REG_INT_EN_FLD_DATA_BUF_EMPTY))
			;
#endif
		len -= cw_size;
	}

	return 0;
}

static void img_nand_configure_ecc(struct img_nand_host *host)
{
	unsigned val;

	if (!host->ecc_data.ecc_opt)
		img_nand_write_reg(host, REG_BCH_CTRL_REG0, NAND_ECC_NONE);
	else {
		if (host->ecc_data.ecc_algo == IMG_NAND_ECC_ALGO_BCH) {
			unsigned int bch_t, bch_n;

			bch_t = (host->ecc_data.err_corr /
				 IMG_MIN_ERR_CORR_BCH) - 1;
			bch_n = ((bch_t * 4) + 4) * IMG_BCH_ALGO_POLYNOMIAL;

			val = (host->ecc_data.ecc_opt |
			       (host->ecc_data.ecc_algo <<
				REG_BCH_CTRL_REG0_FLD_ECC_ALGO_SHIFT) |
			       (bch_t << REG_BCH_CTRL_REG0_BCH_T_SHIFT) |
			       (bch_n << REG_BCH_CTRL_REG0_BCH_N_SHIFT));

			img_nand_write_reg(host, REG_BCH_CTRL_REG1,
					   (host->bch_cw_size /
					    IMG_MIN_BCH_CW_SIZE) - 1);

		} else {
			val = host->ecc_data.ecc_opt |
				(host->ecc_data.ecc_algo <<
				 REG_BCH_CTRL_REG0_FLD_ECC_ALGO_SHIFT);
		}

		img_nand_write_reg(host, REG_BCH_CTRL_REG0, val);
	}
}

/* platform specific dev_ready function */
static int img_nand_dev_ready(struct mtd_info *mtd)
{
	struct nand_chip *chip = mtd->priv;
	struct img_nand_host *host = chip->priv;

	return img_nand_read_reg(host, REG_RDY_BUSY_STATUS) == 0xFF;
}

/* Nand Initialization function */
static int img_nand_setup(struct mtd_info *mtd)
{
	u32 val;
	struct nand_chip *chip = mtd->priv;
	struct img_nand_host *host = chip->priv;

	/* Soft reset */
	img_nand_write_reg(host, REG_RESET, REG_RESET_FLD_SOFT_RESET |
			   REG_RESET_FLD_CMD_FIFO_RESET);

	nand_wait_ready(mtd);

	/* Clear all interrupts in Interrupt status register */
	img_nand_write_reg(host, REG_INT_STATUS,
			   REG_INT_STATUS_VAL_CLR_ALL_NAND_INT);

	host->ctrl.col_addr_cycles = 2;
	if (chip->chipsize > (128 << 20))
		host->ctrl.row_addr_cycles = 3;
	else
		host->ctrl.row_addr_cycles = 2;

	/* Configure Control_Reg0 Params */
	val = host->ctrl.chip_sel |
		(host->ctrl.wp_ctrl << REG_CTRL_REG0_FLD_WR_PROTECT_SHIFT) |
		(host->ctrl.bus_width << REG_CTRL_REG0_FLD_BUS_WIDTH_SHIFT) |
		(host->ctrl.col_addr_cycles <<
		 REG_CTRL_REG0_FLD_COL_ADDR_CYC_SHIFT) |
		(host->ctrl.row_addr_cycles <<
		 REG_CTRL_REG0_FLD_ROW_ADDR_CYC_SHIFT) |
		(((mtd->writesize / IMG_MIN_NAND_PAGE_SUPPORT) >> 1) <<
		 REG_CTRL_REG0_FLD_PAGE_SIZE_SHIFT) |
		(host->ctrl.chip_sel << REG_CTRL_REG0_FLD_CHIP_SELECT_SHIFT);
	img_nand_write_reg(host, REG_CTRL_REG0, val);

	/* Configure ECC details */
	img_nand_configure_ecc(host);

	if (host->use_user_timings)
		host->tmode = IMG_NAND_TMODE_USER;
	else {
		if (host->mode == IMG_NAND_SDR_MODE)
			host->tmode = onfi_get_async_timing_mode(&host->chip);
		else
			host->tmode = onfi_get_sync_timing_mode(&host->chip);

		/* Convert ONFI timing mode to IMG timing mode. */
		if (host->tmode == ONFI_TIMING_MODE_UNKNOWN)
			/* Set default timing mode */
			host->tmode = IMG_NAND_TMODE_0;
		else
			host->tmode = fls(host->tmode) - 1;

		if (host->tmode < 0)
			host->tmode = IMG_NAND_TMODE_0;
	}

#ifdef EUROPA_BUB_WORKAROUND
	/* This is maximal speed Europa BUB can work with. */
	host->mode = IMG_NAND_SDR_MODE;
	host->tmode = IMG_NAND_TMODE_0;
#endif

	dev_info(&mtd->dev, "Configuring mode %d and timing mode %d...\n",
		 host->mode, host->tmode);
	return img_nand_configure_timing_params(host);
}

static void img_nand_command(struct mtd_info *mtd, unsigned int command,
			     int column, int page_addr)
{
	u32 val;
	register struct nand_chip *chip = mtd->priv;
	struct img_nand_host *host = chip->priv;

	host->prev_cmd = host->curr_cmd;
	host->curr_cmd = command;

	/* Emulate NAND_CMD_READOOB */
	if (command == NAND_CMD_READOOB) {
		host->oob_access = true;
		command = NAND_CMD_READ0;
	}

	if (command == NAND_CMD_READ0) {
		if (!host->cmd_post_proc_needed) {
			/* At this stage there is no information about raw
			 * access request in the function input parameters.
			 * Postpone this command until read_page function is
			 * called. */
			host->curr_column = column;
			host->curr_page_addr = page_addr;
			host->cmd_post_proc_needed = true;

			if (host->curr_cmd == NAND_CMD_READ0)
				return;
		}
		host->cmd_post_proc_needed = false;

		if (chip->ecc.mode == NAND_ECC_HW_OOB_FIRST &&
		    !host->raw_access) {
			if (host->oob_access)
				column = mtd->writesize;
			else
				/* Assign ECC data location */
				column = mtd->writesize +
					IMG_ECC_OFFSET_IN_OOB;
		}
	}

	if (command == NAND_CMD_READ0 || command == NAND_CMD_STATUS) {
		/* Set 'Rx_Mode' in Control_reg0' register */
		val = img_nand_read_reg(host, REG_CTRL_REG0);
		img_nand_write_reg(host, REG_CTRL_REG0, (val &
			~(REG_CTRL_REG0_FLD_TX_MODE_VAL_WRITE_TO_FLASH <<
			  REG_CTRL_REG0_FLD_TX_MODE_SHIFT)));
	} else if (command == NAND_CMD_SEQIN) {

		/* Set 'Tx_Mode' in Control_reg0' register */
		val = img_nand_read_reg(host, REG_CTRL_REG0);
		img_nand_write_reg(host, REG_CTRL_REG0, (val |
			(REG_CTRL_REG0_FLD_TX_MODE_VAL_WRITE_TO_FLASH <<
			 REG_CTRL_REG0_FLD_TX_MODE_SHIFT)));
	}

	/* Command latch cycle */
	chip->cmd_ctrl(mtd, command, NAND_NCE | NAND_CLE);

	if (column != -1 || page_addr != -1) {
		int ctrl = NAND_NCE | NAND_ALE;

		if (command == NAND_CMD_SET_FEATURES) {
			chip->cmd_ctrl(mtd, (column & IMG_NAND_BYTE_MASK),
				       ctrl);
			chip->cmd_ctrl(mtd, NAND_CMD_NONE, NAND_NCE);
			return;
		}

		if (command == NAND_CMD_PARAM ||
		    command == NAND_CMD_GET_FEATURES ||
		    command == NAND_CMD_READID) {
			chip->cmd_ctrl(mtd, column & IMG_NAND_BYTE_MASK, ctrl);
			return;
		}

		/* Serially input address */
		if (column != -1) {
			/* Adjust columns for 16 bit buswidth */
			if (chip->options & NAND_BUSWIDTH_16)
				column >>= 1;
			chip->cmd_ctrl(mtd, column & IMG_NAND_BYTE_MASK, ctrl);
			chip->cmd_ctrl(mtd, (column >> IMG_NAND_BYTE_SHIFT) &
				       IMG_NAND_BYTE_MASK, ctrl);
		}
		if (page_addr != -1) {
			chip->cmd_ctrl(mtd, page_addr & IMG_NAND_BYTE_MASK,
				       ctrl);
			chip->cmd_ctrl(mtd,
				(page_addr >> IMG_NAND_BYTE_SHIFT) &
				IMG_NAND_BYTE_MASK, NAND_NCE | NAND_ALE);
			/* One more address cycle for devices > 128MiB */
			if (chip->chipsize > (128 << 20))
				chip->cmd_ctrl(mtd, (page_addr >>
						     IMG_NAND_WORD_SHIFT) &
					       IMG_NAND_BYTE_MASK,
					       NAND_NCE | NAND_ALE);
		}
	}

	chip->cmd_ctrl(mtd, NAND_CMD_NONE, NAND_NCE);

	/*
	 * Erase have its own busy handlers status, sequential
	 * in, and deplete1 need no delay.
	 */
	switch (command) {
	case NAND_CMD_ERASE1:
	case NAND_CMD_ERASE2:
		return;

	case NAND_CMD_RNDIN:
		/* send wait for cycle count cmd to cmd buffer */
		val = (IMG_NAND_PAGE_PROG_WAIT_CYCLE |
		       (WAIT_CYCLE_CNT <<
			REG_CMD_BUF_FLD_CMD_DATA_VAL_CMD_TYPE_SHIFT));
		put_to_cmd_fifo(host, val);

		if (!host->oob_access) {
			val = (chip->ecc.layout->eccbytes &
			       IMG_NAND_BYTE_MASK) |
				(SPARE_WR_TYPE <<
				 REG_CMD_BUF_FLD_CMD_DATA_VAL_CMD_TYPE_SHIFT);
			put_to_cmd_fifo(host, val);

			/* send spare count write byte 2 */
			val = ((chip->ecc.layout->eccbytes >>
				IMG_NAND_BYTE_SHIFT) & IMG_NAND_BYTE_MASK) |
				(SPARE_WR_TYPE <<
				 REG_CMD_BUF_FLD_CMD_DATA_VAL_CMD_TYPE_SHIFT);
			put_to_cmd_fifo(host, val);
		}
		return;

	case NAND_CMD_RESET:
		if (chip->dev_ready)
			break;
		udelay(chip->chip_delay);
		chip->cmd_ctrl(mtd, NAND_CMD_STATUS, NAND_NCE | NAND_CLE);
		chip->cmd_ctrl(mtd, NAND_CMD_NONE, NAND_NCE);
		while (!(chip->read_byte(mtd) & NAND_STATUS_READY))
			;
		return;

	case NAND_CMD_RNDOUT:
		/* No ready / busy check necessary */
		chip->cmd_ctrl(mtd, NAND_CMD_RNDOUTSTART, NAND_NCE | NAND_CLE);
		chip->cmd_ctrl(mtd, NAND_CMD_NONE, NAND_NCE);
		return;

	case NAND_CMD_SEQIN:
		if (column >= mtd->writesize)
			host->oob_access = true;

		/* send no operation cmd */
		val = IMG_NAND_NO_OP_CMD_VAL |
			(NO_OPERATION <<
			 REG_CMD_BUF_FLD_CMD_DATA_VAL_CMD_TYPE_SHIFT);
		put_to_cmd_fifo(host, val);

		/* send wait for cycle count cmd to cmd buffer */
		val = (IMG_NAND_PAGE_PROG_WAIT_CYCLE |
		       (WAIT_CYCLE_CNT <<
			REG_CMD_BUF_FLD_CMD_DATA_VAL_CMD_TYPE_SHIFT));
		put_to_cmd_fifo(host, val);
		return;

	case NAND_CMD_CACHEDPROG:
	case NAND_CMD_PAGEPROG:
		/* send wait for ready cmd to cmd buffer */
		val = IMG_NAND_WAIT_FOR_RDY_CMD_VAL |
			(WAIT_FOR_READY <<
			 REG_CMD_BUF_FLD_CMD_DATA_VAL_CMD_TYPE_SHIFT);
		put_to_cmd_fifo(host, val);
		nand_wait_ready(mtd);
		return;

	case NAND_CMD_READ0:
		chip->cmd_ctrl(mtd, NAND_CMD_READSTART, NAND_NCE | NAND_CLE);
		if ((chip->ecc.mode == NAND_ECC_HW_OOB_FIRST) &&
		    !host->oob_access && !host->raw_access) {
			/* send spare count read byte 1 */
			u32 val = (chip->ecc.layout->eccbytes &
			       IMG_NAND_BYTE_MASK) |
				(SPARE_RD_TYPE <<
				 REG_CMD_BUF_FLD_CMD_DATA_VAL_CMD_TYPE_SHIFT);
			put_to_cmd_fifo(host, val);

			/* send spare count read byte 2 */
			val = ((chip->ecc.layout->eccbytes >>
				IMG_NAND_BYTE_SHIFT) & IMG_NAND_BYTE_MASK) |
				(SPARE_RD_TYPE <<
				 REG_CMD_BUF_FLD_CMD_DATA_VAL_CMD_TYPE_SHIFT);
			put_to_cmd_fifo(host, val);

			/* send change read column cmd-1 */
			val = NAND_CMD_RNDOUT |
				(NORMAL_COMMAND <<
				 REG_CMD_BUF_FLD_CMD_DATA_VAL_CMD_TYPE_SHIFT);
			put_to_cmd_fifo(host, val);

			/* send column addr */
			val = (host->curr_column & IMG_NAND_BYTE_MASK) |
				(ADDR_TYPE <<
				 REG_CMD_BUF_FLD_CMD_DATA_VAL_CMD_TYPE_SHIFT);
			put_to_cmd_fifo(host, val);

			val = ((host->curr_column >> IMG_NAND_BYTE_SHIFT) &
			       IMG_NAND_BYTE_MASK) |
				(ADDR_TYPE <<
				 REG_CMD_BUF_FLD_CMD_DATA_VAL_CMD_TYPE_SHIFT);
			put_to_cmd_fifo(host, val);

			/* send change read column cmd-2 */
			val = NAND_CMD_RNDOUTSTART |
				(NORMAL_COMMAND <<
				 REG_CMD_BUF_FLD_CMD_DATA_VAL_CMD_TYPE_SHIFT);
			put_to_cmd_fifo(host, val);
		}
		break;

	default:
		chip->cmd_ctrl(mtd, NAND_CMD_NONE, NAND_NCE);
		break;
	}
	/*
	 * If we don't have access to the busy pin, we apply the given
	 * command delay.
	 */
	if (!chip->dev_ready) {
		udelay(chip->chip_delay);
		return;
	}

	nand_wait_ready(mtd);
}

static void img_mark_page_written(struct mtd_info *mtd)
{
	register struct nand_chip *chip = mtd->priv;
	struct img_nand_host *host = chip->priv;
	const u32 page_magic_no = PAGE_WRITTEN_MAGIC_NUMBER;

	host->oob_access = true;
	chip->cmdfunc(mtd, NAND_CMD_RNDIN, mtd->writesize + mtd->oobsize -
		      PAGE_WRITTEN_MAGIC_NUMBER_SIZE, -1);
	chip->write_buf(mtd, (const u8 *)&page_magic_no,
			sizeof(page_magic_no));
}

static int img_check_page_written(struct mtd_info *mtd, bool *page_written)
{
	register struct nand_chip *chip = mtd->priv;
	struct img_nand_host *host = chip->priv;
	u32 page_magic_no, val;

	*page_written = false;

	nand_wait_ready(mtd);

	/* Send change read column cmd-1 */
	val = NAND_CMD_RNDOUT | (NORMAL_COMMAND <<
				 REG_CMD_BUF_FLD_CMD_DATA_VAL_CMD_TYPE_SHIFT);
	put_to_cmd_fifo(host, val);

	/* Send column addr */
	val = ((mtd->writesize + mtd->oobsize -
		PAGE_WRITTEN_MAGIC_NUMBER_SIZE) & IMG_NAND_BYTE_MASK) |
		(ADDR_TYPE << REG_CMD_BUF_FLD_CMD_DATA_VAL_CMD_TYPE_SHIFT);
	put_to_cmd_fifo(host, val);

	val = (((mtd->writesize + mtd->oobsize -
		 PAGE_WRITTEN_MAGIC_NUMBER_SIZE) >> IMG_NAND_BYTE_SHIFT) &
	       IMG_NAND_BYTE_MASK) |
		(ADDR_TYPE << REG_CMD_BUF_FLD_CMD_DATA_VAL_CMD_TYPE_SHIFT);
	put_to_cmd_fifo(host, val);

	/* Send change read column cmd-2 */
	val = NAND_CMD_RNDOUTSTART | (NORMAL_COMMAND <<
		 REG_CMD_BUF_FLD_CMD_DATA_VAL_CMD_TYPE_SHIFT);
	put_to_cmd_fifo(host, val);

	host->oob_access = true;
	chip->read_buf(mtd, (u8 *)&page_magic_no, sizeof(page_magic_no));

	if (page_magic_no == PAGE_WRITTEN_MAGIC_NUMBER)
		*page_written = true;

	return 0;
}

/* hardware specific access to control-lines */
static void img_nand_cmd_ctrl(struct mtd_info *mtd, int cmd, unsigned int ctrl)
{
	u32 val;
	struct nand_chip *chip = mtd->priv;
	struct img_nand_host *host = chip->priv;

	if (cmd != NAND_CMD_NONE) {
		if (ctrl & NAND_CLE) {
			/* write the received command into command buffer */
			val = (cmd & IMG_NAND_BYTE_MASK) |
			      (NORMAL_COMMAND <<
			       REG_CMD_BUF_FLD_CMD_DATA_VAL_CMD_TYPE_SHIFT);
			put_to_cmd_fifo(host, val);
		} else if (ctrl & NAND_ALE) {
			/* write the received address byte into cmd buffer */
			val = (cmd & IMG_NAND_BYTE_MASK) |
				(ADDR_TYPE <<
				 REG_CMD_BUF_FLD_CMD_DATA_VAL_CMD_TYPE_SHIFT);
			put_to_cmd_fifo(host, val);
		}
	}
}

/*
 * Trigger NAND DMA operation
 */
static int img_nand_start_dma(struct img_nand_host *host)
{
	u32 val;

	if (!host->dma.dma_en) {
		host->dma.dma_en = true;
		val = img_nand_read_reg(host, REG_DMA_CTRL);
		img_nand_write_reg(host, REG_DMA_CTRL,
				   val | REG_DMA_CTRL_FLD_DMA_EN);
	}

	/* Enable both DMA transaction done & DMA boundary cross interrupts. */
	img_nand_write_reg(host, REG_INT_EN,
			   REG_INT_EN_FLD_DMA_BOUNDARY_CROSS_INT |
			   REG_INT_EN_FLD_DMA_TRANS_DONE |
			   REG_INT_EN_FLD_SLAVE_ERROR);

	if (img_nand_read_reg(host, REG_DMA_ADDR) != host->dma.dma_update_addr)
		img_nand_write_reg(host, REG_DMA_ADDR,
				   host->dma.dma_update_addr);

	if (img_nand_wait_interrupt(host) < 0)
		return -EIO;

	return 0;
}

static int img_dma_transfer(struct mtd_info *mtd, const void *buf, int len,
			    enum dma_transfer_direction dir)
{
	u32 val;
	u32 dma_phys_addr = 0;
	int ret = 0;
	struct nand_chip *chip = mtd->priv;
	struct img_nand_host *host = chip->priv;

	host->bytes_xfered = 0;
	host->dma.dma_en = false;
	host->dma.dma_done = false;

	if (buf == host->dma.dma_buf_virt)
		/* Internal buffer. */
		dma_phys_addr = host->dma.dma_buf_phys;
	else {
		dma_phys_addr = dma_map_single(&mtd->dev, (void *)buf, len,
					       DMA_BIDIRECTIONAL);
		ret = dma_mapping_error(&mtd->dev, dma_phys_addr);
		if (ret) {
			dev_err(&mtd->dev, "DMA mapping error\n");
			goto disable_dma;
		}
	}

	host->dma.dma_start_addr = dma_phys_addr;
	host->dma.dma_update_addr = host->dma.dma_start_addr;

	/* Set DMA direction */
	val = img_nand_read_reg(host, REG_DMA_CTRL) &
		~(IMG_NAND_DMA_WRITE << REG_DMA_CTRL_FLD_DATA_DIR_SHIFT);
	img_nand_write_reg(host, REG_DMA_CTRL, val |
			   (dir == DMA_MEM_TO_DEV ? IMG_NAND_DMA_READ :
			    IMG_NAND_DMA_WRITE <<
			    REG_DMA_CTRL_FLD_DATA_DIR_SHIFT));

	host->uncorrectable_err = false;

	/* Start DMA */
	while (host->bytes_xfered != mtd->writesize) {
		ret = img_nand_start_dma(host);
		if (ret < 0) {
			dev_err(&mtd->dev, "DMA data transfer failed\n");
			goto disable_dma;
		}
		if (host->dma.dma_done)
			break;
	}

	/* Wait for dma_trans_done interrupt. */
	if (!host->dma.dma_done) {
		img_nand_write_reg(host, REG_INT_EN,
				   REG_INT_EN_FLD_DMA_TRANS_DONE);

		ret = img_nand_wait_interrupt(host);
		if (ret < 0) {
			dev_err(&mtd->dev, "DMA transaction not done\n");
			return -EIO;
		}
	}

disable_dma:
	if (buf != host->dma.dma_buf_virt)
		dma_unmap_single(&mtd->dev, dma_phys_addr, len,
				 DMA_BIDIRECTIONAL);
	/* Disable DMA */
	img_nand_write_reg(host, REG_DMA_CTRL, 0);
	img_nand_write_reg(host, REG_DMA_ADDR, 0);

	host->dma.dma_done = false;
	host->dma.dma_en = false;
	host->bytes_xfered = 0;

	return ret;
}

static void img_nand_dma_setup(struct img_nand_host *host)
{
	img_nand_write_reg(host, REG_DMA_CTRL,
			   ((host->dma.dma_buf_size &
			     REG_DMA_CTRL_FLD_BUF_SIZE_MASK) <<
			    REG_DMA_CTRL_FLD_BUF_SIZE_SHIFT));
}

static u8 img_read_byte(struct mtd_info *mtd)
{
	int ret;
	u8 d = 0;
	int rd_byte_cnt;
	u32 val, mask = 0;
	struct nand_chip *chip = mtd->priv;
	struct img_nand_host *host = chip->priv;

	/* Before accessing OOB area, make sure ecc is in off condition */
	if (host->oob_access) {
		rd_byte_cnt = IMG_NAND_MIN_DATA_BUF_RD_CNT;
		mask = img_nand_read_reg(host, REG_BCH_CTRL_REG0);
		if (mask & REG_BCH_CTRL_REG0_FLD_ECC_ON_OFF_MASK)
			img_nand_write_reg(host, REG_BCH_CTRL_REG0,
						(mask & ~IMG_NAND_ECC_ON));

		/* send lower-byte of code word size */
		val = (rd_byte_cnt & IMG_NAND_BYTE_MASK) |
			(RD_BYTE_CNT_TYPE <<
			 REG_CMD_BUF_FLD_CMD_DATA_VAL_CMD_TYPE_SHIFT);
		put_to_cmd_fifo(host, val);

		/* send higher-byte of code word size */
		val = ((rd_byte_cnt >> IMG_NAND_BYTE_SHIFT) &
		       IMG_NAND_BYTE_MASK) |
			(RD_BYTE_CNT_TYPE <<
			 REG_CMD_BUF_FLD_CMD_DATA_VAL_CMD_TYPE_SHIFT);
		put_to_cmd_fifo(host, val);

		/* Enable Buffer_Rd_Ready interrupt and wait for that */
		img_nand_write_reg(host, REG_INT_EN,
				   REG_INT_EN_FLD_DATA_BUF_RD_RDY_INT);
		ret = img_nand_wait_interrupt(host);
		if (ret < 0) {
			if (host->oob_access) {
				img_nand_write_reg(host, REG_BCH_CTRL_REG0,
						   mask);
				host->oob_access = false;
			}
			return -EIO;
		}

		d = img_nand_read_reg(host, REG_DATA_BUF) & IMG_NAND_BYTE_MASK;
	} else if (host->curr_cmd == NAND_CMD_STATUS) {
		img_nand_write_reg(host, REG_INT_EN,
				   REG_INT_EN_FLD_WRRD_DONE_INT);
		ret = img_nand_wait_interrupt(host);
		if (ret < 0) {
			dev_err(&mtd->dev,
				"Failed to wait for wrrd_done irq.\n");
			return ret;
		}

		d = img_nand_read_reg(host, REG_FLASH_STATUS) &
			IMG_NAND_BYTE_MASK;
	} else if (host->curr_cmd == NAND_CMD_GET_FEATURES) {
		rd_byte_cnt = IMG_NAND_MIN_STATUS_RD_CNT;
		val = (rd_byte_cnt & IMG_NAND_BYTE_MASK) |
			(STATUS_RD_CMD <<
			 REG_CMD_BUF_FLD_CMD_DATA_VAL_CMD_TYPE_SHIFT);
		put_to_cmd_fifo(host, val);

		nand_wait_ready(mtd);

		img_nand_write_reg(host, REG_INT_EN,
				   REG_INT_EN_FLD_WRRD_DONE_INT);
		ret = img_nand_wait_interrupt(host);
		if (ret < 0)
			return ret;

		d = img_nand_read_reg(host, REG_FLASH_STATUS);
	} else if (host->curr_cmd == NAND_CMD_READID ||
		   host->curr_cmd == NAND_CMD_PARAM) {
		val = IMG_NAND_MIN_STATUS_RD_CNT |
			(STATUS_RD_CMD <<
			 REG_CMD_BUF_FLD_CMD_DATA_VAL_CMD_TYPE_SHIFT);
		img_nand_write_reg(host, REG_INT_STATUS,
				   REG_INT_EN_FLD_WRRD_DONE_INT);
		img_nand_write_reg(host, REG_INT_EN,
				   REG_INT_EN_FLD_WRRD_DONE_INT);
		put_to_cmd_fifo(host, val);

		ret = img_nand_wait_interrupt(host);
		if (ret < 0)
			return ret;

		img_nand_write_reg(host, REG_INT_STATUS,
				   REG_INT_EN_FLD_WRRD_DONE_INT);
		d = img_nand_read_reg(host, REG_FLASH_STATUS) &
			IMG_NAND_BYTE_MASK;
	}

	/* restore ecc condition */
	if (host->oob_access || host->curr_cmd == NAND_CMD_READOOB) {
		img_nand_write_reg(host, REG_BCH_CTRL_REG0, mask);

		if (host->oob_access)
			host->oob_access = false;
	}

	return d;
}

static u16 img_read_word(struct mtd_info *mtd)
{
	u16 d;
	int ret;
	int rd_byte_cnt;
	u32 val, mask = 0;
	struct nand_chip *chip = mtd->priv;
	struct img_nand_host *host = chip->priv;

	rd_byte_cnt = IMG_NAND_MIN_DATA_BUF_RD_CNT;

	/* Before accessing OOB area, make sure ecc is in off condition */
	mask = img_nand_read_reg(host, REG_BCH_CTRL_REG0);
	if (mask & REG_BCH_CTRL_REG0_FLD_ECC_ON_OFF_MASK)
		img_nand_write_reg(host, REG_BCH_CTRL_REG0,
						(mask & ~IMG_NAND_ECC_ON));
	/* send lower-byte of code word size */
	val = (rd_byte_cnt & IMG_NAND_BYTE_MASK) |
		(RD_BYTE_CNT_TYPE <<
		 REG_CMD_BUF_FLD_CMD_DATA_VAL_CMD_TYPE_SHIFT);
	put_to_cmd_fifo(host, val);

	/* send higher-byte of code word size */
	val = ((rd_byte_cnt >> IMG_NAND_BYTE_SHIFT) & IMG_NAND_BYTE_MASK) |
		(RD_BYTE_CNT_TYPE <<
		 REG_CMD_BUF_FLD_CMD_DATA_VAL_CMD_TYPE_SHIFT);
	put_to_cmd_fifo(host, val);

	/* Enable Buffer_Rd_Ready interrupt and wait for that */
	img_nand_write_reg(host, REG_INT_EN,
			   REG_INT_EN_FLD_DATA_BUF_RD_RDY_INT);
	ret = img_nand_wait_interrupt(host);
	if (ret < 0) {
		img_nand_write_reg(host, REG_BCH_CTRL_REG0, mask);

		if (host->oob_access)
			host->oob_access = false;
		return -EIO;
	}

	d = img_nand_read_reg(host, REG_DATA_BUF) & IMG_NAND_WORD_MASK;

	/* restore ecc condition */
	img_nand_write_reg(host, REG_BCH_CTRL_REG0, mask);

	if (host->oob_access)
		host->oob_access = false;

	return d;
}

static void img_nand_read_buf(struct mtd_info *mtd, u8 *buf, int len)
{
	int ret, i;
	u32 val, mask = 0;
	bool dma_mode_status;
	int cw_cnt = 0, cw_size = 0;
	struct nand_chip *chip = mtd->priv;
	struct img_nand_host *host = chip->priv;
	bool disable_ecc = host->raw_access || host->oob_access ||
		len == mtd->oobsize;

	if (host->curr_cmd == NAND_CMD_PARAM ||
	    (host->curr_cmd == NAND_CMD_RNDOUT &&
	     !(host->prev_cmd == NAND_CMD_PARAM))) {
		cw_size = IMG_NAND_MIN_CW_SIZE;
		cw_cnt = len / cw_size;
	} else if (len == mtd->oobsize || host->oob_access) {
		cw_size = len;
		cw_cnt = IMG_NAND_MIN_CW_CNT;
	} else if (len == mtd->writesize) {
		cw_size = host->cw_size;
		cw_cnt = len / cw_size;
	}

	/* Disable ECC? */
	if (disable_ecc) {
		mask = img_nand_read_reg(host, REG_BCH_CTRL_REG0);
		if (mask & REG_BCH_CTRL_REG0_FLD_ECC_ON_OFF_MASK)
			img_nand_write_reg(host, REG_BCH_CTRL_REG0,
					   mask & ~IMG_NAND_ECC_ON);
	}

	img_nand_write_reg(host, REG_CODEWORD_SIZE_CNT, cw_size |
			   (cw_cnt << REG_CODEWORD_SIZE_CNT_FLD_CNT_SHIFT));

	for (i = 0; i < cw_cnt; i++) {
		if (host->curr_cmd == NAND_CMD_PARAM ||
		    host->curr_cmd == NAND_CMD_RNDOUT) {
			/* send DUT status read command */
			val = (cw_size & IMG_NAND_BYTE_MASK) |
				(STATUS_RD_CMD <<
				 REG_CMD_BUF_FLD_CMD_DATA_VAL_CMD_TYPE_SHIFT);
			put_to_cmd_fifo(host, val);
		} else {
			/* send lower-byte of code word size */
			val = (cw_size & IMG_NAND_BYTE_MASK) |
				(RD_BYTE_CNT_TYPE <<
				 REG_CMD_BUF_FLD_CMD_DATA_VAL_CMD_TYPE_SHIFT);
			put_to_cmd_fifo(host, val);

			/* send higher-byte of code word size */
			val = ((cw_size >> IMG_NAND_BYTE_SHIFT) &
			       IMG_NAND_BYTE_MASK) |
				(RD_BYTE_CNT_TYPE <<
				 REG_CMD_BUF_FLD_CMD_DATA_VAL_CMD_TYPE_SHIFT);
			put_to_cmd_fifo(host, val);
		}
	}

	if (host->dma.has_dma && len == mtd->writesize) {
		ret = img_dma_transfer(mtd, buf, len, DMA_DEV_TO_MEM);
		if (ret < 0) {
			if (host->dma.dma_en) {
				dev_err(&mtd->dev, "DMA Data read failed\n");
				goto restore_ecc_cond;
			} else
				dev_err(&mtd->dev, "Fall back to CPU I/O\n");
		} else {
			dma_mode_status = true;
			goto restore_ecc_cond;
		}
	}

	dma_mode_status = false;
	if (host->curr_cmd == NAND_CMD_RNDOUT) {
		/* Use a temporary buffer as data buffer has not been
		 * allocated yet. */
		unsigned tmp_buf[len];

		if (len % 4) {
			dev_err(&mtd->dev,
				"Incorrect data length(%d).\n", len);
			goto restore_ecc_cond;
		}

		ret = img_nand_read_parameter_page(mtd, tmp_buf, len);
		if (ret < 0) {
			dev_err(&mtd->dev, "Parameter page read failed\n");
			goto restore_ecc_cond;
		}
		memcpy(buf, tmp_buf, len);
		goto restore_ecc_cond;
	}

	dma_mode_status = false;
	ret = img_nand_data_read(mtd, buf, len);
	if (ret < 0) {
		dev_err(&mtd->dev, "Data read failed\n");
		return;
	}

restore_ecc_cond:
	/* restore ecc condition */
	if (disable_ecc) {
		img_nand_write_reg(host, REG_BCH_CTRL_REG0, mask);
		if (host->oob_access)
			host->oob_access = false;
	}
}

static void img_write_byte(struct mtd_info *mtd, u8 byte)
{
	u32 val;
	struct nand_chip *chip = mtd->priv;
	struct img_nand_host *host = chip->priv;

	if (host->curr_cmd == NAND_CMD_SET_FEATURES) {
		/* send direct data */
		val = (byte & IMG_NAND_BYTE_MASK) |
			(STATUS_WR_CMD <<
			 REG_CMD_BUF_FLD_CMD_DATA_VAL_CMD_TYPE_SHIFT);
		put_to_cmd_fifo(host, val);
		return;
	}

	/* write a byte via data_buffer_reg */
	img_nand_write_word(host, REG_DATA_BUF, byte);
}

static void img_write_byte16(struct mtd_info *mtd, u8 byte)
{
	u32 val;
	u16 word = byte;
	struct nand_chip *chip = mtd->priv;
	struct img_nand_host *host = chip->priv;

	if (host->curr_cmd == NAND_CMD_SET_FEATURES) {
		/* send direct data */
		val = (byte & IMG_NAND_BYTE_MASK) |
			(STATUS_WR_CMD <<
			 REG_CMD_BUF_FLD_CMD_DATA_VAL_CMD_TYPE_SHIFT);
		put_to_cmd_fifo(host, val);
		return;
	}

	/* write a byte via data_buffer_reg */
	img_nand_write_word(host, REG_DATA_BUF, word);
}

static void img_nand_write_buf(struct mtd_info *mtd, const u8 *buf,
				   int len)
{
	int ret, i;
	u32 val, mask = 0;
	int cw_cnt = 0, cw_size = 0;
	struct nand_chip *chip = mtd->priv;
	struct img_nand_host *host = chip->priv;

	/* Before accessing OOB area, make sure ecc is in off condition */
	if (host->oob_access || len == mtd->oobsize) {
		mask = img_nand_read_reg(host, REG_BCH_CTRL_REG0);
		if (mask & REG_BCH_CTRL_REG0_FLD_ECC_ON_OFF_MASK)
			img_nand_write_reg(host, REG_BCH_CTRL_REG0,
						(mask & ~IMG_NAND_ECC_ON));
		cw_cnt = IMG_NAND_MIN_CW_CNT;
		cw_size = len;

	} else if (len == mtd->writesize) {
		cw_size = host->cw_size;
		cw_cnt = len / cw_size;
	}

	img_nand_write_reg(host, REG_CODEWORD_SIZE_CNT, cw_size |
			   (cw_cnt << REG_CODEWORD_SIZE_CNT_FLD_CNT_SHIFT));

	for (i = 0; i < cw_cnt; i++) {
		/* send lower-byte of code word size */
		val = (cw_size & IMG_NAND_BYTE_MASK) |
			(WR_BYTE_CNT_TYPE <<
			 REG_CMD_BUF_FLD_CMD_DATA_VAL_CMD_TYPE_SHIFT);
		put_to_cmd_fifo(host, val);

		/* send higher-byte of code word size */
		val = ((cw_size >> IMG_NAND_BYTE_SHIFT) & IMG_NAND_BYTE_MASK) |
			(WR_BYTE_CNT_TYPE <<
			 REG_CMD_BUF_FLD_CMD_DATA_VAL_CMD_TYPE_SHIFT);
		put_to_cmd_fifo(host, val);
	}

	if (host->dma.has_dma && len == mtd->writesize) {
		ret = img_dma_transfer(mtd, buf, len, DMA_MEM_TO_DEV);
		if (ret < 0) {
			if (host->dma.dma_en) {
				dev_err(&mtd->dev, "DMA Data write failed\n");
				return;
			} else
				dev_err(&mtd->dev, "Fall back to CPU I/O\n");
		} else
			goto check_wrrd_done;
	}

	ret = img_nand_data_write(host, buf, len);
	if (ret < 0) {
		dev_err(&mtd->dev, "Data write failed\n");
		return;
	}

check_wrrd_done:
	/* Enable Write_read_done interrupt and wait for that */
	img_nand_write_reg(host, REG_INT_EN, REG_INT_EN_FLD_WRRD_DONE_INT);
	ret = img_nand_wait_interrupt(host);
	if (ret < 0) {
		dev_err(&mtd->dev, "Data write failed\n");
		return;
	}

	/* restore ecc condition */
	if (host->oob_access || len == mtd->oobsize) {
		img_nand_write_reg(host, REG_BCH_CTRL_REG0, mask);
		if (host->oob_access)
			host->oob_access = false;
	}
}

static void img_nand_find_new_dma_addr(struct img_nand_host *host)
{
	u32 val;
	unsigned int dma_buf_size;

	/* Get configured DMA buffer size from DMA Control register */
	val = (img_nand_read_reg(host, REG_DMA_CTRL) >>
	       REG_DMA_CTRL_FLD_BUF_SIZE_SHIFT) &
		REG_DMA_CTRL_FLD_BUF_SIZE_MASK;

	dma_buf_size = IMG_NAND_MIN_DMA_BUF_SIZE * BIT(val);

	host->bytes_xfered += dma_buf_size - (host->dma.dma_update_addr &
					      (dma_buf_size - 1));
	host->dma.dma_update_addr = host->dma.dma_start_addr +
		host->bytes_xfered;
}

static irqreturn_t img_nand_irq(int irq, struct img_nand_host *host)
{
	u32 data, mask;

	mask = img_nand_read_reg(host, REG_INT_EN);
	img_nand_write_reg(host, REG_INT_EN, REG_INT_EN_VAL_DISABLE_ALL_INTS);

	data = img_nand_read_reg(host, REG_INT_STATUS);
	data = data & mask;
	if (data & REG_INT_EN_FLD_UNCORR_ERR_INT)
		host->uncorrectable_err = true;

	if (!data || ((data & IMG_NAND_ALL_INT_SET) == IMG_NAND_ALL_INT_SET))
		return IRQ_NONE;

	if (data & REG_INT_EN_FLD_DMA_TRANS_DONE)
		host->dma.dma_done = true;

	if (data & REG_INT_EN_FLD_DMA_BOUNDARY_CROSS_INT)
		/* calculate new DMA address to be updated */
		img_nand_find_new_dma_addr(host);

	if (data & REG_INT_EN_FLD_SLAVE_ERROR)
		dev_err(&host->mtd.dev, "DMA slave error reported.\n");

	img_nand_write_reg(host, REG_INT_STATUS, data);
	complete(&host->comp_nand);

	return IRQ_HANDLED;
}

/* stubs for ECC functions not used by the NAND core */
static int img_ecc_calculate(struct mtd_info *mtd, const u8 *data,
			     u8 *ecc_code)
{
	dev_err(&mtd->dev, "img_ecc_calculate called unexpectedly\n");
	BUG();
	return -EIO;
}

static int img_ecc_correct(struct mtd_info *mtd, u8 *data, u8 *read_ecc,
			   u8 *calc_ecc)
{
	struct nand_chip *chip = mtd->priv;
	struct img_nand_host *host = chip->priv;

	u32 val = img_nand_read_reg(host, REG_BCH_STATUS) &
		REG_BCH_STATUS_FLD_ERR_CNT_MASK;
	if (host->uncorrectable_err) {
		host->uncorrectable_err = false;
		return -EIO;
	}

	return val;
}

/* enables the hardware ecc functionality */
static void img_ecc_enable_hwecc(struct mtd_info *mtd, int mode)
{
	struct nand_chip *chip = mtd->priv;
	struct img_nand_host *host = chip->priv;

	switch (mode) {
	case NAND_ECC_READ:
	case NAND_ECC_WRITE:
		/* Configure ECC details */
		img_nand_configure_ecc(host);
		break;
	default:
		dev_err(&mtd->dev, "error: unrecognised Mode[%d]!\n", mode);
		break;
	}
}

/**
 * wait until the command is done
 *
 * Wait function is called during Program and erase operations and
 * the way it is called from MTD layer, we should wait till the NAND
 * chip is ready after the programming/erase operation has completed.
 *
 * Erase can take up to 400ms and program up to 20ms according to
 * general NAND and SmartMedia specs
 */
static int img_nand_waitfunc(struct mtd_info *mtd, struct nand_chip *chip)
{
	unsigned long timeo = jiffies;
	int status, state = chip->state;
	struct img_nand_host *host = chip->priv;

	nand_wait_ready(mtd);

	/* Wait for current command to complete... */
	img_nand_write_reg(host, REG_INT_EN, REG_INT_EN_FLD_WRRD_DONE_INT);
	if (img_nand_wait_interrupt(host) < 0) {
		dev_err(&mtd->dev,
			"Failed to wait for curr_cmd to complete.\n");
		status = NAND_STATUS_FAIL;
		goto err_wrrd_done;
	}

	if (state == FL_ERASING)
		timeo += msecs_to_jiffies(400);
	else
		timeo += msecs_to_jiffies(20);

	/* Wait until the Nand chip is ready after the Page program/Block erase
	 * operations has completed, by sending Read_Status_Register command.
	 * As a result of Read status command, the status register value will
	 * reflect on 'Ready_Busy_staus' register.
	 */
	chip->cmd_ctrl(mtd, NAND_CMD_STATUS, NAND_NCE | NAND_CLE);
	chip->cmd_ctrl(mtd, NAND_CMD_NONE, NAND_NCE);

	img_nand_write_reg(host, REG_INT_EN, REG_INT_EN_FLD_WRRD_DONE_INT);
	if (img_nand_wait_interrupt(host) < 0) {
		dev_err(&mtd->dev, "Failed to wait for wrrd_done irq.\n");
		status = NAND_STATUS_FAIL;
		goto err_wrrd_done;
	}

	do {
		status = img_nand_read_reg(host, REG_FLASH_STATUS);
		if (status & FLASH_STATUS_RDY)
			break;
		cond_resched();
	} while (time_before(jiffies, timeo));

err_wrrd_done:
	return status;
}

static int img_ecc_read_page_raw(struct mtd_info *mtd, struct nand_chip *chip,
				 u8 *buf, int oob_required, int page)
{
	struct img_nand_host *host = chip->priv;

	host->raw_access = true;

	/* Command post processing is needed. */
	if (host->cmd_post_proc_needed)
		img_nand_command(mtd, host->curr_cmd, host->curr_column,
				 host->curr_page_addr);

	chip->read_buf(mtd, buf, mtd->writesize);
	host->raw_access = false;

	if (oob_required) {
		host->oob_access = true;
		chip->read_buf(mtd, chip->oob_poi, mtd->oobsize);
	}

	return 0;
}

static bool img_is_bbt(struct nand_chip *chip, int page)
{
	unsigned page_no;

	for (page_no = 0; page_no < NAND_MAX_CHIPS; page_no++) {
		if (chip->bbt_td->pages[page_no] == page ||
		    chip->bbt_md->pages[page_no] == page)
			return true;
	}

	return false;
}

static int img_ecc_read_page(struct mtd_info *mtd, struct nand_chip *chip,
			     u8 *buf, int oob_required, int page)
{
	u32 val;
	int ret, i;
	int cw_size, cw_cnt;
	unsigned int max_bitflips = 0;
	unsigned int data_len = mtd->writesize;
	struct img_nand_host *host = chip->priv;

	/* Clear all interrupts in Interrupt status register */
	img_nand_write_reg(host, REG_INT_STATUS,
			   REG_INT_STATUS_VAL_CLR_ALL_NAND_INT);

	/* Command post processing is needed. */
	if (host->cmd_post_proc_needed)
		img_nand_command(mtd, host->curr_cmd, host->curr_column,
				 host->curr_page_addr);

	cw_size = host->cw_size;
	cw_cnt = data_len / cw_size;

	img_nand_write_reg(host, REG_CODEWORD_SIZE_CNT, cw_size |
			   (cw_cnt << REG_CODEWORD_SIZE_CNT_FLD_CNT_SHIFT));

	for (i = 0; i < cw_cnt; i++) {
		/* send lower-byte of code word size */
		val = (cw_size & IMG_NAND_BYTE_MASK) |
			(RD_BYTE_CNT_TYPE <<
			 REG_CMD_BUF_FLD_CMD_DATA_VAL_CMD_TYPE_SHIFT);
		put_to_cmd_fifo(host, val);

		/* send higher-byte of code word size */
		val = ((cw_size >> IMG_NAND_BYTE_SHIFT) & IMG_NAND_BYTE_MASK) |
			(RD_BYTE_CNT_TYPE <<
			 REG_CMD_BUF_FLD_CMD_DATA_VAL_CMD_TYPE_SHIFT);
		put_to_cmd_fifo(host, val);
	}

	if (host->dma.has_dma) {
		ret = img_dma_transfer(mtd, buf, data_len, DMA_DEV_TO_MEM);
		if (ret < 0) {
			if (host->dma.dma_en) {
				dev_err(&mtd->dev, "DMA Data read failed\n");
				return -EIO;
			} else
				dev_err(&mtd->dev, "Fall back to CPU I/O\n");
		} else
			goto dma_buf_copy;
	}

	ret = img_nand_data_read(mtd, buf, data_len);
	if (ret < 0)
		return -EIO;

dma_buf_copy:
	/* Read accumulated 'corrected error count' value for current page */
	if (host->ecc_data.ecc_opt && !host->oob_access && !host->raw_access) {
		u32 bch_status = img_nand_read_reg(host, REG_BCH_STATUS);

		max_bitflips = (bch_status >>
				REG_BCH_STATUS_FLD_ACC_ERR_CNT_SHIFT) &
			REG_BCH_STATUS_FLD_ACC_ERR_CNT_MASK;

		mtd->ecc_stats.corrected += max_bitflips;

		if (host->uncorrectable_err) {
			bool page_written;

			host->uncorrectable_err = false;

			ret = img_check_page_written(mtd, &page_written);
			if (ret < 0) {
				dev_err(&mtd->dev,
					"Failed to check magic number\n.");
				return -EIO;
			}

			ret = img_uncorrectable_error_handle(host);
			if (ret < 0) {
				dev_err(&mtd->dev,
					"Failed to reset timing engine\n");
				return -EIO;
			}

			/* If the page has been written (not erased page),
			 * report the error.
			 * Don't report the error when BBT is being read since
			 * BBT area isn't protected by ECC. */
			if (page_written && !(oob_required ||
					      img_is_bbt(chip, page))) {
				dev_err(&mtd->dev,
					"Uncorrectable error at page no:%d.\n",
					page);
				mtd->ecc_stats.failed++;
			}
		}
	}

	if (oob_required) {
		u32 val = 0;
		nand_wait_ready(mtd);

		/* send change read column cmd-1 */
		val = NAND_CMD_RNDOUT |
			(NORMAL_COMMAND <<
			 REG_CMD_BUF_FLD_CMD_DATA_VAL_CMD_TYPE_SHIFT);
		put_to_cmd_fifo(host, val);

		/* send column addr */
		val = (mtd->writesize & IMG_NAND_BYTE_MASK) |
			(ADDR_TYPE <<
			 REG_CMD_BUF_FLD_CMD_DATA_VAL_CMD_TYPE_SHIFT);
		put_to_cmd_fifo(host, val);

		val = ((mtd->writesize >> IMG_NAND_BYTE_SHIFT) &
		       IMG_NAND_BYTE_MASK) |
			(ADDR_TYPE <<
			 REG_CMD_BUF_FLD_CMD_DATA_VAL_CMD_TYPE_SHIFT);
		put_to_cmd_fifo(host, val);

		/* send change read column cmd-2 */
		val = NAND_CMD_RNDOUTSTART |
			(NORMAL_COMMAND <<
			 REG_CMD_BUF_FLD_CMD_DATA_VAL_CMD_TYPE_SHIFT);
		put_to_cmd_fifo(host, val);

		host->oob_access = true;
		chip->read_buf(mtd, chip->oob_poi, mtd->oobsize);
	}

	img_nand_write_reg(host, REG_INT_EN, REG_INT_EN_FLD_WRRD_DONE_INT);
	ret = img_nand_wait_interrupt(host);
	if (ret < 0)
		return -EIO;

	return max_bitflips;
}

static int img_ecc_write_page_raw(struct mtd_info *mtd, struct nand_chip *chip,
				  const u8 *buf, int oob_required)
{
	dev_err(&mtd->dev, "Writing raw data not supported.\n");
	BUG();
	return -EIO;
}

static int img_ecc_write_page(struct mtd_info *mtd, struct nand_chip *chip,
			  const u8 *buf, int oob_required)
{
	u32 val;
	int ret, i;
	int cw_size, cw_cnt;
	unsigned int data_len = mtd->writesize;
	struct img_nand_host *host = chip->priv;

	/* Clear all interrupts in Interrupt status register */
	img_nand_write_reg(host, REG_INT_STATUS,
			   REG_INT_STATUS_VAL_CLR_ALL_NAND_INT);

	cw_size = host->cw_size;
	cw_cnt = data_len / cw_size;

	img_nand_write_reg(host, REG_CODEWORD_SIZE_CNT, cw_size |
			   (cw_cnt << REG_CODEWORD_SIZE_CNT_FLD_CNT_SHIFT));

	/* Set 'Tx_Mode' in Control_reg0' register */
	val = img_nand_read_reg(host, REG_CTRL_REG0);
	img_nand_write_reg(host, REG_CTRL_REG0, (val |
			(REG_CTRL_REG0_FLD_TX_MODE_VAL_WRITE_TO_FLASH <<
			 REG_CTRL_REG0_FLD_TX_MODE_SHIFT)));

	for (i = 0; i < cw_cnt; i++) {
		/* send lower-byte of code word size */
		val = (cw_size & IMG_NAND_BYTE_MASK) |
			(WR_BYTE_CNT_TYPE <<
			 REG_CMD_BUF_FLD_CMD_DATA_VAL_CMD_TYPE_SHIFT);
		put_to_cmd_fifo(host, val);

		/* send higher-byte of code word size */
		val = ((cw_size >> IMG_NAND_BYTE_SHIFT) & IMG_NAND_BYTE_MASK) |
			(WR_BYTE_CNT_TYPE <<
			 REG_CMD_BUF_FLD_CMD_DATA_VAL_CMD_TYPE_SHIFT);
		put_to_cmd_fifo(host, val);
	}

	if (!oob_required)
		chip->cmdfunc(mtd, NAND_CMD_RNDIN,
			      mtd->writesize + IMG_ECC_OFFSET_IN_OOB, -1);

	if (host->dma.has_dma) {
		ret = img_dma_transfer(mtd, buf, mtd->writesize,
				       DMA_MEM_TO_DEV);
		if (ret < 0) {
			if (host->dma.dma_en) {
				dev_err(&mtd->dev, "DMA Data Write failed\n");
				return -EIO;
			} else
				dev_err(&mtd->dev, "Fall back to CPU I/O\n");
		} else
			goto check_wrrd_done;

	}

	ret = img_nand_data_write(host, buf, mtd->writesize);
	if (ret < 0)
		return ret;

check_wrrd_done:
	if (oob_required) {
		host->oob_access = true;
		chip->cmdfunc(mtd, NAND_CMD_RNDIN, mtd->writesize, -1);
		chip->write_buf(mtd, chip->oob_poi, mtd->oobsize);
	}

	/* Enable Write_read_done interrupt and wait for that */
	img_nand_write_reg(host, REG_INT_EN, REG_INT_EN_FLD_WRRD_DONE_INT);
	ret = img_nand_wait_interrupt(host);
	if (ret < 0) {
		dev_err(&mtd->dev, "Data write failed\n");
		return ret;
	}

	return 0;
}

static int img_ecc_write_oob_raw(struct mtd_info *mtd, struct nand_chip *chip,
				 int page)
{
	dev_err(&mtd->dev, "Writing raw OOB data not supported.\n");
	BUG();
	return -EIO;
}

static int img_ecc_read_oob_raw(struct mtd_info *mtd, struct nand_chip *chip,
				int page)
{
	struct img_nand_host *host = chip->priv;

	host->raw_access = true;
	chip->cmdfunc(mtd, NAND_CMD_READOOB, 0, page);
	chip->read_buf(mtd, chip->oob_poi, mtd->oobsize);
	host->raw_access = false;

	return 0;
}

static int img_write_page(struct mtd_info *mtd, struct nand_chip *chip,
			  uint32_t offset, int data_len, const uint8_t *buf,
			  int oob_required, int page, int cached, int raw)
{
	int status, subpage;

	if (!(chip->options & NAND_NO_SUBPAGE_WRITE) &&
		chip->ecc.write_subpage)
		subpage = offset || (data_len < mtd->writesize);
	else
		subpage = 0;

	chip->cmdfunc(mtd, NAND_CMD_SEQIN, 0x00, page);

	if (unlikely(raw))
		status = chip->ecc.write_page_raw(mtd, chip, buf,
						  oob_required);
	else if (subpage)
		status = chip->ecc.write_subpage(mtd, chip, offset, data_len,
						 buf, oob_required);
	else
		status = chip->ecc.write_page(mtd, chip, buf, oob_required);

	if (status < 0)
		return status;

	/*
	 * Cached programming disabled for now. Not sure if it's worth the
	 * trouble. The speed gain is not very impressive. (2.3->2.6Mib/s).
	 */
	cached = 0;

	img_mark_page_written(mtd);
	if (!cached || !NAND_HAS_CACHEPROG(chip)) {
		chip->cmdfunc(mtd, NAND_CMD_PAGEPROG, -1, -1);
		status = chip->waitfunc(mtd, chip);

		/*
		 * See if operation failed and additional status checks are
		 * available.
		 */
		if ((status & NAND_STATUS_FAIL) && (chip->errstat))
			status = chip->errstat(mtd, chip, FL_WRITING, status,
					       page);

		if (status & NAND_STATUS_FAIL)
			return -EIO;
	} else {
		chip->cmdfunc(mtd, NAND_CMD_CACHEDPROG, -1, -1);
		status = chip->waitfunc(mtd, chip);
	}

	return 0;
}

static void img_nand_parse_ecc_info(struct device *dev,
				    struct img_nand_host *host)
{
	int ret;
	unsigned int strength, blk_size;
	struct device_node *node = dev->of_node;

	ret = of_property_read_u32(node, "nand-ecc-strength", &strength);
	if (ret != 0)
		host->ecc_data.err_corr = 0;
	else
		host->ecc_data.err_corr = strength;
	ret = of_property_read_u32(node, "nand-ecc-step-size", &blk_size);
	if (ret != 0)
		host->bch_cw_size = 0;
	else
		host->bch_cw_size = blk_size;
}

static int img_nand_parse_dt(struct device *dev, struct img_nand_host *host)
{
	unsigned int val;
	int ecc_mode, nsels, ret;
	struct mtd_info *mtd = &host->mtd;
	struct device_node *node = dev->of_node;

	if (!of_get_property(node, "nand-chip-sel", &nsels))
		return -EINVAL;

	ret = of_property_read_u32_index(node, "nand-chip-sel", 0, &val);
	if (ret) {
		dev_err(&mtd->dev,
			"can't retrieve nand-chip-sel property: %d\n", ret);
		return ret;
	}

	if (val > IMG_MAX_NAND_CHIPS) {
		dev_err(&mtd->dev,
			"invalid nand-chip-sel: %u (max. chips = 7)\n", val);
		return -EINVAL;
	}

	host->ctrl.chip_sel = val;

	if (of_get_nand_bus_width(node) == IMG_NAND_BUS_WIDTH_16)
		host->ctrl.bus_width = 1;
	else
		host->ctrl.bus_width = 0;

	ecc_mode = of_get_nand_ecc_mode(node);
	if (ecc_mode > 0) {
		host->ecc_data.ecc_opt = IMG_NAND_ECC_ON;
		/* invoke an API to parse ecc_info */
		img_nand_parse_ecc_info(dev, host);
	} else
		host->ecc_data.ecc_opt = NAND_ECC_NONE;

	/* If ECC is on, this will be overridden at find_best_ecc_mode() */
	host->cw_size = 512;
	if (of_property_read_u32(node, "nand-codeword-size", &val) == 0) {
		if (val < 0) {
			dev_err(&mtd->dev, "invalid nand-codeword-size: %d\n",
				val);
			return -EINVAL;
		}
		host->cw_size = val;
	}

	host->ctrl.wp_ctrl = of_property_read_bool(node, "nand-wp-ctrl");
	host->flash_bbt = of_get_nand_on_flash_bbt(node);

	if (of_property_read_u32(node, "nand-data-rate-mode", &val) == 0) {
		if ((val > 2) || (val < 0)) {
			dev_err(&mtd->dev,
				"invalid data-rate-mode selection: %d\n", val);
			return -EINVAL;
		}
		host->mode = (unsigned char)val;
	}

	host->use_user_timings = of_property_read_bool(node,
						       "nand-user-timings");

	host->ctrl.dev_ready = of_property_read_bool(node, "nand-dev-ready");
	host->dma.has_dma = of_property_read_bool(node, "nand-has-dma");
	if (host->dma.has_dma) {
		if (of_property_read_u32(node, "nand-dma-buf-size",
					 &val) == 0) {
			if ((val > 7) || (val < 0)) {
				dev_err(&mtd->dev,
					"invalid nand-dma-buf-size: %d\n",
					val);
				return -EINVAL;
			}
			host->dma.dma_buf_size = (unsigned char)val;
		}
	}
	host->bch_byte_align_8bitmode =
		of_property_read_bool(node, "nand-bch-byte-align-8bitmode");

	return 0;
}

/*
 * It's better to use the ECC mode recommended by chip than to always
 * have to set one in DTS. DTS should be an override option.
 */
static int find_best_ecc_mode(struct img_nand_host *host)
{
	struct nand_chip *chip = &host->chip;
	struct mtd_info *mtd = &host->mtd;

	if (host->ecc_data.err_corr == 0 || host->bch_cw_size == 0) {
		if (chip->ecc_strength_ds != 0) {
			host->bch_cw_size = chip->ecc_step_ds;
			host->ecc_data.err_corr = chip->ecc_strength_ds;
		} else
			/* this will only happen if flash is screwed up */
			return -EINVAL;
	}

	if (host->ecc_data.err_corr == 1) {
		host->ecc_data.ecc_algo = IMG_NAND_ECC_ALGO_HAMMING;
		host->bch_cw_size = IMG_HAMMING_CW_SIZE;
	} else {
		host->ecc_data.err_corr = round_up(host->ecc_data.err_corr, 4);

		if (host->ecc_data.err_corr > IMG_MAX_ERR_CORR_BCH) {
			dev_err(&mtd->dev, "invalid ecc strength value %d\n",
				host->ecc_data.err_corr);
			return -EINVAL;
		}

		switch (host->bch_cw_size) {
		case IMG_MIN_BCH_CW_SIZE:
		case IMG_MAX_BCH_CW_SIZE:
			break;
		default:
			dev_err(&mtd->dev,
				"invalid BCH codeword size value %d\n",
				host->bch_cw_size);
			return -EINVAL;
		}
		host->ecc_data.ecc_algo = IMG_NAND_ECC_ALGO_BCH;
	}
	host->cw_size = host->bch_cw_size;
	dev_info(&mtd->dev, "ECC mode used: %d corrections over %d bytes\n",
		 host->ecc_data.err_corr, host->bch_cw_size);
	return 0;
}

/*
 * Probe for NAND controller
 */
static int img_nand_probe(struct platform_device *pdev)
{
	int ret, irq, i;
	unsigned int oob_index;
	struct resource *rc;
	struct mtd_info *mtd;
	struct img_nand_host *host;
	struct nand_chip *chip;
	struct nand_ecclayout *ecclayout;
	struct mtd_part_parser_data ppdata = {};
	struct nand_buffers *nbuf;

	/* Allocate memory for the device structure (and zero it) */
	host = devm_kzalloc(&pdev->dev, sizeof(*host), GFP_KERNEL);
	if (!host)
		return -ENOMEM;

	rc = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	host->io_base = devm_ioremap_resource(&pdev->dev, rc);
	if (IS_ERR(host->io_base))
		return PTR_ERR(host->io_base);

	mtd = &host->mtd;
	chip = &host->chip;

	/* Parsing dt */
	ret = img_nand_parse_dt(&pdev->dev, host);
	if (ret < 0) {
		dev_err(&pdev->dev, "dt parsing failed\n");
		return ret;
	}

	chip->priv = host;	/* link the private data structures */
	mtd->priv = chip;
	mtd->owner = THIS_MODULE;
	mtd->dev.parent	= &pdev->dev;
	mtd->name = DRIVER_NAME;

	/* Get NAND Flash clock */
	host->flash_clk = devm_clk_get(&pdev->dev, "ufc_flash_clk");
	if (IS_ERR(host->flash_clk)) {
		dev_err(&pdev->dev, "failed to get flash clock\n");
		return PTR_ERR(host->flash_clk);
	}

	img_nand_write_reg(host, REG_NFLASH_CONTROL_3, 0x555);

	/* Enable NAND Flash clock */
	ret = clk_prepare_enable(host->flash_clk);
	if (ret < 0) {
		dev_err(&pdev->dev, "failed to enable flash clock\n");
		return ret;
	}

	chip->cmd_ctrl = img_nand_cmd_ctrl;
	chip->cmdfunc = img_nand_command;
	chip->read_byte = img_read_byte;
	chip->read_word = img_read_word;
	chip->read_buf = img_nand_read_buf;
	chip->write_byte = img_write_byte;
	chip->write_buf = img_nand_write_buf;
	chip->write_page = img_write_page;

	init_completion(&host->comp_nand);

	/*
	 * If RDY/BSY line is connected to IMG controller then use the
	 * IMG ready function and the generic nand_wait function which reads
	 * the status register after monitoring the RDY/BSY line. Otherwise,
	 * use a standard chip delay which is slightly more than tR (AC Timing)
	 * of the NAND device and read status register until you get a failure
	 * or success
	 */
	if (host->ctrl.dev_ready) {
		chip->dev_ready = img_nand_dev_ready;
		chip->chip_delay = 0;
	} else {
		chip->waitfunc = img_nand_waitfunc;
		chip->chip_delay = 50;
	}

	chip->options = NAND_NO_SUBPAGE_WRITE | NAND_OWN_BUFFERS |
		NAND_USE_BOUNCE_BUFFER | NAND_AVOID_HIGHMEM_BUFFERS;

	if (host->flash_bbt)
		chip->bbt_options = NAND_BBT_USE_FLASH | NAND_BBT_NO_OOB;
	else
		chip->options |= NAND_SKIP_BBTSCAN;

	if (host->ctrl.bus_width) {   /* 16-bit bus width */
		chip->options |= NAND_BUSWIDTH_16;
		chip->write_byte = img_write_byte16;
	}

	if (nand_scan_ident(mtd, 1, NULL)) {
		dev_err(&pdev->dev,
			"scan failed, may be bus-width mismatch\n");
		ret = -ENXIO;
		goto err_flash_clk_disable;
	}

	if (host->ecc_data.ecc_opt != NAND_ECC_NONE
			&& find_best_ecc_mode(host)) {
		dev_err(&pdev->dev, "No suitable ECC mode found\n");
		ret = -EINVAL;
		goto err_flash_clk_disable;
	}

	/* Init Nand Controller */
	ret = img_nand_setup(mtd);
	if (ret < 0) {
		dev_err(&pdev->dev,
			"Nand controller initialization failure\n");
		goto err_flash_clk_disable;
	}

	host->uncorrectable_err = false;

	/* populate MTD interface based on ECC scheme */
	ecclayout = &img_oobinfo;

	if (host->ecc_data.ecc_opt) {
		chip->ecc.mode = NAND_ECC_HW_OOB_FIRST;
		chip->ecc.size = host->bch_cw_size;
		chip->ecc.strength = host->ecc_data.err_corr;
		/* Number of correctable bit-flips per page. */
		mtd->bitflip_threshold = host->ecc_data.err_corr *
			(mtd->writesize / chip->ecc.size);

		switch (host->ecc_data.ecc_algo) {
		case IMG_NAND_ECC_ALGO_HAMMING:
			chip->ecc.bytes = 3;
			break;

		case IMG_NAND_ECC_ALGO_BCH:
			if (host->bch_byte_align_8bitmode) {
				chip->ecc.bytes =
					DIV_ROUND_UP(IMG_BCH_ALGO_POLYNOMIAL *
						     chip->ecc.strength, 8);
			} else {
				chip->ecc.bytes = chip->ecc.strength * 2;
				if (chip->ecc.strength >= 32)
					chip->ecc.bytes -= 4;
			}
			break;

		default:
			dev_err(&pdev->dev,
				"invalid or unsupported ECC scheme\n");
			ret = -EINVAL;
			goto err_flash_clk_disable;
		}

		/* Define ECC layout. */
		ecclayout->eccbytes = chip->ecc.bytes *
				(mtd->writesize / chip->ecc.size);

		oob_index = IMG_ECC_OFFSET_IN_OOB;

		for (i = 0; i < ecclayout->eccbytes; i++, oob_index++)
			ecclayout->eccpos[i] = oob_index;

		ecclayout->oobfree->offset =
			ecclayout->eccpos[ecclayout->eccbytes - 1] + 1;
	} else
		chip->ecc.mode = NAND_ECC_NONE;

	chip->ecc.hwctl			= img_ecc_enable_hwecc;
	chip->ecc.calculate		= img_ecc_calculate;
	chip->ecc.correct		= img_ecc_correct;
	chip->ecc.read_page		= img_ecc_read_page;
	chip->ecc.write_page		= img_ecc_write_page;
	chip->ecc.read_page_raw		= img_ecc_read_page_raw;
	chip->ecc.write_page_raw	= img_ecc_write_page_raw;
	chip->ecc.read_oob_raw		= img_ecc_read_oob_raw;
	chip->ecc.write_oob_raw		= img_ecc_write_oob_raw;

	/* All OOB bytes from oobfree->offset till end off OOB are free */
	ecclayout->oobfree->length = mtd->oobsize - ecclayout->oobfree->offset;

	/* Check if NAND device's OOB is enough to store ECC signatures */
	if (mtd->oobsize < (ecclayout->eccbytes + IMG_ECC_OFFSET_IN_OOB)) {
		dev_err(&pdev->dev,
			"not enough OOB bytes required = %d, available=%d\n",
			ecclayout->eccbytes + IMG_ECC_OFFSET_IN_OOB,
			mtd->oobsize);
		ret = -EINVAL;
		goto err_flash_clk_disable;
	}
	/* Check if OOB is big enough to store page written magic number. */
	if (ecclayout->eccbytes + IMG_ECC_OFFSET_IN_OOB >
	    mtd->oobsize - PAGE_WRITTEN_MAGIC_NUMBER_SIZE) {
		dev_err(&pdev->dev,
			"not enough OOB bytes required = %d, available=%d\n",
			ecclayout->eccbytes + IMG_ECC_OFFSET_IN_OOB +
			PAGE_WRITTEN_MAGIC_NUMBER_SIZE, mtd->oobsize);
		ret = -EINVAL;
		goto err_flash_clk_disable;
	}

	chip->ecc.layout = ecclayout;

	if (host->dma.has_dma)
		img_nand_dma_setup(host);

	irq = platform_get_irq(pdev, 0);
	if (irq < 0) {
		dev_err(&pdev->dev, "failed to get platform irq\n");
		ret = -ENXIO;
		goto err_flash_clk_disable;
	}

	ret = devm_request_irq(&pdev->dev, irq, (irq_handler_t)&img_nand_irq,
			       IRQF_TRIGGER_HIGH, DRIVER_NAME, host);
	if (ret) {
		dev_err(&pdev->dev, "Error requesting NAND IRQ\n");
		ret = -ENXIO;
		goto err_flash_clk_disable;
	}
	host->irq_initialized = true;

	nbuf = devm_kzalloc(&pdev->dev, sizeof(*nbuf), GFP_KERNEL);

	if (!nbuf) {
		ret = -ENOMEM;
		goto err_flash_clk_disable;
	}

	if (host->dma.has_dma) {
		host->dma.dma_buf_virt =
			dma_alloc_coherent(&pdev->dev,
					   mtd->writesize + mtd->oobsize,
					   &host->dma.dma_buf_phys, GFP_DMA);

		if (!host->dma.dma_buf_virt ||
		    !virt_addr_valid(host->dma.dma_buf_virt))
			dev_err(&pdev->dev,
				"Failed to allocate DMA buffer.\n");
		else
			nbuf->databuf = host->dma.dma_buf_virt;
	} else
		nbuf->databuf = devm_kzalloc(&pdev->dev,
					     mtd->writesize + mtd->oobsize,
					     GFP_KERNEL);

	if (!nbuf->databuf) {
		ret = -ENOMEM;
		goto err_release_nbuf;
	}
	chip->buffers = nbuf;

	if (nand_scan_tail(mtd)) {
		ret = -ENXIO;
		goto err_release_data_buf;
	}

	ppdata.of_node = pdev->dev.of_node;
	ret = mtd_device_parse_register(mtd, NULL, &ppdata, NULL, 0);
	if (!ret) {
		platform_set_drvdata(pdev, host);
		return ret;
	}

	nand_release(mtd);

err_release_data_buf:
	if (host->dma.has_dma)
		dma_free_coherent(&pdev->dev, mtd->writesize,
				  host->dma.dma_buf_virt,
				  host->dma.dma_buf_phys);
	else
		devm_kfree(&pdev->dev, nbuf->databuf);

err_release_nbuf:
	devm_kfree(&pdev->dev, nbuf);
	chip->buffers = NULL;

err_flash_clk_disable:
	clk_disable_unprepare(host->flash_clk);
	return ret;
}

/*
 * Remove NAND device
 */
static int img_nand_remove(struct platform_device *pdev)
{
	struct img_nand_host *host = platform_get_drvdata(pdev);
	struct mtd_info *mtd = &host->mtd;

	if (host->dma.dma_buf_virt && virt_addr_valid(host->dma.dma_buf_virt))
		dma_free_coherent(&pdev->dev, mtd->writesize,
				  host->dma.dma_buf_virt,
				  host->dma.dma_buf_phys);
	if (host->dma.has_dma)
		dma_free_coherent(&pdev->dev, mtd->writesize,
				  host->dma.dma_buf_virt,
				  host->dma.dma_buf_phys);
	else
		devm_kfree(&pdev->dev, host->chip.buffers->databuf);

	devm_kfree(&pdev->dev, host->chip.buffers);

	if (host)
		devm_kfree(&pdev->dev, host);

	nand_release(mtd);
	clk_disable_unprepare(host->flash_clk);
	return 0;
}

static const struct of_device_id img_nand_match[] = {
	{ .compatible = "img,nfc" },
	{ /* sentinel */ },
};
MODULE_DEVICE_TABLE(of, img_nand_match);

static struct platform_driver img_nand_driver = {
	.probe		= img_nand_probe,
	.remove		= img_nand_remove,
	.driver		= {
		.name	= DRIVER_NAME,
		.of_match_table = img_nand_match,
	},
};
module_platform_driver(img_nand_driver);

MODULE_LICENSE("GPL v2");
MODULE_AUTHOR("Arul Ramasamy, Krzysztof Antonowicz");
MODULE_DESCRIPTION("Imagination Technologies Generic NAND driver");
