/*
 * Copyright 2017 RnD Center "ELVEES", JSC
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, version 2,
 * as published by the Free Software Foundation.
 *
 */

#include <linux/mfd/syscon.h>
#include <linux/of.h>
#include <linux/of_address.h>
#include <linux/of_net.h>
#include <linux/phy.h>
#include <linux/regmap.h>
#include <linux/reset.h>
#include <linux/stmmac.h>

#include "stmmac.h"
#include "stmmac_platform.h"

#define CR_PERIP_ETH_PHY_INTF_SEL		0x1b0
#define CR_ETHERNET_CLK_CTRL			0x1ac

#define CR_ETHERNET_MX_ETH_CLK_OSC_SW		BIT(0)
#define CR_ETHERNET_MX_ETH_CLK_TX_SW		BIT(1)
#define CR_ETHERNET_MX_ETH_CTRL_CLK_RX_SW	BIT(2)

struct cashew_dwmac {
	struct plat_stmmacenet_data *plat_dat;
	struct device *dev;
	struct regmap *periph_regs;
	struct regmap *top_secure_regs;
};

static int cashew_dwmac_init(struct cashew_dwmac *dwmac)
{
	int rc;
	u32 iface_sel, clk_rate, clk_cfg;

	dev_info(dwmac->dev, "Connected PHY interface: %s\n",
		 phy_modes(dwmac->plat_dat->interface));

	switch (dwmac->plat_dat->interface) {
	case PHY_INTERFACE_MODE_RMII:
		iface_sel = 4;
		clk_rate = 50000000;
		clk_cfg = 0;
		break;
	case PHY_INTERFACE_MODE_RGMII:
	case PHY_INTERFACE_MODE_RGMII_ID:
	case PHY_INTERFACE_MODE_RGMII_TXID:
	case PHY_INTERFACE_MODE_RGMII_RXID:
		iface_sel = 1;
		clk_rate = 125000000;
		clk_cfg = CR_ETHERNET_MX_ETH_CLK_TX_SW |
			  CR_ETHERNET_MX_ETH_CTRL_CLK_RX_SW;
		break;
	default:
		dev_err(dwmac->dev, "Unsupported PHY interface\n");
		return -ENOTSUPP;
	}

	rc = regmap_write(dwmac->periph_regs, CR_PERIP_ETH_PHY_INTF_SEL,
			  iface_sel);
	if (rc) {
		dev_err(dwmac->dev, "Failed to configure PHY interface\n");
		return rc;
	}

	rc = regmap_write(dwmac->top_secure_regs, CR_ETHERNET_CLK_CTRL,
			  clk_cfg);
	if (rc) {
		dev_err(dwmac->dev, "PHY interface clock\n");
		return rc;
	}

	rc = clk_set_rate(dwmac->plat_dat->stmmac_clk, clk_rate);
	if (rc) {
		dev_err(dwmac->dev, "Failed to set clock rate\n");
		return rc;
	}

	/* TODO: Output enable of ETH_CLK pad should be turned on here
	 * if ETH_CLK pad is used to provide RMII clock for PHY */

	rc = reset_control_reset(dwmac->plat_dat->stmmac_rst);
	if (rc) {
		dev_err(dwmac->dev, "Failed to reset controller\n");
		return rc;
	}

	return 0;
}

static int cashew_dwmac_probe(struct platform_device *pdev)
{
	struct plat_stmmacenet_data *plat_dat;
	struct stmmac_resources stmmac_res;
	struct device *dev = &pdev->dev;
	struct cashew_dwmac *dwmac;
	int ret;

	ret = stmmac_get_platform_resources(pdev, &stmmac_res);
	if (ret)
		return ret;

	plat_dat = stmmac_probe_config_dt(pdev, &stmmac_res.mac);
	if (IS_ERR(plat_dat))
		return PTR_ERR(plat_dat);

	plat_dat->has_gmac4 = 1;
	plat_dat->dma_cfg->aal = 1;
	plat_dat->tso_en = 1;
	plat_dat->pmt = 1;

	dwmac = devm_kzalloc(dev, sizeof(*dwmac), GFP_KERNEL);
	if (!dwmac) {
		ret = -ENOMEM;
		goto err_remove_config_dt;
	}

	dwmac->plat_dat = plat_dat;
	dwmac->dev = dev;
	plat_dat->bsp_priv = dwmac;

	dwmac->periph_regs = syscon_regmap_lookup_by_phandle(dev->of_node,
							     "img,cr-periph");
	if (IS_ERR(dwmac->periph_regs)) {
		ret = PTR_ERR(dwmac->periph_regs);
		goto err_remove_config_dt;
	}

	dwmac->top_secure_regs =
		syscon_regmap_lookup_by_phandle(dev->of_node,
						"img,cr-top-secure");
	if (IS_ERR(dwmac->top_secure_regs)) {
		ret = PTR_ERR(dwmac->top_secure_regs);
		goto err_remove_config_dt;
	}

	ret = stmmac_dvr_probe(dev, plat_dat, &stmmac_res);
	if (ret)
		goto err_remove_config_dt;

	ret = cashew_dwmac_init(dwmac);
	if (ret)
		goto err_dvr_remove;

	return 0;

err_dvr_remove:
	stmmac_dvr_remove(dev);
err_remove_config_dt:
	stmmac_remove_config_dt(pdev, plat_dat);

	return ret;
}

static const struct of_device_id cashew_dwmac_match[] = {
	{ .compatible = "img,cashew-dwmac" },
	{ }
};
MODULE_DEVICE_TABLE(of, cashew_dwmac_match);

static struct platform_driver cashew_dwmac_driver = {
	.probe  = cashew_dwmac_probe,
	.remove = stmmac_pltfr_remove,
	.driver = {
		.name           = "cashew-dwmac",
		.of_match_table = cashew_dwmac_match,
	},
};
module_platform_driver(cashew_dwmac_driver);

MODULE_AUTHOR("Dmitriy Zagrebin <dzagrebin@elvees.com>");
MODULE_DESCRIPTION("ImgTec Cashew DWMAC specific glue layer");
MODULE_LICENSE("GPL v2");
