/*
 * Copyright (c) 2011-2017 Imagination Technologies Ltd.
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <linux/io.h>
#include <linux/kernel.h>
#include <linux/dma-mapping.h>
#include <linux/delay.h>
#include <linux/slab.h>

#include <core.h>
#include <fwldr.h>

struct fwload_priv  *fpriv, fpv;


static enum uccp_mem_region fwldr_chk_region(unsigned int src_addr,
				      int length);

static void dir_mem_set(unsigned int addr,
			unsigned char data,
			unsigned int len);

/* dir_mem_cpy
 *
 * Perform a memcpy of 'len' bytes from 'src_addr' to the UCCP memory location
 * pointed by 'dst_addr'.
 *
 * dst_addr is always a 4 byte aligned address
 * data is always a 4 byte aligned address
 * len is always a multiple of 4 when dst_addr is of type 0xB4xxxxxx
 * len may NOT be a multiple of 4 when dst_addr is of type 0xB7xxxxxx
 *
 *
 * When dst_addr is of type 0xB4xxxxxx, perform only 32 bit writes to these
 * locations
 *
 */
void dir_mem_cpy(unsigned int addr,
		unsigned char *data,
		unsigned int len)
{
	int i;
	unsigned long offset = (unsigned long)addr & UCCP_OFFSET_MASK;
	unsigned long base = ((unsigned long)addr & UCCP_BASE_MASK) >> 24;
	unsigned int *data_addr = (unsigned int *)data;
	unsigned char *gram_byte_addr, *data_byte_addr;

	if ((fpriv->gram_b4_addr ==  NULL) && (base == UCCP_GRAM_MSB)) {

		/* The HAL didn't provide a virtual address for 0xB4xxxxxx alias
		 * Convert into 0xB7 and do the writes by ignoring MSB of the
		 * 32-bit word
		 */

		addr &= 0x00FFFFFF;
		addr |= 0xB7000000;

		data_byte_addr = (unsigned char *)data_addr;
		gram_byte_addr = (void *)(fpriv->gram_addr + (offset / 4) * 3);

		hal_ops.set_mem_region(addr);

		if (len % 4 == 0) {
			for (i = 0; i < len / 4; i++) {
				memcpy(gram_byte_addr, data_byte_addr + 1, 3);
				gram_byte_addr += 3;
				data_byte_addr += 4;
			}
		} else {
			fwldr_dbg_err("%s:Unexpected length(base:%lx)\n",
				      __func__, base);
		}

	} else {

		hal_ops.set_mem_region(addr);
		if (len % 4 == 0) {
			for (i = 0; i < len / 4; i++) {
				fwload_uccp_write(fpriv, base, offset,
						  data_addr[i]);
				offset += 4;
			}
		} else {
			if (base == UCCP_GRAM_PACKED)
				memcpy((void *)(fpriv->gram_addr + offset),
				       (void *)data_addr, len);
			else
				fwldr_dbg_err("%s:Unexpected length(%lx)\n",
					      __func__, base);
		}
	}
}


/* dir_mem_set
 *
 * Perform a memset of 'len' bytes with value of 'val' to the UCCP memory
 * location pointed by 'dst_addr'.
 *
 * dst_addr is always a 4 byte aligned address
 * len is always a multiple of 4 when dst_addr is of type 0xB4xxxxxx
 * len may NOT be a multiple of 4 when dst_addr is of type 0xB7xxxxxx
 *
 *
 * When dst_addr is of type 0xB4xxxxxx, perform only 32 bit writes to these
 * locations
 *
 */
static void dir_mem_set(unsigned int addr,
			unsigned char data,
			unsigned int len)
{
	int i;
	unsigned long offset = (unsigned long)addr & UCCP_OFFSET_MASK;
	unsigned long base = ((unsigned long)addr & UCCP_BASE_MASK) >> 24;
	unsigned char *gram_byte_addr;

	if ((fpriv->gram_b4_addr ==  NULL) && (base == UCCP_GRAM_MSB)) {

		/* The HAL didn't provide a virtual address for 0xB4xxxxxx alias
		 * Convert into 0xB7 and do the writes by ignoring MSB of the
		 * 32-bit word
		 */

		addr &= 0x00FFFFFF;
		addr |= 0xB7000000;

		gram_byte_addr = (void *)(fpriv->gram_addr + (offset / 4) * 3);

		hal_ops.set_mem_region(addr);

		if (len % 4 == 0) {
			memset(gram_byte_addr, data, (len / 4) * 3);
		} else {
			fwldr_dbg_err("%s :Unexpected length (base : %lx)\n",
				      __func__, base);
		}

	} else {

		hal_ops.set_mem_region(addr);

		if (len % 4 == 0) {
			for (i = 0; i <= len / 4; i++) {
				fwload_uccp_write(fpriv, base, offset, data);
				offset += 4;
			}
		} else if (base == UCCP_GRAM_PACKED) {
			memset((void *)(fpriv->gram_addr + offset),
			       data, len);
		} else {
			fwldr_dbg_err("%s: Unexpected length (base %lx)\n",
				      __func__, base);
		}
	}
}

/* Perform 'len' 32 bit reads from a UCCP memory location 'addr'
 * 'addr' is always a 4 byte aligned address
 */
void dir_mem_read(unsigned int addr,
			 unsigned int *data,
			 unsigned int len)
{
	int i = 0;
	unsigned long offset = (unsigned long)addr & UCCP_OFFSET_MASK;
	unsigned long base = ((unsigned long)addr & UCCP_BASE_MASK) >> 24;

	hal_ops.set_mem_region(addr);

	for (i = 0; i <= len / 4; i++) {
		fwload_uccp_read(fpriv, base, offset, data+i);
		offset += 4;
	}
}

/* 32 bit write to UCCP memory location 'addr'
 * 'addr' is always a 4 byte aligned address
 */
void dir_mem_write(unsigned int addr,
			 unsigned int data)
{
	unsigned long offset = (unsigned long)addr & UCCP_OFFSET_MASK;
	unsigned long base = ((unsigned long)addr & UCCP_BASE_MASK) >> 24;

	hal_ops.set_mem_region(addr);
	fwload_uccp_write(fpriv, base, offset, data);
}

static void core_mem_cpy(unsigned int addr,
			 unsigned char *data,
			 unsigned int len)
{
	unsigned int i = 0;
	unsigned int *src_data = (unsigned int *)data;
	unsigned int flag = 0;
	unsigned int val = 0;

	/* Poll MSLVCTRL1 */
	do {
		dir_mem_read(MSLVCTRL1, &val, 1);
	} while (!MSLAVE_READY(val));

	if (len > 1)
		flag = SLAVE_BLOCK_WRITE;
	else
		flag = SLAVE_SINGLE_WRITE;

	dir_mem_write(MSLVCTRL0,
			((addr & SLAVE_ADDR_MODE_MASK) | flag));

	for (i = 0; i < len / 4; i++) {
		do {
			dir_mem_read(MSLVCTRL1, &val, 1);
		} while (!MSLAVE_READY(val));

		if (data != NULL)
			dir_mem_write(MSLVDATAT, src_data[i]);
		else
			dir_mem_write(MSLVDATAT, 0x00);
	}

}

void core_mem_set(unsigned int addr,
		unsigned int data,
		unsigned int len)
{
	unsigned int flag = 0;
	unsigned int val = 0;

	/* Poll MSLVCTRL1 */
	do {
		dir_mem_read(MSLVCTRL1, &val, 1);
	} while (!MSLAVE_READY(val));

	if (len > 1)
		flag = SLAVE_BLOCK_WRITE;
	else
		flag = SLAVE_SINGLE_WRITE;

	dir_mem_write(MSLVCTRL0,
			((addr & SLAVE_ADDR_MODE_MASK) | flag));
	dir_mem_write(MSLVDATAT, data);
}


void core_mem_read(unsigned int addr,
			  unsigned int *data,
			  unsigned int len)
{
	unsigned int i = 0;
	unsigned int val = 0;

	/* Poll MSLVCTRL1 */
	do {
		dir_mem_read(MSLVCTRL1, &val, 1);
	} while (!MSLAVE_READY(val));

	dir_mem_write(MSLVCTRL0,
			((addr & SLAVE_ADDR_MODE_MASK) | SLAVE_BLOCK_READ));

	for (i = 0; i < len-1; i++) {
		do {
			dir_mem_read(MSLVCTRL1, &val, 1);
		} while (!MSLAVE_READY(val));

		dir_mem_read(MSLVDATAT, &data[i], 1);
	}

	/* Read the last word */
	do {
		dir_mem_read(MSLVCTRL1, &val, 1);
	} while (!MSLAVE_READY(val));

	dir_mem_read(MSLVDATAX, &data[len-1], 1);


}

int rpudump_init(void)
{
	fpriv = &fpv;

	hal_ops.request_mem_regions(&fpriv->gram_addr,
				    &fpriv->sysbus_addr,
				    &fpriv->gram_b4_addr);
	return 0;
}

int fwldr_init(void)
{
	int err = FWLDR_SUCCESS;

	fpriv = &fpv;
	hal_ops.request_mem_regions(&fpriv->gram_addr,
				    &fpriv->sysbus_addr,
				    &fpriv->gram_b4_addr);

	return err;
}


void fwldr_load_mem(unsigned int dst_addr,
		    unsigned int len,
		    unsigned char *src_buf)
{
	enum uccp_mem_region mem_region = UCCP_MEM_ERR;
	int i = 0;

	mem_region = fwldr_chk_region(dst_addr, len);

	fwldr_dbg_info("%s dst_addr = 0x%X, length = 0x%X, srcaddr = 0x%X\n",
		       __func__, dst_addr, len, (unsigned int)src_buf);

	fwldr_dbg_info("Dump upto 16 bytes\n");

	if (0 != (dst_addr % 4))
		fwldr_dbg_info("Destination Address is not 4 - byte aligned\n");

	for (i = 0; i < 16; i += 2)
		fwldr_dbg_dump("0x%X \t 0x%X\n", src_buf[i], src_buf[i + 1]);

	switch (mem_region) {
	case UCCP_MEM_CORE:
		core_mem_cpy(dst_addr, src_buf, len);
		break;

	case UCCP_MEM_DIRECT:
		dir_mem_cpy(dst_addr, src_buf, len);
		break;

	default:
		fwldr_dbg_err("Region unknown. Skipped writing\n");
		break;
	}
}


void fwldr_zero_mem(unsigned int dst_addr,
		    unsigned int len)
{
	int mem_region = 0;

	fwldr_dbg_info("%s DestAddr = 0x%X, length = 0x%X\n",
		       __func__, dst_addr, len);

	if (0 != (dst_addr % 4))
		fwldr_dbg_info("Destination Address is not 4 - byte aligned");

	mem_region = fwldr_chk_region(dst_addr, len);

	switch (mem_region) {
	case UCCP_MEM_CORE:
		core_mem_cpy(dst_addr, NULL, len);
		break;

	case UCCP_MEM_DIRECT:
		dir_mem_set(dst_addr, 0x00, len);
		break;

	default:
		fwldr_dbg_err("Region unknown. Skipped setting\n");
		break;
	}
}


unsigned int fwldr_config_read(unsigned int dst_addr)
{
	int mem_region = 0;
	int val = 0;

	fwldr_dbg_info("%s dst_addr = 0x%X\n", __func__, dst_addr);

	if (0 != (dst_addr % 4))
		fwldr_dbg_info("Destination Address is not 4 - byte aligned");

	mem_region = fwldr_chk_region(dst_addr, 0);

	switch (mem_region) {
	case UCCP_MEM_CORE:
		core_mem_read(dst_addr, &val, 1);
		return val;

	case UCCP_MEM_DIRECT:
		dir_mem_read(dst_addr, &val, 1);
		return val;

	default:
		fwldr_dbg_err("Region unknown. Skipped reading\n");
		return 0;
	}

	return 0;
}


void fwldr_config_write(unsigned int dst_addr,
			unsigned int val)
{
	int mem_region = 0;

	fwldr_dbg_info("%s dst_addr = 0x%X,\tValue = 0x%X\n",
		       __func__, dst_addr, val);

	if (0 != (dst_addr % 4))
		fwldr_dbg_info("Destination Address is not 4 - byte aligned");

	mem_region = fwldr_chk_region(dst_addr, 0);


	switch (mem_region) {
	case UCCP_MEM_CORE:
		core_mem_set(dst_addr, val, 1);
		break;

	case UCCP_MEM_DIRECT:
		dir_mem_write(dst_addr, val);
		break;

	default:
		fwldr_dbg_err("Region unknown. Skipped writing\n");
		break;
	}

}


static enum uccp_mem_region fwldr_chk_region(unsigned int src_addr, int len)
{
	unsigned int dst_addr = src_addr + len;

	if (((src_addr >= 0x03000000) && (src_addr <= 0x04FFFFFF))  ||
	    ((src_addr >= 0x02009000) && (src_addr <= 0x0203BFFF))  ||
	    ((src_addr >= 0x80000000) && (src_addr <= 0x87FFFFFF))) {
		if (len != 0) {
			if (((dst_addr >= 0x03000000) &&
			     (dst_addr <= 0x04FFFFFF)) ||
			    ((dst_addr >= 0x02009000) &&
			     (dst_addr <= 0x0203BFFF)) ||
			    ((dst_addr >= 0x80000000) &&
			     (dst_addr <= 0x87FFFFFF)))
				return UCCP_MEM_CORE;
			else
				return UCCP_MEM_ERR;
		}

		return UCCP_MEM_CORE;
	} else if ((src_addr & 0xFF000000) == 0xB0000000) {
		return UCCP_MEM_ERR;
	} else {
		return UCCP_MEM_DIRECT;
	}
}
