/*
 * Copyright (c) 2011-2017 Imagination Technologies Ltd.
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef _CORE_H_
#define _CORE_H_

#include <linux/atomic.h>
#include <linux/delay.h>
#include <linux/dma-mapping.h>
#include <linux/etherdevice.h>
#include <linux/interrupt.h>
#include <linux/jiffies.h>
#include <linux/sched.h>
#include <linux/skbuff.h>
#include <linux/spinlock.h>
#include <linux/timer.h>
#include <linux/version.h>
#include <linux/wireless.h>

#include <net/mac80211.h>

#include "soc.h"
#include "descriptors.h"
#include "rpu_if.h"
#include "hal_common.h"
#include "version.h"

extern int uccp_reinit;
extern unsigned int vht_support;
extern struct cmd_send_recv_cnt cmd_info;
extern int rpu_debug;



enum RPU_SLEEP_CTRL {
	RPU_SLEEP_CTRL_DISABLE = 0,
	RPU_SLEEP_CTRL_ENABLE,
};


#define RPU_DEBUG_MAIN(fmt, ...)	    \
do {					      \
	if (rpu_debug & RPU_DEBUG_MAIN)      \
		pr_debug(fmt, ##__VA_ARGS__);	\
} while (0)


#define RPU_DEBUG_UMACIF(fmt, ...)	  \
do {					     \
	if (rpu_debug & RPU_DEBUG_UMACIF)  \
		pr_debug(fmt, ##__VA_ARGS__);  \
} while (0)

#define RPU_DEBUG_CRYPTO(fmt, ...)	     \
do {					       \
	if (rpu_debug & RPU_DEBUG_CRYPTO)     \
		pr_debug(fmt, ##__VA_ARGS__);	 \
} while (0)
#define RPU_DEBUG_VIF(fmt, ...)			\
do {						 \
	if (rpu_debug & RPU_DEBUG_VIF)		  \
		pr_debug(fmt, ##__VA_ARGS__);	   \
} while (0)

#define RPU_DEBUG_TX(fmt, ...)			\
do {						 \
	if (rpu_debug & RPU_DEBUG_TX)		  \
		pr_debug(fmt, ##__VA_ARGS__);	   \
} while (0)

#define RPU_DEBUG_SCAN(fmt, ...)	    \
do {					      \
	if (rpu_debug & RPU_DEBUG_SCAN)      \
		pr_debug(fmt, ##__VA_ARGS__);	\
} while (0)

#define RPU_DEBUG_ROC(fmt, ...)		   \
do {					     \
	if (rpu_debug & RPU_DEBUG_ROC)	    \
		pr_debug(fmt, ##__VA_ARGS__);  \
} while (0)

#define RPU_DEBUG_TSMC(fmt, ...)	     \
do {					       \
	if (rpu_debug & RPU_DEBUG_TSMC)       \
		pr_debug(fmt, ##__VA_ARGS__);	 \
} while (0)

#define RPU_DEBUG_IF(fmt, ...)		    \
do {					      \
	if (rpu_debug & RPU_DEBUG_IF)	     \
		pr_debug(fmt, ##__VA_ARGS__);	\
} while (0)

#define RPU_DEBUG_HAL(fmt, ...)				  \
do {							\
	if ((rpu_debug & RPU_DEBUG_HAL) && net_ratelimit()) \
		pr_debug(fmt, ##__VA_ARGS__);	 \
} while (0)

#define RPU_DEBUG_DUMP_HAL(fmt, ...)			       \
do {							\
	if (rpu_debug & RPU_DEBUG_DUMP_HAL)			\
		print_hex_dump(KERN_DEBUG, fmt, ##__VA_ARGS__);	 \
} while (0)

#define DUMP_HAL (rpu_debug & RPU_DEBUG_DUMP_HAL)

#define RPU_DEBUG_DUMP_TX(fmt, ...)			      \
do {							\
	if (rpu_debug & RPU_DEBUG_DUMP_TX)			\
		print_hex_dump(KERN_DEBUG, "DUMP_TX", fmt, ##__VA_ARGS__);  \
} while (0)

#define DUMP_TX (rpu_debug & RPU_DEBUG_DUMP_TX)

#define RPU_DEBUG_RUNTIME_PM(fmt, ...)				  \
do {							\
	if ((rpu_debug & RPU_DEBUG_RUNTIME_PM) && net_ratelimit()) \
		pr_debug(fmt, ##__VA_ARGS__);	 \
} while (0)

/* Wrapper to check return values for all
 * umac_if layer calls.
 */
#define CALL_RPU(prog_rpu, ...) \
do {				\
	ret = prog_rpu(__VA_ARGS__);   \
	if (ret != 0)			 \
		goto prog_rpu_fail;	 \
} while (0)

#ifdef CONFIG_NL80211_TESTMODE
#define MAX_NL_DUMP_LEN (PAGE_SIZE-1024)
/* This section contains example code for using netlink
 * attributes with the testmode command in nl80211.
 */

/* These enums need to be kept in sync with userspace */
enum rpu_testmode_attr {
	__RPU_TM_ATTR_INVALID = 0,
	RPU_TM_ATTR_CMD      = 1,
	RPU_TM_ATTR_DUMP      = 2,
	/* keep last */
	__RPU_TM_ATTR_AFTER_LAST,
	RPU_TM_ATTR_MAX       = __RPU_TM_ATTR_AFTER_LAST - 1
};

enum rpu_testmode_cmd {
	RPU_TM_CMD_ALL	= 0,
	RPU_TM_CMD_GRAM  = 1,
	RPU_TM_CMD_COREA  = 2,
	RPU_TM_CMD_COREB  = 3,
	RPU_TM_CMD_PERIP = 4,
	RPU_TM_CMD_SYSBUS = 5,
};

#endif
extern unsigned int system_rev;
extern unsigned int ftm;


extern unsigned char vif_macs[2][ETH_ALEN];

extern spinlock_t tsf_lock;

#define RESET_RX_CONTROL_INFO 1
#define MAX_OUTSTANDING_CTRL_REQ 2
#define RESET_TIMEOUT 5000   /* In milli-seconds*/
#define RESET_TIMEOUT_TICKS msecs_to_jiffies(RESET_TIMEOUT)
/*100: For ROC, 500: For initial*/
#define CH_PROG_TIMEOUT 2000   /* In milli-seconds*/
#define CH_PROG_TIMEOUT_TICKS msecs_to_jiffies(CH_PROG_TIMEOUT)

#define QUEUE_FLUSH_TIMEOUT  2000   /* Specify delay in milli-seconds*/
#define QUEUE_FLUSH_TIMEOUT_TICKS   msecs_to_jiffies(QUEUE_FLUSH_TIMEOUT)

#define TX_DEINIT_TIMEOUT 5000
#define TX_DEINIT_TIMEOUT_TICKS msecs_to_jiffies(TX_DEINIT_TIMEOUT)


#define TX_COMPLETE_TIMEOUT 1000  /* In milli-seconds*/
#define TX_COMPLETE_TIMEOUT_TICKS msecs_to_jiffies(TX_COMPLETE_TIMEOUT)
#define SCAN_ABORT_TIMEOUT 1000
#define SCAN_ABORT_TIMEOUT_TICKS msecs_to_jiffies(SCAN_ABORT_TIMEOUT)
#define CANCEL_HW_ROC_TIMEOUT 1000
#define CANCEL_HW_ROC_TIMEOUT_TICKS msecs_to_jiffies(CANCEL_HW_ROC_TIMEOUT)

#define DEFAULT_TX_ANT_SELECT 3 /* bitmap of antennas for tx, 3=> both first and
				 * second antenna to be used
				 */
#define DEFAULT_TX_POWER 15
#define DEFAULT_RTS_THRESHOLD 2347
#define SUPPORTED_FILTERS (FIF_ALLMULTI | FIF_BCN_PRBRESP_PROMISC)
#define TX_DESC_BUCKET_BOUND 32

#define MAX_DATA_SIZE (0) /* Defined in HAL (or) can be configured from proc */
#define MAX_TX_QUEUE_LEN 192
#define MAX_AUX_ADC_SAMPLES 10

/* Maximum number of Tx streams supported */
/* Maximum number of RX streams supported */
	#define MAX_TX_STREAMS 2
	#define MAX_RX_STREAMS 2

#define   MAX_RSSI_SAMPLES 10
#define   RPU_DBG_DEFAULT		0

#define CLOCK_MASK 0x3FFFFFFF
#define TICK_NUMRATOR 12288 /* 12288 KHz  */
#define TICK_DENOMINATOR 1000 /* 1000 KHz */

#define BTS_AP_24GHZ_ETS 195 /* Microsecs */
#define BTS_AP_5GHZ_ETS 25 /* Microsecs */

#define DEFAULT_MAC_ADDRESS "001122334455"


enum ptype {
	UCAST = 0,
	MCAST
};

enum noa_triggers {
	FROM_TX = 0,
	FROM_TX_DONE,
	FROM_EVENT_NOA
};

enum tx_deinit {
	TX_FLUSH = 0,
	TX_DROP,
};

enum rpu_hw_scan_status {
	HW_SCAN_STATUS_NONE,
	HW_SCAN_STATUS_PROGRESS
};

enum rpu_debug {
	RPU_DEBUG_SCAN			= BIT(1),
	RPU_DEBUG_ROC			= BIT(2),
	RPU_DEBUG_TX			= BIT(3),
	RPU_DEBUG_MAIN			= BIT(4),
	RPU_DEBUG_IF			= BIT(5),
	RPU_DEBUG_UMACIF		= BIT(6),
	RPU_DEBUG_RX			= BIT(7),
	RPU_DEBUG_HAL			= BIT(8),
	RPU_DEBUG_CRYPTO		= BIT(9),
	RPU_DEBUG_DUMP_RX		= BIT(10),
	RPU_DEBUG_DUMP_HAL		= BIT(11),
	RPU_DEBUG_TSMC			= BIT(12),
	RPU_DEBUG_P2P			= BIT(13),
	RPU_DEBUG_VIF			= BIT(14),
	RPU_DEBUG_DUMP_TX		= BIT(15),
	RPU_DEBUG_RUNTIME_PM		= BIT(16),
};

struct wifi_sync {
	unsigned int  status;
	unsigned char ts1[8];
	unsigned long long atu;
	unsigned char  bssid[8];
	unsigned char  name[10];
	unsigned int  ts2;
};

struct wifi_params {
	int ed_sensitivity;
	int num_vifs;
	int tx_fixed_rate;
	int tx_fixed_mcs_indx;
	int mgd_mode_tx_fixed_rate;
	int mgd_mode_tx_fixed_mcs_indx;
	unsigned int peer_ampdu_factor;
	unsigned char is_associated;
	unsigned char rate_protection_type;
	unsigned char num_spatial_streams;
	unsigned char enable_early_agg_checks;
	unsigned char rpu_max_nss;
	/*RF Params: Input to the RF for operation*/
	unsigned char  rf_params[RF_PARAMS_SIZE];
	unsigned char  rf_params_vpd[RF_PARAMS_SIZE];
	unsigned char production_test;
	unsigned int dot11a_support;
	unsigned int dot11g_support;

	/* CHannel config*/
	unsigned int chnl_40_offset;
	unsigned int chnl_80_offset;

	/*Multicast  Rate config options*/
	unsigned int mgd_mode_mcast_fixed_rate_flags;
	int mgd_mode_mcast_fixed_data_rate;
	unsigned int mgd_mode_mcast_fixed_bcc_or_ldpc;
	unsigned int mgd_mode_mcast_fixed_stbc_enabled;
	unsigned int mgd_mode_mcast_fixed_preamble;
	unsigned char mgd_mode_mcast_fixed_nss;
	/*End*/

	unsigned int fixed_rate_flags;
	unsigned int fixed_rate_preamble_type;
	unsigned int fixed_rate_stbc_enabled;
	unsigned int fixed_rate_bcc_or_ldpc;
	unsigned int max_tx_streams;
	unsigned int max_rx_streams;
	unsigned int max_data_size;
	unsigned int disable_power_save;
	unsigned int disable_sm_power_save;
	unsigned int max_tx_cmds;
	unsigned int vht_beamform_enable;
	unsigned int vht_beamform_period;
	unsigned int vht_beamform_support;
	unsigned char bg_scan_channel_list[50];
	unsigned char bg_scan_channel_flags[50];
	unsigned int bg_scan_enable;
	unsigned int bg_scan_intval;
	unsigned int bg_scan_chan_dur;
	unsigned int bg_scan_serv_chan_dur;
	unsigned int bg_scan_num_channels;
	unsigned int hw_scan_status;
	unsigned int scan_type;
	unsigned int set_tx_power;
	unsigned int aux_adc_chain_id;
	unsigned char pdout_voltage[MAX_AUX_ADC_SAMPLES];
	char rssi_average[MAX_RSSI_SAMPLES];
	unsigned int fw_loading;
	struct wifi_sync sync[MAX_VIFS];
	unsigned int bt_state;
	unsigned int antenna_sel;
	int pkt_gen_val;
	int init_pkt_gen;
	int payload_length;
	int start_prod_mode;
	int init_prod;
	unsigned char bypass_vpd;
	unsigned int cont_tx;
	unsigned int minstrel_fastpath;
	unsigned int max_rate_retries;
};

struct cmd_send_recv_cnt {
	int tx_cmd_send_count;
	int tx_done_recv_count;
	int total_cmd_send_count;
	unsigned int outstanding_ctrl_req;
	unsigned long control_path_flags;
	spinlock_t control_path_lock;
	struct sk_buff_head outstanding_cmd;
};

struct wifi_stats {
	unsigned int ht_tx_mcs0_packet_count;
	unsigned int ht_tx_mcs1_packet_count;
	unsigned int ht_tx_mcs2_packet_count;
	unsigned int ht_tx_mcs3_packet_count;
	unsigned int ht_tx_mcs4_packet_count;
	unsigned int ht_tx_mcs5_packet_count;
	unsigned int ht_tx_mcs6_packet_count;
	unsigned int ht_tx_mcs7_packet_count;
	unsigned int ht_tx_mcs8_packet_count;
	unsigned int ht_tx_mcs9_packet_count;
	unsigned int ht_tx_mcs10_packet_count;
	unsigned int ht_tx_mcs11_packet_count;
	unsigned int ht_tx_mcs12_packet_count;
	unsigned int ht_tx_mcs13_packet_count;
	unsigned int ht_tx_mcs14_packet_count;
	unsigned int ht_tx_mcs15_packet_count;
	unsigned int vht_tx_mcs0_packet_count;
	unsigned int vht_tx_mcs1_packet_count;
	unsigned int vht_tx_mcs2_packet_count;
	unsigned int vht_tx_mcs3_packet_count;
	unsigned int vht_tx_mcs4_packet_count;
	unsigned int vht_tx_mcs5_packet_count;
	unsigned int vht_tx_mcs6_packet_count;
	unsigned int vht_tx_mcs7_packet_count;
	unsigned int vht_tx_mcs8_packet_count;
	unsigned int vht_tx_mcs9_packet_count;
	unsigned int tx_cmds_from_stack;
	unsigned int tx_dones_to_stack;
	unsigned int system_rev;
	unsigned int outstanding_cmd_cnt;
	unsigned int pending_tx_cnt;
	unsigned int umac_scan_req;
	unsigned int umac_scan_complete;
	unsigned int gen_cmd_send_count;
	unsigned int tx_cmd_send_count_single;
	unsigned int tx_cmd_send_count_multi;

	unsigned int tx_noagg_not_qos;
	unsigned int tx_noagg_not_ampdu;
	unsigned int tx_noagg_not_addr;
	unsigned int tx_noagg_not_rate;

	unsigned int tx_cmd_send_count_beaconq;
	unsigned int tx_done_recv_count;
	unsigned int rx_packet_mgmt_count;
	unsigned int rx_packet_data_count;
	unsigned int ed_cnt;
	unsigned int mpdu_cnt;
	unsigned int ofdm_crc32_pass_cnt;
	unsigned int ofdm_crc32_fail_cnt;
	unsigned int dsss_crc32_pass_cnt;
	unsigned int dsss_crc32_fail_cnt;
	unsigned int mac_id_pass_cnt;
	unsigned int mac_id_fail_cnt;
	unsigned int ofdm_corr_pass_cnt;
	unsigned int ofdm_corr_fail_cnt;
	unsigned int dsss_corr_pass_cnt;
	unsigned int dsss_corr_fail_cnt;
	unsigned int ofdm_s2l_fail_cnt;
	unsigned int lsig_fail_cnt;
	unsigned int htsig_fail_cnt;
	unsigned int vhtsiga_fail_cnt;
	unsigned int vhtsigb_fail_cnt;
	unsigned int nonht_ofdm_cnt;
	unsigned int nonht_dsss_cnt;
	unsigned int mm_cnt;
	unsigned int gf_cnt;
	unsigned int vht_cnt;
	unsigned int aggregation_cnt;
	unsigned int non_aggregation_cnt;
	unsigned int ndp_cnt;
	unsigned int ofdm_ldpc_cnt;
	unsigned int ofdm_bcc_cnt;
	unsigned int midpacket_cnt;
	unsigned int dsss_sfd_fail_cnt;
	unsigned int dsss_hdr_fail_cnt;
	unsigned int dsss_short_preamble_cnt;
	unsigned int dsss_long_preamble_cnt;
	unsigned int sifs_event_cnt;
	unsigned int cts_cnt;
	unsigned int ack_cnt;
	unsigned int sifs_no_resp_cnt;
	unsigned int unsupported_cnt;
	unsigned int l1_corr_fail_cnt;
	unsigned int sifs_crc_exit_cnt;
	unsigned int low_energy_event_cnt;
	unsigned int deagg_error_cnt;
	unsigned int nsymbols_error_cnt;
	unsigned int mcs32_cnt;
	unsigned int ndpa_cnt;
	unsigned int lsig_duration_error_cnt;
	unsigned int rts_cnt;
	unsigned int non_ht_cts_cnt;
	unsigned int rxp_active_exit_cnt;
	unsigned int beamform_feedback_cnt;
	unsigned int self_cts_cnt;
	unsigned int pop_master_cnt;
	unsigned int pop_error_cnt;
	unsigned int multicast_cnt;
	unsigned int tx_ed_abort_cnt;
	unsigned int mcp_cts_cnt;
	unsigned int deagg_q_post_cnt;
	unsigned int rxp_active_exit_dsss_cnt;
	unsigned int rxp_extreme_error_cnt;
	unsigned int aci_fail_cnt;

	/* TX related */
	unsigned int tx_pkts_from_lmac;
	unsigned int tx_pkts_tx2tx;
	unsigned int tx_pkts_from_rx;
	unsigned int tx_pkts_ofdm;

	unsigned int tx_pkts_dsss;
	unsigned int tx_pkts_reached_end_of_fsm;
	unsigned int tx_unsupported_modulation;
	unsigned int tx_latest_pkt_from_lmac_or_sifs;

	unsigned int tx_abort_bt_confirm_cnt;	    /* Tx abort due to BT
						     * confirm at the start of
						     * Txn
						     */
	unsigned int tx_abort_txstart_timeout_cnt;  /* Tx abort due to Tx start
						     * time-out
						     */
	unsigned int tx_abort_mid_bt_cnt;	    /* Tx abort due to BT during
						     * WLAN txn
						     */
	unsigned int tx_abort_dac_underrun_cnt;     /* Tx abort due to DAC
						     * under-run only
						     */
	unsigned int tx_ofdm_symbols_master;
	unsigned int tx_ofdm_symbols_slave1;
	unsigned int tx_ofdm_symbols_slave2;
	unsigned int tx_dsss_symbols;
	unsigned int cts_received_mcp_cnt;
	/*MAC Stats*/
	unsigned int roc_start;
	unsigned int roc_stop;
	unsigned int roc_complete;
	unsigned int roc_stop_complete;
	/* TX related */
	unsigned int tx_cmd_cnt; /* Num of TX commands received from host */
	unsigned int tx_done_cnt; /* Num of Tx done events sent to host */
	unsigned int tx_edca_trigger_cnt; /* Num times EDCA engine was
					   * triggered
					   */
	unsigned int tx_edca_isr_cnt; /* Num of times EDCA ISR was generated */
	unsigned int tx_start_cnt; /* Num of TX starts to MCP */
	unsigned int tx_abort_cnt; /* Num of TX aborts detected */
	unsigned int tx_abort_isr_cnt; /* Num of TX aborts received from MCP */
	unsigned int tx_underrun_cnt; /* Num of under-runs */
	unsigned int tx_rts_cnt; /* Num of RTS frames Txd */
	unsigned int tx_ampdu_cnt; /* Num of AMPDUs txd incremented by 1 for
				    * each A-MPDU (consisting of one or more
				    * MPDUs)
				    */
	unsigned int tx_mpdu_cnt; /* Num of MPDUs txd  incremented by 1 for
				   * MPDU (1 for each A-MPDU subframe)
				   */
	/* RX related */
	unsigned int rx_isr_cnt; /* Num of RX ISRs */
	unsigned int rx_ack_cts_to_cnt; /* Num of timeouts ACK */
	unsigned int rx_cts_cnt; /* Num of CTS frames received */
	unsigned int rx_ack_resp_cnt; /* Num of Ack frames received */
	unsigned int rx_ba_resp_cnt; /* Num of BA frames received */
	unsigned int rx_fail_in_ba_bitmap_cnt; /* Num of BA frames indicating at
						* least one failure in the BA
						* bitmap
						*/
	unsigned int rx_circular_buffer_free_cnt; /* Num of entries returned to
						   * RX circular buffers
						   */
	unsigned int rx_mic_fail_cnt; /* Num of MIC failures */

	/* HAL related */
	unsigned int hal_cmd_cnt; /* Num of commands received by HAL from the
				   * host
				   */
	unsigned int hal_event_cnt; /* Num of events sent by HAL to the host */
	unsigned int hal_ext_ptr_null_cnt; /* Num of packets dropped due to lack
					    * of Ext Ram buffers from host
					    */
	unsigned int deagg_bufOverFlow_cnt; /* RX buffer overflow count */
	unsigned int hal_int_ptr_null_cnt;
	/*RF Calibration Data*/
	unsigned int rf_calib_data_length;
	unsigned char rf_calib_data[MAX_RF_CALIB_DATA];
	unsigned int pdout_val;
	unsigned char rpu_lmac_version[8];
};


struct tx_pkt_info {
	struct sk_buff_head pkt;
	unsigned int hdr_len;
	unsigned int queue;
	unsigned int vif_index;
	unsigned int rate[4];
	unsigned int retries[4];
	unsigned int curr_retries;
	unsigned int max_retries;
	int roc_peer_id;
	int peer_id;
	bool adjusted_rates;
	struct ieee80211_tx_rate img_latest_rates[4];
};


struct tx_config {
	/* Used to protect the TX pool */
	spinlock_t lock;

	/* Used to store tx tokens(buff pool ids) */
	unsigned long buf_pool_bmp[(NUM_TX_DESCS/TX_DESC_BUCKET_BOUND) + 1];

	unsigned int outstanding_tokens[NUM_ACS];
	unsigned int next_spare_token_ac;

	/* Used to store the address of pending skbs per ac */
	struct sk_buff_head pending_pkt[MAX_RPU_VIF_CHANCTX_TYPES]
				       [MAX_PEND_Q_PER_AC]
				       [NUM_ACS];

	/* Peer which has the opportunity to xmit next on a queue */
	unsigned int curr_peer_opp[MAX_CHANCTX + MAX_OFF_CHANCTX][NUM_ACS];

	/* Used to store the address of tx'ed skb and len of 802.11 hdr
	 * it will be used in tx complete.
	 */
	int desc_chan_map[NUM_TX_DESCS];
	struct tx_pkt_info pkt_info[MAX_CHANCTX + MAX_OFF_CHANCTX]
				   [NUM_TX_DESCS];

	unsigned int queue_stopped_bmp;
	struct sk_buff_head proc_tx_list[NUM_TX_DESCS];
};

enum device_state {
	STOPPED = 0,
	STARTED
};

enum tid_aggr_state {
	TID_STATE_INVALID = 0,
	TID_STATE_AGGR_START,
	TID_STATE_AGGR_STOP,
	TID_STATE_AGGR_OPERATIONAL
};

#define TID_INITIATOR_STA 0x0000
#define TID_INITIATOR_AP 0x0010

struct sta_tid_info {
	unsigned short ssn;
	enum tid_aggr_state tid_state;
};


struct current_channel {
	unsigned int pri_chnl_num;
	unsigned int center_freq1;
	unsigned int center_freq2;
	unsigned int freq_band;
	unsigned int ch_width;
};

struct roc_params {
	unsigned char roc_in_progress;
	unsigned int roc_type;
	bool need_offchan;
	atomic_t roc_mgmt_tx_count;
};

struct img_priv {
	struct proc_dir_entry *umac_proc_dir_entry;
	struct device *dev;
	struct mac_address if_mac_addresses[MAX_VIFS];
	unsigned int current_vif_count;
	unsigned int active_vifs;
	struct mutex mutex;
	int state;
	int txpower;
	unsigned char mc_filters[MCST_ADDR_LIMIT][6];
	int mc_filter_count;

	struct tasklet_struct proc_tx_tasklet;
	/*ROC Work*/
	struct delayed_work roc_complete_work;
	struct roc_params roc_params;
	struct current_channel cur_chan;
	struct tx_config tx;
	struct sk_buff_head pending_pkt[NUM_ACS];

	/* Regulatory stuff */
	char alpha2[2]; /* alpha2 country code */
	struct wifi_params *params;
	struct wifi_stats  *stats;
	char name[20];
	char scan_abort_done;
	char cancel_hw_roc_done;
	char cancel_roc;
	char chan_prog_done;
	char reset_complete;
	char tx_deinit_complete;
	int power_save; /* Will be set only when a single VIF in
			 * STA mode is active
			 */
	struct ieee80211_vif *vifs[MAX_VIFS];
	struct ieee80211_sta *peers[MAX_PEERS];
	struct ieee80211_hw *hw;
	struct sta_tid_info  tid_info[32];
	spinlock_t bcast_lock; /* Used to ensure more_frames bit is set properly
				* when transmitting bcast frames in AP in IBSS
				* modes
				*/
	spinlock_t roc_lock;
	unsigned char tx_antenna;
	unsigned char tx_last_beacon;
	unsigned int rts_threshold;
	spinlock_t chanctx_lock;
	struct ieee80211_chanctx_conf *chanctx[MAX_CHANCTX];
	struct umac_chanctx *off_chanctx[MAX_OFF_CHANCTX];
	int roc_off_chanctx_idx;
	int curr_chanctx_idx;
	int num_active_chanctx;
};

extern struct wifi_dev *wifi;
struct wifi_dev {
	struct proc_dir_entry *umac_proc_dir_entry;
	struct wifi_params params;
	struct wifi_stats stats;
	struct ieee80211_hw *hw;
};

struct edca_params {
	unsigned short txop; /* units of 32us */
	unsigned short cwmin;/* units of 2^n-1 */
	unsigned short cwmax;/* units of 2^n-1 */
	unsigned char aifs;
	unsigned char uapsd;
};

struct umac_vif {
	struct timer_list bcn_timer;
	struct uvif_config {
		unsigned int atim_window;
		unsigned int aid;
		unsigned int bcn_lost_cnt;
		struct edca_params edca_params[NUM_ACS];
	} config;

	unsigned int noa_active;
	struct sk_buff_head noa_que;
	unsigned int noa_tx_allowed;

	int vif_index;
	struct ieee80211_vif *vif;
	struct img_priv *priv;
	unsigned char bssid[ETH_ALEN];
	unsigned int peer_ampdu_factor;

	/*Global Sequence no for non-qos and mgmt frames/vif*/
	__u16 seq_no;

	struct list_head list;
	struct umac_chanctx *chanctx;
	struct umac_chanctx *off_chanctx;
};

struct umac_sta {
	int index;
	int vif_index;
	struct umac_chanctx *chanctx;
};

struct umac_chanctx {
	int index;

	struct list_head vifs;
	short nvifs;
};


struct curr_peer_info {
	int id;
	int op_chan_idx;
};


void rpu_proc_ch_sw_event(struct rpu_event_ch_switch *ch_sw_info,
				  void *context);

extern int  rpu_core_init(struct img_priv *priv, unsigned int ftm);
extern void rpu_core_deinit(struct img_priv *priv, unsigned int ftm);
extern void rpu_vif_add(struct umac_vif  *uvif);
extern void rpu_vif_remove(struct umac_vif *uvif);
extern void rpu_vif_set_edca_params(unsigned short queue,
					    struct umac_vif *uvif,
					    struct edca_params *params,
					    unsigned int vif_active);
extern void rpu_vif_bss_info_changed(struct umac_vif *uvif,
					     struct ieee80211_bss_conf
					     *bss_conf, unsigned int changed);
extern int  rpu_tx_frame(struct sk_buff *skb,
				 struct ieee80211_sta *sta,
				 struct img_priv *priv,
				 int curr_chanctx_idx,
				 bool bcast);
extern int __rpu_tx_frame(struct img_priv *priv,
				  unsigned int queue,
				  unsigned int token_id,
				  int curr_chanctx_idx,
				  unsigned int more_frames,
				  bool retry);
extern void rpu_tx_init(struct img_priv *priv);
extern void rpu_tx_deinit(struct img_priv *priv);
void rpu_tx_proc_send_pend_frms_all(struct img_priv *priv,
					   int chan_id);
extern void proc_bss_info_changed(unsigned char *mac_addr, int value);
extern void packet_generation(unsigned long data);
extern int wait_for_reset_complete(struct img_priv *priv);

extern int rpu_tx_proc_pend_frms(struct img_priv *priv,
				   int queue,
				   int curr_chanctx_idx,
				   int token_id);
void free_token(struct img_priv *priv,
		int token_id,
		int queue);

struct curr_peer_info get_curr_peer_opp(struct img_priv *priv,
		      int curr_chanctx_idx,
		      int queue);

int rpu_flush_vif_queues(struct img_priv *priv,
			     struct umac_vif *uvif,
			     int chanctx_idx,
			     unsigned int hw_queue_map,
			     enum RPU_VIF_CHANCTX_TYPE vif_chanctx_type,
			     bool drop);

int rpu_discard_sta_pend_q(struct img_priv *priv,
				   struct umac_vif *uvif,
				   int peer_id,
				   unsigned int hw_queue_map);

int rpu_discard_sta_tx_q(struct img_priv *priv,
				   struct umac_vif *uvif,
				   int peer_id,
				   unsigned int hw_queue_map,
				   int chanctx_idx);
/* Beacon TimeStamp */
__s32 __attribute__((weak)) frc_to_atu(__u32 frccnt, __u64 *patu, s32 dir);
int __attribute__((weak)) get_evt_timer_freq(unsigned int *mask,
						unsigned int *num,
						unsigned int *denom);

int tx_queue_map(int queue);
int tx_queue_unmap(int queue);

extern unsigned char *rf_params_vpd;
extern int num_streams_vpd;

static __always_inline long param_get_val(unsigned char *buf,
			  unsigned char *str,
			  unsigned long *val)
{
	unsigned char *temp;

	if (strstr(buf, str)) {
		temp = strstr(buf, "=") + 1;
		/*To handle the fixed rate 5.5Mbps case*/
		if (!strncmp(temp, "5.5", 3)) {
			*val = 55;
			return 1;
		} else if (!kstrtoul(temp, 0, val)) {
			return 1;
		} else {
			return 0;
		}
	} else {
		return 0;
	}
}

static __always_inline long param_get_sval(unsigned char *buf,
			   unsigned char *str,
			   long *val)
{

	unsigned char *temp;

	if (strstr(buf, str)) {
		temp = strstr(buf, "=") + 1;
		/*To handle the fixed rate 5.5Mbps case*/
		if (!strncmp(temp, "5.5", 3)) {
			*val = 55;
			return 1;
		} else if (!kstrtol(temp, 0, val)) {
			return 1;
		} else {
			return 0;
		}
	} else {
		return 0;
	}

}

static __always_inline long param_get_match(unsigned char *buf,
				unsigned char *str)
{

	if (strstr(buf, str))
		return 1;
	else
		return 0;
}

static __always_inline char *get_string_from_rate(int rate,
						   unsigned int flags)
{
	if (rate == -1)
		return "Disabled";

	if (flags & ENABLE_VHT_FORMAT)
		return "VHT";

	if (flags & ENABLE_11N_FORMAT)
		return "HT";

	return "Legacy";
}

static __always_inline bool check_valid_rate_flags(struct img_priv *priv,
						   unsigned long val)
{
	bool ret = false;

	do {
		/*Only first 6 flags are defined currently*/
		if (val > 63) {
			pr_err("Invalid parameter value");
			break;
		}

		if ((val & ENABLE_VHT_FORMAT) &&
		    (val & ENABLE_11N_FORMAT)) {
			pr_err("Cannot set HT and VHT both.");
			break;
		}

		if ((val & ENABLE_CHNL_WIDTH_40MHZ) &&
		    (val & ENABLE_CHNL_WIDTH_80MHZ)) {
			pr_err("Cannot set 40 and 80 both.");
			break;
		}

		if ((priv->params->rpu_max_nss == 1)  &&
		    (val & ENABLE_SGI) &&
		    (val & ENABLE_GREEN_FIELD)) {
			pr_err("Cannot set GreenField when SGI is enabled for SISO");
			break;
		}
		ret = true;
	} while (0);

	return ret;
}

static __always_inline bool check_valid_data_rate(struct img_priv *priv,
						  int dr,
						  enum ptype type)
{
	bool is_mcs = dr & 0x80;
	bool ret = false;
	unsigned int rate;
	unsigned int nss;

	if (type == UCAST) {
		rate = priv->params->fixed_rate_flags;
		nss  = priv->params->num_spatial_streams;
	} else {
		rate = priv->params->mgd_mode_mcast_fixed_rate_flags;
		nss  = priv->params->mgd_mode_mcast_fixed_nss;
	}

	if (dr == -1)
		return true;

	if (is_mcs) {
		dr = dr & 0x7F;
		if (vht_support && (rate &
				    ENABLE_VHT_FORMAT)) {
			if ((dr >= 0 && dr <= 9))
				ret = true;
			else
				pr_err("Invalid VHT MCS value: %d\n", dr);
		} else
		if (rate & ENABLE_11N_FORMAT) {
			if (nss == 2) {
				if ((dr >= 8) && (dr <= 15))
					ret = true;
				else
					pr_err("Invalid MIMO HT MCS: %d\n",
					       dr);
			}
			if (nss == 1) {
				if ((dr >= 0) && (dr <= 7))
					ret = true;
				else
					pr_err("Invalid SISO HT MCS: %d\n",
					       dr);
			}
		}

	} else {
		if (priv->params->dot11g_support == 1 &&
		    ((dr == 1) ||
		     (dr == 2) ||
		     (dr == 55) ||
		     (dr == 11))) {
			ret = true;
		} else if ((dr == 6) ||
			   (dr == 9) ||
			   (dr == 12) ||
			   (dr == 18) ||
			   (dr == 24) ||
			   (dr == 36) ||
			   (dr == 48) ||
			   (dr == 54) ||
			   (dr == -1)) {
			ret = true;
		} else
			pr_err("Invalid Legacy Rate value: %d\n", dr);
		if ((rate & ENABLE_11N_FORMAT)
		    || (rate & ENABLE_VHT_FORMAT)
		    ) {
			ret = false;
			pr_err("Invalid rate_flags for legacy: %d\n", dr);
		}
	}
	return ret;
}

static inline int vif_addr_to_index(unsigned char *addr,
				    struct img_priv *priv)
{
	int i;
	struct ieee80211_vif *vif = NULL;

	for (i = 0; i < MAX_VIFS; i++) {
		if (!((i < MAX_VIFS) && (priv->active_vifs & (1 << i))))
			continue;

		rcu_read_lock();
		vif = rcu_dereference(priv->vifs[i]);
		rcu_read_unlock();

		if (ether_addr_equal(addr, vif->addr))
			break;
	}

	if (i < priv->params->num_vifs)
		return i;
	else
		return -1;
}

static inline int ieee80211_is_unicast_robust_mgmt_frame(struct sk_buff *skb)
{
	struct ieee80211_hdr *hdr = (struct ieee80211_hdr *) skb->data;

	if (skb->len < 24 || is_multicast_ether_addr(hdr->addr1))
		return 0;

	return ieee80211_is_robust_mgmt_frame(skb);
}
static inline bool is_bufferable_mgmt_frame(struct ieee80211_hdr *hdr)
{
	__u16 fc = hdr->frame_control;
	/*TODO: Handle Individual Probe Response frame in IBSS*/
	if (ieee80211_is_action(fc) ||
		ieee80211_is_disassoc(fc) ||
		ieee80211_is_deauth(fc))
		return	true;

	return false;
}


static inline void img_ether_addr_copy(char *dst, const char *src)
{
	ether_addr_copy(dst, src);
}

int proc_init(struct proc_dir_entry ***main_dir_entry);
int rpu_init(void);
void rpu_exit(void);
void proc_exit(void);
void update_mcs_packet_stat(int mcs_rate_num,
				   int rate_flags,
				   struct img_priv *priv);
int rpu_proc_tx(struct img_priv *priv);
void rpu_unblock_all_frames(struct img_priv *priv,
					    int ch_id);
int load_fw(struct ieee80211_hw *hw);
void stop(struct ieee80211_hw *hw);
int start_prod_mode(struct img_priv *priv, unsigned int val);
int stop_prod_mode(struct img_priv *priv, unsigned int val);
int start_packet_gen(struct img_priv *priv, int sval);
int stop_packet_gen(struct img_priv *priv, int sval);
int get_rate_prod(struct cmd_tx_ctrl *txcmd,
		  struct img_priv *priv);
int img_resume(struct ieee80211_hw *hw);
int img_suspend(struct ieee80211_hw *hw,
		struct cfg80211_wowlan *wowlan);
void init_beacon(struct umac_vif *uvif);
void deinit_beacon(struct umac_vif *uvif);
void modify_beacon_params(struct umac_vif *uvif,
			  struct ieee80211_bss_conf *bss_conf);

int load_fw(struct ieee80211_hw *hw);
extern void read_mem_region(unsigned int, int);
#define RPU_READY_TIMEOUT 200
#define RPU_READY_TIMEOUT_TICKS msecs_to_jiffies(RPU_READY_TIMEOUT)
#define RPU_EVENT_RPU_READY 0xDEAD


#endif /* _CORE_H_ */
