/*
 * Copyright (c) 2011-2017 Imagination Technologies Ltd.
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "core.h"

void modify_beacon_params(struct umac_vif *uvif,
			  struct ieee80211_bss_conf *bss_conf)
{
	unsigned int bcn_int = 0;
	unsigned long bcn_tim_val = 0;
	int ret = 0;

	if (uvif->vif->bss_conf.enable_beacon == true) {

		bcn_int = bss_conf->beacon_int;
		bcn_tim_val =  msecs_to_jiffies(bcn_int - 10);

		mod_timer(&uvif->bcn_timer,
			  jiffies + bcn_tim_val);

		CALL_RPU(rpu_prog_vif_beacon_int,
			  uvif->vif_index,
			  uvif->vif->addr,
			  bcn_int);
	} else {
		del_timer(&uvif->bcn_timer);
	}
prog_rpu_fail:
	return;

}

static void vif_bcn_timer_expiry(unsigned long data)
{
	struct umac_vif *uvif = (struct umac_vif *)data;
	struct sk_buff *skb, *temp;
	struct sk_buff_head bcast_frames;

	if (uvif->vif->bss_conf.enable_beacon == false)
		return;

	if (uvif->vif->type == NL80211_IFTYPE_AP) {
		temp = skb = ieee80211_beacon_get(uvif->priv->hw, uvif->vif);

		if (!skb) {
			/* No beacon, so dont transmit braodcast frames*/
			goto reschedule_timer;
		}

		skb_queue_head_init(&bcast_frames);
		skb->priority = 1;
		skb_queue_tail(&bcast_frames, skb);

		skb = ieee80211_get_buffered_bc(uvif->priv->hw, uvif->vif);

		while (skb) {
			/* Hack: skb->priority is used to indicate more
			 * frames
			 */
			skb->priority = 1;
			skb_queue_tail(&bcast_frames, skb);
			temp = skb;
			skb = ieee80211_get_buffered_bc(uvif->priv->hw,
							uvif->vif);
		}

		if (temp)
			temp->priority = 0;

		spin_lock_bh(&uvif->priv->bcast_lock);

		while ((skb = skb_dequeue(&bcast_frames))) {
			/* For a Beacon queue we will let the frames pass
			 * through irrespective of the current channel context.
			 * The FW will take care of transmitting them in the
			 * appropriate channel. Hence pass the interfaces
			 * channel context instead of the actual current channel
			 * context.
			 */
			rpu_tx_frame(skb,
					     NULL,
					     uvif->priv,
					     uvif->chanctx->index,
					     true);
		}

		spin_unlock_bh(&uvif->priv->bcast_lock);

		if (uvif->vif->csa_active) {
			if (ieee80211_csa_is_complete(uvif->vif))
				ieee80211_csa_finish(uvif->vif);
		}
	} else {
		skb = ieee80211_beacon_get(uvif->priv->hw, uvif->vif);

		if (!skb)
			goto reschedule_timer;

		/* For a Beacon queue we will let the frames pass through
		 * irrespective of the current channel context. The FW will take
		 * care of transmitting them in the appropriate channel.  Hence
		 * pass the interfaces channel context instead of the actual
		 * current channel context.
		 */
		rpu_tx_frame(skb,
				     NULL,
				     uvif->priv,
				     uvif->chanctx->index,
				     true);

	}
reschedule_timer:
	return;

}

void init_beacon(struct umac_vif *uvif)
{

	init_timer(&uvif->bcn_timer);
	uvif->bcn_timer.data = (unsigned long)uvif;
	uvif->bcn_timer.function = vif_bcn_timer_expiry;

}

void deinit_beacon(struct umac_vif *uvif)
{
	del_timer(&uvif->bcn_timer);

}
