/*
 * Copyright (c) 2011-2017 Imagination Technologies Ltd.
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <asm/unaligned.h>

#include <linux/clk.h>
#include <linux/etherdevice.h>
#include <linux/iio/consumer.h>
#include <linux/interrupt.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/netdevice.h>
#include <linux/of.h>
#include <linux/of_net.h>
#include <linux/of_device.h>
#include <linux/proc_fs.h>
#include <linux/skbuff.h>
#include <linux/slab.h>
#include <linux/sort.h>
#include <linux/time.h>


#include "core.h"
#include "hal.h"
#include "hal_common.h"
#include "soc.h"

#define COMMAND_START_MAGIC 0xDEAD
#define MAX_ATTEMPTS_FOR_AUXADC_VALID 500
static int init_rx_buf(int pkt_desc,
		       unsigned int max_data_size,
		       dma_addr_t *dma_buf);
const char *hal_name = "RPU_WIFI_HAL";
static unsigned long shm_offset = HAL_SHARED_MEM_OFFSET;
module_param(shm_offset, ulong, S_IRUSR|S_IWUSR);

unsigned int hal_cmd_sent;
unsigned int hal_event_recv;
struct timer_list stats_timer;
unsigned int alloc_skb_failures;
unsigned int alloc_skb_dma_region;
unsigned int alloc_skb_priv_tx_region;
unsigned int alloc_skb_priv_rx_region;
unsigned int alloc_skb_priv_runtime_fails;

static unsigned int rpu_ddr_base;


/* for send and receive count. */
static unsigned long tx_cnt;
static unsigned long rx_cnt;
/*RPU_DEBUG_HAL */

bool waiting_for_rpu_ready;

unsigned char vif_macs[2][ETH_ALEN];
char *mac_addr;

module_param(mac_addr, charp, 0000);
MODULE_PARM_DESC(mac_addr, "Configure wifi base mac address");


/* Range check */
#define CHECK_EVENT_ADDR_RPU(x) ((x) >= HAL_RPU_GRAM_BASE && (x) <=\
				  (HAL_RPU_GRAM_BASE + \
				  hpriv->rpu_pkd_gram_len))

#define CHECK_EVENT_STATUS_ADDR_RPU(x) ((x) >= HAL_RPU_GRAM_BASE && (x) <=\
					 (HAL_RPU_GRAM_BASE + \
					 hpriv->rpu_pkd_gram_len))

#define CHECK_EVENT_LEN(x) ((x) < 0x5000)
#define CHECK_RX_PKT_CNT(x) ((x) >= 1 && (x) <= 16)
/* #define CHECK_SRC_PTR(x, y) ((x) >= (y) && (x) <= (y) +
 * HAL_HOST_BOUNCE_BUF_LEN)
 */
/* MAX_RX_BUFS */
#define CHECK_PKT_DESC(x) ((x) < (hpriv->rx_bufs_2k + hpriv->rx_bufs_12k))




static int hal_reset_hal_params(void)
{
	hpriv->cmd_cnt = COMMAND_START_MAGIC;
	hpriv->event_cnt = 0;
	return 0;
}


static int hal_ready(struct hal_priv *priv)
{
	unsigned int value = 0;

	/* Check the ACK register bit */
	value =  readl((void __iomem *)(HOST_TO_MTX_CMD_ADDR));

	if (value & BIT(MTX_HOST_INT_SHIFT))
		return 0;
	else
		return 1;
}


static void tx_tasklet_fn(unsigned long data)
{
	struct hal_priv *priv = (struct hal_priv *)data;
	struct sk_buff *skb;
	unsigned int value = 0;
	unsigned long start_addr;
	unsigned long start = 0;

	while (1) {
		if (skb_peek(&priv->txq) == NULL)
			break;
		skb = skb_dequeue(&priv->txq);

		if (skb == NULL)
			break;

		tx_cnt++;
		RPU_DEBUG_HAL("%s: tx_cnt=%ld cmd_cnt=0x%X event_cnt=0x%X\n",
				hal_name,
				tx_cnt,
				priv->cmd_cnt,
				priv->event_cnt);
		if (DUMP_HAL) {
			RPU_DEBUG_HAL("%s: xmit dump\n", hal_name);
			RPU_DEBUG_DUMP_HAL(" ", DUMP_PREFIX_NONE, 16, 1,
					 skb->data, skb->len, 1);
		}

		start = jiffies;

		while (!hal_ready(priv) &&
		     time_before(jiffies, start + msecs_to_jiffies(1000))) {
			cpu_relax();
		}

		if (!hal_ready(priv)) {
			pr_err("%s: Intf not ready for 1000ms, dropping cmd\n",
			       hal_name);
			dev_kfree_skb_any(skb);
			skb = NULL;
		}

		if (!skb)
			continue;

		if (priv->hal_disabled)
			break;

		/* Write the command buffer in GRAM */
		start_addr = readl((void __iomem *)
				   (hpriv->gram_mem_addr +
				   HAL_GRAM_CMD_START));

		RPU_DEBUG_HAL("%s: Command address = 0x%08x\n",
			 hal_name, (unsigned int)start_addr);

		start_addr -= HAL_RPU_GRAM_BASE;
		start_addr += ((priv->gram_mem_addr)-(priv->shm_offset));

		if ((start_addr < priv->gram_mem_addr) ||
		    (start_addr > (priv->gram_mem_addr + HAL_RPU_GRAM_LEN))) {
			pr_err("%s: Invalid cmd addr 0x%08x, dropping cmd\n",
			       hal_name, (unsigned int)start_addr);
			dev_kfree_skb_any(skb);
			skb = NULL;
			continue;
		}

		memcpy((unsigned char *)start_addr, skb->data, skb->len);

		writel(skb->len, (void __iomem *)(hpriv->gram_mem_addr +
						  HAL_GRAM_CMD_LEN));

		value = (unsigned int) (priv->cmd_cnt);
		value |= 0x7fff0000;
		writel(value, (void __iomem *)(HOST_TO_MTX_CMD_ADDR));
		priv->cmd_cnt++;
		hal_cmd_sent++;

		dev_kfree_skb_any(skb);
	}
}


static void _hal_send(struct hal_priv  *priv,
			  struct sk_buff   *skb)
{
	skb_queue_tail(&priv->txq, skb);
	tasklet_schedule(&priv->tx_tasklet);
}

static void _hal_send_head(struct hal_priv  *priv,
			  struct sk_buff   *skb)
{
	skb_queue_head(&priv->txq, skb);
	tasklet_schedule(&priv->tx_tasklet);
}


static void hal_send(void *msg,
		     void *payload,
		     unsigned int descriptor_id)
{

	struct sk_buff *skb, *tmp;
	struct sk_buff_head *skb_list;
	struct hal_hdr *hdr;
	unsigned long dcp_start_addr;
	unsigned int pkt = 0, desc_id = 0, frame_id = 0;
	struct hal_tx_data *hal_tx_data = NULL;
	struct buf_info *tx_buf_info = NULL;
	dma_addr_t dma_buf;

	if (payload) {
		hdr = (struct hal_hdr *)msg;
		skb_list = (struct sk_buff_head *)payload;

		desc_id = (descriptor_id) & 0x0000FFFF;

		skb_queue_walk_safe(skb_list, skb, tmp)
			{
			frame_id = (desc_id * NUM_FRAMES_IN_TX_DESC) + pkt;
			hal_tx_data = &hpriv->hal_tx_data[frame_id];
			tx_buf_info = &hpriv->tx_buf_info[frame_id];

			hal_tx_data->data_len = tx_buf_info->dma_buf_len;

			dma_buf = tx_buf_info->dma_buf;
			dma_buf -= rpu_ddr_base;
			/* [2:24] = 23 bits*/
			hal_tx_data->address = (dma_buf >> 2) & 0x007fffff;
			/*[0:1] = 2 bits*/
			hal_tx_data->offset = dma_buf & 0x00000003;
			/*[31:35] = 7 bits stored at MSB position*/
			hal_tx_data->msb = ((dma_buf) >> 25) << 17;
			pkt++;
			}

		dcp_start_addr = hpriv->gram_mem_addr +
				 HAL_GRAM_TX_DATA_START +
				 (desc_id * TX_DESC_HAL_SIZE);

		memcpy((void *)dcp_start_addr,
		       &hpriv->hal_tx_data[(desc_id * NUM_FRAMES_IN_TX_DESC)],
		       TX_DESC_HAL_SIZE);
	}
	_hal_send(hpriv, msg);

}

static void recv_tasklet_fn(unsigned long data)
{
	struct hal_priv *priv = (struct hal_priv *)data;
	struct sk_buff *skb;

	while ((skb = skb_dequeue(&priv->refillq))) {
		/* As we refilled the buffers, now pass them UP */
		priv->rcv_handler(skb);
	}
}

static void rx_tasklet_fn(unsigned long data)
{
	struct hal_priv *priv = (struct hal_priv *)data;
	struct sk_buff	*skb;
	unsigned char *buf;
	unsigned long temp;
	unsigned char *nbuff;
	struct event_hal *evnt;
	struct cmd_hal cmd_rx;
	struct sk_buff *nbuf, *rx_skb;
	unsigned char *cmd_data;
	unsigned int data_length;
	void __iomem *src_ptr;
	int count = 0;
	unsigned int pkt_desc = 0, max_data_size = MAX_DATA_SIZE_2K;
	dma_addr_t dma_buf = 0;
	unsigned long event_addr, event_status_addr, event_len;
	unsigned long dbg_src_ptr;

	struct buf_info *rx_buf_info = NULL;
	struct buf_info temp_rx_buf_info;
	int ret = 0;
	unsigned int payload_length, length;

	while ((skb = skb_dequeue(&priv->rxq))) {
		event_addr = *((unsigned long *)(skb->cb));
		event_status_addr = *((unsigned long *)(skb->cb + 4));
		event_len = *(unsigned long *)(skb->cb + 8);

		/* Range check */
		if (skb->len > event_len) {
			pr_err("%s: Err! skb->len=%d, event_len =%d\n",
			       __func__, skb->len, (int)event_len);
			dev_kfree_skb_any(skb);
			continue;
		}

		temp = event_addr;
		buf = skb_put(skb, event_len);
		memcpy(buf, (unsigned char *)temp, skb->len);

		/* Mark the buffer free */
		temp = event_status_addr;

		RPU_DEBUG_HAL("%s: Freeing event buffer at 0x%08x\n",
			 hal_name, (unsigned int)temp);

		*((unsigned long *)temp) = 0;

		rx_cnt++;
		RPU_DEBUG_HAL("%s:rx_cnt=%ld cmd_cnt=0x%X event_cnt=0x%X\n",
			 hal_name, rx_cnt, priv->cmd_cnt, priv->event_cnt);
		if (DUMP_HAL) {
			RPU_DEBUG_HAL("%s: recv dump\n", hal_name);
			RPU_DEBUG_DUMP_HAL(" ", DUMP_PREFIX_NONE, 16, 1,
						skb->data, skb->len, 1);
		}
		nbuff = skb->data;
		evnt = (struct event_hal *)nbuff;

		/* Message from HAL after the DMA completion,
		 * Fetch the buffer addrs from RPU HOST RAM
		 * Copy them to the skb
		 * Pass them up
		 * Refresh the RX descriptor in firmware
		 */
		if (evnt->hdr.id == 0xffffffff) {
			/* HAL_INTERNAL CMD */
			memset(&cmd_rx, 0, sizeof(struct cmd_hal));

			if (!CHECK_RX_PKT_CNT(evnt->rx_pkt_cnt)) {
				/* Range check */
				pr_err("%s: Error!!! rx_pkt_cnt = %d\n",
				       __func__, evnt->rx_pkt_cnt);
				dev_kfree_skb_any(skb);
				continue;
			}

			for (count = 0; count < evnt->rx_pkt_cnt; count++) {
				pkt_desc = evnt->rx_pkt_desc[count];

				/* Range check */
				if (!CHECK_PKT_DESC(pkt_desc)) {
					pr_err("%s: Error!!! pkt_desc = %d\n",
					       __func__, pkt_desc);

					/* Drop all the remaining buffers: As
					 * per Design They will not be reclaimed
					 * by FW.
					 */
					break;
				}

				if (pkt_desc < hpriv->rx_bufs_12k)
					max_data_size = MAX_DATA_SIZE_12K;
				else
					max_data_size = MAX_DATA_SIZE_2K;

				if (hpriv->rx_buf_info == NULL)
					break;

				rx_buf_info = hpriv->rx_buf_info + pkt_desc;

				memcpy(&temp_rx_buf_info,
				       rx_buf_info,
				       sizeof(struct buf_info));

				dma_unmap_single(NULL,
						 rx_buf_info->dma_buf,
						 rx_buf_info->dma_buf_len,
						 DMA_FROM_DEVICE);

				dma_buf = rx_buf_info->dma_buf;
				src_ptr = rx_buf_info->src_ptr;

				RPU_DEBUG_HAL("%s: dma_buf = 0x%08X\n",
					       hal_name,
					       (unsigned int)dma_buf);

				dbg_src_ptr = (unsigned long)src_ptr;

				RPU_DEBUG_HAL("%s: src_ptr = 0x%08X\n",
					       hal_name,
					       (unsigned int)dbg_src_ptr);

				if (DUMP_HAL) {
					RPU_DEBUG_HAL("DMA data dump:");
					RPU_DEBUG_HAL(" size=200\n");
					RPU_DEBUG_DUMP_HAL(" ",
							DUMP_PREFIX_NONE, 16,
							1, src_ptr, 200, 1);
				}
				rx_skb = temp_rx_buf_info.skb;
				/* Offset in UMAC_LMAC_MSG_HDR, points to
				 * payload_length
				 */

				/* 802.11hdr + payload Len*/
				payload_length = *(((unsigned int *)src_ptr) +
						   3);
				length = *(((unsigned int *)src_ptr) + 5);

				/* Control Info Len*/
				data_length = payload_length + length;

				/* Complete data length to be copied */
				RPU_DEBUG_HAL("%s: Payload Len =%d(0x%x), ",
					       hal_name,
					       payload_length,
					       payload_length);

				RPU_DEBUG_HAL("Len=%d(0x%x), ",
					       length,
					       length);

				RPU_DEBUG_HAL("Data Len = %d(0x%x)\n",
					      data_length,
					      data_length);
				if (data_length > max_data_size) {
					pr_err("Max length exceeded:");
					pr_err(" payload_len: %d len:%d",
						payload_length,
						length);
					pr_err(" data_len:%d desc:%d\n",
						data_length,
						pkt_desc);

					pr_err("%s: dma_buf = 0x%08X\n",
						hal_name,
						(unsigned int)dma_buf);
					dbg_src_ptr = (unsigned long)src_ptr;
					pr_err("%s: src_ptr = 0x%08X\n",
						hal_name,
						(unsigned int)dbg_src_ptr);

					pr_err("Event from LMAC:");
					print_hex_dump(KERN_DEBUG,
						       "",
						       DUMP_PREFIX_NONE,
						       16,
						       1,
						       skb->data,
						       skb->len,
						       1);

					pr_err("DMA Data from LMAC:");
					print_hex_dump(KERN_DEBUG,
						       "",
						       DUMP_PREFIX_NONE,
						       16,
						       1,
						       src_ptr,
						       200,
						       1);

					/* Do not send the packet UP,
					 * just refill the buffer
					 * and give it to HW, for
					 * non-DMA case give the same
					 * buffer.
					 */
					dma_map_single(NULL,
						       rx_buf_info->src_ptr,
						       max_data_size,
						       DMA_FROM_DEVICE);
					cmd_rx.rx_pkt_data.rx_pkt_cnt++;
					cmd_rx.rx_pkt_data.rx_pkt[count].desc =
						evnt->rx_pkt_desc[count];
					cmd_rx.rx_pkt_data.rx_pkt[count].ptr  =
						dma_buf - rpu_ddr_base;
					continue;
				}

				if (temp_rx_buf_info.dma_buf_priv) {
					memcpy(skb_put(rx_skb,
					       data_length),
					       src_ptr,
					       data_length);
				} else {
					skb_put(rx_skb, data_length);
				}

				ret = init_rx_buf(pkt_desc, max_data_size,
					    &dma_buf);

				if (ret < 0) {
					/* Drop Packet, refill same dma_buf
					 */
					memcpy(rx_buf_info,
					       &temp_rx_buf_info,
					       sizeof(struct buf_info));

					dma_map_single(NULL,
						       rx_buf_info->src_ptr,
						       max_data_size,
						       DMA_FROM_DEVICE);
					alloc_skb_priv_runtime_fails++;
				} else
					skb_queue_tail(&hpriv->refillq, rx_skb);
				cmd_rx.rx_pkt_data.rx_pkt_cnt++;
				cmd_rx.rx_pkt_data.rx_pkt[count].desc =
					evnt->rx_pkt_desc[count];
				cmd_rx.rx_pkt_data.rx_pkt[count].ptr =
					dma_buf - rpu_ddr_base;

			}

			if (cmd_rx.rx_pkt_data.rx_pkt_cnt != 0) {
				cmd_rx.hdr.id = 0xffffffff;

				/* Inform HAL about the newly allocated
				 * buffers
				 */
				nbuf = alloc_skb(sizeof(struct cmd_hal),
						 GFP_ATOMIC);
				if (nbuf) {
					cmd_data = skb_put(nbuf,
						   sizeof(struct cmd_hal));

					memcpy(cmd_data,
					       (unsigned char *)&cmd_rx,
					       sizeof(struct cmd_hal));
					hal_cmd_sent--;
					_hal_send_head(hpriv, nbuf);

				}
			}
			tasklet_schedule(&priv->recv_tasklet);

			/* Internal CMD, Free it */
			dev_kfree_skb_any(skb);

		} else	{
			/* MSG from LMAC, non-data*/
			hal_event_recv++;
			priv->rcv_handler(skb);
		}
	}
}


static void hal_register_callback(msg_handler handler)
{
	hpriv->rcv_handler = handler;
}


static irqreturn_t hal_irq_handler(int irq, void  *p)
{

	unsigned int value, rpu_ready_sig;
	unsigned long event_addr, event_status_addr, event_len;
	unsigned char spurious;
	struct sk_buff *skb;
	struct hal_priv *priv = (struct hal_priv *)p;
	int is_err = 0;

	spurious = 0;

	value = readl((void __iomem *)(MTX_TO_HOST_CMD_ADDR));

	if (!(value & 0x80000000)) {
		pr_debug("%s: No Interrupt from RPU\n", hal_name);
		return IRQ_NONE;
	}

	value &= 0x7fffffff;
	rpu_ready_sig = value;
	/* RPU READY Sleep Controller Interrupt
	 */
	if (rpu_ready_sig == 0x7DEADBAD) {
		pr_debug("%s:%d Got HP_READY Interrupt\n", __func__, __LINE__);
		waiting_for_rpu_ready = true;
	} else if (value == (0x7FFF0000 | priv->event_cnt)) {
		event_addr = readl((void __iomem *)(hpriv->gram_mem_addr +
						    HAL_GRAM_EVENT_START));
		event_status_addr =
			readl((void __iomem *)(hpriv->gram_mem_addr +
					       HAL_GRAM_EVENT_STATUS));
		event_len = readl((void __iomem *)(hpriv->gram_mem_addr +
						   HAL_GRAM_EVENT_LEN));

		/* Range check */
		if (!(CHECK_EVENT_ADDR_RPU(event_addr)) ||
		    !(CHECK_EVENT_STATUS_ADDR_RPU(event_status_addr)) ||
		    !CHECK_EVENT_LEN(event_len)) {
			pr_err("%s: Error!!! event_addr = 0x%08x\n",
			       __func__,
			       (unsigned int)event_addr);

			pr_err("%s: Error!!! event_len =%d\n",
			       __func__,
			       (int)event_len);

			pr_err("%s: Error!!! event_status_addr = 0x%08x\n",
			       __func__,
			       (unsigned int)event_status_addr);

			is_err = 1;
		}
		RPU_DEBUG_HAL("%s: event address = 0x%08x\n",
			hal_name,
			(unsigned int)event_addr);
		RPU_DEBUG_HAL("%s: event status address = 0x%08x\n",
			hal_name,
			(unsigned int)event_status_addr);
		RPU_DEBUG_HAL("%s: event len = %d\n",
			hal_name,
			(int)event_len);

		if (unlikely(is_err)) {
			/* If addr is valid try to clear */
			if (CHECK_EVENT_STATUS_ADDR_RPU(event_status_addr)) {
				event_status_addr -= HAL_RPU_GRAM_BASE;
				event_status_addr += ((priv->gram_mem_addr) -
						      (priv->shm_offset));
				*((unsigned long *)event_status_addr) = 0;
			} else
				pr_err("%s: RPU status addr invalid, not clearing it\n",
				       hal_name);

			return IRQ_HANDLED;
		}

		event_addr -= HAL_RPU_GRAM_BASE;
		event_status_addr -= HAL_RPU_GRAM_BASE;
		event_addr += ((priv->gram_mem_addr) - (priv->shm_offset));
		event_status_addr += ((priv->gram_mem_addr) -
				      (priv->shm_offset));

		skb = dev_alloc_skb(event_len);

		if (!skb) {
			*((unsigned long *)event_status_addr) = 0;
		} else {
			*(unsigned long *)(skb->cb) = event_addr;

			/* Address of event payload */
			*(unsigned long *)(skb->cb + 4) = event_status_addr;

			/* Address to mark free */
			*(unsigned long *)(skb->cb + 8) = event_len;

			/* Length of event payload */
			skb_queue_tail(&priv->rxq, skb);
			tasklet_schedule(&priv->rx_tasklet);
		}

		priv->event_cnt++;
	} else {
		spurious = 1;
	}

	if (!spurious) {
		/* Clear the rpu interrupt */
		value = 0;
		value |= BIT(MTX_INT_CLR_SHIFT);
		writel(*((unsigned long   *)&(value)),
		(void __iomem *)(HOST_TO_MTX_ACK_ADDR));
	} else {
		pr_warn("%s: Spurious interrupt received\n", hal_name);

	}
	return IRQ_HANDLED;
}


static void hal_enable_int(void)
{
	unsigned int   value = 0;

	/* Set external pin irq enable for host_irq and rpu_irq */
	value = readl((void __iomem *)SYS_INT_ENAB_ADDR);
	value |= BIT(SYS_INT_MTX_IRQ_ENAB_SHIFT);

	writel(*((unsigned long   *)&(value)),
	       (void __iomem *)(SYS_INT_ENAB_ADDR));

	/* Enable raising rpu_int when RPU_INT = 1 */
	value = 0;
	value |= BIT(MTX_INT_EN_SHIFT);
	writel(*((unsigned long *)&(value)),
	       (void __iomem *)(MTX_INT_ENABLE_ADDR));
}


static void hal_disable_int(void)
{
	unsigned int value = 0;

	/* Reset external pin irq enable for host_irq and rpu_irq */
	value = readl((void __iomem *)SYS_INT_ENAB_ADDR);
	value &= ~(BIT(SYS_INT_MTX_IRQ_ENAB_SHIFT));
	writel(*((unsigned long   *)&(value)),
	       (void __iomem *)(SYS_INT_ENAB_ADDR));

	/* Disable raising rpu_int when RPU_INT = 1 */
	value = 0;
	value &= ~(BIT(MTX_INT_EN_SHIFT));
	writel(*((unsigned long *)&(value)),
	       (void __iomem *)(MTX_INT_ENABLE_ADDR));
}


int chg_irq_register(int val)
{
	RPU_DEBUG_HAL("%s: change irq regist state %s.\n",
		 hal_name, ((val == 1) ? "ON" : "OFF"));

	if (val == 0) {
		/* Unregister irq handler */
		free_irq(hpriv->irq, hpriv);

	} else if (val == 1) {
		/* Register irq handler */
		if (request_irq(hpriv->irq,
				hal_irq_handler,
				hpriv->irq_flags,
				"wlan",
				hpriv) != 0) {
			return -1;
		}
	}

	return 0;
}



static ssize_t proc_write_hal_stats(struct file *file,
		const char __user    *buffer,
		size_t		     count,
		loff_t		     *ppos)
{
	char buf[50];

	if (count >= sizeof(buf))
		count = sizeof(buf)-1;

	if (copy_from_user(buf, buffer, count))
		return -EFAULT;
	buf[count] = '\0';

	return count;
}

static int proc_read_hal_stats(struct seq_file *m, void *v)
{
	seq_printf(m, "Alloc SKB Failures: %d\n",
		   alloc_skb_failures);

	seq_printf(m, "Alloc SKB in 60 MB DMA Region  %d\n",
		   alloc_skb_dma_region);

	seq_printf(m, "Alloc SKB in Priv 4 MB TX Region: %d\n",
		   alloc_skb_priv_tx_region);

	seq_printf(m, "Alloc SKB in Priv 4 MB RX Region: %d\n",
		   alloc_skb_priv_rx_region);

	seq_printf(m, "Alloc SKB Run time: %d\n", alloc_skb_priv_runtime_fails);

	seq_printf(m, "hal_cmd_sent_cnt: %d\n",
		   hal_cmd_sent);

	seq_printf(m, "hal_event_recv_cnt: %d\n",
		   hal_event_recv);

	return 0;
}


static int proc_open_hal_stats(struct inode *inode, struct file *file)
{
	return single_open(file, proc_read_hal_stats, NULL);
}


static const struct file_operations params_fops_hal_stats = {
	.open = proc_open_hal_stats,
	.read = seq_read,
	.llseek = seq_lseek,
	.write = proc_write_hal_stats,
	.release = single_release
};


static int hal_proc_init(struct proc_dir_entry *hal_proc_dir_entry)
{
	struct proc_dir_entry *entry;
	int err = 0;

	entry = proc_create("hal_stats",
			    0444,
			    hal_proc_dir_entry,
			    &params_fops_hal_stats);

	if (!entry) {
		pr_err("Failed to create HAL proc entry\n");
		err = -ENOMEM;
	}

	return err;
}




int hal_start(void)
{

	hpriv->hal_disabled = 0;

	/* Enable host_int and rpu_int */
	hal_enable_int();

	return 0;
}


int hal_stop(void)
{
	/* Disable host_int and rpu_irq */
	hal_disable_int();
	return 0;
}


/* Unmap and release all resoruces*/
static int cleanup_all_resources(void)
{
	/* Free RPU Host RAM */
	kfree(hpriv->base_addr_rpu_host_ram);
	hpriv->base_addr_rpu_host_ram = NULL;
	/* Free RPU HAL TX data */
	kfree(hpriv->hal_tx_data);
	hpriv->hal_tx_data = NULL;
	return 0;
}
static int hal_deinit(void *dev)
{
	struct sk_buff *skb;

	(void)(dev);

	_rpu_umac_if_exit();

	/* Free irq line */
	chg_irq_register(0);

	/* Kill the HAL tasklet */
	tasklet_kill(&hpriv->tx_tasklet);
	tasklet_kill(&hpriv->rx_tasklet);
	tasklet_kill(&hpriv->recv_tasklet);
	while ((skb = skb_dequeue(&hpriv->rxq)))
		dev_kfree_skb_any(skb);

	while ((skb = skb_dequeue(&hpriv->refillq)))
		dev_kfree_skb_any(skb);

	while ((skb = skb_dequeue(&hpriv->txq)))
		dev_kfree_skb_any(skb);
#ifndef CONFIG_PM_SLEEP
	if (soc_ops.program_soc_regs_deinit(dev) < 0)
		pr_err("soc register deinit failed\n");
#endif

	cleanup_all_resources();

	return 0;
}


static int hal_init(void *dev)
{
	struct proc_dir_entry *main_dir_entry;
	int err = 0;

	(void) (dev);

	hpriv->shm_offset =  shm_offset;

	if (hpriv->shm_offset != HAL_SHARED_MEM_OFFSET)
		RPU_DEBUG_HAL("%s: Using shared memory offset 0x%lx\n",
			 hal_name, hpriv->shm_offset);

	hpriv->gram_mem_addr = hpriv->gram_base_addr + hpriv->shm_offset;
	hpriv->rpu_mem_addr = hpriv->rpu_sysbus_base_addr +
			       HAL_RPU_CORE_REG_OFFSET;


	hpriv->base_addr_rpu_host_ram = kmalloc(HAL_HOST_BOUNCE_BUF_LEN,
						 GFP_DMA);

	if (!hpriv->base_addr_rpu_host_ram) {
		pr_err("%s: rpu host ram: failed to allocate memory\n",
			       hal_name);
		err = -ENOMEM;
		goto error;
	}
	soc_ops.program_rpu_dma_start(hpriv);

#ifndef CONFIG_PM_SLEEP
	err = soc_ops.program_soc_regs_init(dev);

	if (err < 0) {
		pr_err("%s:SOC register init failed\n", hal_name);
		goto free_bounce_buf;
	}
#endif
	/* Register irq handler */
	if (chg_irq_register(1)) {
		pr_err("%s: Unable to register Interrupt handler with kernel\n",
		       hal_name);
		err = -ENOMEM;
		goto free_bounce_buf;
	}

	/*Allocate space do update data pointers to DCP*/
	hpriv->hal_tx_data = kzalloc((NUM_TX_DESC * NUM_FRAMES_IN_TX_DESC *
				      sizeof(struct hal_tx_data)), GFP_KERNEL);

	if (!hpriv->hal_tx_data) {
		pr_err("%s: hal_tx_data: unable to allocate memory\n",
		       hal_name);
		err = -ENOMEM;
		goto free_bounce_buf;
	}
	/* Intialize HAL tasklets */
	tasklet_init(&hpriv->tx_tasklet,
		     tx_tasklet_fn,
		     (unsigned long)hpriv);
	tasklet_init(&hpriv->rx_tasklet,
		     rx_tasklet_fn,
		     (unsigned long)hpriv);
	tasklet_init(&hpriv->recv_tasklet,
		     recv_tasklet_fn,
		     (unsigned long)hpriv);
	skb_queue_head_init(&hpriv->rxq);
	skb_queue_head_init(&hpriv->txq);
	skb_queue_head_init(&hpriv->refillq);
	if (_rpu_umac_if_init(&main_dir_entry) < 0) {
		pr_err("%s: wlan_init failed\n", hal_name);
		err = -ENOMEM;
		goto free_tx_data;
	}

	err = hal_proc_init(main_dir_entry);

	if (err)
		goto umac_if_deinit;

	hpriv->cmd_cnt = COMMAND_START_MAGIC;
	hpriv->event_cnt = 0;

	return err;
umac_if_deinit:
	hal_deinit(NULL);
free_tx_data:
	kfree(hpriv->hal_tx_data);
	hpriv->hal_tx_data = NULL;
free_bounce_buf:
	kfree(hpriv->base_addr_rpu_host_ram);
	hpriv->base_addr_rpu_host_ram = NULL;
error:
	return err;
}

static void hal_deinit_bufs(void)
{
	int i = 0, j = 0;
	struct buf_info *info = NULL;

	tasklet_disable(&hpriv->rx_tasklet);
	tasklet_disable(&hpriv->recv_tasklet);

	if (hpriv->rx_buf_info) {
		for (i = 0; i < hpriv->rx_bufs_2k + hpriv->rx_bufs_12k; i++) {
			info = &hpriv->rx_buf_info[i];

			if (info->dma_buf) {
				dma_unmap_single(NULL,
						 info->dma_buf,
						 info->dma_buf_len,
						 DMA_FROM_DEVICE);

				info->dma_buf = 0;
				info->dma_buf_len = 0;
			}

			if (hpriv->rx_buf_info[i].skb) {
				kfree_skb(hpriv->rx_buf_info[i].skb);
				hpriv->rx_buf_info[i].skb = NULL;
			}
		}

		kfree(hpriv->rx_buf_info);
		hpriv->rx_buf_info = NULL;
	}

	if (hpriv->tx_buf_info) {
		for (i = 0; i < hpriv->tx_bufs; i++) {
			for (j = 0; i < NUM_FRAMES_IN_TX_DESC; i++) {
				info = &hpriv->tx_buf_info[i + j];

				if (info->dma_buf) {
					dma_unmap_single(NULL,
							 info->dma_buf,
							 info->dma_buf_len,
							 DMA_TO_DEVICE);

					info->dma_buf = 0;
					info->dma_buf_len = 0;
				}
			}
		}

		kfree(hpriv->tx_buf_info);
		hpriv->tx_buf_info = NULL;
	}

	hpriv->hal_disabled = 1;
	tasklet_enable(&hpriv->rx_tasklet);
	tasklet_enable(&hpriv->recv_tasklet);
}


static int hal_init_bufs(unsigned int tx_bufs,
			 unsigned int rx_bufs_2k,
			 unsigned int rx_bufs_12k,
			 unsigned int tx_max_data_size)
{
	struct cmd_hal cmd_rx;
	struct sk_buff *nbuf = NULL;
	unsigned int count = 0, cmd_count = 0, pkt_desc = 0;
	unsigned int rx_max_data_size;
	dma_addr_t dma_buf = 0;
	unsigned int cmd_buf_count = ((rx_bufs_2k + rx_bufs_12k) /
				      HAL_INT_EVENT_MAX_RX);
	int result = -1;


	hpriv->tx_bufs = tx_bufs;
	hpriv->rx_bufs_2k = rx_bufs_2k;
	hpriv->rx_bufs_12k = rx_bufs_12k;
	hpriv->max_data_size = tx_max_data_size;
	hpriv->tx_base_addr_rpu_host_ram = hpriv->base_addr_rpu_host_ram;
	hpriv->rx_base_addr_rpu_host_ram = hpriv->base_addr_rpu_host_ram +
		(tx_bufs * NUM_FRAMES_IN_TX_DESC * tx_max_data_size);

	if (((tx_bufs * NUM_FRAMES_IN_TX_DESC * tx_max_data_size) +
	     ((rx_bufs_2k * MAX_DATA_SIZE_2K + rx_bufs_12k *
	       MAX_DATA_SIZE_12K))) > HAL_HOST_BOUNCE_BUF_LEN) {
		pr_err("%s Cannot accomodate tx_bufs: %d, frames/desc: %d and rx_bufs_2k: %d rx_bufs_12k: %d in %d RPU Host RAM\n",
		       hal_name, tx_bufs, NUM_FRAMES_IN_TX_DESC,
		       rx_bufs_2k, rx_bufs_12k, HAL_HOST_BOUNCE_BUF_LEN);

		goto err;
	}

	hpriv->rx_buf_info = kzalloc(((rx_bufs_2k + rx_bufs_12k) *
				      sizeof(struct buf_info)), GFP_KERNEL);

	if (!hpriv->rx_buf_info) {
		pr_err("%s out of memory\n", hal_name);
		goto err;
	}

	hpriv->tx_buf_info = kzalloc((tx_bufs * NUM_FRAMES_IN_TX_DESC *
				      sizeof(struct buf_info)),
				     GFP_KERNEL);

	if (!hpriv->tx_buf_info) {
		pr_err("%s out of memory\n", hal_name);
		goto err;
	}


	for (cmd_count = 0; cmd_count < cmd_buf_count; cmd_count++) {
		memset(&cmd_rx, 0, sizeof(struct cmd_hal));


		for (count = 0; count < HAL_INT_EVENT_MAX_RX; count++,
		     pkt_desc++) {

			if (pkt_desc < hpriv->rx_bufs_12k)
				rx_max_data_size = MAX_DATA_SIZE_12K;
			else
				rx_max_data_size = MAX_DATA_SIZE_2K;

			RPU_DEBUG_HAL("%s: Loop :%d: rx_max_data_size: %d\n",
				 hal_name, cmd_count, rx_max_data_size);
			result = init_rx_buf(pkt_desc,
					     rx_max_data_size,
					     &dma_buf);

			if (result) {
				pr_err("%s Failed to initialize RX buf %d\n",
				       hal_name, pkt_desc);
				goto err;
			}

			cmd_rx.rx_pkt_data.rx_pkt_cnt++;
			cmd_rx.rx_pkt_data.rx_pkt[count].desc = pkt_desc;
			cmd_rx.rx_pkt_data.rx_pkt[count].ptr = dma_buf -
							       rpu_ddr_base;
		}

		cmd_rx.hdr.id = 0xFFFFFFFF;

		nbuf = alloc_skb(sizeof(struct cmd_hal), GFP_ATOMIC);

		if (!nbuf)
			goto err;

		memcpy(skb_put(nbuf, sizeof(struct cmd_hal)),
		       (unsigned char *)&cmd_rx, sizeof(struct cmd_hal));
		hal_cmd_sent--;
		_hal_send_head(hpriv, nbuf);
	}

	return 0;
err:
	if (nbuf) {
		kfree_skb(nbuf);
		nbuf = NULL;
	}

	hal_deinit_bufs();

	return -1;
}


int hal_map_tx_buf(int pkt_desc,
		   int frame_id,
		   unsigned char *data,
		   int len, dma_addr_t *phy_addr)
{
	unsigned int index = (pkt_desc * NUM_FRAMES_IN_TX_DESC) + frame_id;
	void __iomem  *tx_address = NULL;
	dma_addr_t dma_buf = 0;
	unsigned int dbg_dma_buf = 0;
	int i, j;
	dma_addr_t curr_buf = 0;

	/* For QoS Null frames we dont try to map the frame since the data len
	 * will be 0 and there is nothing for the FW to process
	 */
	if (len == 0)
		return 0;

	/* Sanity check */
	dma_buf = ((struct buf_info)(hpriv->tx_buf_info[index])).dma_buf;

	if (dma_buf) {
		pr_err("%s: Already mapped pkt descriptor: %d and frame: %d dma_buf: 0x%x dma_buf: 0x%x index: %d\n",
		       __func__,
		       pkt_desc,
		       frame_id,
		       (unsigned int)hpriv->tx_buf_info[index].dma_buf,
		       (unsigned int)dma_buf,
		       index);

		for (i = 0; i < NUM_TX_DESC; i++) {
			for (j = 0; j < NUM_FRAMES_IN_TX_DESC; j++) {
				RPU_DEBUG_HAL("%s: TX: descriptor: %d ",
					       __func__, i);
				curr_buf = hpriv->tx_buf_info[i + j].dma_buf;
				RPU_DEBUG_HAL("and frame: %d dma_buf: 0x%x\n",
					       j,
					       (unsigned int)curr_buf);
			}
		}

		for (i = 0; i < 80; i++) {
			dbg_dma_buf =
				(unsigned int)(hpriv->rx_buf_info[i].dma_buf);
			RPU_DEBUG_HAL("%s: RX: descriptor: %d dma_buf: 0x%x\n",
				       __func__,
				       i,
				       dbg_dma_buf);
		}

		return -1;
	}
	tx_address = data;
	alloc_skb_dma_region++;
	dma_buf = dma_map_single(NULL,
				 tx_address,
				 len,
				 DMA_TO_DEVICE);

	if (unlikely(dma_mapping_error(NULL,
				       dma_buf))) {
		pr_err("%s Unable to map DMA on TX\n", hal_name);
		return -1;
	}

	hpriv->tx_buf_info[index].dma_buf = dma_buf;

	hpriv->tx_buf_info[index].dma_buf_len = len;

	*phy_addr = dma_buf;
	return 0;
}


int hal_unmap_tx_buf(int pkt_desc, int frame_id)
{
	unsigned int index = (pkt_desc * NUM_FRAMES_IN_TX_DESC) + frame_id;

	/* For QoS Null frames we did not map the frame (since the data len
	 * will be 0 and there is nothing for the FW to process), hence no need
	 * to try and unmap
	 */
	if (!hpriv->tx_buf_info[index].dma_buf_len)
		return 0;

	/* Sanity check */
	if (!hpriv->tx_buf_info[index].dma_buf) {
		pr_err("%s called for unmapped pkt desc: %d , frame: %d\n",
		       __func__, pkt_desc, frame_id);
		return -1;
	}

	dma_unmap_single(NULL,
			 hpriv->tx_buf_info[index].dma_buf,
			 hpriv->tx_buf_info[index].dma_buf_len,
			 DMA_TO_DEVICE);

	memset(&hpriv->tx_buf_info[index], 0, sizeof(struct buf_info));

	return 0;
}

static int init_rx_buf(int pkt_desc,
		       unsigned int max_data_size,
		       dma_addr_t *dma_buf)
{
	struct sk_buff *rx_skb = NULL;
	void __iomem *src_ptr = NULL;

	memset(&hpriv->rx_buf_info[pkt_desc], 0, sizeof(struct buf_info));

	rx_skb = alloc_skb(max_data_size, GFP_ATOMIC);

	if (!rx_skb) {
		alloc_skb_failures++;
		return -1;
	}
	src_ptr = rx_skb->data;
	alloc_skb_dma_region++;
	*dma_buf = dma_map_single(NULL,
				  src_ptr,
				  max_data_size,
				  DMA_FROM_DEVICE);

	if (unlikely(dma_mapping_error(NULL,
				       *dma_buf))) {
		pr_err("%s Unable to map DMA on RX\n", hal_name);

		if (rx_skb)
			dev_kfree_skb_any(rx_skb);

		return -1;
	}

	hpriv->rx_buf_info[pkt_desc].skb = rx_skb;
	hpriv->rx_buf_info[pkt_desc].src_ptr = src_ptr;
	hpriv->rx_buf_info[pkt_desc].dma_buf = *dma_buf;
	hpriv->rx_buf_info[pkt_desc].dma_buf_len = max_data_size;

	return 0;
}


void wow_enable_irq_wake(void)
{
	enable_irq_wake(hpriv->irq);
}

void wow_disable_irq_wake(void)
{
	disable_irq_wake(hpriv->irq);
}

void set_mem_region(unsigned int addr)
{
	soc_ops.set_mem_region(addr);
}

void request_mem_regions(unsigned char **gram_addr,
			 unsigned char **sysbus_addr,
			 unsigned char **gram_b4_addr)
{

	*gram_addr = (unsigned char *)hpriv->gram_base_addr;
	*sysbus_addr = (unsigned char *)hpriv->rpu_sysbus_base_addr;
}

struct hal_ops_tag hal_ops = {
	.init = hal_init,
	.deinit	= hal_deinit,
	.start = hal_start,
	.stop = hal_stop,
	.register_callback = hal_register_callback,
	.send = hal_send,
	.init_bufs = hal_init_bufs,
	.deinit_bufs = hal_deinit_bufs,
	.map_tx_buf = hal_map_tx_buf,
	.unmap_tx_buf = hal_unmap_tx_buf,
	.reset_hal_params	= hal_reset_hal_params,
	.get_dev = hal_get_dev,
	.request_mem_regions = request_mem_regions,
	.set_mem_region = set_mem_region,
};
