/*
 * Copyright (c) 2011-2017 Imagination Technologies Ltd.
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <linux/proc_fs.h>
#include "core.h"
#include "hal_common.h"

int _rpu_umac_if_init(struct proc_dir_entry **main_dir_entry)
{
	int error;

	error = proc_init(&main_dir_entry);
	if (error)
		return error;

	error = rpu_init();

	return error;
}

void _rpu_umac_if_exit(void)
{
	rpu_exit();
	proc_exit();
}


int rpu_core_init(struct img_priv *priv, unsigned int ftm)
{
	int ret = 0;
	unsigned int reset_type = LMAC_ENABLE;

	RPU_DEBUG_MAIN("%s-UMAC: Init called\n", priv->name);
	spin_lock_init(&tsf_lock);
	rpu_if_init(priv, priv->name);

	/* Enable the LMAC, set defaults and initialize TX */
	priv->reset_complete = 0;

	pr_info("%s-UMAC: Reset (ENABLE)\n", priv->name);

	if (hal_ops.start()) {
		ret = -1;
		goto rpu_if_deinit;
	}

	if (hal_ops.init_bufs(NUM_TX_DESCS,
			      NUM_RX_BUFS_2K,
			      NUM_RX_BUFS_12K,
			      priv->params->max_data_size) < 0) {
		ret = -1;
		goto hal_stop;
	}

	if (ftm)
		CALL_RPU(rpu_prog_reset,
			  reset_type,
			  LMAC_MODE_FTM);
	else
		CALL_RPU(rpu_prog_reset,
			  reset_type,
			  LMAC_MODE_NORMAL);

	if (wait_for_reset_complete(priv) < 0) {
		ret = -1;
		goto hal_deinit_bufs;
	}

	CALL_RPU(rpu_prog_btinfo, priv->params->bt_state);

	CALL_RPU(rpu_prog_txpower, priv->txpower);

	rpu_tx_init(priv);


	return 0;
hal_deinit_bufs:
	hal_ops.deinit_bufs();
prog_rpu_fail:
hal_stop:
	hal_ops.stop();
rpu_if_deinit:
	rpu_if_deinit();
	return ret;
}


void rpu_core_deinit(struct img_priv *priv, unsigned int ftm)
{
	int ret = 0;

	RPU_DEBUG_MAIN("%s-UMAC: De-init called\n", priv->name);

	/* De initialize tx  and disable LMAC*/
	rpu_tx_deinit(priv);

	/* Disable the LMAC */
	priv->reset_complete = 0;
	pr_info("%s-UMAC: Reset (DISABLE)\n", priv->name);

	if (ftm)
		CALL_RPU(rpu_prog_reset,
			  LMAC_DISABLE,
			  LMAC_MODE_FTM);
	else
		CALL_RPU(rpu_prog_reset,
			  LMAC_DISABLE,
			  LMAC_MODE_NORMAL);

	wait_for_reset_complete(priv);

prog_rpu_fail:
	rpu_if_free_outstnding();

	hal_ops.stop();
	hal_ops.deinit_bufs();

	rpu_if_deinit();
}


void rpu_reset_complete(char *lmac_version, void *context)
{
	struct img_priv *priv = (struct img_priv *)context;

	memcpy(priv->stats->rpu_lmac_version, lmac_version, 5);
	priv->stats->rpu_lmac_version[5] = '\0';
	priv->reset_complete = 1;
}


void rpu_mib_stats(struct rpu_event_mib_stats *mib_stats,
			   void *context)
{
	struct img_priv *priv = (struct img_priv *)context;

	memcpy(&priv->stats->ed_cnt,
	       &mib_stats->ed_cnt,
	       sizeof(struct rpu_event_mib_stats));

}

void rpu_mac_stats(struct rpu_event_mac_stats *mac_stats,
			   void *context)
{
	struct img_priv *priv = (struct img_priv *)context;

	/* TX related */
	priv->stats->roc_start = mac_stats->roc_start;
	priv->stats->roc_stop = mac_stats->roc_stop;
	priv->stats->roc_complete = mac_stats->roc_complete;
	priv->stats->roc_stop_complete = mac_stats->roc_stop_complete;
	priv->stats->tx_cmd_cnt = mac_stats->tx_cmd_cnt;
	priv->stats->tx_done_cnt = mac_stats->tx_done_cnt;
	priv->stats->tx_edca_trigger_cnt = mac_stats->tx_edca_trigger_cnt;
	priv->stats->tx_edca_isr_cnt = mac_stats->tx_edca_isr_cnt;
	priv->stats->tx_start_cnt = mac_stats->tx_start_cnt;
	priv->stats->tx_abort_cnt = mac_stats->tx_abort_cnt;
	priv->stats->tx_abort_isr_cnt = mac_stats->tx_abort_isr_cnt;
	priv->stats->tx_underrun_cnt = mac_stats->tx_underrun_cnt;
	priv->stats->tx_rts_cnt = mac_stats->tx_rts_cnt;
	priv->stats->tx_ampdu_cnt = mac_stats->tx_ampdu_cnt;
	priv->stats->tx_mpdu_cnt = mac_stats->tx_mpdu_cnt;
	/* RX related */
	priv->stats->rx_isr_cnt = mac_stats->rx_isr_cnt;
	priv->stats->rx_ack_cts_to_cnt = mac_stats->rx_ack_cts_to_cnt;
	priv->stats->rx_cts_cnt = mac_stats->rx_cts_cnt;
	priv->stats->rx_ack_resp_cnt = mac_stats->rx_ack_resp_cnt;
	priv->stats->rx_ba_resp_cnt = mac_stats->rx_ba_resp_cnt;
	priv->stats->rx_fail_in_ba_bitmap_cnt =
		mac_stats->rx_fail_in_ba_bitmap_cnt;
	priv->stats->rx_circular_buffer_free_cnt =
		mac_stats->rx_circular_buffer_free_cnt;
	priv->stats->rx_mic_fail_cnt = mac_stats->rx_mic_fail_cnt;
	priv->stats->deagg_bufOverFlow_cnt = mac_stats->deagg_bufOverFlow_cnt;


	/* HAL related */
	priv->stats->hal_cmd_cnt = mac_stats->hal_cmd_cnt;
	priv->stats->hal_event_cnt = mac_stats->hal_event_cnt;
	priv->stats->hal_ext_ptr_null_cnt = mac_stats->hal_ext_ptr_null_cnt;
	priv->stats->hal_int_ptr_null_cnt = mac_stats->hal_int_ptr_null_cnt;


}
void rpu_rf_calib_data(struct rpu_event_rf_calib_data *rf_data,
			       void *context)
{
	struct img_priv  *priv = (struct img_priv *)context;

	if (rf_data->rf_calib_data_length > MAX_RF_CALIB_DATA) {
		printk_once("%s: RF calib data exceeded the max size: %d\n",
			    __func__,
			    MAX_RF_CALIB_DATA);
		return;
	}
	priv->stats->rf_calib_data_length = rf_data->rf_calib_data_length;
	memset(priv->stats->rf_calib_data, 0x00,
	       MAX_RF_CALIB_DATA);
	memcpy(priv->stats->rf_calib_data, rf_data->rf_calib_data,
	       rf_data->rf_calib_data_length);
}



void rpu_ch_prog_complete(int event,
				  struct rpu_event_ch_prog_complete *prog_ch,
				  void *context)
{
	struct img_priv *priv = (struct img_priv *)context;

	priv->chan_prog_done = 1;
}

