/*
 * Copyright (c) 2011-2017 Imagination Technologies Ltd.
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <net/cfg80211.h>
#include <net/mac80211.h>

#include "core.h"
#include "p2p.h"
#include "utils.h"

#define RPU_DEBUG_P2P(fmt, ...)        \
do {                                         \
	if (rpu_debug & RPU_DEBUG_P2P)  \
		pr_debug(fmt, ##__VA_ARGS__);  \
} while (0)



void rpu_roc_complete_work(struct work_struct *work)
{
	struct delayed_work *dwork = NULL;
	struct img_priv *priv = NULL;
	struct umac_chanctx *off_chanctx = NULL;
	struct umac_vif *uvif = NULL, *tmp = NULL;
	struct tx_config *tx = NULL;
	u32 roc_queue = 0;
	bool need_offchan;
	int roc_off_chanctx_idx = -1;
	int chan_id = 0;

	dwork = container_of(work, struct delayed_work, work);
	priv = container_of(dwork, struct img_priv, roc_complete_work);
	tx = &priv->tx;

	if (priv->roc_params.roc_in_progress == 0)
		return;

	mutex_lock(&priv->mutex);
	need_offchan = priv->roc_params.need_offchan;

	roc_queue = tx_queue_unmap(RPU_ROC_AC);
	roc_off_chanctx_idx = priv->roc_off_chanctx_idx;

	/* Stop the ROC queue */
	ieee80211_stop_queue(priv->hw, roc_queue);
	/* Unlock RCU immediately as we are freeing off_chanctx in this funciton
	 * only and because flush_vif_queues sleep
	 */
	rcu_read_lock();
	off_chanctx = rcu_dereference(priv->off_chanctx[roc_off_chanctx_idx]);
	rcu_read_unlock();

	list_for_each_entry_safe(uvif, tmp, &off_chanctx->vifs, list) {
		if (uvif == NULL || uvif->off_chanctx  == NULL)
			continue;
		/* Flush the TX queues */
		rpu_flush_vif_queues(priv,
					 uvif,
					 uvif->off_chanctx->index,
					 BIT(RPU_ROC_AC),
					 RPU_VIF_CHANCTX_TYPE_OFF,
					 TX_DROP);


		spin_lock_bh(&tx->lock);
		spin_lock(&priv->chanctx_lock);

		/* ROC DONE: Move the channel context */
		if (uvif->chanctx)
			priv->curr_chanctx_idx = uvif->chanctx->index;
		else
			priv->curr_chanctx_idx = -1;

		spin_unlock(&priv->chanctx_lock);
		spin_unlock_bh(&tx->lock);

		if (need_offchan) {
			/* DEL from OFF chan list */
			list_del_init(&uvif->list);
			if (uvif->chanctx) {
				/* Add it back to OP chan list */
				list_add_tail(&uvif->list,
					      &uvif->chanctx->vifs);

				/* !need_offchan: In this case, the frames are
				 * transmitted, so trigger is not needed.
				 *
				 * need_offchan: In this case, frames are
				 * buffered so we need trigger in case no frames
				 * come from mac80211.
				 */
				/* Process OPER pending frames only.
				 * TXQ is flushed before start of ROC
				 */
				chan_id = uvif->chanctx->index;
				rpu_tx_proc_send_pend_frms_all(priv,
								       chan_id);
			}
			off_chanctx->nvifs--;
		}
		uvif->off_chanctx = NULL;
	}

	if (need_offchan)
		kfree(off_chanctx);


	rcu_assign_pointer(priv->off_chanctx[roc_off_chanctx_idx], NULL);
	priv->roc_off_chanctx_idx = -1;
	priv->roc_params.roc_in_progress = 0;

	if (priv->cancel_roc == 0) {
		ieee80211_remain_on_channel_expired(priv->hw);
		RPU_DEBUG_ROC("%s:%d ROC STOPPED..\n", __func__, __LINE__);
	} else {
		priv->cancel_hw_roc_done = 1;
		priv->cancel_roc = 0;
		RPU_DEBUG_ROC("%s:%d ROC CANCELLED..\n", __func__, __LINE__);
	}

	/* Start the ROC queue */
	ieee80211_wake_queue(priv->hw, roc_queue);
	mutex_unlock(&priv->mutex);
}

int remain_on_channel(struct ieee80211_hw *hw,
			     struct ieee80211_vif *vif,
			     struct ieee80211_channel *channel,
			     int duration,
			     enum ieee80211_roc_type type)
{
	struct img_priv *priv = (struct img_priv *)hw->priv;
	unsigned int pri_chnl_num =
		ieee80211_frequency_to_channel(channel->center_freq);
	struct umac_vif *uvif = (struct umac_vif *)vif->drv_priv;
	struct umac_chanctx *off_chanctx = NULL;
	int off_chanctx_id = 0, i = 0;
	struct tx_config *tx = &priv->tx;
	u32 hw_queue_map = 0;
	struct ieee80211_chanctx_conf *vif_chanctx;
	bool need_offchan = true;
	int ret = 0;

	mutex_lock(&priv->mutex);
	RPU_DEBUG_ROC("%s:%d The Params are:",
					__func__,
					__LINE__);
	RPU_DEBUG_ROC(" channel:%d duration:%d type: %d\n",
			ieee80211_frequency_to_channel(channel->center_freq),
			duration,
			type);

	if (priv->roc_params.roc_in_progress) {
		RPU_DEBUG_ROC("%s:%d Dropping roc...Busy\n",
				__func__,
				__LINE__);
		mutex_unlock(&priv->mutex);
		return -EBUSY;
	}

	if (priv->num_active_chanctx == 2) {
		RPU_DEBUG_ROC("%s:%d RoC is not supported in TSMC Mode\n",
				__func__,
				__LINE__);
		mutex_unlock(&priv->mutex);
		return -ENOTSUPP;
	}

	/* Inform FW that ROC is started:
	 * For pure TX we send OFFCHANNEL_TX so that driver can terminate ROC
	 * For Tx + Rx we use NORMAL, FW will terminate ROC based on duration.
	 */
	if (duration != 10 && type == ROC_TYPE_OFFCHANNEL_TX)
		type = ROC_TYPE_NORMAL;

	/* uvif is in connected state
	 */
	if (uvif->chanctx) {
		rcu_read_lock();

		vif_chanctx =
			rcu_dereference(priv->chanctx[uvif->chanctx->index]);

		/* AS ROC frames are MGMT frames, checking only for Primary
		 * Channel.
		 */
		if (vif_chanctx->def.chan->center_freq == channel->center_freq)
			need_offchan = false;

		rcu_read_unlock();
	}

	RPU_DEBUG_ROC("%s:%d need_offchan: %d\n",
			__func__,
			__LINE__,
			need_offchan);
	priv->roc_params.need_offchan = need_offchan;

	if (need_offchan) {
		/* Different chan context than the uvif */
		off_chanctx = kmalloc(sizeof(struct umac_chanctx),
				      GFP_KERNEL);

		if (!off_chanctx) {
			pr_err("%s: Unable to alloc mem for channel context\n",
			       __func__);
			mutex_unlock(&priv->mutex);
			return -ENOMEM;
		}

		/** Currently OFFCHAN is limited to handling ROC case
		 *  but it is meant for a generic case.
		 *  ideally we should look for existing offchan context
		 *  and re-use/create.
		 */
		for (i = 0; i < MAX_OFF_CHANCTX; i++) {
			if (!priv->off_chanctx[i]) {
				off_chanctx_id = i;
				break;
			}
		}

		if (uvif->chanctx) {
			ieee80211_stop_queues(hw);

			hw_queue_map = BIT(WLAN_AC_BK) |
				BIT(WLAN_AC_BE) |
				BIT(WLAN_AC_VI) |
				BIT(WLAN_AC_VO) |
				BIT(WLAN_AC_BCN);

			rpu_flush_vif_queues(priv,
					uvif,
					uvif->chanctx->index,
					hw_queue_map,
					RPU_VIF_CHANCTX_TYPE_OPER,
					TX_FLUSH);
		}


		off_chanctx->index = OFF_CHANCTX_IDX_BASE + off_chanctx_id;
		priv->roc_off_chanctx_idx = off_chanctx_id;
		INIT_LIST_HEAD(&off_chanctx->vifs);
		off_chanctx->nvifs = 0;

		if (uvif->chanctx) {
			/* Delete the uvif from OP channel list */
			list_del_init(&uvif->list);
		}
		/* Add the vif to the off_chanctx */
		list_add_tail(&uvif->list, &off_chanctx->vifs);
		off_chanctx->nvifs++;
		rcu_assign_pointer(priv->off_chanctx[off_chanctx_id],
				   off_chanctx);
		synchronize_rcu();


		/* Move the channel context */
		spin_lock_bh(&priv->chanctx_lock);
		priv->curr_chanctx_idx = off_chanctx->index;
		spin_unlock_bh(&priv->chanctx_lock);
	} else {
		/* Same channel context, just update off_chanctx
		 * to chanctx
		 */
		off_chanctx = uvif->chanctx;

		for (i = 0; i < MAX_OFF_CHANCTX; i++) {
			if (!priv->off_chanctx[i]) {
				off_chanctx_id = i;
				break;
			}
		}
		priv->roc_off_chanctx_idx = off_chanctx->index;
		rcu_assign_pointer(priv->off_chanctx[off_chanctx_id],
				   off_chanctx);
		synchronize_rcu();
	}
	spin_lock_bh(&tx->lock);
	uvif->off_chanctx = off_chanctx;
	spin_unlock_bh(&tx->lock);
	CALL_RPU(rpu_prog_roc,
		  ROC_START,
		  pri_chnl_num,
		  duration,
		  type);

	if (uvif->chanctx)
		ieee80211_wake_queues(hw);

prog_rpu_fail:
	mutex_unlock(&priv->mutex);
	return ret;
}


int cancel_remain_on_channel(struct ieee80211_hw *hw)
{
	struct img_priv *priv = (struct img_priv *)hw->priv;
	int ret = 0;

	mutex_lock(&priv->mutex);

	if (priv->roc_params.roc_in_progress) {
		priv->cancel_hw_roc_done = 0;
		priv->cancel_roc = 1;
		RPU_DEBUG_ROC("%s:%d Cancelling HW ROC....\n",
				__func__, __LINE__);
		CALL_RPU(rpu_prog_roc, ROC_STOP, 0, 0, 0);

		mutex_unlock(&priv->mutex);

		if (!wait_for_cancel_hw_roc(priv)) {
			RPU_DEBUG_ROC("%s:%d Cancel HW ROC....done\n",
							__func__,
							__LINE__);
			ret = 0;
		} else {
			RPU_DEBUG_ROC("%s:%d Cancel HW ROC..timedout\n",
							__func__,
							__LINE__);
			ret = -1;
		}
	}
prog_rpu_fail:
	mutex_unlock(&priv->mutex);
	return ret;
}


void rpu_noa_event(int event, struct rpu_event_noa *noa, void *context,
			   struct sk_buff *skb)
{
	struct img_priv  *priv = (struct img_priv *)context;
	struct ieee80211_vif *vif;
	struct umac_vif *uvif;
	bool transmit = false;
	int curr_chanctx_idx = -1;

	rcu_read_lock();

	vif = (struct ieee80211_vif *)rcu_dereference(
					priv->vifs[noa->if_index]);

	if (vif == NULL) {
		rcu_read_unlock();
		return;
	}

	uvif = (struct umac_vif *)vif->drv_priv;

	spin_lock_bh(&uvif->noa_que.lock);

	if (event == FROM_TX) {
		if (uvif->noa_active) {
			if (!uvif->noa_tx_allowed || skb_peek(&uvif->noa_que))
				__skb_queue_tail(&uvif->noa_que, skb);
			else
				transmit = true;
		} else
			transmit = true;
	} else if (event == FROM_TX_DONE) {
		if (uvif->noa_active && uvif->noa_tx_allowed) {
			skb = __skb_dequeue(&uvif->noa_que);

			if (skb)
				transmit = true;
		}
	} else { /* event = FROM_EVENT_NOA */

		uvif->noa_active = noa->noa_active;

		if (uvif->noa_active) {
			RPU_DEBUG_P2P("%s: noa active = %d, ",
					priv->name, noa->noa_active);
			RPU_DEBUG_P2P("ap_present = %d\n",
					noa->ap_present);

			uvif->noa_tx_allowed = noa->ap_present;

			if (uvif->noa_tx_allowed) {
				skb = __skb_dequeue(&uvif->noa_que);
				if (skb)
					transmit = true;
			}
		} else {
			RPU_DEBUG_P2P("%s: noa active = %d\n",
				 priv->name, noa->noa_active);

			uvif->noa_tx_allowed = 1;

			/* Can be done in a better way. For now, just flush the
			 * NoA Queue
			 */
			while ((skb = __skb_dequeue(&uvif->noa_que)))
				dev_kfree_skb_any(skb);
		}
	}

	spin_unlock_bh(&uvif->noa_que.lock);

	rcu_read_unlock();

	if (transmit) {
		spin_lock_bh(&priv->chanctx_lock);
		curr_chanctx_idx = priv->curr_chanctx_idx;
		spin_unlock_bh(&priv->chanctx_lock);
		rpu_tx_frame(skb,
				     NULL,
				     priv,
				     curr_chanctx_idx,
				     false);
	}
}

