/*
 * Copyright (c) 2011-2017 Imagination Technologies Ltd.
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <linux/proc_fs.h>

#include "core.h"
#include "utils.h"
#include "version.h"
#include "fwldr.h"

struct wifi_dev *wifi;


static char *rpu_get_vif_name(int vif_idx)
{
	struct img_priv *priv = ((struct img_priv *)(wifi->hw->priv));
	struct wireless_dev *wdev = NULL;
	struct ieee80211_vif *vif = NULL;

	if ((priv->active_vifs & (1 << vif_idx))) {
		rcu_read_lock();
		vif = rcu_dereference(priv->vifs[vif_idx]);
		rcu_read_unlock();

		wdev = ieee80211_vif_to_wdev(vif);
		return wdev->netdev->name;
	}

	return NULL;
}


int ddr_test_64mb_boundary(unsigned int start_addr, unsigned int len)
{
	void __iomem *host_ram_addr;
	unsigned int physical_addr, data_start_addr, i;
	unsigned char j = 0;

	/* Map the reserved Host RAM for RPU */
	if (!request_mem_region(start_addr,
				len,
				"host_rpu_ram")) {
		pr_err("%s: request_mem_region failed.\n", __func__);
		goto free_hpriv;
	}
	/*pr_debug("ioremapped: 0x%x\n", start_addr + 60 * 1024 * 1024);*/
	host_ram_addr = ioremap(start_addr, 70 * 1024 * 1024);
	pr_debug("ioremapped: b4:0x%x after: %p\n", start_addr, host_ram_addr);

	if (host_ram_addr == 0) {
		pr_err("%s: ioremap failed\n", __func__);
		release_mem_region(start_addr, len);
		goto free_hpriv;
	} else {
		physical_addr = virt_to_phys(host_ram_addr);
		pr_debug("%p = physical 0x%x\n",  host_ram_addr,
			  physical_addr);

		data_start_addr = physical_addr + (64 * 1024 * 1024) - 1000;
		pr_debug("data_start_addr=0x%x\n", data_start_addr);


		for (i = 0; i < 500; i++) {
			*((unsigned char *) (data_start_addr + i)) = j++;
			if (j > 255)
				j = 0;
		}
		pr_debug("Dumping Data\n");
		print_hex_dump(KERN_DEBUG,
			       "",
			       DUMP_PREFIX_NONE,
			       16,
			       1,
			       (unsigned char *)data_start_addr,
			       1500,
			       1);
	}

	return 0;
free_hpriv:
	return -1;
}

static int proc_read_config(struct seq_file *m, void *v)
{
	int i = 0;
	int rf_params_size = sizeof(wifi->params.rf_params) /
			     sizeof(wifi->params.rf_params[0]);
	struct img_priv *priv;
	unsigned int nss, flags;
	int mcs;

	if (!wifi->hw)
		return -ENODEV;

	priv = ((struct img_priv *)(wifi->hw->priv));

	seq_puts(m, "************* Configurable Parameters ***********\n");
	seq_printf(m, "dot11g_support = %d\n", wifi->params.dot11g_support);
	seq_printf(m, "dot11a_support = %d\n", wifi->params.dot11a_support);
	/*RF Input params*/
	seq_puts(m, "rf_params =");
	for (i = 0; i < rf_params_size; i++)
		seq_printf(m, " %02X", wifi->params.rf_params[i]);

	seq_puts(m, "\n");

	seq_puts(m, "rf_params_vpd =");
	for (i = 0; i < rf_params_size; i++)
		seq_printf(m, " %02X", wifi->params.rf_params_vpd[i]);

	seq_puts(m, "\n");
	seq_printf(m, "bypass_vpd = %d\n", wifi->params.bypass_vpd);
	seq_printf(m, "production_test = %d\n", wifi->params.production_test);
	seq_printf(m, "rpu_max_nss (RPU Init) = %d\n",
		   wifi->params.rpu_max_nss);
	seq_printf(m, "enable_early_agg_checks = %d\n",
		   wifi->params.enable_early_agg_checks);
	seq_printf(m, "antenna_sel (RPU Init) = %d\n",
		   wifi->params.antenna_sel);
	seq_printf(m, "max_data_size = %d (%dK)\n",
		   wifi->params.max_data_size,
		   wifi->params.max_data_size/1024);
	seq_printf(m, "max_tx_cmds = %d\n",
		   wifi->params.max_tx_cmds);
	seq_printf(m, "disable_power_save (Disables all power save's) = %d\n",
		   wifi->params.disable_power_save);
	seq_printf(m, "disable_sm_power_save (Disables MIMO PS only) = %d\n",
		   wifi->params.disable_sm_power_save);

	seq_printf(m, "num_vifs = %d\n",
		   wifi->params.num_vifs);

	seq_printf(m, "chnl_40_offset= %d\n", wifi->params.chnl_40_offset);
	if (vht_support)
		seq_printf(m, "chnl_80_offset= %d\n",
			   wifi->params.chnl_80_offset);
	seq_printf(m, "rate_protection_type = %d (0: Disable, 1: Enable)\n",
		   wifi->params.rate_protection_type);
	if (wifi->params.production_test) {
		mcs = wifi->params.tx_fixed_mcs_indx;
		flags = wifi->params.fixed_rate_flags;
		seq_puts(m, "***Production Mode Rate config options\n");
		seq_printf(m, "tx_fixed_mcs_indx = %d (%s)\n",
			   mcs, get_string_from_rate(mcs, flags));
		if (wifi->params.tx_fixed_rate > -1) {
			if (wifi->params.tx_fixed_rate == 55)
				seq_puts(m, "tx_fixed_rate = 5.5\n");
			else
				seq_printf(m, "tx_fixed_rate = %d\n",
					   wifi->params.tx_fixed_rate);
		} else
			seq_printf(m, "tx_fixed_rate = %d\n",
				   wifi->params.tx_fixed_rate);
	} else {
		seq_puts(m, "***Unicast Fixed rate config options\n");
		mcs = wifi->params.mgd_mode_tx_fixed_mcs_indx;
		flags = wifi->params.fixed_rate_flags;
		seq_printf(m, "mgd_mode_tx_fixed_mcs_indx = %d (%s)\n",
			   mcs, get_string_from_rate(mcs, flags));
		if (wifi->params.mgd_mode_tx_fixed_rate > -1) {
			if (wifi->params.mgd_mode_tx_fixed_rate == 55)
				seq_puts(m, "mgd_mode_tx_fixed_rate = 5.5\n");
			else
				seq_printf(m, "mgd_mode_tx_fixed_rate = %d\n",
					   wifi->params.mgd_mode_tx_fixed_rate);
		} else
			seq_printf(m, "mgd_mode_tx_fixed_rate = %d\n",
				   wifi->params.mgd_mode_tx_fixed_rate);
	}

	if (wifi->params.fixed_rate_flags & ENABLE_11N_FORMAT)
		nss = wifi->params.mgd_mode_tx_fixed_mcs_indx/8 + 1;
	else
		nss = wifi->params.num_spatial_streams;

	seq_puts(m, "Bits:80MHz-VHT-11N-SGI-40MHz-GF\n");
	seq_printf(m, "fixed_rate_flags = %d\n",
		   wifi->params.fixed_rate_flags);
	seq_printf(m, "fixed_rate_preamble_type (0: Short, 1: Long) = %d\n",
		   wifi->params.fixed_rate_preamble_type);
	seq_printf(m, "fixed_rate_stbc_enabled = %d\n",
		   wifi->params.fixed_rate_stbc_enabled);
	seq_printf(m, "fixed_rate_bcc_or_ldpc = %d\n",
		   wifi->params.fixed_rate_bcc_or_ldpc);
	seq_printf(m, "num_spatial_streams = %d\n", nss);

	seq_puts(m, "***Multicast Fixed rate config options\n");

	mcs = wifi->params.mgd_mode_mcast_fixed_data_rate;
	flags = wifi->params.mgd_mode_mcast_fixed_rate_flags;
	if ((mcs != -1) && (mcs & 0x80))
		mcs = mcs & 0x7F;

	if (flags & ENABLE_11N_FORMAT)
		nss = mcs/8 + 1;
	else
		nss = wifi->params.mgd_mode_mcast_fixed_nss;


	seq_printf(m, "mgd_mode_mcast_fixed_data_rate = %d (%s)\n",
		   mcs, get_string_from_rate(mcs, flags));
	seq_puts(m, "Bits:80MHz-VHT-11N-SGI-40MHz-GF\n");
	seq_printf(m, "mgd_mode_mcast_fixed_rate_flags = %d\n",
		   wifi->params.mgd_mode_mcast_fixed_rate_flags);
	seq_printf(m, "mgd_mode_mcast_fixed_preamble = %d\n",
		   wifi->params.mgd_mode_mcast_fixed_preamble);
	seq_printf(m, "mgd_mode_mcast_fixed_bcc_or_ldpc = %d\n",
		   wifi->params.mgd_mode_mcast_fixed_bcc_or_ldpc);
	seq_printf(m, "mgd_mode_mcast_fixed_stbc_enabled = %d\n",
		   wifi->params.mgd_mode_mcast_fixed_stbc_enabled);
	seq_printf(m, "mgd_mode_mcast_fixed_nss = %d\n", nss);
	seq_puts(m, "***\n");

	seq_printf(m, "vht_beamformer_enable = %d\n",
		   wifi->params.vht_beamform_enable);
	seq_printf(m, "vht_beamformer_period = %dms\n",
		   wifi->params.vht_beamform_period);
	seq_printf(m, "bg_scan_enable = %d\n",
		   wifi->params.bg_scan_enable);
	seq_puts(m, "bg_scan_channel_list =");

	for (i = 0; i < wifi->params.bg_scan_num_channels;  i++) {
		if (wifi->params.bg_scan_channel_list[i])
			seq_printf(m, " %d",
				   wifi->params.bg_scan_channel_list[i]);
	}

	seq_puts(m, "\n");
	seq_puts(m, "bg_scan_channel_flags =");

	for (i = 0; i < wifi->params.bg_scan_num_channels;  i++) {
		if (wifi->params.bg_scan_channel_flags[i])
			seq_printf(m, " %d",
				   wifi->params.bg_scan_channel_flags[i]);
	}

	seq_puts(m, "\n");
	seq_printf(m, "bg_scan_intval = %dms\n",
		   wifi->params.bg_scan_intval/1000);
	seq_printf(m, "bg_scan_num_channels = %d\n",
		   wifi->params.bg_scan_num_channels);
	seq_printf(m, "scan_type = %d (PASSIVE: 0, ACTIVE: 1)\n",
		   wifi->params.scan_type);
	seq_printf(m, "fw_loading = %d\n", wifi->params.fw_loading);
	seq_printf(m, "bt_state = %d\n", wifi->params.bt_state);

	/* Beacon Time Stamp */
	if (priv->state == STARTED) {
		unsigned int cnt;

		for (cnt = 0; cnt < MAX_VIFS; cnt++) {
			unsigned long long ts1;
			unsigned long long bssid, atu;
			int status;
			unsigned int t2;

			spin_lock_bh(&tsf_lock);
			ts1 = get_unaligned_le64(wifi->params.sync[cnt].ts1);
			bssid =
			    get_unaligned_le64(wifi->params.sync[cnt].bssid);
			status = wifi->params.sync[cnt].status;
			atu = wifi->params.sync[cnt].atu;
			t2 = wifi->params.sync[cnt].ts2;
			spin_unlock_bh(&tsf_lock);
			if (status)
				seq_printf(m,
					   "sync=%s %d %llu %llu %llx t2=%u\n",
					   rpu_get_vif_name(cnt),
					   status,
					   (unsigned long long)ts1,
					   atu,
					   (unsigned long long)bssid,
					   t2);
		}
	}
	seq_printf(m, "minstrel_fastpath= %s\n",
		   wifi->params.minstrel_fastpath ? "Enabled" : "Disabled");

	seq_puts(m, "****** Production Test (or) FTM Parameters *******\n");
	seq_printf(m, "start_packet_gen = %d (-1: Infinite loop)\n",
		   wifi->params.pkt_gen_val);
	seq_printf(m, "payload_length = %d bytes\n",
		   wifi->params.payload_length);
	seq_printf(m, "start_prod_mode = channel: %d\n",
		   wifi->params.start_prod_mode);
	seq_printf(m, "continuous_tx = %d\n",
		   wifi->params.cont_tx);

	if (ftm || wifi->params.production_test)
		seq_printf(m, "set_tx_power = %d dB\n",
			   wifi->params.set_tx_power);

	seq_printf(m, "center_frequency = %d\n",
		   ieee80211_frequency_to_channel(priv->cur_chan.center_freq1));

	if (ftm)
		seq_printf(m, "aux_adc_chain_id = %d\n",
			   wifi->params.aux_adc_chain_id);
	seq_puts(m, "RPU Sleep Type 0: Sleep Enabled, 1: Sleep Disabled)\n");
	seq_puts(m, "RPU Runtime Debug Support Configuration.\n");
	seq_printf(m, "rpu_debug = %d.\n", rpu_debug);
#define PRINT_DEBUG_MOD(MOD_NAME, MOD_ID) \
		seq_printf(m, "***rpu_debug: %s: val= %d, status= %s\n",\
			   MOD_NAME,\
			   MOD_ID,\
			   rpu_debug & MOD_ID ? "ENABLED" : "DISABLED")\

	PRINT_DEBUG_MOD("SCAN", RPU_DEBUG_SCAN);
	PRINT_DEBUG_MOD("ROC", RPU_DEBUG_ROC);
	PRINT_DEBUG_MOD("TX", RPU_DEBUG_TX);
	PRINT_DEBUG_MOD("MAIN", RPU_DEBUG_MAIN);
	PRINT_DEBUG_MOD("RPU_IF", RPU_DEBUG_IF);
	PRINT_DEBUG_MOD("UMACIF", RPU_DEBUG_UMACIF);
	PRINT_DEBUG_MOD("RX", RPU_DEBUG_RX);
	PRINT_DEBUG_MOD("HAL", RPU_DEBUG_HAL);
	PRINT_DEBUG_MOD("CRYPTO", RPU_DEBUG_CRYPTO);
	PRINT_DEBUG_MOD("TX_DUMP", RPU_DEBUG_DUMP_TX);
	PRINT_DEBUG_MOD("RX_DUMP", RPU_DEBUG_DUMP_RX);
	PRINT_DEBUG_MOD("HAL_DUMP", RPU_DEBUG_DUMP_HAL);
	PRINT_DEBUG_MOD("HAL_DUMP", RPU_DEBUG_DUMP_HAL);
	PRINT_DEBUG_MOD("RUNTIME_PM", RPU_DEBUG_RUNTIME_PM);
	PRINT_DEBUG_MOD("TSMC", RPU_DEBUG_TSMC);

	seq_puts(m, "HELP: Add the values beside each module and\n");
	seq_puts(m, " echo rpu_debug=<SUM> to enable logging\n");
	seq_puts(m, " for those modules.\n");

	seq_puts(m, "To see the updated stats\n");
	seq_puts(m, "please run: echo get_stats=1 > /proc/rpu/params\n");
	seq_puts(m, "To see the cleared phy stats\n");
	seq_puts(m, "please run: echo clear_stats=1 > /proc/rpu/params\n");
	seq_puts(m, "************* VERSION ***********\n");
	seq_printf(m, "RPU_DRIVER_VERSION = %s\n", RPU_DRIVER_VERSION);

	if (wifi->hw &&
	    (((struct img_priv *)(wifi->hw->priv))->state != STARTED)) {
		seq_printf(m, "LMAC_VERSION = %s\n", "UNKNOWN");
		seq_printf(m, "Firmware version = %s\n", "UNKNOWN");
	} else {
		seq_printf(m, "LMAC_VERSION = %s\n",
			   wifi->stats.rpu_lmac_version);
		seq_printf(m, "Firmware version= %d.%d\n",
			   (wifi->stats.rpu_lmac_version[0] - '0'),
			   (wifi->stats.rpu_lmac_version[2] - '0'));
	}

	return 0;
}


static int proc_read_phy_stats(struct seq_file *m, void *v)
{
	int i = 0;

	seq_puts(m, "************* BB Stats ***********\n");

	seq_printf(m, "ed_cnt=%d\n",
		   wifi->stats.ed_cnt);
	seq_printf(m, "mpdu_cnt=%d\n",
		   wifi->stats.mpdu_cnt);
	seq_printf(m, "ofdm_crc32_pass_cnt=%d\n",
		   wifi->stats.ofdm_crc32_pass_cnt);
	seq_printf(m, "ofdm_crc32_fail_cnt=%d\n",
		   wifi->stats.ofdm_crc32_fail_cnt);
	seq_printf(m, "dsss_crc32_pass_cnt=%d\n",
		   wifi->stats.dsss_crc32_pass_cnt);
	seq_printf(m, "dsss_crc32_fail_cnt=%d\n",
		   wifi->stats.dsss_crc32_fail_cnt);
	seq_printf(m, "mac_id_pass_cnt=%d\n",
		   wifi->stats.mac_id_pass_cnt);
	seq_printf(m, "mac_id_fail_cnt=%d\n",
		   wifi->stats.mac_id_fail_cnt);
	seq_printf(m, "ofdm_corr_pass_cnt=%d\n",
		   wifi->stats.ofdm_corr_pass_cnt);
	seq_printf(m, "ofdm_corr_fail_cnt=%d\n",
		   wifi->stats.ofdm_corr_fail_cnt);
	seq_printf(m, "dsss_corr_pass_cnt=%d\n",
		   wifi->stats.dsss_corr_pass_cnt);
	seq_printf(m, "dsss_corr_fail_cnt=%d\n",
		   wifi->stats.dsss_corr_fail_cnt);
	seq_printf(m, "ofdm_s2l_fail_cnt=%d\n",
		   wifi->stats.ofdm_s2l_fail_cnt);
	seq_printf(m, "lsig_fail_cnt=%d\n",
		   wifi->stats.lsig_fail_cnt);
	seq_printf(m, "htsig_fail_cnt=%d\n",
		   wifi->stats.htsig_fail_cnt);
	seq_printf(m, "vhtsiga_fail_cnt=%d\n",
		   wifi->stats.vhtsiga_fail_cnt);
	seq_printf(m, "vhtsigb_fail_cnt=%d\n",
		   wifi->stats.vhtsigb_fail_cnt);
	seq_printf(m, "nonht_ofdm_cnt=%d\n",
		   wifi->stats.nonht_ofdm_cnt);
	seq_printf(m, "nonht_dsss_cnt=%d\n",
		   wifi->stats.nonht_dsss_cnt);
	seq_printf(m, "mm_cnt=%d\n",
		   wifi->stats.mm_cnt);
	seq_printf(m, "gf_cnt=%d\n",
		   wifi->stats.gf_cnt);
	seq_printf(m, "vht_cnt=%d\n",
		   wifi->stats.vht_cnt);
	seq_printf(m, "aggregation_cnt=%d\n",
		   wifi->stats.aggregation_cnt);
	seq_printf(m, "non_aggregation_cnt=%d\n",
		   wifi->stats.non_aggregation_cnt);
	seq_printf(m, "ndp_cnt=%d\n",
		   wifi->stats.ndp_cnt);
	seq_printf(m, "ofdm_ldpc_cnt=%d\n",
		   wifi->stats.ofdm_ldpc_cnt);
	seq_printf(m, "ofdm_bcc_cnt=%d\n",
		   wifi->stats.ofdm_bcc_cnt);
	seq_printf(m, "midpacket_cnt=%d\n",
		   wifi->stats.midpacket_cnt);
	seq_printf(m, "dsss_sfd_fail_cnt=%d\n",
		   wifi->stats.dsss_sfd_fail_cnt);
	seq_printf(m, "dsss_hdr_fail_cnt=%d\n",
		   wifi->stats.dsss_hdr_fail_cnt);
	seq_printf(m, "dsss_short_preamble_cnt=%d\n",
		   wifi->stats.dsss_short_preamble_cnt);
	seq_printf(m, "dsss_long_preamble_cnt=%d\n",
		   wifi->stats.dsss_long_preamble_cnt);
	seq_printf(m, "sifs_event_cnt=%d\n",
		   wifi->stats.sifs_event_cnt);
	seq_printf(m, "cts_cnt=%d\n",
		   wifi->stats.cts_cnt);
	seq_printf(m, "ack_cnt=%d\n",
		   wifi->stats.ack_cnt);
	seq_printf(m, "sifs_no_resp_cnt=%d\n",
		   wifi->stats.sifs_no_resp_cnt);
	seq_printf(m, "unsupported_cnt=%d\n",
		   wifi->stats.unsupported_cnt);
	seq_printf(m, "l1_corr_fail_cnt=%d\n",
		   wifi->stats.l1_corr_fail_cnt);
	seq_printf(m, "sifs_crc_exit_cnt=%d\n",
		wifi->stats.sifs_crc_exit_cnt);
	seq_printf(m, "low_energy_event_cnt=%d\n",
		wifi->stats.low_energy_event_cnt);
	seq_printf(m, "deagg_error_cnt=%d\n",
		wifi->stats.deagg_error_cnt);
	seq_printf(m, "nsymbols_error_cnt=%d\n",
		wifi->stats.nsymbols_error_cnt);
	seq_printf(m, "mcs32_cnt=%d\n",
		wifi->stats.mcs32_cnt);
	seq_printf(m, "ndpa_cnt=%d\n",
		wifi->stats.ndpa_cnt);
	seq_printf(m, "lsig_duration_error_cnt=%d\n",
		wifi->stats.lsig_duration_error_cnt);
	seq_printf(m, "rts_cnt=%d\n",
		wifi->stats.rts_cnt);
	seq_printf(m, "non_ht_cts_cnt=%d\n",
		wifi->stats.non_ht_cts_cnt);
	seq_printf(m, "rxp_active_exit_cnt=%d\n",
		wifi->stats.rxp_active_exit_cnt);
	seq_printf(m, "beamform_feedback_cnt=%d\n",
		wifi->stats.beamform_feedback_cnt);
	seq_printf(m, "self_cts_cnt=%d\n",
		wifi->stats.self_cts_cnt);
	seq_printf(m, "pop_master_cnt=%d\n",
		wifi->stats.pop_master_cnt);
	seq_printf(m, "pop_error_cnt=%d\n",
		wifi->stats.pop_error_cnt);
	seq_printf(m, "multicast_cnt=%d\n",
		wifi->stats.multicast_cnt);
	seq_printf(m, "tx_ed_abort_cnt=%d\n",
		wifi->stats.tx_ed_abort_cnt);
	seq_printf(m, "mcp_cts_cnt=%d\n",
		wifi->stats.mcp_cts_cnt);
	seq_printf(m, "deagg_q_post_cnt=%d\n",
		wifi->stats.deagg_q_post_cnt);
	seq_printf(m, "rxp_active_exit_dsss_cnt=%d\n",
		wifi->stats.rxp_active_exit_dsss_cnt);
	seq_printf(m, "rxp_extreme_error_cnt=%d\n",
		wifi->stats.rxp_extreme_error_cnt);
	seq_printf(m, "aci_fail_cnt=%d\n",
		wifi->stats.aci_fail_cnt);
	/* TX stats*/
	seq_printf(m, "tx_pkts_from_lmac = %d\n",
		   wifi->stats.tx_pkts_from_lmac);
	seq_printf(m, "tx_pkts_tx2tx = %d\n",
		   wifi->stats.tx_pkts_tx2tx);
	seq_printf(m, "tx_pkts_from_rx = %d\n",
		   wifi->stats.tx_pkts_from_rx);
	seq_printf(m, "tx_pkts_ofdm = %d\n",
		   wifi->stats.tx_pkts_ofdm);
	seq_printf(m, "tx_pkts_dsss = %d\n",
		   wifi->stats.tx_pkts_dsss);
	seq_printf(m, "tx_pkts_reached_end_of_fsm = %d\n",
		   wifi->stats.tx_pkts_reached_end_of_fsm);
	seq_printf(m, "tx_unsupported_modulation = %d\n",
		   wifi->stats.tx_unsupported_modulation);
	seq_printf(m, "tx_latest_pkt_from_lmac_or_sifs = %d\n",
		   wifi->stats.tx_latest_pkt_from_lmac_or_sifs);
	seq_printf(m, "tx_abort_bt_confirm_cnt = %d\n",
		   wifi->stats.tx_abort_bt_confirm_cnt);
	seq_printf(m, "tx_abort_txstart_timeout_cnt = %d\n",
		   wifi->stats.tx_abort_txstart_timeout_cnt);
	seq_printf(m, "tx_abort_midBT_cnt = %d\n",
		   wifi->stats.tx_abort_mid_bt_cnt);
	seq_printf(m, "tx_abort_dac_underrun_cnt = %d\n",
		   wifi->stats.tx_abort_dac_underrun_cnt);
	seq_printf(m, "tx_ofdm_symbols_master = %d\n",
		   wifi->stats.tx_ofdm_symbols_master);
	seq_printf(m, "tx_ofdm_symbols_slave1 = %d\n",
		   wifi->stats.tx_ofdm_symbols_slave1);
	seq_printf(m, "tx_ofdm_symbols_slave2 = %d\n",
		   wifi->stats.tx_ofdm_symbols_slave2);
	seq_printf(m, "tx_dsss_symbols = %d\n",
		   wifi->stats.tx_dsss_symbols);

	seq_puts(m, "************* RF Stats ***********\n");
	/*RF output data*/
	seq_puts(m, "rf_calib_data =");
	for (i = 0; i < wifi->stats.rf_calib_data_length; i++)
		seq_printf(m, "%02X", wifi->stats.rf_calib_data[i]);

	seq_puts(m, "\n");
	return 0;
}

static int proc_read_mac_stats(struct seq_file *m, void *v)
{
	unsigned int index;
	unsigned int total_samples = 0;
	unsigned int total_value = 0;
	int total_rssi_samples = 0;
	int total_rssi_value = 0;
	struct img_priv *priv = NULL;

	if (!wifi->hw)
		return -ENODEV;

	priv = (struct img_priv *)(wifi->hw->priv);

	if (ftm) {
		for (index = 0; index < MAX_AUX_ADC_SAMPLES; index++) {
			if (!wifi->params.pdout_voltage[index])
				continue;

			total_value += wifi->params.pdout_voltage[index];
			total_samples++;
		}
	}

	for (index = 0; index < MAX_RSSI_SAMPLES; index++) {

		if (!wifi->params.production_test)
			break;

		if (!wifi->params.rssi_average[index])
			continue;

		total_rssi_value += wifi->params.rssi_average[index];
		total_rssi_samples++;
	}

	seq_puts(m, "************* UMAC STATS ***********\n");
	seq_printf(m, "rx_packet_mgmt_count = %d\n",
		   wifi->stats.rx_packet_mgmt_count);
	seq_printf(m, "rx_packet_data_count = %d\n",
		   wifi->stats.rx_packet_data_count);
	seq_printf(m, "tx_packet_count(HT MCS0) = %d\n",
		   wifi->stats.ht_tx_mcs0_packet_count);
	seq_printf(m, "tx_packet_count(HT MCS1) = %d\n",
		   wifi->stats.ht_tx_mcs1_packet_count);
	seq_printf(m, "tx_packet_count(HT MCS2) = %d\n",
		   wifi->stats.ht_tx_mcs2_packet_count);
	seq_printf(m, "tx_packet_count(HT MCS3) = %d\n",
		   wifi->stats.ht_tx_mcs3_packet_count);
	seq_printf(m, "tx_packet_count(HT MCS4) = %d\n",
		   wifi->stats.ht_tx_mcs4_packet_count);
	seq_printf(m, "tx_packet_count(HT MCS5) = %d\n",
		   wifi->stats.ht_tx_mcs5_packet_count);
	seq_printf(m, "tx_packet_count(HT MCS6) = %d\n",
		   wifi->stats.ht_tx_mcs6_packet_count);
	seq_printf(m, "tx_packet_count(HT MCS7) = %d\n",
		   wifi->stats.ht_tx_mcs7_packet_count);

	if (wifi->params.rpu_max_nss == 2) {
		seq_printf(m, "tx_packet_count(HT MCS8) = %d\n",
			   wifi->stats.ht_tx_mcs8_packet_count);
		seq_printf(m, "tx_packet_count(HT MCS9) = %d\n",
			   wifi->stats.ht_tx_mcs9_packet_count);
		seq_printf(m, "tx_packet_count(HT MCS10) = %d\n",
			   wifi->stats.ht_tx_mcs10_packet_count);
		seq_printf(m, "tx_packet_count(HT MCS11) = %d\n",
			   wifi->stats.ht_tx_mcs11_packet_count);
		seq_printf(m, "tx_packet_count(HT MCS12) = %d\n",
			   wifi->stats.ht_tx_mcs12_packet_count);
		seq_printf(m, "tx_packet_count(HT MCS13) = %d\n",
			   wifi->stats.ht_tx_mcs13_packet_count);
		seq_printf(m, "tx_packet_count(HT MCS14) = %d\n",
			   wifi->stats.ht_tx_mcs14_packet_count);
		seq_printf(m, "tx_packet_count(HT MCS15) = %d\n",
			   wifi->stats.ht_tx_mcs15_packet_count);
	}
	if (vht_support) {
		seq_printf(m, "tx_packet_count(VHT MCS0) = %d\n",
			   wifi->stats.vht_tx_mcs0_packet_count);
		seq_printf(m, "tx_packet_count(VHT MCS1) = %d\n",
			   wifi->stats.vht_tx_mcs1_packet_count);
		seq_printf(m, "tx_packet_count(VHT MCS2) = %d\n",
			   wifi->stats.vht_tx_mcs2_packet_count);
		seq_printf(m, "tx_packet_count(VHT MCS3) = %d\n",
			   wifi->stats.vht_tx_mcs3_packet_count);
		seq_printf(m, "tx_packet_count(VHT MCS4) = %d\n",
			   wifi->stats.vht_tx_mcs4_packet_count);
		seq_printf(m, "tx_packet_count(VHT MCS5) = %d\n",
			   wifi->stats.vht_tx_mcs5_packet_count);
		seq_printf(m, "tx_packet_count(VHT MCS6) = %d\n",
			   wifi->stats.vht_tx_mcs6_packet_count);
		seq_printf(m, "tx_packet_count(VHT MCS7) = %d\n",
			   wifi->stats.vht_tx_mcs7_packet_count);
		seq_printf(m, "tx_packet_count(VHT MCS8) = %d\n",
			   wifi->stats.vht_tx_mcs8_packet_count);
		seq_printf(m, "tx_packet_count(VHT MCS9) = %d\n",
			   wifi->stats.vht_tx_mcs9_packet_count);
	}
	seq_printf(m, "tx_cmds_from_stack= %d\n",
		   wifi->stats.tx_cmds_from_stack);
	seq_printf(m, "tx_dones_to_stack= %d\n",
		   wifi->stats.tx_dones_to_stack);
	seq_printf(m, "tx_noagg_not_addr= %d\n",
		   wifi->stats.tx_noagg_not_addr);
	seq_printf(m, "tx_noagg_not_ampdu= %d\n",
		   wifi->stats.tx_noagg_not_ampdu);
	seq_printf(m, "tx_noagg_not_qos= %d\n",
		   wifi->stats.tx_noagg_not_qos);
	seq_printf(m, "tx_noagg_not_rate= %d\n",
		   wifi->stats.tx_noagg_not_rate);
	seq_printf(m, "oustanding_cmd_cnt = %d\n",
		   wifi->stats.outstanding_cmd_cnt);
	seq_printf(m, "gen_cmd_send_count = %d\n",
		   wifi->stats.gen_cmd_send_count);
	seq_printf(m, "umac_scan_req = %d\n",
		   wifi->stats.umac_scan_req);
	seq_printf(m, "umac_scan_complete = %d\n",
		   wifi->stats.umac_scan_complete);
	seq_printf(m, "tx_cmd_send_count_single = %d\n",
		   wifi->stats.tx_cmd_send_count_single);
	seq_printf(m, "tx_cmd_send_count_multi = %d\n",
		   wifi->stats.tx_cmd_send_count_multi);
	seq_printf(m, "tx_cmd_send_count_beacon_q = %d\n",
		   wifi->stats.tx_cmd_send_count_beaconq);
	seq_printf(m, "tx_done_recv_count = %d\n",
		   wifi->stats.tx_done_recv_count);

	seq_printf(m, "tx_buff_pool_map = %ld\n",
		   priv->tx.buf_pool_bmp[0]);
	{
		int i, j;
		struct sk_buff_head *pend_pkt_q;

		seq_puts(m, "Pending Qs status\n");
		for (i = 0; i < MAX_PEERS; i++) {

			if (!priv->peers[i])
				continue;

			for (j = 0; j < WLAN_AC_MAX_CNT; j++) {
				spin_lock_bh(&priv->tx.lock);
				pend_pkt_q = &priv->tx.pending_pkt[0][i][j];
				if (skb_queue_len(pend_pkt_q))
					seq_printf(m,
						   "ac:%d peer:%d = %d\n",
						   j,
						   i,
						   skb_queue_len(pend_pkt_q));
				spin_unlock_bh(&priv->tx.lock);
			}
		}
		seq_puts(m, "\n");
	}

	if (ftm)
		seq_printf(m, "pdout_val = %d (total samples: %d)\n",
			   total_samples ? (total_value/total_samples) : 0,
			   total_samples);
	if (wifi->params.production_test)
		seq_printf(m,
			   "rssi_average = %d dBm (total rssi samples: %d)\n",
			   total_rssi_samples ?
			   (total_rssi_value/total_rssi_samples) : 0,
			   total_rssi_samples);

	seq_puts(m, "************* LMAC STATS ***********\n");
	seq_printf(m, "roc_start =%d\n",
		   wifi->stats.roc_start);
	seq_printf(m, "roc_stop =%d\n",
		   wifi->stats.roc_stop);
	seq_printf(m, "roc_complete =%d\n",
		   wifi->stats.roc_complete);
	seq_printf(m, "roc_stop_complete =%d\n",
		   wifi->stats.roc_stop_complete);
	/* TX related */
	seq_printf(m, "tx_cmd_cnt =%d\n",
		   wifi->stats.tx_cmd_cnt);
	seq_printf(m, "tx_done_cnt =%d\n",
		   wifi->stats.tx_done_cnt);
	seq_printf(m, "tx_edca_trigger_cnt =%d\n",
		   wifi->stats.tx_edca_trigger_cnt);
	seq_printf(m, "tx_edca_isr_cnt =%d\n",
		   wifi->stats.tx_edca_isr_cnt);
	seq_printf(m, "tx_start_cnt =%d\n",
		   wifi->stats.tx_start_cnt);
	seq_printf(m, "tx_abort_cnt =%d\n",
		   wifi->stats.tx_abort_cnt);
	seq_printf(m, "tx_abort_isr_cnt =%d\n",
		   wifi->stats.tx_abort_isr_cnt);
	seq_printf(m, "tx_underrun_cnt =%d\n",
		   wifi->stats.tx_underrun_cnt);
	seq_printf(m, "tx_rts_cnt =%d\n",
		   wifi->stats.tx_rts_cnt);
	seq_printf(m, "tx_ampdu_cnt =%d\n",
		   wifi->stats.tx_ampdu_cnt);
	seq_printf(m, "tx_mpdu_cnt =%d\n",
		   wifi->stats.tx_mpdu_cnt);
	/* RX related */
	seq_printf(m, "rx_isr_cnt  =%d\n",
		   wifi->stats.rx_isr_cnt);
	seq_printf(m, "rx_ack_cts_to_cnt =%d\n",
		   wifi->stats.rx_ack_cts_to_cnt);
	seq_printf(m, "rx_cts_cnt =%d\n",
		   wifi->stats.rx_cts_cnt);
	seq_printf(m, "rx_ack_resp_cnt =%d\n",
		   wifi->stats.rx_ack_resp_cnt);
	seq_printf(m, "rx_ba_resp_cnt =%d\n",
		   wifi->stats.rx_ba_resp_cnt);
	seq_printf(m, "rx_fail_in_ba_bitmap_cnt =%d\n",
		   wifi->stats.rx_fail_in_ba_bitmap_cnt);
	seq_printf(m, "rx_circular_buffer_free_cnt =%d\n",
		   wifi->stats.rx_circular_buffer_free_cnt);
	seq_printf(m, "rx_mic_fail_cnt =%d\n",
		   wifi->stats.rx_mic_fail_cnt);

	/* HAL related */
	seq_printf(m, "hal_cmd_cnt  =%d\n",
		   wifi->stats.hal_cmd_cnt);
	seq_printf(m, "hal_event_cnt =%d\n",
		   wifi->stats.hal_event_cnt);

	/* RPU Frame Drops*/
	seq_puts(m, "================\n");
	seq_puts(m, "RPU Frame Drops\n");
	seq_puts(m, "================\n");
	seq_printf(m, "rpu_ext_drops =%d\n",
		   wifi->stats.hal_ext_ptr_null_cnt);
	seq_printf(m, "rpu_int_drops =%d\n",
		   wifi->stats.hal_int_ptr_null_cnt);
	seq_printf(m, "rpu_deagg_drops =%d\n",
		   wifi->stats.deagg_bufOverFlow_cnt);

	return 0;

}

void rpu_reinit(void)
{

	if (wifi->hw)
		rpu_exit();
	rpu_init();
	uccp_reinit = 1;
}

static ssize_t proc_write_config(struct file *file,
				 const char __user *buffer,
				 size_t count,
				 loff_t *ppos)
{
	char buf[(RF_PARAMS_SIZE * 2) + 50];
	unsigned long val;
	long sval;
	struct img_priv *priv;
	int ret = 0;

	if (!wifi->hw)
		return -ENODEV;

	priv = (struct img_priv *)(wifi->hw->priv);

	if (count >= sizeof(buf))
		count = sizeof(buf) - 1;

	if (copy_from_user(buf, buffer, count))
		return -EFAULT;

	buf[count] = '\0';


	if (param_get_val(buf, "dot11a_support=", &val)) {
		if (((val == 0) || (val == 1)) &&
		    (wifi->params.dot11a_support != val)) {
			wifi->params.dot11a_support = val;

			if ((wifi->params.dot11g_support == 0) &&
			    (wifi->params.dot11a_support == 0)) {
				pr_err("Invalid parameter value. Both bands can't be disabled, at least 1 is needed\n");
			} else {
					rpu_reinit();
					pr_info("Re-initializing UMAC ..with 2.4GHz support %s and 5GHz support %s\n",
					wifi->params.dot11g_support == 0 ?
					"disabled" : "enabled",
					wifi->params.dot11a_support == 0 ?
					"disabled" : "enabled");

			}
		} else
			pr_err("Invalid parameter value\n");
	} else if (param_get_val(buf, "dot11g_support=", &val)) {
		if (((val == 0) || (val == 1)) &&
		    (wifi->params.dot11g_support != val)) {
			wifi->params.dot11g_support = val;

			if ((wifi->params.dot11g_support == 0) &&
			    (wifi->params.dot11a_support == 0)) {
				pr_err("Invalid parameter value. Both bands can't be disabled, at least 1 is needed\n");
			} else {
					rpu_reinit();
					pr_info("Re-initializing UMAC ..with 2.4GHz support %s and 5GHz support %s\n",
					wifi->params.dot11g_support == 0 ?
					"disabled" : "enabled",
					wifi->params.dot11a_support == 0 ?
					"disabled" : "enabled");

			}
		} else
			pr_err("Invalid parameter value\n");
	} else if (param_get_val(buf, "production_test=", &val)) {
		if ((val == 0) || (val == 1)) {
			if (wifi->params.production_test != val) {
				if (wifi->params.production_test)
					wifi->params.num_vifs = 1;

				wifi->params.production_test = val;

				rpu_reinit();
				pr_err("Re-initializing UMAC ..\n");
			}
		} else
			pr_err("Invalid parameter value\n");
	} else if (param_get_val(buf, "bypass_vpd=", &val)) {
		if ((val == 0) || (val == 1)) {
			if (wifi->params.bypass_vpd != val)
				wifi->params.bypass_vpd = val;
		} else
			pr_err("Invalid parameter value\n");
	} else if (param_get_val(buf, "num_vifs=", &val)) {
		if (val > 0 && val <= MAX_VIFS) {
			if (wifi->params.num_vifs != val) {
				rpu_reinit();
				pr_err("Re-initializing UMAC ..\n");
				wifi->params.num_vifs = val;
			}
		}
	} else if (param_get_match(buf, "rf_params=")) {
		conv_str_to_byte(wifi->params.rf_params,
				strstr(buf, "=") + 1,
				RF_PARAMS_SIZE);
	} else if (param_get_val(buf, "rx_packet_mgmt_count=", &val)) {
		wifi->stats.rx_packet_mgmt_count = val;
	} else if (param_get_val(buf, "rx_packet_data_count=", &val)) {
		wifi->stats.rx_packet_data_count = val;
	} else if (param_get_val(buf, "pdout_val=", &val)) {
		wifi->stats.pdout_val = val;
	} else if (param_get_val(buf, "get_stats=", &val)) {
		if (priv->state != STARTED) {
			pr_err("Interface is not initialized\n");
			goto error;
		}
		CALL_RPU(rpu_prog_mib_stats);
	} else if (param_get_val(buf, "max_data_size=", &val)) {
		if (wifi->params.max_data_size != val) {
			if ((wifi->params.max_data_size >= 2 * 1024) &&
			    (wifi->params.max_data_size <= (12 * 1024))) {
				wifi->params.max_data_size = val;

				rpu_reinit();
				pr_err("Re-initalizing RPU with %ld as max data size\n",
				       val);

			} else
				pr_err("Invalid Value for max data size: should be (2K-12K)\n");
		}
	} else if (param_get_val(buf, "max_tx_cmds=", &val)) {
		int max_tx_cmd_limit = MAX_SUBFRAMES_IN_AMPDU_HT;

		if (vht_support)
			max_tx_cmd_limit = MAX_SUBFRAMES_IN_AMPDU_VHT;

		if (val >= 1 && val <= max_tx_cmd_limit)
			wifi->params.max_tx_cmds = val;
		else
			pr_err("Please enter value between 1 and %d\n",
			       max_tx_cmd_limit);
	} else if (param_get_val(buf, "disable_power_save=", &val)) {
		if ((val == 0) || (val == 1)) {
			if (val != wifi->params.disable_power_save) {
				wifi->params.disable_power_save = val;

				rpu_reinit();
				pr_err("Re-initalizing RPU with global powerave %s\n",
				       val ? "DISABLED" : "ENABLED");
			}
		}
	} else if (param_get_val(buf, "disable_sm_power_save=", &val)) {
		if ((val == 0) || (val == 1)) {
			if (val != wifi->params.disable_sm_power_save) {
				wifi->params.disable_sm_power_save = val;

				rpu_reinit();
				pr_err("Re-initalizing RPU with smps %s\n",
				       val ? "DISABLED" : "ENABLED");

			}
		}
	} else if (param_get_val(buf, "rpu_max_nss=", &val)) {
		if (val > 0 && val <= min(MAX_TX_STREAMS, MAX_RX_STREAMS)) {
			if (val != wifi->params.rpu_max_nss) {
				wifi->params.rpu_max_nss = val;
				wifi->params.num_spatial_streams = val;
				wifi->params.max_tx_streams = val;
				wifi->params.max_rx_streams = val;
				rpu_reinit();
				pr_err("Re-initalizing RPU with %ld spatial streams\n",
				       val);
			}
		} else
			pr_err("Invalid parameter value: Allowed Range: 1 to %d\n",
			       min(MAX_TX_STREAMS, MAX_RX_STREAMS));
	} else if (param_get_val(buf, "enable_early_agg_checks=", &val)) {
		if ((val == 0) || (val == 1)) {
			if (val != wifi->params.enable_early_agg_checks)
				wifi->params.enable_early_agg_checks = val;
		} else
			pr_err("Invalid parameter value: Allowed: 0/1\n");
	} else if (param_get_val(buf, "antenna_sel=", &val)) {
		if (val == 1 || val == 2) {
			if (val != wifi->params.antenna_sel) {
				wifi->params.antenna_sel = val;
				rpu_reinit();
				pr_err("Re-initalizing RPU with %ld antenna selection\n",
				       val);
			}
		} else
			pr_err("Invalid parameter value: Allowed Values: 1 or 2\n");
	} else if (param_get_val(buf, "num_spatial_streams=", &val)) {
		if (val > 0 && val <= wifi->params.rpu_max_nss)
			wifi->params.num_spatial_streams = val;
		else
			pr_err("Invalid parameter value, should be less than or equal to rpu_max_nss\n");
	} else if (param_get_sval(buf, "mgd_mode_tx_fixed_mcs_indx=", &sval)) {
		if (wifi->params.mgd_mode_tx_fixed_rate == -1) {
			if (check_valid_data_rate(priv, sval | 0x80, UCAST))
				wifi->params.mgd_mode_tx_fixed_mcs_indx = sval;
		} else
			pr_err("Fixed rate other than MCS index is currently set\n");
	} else if (param_get_sval(buf, "mgd_mode_tx_fixed_rate=", &sval)) {
		if (wifi->params.mgd_mode_tx_fixed_mcs_indx == -1) {
			if (check_valid_data_rate(priv, sval, UCAST))
				wifi->params.mgd_mode_tx_fixed_rate = sval;
		} else
			pr_err("MCS data rate(index) is currently set\n");
	/* Multicast Rate configuration options.
	 */
	} else if (param_get_sval(buf, "mgd_mode_mcast_fixed_data_rate=",
		   &sval)) {
		if (check_valid_data_rate(priv, sval, MCAST))
			wifi->params.mgd_mode_mcast_fixed_data_rate = sval;
	} else if (param_get_val(buf, "mgd_mode_mcast_fixed_rate_flags=",
		   &val)) {
		if (check_valid_rate_flags(priv, val))
			wifi->params.mgd_mode_mcast_fixed_rate_flags = val;
	} else if (param_get_val(buf, "mgd_mode_mcast_fixed_bcc_or_ldpc=",
		   &val)) {
		wifi->params.mgd_mode_mcast_fixed_bcc_or_ldpc = val;
	} else if (param_get_val(buf, "mgd_mode_mcast_fixed_stbc_enabled=",
		   &val)) {
		wifi->params.mgd_mode_mcast_fixed_stbc_enabled = val;
	} else if (param_get_val(buf, "mgd_mode_mcast_fixed_preamble=",
		   &val)) {
		wifi->params.mgd_mode_mcast_fixed_preamble = val;
	} else if (param_get_val(buf, "mgd_mode_mcast_fixed_nss=", &val)) {
		wifi->params.mgd_mode_mcast_fixed_nss = val;

	/* Production mode rate configuration
	 */
	} else if (param_get_sval(buf, "tx_fixed_mcs_indx=", &sval)) {
		if (wifi->params.production_test != 1) {
			pr_err("Only can be set in production mode.\n");
			goto error;
		}

		if (sval == -1) {
			wifi->params.tx_fixed_mcs_indx = -1;
			goto error;
		}

		if (wifi->params.tx_fixed_rate != -1) {
			pr_err("Fixed rate other than MCS index is currently set\n");
			goto error;
		}

		if (check_valid_data_rate(priv, sval | 0x80, UCAST))
			wifi->params.tx_fixed_mcs_indx = sval;


	} else if (param_get_sval(buf, "tx_fixed_rate=", &sval)) {
		if (wifi->params.production_test != 1) {
			pr_err("Only can be set in production mode.\n");
			goto error;
		}

		if (sval == -1) {
			wifi->params.tx_fixed_rate = -1;
			goto error;
		}
		if (wifi->params.tx_fixed_mcs_indx != -1) {
			pr_err("MCS Index is currently set.\n");
			goto error;
		}

		if (check_valid_data_rate(priv, sval, UCAST))
			wifi->params.tx_fixed_rate = sval;
	} else if (param_get_sval(buf, "chnl_40_offset=", &sval)) {

		do {
			if (wifi->params.production_test != 1) {
				pr_err("Can be set in only in production mode.\n");
				break;
			}

			if (!((sval == 0) || (sval == 1) || (sval == -1))) {
				pr_err("Invalid parameter value: %ld\n", sval);
				break;
			}

			wifi->params.chnl_40_offset = sval;
		} while (0);
	} else if (vht_support &&
		   param_get_sval(buf, "chnl_80_offset=", &sval)) {

		do {
			if (wifi->params.production_test != 1) {
				pr_err("Can be set in only in production mode.\n");
				break;
			}

			if (!((sval == 0) || (sval == 1) || (sval == -1))) {
				pr_err("Invalid parameter value: %ld\n", sval);
				break;
			}

			wifi->params.chnl_80_offset = sval;
		} while (0);
	} else if (param_get_val(buf, "fixed_rate_flags=", &val)) {
		if (check_valid_rate_flags(priv, val))
			wifi->params.fixed_rate_flags = val;
	} else if (param_get_val(buf, "rate_protection_type=", &val)) {
		/* 0 is None, 1 is RTS/CTS, 2 is for CTS2SELF */
		if ((val == 0) || (val == 1) /*|| (val == 2)*/)
			wifi->params.rate_protection_type = val;
		else
			pr_err("Invalid parameter value");
	} else if (param_get_val(buf, "fixed_rate_preamble_type=", &val)) {
		/*0 is short, 1 is Long*/
		if ((val == 0) || (val == 1))
			wifi->params.fixed_rate_preamble_type = val;
		else
			pr_err("Invalid parameter value");
	} else if (param_get_val(buf, "fixed_rate_stbc_enabled=", &val)) {
		if (val <= 1)
			wifi->params.fixed_rate_stbc_enabled = val;
		else
			pr_err("Invalid parameter value\n");
	} else if (param_get_val(buf, "fixed_rate_bcc_or_ldpc=", &val)) {
		if (val <= 1)
			wifi->params.fixed_rate_bcc_or_ldpc = val;
		else
			pr_err("Invalid parameter value\n");
	} else if (param_get_val(buf, "reset_hal_params=", &val)) {
		if (priv->state != STARTED) {
			if (val != 1)
				pr_err("Invalid parameter value\n");
			else
				hal_ops.reset_hal_params();
		} else
			pr_err("HAL parameters reset can be done only when all interface are down\n");
	} else if (param_get_val(buf, "vht_beamformer_enable=", &val)) {
		do {
			int vht_beamform_period;

			if (wifi->params.vht_beamform_enable == val)
				break;

			if (!((val == VHT_BEAMFORM_ENABLE) ||
			      (val == VHT_BEAMFORM_DISABLE))) {
				pr_err("Invalid VHT Beamforming Enable value should be 1 or 0\n");
				break;
			}

			wifi->params.vht_beamform_enable = val;

			/* If not associated, it will be sent upon
			 * association
			 */
			if (!wifi->params.is_associated)
				break;

			if (!wifi->params.vht_beamform_support) {
				pr_err("Peer doesn't support VHT Beamformee.\n");
				break;
			}

			vht_beamform_period = wifi->params.vht_beamform_period;

			if (priv->state != STARTED) {
				pr_err("Interface is not initialized\n");
				goto error;
			}

			CALL_RPU(rpu_prog_vht_bform,
				  val,
				  vht_beamform_period);
		} while (0);

	} else if (param_get_val(buf, "vht_beamformer_period=", &val)) {

		do {
			int vht_beamform_enable;

			if (wifi->params.vht_beamform_enable !=
			    VHT_BEAMFORM_ENABLE) {
				pr_err("VHT Beamforming is disabled, please enable it first\n");
				break;
			}

			if (wifi->params.vht_beamform_period == val)
				break;

			if (!((val > 100) || (val < 10000))) {
				pr_err("Invalid VHT Beamforming Period must be between 100-10000ms\n");
				break;
			}

			wifi->params.vht_beamform_period = val;

			/* If not associated, it will be sent upon
			 * association
			 */
			if (!wifi->params.is_associated)
				break;

			if (!wifi->params.vht_beamform_support) {
				pr_err("Peer doesn't support VHT Beamformee.\n");
				break;
			}

			vht_beamform_enable = wifi->params.vht_beamform_enable;

			if (priv->state != STARTED) {
				pr_err("Interface is not initialized\n");
				goto error;
			}

			CALL_RPU(rpu_prog_vht_bform,
				  vht_beamform_enable,
				  val);
		} while (0);
	} else if (param_get_val(buf, "scan_type=", &val)) {
		if ((val == 0) || (val == 1))
			wifi->params.scan_type = val;
		else
			pr_err("Invalid scan type value %d, should be 0 or 1\n",
			       (unsigned int)val);
	} else if (ftm && param_get_val(buf, "aux_adc_chain_id=", &val)) {

		if (priv->state != STARTED) {
			pr_err("Interface is not initialized\n");
			goto error;
		}

		memset(wifi->params.pdout_voltage, 0,
		       sizeof(char) * MAX_AUX_ADC_SAMPLES);
		if ((val == AUX_ADC_CHAIN1) || (val == AUX_ADC_CHAIN2)) {
			wifi->params.aux_adc_chain_id = val;
			CALL_RPU(rpu_prog_aux_adc_chain, val);
		} else
			pr_err("Invalid chain id %d, should be %d or %d\n",
			       (unsigned int) val,
			       AUX_ADC_CHAIN1,
			       AUX_ADC_CHAIN2);
	} else if (param_get_val(buf, "continuous_tx=", &val)) {
		if (wifi->params.production_test != 1) {
			pr_err("continuous_tx: Can be set in only in production mode.\n");
			goto error;
		}

		if (priv->state != STARTED) {
			pr_err("Interface is not initialized\n");
			goto error;
		}

		if (val == 0 || val == 1) {
			wifi->params.cont_tx = val;
			CALL_RPU(rpu_prog_cont_tx, val);
		} else
			pr_err("Invalid tx_continuous parameter\n");
	} else if (param_get_val(buf, "start_prod_mode=", &val)) {
		start_prod_mode(priv, val);
	} else if (param_get_sval(buf, "stop_prod_mode=", &sval)) {
		stop_prod_mode(priv, sval);
	} else if (param_get_sval(buf, "start_packet_gen=", &sval)) {
		start_packet_gen(priv, sval);
	} else if (param_get_sval(buf, "stop_packet_gen=", &sval)) {
		stop_packet_gen(priv, sval);
	} else if (param_get_val(buf, "payload_length=", &val)) {
		wifi->params.payload_length = val;
	} else if (param_get_sval(buf, "set_tx_power=", &sval)) {
		if (wifi->params.production_test != 1 && !ftm) {
			pr_err("set_tx_power: Can be set in only in FTM/production mode.\n");
			goto error;
		}

		if (!wifi->params.init_prod) {
			pr_err("NEW Production Mode is not Initialized\n");
			goto error;
		}

		memset(wifi->params.pdout_voltage, 0,
		       sizeof(char) * MAX_AUX_ADC_SAMPLES);
		wifi->params.set_tx_power = sval;
		CALL_RPU(rpu_prog_txpower, sval);
	} else if (param_get_val(buf, "fw_loading=", &val)) {
		wifi->params.fw_loading = val;
	} else if (param_get_val(buf, "bt_state=", &val)) {
		if (priv->state != STARTED) {
			pr_err("Interface is not initialized\n");
			goto error;
		}

		if (val == 0 || val == 1) {
			if (val != wifi->params.bt_state) {
				wifi->params.bt_state = val;
				CALL_RPU(rpu_prog_btinfo, val);
			}
		} else
			pr_err("Invalid parameter value: Allowed values: 0 or 1\n");
	} else if (param_get_val(buf, "clear_stats=", &val)) {
		if (priv->state != STARTED) {
			pr_err("Interface is not initialized\n");
			goto error;
		}
		CALL_RPU(rpu_prog_clear_stats);
	} else if (param_get_val(buf, "rpu_debug=", &val)) {
		rpu_debug = val;
	} else if (param_get_val(buf, "rpu_mtp_soft_reset=", &val)) {
		fwldr_init();
		fwldr_soft_reset(0);
	} else if (param_get_val(buf, "rpu_mtp_stop_thread=", &val)) {
		fwldr_stop_thrd(0);
	} else if (param_get_val(buf, "ddr_get_64mb_addr=", &val)) {
		ddr_test_64mb_boundary(val, 70 * 1024 * 1024);
	} else if (param_get_val(buf, "load_dbg_fw=", &val)) {
		load_fw(wifi->hw);
	} else if (param_get_val(buf, "minstrel_fastpath=", &val)) {
		wifi->params.minstrel_fastpath = val;
		rpu_reinit();
		pr_info("Re-initializing UMAC ..with RC_TABLE support %s\n",
			val ? "Enabled" : "Disabled");
	} else if (param_get_val(buf, "max_rate_retries=", &val)) {
		wifi->params.max_rate_retries = val;
		rpu_reinit();
		pr_info("Re-initializing UMAC ..with max_rate_retries/rate-series=%ld\n",
			val);
	} else
		pr_err("Invalid parameter name: %s\n", buf);
error:
	return count;
prog_rpu_fail:
	return ret;
}


static int proc_open_config(struct inode *inode, struct file *file)
{
	return single_open(file, proc_read_config, NULL);
}


static int proc_open_phy_stats(struct inode *inode, struct file *file)
{
	return single_open(file, proc_read_phy_stats, NULL);
}

static int proc_open_mac_stats(struct inode *inode, struct file *file)
{
	return single_open(file, proc_read_mac_stats, NULL);
}


static const struct file_operations params_fops_config = {
	.open = proc_open_config,
	.read = seq_read,
	.llseek = seq_lseek,
	.write = proc_write_config,
	.release = single_release
};

static const struct file_operations params_fops_phy_stats = {
	.open = proc_open_phy_stats,
	.read = seq_read,
	.llseek = seq_lseek,
	.write = NULL,
	.release = single_release
};

static const struct file_operations params_fops_mac_stats = {
	.open = proc_open_mac_stats,
	.read = seq_read,
	.llseek = seq_lseek,
	.write = NULL,
	.release = single_release
};
int proc_init(struct proc_dir_entry ***main_dir_entry)
{
	struct proc_dir_entry *entry;
	int err = 0;
	unsigned int i = 0;
	/*2.4GHz and 5 GHz PD and TX-PWR calibration params*/
	unsigned char rf_params[RF_PARAMS_SIZE * 2];

	strncpy(rf_params, DEFAULT_RF_PARAMS, (RF_PARAMS_SIZE * 2));

	wifi = kzalloc(sizeof(struct wifi_dev), GFP_KERNEL);
	if (!wifi) {
		err = -ENOMEM;
		goto out;
	}

	wifi->umac_proc_dir_entry = proc_mkdir("rpu", NULL);
	if (!wifi->umac_proc_dir_entry) {
		pr_err("Failed to create proc dir\n");
		err = -ENOMEM;
		goto  proc_dir_fail;
	}

	entry = proc_create("params", 0644, wifi->umac_proc_dir_entry,
			    &params_fops_config);
	if (!entry) {
		pr_err("Failed to create proc entry\n");
		err = -ENOMEM;
		goto  proc_entry1_fail;
	}

	entry = proc_create("phy_stats", 0444, wifi->umac_proc_dir_entry,
			    &params_fops_phy_stats);
	if (!entry) {
		pr_err("Failed to create proc entry\n");
		err = -ENOMEM;
		goto  proc_entry2_fail;
	}

	entry = proc_create("mac_stats", 0444, wifi->umac_proc_dir_entry,
			    &params_fops_mac_stats);
	if (!entry) {
		pr_err("Failed to create proc entry\n");
		err = -ENOMEM;
		goto  proc_entry3_fail;
	}
	/* Initialize WLAN params */
	memset(&wifi->params, 0, sizeof(struct wifi_params));

	/* TODO: Make this a struct */
	memset(wifi->params.rf_params, 0xFF, sizeof(wifi->params.rf_params));
	conv_str_to_byte(wifi->params.rf_params, rf_params, RF_PARAMS_SIZE);

	if (!rf_params_vpd)
		rf_params_vpd = wifi->params.rf_params;

	memcpy(wifi->params.rf_params_vpd, rf_params_vpd, RF_PARAMS_SIZE);

	wifi->params.is_associated = 0;
	wifi->params.dot11a_support = 1;
	wifi->params.dot11g_support = 1;
	wifi->params.num_vifs = 2;

	/* Check, if required add it */
	wifi->params.tx_fixed_mcs_indx = -1;
	wifi->params.tx_fixed_rate = -1;
	wifi->params.num_spatial_streams = min(MAX_TX_STREAMS, MAX_RX_STREAMS);
	wifi->params.rpu_max_nss = min(MAX_TX_STREAMS,
						    MAX_RX_STREAMS);
	wifi->params.antenna_sel = 1;

	if (num_streams_vpd > 0)
		wifi->params.rpu_max_nss = num_streams_vpd;

	wifi->params.enable_early_agg_checks = 1;
	wifi->params.bt_state = 1;

	/* Defaults optimized for all IMG clients
	 */
	wifi->params.mgd_mode_tx_fixed_mcs_indx = -1;
	wifi->params.mgd_mode_mcast_fixed_data_rate = -1;
	wifi->params.mgd_mode_tx_fixed_rate = -1;
	wifi->params.mgd_mode_mcast_fixed_nss = 1;
	wifi->params.mgd_mode_mcast_fixed_bcc_or_ldpc = 1;
	wifi->params.mgd_mode_mcast_fixed_stbc_enabled = 1;
	wifi->params.vht_beamform_enable = VHT_BEAMFORM_DISABLE;
	wifi->params.vht_beamform_period = 1000; /* ms */
	wifi->params.vht_beamform_support = 0;
	wifi->params.max_tx_streams = MAX_TX_STREAMS;
	wifi->params.max_rx_streams = MAX_RX_STREAMS;
	wifi->params.max_data_size  = 8 * 1024;

	wifi->params.max_tx_cmds = MAX_SUBFRAMES_IN_AMPDU_HT;
	if (vht_support)
		wifi->params.max_tx_cmds = MAX_SUBFRAMES_IN_AMPDU_VHT;
	wifi->params.disable_power_save = 0;
	wifi->params.disable_sm_power_save = 0;
	wifi->params.rate_protection_type = 1;
	wifi->params.fixed_rate_preamble_type = 1; /* LONG */
	wifi->params.fixed_rate_stbc_enabled = 0;
	wifi->params.fixed_rate_bcc_or_ldpc = 0;
	wifi->params.bg_scan_enable = 0;
	memset(wifi->params.bg_scan_channel_list, 0, 50);
	memset(wifi->params.bg_scan_channel_flags, 0, 50);

	if (wifi->params.dot11g_support) {
		wifi->params.bg_scan_num_channels = 3;

		wifi->params.bg_scan_channel_list[i] = 1;
		wifi->params.bg_scan_channel_flags[i++] = ACTIVE;

		wifi->params.bg_scan_channel_list[i] = 6;
		wifi->params.bg_scan_channel_flags[i++] = ACTIVE;

		wifi->params.bg_scan_channel_list[i] = 11;
		wifi->params.bg_scan_channel_flags[i++] = ACTIVE;
	}

	if (wifi->params.dot11a_support) {
		wifi->params.bg_scan_num_channels += 4;

		wifi->params.bg_scan_channel_list[i] = 36;
		wifi->params.bg_scan_channel_flags[i++] = ACTIVE;

		wifi->params.bg_scan_channel_list[i] = 40;
		wifi->params.bg_scan_channel_flags[i++] = ACTIVE;

		wifi->params.bg_scan_channel_list[i] = 44;
		wifi->params.bg_scan_channel_flags[i++] = ACTIVE;

		wifi->params.bg_scan_channel_list[i] = 48;
		wifi->params.bg_scan_channel_flags[i++] = ACTIVE;
	}
	wifi->params.pkt_gen_val = -1;
	wifi->params.init_pkt_gen = 0;
	wifi->params.payload_length = 4000;
	wifi->params.start_prod_mode = 0;
	wifi->params.init_prod = 0;
	wifi->params.bg_scan_intval = 5000 * 1000; /* Once in 5 seconds */
	wifi->params.bg_scan_chan_dur = 300; /* Channel spending time */
	wifi->params.bg_scan_serv_chan_dur = 100; /* Oper chan spending time */
	wifi->params.scan_type = ACTIVE;
	wifi->params.hw_scan_status = HW_SCAN_STATUS_NONE;
	wifi->params.fw_loading = 1;
	wifi->params.minstrel_fastpath = false;
	wifi->params.max_rate_retries = 2;

	**main_dir_entry = wifi->umac_proc_dir_entry;
	return err;
proc_entry3_fail:
	remove_proc_entry("phy_stats", wifi->umac_proc_dir_entry);
proc_entry2_fail:
	remove_proc_entry("params", wifi->umac_proc_dir_entry);
proc_entry1_fail:
	remove_proc_entry("rpu", NULL);
proc_dir_fail:
	kfree(wifi);
out:
	return err;

}

void proc_exit(void)
{
	remove_proc_entry("hal_stats", wifi->umac_proc_dir_entry);
	remove_proc_entry("mac_stats", wifi->umac_proc_dir_entry);
	remove_proc_entry("phy_stats", wifi->umac_proc_dir_entry);
	remove_proc_entry("params", wifi->umac_proc_dir_entry);
	remove_proc_entry("rpu", NULL);
	kfree(wifi);
}


