/*
 * Copyright (c) 2011-2017 Imagination Technologies Ltd.
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */


#include <linux/netdevice.h>
#include <linux/rcupdate.h>
#include <linux/slab.h>
#include <linux/spinlock.h>

#include "core.h"
#include "utils.h"

unsigned char wildcard_ssid[7] = "DIRECT-";

struct cmd_send_recv_cnt cmd_info;

struct rpu_if_data {
	char *name;
	void *context;
};

static struct rpu_if_data __rcu *rpu_if;

void update_mcs_packet_stat(int mcs_rate_num,
			    int rate_flags,
			    struct img_priv *priv)
{
	if (rate_flags & ENABLE_11N_FORMAT) {
		switch (mcs_rate_num) {
		case 0:
			priv->stats->ht_tx_mcs0_packet_count++;
			break;
		case 1:
			priv->stats->ht_tx_mcs1_packet_count++;
			break;
		case 2:
			priv->stats->ht_tx_mcs2_packet_count++;
			break;
		case 3:
			priv->stats->ht_tx_mcs3_packet_count++;
			break;
		case 4:
			priv->stats->ht_tx_mcs4_packet_count++;
			break;
		case 5:
			priv->stats->ht_tx_mcs5_packet_count++;
			break;
		case 6:
			priv->stats->ht_tx_mcs6_packet_count++;
			break;
		case 7:
			priv->stats->ht_tx_mcs7_packet_count++;
			break;
		case 8:
			priv->stats->ht_tx_mcs8_packet_count++;
			break;
		case 9:
			priv->stats->ht_tx_mcs9_packet_count++;
			break;
		case 10:
			priv->stats->ht_tx_mcs10_packet_count++;
			break;
		case 11:
			priv->stats->ht_tx_mcs11_packet_count++;
			break;
		case 12:
			priv->stats->ht_tx_mcs12_packet_count++;
			break;
		case 13:
			priv->stats->ht_tx_mcs13_packet_count++;
			break;
		case 14:
			priv->stats->ht_tx_mcs14_packet_count++;
			break;
		case 15:
			priv->stats->ht_tx_mcs15_packet_count++;
			break;
		default:
			break;
		}
	} else if (rate_flags & ENABLE_VHT_FORMAT) {
		switch (mcs_rate_num) {
		case 0:
			priv->stats->vht_tx_mcs0_packet_count++;
			break;
		case 1:
			priv->stats->vht_tx_mcs1_packet_count++;
			break;
		case 2:
			priv->stats->vht_tx_mcs2_packet_count++;
			break;
		case 3:
			priv->stats->vht_tx_mcs3_packet_count++;
			break;
		case 4:
			priv->stats->vht_tx_mcs4_packet_count++;
			break;
		case 5:
			priv->stats->vht_tx_mcs5_packet_count++;
			break;
		case 6:
			priv->stats->vht_tx_mcs6_packet_count++;
			break;
		case 7:
			priv->stats->vht_tx_mcs7_packet_count++;
			break;
		case 8:
			priv->stats->vht_tx_mcs8_packet_count++;
			break;
		case 9:
			priv->stats->vht_tx_mcs9_packet_count++;
			break;
		default:
			break;
		}
	}
}


static void get_rate(struct sk_buff *skb,
		     struct cmd_tx_ctrl *txcmd,
		     struct tx_pkt_info *pkt_info,
		     bool retry,
		     struct img_priv *priv)
{
	struct ieee80211_rate *rate;
	struct ieee80211_tx_info *c = IEEE80211_SKB_CB(skb);
	unsigned int index, min_rate;
	bool is_mcs = false, is_mgd = false;
	struct ieee80211_tx_rate *txrate;
	unsigned char mcs_rate_num = 0;
	struct ieee80211_hdr *hdr = (struct ieee80211_hdr *) skb->data;
	int mcs_indx;
	int mgd_rate;
	int mgd_mcast_rate;
	int prot_type;
	unsigned char nss = 1;
	bool all_rates_invalid = true;


	/* Normal Mode*/
	rate = ieee80211_get_tx_rate(priv->hw, c);
	min_rate = priv->hw->wiphy->bands[c->band]->bitrates[0].hw_value;

	if (rate == NULL) {
		RPU_DEBUG_IF("%s:%d rate is null taking defaults: min: %d\n",
			      __func__,
			      __LINE__,
			      c->control.rates[0].idx);
		txcmd->num_rates = 1;
		txcmd->rate[0] = min_rate;
		txcmd->rate_retries[0] = 5;
		txcmd->rate_protection_type[0] = USE_PROTECTION_NONE;
		txcmd->rate_preamble_type[0] = DONT_USE_SHORT_PREAMBLE;
		txcmd->num_spatial_streams[0] = 1;
		txcmd->bcc_or_ldpc = 0;
		txcmd->stbc_enabled = 0;
		txcmd->rate_flags[0] = 0;
		return;
	}

	/* Some defaults*/
	txcmd->num_rates = 0;
	txcmd->stbc_enabled = 0;

	/* BCC (or) LDPC */
	if (c->flags & IEEE80211_TX_CTL_LDPC)
		txcmd->bcc_or_ldpc = 1;
	else
		txcmd->bcc_or_ldpc = 0;

	if (ieee80211_is_data(hdr->frame_control) &&
	    c->flags & IEEE80211_TX_CTL_AMPDU) {
		txcmd->aggregate_mpdu = AMPDU_AGGR_ENABLED;
	}

	for (index = 0; index < 4; index++) {
		bool skip_rate = false;

		txrate = (&c->control.rates[index]);
		txcmd->rate_flags[index] = 0;

		if (txrate->idx < 0)
			continue;

		txcmd->num_rates++;
		txcmd->num_spatial_streams[index] = 1;

		/* No input from production_test proc, continue and use
		 * info from mac80211 RC
		 */

		/* It is an VHT MCS rate */
		if (txrate->flags & IEEE80211_TX_RC_VHT_MCS) {
			/*idx field is split
			 * into a higher 4 bits (Nss), starts
			 * with 0 and lower 4 bits (MCS number)
			 */
			is_mcs = true;
			mcs_rate_num = ieee80211_rate_get_vht_mcs(txrate);
			nss =  ieee80211_rate_get_vht_nss(txrate);
			txcmd->rate_flags[index] |= ENABLE_VHT_FORMAT;
		} else
		if (txrate->flags & IEEE80211_TX_RC_MCS) {
			is_mcs = true;
			mcs_rate_num  = txrate->idx;
			nss = mcs_rate_num/8 + 1;
			txcmd->rate_flags[index] |= ENABLE_11N_FORMAT;
		}

		mcs_indx = priv->params->mgd_mode_tx_fixed_mcs_indx;
		mgd_rate = priv->params->mgd_mode_tx_fixed_rate;
		mgd_mcast_rate = priv->params->mgd_mode_mcast_fixed_data_rate;

		/* Rate Index:
		 * From proc:
		 *    ** Multicast data packets
		 *    ** Unicast data packets
		 * From RC in mac80211
		 * Can be MCS(HT/VHT) or Rate (11abg)
		 */
		if (ieee80211_is_data(hdr->frame_control) &&
		    is_multicast_ether_addr(hdr->addr1) &&
		    (mgd_mcast_rate != -1)) {
			/* proc: Fixed MCS/Legacy rate for Multicast packets
			 */
			is_mgd = true;
			is_mcs = (mgd_mcast_rate & 0x80) == 0x80 ? true : false;

			if (!is_mcs) {
				if (mgd_mcast_rate == 55)
					mgd_mcast_rate = 11;
				else
					mgd_mcast_rate *= 2;
			}

			txcmd->rate[index] = mgd_mcast_rate;
			txcmd->rate_flags[index] =
				priv->params->mgd_mode_mcast_fixed_rate_flags;
			txcmd->bcc_or_ldpc =
				priv->params->mgd_mode_mcast_fixed_bcc_or_ldpc;
			if (txcmd->rate_flags[index] & ENABLE_11N_FORMAT)
				nss = (mgd_mcast_rate & 0x7F)/8 + 1;
			else
				nss = priv->params->mgd_mode_mcast_fixed_nss;
			txcmd->stbc_enabled =
				priv->params->mgd_mode_mcast_fixed_stbc_enabled;
			txcmd->rate_preamble_type[index] =
				priv->params->mgd_mode_mcast_fixed_preamble;
			if (is_mcs)
				update_mcs_packet_stat(mgd_mcast_rate & 0x7F,
						       txcmd->rate_flags[index],
						       priv);
		} else if (ieee80211_is_data(hdr->frame_control) &&
			   mcs_indx != -1) {
			/* proc: Fixed MCS for unicast
			 */
			is_mgd = true;

			txcmd->rate[index] = 0x80;
			txcmd->rate[index] |= (mcs_indx);
			txcmd->rate_flags[index] =
				priv->params->fixed_rate_flags;
			if (txcmd->rate_flags[index] & ENABLE_11N_FORMAT)
				nss = (mcs_indx)/8 + 1;
			else
				nss = priv->params->num_spatial_streams;
			txcmd->bcc_or_ldpc =
				priv->params->fixed_rate_bcc_or_ldpc;
			txcmd->stbc_enabled =
				priv->params->fixed_rate_stbc_enabled;

			update_mcs_packet_stat(mcs_indx,
					       txcmd->rate_flags[index],
					       priv);
		} else if (ieee80211_is_data(hdr->frame_control) &&
			   mgd_rate != -1) {
			/* proc: Fixed Legacy Rate for unicast
			 */
			is_mgd = true;
			txcmd->rate[index] = 0x80;
			txcmd->rate[index] = 0x00;

			if (mgd_rate == 55)
				txcmd->rate[index] |= ((mgd_rate) / 5);
			else
				txcmd->rate[index] |= ((mgd_rate * 10) / 5);

			txcmd->rate_flags[index] = 0;
			nss = 1;
			txcmd->bcc_or_ldpc	   = 0;
			txcmd->stbc_enabled	   = 0;
			txcmd->rate_preamble_type[index] =
				priv->params->fixed_rate_preamble_type;
		} else if (is_mcs) {
			txcmd->rate[index] = MARK_RATE_AS_MCS_INDEX;
			txcmd->rate[index] |= mcs_rate_num;
			update_mcs_packet_stat(mcs_rate_num,
					      txcmd->rate_flags[index],
					      priv);
		} else if (!is_mcs) {
			rate = &priv->hw->wiphy->bands[
				c->band]->bitrates[
				c->control.rates[index].idx];
			txcmd->rate[index] = MARK_RATE_AS_RATE;
			txcmd->rate[index] |= rate->hw_value;
			nss = 1;
		}

		txcmd->num_spatial_streams[index] = nss;

		if (is_mgd) {
			if (priv->params->rate_protection_type)
				txcmd->rate_protection_type[index] =
					USE_PROTECTION_RTS;
			else
				txcmd->rate_protection_type[index] =
					USE_PROTECTION_NONE;
			txcmd->rate_retries[index] = 1;
			all_rates_invalid = false;
			break;
		}

		/* STBC Enabled/Disabled: valid if n_antennas > Nss */
		if (priv->params->rpu_max_nss > nss &&
		    (c->flags & IEEE80211_TX_CTL_STBC))
			txcmd->stbc_enabled = 1;

		txcmd->rate_retries[index] =
			c->control.rates[index].count;

		if (c->control.rates[index].flags &
		    IEEE80211_TX_RC_USE_SHORT_PREAMBLE)
			txcmd->rate_preamble_type[index] =
				USE_SHORT_PREAMBLE;
		else
			txcmd->rate_preamble_type[index] =
				DONT_USE_SHORT_PREAMBLE;

		prot_type = USE_PROTECTION_NONE;
		if (priv->params->rate_protection_type == 1) {
			/* Protection*/
			if (c->control.rates[index].flags &
			    IEEE80211_TX_RC_USE_CTS_PROTECT)
				prot_type = USE_PROTECTION_CTS2SELF;
			else if (c->control.rates[index].flags &
				 IEEE80211_TX_RC_USE_RTS_CTS)
				prot_type = USE_PROTECTION_RTS;
			else
				prot_type = USE_PROTECTION_NONE;

			/*RTS threshold: Check for PSDU length
			 * Need to add all HW added lenghts to skb,
			 * sw added lengths are already part of skb->len
			 * IV ==> Always SW
			 * MIC for CCMP ==> HW (MMIC for TKIP ==> SW)
			 * ICV ==> HW
			 * FCS ==> HW
			*/
			if (ieee80211_is_data(hdr->frame_control) &&
			    !is_multicast_ether_addr(hdr->addr1) &&
			    ieee80211_has_protected(hdr->frame_control)) {
				if (skb->len +
				    c->control.hw_key->icv_len +
				    priv->rts_threshold < FCS_LEN)
					prot_type = USE_PROTECTION_RTS;
			}

			if (ieee80211_is_data(hdr->frame_control) &&
			    !is_multicast_ether_addr(hdr->addr1) &&
			    !ieee80211_has_protected(hdr->frame_control) &&
			    (skb->len + FCS_LEN > priv->rts_threshold))
				prot_type = USE_PROTECTION_RTS;
		}

		txcmd->rate_protection_type[index] = prot_type;

		if (c->control.rates[index].flags &
				IEEE80211_TX_RC_GREEN_FIELD)
			txcmd->rate_flags[index] |=
					ENABLE_GREEN_FIELD;
		if (c->control.rates[index].flags &
				IEEE80211_TX_RC_40_MHZ_WIDTH)
			txcmd->rate_flags[index] |=
					ENABLE_CHNL_WIDTH_40MHZ;
		if (c->control.rates[index].flags &
				IEEE80211_TX_RC_80_MHZ_WIDTH)
			txcmd->rate_flags[index] |=
				ENABLE_CHNL_WIDTH_80MHZ;
		if (c->control.rates[index].flags &
				IEEE80211_TX_RC_SHORT_GI)
			txcmd->rate_flags[index] |= ENABLE_SGI;

		/*Some Sanity Checks*/
		if (nss <= max(MAX_TX_STREAMS, MAX_RX_STREAMS))
			/*Got at-least one valid rate*/
			all_rates_invalid = false;
		else {
			if (net_ratelimit())
				RPU_DEBUG_IF("RPU_WIFI:Skip Nss: %d\n",
					      nss);
			skip_rate = true;
		}

		/* VHT 20MHz MCS9 is not valid*/
		if (txrate->flags & IEEE80211_TX_RC_VHT_MCS &&
			((txcmd->rate[index] & 0x7F) == 9) &&
			!(txcmd->rate_flags[index] &
			  ENABLE_CHNL_WIDTH_40MHZ) &&
			!(txcmd->rate_flags[index] &
			  ENABLE_CHNL_WIDTH_80MHZ))
			skip_rate = true;
		/*First Time*/
			if (!index)
				pkt_info->max_retries = 0;
			pkt_info->max_retries +=
				txcmd->rate_retries[index];
		if (skip_rate)
			txcmd->rate_retries[index] = 0;

	}

	if (all_rates_invalid) {
		/*use min supported rate*/
		if (net_ratelimit())
			RPU_DEBUG_TX("RPU_WIFI:invalid rates\n");
		txcmd->num_rates = 1;
		txcmd->rate[0] = min_rate;
		txcmd->rate_retries[0] = 4;
		txcmd->rate_protection_type[0] = USE_PROTECTION_NONE;
		txcmd->rate_preamble_type[0] = DONT_USE_SHORT_PREAMBLE;
		txcmd->num_spatial_streams[0] = 1;
		txcmd->bcc_or_ldpc = 0;
		txcmd->stbc_enabled = 0;
		txcmd->rate_flags[0] = 0;
	}
}

static void get_rate_local(struct sk_buff *skb,
			   struct cmd_tx_ctrl *txcmd,
			   struct tx_pkt_info *pkt_info,
			   bool retry,
			   struct img_priv *priv,
			   struct ieee80211_tx_rate *img_latest_rates)
{
	unsigned int index, min_rate;
	bool is_mcs = false, is_mgd = false;
	struct ieee80211_tx_rate *txrate;
	struct ieee80211_rate *rate;
	unsigned char mcs_rate_num = 0;
	int mcs_indx;
	int mgd_rate;
	int mgd_mcast_rate;
	int prot_type;
	unsigned char nss = 1;
	bool all_rates_invalid = true;
	struct ieee80211_tx_info *c = IEEE80211_SKB_CB(skb);
	struct ieee80211_hdr *hdr = (struct ieee80211_hdr *) skb->data;

	min_rate = priv->hw->wiphy->bands[c->band]->bitrates[0].hw_value;

	/* Some defaults*/
	txcmd->num_rates = 0;
	txcmd->stbc_enabled = 0;

	/* BCC (or) LDPC */
	if (c->flags & IEEE80211_TX_CTL_LDPC)
		txcmd->bcc_or_ldpc = 1;
	else
		txcmd->bcc_or_ldpc = 0;

	if (ieee80211_is_data(hdr->frame_control) &&
	    c->flags & IEEE80211_TX_CTL_AMPDU) {
		txcmd->aggregate_mpdu = AMPDU_AGGR_ENABLED;
	}

	for (index = 0; index < 4; index++) {
		bool skip_rate = false;

		txrate = &img_latest_rates[index];
		txcmd->rate_flags[index] = 0;

		if (txrate->idx < 0)
			continue;

		txcmd->num_rates++;
		txcmd->num_spatial_streams[index] = 1;

		/* No input from production_test proc, continue and use
		 * info from mac80211 RC
		 */

		/* It is an VHT MCS rate */
		if (txrate->flags & IEEE80211_TX_RC_VHT_MCS) {
			/*idx field is split
			 * into a higher 4 bits (Nss), starts
			 * with 0 and lower 4 bits (MCS number)
			 */
			is_mcs = true;
			mcs_rate_num = ieee80211_rate_get_vht_mcs(txrate);
			nss =  ieee80211_rate_get_vht_nss(txrate);
			txcmd->rate_flags[index] |= ENABLE_VHT_FORMAT;
		} else
		if (txrate->flags & IEEE80211_TX_RC_MCS) {
			is_mcs = true;
			mcs_rate_num  = txrate->idx;
			nss = mcs_rate_num/8 + 1;
			txcmd->rate_flags[index] |= ENABLE_11N_FORMAT;
		}

		mcs_indx = priv->params->mgd_mode_tx_fixed_mcs_indx;
		mgd_rate = priv->params->mgd_mode_tx_fixed_rate;
		mgd_mcast_rate = priv->params->mgd_mode_mcast_fixed_data_rate;

		/* Rate Index:
		 * From proc:
		 *    ** Multicast data packets
		 *    ** Unicast data packets
		 * From RC in mac80211
		 * Can be MCS(HT/VHT) or Rate (11abg)
		 */
		if (ieee80211_is_data(hdr->frame_control) &&
		    is_multicast_ether_addr(hdr->addr1) &&
		    (mgd_mcast_rate != -1)) {
			/* proc: Fixed MCS/Legacy rate for Multicast packets
			 */
			is_mgd = true;
			is_mcs = (mgd_mcast_rate & 0x80) == 0x80 ? true : false;

			if (!is_mcs) {
				if (mgd_mcast_rate == 55)
					mgd_mcast_rate = 11;
				else
					mgd_mcast_rate *= 2;
			}

			txcmd->rate[index] = mgd_mcast_rate;
			txcmd->rate_flags[index] =
				priv->params->mgd_mode_mcast_fixed_rate_flags;
			txcmd->bcc_or_ldpc =
				priv->params->mgd_mode_mcast_fixed_bcc_or_ldpc;
			if (txcmd->rate_flags[index] & ENABLE_11N_FORMAT)
				nss = (mgd_mcast_rate & 0x7F)/8 + 1;
			else
				nss = priv->params->mgd_mode_mcast_fixed_nss;
			txcmd->stbc_enabled =
				priv->params->mgd_mode_mcast_fixed_stbc_enabled;
			txcmd->rate_preamble_type[index] =
				priv->params->mgd_mode_mcast_fixed_preamble;
			if (is_mcs)
				update_mcs_packet_stat(mgd_mcast_rate & 0x7F,
						       txcmd->rate_flags[index],
						       priv);
		} else if (ieee80211_is_data(hdr->frame_control) &&
			   mcs_indx != -1) {
			/* proc: Fixed MCS for unicast
			 */
			is_mgd = true;

			txcmd->rate[index] = 0x80;
			txcmd->rate[index] |= (mcs_indx);
			txcmd->rate_flags[index] =
				priv->params->fixed_rate_flags;
			if (txcmd->rate_flags[index] & ENABLE_11N_FORMAT)
				nss = (mcs_indx)/8 + 1;
			else
				nss = priv->params->num_spatial_streams;
			txcmd->bcc_or_ldpc =
				priv->params->fixed_rate_bcc_or_ldpc;
			txcmd->stbc_enabled =
				priv->params->fixed_rate_stbc_enabled;

			update_mcs_packet_stat(mcs_indx,
					       txcmd->rate_flags[index],
					       priv);
		} else if (ieee80211_is_data(hdr->frame_control) &&
			   mgd_rate != -1) {
			/* proc: Fixed Legacy Rate for unicast
			 */
			is_mgd = true;
			txcmd->rate[index] = 0x80;
			txcmd->rate[index] = 0x00;

			if (mgd_rate == 55)
				txcmd->rate[index] |= ((mgd_rate) / 5);
			else
				txcmd->rate[index] |= ((mgd_rate * 10) / 5);

			txcmd->rate_flags[index] = 0;
			nss = 1;
			txcmd->bcc_or_ldpc	   = 0;
			txcmd->stbc_enabled	   = 0;
			txcmd->rate_preamble_type[index] =
				priv->params->fixed_rate_preamble_type;
		} else if (is_mcs) {
			txcmd->rate[index] = MARK_RATE_AS_MCS_INDEX;
			txcmd->rate[index] |= mcs_rate_num;
			update_mcs_packet_stat(mcs_rate_num,
					      txcmd->rate_flags[index],
					      priv);
		} else if (!is_mcs) {
			rate = &priv->hw->wiphy->bands[
				c->band]->bitrates[
				txrate->idx];
			txcmd->rate[index] = MARK_RATE_AS_RATE;
			txcmd->rate[index] |= rate->hw_value;
			nss = 1;
		}

		txcmd->num_spatial_streams[index] = nss;

		if (is_mgd) {
			if (priv->params->rate_protection_type)
				txcmd->rate_protection_type[index] =
					USE_PROTECTION_RTS;
			else
				txcmd->rate_protection_type[index] =
					USE_PROTECTION_NONE;
			txcmd->rate_retries[index] = 1;
			all_rates_invalid = false;
			break;
		}

		/* STBC Enabled/Disabled: valid if n_antennas > Nss */
		if (priv->params->rpu_max_nss > nss &&
		    (c->flags & IEEE80211_TX_CTL_STBC))
			txcmd->stbc_enabled = 1;

		txcmd->rate_retries[index] = txrate->count;

		if (txrate->flags &
		    IEEE80211_TX_RC_USE_SHORT_PREAMBLE)
			txcmd->rate_preamble_type[index] =
				USE_SHORT_PREAMBLE;
		else
			txcmd->rate_preamble_type[index] =
				DONT_USE_SHORT_PREAMBLE;

		prot_type = USE_PROTECTION_NONE;
		if (priv->params->rate_protection_type == 1) {
			/* Protection*/
			if (txrate->flags &
			    IEEE80211_TX_RC_USE_CTS_PROTECT)
				prot_type = USE_PROTECTION_CTS2SELF;
			else if (txrate->flags &
				 IEEE80211_TX_RC_USE_RTS_CTS)
				prot_type = USE_PROTECTION_RTS;
			else
				prot_type = USE_PROTECTION_NONE;

		}

		txcmd->rate_protection_type[index] = prot_type;

		if (txrate->flags &
				IEEE80211_TX_RC_GREEN_FIELD)
			txcmd->rate_flags[index] |=
					ENABLE_GREEN_FIELD;
		if (txrate->flags &
				IEEE80211_TX_RC_40_MHZ_WIDTH)
			txcmd->rate_flags[index] |=
					ENABLE_CHNL_WIDTH_40MHZ;
		if (txrate->flags &
				IEEE80211_TX_RC_80_MHZ_WIDTH)
			txcmd->rate_flags[index] |=
				ENABLE_CHNL_WIDTH_80MHZ;
		if (txrate->flags &
				IEEE80211_TX_RC_SHORT_GI)
			txcmd->rate_flags[index] |= ENABLE_SGI;

		/*Some Sanity Checks*/
		if (nss <= max(MAX_TX_STREAMS, MAX_RX_STREAMS))
			/*Got at-least one valid rate*/
			all_rates_invalid = false;
		else {
			if (net_ratelimit())
				RPU_DEBUG_IF("RPU_WIFI:Skip Nss: %d\n",
					      nss);
			skip_rate = true;
		}
		/* VHT 20MHz MCS9 is not valid*/
		if (txrate->flags & IEEE80211_TX_RC_VHT_MCS &&
			((txcmd->rate[index] & 0x7F) == 9) &&
			!(txcmd->rate_flags[index] &
			  ENABLE_CHNL_WIDTH_40MHZ) &&
			!(txcmd->rate_flags[index] &
			  ENABLE_CHNL_WIDTH_80MHZ))
			skip_rate = true;
				pkt_info->max_retries = 0;
			pkt_info->max_retries +=
				txcmd->rate_retries[index];
		if (skip_rate)
			txcmd->rate_retries[index] = 0;

	}

	if (all_rates_invalid) {
		/*use min supported rate*/
		if (net_ratelimit())
			RPU_DEBUG_TX("RPU_WIFI:invalid rates\n");
		txcmd->num_rates = 1;
		txcmd->rate[0] = min_rate;
		txcmd->rate_retries[0] = 4;
		txcmd->rate_protection_type[0] = USE_PROTECTION_NONE;
		txcmd->rate_preamble_type[0] = DONT_USE_SHORT_PREAMBLE;
		txcmd->num_spatial_streams[0] = 1;
		txcmd->bcc_or_ldpc = 0;
		txcmd->stbc_enabled = 0;
		txcmd->rate_flags[0] = 0;
	}
}

static int rpu_send_cmd(unsigned char *buf,
			unsigned int len,
			unsigned char id)
{
	struct host_rpu_msg_hdr *hdr = (struct host_rpu_msg_hdr *)buf;
	struct sk_buff *nbuf;
	struct rpu_if_data *p;
	struct img_priv *priv;

	rcu_read_lock();

	p = (struct rpu_if_data *)(rcu_dereference(rpu_if));

	if (!p) {
		pr_err("%s: Unable to retrieve rpu_if\n", __func__);
		WARN_ON(1);
		rcu_read_unlock();
		return -1;
	}
	priv = p->context;
	nbuf = alloc_skb(len, GFP_ATOMIC);

	if (!nbuf) {
		rcu_read_unlock();
		WARN_ON(1);
		return -ENOMEM;
	}
	hdr->id = id;
	RPU_DEBUG_IF("%s-RPUIF: Sending command:%d, outstanding_cmds: %d\n",
		     p->name, hdr->id, cmd_info.outstanding_ctrl_req);
	hdr->length = len;
	hdr->descriptor_id = 0;
	hdr->descriptor_id |= 0x0000ffff;

	memcpy(skb_put(nbuf, len), buf, len);
	priv->stats->outstanding_cmd_cnt = cmd_info.outstanding_ctrl_req;

	/* Take lock to make the control commands sequential in case of SMP*/
	spin_lock_bh(&cmd_info.control_path_lock);

	if (
	    (cmd_info.outstanding_ctrl_req < MAX_OUTSTANDING_CTRL_REQ)) {
		RPU_DEBUG_IF("Sending the CMD, got Access\n");
		hal_ops.send((void *)nbuf, NULL, 0);
		priv->stats->gen_cmd_send_count++;
	} else {
		RPU_DEBUG_IF("Sending the CMD, Waiting in Queue: %d\n",
			     cmd_info.outstanding_ctrl_req);
		skb_queue_tail(&cmd_info.outstanding_cmd, nbuf);
	}

	/* sent but still no proc_done / unsent due to pending requests */
	cmd_info.outstanding_ctrl_req++;
	spin_unlock_bh(&cmd_info.control_path_lock);
	rcu_read_unlock();

	return 0;
}


int rpu_prog_reset(unsigned int reset_type,
		   unsigned int rpu_mode)
{
	struct cmd_reset reset;
	struct img_priv *priv;
	struct rpu_if_data *p;
	unsigned int i;

	rcu_read_lock();
	p = (struct rpu_if_data *)(rcu_dereference(rpu_if));

	if (!p) {
		WARN_ON(1);
		rcu_read_unlock();
		return -1;
	}
	rcu_read_unlock();
	priv = p->context;

	memset(&reset, 0, sizeof(struct cmd_reset));

	reset.type = reset_type;

	if (reset_type == LMAC_ENABLE) {
		reset.num_spatial_streams =
			priv->params->rpu_max_nss;
		reset.lmac_mode = rpu_mode;
		reset.antenna_sel = priv->params->antenna_sel;

		if (priv->params->production_test == 0 &&
			priv->params->bypass_vpd == 0) {
			memcpy(reset.rf_params, priv->params->rf_params_vpd,
			       RF_PARAMS_SIZE);
		} else {
			memcpy(reset.rf_params, priv->params->rf_params,
			       RF_PARAMS_SIZE);
		}

		reset.system_rev = priv->stats->system_rev;
		reset.bg_scan.enabled = priv->params->bg_scan_enable;

		if (reset.bg_scan.enabled) {
			for (i = 0; i < priv->params->bg_scan_num_channels;
			     i++) {
				reset.bg_scan.channel_list[i] =
					priv->params->bg_scan_channel_list[i];
				reset.bg_scan.channel_flags[i] =
					priv->params->bg_scan_channel_flags[i];
			}
			reset.bg_scan.num_channels =
				priv->params->bg_scan_num_channels;
			reset.bg_scan.scan_intval =
				priv->params->bg_scan_intval;
			reset.bg_scan.channel_dur =
				/* Channel spending time */
				priv->params->bg_scan_chan_dur;

			reset.bg_scan.serv_channel_dur =
				/* operating channel spending time */
				priv->params->bg_scan_serv_chan_dur;
		}
	}

	return rpu_send_cmd((unsigned char *) &reset,
				    sizeof(struct cmd_reset), RPU_CMD_RESET);
}

int rpu_prog_txpower(unsigned int txpower)
{
	struct cmd_tx_pwr power;

	memset(&power, 0, sizeof(struct cmd_tx_pwr));
	power.tx_pwr = txpower;
	power.if_index = 0;

	return rpu_send_cmd((unsigned char *) &power,
				    sizeof(struct cmd_tx_pwr),
				    RPU_CMD_TX_POWER);
}

int rpu_prog_btinfo(unsigned int bt_state)
{
	struct cmd_bt_info bt_info;

	memset(&bt_info, 0, sizeof(struct cmd_bt_info));
	bt_info.bt_state = bt_state;

	return rpu_send_cmd((unsigned char *) &bt_info,
					sizeof(struct cmd_bt_info),
					RPU_CMD_BT_INFO);
}

int rpu_prog_vif_ctrl(int index,
		      unsigned char *mac_addr,
		      unsigned int vif_type,
		      unsigned int op)
{
	struct cmd_vifctrl vif_ctrl;

	memset(&vif_ctrl, 0, sizeof(struct cmd_vifctrl));
	vif_ctrl.mode = vif_type;
	memcpy(vif_ctrl.mac_addr, mac_addr, 6);
	vif_ctrl.if_index = index;
	vif_ctrl.if_ctrl = op;

	return rpu_send_cmd((unsigned char *) &vif_ctrl,
				    sizeof(struct cmd_vifctrl),
				    RPU_CMD_VIF_CTRL);
}


int rpu_prog_mcast_addr_cfg(unsigned char *mcast_addr,
			    unsigned int op)
{
	struct cmd_mcst_addr_cfg mcast_config;

	memset(&mcast_config, 0, sizeof(struct cmd_mcst_addr_cfg));

	mcast_config.op = op;
	memcpy(mcast_config.mac_addr, mcast_addr, 6);

	return rpu_send_cmd((unsigned char *) &mcast_config,
				    sizeof(struct cmd_mcst_addr_cfg),
				    RPU_CMD_MCST_ADDR_CFG);
}


int rpu_prog_mcast_filter_control(unsigned int mcast_filter_enable)
{
	struct cmd_mcst_filter_ctrl mcast_ctrl;

	memset(&mcast_ctrl, 0, sizeof(struct cmd_mcst_filter_ctrl));
	mcast_ctrl.ctrl = mcast_filter_enable;

	return rpu_send_cmd((unsigned char *) &mcast_ctrl,
				    sizeof(struct cmd_mcst_filter_ctrl),
				    RPU_CMD_MCST_FLTR_CTRL);
}


int rpu_prog_vht_bform(unsigned int vht_beamform_status,
		       unsigned int vht_beamform_period)
{
	struct cmd_vht_beamform vht_beamform;

	memset(&vht_beamform, 0, sizeof(struct cmd_vht_beamform));

	vht_beamform.vht_beamform_status = vht_beamform_status;
	vht_beamform.vht_beamform_period = vht_beamform_period;

	return rpu_send_cmd((unsigned char *) &vht_beamform,
				    sizeof(struct cmd_vht_beamform),
				    RPU_CMD_VHT_BEAMFORM_CTRL);
}

int rpu_prog_roc(unsigned int roc_ctrl,
		 unsigned int roc_channel,
		 unsigned int roc_duration,
		 unsigned int roc_type)
{
	struct cmd_roc cmd_roc;

	memset(&cmd_roc, 0, sizeof(struct cmd_roc));

	cmd_roc.roc_ctrl = roc_ctrl;
	cmd_roc.roc_channel	= roc_channel;
	cmd_roc.roc_duration = roc_duration;
	cmd_roc.roc_type = roc_type;

	return rpu_send_cmd((unsigned char *) &cmd_roc,
			sizeof(struct cmd_roc), RPU_CMD_ROC_CTRL);
}


int rpu_prog_peer_key(int vif_index,
		      unsigned char *vif_addr,
		      unsigned int op,
		      unsigned int key_id,
		      unsigned int key_type,
		      unsigned int cipher_type,
		      struct umac_key *key)
{
	struct cmd_setkey peer_key;

	memset(&peer_key, 0, sizeof(struct cmd_setkey));

	peer_key.if_index = vif_index;
	/* memcpy(peer_key.vif_addr, vif_addr, ETH_ALEN); */
	peer_key.ctrl = op;
	peer_key.key_id = key_id;
	img_ether_addr_copy(peer_key.mac_addr, key->peer_mac);

	peer_key.key_type = key_type;
	peer_key.cipher_type = cipher_type;
	memcpy(peer_key.key, key->key, MAX_KEY_LEN);
	peer_key.key_len = MAX_KEY_LEN;

	if (key->tx_mic) {
		memcpy(peer_key.key + MAX_KEY_LEN, key->tx_mic, TKIP_MIC_LEN);
		peer_key.key_len += TKIP_MIC_LEN;
	}
	if (key->rx_mic) {
		memcpy(peer_key.key + MAX_KEY_LEN + TKIP_MIC_LEN, key->rx_mic,
		       TKIP_MIC_LEN);
		peer_key.key_len += TKIP_MIC_LEN;
	}
	peer_key.rsc_len = 6;
	memset(peer_key.rsc, 0, 6);

	return rpu_send_cmd((unsigned char *) &peer_key,
				    sizeof(struct cmd_setkey), RPU_CMD_SETKEY);
}


int rpu_prog_if_key(int vif_index,
		    unsigned char *vif_addr,
		    unsigned int op,
		    unsigned int key_id,
		    unsigned int cipher_type,
		    struct umac_key *key)
	{
	struct cmd_setkey if_key;

	memset(&if_key, 0, sizeof(struct cmd_setkey));

	if_key.if_index = vif_index;
	/* memcpy(if_key.vif_addr, vif_addr, 6); */
	if_key.key_id = key_id;
	if_key.ctrl = op;

	if (op == KEY_CTRL_ADD) {
		if_key.cipher_type = cipher_type;

		if (cipher_type == CIPHER_TYPE_TKIP ||	cipher_type ==
		    CIPHER_TYPE_CCMP) {
			memcpy(if_key.key, key->key, MAX_KEY_LEN);
			if_key.key_len = MAX_KEY_LEN;

			if (key->tx_mic) {
				memcpy(if_key.key + MAX_KEY_LEN, key->tx_mic,
				       TKIP_MIC_LEN);
				if_key.key_len += TKIP_MIC_LEN;
			}
		} else {
			if_key.key_len =
				(cipher_type == CIPHER_TYPE_WEP40) ? 5 : 13;
			memcpy(if_key.key, key->key, if_key.key_len);
		}
	}

	if_key.rsc_len = 6;
	if_key.key_type = KEY_TYPE_BCAST;
	memset(if_key.rsc, 0, 6);
	memset(if_key.mac_addr, 0xff, 6);

	return rpu_send_cmd((unsigned char *) &if_key,
				    sizeof(struct cmd_setkey), RPU_CMD_SETKEY);
}

int rpu_prog_ba_session_data(unsigned int op,
			     unsigned short tid,
			     unsigned short *ssn,
			     unsigned short ba_policy,
			     unsigned char *vif_addr,
			     unsigned char *peer_addr)
{
	struct cmd_ht_ba ba_cmd;
	int index;
	struct img_priv *priv;
	struct rpu_if_data *p;
	struct ieee80211_vif *vif = NULL;

	rcu_read_lock();
	p = (struct rpu_if_data *)(rcu_dereference(rpu_if));

	if (!p) {
		WARN_ON(1);
		rcu_read_unlock();
		return -1;
	}

	priv = p->context;

	memset(&ba_cmd, 0, sizeof(struct cmd_ht_ba));

	for (index = 0; index < priv->params->num_vifs; index++) {
		if (!(priv->active_vifs & (1 << index)))
			continue;

		vif = rcu_dereference(priv->vifs[index]);

		if (ether_addr_equal(vif->addr, vif_addr))
			break;
	}

	if (index == priv->params->num_vifs) {
		RPU_DEBUG_IF("no VIF found\n");
		return -1;
	}

	ba_cmd.if_index = index;
	ba_cmd.op = op;
	ba_cmd.policy = ba_policy;
	ba_cmd.tid = tid;
	ba_cmd.ssn = *ssn;
	img_ether_addr_copy(ba_cmd.vif_addr, vif_addr);
	img_ether_addr_copy(ba_cmd.peer_addr, peer_addr);

	rcu_read_unlock();

	return rpu_send_cmd((unsigned char *) &ba_cmd,
				    sizeof(struct cmd_ht_ba),
				    RPU_CMD_BA_SESSION_INFO);
}


int rpu_scan(int index, struct scan_req *req)
{
	struct cmd_scan *scan;
	unsigned char i;
	struct img_priv *priv;
	struct rpu_if_data *p;

	rcu_read_lock();
	p = (struct rpu_if_data *)(rcu_dereference(rpu_if));

	if (!p) {
		WARN_ON(1);
		rcu_read_unlock();
		return -1;
	}

	rcu_read_unlock();
	priv = p->context;

	scan = kmalloc(sizeof(struct cmd_scan) +
		       req->ie_len, GFP_KERNEL);

	if (scan == NULL) {
		RPU_DEBUG_IF("%s: Failed to allocate memory\n", __func__);
		return -ENOMEM;
	}

	memset(scan, 0, sizeof(struct cmd_scan));

	scan->if_index = index;

	/* We support 4 SSIDs */
	scan->n_ssids = req->n_ssids;
	scan->n_channel = req->n_channels;
	scan->type = priv->params->scan_type;

	for (i = 0; i < scan->n_channel; i++) {

		enum ieee80211_channel_flags passive_scan;

		passive_scan = IEEE80211_CHAN_NO_IR;

		scan->channel_list[i] =
			(ieee80211_frequency_to_channel(req->center_freq[i]));
		scan->chan_max_power[i] = req->freq_max_power[i];

		/* In mac80211 the flags are u32 but for scanning we need
		 * only first PASSIVE_SCAN flag, remaining flags may be used
		 * in future.
		 */
		if ((req->chan_flags[i] & passive_scan)
		    || (req->chan_flags[i] & IEEE80211_CHAN_RADAR)
		) {
			scan->chan_flags[i] = PASSIVE;
		} else {
			scan->chan_flags[i] = ACTIVE;
		}
	}

	scan->p2p_probe = req->p2p_probe;

	scan->extra_ies_len = req->ie_len;

	if (req->ie_len)
		memcpy(scan->extra_ies, req->ie, req->ie_len);

	if (req->n_ssids > 0) {
		for (i = 0; i < scan->n_ssids; i++) {
			scan->ssids[i].len = req->ssids[i].ssid_len;
			if (scan->ssids[i].len > 0)
				memcpy(scan->ssids[i].ssid, req->ssids[i].ssid,
				       req->ssids[i].ssid_len);
		}
	}
	RPU_DEBUG_SCAN("Scan request ie len = %d n_channel = %d,",
						req->ie_len,
						scan->n_channel);
	RPU_DEBUG_SCAN(" n_ssids = %d, if_index = %d type = %d p2p = %d\n",
						scan->n_ssids,
						scan->if_index,
						scan->type,
						scan->p2p_probe);

	for (i = 0; i < scan->n_ssids; i++) {
		if (scan->ssids[i].len != 0)
			RPU_DEBUG_SCAN("SSID: %s\n", scan->ssids[i].ssid);
		else
			RPU_DEBUG_SCAN("SSID: EMPTY\n");
	}

	RPU_DEBUG_SCAN("CHANNEL_LIST: Channel ==> Channel Flags\n");

	for (i = 0; i < scan->n_channel; i++)
		RPU_DEBUG_SCAN("Index %d: %d ==> %d\n", i,
				scan->channel_list[i], scan->chan_flags[i]);

	priv->stats->umac_scan_req++;

	rpu_send_cmd((unsigned char *)scan, sizeof(struct cmd_scan) +
			     req->ie_len, RPU_CMD_SCAN);
	kfree(scan);

	return 0;
}


int rpu_scan_abort(int index)
{
	struct cmd_scan_abort *scan_abort = NULL;

	scan_abort = (struct cmd_scan_abort *)
		kmalloc(sizeof(struct cmd_scan_abort), GFP_KERNEL);

	if (scan_abort == NULL) {
		RPU_DEBUG_IF("%s: Failed to allocate memory\n", __func__);
		return -ENOMEM;
	}

	memset(scan_abort, 0, sizeof(struct cmd_scan_abort));

	scan_abort->if_index = index;

	rpu_send_cmd((unsigned char *)scan_abort,
		     sizeof(struct cmd_scan_abort),
		     RPU_CMD_SCAN_ABORT);

	kfree(scan_abort);
	scan_abort = NULL;

	return 0;
}


int rpu_prog_channel(unsigned int prim_ch,
		     unsigned int center_freq1,
		     unsigned int center_freq2,
		     unsigned int ch_width,
		     unsigned int vif_index,
		     unsigned int freq_band)
{
	struct cmd_channel channel;
	struct rpu_if_data *p;
	struct img_priv *priv;
	unsigned int bw = ch_width;
	int ch_no1, ch_no2;
	int err = 0;
	unsigned int cf_offset = center_freq1;

	memset(&channel, 0, sizeof(struct cmd_channel));

	rcu_read_lock();
	p = (struct rpu_if_data *)(rcu_dereference(rpu_if));

	if (!p) {
		WARN_ON(1);
		rcu_read_unlock();
		return -1;
	}
	priv = p->context;

	if (priv->params->production_test) {
		/*Initialize - 20MHz*/
		bw = NL80211_CHAN_WIDTH_20;

		/* 40MHz BW: cfreq +/- 10MHz */
		cf_offset += (priv->params->chnl_40_offset * 10);
		bw += (priv->params->chnl_40_offset *
		       priv->params->chnl_40_offset);

		/* 80MHz BW: cfreq +/- 20MHz */
		if (vht_support) {
			bw += (priv->params->chnl_80_offset *
			       priv->params->chnl_80_offset);
			cf_offset += (priv->params->chnl_80_offset * 20);
		}
	}

	ch_no1 = ieee80211_frequency_to_channel(cf_offset);
	ch_no2 = 0;

	channel.primary_ch_number = prim_ch;
	channel.channel_number1 = ch_no1;
	channel.channel_number2 = ch_no2;
	channel.freq_band = freq_band;
	channel.vif_index = vif_index;

	switch (bw) {
	case NL80211_CHAN_WIDTH_20_NOHT:
	case NL80211_CHAN_WIDTH_20:
		channel.channel_bw = RPU_CHAN_WIDTH_20_NOHT;
		break;
	case NL80211_CHAN_WIDTH_40:
		channel.channel_bw = RPU_CHAN_WIDTH_40;
		break;
	case NL80211_CHAN_WIDTH_80:
		channel.channel_bw = RPU_CHAN_WIDTH_80;
		break;
	default:
		WARN_ON(1);
		break;
	}
	rcu_read_unlock();

	priv->chan_prog_done = 0;

	err = rpu_send_cmd((unsigned char *) &channel,
			   sizeof(struct cmd_channel),
			   RPU_CMD_CHANNEL);

	if (err)
		return err;

	if (wait_for_channel_prog_complete(priv))
		return -1;

	/*Store Locally*/
	priv->cur_chan.center_freq1 = ieee80211_channel_to_frequency(ch_no1,
								     freq_band);
	priv->cur_chan.center_freq2 = ieee80211_channel_to_frequency(ch_no2,
								     freq_band);
	priv->cur_chan.pri_chnl_num = prim_ch;
	priv->cur_chan.ch_width  = ch_width;
	priv->cur_chan.freq_band = freq_band;


	return 0;
}


int rpu_prog_chanctx_time_info(void)
{
	struct cmd_chanctx_time_config time_cfg;
	int i = 0;
	int j = 0;
	struct img_priv *priv = NULL;
	struct rpu_if_data *p = NULL;
	struct ieee80211_chanctx_conf *curr_conf = NULL;
	struct umac_chanctx *curr_ctx = NULL;
	int freq = 0;

	rcu_read_lock();

	p = (struct rpu_if_data *)(rcu_dereference(rpu_if));

	if (!p) {
		WARN_ON(1);
		rcu_read_unlock();
		return -1;
	}

	rcu_read_unlock();

	priv = p->context;

	memset(&time_cfg, 0, sizeof(struct cmd_chanctx_time_config));

	rcu_read_lock();

	for (i = 0; i < MAX_CHANCTX; i++) {
		curr_conf = rcu_dereference(priv->chanctx[i]);

		if (curr_conf) {
			curr_ctx = (struct umac_chanctx *)curr_conf->drv_priv;

			if (curr_ctx->nvifs) {
				freq = curr_conf->def.chan->center_freq;

				time_cfg.info[j].chan =
					ieee80211_frequency_to_channel(freq);
				time_cfg.info[j].percentage =
					(100 / priv->num_active_chanctx);
				j++;
			}
		}
	}

	rcu_read_unlock();

	return rpu_send_cmd((unsigned char *)&time_cfg,
			    sizeof(struct cmd_chanctx_time_config),
			    RPU_CMD_CHANCTX_TIME_INFO);
}


int rpu_prog_ps_state(int index,
		      unsigned char *vif_addr,
		      unsigned int powersave_state)
{
	struct cmd_ps ps_cfg;

	memset(&ps_cfg, 0, sizeof(struct cmd_ps));
	ps_cfg.mode = powersave_state;
	ps_cfg.if_index = index;

	return rpu_send_cmd((unsigned char *)&ps_cfg,
			    sizeof(struct cmd_ps), RPU_CMD_PS);
}


int rpu_prog_tx(unsigned int queue,
		unsigned int more_frms,
		int curr_chanctx_idx,
		unsigned int descriptor_id,
		bool retry)
{
	struct cmd_tx_ctrl *tx_cmd;
	struct rpu_if_data *p;
	struct img_priv *priv;
	struct umac_vif *uvif;
	struct sk_buff *skb, *skb_first, *tmp, *tx_cmd_skb;
	struct sk_buff_head *txq = NULL;
	struct ieee80211_hdr *mac_hdr;
	struct ieee80211_tx_info *tx_info_first;
	unsigned int hdrlen, pkt = 0, extra_copy_gram = 0;
	int vif_index;
	__u16 fc;
	dma_addr_t phy_addr;
	struct tx_config *tx;
	struct tx_pkt_info *pkt_info = NULL;
	unsigned int cmd_tx_size;
	struct ieee80211_sta *sta;
	struct ieee80211_tx_rate *img_latest_rates;

	rcu_read_lock();
	p = (struct rpu_if_data *)(rcu_dereference(rpu_if));

	if (!p) {
		WARN_ON(1);
		rcu_read_unlock();
		return -1;
	}

	priv = p->context;
	spin_lock_bh(&priv->tx.lock);
	tx = &priv->tx;
	txq = &priv->tx.pkt_info[curr_chanctx_idx][descriptor_id].pkt;
	pkt_info = &priv->tx.pkt_info[curr_chanctx_idx][descriptor_id];
	skb_first = skb_peek(txq);

	if (!skb_first) {
		spin_unlock_bh(&priv->tx.lock);
		rcu_read_unlock();
		return -10;
	}

	cmd_tx_size = sizeof(struct cmd_tx_ctrl) +
			(skb_queue_len(txq) * MAX_GRAM_PAYLOAD_LEN);

	tx_cmd_skb = alloc_skb(cmd_tx_size, GFP_KERNEL);

	if (!tx_cmd_skb) {
		spin_unlock_bh(&priv->tx.lock);
		rcu_read_unlock();
		return -ENOMEM;
	}

	tx_cmd = kzalloc(cmd_tx_size, GFP_KERNEL);

	if (!tx_cmd) {
		spin_unlock_bh(&priv->tx.lock);
		rcu_read_unlock();
		kfree(tx_cmd_skb);
		return -ENOMEM;
	}

	skb_put(tx_cmd_skb, cmd_tx_size);

	tx_info_first = IEEE80211_SKB_CB(skb_first);

	mac_hdr = (struct ieee80211_hdr *)skb_first->data;
	fc = mac_hdr->frame_control;
	hdrlen = ieee80211_hdrlen(fc);
	vif_index = vif_addr_to_index(mac_hdr->addr2, priv);

	/* GET The security Header Length only for data/qos-data/unicast PMF
	 * for 11W case.
	 */
	if ((ieee80211_is_data(fc) ||
	     ieee80211_is_data_qos(fc))
	    && ieee80211_has_protected(fc)) {
		/* hw_key == NULL: Encrypted in SW (injected frames)
		 * iv_len = 0: treat as SW encryption.
		 */
		if (tx_info_first->control.hw_key == NULL ||
		    !tx_info_first->control.hw_key->iv_len) {
			RPU_DEBUG_IF("%s: hw_key is %s and iv_len: 0\n",
			  __func__,
			  tx_info_first->control.hw_key?"valid":"NULL");
			tx_cmd->encrypt = ENCRYPT_DISABLE;
		 } else {
			RPU_DEBUG_IF("%s: cipher: %d, icv: %d",
				  __func__,
				  tx_info_first->control.hw_key->cipher,
				  tx_info_first->control.hw_key->icv_len);
			RPU_DEBUG_IF("iv: %d, key: %d\n",
				  tx_info_first->control.hw_key->iv_len,
				  tx_info_first->control.hw_key->keylen);
			/* iv_len is always the header and icv_len is always
			 * the trailer include only iv_len
			 */
			extra_copy_gram +=
				tx_info_first->control.hw_key->iv_len;
			tx_cmd->encrypt = ENCRYPT_ENABLE;
		}
	}

	if (tx_info_first->flags & IEEE80211_TX_CTL_TX_OFFCHAN)
		tx_cmd->tx_flags |= (1 << RPU_TX_FLAG_OFFCHAN_FRM);

	/* For injected frames (wlantest) hw_key is not set,as PMF uses
	 * CCMP always so hardcode this to CCMP IV LEN 8.
	 * For Auth3: It is completely handled in SW (mac80211).
	 */
	if (ieee80211_is_unicast_robust_mgmt_frame(skb_first) &&
	    ieee80211_has_protected(fc)) {
		extra_copy_gram += 8;
		tx_cmd->encrypt = ENCRYPT_ENABLE;
	}

	/* separate in to up to TSF and From TSF*/
	if (ieee80211_is_beacon(fc) || ieee80211_is_probe_resp(fc))
		extra_copy_gram += 8; /* Timestamp*/

	/* HAL Host-RPU HDR*/
	tx_cmd->hdr.id = RPU_CMD_TX;
	/* Keep the queue num and pool id in descriptor id */
	tx_cmd->hdr.descriptor_id = 0;
	tx_cmd->hdr.descriptor_id |= ((queue & 0x0000FFFF) << 16);
	tx_cmd->hdr.descriptor_id |= (descriptor_id & 0x0000FFFF);
	/* Not used anywhere currently */
	tx_cmd->hdr.length = sizeof(struct cmd_tx_ctrl);

	/* RPU_CMD_TX*/
	tx_cmd->if_index = vif_index;
	tx_cmd->queue_num = queue;
	tx_cmd->more_frms = more_frms;
	tx_cmd->descriptor_id = descriptor_id;
	tx_cmd->num_frames_per_desc = skb_queue_len(txq);
	tx_cmd->aggregate_mpdu = AMPDU_AGGR_DISABLED;

	/* These extra fields will be copied to GRAM along with
	 * tx_cmd and 802.11 header
	 */
	hdrlen += extra_copy_gram;
	tx_cmd->config_mac_hdr_len = hdrlen;

	priv->tx.pkt_info[curr_chanctx_idx][descriptor_id].vif_index =
		vif_index;
	priv->tx.pkt_info[curr_chanctx_idx][descriptor_id].queue = queue;

	uvif = (struct umac_vif *) (tx_info_first->control.vif->drv_priv);


	if (priv->params->minstrel_fastpath) {
		img_latest_rates = pkt_info->img_latest_rates;

		if (priv->params->production_test == 1)
			get_rate_prod(tx_cmd, priv);
		else {
			struct ieee80211_sta_rates *ratetbl = NULL;

			sta = ieee80211_find_sta(tx_info_first->control.vif,
						 mac_hdr->addr1);
			if (!sta) {
				spin_unlock_bh(&priv->tx.lock);
				rcu_read_unlock();
				kfree(tx_cmd_skb);
				return -ENOENT;
			}

			{
				int i;
				unsigned char rc;
				struct ieee80211_tx_rate rate;

				ratetbl = rcu_dereference(sta->rates);

				for (i = 0; i < 4; i++) {
					rc = tx_info_first->control.skip_table;
					rate = img_latest_rates[i];

					RPU_DEBUG_TX("PRE_IMG: skip_rc: %d ",
						      rc);
					RPU_DEBUG_TX("rateid: %d count: %d ",
						     rate.idx, rate.count);
					RPU_DEBUG_TX("flags: %d\n",
						     rate.flags);
					rate = tx_info_first->control.rates[i];

					RPU_DEBUG_TX("PRE_tx_info: rateid: %d ",
						      rate.idx);
					RPU_DEBUG_TX("count: %d flags: %d\n",
						     rate.count, rate.flags);

					RPU_DEBUG_TX("PRE_STA: rateid: %d  ",
						     ratetbl->rate[i].idx);
					RPU_DEBUG_TX("count: %d flags: %d\n",
						     ratetbl->rate[i].count,
						     ratetbl->rate[i].flags);
				}
			}

			ieee80211_get_tx_rates(tx_info_first->control.vif,
					       sta,
					       skb_first,
					       img_latest_rates,
					       /*ARRAY_SIZE(img_latest_rates)*/
					       4);
			{
				int i;
				unsigned char rc;
				struct ieee80211_tx_rate rate;

				for (i = 0; i < 4; i++) {
					rc = tx_info_first->control.skip_table;
					rate = img_latest_rates[i];

					RPU_DEBUG_TX("PRE_IMG: skip_rc: %d ",
						      rc);
					RPU_DEBUG_TX("rateid: %d count: %d ",
						     rate.idx, rate.count);
					RPU_DEBUG_TX("flags: %d\n",
						     rate.flags);
					rate = tx_info_first->control.rates[i];
					RPU_DEBUG_TX("tx_info: rateid: %d ",
						     rate.idx);
					RPU_DEBUG_TX("count: %d flags: %d\n",
						     rate.count, rate.flags);
				}
			}

			get_rate_local(skb_first,
				       tx_cmd,
				       pkt_info,
				       retry,
				       priv,
				       img_latest_rates);
		}

	} else {
		if (priv->params->production_test == 1)
			get_rate_prod(tx_cmd, priv);
		else
			/* Get the rate for first packet as all
			 * packets have same rate
			 */
			get_rate(skb_first,
				 tx_cmd,
				 pkt_info,
				 retry,
				 priv);
	}

	RPU_DEBUG_TX("%s-UMACTX: TX Frame, Queue = %d, descriptord_id = %d\n",
		     priv->name,
		     tx_cmd->queue_num, tx_cmd->descriptor_id);
	RPU_DEBUG_TX("num_frames= %d qlen: %d\n",
		     tx_cmd->num_frames_per_desc, skb_queue_len(txq));

	RPU_DEBUG_TX("%s-UMACTX: Num rates = %d, %x, %x, %x, %x\n",
		     priv->name,
		     tx_cmd->num_rates,
		     tx_cmd->rate[0],
		     tx_cmd->rate[1],
		     tx_cmd->rate[2],
		     tx_cmd->rate[3]);

	RPU_DEBUG_TX("%s-UMACTX: Retries   = %d, %d, %d, %d, %d\n",
		  priv->name,
		  pkt_info->max_retries,
		  tx_cmd->rate_retries[0],
		  tx_cmd->rate_retries[1],
		  tx_cmd->rate_retries[2],
		  tx_cmd->rate_retries[3]);

	tx->desc_chan_map[descriptor_id] = curr_chanctx_idx;

	skb_queue_walk_safe(txq, skb, tmp) {
		struct ieee80211_tx_info *tx_info;

		if (!skb || (pkt > tx_cmd->num_frames_per_desc))
			break;

		tx_info = IEEE80211_SKB_CB(skb_first);

		mac_hdr = (struct ieee80211_hdr *)skb->data;

		/* Only for Non-Qos and MGMT frames, for Qos-Data
		 * mac80211 handles the sequence no generation
		 */
		if (!retry &&
		    tx_info->flags &
		    IEEE80211_TX_CTL_ASSIGN_SEQ) {
			if (tx_info->flags &
			    IEEE80211_TX_CTL_FIRST_FRAGMENT) {
				uvif->seq_no += 0x10;
			}

			mac_hdr->seq_ctrl &= cpu_to_le16(IEEE80211_SCTL_FRAG);
			mac_hdr->seq_ctrl |= cpu_to_le16(uvif->seq_no);
		}

		/* Need it for tx_status later */
		priv->tx.pkt_info[curr_chanctx_idx][descriptor_id].hdr_len =
			hdrlen;
		/* Complete packet length */
		tx_cmd->pkt_length[pkt] = skb->len;

		RPU_DEBUG_DUMP_TX(DUMP_PREFIX_NONE,
				  16,
				  1,
				  skb->data,
				  skb->len,
				  1);
		/* We move the 11hdr from skb to RPU_CMD_TX, this is part of
		 * online DMA changes, HW expects only data portion
		 * While DMA. Not requried for loopback
		 */
		memcpy((unsigned char *)tx_cmd +
			sizeof(struct cmd_tx_ctrl)+
			(pkt * MAX_GRAM_PAYLOAD_LEN),
			mac_hdr, hdrlen);

		skb_pull(skb, hdrlen);
		if (hal_ops.map_tx_buf(descriptor_id, pkt,
				       skb->data, skb->len, &phy_addr)) {
			spin_unlock_bh(&priv->tx.lock);
			rcu_read_unlock();
			kfree(tx_cmd_skb);
			kfree(tx_cmd);
			return -30;
		}

		pkt++;
	}

		memcpy(tx_cmd_skb->data, tx_cmd, cmd_tx_size);

		/* SDK: Check if we can use the same txq initialized before in
		 * the function here
		 */
		txq = &priv->tx.pkt_info[curr_chanctx_idx][descriptor_id].pkt;

		RPU_DEBUG_TX("%s:%d Sending TX_CMD\n", __func__, __LINE__);
		RPU_DEBUG_DUMP_TX(DUMP_PREFIX_NONE,
				  16,
				  1,
				  tx_cmd_skb->data,
				  tx_cmd_skb->len,
				  1);

		spin_lock_bh(&cmd_info.control_path_lock);
		hal_ops.send((void *)tx_cmd_skb,
			     (void *)txq,
			     tx_cmd->descriptor_id);
		spin_unlock_bh(&cmd_info.control_path_lock);

		/* increment tx_cmd_send_count to keep track of number of
		 * tx_cmd send
		 */
		if (queue != WLAN_AC_BCN) {
			if (skb_queue_len(txq) == 1)
				priv->stats->tx_cmd_send_count_single++;
			else if (skb_queue_len(txq) > 1)
				priv->stats->tx_cmd_send_count_multi++;
		} else
			priv->stats->tx_cmd_send_count_beaconq++;

	spin_unlock_bh(&priv->tx.lock);
	rcu_read_unlock();
	kfree(tx_cmd);

	return 0;
}


int rpu_prog_vif_short_slot(int index,
			    unsigned char *vif_addr,
			    unsigned int use_short_slot)
{
	struct cmd_vif_cfg vif_cfg;

	memset(&vif_cfg, 0, sizeof(struct cmd_vif_cfg));
	vif_cfg.changed_bitmap = SHORTSLOT_CHANGED;
	vif_cfg.use_short_slot = use_short_slot;
	vif_cfg.if_index = index;
	img_ether_addr_copy(vif_cfg.vif_addr, vif_addr);

	return rpu_send_cmd((unsigned char *)&vif_cfg,
			    sizeof(struct cmd_vif_cfg),
			    RPU_CMD_VIF_CFG);
}


int rpu_prog_vif_atim_window(int index,
			     unsigned char *vif_addr,
			     unsigned int atim_window)
{
	struct cmd_vif_cfg vif_cfg;

	memset(&vif_cfg, 0, sizeof(struct cmd_vif_cfg));
	vif_cfg.changed_bitmap = ATIMWINDOW_CHANGED;
	vif_cfg.atim_window = atim_window;
	vif_cfg.if_index = index;
	img_ether_addr_copy(vif_cfg.vif_addr, vif_addr);

	return rpu_send_cmd((unsigned char *)&vif_cfg,
			    sizeof(struct cmd_vif_cfg),
			    RPU_CMD_VIF_CFG);
}


int rpu_prog_long_retry(int index,
			unsigned char *vif_addr,
			unsigned int long_retry)
{
	struct cmd_vif_cfg vif_cfg;

	memset(&vif_cfg, 0, sizeof(struct cmd_vif_cfg));
	vif_cfg.changed_bitmap = LONGRETRY_CHANGED;
	vif_cfg.long_retry = long_retry;
	vif_cfg.if_index = index;
	img_ether_addr_copy(vif_cfg.vif_addr, vif_addr);

	return rpu_send_cmd((unsigned char *)&vif_cfg,
			    sizeof(struct cmd_vif_cfg),
			    RPU_CMD_VIF_CFG);
}


int rpu_prog_short_retry(int index,
			 unsigned char *vif_addr,
			 unsigned int short_retry)
{

	struct cmd_vif_cfg vif_cfg;

	memset(&vif_cfg, 0, sizeof(struct cmd_vif_cfg));
	vif_cfg.changed_bitmap = SHORTRETRY_CHANGED;
	vif_cfg.short_retry = short_retry;
	vif_cfg.if_index = index;
	img_ether_addr_copy(vif_cfg.vif_addr, vif_addr);

	return rpu_send_cmd((unsigned char *)&vif_cfg,
			    sizeof(struct cmd_vif_cfg),
			    RPU_CMD_VIF_CFG);
}


int rpu_prog_vif_basic_rates(int index,
			     unsigned char *vif_addr,
			     unsigned int basic_rate_set)
{
	struct cmd_vif_cfg vif_cfg;

	memset(&vif_cfg, 0, sizeof(struct cmd_vif_cfg));
	vif_cfg.changed_bitmap = BASICRATES_CHANGED;
	vif_cfg.basic_rate_set = basic_rate_set;
	vif_cfg.if_index = index;
	img_ether_addr_copy(vif_cfg.vif_addr, vif_addr);

	return rpu_send_cmd((unsigned char *)&vif_cfg,
			    sizeof(struct cmd_vif_cfg),
			    RPU_CMD_VIF_CFG);


}


int rpu_prog_vif_aid(int index,
		     unsigned char *vif_addr,
		     unsigned int aid)
{
	struct cmd_vif_cfg vif_cfg;

	memset(&vif_cfg, 0, sizeof(struct cmd_vif_cfg));
	vif_cfg.changed_bitmap = AID_CHANGED;
	vif_cfg.aid = aid;
	vif_cfg.if_index = index;
	img_ether_addr_copy(vif_cfg.vif_addr, vif_addr);

	return rpu_send_cmd((unsigned char *)&vif_cfg,
			    sizeof(struct cmd_vif_cfg),
			    RPU_CMD_VIF_CFG);
}


int rpu_prog_vif_op_channel(int index,
			    unsigned char *vif_addr,
			    unsigned char op_channel)
{

	struct cmd_vif_cfg vif_cfg;

	memset(&vif_cfg, 0, sizeof(struct cmd_vif_cfg));
	vif_cfg.changed_bitmap = OP_CHAN_CHANGED;
	vif_cfg.op_channel = op_channel;
	vif_cfg.if_index = index;
	img_ether_addr_copy(vif_cfg.vif_addr, vif_addr);

	return rpu_send_cmd((unsigned char *)&vif_cfg,
			    sizeof(struct cmd_vif_cfg),
			    RPU_CMD_VIF_CFG);
}


int rpu_prog_vif_conn_state(int index,
			    unsigned char *vif_addr,
			    unsigned int connect_state)
{

	struct cmd_vif_cfg vif_cfg;

	memset(&vif_cfg, 0, sizeof(struct cmd_vif_cfg));
	vif_cfg.changed_bitmap = CONNECT_STATE_CHANGED;
	vif_cfg.connect_state = connect_state;
	vif_cfg.if_index = index;
	img_ether_addr_copy(vif_cfg.vif_addr, vif_addr);
	return rpu_send_cmd((unsigned char *)&vif_cfg,
			    sizeof(struct cmd_vif_cfg),
			    RPU_CMD_VIF_CFG);
}


int rpu_prog_vif_assoc_cap(int index,
			   unsigned char *vif_addr,
			   unsigned int caps)
{
	struct cmd_vif_cfg vif_cfg;


	memset(&vif_cfg, 0, sizeof(struct cmd_vif_cfg));
	vif_cfg.changed_bitmap = CAPABILITY_CHANGED;
	vif_cfg.capability = caps;
	vif_cfg.if_index = index;
	img_ether_addr_copy(vif_cfg.vif_addr, vif_addr);

	return rpu_send_cmd((unsigned char *)&vif_cfg,
			    sizeof(struct cmd_vif_cfg),
			    RPU_CMD_VIF_CFG);

}


int rpu_prog_vif_beacon_int(int index,
			    unsigned char *vif_addr,
			    unsigned int bcn_int)
{
	struct cmd_vif_cfg vif_cfg;

	memset(&vif_cfg, 0, sizeof(struct cmd_vif_cfg));

	vif_cfg.changed_bitmap = BCN_INT_CHANGED;
	vif_cfg.beacon_interval = bcn_int;
	vif_cfg.if_index = index;
	img_ether_addr_copy(vif_cfg.vif_addr, vif_addr);

	return rpu_send_cmd((unsigned char *)&vif_cfg,
			    sizeof(struct cmd_vif_cfg),
			    RPU_CMD_VIF_CFG);
}


int rpu_prog_vif_dtim_period(int index,
			     unsigned char *vif_addr,
			     unsigned int dtim_period)
{
	struct cmd_vif_cfg vif_cfg;

	memset(&vif_cfg, 0, sizeof(struct cmd_vif_cfg));

	vif_cfg.changed_bitmap = DTIM_PERIOD_CHANGED;
	vif_cfg.dtim_period = dtim_period;
	vif_cfg.if_index = index;
	img_ether_addr_copy(vif_cfg.vif_addr, vif_addr);

	return rpu_send_cmd((unsigned char *)&vif_cfg,
			    sizeof(struct cmd_vif_cfg),
			    RPU_CMD_VIF_CFG);
}


int rpu_prog_vif_bssid(int index,
		       unsigned char *vif_addr,
		       unsigned char *bssid)
{
	struct cmd_vif_cfg vif_cfg;

	memset(&vif_cfg, 0, sizeof(struct cmd_vif_cfg));
	vif_cfg.changed_bitmap = BSSID_CHANGED;
	img_ether_addr_copy(vif_cfg.bssid, bssid);
	img_ether_addr_copy(vif_cfg.vif_addr, vif_addr);
	vif_cfg.if_index = index;

	return rpu_send_cmd((unsigned char *)&vif_cfg,
			    sizeof(struct cmd_vif_cfg),
			    RPU_CMD_VIF_CFG);
}


int rpu_prog_vif_smps(int index,
		      unsigned char *vif_addr,
		      unsigned char smps_mode)
{
	struct cmd_vif_cfg vif_cfg;

	memset(&vif_cfg, 0, sizeof(struct cmd_vif_cfg));
	vif_cfg.changed_bitmap = SMPS_CHANGED;
	vif_cfg.if_index = index;
	img_ether_addr_copy(vif_cfg.vif_addr, vif_addr);

	switch (smps_mode) {
	case IEEE80211_SMPS_STATIC:
		vif_cfg.smps_info |= SMPS_ENABLED;
		break;
	case IEEE80211_SMPS_DYNAMIC:
		vif_cfg.smps_info |= SMPS_ENABLED;
		vif_cfg.smps_info |= SMPS_MODE;
		break;
	case IEEE80211_SMPS_AUTOMATIC:/* will be one of the above*/
	case IEEE80211_SMPS_OFF:
		break;
	default:
		WARN(1, "Invalid SMPS Mode: %d\n", smps_mode);
	}

	return rpu_send_cmd((unsigned char *)&vif_cfg,
			    sizeof(struct cmd_vif_cfg),
			    RPU_CMD_VIF_CFG);
}

int rpu_prog_txq_params(int index,
			unsigned char *addr,
			unsigned int queue,
			unsigned int aifs,
			unsigned int txop,
			unsigned int cwmin,
			unsigned int cwmax,
			unsigned int uapsd)
{
	struct cmd_txq_params params;

	memset(&params, 0, (sizeof(struct cmd_txq_params)));

	params.if_index = index;
	img_ether_addr_copy(params.vif_addr, addr);
	params.queue_num = queue;
	params.aifsn = aifs;
	params.txop = txop;
	params.cwmin = cwmin;
	params.cwmax = cwmax;
	params.uapsd = uapsd;

	return rpu_send_cmd((unsigned char *) &params,
			    sizeof(struct cmd_txq_params),
			    RPU_CMD_TXQ_PARAMS);
}

int rpu_prog_rcv_bcn_mode(unsigned int bcn_rcv_mode)
{
	struct cmd_vif_cfg vif_cfg;

	memset(&vif_cfg, 0, sizeof(struct cmd_vif_cfg));
	vif_cfg.changed_bitmap = RCV_BCN_MODE_CHANGED;
	vif_cfg.bcn_mode = bcn_rcv_mode;

	return rpu_send_cmd((unsigned char *)&vif_cfg,
			    sizeof(struct cmd_vif_cfg),
			    RPU_CMD_VIF_CFG);

}

int rpu_prog_aux_adc_chain(unsigned int chain_id)
{
	struct cmd_aux_adc_chain_sel aadc_chain_sel;

	memset(&aadc_chain_sel, 0, sizeof(struct cmd_aux_adc_chain_sel));
	aadc_chain_sel.chain_id = chain_id;

	return rpu_send_cmd((unsigned char *)&aadc_chain_sel,
			    sizeof(struct cmd_aux_adc_chain_sel),
			    RPU_CMD_AUX_ADC_CHAIN_SEL);
}

int rpu_prog_cont_tx(int val)
{
	struct cmd_cont_tx status;

	memset(&status, 0, sizeof(struct cmd_cont_tx));
	status.op = val;

	return rpu_send_cmd((unsigned char *)&status,
			    sizeof(struct cmd_cont_tx),
			    RPU_CMD_CONT_TX);
}

int rpu_prog_mib_stats(void)
{
	struct host_rpu_msg_hdr mib_stats_cmd;

	RPU_DEBUG_IF("cmd mib stats\n");
	memset(&mib_stats_cmd, 0, sizeof(struct host_rpu_msg_hdr));

	return rpu_send_cmd((unsigned char *)&mib_stats_cmd,
			    sizeof(struct host_rpu_msg_hdr),
			    RPU_CMD_MIB_STATS);
}

int rpu_prog_clear_stats(void)
{
	struct host_rpu_msg_hdr clear_stats_cmd;

	RPU_DEBUG_IF("cmd clear stats\n");
	memset(&clear_stats_cmd, 0, sizeof(struct host_rpu_msg_hdr));

	return rpu_send_cmd((unsigned char *)&clear_stats_cmd,
			    sizeof(struct host_rpu_msg_hdr),
			    RPU_CMD_CLEAR_STATS);
}

int rpu_prog_phy_stats(void)
{
	struct host_rpu_msg_hdr phy_stats_cmd;

	RPU_DEBUG_IF("cmd phy stats\n");
	memset(&phy_stats_cmd, 0, sizeof(struct host_rpu_msg_hdr));

	return rpu_send_cmd((unsigned char *)&phy_stats_cmd,
			    sizeof(struct host_rpu_msg_hdr),
			    RPU_CMD_PHY_STATS);
}


int rpu_prog_radar_detect(unsigned int op_code)
{
	struct cmd_detect_radar dfs_op;

	RPU_DEBUG_IF("cmd radar detect\n");
	dfs_op.radar_detect_op = op_code;

	return rpu_send_cmd((unsigned char *) &dfs_op,
			    sizeof(struct cmd_detect_radar),
			    RPU_CMD_DETECT_RADAR);
}


int rpu_prog_tx_deinit(int vif_index, char *peer_addr)
{
	struct cmd_tx_deinit cmd_tx_deinit;

	memset(&cmd_tx_deinit, 0, (sizeof(struct cmd_tx_deinit)));
	cmd_tx_deinit.if_index = vif_index;
	img_ether_addr_copy(cmd_tx_deinit.peer_addr, peer_addr);
	return rpu_send_cmd((unsigned char *) &cmd_tx_deinit,
			    sizeof(struct cmd_tx_deinit),
			    RPU_CMD_TX_DEINIT);
}


int rpu_msg_handler(void *nbuff)
{
	unsigned int event;
	unsigned char *buff;
	struct host_rpu_msg_hdr *hdr;
	struct rpu_if_data *p;
	struct sk_buff *skb = (struct sk_buff *)nbuff;
	struct sk_buff *pending_cmd;
	struct img_priv *priv;
	int curr_chanctx_idx = -1;

	rcu_read_lock();

	p = (struct rpu_if_data *)(rcu_dereference(rpu_if));

	if (!p) {
		WARN_ON(1);
		dev_kfree_skb_any(skb);
		rcu_read_unlock();
		return 0;
	}

	buff = skb->data;
	hdr = (struct host_rpu_msg_hdr *)buff;

	event = hdr->id & 0xffff;

	priv = (struct img_priv *)p->context;

	/* RPU_DEBUG_IF("%s-RPUIF: event %d received\n", p->name, event); */
	if (event == RPU_EVENT_RESET_COMPLETE) {
		struct rpu_event_reset_complete *r =
				(struct rpu_event_reset_complete *)buff;

		rpu_reset_complete(r->version, p->context);
		spin_lock_bh(&cmd_info.control_path_lock);

		if (cmd_info.outstanding_ctrl_req == 0) {
			pr_err("%s-RPUIF: Unexpected: Spurious reset_complete received. Ignoring and continuing.\n",
			       p->name);
		} else {
			cmd_info.outstanding_ctrl_req--;

			RPU_DEBUG_IF("After DEC: outstanding cmd: %d\n",
				     cmd_info.outstanding_ctrl_req);
			pending_cmd = skb_dequeue(&cmd_info.outstanding_cmd);

			if (likely(pending_cmd != NULL)) {
				RPU_DEBUG_IF("Send 1 outstanding cmd\n");
				hal_ops.send((void *)pending_cmd, NULL, 0);
				priv->stats->gen_cmd_send_count++;
			}
		}

		spin_unlock_bh(&cmd_info.control_path_lock);
	} else if (event == RPU_EVENT_SCAN_ABORT_COMPLETE) {
		priv->scan_abort_done = 1;
	} else if (event == RPU_EVENT_SCAN_COMPLETE) {
		rpu_scan_complete(p->context,
			(struct rpu_event_scanres *) buff,
			buff +	sizeof(struct rpu_event_scanres), skb->len);
	} else if (event == RPU_EVENT_RX) {
		if (priv->params->production_test) {
			priv->stats->rx_packet_data_count++;
			dev_kfree_skb_any(skb);
		} else {
			rpu_rx_frame(skb, p->context);
		}
	} else if (event == RPU_EVENT_TX_DONE) {
		if (priv->params->production_test &&
		    priv->params->start_prod_mode)
			rpu_proc_tx_complete((void *)buff,
						     p->context);
		else {
			/* Increment tx_done_recv_count to keep track of number
			 * of tx_done received do not count tx dones from host.
			 */
			priv->stats->tx_done_recv_count++;

			spin_lock(&priv->chanctx_lock);
			curr_chanctx_idx = priv->curr_chanctx_idx;
			spin_unlock(&priv->chanctx_lock);
			rpu_tx_complete((void *)buff,
						curr_chanctx_idx,
						p->context);
		}

		cmd_info.tx_done_recv_count++;
	} else if (event == RPU_EVENT_DISCONNECTED) {
		struct rpu_event_disconnect *dis =
			(struct rpu_event_disconnect *)buff;
		struct img_priv *priv = (struct img_priv *)p->context;
		struct ieee80211_vif *vif = NULL;
		int i = 0;

		if (dis->reason_code == REASON_NW_LOST) {
			for (i = 0; i < MAX_VIFS; i++) {
				if (!(priv->active_vifs & (1 << i)))
					continue;

				vif = rcu_dereference(priv->vifs[i]);

				if (ether_addr_equal(vif->addr,
						     dis->mac_addr)) {
					ieee80211_connection_loss(vif);
					break;
				}
			}
		}
	} else if (event == RPU_EVENT_MIB_STAT) {
		struct rpu_event_mib_stats  *mib_stats =
			(struct rpu_event_mib_stats *) buff;

		rpu_mib_stats(mib_stats, p->context);
	} else if (event == RPU_EVENT_MAC_STATS) {
		struct rpu_event_mac_stats  *mac_stats =
			(struct rpu_event_mac_stats *) buff;

		rpu_mac_stats(mac_stats, p->context);
	} else if (event == RPU_EVENT_PHY_STAT) {
		int i;
		struct rpu_event_phy_stats *phy =
			(struct rpu_event_phy_stats *)buff;
		RPU_DEBUG_IF("received phy stats event\n");
		RPU_DEBUG_IF("phy stats are\n");

		for (i = 0; i < 32; i++)
			RPU_DEBUG_IF("%x ", phy->phy_stats[i]);

		RPU_DEBUG_IF("\n\n\n");
	} else if (event == RPU_EVENT_NOA) {
		rpu_noa_event(FROM_EVENT_NOA, (void *)buff,
				      p->context, NULL);

	} else if (event == RPU_EVENT_COMMAND_PROC_DONE) {
		/*struct rpu_event_command_complete *cmd =
		 * (struct rpu_event_command_complete*)buff;
		 */
		RPU_DEBUG_IF("Received	PROC_DONE\n");

		spin_lock_bh(&cmd_info.control_path_lock);

		if (cmd_info.outstanding_ctrl_req == 0) {
			if (event == RPU_EVENT_COMMAND_PROC_DONE)
				pr_err("%s-RPUIF: Unexpected: Spurious proc_done received. Ignoring and continuing\n",
				       p->name);
		} else {
			if (event == RPU_EVENT_COMMAND_PROC_DONE)
				cmd_info.outstanding_ctrl_req--;

			RPU_DEBUG_IF("After DEC: outstanding cmd: %d\n",
				     cmd_info.outstanding_ctrl_req);

			pending_cmd = skb_dequeue(&cmd_info.outstanding_cmd);

			if (unlikely(pending_cmd != NULL)) {
				RPU_DEBUG_IF("Send 1 outstanding cmd\n");
				hal_ops.send((void *)pending_cmd, NULL, 0);
				priv->stats->gen_cmd_send_count++;
			}
		}
		spin_unlock_bh(&cmd_info.control_path_lock);

	} else if (event == RPU_EVENT_CH_PROG_DONE) {
		rpu_ch_prog_complete(event,
			(struct rpu_event_ch_prog_complete *)buff, p->context);
	} else if (event == RPU_EVENT_RADAR_DETECTED) {
		ieee80211_radar_detected(priv->hw);
	} else if (event == RPU_EVENT_RF_CALIB_DATA) {
		struct rpu_event_rf_calib_data  *rf_data = (void *)buff;

		rpu_rf_calib_data(rf_data, p->context);
	} else if (event == RPU_EVENT_ROC_STATUS) {
		struct rpu_event_roc_status *roc_status = (void *)buff;
		struct delayed_work *work = NULL;

		RPU_DEBUG_ROC("%s:%d ROC status is %d\n",
			__func__, __LINE__, roc_status->roc_status);

		switch (roc_status->roc_status) {
		case RPU_ROC_STAT_STARTED:
			if (priv->roc_params.roc_in_progress == 0) {
				priv->roc_params.roc_in_progress = 1;
				ieee80211_ready_on_channel(priv->hw);
				RPU_DEBUG_ROC("%s-RPUIF: ROC READY..\n",
					  priv->name);
			}
			break;
		case RPU_ROC_STAT_DONE:
		case RPU_ROC_STAT_STOPPED:
			if (priv->roc_params.roc_in_progress == 1) {
				work = &priv->roc_complete_work;
				ieee80211_queue_delayed_work(priv->hw,
							     work,
							     0);
			}
			break;
		}
	} else if (event == RPU_EVENT_CHAN_SWITCH) {
		rpu_proc_ch_sw_event((void *)buff,
					     p->context);

	} else if (event == RPU_EVENT_TX_DEINIT_DONE) {
		priv->tx_deinit_complete = 1;
	} else if (event == RPU_EVENT_FW_ERROR) {
		pr_err("%s: FW is in Error State, please reload.\n", __func__);
	} else {
		pr_warn("%s: Unknown event received %d\n", __func__, event);
	}

	if (event != RPU_EVENT_RX)
		dev_kfree_skb_any(skb);

	rcu_read_unlock();

	return 0;
}


int rpu_if_init(void *context, const char *name)
{
	struct rpu_if_data *p;

	RPU_DEBUG_IF("%s-RPUIF: rpu_if init called\n", name);

	p = kzalloc(sizeof(struct rpu_if_data), GFP_KERNEL);

	if (!p)
		return -ENOMEM;

	p->name = (char *)name;
	p->context = context;
	hal_ops.register_callback(rpu_msg_handler);
	rcu_assign_pointer(rpu_if, p);
	skb_queue_head_init(&cmd_info.outstanding_cmd);
	spin_lock_init(&cmd_info.control_path_lock);
	cmd_info.outstanding_ctrl_req = 0;

	return 0;
}


void rpu_if_deinit(void)
{
	struct rpu_if_data *p;

	RPU_DEBUG_IF("%s-RPUIF: Deinit called\n", rpu_if->name);

	p = rcu_dereference(rpu_if);
	rcu_assign_pointer(rpu_if, NULL);
	synchronize_rcu();
	kfree(p);
}


void rpu_if_free_outstnding(void)
{

	struct sk_buff *skb;

	/* First free the outstanding commands, we are not sending
	 * anymore commands to the FW except RESET.
	 */
	while ((skb = __skb_dequeue(&cmd_info.outstanding_cmd)))
		dev_kfree_skb_any(skb);

	cmd_info.outstanding_ctrl_req = 0;
}

