/* Copyright (c) 2011-2017 Imagination Technologies Ltd.
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "core.h"

#define RPU_DEBUG_RX(fmt, ...)				  \
do {							    \
	if ((rpu_debug & RPU_DEBUG_RX) && net_ratelimit()) \
		pr_debug(fmt, ##__VA_ARGS__);		      \
} while (0)


#define RPU_DEBUG_DUMP_RX(fmt, ...)			      \
do {								\
	if (rpu_debug & RPU_DEBUG_DUMP_RX)		       \
		print_hex_dump(KERN_DEBUG, fmt, ##__VA_ARGS__);   \
} while (0)


#define DUMP_RX (rpu_debug & RPU_DEBUG_DUMP_RX)

struct wlan_rx_pkt rx_control_info;

spinlock_t tsf_lock;


void rpu_rx_frame(struct sk_buff *skb, void *context)
{
	struct img_priv *priv = (struct img_priv *)context;
	struct ieee80211_hdr *hdr;
	struct ieee80211_rx_status rx_status;
	struct ieee80211_supported_band *band = NULL;
	int i;
	static unsigned int rssi_index;
	unsigned char mic_status;

	memcpy(&rx_control_info, skb->data, sizeof(struct wlan_rx_pkt));
	/* Remove RX control information:
	 * unused more_cmd_data in RX direction is used to indicate QoS/Non-Qos
	 * frames
	 */
	if (rx_control_info.hdr.more_cmd_data == 0) {
		/* Non-QOS case*/
		skb_pull(skb, sizeof(struct wlan_rx_pkt));
	} else {
		/* Qos Case: The RPU overwrites the 2 reserved bytes with data
		 * to maintain the 4 byte alignment of total length and 2 byte
		 * alignment
		 * of starting address (as expected by mac80211).
		 */
		skb_pull(skb, sizeof(struct wlan_rx_pkt) - 2);
		skb_trim(skb, skb->len - 2);
	}

	hdr = (struct ieee80211_hdr *)skb->data;

	/* Stats for debugging */
	if (ieee80211_is_data(hdr->frame_control))
		priv->stats->rx_packet_data_count++;
	else if (ieee80211_is_mgmt(hdr->frame_control))
		priv->stats->rx_packet_mgmt_count++;

	memset(&rx_status, 0, sizeof(struct ieee80211_rx_status));

	/* Remove this once hardware supports bip(11w) is available*/
	if (!ieee80211_is_robust_mgmt_frame(skb))
		rx_status.flag |= RX_FLAG_DECRYPTED;

	rx_status.flag |= RX_FLAG_MMIC_STRIPPED;

	mic_status = rx_control_info.rx_pkt_status;

	if (mic_status == RX_MIC_FAILURE_TKIP) {
		rx_status.flag |= RX_FLAG_MMIC_ERROR;
	} else if (mic_status == RX_MIC_FAILURE_CCMP) {
		/*Drop the Frame*/
		dev_kfree_skb_any(skb);
		return;
	}

	if (rx_control_info.channel < 15)
		rx_status.band = IEEE80211_BAND_2GHZ;
	else
		rx_status.band = IEEE80211_BAND_5GHZ;

	rx_status.freq = ieee80211_channel_to_frequency(rx_control_info.channel,
							rx_status.band);
	rx_status.signal = rx_control_info.rssi;

	/* RSSI Average for Production Mode*/
	if (priv->params->production_test == 1) {
		priv->params->rssi_average[rssi_index++] =
			(char)(rx_control_info.rssi);
		if (rssi_index >= MAX_RSSI_SAMPLES)
			rssi_index = 0;
	}

	rx_status.antenna = 0;

	if (rx_control_info.rate_flags & ENABLE_VHT_FORMAT) {
		/* Rate */
		if ((rx_control_info.rate_or_mcs & MARK_RATE_AS_MCS_INDEX) !=
			0x80) {
			RPU_DEBUG_RX("Invalid VHT MCS Information\n");
			rx_control_info.rate_or_mcs = 0;/*default to MCS0*/
		} else {
			rx_status.rate_idx =
				(rx_control_info.rate_or_mcs & 0x7f);
		}

		/* NSS */
		if (!rx_control_info.nss || rx_control_info.nss > 8)
			rx_status.vht_nss = 1;
		else
			 rx_status.vht_nss = rx_control_info.nss;

		/* CBW */
		if (rx_control_info.rate_flags & ENABLE_CHNL_WIDTH_80MHZ)
			rx_status.flag |= RX_VHT_FLAG_80MHZ;
		else if (rx_control_info.rate_flags & ENABLE_CHNL_WIDTH_40MHZ)
			rx_status.flag |= RX_FLAG_40MHZ;

		/* SGI */
		if (rx_control_info.rate_flags & ENABLE_SGI)
			rx_status.flag |= RX_FLAG_SHORT_GI;

		rx_status.flag |= RX_FLAG_VHT;
	} else
	if (rx_control_info.rate_flags & ENABLE_11N_FORMAT) {
		/* Rate */
		if ((rx_control_info.rate_or_mcs & MARK_RATE_AS_MCS_INDEX) !=
			0x80) {
			RPU_DEBUG_RX("Invalid HT MCS Information\n");
			rx_control_info.rate_or_mcs = 0;/*default to MCS0*/
		} else {
			rx_status.rate_idx =
				(rx_control_info.rate_or_mcs & 0x7f);
		}

		/* CBW */
		if (rx_control_info.rate_flags & ENABLE_CHNL_WIDTH_40MHZ)
			rx_status.flag |= RX_FLAG_40MHZ;

		/* SGI */
		if (rx_control_info.rate_flags & ENABLE_SGI)
			rx_status.flag |= RX_FLAG_SHORT_GI;

		/* HT Greenfield */
		if (rx_control_info.rate_flags & ENABLE_GREEN_FIELD)
			rx_status.flag |= RX_FLAG_HT_GF;

		rx_status.flag |= RX_FLAG_HT;
	} else {
		band = priv->hw->wiphy->bands[rx_status.band];

		if (!WARN_ON_ONCE(!band)) {
			for (i = 0; i < band->n_bitrates; i++) {
				if (rx_control_info.rate_or_mcs ==
				    band->bitrates[i].hw_value) {
					rx_status.rate_idx = i;
					break;
				}
			}
		} else {
			RPU_DEBUG_DUMP_RX(" ",
					  DUMP_PREFIX_NONE, 16, 1,
					  &rx_control_info,
					  sizeof(struct wlan_rx_pkt),
					  1);

			dev_kfree_skb_any(skb);
			return;
		}
	}

	if (((hdr->frame_control & IEEE80211_FCTL_FTYPE) ==
	     IEEE80211_FTYPE_MGMT) &&
	    ((hdr->frame_control & IEEE80211_FCTL_STYPE) ==
	     IEEE80211_STYPE_BEACON)) {
		rx_status.mactime =
			get_unaligned_le64(rx_control_info.timestamp);
		rx_status.flag |= RX_FLAG_MACTIME_START;
	}
	/* Beacon Time Stamp */
	if (((hdr->frame_control & IEEE80211_FCTL_FTYPE) ==
	     IEEE80211_FTYPE_MGMT) &&
	    ((hdr->frame_control & IEEE80211_FCTL_STYPE) ==
	     IEEE80211_STYPE_BEACON)) {

		for (i = 0; i < MAX_VIFS; i++) {
			struct ieee80211_vif *vif = NULL;

			vif = NULL;
			vif = priv->vifs[i];
			if (vif &&
			ether_addr_equal(hdr->addr2, vif->bss_conf.bssid)) {
				unsigned int ts2;
				unsigned int ldelta;

				spin_lock(&tsf_lock);
				priv->params->sync[i].status = 1;
				memcpy(priv->params->sync[i].bssid,
				       vif->bss_conf.bssid, 6);
				memcpy(priv->params->sync[i].ts1,
				       &rx_control_info.reserved1, 8);
				memcpy(&ts2,
				       &rx_control_info.reserved1[8], 4);
				memcpy(&priv->params->sync[i].ts2,
				       &rx_control_info.reserved1[8], 4);
				memcpy(&ldelta,
				       &rx_control_info.reserved1[12], 4);
				priv->params->sync[i].atu = 0;
				/* ts2 = get_real_ts2(ts2, ldelta); */
				if (frc_to_atu) {
					frc_to_atu(ts2,
						&priv->params->sync[i].atu, 0);
				priv->params->sync[i].atu -= ldelta * 1000;
				}
				spin_unlock(&tsf_lock);
			}
		}
	}

	RPU_DEBUG_RX(KERN_DEBUG
		      "%s-RX: RX frame, length = %d, RSSI = %d, rate = %d\n",
		      priv->name,
		      skb->len,
		      rx_control_info.rssi,
		      rx_control_info.rate_or_mcs);

	RPU_DEBUG_DUMP_RX(" ",
			DUMP_PREFIX_NONE, 16, 1,
			skb->data, skb->len, 1);

	memcpy(IEEE80211_SKB_RXCB(skb), &rx_status, sizeof(rx_status));
	ieee80211_rx(priv->hw, skb);
}
