/*
 * Copyright (c) 2011-2017 Imagination Technologies Ltd.
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#include "core.h"
#include "soc.h"
#include "utils.h"

#define MAX_ATTEMPTS_FOR_AUXADC_VALID 500
static unsigned int init_auxadc_status;

void clock_init(void *dev)
{
	clk_prepare_enable(devm_clk_get(dev, "rpu_fast_clk"));
	clk_prepare_enable(devm_clk_get(dev, "rpu_core_clk"));
	/* Aux ADC clock*/
	clk_prepare_enable(devm_clk_get(dev, "aux_adc_clk"));
}

void clock_deinit(void *dev)
{
	clk_disable_unprepare(devm_clk_get(dev, "rpu_fast_clk"));
	clk_disable_unprepare(devm_clk_get(dev, "rpu_core_clk"));
	/* Aux ADC clock*/
	clk_disable_unprepare(devm_clk_get(dev, "aux_adc_clk"));
}

void program_rpu_dma_start(struct hal_priv *priv)
{
}




int parse_dtb_config(struct hal_priv *priv)
{
	struct platform_device *pdev = priv->bus_dev;
	struct resource *res = NULL;
	int irq;
	struct device_node *np = pdev->dev.of_node;
	struct property *pp = NULL;
	int size;

	irq = platform_get_irq_byname(pdev, "rpu_irq");

	hpriv->irq = irq;

	res = platform_get_resource_byname(pdev, IORESOURCE_MEM,
					   "rpu_sysbus_base");
	if (!res)
		return pr_err("No dts entry : rpu_sysbus_base");

	hpriv->rpu_sysbus_base = res->start;
	hpriv->rpu_sysbus_len = res->end - res->start + 1;


	res = platform_get_resource_byname(pdev, IORESOURCE_MEM,
					   "rpu_pkd_gram_base");

	if (!res)
		return pr_err("No dts entry : rpu_pkd_gram_base");

	hpriv->rpu_pkd_gram_base = res->start;
	hpriv->rpu_pkd_gram_len = res->end - res->start + 1;

	res = platform_get_resource_byname(pdev, IORESOURCE_MEM,
					   "rpu_gram_base");

	if (res) {
		hpriv->rpu_gram_base = res->start;
		hpriv->rpu_gram_len = res->end - res->start + 1;
	}

	pp = of_find_property(np, "mac-address0", NULL);

	if (pp && (pp->length == ETH_ALEN) && pp->value)
		memcpy(&vif_macs[0], (void *)pp->value, ETH_ALEN);
	else if (mac_addr == NULL)
		mac_addr = DEFAULT_MAC_ADDRESS;

	pp = of_find_property(np, "mac-address1", NULL);

	if (pp && (pp->length == ETH_ALEN) && pp->value)
		memcpy(&vif_macs[1], (void *)pp->value, ETH_ALEN);

	if (mac_addr != NULL) {

		conv_str_to_byte(vif_macs[0], mac_addr, ETH_ALEN);

		img_ether_addr_copy(vif_macs[1], vif_macs[0]);

		/* Set the Locally Administered bit*/
		vif_macs[1][0] |= 0x02;

		/* Increment the MSB by 1 (excluding 2 special bits)*/
		vif_macs[1][0] += (1 << 2);
	}

	pp = of_find_property(np, "rf-params", &size);

	if (pp && pp->value) {
		hpriv->rf_params = kzalloc(size, GFP_KERNEL);

		if (!hpriv->rf_params)
			return -ENOMEM;

		memcpy(hpriv->rf_params, pp->value, size);
		rf_params_vpd = hpriv->rf_params;
	}

	pp = of_find_property(np, "num_streams", &size);

	if (pp && pp->value)
		num_streams_vpd = *((int *)pp->value);

	return 0;

}

void config_mem_region(unsigned int addr)
{

}

int program_soc_regs_deinit(void *dev)
{
	void __iomem *base_addr, *addr;
	unsigned long value;

	 /* disabling the AUXADC during deinit */
	 /* configure the registers of AUXADC */
	 base_addr = devm_ioremap(dev,
				  RPU_WRAPPER_BASE,
				  RPU_WRAPPER_LEN);

	if (!base_addr) {
		pr_err("%s:%d ioremap failed, ret: 0x%p\n",
			__func__, __LINE__, base_addr);
		return -ENOMEM;
	}

	/* Bit0=0 Disable AUXADC.*/
	addr = base_addr + CR_UCCP_LOW_RATE_ADC_CONTROL_OFFSET;
	value = readl(addr);
	value = (value & (~AUXADC_MASK_BIT0)) | AUXADC_DISBLE_BIT0;
	writel(value, addr);

	devm_iounmap(dev, base_addr);
	return 0;
}

int program_soc_regs_init(void *dev)
{
	void __iomem *base_addr, *addr;
	unsigned long value;
	int auxadc_loop_cnt = 0;
	int auxadc_status = 0;
	int attempts_auxadc_valid = 1;

	/* REG1: BT_RM */
	base_addr = devm_ioremap(dev,
				 SWITCHED_TOP_REGBANK_SECURE_BASE,
				 SWITCHED_TOP_REGBANK_SECURE_LEN);

	if (!base_addr) {
		pr_err("%s:%d ioremap failed, ret: 0x%p\n",
			__func__, __LINE__, base_addr);
		return -ENOMEM;
	}

	addr = base_addr + CR_ALT_SWITCHED_GPIO_SEL_PADS7_OFFSET;

	__set_bit(GPIO_SEL_PADS7_BT_RM_POS, addr);

	devm_iounmap(dev, base_addr);

	base_addr = devm_ioremap(dev,
				 PAD_CONTROL_BASE,
				 PAD_CONTROL_LEN);

	if (!base_addr) {
		pr_err("%s:%d ioremap failed, ret: 0x%p\n",
			__func__, __LINE__, base_addr);
		return -ENOMEM;
	}

	addr = base_addr + CR_ST_PAD_CONTROL_PU_PD15_OFFSET;

	value = readl(addr);
	value = (value & ~(GPIO_SEL_PAD15_BT_RM_MASK)) |
		 GPIO_SEL_PAD15_BT_RM_VAL;
	writel(value, addr);

	devm_iounmap(dev, base_addr);

reprogram_aux_adc:
	/* org sequence according to python.
	 * 1. Clear EOC
	 * 2. Enable AUXADC
	 * 3. reset auxadc
	 * 4. configure the registers
	 * 5. check validity of outout
	 * note: Extra thing added is that a delay of 10us is added after reset.
	 */

	/*configure the registers of AUXADC */
	base_addr = devm_ioremap(dev,
				 RPU_WRAPPER_BASE,
				 RPU_WRAPPER_LEN);

	if (!base_addr) {
		pr_err("%s:%d ioremap failed, ret: 0x%p\n",
			__func__, __LINE__, base_addr);
		return -ENOMEM;
	}

	/* bit0: Clear EOC */
	addr = base_addr + CR_UCCP_LOW_RATE_ADC_EOC_CLEAR_OFFSET;
	value = readl(addr);
	value = (value & (~AUXADC_MASK_BIT0)) | AUXADC_ENABLE_BIT0;
	writel(value, addr);

	/* Bit0=1 Enable AUXADC.
	 * setting different parameters.
	 * Bit12=0: Select Continuous conversion (STARTMODE)
	 * 9=0: the voltage applied to vref (SELREF)
	 * 8=0: Select single-ended inputs. (SELDIFF)
	 * 7=0: internal bandgap power down. (SELBG)
	 */
	addr = base_addr + CR_UCCP_LOW_RATE_ADC_CONTROL_OFFSET;
	value = readl(addr);
	value = (value & (~AUXADC_ENABLE_MODE_REF_IN_BG_MASK)) |
		 AUXADC_ENABLE_BIT0;
	writel(value, addr);

	devm_iounmap(dev, base_addr);

	/*RESET AUXADC */
	base_addr = devm_ioremap(dev,
				 SWITCHED_TOP_REGBANK_GENERAL_BASE,
				 SWITCHED_TOP_REGBANK_GENERAL_LEN);

	if (!base_addr) {
		pr_err("%s:%d ioremap failed, ret: 0x%p\n",
			__func__, __LINE__, base_addr);
		return -ENOMEM;
	}
	addr = base_addr + CR_TOP_RSVD1_OFFSET;

	/* Bit0: reset AUXADC */
	value = readl(addr);
	value = (value & (~AUXADC_MASK_BIT0)) | AUXADC_ENABLE_BIT0;
	writel(value, addr);

	udelay(AUXADC_DELAY_10); /* delay of 10us */

	/* bring out of reset */
	value = readl(addr);
	value = (value & (~AUXADC_MASK_BIT0)) | AUXADC_DISBLE_BIT0;
	writel(value, addr);

	devm_iounmap(dev, base_addr);

	/*configure the registers of AUXADC */

	base_addr = devm_ioremap(dev,
				 RPU_WRAPPER_BASE,
				 RPU_WRAPPER_LEN);

	if (!base_addr) {
		pr_err("%s:%d ioremap failed, ret: 0x%p\n",
			__func__, __LINE__, base_addr);
		return -ENOMEM;
	}

	/* Bits[6:4] Select the channel */
	addr = base_addr + CR_UCCP_LOW_RATE_ADC_CONTROL_OFFSET;
	value = readl(addr);
	value = (value & (~AUXADC_CHANNEL_SELECT_MASK)) |
		 (AUXADC_CHANNEL_SELECTION << 4);
	writel(value, addr);

	/* AUX ADC channel expansion */
	addr = base_addr + CR_AUX_ADC_RPU_CHSEL_EXPAND_OFFSET;
	value = readl(addr);

	/*Reset all expansions */
	value = value & (~(AUX_ADC_RPU_CHSEL_CH0_MASK |
			   AUX_ADC_RPU_CHSEL_CH1_MASK |
			   AUX_ADC_RPU_CHSEL_CH2_MASK |
			   AUX_ADC_RPU_CHSEL_CH3_MASK));

	value = value | (AUXADC_CHANNEL_SELECTION << 0) |
			(AUXADC_CHANNEL_SELECTION << 4) |
			(AUXADC_CHANNEL_SELECTION << 8) |
			(AUXADC_CHANNEL_SELECTION << 12);
	writel(value, addr);

	/*
	 * Bit31=1: SOC.  set start of conversion
	 * bits(11:10)=(1:1)=3: Select 12-bit resolution.
	 * BIT[11:10] Selects the ADC resolution: 0/1/2/3 => 6/8/10/12bit
	 */
	addr = base_addr + CR_UCCP_LOW_RATE_ADC_CONTROL_OFFSET;
	value = readl(addr);
	value = (value & (~AUXADC_SOC_RESOLUTION_MASK)) |
			 (AUXADC_SOC_RESOLUTION_VALUE);
	writel(value, addr);

	/* delay of 10us. is it neg-edge triggered?
	 * python has 0.1 sec wait time
	 */
	udelay(AUXADC_DELAY_10);

	/* Bit31: Clear SOC; clear start of conversion */
	addr = base_addr + CR_UCCP_LOW_RATE_ADC_CONTROL_OFFSET;
	value = readl(addr);
	value = (value & (~AUXADC_CLEAR_SOC_MASK));
	writel(value, addr);

	/* conversion cycle, Checking the validity of AUXADC output before
	 * it is used for power calibration
	 */

	for (auxadc_loop_cnt = 0; auxadc_loop_cnt < 5; auxadc_loop_cnt++) {

		addr = base_addr + CR_UCCP_LOW_RATE_ADC_STATUS_OFFSET;

		auxadc_status = readl(addr); /* read AUXADC status*/

		/* Bit31: check End Of Conversion bit. value of 1 indicates
		 * end of conversion pulse of the Low Rate ADC..
		 */

		if ((auxadc_status & AUXADC_CHECK_EOC_MASK) != 0) {

			/* Bit0: Clear End Of Conversion */
			addr = base_addr +
			       CR_UCCP_LOW_RATE_ADC_EOC_CLEAR_OFFSET;
			value = readl(addr);
			value = (value & (~AUXADC_MASK_BIT0)) |
				 AUXADC_ENABLE_BIT0;
			writel(value, addr);
			init_auxadc_status = 1;
			break;
		}
	}

	devm_iounmap(dev, base_addr);

	/* if onetmie auxadc output is not valid, return an error */
	if (init_auxadc_status == 0) {
		attempts_auxadc_valid++;
		program_soc_regs_deinit(dev);
		/* current measurement shows that 1 attempt is
		 * taking approx 0.2ms
		 */
		if (attempts_auxadc_valid > MAX_ATTEMPTS_FOR_AUXADC_VALID) {
			pr_err("AUXADC output validity fail\n");
			return -1;
		}
		goto reprogram_aux_adc;
	}

	return 0;
}


struct soc_ops_tag soc_ops = {
	.program_rpu_dma_start = program_rpu_dma_start,
	.parse_dtb = parse_dtb_config,
	.clock_init = clock_init,
	.clock_deinit = clock_deinit,
	.set_mem_region = config_mem_region,
	.program_soc_regs_init = program_soc_regs_init,
	.program_soc_regs_deinit = program_soc_regs_deinit,
};
