/*
 * Copyright (c) 2011-2017 Imagination Technologies Ltd.
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "core.h"
#include "utils.h"

#define TX_TO_MACDEV(x) ((struct img_priv *) \
			 (container_of(x, struct img_priv, tx)))

static void wait_for_tx_complete(struct tx_config *tx)
{
	int count = 0;
	struct img_priv *priv = NULL;

	/* Find_last_bit: Returns the bit number of the first set bit,
	 * or size.
	 */
	while (find_last_bit(tx->buf_pool_bmp,
			     NUM_TX_DESCS) != NUM_TX_DESCS) {
		count++;

		if (count < TX_COMPLETE_TIMEOUT_TICKS) {
			current->state = TASK_INTERRUPTIBLE;
			schedule_timeout(1);
		} else {
			priv = TX_TO_MACDEV(tx);

			RPU_DEBUG_TX("%s-UMACTX:WARNING: ", priv->name);
			RPU_DEBUG_TX("TX complete failed!!\n");
			RPU_DEBUG_TX("%s-UMACTX:After ", priv->name);
			RPU_DEBUG_TX("%ld: bitmap is: 0x%lx\n",
			       TX_COMPLETE_TIMEOUT_TICKS,
			       tx->buf_pool_bmp[0]);
			break;
		}
	}

	if (count && (count < TX_COMPLETE_TIMEOUT_TICKS)) {
		RPU_DEBUG_TX("%s-UMACTX:TX complete after %d timer ticks\n",
			priv->name, count);
	}
}

int tx_queue_map(int queue)
{
	unsigned int ac[4] = {WLAN_AC_VO, WLAN_AC_VI, WLAN_AC_BE, WLAN_AC_BK};

	if (queue < 4)
		return ac[queue];

	return WLAN_AC_VO;
}

int tx_queue_unmap(int queue)
{
	unsigned int ac[4] = {3, 2, 1, 0};

	return ac[queue];
}


static int check_80211_aggregation(struct img_priv *priv,
				struct sk_buff *skb,
			       int ac,
			       int off_chanctx_idx,
			       int peer_id)
{
	struct ieee80211_tx_info *tx_info = IEEE80211_SKB_CB(skb);
	struct ieee80211_tx_info *tx_info_first = NULL;
	struct ieee80211_hdr *mac_hdr = NULL, *mac_hdr_first = NULL;
	struct sk_buff *skb_first;
	struct sk_buff_head *pend_pkt_q = NULL;
	struct tx_config *tx = &priv->tx;
	bool ampdu = false, is_qos = false, addr = true, rate = true;
	bool is_fixed_rate = false;
	int mcs_indx;
	int mgd_rate;

	mac_hdr = (struct ieee80211_hdr *)skb->data;
	pend_pkt_q = &tx->pending_pkt[off_chanctx_idx][peer_id][ac];
	skb_first = skb_peek(pend_pkt_q);

	mac_hdr_first = (struct ieee80211_hdr *)skb_first->data;
	tx_info_first = IEEE80211_SKB_CB(skb_first);

	ampdu = (tx_info->flags & IEEE80211_TX_CTL_AMPDU) ? true : false;

	if (tx_info->flags & IEEE80211_TX_CTL_RATE_CTRL_PROBE)
		ampdu = false;

	is_qos = ieee80211_is_data_qos(mac_hdr->frame_control);


	mcs_indx = priv->params->mgd_mode_tx_fixed_mcs_indx;
	mgd_rate = priv->params->mgd_mode_tx_fixed_rate;

	if ((priv->params->production_test == 1) ||
	    (mcs_indx != -1) ||
	    (mgd_rate != -1))
		is_fixed_rate = true;

	if (priv->params->minstrel_fastpath)
		is_fixed_rate = true;

	/* RPU has a limitation, it expects A1-A2-A3 to be same
	 * for all MPDU's within an AMPDU. This is a temporary
	 * solution, remove it when RPU has fix for this.
	 */
	if (skb_first &&
	    (((!ether_addr_equal(mac_hdr->addr1,
		       mac_hdr_first->addr1)) ||
	    (!ether_addr_equal(mac_hdr->addr2,
		       mac_hdr_first->addr2)) ||
	    (!ether_addr_equal(mac_hdr->addr3,
		       mac_hdr_first->addr3)))))
		addr = false;

	if (!is_fixed_rate &&
	   ((memcmp(&tx_info_first->control.rates[0],
		    &tx_info->control.rates[0],
		    sizeof(struct ieee80211_tx_rate) *
		    IEEE80211_TX_MAX_RATES) != 0) ||
	    (tx_info_first->flags != tx_info->flags)))
		rate = false;

	/*stats and debug*/
	if (!is_qos) {
		RPU_DEBUG_TX("Not Qos\n");
		priv->stats->tx_noagg_not_qos++;
	} else if (!ampdu) {
		RPU_DEBUG_TX("Not AMPDU\n");
		priv->stats->tx_noagg_not_ampdu++;
	} else if (!addr) {
		if (skb_first) {
			RPU_DEBUG_TX("first: A1: %pM-A2:%pM -A3%pM not same\n",
				      mac_hdr_first->addr1,
				      mac_hdr_first->addr2,
				      mac_hdr_first->addr3);
			RPU_DEBUG_TX("curr:  A1: %pM-A2:%pM -A3%pM not same\n",
				      mac_hdr->addr1,
				      mac_hdr->addr2,
				      mac_hdr->addr3);
		}
		priv->stats->tx_noagg_not_addr++;
	} else if (!rate) {
		RPU_DEBUG_TX("Rate info mistmatch\n");
		priv->stats->tx_noagg_not_rate++;
	}

	return (ampdu && is_qos && addr && rate);
}


static void tx_status(struct sk_buff *skb,
		      struct rpu_event_tx_done *tx_done,
		      unsigned int frame_idx,
		      struct img_priv *priv,
		      unsigned int q_len, struct tx_pkt_info *pkt_info)
{
	int index, i;
	struct ieee80211_tx_rate *txrate;
	struct ieee80211_tx_rate *tx_inf_rate = NULL;
	struct ieee80211_tx_info *tx_info = IEEE80211_SKB_CB(skb);
	struct ieee80211_supported_band *band = NULL;
	struct umac_vif *uvif = NULL;
	int ret = 0;
	int tx_done_retry_cnt = tx_done->retries_num[frame_idx] + 1;

	uvif = (struct umac_vif *)(tx_info->control.vif->drv_priv);

	/*Just inform ma8c0211, it will free the skb*/
	if (tx_done->frm_status[frame_idx] == TX_DONE_STAT_DISCARD) {
		ieee80211_free_txskb(priv->hw, skb);
		priv->stats->tx_dones_to_stack++;
		return;
	}

	if (tx_done->frm_status[frame_idx] == TX_DONE_STAT_SUCCESS)
		tx_info->flags |= IEEE80211_TX_STAT_ACK;
	else if (tx_info->flags & IEEE80211_TX_CTL_AMPDU)
		tx_info->flags |= IEEE80211_TX_STAT_AMPDU_NO_BACK;

	if (priv->params->minstrel_fastpath) {
		/* Per-AMPDU reporting logic: WIP
		 * TODO: aggrgeate BACK failures with optimized SSN.
		 * (or) do not inform mac80211 and send our own BAR.
		 */
		if (frame_idx)
			goto submit_skb;

		if (tx_info->flags & IEEE80211_TX_CTL_AMPDU)
			tx_info->flags |= IEEE80211_TX_STAT_AMPDU;

		/*TODO: optimize loop*/
		for (i = 0; i < q_len; i++) {
			if (tx_done->frm_status[i] == TX_DONE_STAT_SUCCESS)
				tx_info->status.ampdu_ack_len += 1;
		}

		tx_info->status.ampdu_len = q_len;
		priv->stats->tx_dones_to_stack += q_len;
	} else {
		tx_info->flags &= ~IEEE80211_TX_STAT_AMPDU;
		tx_info->flags &= ~IEEE80211_TX_CTL_AMPDU;
		priv->stats->tx_dones_to_stack++;
	}

	band = priv->hw->wiphy->bands[tx_info->band];


	for (index = 0; index < 4; index++) {
		tx_inf_rate = &tx_info->status.rates[index];
		if (priv->params->minstrel_fastpath) {
			txrate = &pkt_info->img_latest_rates[index];
			memcpy(tx_inf_rate, txrate,
			       sizeof(struct ieee80211_tx_rate));
		} else {
			txrate = (&tx_info->control.rates[index]);
		}


		if (txrate->idx < 0)
			break;

		if (tx_done_retry_cnt > txrate->count) {
			tx_inf_rate->count = txrate->count;
			tx_done_retry_cnt -= txrate->count;
		} else {
			tx_inf_rate->count = tx_done_retry_cnt;
			if (tx_inf_rate->count == 0)
				tx_inf_rate->count = 1;
			break;
		}
	}

	/* Invalidate the remaining indices */
	while (((index + 1) < 4)) {
		tx_info->status.rates[index + 1].idx = -1;
		tx_info->status.rates[index + 1].count = 0;
		index++;
	}


	{
		RPU_DEBUG_TX("%s-: tx_status\n", __func__);
		for (i = 0; i < 4; i++) {
			RPU_DEBUG_TX("ctrl_mcs: %x rate_index: %d ra_idx: %x ",
				tx_info->control.rates[i].idx,
				i,
				tx_info->status.rates[i].idx);
			RPU_DEBUG_TX("ra_count: %d done_mcs: %x ",
				tx_info->status.rates[i].count,
				tx_done->rate[frame_idx]);
			RPU_DEBUG_TX(" done_retries: %d\n",
				tx_done->retries_num[frame_idx]);
			RPU_DEBUG_TX("%s-: tx_status done\n", __func__);
		}
	}

	if (((tx_info->flags & IEEE80211_TX_CTL_TX_OFFCHAN)
	     || (uvif->chanctx &&
		 (uvif->chanctx->index == priv->roc_off_chanctx_idx))
	    ) &&
	    (atomic_dec_return(&priv->roc_params.roc_mgmt_tx_count) == 0)) {
		RPU_DEBUG_ROC("%s:%d TXDONE Frame: %d\n",
			__func__,
			__LINE__,
			atomic_read(&priv->roc_params.roc_mgmt_tx_count));
		if (priv->roc_params.roc_in_progress &&
		    priv->roc_params.roc_type == ROC_TYPE_OFFCHANNEL_TX) {
			CALL_RPU(rpu_prog_roc, ROC_STOP, 0, 0, 0);
			RPU_DEBUG_ROC("%s:%d", __func__, __LINE__);
			RPU_DEBUG_ROC("all offchan pending frames cleared\n");
		}
	}

submit_skb:
	ieee80211_tx_status(priv->hw, skb);
prog_rpu_fail:
	return;
}


static int get_token(struct img_priv *priv,
		     int curr_chanctx_idx,
		     int queue)
{
	int cnt = 0;
	int curr_bit = 0;
	int pool_id = 0;
	int token_id = NUM_TX_DESCS;
	struct tx_config *tx = &priv->tx;

	/* First search for a reserved token */
	for (cnt = 0; cnt < NUM_TX_DESCS_PER_AC; cnt++) {
		curr_bit = ((queue + (NUM_ACS * cnt)) % TX_DESC_BUCKET_BOUND);
		pool_id = ((queue + (NUM_ACS * cnt)) / TX_DESC_BUCKET_BOUND);

		if (!test_and_set_bit(curr_bit, &tx->buf_pool_bmp[pool_id])) {
			token_id = queue + (NUM_ACS * cnt);
			tx->outstanding_tokens[queue]++;
			break;
		}
	}

	/* If reserved token is not found search for a spare token
	 * (only for non beacon queues)
	 */
	if ((cnt == NUM_TX_DESCS_PER_AC) && (queue != WLAN_AC_BCN)) {
		for (token_id = NUM_TX_DESCS_PER_AC * NUM_ACS;
		     token_id < NUM_TX_DESCS;
		     token_id++) {
			curr_bit = (token_id % TX_DESC_BUCKET_BOUND);
			pool_id = (token_id / TX_DESC_BUCKET_BOUND);
			if (!test_and_set_bit(curr_bit,
					      &tx->buf_pool_bmp[pool_id])) {
				tx->outstanding_tokens[queue]++;
				break;
			}
		}
	}

	return token_id;
}

void free_token(struct img_priv *priv,
		int token_id,
		int queue)
{
	struct tx_config *tx = &priv->tx;
	int bit = -1;
	int pool_id = -1;
	int test = 0;
	unsigned int old_token = tx->outstanding_tokens[queue];

	bit = (token_id % TX_DESC_BUCKET_BOUND);
	pool_id = (token_id / TX_DESC_BUCKET_BOUND);

	__clear_bit(bit, &tx->buf_pool_bmp[pool_id]);

	tx->outstanding_tokens[queue]--;

	test = tx->outstanding_tokens[queue];
	if (WARN_ON_ONCE(test < 0 || test > 4)) {
		RPU_DEBUG_TX("%s: invalid outstanding_tokens: %d, old:%d\n",
			      __func__,
			      test,
			      old_token);
	}
}


struct curr_peer_info get_curr_peer_opp(struct img_priv *priv,
					int curr_chanctx_idx,
					int ac)
{
	unsigned int curr_peer_opp = 0;
	unsigned int curr_vif_op_chan = RPU_VIF_CHANCTX_TYPE_OPER;
	unsigned int i = 0;
	struct tx_config *tx = NULL;
	struct ieee80211_sta *sta = NULL;
	struct ieee80211_vif *vif = NULL;
	struct umac_sta *usta = NULL;
	struct umac_vif *uvif = NULL;
	int vif_index = -1;
	unsigned int init_peer_opp = 0;
	struct curr_peer_info peer_info;
	unsigned int pend_q_len;
	struct sk_buff_head *pend_q = NULL;

	tx = &priv->tx;

	init_peer_opp = tx->curr_peer_opp[curr_chanctx_idx][ac];
	/*TODO: Optimize this loop for BCN_Q
	 */
	for (i = 0; i < MAX_PEND_Q_PER_AC; i++) {
		curr_peer_opp = (init_peer_opp + i) % MAX_PEND_Q_PER_AC;

		rcu_read_lock();

		/* RoC Frame do not have a "sta" entry.
		 * so we need not handle RoC stuff here
		 */
		if (curr_peer_opp < MAX_PEERS) {
			sta = rcu_dereference(priv->peers[curr_peer_opp]);

			if (!sta) {
				rcu_read_unlock();
				continue;
			}

			usta = (struct umac_sta *)(sta->drv_priv);

			vif = rcu_dereference(priv->vifs[usta->vif_index]);

			if (!vif) {
				rcu_read_unlock();
				continue;
			}


			uvif = (struct umac_vif *)(vif->drv_priv);

			if (!uvif->chanctx && !uvif->off_chanctx) {
				rcu_read_unlock();
				continue;
			}

			if ((uvif->chanctx &&
			     (uvif->chanctx->index != curr_chanctx_idx)) ||
			    !uvif->chanctx) {
				if ((uvif->off_chanctx &&
				     (uvif->off_chanctx->index !=
				      curr_chanctx_idx)) ||
				    !uvif->off_chanctx) {
					rcu_read_unlock();
					continue;
				} else {
					curr_vif_op_chan =
						RPU_VIF_CHANCTX_TYPE_OFF;
				}
			} else {
				if (priv->roc_params.roc_in_progress &&
				    !priv->roc_params.need_offchan)
					curr_vif_op_chan =
						RPU_VIF_CHANCTX_TYPE_OFF;
				else
					curr_vif_op_chan =
						RPU_VIF_CHANCTX_TYPE_OPER;
			}
		} else {
			vif_index = (curr_peer_opp - MAX_PEERS);

			vif = rcu_dereference(priv->vifs[vif_index]);

			if (!vif) {
				rcu_read_unlock();
				continue;
			}

			uvif = (struct umac_vif *)(vif->drv_priv);

			if (!uvif->chanctx && !uvif->off_chanctx) {
				rcu_read_unlock();
				continue;
			}

			/* For a beacon queue we will process the frames
			 * irrespective of the current channel context.
			 * The FW will take care of transmitting them in the
			 * appropriate channel.
			 */

			if (ac != WLAN_AC_BCN &&
			    ((uvif->chanctx &&
			      (uvif->chanctx->index != curr_chanctx_idx)) ||
			     !uvif->chanctx)) {
				if ((uvif->off_chanctx &&
				     (uvif->off_chanctx->index !=
				      curr_chanctx_idx)) ||
				    !uvif->off_chanctx) {
					rcu_read_unlock();
					continue;
				} else {
					curr_vif_op_chan =
						RPU_VIF_CHANCTX_TYPE_OFF;
				}
			} else {
				if (priv->roc_params.roc_in_progress &&
				    !priv->roc_params.need_offchan)
					curr_vif_op_chan =
						RPU_VIF_CHANCTX_TYPE_OFF;
				else
					curr_vif_op_chan =
						RPU_VIF_CHANCTX_TYPE_OPER;
			}
		}

		rcu_read_unlock();
		pend_q = &tx->pending_pkt[curr_vif_op_chan][curr_peer_opp][ac];
		pend_q_len = skb_queue_len(pend_q);

		if (pend_q_len) {
			tx->curr_peer_opp[curr_chanctx_idx][ac] =
				(curr_peer_opp + 1) % MAX_PEND_Q_PER_AC;
			break;
		}
	}

	if (i == MAX_PEND_Q_PER_AC) {
		peer_info.id = -1;
		peer_info.op_chan_idx = -1;
	} else {
		peer_info.id = curr_peer_opp;
		peer_info.op_chan_idx = curr_vif_op_chan;
		RPU_DEBUG_TX("%s: Queue: %d Peer: %d op_chan: %d ",
			__func__,
			ac,
			curr_peer_opp,
			curr_vif_op_chan);
		RPU_DEBUG_TX("chanctx: %d got opportunity, pending: %d\n",
			curr_chanctx_idx,
			pend_q_len);
	}

	return peer_info;
}


void rpu_tx_proc_send_pend_frms_all(struct img_priv *priv,
					    int ch_id)
{
	int txq_len = 0;
	int i = 0, cnt = 0;
	int queue = 0;
	int curr_bit = 0;
	int pool_id = 0;
	int ret = 0;
	int start_ac, end_ac;
	unsigned int pkts_pend = 0;
	struct tx_config *tx = NULL;
	struct sk_buff_head *txq = NULL;

	tx = &priv->tx;

	for (i = 0; i < NUM_TX_DESCS; i++) {
		spin_lock_bh(&tx->lock);

		curr_bit = (i % TX_DESC_BUCKET_BOUND);
		pool_id = (i / TX_DESC_BUCKET_BOUND);

		if (test_and_set_bit(curr_bit, &tx->buf_pool_bmp[pool_id])) {
			spin_unlock_bh(&tx->lock);
			continue;
		}

		txq = &tx->pkt_info[ch_id][i].pkt;
		txq_len = skb_queue_len(txq);

		/* Not valid when txq len is 0 */
		queue = tx->pkt_info[ch_id][i].queue;

		if (!txq_len) {
			/* Reserved token */
			if (i < (NUM_TX_DESCS_PER_AC * NUM_ACS)) {
				queue = (i % NUM_ACS);
				start_ac = end_ac = queue;
			} else {
				/* Spare token:
				 * Loop through all AC's
				 */
				start_ac = WLAN_AC_VO;
				end_ac = WLAN_AC_BK;
			}

			for (cnt = start_ac; cnt >= end_ac; cnt--) {
				pkts_pend = rpu_tx_proc_pend_frms(priv,
									  cnt,
									  ch_id,
									  i);
				if (pkts_pend) {
					queue = cnt;
					break;
				}
			}

			if (pkts_pend == 0) {
				__clear_bit(curr_bit,
					    &tx->buf_pool_bmp[pool_id]);
				spin_unlock_bh(&tx->lock);
				continue;
			}
		}

		tx->outstanding_tokens[queue]++;
		spin_unlock_bh(&tx->lock);

		ret = __rpu_tx_frame(priv,
					     queue,
					     i,
					     ch_id,
					     0,
					     0); /* TODO: Currently sending 0
						  * since this param is not used
						  * as expected in the orig
						  * code for multiple frames etc
						  * Need to set this
						  * properly when the orig code
						  * logic is corrected
						  */
		if (ret < 0) {
			pr_err("%s: Queueing of TX frame to FW failed\n",
			       __func__);
		}
	}
}


int rpu_tx_proc_pend_frms(struct img_priv *priv,
				  int ac,
				  int curr_chanctx_idx,
				  int token_id)
{
	struct tx_config *tx = &priv->tx;
	unsigned long ampdu_len = 0;
	struct sk_buff *loop_skb = NULL;
	struct sk_buff *tmp = NULL;
	struct ieee80211_hdr *mac_hdr = NULL;
	struct ieee80211_tx_info *tx_info = NULL;
	struct umac_vif *uvif = NULL;
	struct ieee80211_vif *ivif = NULL;
	unsigned char *data = NULL;
	unsigned int max_tx_cmds = priv->params->max_tx_cmds;
	struct sk_buff_head *txq = NULL;
	struct sk_buff_head *pend_pkt_q = NULL;
	unsigned int total_pending_processed = 0;
	int pend_pkt_q_len = 0;
	struct curr_peer_info peer_info;
	int loop_cnt = 0;
	struct tx_pkt_info *pkt_info = NULL;

	peer_info = get_curr_peer_opp(priv,
				       curr_chanctx_idx,
				       ac);

	/* No pending frames for any peer in that AC.
	 */
	if (peer_info.id == -1)
		return 0;

	pend_pkt_q = &tx->pending_pkt[peer_info.op_chan_idx][peer_info.id][ac];

	pkt_info = &priv->tx.pkt_info[curr_chanctx_idx][token_id];
	txq = &pkt_info->pkt;


	/* Aggregate Only MPDU's with same RA, same Rate,
	 * same Rate flags, same Tx Info flags
	 */
	skb_queue_walk_safe(pend_pkt_q,
			    loop_skb,
			    tmp) {
		data = loop_skb->data;
		mac_hdr = (struct ieee80211_hdr *)data;

		tx_info = IEEE80211_SKB_CB(loop_skb);

		ivif = tx_info->control.vif;
		uvif = (struct umac_vif *)(ivif->drv_priv);

		ampdu_len += loop_skb->len;

		if (vht_support && !loop_cnt)
			if ((tx_info->control.rates[0].flags &
			     IEEE80211_TX_RC_MCS) &&
			    max_tx_cmds > MAX_SUBFRAMES_IN_AMPDU_HT)
				max_tx_cmds = MAX_SUBFRAMES_IN_AMPDU_HT;

		if (!check_80211_aggregation(priv,
					     loop_skb,
					     ac,
					     peer_info.op_chan_idx,
					     peer_info.id) ||
		    (skb_queue_len(txq) >= max_tx_cmds)) {
			break;
		}
		loop_cnt++;
		__skb_unlink(loop_skb, pend_pkt_q);
		skb_queue_tail(txq, loop_skb);
	}

	/* If our criterion rejects all pending frames, or
	 * pend_q is empty, send only 1
	 */
	if (!skb_queue_len(txq))
		skb_queue_tail(txq, skb_dequeue(pend_pkt_q));

	total_pending_processed = skb_queue_len(txq);

	pend_pkt_q_len = skb_queue_len(pend_pkt_q);
	if ((ac != WLAN_AC_BCN) &&
	    (tx->queue_stopped_bmp & (1 << ac)) &&
	    pend_pkt_q_len < (MAX_TX_QUEUE_LEN / 2)) {
		ieee80211_wake_queue(priv->hw, tx_queue_unmap(ac));
		tx->queue_stopped_bmp &= ~(1 << (ac));
	}

	pkt_info->peer_id = peer_info.id;
	RPU_DEBUG_TX("%s-UMACTX: token_id: %d ",
				priv->name,
				token_id);
	RPU_DEBUG_TX("total_pending_packets_process: %d\n",
		skb_queue_len(txq));

	return total_pending_processed;
}


int rpu_tx_alloc_token(struct img_priv *priv,
			       int ac,
			       int off_chanctx_idx,
			       int curr_chanctx_idx,
			       int peer_id,
			       struct sk_buff *skb)
{
	int token_id = NUM_TX_DESCS;
	struct tx_config *tx = &priv->tx;
	struct sk_buff_head *pend_pkt_q = NULL;
	unsigned int pkts_pend = 0;
	struct ieee80211_tx_info *tx_info;

	spin_lock_bh(&tx->lock);
	pend_pkt_q = &tx->pending_pkt[off_chanctx_idx][peer_id][ac];

	RPU_DEBUG_TX("%s-UMACTX:Alloc Req q = %d off_chan: %d out_tok:%d\n",
		      priv->name,
		      ac,
		      off_chanctx_idx, tx->outstanding_tokens[ac]);
	RPU_DEBUG_TX("peerid: %d,\n", peer_id);

	/* Queue the frame to the pending frames queue */
	skb_queue_tail(pend_pkt_q, skb);

	tx_info = IEEE80211_SKB_CB(skb);

	if (tx->outstanding_tokens[ac] >= NUM_TX_DESCS_PER_AC) {
		bool agg_status = false;

		agg_status = check_80211_aggregation(priv,
						     skb,
						     ac,
						     off_chanctx_idx,
						     peer_id);

		if (agg_status || !priv->params->enable_early_agg_checks) {
			int max_cmds = priv->params->max_tx_cmds;

			/* encourage aggregation to the max size
			 * supported (priv->params->max_tx_cmds)
			 */
			if (skb_queue_len(pend_pkt_q) < max_cmds) {
				RPU_DEBUG_TX("pend_q not full out_tok:%d\n",
					      tx->outstanding_tokens[ac]);
				goto out;
			 } else {
				RPU_DEBUG_TX("pend_q full out_tok:%d\n",
					      tx->outstanding_tokens[ac]);
			}
		}
	}

	/* Take steps to stop the TX traffic if we have reached
	 * the queueing limit.
	 * We dont this for the ROC queue to avoid the case where we are in the
	 * OFF channel but there is lot of traffic for the operating channel on
	 * the shared ROC queue (which is VO right now), since this would block
	 * ROC traffic too.
	 */
	if (skb_queue_len(pend_pkt_q) >= MAX_TX_QUEUE_LEN) {
		if ((!priv->roc_params.roc_in_progress) ||
		    (priv->roc_params.roc_in_progress &&
		     (ac != RPU_ROC_AC))) {
			ieee80211_stop_queue(priv->hw,
					     skb->queue_mapping);
			tx->queue_stopped_bmp |= (1 << ac);
		}
	}

	token_id = get_token(priv,
			     curr_chanctx_idx,
			     ac);

	RPU_DEBUG_TX("%s-UMACTX:Alloc buf Result *id= %d q = %d out_tok: %d",
					priv->name,
					token_id,
					ac, tx->outstanding_tokens[ac]);
	RPU_DEBUG_TX(", peerid: %d,\n", peer_id);

	if (token_id == NUM_TX_DESCS)
		goto out;

	pkts_pend = rpu_tx_proc_pend_frms(priv,
						  ac,
						  curr_chanctx_idx,
						  token_id);

	/* We have just added a frame to pending_q but channel context is
	 * mismatch.
	 */

	if (!pkts_pend) {
		free_token(priv, token_id, ac);
		token_id = NUM_TX_DESCS;
	}

out:
	spin_unlock_bh(&tx->lock);

	RPU_DEBUG_TX("%s-UMACTX:Alloc buf Result *id= %d out_tok:%d\n",
					priv->name,
					token_id, tx->outstanding_tokens[ac]);
	/* If token is available, just return tokenid, list will be sent*/
	return token_id;
}


int get_band_chanctx(struct img_priv *priv, struct umac_vif *uvif)
{
	struct ieee80211_chanctx_conf *chanctx = NULL;
	int index = 0;
	int band = 0;

	rcu_read_lock();
	index = uvif->chanctx->index;
	chanctx = rcu_dereference(priv->chanctx[index]);
	band = (chanctx->def.chan)->band;
	rcu_read_unlock();

	return band;
}

int rpu_tx_free_buff_req(struct img_priv *priv,
				 struct rpu_event_tx_done *tx_done,
				 unsigned char *ac,
				 int curr_chanctx_idx,
				 int *vif_index_bitmap)
{
	int i = 0;
	unsigned int pkts_pend = 0;
	struct tx_config *tx = &priv->tx;
	struct ieee80211_hdr *mac_hdr;
	struct ieee80211_tx_info *tx_info_bcn;
	struct ieee80211_tx_info tx_info_1st_mpdu;
	struct sk_buff *skb, *tmp, *skb_first = NULL;
	struct sk_buff_head *skb_list, tx_done_list;
	int vif_index = -1;
	unsigned int pkt = 0;
	int cnt = 0;
	unsigned int desc_id = tx_done->descriptor_id;
	struct umac_vif *uvif = NULL;
	struct ieee80211_vif *ivif = NULL;
	unsigned long bcn_int = 0;
	int chanctx_idx = 0;
	struct tx_pkt_info *pkt_info = NULL;
	int start_ac, end_ac;
	unsigned int q_len;

	skb_queue_head_init(&tx_done_list);

	spin_lock_bh(&tx->lock);

	chanctx_idx = tx->desc_chan_map[desc_id];
	if (chanctx_idx == -1) {
		spin_unlock_bh(&tx->lock);
		if (net_ratelimit())
			pr_err("%s: Unexpected channel context:tok:%d q:%d\n",
			       __func__,
			       desc_id,
			       tx_done->queue);
		goto out;
	}
	pkt_info = &priv->tx.pkt_info[chanctx_idx][desc_id];
	RPU_DEBUG_TX("%s-UMACTX:Free buf Req q = %d",
				priv->name,
				tx_done->queue);
	RPU_DEBUG_TX(", desc_id: %d chanctx: %d out_tok: %d\n",
				desc_id,
				chanctx_idx,
				priv->tx.outstanding_tokens[tx_done->queue]);


	/* Defer Tx Done Processsing */
	skb_list = &priv->tx.pkt_info[chanctx_idx][desc_id].pkt;

	if (skb_queue_len(skb_list)) {
		/* Cut the list to new one, tx_pkt will be re-initialized */
		skb_queue_splice_tail_init(skb_list, &tx_done_list);
	} else {
		RPU_DEBUG_TX("%s-UMACTX:Got Empty List: list_addr: %p\n",
						priv->name,
						skb_list);
	}

	/* Reserved token */
	if (desc_id < (NUM_TX_DESCS_PER_AC * NUM_ACS)) {
		start_ac = end_ac = tx_done->queue;
	} else {
		/* Spare token:
		 * Loop through all AC's
		 */
		start_ac = WLAN_AC_VO;
		end_ac = WLAN_AC_BK;
	}
	for (cnt = start_ac; cnt >= end_ac; cnt--) {
		pkts_pend = rpu_tx_proc_pend_frms(priv,
					      cnt,
					      curr_chanctx_idx,
					      desc_id);

		if (pkts_pend) {
			*ac = cnt;
			/* Spare Token Case*/
			if (tx_done->queue != *ac) {
				/*Adjust the counters*/
				tx->outstanding_tokens[tx_done->queue]--;
				tx->outstanding_tokens[*ac]++;
			}
			break;
		}
	}

	/* Unmap here before release lock to avoid race */
	if (skb_queue_len(&tx_done_list)) {
		skb_queue_walk_safe(&tx_done_list, skb, tmp) {
			hal_ops.unmap_tx_buf(tx_done->descriptor_id, pkt);
			RPU_DEBUG_TX("%s-UMACTX:TXDONE: ID=%d",
				priv->name,
				tx_done->descriptor_id);
			RPU_DEBUG_TX("Stat=%d (%d, %d)\n",
				tx_done->frm_status[pkt],
				tx_done->rate[pkt],
				tx_done->retries_num[pkt]);

			pkt++;
		}
	}

	if (!pkts_pend) {
		/* Mark the token as available */
		free_token(priv, desc_id, tx_done->queue);
		priv->tx.desc_chan_map[desc_id] = -1;
	}

	/* Unlock: Give a chance for Tx to add to pending lists */
	spin_unlock_bh(&tx->lock);

	/* Protection from mac80211 _ops especially stop */
	if (priv->state != STARTED)
		goto out;

	if (!skb_queue_len(&tx_done_list))
		goto out;

	q_len = skb_queue_len(&tx_done_list);
	skb_first = skb_peek(&tx_done_list);

	memcpy(&tx_info_1st_mpdu,
	       (struct ieee80211_tx_info *)IEEE80211_SKB_CB(skb_first),
	       sizeof(struct ieee80211_tx_info));

	pkt = 0;

	skb_queue_walk_safe(&tx_done_list, skb, tmp) {
		__skb_unlink(skb, &tx_done_list);

		if (!skb)
			continue;
		/* In the Tx path we move the .11hdr from skb to CMD_TX
		 * Hence pushing it here, not required for loopback case
		 */
		skb_push(skb,
			 priv->tx.pkt_info[chanctx_idx][desc_id].hdr_len);
		mac_hdr = (struct ieee80211_hdr *)(skb->data);

		if (!ieee80211_is_beacon(mac_hdr->frame_control)) {
			vif_index = vif_addr_to_index(mac_hdr->addr2,
						      priv);
			if (vif_index > -1)
				*vif_index_bitmap |= (1 << vif_index);

			/* Same Rate info for all packets */
			tx_status(skb,
				  tx_done,
				  pkt,
				  priv,
				  q_len,
				  pkt_info);
		} else {
			struct ieee80211_bss_conf *bss_conf;
			bool bcn_status;

			if (tx_done->frm_status[pkt] ==
			    TX_DONE_STAT_DISCARD_BCN) {
				/* We did not send beacon */
				priv->tx_last_beacon = 0;
			} else if (tx_done->frm_status[pkt] ==
				   TX_DONE_STAT_SUCCESS) {
				/* We did send beacon */
				priv->tx_last_beacon = 1;
			}

			tx_info_bcn = IEEE80211_SKB_CB(skb);
			ivif = tx_info_bcn->control.vif;
			uvif = (struct umac_vif *)(ivif->drv_priv);

			bss_conf = &uvif->vif->bss_conf;
			bcn_status = bss_conf->enable_beacon;
			bcn_int = bss_conf->beacon_int - 10;
			bcn_int = msecs_to_jiffies(bcn_int);

			for (i = 0; i < MAX_VIFS; i++) {
				if (priv->active_vifs & (1 << i)) {
					if ((priv->vifs[i] == ivif) &&
					    (bcn_status == true)) {
						mod_timer(&uvif->bcn_timer,
							  jiffies +
							  bcn_int);
					}
				}
			}

			/* Beacon Time Stamp */
			if (tx_done->frm_status[pkt] == TX_DONE_STAT_SUCCESS) {
				unsigned int ts2;
				unsigned int ldelta = 0;
				int ets_band;
				int bts_vif = uvif->vif_index;
				struct wifi_sync *sync =
						&priv->params->sync[bts_vif];
				ets_band = get_band_chanctx(priv, uvif);
				spin_lock(&tsf_lock);
				sync->status = 1;
				img_ether_addr_copy(sync->bssid,
					ivif->bss_conf.bssid);
				memcpy(sync->ts1, tx_done->reserved, 8);
				memcpy(&sync->ts2, (tx_done->reserved + 8), 4);
				ts2 = sync->ts2;
				sync->atu = 0;

				if (ets_band == IEEE80211_BAND_2GHZ)
					ldelta = BTS_AP_24GHZ_ETS;
				else if (ets_band == IEEE80211_BAND_5GHZ)
					ldelta = BTS_AP_5GHZ_ETS;

				if (frc_to_atu) {
					frc_to_atu(ts2, &sync->atu, 0);
					sync->atu += ldelta * 1000;
				}
				spin_unlock(&tsf_lock);
			}
			dev_kfree_skb_any(skb);
		}

		pkt++;
	}
out:
	return pkts_pend;
}


void rpu_proc_ch_sw_event(struct rpu_event_ch_switch *ch_sw_info,
				  void *context)
{
	struct img_priv *priv = NULL;
	int chan = 0;
	int curr_freq = 0;
	int chan_id = 0;
	struct ieee80211_chanctx_conf *curr_chanctx = NULL;
	int i = 0;

	if (!ch_sw_info || !context) {
		pr_err("%s: Invalid Parameters:\n", __func__);
		return;
	}

	priv = (struct img_priv *)context;
	chan = ch_sw_info->chan;

	rcu_read_lock();

	for (i = 0; i < MAX_CHANCTX; i++) {
		curr_chanctx = rcu_dereference(priv->chanctx[i]);

		if (curr_chanctx) {
			curr_freq = curr_chanctx->def.chan->center_freq;

			if (ieee80211_frequency_to_channel(curr_freq) == chan) {
				chan_id = i;
				break;
			}
		}
	}

	rcu_read_unlock();

	if (i == MAX_CHANCTX) {
		pr_err("%s: Invalid Channel Context: chan: %d\n",
		       __func__,
		       chan);
		return;
	}

	/* Switch to the new channel context */
	spin_lock(&priv->chanctx_lock);
	priv->curr_chanctx_idx = chan_id;
	spin_unlock(&priv->chanctx_lock);

	/* We now try to xmit any frames whose xmission got cancelled due to a
	 * previous channel switch
	 */
	rpu_tx_proc_send_pend_frms_all(priv, chan_id);
}


unsigned int rpu_proc_tx_dscrd_chsw(struct img_priv *priv,
					    int curr_chanctx_idx,
					    struct rpu_event_tx_done *tx_done)
{
	struct tx_config *tx = &priv->tx;
	struct sk_buff_head *txq = NULL, tx_done_list;
	int chanctx_idx = -1;
	int pkt = 0;
	int txq_len = 0;
	struct sk_buff *skb = NULL;
	struct sk_buff *skb_first = NULL;
	struct sk_buff *tmp = NULL;
	int queue = 0;
	int ret = 0, cnt = 0;
	unsigned int desc_id = 0;
	unsigned int *curr_retries = NULL;
	unsigned int max_retries = 0;
	struct ieee80211_tx_info tx_info_1st_mpdu;
	struct ieee80211_hdr *mac_hdr = NULL;
	bool retries_exceeded = false;
	unsigned int *rate = NULL;
	unsigned int *retries = NULL;
	int start_ac, end_ac;
	unsigned int pkts_pend = 0;

	skb_queue_head_init(&tx_done_list);

	spin_lock_bh(&tx->lock);

	desc_id = tx_done->descriptor_id;

	/* We keep the frames which were not consumed by the FW in the
	 * tx_pkt queue. These frames will then be requeued to the FW when this
	 * channel context is scheduled again
	 */
	chanctx_idx = tx->desc_chan_map[desc_id];

	if ((chanctx_idx == -1) ||
	    (chanctx_idx > (MAX_CHANCTX + MAX_OFF_CHANCTX))) {
		pr_err("%s: Unexpected channel context: %d\n",
		       __func__,
		       chanctx_idx);
		goto out;
	}

	txq = &tx->pkt_info[chanctx_idx][desc_id].pkt;
	txq_len = skb_queue_len(txq);

	if (!txq_len) {
		pr_err("%s: TX_DONE received for empty queue: chan: %d desc_id: %d\n",
		       __func__,
		       chanctx_idx,
		       desc_id);
		goto out;
	}

	RPU_DEBUG_TX("%s: %d retries: %d rate: %d\n",
			__func__,
			__LINE__,
			tx_done->retries_num[0],
			tx_done->rate[0]);
	pkt = 0;

	skb_first = skb_peek(txq);

	if (!skb_first) {
		pr_err("%s: Empty txq: chan: %d desc_id: %d\n",
		       __func__,
		       chanctx_idx,
		       desc_id);
		goto out;
	}

	curr_retries = &tx->pkt_info[chanctx_idx][desc_id].curr_retries;
	max_retries = tx->pkt_info[chanctx_idx][desc_id].max_retries;
	retries = tx->pkt_info[chanctx_idx][desc_id].retries;
	rate = tx->pkt_info[chanctx_idx][desc_id].rate;
	tx->pkt_info[chanctx_idx][desc_id].adjusted_rates = true;

	if ((tx_done->retries_num[0] + *curr_retries) > max_retries)
		retries_exceeded = true;
	else
		*curr_retries += tx_done->retries_num[0];

	memcpy(&tx_info_1st_mpdu,
	       (struct ieee80211_tx_info *)IEEE80211_SKB_CB(skb_first),
	       sizeof(struct ieee80211_tx_info));

	skb_queue_walk_safe(txq, skb, tmp) {
		if (!skb)
			continue;

		hal_ops.unmap_tx_buf(desc_id, pkt);

		/* In the Tx path we move the .11hdr from skb to CMD_TX
		 * Hence pushing it here
		 */
		skb_push(skb,
			 tx->pkt_info[chanctx_idx][desc_id].hdr_len);

		mac_hdr = (struct ieee80211_hdr *)skb->data;

		if (retries_exceeded) {
			__skb_unlink(skb, txq);

			if (!skb)
				continue;
			skb_queue_tail(&tx_done_list, skb);

			RPU_DEBUG_TX("%s: %d ", __func__, __LINE__);
			RPU_DEBUG_TX("Freeing the skb MAX retries reached.\n");
		} else {
			RPU_DEBUG_TX("%s: %d ", __func__, __LINE__);
			RPU_DEBUG_TX("Re-programming the skb when ");
			RPU_DEBUG_TX("CTX is right with retry bit set.\n");
		mac_hdr->frame_control |= cpu_to_le16(IEEE80211_FCTL_RETRY);
		}
		pkt++;
	}

	/* First check if there is a packet in the txq of the current
	 * chanctx that needs to be transmitted
	 */
	txq = &tx->pkt_info[curr_chanctx_idx][desc_id].pkt;
	txq_len = skb_queue_len(txq);
	queue = tx->pkt_info[curr_chanctx_idx][desc_id].queue;
	pkts_pend = txq_len;

	if (txq_len) {
		spin_unlock_bh(&tx->lock);

		/* TODO: Currently sending 0 since this param is not
		 * used as expected in the orig code for multiple
		 * frames etc Need to set this properly when the orig
		 * code logic is corrected
		 */
		ret = __rpu_tx_frame(priv,
					     queue,
					     desc_id,
					     curr_chanctx_idx,
					     0,
					     1);
		if (ret < 0) {
			/* TODO: Check if we need to clear the TX bitmap
			 * and desc_chan_map here
			 */
			pr_err("%s: Queueing of TX frame to FW failed\n",
			       __func__);
		}

		/* This is needed to avoid freeing up the token
		 */
		pkts_pend = 1;

		goto tx_done;
	} else {
		/* Check pending queue */
		/* Reserved token */
		if (desc_id < (NUM_TX_DESCS_PER_AC * NUM_ACS)) {
			queue = (desc_id % NUM_ACS);
			start_ac = end_ac = queue;
		} else {
			/* Spare token:
			 * Loop through all AC's
			 */
			start_ac = WLAN_AC_VO;
			end_ac = WLAN_AC_BK;
		}

		for (cnt = start_ac; cnt >= end_ac; cnt--) {
			pkts_pend = rpu_tx_proc_pend_frms(priv,
						      cnt,
						      curr_chanctx_idx,
						      desc_id);
			if (pkts_pend) {
				queue = cnt;
				if (tx_done->queue != queue) {
					unsigned int txd_q = tx_done->queue;
					/*Adjust the counters*/
					tx->outstanding_tokens[txd_q]--;
					tx->outstanding_tokens[queue]++;
				}
				break;
			}
		}

		spin_unlock_bh(&tx->lock);

		if (pkts_pend > 0) {
			/* TODO: Currently sending 0 since this param is not
			 * used as expected in the orig code for multiple
			 * frames etc. Need to set this properly when the orig
			 * code logic is corrected
			 */
			ret = __rpu_tx_frame(priv,
						     queue,
						     desc_id,
						     curr_chanctx_idx,
						     0,
						     0);

			if (ret < 0) {
				pr_err("%s: Queueing of TX frame to FW failed\n",
				       __func__);
			}
		}
		goto tx_done;
	}

tx_done:
	skb_queue_walk_safe(&tx_done_list, skb, tmp) {
		unsigned int q_len = skb_queue_len(&tx_done_list);

			tx_status(skb,
				  tx_done,
				  pkt,
				  priv,
				  q_len,
				  &tx->pkt_info[chanctx_idx][desc_id]);
	}

	spin_lock_bh(&tx->lock);

	if (!pkts_pend) {
		/* Mark the token as available */
		free_token(priv, desc_id, tx_done->queue);
		priv->tx.desc_chan_map[desc_id] = -1;
	}
out:
	spin_unlock_bh(&tx->lock);

	return pkts_pend;
}




void rpu_tx_init(struct img_priv *priv)
{
	int i = 0;
	int j = 0;
	int k = 0;
	struct tx_config *tx = &priv->tx;

	memset(&tx->buf_pool_bmp,
	       0,
	       sizeof(long) * ((NUM_TX_DESCS/TX_DESC_BUCKET_BOUND) + 1));

	tx->queue_stopped_bmp = 0;
	tx->next_spare_token_ac = WLAN_AC_BE;

	for (i = 0; i < NUM_ACS; i++) {
		for (j = 0; j < MAX_PEND_Q_PER_AC; j++) {
			for (k = 0; k < MAX_RPU_VIF_CHANCTX_TYPES; k++)
				skb_queue_head_init(&tx->pending_pkt[k][j][i]);
		}

		tx->outstanding_tokens[i] = 0;
	}

	for (i = 0; i < NUM_TX_DESCS; i++) {
		tx->desc_chan_map[i] = -1;

		for (j = 0; j < MAX_CHANCTX + MAX_OFF_CHANCTX ; j++)
			skb_queue_head_init(&tx->pkt_info[j][i].pkt);
	}

	for (j = 0; j < NUM_ACS; j++)
		for (i = 0; i < MAX_CHANCTX; i++)
			tx->curr_peer_opp[i][j] = 0;

	priv->curr_chanctx_idx = -1;
	spin_lock_init(&tx->lock);
	ieee80211_wake_queues(priv->hw);

	RPU_DEBUG_TX("%s-UMACTX: initialization successful\n",
			TX_TO_MACDEV(tx)->name);
}


void rpu_tx_deinit(struct img_priv *priv)
{
	int i = 0;
	int j = 0;
	int k = 0;
	struct tx_config *tx = &priv->tx;
	struct sk_buff *skb = NULL;
	unsigned int qlen = 0;
	struct sk_buff_head *pend_q = NULL;

	ieee80211_stop_queues(priv->hw);

	wait_for_tx_complete(tx);

	spin_lock_bh(&tx->lock);

	for (i = 0; i < NUM_TX_DESCS; i++) {
		for (j = 0; j < MAX_CHANCTX + MAX_OFF_CHANCTX; j++) {
			qlen = skb_queue_len(&tx->pkt_info[j][i].pkt);

			if (qlen) {
				while ((skb =
					skb_dequeue(&tx->pkt_info[j][i].pkt)) !=
				       NULL) {
					dev_kfree_skb_any(skb);
				}
			}
		}
	}

	for (i = 0; i < NUM_ACS; i++) {
		for (j = 0; j < MAX_PEND_Q_PER_AC; j++) {
			for (k = 0; k < MAX_RPU_VIF_CHANCTX_TYPES; k++)
				pend_q = &tx->pending_pkt[k][j][i];

			while ((skb = skb_dequeue(pend_q)) != NULL)
				dev_kfree_skb_any(skb);
		}
	}

	spin_unlock_bh(&tx->lock);

	RPU_DEBUG_TX("%s-UMACTX: deinitialization successful\n",
			TX_TO_MACDEV(tx)->name);
}


int __rpu_tx_frame(struct img_priv *priv,
			   unsigned int queue,
			   unsigned int token_id,
			   int curr_chanctx_idx,
			   unsigned int more_frames,
			   bool retry)
{
	struct rpu_event_tx_done tx_done;
	struct sk_buff_head *txq = NULL;
	int ret = 0;
	int pkt = 0;

	ret = rpu_prog_tx(queue,
				  more_frames,
				  curr_chanctx_idx,
				  token_id,
				  retry);

	if (ret < 0) {
		pr_err("%s-UMACTX: Unable to send frame, dropping ..%d\n",
		       priv->name, ret);

		tx_done.descriptor_id = token_id;
		tx_done.queue = queue;
		priv->tx.desc_chan_map[token_id] = curr_chanctx_idx;

		txq = &priv->tx.pkt_info[curr_chanctx_idx][token_id].pkt;

		for (pkt = 0; pkt < skb_queue_len(txq); pkt++) {
			tx_done.frm_status[pkt] = TX_DONE_STAT_ERR_RETRY_LIM;
			tx_done.rate[pkt] = 0;
		}

		rpu_tx_complete(&tx_done,
					curr_chanctx_idx,
					priv);
	}

	return ret;
}


int rpu_tx_frame(struct sk_buff *skb,
			 struct ieee80211_sta *sta,
			 struct img_priv *priv,
			 int curr_chanctx_idx,
			 bool bcast)
{
	unsigned int queue = 0;
	unsigned int token_id = 0;
	unsigned int more_frames = 0;
	int ret = 0;
	struct ieee80211_tx_info *tx_info = IEEE80211_SKB_CB(skb);
	struct ieee80211_hdr *mac_hdr = NULL;
	struct umac_vif *uvif = NULL;
	struct umac_sta *usta = NULL;
	int peer_id = -1;
	int off_chanctx_idx;

	uvif = (struct umac_vif *)(tx_info->control.vif->drv_priv);
	mac_hdr = (struct ieee80211_hdr *)(skb->data);

	if (sta) {
		usta = (struct umac_sta *)sta->drv_priv;
		peer_id = usta->index;
	} else {
		peer_id = MAX_PEERS + uvif->vif_index;
	}

	if (bcast == false) {
		queue = tx_queue_map(skb->queue_mapping);
		more_frames = 0;
	} else {
		queue = WLAN_AC_BCN;
		/* Hack: skb->priority is used to indicate more frames */
		more_frames = skb->priority;
	}

	if (!ieee80211_is_beacon(mac_hdr->frame_control))
		priv->stats->tx_cmds_from_stack++;

	if (priv->params->production_test == 1)
		tx_info->flags |= IEEE80211_TX_CTL_AMPDU;

	if ((tx_info->flags & IEEE80211_TX_CTL_TX_OFFCHAN) ||
	    (uvif->chanctx &&
	    uvif->chanctx->index == priv->roc_off_chanctx_idx))  {
		atomic_inc(&priv->roc_params.roc_mgmt_tx_count);
		off_chanctx_idx = RPU_VIF_CHANCTX_TYPE_OFF;
		RPU_DEBUG_ROC("%s:%d Sending OFFCHAN Frame: %d\n",
			__func__, __LINE__,
			atomic_read(&priv->roc_params.roc_mgmt_tx_count));
	} else {
		off_chanctx_idx = RPU_VIF_CHANCTX_TYPE_OPER;
	}


	RPU_DEBUG_TX("%s-UMACTX:%s:%d ",
			priv->name,
			 __func__,
			 __LINE__);
	RPU_DEBUG_TX("Wait Alloc:queue: %d qmap: %d is_bcn: %d bcast:%d\n",
			queue,
			skb->queue_mapping,
			ieee80211_is_beacon(mac_hdr->frame_control),
			is_multicast_ether_addr(mac_hdr->addr1) ? true : false);

	token_id = rpu_tx_alloc_token(priv,
						 queue,
						 off_chanctx_idx,
						 curr_chanctx_idx,
						 peer_id,
						 skb);

	/* The frame was unable to find a reserved token */
	if (token_id == NUM_TX_DESCS) {
		RPU_DEBUG_TX("%s-UMACTX:%s:%d Token Busy Queued:\n",
			priv->name, __func__, __LINE__);
		return NETDEV_TX_OK;
	}

	ret = __rpu_tx_frame(priv,
				     queue,
				     token_id,
				     curr_chanctx_idx,
				     more_frames,
				     0);


	return NETDEV_TX_OK;
}




void rpu_tx_complete(struct rpu_event_tx_done *tx_done,
			     int curr_chanctx_idx,
			     void *context)
{
	struct img_priv *priv = (struct img_priv *)context;
	unsigned int more_frames = 0;
	int vif_index = 0, vif_index_bitmap = 0, ret = 0;
	unsigned int pkts_pending = 0;
	unsigned char queue = 0;
	struct rpu_event_noa noa_event;
	int token_id = 0;
	int qlen = 0;
	struct tx_pkt_info *pkt_info = NULL;

	token_id = tx_done->descriptor_id;

	if (token_id < 0 || token_id > 11) {
		RPU_DEBUG_TX("%s:%d Invalid token_id: %d\n",
			     __func__,
			     __LINE__,
			     token_id);
		RPU_DEBUG_DUMP_TX(DUMP_PREFIX_NONE,
				  16,
				  1,
				  tx_done,
				  sizeof(struct rpu_event_tx_done),
				  1);
		return;
	}

	pkt_info = &priv->tx.pkt_info[curr_chanctx_idx][token_id];
	qlen = skb_queue_len(&pkt_info->pkt);

	RPU_DEBUG_TX("%s-UMACTX:TX Done Rx for desc_id: %d",
			  priv->name,
			  tx_done->descriptor_id);
	RPU_DEBUG_TX("Q: %d qlen: %d status: %d chactx: %d out_tok: %d\n",
			  tx_done->queue,
			  qlen,
			  tx_done->frm_status[0],
			  curr_chanctx_idx,
			  priv->tx.outstanding_tokens[tx_done->queue]);

	update_aux_adc_voltage(priv, tx_done->pdout_voltage);

	if (tx_done->frm_status[0] == TX_DONE_STAT_DISCARD_CHSW) {
		pkts_pending = rpu_proc_tx_dscrd_chsw(priv,
							      curr_chanctx_idx,
							      tx_done);
		goto out;
	}
	pkts_pending = rpu_tx_free_buff_req(priv,
						    tx_done,
						    &queue,
						    curr_chanctx_idx,
						    &vif_index_bitmap);

	if (pkts_pending) {
		/*TODO..Do we need to check each skb for more_frames?*/
		more_frames = 0;

		RPU_DEBUG_TX("%s-UMACTX:%s:%d Transfer Pending Frames:\n",
			       priv->name,
			       __func__,
			       __LINE__);

		ret = __rpu_tx_frame(priv,
					     queue,
					     token_id,
					     curr_chanctx_idx,
					     more_frames,
					     0);

	} else {
		RPU_DEBUG_TX("%s-UMACTX:No Pending Packets\n", priv->name);
	}

out:

	for (vif_index = 0; vif_index < MAX_VIFS; vif_index++) {
		if (vif_index_bitmap & (1 << vif_index)) {
			memset(&noa_event, 0, sizeof(noa_event));
			noa_event.if_index = vif_index;
			rpu_noa_event(FROM_TX_DONE,
					      &noa_event,
					      (void *)priv,
					      NULL);
		}
	}
}

void rpu_purge_tx_queue(struct img_priv *priv,
			   struct sk_buff_head *skbs)
{


	struct sk_buff *loop_skb = NULL, *tmp = NULL;

	skb_queue_walk_safe(skbs,
			    loop_skb,
			    tmp) {
		struct ieee80211_hdr *hdr = NULL;

		__skb_unlink(loop_skb,
			     skbs);

		if (!loop_skb)
			continue;

		hdr = (struct ieee80211_hdr *)loop_skb->data;

		if (!ieee80211_is_beacon(hdr->frame_control))
			priv->stats->tx_dones_to_stack++;

		ieee80211_free_txskb(priv->hw,
				     loop_skb);
	}

}

static int rpu_flush_vif_all_pend_q(struct img_priv *priv,
					struct umac_vif *uvif,
					unsigned int hw_queue_map,
					enum RPU_VIF_CHANCTX_TYPE chanctx_type)
{
	unsigned int pending = 0;
	int count = 0;
	unsigned int queue = 0;
	int pend_q = 0;
	struct sk_buff_head *pend_pkt_q = NULL;
	struct tx_config *tx = NULL;
	struct ieee80211_sta *sta = NULL;
	struct umac_sta *usta = NULL;
	bool warned = false;

	tx = &priv->tx;

	if (!uvif->chanctx) {
		RPU_DEBUG_TX("%s: Chanctx NULL, returning\n",
						__func__);
		return -1;
	}

	for (queue = 0; queue < NUM_ACS; queue++) {
		if (!(BIT(queue) & hw_queue_map))
			continue;

		for (pend_q = 0; pend_q < MAX_PEND_Q_PER_AC; pend_q++) {
			rcu_read_lock();
			if (pend_q < MAX_PEERS) {
				sta = rcu_dereference(priv->peers[pend_q]);

				if (!sta) {
					rcu_read_unlock();
					continue;
				}

				usta = (struct umac_sta *)(sta->drv_priv);

				if (usta->vif_index != uvif->vif_index) {
					rcu_read_unlock();
					continue;
				}
				rcu_read_unlock();
			} else if (pend_q >= MAX_PEERS) {
				int vif_index;
				struct umac_vif *tvif;
				struct ieee80211_vif *vif = NULL;

				vif_index = pend_q - MAX_PEERS;
				vif = rcu_dereference(priv->vifs[vif_index]);

				if (!vif) {
					rcu_read_unlock();
					continue;
				}

				tvif = (struct umac_vif *)(vif->drv_priv);

				if (uvif->vif_index != tvif->vif_index) {
					rcu_read_unlock();
					continue;
				}
				rcu_read_unlock();
			} else {
				rcu_read_unlock();
				continue;
			}

			while (1) {
				spin_lock_bh(&tx->lock);

				pend_pkt_q =
					&tx->pending_pkt[chanctx_type]
							[pend_q]
							[queue];

				/* Assuming all packets for the peer have same
				 * channel context
				 */
				pending = skb_queue_len(pend_pkt_q);

				spin_unlock_bh(&tx->lock);

				if (!pending)
					break;

				if (!warned &&
				    count >= QUEUE_FLUSH_TIMEOUT_TICKS) {
					pr_err("%s: Timeout: VIF: %d Queue: %d pending: %d\n",
					       priv->name,
					       uvif->vif_index,
					       queue,
					       pending);
					WARN_ON(1);
					warned = true;
					return -1;
				}

				current->state = TASK_INTERRUPTIBLE;

				if (schedule_timeout(1) == 0)
					count++;

			}

			if (pending) {
				pr_err("%s-UMACTX: Failed for VIF: %d ",
					       priv->name,
					       uvif->vif_index);
				pr_err(" and Queue: %d, pending: %d\n",
					       queue,
					       pending);

				return -1;
			}
		}
	}

	RPU_DEBUG_TX("%s: Success for VIF: %d and Queue: %d\n",
					__func__,
					uvif->vif_index,
					queue);
	return 0;
}

int rpu_discard_sta_pend_q(struct img_priv *priv,
				   struct umac_vif *uvif,
				   int peer_id,
				   unsigned int hw_queue_map)
{
	unsigned int pending = 0;
	unsigned int queue = 0;
	int pend_q = 0;
	struct sk_buff_head *pend_pkt_q = NULL, tx_discard_list;
	struct tx_config *tx = &priv->tx;

	RPU_DEBUG_TX("%s:%d Enter..:tx:%d txd:%d hqmap:%d vif:%d\n",
		      __func__,
		      __LINE__,
		      priv->stats->tx_cmds_from_stack,
		      priv->stats->tx_dones_to_stack,
		      hw_queue_map,
		      uvif->vif_index);
	skb_queue_head_init(&tx_discard_list);

	for (queue = 0; queue < NUM_ACS; queue++) {
		int ac = 0;

		if (!(BIT(queue) & hw_queue_map))
			continue;

		RPU_DEBUG_TX("%s:%d MATCH queue: %d pend_q:%d\n",
			      __func__,
			      __LINE__,
			      queue,
			      pend_q);

		pend_pkt_q =
			&tx->pending_pkt[0]
					[peer_id]
					[queue];

		pending = skb_queue_len(pend_pkt_q);

		if (!pending)
			continue;

		RPU_DEBUG_TX("%s:%d Free the skbs..\n",
			      __func__,
			      __LINE__);

		skb_queue_splice_tail_init(pend_pkt_q,
					   &tx_discard_list);
		rpu_purge_tx_queue(priv, &tx_discard_list);

		ac = queue;

		if (tx->queue_stopped_bmp & (1 << ac)) {
			RPU_DEBUG_TX("%s:%d wakeup the queue: %d ac: %d\n",
				     __func__, __LINE__, queue, ac);
			ieee80211_wake_queue(priv->hw, tx_queue_unmap(ac));
		}

	}
	RPU_DEBUG_TX("%s:%d Exit..:tx:%d txd:%d\n", __func__, __LINE__,
		      priv->stats->tx_cmds_from_stack,
		      priv->stats->tx_dones_to_stack);
	return 0;
}

int rpu_discard_sta_tx_q(struct img_priv *priv,
				   struct umac_vif *uvif,
				   int peer_id,
				   unsigned int hw_queue_map,
				   int chanctx_idx)
{
	struct tx_config *tx = &priv->tx;
	struct tx_pkt_info *pkt_info = NULL;
	struct sk_buff_head *txq = NULL;
	struct sk_buff *loop_skb = NULL, *tmp = NULL;
	int i = 0, pkt;

	RPU_DEBUG_TX("%s:%d Enter..:tx:%d txd:%d hqmap:%d vif:%d\n",
		      __func__,
		      __LINE__,
		      priv->stats->tx_cmds_from_stack,
		      priv->stats->tx_dones_to_stack,
		      hw_queue_map,
		      uvif->vif_index);

	for (i = 0; i < NUM_TX_DESCS; i++) {
		pkt_info = &tx->pkt_info[chanctx_idx][i];
		RPU_DEBUG_TX("%s:%d pvif: %d uvif:%d q:%d hq:%d peer_id:%d\n",
			      __func__,
			      __LINE__,
			      pkt_info->vif_index,
			      uvif->vif_index,
			      pkt_info->queue,
			      hw_queue_map,
			      peer_id);

		if ((pkt_info->vif_index == uvif->vif_index) &&
		    ((peer_id == -1) || (pkt_info->peer_id == peer_id)) &&
		    (BIT(pkt_info->queue) & hw_queue_map)) {
			int pool_id = i/TX_DESC_BUCKET_BOUND;

			txq = &pkt_info->pkt;

			if (!skb_queue_len(txq) ||
			    !test_bit(i, &tx->buf_pool_bmp[pool_id]))
				continue;

			RPU_DEBUG_TX("%s: Free the skbs:%d\n", __func__, i);


			/* In the Tx path we move the .11hdr from skb to CMD_TX
			 * Hence pushing it here, not required for loopback case
			 */
			pkt = 0;
			skb_queue_walk_safe(txq,
					    loop_skb,
					    tmp) {
				skb_push(loop_skb, pkt_info->hdr_len);
				hal_ops.unmap_tx_buf(i, pkt);
				pkt++;
			}
			rpu_purge_tx_queue(priv, txq);
			free_token(priv, i, pkt_info->queue);
			priv->tx.desc_chan_map[i] = -1;
		}
	}
	RPU_DEBUG_TX("%s:%d Exit..:tx:%d txd:%d\n", __func__, __LINE__,
		      priv->stats->tx_cmds_from_stack,
		      priv->stats->tx_dones_to_stack);
	return 0;
}
static int rpu_discard_vif_tx_queues(struct img_priv *priv,
				       struct umac_vif *uvif,
				       int chanctx_idx,
				       unsigned int hw_queue_map)
{
	struct tx_config *tx = NULL;

	tx = &priv->tx;
	RPU_DEBUG_TX("%s:%d Enter..:tx:%d txd:%d\n", __func__, __LINE__,
		      priv->stats->tx_cmds_from_stack,
		      priv->stats->tx_dones_to_stack);
	spin_lock_bh(&tx->lock);
	rpu_discard_sta_tx_q(priv, uvif, -1, hw_queue_map, chanctx_idx);
	spin_unlock_bh(&tx->lock);


	RPU_DEBUG_TX("%s: Success for VIF: %d",
					__func__,
					uvif->vif_index);
	RPU_DEBUG_TX("%s:%d Enter..:tx:%d txd:%d\n", __func__, __LINE__,
		      priv->stats->tx_cmds_from_stack,
		      priv->stats->tx_dones_to_stack);
	return 0;

}
static int rpu_flush_vif_tx_queues(struct img_priv *priv,
				       struct umac_vif *uvif,
				       int chanctx_idx,
				       unsigned int hw_queue_map)
{
	unsigned int tokens = 0;
	unsigned int i = 0;
	unsigned long buf_pool_bmp = 0;
	struct tx_pkt_info *pkt_info = NULL;
	struct tx_config *tx = NULL;
	int count = 0;
	bool warned = false;

	tx = &priv->tx;

	spin_lock_bh(&tx->lock);

	for (i = 0; i < NUM_TX_DESCS; i++) {
		pkt_info = &tx->pkt_info[chanctx_idx][i];

		if ((pkt_info->vif_index == uvif->vif_index) &&
		    (BIT(pkt_info->queue) & hw_queue_map))
			tokens |= BIT(i);
	}

	spin_unlock_bh(&tx->lock);

	if (!tokens)
		return 0;

	while (1) {
		spin_lock_bh(&tx->lock);
		buf_pool_bmp = tx->buf_pool_bmp[0];
		spin_unlock_bh(&tx->lock);

		if (!(buf_pool_bmp & tokens))
			break;

		if (!warned &&
		    count >= QUEUE_FLUSH_TIMEOUT_TICKS) {
			pr_err("%s-UMACTX: TXQ: Failed for VIF: %d, buf_pool_bmp : 0x%lx:\n",
			       priv->name,
			       uvif->vif_index,
			       buf_pool_bmp);
			WARN_ON(1);
			warned = true;
			return -1;
		}

		current->state = TASK_INTERRUPTIBLE;

		if (schedule_timeout(1) == 0)
			count++;
	}

	RPU_DEBUG_TX("%s: Success for VIF: %d, buf_pool_bmp : 0x%lx\n",
					__func__,
					uvif->vif_index,
					buf_pool_bmp);
	return 0;
}
int rpu_discard_vif_all_pend_q(struct img_priv *priv,
				   struct umac_vif *uvif,
				   unsigned int hw_queue_map)
{
	struct tx_config *tx = &priv->tx;
	struct ieee80211_sta *sta = NULL;
	struct umac_sta *usta = NULL;
	unsigned int pend_q;

	RPU_DEBUG_TX("%s:%d Enter..:tx:%d txd:%d hqmap:%d vif:%d\n",
		      __func__,
		      __LINE__,
		      priv->stats->tx_cmds_from_stack,
		      priv->stats->tx_dones_to_stack,
		      hw_queue_map,
		      uvif->vif_index);
	spin_lock_bh(&tx->lock);

	for (pend_q = 0; pend_q < MAX_PEND_Q_PER_AC; pend_q++) {
		rcu_read_lock();
		if (pend_q < MAX_PEERS) {
			sta = rcu_dereference(priv->peers[pend_q]);

			if (!sta) {
				rcu_read_unlock();
				continue;
			}

			usta = (struct umac_sta *)(sta->drv_priv);

			if (usta->vif_index != uvif->vif_index) {
				rcu_read_unlock();
				continue;
			}
		} else if (pend_q >= MAX_PEERS) {
			int vif_index;
			struct umac_vif *tvif;
			struct ieee80211_vif *vif = NULL;

			vif_index = pend_q - MAX_PEERS;
			vif = rcu_dereference(priv->vifs[vif_index]);

			if (!vif) {
				rcu_read_unlock();
				continue;
			}

			tvif = (struct umac_vif *)(vif->drv_priv);

			if (uvif->vif_index != tvif->vif_index) {
				rcu_read_unlock();
				continue;
			}
		} else {
			rcu_read_unlock();
			continue;
		}
		rcu_read_unlock();
		rpu_discard_sta_pend_q(priv, uvif, pend_q, hw_queue_map);
	}

	spin_unlock_bh(&tx->lock);
	RPU_DEBUG_TX("%s:%d Exit..:tx:%d txd:%d\n", __func__, __LINE__,
		      priv->stats->tx_cmds_from_stack,
		      priv->stats->tx_dones_to_stack);
	return 0;
}

int rpu_flush_vif_queues(struct img_priv *priv,
			     struct umac_vif *uvif,
			     int chanctx_idx,
			     unsigned int hw_queue_map,
			     enum RPU_VIF_CHANCTX_TYPE vif_chanctx_type,
			     bool drop)
{
	int result  = -1;
	char peer_addr[ETH_ALEN] = {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};

	RPU_DEBUG_TX("%s:%d Enter..:tx:%d txd:%d\n", __func__, __LINE__,
		      priv->stats->tx_cmds_from_stack,
		      priv->stats->tx_dones_to_stack);

	if (drop) {
		/*Discard: clear pend_q, send tx_deinit to LMAC*/
		RPU_DEBUG_TX("%s: DISCARD\n", __func__);
		result = rpu_discard_vif_all_pend_q(priv,
						      uvif,
						      hw_queue_map);
		if (result)
			return result;

		priv->tx_deinit_complete = 0;
		rpu_prog_tx_deinit(uvif->vif_index, peer_addr);
		if (wait_for_tx_deinit_complete(priv) < 0) {
			/*Stuck: Reload FW??*/
			WARN_ON(1);
			result = rpu_discard_vif_tx_queues(priv,
							       uvif,
							       chanctx_idx,
							       hw_queue_map);
		}
	} else {
		RPU_DEBUG_TX("%s: FLUSH\n", __func__);
		/*Flush: Wait for frames*/
		result = rpu_flush_vif_all_pend_q(priv,
						      uvif,
						      hw_queue_map,
						      vif_chanctx_type);

		if (result == 0) {
			result = rpu_flush_vif_tx_queues(priv,
							     uvif,
							     chanctx_idx,
							     hw_queue_map);
		}

		if (result) {
			result = rpu_discard_vif_all_pend_q(priv,
								uvif,
								hw_queue_map);

			result = rpu_discard_vif_tx_queues(priv,
							     uvif,
							     chanctx_idx,
							     hw_queue_map);
		}
	}
	RPU_DEBUG_TX("%s:%d Enter..:tx:%d txd:%d\n", __func__, __LINE__,
		      priv->stats->tx_cmds_from_stack,
		      priv->stats->tx_dones_to_stack);
	return result;
}
