/*
 * Copyright (c) 2011-2017 Imagination Technologies Ltd.
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <linux/device.h>
#include <linux/etherdevice.h>
#include <linux/firmware.h>
#include <linux/interrupt.h>
#include <linux/ip.h>
#include <linux/kernel.h>
#include <linux/moduleparam.h>
#ifdef CONFIG_PM_SLEEP
#include <linux/pm_runtime.h>
#endif
#include <linux/udp.h>
#include <linux/version.h>

#include <net/cfg80211.h>
#include <net/mac80211.h>

#include "core.h"
#include "p2p.h"
#include "utils.h"
#include "fwldr.h"


/* Its value will be the default mac address and it can only be updated with the
 * command line arguments
 */
unsigned int vht_support = 1;
module_param(vht_support, int, 0);
MODULE_PARM_DESC(vht_support, "Configure the 11ac support for this device");

unsigned int ftm;
module_param(ftm, int, 0);
MODULE_PARM_DESC(ftm, "Factory Test Mode, should be used only for calibrations.");

unsigned int system_rev = 0x494D47; /*ASCII: IMG*/

int rpu_debug = RPU_DEBUG_UMACIF + RPU_DEBUG_CRYPTO + RPU_DEBUG_MAIN;

module_param(rpu_debug, uint, 0);
MODULE_PARM_DESC(rpu_debug, " rpu_debug: Configure Debugging Mask");
int uccp_reinit;

#define CHAN2G(_freq, _idx)  {		\
	.band = IEEE80211_BAND_2GHZ,	\
	.center_freq = (_freq),		\
	.hw_value = (_idx),		\
	.max_power = 20,		\
}

#define CHAN5G(_freq, _idx, _flags) {	\
	.band = IEEE80211_BAND_5GHZ,	\
	.center_freq = (_freq),		\
	.hw_value = (_idx),		\
	.max_power = 20,		\
	.flags = (_flags),		\
}

static struct ieee80211_channel dsss_chantable[] = {
	CHAN2G(2412, 0),  /* Channel 1 */
	CHAN2G(2417, 1),  /* Channel 2 */
	CHAN2G(2422, 2),  /* Channel 3 */
	CHAN2G(2427, 3),  /* Channel 4 */
	CHAN2G(2432, 4),  /* Channel 5 */
	CHAN2G(2437, 5),  /* Channel 6 */
	CHAN2G(2442, 6),  /* Channel 7 */
	CHAN2G(2447, 7),  /* Channel 8 */
	CHAN2G(2452, 8),  /* Channel 9 */
	CHAN2G(2457, 9),  /* Channel 10 */
	CHAN2G(2462, 10), /* Channel 11 */
	CHAN2G(2467, 11), /* Channel 12 */
	CHAN2G(2472, 12), /* Channel 13 */
	CHAN2G(2484, 13), /* Channel 14 */
};

static struct ieee80211_channel ofdm_chantable[] = {
	CHAN5G(5180, 14, 0), /* Channel 36 */
	CHAN5G(5200, 15, 0), /* Channel 40 */
	CHAN5G(5220, 16, 0), /* Channel 44 */
	CHAN5G(5240, 17, 0), /* Channel 48 */
	CHAN5G(5260, 18, IEEE80211_CHAN_RADAR), /* Channel 52 */
	CHAN5G(5280, 19, IEEE80211_CHAN_RADAR), /* Channel 56 */
	CHAN5G(5300, 20, IEEE80211_CHAN_RADAR), /* Channel 60 */
	CHAN5G(5320, 21, IEEE80211_CHAN_RADAR), /* Channel 64 */
	CHAN5G(5500, 22, IEEE80211_CHAN_RADAR), /* Channel 100 */
	CHAN5G(5520, 23, IEEE80211_CHAN_RADAR), /* Channel 104 */
	CHAN5G(5540, 24, IEEE80211_CHAN_RADAR), /* Channel 108 */
	CHAN5G(5560, 25, IEEE80211_CHAN_RADAR), /* Channel 112 */
	CHAN5G(5580, 26, IEEE80211_CHAN_RADAR), /* Channel 116 */
	CHAN5G(5600, 27, IEEE80211_CHAN_RADAR), /* Channel 120 */
	CHAN5G(5620, 28, IEEE80211_CHAN_RADAR), /* Channel 124 */
	CHAN5G(5640, 29, IEEE80211_CHAN_RADAR), /* Channel 128 */
	CHAN5G(5660, 30, IEEE80211_CHAN_RADAR), /* Channel 132 */
	CHAN5G(5680, 31, IEEE80211_CHAN_RADAR), /* Channel 136 */
	CHAN5G(5700, 32, IEEE80211_CHAN_RADAR), /* Channel 140 */
	CHAN5G(5720, 33, IEEE80211_CHAN_RADAR), /* Channel 144 */
	CHAN5G(5745, 34, 0), /* Channel 149 */
	CHAN5G(5765, 35, 0), /* Channel 153 */
	CHAN5G(5785, 36, 0), /* Channel 157 */
	CHAN5G(5805, 37, 0), /* Channel 161 */
	CHAN5G(5825, 38, 0), /* Channel 165 */
};

static struct ieee80211_rate ofdm_rates[] = {
	{ .bitrate = 60, .hw_value = 12},
	{ .bitrate = 90, .hw_value = 18},
	{ .bitrate = 120, .hw_value = 24},
	{ .bitrate = 180, .hw_value = 36},
	{ .bitrate = 240, .hw_value = 48},
	{ .bitrate = 360, .hw_value = 72},
	{ .bitrate = 480, .hw_value = 96},
	{ .bitrate = 540, .hw_value = 108}
};

static struct ieee80211_rate dsss_rates[] = {
	{ .bitrate = 10, .hw_value = 2},
	{ .bitrate = 20, .hw_value = 4,
	.flags = IEEE80211_RATE_SHORT_PREAMBLE},
	{ .bitrate = 55, .hw_value = 11,
	.flags = IEEE80211_RATE_SHORT_PREAMBLE},
	{ .bitrate = 110, .hw_value = 22,
	.flags = IEEE80211_RATE_SHORT_PREAMBLE},
	{ .bitrate = 60, .hw_value = 12},
	{ .bitrate = 90, .hw_value = 18},
	{ .bitrate = 120, .hw_value = 24},
	{ .bitrate = 180, .hw_value = 36},
	{ .bitrate = 240, .hw_value = 48},
	{ .bitrate = 360, .hw_value = 72},
	{ .bitrate = 480, .hw_value = 96},
	{ .bitrate = 540, .hw_value = 108}
};

static struct ieee80211_supported_band band_2ghz = {
	.channels = dsss_chantable,
	.n_channels = ARRAY_SIZE(dsss_chantable),
	.band = IEEE80211_BAND_2GHZ,
	.bitrates = dsss_rates,
	.n_bitrates = ARRAY_SIZE(dsss_rates),
};

static struct ieee80211_supported_band band_5ghz = {
	.channels = ofdm_chantable,
	.n_channels = ARRAY_SIZE(ofdm_chantable),
	.band = IEEE80211_BAND_5GHZ,
	.bitrates = ofdm_rates,
	.n_bitrates = ARRAY_SIZE(ofdm_rates),
};

/* Interface combinations for Virtual interfaces */
static const struct ieee80211_iface_limit if_limit1[] = {
		{ .max = 2, .types = BIT(NL80211_IFTYPE_STATION)}
};

static const struct ieee80211_iface_limit if_limit2[] = {
		{ .max = 1, .types = BIT(NL80211_IFTYPE_STATION)},
		{ .max = 1, .types = BIT(NL80211_IFTYPE_AP) |
				     BIT(NL80211_IFTYPE_P2P_CLIENT) |
				     BIT(NL80211_IFTYPE_ADHOC) |
				     BIT(NL80211_IFTYPE_P2P_GO)}
};

static const struct ieee80211_iface_limit if_limit3[] = {
		{ .max = 2, .types = BIT(NL80211_IFTYPE_P2P_CLIENT)}
};

static const struct ieee80211_iface_limit if_limit4[] = {
		{ .max = 1, .types = BIT(NL80211_IFTYPE_ADHOC)},
		{ .max = 1, .types = BIT(NL80211_IFTYPE_P2P_CLIENT)}
};

static const struct ieee80211_iface_limit if_limit5[] = {
		{ .max = 1, .types = BIT(NL80211_IFTYPE_STATION)},
		{ .max = 1, .types = BIT(NL80211_IFTYPE_AP) |
				     BIT(NL80211_IFTYPE_P2P_GO) |
				     BIT(NL80211_IFTYPE_P2P_CLIENT)}
};

static const struct ieee80211_iface_limit if_limit6[] = {
		{ .max = 1, .types = BIT(NL80211_IFTYPE_AP)}
};

static const struct ieee80211_iface_limit if_limit7[] = {
		{ .max = 1, .types = BIT(NL80211_IFTYPE_STATION)}
};

static const struct ieee80211_iface_combination if_comb[] = {
	{ .limits = if_limit1,
	  .n_limits = ARRAY_SIZE(if_limit1),
	  .max_interfaces = 2,
	  .num_different_channels = 1},
	{ .limits = if_limit2,
	  .n_limits = ARRAY_SIZE(if_limit2),
	  .max_interfaces = 2,
	  .num_different_channels = 1},
	{ .limits = if_limit3,
	  .n_limits = ARRAY_SIZE(if_limit3),
	  .max_interfaces = 2,
	  .num_different_channels = 1},
	{ .limits = if_limit5,
	  .n_limits = ARRAY_SIZE(if_limit5),
	  .max_interfaces = 2,
	  .num_different_channels = 2},
	{ .limits = if_limit1,
	  .n_limits = ARRAY_SIZE(if_limit1),
	  .max_interfaces = 2,
	  .num_different_channels = 2},
	{ .limits = if_limit4,
	  .n_limits = ARRAY_SIZE(if_limit4),
	  .max_interfaces = 2,
	  .num_different_channels = 1},
	{ .limits = if_limit7,
	  .n_limits = ARRAY_SIZE(if_limit7),
	  .max_interfaces = 1,
	  .num_different_channels = 1,
	  .radar_detect_widths = BIT(NL80211_CHAN_WIDTH_20_NOHT) |
				 BIT(NL80211_CHAN_WIDTH_20) |
				 BIT(NL80211_CHAN_WIDTH_40) |
				 BIT(NL80211_CHAN_WIDTH_80) |
				 BIT(NL80211_CHAN_WIDTH_160)}
};


static const struct wiphy_wowlan_support uccp_wowlan_support = {
	.flags = WIPHY_WOWLAN_ANY,
};

static void tx(struct ieee80211_hw *hw,
	       struct ieee80211_tx_control *txctl,
	       struct sk_buff *skb)
{
	struct img_priv *priv = hw->priv;
	struct ieee80211_hdr *hdr = (struct ieee80211_hdr *) skb->data;
	struct ieee80211_tx_info *tx_info = IEEE80211_SKB_CB(skb);
	struct umac_vif *uvif;
	unsigned char null_bssid[6] = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
	struct iphdr *iphdr;
	unsigned char *pktgen_magic;
	unsigned int orig_pktgen_magic = 0x55e99bbe; /*Endianness 0xbe9be955*/
	struct rpu_event_noa noa_event;
	int curr_chanctx_idx = -1;

	if (tx_info->control.vif == NULL) {
		RPU_DEBUG_UMACIF("%s: Dropping injected TX frame\n",
			 priv->name);
		dev_kfree_skb_any(skb);
		return;
	}

	uvif = (struct umac_vif *)(tx_info->control.vif->drv_priv);

	if (wifi->params.production_test) {
		if (((hdr->frame_control &
		      IEEE80211_FCTL_FTYPE) != IEEE80211_FTYPE_DATA) ||
		    (tx_info->control.vif == NULL))
			goto tx_status;

		iphdr = (struct iphdr *) skb_network_header(skb);
		if (iphdr->protocol == IPPROTO_UDP) {
			pktgen_magic = skb_transport_header(skb);
			pktgen_magic += sizeof(struct udphdr);
			/*If not PKTGEN, then drop it*/
			if (memcmp(pktgen_magic, &orig_pktgen_magic, 4) != 0) {
				RPU_DEBUG_UMACIF("%s:%d Prod_Mode: The pkt ",
						   __func__, __LINE__);
				RPU_DEBUG_UMACIF("is NOT PKTGEN so ");
				RPU_DEBUG_UMACIF("dropping it\n");
				goto tx_status;
			}
		} else {
			RPU_DEBUG_UMACIF("%s:%d prod_mode: The pkt is NOT ",
					   __func__, __LINE__);
			RPU_DEBUG_UMACIF("PKTGEN so dropping it\n");
			goto tx_status;
		}
	}
	if (ether_addr_equal(hdr->addr3, null_bssid))
		goto tx_status;

	if (uvif->vif->type != NL80211_IFTYPE_AP) {
		if ((priv->power_save == PWRSAVE_STATE_DOZE) &&
		    (!wifi->params.disable_power_save) &&
		    (((hdr->frame_control &
		      IEEE80211_FCTL_FTYPE) == IEEE80211_FTYPE_DATA) ||
			 is_bufferable_mgmt_frame(hdr))) {
			hdr->frame_control |= IEEE80211_FCTL_PM;
		}
	}

	if (uvif->noa_active) {
		memset(&noa_event, 0, sizeof(noa_event));
		noa_event.if_index = uvif->vif_index;
		rpu_noa_event(FROM_TX, &noa_event, priv, skb);
		return;
	}

	spin_lock_bh(&priv->chanctx_lock);
	curr_chanctx_idx = priv->curr_chanctx_idx;
	spin_unlock_bh(&priv->chanctx_lock);

	rpu_tx_frame(skb,
			     txctl->sta,
			     priv,
			     curr_chanctx_idx,
			     false);

	return;

tx_status:
	tx_info->flags |= IEEE80211_TX_STAT_ACK;
	tx_info->status.rates[0].count = 1;
	ieee80211_tx_status(hw, skb);
}

static int start(struct ieee80211_hw *hw)
{
	struct img_priv *priv = (struct img_priv *)hw->priv;
	int ret = 0;

	RPU_DEBUG_UMACIF("%s-80211IF: In start\n", priv->name);

	mutex_lock(&priv->mutex);

#ifdef CONFIG_PM_SLEEP
	RPU_DEBUG_RUNTIME_PM("%s runtime_pm: waking up the chip\n", priv->name);

	ret = pm_runtime_get_sync(priv->dev);

	if (ret < 0 && ret != -EACCES) {
		ret = -ENODEV;
		pr_warn("%s: pm_runtime_get_sync failed: %d\n", __func__, ret);
		goto out;
	} else {
		RPU_DEBUG_RUNTIME_PM("%s: pm_runtime_get_sync succees: %d\n",
				     __func__, ret);
		ret = 0;
	}
#endif

	if (wifi->params.fw_loading == 1) {
		if (load_fw(hw)) {
			RPU_DEBUG_UMACIF("%s-80211IF: FW load failed\n",
					 priv->name);
			ret = -ENODEV;
			goto out;
		}
	}

	if ((rpu_core_init(priv, ftm)) < 0) {
		RPU_DEBUG_UMACIF("%s-80211IF: umac init failed\n", priv->name);
		ret = -ENODEV;
		goto out;
	}

	INIT_DELAYED_WORK(&priv->roc_complete_work, rpu_roc_complete_work);

	priv->state = STARTED;
	memset(priv->params->pdout_voltage, 0,
	       sizeof(char) * MAX_AUX_ADC_SAMPLES);
	priv->roc_off_chanctx_idx = -1;
out:
	mutex_unlock(&priv->mutex);
#ifdef CONFIG_PM_SLEEP
	if (ret < 0)
		pm_runtime_put_sync(priv->dev);
#endif
	return ret;
}

void stop(struct ieee80211_hw *hw)
{
	struct img_priv *priv = (struct img_priv *)hw->priv;
#ifdef CONFIG_PM_SLEEP
	int ret = 0;
#endif

	RPU_DEBUG_UMACIF("%s-80211IF:In stop\n", priv->name);
	mutex_lock(&priv->mutex);
	rpu_core_deinit(priv, ftm);
	priv->state = STOPPED;
	mutex_unlock(&priv->mutex);

	hal_ops.reset_hal_params();

#ifdef CONFIG_PM_SLEEP
	RPU_DEBUG_RUNTIME_PM("%s runtime_pm: powering off the chip\n",
			     priv->name);
	ret = pm_runtime_put_sync(priv->dev);

	if (ret)
		pr_warn("%s: pm_runtime_put_sync failed: %d\n", __func__, ret);
#endif

}

static int add_interface(struct ieee80211_hw *hw,
		struct ieee80211_vif *vif)
{
	struct img_priv *priv = hw->priv;
	struct ieee80211_vif *v;
	struct umac_vif *uvif;
	int vif_index, iftype;

	mutex_lock(&priv->mutex);
	iftype = vif->type;
	v = vif;
	vif->driver_flags |= IEEE80211_VIF_BEACON_FILTER;
	vif->driver_flags |= IEEE80211_VIF_SUPPORTS_UAPSD;

	if (priv->current_vif_count == wifi->params.num_vifs) {
		pr_err("%s: Exceeded Maximum supported VIF's cur:%d max: %d.\n",
		       __func__,
		       priv->current_vif_count,
		       wifi->params.num_vifs);

		mutex_unlock(&priv->mutex);
		return -ENOTSUPP;
	}

	if (!(iftype == NL80211_IFTYPE_STATION ||
	      iftype == NL80211_IFTYPE_ADHOC ||
	      iftype == NL80211_IFTYPE_AP)) {
		pr_err("Invalid Interface type\n");
		return -ENOTSUPP;
	}

	if (wifi->params.production_test) {
		if (priv->active_vifs || iftype != NL80211_IFTYPE_ADHOC) {
			mutex_unlock(&priv->mutex);
			return -EBUSY;
		}
	}

	for (vif_index = 0; vif_index < wifi->params.num_vifs; vif_index++) {
		if (!(priv->active_vifs & (1 << vif_index)))
			break;
	}

	/* This should never happen, we have taken care of this above */
	if (vif_index == wifi->params.num_vifs) {
		pr_err("%s: All VIF's are busy: %pM\n", __func__, vif->addr);
		mutex_unlock(&priv->mutex);
		return -EINVAL;
	}

	uvif = (struct umac_vif *)&v->drv_priv;
	uvif->vif_index = vif_index;
	uvif->vif = v;
	uvif->priv = priv;
	uvif->seq_no = 0;
	rpu_vif_add(uvif);
	priv->active_vifs |= (1 << vif_index);
	priv->current_vif_count++;

	if (iftype == NL80211_IFTYPE_ADHOC)
		priv->tx_last_beacon = 0;

	rcu_assign_pointer(priv->vifs[vif_index], v);
	synchronize_rcu();

	mutex_unlock(&priv->mutex);

	return 0;
}

static void remove_interface(struct ieee80211_hw *hw,
		struct ieee80211_vif *vif)
{
	struct img_priv *priv = hw->priv;
	struct ieee80211_vif *v;
	int vif_index;

	mutex_lock(&priv->mutex);
	v = vif;
	vif_index = ((struct umac_vif *)&v->drv_priv)->vif_index;

	rpu_vif_remove((struct umac_vif *)&v->drv_priv);
	priv->active_vifs &= ~(1 << vif_index);
	rcu_assign_pointer(priv->vifs[vif_index], NULL);
	synchronize_rcu();
	wifi->params.sync[vif_index].status = 0;
	priv->current_vif_count--;
	mutex_unlock(&priv->mutex);

}


static int config(struct ieee80211_hw *hw,
		unsigned int changed)
{
	struct img_priv *priv = hw->priv;
	struct ieee80211_conf *conf = &hw->conf;
	unsigned int pri_chnl_num;
	unsigned int freq_band;
	unsigned int ch_width;
	unsigned int center_freq = 0;
	unsigned int center_freq1 = 0;
	unsigned int center_freq2 = 0;
	int i = 0;
	int err = 0;
	struct ieee80211_vif *vif = NULL;
	int ret = 0;

	mutex_lock(&priv->mutex);

	if (changed & IEEE80211_CONF_CHANGE_POWER) {
		priv->txpower = conf->power_level;
		CALL_RPU(rpu_prog_txpower, priv->txpower);
	}

	/* Check for change in channel */
	if (changed & IEEE80211_CONF_CHANGE_CHANNEL) {
		center_freq = conf->chandef.chan->center_freq;
		center_freq1 = conf->chandef.center_freq1;
		center_freq2 = conf->chandef.center_freq2;
		freq_band = conf->chandef.chan->band;
		ch_width = conf->chandef.width;

		pri_chnl_num = ieee80211_frequency_to_channel(center_freq);
		RPU_DEBUG_UMACIF("%s-80211IF:Primary Channel is %d\n",
				   priv->name,
				   pri_chnl_num);

		err = rpu_prog_channel(pri_chnl_num,
					       center_freq1, center_freq2,
					       ch_width,
					       0,
					       freq_band);

		if (err) {
			mutex_unlock(&priv->mutex);
			return err;
		}
		if (conf->radar_enabled) {
			RPU_DEBUG_UMACIF("RADAR Detection ENABLED on ");
			RPU_DEBUG_UMACIF("PriChannel=%d with ch_width=%d\n",
					   pri_chnl_num,
					   ch_width);

			CALL_RPU(rpu_prog_radar_detect,
				  RADAR_DETECT_OP_START);
		}
	}

	/* Check for change in Power save state */
	for (i = 0; i < MAX_VIFS; i++) {
		if (!(changed & IEEE80211_CONF_CHANGE_PS))
			break;

		if (!(priv->active_vifs & (1 << i)))
			continue;

		/* When ROC is in progress, do not mess with
		 * PS state
		 */
		if (priv->roc_params.roc_in_progress)
			continue;

		if (wifi->params.disable_power_save)
			continue;

		if (conf->flags & IEEE80211_CONF_PS)
			priv->power_save = PWRSAVE_STATE_DOZE;
		else
			priv->power_save = PWRSAVE_STATE_AWAKE;

		RPU_DEBUG_UMACIF("%s-80211IF:PS state of VIF", priv->name);
		RPU_DEBUG_UMACIF(" %d changed to %d\n", i, priv->power_save);

		rcu_read_lock();
		vif = rcu_dereference(priv->vifs[i]);
		rcu_read_unlock();

		rpu_prog_ps_state(i,
					  vif->addr,
					  priv->power_save);
	}

	/* TODO: Make this global config as it effects all VIF's */
	for (i = 0; i < MAX_VIFS; i++) {
		if (!(changed & IEEE80211_CONF_CHANGE_SMPS))
			break;

		if (wifi->params.production_test == 1)
			break;

		if (!(priv->active_vifs & (1 << i)))
			continue;

		RPU_DEBUG_UMACIF("%s-80211IF:MIMO PS state of VIF %d -> %d\n",
				   priv->name,
				   i,
				   conf->smps_mode);

		rcu_read_lock();
		vif = rcu_dereference(priv->vifs[i]);
		rcu_read_unlock();

		rpu_prog_vif_smps(i,
					  vif->addr,
					  conf->smps_mode);
	}

	/* Check for change in Retry Limits */
	if (changed & IEEE80211_CONF_CHANGE_RETRY_LIMITS) {
		RPU_DEBUG_UMACIF("%s-80211IF:Retry Limits changed",
				   priv->name);
		RPU_DEBUG_UMACIF(" to %d and %d\n",
				   conf->short_frame_max_tx_count,
				   conf->long_frame_max_tx_count);
	}

	for (i = 0; i < MAX_VIFS; i++) {
		if (!(changed & IEEE80211_CONF_CHANGE_RETRY_LIMITS))
			break;

		if (!(priv->active_vifs & (1 << i)))
			continue;

		rcu_read_lock();
		vif = rcu_dereference(priv->vifs[i]);
		rcu_read_unlock();

		rpu_prog_short_retry(i,
					     vif->addr,
					     conf->short_frame_max_tx_count);
		rpu_prog_long_retry(i,
					    vif->addr,
					    conf->long_frame_max_tx_count);
	}

prog_rpu_fail:
	mutex_unlock(&priv->mutex);
	return ret;
}


static u64 prepare_multicast(struct ieee80211_hw *hw,
			     struct netdev_hw_addr_list *mc_list)
{
	struct img_priv *priv = hw->priv;
	int i;
	struct netdev_hw_addr *ha;
	int mc_count = 0;
	int ret = 0;

	if (priv->state != STARTED)
		return 0;

	mc_count = netdev_hw_addr_list_count(mc_list);
	{
		if (mc_count > MCST_ADDR_LIMIT) {
			mc_count = 0;
			pr_warn("%s-80211IF:Disabling MCAST filter (cnt=%d)\n",
				priv->name, mc_count);
			goto out;
		}
	}
	RPU_DEBUG_UMACIF("%s-80211IF: Multicast filter count", priv->name);
	RPU_DEBUG_UMACIF("adding: %d removing: %d\n", mc_count,
			priv->mc_filter_count);

	if (priv->mc_filter_count > 0) {
		/* Remove all previous multicast addresses from the LMAC */
		for (i = 0; i < priv->mc_filter_count; i++)
			rpu_prog_mcast_addr_cfg(priv->mc_filters[i],
							WLAN_MCAST_ADDR_REM);
	}

	i = 0;

	netdev_hw_addr_list_for_each(ha, mc_list) {
		/* Prog the multicast address into the LMAC */
		CALL_RPU(rpu_prog_mcast_addr_cfg,
			  ha->addr,
			  WLAN_MCAST_ADDR_ADD);
		memcpy(priv->mc_filters[i], ha->addr, 6);
		i++;
	}

	priv->mc_filter_count = mc_count;
out:
	return mc_count;
prog_rpu_fail:
	return ret;
}


static void configure_filter(struct ieee80211_hw *hw,
		unsigned int changed_flags,
		unsigned int *new_flags,
		u64 mc_count)
{
	struct img_priv *priv = hw->priv;
	int ret = 0;

	mutex_lock(&priv->mutex);

	changed_flags &= SUPPORTED_FILTERS;
	*new_flags &= SUPPORTED_FILTERS;

	if (priv->state != STARTED) {
		mutex_unlock(&priv->mutex);
		return;
	}

	if ((*new_flags & FIF_ALLMULTI) || (mc_count == 0)) {
		/* Disable the multicast filter in LMAC */
		RPU_DEBUG_UMACIF("%s-80211IF: Multicast filters disabled\n",
				   priv->name);
		CALL_RPU(rpu_prog_mcast_filter_control,
			  MCAST_FILTER_DISABLE);
	} else if (mc_count) {
		/* Enable the multicast filter in LMAC */
		RPU_DEBUG_UMACIF("%s-80211IF: Multicast filters enabled\n",
			       priv->name);
		CALL_RPU(rpu_prog_mcast_filter_control,
			  MCAST_FILTER_ENABLE);
	}

	if (changed_flags == 0)
		/* No filters which we support changed */
		goto out;

	if (wifi->params.production_test == 0) {
		if (*new_flags & FIF_BCN_PRBRESP_PROMISC) {
			/* Receive all beacons and probe responses */
			RPU_DEBUG_UMACIF("%s-80211IF: RCV ALL bcns\n",
				       priv->name);
			CALL_RPU(rpu_prog_rcv_bcn_mode, RCV_ALL_BCNS);
		} else {
			/* Receive only network beacons and probe responses */
			RPU_DEBUG_UMACIF("%s-80211IF: RCV NW bcns\n",
					   priv->name);
			CALL_RPU(rpu_prog_rcv_bcn_mode,
				  RCV_ALL_NETWORK_ONLY);
		}
	}
out:
	if (wifi->params.production_test == 1) {
		RPU_DEBUG_UMACIF("%s-80211IF: RCV ALL bcns\n", priv->name);
		CALL_RPU(rpu_prog_rcv_bcn_mode, RCV_ALL_BCNS);
	}

prog_rpu_fail:
	mutex_unlock(&priv->mutex);
}


static int conf_vif_tx(struct ieee80211_hw  *hw,
		       struct ieee80211_vif *vif,
		       unsigned short queue,
		       const struct ieee80211_tx_queue_params *txq_params)
{
	struct img_priv *priv = hw->priv;
	int vif_index, vif_active;
	struct edca_params params;
	struct ieee80211_vif *vif_local = NULL;

	mutex_lock(&priv->mutex);

	for (vif_index = 0; vif_index < wifi->params.num_vifs; vif_index++) {
		if (!(priv->active_vifs & (1 << vif_index)))
			continue;

		rcu_read_lock();
		vif_local = rcu_dereference(priv->vifs[vif_index]);
		rcu_read_unlock();

		if (ether_addr_equal(vif_local->addr,
				     vif->addr))
			break;
	}

	if (WARN_ON(vif_index == wifi->params.num_vifs)) {
		mutex_unlock(&priv->mutex);
		return -EINVAL;
	}

	vif_active = 0;

	if ((priv->active_vifs & (1 << vif_index)))
		vif_active = 1;

	memset(&params, 0, sizeof(params));
	params.aifs = txq_params->aifs;
	params.txop = txq_params->txop;
	params.cwmin = txq_params->cw_min;
	params.cwmax = txq_params->cw_max;
	params.uapsd = txq_params->uapsd;

	rpu_vif_set_edca_params(queue,
					(struct umac_vif *)&vif->drv_priv,
					&params,
					vif_active);
	mutex_unlock(&priv->mutex);
	return 0;
}


static int set_key(struct ieee80211_hw *hw,
		   enum set_key_cmd cmd,
		   struct ieee80211_vif *vif,
		   struct ieee80211_sta *sta,
		   struct ieee80211_key_conf *key_conf)
{

	struct umac_key sec_key;
	unsigned int result = 0;
	struct img_priv *priv = hw->priv;
	unsigned int cipher_type, key_type;
	int vif_index;
	struct umac_vif *uvif;

	uvif = ((struct umac_vif *)&vif->drv_priv);

	memset(&sec_key, 0, sizeof(struct umac_key));

	switch (key_conf->cipher) {
	case WLAN_CIPHER_SUITE_WEP40:
		sec_key.key = key_conf->key;
		cipher_type = CIPHER_TYPE_WEP40;
		break;
	case WLAN_CIPHER_SUITE_WEP104:
		sec_key.key = key_conf->key;
		cipher_type = CIPHER_TYPE_WEP104;
		break;
	case WLAN_CIPHER_SUITE_TKIP:
		key_conf->flags |= IEEE80211_KEY_FLAG_GENERATE_MMIC;
		/* We get the key in the following form:
		 * KEY (16 bytes) - TX MIC (8 bytes) - RX MIC (8 bytes)
		 */
		sec_key.key = key_conf->key;
		sec_key.tx_mic = key_conf->key + 16;
		sec_key.rx_mic = key_conf->key + 24;
		cipher_type = CIPHER_TYPE_TKIP;
		break;
	case WLAN_CIPHER_SUITE_CCMP:
		sec_key.key = key_conf->key;
		cipher_type = CIPHER_TYPE_CCMP;
		break;
	default:
		result = -EOPNOTSUPP;
		mutex_unlock(&priv->mutex);
		goto out;
	}

	vif_index = ((struct umac_vif *)&vif->drv_priv)->vif_index;

	mutex_lock(&priv->mutex);

	if (cmd == SET_KEY) {
		key_conf->hw_key_idx = 0; /* Don't really use this */

		/* This flag indicate that it requires IV generation */
		key_conf->flags |= IEEE80211_KEY_FLAG_GENERATE_IV;


		if (cipher_type == CIPHER_TYPE_WEP40 ||
		    cipher_type == CIPHER_TYPE_WEP104) {
			RPU_DEBUG_CRYPTO("%s-80211IF: ADD IF KEY (WEP).",
					  priv->name);
			RPU_DEBUG_CRYPTO(" vif_index = %d,", vif_index);
			RPU_DEBUG_CRYPTO(" keyidx = %d, cipher_type = %d\n",
					  key_conf->keyidx, cipher_type);

			rpu_prog_if_key(vif_index,
						vif->addr,
						KEY_CTRL_ADD,
						key_conf->keyidx,
						cipher_type,
						&sec_key);
		} else if (sta) {
			sec_key.peer_mac = sta->addr;

			if (key_conf->flags & IEEE80211_KEY_FLAG_PAIRWISE)
				key_type = KEY_TYPE_UCAST;
			else
				key_type = KEY_TYPE_BCAST;
			RPU_DEBUG_CRYPTO("%s-80211IF: ADD PEER KEY (WPA/WPA2)",
					  priv->name);
			RPU_DEBUG_CRYPTO(" vif_index = %d,", vif_index);
			RPU_DEBUG_CRYPTO(" keyidx = %d, keytype = %d,",
					  key_conf->keyidx, key_type);
			RPU_DEBUG_CRYPTO(" cipher_type = %d\n", cipher_type);

			rpu_prog_peer_key(vif_index,
						  vif->addr,
						  KEY_CTRL_ADD,
						  key_conf->keyidx,
						  key_type,
						  cipher_type,
						  &sec_key);
		} else {
			key_type = KEY_TYPE_BCAST;

			if (vif->type == NL80211_IFTYPE_STATION) {
				sec_key.peer_mac =
					(unsigned char *)vif->bss_conf.bssid;

				memcpy(uvif->bssid,
				       (vif->bss_conf.bssid),
				       ETH_ALEN);
				RPU_DEBUG_CRYPTO("%s-80211IF: ADD PEER KEY ",
						  priv->name);
				RPU_DEBUG_CRYPTO("(BCAST-STA). vif_index = %d",
						  vif_index);
				RPU_DEBUG_CRYPTO(", keyidx = %d, keytype = %d",
						key_conf->keyidx, key_type);
				RPU_DEBUG_CRYPTO(", cipher_type = %d\n",
						  cipher_type);

				rpu_prog_peer_key(vif_index,
							  vif->addr,
							  KEY_CTRL_ADD,
							  key_conf->keyidx,
							  key_type, cipher_type,
							  &sec_key);

			} else if (vif->type == NL80211_IFTYPE_AP) {
				RPU_DEBUG_CRYPTO("%s-80211IF: ADD IF KEY ",
						  priv->name);
				RPU_DEBUG_CRYPTO("(BCAST-AP). vif_index = %d",
						  vif_index);
				RPU_DEBUG_CRYPTO(", keyidx = %d",
						  key_conf->keyidx);
				RPU_DEBUG_CRYPTO(", cipher_type = %d\n",
						  cipher_type);

				rpu_prog_if_key(vif_index,
							vif->addr,
							KEY_CTRL_ADD,
							key_conf->keyidx,
							cipher_type,
							&sec_key);
			} else {
				/* ADHOC */
				/* TODO: Check this works for IBSS RSN */
				RPU_DEBUG_CRYPTO("%s-80211IF: ADD IF KEY ",
						  priv->name);
				RPU_DEBUG_CRYPTO("(BCAST-IBSS).vif_index = %d",
						  vif_index);
				RPU_DEBUG_CRYPTO(", keyidx = %d",
						  key_conf->keyidx);
				RPU_DEBUG_CRYPTO(", cipher_type = %d\n",
						  cipher_type);

				rpu_prog_if_key(vif_index,
							vif->addr,
							KEY_CTRL_ADD,
							key_conf->keyidx,
							cipher_type,
							&sec_key);
			}
		}
	} else if (cmd == DISABLE_KEY) {
		if ((cipher_type == CIPHER_TYPE_WEP40) ||
		    (cipher_type == CIPHER_TYPE_WEP104)) {
			rpu_prog_if_key(vif_index,
						vif->addr,
						KEY_CTRL_DEL,
						key_conf->keyidx,
						cipher_type,
						&sec_key);
			RPU_DEBUG_CRYPTO("%s-80211IF: DEL IF KEY (WEP).",
					  priv->name);
			RPU_DEBUG_CRYPTO(" vif_index = %d, keyidx = %d",
					  vif_index, key_conf->keyidx);
			RPU_DEBUG_CRYPTO(", cipher_type = %d\n", cipher_type);
		} else if (sta) {
			sec_key.peer_mac = sta->addr;

			if (key_conf->flags & IEEE80211_KEY_FLAG_PAIRWISE)
				key_type = KEY_TYPE_UCAST;
			else
				key_type = KEY_TYPE_BCAST;
			RPU_DEBUG_CRYPTO("%s-80211IF: DEL IF KEY (WPA/WPA2).",
					  priv->name);
			RPU_DEBUG_CRYPTO(" vif_index = %d, keyidx = %d",
					  vif_index, key_conf->keyidx);
			RPU_DEBUG_CRYPTO(", cipher_type = %d\n", cipher_type);

			rpu_prog_peer_key(vif_index,
						  vif->addr,
						  KEY_CTRL_DEL,
						  key_conf->keyidx,
						  key_type,
						  cipher_type,
						  &sec_key);
		} else {
			if (vif->type == NL80211_IFTYPE_STATION) {
				sec_key.peer_mac = uvif->bssid;
				RPU_DEBUG_CRYPTO("%s-80211IF: DEL IF KEY ",
						  priv->name);
				RPU_DEBUG_CRYPTO("(BCAST-STA). vif_index = %d",
						  vif_index);
				RPU_DEBUG_CRYPTO(", keyidx = %d",
						  key_conf->keyidx);
				RPU_DEBUG_CRYPTO(", cipher_type = %d\n",
						  cipher_type);

				rpu_prog_peer_key(vif_index,
							  vif->addr,
							  KEY_CTRL_DEL,
							  key_conf->keyidx,
							  KEY_TYPE_BCAST,
							  cipher_type,
							  &sec_key);

			} else if (vif->type == NL80211_IFTYPE_AP) {
				RPU_DEBUG_CRYPTO("%s-80211IF: DEL IF KEY ",
						  priv->name);
				RPU_DEBUG_CRYPTO("(BCAST-AP). vif_index = %d",
						  vif_index);
				RPU_DEBUG_CRYPTO(", keyidx = %d",
						  key_conf->keyidx);
				RPU_DEBUG_CRYPTO(", cipher_type = %d\n",
						  cipher_type);

				rpu_prog_if_key(vif_index,
							vif->addr,
							KEY_CTRL_DEL,
							key_conf->keyidx,
							cipher_type,
							&sec_key);
			} else {
				RPU_DEBUG_CRYPTO("%s-80211IF: DEL IF KEY ",
						  priv->name);
				RPU_DEBUG_CRYPTO("(BCAST-IBSS).vif_index = %d",
						  vif_index);
				RPU_DEBUG_CRYPTO(", keyidx = %d",
						  key_conf->keyidx);
				RPU_DEBUG_CRYPTO(", cipher_type = %d\n",
						  cipher_type);

				rpu_prog_if_key(vif_index,
							vif->addr,
							KEY_CTRL_DEL,
							key_conf->keyidx,
							cipher_type,
							&sec_key);
			}
		}
	}

	mutex_unlock(&priv->mutex);

out:
	return result;
}


static void bss_info_changed(struct ieee80211_hw *hw,
			     struct ieee80211_vif *vif,
			     struct ieee80211_bss_conf *bss_conf,
			     unsigned int changed)
{
	struct img_priv *priv = hw->priv;

	mutex_lock(&priv->mutex);

	if (wifi->params.production_test) {
		/* Disable beacon generation when running pktgen
		 * for performance
		 */
		changed &= ~BSS_CHANGED_BEACON_INT;
		changed &= ~BSS_CHANGED_BEACON_ENABLED;
	}

	rpu_vif_bss_info_changed((struct umac_vif *)&vif->drv_priv,
					 bss_conf,
					 changed);
	mutex_unlock(&priv->mutex);
}


static void setup_ht_cap(struct ieee80211_sta_ht_cap *ht_info)
{
	int i;

	memset(ht_info, 0, sizeof(*ht_info));
	ht_info->ht_supported = true;
	pr_info("SETUP HT CALLED\n");

	ht_info->cap = 0;
	ht_info->cap |= IEEE80211_HT_CAP_MAX_AMSDU;
	ht_info->cap |= IEEE80211_HT_CAP_SGI_40;
	ht_info->cap |= IEEE80211_HT_CAP_SGI_20;
	ht_info->cap |= IEEE80211_HT_CAP_SUP_WIDTH_20_40;
	ht_info->cap |= IEEE80211_HT_CAP_GRN_FLD;
	ht_info->cap |= IEEE80211_HT_CAP_LDPC_CODING;
	ht_info->cap |= IEEE80211_HT_CAP_TX_STBC;
	ht_info->cap |= (1 << IEEE80211_HT_CAP_RX_STBC_SHIFT);
	ht_info->cap |= IEEE80211_HT_CAP_LSIG_TXOP_PROT;
	/*We support SMPS*/

	ht_info->ampdu_factor = IEEE80211_HT_MAX_AMPDU_64K;
	ht_info->ampdu_density = IEEE80211_HT_MPDU_DENSITY_4;

	memset(&ht_info->mcs, 0, sizeof(ht_info->mcs));

	if (wifi->params.max_tx_streams != wifi->params.max_rx_streams) {
		ht_info->mcs.tx_params |= IEEE80211_HT_MCS_TX_RX_DIFF;
		ht_info->mcs.tx_params |= ((wifi->params.max_tx_streams - 1)
				<< IEEE80211_HT_MCS_TX_MAX_STREAMS_SHIFT);
	}

	ht_info->mcs.tx_params |= IEEE80211_HT_MCS_TX_DEFINED;

	for (i = 0; i < wifi->params.max_rx_streams; i++)
		ht_info->mcs.rx_mask[i] = 0xff;
	ht_info->mcs.rx_mask[4] = 0x1;
}


#define IEEE80211_VHT_CAP_BEAMFORMEE_STS_SHIFT 13
#define IEEE80211_VHT_CAP_SOUNDING_DIMENSIONS_SHIFT 16
static void setup_vht_cap(struct ieee80211_sta_vht_cap *vht_info)
{
	if (!vht_support)
		return;

	memset(vht_info, 0, sizeof(*vht_info));
	vht_info->vht_supported = true;
	pr_info("SETUP VHT CALLED\n");

	vht_info->cap = IEEE80211_VHT_CAP_MAX_MPDU_LENGTH_11454 |
			/*64KB Rx buffer size*/
			(3 <<
			IEEE80211_VHT_CAP_MAX_A_MPDU_LENGTH_EXPONENT_SHIFT) |
			IEEE80211_VHT_CAP_SHORT_GI_80 |
			IEEE80211_VHT_CAP_RXLDPC |
			IEEE80211_VHT_CAP_TXSTBC |
			IEEE80211_VHT_CAP_RXSTBC_1 |
			IEEE80211_VHT_CAP_HTC_VHT;
	/* 1x1 */
	if ((wifi->params.max_tx_streams == 1) &&
	    (wifi->params.max_rx_streams == 1)) {
		vht_info->vht_mcs.rx_mcs_map =
			((IEEE80211_VHT_MCS_SUPPORT_0_7) << (2*0)) |
			((IEEE80211_VHT_MCS_NOT_SUPPORTED) << (2*1)) |
			((IEEE80211_VHT_MCS_NOT_SUPPORTED) << (2*2)) |
			((IEEE80211_VHT_MCS_NOT_SUPPORTED) << (2*3)) |
			((IEEE80211_VHT_MCS_NOT_SUPPORTED) << (2*4)) |
			((IEEE80211_VHT_MCS_NOT_SUPPORTED) << (2*5)) |
			((IEEE80211_VHT_MCS_NOT_SUPPORTED) << (2*6)) |
			((IEEE80211_VHT_MCS_NOT_SUPPORTED) << (2*7));
	}

	/*2x2 */
	if ((wifi->params.max_tx_streams == 2) &&
	    (wifi->params.max_rx_streams == 2)) {
		vht_info->vht_mcs.rx_mcs_map =
			((IEEE80211_VHT_MCS_SUPPORT_0_7) << (2*0)) |
			((IEEE80211_VHT_MCS_SUPPORT_0_7) << (2*1)) |
			((IEEE80211_VHT_MCS_NOT_SUPPORTED) << (2*2)) |
			((IEEE80211_VHT_MCS_NOT_SUPPORTED) << (2*3)) |
			((IEEE80211_VHT_MCS_NOT_SUPPORTED) << (2*4)) |
			((IEEE80211_VHT_MCS_NOT_SUPPORTED) << (2*5)) |
			((IEEE80211_VHT_MCS_NOT_SUPPORTED) << (2*6)) |
			((IEEE80211_VHT_MCS_NOT_SUPPORTED) << (2*7));
	}

	vht_info->vht_mcs.tx_mcs_map = vht_info->vht_mcs.rx_mcs_map;
}

static void set_hw_flags(struct ieee80211_hw *hw)
{
	hw->flags = IEEE80211_HW_SIGNAL_DBM;
	hw->flags |= IEEE80211_HW_SUPPORTS_PS;
	hw->flags |= IEEE80211_HW_HOST_BROADCAST_PS_BUFFERING;
	hw->flags |= IEEE80211_HW_AMPDU_AGGREGATION;
	hw->flags |= IEEE80211_HW_MFP_CAPABLE;
	hw->flags |= IEEE80211_HW_REPORTS_TX_ACK_STATUS;
	if (wifi->params.minstrel_fastpath)
		hw->flags |= IEEE80211_HW_SUPPORTS_RC_TABLE;
	if (wifi->params.dot11a_support)
		hw->flags |= IEEE80211_HW_SPECTRUM_MGMT;
	hw->flags |= IEEE80211_HW_SUPPORTS_PER_STA_GTK;
	hw->flags |= IEEE80211_HW_CONNECTION_MONITOR;
	hw->flags |= IEEE80211_HW_CHANCTX_STA_CSA;
}

static void init_hw(struct ieee80211_hw *hw)
{
	struct img_priv *priv = (struct img_priv *)hw->priv;
	int num_if_comb = 0;

	/* Supported Interface Types and other Default values*/
	hw->wiphy->interface_modes = BIT(NL80211_IFTYPE_STATION) |
				     BIT(NL80211_IFTYPE_ADHOC) |
				     BIT(NL80211_IFTYPE_AP) |
				     BIT(NL80211_IFTYPE_P2P_CLIENT) |
				     BIT(NL80211_IFTYPE_P2P_GO);

	hw->wiphy->iface_combinations = if_comb;

	num_if_comb = (sizeof(if_comb) /
		       sizeof(struct ieee80211_iface_combination));
	hw->wiphy->n_iface_combinations = num_if_comb;

	set_hw_flags(hw);
	hw->wiphy->max_scan_ssids = MAX_NUM_SSIDS; /* 4 */
	 /* Low priority bg scan */
	hw->wiphy->features |= NL80211_FEATURE_LOW_PRIORITY_SCAN;
	hw->wiphy->max_scan_ie_len = IEEE80211_MAX_DATA_LEN;
	hw->max_listen_interval = 10;
	hw->wiphy->max_remain_on_channel_duration = 5000; /*ROC*/
	hw->offchannel_tx_hw_queue = WLAN_AC_VO;
	hw->max_rates = 4;
	hw->max_rate_tries = wifi->params.max_rate_retries;
	hw->queues = 4;

	/* Size */
	hw->extra_tx_headroom = 0;
	hw->vif_data_size = sizeof(struct umac_vif);
	hw->sta_data_size = sizeof(struct umac_sta);
	hw->chanctx_data_size = sizeof(struct umac_chanctx);

	if (wifi->params.dot11g_support) {
		hw->wiphy->bands[IEEE80211_BAND_2GHZ] = &band_2ghz;
			setup_ht_cap(&band_2ghz.ht_cap);
	}

	if (wifi->params.dot11a_support) {
		hw->wiphy->bands[IEEE80211_BAND_5GHZ] = &band_5ghz;
		setup_ht_cap(&band_5ghz.ht_cap);
		if (vht_support)
			setup_vht_cap(&band_5ghz.vht_cap);
	}

	memset(hw->wiphy->addr_mask, 0, sizeof(hw->wiphy->addr_mask));

	if (wifi->params.num_vifs == 1) {
		hw->wiphy->addresses = NULL;
		SET_IEEE80211_PERM_ADDR(hw, priv->if_mac_addresses[0].addr);
	} else {
		hw->wiphy->n_addresses = wifi->params.num_vifs;
		hw->wiphy->addresses = priv->if_mac_addresses;
	}

	hw->wiphy->flags |= WIPHY_FLAG_AP_UAPSD;
	hw->wiphy->flags |= WIPHY_FLAG_IBSS_RSN;
	hw->wiphy->flags |= WIPHY_FLAG_HAS_REMAIN_ON_CHANNEL;
	hw->wiphy->flags |= WIPHY_FLAG_HAS_CHANNEL_SWITCH;

	if (!wifi->params.disable_power_save &&
	    !wifi->params.disable_sm_power_save) {
		/* SMPS Support both Static and Dynamic */
		hw->wiphy->features |= NL80211_FEATURE_STATIC_SMPS;
		hw->wiphy->features |= NL80211_FEATURE_DYNAMIC_SMPS;
	}
}


static int ampdu_action(struct ieee80211_hw *hw,
				struct ieee80211_vif *vif,
				enum ieee80211_ampdu_mlme_action action,
				struct ieee80211_sta *sta,
				u16 tid, u16 *ssn, u8 buf_size)
{
	int ret = 0;
	unsigned int val = 0;
	struct img_priv *priv = (struct img_priv *)hw->priv;

	switch (action) {
	case IEEE80211_AMPDU_RX_START:
		{
		val = tid | TID_INITIATOR_AP;
		priv->tid_info[val].tid_state = TID_STATE_AGGR_START;
		priv->tid_info[val].ssn = *ssn;
		rpu_prog_ba_session_data(1,
						 tid,
						 &priv->tid_info[val].ssn,
						 1,
						 vif->addr,
				   (unsigned char *)(vif->bss_conf.bssid));
		}
		break;
	case IEEE80211_AMPDU_RX_STOP:
		{
		val = tid | TID_INITIATOR_AP;
		priv->tid_info[val].tid_state = TID_STATE_AGGR_STOP;
		rpu_prog_ba_session_data(0,
						 tid,
						 &priv->tid_info[val].ssn,
						 1,
						 vif->addr,
				   (unsigned char *)(vif->bss_conf.bssid));
		}
		break;
	case IEEE80211_AMPDU_TX_START:
		{
		val = tid | TID_INITIATOR_STA;
		ieee80211_start_tx_ba_cb_irqsafe(vif, sta->addr, tid);
		priv->tid_info[val].tid_state = TID_STATE_AGGR_START;
		priv->tid_info[val].ssn = *ssn;
		}
		break;
	case IEEE80211_AMPDU_TX_STOP_FLUSH:
	case IEEE80211_AMPDU_TX_STOP_FLUSH_CONT:
	case IEEE80211_AMPDU_TX_STOP_CONT:
		{
		val = tid | TID_INITIATOR_STA;
		priv->tid_info[val].tid_state = TID_STATE_AGGR_STOP;
		ieee80211_stop_tx_ba_cb_irqsafe(vif, sta->addr, tid);
		}
		break;
	case IEEE80211_AMPDU_TX_OPERATIONAL:
		{
		val = tid | TID_INITIATOR_STA;
		priv->tid_info[val].tid_state = TID_STATE_AGGR_OPERATIONAL;
		}
		break;
	default:
		pr_err("%s: Invalid command, ignoring\n",
		       __func__);
	}
	return ret;
}


static int set_antenna(struct ieee80211_hw *hw, u32 tx_ant, u32 rx_ant)
{
	struct img_priv *priv = (struct img_priv *)hw->priv;

	/* Maximum no of antenna supported =2 */
	if (!tx_ant || (tx_ant & ~3) || !rx_ant || (rx_ant & ~3))
		return -EINVAL;

	priv->tx_antenna = (tx_ant & 3);

	return 0;
}



/* Needed in case of IBSS to send out probe responses when we are beaconing */
static int tx_last_beacon(struct ieee80211_hw *hw)
{
	struct img_priv *priv = (struct img_priv *)hw->priv;

	return priv->tx_last_beacon;
}
int scan(struct ieee80211_hw *hw,
	 struct ieee80211_vif *vif,
	 struct ieee80211_scan_request *ireq)
{
	struct umac_vif *uvif = (struct umac_vif *)vif->drv_priv;
	struct scan_req scan_req = {0};
	int i = 0;

	struct cfg80211_scan_request *req;

	req = &ireq->req;

	scan_req.n_ssids = req->n_ssids;
	scan_req.n_channels = req->n_channels;
	scan_req.ie_len = req->ie_len;

	if (wifi->params.hw_scan_status != HW_SCAN_STATUS_NONE)
		return -EBUSY; /* Already in HW SCAN State */

	/* Keep track of HW Scan requests and compeltes */
	wifi->params.hw_scan_status = HW_SCAN_STATUS_PROGRESS;

	if (uvif->priv->params->production_test == 1) {
		/* Drop scan, its just intended for IBSS
		 * and some data traffic
		 */
		if (wifi->params.hw_scan_status != HW_SCAN_STATUS_NONE) {
			ieee80211_scan_completed(uvif->priv->hw, false);
			wifi->params.hw_scan_status = HW_SCAN_STATUS_NONE;
		}

		return 0;
	}

	if (req->ie_len)
		memcpy(scan_req.ie, req->ie, req->ie_len);

	for (i = 0; i < req->n_channels; i++) {
		scan_req.center_freq[i] = req->channels[i]->center_freq;
		scan_req.freq_max_power[i] = req->channels[i]->max_power;
		scan_req.chan_flags[i] = req->channels[i]->flags;
		/* The type of scan comes from mac80211 so its taken care of */
	}

	scan_req.p2p_probe = req->no_cck;

	/* For hostapd scan (40MHz) and scan_type=passive, n_ssids=0
	 * and req->ssids is NULL
	 */
	if (req->n_ssids > 0) {
		for (i = 0; i < req->n_ssids; i++) {
			scan_req.ssids[i].ssid_len = req->ssids[i].ssid_len;
			if (req->ssids[i].ssid_len > 0)
				memcpy(scan_req.ssids[i].ssid,
				       req->ssids[i].ssid,
				       req->ssids[i].ssid_len);
		}
	}

	return rpu_scan(uvif->vif_index, &scan_req);
}


void rpu_scan_complete(void *context,
			       struct rpu_event_scanres *scan_res,
			       unsigned char *skb,
			       unsigned int len)
{
	struct img_priv *priv = (struct img_priv *)context;

	RPU_DEBUG_SCAN("Event Scan Complete from RPU:");
	RPU_DEBUG_SCAN(" More_results: 0, Scan is Completed\n");
	/* There can be a race where we receive remove_interface and
	 * abort the scan(1)
	 * But we get scan_complete from the FW(2), this check will make
	 * sure we are not calling scan_complete when we have already
	 * aborted the scan. Eg: Killing wpa_supplicant in middle of
	 * scanning
	 */
	if (wifi->params.hw_scan_status != HW_SCAN_STATUS_NONE) {
		priv->stats->umac_scan_complete++;
		ieee80211_scan_completed(priv->hw, false);

		/* Keep track of HW Scan requests and compeltes */
		wifi->params.hw_scan_status = HW_SCAN_STATUS_NONE;
	}

}


void cancel_hw_scan(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct umac_vif *uvif = (struct umac_vif *)vif->drv_priv;
	struct img_priv *priv = NULL;

	priv = (struct img_priv *)hw->priv;

	if (wifi->params.hw_scan_status == HW_SCAN_STATUS_PROGRESS) {
		pr_info("Aborting pending scan request...\n");

		priv->scan_abort_done = 0;

		if (rpu_scan_abort(uvif->vif_index))
			return;

		if (!wait_for_scan_abort(priv)) {
			ieee80211_scan_completed(hw, true);
			wifi->params.hw_scan_status = HW_SCAN_STATUS_NONE;
			priv->stats->umac_scan_complete++;
			return;
		}
	}
}


int set_rts_threshold(struct ieee80211_hw *hw,
		      u32 value)
{
	struct img_priv *priv = NULL;

	priv = (struct img_priv *)hw->priv;
	/*if thres>=2347 (default case) hostapd sends down (u32) -1*/
	if (value > 65536)
		priv->rts_threshold = 65536;
	else
		priv->rts_threshold = value;
	return 0;

}


int sta_add(struct ieee80211_hw *hw,
	    struct ieee80211_vif *vif,
	    struct ieee80211_sta *sta)
{
	struct umac_vif *uvif = (struct umac_vif *)vif->drv_priv;
	int i;
	int result = 0;
	struct img_priv *priv = hw->priv;
	struct umac_sta *usta = (struct umac_sta *)sta->drv_priv;
	unsigned int peer_id = 0;

	for (i = 0; i < MAX_PEERS; i++) {
		if (!priv->peers[i]) {
			peer_id = i;
			break;
		}
	}

	if (i == MAX_PEERS) {
		pr_err("Exceeded Max STA limit(%d)\n", MAX_PEERS);
		return -1;
	}

	rcu_assign_pointer(priv->peers[peer_id], sta);
	synchronize_rcu();

	usta->index = peer_id;
	usta->chanctx = uvif->chanctx;
	usta->vif_index = uvif->vif_index;
	return result;
}


int sta_remove(struct ieee80211_hw *hw,
	       struct ieee80211_vif *vif,
	       struct ieee80211_sta *sta)
{
	struct umac_vif *uvif = (struct umac_vif *)vif->drv_priv;
	int i;
	int result = 0;
	struct img_priv *priv = hw->priv;
	struct umac_sta *usta = (struct umac_sta *)sta->drv_priv;
	struct tx_config *tx = &priv->tx;
	u32 hw_queue_map = 0;

	/*purge the queues*/
	for (i = 0; i < NUM_ACS; i++)
		hw_queue_map |= BIT(i);

	spin_lock_bh(&tx->lock);
	RPU_DEBUG_TX("%s:%d discard tx\n", __func__, __LINE__);
	rpu_discard_sta_pend_q(priv, uvif, usta->index, hw_queue_map);
	spin_unlock_bh(&tx->lock);
	priv->tx_deinit_complete = 0;
	rpu_prog_tx_deinit(usta->vif_index, sta->addr);

	if (wait_for_tx_deinit_complete(priv) < 0) {
		WARN_ON(1);
		spin_lock_bh(&tx->lock);
		RPU_DEBUG_TX("%s:%d discarding\n", __func__, __LINE__);
		rpu_discard_sta_tx_q(priv,
					 uvif,
					 usta->index,
					 hw_queue_map,
					 usta->chanctx->index);
		spin_unlock_bh(&tx->lock);
	}

	rcu_assign_pointer(priv->peers[usta->index], NULL);
	synchronize_rcu();

	usta->index = -1;
	return result;
}

int load_fw(struct ieee80211_hw *hw)
{
	int err = 0;
	int i = 0;
	struct img_priv *dev = (struct img_priv *)hw->priv;
	const struct firmware *fw = NULL;

	do {
		err = request_firmware(&fw, bin_name[i], dev->dev);

		if (err) {
			pr_err("get failed %s, err= %d\n", bin_name[i], err);
			break;
		}

		err = fwldr_load_fw(fw->data, fw->size, i);
		if (err == FWLDR_SUCCESS)
			pr_info("%s is loaded\n", bin_name[i]);
		else
			pr_err("Loading of %s failed\n", bin_name[i]);

		release_firmware(fw);

		i++;
	} while ((i < FWLDR_NUM_BINS) && (!err));

	return err;
}

static void channel_switch_beacon(struct ieee80211_hw *hw,
				  struct ieee80211_vif *vif,
				  struct cfg80211_chan_def *chandef)
{
	pr_err("RECEIVED CHANNEL SWITCH BEACON\n");
}


static int umac_chanctx_set_channel(struct img_priv *priv,
				     struct umac_vif *uvif,
				     struct cfg80211_chan_def *chandef)
{
	unsigned int freq_band = 0;
	unsigned int ch_width = 0;
	int center_freq1 = 0;
	int center_freq2 = 0;
	unsigned int pri_chan;
	int err = 0;

	pri_chan = ieee80211_frequency_to_channel(chandef->chan->center_freq);
	center_freq1 = chandef->center_freq1;
	center_freq2 = chandef->center_freq2;

	freq_band = chandef->chan->band;
	ch_width = chandef->width;
	RPU_DEBUG_UMACIF("%s: Primary Channel is: %d\n", __func__, pri_chan);
	err = rpu_prog_channel(pri_chan, center_freq1,
				       center_freq2,
				       ch_width,
				       uvif->vif_index,
				       freq_band);

	if (!err) {
		/* RPU expects to program the associated channel
		 * every time it changes, else it leads to
		 * disconnections.
		 */
		rpu_prog_vif_op_channel(uvif->vif_index,
						uvif->vif->addr,
						pri_chan);
	}

	return err;
}


static int add_chanctx(struct ieee80211_hw *hw,
		       struct ieee80211_chanctx_conf *conf)
{
	struct img_priv *priv = NULL;
	struct umac_chanctx *ctx = NULL;
	int chanctx_id = 0;
	int i = 0;

	priv = hw->priv;

	RPU_DEBUG_UMACIF("GOT add chanctx\n");

	for (i = 0; i < MAX_CHANCTX; i++) {
		if (!priv->chanctx[i]) {
			chanctx_id = i;
			break;
		}
	}

	if (i == MAX_CHANCTX) {
		pr_err("Exceeded Max chan contexts limit(%d)\n", MAX_CHANCTX);
		return -1;
	}

	RPU_DEBUG_UMACIF("%s: %d MHz\n",
			__func__,
			conf->def.chan->center_freq);

	mutex_lock(&priv->mutex);

	ctx = (struct umac_chanctx *)conf->drv_priv;
	ctx->index = chanctx_id;
	INIT_LIST_HEAD(&ctx->vifs);
	ctx->nvifs = 0;

	rcu_assign_pointer(priv->chanctx[i], conf);
	synchronize_rcu();

	mutex_unlock(&priv->mutex);
	return 0;
}


static void remove_chanctx(struct ieee80211_hw *hw,
			   struct ieee80211_chanctx_conf *conf)
{
	struct img_priv *priv = NULL;
	struct umac_chanctx *ctx = NULL;

	priv = hw->priv;
	ctx = (struct umac_chanctx *)conf->drv_priv;
	RPU_DEBUG_UMACIF("GOT remove chanctx\n");

	RPU_DEBUG_UMACIF("%s: %d MHz\n",
					 __func__,
					 conf->def.chan->center_freq);

	mutex_lock(&priv->mutex);

	/* Unassign_vif_chanctx should have been called to free all the assigned
	 * vifs before this call is called, hence we dont need to specifically
	 * free the vifs here
	 */
	rcu_assign_pointer(priv->chanctx[ctx->index], NULL);
	synchronize_rcu();

	ctx->index = -1;

	mutex_unlock(&priv->mutex);
}


static void change_chanctx(struct ieee80211_hw *hw,
			   struct ieee80211_chanctx_conf *conf,
			   u32 changed)
{
	struct umac_vif *uvif = NULL;
	struct img_priv *priv = NULL;
	struct umac_chanctx *ctx = NULL;
	int i = 0;
	int center_freq = 0;
	int chan = 0;
	int err = 0, ret = 0;

	priv = hw->priv;
	ctx = (struct umac_chanctx *)conf->drv_priv;

	RPU_DEBUG_UMACIF("Got change_chanctx: %d\n", changed);
	RPU_DEBUG_UMACIF("%s: %d MHz\n", __func__,
			   conf->def.chan->center_freq);

	if (changed & IEEE80211_CHANCTX_CHANGE_WIDTH ||
	    changed & IEEE80211_CHANCTX_CHANGE_CHANNEL) {
		RPU_DEBUG_UMACIF("%s channel width = %d channel = %d\n",
				   __func__,
				   conf->def.width, conf->def.center_freq1);

		center_freq = conf->def.chan->center_freq;
		chan = ieee80211_frequency_to_channel(center_freq);

		list_for_each_entry(uvif, &ctx->vifs, list) {
			err = umac_chanctx_set_channel(priv, uvif, &conf->def);

			if (err) {
				pr_err("%s: Failed to set channel/width\n",
				       __func__);
				return;
			}
		}
	}

	if (changed & IEEE80211_CHANCTX_CHANGE_MIN_WIDTH) {
		RPU_DEBUG_UMACIF("%s Minimum channel width = %d\n", __func__,
			conf->min_def.width);

		center_freq = conf->min_def.chan->center_freq;
		chan = ieee80211_frequency_to_channel(center_freq);

		list_for_each_entry(uvif, &ctx->vifs, list) {
			err = umac_chanctx_set_channel(priv, uvif,
						       &conf->min_def);

			if (err) {
				pr_err("%s: Failed to set channel/width\n",
					__func__);

				return;
			}
		}
	}

	/* TODO: Make this global config as it effects all VIF's */
	if (changed & IEEE80211_CHANCTX_CHANGE_RX_CHAINS) {
		RPU_DEBUG_UMACIF("%s rx_chains static=%d dynamic=%d\n",
				   __func__,
				   conf->rx_chains_static,
				   conf->rx_chains_dynamic);

		list_for_each_entry(uvif, &ctx->vifs, list) {
			for (i = 0; i < MAX_VIFS; i++) {
				if (!(priv->active_vifs & (1 << i)))
					continue;

				if (priv->vifs[i] &&
				    !ether_addr_equal(priv->vifs[i]->addr,
						      uvif->vif->addr))
					continue;

				if (conf->rx_chains_static > 1)
					CALL_RPU(rpu_prog_vif_smps, i,
						uvif->vif->addr,
						IEEE80211_SMPS_OFF);
				else if (conf->rx_chains_dynamic > 1)
					CALL_RPU(rpu_prog_vif_smps, i,
						uvif->vif->addr,
						IEEE80211_SMPS_DYNAMIC);
				else
					CALL_RPU(rpu_prog_vif_smps, i,
						uvif->vif->addr,
						IEEE80211_SMPS_STATIC);
			}
		}
	}
	if (changed & IEEE80211_CHANCTX_CHANGE_RADAR) {

		RPU_DEBUG_UMACIF("%s radar enabled =%d\n",
				   __func__,
				   conf->radar_enabled);

		if (conf->radar_enabled)
			CALL_RPU(rpu_prog_radar_detect,
				  RADAR_DETECT_OP_START);
		else
			CALL_RPU(rpu_prog_radar_detect,
				  RADAR_DETECT_OP_STOP);
	}
prog_rpu_fail:
	return;
}


static int assign_vif_chanctx(struct ieee80211_hw *hw,
			      struct ieee80211_vif *vif,
			      struct ieee80211_chanctx_conf *conf)
{
	struct img_priv *priv = NULL;
	struct umac_vif *uvif = NULL;
	struct umac_chanctx *ctx = NULL;
	int prog_chanctx_time_info = 0;
	int ret = 0;

	priv = hw->priv;
	uvif = (struct umac_vif *)vif->drv_priv;
	ctx = (struct umac_chanctx *)conf->drv_priv;
	RPU_DEBUG_UMACIF("Got assign_vif_chanctx\n");

	RPU_DEBUG_UMACIF("%s: addr: %pM, type: %d, p2p: %d chan: %d MHz\n",
		  __func__,
		  vif->addr,
		  vif->type,
		  vif->p2p,
		  conf->def.chan->center_freq);

	mutex_lock(&priv->mutex);

	uvif->chanctx = ctx;
	list_add_tail(&uvif->list, &ctx->vifs);

	prog_chanctx_time_info = !(ctx->nvifs);
	ctx->nvifs++;

	/* If this is the first vif being assigned to the channel context then
	 * increment our count of the active channel contexts
	 */
	if (prog_chanctx_time_info) {
		if (!priv->num_active_chanctx)
			priv->curr_chanctx_idx = ctx->index;

		priv->num_active_chanctx++;
		CALL_RPU(rpu_prog_chanctx_time_info);
	}

	ret = umac_chanctx_set_channel(priv, uvif, &conf->def);

prog_rpu_fail:
	mutex_unlock(&priv->mutex);
	return ret;
}


static void unassign_vif_chanctx(struct ieee80211_hw *hw,
				 struct ieee80211_vif *vif,
				 struct ieee80211_chanctx_conf *conf)
{
	struct img_priv *priv = NULL;
	struct umac_vif *uvif = NULL;
	struct umac_chanctx *ctx = NULL;
	u32 hw_queue_map = 0;
	int i = 0, ret = 0;

	priv = hw->priv;
	uvif = (struct umac_vif *)vif->drv_priv;
	ctx = (struct umac_chanctx *)conf->drv_priv;

	RPU_DEBUG_UMACIF("Got unassign_vif_chanctx\n");

	RPU_DEBUG_UMACIF("%s: addr: %pM, type: %d, p2p: %d chan: %d MHz\n",
		  __func__,
		  vif->addr,
		  vif->type,
		  vif->p2p,
		  conf->def.chan->center_freq);

	mutex_lock(&priv->mutex);

	/* We need to specifically handle flushing tx queues for the AP VIF
	 * here (for STA VIF, mac80211 handles this via flush_queues)
	 */
	if (vif->type == NL80211_IFTYPE_AP) {
		/* Flush all queues for this VIF */
		for (i = 0; i < NUM_ACS; i++)
			hw_queue_map |= BIT(i);

		ieee80211_stop_queues(hw);
		rpu_flush_vif_queues(priv,
					 uvif,
					 uvif->chanctx->index,
					 hw_queue_map,
					 RPU_VIF_CHANCTX_TYPE_OPER,
					 TX_DROP);

		ieee80211_wake_queues(hw);
	}

	uvif->chanctx = NULL;

	list_del(&uvif->list);
	ctx->nvifs--;

	if (!ctx->nvifs) {
		priv->num_active_chanctx--;

		if (priv->num_active_chanctx)
			CALL_RPU(rpu_prog_chanctx_time_info);
	}

prog_rpu_fail:
	mutex_unlock(&priv->mutex);
}


static int switch_vif_chanctx(struct ieee80211_hw *hw,
			      struct ieee80211_vif_chanctx_switch *vifs,
			      int n_vifs,
			      enum ieee80211_chanctx_switch_mode mode)
{
	struct img_priv *priv = NULL;
	int ret = 0;

	RPU_DEBUG_UMACIF("Got switch_vif_chanctx\n");
	priv = hw->priv;

	/*TODO*/
	if (n_vifs > 1)
		return -EOPNOTSUPP;

	RPU_DEBUG_UMACIF("%s switch freq %hu->%hu width %d->%d\n",
		 __func__,
		vifs[0].old_ctx->def.chan->center_freq,
		vifs[0].new_ctx->def.chan->center_freq,
		vifs[0].old_ctx->def.width,
		vifs[0].new_ctx->def.width);

	switch (mode) {
	case CHANCTX_SWMODE_SWAP_CONTEXTS:
		unassign_vif_chanctx(hw, vifs[0].vif, vifs[0].old_ctx);
		remove_chanctx(hw, vifs[0].old_ctx);
		add_chanctx(hw, vifs[0].new_ctx);
		assign_vif_chanctx(hw, vifs[0].vif, vifs[0].new_ctx);
		break;
	case CHANCTX_SWMODE_REASSIGN_VIF:
		unassign_vif_chanctx(hw, vifs[0].vif, vifs[0].old_ctx);
		assign_vif_chanctx(hw, vifs[0].vif, vifs[0].new_ctx);
		break;
	default:
		ret = -EOPNOTSUPP;
		break;
	}
	return ret;
}


static void flush_queues(struct ieee80211_hw *hw,
			 struct ieee80211_vif *vif,
			 u32 queues,
			 bool drop)
{
	struct img_priv *priv = NULL;
	struct umac_vif *uvif = NULL;
	u32 hw_queue_map = 0;
	int i = 0;

	priv = hw->priv;

	mutex_lock(&priv->mutex);

	if (!vif)
		goto out;

	uvif = (struct umac_vif *)vif->drv_priv;

	if (!uvif->chanctx)
		goto out;

	/* Convert the mac80211 queue map to our hw queue map */
	for (i = 0; i < IEEE80211_NUM_ACS; i++) {
		if (queues & BIT(i))
			hw_queue_map |= BIT(tx_queue_map(i));
	}
	/* This op should not get called during ROC operation, so we can assume
	 * that the vif_chanctx_type will be RPU_VIF_CHANCTX_TYPE_OPER. As for
	 * TSMC operation the VIF can only be associated to one channel context,
	 * so we pass uvif->chanctx->index as the parameter for chanctx_idx
	 */
	rpu_flush_vif_queues(priv,
				 uvif,
				 uvif->chanctx->index,
				 hw_queue_map,
				 RPU_VIF_CHANCTX_TYPE_OPER,
				 drop ? TX_DROP : TX_FLUSH);

out:
	mutex_unlock(&priv->mutex);
}


static struct ieee80211_ops ops = {
	.tx                 = tx,
	.start              = start,
	.stop               = stop,
	.add_interface      = add_interface,
	.remove_interface   = remove_interface,
	.config             = config,
	.prepare_multicast  = prepare_multicast,
	.configure_filter   = configure_filter,
	.sw_scan_start      = NULL,
	.sw_scan_complete   = NULL,
	.get_stats          = NULL,
	.sta_notify         = NULL,
	.conf_tx            = conf_vif_tx,
	.bss_info_changed   = bss_info_changed,
	.set_tim            = NULL,
	.set_key            = set_key,
	.tx_last_beacon     = tx_last_beacon,
	.ampdu_action       = ampdu_action,
	.set_antenna	    = set_antenna,
	.remain_on_channel = remain_on_channel,
	.cancel_remain_on_channel = cancel_remain_on_channel,
	.hw_scan	    = scan,
	.cancel_hw_scan	    = cancel_hw_scan,
	.set_rekey_data     = NULL,
	.set_rts_threshold  = set_rts_threshold,
	.channel_switch_beacon = channel_switch_beacon,
	.sta_add	    = sta_add,
	.sta_remove	    = sta_remove,
	.add_chanctx              = add_chanctx,
	.remove_chanctx           = remove_chanctx,
	.change_chanctx           = change_chanctx,
	.assign_vif_chanctx       = assign_vif_chanctx,
	.unassign_vif_chanctx     = unassign_vif_chanctx,
	.switch_vif_chanctx       = switch_vif_chanctx,
	.flush			  = flush_queues,
};

void rpu_exit(void)
{
	if (wifi->hw) {
		ieee80211_unregister_hw(wifi->hw);
		ieee80211_free_hw(wifi->hw);
		wifi->hw = NULL;
	}
}

int rpu_init(void)
{
	struct ieee80211_hw *hw;
	int error;
	struct img_priv *priv = NULL;
	int i;

	/* Allocate new hardware device */
	hw = ieee80211_alloc_hw(sizeof(struct img_priv), &ops);

	if (hw == NULL) {
		pr_err("Failed to allocate memory for ieee80211_hw\n");
		error = -ENOMEM;
		goto out;
	}

	priv = (struct img_priv *)hw->priv;
	memset(priv, 0, sizeof(struct img_priv));

	priv->dev = hal_ops.get_dev();

	pr_info("MAC ADDR: %pM\n", vif_macs);
	SET_IEEE80211_DEV(hw, priv->dev);

	mutex_init(&priv->mutex);
	spin_lock_init(&priv->bcast_lock);
	spin_lock_init(&priv->chanctx_lock);

	spin_lock_init(&priv->roc_lock);
	priv->state = STOPPED;
	priv->active_vifs = 0;
	priv->txpower = DEFAULT_TX_POWER;
	priv->tx_antenna = DEFAULT_TX_ANT_SELECT;
	priv->rts_threshold = DEFAULT_RTS_THRESHOLD;
	strncpy(priv->name, RPU_DRIVER_NAME, 11);
	priv->name[11] = '\0';

	for (i = 0; i < wifi->params.num_vifs; i++)
		img_ether_addr_copy(priv->if_mac_addresses[i].addr,
				    vif_macs[i]);

	priv->hw = hw;
	priv->params = &wifi->params;
	priv->stats = &wifi->stats;
	priv->umac_proc_dir_entry = wifi->umac_proc_dir_entry;
	priv->current_vif_count = 0;
	priv->stats->system_rev = system_rev;
	priv->num_active_chanctx = 0;

	for (i = 0; i < MAX_VIFS; i++)
		priv->vifs[i] = NULL;

	/* Initialize HW parameters */
	init_hw(hw);
	/*Register hardware*/
	error = ieee80211_register_hw(hw);

	/* Production test hack: Set all channel flags to 0 to allow IBSS
	 * creation in all channels
	 */
	if (wifi->params.production_test && !error) {
		enum ieee80211_band band;
		struct ieee80211_supported_band *sband;

		for (band = 0; band < IEEE80211_NUM_BANDS; band++) {
			sband = hw->wiphy->bands[band];
			if (sband)
				for (i = 0; i < sband->n_channels; i++)
					sband->channels[i].flags = 0;
		}
	}

	if (!error) {
		wifi->hw = hw;
		rpu_if_init(priv, priv->name);
		goto out;
	} else {
		rpu_exit();
		goto out;
	}

out:
	return error;
}





