/*
 * Copyright (c) 2011-2017 Imagination Technologies Ltd.
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "core.h"

int conv_str_to_byte(unsigned char *byte,
		     unsigned char *str,
		     int len)
{
	int  i, j = 0;
	unsigned char ch, val = 0;

	for (i = 0; i < (len * 2); i++) {
		/*convert to lower*/
		ch = ((str[i] >= 'A' && str[i] <= 'Z') ? str[i] + 32 : str[i]);

		if ((ch < '0' || ch > '9') && (ch < 'a' || ch > 'f'))
			return -1;

		if (ch >= '0' && ch <= '9')  /*check is digit*/
			ch = ch - '0';
		else
			ch = ch - 'a' + 10;

		val += ch;

		if (!(i%2))
			val <<= 4;
		else {
			byte[j] = val;
			j++;
			val = 0;
		}
	}

	return 0;
}

int wait_for_scan_abort(struct img_priv *priv)
{
	int count;

	count = 0;

check_scan_abort_complete:
	if (!priv->scan_abort_done && (count < SCAN_ABORT_TIMEOUT_TICKS)) {
		current->state = TASK_INTERRUPTIBLE;

		if (schedule_timeout(1) == 0)
			count++;

		goto check_scan_abort_complete;
	}

	if (!priv->scan_abort_done) {
		pr_debug("%s-UMAC: No SCAN_ABORT_DONE after %ld ticks\n",
			   priv->name, SCAN_ABORT_TIMEOUT_TICKS);
		return -1;
	}

	RPU_DEBUG_SCAN("%s-UMAC: Scan abort complete after %d timer ticks\n",
					priv->name,
					count);

	return 0;

}

int wait_for_cancel_hw_roc(struct img_priv *priv)
{
	int count = 0;

check_cancel_hw_roc_complete:
	if (!priv->cancel_hw_roc_done &&
	    (count < CANCEL_HW_ROC_TIMEOUT_TICKS)) {
		current->state = TASK_INTERRUPTIBLE;
		if (schedule_timeout(1) == 0)
			count++;
		goto check_cancel_hw_roc_complete;
	}

	if (!priv->cancel_hw_roc_done) {
		pr_err("%s-UMAC: Warning: Didn't get CANCEL_HW_ROC_DONE after %ld timer ticks\n",
		       priv->name,
		       CANCEL_HW_ROC_TIMEOUT_TICKS);
		return -1;
	}

	RPU_DEBUG_ROC("%s-UMAC: Cancel HW RoC complet after %d timer ticks\n",
					priv->name,
					count);

	return 0;

}

int wait_for_channel_prog_complete(struct img_priv *priv)
{
	int count;

	count = 0;

check_ch_prog_complete:
	if (!priv->chan_prog_done && (count < CH_PROG_TIMEOUT_TICKS)) {
		current->state = TASK_INTERRUPTIBLE;

		if (schedule_timeout(1) == 0)
			count++;

		goto check_ch_prog_complete;
	}

	if (!priv->chan_prog_done) {
		pr_debug("%s-UMAC: No channel prog done after %ld ticks\n",
			   priv->name, CH_PROG_TIMEOUT_TICKS);
		return -1;
	}

	pr_debug("%s-UMAC: Channel Prog Complete after %d timer ticks\n",
			priv->name, count);

	return 0;

}

int wait_for_tx_deinit_complete(struct img_priv *priv)
{
	int count = 0;

check_tx_deinit_complete:
	if (!priv->tx_deinit_complete &&
	    (count < TX_DEINIT_TIMEOUT_TICKS)) {
		current->state = TASK_INTERRUPTIBLE;
		if (schedule_timeout(1) == 0)
			count++;
		goto check_tx_deinit_complete;
	}

	if (!priv->tx_deinit_complete) {
		pr_err("%s-UMAC: Warning: Tx discard failed after %ld timer ticks\n",
		       priv->name,
		       TX_DEINIT_TIMEOUT_TICKS);
		return -1;
	}

	pr_debug("Discarded Tx successfully in %d timer ticks\n",
			count);

	return 0;
}

int wait_for_tx_queue_flush_complete(struct img_priv *priv,
				     unsigned int queue)
{
	int count = 0;

check_tx_queue_flush_complete:
	if (priv->tx.outstanding_tokens[queue] &&
	    (count < QUEUE_FLUSH_TIMEOUT_TICKS)) {
		current->state = TASK_INTERRUPTIBLE;
		if (schedule_timeout(1) == 0)
			count++;
		goto check_tx_queue_flush_complete;
	}

	if (priv->tx.outstanding_tokens[queue]) {
		pr_err("%s-UMAC: Warning: Tx Queue %d flush failed pending: %d after %ld timer ticks\n",
		       priv->name,
		       queue,
		       priv->tx.outstanding_tokens[queue],
		       QUEUE_FLUSH_TIMEOUT_TICKS);
		return -1;
	}

	RPU_DEBUG_ROC("%s-UMAC:", priv->name);
	RPU_DEBUG_ROC("Flushed Tx queue %d successfully in %d timer ticks\n",
		       queue,
			   count);

	return 0;

}

int wait_for_reset_complete(struct img_priv *priv)
{
	int count;

	count = 0;

check_reset_complete:
	if (!priv->reset_complete && (count < RESET_TIMEOUT_TICKS)) {
		current->state = TASK_INTERRUPTIBLE;

		if (schedule_timeout(1) == 0)
			count++;

		goto check_reset_complete;
	}

	if (!priv->reset_complete) {
		pr_debug("%s-UMAC: No reset complete after %ld ticks\n",
			   priv->name, RESET_TIMEOUT_TICKS);
		return -1;
	}

	pr_debug("%s-UMAC: Reset complete after %d timer ticks\n",
		   priv->name, count);
	return 0;

}





void update_aux_adc_voltage(struct img_priv *priv,
				   unsigned char pdout)
{
	static unsigned int index;

	if (index > MAX_AUX_ADC_SAMPLES)
		index = 0;

	priv->params->pdout_voltage[index++] = pdout;
}
