/*
 * IMG Cashew USB PHY driver
 *
 * Copyright (C) 2016 Imagination Technologies Ltd.
 *
 * Author: Sifan Naeem <sifan.naeem@imgtec.com>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 */

#include <linux/clk.h>
#include <linux/delay.h>
#include <linux/io.h>
#include <linux/kernel.h>
#include <linux/mfd/syscon.h>
#include <linux/module.h>
#include <linux/of.h>
#include <linux/phy/phy.h>
#include <linux/platform_device.h>
#include <linux/regmap.h>

#define CR_USB_PHY_OFS					0x4

#define CR_USB0_PHY_CLK_CTRL				0x25c
#define CR_USB0_PHY_FSEL_SHIFT		23
#define CR_USB0_PHY_FSEL_MASK			0x3f
#define CR_USB0_PHY_REF_SSP_EN_SHIFT	14
#define CR_USB0_PHY_REF_SSP_EN_MASK		0x1

struct cashew_usb_phy {
	struct device *dev;
	struct regmap *cr_top;
	struct clk *phy_clk;
	unsigned int phy_id;
};

static const unsigned long fsel_rate_map[] = {
	19200000,
	20000000,
	24000000,
	100000000,
};

/* fsel[5:0] values from PHY databook table 3-1 */
static const unsigned long fsel_reg_map[] = {
	0x38,
	0x31,
	0x2A,
	0x27,
};

static int cashew_usb_phy_power_on(struct phy *phy)
{
	struct cashew_usb_phy *p_phy = phy_get_drvdata(phy);
	unsigned long rate;
	unsigned int i, clk_ctrl_ofs;
	int ret;

	clk_ctrl_ofs = CR_USB0_PHY_CLK_CTRL + (p_phy->phy_id * CR_USB_PHY_OFS);

	/*
	 * Disable the PHY reference clock buffer for the USB SuperSpeed PLL,
	 * until the reference clock setup is done
	 */
	regmap_update_bits(p_phy->cr_top, clk_ctrl_ofs,
			   CR_USB0_PHY_REF_SSP_EN_MASK <<
			   CR_USB0_PHY_REF_SSP_EN_SHIFT,
			   0 << CR_USB0_PHY_REF_SSP_EN_SHIFT);

	rate = clk_get_rate(p_phy->phy_clk);

	for (i = 0; i < ARRAY_SIZE(fsel_rate_map); i++) {
		if (rate == fsel_rate_map[i])
			break;
	}

	if (i == ARRAY_SIZE(fsel_rate_map)) {
		dev_err(p_phy->dev, "Unsupported clock rate: %lu\n", rate);
		return -EINVAL;
	}

	regmap_update_bits(p_phy->cr_top, clk_ctrl_ofs,
			   CR_USB0_PHY_FSEL_MASK <<
			   CR_USB0_PHY_FSEL_SHIFT,
			   fsel_reg_map[i] << CR_USB0_PHY_FSEL_SHIFT);

	ret = clk_prepare_enable(p_phy->phy_clk);
	if (ret < 0) {
		dev_err(p_phy->dev, "Failed to enable PHY clock: %d\n", ret);
		return ret;
	}

	regmap_update_bits(p_phy->cr_top, clk_ctrl_ofs,
			   CR_USB0_PHY_REF_SSP_EN_MASK <<
			   CR_USB0_PHY_REF_SSP_EN_SHIFT,
			   1 << CR_USB0_PHY_REF_SSP_EN_SHIFT);

	return 0;
}

static int cashew_usb_phy_power_off(struct phy *phy)
{
	struct cashew_usb_phy *p_phy = phy_get_drvdata(phy);

	clk_disable_unprepare(p_phy->phy_clk);

	return 0;
}

static const struct phy_ops cashew_usb_phy_ops = {
	.power_on = cashew_usb_phy_power_on,
	.power_off = cashew_usb_phy_power_off,
	.owner = THIS_MODULE,
};

static int cashew_usb_phy_probe(struct platform_device *pdev)
{
	struct cashew_usb_phy *p_phy;
	struct phy_provider *provider;
	struct phy *phy;
	int ret;

	p_phy = devm_kzalloc(&pdev->dev, sizeof(*p_phy), GFP_KERNEL);
	if (!p_phy)
		return -ENOMEM;
	p_phy->dev = &pdev->dev;
	platform_set_drvdata(pdev, p_phy);

	p_phy->cr_top = syscon_regmap_lookup_by_phandle(p_phy->dev->of_node,
							"img,cr-top-secure");
	if (IS_ERR(p_phy->cr_top)) {
		dev_err(p_phy->dev, "Failed to get SWITCHED_TOP_REGBANK_SECURE registers: %ld\n",
			PTR_ERR(p_phy->cr_top));
		return PTR_ERR(p_phy->cr_top);
	}

	p_phy->phy_clk = devm_clk_get(p_phy->dev, "usb_phy");
	if (IS_ERR(p_phy->phy_clk)) {
		dev_err(p_phy->dev, "Failed to get usb_phy clock: %ld\n",
			PTR_ERR(p_phy->phy_clk));
		return PTR_ERR(p_phy->phy_clk);
	}

	ret = of_property_read_u32(p_phy->dev->of_node, "img,phy_id",
				   &p_phy->phy_id);
	if (ret < 0) {
		dev_err(p_phy->dev, "No phy_id specified\n");
		return ret;
	}

	phy = devm_phy_create(p_phy->dev, NULL, &cashew_usb_phy_ops);
	if (IS_ERR(phy)) {
		dev_err(p_phy->dev, "Failed to create PHY: %ld\n",
			PTR_ERR(phy));
		return PTR_ERR(phy);
	}
	phy_set_drvdata(phy, p_phy);

	provider = devm_of_phy_provider_register(p_phy->dev,
						 of_phy_simple_xlate);
	if (IS_ERR(provider)) {
		dev_err(p_phy->dev, "Failed to register PHY provider: %ld\n",
			PTR_ERR(provider));
		return PTR_ERR(provider);
	}

	return 0;
}

static const struct of_device_id cashew_usb_phy_of_match[] = {
	{ .compatible = "img,cashew-usb-phy", },
	{ },
};
MODULE_DEVICE_TABLE(of, cashew_usb_phy_of_match);

static struct platform_driver cashew_usb_phy_driver = {
	.probe		= cashew_usb_phy_probe,
	.driver		= {
		.name	= "cashew-usb-phy",
		.of_match_table = cashew_usb_phy_of_match,
	},
};
module_platform_driver(cashew_usb_phy_driver);

MODULE_AUTHOR("Sifan Naeem <sifan.naeem@imgtec.com>");
MODULE_DESCRIPTION("IMG Cashew USB3.0 PHY driver");
MODULE_LICENSE("GPL v2");
