/*
 * Cashew SoC pinctrl driver
 *
 * Copyright (C) 2016 Imagination Technologies Ltd.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 */

#include <linux/err.h>
#include <linux/gpio.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/kernel.h>
#include <linux/of_device.h>
#include <linux/of_irq.h>
#include <linux/pinctrl/pinconf.h>
#include <linux/pinctrl/pinconf-generic.h>
#include <linux/pinctrl/pinctrl.h>
#include <linux/pinctrl/pinmux.h>

#include "core.h"
#include "pinctrl-cashew.h"
#include "pinctrl-utils.h"

#define GPIO_IO_MODE			0x00
#define GPIO_BIT_EN			0x04
#define GPIO_IRQ_POLARITY		0x08
#define GPIO_IRQ_TYPE			0x0c
#define GPIO_IRQ_TYPE_LEVEL	0x0
#define GPIO_IRQ_TYPE_EDGE	0x1
#define GPIO_IRQ_EDGE			0x10
#define GPIO_IRQ_EDGE_SINGLE	0x0
#define GPIO_IRQ_EDGE_DUAL	0x1
#define GPIO_IRQ_STATUS			0x14
#define GPIO_IRQ_MASK			0x18
#define GPIO_IRQ_ENABLE			0x1c
#define GPIO_INPUT			0x20
#define GPIO_OUTPUT			0x24
#define GPIO2_OFFSET_IN			0
#define GPIO2_OFFSET_OUT		0x24
#define GPIO2_OFFSET_OEN		0x48
#define GPIO2_OFFSET_SEL		0x6c

void pwrite(struct img_pinctrl *pctl, u32 reg, u32 val)
{
	writel(val, pctl->base + reg);
}

u32 pread(struct img_pinctrl *pctl, u32 reg)
{
	return readl(pctl->base + reg);
}

static int img_get_groups_count(struct pinctrl_dev *pctldev)
{
	struct img_pinctrl *pctl = pinctrl_dev_get_drvdata(pctldev);
	const struct img_pinctrl_soc_data *data = pctl->data;

	return data->ngroups;
}

static const char *img_get_group_name(struct pinctrl_dev *pctldev,
				      unsigned selector)
{
	struct img_pinctrl *pctl = pinctrl_dev_get_drvdata(pctldev);
	const struct img_pinctrl_soc_data *data = pctl->data;

	return data->groups[selector].name;
}

static int img_get_group_pins(struct pinctrl_dev *pctldev, unsigned selector,
			      unsigned const  **pins,
			      unsigned * const num_pins)
{
	struct img_pinctrl *pctl = pinctrl_dev_get_drvdata(pctldev);
	const struct img_pinctrl_soc_data *data = pctl->data;

	*pins = data->groups[selector].pins;
	*num_pins = data->groups[selector].npins;

	return 0;
}

static struct pinctrl_ops img_pctrl_ops = {
	.get_groups_count = img_get_groups_count,
	.get_group_name = img_get_group_name,
	.get_group_pins = img_get_group_pins,
	.dt_node_to_map =  pinconf_generic_dt_node_to_map_pin,
	.dt_free_map = pinctrl_utils_dt_free_map,
};

static int img_get_functions_count(struct pinctrl_dev *pctldev)
{
	struct img_pinctrl *pctl = pinctrl_dev_get_drvdata(pctldev);
	const struct img_pinctrl_soc_data *data = pctl->data;

	return data->nfunctions;
}

static const char *img_get_fname(struct pinctrl_dev *pctldev, unsigned selector)
{
	struct img_pinctrl *pctl = pinctrl_dev_get_drvdata(pctldev);
	const struct img_pinctrl_soc_data *data = pctl->data;

	return data->functions[selector].name;
}

static int img_get_groups(struct pinctrl_dev *pctldev, unsigned selector,
			  const char * const **groups,
			  unsigned * const num_groups)
{
	struct img_pinctrl *pctl = pinctrl_dev_get_drvdata(pctldev);
	const struct img_pinctrl_soc_data *data = pctl->data;

	*groups = data->functions[selector].groups;
	*num_groups = data->functions[selector].ngroups;

	return 0;
}

static int pin_to_mux_pin(struct img_pinctrl *pctl, unsigned pin)
{
	int i;

	for (i = 0; i < pctl->data->nmuxed_pads; i++) {
		if (pin == pctl->data->muxed_pads[i])
			return i;
	}

	WARN(true, "Pin %u does not support muxing\n", pin);
	return -1;
}

static int img_set_mux(struct pinctrl_dev *pctldev, unsigned selector,
		       unsigned group)
{
	struct img_pinctrl *pctl = pinctrl_dev_get_drvdata(pctldev);
	const struct img_pinctrl_soc_data *data = pctl->data;
	const struct img_group *grp = &data->groups[group];
	int i;

	for (i = 0; i < grp->npins; i++) {
		unsigned pin = grp->pins[i];
		unsigned int mux_id = pin_to_mux_pin(pctl, pin);

		dev_dbg(pctldev->dev, "Muxing pin %s for function %s\n",
			pin_get_name(pctldev, pin),
			img_get_fname(pctldev, selector));

		data->funcs->set_mux(pctl, mux_id, grp->mux_id);
	}

	return 0;
}

static struct pinmux_ops img_pmx_ops = {
	.get_functions_count = img_get_functions_count,
	.get_function_name = img_get_fname,
	.get_function_groups = img_get_groups,
	.set_mux = img_set_mux,
};

static int img_pinconf_get(struct pinctrl_dev *pctldev, unsigned pin,
			   unsigned long *config)
{
	struct img_pinctrl *pctl = pinctrl_dev_get_drvdata(pctldev);
	enum pin_config_param param = pinconf_to_config_param(*config);
	u32 arg;

	arg = pctl->data->funcs->get_pinconf(pctl, pin, param);

	*config = pinconf_to_config_packed(param, arg);

	return 0;
}

static int img_pinconf_set(struct pinctrl_dev *pctldev, unsigned pin,
			   unsigned long *configs, unsigned num_configs)
{
	struct img_pinctrl *pctl = pinctrl_dev_get_drvdata(pctldev);
	enum pin_config_param param;
	u32 arg;
	unsigned int i;

	for (i = 0; i < num_configs; i++) {
		param = pinconf_to_config_param(configs[i]);
		arg = pinconf_to_config_argument(configs[i]);

		pctl->data->funcs->set_pinconf(pctl, pin, param, arg);
	}
	return 0;
}

static const struct pinconf_ops img_pinconf_ops = {
	.pin_config_get = img_pinconf_get,
	.pin_config_set = img_pinconf_set,
	.is_generic = true,
};

static inline struct img_gpio_bank *gc_to_bank(struct gpio_chip *gc)
{
	return container_of(gc, struct img_gpio_bank, gpio_chip);
}

static inline struct img_gpio_bank *irqd_to_bank(struct irq_data *d)
{
	return gc_to_bank(irq_data_get_irq_chip_data(d));
}

static inline u32 gpio_readl(struct img_gpio_bank *bank, u32 reg)
{
	return readl(bank->base + reg);
}

static inline int gpio2_get_bit(struct img_gpio_bank *bank, u32 reg,
				unsigned int offset)
{
	reg += BIT_WORD(offset) * sizeof(u32);

	return !!(readl(bank->base + reg) & BIT_MASK(offset));
}

static inline void gpio_writel(struct img_gpio_bank *bank, u32 val,
			       u32 reg)
{
	writel(val, bank->base + reg);
}

static inline void gpio2_set_bit(struct img_gpio_bank *bank,
				 u32 reg, unsigned int offset, u32 val)
{
	u32 cont_reg;

	reg += BIT_WORD(offset) * sizeof(u32);
	cont_reg = gpio_readl(bank, reg);

	if (val)
		cont_reg |= BIT_MASK(offset);
	else
		cont_reg &= ~BIT_MASK(offset);

	gpio_writel(bank, cont_reg, reg);
}

static inline void gpio_mask_writel(struct img_gpio_bank *bank,
				    u32 reg, unsigned int bit, u32 val)
{
	/*
	 * For most of the GPIO registers, bit 16 + X must be set in order to
	 * write bit X.
	 */
	gpio_writel(bank, (0x10000 | val) << bit, reg);
}

static inline void gpio_enable(struct img_gpio_bank *bank,
			       unsigned offset)
{
	gpio_mask_writel(bank, GPIO_BIT_EN, offset, 1);
}

static inline void gpio_disable(struct img_gpio_bank *bank,
				unsigned offset)
{
	gpio_mask_writel(bank, GPIO_BIT_EN, offset, 0);
}

static int img_gpio2_request(struct gpio_chip *chip, unsigned int offset)
{
	struct img_gpio_bank *bank = gc_to_bank(chip);
	int ret;

	ret = pinctrl_request_gpio(chip->base + offset);
	if (ret)
		return ret;

	/*
	 * TODO: This function must be called from capri_set_mux()
	 * in pinctrl-cashew-capri driver.
	 */

	gpio2_set_bit(bank, GPIO2_OFFSET_SEL, offset, 1);

	return 0;
}

static int img_gpio_request(struct gpio_chip *chip, unsigned offset)
{
	return pinctrl_request_gpio(chip->base + offset);
}

static void img_gpio2_free(struct gpio_chip *chip, unsigned int offset)
{
	struct img_gpio_bank *bank = gc_to_bank(chip);

	gpio2_set_bit(bank, GPIO2_OFFSET_SEL, offset, 0);
	pinctrl_free_gpio(chip->base + offset);
}

static void img_gpio_free(struct gpio_chip *chip, unsigned offset)
{
	pinctrl_free_gpio(chip->base + offset);
}

static int img_gpio2_get_direction(struct gpio_chip *chip, unsigned int offset)
{
	struct img_gpio_bank *bank = gc_to_bank(chip);

	return gpio2_get_bit(bank, GPIO2_OFFSET_OEN, offset);
}

static int img_gpio_get_direction(struct gpio_chip *chip, unsigned offset)
{
	struct img_gpio_bank *bank = gc_to_bank(chip);

	return !(gpio_readl(bank, GPIO_IO_MODE) & BIT(offset));
}

static int img_gpio2_get(struct gpio_chip *chip, unsigned int offset)
{
	struct img_gpio_bank *bank = gc_to_bank(chip);

	return gpio2_get_bit(bank, GPIO2_OFFSET_IN, offset);
}

static int img_gpio_get(struct gpio_chip *chip, unsigned offset)
{
	struct img_gpio_bank *bank = gc_to_bank(chip);
	u32 reg;

	if (img_gpio_get_direction(chip, offset))
		reg = GPIO_INPUT;
	else
		reg = GPIO_OUTPUT;

	return !!(gpio_readl(bank, reg) & BIT(offset));
}

static void img_gpio2_set(struct gpio_chip *chip, unsigned int offset,
			  int value)
{
	struct img_gpio_bank *bank = gc_to_bank(chip);

	gpio2_set_bit(bank, GPIO2_OFFSET_OUT, offset, value);
}

static void img_gpio_set(struct gpio_chip *chip, unsigned offset,
			 int value)
{
	struct img_gpio_bank *bank = gc_to_bank(chip);

	gpio_mask_writel(bank, GPIO_OUTPUT, offset, !!value);
}

static int img_gpio2_direction_input(struct gpio_chip *chip,
				     unsigned int offset)
{
	struct img_gpio_bank *bank = gc_to_bank(chip);

	gpio2_set_bit(bank, GPIO2_OFFSET_OEN, offset, 1);

	return 0;
}

static int img_gpio_direction_input(struct gpio_chip *chip,
				    unsigned offset)
{
	struct img_gpio_bank *bank = gc_to_bank(chip);

	gpio_mask_writel(bank, GPIO_IO_MODE, offset, 0);
	gpio_enable(bank, offset);

	return 0;
}

static int img_gpio2_direction_output(struct gpio_chip *chip,
				      unsigned int offset, int value)
{
	struct img_gpio_bank *bank = gc_to_bank(chip);

	img_gpio2_set(chip, offset, value);
	gpio2_set_bit(bank, GPIO2_OFFSET_OEN, offset, 0);

	return 0;
}

static int img_gpio_direction_output(struct gpio_chip *chip,
				     unsigned offset, int value)
{
	struct img_gpio_bank *bank = gc_to_bank(chip);

	img_gpio_set(chip, offset, value);
	gpio_mask_writel(bank, GPIO_IO_MODE, offset, 1);
	gpio_enable(bank, offset);

	return 0;
}

static void img_gpio_irq_ack(struct irq_data *data)
{
	struct img_gpio_bank *bank = irqd_to_bank(data);

	gpio_writel(bank, 0xffff & ~BIT(data->hwirq), GPIO_IRQ_STATUS);
}

static void img_gpio_irq_mask(struct irq_data *data)
{
	struct img_gpio_bank *bank = irqd_to_bank(data);

	gpio_mask_writel(bank, GPIO_IRQ_MASK, data->hwirq, 0);
}

static void img_gpio_irq_unmask(struct irq_data *data)
{
	struct img_gpio_bank *bank = irqd_to_bank(data);

	gpio_mask_writel(bank, GPIO_IRQ_MASK, data->hwirq, 1);
}

static unsigned int img_gpio_irq_startup(struct irq_data *data)
{
	struct gpio_chip *chip = irq_data_get_irq_chip_data(data);

	img_gpio_direction_input(chip, data->hwirq);
	img_gpio_irq_unmask(data);

	return 0;
}

static int img_gpio_irq_set_type(struct irq_data *data, unsigned int type)
{
	struct img_gpio_bank *bank = irqd_to_bank(data);

	switch (type & IRQ_TYPE_SENSE_MASK) {
	case IRQ_TYPE_EDGE_RISING:
		gpio_mask_writel(bank, GPIO_IRQ_POLARITY, data->hwirq, 1);
		gpio_mask_writel(bank, GPIO_IRQ_TYPE, data->hwirq,
				 GPIO_IRQ_TYPE_EDGE);
		gpio_mask_writel(bank, GPIO_IRQ_EDGE, data->hwirq,
				 GPIO_IRQ_EDGE_SINGLE);
		break;
	case IRQ_TYPE_EDGE_FALLING:
		gpio_mask_writel(bank, GPIO_IRQ_POLARITY, data->hwirq, 0);
		gpio_mask_writel(bank, GPIO_IRQ_TYPE, data->hwirq,
				 GPIO_IRQ_TYPE_EDGE);
		gpio_mask_writel(bank, GPIO_IRQ_EDGE, data->hwirq,
				 GPIO_IRQ_EDGE_SINGLE);
		break;
	case IRQ_TYPE_EDGE_BOTH:
		gpio_mask_writel(bank, GPIO_IRQ_TYPE, data->hwirq,
				 GPIO_IRQ_TYPE_EDGE);
		gpio_mask_writel(bank, GPIO_IRQ_EDGE, data->hwirq,
				 GPIO_IRQ_EDGE_DUAL);
		break;
	case IRQ_TYPE_LEVEL_HIGH:
		gpio_mask_writel(bank, GPIO_IRQ_POLARITY, data->hwirq, 1);
		gpio_mask_writel(bank, GPIO_IRQ_TYPE, data->hwirq,
				 GPIO_IRQ_TYPE_LEVEL);
		break;
	case IRQ_TYPE_LEVEL_LOW:
		gpio_mask_writel(bank, GPIO_IRQ_POLARITY, data->hwirq, 0);
		gpio_mask_writel(bank, GPIO_IRQ_TYPE, data->hwirq,
				 GPIO_IRQ_TYPE_LEVEL);
		break;
	default:
		return -EINVAL;
	}

	if (type & IRQ_TYPE_LEVEL_MASK)
		__irq_set_handler_locked(data->irq, handle_level_irq);
	else
		__irq_set_handler_locked(data->irq, handle_edge_irq);

	return 0;
}

static int img_gpio_irq_set_wake(struct irq_data *data, unsigned int on)
{
	struct img_gpio_bank *bank = irqd_to_bank(data);

	return irq_set_irq_wake(bank->irq, on);
}

static void img_gpio_irq_handler(unsigned int irq, struct irq_desc *desc)
{
	struct gpio_chip *gc = irq_get_handler_data(irq);
	struct img_gpio_bank *bank = gc_to_bank(gc);
	struct irq_chip *chip = irq_get_chip(irq);
	unsigned long pending;
	unsigned int pin;

	chained_irq_enter(chip, desc);
	pending = gpio_readl(bank, GPIO_IRQ_STATUS);
	for_each_set_bit(pin, &pending, 16)
		generic_handle_irq(irq_linear_revmap(gc->irqdomain, pin));
	chained_irq_exit(chip, desc);
}

#define GPIO_BANK(_bank)				\
	{								\
		.gpio_chip = {						\
			.label = "GPIO" #_bank,				\
			.request = img_gpio_request,			\
			.free = img_gpio_free,				\
			.get_direction = img_gpio_get_direction,	\
			.direction_input = img_gpio_direction_input,	\
			.direction_output = img_gpio_direction_output,	\
			.get = img_gpio_get,				\
			.set = img_gpio_set,				\
			.base = -1,					\
			.ngpio = 16,					\
		},							\
		.irq_chip = {						\
			.name = "GPIO" #_bank,				\
			.irq_startup = img_gpio_irq_startup,		\
			.irq_ack = img_gpio_irq_ack,			\
			.irq_mask = img_gpio_irq_mask,			\
			.irq_unmask = img_gpio_irq_unmask,		\
			.irq_set_type = img_gpio_irq_set_type,		\
			.irq_set_wake = img_gpio_irq_set_wake,		\
		},							\
	}

static struct img_gpio_bank img_gpio_banks[] = {
	GPIO_BANK(0),
	GPIO_BANK(1),
	{
		.gpio_chip = {
			.label = "GPIO2",
			.request = img_gpio2_request,
			.free = img_gpio2_free,
			.get_direction = img_gpio2_get_direction,
			.direction_input = img_gpio2_direction_input,
			.direction_output = img_gpio2_direction_output,
			.get = img_gpio2_get,
			.set = img_gpio2_set,
			.base = -1,
			.ngpio = 278,
		},
	},
};

struct pinctrl_desc img_pin_desc = {
	.name = "img-pin-controller",
	.pctlops = &img_pctrl_ops,
	.pmxops = &img_pmx_ops,
	.confops = &img_pinconf_ops,
	.owner = THIS_MODULE,
};

static int img_gpio_register(struct platform_device *pdev,
			     struct img_pinctrl *pctl)
{
	struct device_node *node = pctl->dev->of_node;
	struct img_gpio_bank *bank;
	u32 regs;
	unsigned int i;
	int ret = 0;

	for (i = 0; i < pctl->num_banks; i++) {
		char child_name[sizeof("gpioXX")];
		struct device_node *child;

		snprintf(child_name, sizeof(child_name), "gpio%u", i);
		child = of_get_child_by_name(node, child_name);
		if (!child) {
			dev_err(pctl->dev, "No node for bank %u\n", i);
			ret = -ENODEV;
			goto err;
		}

		if (!of_find_property(child, "gpio-controller", NULL)) {
			dev_err(pctl->dev,
				"No gpio-controller property for bank %u\n", i);
			ret = -ENODEV;
			goto err;
		}

		ret = of_property_read_u32(child, "offset", &regs);
		if (ret < 0) {
			dev_err(pctl->dev, "No offset for bank %u\n", i);
			goto err;
		}

		bank = &pctl->gpio_banks[i];
		bank->pctl = pctl;

		bank->irq = irq_of_parse_and_map(child, 0);
		if (bank->irq < 0) {
			dev_err(pctl->dev, "No IRQ for bank %u: %d\n", i,
				bank->irq);
			ret = bank->irq;
			goto err;
		}

		bank->base = devm_ioremap_wc(&pdev->dev, regs,
					     GPIO_BLOCK_LENGTH);
		if (IS_ERR(bank->base))
			return PTR_ERR(bank->base);

		gpio_writel(bank, 1, GPIO_IRQ_ENABLE);

		bank->gpio_chip.dev = pctl->dev;
		bank->gpio_chip.of_node = child;
		ret = gpiochip_add(&bank->gpio_chip);
		if (ret < 0) {
			dev_err(pctl->dev, "Failed to add GPIO chip %u: %d\n",
				i, ret);
			goto err;
		}

		ret = gpiochip_irqchip_add(&bank->gpio_chip, &bank->irq_chip,
					   0, handle_level_irq, IRQ_TYPE_NONE);
		if (ret < 0) {
			dev_err(pctl->dev, "Failed to add IRQ chip %u: %d\n",
				i, ret);
			gpiochip_remove(&bank->gpio_chip);
			goto err;
		}
		gpiochip_set_chained_irqchip(&bank->gpio_chip, &bank->irq_chip,
					     bank->irq, img_gpio_irq_handler);
	}
	return 0;
err:
	for (; i > 0; i--) {
		bank = &pctl->gpio_banks[i - 1];
		gpiochip_remove(&bank->gpio_chip);
	}
	return ret;
}

int img_pinctrl_probe(struct platform_device *pdev,
		      struct img_pinctrl_soc_data *data,
		      void *soc_ctx)
{
	struct img_pinctrl *pctl;
	struct resource *res;
	int ret;

	pctl = devm_kzalloc(&pdev->dev, sizeof(*pctl), GFP_KERNEL);
	if (!pctl)
		return -ENOMEM;

	pctl->dev = &pdev->dev;
	dev_set_drvdata(&pdev->dev, pctl);

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	pctl->base = devm_ioremap_resource(&pdev->dev, res);
	if (IS_ERR(pctl->base))
		return PTR_ERR(pctl->base);

	pctl->gpio_banks = img_gpio_banks;
	pctl->num_banks = ARRAY_SIZE(img_gpio_banks);
	pctl->data = data;
	pctl->soc_ctx = soc_ctx;

	img_pin_desc.pins = data->pins;
	img_pin_desc.npins = data->npins;

	pctl->pctldev = pinctrl_register(&img_pin_desc, &pdev->dev, pctl);
	if (!pctl->pctldev) {
		dev_err(&pdev->dev, "Failed to register pinctrl device\n");
		return -EINVAL;
	}

	ret = img_gpio_register(pdev, pctl);
	if (ret < 0) {
		pinctrl_unregister(pctl->pctldev);
		return ret;
	}

	return 0;
}

#ifdef CONFIG_PM_SLEEP
int img_pinctrl_suspend(struct device *dev)
{
	struct img_pinctrl *pctl = dev_get_drvdata(dev);
	struct img_gpio_bank *bank;
	struct img_gpio_ctx *ctx;
	int i;

	for (i = 0; i < pctl->num_banks; i++) {
		bank = &pctl->gpio_banks[i];
		ctx = &bank->suspend_ctx;
		ctx->io_mode      = gpio_readl(bank, GPIO_IO_MODE);
		ctx->bit_en       = gpio_readl(bank, GPIO_BIT_EN);
		ctx->irq_mask     = gpio_readl(bank, GPIO_IRQ_MASK);
		ctx->irq_polarity = gpio_readl(bank, GPIO_IRQ_POLARITY);
		ctx->irq_type     = gpio_readl(bank, GPIO_IRQ_TYPE);
		ctx->irq_edge     = gpio_readl(bank, GPIO_IRQ_EDGE);
		ctx->output       = gpio_readl(bank, GPIO_OUTPUT);
	}

	/* Save SoC specific context */
	pctl->data->funcs->save_ctx(pctl, pctl->soc_ctx);
	return 0;
}

int img_pinctrl_resume(struct device *dev)
{
	struct img_pinctrl *pctl = dev_get_drvdata(dev);
	struct img_gpio_bank *bank;
	struct img_gpio_ctx *ctx;
	int i;

	for (i = 0; i < pctl->num_banks; i++) {
		bank = &pctl->gpio_banks[i];
		ctx = &bank->suspend_ctx;
		gpio_writel(bank,
			    0xffff0000 | ctx->io_mode, GPIO_IO_MODE);
		gpio_writel(bank,
			    0xffff0000 | ctx->bit_en, GPIO_BIT_EN);
		gpio_writel(bank,
			    0xffff0000 | ctx->irq_mask, GPIO_IRQ_MASK);
		gpio_writel(bank,
			    0xffff0000 | ctx->irq_polarity, GPIO_IRQ_POLARITY);
		gpio_writel(bank,
			    0xffff0000 | ctx->irq_type, GPIO_IRQ_TYPE);
		gpio_writel(bank,
			    0xffff0000 | ctx->irq_edge, GPIO_IRQ_EDGE);
		gpio_writel(bank,
			    0xffff0000 | ctx->output, GPIO_OUTPUT);
		gpio_writel(bank,
			    1, GPIO_IRQ_ENABLE);
	}

	/* Restore SoC specific context */
	pctl->data->funcs->restore_ctx(pctl, pctl->soc_ctx);
	return 0;
}
#endif /* CONFIG_PM_SLEEP */
