/*
 * Cashew SoC Reset Controller driver
 *
 * Copyright (C) 2015 Imagination Technologies Ltd.
 * Copyright 2016 ELVEES NeoTek JSC
 *
 * Author: Sifan Naeem <sifan.naeem@imgtec.com>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 */

#include <linux/delay.h>
#include <linux/module.h>
#include <linux/of.h>
#include <linux/platform_device.h>
#include <linux/regmap.h>
#include <linux/reset-controller.h>
#include <linux/slab.h>
#include <linux/mfd/syscon.h>

#include <dt-bindings/reset/cashew-resets.h>

/* Peripheral registers */
#define	CASHEW_SOFT_RESET		0

/* Top level registers */
#define	CR_TOP_SOFT_RESET		0x1a4

#define CR_TOP_PERIPH_SOFT_RESET	0x678

struct cashew_reset_data {
	struct reset_controller_dev	rcdev;
	struct regmap			*periph_regs;
	struct regmap			*top_regs;
};

static inline int cashew_reset_shift(unsigned long id)
{
	switch (id) {
	case CASHEW_RESET_TIMER:
	case CASHEW_RESET_ETH_PKT_SNIF:
	case CASHEW_RESET_MDC:
	case CASHEW_RESET_RSHF:
	case CASHEW_RESET_UART0:
	case CASHEW_RESET_UART1:
	case CASHEW_RESET_GPIO1:
	case CASHEW_RESET_GPIO0:
	case CASHEW_RESET_RNG:
	case CASHEW_RESET_ETS:
	case CASHEW_RESET_SPIS:
	case CASHEW_RESET_I2S_OUT0:
	case CASHEW_RESET_I2S_OUT1:
	case CASHEW_RESET_I2S_IN1:
	case CASHEW_RESET_I2S_IN0:
	case CASHEW_RESET_PWM:
	case CASHEW_RESET_I2C7:
	case CASHEW_RESET_I2C6:
	case CASHEW_RESET_I2C5:
	case CASHEW_RESET_I2C4:
	case CASHEW_RESET_I2C3:
	case CASHEW_RESET_I2C2:
	case CASHEW_RESET_I2C1:
	case CASHEW_RESET_I2C0:
	case CASHEW_RESET_SPDIF_IN:
	case CASHEW_RESET_SPDIF:
		return id + 6;
	case CR_TOP_RESET_CAMERA:
	case CR_TOP_RESET_VXE:
	case CR_TOP_RESET_VXD:
	case CR_TOP_RESET_GPU:
	case CR_TOP_RESET_DISPLAY:
	case CR_TOP_RESET_SECURITY:
	case CR_TOP_RESET_HDMI_RX:
	case CR_TOP_RESET_CPU1:
	case CR_TOP_RESET_CPU2:
	case CR_TOP_RESET_CIP2:
	case CR_TOP_RESET_CIP3:
	case CR_TOP_RESET_CIP4:
	case CR_TOP_RESET_CIP5:
		return id - 100;
	case RESET_SDHOST0:
	case RESET_SDHOST1:
	case RESET_SDHOST2:
	case RESET_USB0:
	case RESET_USB1:
	case RESET_ETHERNET:
	case RESET_UFC:
	case RESET_SFC0:
	case RESET_SFC1:
		return id - RESET_PERIPH_BASE;
	default:
		return -1;
	}
}

static int cashew_reset_set(struct reset_controller_dev *rcdev,
			    unsigned long id, unsigned int val)
{
	struct cashew_reset_data *rd;
	u32 mask;
	int shift;

	rd = container_of(rcdev, struct cashew_reset_data, rcdev);
	shift = cashew_reset_shift(id);
	if (shift < 0)
		return shift;
	mask = 1UL << shift;
	val <<= shift;
	if (id <= CASHEW_RESET_MAX)
		regmap_update_bits(rd->periph_regs,
				   CASHEW_SOFT_RESET,
				   mask, val);
	else if (id <= CR_TOP_RESET_MAX)
		regmap_update_bits(rd->top_regs,
				   CR_TOP_SOFT_RESET,
				   mask, val);
	else
		regmap_update_bits(rd->top_regs,
				   CR_TOP_PERIPH_SOFT_RESET,
				   mask, val);

	return 0;
}

static int cashew_reset_assert(struct reset_controller_dev *rcdev,
			       unsigned long id)
{
	return cashew_reset_set(rcdev, id, 1);
}

static int cashew_reset_deassert(struct reset_controller_dev *rcdev,
				 unsigned long id)
{
	return cashew_reset_set(rcdev, id, 0);
}

static int cashew_reset_reset(struct reset_controller_dev *rcdev,
			      unsigned long id)
{
	int rc;
	rc = cashew_reset_assert(rcdev, id);
	if (rc < 0)
		return rc;

	usleep_range(20, 100);

	return cashew_reset_deassert(rcdev, id);
}

static struct reset_control_ops cashew_reset_ops = {
	.assert		= cashew_reset_assert,
	.deassert	= cashew_reset_deassert,
	.reset		= cashew_reset_reset
};

static int cashew_reset_probe(struct platform_device *pdev)
{
	struct cashew_reset_data *rd;
	int ret;
	struct device *dev = &pdev->dev;
	struct device_node *np = pdev->dev.of_node;

	rd = devm_kzalloc(dev, sizeof(*rd), GFP_KERNEL);
	if (!rd)
		return -ENOMEM;

	rd->periph_regs = syscon_regmap_lookup_by_phandle(np, "img,cr-periph");
	if (IS_ERR(rd->periph_regs))
		return PTR_ERR(rd->periph_regs);

	rd->top_regs = syscon_regmap_lookup_by_phandle(np, "img,cr-top-secure");
	if (IS_ERR(rd->top_regs))
		return PTR_ERR(rd->top_regs);

	rd->rcdev.owner = THIS_MODULE;

	/*
	 * Although we only have 48 resets, nr_resets has to be set to
	 * RESET_MAX as devm_reset_control_get() returns an error for reset
	 * requests with a value larger than what nr_resets is set to.
	 */
	rd->rcdev.nr_resets = RESET_MAX + 1;
	rd->rcdev.ops = &cashew_reset_ops;
	rd->rcdev.of_node = np;

	ret = reset_controller_register(&rd->rcdev);
	if (ret)
		return ret;

	return 0;
}

static int cashew_reset_remove(struct platform_device *pdev)
{
	struct cashew_reset_data *data = platform_get_drvdata(pdev);

	reset_controller_unregister(&data->rcdev);

	return 0;
}

static const struct of_device_id cashew_reset_dt_ids[] = {
	 { .compatible = "img,cashew-reset", },
	 { /* sentinel */ },
};
MODULE_DEVICE_TABLE(of, cashew_reset_dt_ids);

static struct platform_driver cashew_reset_driver = {
	.probe	= cashew_reset_probe,
	.remove	= cashew_reset_remove,
	.driver = {
		.name		= "cashew-reset",
		.of_match_table	= cashew_reset_dt_ids,
	},
};
module_platform_driver(cashew_reset_driver);

MODULE_AUTHOR("Sifan Naeem <sifan.naeem@imgtec.com>");
MODULE_DESCRIPTION("Cashew Reset Controller Driver");
MODULE_LICENSE("GPL v2");
