/*
 * ImgTec PowerDown Controller (PDC) RTC
 *
 * Copyright 2010-2016 Imagination Technologies Ltd.
 *
 */

#include <linux/clk.h>
#include <linux/init.h>
#include <linux/rtc.h>
#include <linux/io.h>
#include <linux/math64.h>
#include <linux/module.h>
#include <linux/notifier.h>
#include <linux/of.h>
#include <linux/platform_device.h>
#include <linux/spinlock.h>
#include <linux/slab.h>

#define HW_EPOCH			2000
#define HW_YEARS			100
#define LIN_EPOCH			1900
#define HW2LIN_EPOCH			(HW_EPOCH - LIN_EPOCH)
#define LIN2HW_EPOCH			(LIN_EPOCH - HW_EPOCH)

/* Registers */
#define PDC_RTC_CONTROL			0x00
#define PDC_RTC_SEC			0x04
#define PDC_RTC_MIN			0x08
#define PDC_RTC_HOUR			0x0c
#define PDC_RTC_DAY			0x10
#define PDC_RTC_MON			0x14
#define PDC_RTC_YEAR			0x18
#define PDC_RTC_ASEC			0x1c
#define PDC_RTC_AMIN			0x20
#define PDC_RTC_AHOUR			0x24
#define PDC_RTC_ADAY			0x28
#define PDC_RTC_AMON			0x2c
#define PDC_RTC_AYEAR			0x30
#define PDC_RTC_IRQ_STATUS		0x34
#define PDC_RTC_IRQ_CLEAR		0x38
#define PDC_RTC_IRQ_EN			0x3c
#define PDC_RTC_CLK_GEN_DIV		0x40

/* Register field masks */
#define PDC_RTC_CONTROL_GAE		0x08	/* global alarm enable */
#define PDC_RTC_CONTROL_FAST		0x04
#define PDC_RTC_CONTROL_UPDATE		0x02
#define PDC_RTC_CONTROL_CE		0x01	/* clock enable */
#define PDC_RTC_SEC_SEC			0x3f
#define PDC_RTC_MIN_MIN			0x3f
#define PDC_RTC_HOUR_HOUR		0x1f
#define PDC_RTC_DAY_DAY			0x1f
#define PDC_RTC_MON_MON			0x0f
#define PDC_RTC_YEAR_YEAR		0x7f
#define PDC_RTC_ASEC_EN			0x40
#define PDC_RTC_ASEC_ASEC		0x3f
#define PDC_RTC_AMIN_EN			0x40
#define PDC_RTC_AMIN_AMIN		0x3f
#define PDC_RTC_AHOUR_EN		0x20
#define PDC_RTC_AHOUR_AHOUR		0x1f
#define PDC_RTC_ADAY_EN			0x20
#define PDC_RTC_ADAY_ADAY		0x1f
#define PDC_RTC_AMON_EN			0x10
#define PDC_RTC_AMON_AMON		0x0f
#define PDC_RTC_AYEAR_EN		0x80
#define PDC_RTC_AYEAR_AYEAR		0x7f
#define PDC_RTC_IRQ_ALARM		0x04
#define PDC_RTC_IRQ_MIN			0x02
#define PDC_RTC_IRQ_SEC			0x01

/* Register field shifts */
#define PDC_RTC_CLK_DIV_32KHZ_SHIFT	0
#define PDC_RTC_CLK_DIV_NEG_EN_SHIFT	7
#define PDC_RTC_CLK_DIV_2HZ_SHIFT	8

/**
 * struct pdc_rtc_priv - Private PDC RTC data.
 * @rtc_dev:		RTC device structure.
 * @dev:		Platform device (used for dev_dbg messages etc).
 * @clk:		RTC core clock.
 * @irq:		IRQ number of RTC device.
 * @reg_base:		Base of registers memory.
 * @time_set_delay:	Number of seconds it takes to set the time.
 * @alarm_irq_delay:	Number of seconds the alarm IRQ is delayed.
 * @clk_nb:		Notifier block for clock notify events.
 * @alarm_pending:	Whether an alarm has fired and hasn't been handled.
 * @lock:		Protects PDC_RTC_CONTROL, control_reg, and
 *			softalrm_sec.
 * @control_reg:	Back up of PDC_RTC_CONTROL to work around buggy
 *			hardware. It takes time for some last written values to
 *			take effect, and later writes can replace a write that
 *			hasn't taken effect yet.
 * @softalrm_sec:	Software alarm second, for emulating alarms which need
 *			to fire very soon, which would otherwise foil our work
 *			around for late alarm interrupts.
 * @hardalrm_offset:	Offset from desired alarm time given to hardware.
 * @hardstop_time:	Time the alarm was stopped (or possibly second before).
 * @alrm_time:		Time of current alarm, for filtering out delayed
 *			cancelled alarm interrupts.
 * @adj_alrm_time:	Time of current adjusted alarm (for wrong clock rate).
 * @time_update_diff:	Difference between the old time and the new time, which
 *			should be used while hardware is still updating the
 *			time.
 * @suspended:		Whether the device is in suspend mode, in which case rtc
 *			interrupt events should be postponed until resume (see
 *			postponed_rtc_int).
 * @wakeup:		Whether the device can wake the system from a sleep
 *			state.
 * @postponed_rtc_int:	Postponed rtc interrupt flags to submit on resume.
 * @last_irq_en:	Preserved IRQ enable state when wakeup is in use.
 */
struct pdc_rtc_priv {
	struct rtc_device *rtc_dev;
	struct device *dev;
	struct clk *clk;
	int irq;
	void __iomem *reg_base;
	unsigned int time_set_delay;
	unsigned int alarm_irq_delay;
	struct notifier_block clk_nb;

	int alarm_pending;
	spinlock_t lock;
	u32 control_reg;
	int softalrm_sec;
	int hardalrm_offset;
	unsigned long hardstop_time;
	unsigned long alrm_time;
	unsigned long adj_alrm_time;
	unsigned long time_update_diff;

	/* suspend data */
	bool suspended;
	bool wakeup;
	unsigned int postponed_rtc_int;
	unsigned int last_irq_en;
};

static void pdc_rtc_write(struct pdc_rtc_priv *priv,
			  unsigned int reg_offs, unsigned int data)
{
	iowrite32(data, priv->reg_base + reg_offs);
}

static unsigned int pdc_rtc_read(struct pdc_rtc_priv *priv,
				 unsigned int reg_offs)
{
	return ioread32(priv->reg_base + reg_offs);
}

/* caller must hold lock. does not handle read during time update. */
static void _pdc_rtc_read_time_raw(struct pdc_rtc_priv *priv,
				   struct rtc_time *tm, int *updating)
{
	int min, upd = 0;

	/*
	 * If it takes time for the time to get set and an update is in
	 * progress, we need to check that the update is still in progress
	 * afterwards otherwise it will have changed while we were reading it.
	 */
	if (priv->time_set_delay)
		upd = pdc_rtc_read(priv, PDC_RTC_CONTROL)
			& PDC_RTC_CONTROL_UPDATE;
start_again:

	/*
	 * We re-read the minute at the end of the loop to check it hasn't
	 * changed while we were reading the others. If it has then we didn't
	 * read atomically and should try again. The second is allowed to
	 * change by itself as that won't result in an inconsistent time.
	 */
	min = pdc_rtc_read(priv, PDC_RTC_MIN) & PDC_RTC_MIN_MIN;
	do {
		tm->tm_sec  = (pdc_rtc_read(priv, PDC_RTC_SEC) &
			       PDC_RTC_SEC_SEC);
		tm->tm_min  = min;
		tm->tm_hour = (pdc_rtc_read(priv, PDC_RTC_HOUR) &
			       PDC_RTC_HOUR_HOUR);
		tm->tm_mday = (pdc_rtc_read(priv, PDC_RTC_DAY) &
			       PDC_RTC_DAY_DAY);
		tm->tm_mon  = (pdc_rtc_read(priv, PDC_RTC_MON) &
			       PDC_RTC_MON_MON) - 1;
		tm->tm_year = (pdc_rtc_read(priv, PDC_RTC_YEAR) &
			       PDC_RTC_YEAR_YEAR) + HW2LIN_EPOCH;
		if (upd) {
			upd = pdc_rtc_read(priv, PDC_RTC_CONTROL)
				& PDC_RTC_CONTROL_UPDATE;
			/* did the update finish while we were reading */
			if (!upd)
				goto start_again;
		}
		min = pdc_rtc_read(priv, PDC_RTC_MIN) & PDC_RTC_MIN_MIN;
		if (min != tm->tm_min)
			dev_dbg(priv->dev,
				"time read %02d:%02d:%02d nonatomic, retrying\n",
				tm->tm_hour, tm->tm_min, tm->tm_sec);
	} while (min != tm->tm_min);

	if (updating)
		*updating = upd;
}

/*
 * caller must hold lock.
 * does handle read during time update.
 * does not initialise tm->wday, tm->yday, or tm->tm_isdst
 */
static void _pdc_rtc_read_time(struct pdc_rtc_priv *priv, struct rtc_time *tm)
{
	int upd;
	unsigned long time;

	_pdc_rtc_read_time_raw(priv, tm, &upd);

	dev_dbg(priv->dev, "time read %02d:%02d:%02d\n",
		tm->tm_hour, tm->tm_min, tm->tm_sec);

	/* if we got the old time during an update, add the difference */
	if (upd && priv->time_update_diff) {
		rtc_tm_to_time(tm, &time);
		time += priv->time_update_diff;
		rtc_time_to_tm(time, tm);

		dev_dbg(priv->dev,
			"update was in progress, adjusting to %02d:%02d:%02d\n",
			tm->tm_hour, tm->tm_min, tm->tm_sec);
	}
}

static int pdc_rtc_read_time(struct device *dev, struct rtc_time *tm)
{
	struct pdc_rtc_priv *priv = dev_get_drvdata(dev);
	unsigned long flags;

	spin_lock_irqsave(&priv->lock, flags);
	_pdc_rtc_read_time(priv, tm);
	spin_unlock_irqrestore(&priv->lock, flags);

	tm->tm_wday  = -1;
	tm->tm_yday  = -1;
	tm->tm_isdst = -1;

	return 0;
}

static int pdc_rtc_set_time(struct device *dev, struct rtc_time *tm)
{
	struct pdc_rtc_priv *priv = dev_get_drvdata(dev);
	unsigned int hw_year;
	unsigned int ctrl;
	unsigned long flags;
	unsigned long time = 0;
	unsigned long rtime;
	struct rtc_time tm_adj;

	dev_dbg(priv->dev, "time set %02d:%02d:%02d\n",
		tm->tm_hour, tm->tm_min, tm->tm_sec);
	/*
	 * Due to a hardware quirk the time may only be set after several
	 * seconds.
	 */
	if (priv->time_set_delay) {
		rtc_tm_to_time(tm, &time);
		rtc_time_to_tm(time + priv->time_set_delay, &tm_adj);
		tm = &tm_adj;
	}

	hw_year = tm->tm_year + LIN2HW_EPOCH;
	/* year must be in range */
	if (hw_year >= HW_YEARS)
		return -EINVAL;

	spin_lock_irqsave(&priv->lock, flags);

	/* write out the values */
	pdc_rtc_write(priv, PDC_RTC_SEC, tm->tm_sec);
	pdc_rtc_write(priv, PDC_RTC_MIN, tm->tm_min);
	pdc_rtc_write(priv, PDC_RTC_HOUR, tm->tm_hour);
	pdc_rtc_write(priv, PDC_RTC_DAY, tm->tm_mday);
	pdc_rtc_write(priv, PDC_RTC_MON, tm->tm_mon + 1);
	pdc_rtc_write(priv, PDC_RTC_YEAR, hw_year);

	/* update the clock with the written values */
	ctrl = priv->control_reg | PDC_RTC_CONTROL_UPDATE;
	pdc_rtc_write(priv, PDC_RTC_CONTROL, ctrl);

	/*
	 * Record the offset of the new time so that we can calculate the
	 * current time before the update is complete.
	 */
	if (priv->time_set_delay) {
		_pdc_rtc_read_time_raw(priv, &tm_adj, NULL);
		rtc_tm_to_time(&tm_adj, &rtime);
		priv->time_update_diff = time - rtime;
	}
	spin_unlock_irqrestore(&priv->lock, flags);

	return 0;
}

static int pdc_rtc_alarm_enabled(struct pdc_rtc_priv *priv)
{
	return !!(priv->control_reg & PDC_RTC_CONTROL_GAE);
}

static int pdc_rtc_read_alarm(struct device *dev, struct rtc_wkalrm *alrm)
{
	struct pdc_rtc_priv *priv = dev_get_drvdata(dev);
	unsigned long flags;
	unsigned long scheduled;
	int offset;

	spin_lock_irqsave(&priv->lock, flags);

	offset = priv->hardalrm_offset;

	/* Just get the register values */
	alrm->enabled = pdc_rtc_alarm_enabled(priv);
	alrm->pending = priv->alarm_pending;

	alrm->time.tm_sec  = pdc_rtc_read(priv, PDC_RTC_ASEC);
	alrm->time.tm_min  = pdc_rtc_read(priv, PDC_RTC_AMIN);
	alrm->time.tm_hour = pdc_rtc_read(priv, PDC_RTC_AHOUR);
	alrm->time.tm_mday = pdc_rtc_read(priv, PDC_RTC_ADAY);
	alrm->time.tm_mon  = pdc_rtc_read(priv, PDC_RTC_AMON);
	alrm->time.tm_year = pdc_rtc_read(priv, PDC_RTC_AYEAR);

	spin_unlock_irqrestore(&priv->lock, flags);

	/* Any misisng _EN bit translated to -1 */

	if (alrm->time.tm_sec & PDC_RTC_ASEC_EN)
		alrm->time.tm_sec &= PDC_RTC_ASEC_ASEC;
	else
		alrm->time.tm_sec = -1;

	if (alrm->time.tm_min & PDC_RTC_AMIN_EN)
		alrm->time.tm_min &= PDC_RTC_AMIN_AMIN;
	else
		alrm->time.tm_min = -1;

	if (alrm->time.tm_hour & PDC_RTC_AHOUR_EN)
		alrm->time.tm_hour &= PDC_RTC_AHOUR_AHOUR;
	else
		alrm->time.tm_hour = -1;

	if (alrm->time.tm_mday & PDC_RTC_ADAY_EN)
		alrm->time.tm_mday &= PDC_RTC_ADAY_ADAY;
	else
		alrm->time.tm_mday = -1;

	if (alrm->time.tm_wday & PDC_RTC_ADAY_EN)
		alrm->time.tm_wday &= PDC_RTC_ADAY_ADAY;
	else
		alrm->time.tm_wday = -1;

	if (alrm->time.tm_mon & PDC_RTC_AMON_EN)
		alrm->time.tm_mon = (alrm->time.tm_mon & PDC_RTC_AMON_AMON) - 1;
	else
		alrm->time.tm_mon = -1;

	if (alrm->time.tm_year & PDC_RTC_AYEAR_EN)
		alrm->time.tm_year = (alrm->time.tm_year & PDC_RTC_AYEAR_AYEAR)
					+ HW2LIN_EPOCH;
	else
		alrm->time.tm_year = -1;

	alrm->time.tm_wday  = -1;
	alrm->time.tm_yday  = -1;
	alrm->time.tm_isdst = -1;

	/*
	 * The alarm time in the hardware is offset to compensate for the late
	 * alarm interrupts, so we need to adjust it to get the original alarm.
	 */

	rtc_tm_to_time(&alrm->time, &scheduled);
	scheduled -= offset;
	rtc_time_to_tm(scheduled, &alrm->time);

	return 0;
}

/* caller must hold priv->lock */
static void _pdc_rtc_stop_alarm(struct pdc_rtc_priv *priv, unsigned long now)
{
	/* disable the secondly interrupt */
	pdc_rtc_write(priv, PDC_RTC_IRQ_EN, PDC_RTC_IRQ_ALARM);
	priv->softalrm_sec = -1;
	/* disable the global alarm enable bit */
	priv->control_reg &= ~PDC_RTC_CONTROL_GAE;
	pdc_rtc_write(priv, PDC_RTC_CONTROL, priv->control_reg);

	priv->hardstop_time = now;
}

static void pdc_rtc_stop_alarm(struct pdc_rtc_priv *priv)
{
	unsigned long flags;
	struct rtc_time tm;
	unsigned long now = 0;

	spin_lock_irqsave(&priv->lock, flags);
	if (priv->alarm_irq_delay) {
		_pdc_rtc_read_time(priv, &tm);
		rtc_tm_to_time(&tm, &now);
	}
	_pdc_rtc_stop_alarm(priv, now);
	spin_unlock_irqrestore(&priv->lock, flags);
}

static void pdc_rtc_start_alarm(struct pdc_rtc_priv *priv)
{
	unsigned long flags;

	spin_lock_irqsave(&priv->lock, flags);
	priv->control_reg |= PDC_RTC_CONTROL_GAE;
	pdc_rtc_write(priv, PDC_RTC_CONTROL, priv->control_reg);
	spin_unlock_irqrestore(&priv->lock, flags);
}

static int _pdc_rtc_set_alarm(struct pdc_rtc_priv *priv, bool temporary,
			      unsigned long scheduled, struct rtc_wkalrm *alrm)
{
	struct rtc_time tm;
	unsigned long flags;
	unsigned long now, adjusted;
	struct rtc_wkalrm alrm_adj, *alrm_orig = alrm;

	spin_lock_irqsave(&priv->lock, flags);
	/*
	 * If we're compensating for hardware bugs, don't change the stored
	 * alarm time.
	 */
	if (!temporary)
		priv->alrm_time = scheduled;
	/*
	 * But still record the alarm being set so we can tell whether we need
	 * to change it back again.
	 */
	priv->adj_alrm_time = scheduled;

	/*
	 * Due to a hardware quirk the alarm may fire several seconds late, so
	 * rewind the scheduled alarm time to compensate.
	 */
	if (priv->alarm_irq_delay) {
		alrm_adj = *alrm;

		_pdc_rtc_read_time(priv, &tm);
		rtc_tm_to_time(&tm, &now);

try_again_locked:
		adjusted = scheduled - priv->alarm_irq_delay;

		/* disable the alarm while we set it up */
		_pdc_rtc_stop_alarm(priv, now);

		/* Make sure we're not setting the alarm in the past */
		if (scheduled <= now) {
			spin_unlock_irqrestore(&priv->lock, flags);
			return -ETIME;
		}
		/* If adjusted time is in past, emulate with a soft alarm */
		if (adjusted <= now && alrm->enabled) {
			/* clear and enable secondly interrupt */
			pdc_rtc_write(priv, PDC_RTC_IRQ_CLEAR, PDC_RTC_IRQ_SEC);
			pdc_rtc_write(priv, PDC_RTC_IRQ_EN,
				      PDC_RTC_IRQ_ALARM | PDC_RTC_IRQ_SEC);
			priv->softalrm_sec = alrm_orig->time.tm_sec;
			/* still set the real alarm as early as possible */
			adjusted = now + 1;
		}
		priv->hardalrm_offset = adjusted - scheduled;
		spin_unlock_irqrestore(&priv->lock, flags);

		rtc_time_to_tm(adjusted, &alrm_adj.time);
		alrm = &alrm_adj;

		dev_dbg(priv->dev, "alarm setting %02d:%02d:%02d\n",
			alrm->time.tm_hour, alrm->time.tm_min,
			alrm->time.tm_sec);
	} else {
		/* disable the alarm while we set it up */
		_pdc_rtc_stop_alarm(priv, 0);

		adjusted = scheduled;
		priv->hardalrm_offset = 0;
		spin_unlock_irqrestore(&priv->lock, flags);
	}

	/*
	 * don't use fields set to -1
	 * all smaller fields than a field in use must also be in use
	 */

	if (alrm->time.tm_year >= 0) {
		tm.tm_year = alrm->time.tm_year + LIN2HW_EPOCH;
		/* year must be in range */
		if ((unsigned int)tm.tm_year > HW_YEARS)
			return -EINVAL;
		tm.tm_year |= PDC_RTC_AYEAR_EN;
	} else
		tm.tm_year = 0;

	if (alrm->time.tm_mon >= 0) {
		tm.tm_mon = (alrm->time.tm_mon + 1) | PDC_RTC_AMON_EN;
	} else {
		if (tm.tm_year & PDC_RTC_AYEAR_EN)
			return -EINVAL;
		tm.tm_mon = 0;
	}

	if (alrm->time.tm_mday >= 0) {
		tm.tm_mday = alrm->time.tm_mday | PDC_RTC_ADAY_EN;
	} else {
		if (tm.tm_mon & PDC_RTC_AMON_EN)
			return -EINVAL;
		tm.tm_mday = 0;
	}

	if (alrm->time.tm_hour >= 0) {
		tm.tm_hour = alrm->time.tm_hour | PDC_RTC_AHOUR_EN;
	} else {
		if (tm.tm_mday & PDC_RTC_ADAY_EN)
			return -EINVAL;
		tm.tm_hour = 0;
	}

	if (alrm->time.tm_min >= 0) {
		tm.tm_min = alrm->time.tm_min | PDC_RTC_AMIN_EN;
	} else {
		if (tm.tm_hour & PDC_RTC_AHOUR_EN)
			return -EINVAL;
		tm.tm_min = 0;
	}

	if (alrm->time.tm_sec >= 0) {
		tm.tm_sec = alrm->time.tm_sec | PDC_RTC_ASEC_EN;
	} else {
		if (tm.tm_min & PDC_RTC_AMIN_EN)
			return -EINVAL;
		tm.tm_sec = 0;
	}

	spin_lock_irqsave(&priv->lock, flags);
	pdc_rtc_write(priv, PDC_RTC_ASEC,  tm.tm_sec);
	pdc_rtc_write(priv, PDC_RTC_AMIN,  tm.tm_min);
	pdc_rtc_write(priv, PDC_RTC_AHOUR, tm.tm_hour);
	pdc_rtc_write(priv, PDC_RTC_ADAY,  tm.tm_mday);
	pdc_rtc_write(priv, PDC_RTC_AMON,  tm.tm_mon);
	pdc_rtc_write(priv, PDC_RTC_AYEAR, tm.tm_year);
	priv->alarm_pending = 0;
	spin_unlock_irqrestore(&priv->lock, flags);

	/* re-enable the alarm if applicable */
	if (alrm->enabled) {
		pdc_rtc_start_alarm(priv);

		/*
		 * Check that setting the alarm didn't lose the race against the
		 * next clock tick (which may be the one we're trying to set the
		 * alarm on). It may be that the interrupt just hasn't been
		 * handled yet (e.g. on another CPU), but it does no harm to
		 * handle it here instead.
		 */
		spin_lock_irqsave(&priv->lock, flags);
		if (!priv->alarm_pending) {
			_pdc_rtc_read_time(priv, &tm);
			rtc_tm_to_time(&tm, &now);
			/* If it's too late, immediately trigger the alarm */
			if (scheduled <= now) {
				_pdc_rtc_stop_alarm(priv, now);
				dev_dbg(priv->dev,
					"alarm set race lost, triggering immediately\n");
				spin_unlock_irqrestore(&priv->lock, flags);
				return -ETIME;
			}
			/*
			 * If we've missed the window of oportunity to set the
			 * alarm interrupt, we need to reconsider.
			 */
			if (adjusted <= now) {
				dev_dbg(priv->dev,
					"alarm set race lost, retrying\n");
				goto try_again_locked;
			}
		}
		spin_unlock_irqrestore(&priv->lock, flags);
	}

	return 0;
}

static int pdc_rtc_set_alarm(struct device *dev, struct rtc_wkalrm *alrm)
{
	struct pdc_rtc_priv *priv = dev_get_drvdata(dev);
	unsigned long scheduled;

	dev_dbg(priv->dev, "alarm set %02d:%02d:%02d\n",
		alrm->time.tm_hour, alrm->time.tm_min, alrm->time.tm_sec);

	rtc_tm_to_time(&alrm->time, &scheduled);
	return _pdc_rtc_set_alarm(priv, false, scheduled, alrm);
}

static int pdc_rtc_alarm_irq_enable(struct device *dev, unsigned int enabled)
{
	struct pdc_rtc_priv *priv = dev_get_drvdata(dev);

	dev_dbg(priv->dev, "alarm irq enable %d\n", enabled);

	if (enabled)
		pdc_rtc_start_alarm(priv);
	else
		pdc_rtc_stop_alarm(priv);

	return 0;
}

static struct rtc_class_ops pdc_rtc_ops = {
	.read_time		= pdc_rtc_read_time,
	.set_time		= pdc_rtc_set_time,
	.read_alarm		= pdc_rtc_read_alarm,
	.set_alarm		= pdc_rtc_set_alarm,
	.alarm_irq_enable	= pdc_rtc_alarm_irq_enable,
};

static irqreturn_t pdc_rtc_isr(int irq, void *dev_id)
{
	struct pdc_rtc_priv *priv = dev_id;
	unsigned int status;
	unsigned long events = RTC_IRQF;
	unsigned long flags;
	struct rtc_time tm;
	unsigned long now = 0;
	u32 sec;

	spin_lock_irqsave(&priv->lock, flags);
	status = pdc_rtc_read(priv, PDC_RTC_IRQ_STATUS);
	pdc_rtc_write(priv, PDC_RTC_IRQ_CLEAR, status);

	/* ignore delayed alarm interrupts after turned alarm off */
	if (pdc_rtc_alarm_enabled(priv)) {
		if (status & PDC_RTC_IRQ_ALARM) {
			/*
			 * Alarm interrupt.
			 * If an alarm was cancelled, we may still get the
			 * delayed interrupt, so we need to check the current
			 * time has actually exceeded the alarm time.
			 * Of course if we've corrected the time we may also get
			 * a legitimate interrupt early, so we use hardstop_time
			 * to check whether the alarm could be the second after
			 * we've disabled it, which would indicate an echo.
			 */
			_pdc_rtc_read_time(priv, &tm);
			rtc_tm_to_time(&tm, &now);
			if (now >= priv->alrm_time) {
				events |= RTC_AF;
				dev_dbg(priv->dev, "isr alarm %02d:%02d:%02d\n",
					tm.tm_hour, tm.tm_min, tm.tm_sec);
			} else if (now > priv->hardstop_time &&
				   now <= priv->hardstop_time + 2) {
				dev_dbg(priv->dev,
					"isr alarm %02d:%02d:%02d ignored (echo %lu disabled %lu)\n",
					tm.tm_hour, tm.tm_min, tm.tm_sec,
					now, priv->hardstop_time);
			} else {
				/*
				 * It looks like a legitimate interrupt since we
				 * haven't just cancelled an alarm.
				 */
				events |= RTC_AF;
				dev_dbg(priv->dev,
					"isr alarm %02d:%02d:%02d (early %lu < %lu)\n",
					tm.tm_hour, tm.tm_min, tm.tm_sec,
					now, priv->alrm_time);
			}
		} else if (status & PDC_RTC_IRQ_SEC) {
			/*
			 * Secondly interrupt.
			 * Check if the current second matches the soft alarm.
			 */
			sec = pdc_rtc_read(priv, PDC_RTC_SEC);
			if (priv->softalrm_sec == sec)
				events |= RTC_AF;
			dev_dbg(priv->dev,
				"isr second %02d:%02d:%02d (compare %02d)\n",
				pdc_rtc_read(priv, PDC_RTC_HOUR),
				pdc_rtc_read(priv, PDC_RTC_MIN),
				sec, priv->softalrm_sec);
		}

		/* rtc alarms are one-shot */
		priv->hardstop_time = 0;
		if (events & RTC_AF) {
			if (!now && priv->alarm_irq_delay) {
				_pdc_rtc_read_time(priv, &tm);
				rtc_tm_to_time(&tm, &now);
			}
			priv->alarm_pending = 1;
			_pdc_rtc_stop_alarm(priv, now);
		}
	} else {
		/* make absolutely sure that the alarm is properly stopped */
		if (priv->alarm_irq_delay) {
			_pdc_rtc_read_time(priv, &tm);
			rtc_tm_to_time(&tm, &now);
		}
		_pdc_rtc_stop_alarm(priv, now);
		dev_dbg(priv->dev,
			"isr irq %#x ignored (alarm disabled)\n",
			status);
	}

	spin_unlock_irqrestore(&priv->lock, flags);

	if (events != RTC_IRQF)
		rtc_update_irq(priv->rtc_dev, 1, events);

	return IRQ_HANDLED;
}

static int pdc_rtc_setup(struct pdc_rtc_priv *priv)
{
	unsigned long clk_rate, flags;
	u32 reg;

	/* setup clock dividers */
	clk_rate = clk_get_rate(priv->clk);

	reg = (clk_rate / 2) << PDC_RTC_CLK_DIV_2HZ_SHIFT;
	if (clk_rate % 32768) {
		reg |= (2 * clk_rate / 32768) << PDC_RTC_CLK_DIV_32KHZ_SHIFT;
		reg |= 1 << PDC_RTC_CLK_DIV_NEG_EN_SHIFT;
	} else {
		reg |= (clk_rate / 32768) << PDC_RTC_CLK_DIV_32KHZ_SHIFT;
	}
	pdc_rtc_write(priv, PDC_RTC_CLK_GEN_DIV, reg);

	/* enable appropriate interrupts */
	pdc_rtc_write(priv, PDC_RTC_IRQ_EN, PDC_RTC_IRQ_ALARM);

	/* make sure clock is enabled */

	spin_lock_irqsave(&priv->lock, flags);
	priv->control_reg |= PDC_RTC_CONTROL_CE;
	pdc_rtc_write(priv, PDC_RTC_CONTROL, priv->control_reg);
	spin_unlock_irqrestore(&priv->lock, flags);

	return 0;
}

static int pdc_rtc_probe(struct platform_device *pdev)
{
	struct pdc_rtc_priv *priv;
	struct resource *res_regs;
	struct device_node *node = pdev->dev.of_node;
	int irq, ret, error;
	u32 val;

	if (!node)
		return -ENOENT;

	/* Get resources from platform device */
	irq = platform_get_irq(pdev, 0);
	if (irq < 0) {
		dev_err(&pdev->dev, "cannot find IRQ resource\n");
		return irq;
	}

	res_regs = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (res_regs == NULL) {
		dev_err(&pdev->dev, "cannot find registers resource\n");
		return -ENOENT;
	}

	/* Private driver data */
	priv = devm_kzalloc(&pdev->dev, sizeof(*priv), GFP_KERNEL);
	if (!priv)
		return -ENOMEM;

	spin_lock_init(&priv->lock);
	platform_set_drvdata(pdev, priv);
	priv->dev = &pdev->dev;

	/* Get devicetree properties */
	ret = of_property_read_u32(node, "time-set-delay", &val);
	if (!ret)
		priv->time_set_delay = val;
	ret = of_property_read_u32(node, "alarm-irq-delay", &val);
	if (!ret)
		priv->alarm_irq_delay = val;

	/* Ioremap the registers */
	priv->reg_base = devm_ioremap(&pdev->dev, res_regs->start,
				      res_regs->end - res_regs->start);
	if (!priv->reg_base) {
		dev_err(&pdev->dev,
			"cannot ioremap registers\n");
		return -EIO;
	}

	/* Get clock */
	priv->clk = devm_clk_get(&pdev->dev, "core");
	if (IS_ERR(priv->clk)) {
		dev_err(&pdev->dev, "could not get core clock resource\n");
		return PTR_ERR(priv->clk);
	}
	clk_prepare_enable(priv->clk);

	/* disable interrupts */
	pdc_rtc_write(priv, PDC_RTC_IRQ_EN, 0);

	priv->softalrm_sec = -1;
	priv->irq = irq;
	error = devm_request_irq(&pdev->dev, priv->irq, pdc_rtc_isr, 0,
				 "pdc-rtc", priv);
	if (error) {
		dev_err(&pdev->dev, "cannot register IRQ %u\n",
			priv->irq);
		return -EIO;
	}

	/* Register our RTC with the RTC framework */
	device_init_wakeup(&pdev->dev, 1);
	priv->rtc_dev = rtc_device_register(pdev->name, &pdev->dev,
					    &pdc_rtc_ops,
					    THIS_MODULE);
	if (unlikely(IS_ERR(priv->rtc_dev)))
		return PTR_ERR(priv->rtc_dev);

	pdc_rtc_setup(priv);

	return 0;
}

static int pdc_rtc_remove(struct platform_device *pdev)
{
	struct pdc_rtc_priv *priv = platform_get_drvdata(pdev);

	rtc_device_unregister(priv->rtc_dev);
	clk_disable_unprepare(priv->clk);

	return 0;
}

#ifdef CONFIG_PM_SLEEP
/*
 * We use noirq callbacks because an ISR after the normal callbacks can clear
 * the secondly interrupt which would then get restored on resume and keep
 * firing.
 */
static int pdc_rtc_suspend_noirq(struct device *dev)
{
	struct pdc_rtc_priv *priv = dev_get_drvdata(dev);
	unsigned int irq_en;

	priv->suspended = true;
	/* only wake if the alarm is enabled */
	if (device_may_wakeup(dev) && pdc_rtc_alarm_enabled(priv)) {
		/* disable interrupts other than the alarm */
		irq_en = priv->last_irq_en = pdc_rtc_read(priv, PDC_RTC_IRQ_EN);
		irq_en &= PDC_RTC_IRQ_ALARM;
		pdc_rtc_write(priv, PDC_RTC_IRQ_EN, irq_en);
		/* and enable wakeup on the interrupt */
		enable_irq_wake(priv->irq);
		priv->wakeup = true;
	}
	return 0;
}

static int pdc_rtc_resume_noirq(struct device *dev)
{
	struct pdc_rtc_priv *priv = dev_get_drvdata(dev);

	if (priv->wakeup) {
		/* disable wakeup */
		disable_irq_wake(priv->irq);
		/* and restore the previous interrupt enable bits */
		pdc_rtc_write(priv, PDC_RTC_IRQ_EN, priv->last_irq_en);
		priv->wakeup = false;
	}
	/* submit any postponed rtc interrupt */
	priv->suspended = false;
	if (priv->postponed_rtc_int) {
		dev_dbg(priv->dev,
			"submitting postponed rtc interrupt %x\n",
			priv->postponed_rtc_int);
		rtc_update_irq(priv->rtc_dev, 1, priv->postponed_rtc_int);
		priv->postponed_rtc_int = 0;
	}
	return 0;
}
#else
#define pdc_rtc_suspend NULL
#define pdc_rtc_resume NULL
#endif	/* CONFIG_PM_SLEEP */

static const struct dev_pm_ops pdc_rtc_pmops = {
#ifdef CONFIG_PM_SLEEP
	.suspend_noirq = pdc_rtc_suspend_noirq,
	.resume_noirq = pdc_rtc_resume_noirq,
	.freeze_noirq = pdc_rtc_suspend_noirq,
	.thaw_noirq = pdc_rtc_resume_noirq,
	.poweroff_noirq = pdc_rtc_suspend_noirq,
	.restore_noirq = pdc_rtc_resume_noirq,
#endif
};

static const struct of_device_id pdc_rtc_match[] = {
	{ .compatible = "img,pdc-rtc" },
	{}
};
MODULE_DEVICE_TABLE(of, pdc_rtc_match);

static struct platform_driver pdc_rtc_driver = {
	.driver = {
		.name = "pdc-rtc",
		.owner	= THIS_MODULE,
		.of_match_table	= pdc_rtc_match,
		.pm = &pdc_rtc_pmops,
	},
	.probe = pdc_rtc_probe,
	.remove = pdc_rtc_remove,
};

module_platform_driver(pdc_rtc_driver);

MODULE_AUTHOR("Ed Blake <ed.blake@imgtec.com>");
MODULE_DESCRIPTION("ImgTec PowerDown Controller RTC");
MODULE_LICENSE("GPL");
