/*
* Imagination Technologies Serial Flash Controller driver
*
* Copyright (C) 2015 Imagination Technologies Ltd.
*
* This program is free software; you can redistribute it and/or modify it
* under the terms and conditions of the GNU General Public License,
* version 2, as published by the Free Software Foundation.
*/

#include <linux/clk.h>
#include <linux/completion.h>
#include <linux/dma-mapping.h>
#include <linux/dmapool.h>
#include <linux/interrupt.h>
#include <linux/io.h>
#include <linux/irq.h>
#include <linux/list.h>
#include <linux/module.h>
#include <linux/of.h>
#include <linux/platform_device.h>
#include <linux/pm_runtime.h>
#include <linux/scatterlist.h>
#include <linux/delay.h>
#include <linux/spi/spi.h>

/* Configuration/Control Registers. */
#define SFC_CONFIG			0x08
#define SFC_CONFIG_MODE_SEL_MASK	0x7
#define SFC_CONFIG_CPHA_EN		BIT(3)
#define SFC_CONFIG_CPOL_EN		BIT(4)
#define SFC_CONFIG_LSBF_EN		BIT(5)
#define SFC_CONFIG_SOFT_RESET		BIT(16)

#define SFC_CONTROL			0x0C
#define SFC_CONTROL_AHB_ALIGN		8
#define SFC_CONTROL_CMD_INIT_SHIFT	16
#define SFC_CONTROL_XFER_MODE_SHIFT	1
#define SFC_CONTROL_LANE_MODE_MASK	0x3
#define SFC_CONTROL_LANE_MODE_SHIFT	18
#define SFC_CONTROL_CHIP_DEASSERT_SHIFT	22
#define SFC_CONTROL_DEV_SEL_2LSB_MASK	0x3
#define SFC_CONTROL_DEV_SEL_2LSB_SHIFT	20
#define SFC_CONTROL_DEV_SEL_MSB_MASK	0x1
#define SFC_CONTROL_DEV_SEL_MSB_SHIFT	25

#define SFC_CLK_CTRL			0x10
#define SFC_CLK_CTRL_CLK_ENABLE		BIT(0)
#define SFC_CLK_CTRL_CLK_STABLE		BIT(1)
#define SFC_CLK_CTRL_DIV_SHIFT		8
#define SFC_CLK_CTRL_DIV_MASK		0x7FF

/* Buffer thresholds */
#define SFC_CMD_THRESHOLD		0x14
#define SFC_CMD_THRESHOLD_DEFAULT_VAL	0xFFFFFFFF

#define SFC_INT_THRESHOLD		0x18
#define SFC_INT_THRESHOLD_TX_VAL_SHIFT	16
#define SFC_INT_THRESHOLD_RX_VAL_SHIFT	0
#define SFC_INT_THRESHOLD_RX_MASK	0xFFFF
#define SFC_INT_THRESHOLD_TX_MASK	0xFFFF

#define SFC_CONTROL_THRESHOLD		0x54
#define SFC_CONTROL_THRESHOLD_BUF_VAL	1

/* PIO TX/RX data registers */
#define SFC_TX_CMD_DATA			0x24
#define SFC_RX_CMD_DATA			0x28

/* DMA Registers */
#define SFC_BD_BASE_ADDR			0x40
#define SFC_BD_CTRL				0x34
#define SFC_BD_CTRL_CSR_DMA_EN			BIT(0)
#define SFC_BD_CTRL_CSR_BDP_CH_START_WSTB	BIT(2)

/* Interrupt Registers */
#define SFC_INT_STATUS			0x20
#define SFC_INT_ENABLE			0x1c
#define SFC_INT_SIG_ENABLE		0x58

/*
 * Registers that deal with handling/enabling interrupts have an identical
 * bit-fields, the following macros are used for Enabling and checking interrupt
 * status.
 */
#define SFC_TX_BUFF_EMPTY_INT		BIT(0)
#define SFC_TX_THRESH_INT		BIT(2)
#define SFC_RX_BUFF_EMPTY_INT		BIT(3)
#define SFC_RX_THRESH_INT		BIT(5)
#define SFC_CTRL_BUF_THRES_INT		BIT(8)
#define SFC_BDP_CSR_CBD_INT		BIT(9)
#define SFC_BDP_CSR_PKT_INT		BIT(10)
#define MAS_ERR_INT			BIT(11)

/* Non-register related macros */
#define MAX_SFC_CLK_DIV			2048
#define TX_RX_FIFO_DEPTH		256
#define TX_RX_FIFO_WIDTH		4
#define TX_RX_FIFO_BYTE_SIZE		(TX_RX_FIFO_DEPTH * TX_RX_FIFO_WIDTH)
#define MAX_PIO_CHUNK			TX_RX_FIFO_BYTE_SIZE
#define MAX_PIO_XFER_SIZE		16
#define MAX_BD_DATA_SIZE		256
#define SGT_PAGE_SIZE			PAGE_SIZE
#define SFC_NUM_CS			8
#define RESET_TIMEOUT_JIFFIES		msecs_to_jiffies(1000)
#define CLK_STABLE_TIMEOUT_JIFFIES	msecs_to_jiffies(1000)
#define DMA_XFER_TIMEOUT_JIFFIES(size)	msecs_to_jiffies(((size) * 10) + 50)
#define PIO_THRESH_TIMEOUT_JIFFIES	msecs_to_jiffies(10000)
#define CMD_THRESH_TIMEOUT_JIFFIES	msecs_to_jiffies(1000)

#define SUPPORTED_INTERRUPTS	(SFC_TX_BUFF_EMPTY_INT	|\
				 SFC_TX_THRESH_INT	|\
				 SFC_RX_BUFF_EMPTY_INT	|\
				 SFC_RX_THRESH_INT	|\
				 SFC_CTRL_BUF_THRES_INT	|\
				 SFC_BDP_CSR_CBD_INT	|\
				 SFC_BDP_CSR_PKT_INT)

#define ERR_INT			MAS_ERR_INT

struct img_sfc {
	int			irq;
	bool			int_err;
	spinlock_t		lock;
	void __iomem		*regs;
	struct device		*dev;
	struct clk		*sfc_clk;
	struct clk		*sys_clk;
	struct completion	irq_done;
};

struct sfc_bd_ctrl {
	u32	bd_buf_len:14;
	u32	bdp_ecc_cal_en:1;
	u32	bdp_ecc_access_en:1;
	u32	cbd_int_en:1;
	u32	pkt_int_en:1;
	u32	lifm:1;
	u32	last_bd:1;
	u32	dir_bit:1;
	u32	sdr_ddr_sel:1;
	u32	mode:2;
	u32	sfc_dev_sel_msb:1;
	u32	lsbf:1;
	u32	rx_ecc_cal_start:1;
	u32	status_check:1;
	u32	sfc_dev_sel_2lsb:2;
	u32	chip_sel:1;
	u32	desc_en:1;
};

struct sfc_bd {
	/* Hardware buffer descriptor fields */
	struct sfc_bd_ctrl	bd_ctrl;
	u32			bd_status;
	u32			bd_data_addr;
	dma_addr_t		nextaddr;
	/* Software buffer descriptor fields */
	struct sfc_bd		*next;
} ____cacheline_aligned;

enum sfc_mode {
	SFC_BOOT_MODE,
	SFC_PIO_MODE,
	SFC_DMA_MODE,
	SFC_XIP_MODE,
};

enum sfc_cmd_init {
	SFC_CMD_IDLE,
	SFC_CMD_TX,
	SFC_CMD_RX,
};

static inline u32 sfc_readl(struct img_sfc *sfc, u32 reg)
{
	return readl(sfc->regs + reg);
}

static inline void sfc_writel(struct img_sfc *sfc, u32 val, u32 reg)
{
	writel(val, sfc->regs + reg);
}

static int sfc_reset(struct img_sfc *sfc)
{
	u32 reset_bit = 0;
	unsigned long timeout;

	sfc_writel(sfc, SFC_PIO_MODE, SFC_CONFIG);
	sfc_writel(sfc, 0, SFC_INT_ENABLE);
	sfc_writel(sfc, 0, SFC_INT_SIG_ENABLE);
	sfc_writel(sfc, 0, SFC_CONFIG);
	sfc_writel(sfc, 0, SFC_CLK_CTRL);
	sfc_writel(sfc, SFC_CONFIG_SOFT_RESET | SFC_PIO_MODE, SFC_CONFIG);

	/* Wait for self clear reset bit to clear */
	timeout = jiffies + RESET_TIMEOUT_JIFFIES;
	do {
		reset_bit = sfc_readl(sfc, SFC_CONFIG) & SFC_CONFIG_SOFT_RESET;
		cpu_relax();
	} while (reset_bit && time_before(jiffies, timeout));

	if (!time_before(jiffies, timeout)) {
		dev_err(sfc->dev, "Timed out waiting for Reset bit to clear.\n");
		return -ETIMEDOUT;
	}
	/* Disable clock as soft reset re-enables the clock */
	sfc_writel(sfc, 0, SFC_CLK_CTRL);
	return 0;
}

static int img_sfc_clk_config(struct spi_master *master,
			      struct spi_transfer *xfer)
{
	struct img_sfc *sfc = spi_master_get_devdata(master);
	unsigned long timeout;
	u32 reg_val, clk_div_config, clk_div_val, clk_enable, clk_stable, ret;

	reg_val = sfc_readl(sfc, SFC_CLK_CTRL);

	clk_div_config = roundup_pow_of_two(DIV_ROUND_UP(master->max_speed_hz,
							 xfer->speed_hz)) >> 1;

	/* Return clk divider value used upon successful clk configuration */
	ret = clk_div_config << 1;

	clk_div_config = (clk_div_config & SFC_CLK_CTRL_DIV_MASK)
			   << SFC_CLK_CTRL_DIV_SHIFT;

	clk_div_val = reg_val & (SFC_CLK_CTRL_DIV_MASK
				 << SFC_CLK_CTRL_DIV_SHIFT);

	clk_enable = reg_val & SFC_CLK_CTRL_CLK_ENABLE;

	/*
	 * No need to reconfigure clock if it is already enabled with the same
	 * divisor value
	 */
	if (clk_enable && (clk_div_config == clk_div_val))
		return ret;

	sfc_writel(sfc, clk_div_config | SFC_CLK_CTRL_CLK_ENABLE, SFC_CLK_CTRL);
	timeout = jiffies + CLK_STABLE_TIMEOUT_JIFFIES;
	do {
		clk_stable = sfc_readl(sfc, SFC_CLK_CTRL)
				      & SFC_CLK_CTRL_CLK_STABLE;
		cpu_relax();
	} while (!clk_stable && time_before(jiffies, timeout));

	if (!time_before(jiffies, timeout)) {
		dev_err(sfc->dev, "Timed out waiting for stable clock.\n");
		return -ETIMEDOUT;
	}

	return ret;
}

static void img_sfc_control_reg_params(struct img_sfc *sfc,
				       struct spi_transfer *xfer,
				       struct spi_device *spi_dev,
				       bool last_chunk, u32 chunk_size)
{
	u32 cmd_init, reg_val = 0, lane_mode, cs_deasrt, cs_field_value;

	/*
	 * Configure lane mode.
	 * Mode			n_bits	lane_mode
	 * SINGLE_MODE		0x01	0x00
	 * DUAL_MODE		0x02	0x01
	 * QUAD_MODE		0x04	0x02
	 */

	lane_mode = ((xfer->rx_nbits | xfer->tx_nbits)
		     >> SFC_CONTROL_XFER_MODE_SHIFT)
		     & SFC_CONTROL_LANE_MODE_MASK;

	cmd_init = xfer->rx_buf ? SFC_CMD_RX : SFC_CMD_TX;

	cs_deasrt = xfer->cs_change && last_chunk ? 1 : 0;

	cs_field_value = (spi_dev->chip_select & SFC_CONTROL_DEV_SEL_2LSB_MASK)
			 << SFC_CONTROL_DEV_SEL_2LSB_SHIFT;

	cs_field_value |= ((spi_dev->chip_select >> 2)
			   & SFC_CONTROL_DEV_SEL_MSB_MASK)
			  << SFC_CONTROL_DEV_SEL_MSB_SHIFT;
	reg_val = (cs_deasrt << SFC_CONTROL_CHIP_DEASSERT_SHIFT)
		  | (cmd_init << SFC_CONTROL_CMD_INIT_SHIFT)
		  | chunk_size
		  | ((lane_mode & SFC_CONTROL_LANE_MODE_MASK)
		     << SFC_CONTROL_LANE_MODE_SHIFT)
		  | cs_field_value;

	sfc_writel(sfc, reg_val, SFC_CONTROL);
}

static int img_sfc_wait_xfer_complete(struct img_sfc *sfc, u32 dir)
{
	unsigned long flags;
	u32 buf_2_check = dir ? SFC_RX_BUFF_EMPTY_INT : SFC_TX_BUFF_EMPTY_INT;

	sfc_writel(sfc, buf_2_check | ERR_INT, SFC_INT_ENABLE);
	sfc_writel(sfc, buf_2_check | ERR_INT, SFC_INT_SIG_ENABLE);
	if (!wait_for_completion_timeout(&sfc->irq_done,
					 PIO_THRESH_TIMEOUT_JIFFIES)) {
		reinit_completion(&sfc->irq_done);
		if (sfc->int_err) {
			spin_lock_irqsave(&sfc->lock, flags);
			sfc->int_err = false;
			spin_unlock_irqrestore(&sfc->lock, flags);
			dev_err(sfc->dev, "Master Error occurred while waiting for %s threshold interrupt\n",
				dir ? "RX" : "TX");
			return -EIO;
		}
		dev_err(sfc->dev, "Waiting for %s FIFO empty interrupt timed out.\n",
			dir ? "RX" : "TX");
		return -ETIMEDOUT;
	}

	return 0;
}

static int img_sfc_fifo_thresh(struct img_sfc *sfc, u32 byte_size, u32 dir)
{
	unsigned long flags;
	u32 thresh_int_en, thresh_val;

	/* Threshold value is configured in 4-byte multiples */
	thresh_val = DIV_ROUND_UP(byte_size, 4);

	if (!dir) {	/* TX Thresh Configuration */
		thresh_int_en = SFC_TX_THRESH_INT;
		thresh_val <<= SFC_INT_THRESHOLD_TX_VAL_SHIFT;
		thresh_val &= SFC_INT_THRESHOLD_TX_MASK
				<< SFC_INT_THRESHOLD_TX_VAL_SHIFT;
	} else {	/* RX Thresh Configuration */
		thresh_int_en = SFC_RX_THRESH_INT;
		thresh_val <<= SFC_INT_THRESHOLD_RX_VAL_SHIFT;
		thresh_val &= SFC_INT_THRESHOLD_RX_MASK
				<< SFC_INT_THRESHOLD_RX_VAL_SHIFT;
	}

	sfc_writel(sfc, thresh_val, SFC_INT_THRESHOLD);
	sfc_writel(sfc, thresh_int_en | ERR_INT, SFC_INT_ENABLE);
	sfc_writel(sfc, thresh_int_en | ERR_INT, SFC_INT_SIG_ENABLE);

	if (!wait_for_completion_timeout(&sfc->irq_done,
					 PIO_THRESH_TIMEOUT_JIFFIES)) {
		reinit_completion(&sfc->irq_done);
		if (sfc->int_err) {
			spin_lock_irqsave(&sfc->lock, flags);
			sfc->int_err = false;
			spin_unlock_irqrestore(&sfc->lock, flags);
			dev_err(sfc->dev, "Master Error occurred while waiting for %s threshold interrupt.\n",
				dir ? "RX" : "TX");
			return -EIO;
		}
		dev_err(sfc->dev, "Waiting for %s threshold interrupt timed out.\n",
			dir ? "RX" : "TX");
		return -ETIMEDOUT;
	}

	return 0;
}

static int img_sfc_ctrl_buf_thresh_check(struct img_sfc *sfc)
{
	int ret = 0;
	unsigned long flags;

	sfc_writel(sfc, SFC_CONTROL_THRESHOLD_BUF_VAL, SFC_CONTROL_THRESHOLD);
	sfc_writel(sfc, SFC_CTRL_BUF_THRES_INT | ERR_INT, SFC_INT_ENABLE);
	sfc_writel(sfc, SFC_CTRL_BUF_THRES_INT | ERR_INT, SFC_INT_SIG_ENABLE);

	/* Wait for interrupt */
	if (!wait_for_completion_timeout(&sfc->irq_done,
					 CMD_THRESH_TIMEOUT_JIFFIES)) {
		reinit_completion(&sfc->irq_done);
		if (sfc->int_err) {
			spin_lock_irqsave(&sfc->lock, flags);
			sfc->int_err = false;
			spin_unlock_irqrestore(&sfc->lock, flags);
			dev_err(sfc->dev, "Master Error occurred while waiting for Control Buffer Threshold interrupt.\n");
			ret = -EIO;
		} else {
			dev_err(sfc->dev, "Waiting for Control Buffer interrupt timed out.\n");
			ret = -ETIMEDOUT;
		}
	}

	return ret;
}

/*
 * Both img_sfc_pio_write and img_sfc_pio_read are customised functions for
 * performing read and write operations for the SFC. These two functions take
 * into consideration the limitations of the SFC module. These functions will
 * only be able to accept a transfer size (xfer_size) of 1, 2, 4, or x4 bytes.
 */
static int img_sfc_pio_write(struct img_sfc *sfc, const void *buf,
			     u32 buf_offset, u32 xfer_size)
{
	u32 buf_end_index = buf_offset + xfer_size;
	/*
	 * copy_size is determined by the value of xfer_size. at this stage of
	 * the PIO transfer sequence, xfer_size will either be 16, 8, 4, 2, or 1
	 * bytes assuring that if the xfer_size is larger than 4 bytes, it is
	 * guaranteed to be a value that is in multiples of 4, therefore
	 * copy_size is set to 4 as the loop processes 4 bytes at a
	 * time. If xfer_size is less than 4, it could either be 2 or 1 bytes,
	 * in each case the loop will only iterate once, and copy_size is the
	 * size of the transfer.
	 * the same description applies for copy_size in img_sfc_pio_read.
	 */
	u32 copy_size = xfer_size >= 4 ? 4 : xfer_size;
	u32 data = 0;

	if ((xfer_size > 4) && (xfer_size % 4)) {
		dev_err(sfc->dev, "Invalid transfer size to be written - %d bytes.\n",
			xfer_size);
		return -EINVAL;
	}
	while (buf_offset < buf_end_index) {
		memcpy(&data, buf + buf_offset, copy_size);
		sfc_writel(sfc, data, SFC_TX_CMD_DATA);
		buf_offset += copy_size;
	}

	return 0;
}

static int img_sfc_pio_read(struct img_sfc *sfc, void *buf,
			    u32 buf_offset, u32 xfer_size)
{
	u32 buf_end_index = buf_offset + xfer_size;
	u32 rx_data_reg_val;
	u32 copy_size = xfer_size >= 4 ? 4 : xfer_size;

	if ((xfer_size > 4) && (xfer_size % 4)) {
		dev_err(sfc->dev, "Invalid transfer size to be read - %d bytes.\n"
			, xfer_size);
		return -EINVAL;
	}
	while (buf_offset < buf_end_index) {
		rx_data_reg_val = sfc_readl(sfc, SFC_RX_CMD_DATA);
		memcpy(buf + buf_offset, &rx_data_reg_val, copy_size);
		buf_offset += copy_size;
	}

	return 0;
}

static inline u32 img_sfc_calc_xfer_chunk_size(u32 size)
{
	u32 res = 0;

	if (size >= MAX_PIO_CHUNK) {
		res = MAX_PIO_CHUNK;
	} else {
		if (size >= SFC_CONTROL_AHB_ALIGN) {
			res = size & ~(SFC_CONTROL_AHB_ALIGN - 1);
		} else {
			if (size >= 4)
				res = 4;
			else if (size >= 2)
				res = 2;
			else
				res = 1;
		}
	}

	return res;
}

static int img_sfc_start_pio(struct spi_master *master,
			     struct spi_device *spi_dev,
			     struct spi_transfer *xfer)
{
	struct img_sfc *sfc = spi_master_get_devdata(spi_dev->master);
	u32 pending_bytes = 0, chunk_size = 0, bytes_done = 0;
	bool last_chunk = 0;
	int err;

	/* Check direction of Transfer (RX = 1 , TX = 0) */
	u32 dir = xfer->tx_buf > xfer->rx_buf ? 0 : 1;

	pending_bytes = xfer->len;
	while (pending_bytes) {
		/* Split transfer to suitable chunks */
		chunk_size = img_sfc_calc_xfer_chunk_size(pending_bytes);

		/* Check Control Buffer for  space availability */
		err = img_sfc_ctrl_buf_thresh_check(sfc);
		if (err)
			return err;

		/* Configure the Control register to prepare for Transfer */
		last_chunk = pending_bytes == chunk_size ? 1 : 0;
		img_sfc_control_reg_params(sfc, xfer, spi_dev, last_chunk,
					   chunk_size);

		/* Configure TX/RX FIFO threshold */
		err = img_sfc_fifo_thresh(sfc, chunk_size, dir);
		if (err)
			return err;

		/*Start Transaction*/
		if (dir) {	/* RX Transaction */
			err = img_sfc_pio_read(sfc, xfer->rx_buf, bytes_done,
					       chunk_size);
			if (err)
				return err;
		} else {	/* TX Transaction */
			err = img_sfc_pio_write(sfc, xfer->tx_buf, bytes_done,
						chunk_size);
			if (err)
				return err;
		}
		/* Wait for transaction to finish */
		err = img_sfc_wait_xfer_complete(sfc, dir);
		if (err)
			return err;

		bytes_done += chunk_size;
		pending_bytes -= chunk_size;
	}

	return 0;
}

static int img_sfc_map_sgt(struct spi_device *spi_dev,
			   struct sg_table *sgt,
			   struct spi_transfer *xfer)
{
	struct img_sfc *sfc = spi_master_get_devdata(spi_dev->master);
	u32 sgl_len, sgl_index = 0, buf_index = 0;
	u32 len = xfer->len;
	u32 dir = xfer->tx_buf ? DMA_TO_DEVICE : DMA_FROM_DEVICE;
	u32 sgt_size = DIV_ROUND_UP(xfer->len, SGT_PAGE_SIZE);

	for (sgl_index = 0; sgl_index < sgt_size; sgl_index++) {
		sgl_len = len < SGT_PAGE_SIZE ? len : SGT_PAGE_SIZE;
		if (dir == DMA_TO_DEVICE) {
			/* TX transaction, use xfer->tx_buf */
			sg_set_page(&sgt->sgl[sgl_index],
				    virt_to_page(xfer->tx_buf + buf_index),
				    sgl_len,
				    offset_in_page(xfer->tx_buf + buf_index));
		} else {
			/* RX transaction, use xfer->rx_buf */
			sg_set_page(&sgt->sgl[sgl_index],
				    virt_to_page(xfer->rx_buf + buf_index),
				    sgl_len,
				    offset_in_page(xfer->rx_buf + buf_index));
		}
		buf_index += sgl_len;
		len -= sgl_len;
	}
	/* DMA map the scatter list */
	if (!dma_map_sg(&spi_dev->dev, sgt->sgl, sgt->nents, dir)) {
		dev_err(sfc->dev, "Failed to DMA map Scatter Table.\n");
		return -ENOMEM;
	}

	return 0;
}

static inline void img_sfc_config_default_bd_ctrl(struct spi_device *spi_dev,
						  struct spi_transfer *xfer,
						  struct sfc_bd_ctrl *bd_ctrl)
{
	/* Fill in the buffer descriptor control field */
	bd_ctrl->bd_buf_len = MAX_BD_DATA_SIZE;
	bd_ctrl->bdp_ecc_cal_en = 0;
	bd_ctrl->bdp_ecc_access_en = 0;
	bd_ctrl->cbd_int_en = 0;
	bd_ctrl->pkt_int_en = 0;
	bd_ctrl->lifm = 0;
	bd_ctrl->last_bd = 0;
	bd_ctrl->dir_bit = xfer->tx_buf > xfer->rx_buf ? 0 : 1;
	bd_ctrl->sdr_ddr_sel = 0;
	bd_ctrl->mode = ((xfer->rx_nbits | xfer->tx_nbits)
			>> SFC_CONTROL_XFER_MODE_SHIFT)
			& SFC_CONTROL_LANE_MODE_MASK;
	bd_ctrl->sfc_dev_sel_msb = (spi_dev->chip_select >> 2)
				& SFC_CONTROL_DEV_SEL_MSB_MASK;
	bd_ctrl->lsbf = spi_dev->mode & SPI_LSB_FIRST ? 1 : 0;
	bd_ctrl->rx_ecc_cal_start = 0;
	bd_ctrl->status_check = 0;
	bd_ctrl->sfc_dev_sel_2lsb = spi_dev->chip_select &
			SFC_CONTROL_DEV_SEL_2LSB_MASK;
	bd_ctrl->chip_sel = 0;
	bd_ctrl->desc_en = 1;
}

static inline void img_sfc_config_last_bd_ctrl(struct sfc_bd_ctrl *bd_ctrl,
					       struct spi_transfer *xfer,
					       u32 bd_buf_len)
{
	bd_ctrl->bd_buf_len = bd_buf_len;
	bd_ctrl->cbd_int_en = 1;
	bd_ctrl->lifm = 1;
	bd_ctrl->last_bd = 1;
	bd_ctrl->chip_sel = xfer->cs_change;
}

static inline void img_sfc_free_bd_list(struct img_sfc *sfc,
					struct dma_pool *bd_pool,
					struct sfc_bd *list_head,
					dma_addr_t bd_head_phys_addr)
{
	struct sfc_bd *curr_bd;
	dma_addr_t curr_bd_dma_addr = bd_head_phys_addr;
	dma_addr_t next_bd_dma_addr;

	while (list_head) {
		curr_bd = list_head;
		list_head = list_head->next;
		next_bd_dma_addr = curr_bd->nextaddr;
		dma_pool_free(bd_pool, (void *)curr_bd, curr_bd_dma_addr);
		curr_bd_dma_addr = next_bd_dma_addr;
	}
}

static inline void img_sfc_start_bdp(struct img_sfc *sfc,
				     dma_addr_t bd_head_phys_addr)
{
	/* Provide the physical address to buffer descriptor */
	sfc_writel(sfc, (u32)bd_head_phys_addr, SFC_BD_BASE_ADDR);
	/* Enable interrupts */
	sfc_writel(sfc, SFC_BDP_CSR_CBD_INT | ERR_INT, SFC_INT_ENABLE);
	sfc_writel(sfc, SFC_BDP_CSR_CBD_INT | ERR_INT, SFC_INT_SIG_ENABLE);
	/* Enable and start the DMA transaction */
	sfc_writel(sfc, SFC_BD_CTRL_CSR_DMA_EN |
			SFC_BD_CTRL_CSR_BDP_CH_START_WSTB, SFC_BD_CTRL);
}

static int img_sfc_start_dma(struct spi_master *master,
			     struct spi_device *spi_dev,
			     struct spi_transfer *xfer)
{
	struct img_sfc *sfc = spi_master_get_devdata(spi_dev->master);
	struct sg_table *sgt;
	struct scatterlist *sg;
	struct sfc_bd_ctrl bd_ctrl_tmplt;
	struct sfc_bd *curr_bd = NULL, *bd_list_head = NULL;
	struct dma_pool *bd_pool;
	unsigned long flags, timeout;
	bool xfer_dir = xfer->tx_buf > xfer->rx_buf ? 0 : 1;
	u32 dma_dir = xfer->tx_buf > xfer->rx_buf ?
			DMA_TO_DEVICE : DMA_FROM_DEVICE;
	u32 sgl_count = 0, bd_count, no_of_bds = 0, len, bd_size;
	dma_addr_t sg_addr, bd_dma_addr, phys_addr;
	u32 sgt_size = DIV_ROUND_UP(xfer->len, SGT_PAGE_SIZE);
	int ret = 0;

	/* Allocate scatter table */
	sgt = kmalloc(sizeof(struct sg_table *), GFP_KERNEL);
	if (!sgt)
		return -ENOMEM;

	ret = sg_alloc_table(sgt, sgt_size, GFP_KERNEL);
	if (ret) {
		dev_err(sfc->dev, "Failed to allocate scatter table error.\n");
		goto free_sgt_malloc;
	}

	/* Map the transfer buffer into a scatter/gather table */
	ret = img_sfc_map_sgt(spi_dev, sgt, xfer);
	if (ret)
		goto free_sgt;
	/*
	 * The IMG SFC requires a buffer descriptor (bd) linked list where each
	 * bd references 256 bytes of data. Each descriptor contains 4 fields
	 * bd_ctrl: contains information about the transfer.
	 * bs_status: not implemented.
	 * bd_data_addr: physical address of the data referenced.
	 * nextaddr: physical address of next bd.
	 *
	 * The buffer descriptors also contain a fifth field (*next) that is not
	 * present to hardware, this field is used to link the bd's together
	 * using their virtual addresses.
	 *
	 * The bd list references the scatter table (sgt) which, based on the
	 * size of the transaction buffer, may contain more than one scatter
	 * list (sg) entry. The bd list is continuous irregardless of how many
	 * sg entries there are (i.e. the last buffer descriptor of sg[i] is
	 * linked to the first descriptor of sg[i+1]).
	 *
	 * the bd_ctrl field is identical for all buffer descriptors except for
	 * the last one where it needs to contain information indicating to the
	 * hardware that it is the last descriptor, hence why a bd_ctrl template
	 * is created a the beginning which is copied for every bd generated,
	 * and that last one is simply a modified template.
	 *
	 * once the bd list is mapped, the physical address of the first bd is
	 * written to the SFC bd processor, an interrupt is generated when the
	 * transaction is completed.
	 */

	img_sfc_config_default_bd_ctrl(spi_dev, xfer, &bd_ctrl_tmplt);
	bd_size = sizeof(struct sfc_bd);
	bd_pool = dma_pool_create("IMG-SFC-BD-POOL", sfc->dev, bd_size,
				  __alignof__(struct sfc_bd), 0);
	if (!bd_pool) {
		dev_err(sfc->dev, "Failed to create buffer descriptor pool.\n");
		ret = -ENOMEM;
		goto unmap_sgt;
	}
	bd_list_head = (struct sfc_bd *)dma_pool_alloc(bd_pool, GFP_KERNEL,
						       &bd_dma_addr);
	if (!bd_list_head) {
		dev_err(sfc->dev, "Buffer Descriptor list head allocation failed.\n");
		ret = -ENOMEM;
		goto free_bd_pool;
	}
	curr_bd = bd_list_head;
	for_each_sg(sgt->sgl, sg, sgt->nents, sgl_count) {
		sg_addr = sg_dma_address(sg);
		len = sg_dma_len(sg);
		no_of_bds = DIV_ROUND_UP(len, MAX_BD_DATA_SIZE);
		bd_count = 0;
		while (bd_count < no_of_bds) {
			if (sg_is_last(sg) && (len <= MAX_BD_DATA_SIZE)) {
				img_sfc_config_last_bd_ctrl(&bd_ctrl_tmplt,
							    xfer, len);
				curr_bd->next = 0;
				curr_bd->nextaddr = 0;
			} else {
				curr_bd->next =
				(struct sfc_bd *)dma_pool_alloc(bd_pool,
								GFP_KERNEL,
								&phys_addr);
				if (!curr_bd->next) {
					dev_err(sfc->dev, "Buffer Descriptor allocation failed.\n");
					ret = -ENOMEM;
					goto free_bd_list;
				}
				curr_bd->nextaddr = phys_addr;
			}
			curr_bd->bd_ctrl = bd_ctrl_tmplt;
			curr_bd->bd_data_addr = (sg_addr +
						(bd_count * MAX_BD_DATA_SIZE));
			curr_bd->bd_status = 0;
			if (curr_bd->next)
				curr_bd = curr_bd->next;
			bd_count++;
			len -= len > MAX_BD_DATA_SIZE ? MAX_BD_DATA_SIZE : len;
		}
	}
	timeout = DMA_XFER_TIMEOUT_JIFFIES(DIV_ROUND_UP(xfer->len,
							MAX_BD_DATA_SIZE));
	img_sfc_start_bdp(sfc, bd_dma_addr);
	if (!wait_for_completion_timeout(&sfc->irq_done, timeout)) {
		reinit_completion(&sfc->irq_done);
		if (sfc->int_err) {
			spin_lock_irqsave(&sfc->lock, flags);
			sfc->int_err = false;
			spin_unlock_irqrestore(&sfc->lock, flags);
			dev_err(sfc->dev, "Master Error occurred while waiting for DMA packet completion.\n");
			ret = -EIO;
		} else {
			dev_err(sfc->dev, "Timed out waiting for DMA packet completion.\n");
			ret = -ETIMEDOUT;
		}
	}

	/* Wait for buffers to clear */
	ret = img_sfc_wait_xfer_complete(sfc, xfer_dir);
	if (ret)
		return ret;
	/* Disable DMA mode and stop buffer descriptor processor */
	sfc_writel(sfc, 0, SFC_BD_CTRL);

free_bd_list:
	img_sfc_free_bd_list(sfc, bd_pool, bd_list_head, bd_dma_addr);
free_bd_pool:
	dma_pool_destroy(bd_pool);
unmap_sgt:
	dma_unmap_sg(&spi_dev->dev, sgt->sgl, sgt->nents, dma_dir);
free_sgt:
	sg_free_table(sgt);
free_sgt_malloc:
	kfree(sgt);

	return ret;
}

static int img_sfc_transfer_one(struct spi_master *master,
				struct spi_device *spi_dev,
				struct spi_transfer *xfer)
{
	struct img_sfc *sfc = spi_master_get_devdata(master);
	int ret;
	u32 mode_config, xfer_rate = 0;

	ret = img_sfc_clk_config(master, xfer);
	if (ret < 0)
		return ret;
	else if (ret > 0)
		xfer_rate = master->max_speed_hz / ret;

	if (xfer->cs_change ||
	    list_is_last(&xfer->transfer_list, &master->cur_msg->transfers)) {
		xfer->cs_change = true;
		/* Add pre-determined chip_sel delays for low speed TX xfers */
		if (xfer->tx_buf) {
			if (xfer_rate <= 100000)
				xfer->delay_usecs = 20000;
			else if (xfer_rate <= 500000)
				xfer->delay_usecs = 15000;
			else if (xfer_rate <= 1000000)
				xfer->delay_usecs = 1000;
			else if (xfer_rate < 5000000)
				xfer->delay_usecs = 10;
		}
	} else {
		xfer->cs_change = false;
	}

	mode_config = sfc_readl(sfc, SFC_CONFIG) & ~SFC_CONFIG_MODE_SEL_MASK;

	if (xfer->len > MAX_PIO_XFER_SIZE) {
		mode_config |= SFC_DMA_MODE;
		sfc_writel(sfc, mode_config, SFC_CONFIG);
		ret = img_sfc_start_dma(master, spi_dev, xfer);
	} else {
		mode_config |= SFC_PIO_MODE;
		sfc_writel(sfc, mode_config, SFC_CONFIG);
		ret = img_sfc_start_pio(master, spi_dev, xfer);
	}

	return ret;
}

static irqreturn_t img_sfc_irq(int irq, void *dev_id)
{
	irqreturn_t ret;
	unsigned long flags;
	struct img_sfc *sfc = (struct img_sfc *)dev_id;
	u32 int_status = sfc_readl(sfc, SFC_INT_STATUS);
	u32 int_check = sfc_readl(sfc, SFC_INT_ENABLE) & int_status;

	if (!int_check) {
		return IRQ_NONE;
	} else if (int_check & SUPPORTED_INTERRUPTS) {
		complete(&sfc->irq_done);
		ret = IRQ_HANDLED;
	} else if (int_check & ERR_INT) {
		spin_lock_irqsave(&sfc->lock, flags);
		sfc->int_err = true;
		spin_unlock_irqrestore(&sfc->lock, flags);
		ret =  IRQ_HANDLED;
	} else {
		dev_err(sfc->dev, "Non supported Interrupt triggered.\n");
		ret = IRQ_NONE;
	}

	sfc_writel(sfc, 0, SFC_INT_ENABLE);
	sfc_writel(sfc, 0, SFC_INT_SIG_ENABLE);

	return ret;
}

static int img_sfc_prepare_message(struct spi_master *master,
				   struct spi_message *msg)
{
	struct img_sfc *sfc = spi_master_get_devdata(master);
	u32 sfc_config = 0;

	/* Configure CPOL, CPHA, and LSB */
	if (msg->spi->mode & SPI_CPHA)
		sfc_config |= SFC_CONFIG_CPHA_EN;
	if (msg->spi->mode & SPI_CPOL)
		sfc_config |= SFC_CONFIG_CPOL_EN;
	if (msg->spi->mode & SPI_LSB_FIRST)
		sfc_config |= SFC_CONFIG_LSBF_EN;
	/*
	 * Enable PIO Mode by default to prevent SFC from getting into boot
	 * mode.
	 */
	sfc_config |= SFC_PIO_MODE;
	sfc_writel(sfc, sfc_config, SFC_CONFIG);
	/* Reset Interrupts and Thresholds */
	sfc_writel(sfc, 0, SFC_INT_ENABLE);
	sfc_writel(sfc, 0, SFC_INT_SIG_ENABLE);
	sfc_writel(sfc, SFC_CONTROL_THRESHOLD_BUF_VAL, SFC_CONTROL_THRESHOLD);
	sfc_writel(sfc, SFC_CMD_THRESHOLD_DEFAULT_VAL, SFC_CMD_THRESHOLD);

	return 0;
}

static int img_sfc_unprepare_message(struct spi_master *master,
				     struct spi_message *msg)
{
	struct img_sfc *sfc = spi_master_get_devdata(master);

	return sfc_reset(sfc);
}

static void img_sfc_handle_err(struct spi_master *master,
			       struct spi_message *msg)
{
	struct img_sfc *sfc = spi_master_get_devdata(master);

	sfc_reset(sfc);
}

static int img_sfc_probe(struct platform_device *pdev)
{
	struct img_sfc *sfc;
	struct resource *res;
	struct spi_master *master;
	unsigned int num_cs, bus_num;
	int ret, warn;

	master = spi_alloc_master(&pdev->dev, sizeof(*sfc));
	if (!master)
		return -ENOMEM;

	platform_set_drvdata(pdev, master);
	sfc = spi_master_get_devdata(master);

	sfc->dev = &pdev->dev;
	spin_lock_init(&sfc->lock);
	init_completion(&sfc->irq_done);

	sfc->int_err = false;

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	sfc->regs = devm_ioremap_resource(sfc->dev, res);
	if (IS_ERR(sfc->regs)) {
		ret = PTR_ERR(sfc->regs);
		goto put_master;
	}

	sfc->irq = platform_get_irq(pdev, 0);
	if (sfc->irq < 0) {
		ret = sfc->irq;
		goto put_master;
	}

	sfc->sys_clk = devm_clk_get(sfc->dev, "sys");
	if (IS_ERR(sfc->sys_clk)) {
		ret = PTR_ERR(sfc->sys_clk);
		goto put_master;
	}

	sfc->sfc_clk = devm_clk_get(sfc->dev, "sfc");
	if (IS_ERR(sfc->sfc_clk)) {
		ret = PTR_ERR(sfc->sfc_clk);
		goto put_master;
	}
	ret = clk_prepare_enable(sfc->sys_clk);
	if (ret)
		goto put_master;
	ret = clk_prepare_enable(sfc->sfc_clk);
	if (ret)
		goto disable_sys_clk;

	ret = sfc_reset(sfc);
	if (ret)
		goto put_master;
	/* Disable interrupts to avoid spurious start-up interrupt */
	sfc_writel(sfc, 0, SFC_INT_ENABLE);
	sfc_writel(sfc, 0, SFC_INT_SIG_ENABLE);

	ret = devm_request_irq(sfc->dev, sfc->irq, img_sfc_irq,
			       IRQ_TYPE_LEVEL_HIGH, dev_name(sfc->dev), sfc);
	if (ret)
		goto disable_sfc_clk;

	master->auto_runtime_pm = true;

	warn = of_property_read_u32(sfc->dev->of_node, "bus-num", &bus_num);
	if (!warn) {
		master->bus_num = bus_num;
	} else {
		dev_warn(sfc->dev, "Failed to read bus-num property, using default value %d.\n",
			 pdev->id);
		master->bus_num = pdev->id;
	}

	master->mode_bits = SPI_CPOL | SPI_CPHA | SPI_TX_DUAL | SPI_RX_DUAL;
	if (of_property_read_bool(sfc->dev->of_node, "img,supports-quad-mode"))
		master->mode_bits |= SPI_TX_QUAD | SPI_RX_QUAD
				     | SPI_SINGLE_MODE_CMD;

	master->flags = SPI_MASTER_HALF_DUPLEX;
	master->dev.of_node = pdev->dev.of_node;
	master->bits_per_word_mask = SPI_BPW_MASK(32) | SPI_BPW_MASK(8);
	master->max_speed_hz = clk_get_rate(sfc->sfc_clk);
	master->min_speed_hz = master->max_speed_hz / MAX_SFC_CLK_DIV;

	warn = of_property_read_u32(sfc->dev->of_node, "num-cs", &num_cs);
	if (!warn) {
		master->num_chipselect = num_cs;
	} else {
		dev_warn(sfc->dev, "Failed to read num-cs property, using default value %d.",
			 SFC_NUM_CS);
		master->num_chipselect = SFC_NUM_CS;
	}

	master->transfer_one = img_sfc_transfer_one;
	master->prepare_message = img_sfc_prepare_message;
	master->unprepare_message = img_sfc_unprepare_message;
	master->handle_err = img_sfc_handle_err;

	pm_runtime_set_autosuspend_delay(sfc->dev, MSEC_PER_SEC);
	pm_runtime_use_autosuspend(sfc->dev);
	pm_runtime_set_active(sfc->dev);
	pm_runtime_enable(sfc->dev);

	ret = devm_spi_register_master(sfc->dev, master);
	if (ret)
		goto disable_pm;

	return 0;

disable_pm:
	pm_runtime_disable(sfc->dev);
	pm_runtime_set_suspended(sfc->dev);
disable_sfc_clk:
	clk_disable_unprepare(sfc->sfc_clk);
disable_sys_clk:
	clk_disable_unprepare(sfc->sys_clk);
put_master:
	spi_master_put(master);

	return ret;
}

static int img_sfc_remove(struct platform_device *pdev)
{
	struct spi_master *master = platform_get_drvdata(pdev);
	struct img_sfc *sfc = spi_master_get_devdata(master);

	pm_runtime_disable(sfc->dev);
	if (!pm_runtime_status_suspended(sfc->dev)) {
		clk_disable_unprepare(sfc->sfc_clk);
		clk_disable_unprepare(sfc->sys_clk);
	}

	spi_master_put(master);

	return 0;
}

#ifdef CONFIG_PM
static int img_sfc_runtime_suspend(struct device *dev)
{
	struct spi_master *master = dev_get_drvdata(dev);
	struct img_sfc *sfc = spi_master_get_devdata(master);

	clk_disable_unprepare(sfc->sfc_clk);
	clk_disable_unprepare(sfc->sys_clk);

	return 0;
}

static int img_sfc_runtime_resume(struct device *dev)
{
	struct spi_master *master = dev_get_drvdata(dev);
	struct img_sfc *sfc = spi_master_get_devdata(master);
	int ret;

	ret = clk_prepare_enable(sfc->sys_clk);
	if (ret)
		return ret;

	ret = clk_prepare_enable(sfc->sfc_clk);
	if (ret) {
		clk_disable_unprepare(sfc->sys_clk);
		return ret;
	}

	return 0;
}
#endif /* CONFIG_PM */

#ifdef CONFIG_PM_SLEEP
static int img_sfc_suspend(struct device *dev)
{
	struct spi_master *master = dev_get_drvdata(dev);
	int ret;

	ret = spi_master_suspend(master);
	if (ret)
		return ret;

	if (!pm_runtime_suspended(dev))
		img_sfc_runtime_suspend(dev);

	return 0;
}

static int img_sfc_resume(struct device *dev)
{
	struct spi_master *master = dev_get_drvdata(dev);

	return spi_master_resume(master);
}

static int img_sfc_resume_noirq(struct device *dev)
{
	struct spi_master *master = dev_get_drvdata(dev);
	struct img_sfc *sfc = spi_master_get_devdata(master);
	int ret;

	ret = img_sfc_runtime_resume(dev);
	if (ret)
		return ret;

	ret = sfc_reset(sfc);

	if (pm_runtime_suspended(dev))
		img_sfc_runtime_suspend(dev);

	return ret;
}
#endif /* CONFIG_PM_SLEEP */

static const struct dev_pm_ops img_sfc_pm_ops = {
	SET_RUNTIME_PM_OPS(img_sfc_runtime_suspend, img_sfc_runtime_resume,
			   NULL)
	SET_SYSTEM_SLEEP_PM_OPS(img_sfc_suspend, img_sfc_resume)
#ifdef CONFIG_PM_SLEEP
	.resume_noirq = img_sfc_resume_noirq,
#endif
};

static const struct of_device_id img_sfc_of_match[] = {
	{ .compatible = "img,sfc", },
	{ },
};
MODULE_DEVICE_TABLE(of, img_sfc_of_match);

static struct platform_driver img_sfc_driver = {
	.driver = {
		.name = "img-sfc",
		.pm = &img_sfc_pm_ops,
		.of_match_table = of_match_ptr(img_sfc_of_match),
	},
	.probe = img_sfc_probe,
	.remove = img_sfc_remove,
};
module_platform_driver(img_sfc_driver);

MODULE_DESCRIPTION("IMG Serial Flash Controller (SFC) driver");
MODULE_AUTHOR("Seraj Mohammed <seraj.mohammed@imgtec.com>");
MODULE_LICENSE("GPL v2");
