/*
 * PDP Framebuffer
 *
 * Copyright (c) 2008 Imagination Technologies Ltd.
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 */

#ifndef _PDPFB_H
#define _PDPFB_H

#ifndef __PDPFB_USE_OSAL__
	#include <linux/fb.h>
	#include <video/pdpfb2.h>
#else
	#include "pdpfb2.h"
#endif
	#include "pdpfb2_regs.h"

#ifdef __PDPFB_USE_OSAL__
	#define PDPFB_STATIC
#else
	#define PDPFB_STATIC static
#endif

#define PDPFB_STREAM_GFX1				0
#define PDPFB_STREAM_GFX2				1
#define PDPFB_STREAM_GFX3				2
#define PDPFB_STREAM_GFX4				3
#define PDPFB_STREAM_VID1				4
#define PDPFB_STREAM_VID2				5
#define PDPFB_STREAM_VID3				6
#define PDPFB_STREAM_VID4				7
#define PDPFB_MAX_NR_STREAMS			8

#define	PDPFB_NO_REG					1
#define PDPFB_MAX_TAPS					8
#define PDPFB_MAX_INTERPOLATION_POINTS 16

/* Video memory pools */
#define PDPFB_MEMPOOL_MEM		0x00	/* Combined video memory */
#define PDPFB_MEMPOOL_GFX1MEM	0x01	/* Graphics 1 plane memory */
#define PDPFB_MEMPOOL_GFX2MEM	0x02	/* Graphics 2 plane memory */
#define PDPFB_MEMPOOL_GFX3MEM	0x03	/* Graphics 3 plane memory */
#define PDPFB_MEMPOOL_GFX4MEM	0x04	/* Graphics 4 plane memory */
#define PDPFB_MEMPOOL_VID1MEM	0x05	/* Video 1 plane memory */
#define PDPFB_MEMPOOL_VID2MEM	0x06	/* Video 2 plane memory */
#define PDPFB_MEMPOOL_VID3MEM	0x07	/* Video 3 plane memory */
#define PDPFB_MEMPOOL_VID4MEM	0x08	/* Video 4 plane memory */
#define PDPFB_MEMPOOL_USER		0x09	/* User provided memory */
#define PDPFB_MEMPOOL_NR_POOLS	10
#define PDPFB_MEMPOOL_USERPLANE	0xfd	/* User provided memory for one plane */
#define PDPFB_MEMPOOL_KERNEL	0xfe	/* Kernel allocated memory */
#define PDPFB_MEMPOOL_NONE	0xff

#define PDP_PAGE_SIZE       4096
#define PDP_PAGE_MASK       (~(PDP_PAGE_SIZE-1))

/* Maximum 6 bytes per pixel for RGB161616 */
#define PDP_MAX_IMAGE_BYTES (PDP_MAX_INPUT_WIDTH * PDP_MAX_INPUT_HEIGHT * 6)

/* Round up */
#define PDP_MAX_IMAGE_PAGES \
            ((PDP_MAX_IMAGE_BYTES + PDP_PAGE_SIZE - 1) / PDP_PAGE_SIZE)

/* The number of pointers per page table */
#define PDPFB_PTRS_PER_PTBL	(PDP_PAGE_SIZE / 4)

/* 
 * The number of pointers per page directory, sized to allow the maximum
 * supported image size.
 * 1 is added to allow for buffers not starting on a page boundary
 */
#define PDPFB_PTRS_PER_PDIR	((PDP_MAX_IMAGE_PAGES / PDPFB_PTRS_PER_PTBL) + 1)

/*
 * The total size of the page directories required for all streams in bytes
 * (there are two directories per stream - current and next)
 */
#define PDPFB_PAGE_DIR_BYTES (PDPFB_MAX_NR_STREAMS * 2 * PDPFB_PTRS_PER_PDIR)

#define PDPFB_REG_EXISTS(reg)	((reg) == PDPFB_NO_REG ? 0 : 1)
#define PDPFB_IS_PLANAR(nonstd)	((nonstd) == PDP_PIXFMT_420_PL8 \
								|| (nonstd) == PDP_PIXFMT_420_PL8IVU  \
								|| (nonstd) == PDP_PIXFMT_420_PL8IUV \
								|| (nonstd) == PDP_PIXFMT_420_PL10IVU \
								|| (nonstd) == PDP_PIXFMT_420_PL10IUV \
								|| (nonstd) == PDP_PIXFMT_422_PL10IVU \
								|| (nonstd) == PDP_PIXFMT_422_PL10IUV)

/* Stream interface */

struct platform_device;

struct pdpfb_lcd_size {			/* width and height of panel in mm */
	unsigned int dynamic_mode;	/* the screen accepts different modes */
	unsigned int width;
	unsigned int height;
};

struct pdpfb_sync_cfg {
	unsigned int force_vsyncs;
	unsigned int hsync_dis;			/* sync_ctrl.hsdis */
	unsigned int vsync_dis;			/* sync_ctrl.vsdis */
	unsigned int blank_dis;			/* sync_ctrl.blnkdis */
	unsigned int de_active_high;	/* sync_ctrl.blnkpol */
	unsigned int clock_inverted;	/* sync_ctrl.clkpol */
};

struct pdpfb_cfg {
	__u32 bpp;
	struct fb_videomode lcd_cfg;
	struct pdpfb_lcd_size lcd_size;
	struct pdpfb_sync_cfg sync_cfg;
	__u32 linestore_len;
};

struct pdpfb_priv;
struct pdpfb_stream;

struct pdpfb_stream_ops {
	int (*probe)(struct pdpfb_priv *priv, struct platform_device *pdev);
	int (*remove)(struct pdpfb_priv *priv, struct platform_device *pdev);
};

struct pdpfb_stream_regs {
	u16 surf;
	u16 blend2_r;
	u16 blend2_gb;
	u16 ctrl;
	u16 uctrl;
	u16 stride;
	u16 size;
	u16 posn;
	u16 galpha;
	u16 ckey_r;
	u16 ckey_gb;
	u16 interleave_ctrl;
	u16 base_addr;
	u16 u_base_addr;
	u16 v_base_addr;
	u16 memctrl;
	u16 memthresh;
	u16 skip_ctrl;
	u16 postskip_ctrl;
	u16 decimate_ctrl;
	u16 scale_ctrl;
	u16 vsinit;
	u16 first_vscoeff;
	u16 last_vscoeff;
	u16 hsinit;
	u16 first_hscoeff;
	u16 last_hscoeff;
	u16 scale_size;
	u16 lumakey_alpha_trans_opaque;
	u16 lumakey_luma_max_min;
	u16 lumakey_c_rg;
	u16 lumakey_c_b;
};

struct pdpfb_stream_caps {
	u32 has_scaler;
	u32 vs_taps;
	u32 hs_taps;
	u32 vs_interpolation_points;
	u32 hs_interpolation_points;
	u32 max_scaler_input_width;
	u32 has_doubler; /* Has pixel / line doubling */
	u32 has_decimator; /* Has pixel halving / line decimation */
	u32 has_cosited_chroma_control;
	u32 has_csc;
};

struct pdpfb_stream {
	struct fb_info info;
	int probed;

	unsigned char mem_pool;
	void *videomem_cpu_virt; /* The CPU virtual address of video memory */
	/*
	* The PDP base address of video memory (the physical address if not using
	* the MMU, otherwise the PDP virtual address)
	*/
	u32 videomem_pdp_base;
	unsigned long videomem_len;
	/*
	 * A pointer to this stream's current and next page directories (containing
	 * physical page table entries for use by the PDP MMU h/w)
	 */
	u32 (*pdir)[2][PDPFB_PTRS_PER_PDIR];
	/*
	 * Shadow directories containing page table kernel virtual addresses,
	 * retained for freeing
	 */
	u32 *pdir_kvirt[2][PDPFB_PTRS_PER_PDIR];
	/*
	 * Shadow directories containing page table bus addresses, retained for 
	 * unmapping 
	 */
	dma_addr_t pdir_bus[2][PDPFB_PTRS_PER_PDIR];
	/*
	 * Shadow directories containing user page pointers, retained for
	 * releasing
	 */
	struct page *pdir_user_page[2][PDPFB_PTRS_PER_PDIR][PDPFB_PTRS_PER_PTBL];
	/*
	 * The offset of this stream's current and next page directory from the MMU
	 * directory base (in directory entries, not bytes)
	 */
	u32 pdir_offset[2];
	/*
	 * The index of the next directory (the directory that is not being used
	 * for the current frame being displayed)
	 */
	int next_dir;
	/* 
	 * Flag to indicate if the next directory has been updated and should
	 * become the current directory for the next frame
	 */
	int next_dir_updated;
#ifdef CONFIG_FB_PDP2_PDUMP
	/*
	 * Shadow directories containing pdump memory block IDs, retained for
	 * freeing
	 */
	int pdump_pdir_mem_block[2][PDPFB_PTRS_PER_PDIR];
	/* The pdump memory block ID for the memory containing the current image */
	int pdump_image_mem_block;
	struct pdpfb_usermem_file usermem_file;
#endif

	int enable;
	int pos;
	int ckey_en;
	int ckey_src;	/* see PDP_STRXCTRL_CKEYSRC_* */
	struct pdpfb_ckey ckey;
	struct pdpfb_blend_config blend_config;
	u16 global_alpha; /* 10-bit alpha */
	struct pdpfb_geom geom;
	struct pdpfb_lumakey lumakey;
	struct pdpfb_geom clip_geom;
	int gamma_en;
	int dithering_en;
	int mode_master; /* framebuffer controls screen mode */
	int bilinear_en; /* use bilinear filtering for horiz and vert scaling */
	int split_scale;
	int hskip;

	struct pdpfb_stream_ops ops;
	struct pdpfb_stream_regs regs;
	struct pdpfb_stream_caps caps;

	/* horizontal scaler */
	int hscoeffs[(PDPFB_MAX_TAPS * PDPFB_MAX_INTERPOLATION_POINTS) + 1];
	int hs_oldsize, hs_oldres;
	int hs_en;
	/* vertical scaler */
	int vscoeffs[(PDPFB_MAX_TAPS * PDPFB_MAX_INTERPOLATION_POINTS) + 1];
	int vs_oldsize, vs_oldres;
	int vs_en;
	/* pixel format info */
	struct pdpfb_csc csc;
	int planar_override;
	int planar_ov_virtres[2];
	int planar_ov_pixfmt;
	struct pdpfb_planar planar;
	int pixfmt;
};

struct pdpfb_cfg *pdpfb_get_config(struct pdpfb_priv *priv);

struct pdpfb_stream *pdpfb_get_stream(struct pdpfb_priv *priv,
									unsigned int stream_id);

void pdpfb_str_videomem_free(struct pdpfb_priv *priv,
							struct pdpfb_stream *stream);

int pdpfb_str_videomem_alloc(struct pdpfb_priv *priv,
							struct pdpfb_stream *stream);

int pdpfb_str_ioctl(struct pdpfb_priv *priv, struct pdpfb_stream *stream,
					unsigned int cmd, unsigned long arg );

int pdpfb_blank(int blank, struct fb_info *info);

int	pdpfb_str_pan_display(struct pdpfb_priv *priv, struct pdpfb_stream *stream,
						struct fb_var_screeninfo *var, struct fb_info *info );

int	pdpfb_str_check_var(struct pdpfb_priv *priv, struct pdpfb_stream *stream,
						struct fb_var_screeninfo *var, struct fb_info *info );

unsigned long pdpfb_str_required_mem(struct fb_var_screeninfo *var);

void pdpfb_str_match_var(struct pdpfb_priv *priv, struct pdpfb_stream *stream,
						struct fb_var_screeninfo *var);

void pdpfb_str_get_planar(struct pdpfb_stream *stream,
						struct pdpfb_planar *pl);

int pdpfb_str_set_planar(struct pdpfb_priv *priv, struct pdpfb_stream *stream,
						struct fb_info *info, struct pdpfb_planar *pl );

void pdpfb_str_get_lumakey(struct pdpfb_stream *stream,
						struct pdpfb_lumakey *lumakey);

int pdpfb_str_set_lumakey(struct pdpfb_priv *priv, struct pdpfb_stream *stream,
						struct pdpfb_lumakey *lumakey);

void pdpfb_str_get_clip_geom(struct pdpfb_stream *stream,
							struct pdpfb_geom *clip_geom);

int pdpfb_str_set_clip_geom(struct pdpfb_priv *priv,
					struct pdpfb_stream *stream, struct pdpfb_geom *clip_geom);

int pdpfb_str_set_par(struct pdpfb_priv *priv, struct pdpfb_stream *stream,
					struct fb_info *info );

int pdpfb_str_get_split_scale(struct pdpfb_stream *stream);

int pdpfb_str_set_split_scale(struct pdpfb_priv *priv,
							struct pdpfb_stream *stream, int split_scale);

#ifdef __PDPFB_USE_OSAL__
int pdpfb_probe(struct platform_device *pdev);
int pdpfb_remove(struct platform_device *pdev);
irqreturn_t pdpfb_interrupt(int irq, void *dev_id);
#endif

#endif
