/*
 * PDP Framebuffer Streams
 *
 * Copyright (c) 2008 Imagination Technologies Ltd.
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 */

#ifndef __PDPFB_USE_OSAL__
	#include <linux/fb.h>
	#include <linux/platform_device.h>
	#include <linux/uaccess.h>
	#include "pdpfb2.h"
	#include "pdpfb2_regs.h"
	#include "pdpfb2_str.h"
#else
	#include "pdpfb2_linuxal.h"
	#define __KERNEL__
	#include "pdpfb2_internal.h"
	#include "pdpfb2_regs.h"
	#include "pdpfb2_str.h"
#endif

/* GFX1 Stream */
static int pdpfb_gfx1_probe(struct pdpfb_priv *priv,
							struct platform_device *pdev);
static int pdpfb_gfx1_remove(struct pdpfb_priv *priv,
							struct platform_device *pdev);

static const struct pdpfb_stream pdpfb_gfx1_stream_template = {
	.mem_pool = PDPFB_MEMPOOL_GFX1MEM,
	.videomem_len = 0,
	.enable = 0,
	.pos = 0,
	.ckey = {
		.ckey = 0x00000000,
		.mask = 0x3FFFFFFF,
	},
	.global_alpha = 0x3FF,
	.ops = {
		.probe = pdpfb_gfx1_probe,
		.remove = pdpfb_gfx1_remove,
	},
	.caps = {
		.has_scaler = 0,
		.has_doubler = 1,
		.has_decimator = 1,
		.has_cosited_chroma_control = 0,
		.has_csc = 1,
	},
	.regs = {
		.surf = PDP_GRPH1SURF_OFFSET,
		.blend2_r = PDP_GRPH1BLND2_R_OFFSET,
		.blend2_gb = PDP_GRPH1BLND2_GB_OFFSET,
		.ctrl = PDP_GRPH1CTRL_OFFSET,
		.uctrl = PDPFB_NO_REG,
		.stride = PDP_GRPH1STRIDE_OFFSET,
		.size = PDP_GRPH1SIZE_OFFSET,
		.posn = PDP_GRPH1POSN_OFFSET,
		.galpha = PDP_GRPH1GALPHA_OFFSET,
		.ckey_r = PDP_GRPH1CKEY_R_OFFSET,
		.ckey_gb = PDP_GRPH1CKEY_GB_OFFSET,
		.interleave_ctrl = PDP_GRPH1INTERLEAVE_CTRL_OFFSET,
		.base_addr = PDP_GRPH1BASEADDR_OFFSET,
		.u_base_addr = PDPFB_NO_REG,
		.v_base_addr = PDPFB_NO_REG,
		.memctrl = PDP_GRPH1_MEMCTRL_OFFSET,
		.memthresh = PDP_GRPH1_MEM_THRESH_OFFSET,
		.skip_ctrl = PDPFB_NO_REG,
		.postskip_ctrl = PDPFB_NO_REG,
		.decimate_ctrl = PDP_GRPH1DECIMATE_CTRL_OFFSET,
		.scale_ctrl = PDPFB_NO_REG,
		.vsinit = PDPFB_NO_REG,
		.first_vscoeff = PDPFB_NO_REG,
		.last_vscoeff = PDPFB_NO_REG,
		.hsinit = PDPFB_NO_REG,
		.first_hscoeff = PDPFB_NO_REG,
		.last_hscoeff = PDPFB_NO_REG,
		.scale_size = PDPFB_NO_REG,
		.lumakey_alpha_trans_opaque = PDPFB_NO_REG,
		.lumakey_luma_max_min = PDPFB_NO_REG,
		.lumakey_c_rg = PDPFB_NO_REG,
		.lumakey_c_b = PDPFB_NO_REG,
	}
};

static int pdpfb_gfx1_check_var(struct fb_var_screeninfo *var,
								struct fb_info *info)
{
	struct pdpfb_priv *priv = dev_get_drvdata(info->device);
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_GFX1);
	return pdpfb_str_check_var(priv, stream, var, info);
}

static int pdpfb_gfx1_ioctl(struct fb_info *info, unsigned int cmd,
							unsigned long arg)
{
	struct pdpfb_priv *priv = dev_get_drvdata(info->device);
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_GFX1);
	return pdpfb_str_ioctl(priv, stream, cmd, arg);
}

static int pdpfb_gfx1_pan_display(struct fb_var_screeninfo *var,
								struct fb_info *info)
{
	struct pdpfb_priv *priv = dev_get_drvdata(info->device);
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_GFX1);
	return pdpfb_str_pan_display(priv, stream, var, info);
}

static int pdpfb_gfx1_set_par(struct fb_info *info)
{
	struct pdpfb_priv *priv = dev_get_drvdata(info->device);
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_GFX1);
	return pdpfb_str_set_par(priv, stream, info);
}

static struct fb_fix_screeninfo pdpfb_gfx1_fix	= {
	.id =		"pdp_gfx1",
	.type =		FB_TYPE_PACKED_PIXELS,
	.xpanstep =	8,
	.ypanstep =	1,
	.accel =	FB_ACCEL_IMG_PDP_1,
};

static struct fb_ops pdpfb_gfx1_ops = {
	.fb_blank	= pdpfb_blank,
	.fb_pan_display	= pdpfb_gfx1_pan_display,
	.fb_check_var	= pdpfb_gfx1_check_var,
	.fb_set_par	= pdpfb_gfx1_set_par,
	.fb_ioctl	= pdpfb_gfx1_ioctl,
#ifndef __PDPFB_USE_OSAL__
	.fb_fillrect	= cfb_fillrect,
	.fb_copyarea	= cfb_copyarea,
	.fb_imageblit	= cfb_imageblit,
#endif
};

static int pdpfb_gfx1_probe(struct pdpfb_priv *priv,
							struct platform_device *pdev)
{
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_GFX1);
	struct fb_info *info = &stream->info;
	struct pdpfb_cfg *cfg = pdpfb_get_config(priv);
	int error;

	info->device = &pdev->dev;
	info->fbops = &pdpfb_gfx1_ops;

	info->var.xres = info->var.xres_virtual = cfg->lcd_cfg.xres;
	info->var.yres = info->var.yres_virtual = cfg->lcd_cfg.yres;
	info->var.width = cfg->lcd_size.width;
	info->var.height = cfg->lcd_size.height;
	info->var.activate = FB_ACTIVATE_NOW;
	info->var.bits_per_pixel = cfg->bpp;
	pdpfb_str_match_var(priv, stream, &info->var);

	info->var.hsync_len = cfg->lcd_cfg.hsync_len;
	info->var.left_margin = cfg->lcd_cfg.left_margin;
	info->var.right_margin = cfg->lcd_cfg.right_margin;
	info->var.vsync_len = cfg->lcd_cfg.vsync_len;
	info->var.upper_margin = cfg->lcd_cfg.upper_margin;
	info->var.lower_margin = cfg->lcd_cfg.lower_margin;

	info->fix = pdpfb_gfx1_fix;
	info->fix.xpanstep = 16 * 8 / info->var.bits_per_pixel;

	error = pdpfb_str_videomem_alloc(priv, stream);
	if (error)
		goto err0;

	if (pdpfb_str_required_mem(&info->var) > stream->videomem_len &&
		stream->videomem_len != 0) {
		error = -ENOMEM;
		goto err1;
	}

	/* Graphics 1 is enabled by default if videomem_len is non-zero */
	if (stream->videomem_len != 0)
		stream->enable = 1;

	info->flags = FBINFO_FLAG_DEFAULT
				| FBINFO_HWACCEL_XPAN
				| FBINFO_HWACCEL_YPAN;

#ifndef __PDPFB_USE_OSAL__
	error = register_framebuffer(info);
	if (error < 0)
		goto err1;
#endif

	dev_info(&pdev->dev, "registered graphics 1 framebuffer (len=0x%lx)\n",
			stream->videomem_len);

	return 0;
err1:
	pdpfb_str_videomem_free(priv, stream);
err0:
	return error;
}

static int pdpfb_gfx1_remove(struct pdpfb_priv *priv,
							struct platform_device *pdev)
{
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_GFX1);
#ifndef __PDPFB_USE_OSAL__
	unregister_framebuffer(&stream->info);
#endif

	pdpfb_str_videomem_free(priv, stream);
	return 0;
}

void pdpfb_gfx1_init_stream(struct pdpfb_stream *stream)
{
	*stream = pdpfb_gfx1_stream_template;
}

/* GFX2 Stream */
static int pdpfb_gfx2_probe(struct pdpfb_priv *priv,
							struct platform_device *pdev);
static int pdpfb_gfx2_remove(struct pdpfb_priv *priv,
							struct platform_device *pdev);

static const struct pdpfb_stream pdpfb_gfx2_stream_template = {
	.mem_pool = PDPFB_MEMPOOL_GFX2MEM,
	.videomem_len = 0,
	.enable = 0,
	.pos = 1,
	.ckey = {
		.ckey = 0x00000000,
		.mask = 0x3FFFFFFF,
	},
	.global_alpha = 0x3FF,
	.ops = {
		.probe = pdpfb_gfx2_probe,
		.remove = pdpfb_gfx2_remove,
	},
	.caps = {
		.has_scaler = 0,
		.has_doubler = 1,
		.has_decimator = 1,
		.has_cosited_chroma_control = 0,
		.has_csc = 1,
	},
	.regs = {
		.surf = PDP_GRPH2SURF_OFFSET,
		.blend2_r = PDP_GRPH2BLND2_R_OFFSET,
		.blend2_gb = PDP_GRPH2BLND2_GB_OFFSET,
		.ctrl = PDP_GRPH2CTRL_OFFSET,
		.uctrl = PDPFB_NO_REG,
		.stride = PDP_GRPH2STRIDE_OFFSET,
		.size = PDP_GRPH2SIZE_OFFSET,
		.posn = PDP_GRPH2POSN_OFFSET,
		.galpha = PDP_GRPH2GALPHA_OFFSET,
		.ckey_r = PDP_GRPH2CKEY_R_OFFSET,
		.ckey_gb = PDP_GRPH2CKEY_GB_OFFSET,
		.interleave_ctrl = PDP_GRPH2INTERLEAVE_CTRL_OFFSET,
		.base_addr = PDP_GRPH2BASEADDR_OFFSET,
		.u_base_addr = PDPFB_NO_REG,
		.v_base_addr = PDPFB_NO_REG,
		.memctrl = PDP_GRPH2_MEMCTRL_OFFSET,
		.memthresh = PDP_GRPH2_MEM_THRESH_OFFSET,
		.skip_ctrl = PDPFB_NO_REG,
		.postskip_ctrl = PDPFB_NO_REG,
		.decimate_ctrl = PDP_GRPH2DECIMATE_CTRL_OFFSET,
		.scale_ctrl = PDPFB_NO_REG,
		.vsinit = PDPFB_NO_REG,
		.first_vscoeff = PDPFB_NO_REG,
		.last_vscoeff = PDPFB_NO_REG,
		.hsinit = PDPFB_NO_REG,
		.first_hscoeff = PDPFB_NO_REG,
		.last_hscoeff = PDPFB_NO_REG,
		.scale_size = PDPFB_NO_REG,
		.lumakey_alpha_trans_opaque = PDPFB_NO_REG,
		.lumakey_luma_max_min = PDPFB_NO_REG,
		.lumakey_c_rg = PDPFB_NO_REG,
		.lumakey_c_b = PDPFB_NO_REG,
	}
};

static int pdpfb_gfx2_check_var(struct fb_var_screeninfo *var,
								struct fb_info *info)
{
	struct pdpfb_priv *priv = dev_get_drvdata(info->device);
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_GFX2);
	return pdpfb_str_check_var(priv, stream, var, info);
}

static int pdpfb_gfx2_ioctl(struct fb_info *info, unsigned int cmd,
							unsigned long arg)
{
	struct pdpfb_priv *priv = dev_get_drvdata(info->device);
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_GFX2);
	return pdpfb_str_ioctl(priv, stream, cmd, arg);
}

static int pdpfb_gfx2_pan_display(struct fb_var_screeninfo *var,
								struct fb_info *info)
{
	struct pdpfb_priv *priv = dev_get_drvdata(info->device);
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_GFX2);
	return pdpfb_str_pan_display(priv, stream, var, info);
}

static int pdpfb_gfx2_set_par(struct fb_info *info)
{
	struct pdpfb_priv *priv = dev_get_drvdata(info->device);
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_GFX2);
	return pdpfb_str_set_par(priv, stream, info);
}

static struct fb_fix_screeninfo pdpfb_gfx2_fix	= {
	.id =		"pdp_gfx2",
	.type =		FB_TYPE_PACKED_PIXELS,
	.xpanstep =	8,
	.ypanstep =	1,
	.accel =	FB_ACCEL_IMG_PDP_1,
};

static struct fb_ops pdpfb_gfx2_ops = {
	.fb_blank	= pdpfb_blank,
	.fb_pan_display	= pdpfb_gfx2_pan_display,
	.fb_check_var	= pdpfb_gfx2_check_var,
	.fb_set_par	= pdpfb_gfx2_set_par,
	.fb_ioctl	= pdpfb_gfx2_ioctl,
#ifndef __PDPFB_USE_OSAL__
	.fb_fillrect	= cfb_fillrect,
	.fb_copyarea	= cfb_copyarea,
	.fb_imageblit	= cfb_imageblit,
#endif
};

static int pdpfb_gfx2_probe(struct pdpfb_priv *priv,
							struct platform_device *pdev)
{
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_GFX2);
	struct fb_info *info = &stream->info;
	struct pdpfb_cfg *cfg = pdpfb_get_config(priv);
	int error;

	info->device = &pdev->dev;
	info->fbops = &pdpfb_gfx2_ops;

	info->var.xres = info->var.xres_virtual = cfg->lcd_cfg.xres;
	info->var.yres = info->var.yres_virtual = cfg->lcd_cfg.yres;
	info->var.width = cfg->lcd_size.width;
	info->var.height = cfg->lcd_size.height;
	info->var.activate = FB_ACTIVATE_NOW;
	info->var.bits_per_pixel = cfg->bpp;
	pdpfb_str_match_var(priv, stream, &info->var);

	info->var.hsync_len = cfg->lcd_cfg.hsync_len;
	info->var.left_margin = cfg->lcd_cfg.left_margin;
	info->var.right_margin = cfg->lcd_cfg.right_margin;
	info->var.vsync_len = cfg->lcd_cfg.vsync_len;
	info->var.upper_margin = cfg->lcd_cfg.upper_margin;
	info->var.lower_margin = cfg->lcd_cfg.lower_margin;

	info->fix = pdpfb_gfx2_fix;
	info->fix.xpanstep = 16 * 8 / info->var.bits_per_pixel;

	error = pdpfb_str_videomem_alloc(priv, stream);
	if (error)
		goto err0;

	if (pdpfb_str_required_mem(&info->var) > stream->videomem_len &&
		stream->videomem_len != 0) {
		error = -ENOMEM;
		goto err1;
	}

	info->flags = FBINFO_FLAG_DEFAULT
				| FBINFO_HWACCEL_XPAN
				| FBINFO_HWACCEL_YPAN;

#ifndef __PDPFB_USE_OSAL__
	error = register_framebuffer(info);
	if (error < 0)
		goto err1;
#endif

	dev_info(&pdev->dev, "registered graphics 2 framebuffer (len=0x%lx)\n",
			stream->videomem_len);

	return 0;
err1:
	pdpfb_str_videomem_free(priv, stream);
err0:
	return error;
}

static int pdpfb_gfx2_remove(struct pdpfb_priv *priv,
							struct platform_device *pdev)
{
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_GFX2);
#ifndef __PDPFB_USE_OSAL__
	unregister_framebuffer(&stream->info);
#endif

	pdpfb_str_videomem_free(priv, stream);
	return 0;
}

void pdpfb_gfx2_init_stream(struct pdpfb_stream *stream)
{
	*stream = pdpfb_gfx2_stream_template;
}

/* GFX3 Stream */
static int pdpfb_gfx3_probe(struct pdpfb_priv *priv,
							struct platform_device *pdev);
static int pdpfb_gfx3_remove(struct pdpfb_priv *priv,
							struct platform_device *pdev);

static const struct pdpfb_stream pdpfb_gfx3_stream_template = {
	.mem_pool = PDPFB_MEMPOOL_GFX3MEM,
	.videomem_len = 0,
	.enable = 0,
	.pos = 2,
	.ckey = {
		.ckey = 0x00000000,
		.mask = 0x3FFFFFFF,
	},
	.global_alpha = 0x3FF,
	.ops = {
		.probe = pdpfb_gfx3_probe,
		.remove = pdpfb_gfx3_remove,
	},
	.caps = {
		.has_scaler = 0,
		.has_doubler = 1,
		.has_decimator = 1,
		.has_cosited_chroma_control = 0,
		.has_csc = 1,
	},
	.regs = {
		.surf = PDP_GRPH3SURF_OFFSET,
		.blend2_r = PDP_GRPH3BLND2_R_OFFSET,
		.blend2_gb = PDP_GRPH3BLND2_GB_OFFSET,
		.ctrl = PDP_GRPH3CTRL_OFFSET,
		.uctrl = PDPFB_NO_REG,
		.stride = PDP_GRPH3STRIDE_OFFSET,
		.size = PDP_GRPH3SIZE_OFFSET,
		.posn = PDP_GRPH3POSN_OFFSET,
		.galpha = PDP_GRPH3GALPHA_OFFSET,
		.ckey_r = PDP_GRPH3CKEY_R_OFFSET,
		.ckey_gb = PDP_GRPH3CKEY_GB_OFFSET,
		.interleave_ctrl = PDP_GRPH3INTERLEAVE_CTRL_OFFSET,
		.base_addr = PDP_GRPH3BASEADDR_OFFSET,
		.u_base_addr = PDPFB_NO_REG,
		.v_base_addr = PDPFB_NO_REG,
		.memctrl = PDP_GRPH3_MEMCTRL_OFFSET,
		.memthresh = PDP_GRPH3_MEM_THRESH_OFFSET,
		.skip_ctrl = PDPFB_NO_REG,
		.postskip_ctrl = PDPFB_NO_REG,
		.decimate_ctrl = PDP_GRPH3DECIMATE_CTRL_OFFSET,
		.scale_ctrl = PDPFB_NO_REG,
		.vsinit = PDPFB_NO_REG,
		.first_vscoeff = PDPFB_NO_REG,
		.last_vscoeff = PDPFB_NO_REG,
		.hsinit = PDPFB_NO_REG,
		.first_hscoeff = PDPFB_NO_REG,
		.last_hscoeff = PDPFB_NO_REG,
		.scale_size = PDPFB_NO_REG,
		.lumakey_alpha_trans_opaque = PDPFB_NO_REG,
		.lumakey_luma_max_min = PDPFB_NO_REG,
		.lumakey_c_rg = PDPFB_NO_REG,
		.lumakey_c_b = PDPFB_NO_REG,
	}
};

static int pdpfb_gfx3_check_var(struct fb_var_screeninfo *var,
								struct fb_info *info)
{
	struct pdpfb_priv *priv = dev_get_drvdata(info->device);
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_GFX3);
	return pdpfb_str_check_var(priv, stream, var, info);
}

static int pdpfb_gfx3_ioctl(struct fb_info *info, unsigned int cmd,
							unsigned long arg)
{
	struct pdpfb_priv *priv = dev_get_drvdata(info->device);
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_GFX3);
	return pdpfb_str_ioctl(priv, stream, cmd, arg);
}

static int pdpfb_gfx3_pan_display(struct fb_var_screeninfo *var,
								struct fb_info *info)
{
	struct pdpfb_priv *priv = dev_get_drvdata(info->device);
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_GFX3);
	return pdpfb_str_pan_display(priv, stream, var, info);
}

static int pdpfb_gfx3_set_par(struct fb_info *info)
{
	struct pdpfb_priv *priv = dev_get_drvdata(info->device);
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_GFX3);
	return pdpfb_str_set_par(priv, stream, info);
}

static struct fb_fix_screeninfo pdpfb_gfx3_fix	= {
	.id =		"pdp_gfx3",
	.type =		FB_TYPE_PACKED_PIXELS,
	.xpanstep =	8,
	.ypanstep =	1,
	.accel =	FB_ACCEL_IMG_PDP_1,
};

static struct fb_ops pdpfb_gfx3_ops = {
	.fb_blank	= pdpfb_blank,
	.fb_pan_display	= pdpfb_gfx3_pan_display,
	.fb_check_var	= pdpfb_gfx3_check_var,
	.fb_set_par	= pdpfb_gfx3_set_par,
	.fb_ioctl	= pdpfb_gfx3_ioctl,
#ifndef __PDPFB_USE_OSAL__
	.fb_fillrect	= cfb_fillrect,
	.fb_copyarea	= cfb_copyarea,
	.fb_imageblit	= cfb_imageblit,
#endif
};

static int pdpfb_gfx3_probe(struct pdpfb_priv *priv,
							struct platform_device *pdev)
{
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_GFX3);
	struct fb_info *info = &stream->info;
	struct pdpfb_cfg *cfg = pdpfb_get_config(priv);
	int error;

	info->device = &pdev->dev;
	info->fbops = &pdpfb_gfx3_ops;

	info->var.xres = info->var.xres_virtual = cfg->lcd_cfg.xres;
	info->var.yres = info->var.yres_virtual = cfg->lcd_cfg.yres;
	info->var.width = cfg->lcd_size.width;
	info->var.height = cfg->lcd_size.height;
	info->var.activate = FB_ACTIVATE_NOW;
	info->var.bits_per_pixel = cfg->bpp;
	pdpfb_str_match_var(priv, stream, &info->var);

	info->var.hsync_len = cfg->lcd_cfg.hsync_len;
	info->var.left_margin = cfg->lcd_cfg.left_margin;
	info->var.right_margin = cfg->lcd_cfg.right_margin;
	info->var.vsync_len = cfg->lcd_cfg.vsync_len;
	info->var.upper_margin = cfg->lcd_cfg.upper_margin;
	info->var.lower_margin = cfg->lcd_cfg.lower_margin;

	info->fix = pdpfb_gfx3_fix;
	info->fix.xpanstep = 16 * 8 / info->var.bits_per_pixel;

	error = pdpfb_str_videomem_alloc(priv, stream);
	if (error)
		goto err0;

	if (pdpfb_str_required_mem(&info->var) > stream->videomem_len &&
		stream->videomem_len != 0) {
		error = -ENOMEM;
		goto err1;
	}

	info->flags = FBINFO_FLAG_DEFAULT
				| FBINFO_HWACCEL_XPAN
				| FBINFO_HWACCEL_YPAN;

#ifndef __PDPFB_USE_OSAL__
	error = register_framebuffer(info);
	if (error < 0)
		goto err1;
#endif

	dev_info(&pdev->dev, "registered graphics 3 framebuffer (len=0x%lx)\n",
			stream->videomem_len);

	return 0;
err1:
	pdpfb_str_videomem_free(priv, stream);
err0:
	return error;
}

static int pdpfb_gfx3_remove(struct pdpfb_priv *priv,
							struct platform_device *pdev)
{
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_GFX3);
#ifndef __PDPFB_USE_OSAL__
	unregister_framebuffer(&stream->info);
#endif

	pdpfb_str_videomem_free(priv, stream);
	return 0;
}

void pdpfb_gfx3_init_stream(struct pdpfb_stream *stream)
{
	*stream = pdpfb_gfx3_stream_template;
}

/* GFX4 Stream */
static int pdpfb_gfx4_probe(struct pdpfb_priv *priv,
							struct platform_device *pdev);
static int pdpfb_gfx4_remove(struct pdpfb_priv *priv,
							struct platform_device *pdev);

static const struct pdpfb_stream pdpfb_gfx4_stream_template = {
	.mem_pool = PDPFB_MEMPOOL_GFX4MEM,
	.videomem_len = 0,
	.enable = 0,
	.pos = 3,
	.ckey = {
		.ckey = 0x00000000,
		.mask = 0x3FFFFFFF,
	},
	.global_alpha = 0x3FF,
	.ops = {
		.probe = pdpfb_gfx4_probe,
		.remove = pdpfb_gfx4_remove,
	},
	.caps = {
		.has_scaler = 0,
		.has_doubler = 1,
		.has_decimator = 1,
		.has_cosited_chroma_control = 0,
		.has_csc = 1,
	},
	.regs = {
		.surf = PDP_GRPH4SURF_OFFSET,
		.blend2_r = PDP_GRPH4BLND2_R_OFFSET,
		.blend2_gb = PDP_GRPH4BLND2_GB_OFFSET,
		.ctrl = PDP_GRPH4CTRL_OFFSET,
		.uctrl = PDPFB_NO_REG,
		.stride = PDP_GRPH4STRIDE_OFFSET,
		.size = PDP_GRPH4SIZE_OFFSET,
		.posn = PDP_GRPH4POSN_OFFSET,
		.galpha = PDP_GRPH4GALPHA_OFFSET,
		.ckey_r = PDP_GRPH4CKEY_R_OFFSET,
		.ckey_gb = PDP_GRPH4CKEY_GB_OFFSET,
		.interleave_ctrl = PDP_GRPH4INTERLEAVE_CTRL_OFFSET,
		.base_addr = PDP_GRPH4BASEADDR_OFFSET,
		.u_base_addr = PDPFB_NO_REG,
		.v_base_addr = PDPFB_NO_REG,
		.memctrl = PDP_GRPH4_MEMCTRL_OFFSET,
		.memthresh = PDP_GRPH4_MEM_THRESH_OFFSET,
		.skip_ctrl = PDPFB_NO_REG,
		.postskip_ctrl = PDPFB_NO_REG,
		.decimate_ctrl = PDP_GRPH4DECIMATE_CTRL_OFFSET,
		.scale_ctrl = PDPFB_NO_REG,
		.vsinit = PDPFB_NO_REG,
		.first_vscoeff = PDPFB_NO_REG,
		.last_vscoeff = PDPFB_NO_REG,
		.hsinit = PDPFB_NO_REG,
		.first_hscoeff = PDPFB_NO_REG,
		.last_hscoeff = PDPFB_NO_REG,
		.scale_size = PDPFB_NO_REG,
		.lumakey_alpha_trans_opaque = PDPFB_NO_REG,
		.lumakey_luma_max_min = PDPFB_NO_REG,
		.lumakey_c_rg = PDPFB_NO_REG,
		.lumakey_c_b = PDPFB_NO_REG,
	}
};

static int pdpfb_gfx4_check_var(struct fb_var_screeninfo *var,
								struct fb_info *info)
{
	struct pdpfb_priv *priv = dev_get_drvdata(info->device);
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_GFX4);
	return pdpfb_str_check_var(priv, stream, var, info);
}

static int pdpfb_gfx4_ioctl(struct fb_info *info, unsigned int cmd,
							unsigned long arg)
{
	struct pdpfb_priv *priv = dev_get_drvdata(info->device);
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_GFX4);
	return pdpfb_str_ioctl(priv, stream, cmd, arg);
}

static int pdpfb_gfx4_pan_display(struct fb_var_screeninfo *var,
								struct fb_info *info)
{
	struct pdpfb_priv *priv = dev_get_drvdata(info->device);
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_GFX4);
	return pdpfb_str_pan_display(priv, stream, var, info);
}

static int pdpfb_gfx4_set_par(struct fb_info *info)
{
	struct pdpfb_priv *priv = dev_get_drvdata(info->device);
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_GFX4);
	return pdpfb_str_set_par(priv, stream, info);
}

static struct fb_fix_screeninfo pdpfb_gfx4_fix	= {
	.id =		"pdp_gfx4",
	.type =		FB_TYPE_PACKED_PIXELS,
	.xpanstep =	8,
	.ypanstep =	1,
	.accel =	FB_ACCEL_IMG_PDP_1,
};

static struct fb_ops pdpfb_gfx4_ops = {
	.fb_blank	= pdpfb_blank,
	.fb_pan_display	= pdpfb_gfx4_pan_display,
	.fb_check_var	= pdpfb_gfx4_check_var,
	.fb_set_par	= pdpfb_gfx4_set_par,
	.fb_ioctl	= pdpfb_gfx4_ioctl,
#ifndef __PDPFB_USE_OSAL__
	.fb_fillrect	= cfb_fillrect,
	.fb_copyarea	= cfb_copyarea,
	.fb_imageblit	= cfb_imageblit,
#endif
};

static int pdpfb_gfx4_probe(struct pdpfb_priv *priv,
							struct platform_device *pdev)
{
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_GFX4);
	struct fb_info *info = &stream->info;
	struct pdpfb_cfg *cfg = pdpfb_get_config(priv);
	int error;

	info->device = &pdev->dev;
	info->fbops = &pdpfb_gfx4_ops;

	info->var.xres = info->var.xres_virtual = cfg->lcd_cfg.xres;
	info->var.yres = info->var.yres_virtual = cfg->lcd_cfg.yres;
	info->var.width = cfg->lcd_size.width;
	info->var.height = cfg->lcd_size.height;
	info->var.activate = FB_ACTIVATE_NOW;
	info->var.bits_per_pixel = cfg->bpp;
	pdpfb_str_match_var(priv, stream, &info->var);

	info->var.hsync_len = cfg->lcd_cfg.hsync_len;
	info->var.left_margin = cfg->lcd_cfg.left_margin;
	info->var.right_margin = cfg->lcd_cfg.right_margin;
	info->var.vsync_len = cfg->lcd_cfg.vsync_len;
	info->var.upper_margin = cfg->lcd_cfg.upper_margin;
	info->var.lower_margin = cfg->lcd_cfg.lower_margin;

	info->fix = pdpfb_gfx4_fix;
	info->fix.xpanstep = 16 * 8 / info->var.bits_per_pixel;

	error = pdpfb_str_videomem_alloc(priv, stream);
	if (error)
		goto err0;

	if (pdpfb_str_required_mem(&info->var) > stream->videomem_len &&
		stream->videomem_len != 0) {
		error = -ENOMEM;
		goto err1;
	}

	info->flags = FBINFO_FLAG_DEFAULT
				| FBINFO_HWACCEL_XPAN
				| FBINFO_HWACCEL_YPAN;

#ifndef __PDPFB_USE_OSAL__
	error = register_framebuffer(info);
	if (error < 0)
		goto err1;
#endif

	dev_info(&pdev->dev, "registered graphics 4 framebuffer (len=0x%lx)\n",
			stream->videomem_len);

	return 0;
err1:
	pdpfb_str_videomem_free(priv, stream);
err0:
	return error;
}

static int pdpfb_gfx4_remove(struct pdpfb_priv *priv,
							struct platform_device *pdev)
{
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_GFX4);
#ifndef __PDPFB_USE_OSAL__
	unregister_framebuffer(&stream->info);
#endif

	pdpfb_str_videomem_free(priv, stream);
	return 0;
}

void pdpfb_gfx4_init_stream(struct pdpfb_stream *stream)
{
	*stream = pdpfb_gfx4_stream_template;
}

/* VID1 Stream */
static int pdpfb_vid1_probe(struct pdpfb_priv *priv,
							struct platform_device *pdev);
static int pdpfb_vid1_remove(struct pdpfb_priv *priv,
							struct platform_device *pdev);

static const struct pdpfb_stream pdpfb_vid1_stream_template = {
	.mem_pool = PDPFB_MEMPOOL_VID1MEM,
	.videomem_len = 0,
	.enable = 0,
	.pos = 4,
	.csc = {
		.enable = 1,
		.cosited = 1,
	},
	.ckey = {
		.ckey = 0x00000000,
		.mask = 0x3FFFFFFF,
	},
	.global_alpha = 0x3FF,
	.ops = {
		.probe = pdpfb_vid1_probe,
		.remove = pdpfb_vid1_remove,
	},
	.caps = {
		.has_scaler = 1,
		.vs_taps = 4,
		.hs_taps = 8,
		.vs_interpolation_points = 16,
		.hs_interpolation_points = 16,
		.max_scaler_input_width = 2560,
		.has_doubler = 0,
		.has_decimator = 1,
		.has_cosited_chroma_control = 1,
		.has_csc = 1,
	},
	.regs = {
		.surf = PDP_VID1SURF_OFFSET,
		.blend2_r = PDP_VID1BLND2_R_OFFSET,
		.blend2_gb = PDP_VID1BLND2_GB_OFFSET,
		.ctrl = PDP_VID1CTRL_OFFSET,
		.uctrl = PDP_VID1UCTRL_OFFSET,
		.stride = PDP_VID1STRIDE_OFFSET,
		.size = PDP_VID1SIZE_OFFSET,
		.posn = PDP_VID1POSN_OFFSET,
		.galpha = PDP_VID1GALPHA_OFFSET,
		.ckey_r = PDP_VID1CKEY_R_OFFSET,
		.ckey_gb = PDP_VID1CKEY_GB_OFFSET,
		.interleave_ctrl = PDP_VID1INTERLEAVE_CTRL_OFFSET,
		.base_addr = PDP_VID1BASEADDR_OFFSET,
		.u_base_addr = PDP_VID1UBASEADDR_OFFSET,
		.v_base_addr = PDP_VID1VBASEADDR_OFFSET,
		.memctrl = PDP_VID1_MEMCTRL_OFFSET,
		.memthresh = PDP_VID1_MEM_THRESH_OFFSET,
		.skip_ctrl = PDP_VID1SKIPCTRL_OFFSET,
		.postskip_ctrl = PDP_VID1POSTSKIPCTRL_OFFSET,
		.decimate_ctrl = PDP_VID1DECIMATE_CTRL_OFFSET,
		.scale_ctrl = PDP_VID1SCALECTRL_OFFSET,
		.vsinit = PDP_VID1VSINIT_OFFSET,
		.first_vscoeff = PDP_VID1VCOEFF0_OFFSET,
		.last_vscoeff = PDP_VID1VCOEFF8_OFFSET,
		.hsinit = PDP_VID1HSINIT_OFFSET,
		.first_hscoeff = PDP_VID1HCOEFF0_OFFSET,
		.last_hscoeff = PDP_VID1HCOEFF16_OFFSET,
		.scale_size = PDP_VID1SCALESIZE_OFFSET,
		.lumakey_alpha_trans_opaque =
				PDP_VID1LUMAKEY_ALPHA_TRANS_OPAQUE_OFFSET,
		.lumakey_luma_max_min = PDP_VID1LUMAKEY_LUMA_MAX_MIN_OFFSET,
		.lumakey_c_rg = PDP_VID1LUMAKEY_C_RG_OFFSET,
		.lumakey_c_b = PDP_VID1LUMAKEY_C_B_OFFSET,
	}
};

static int pdpfb_vid1_check_var(struct fb_var_screeninfo *var,
								struct fb_info *info)
{
	struct pdpfb_priv *priv = dev_get_drvdata(info->device);
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_VID1);
	return pdpfb_str_check_var(priv, stream, var, info);
}

static int pdpfb_vid1_ioctl(struct fb_info *info, unsigned int cmd,
							unsigned long arg)
{
	struct pdpfb_priv *priv = dev_get_drvdata(info->device);
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_VID1);
	void __user *argp = (void __user *)arg;
	struct pdpfb_planar pl;
	struct pdpfb_lumakey lumakey;
	struct pdpfb_geom geom;
	int split_scale;
	int error;

	switch (cmd) {
	/* get planar YUV information */
	case PDPIO_GETPLANAR:
		pdpfb_str_get_planar(stream, &pl);
		if (copy_to_user(argp, &pl,	sizeof(pl)))
			return -EFAULT;
		return 0;
	/* set planar YUV information */
	case PDPIO_SETPLANAR:
		if (copy_from_user(&pl, argp, sizeof(pl)))
			return -EFAULT;
		error = pdpfb_str_set_planar(priv, stream, info, &pl);
		if (copy_to_user(argp, &pl, sizeof(pl)))
			return -EFAULT;
		return error;
	/* get luma key settings */
	case PDPIO_GETLUMAKEY:
		pdpfb_str_get_lumakey(stream, &lumakey);
		if (copy_to_user(argp, &lumakey, sizeof(lumakey)))
			return -EFAULT;
		return 0;
	/* set luma key settings */
	case PDPIO_SETLUMAKEY:
		if (copy_from_user(&lumakey, argp, sizeof(lumakey)))
			return -EFAULT;
		return pdpfb_str_set_lumakey(priv, stream, &lumakey);
	/* get post scale clipping geometry */
	case PDPIO_GETCLIPGEOM:
		pdpfb_str_get_clip_geom(stream, &geom);
		if (copy_to_user(argp, &geom, sizeof(geom)))
			return -EFAULT;
		return 0;
	/* set post scale clipping geometry */
	case PDPIO_SETCLIPGEOM:
		if (copy_from_user(&geom, argp, sizeof(geom)))
			return -EFAULT;
		return pdpfb_str_set_clip_geom(priv, stream, &geom);
	/* get split scaling setting */
	case PDPIO_GETSPLITSCALE:
		split_scale = pdpfb_str_get_split_scale(stream);
		if (copy_to_user(argp, &split_scale, sizeof(split_scale)))
			return -EFAULT;
		return 0;
	/* set split scaling setting */
	case PDPIO_SETSPLITSCALE:
		if (copy_from_user(&split_scale, argp, sizeof(split_scale)))
			return -EFAULT;
		return pdpfb_str_set_split_scale(priv, stream, split_scale);
	default:
		return pdpfb_str_ioctl(priv, stream, cmd, arg);
	}
	return 0;
}

static int pdpfb_vid1_pan_display(struct fb_var_screeninfo *var,
								struct fb_info *info)
{
	struct pdpfb_priv *priv = dev_get_drvdata(info->device);
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_VID1);
	return pdpfb_str_pan_display(priv, stream, var, info);
}

static int pdpfb_vid1_set_par(struct fb_info *info)
{
	struct pdpfb_priv *priv = dev_get_drvdata(info->device);
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_VID1);
	return pdpfb_str_set_par(priv, stream, info);
}

static struct fb_fix_screeninfo pdpfb_vid1_fix	= {
	.id =		"pdp_vid1",
	.type =		FB_TYPE_PACKED_PIXELS,
	.visual =	FB_VISUAL_TRUECOLOR,
	.xpanstep =	2,
	.ypanstep =	2,
	.accel =	FB_ACCEL_IMG_PDP_1,
};

static struct fb_ops pdpfb_vid1_ops = {
	.fb_blank	= pdpfb_blank,
	.fb_pan_display	= pdpfb_vid1_pan_display,
	.fb_check_var	= pdpfb_vid1_check_var,
	.fb_set_par	= pdpfb_vid1_set_par,
	.fb_ioctl	= pdpfb_vid1_ioctl,
#ifndef __PDPFB_USE_OSAL__
	.fb_fillrect	= cfb_fillrect,
	.fb_copyarea	= cfb_copyarea,
	.fb_imageblit	= cfb_imageblit,
#endif
};

static int pdpfb_vid1_probe(struct pdpfb_priv *priv,
							struct platform_device *pdev)
{
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_VID1);
	struct fb_info *info = &stream->info;
	struct pdpfb_cfg *cfg = pdpfb_get_config(priv);
	int error;

	info->device = &pdev->dev;
	info->fbops = &pdpfb_vid1_ops;

	info->var.xres = info->var.xres_virtual = (cfg->lcd_cfg.xres / 2);
	info->var.yres = info->var.yres_virtual = (cfg->lcd_cfg.yres / 2);
	info->var.width = cfg->lcd_size.width;
	info->var.height = cfg->lcd_size.height;
	info->var.activate = FB_ACTIVATE_NOW;
	info->var.nonstd = PDP_PIXFMT_422_Y0VY1U_8888;
	pdpfb_str_match_var(priv, stream, &info->var);

	info->var.hsync_len = cfg->lcd_cfg.hsync_len;
	info->var.left_margin = cfg->lcd_cfg.left_margin;
	info->var.right_margin = cfg->lcd_cfg.right_margin;
	info->var.vsync_len = cfg->lcd_cfg.vsync_len;
	info->var.upper_margin = cfg->lcd_cfg.upper_margin;
	info->var.lower_margin = cfg->lcd_cfg.lower_margin;

	info->fix = pdpfb_vid1_fix;

	error = pdpfb_str_videomem_alloc(priv, stream);
	if (error)
		goto err0;

	if (pdpfb_str_required_mem(&info->var) > stream->videomem_len &&
		stream->videomem_len != 0) {
		error = -ENOMEM;
		goto err1;
	}

	info->flags = FBINFO_FLAG_DEFAULT
				| FBINFO_HWACCEL_XPAN
				| FBINFO_HWACCEL_YPAN;

#ifndef __PDPFB_USE_OSAL__
	error = register_framebuffer(info);
	if (error < 0)
		goto err1;
#endif

	dev_info(&pdev->dev, "registered video 1 framebuffer (len=0x%lx)\n",
			stream->videomem_len);

	return 0;
err1:
	pdpfb_str_videomem_free(priv, stream);
err0:
	return error;
}

static int pdpfb_vid1_remove(struct pdpfb_priv *priv,
							struct platform_device *pdev)
{
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_VID1);
#ifndef __PDPFB_USE_OSAL__
	unregister_framebuffer(&stream->info);
#endif

	pdpfb_str_videomem_free(priv, stream);
	return 0;
}

void pdpfb_vid1_init_stream(struct pdpfb_stream *stream)
{
	*stream = pdpfb_vid1_stream_template;
}

/* VID2 Stream */
static int pdpfb_vid2_probe(struct pdpfb_priv *priv,
							struct platform_device *pdev);
static int pdpfb_vid2_remove(struct pdpfb_priv *priv,
							struct platform_device *pdev);

static const struct pdpfb_stream pdpfb_vid2_stream_template = {
	.mem_pool = PDPFB_MEMPOOL_VID2MEM,
	.videomem_len = 0,
	.enable = 0,
	.pos = 5,
	.csc = {
		.enable = 1,
		.cosited = 1,
	},
	.ckey = {
		.ckey = 0x00000000,
		.mask = 0x3FFFFFFF,
	},
	.global_alpha = 0x3FF,
	.ops = {
		.probe = pdpfb_vid2_probe,
		.remove = pdpfb_vid2_remove,
	},
	.caps = {
		.has_scaler = 1,
		.vs_taps = 4,
		.hs_taps = 8,
		.vs_interpolation_points = 16,
		.hs_interpolation_points = 16,
		.max_scaler_input_width = 2560,
		.has_doubler = 0,
		.has_decimator = 1,
		.has_cosited_chroma_control = 1,
		.has_csc = 1,
	},
	.regs = {
		.surf = PDP_VID2SURF_OFFSET,
		.blend2_r = PDP_VID2BLND2_R_OFFSET,
		.blend2_gb = PDP_VID2BLND2_GB_OFFSET,
		.ctrl = PDP_VID2CTRL_OFFSET,
		.uctrl = PDP_VID2UCTRL_OFFSET,
		.stride = PDP_VID2STRIDE_OFFSET,
		.size = PDP_VID2SIZE_OFFSET,
		.posn = PDP_VID2POSN_OFFSET,
		.galpha = PDP_VID2GALPHA_OFFSET,
		.ckey_r = PDP_VID2CKEY_R_OFFSET,
		.ckey_gb = PDP_VID2CKEY_GB_OFFSET,
		.interleave_ctrl = PDP_VID2INTERLEAVE_CTRL_OFFSET,
		.base_addr = PDP_VID2BASEADDR_OFFSET,
		.u_base_addr = PDP_VID2UBASEADDR_OFFSET,
		.v_base_addr = PDP_VID2VBASEADDR_OFFSET,
		.memctrl = PDP_VID2_MEMCTRL_OFFSET,
		.memthresh = PDP_VID2_MEM_THRESH_OFFSET,
		.skip_ctrl = PDP_VID2SKIPCTRL_OFFSET,
		.postskip_ctrl = PDP_VID2POSTSKIPCTRL_OFFSET,
		.decimate_ctrl = PDP_VID2DECIMATE_CTRL_OFFSET,
		.scale_ctrl = PDP_VID2SCALECTRL_OFFSET,
		.vsinit = PDP_VID2VSINIT_OFFSET,
		.first_vscoeff = PDP_VID2VCOEFF0_OFFSET,
		.last_vscoeff = PDP_VID2VCOEFF8_OFFSET,
		.hsinit = PDP_VID2HSINIT_OFFSET,
		.first_hscoeff = PDP_VID2HCOEFF0_OFFSET,
		.last_hscoeff = PDP_VID2HCOEFF16_OFFSET,
		.scale_size = PDP_VID2SCALESIZE_OFFSET,
		.lumakey_alpha_trans_opaque =
				PDP_VID2LUMAKEY_ALPHA_TRANS_OPAQUE_OFFSET,
		.lumakey_luma_max_min = PDP_VID2LUMAKEY_LUMA_MAX_MIN_OFFSET,
		.lumakey_c_rg = PDP_VID2LUMAKEY_C_RG_OFFSET,
		.lumakey_c_b = PDP_VID2LUMAKEY_C_B_OFFSET,
	}
};

static int pdpfb_vid2_check_var(struct fb_var_screeninfo *var,
								struct fb_info *info)
{
	struct pdpfb_priv *priv = dev_get_drvdata(info->device);
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_VID2);
	return pdpfb_str_check_var(priv, stream, var, info);
}

static int pdpfb_vid2_ioctl(struct fb_info *info, unsigned int cmd,
							unsigned long arg)
{
	struct pdpfb_priv *priv = dev_get_drvdata(info->device);
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_VID2);
	void __user *argp = (void __user *)arg;
	struct pdpfb_planar pl;
	struct pdpfb_lumakey lumakey;
	struct pdpfb_geom geom;
	int split_scale;
	int error;

	switch (cmd) {
	/* get planar YUV information */
	case PDPIO_GETPLANAR:
		pdpfb_str_get_planar(stream, &pl);
		if (copy_to_user(argp, &pl,	sizeof(pl)))
			return -EFAULT;
		return 0;
	/* set planar YUV information */
	case PDPIO_SETPLANAR:
		if (copy_from_user(&pl, argp, sizeof(pl)))
			return -EFAULT;
		error = pdpfb_str_set_planar(priv, stream, info, &pl);
		if (copy_to_user(argp, &pl, sizeof(pl)))
			return -EFAULT;
		return error;
	/* get luma key settings */
	case PDPIO_GETLUMAKEY:
		pdpfb_str_get_lumakey(stream, &lumakey);
		if (copy_to_user(argp, &lumakey, sizeof(lumakey)))
			return -EFAULT;
		return 0;
	/* set luma key settings */
	case PDPIO_SETLUMAKEY:
		if (copy_from_user(&lumakey, argp, sizeof(lumakey)))
			return -EFAULT;
		return pdpfb_str_set_lumakey(priv, stream, &lumakey);
	/* get post scale clipping geometry */
	case PDPIO_GETCLIPGEOM:
		pdpfb_str_get_clip_geom(stream, &geom);
		if (copy_to_user(argp, &geom, sizeof(geom)))
			return -EFAULT;
		return 0;
	/* set post scale clipping geometry */
	case PDPIO_SETCLIPGEOM:
		if (copy_from_user(&geom, argp, sizeof(geom)))
			return -EFAULT;
		return pdpfb_str_set_clip_geom(priv, stream, &geom);
	/* get split scaling setting */
	case PDPIO_GETSPLITSCALE:
		split_scale = pdpfb_str_get_split_scale(stream);
		if (copy_to_user(argp, &split_scale, sizeof(split_scale)))
			return -EFAULT;
		return 0;
	/* set split scaling setting */
	case PDPIO_SETSPLITSCALE:
		if (copy_from_user(&split_scale, argp, sizeof(split_scale)))
			return -EFAULT;
		return pdpfb_str_set_split_scale(priv, stream, split_scale);
	default:
		return pdpfb_str_ioctl(priv, stream, cmd, arg);
	}
	return 0;
}

static int pdpfb_vid2_pan_display(struct fb_var_screeninfo *var,
								struct fb_info *info)
{
	struct pdpfb_priv *priv = dev_get_drvdata(info->device);
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_VID2);
	return pdpfb_str_pan_display(priv, stream, var, info);
}

static int pdpfb_vid2_set_par(struct fb_info *info)
{
	struct pdpfb_priv *priv = dev_get_drvdata(info->device);
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_VID2);
	return pdpfb_str_set_par(priv, stream, info);
}

static struct fb_fix_screeninfo pdpfb_vid2_fix	= {
	.id =		"pdp_vid2",
	.type =		FB_TYPE_PACKED_PIXELS,
	.visual =	FB_VISUAL_TRUECOLOR,
	.xpanstep =	2,
	.ypanstep =	2,
	.accel =	FB_ACCEL_IMG_PDP_1,
};

static struct fb_ops pdpfb_vid2_ops = {
	.fb_blank	= pdpfb_blank,
	.fb_pan_display	= pdpfb_vid2_pan_display,
	.fb_check_var	= pdpfb_vid2_check_var,
	.fb_set_par	= pdpfb_vid2_set_par,
	.fb_ioctl	= pdpfb_vid2_ioctl,
#ifndef __PDPFB_USE_OSAL__
	.fb_fillrect	= cfb_fillrect,
	.fb_copyarea	= cfb_copyarea,
	.fb_imageblit	= cfb_imageblit,
#endif
};

static int pdpfb_vid2_probe(struct pdpfb_priv *priv,
							struct platform_device *pdev)
{
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_VID2);
	struct fb_info *info = &stream->info;
	struct pdpfb_cfg *cfg = pdpfb_get_config(priv);
	int error;

	info->device = &pdev->dev;
	info->fbops = &pdpfb_vid2_ops;

	info->var.xres = info->var.xres_virtual = (cfg->lcd_cfg.xres / 2);
	info->var.yres = info->var.yres_virtual = (cfg->lcd_cfg.yres / 2);
	info->var.width = cfg->lcd_size.width;
	info->var.height = cfg->lcd_size.height;
	info->var.activate = FB_ACTIVATE_NOW;
	info->var.nonstd = PDP_PIXFMT_422_Y0VY1U_8888;
	pdpfb_str_match_var(priv, stream, &info->var);

	info->var.hsync_len = cfg->lcd_cfg.hsync_len;
	info->var.left_margin = cfg->lcd_cfg.left_margin;
	info->var.right_margin = cfg->lcd_cfg.right_margin;
	info->var.vsync_len = cfg->lcd_cfg.vsync_len;
	info->var.upper_margin = cfg->lcd_cfg.upper_margin;
	info->var.lower_margin = cfg->lcd_cfg.lower_margin;

	info->fix = pdpfb_vid2_fix;

	error = pdpfb_str_videomem_alloc(priv, stream);
	if (error)
		goto err0;

	if (pdpfb_str_required_mem(&info->var) > stream->videomem_len &&
		stream->videomem_len != 0) {
		error = -ENOMEM;
		goto err1;
	}

	info->flags = FBINFO_FLAG_DEFAULT
				| FBINFO_HWACCEL_XPAN
				| FBINFO_HWACCEL_YPAN;

#ifndef __PDPFB_USE_OSAL__
	error = register_framebuffer(info);
	if (error < 0)
		goto err1;
#endif

	dev_info(&pdev->dev, "registered video 2 framebuffer (len=0x%lx)\n",
			stream->videomem_len);

	return 0;
err1:
	pdpfb_str_videomem_free(priv, stream);
err0:
	return error;
}

static int pdpfb_vid2_remove(struct pdpfb_priv *priv,
							struct platform_device *pdev)
{
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_VID2);
#ifndef __PDPFB_USE_OSAL__
	unregister_framebuffer(&stream->info);
#endif

	pdpfb_str_videomem_free(priv, stream);
	return 0;
}

void pdpfb_vid2_init_stream(struct pdpfb_stream *stream)
{
	*stream = pdpfb_vid2_stream_template;
}

/* VID3 Stream */
static int pdpfb_vid3_probe(struct pdpfb_priv *priv,
							struct platform_device *pdev);
static int pdpfb_vid3_remove(struct pdpfb_priv *priv,
							struct platform_device *pdev);

static const struct pdpfb_stream pdpfb_vid3_stream_template = {
	.mem_pool = PDPFB_MEMPOOL_VID3MEM,
	.videomem_len = 0,
	.enable = 0,
	.pos = 6,
	.csc = {
		.enable = 1,
		.cosited = 1,
	},
	.ckey = {
		.ckey = 0x00000000,
		.mask = 0x3FFFFFFF,
	},
	.global_alpha = 0x3FF,
	.ops = {
		.probe = pdpfb_vid3_probe,
		.remove = pdpfb_vid3_remove,
	},
	.caps = {
		.has_scaler = 1,
		.vs_taps = 4,
		.hs_taps = 8,
		.vs_interpolation_points = 16,
		.hs_interpolation_points = 16,
		.max_scaler_input_width = 2560,
		.has_doubler = 0,
		.has_decimator = 1,
		.has_cosited_chroma_control = 1,
		.has_csc = 1,
	},
	.regs = {
		.surf = PDP_VID3SURF_OFFSET,
		.blend2_r = PDP_VID3BLND2_R_OFFSET,
		.blend2_gb = PDP_VID3BLND2_GB_OFFSET,
		.ctrl = PDP_VID3CTRL_OFFSET,
		.uctrl = PDP_VID3UCTRL_OFFSET,
		.stride = PDP_VID3STRIDE_OFFSET,
		.size = PDP_VID3SIZE_OFFSET,
		.posn = PDP_VID3POSN_OFFSET,
		.galpha = PDP_VID3GALPHA_OFFSET,
		.ckey_r = PDP_VID3CKEY_R_OFFSET,
		.ckey_gb = PDP_VID3CKEY_GB_OFFSET,
		.interleave_ctrl = PDP_VID3INTERLEAVE_CTRL_OFFSET,
		.base_addr = PDP_VID3BASEADDR_OFFSET,
		.u_base_addr = PDP_VID3UBASEADDR_OFFSET,
		.v_base_addr = PDP_VID3VBASEADDR_OFFSET,
		.memctrl = PDP_VID3_MEMCTRL_OFFSET,
		.memthresh = PDP_VID3_MEM_THRESH_OFFSET,
		.skip_ctrl = PDP_VID3SKIPCTRL_OFFSET,
		.postskip_ctrl = PDP_VID3POSTSKIPCTRL_OFFSET,
		.decimate_ctrl = PDP_VID3DECIMATE_CTRL_OFFSET,
		.scale_ctrl = PDP_VID3SCALECTRL_OFFSET,
		.vsinit = PDP_VID3VSINIT_OFFSET,
		.first_vscoeff = PDP_VID3VCOEFF0_OFFSET,
		.last_vscoeff = PDP_VID3VCOEFF8_OFFSET,
		.hsinit = PDP_VID3HSINIT_OFFSET,
		.first_hscoeff = PDP_VID3HCOEFF0_OFFSET,
		.last_hscoeff = PDP_VID3HCOEFF16_OFFSET,
		.scale_size = PDP_VID3SCALESIZE_OFFSET,
		.lumakey_alpha_trans_opaque =
				PDP_VID3LUMAKEY_ALPHA_TRANS_OPAQUE_OFFSET,
		.lumakey_luma_max_min = PDP_VID3LUMAKEY_LUMA_MAX_MIN_OFFSET,
		.lumakey_c_rg = PDP_VID3LUMAKEY_C_RG_OFFSET,
		.lumakey_c_b = PDP_VID3LUMAKEY_C_B_OFFSET,
	}
};

static int pdpfb_vid3_check_var(struct fb_var_screeninfo *var,
								struct fb_info *info)
{
	struct pdpfb_priv *priv = dev_get_drvdata(info->device);
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_VID3);
	return pdpfb_str_check_var(priv, stream, var, info);
}

static int pdpfb_vid3_ioctl(struct fb_info *info, unsigned int cmd,
							unsigned long arg)
{
	struct pdpfb_priv *priv = dev_get_drvdata(info->device);
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_VID3);
	void __user *argp = (void __user *)arg;
	struct pdpfb_planar pl;
	struct pdpfb_lumakey lumakey;
	struct pdpfb_geom geom;
	int split_scale;
	int error;

	switch (cmd) {
	/* get planar YUV information */
	case PDPIO_GETPLANAR:
		pdpfb_str_get_planar(stream, &pl);
		if (copy_to_user(argp, &pl,	sizeof(pl)))
			return -EFAULT;
		return 0;
	/* set planar YUV information */
	case PDPIO_SETPLANAR:
		if (copy_from_user(&pl, argp, sizeof(pl)))
			return -EFAULT;
		error = pdpfb_str_set_planar(priv, stream, info, &pl);
		if (copy_to_user(argp, &pl, sizeof(pl)))
			return -EFAULT;
		return error;
	/* get luma key settings */
	case PDPIO_GETLUMAKEY:
		pdpfb_str_get_lumakey(stream, &lumakey);
		if (copy_to_user(argp, &lumakey, sizeof(lumakey)))
			return -EFAULT;
		return 0;
	/* set luma key settings */
	case PDPIO_SETLUMAKEY:
		if (copy_from_user(&lumakey, argp, sizeof(lumakey)))
			return -EFAULT;
		return pdpfb_str_set_lumakey(priv, stream, &lumakey);
	/* get post scale clipping geometry */
	case PDPIO_GETCLIPGEOM:
		pdpfb_str_get_clip_geom(stream, &geom);
		if (copy_to_user(argp, &geom, sizeof(geom)))
			return -EFAULT;
		return 0;
	/* set post scale clipping geometry */
	case PDPIO_SETCLIPGEOM:
		if (copy_from_user(&geom, argp, sizeof(geom)))
			return -EFAULT;
		return pdpfb_str_set_clip_geom(priv, stream, &geom);
	/* get split scaling setting */
	case PDPIO_GETSPLITSCALE:
		split_scale = pdpfb_str_get_split_scale(stream);
		if (copy_to_user(argp, &split_scale, sizeof(split_scale)))
			return -EFAULT;
		return 0;
	/* set split scaling setting */
	case PDPIO_SETSPLITSCALE:
		if (copy_from_user(&split_scale, argp, sizeof(split_scale)))
			return -EFAULT;
		return pdpfb_str_set_split_scale(priv, stream, split_scale);
	default:
		return pdpfb_str_ioctl(priv, stream, cmd, arg);
	}
	return 0;
}

static int pdpfb_vid3_pan_display(struct fb_var_screeninfo *var,
								struct fb_info *info)
{
	struct pdpfb_priv *priv = dev_get_drvdata(info->device);
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_VID3);
	return pdpfb_str_pan_display(priv, stream, var, info);
}

static int pdpfb_vid3_set_par(struct fb_info *info)
{
	struct pdpfb_priv *priv = dev_get_drvdata(info->device);
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_VID3);
	return pdpfb_str_set_par(priv, stream, info);
}

static struct fb_fix_screeninfo pdpfb_vid3_fix	= {
	.id =		"pdp_vid3",
	.type =		FB_TYPE_PACKED_PIXELS,
	.visual =	FB_VISUAL_TRUECOLOR,
	.xpanstep =	2,
	.ypanstep =	2,
	.accel =	FB_ACCEL_IMG_PDP_1,
};

static struct fb_ops pdpfb_vid3_ops = {
	.fb_blank	= pdpfb_blank,
	.fb_pan_display	= pdpfb_vid3_pan_display,
	.fb_check_var	= pdpfb_vid3_check_var,
	.fb_set_par	= pdpfb_vid3_set_par,
	.fb_ioctl	= pdpfb_vid3_ioctl,
#ifndef __PDPFB_USE_OSAL__
	.fb_fillrect	= cfb_fillrect,
	.fb_copyarea	= cfb_copyarea,
	.fb_imageblit	= cfb_imageblit,
#endif
};

static int pdpfb_vid3_probe(struct pdpfb_priv *priv,
							struct platform_device *pdev)
{
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_VID3);
	struct fb_info *info = &stream->info;
	struct pdpfb_cfg *cfg = pdpfb_get_config(priv);
	int error;

	info->device = &pdev->dev;
	info->fbops = &pdpfb_vid3_ops;

	info->var.xres = info->var.xres_virtual = (cfg->lcd_cfg.xres / 2);
	info->var.yres = info->var.yres_virtual = (cfg->lcd_cfg.yres / 2);
	info->var.width = cfg->lcd_size.width;
	info->var.height = cfg->lcd_size.height;
	info->var.activate = FB_ACTIVATE_NOW;
	info->var.nonstd = PDP_PIXFMT_422_Y0VY1U_8888;
	pdpfb_str_match_var(priv, stream, &info->var);

	info->var.hsync_len = cfg->lcd_cfg.hsync_len;
	info->var.left_margin = cfg->lcd_cfg.left_margin;
	info->var.right_margin = cfg->lcd_cfg.right_margin;
	info->var.vsync_len = cfg->lcd_cfg.vsync_len;
	info->var.upper_margin = cfg->lcd_cfg.upper_margin;
	info->var.lower_margin = cfg->lcd_cfg.lower_margin;

	info->fix = pdpfb_vid3_fix;

	error = pdpfb_str_videomem_alloc(priv, stream);
	if (error)
		goto err0;

	if (pdpfb_str_required_mem(&info->var) > stream->videomem_len &&
		stream->videomem_len != 0) {
		error = -ENOMEM;
		goto err1;
	}

	info->flags = FBINFO_FLAG_DEFAULT
				| FBINFO_HWACCEL_XPAN
				| FBINFO_HWACCEL_YPAN;

#ifndef __PDPFB_USE_OSAL__
	error = register_framebuffer(info);
	if (error < 0)
		goto err1;
#endif

	dev_info(&pdev->dev, "registered video 3 framebuffer (len=0x%lx)\n",
			stream->videomem_len);

	return 0;
err1:
	pdpfb_str_videomem_free(priv, stream);
err0:
	return error;
}

static int pdpfb_vid3_remove(struct pdpfb_priv *priv,
							struct platform_device *pdev)
{
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_VID3);
#ifndef __PDPFB_USE_OSAL__
	unregister_framebuffer(&stream->info);
#endif

	pdpfb_str_videomem_free(priv, stream);
	return 0;
}

void pdpfb_vid3_init_stream(struct pdpfb_stream *stream)
{
	*stream = pdpfb_vid3_stream_template;
}

/* VID4 Stream */
static int pdpfb_vid4_probe(struct pdpfb_priv *priv,
							struct platform_device *pdev);
static int pdpfb_vid4_remove(struct pdpfb_priv *priv,
							struct platform_device *pdev);

static const struct pdpfb_stream pdpfb_vid4_stream_template = {
	.mem_pool = PDPFB_MEMPOOL_VID4MEM,
	.videomem_len = 0,
	.enable = 0,
	.pos = 7,
	.csc = {
		.enable = 1,
		.cosited = 1,
	},
	.ckey = {
		.ckey = 0x00000000,
		.mask = 0x3FFFFFFF,
	},
	.global_alpha = 0x3FF,
	.ops = {
		.probe = pdpfb_vid4_probe,
		.remove = pdpfb_vid4_remove,
	},
	.caps = {
		.has_scaler = 1,
		.vs_taps = 4,
		.hs_taps = 8,
		.vs_interpolation_points = 16,
		.hs_interpolation_points = 16,
		.max_scaler_input_width = 2560,
		.has_doubler = 0,
		.has_decimator = 1,
		.has_cosited_chroma_control = 1,
		.has_csc = 1,
	},
	.regs = {
		.surf = PDP_VID4SURF_OFFSET,
		.blend2_r = PDP_VID4BLND2_R_OFFSET,
		.blend2_gb = PDP_VID4BLND2_GB_OFFSET,
		.ctrl = PDP_VID4CTRL_OFFSET,
		.uctrl = PDP_VID4UCTRL_OFFSET,
		.stride = PDP_VID4STRIDE_OFFSET,
		.size = PDP_VID4SIZE_OFFSET,
		.posn = PDP_VID4POSN_OFFSET,
		.galpha = PDP_VID4GALPHA_OFFSET,
		.ckey_r = PDP_VID4CKEY_R_OFFSET,
		.ckey_gb = PDP_VID4CKEY_GB_OFFSET,
		.interleave_ctrl = PDP_VID4INTERLEAVE_CTRL_OFFSET,
		.base_addr = PDP_VID4BASEADDR_OFFSET,
		.u_base_addr = PDP_VID4UBASEADDR_OFFSET,
		.v_base_addr = PDP_VID4VBASEADDR_OFFSET,
		.memctrl = PDP_VID4_MEMCTRL_OFFSET,
		.memthresh = PDP_VID4_MEM_THRESH_OFFSET,
		.skip_ctrl = PDP_VID4SKIPCTRL_OFFSET,
		.postskip_ctrl = PDP_VID4POSTSKIPCTRL_OFFSET,
		.decimate_ctrl = PDP_VID4DECIMATE_CTRL_OFFSET,
		.scale_ctrl = PDP_VID4SCALECTRL_OFFSET,
		.vsinit = PDP_VID4VSINIT_OFFSET,
		.first_vscoeff = PDP_VID4VCOEFF0_OFFSET,
		.last_vscoeff = PDP_VID4VCOEFF8_OFFSET,
		.hsinit = PDP_VID4HSINIT_OFFSET,
		.first_hscoeff = PDP_VID4HCOEFF0_OFFSET,
		.last_hscoeff = PDP_VID4HCOEFF16_OFFSET,
		.scale_size = PDP_VID4SCALESIZE_OFFSET,
		.lumakey_alpha_trans_opaque =
				PDP_VID4LUMAKEY_ALPHA_TRANS_OPAQUE_OFFSET,
		.lumakey_luma_max_min = PDP_VID4LUMAKEY_LUMA_MAX_MIN_OFFSET,
		.lumakey_c_rg = PDP_VID4LUMAKEY_C_RG_OFFSET,
		.lumakey_c_b = PDP_VID4LUMAKEY_C_B_OFFSET,
	}
};

static int pdpfb_vid4_check_var(struct fb_var_screeninfo *var,
								struct fb_info *info)
{
	struct pdpfb_priv *priv = dev_get_drvdata(info->device);
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_VID4);
	return pdpfb_str_check_var(priv, stream, var, info);
}

static int pdpfb_vid4_ioctl(struct fb_info *info, unsigned int cmd,
							unsigned long arg)
{
	struct pdpfb_priv *priv = dev_get_drvdata(info->device);
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_VID4);
	void __user *argp = (void __user *)arg;
	struct pdpfb_planar pl;
	struct pdpfb_lumakey lumakey;
	struct pdpfb_geom geom;
	int split_scale;
	int error;

	switch (cmd) {
	/* get planar YUV information */
	case PDPIO_GETPLANAR:
		pdpfb_str_get_planar(stream, &pl);
		if (copy_to_user(argp, &pl,	sizeof(pl)))
			return -EFAULT;
		return 0;
	/* set planar YUV information */
	case PDPIO_SETPLANAR:
		if (copy_from_user(&pl, argp, sizeof(pl)))
			return -EFAULT;
		error = pdpfb_str_set_planar(priv, stream, info, &pl);
		if (copy_to_user(argp, &pl, sizeof(pl)))
			return -EFAULT;
		return error;
	/* get luma key settings */
	case PDPIO_GETLUMAKEY:
		pdpfb_str_get_lumakey(stream, &lumakey);
		if (copy_to_user(argp, &lumakey, sizeof(lumakey)))
			return -EFAULT;
		return 0;
	/* set luma key settings */
	case PDPIO_SETLUMAKEY:
		if (copy_from_user(&lumakey, argp, sizeof(lumakey)))
			return -EFAULT;
		return pdpfb_str_set_lumakey(priv, stream, &lumakey);
	/* get post scale clipping geometry */
	case PDPIO_GETCLIPGEOM:
		pdpfb_str_get_clip_geom(stream, &geom);
		if (copy_to_user(argp, &geom, sizeof(geom)))
			return -EFAULT;
		return 0;
	/* set post scale clipping geometry */
	case PDPIO_SETCLIPGEOM:
		if (copy_from_user(&geom, argp, sizeof(geom)))
			return -EFAULT;
		return pdpfb_str_set_clip_geom(priv, stream, &geom);
	/* get split scaling setting */
	case PDPIO_GETSPLITSCALE:
		split_scale = pdpfb_str_get_split_scale(stream);
		if (copy_to_user(argp, &split_scale, sizeof(split_scale)))
			return -EFAULT;
		return 0;
	/* set split scaling setting */
	case PDPIO_SETSPLITSCALE:
		if (copy_from_user(&split_scale, argp, sizeof(split_scale)))
			return -EFAULT;
		return pdpfb_str_set_split_scale(priv, stream, split_scale);
	default:
		return pdpfb_str_ioctl(priv, stream, cmd, arg);
	}
	return 0;
}

static int pdpfb_vid4_pan_display(struct fb_var_screeninfo *var,
								struct fb_info *info)
{
	struct pdpfb_priv *priv = dev_get_drvdata(info->device);
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_VID4);
	return pdpfb_str_pan_display(priv, stream, var, info);
}

static int pdpfb_vid4_set_par(struct fb_info *info)
{
	struct pdpfb_priv *priv = dev_get_drvdata(info->device);
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_VID4);
	return pdpfb_str_set_par(priv, stream, info);
}

static struct fb_fix_screeninfo pdpfb_vid4_fix	= {
	.id =		"pdp_VID4",
	.type =		FB_TYPE_PACKED_PIXELS,
	.visual =	FB_VISUAL_TRUECOLOR,
	.xpanstep =	2,
	.ypanstep =	2,
	.accel =	FB_ACCEL_IMG_PDP_1,
};

static struct fb_ops pdpfb_vid4_ops = {
	.fb_blank	= pdpfb_blank,
	.fb_pan_display	= pdpfb_vid4_pan_display,
	.fb_check_var	= pdpfb_vid4_check_var,
	.fb_set_par	= pdpfb_vid4_set_par,
	.fb_ioctl	= pdpfb_vid4_ioctl,
#ifndef __PDPFB_USE_OSAL__
	.fb_fillrect	= cfb_fillrect,
	.fb_copyarea	= cfb_copyarea,
	.fb_imageblit	= cfb_imageblit,
#endif
};

static int pdpfb_vid4_probe(struct pdpfb_priv *priv,
							struct platform_device *pdev)
{
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_VID4);
	struct fb_info *info = &stream->info;
	struct pdpfb_cfg *cfg = pdpfb_get_config(priv);
	int error;

	info->device = &pdev->dev;
	info->fbops = &pdpfb_vid4_ops;

	info->var.xres = info->var.xres_virtual = (cfg->lcd_cfg.xres / 2);
	info->var.yres = info->var.yres_virtual = (cfg->lcd_cfg.yres / 2);
	info->var.width = cfg->lcd_size.width;
	info->var.height = cfg->lcd_size.height;
	info->var.activate = FB_ACTIVATE_NOW;
	info->var.nonstd = PDP_PIXFMT_422_Y0VY1U_8888;
	pdpfb_str_match_var(priv, stream, &info->var);

	info->var.hsync_len = cfg->lcd_cfg.hsync_len;
	info->var.left_margin = cfg->lcd_cfg.left_margin;
	info->var.right_margin = cfg->lcd_cfg.right_margin;
	info->var.vsync_len = cfg->lcd_cfg.vsync_len;
	info->var.upper_margin = cfg->lcd_cfg.upper_margin;
	info->var.lower_margin = cfg->lcd_cfg.lower_margin;

	info->fix = pdpfb_vid4_fix;

	error = pdpfb_str_videomem_alloc(priv, stream);
	if (error)
		goto err0;

	if (pdpfb_str_required_mem(&info->var) > stream->videomem_len &&
		stream->videomem_len != 0) {
		error = -ENOMEM;
		goto err1;
	}

	info->flags = FBINFO_FLAG_DEFAULT
				| FBINFO_HWACCEL_XPAN
				| FBINFO_HWACCEL_YPAN;

#ifndef __PDPFB_USE_OSAL__
	error = register_framebuffer(info);
	if (error < 0)
		goto err1;
#endif

	dev_info(&pdev->dev, "registered video 4 framebuffer (len=0x%lx)\n",
			stream->videomem_len);

	return 0;
err1:
	pdpfb_str_videomem_free(priv, stream);
err0:
	return error;
}

static int pdpfb_vid4_remove(struct pdpfb_priv *priv,
							struct platform_device *pdev)
{
	struct pdpfb_stream *stream = pdpfb_get_stream(priv, PDPFB_STREAM_VID4);
#ifndef __PDPFB_USE_OSAL__
	unregister_framebuffer(&stream->info);
#endif

	pdpfb_str_videomem_free(priv, stream);
	return 0;
}

void pdpfb_vid4_init_stream(struct pdpfb_stream *stream)
{
	*stream = pdpfb_vid4_stream_template;
}
