/*
 * Cashew event timer driver
 *
 * Copyright (C) 2016 Imagination Technologies Ltd.
 *
 * Author: Sifan Naeem <sifan.naeem@imgtec.com>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 */

#include <linux/clk.h>
#include <linux/clk-provider.h>
#include <linux/clocksource.h>
#include <linux/delay.h>
#include <linux/hrtimer.h>
#include <linux/interrupt.h>
#include <linux/io.h>
#include <linux/irq.h>
#include <linux/kernel.h>
#include <linux/list.h>
#include <linux/module.h>
#include <linux/of.h>
#include <linux/of_address.h>
#include <linux/of_device.h>
#include <linux/of_irq.h>
#include <linux/platform_device.h>
#include <linux/regmap.h>
#include <linux/slab.h>
#include <linux/spinlock.h>
#include <linux/timecounter.h>

#include <linux/mfd/syscon.h>

#include "cashew-event-timer.h"
#include "cashew-event-timer-internal.h"

#define	CASHEW_EVT_FIFO_DEPTH			16

/* Event Time Stamp Registers */
#define CASHEW_EVT_COUNTER			0x0
#define CASHEW_EVT_COUNTER_MASK			0x3fffffff
#define CASHEW_EVT_COUNTER_NOT_SYS_CLK_MASK	BIT(30)
#define CASHEW_EVT_COUNTER_ENABLE_MASK		BIT(31)

#define CASHEW_EVT_TIMESTAMP_STS		0x4

#define CASHEW_EVT_TIMESTAMP_CLR		0x8

#define CASHEW_EVT_CLKSRC_SELECT		0xc
#define CASHEW_EVT_CLKSRC_SELECT_SHIFT		0
/* We only use 2 refclocks upper 7 bits are ignored. */
#define CASHEW_EVT_CLKSRC_SELECT_WIDTH		1

#define CASHEW_EVT_SOURCE_START			0x14

#define CASHEW_EVT_TIME_STAMP_MASK		0x3fffffff

/* Event Timer Registers */
#define CASHEW_EVT_TIMER_ENABLE			0x100
#define CASHEW_EVT_TIMER_ENABLE_MASK		0x1
#define CASHEW_EVT_TIMER_SOFT_RESET_MASK	0x2

#define CASHEW_EVT_SOURCES			0x108
#define CASHEW_EVT_SOURCES_SHIFT		16
#define	CASHEW_EVT_SOURCES_MASK_LSB		0xffffUL

#define CASHEW_EVT_PHASE_FIFO			0x110

#define	CASHEW_EVT_SAMPLE_FIFO(id)		(0x114 + ((id) * 0x4))

#define CASHEW_EVT_EVENT_CTL			0x120
#define CASHEW_EVT_EVENT_CTL_MASK		0x3
#define CASHEW_EVT_EVENT_CTL_WIDTH		2

#define CASHEW_EVT_TIME_REG(en)			(0x130 + ((en) * 0x4))

#define CASHEW_EVT_INT_STATUS			0x170

#define CASHEW_EVT_INT_ENABLE			0x174

#define CASHEW_EVT_INT_CLEAR			0x178

#define CASHEW_EVT_ET_EVENTS(index)		(0x190 + 0x4 * ((index) / 4))
#define CASHEW_EVT_ET_EVENTS_SHIFT(index)	(8 * ((index) % 4))
#define CASHEW_EVT_ET_EVENTS_MASK		0xff

#define	CASHEW_EVT_INT_SAMPLE_0_FNE_MASK	BIT(5)
#define	CASHEW_EVT_INT_SAMPLE_1_FNE_MASK	BIT(9)
#define	CASHEW_EVT_INT_PHASE_FNE_MASK		BIT(1)

#define	CASHEW_EVT_MIN_EVENT_DELTA_NS		10000

static LIST_HEAD(cashew_evt_list);
static DEFINE_SPINLOCK(cashew_evt_list_spinlock);

static inline u32 cashew_evt_readl(struct cashew_evt *evt, u32 reg)
{
	return readl(evt->base + reg);
}

static inline void cashew_evt_writel(struct cashew_evt *evt, u32 val, u32 reg)
{
	writel(val, evt->base + reg);
}

static inline void cashew_evt_soft_reset(struct cashew_evt *evt)
{
	u32 reg = cashew_evt_readl(evt, CASHEW_EVT_TIMER_ENABLE);

	reg |= CASHEW_EVT_TIMER_SOFT_RESET_MASK;
	cashew_evt_writel(evt, reg, CASHEW_EVT_TIMER_ENABLE);
}

static inline void cashew_evt_stop_count(struct cashew_evt *evt)
{
	u32 reg = cashew_evt_readl(evt, CASHEW_EVT_COUNTER);

	reg &= ~CASHEW_EVT_COUNTER_ENABLE_MASK;
	cashew_evt_writel(evt, reg, CASHEW_EVT_COUNTER);
}

static inline void cashew_evt_start_count(struct cashew_evt *evt)
{
	u32 reg = cashew_evt_readl(evt, CASHEW_EVT_COUNTER);

	reg |= CASHEW_EVT_COUNTER_ENABLE_MASK;
	cashew_evt_writel(evt, reg, CASHEW_EVT_COUNTER);
}

static inline u32 cashew_evt_get_count(struct cashew_evt *evt)
{
	u32 reg = cashew_evt_readl(evt, CASHEW_EVT_COUNTER);

	return reg & CASHEW_EVT_COUNTER_MASK;
}

static cycle_t cashew_evt_cc_read(const struct cyclecounter *cc)
{
	struct cashew_evt *evt;

	evt = container_of(cc, struct cashew_evt, cc);

	return (cycle_t)cashew_evt_get_count(evt);
}

void cashew_evt_get_time_ts(struct cashew_evt *evt, struct timespec *ts)
{
	u64 tmp;

	tmp = cashew_evt_get_time(evt);
	ts->tv_sec = do_div(tmp, NSEC_PER_SEC);
	ts->tv_nsec = tmp;
}
EXPORT_SYMBOL_GPL(cashew_evt_get_time_ts);

static inline bool cashew_evt_bad_event(enum cashew_evt_enable event)
{
	switch (event) {
	case CASHEW_EVT_ENABLE_ETS:
	case CASHEW_EVT_ENABLE_PARALLEL_IN:
	case CASHEW_EVT_ENABLE_I2S_OUT:
	case CASHEW_EVT_ENABLE_SPDIF_OUT:
		return false;
	default:
		return true;
	}
}

static struct cashew_evt_callback *cashew_evt_get_next_trigger(
		struct cashew_evt *evt, u64 *p_next_trigger)
{
	u64 next_trigger, tmp;
	int i;
	struct cashew_evt_callback *cbr = NULL, *cb;

	cb = &evt->trigger_cbs[0];
	next_trigger = ULLONG_MAX;

	for (i = 0; i < CASHEW_EVT_NUM_ENABLES; i++, cb++) {
		if (!cashew_evt_bad_event(i)) {
			tmp = cb->trigger_time;
			if (tmp && (tmp < next_trigger)) {
				next_trigger = tmp;
				cbr = cb;
			}
		}
	}

	*p_next_trigger = next_trigger;

	return cbr;
}

struct cashew_evt *cashew_evt_get(struct device_node *np)
{
	struct cashew_evt *evt, *ret = ERR_PTR(-EPROBE_DEFER);

	spin_lock(&cashew_evt_list_spinlock);
	list_for_each_entry(evt, &cashew_evt_list, list) {
		if (evt->np == np) {
			ret = evt;
			break;
		}
	}
	spin_unlock(&cashew_evt_list_spinlock);

	return ret;
}
EXPORT_SYMBOL_GPL(cashew_evt_get);

void _cashew_evt_disable_event(struct cashew_evt *evt,
			       enum cashew_evt_enable event)
{
	u32 reg;

	dev_dbg(evt->dev, "Disable event %u\n", (unsigned int)event);

	if (cashew_evt_bad_event(event)) {
		dev_err(evt->dev, "Disable event %u failed (bad event %u)\n",
			(unsigned int)event, (unsigned int)event);
		return;
	}

	reg = cashew_evt_readl(evt, CASHEW_EVT_EVENT_CTL);

	reg &= ~(CASHEW_EVT_EVENT_CTL_MASK <<
		 (CASHEW_EVT_EVENT_CTL_WIDTH * event));

	cashew_evt_writel(evt, reg, CASHEW_EVT_EVENT_CTL);

	evt->trigger_cbs[event].trigger_time = 0;
}
EXPORT_SYMBOL_GPL(_cashew_evt_disable_event);

void cashew_evt_disable_event(struct cashew_evt *evt,
			      enum cashew_evt_enable event)
{
	unsigned long flags;

	spin_lock_irqsave(&evt->lock, flags);
	_cashew_evt_disable_event(evt, event);
	spin_unlock_irqrestore(&evt->lock, flags);
}
EXPORT_SYMBOL_GPL(cashew_evt_disable_event);

int cashew_evt_set_event(struct cashew_evt *evt, enum cashew_evt_enable event,
		enum cashew_evt_type type, struct timespec *ts,
		void (*event_trigger_callback)(struct cashew_evt *, void *),
		void *context)
{
	u32 reg, cyc, event_reg_addr, irq_reg_addr;
	u64 trigger_time, next_trigger;
	unsigned long flags;
	struct cashew_evt_callback *cb;
	int ret;

	dev_dbg(evt->dev, "Set event %u type %u time %u,%ld\n",
		(unsigned int)event, (unsigned int)type,
		(unsigned int)ts->tv_sec, ts->tv_nsec);

	if (cashew_evt_bad_event(event)) {
		dev_err(evt->dev, "Set event %u failed (bad event %u)\n",
			(unsigned int)event, (unsigned int)event);
		return -EINVAL;
	}

	if (type != CASHEW_EVT_TYPE_LEVEL && type != CASHEW_EVT_TYPE_PULSE) {
		dev_err(evt->dev, "Set event %u failed (bad event type %u)\n",
			(unsigned int)event, (unsigned int)type);
		return -EINVAL;
	}

	if (!ts) {
		dev_err(evt->dev, "Set event %u failed (ts == NULL)\n",
			(unsigned int)event);
		return -EINVAL;
	}

	event_reg_addr = CASHEW_EVT_TIME_REG(event);
	irq_reg_addr = CASHEW_EVT_TIME_REG(CASHEW_EVT_ENABLE_IRQ_0);

	trigger_time = (u64)ts->tv_sec * NSEC_PER_SEC + ts->tv_nsec;

	spin_lock_irqsave(&evt->lock, flags);

	/* Trigger already pending for this event? */
	if (evt->trigger_cbs[event].trigger_time) {
		spin_unlock_irqrestore(&evt->lock, flags);
		dev_err(evt->dev, "Set event %u failed (trigger already pending at %lluns)\n",
			(unsigned int)event,
			evt->trigger_cbs[event].trigger_time);
		return -EINVAL;
	}

	reg = cashew_evt_readl(evt, CASHEW_EVT_EVENT_CTL);

	/*
	 * This event may have triggered previously. The control bits need to
	 * be cleared before programming a new trigger
	 */
	reg &= ~(CASHEW_EVT_EVENT_CTL_MASK <<
		 (CASHEW_EVT_EVENT_CTL_WIDTH * event));

	cashew_evt_writel(evt, reg, CASHEW_EVT_EVENT_CTL);

	ret = cashew_evt_time_to_reg(evt, trigger_time, &cyc,
				     CASHEW_EVT_MIN_EVENT_DELTA_NS);
	if (ret) {
		spin_unlock_irqrestore(&evt->lock, flags);
		dev_err(evt->dev, "Set event %u failed (%d)\n",
			(unsigned int)event, ret);
		return ret;
	}

	cashew_evt_writel(evt, cyc, event_reg_addr);

	cb = cashew_evt_get_next_trigger(evt, &next_trigger);

	reg |= (type << (CASHEW_EVT_EVENT_CTL_WIDTH * event));

	/*
	 * No irq trigger currently set or the new trigger time is
	 * earlier than the next trigger time?
	 */
	if (!cb || (next_trigger > trigger_time)) {
		cashew_evt_writel(evt, cyc, irq_reg_addr);
		reg |= CASHEW_EVT_TYPE_LEVEL <<
			(CASHEW_EVT_EVENT_CTL_WIDTH *
			 CASHEW_EVT_ENABLE_IRQ_0);
	}

	cashew_evt_writel(evt, reg, CASHEW_EVT_EVENT_CTL);

	evt->trigger_cbs[event].callback = event_trigger_callback;
	evt->trigger_cbs[event].trigger_time = trigger_time;
	evt->trigger_cbs[event].cyc = cyc;
	evt->trigger_cbs[event].context = context;

	spin_unlock_irqrestore(&evt->lock, flags);

	return 0;
}
EXPORT_SYMBOL_GPL(cashew_evt_set_event);

static bool cashew_evt_retrigger(struct cashew_evt *evt,
				 struct cashew_evt_callback *cb)
{
	u32 reg, trig_reg_addr;
	u64 cur_time;

	trig_reg_addr = CASHEW_EVT_TIME_REG(CASHEW_EVT_ENABLE_IRQ_0);

	reg = cashew_evt_readl(evt, CASHEW_EVT_EVENT_CTL);

	reg |= (CASHEW_EVT_TYPE_LEVEL <<
		(CASHEW_EVT_EVENT_CTL_WIDTH * CASHEW_EVT_ENABLE_IRQ_0));

	cashew_evt_writel(evt, cb->cyc, trig_reg_addr);
	cashew_evt_writel(evt, reg, CASHEW_EVT_EVENT_CTL);

	cur_time = _cashew_evt_get_time(evt);

	/* Trigger passed while writing? */
	if (cb->trigger_time < cur_time)
		return false;

	return true;
}

static irqreturn_t cashew_evt_trigger_0_irq(int irq, void *dev_id)
{
	struct cashew_evt *evt = (struct cashew_evt *)dev_id;
	u64 next_trigger, cur_time;
	struct cashew_evt_callback *cb;
	unsigned long flags;
	u32 reg;

	dev_dbg(evt->dev, "Trigger IRQ\n");

	spin_lock_irqsave(&evt->lock, flags);

	while (1) {
		cb = cashew_evt_get_next_trigger(evt, &next_trigger);

		/* Disable the irq trigger */
		reg = cashew_evt_readl(evt, CASHEW_EVT_EVENT_CTL);
		reg &= ~(CASHEW_EVT_EVENT_CTL_MASK <<
			 (CASHEW_EVT_EVENT_CTL_WIDTH *
			  CASHEW_EVT_ENABLE_IRQ_0));
		cashew_evt_writel(evt, reg, CASHEW_EVT_EVENT_CTL);

		if (!cb)
			goto loop_exit;

		cur_time = _cashew_evt_get_time(evt);

		if (cur_time >= next_trigger) {
			if (cb->callback)
				cb->callback(evt, cb->context);
			cb->trigger_time = 0;
		} else if (cashew_evt_retrigger(evt, cb)) {
			goto loop_exit;
		} else {
			if (cb->callback)
				cb->callback(evt, cb->context);
			cb->trigger_time = 0;
		}
	}

loop_exit:
	spin_unlock_irqrestore(&evt->lock, flags);

	return IRQ_HANDLED;
}

int cashew_evt_set_source(struct cashew_evt *evt,
			  int id, enum cashew_evt_source source)
{
	unsigned long flags;
	u32 reg;

	if (id >= CASHEW_EVT_MAX_SOURCES || source >= CASHEW_EVT_NUM_SOURCES)
		return -EINVAL;

	spin_lock_irqsave(&evt->lock, flags);

	reg = cashew_evt_readl(evt, CASHEW_EVT_SOURCES);
	reg &= ~(CASHEW_EVT_SOURCES_MASK_LSB <<
		 (id * CASHEW_EVT_SOURCES_SHIFT));
	reg |= source << (id * CASHEW_EVT_SOURCES_SHIFT);
	cashew_evt_writel(evt, reg, CASHEW_EVT_SOURCES);

	/*
	 * Changing one of the sources invalidates the active sample rate
	 * measurement for the source in question, and the active phase
	 * difference measurement, so reset these states and mask the
	 * interrupts
	 */
	evt->sample_rates[id].state = CASHEW_EVT_STATE_IDLE;
	evt->phase_difference.state = CASHEW_EVT_STATE_IDLE;

	reg = cashew_evt_readl(evt, CASHEW_EVT_INT_ENABLE);
	if (id == 0)
		reg &= ~CASHEW_EVT_INT_SAMPLE_0_FNE_MASK;
	else
		reg &= ~CASHEW_EVT_INT_SAMPLE_1_FNE_MASK;
	reg &= ~CASHEW_EVT_INT_PHASE_FNE_MASK;
	cashew_evt_writel(evt, reg, CASHEW_EVT_INT_ENABLE);

	spin_unlock_irqrestore(&evt->lock, flags);

	return 0;
}
EXPORT_SYMBOL_GPL(cashew_evt_set_source);

int cashew_evt_get_source(struct cashew_evt *evt,
			  int id, enum cashew_evt_source *source)
{
	u32 reg;

	if (id >= CASHEW_EVT_MAX_SOURCES)
		return -EINVAL;

	reg = cashew_evt_readl(evt, CASHEW_EVT_SOURCES);

	*source = (reg >> (id * CASHEW_EVT_SOURCES_SHIFT)) &
		  CASHEW_EVT_SOURCES_MASK_LSB;

	return 0;
}
EXPORT_SYMBOL_GPL(cashew_evt_get_source);

static void cashew_evt_clear_fifo(struct cashew_evt *evt, u32 fifo_offset,
				  u32 mask, bool enable_int)
{
	u32 reg;

	reg = cashew_evt_readl(evt, CASHEW_EVT_INT_ENABLE);
	if (enable_int)
		reg |= mask;
	else
		reg &= ~mask;
	cashew_evt_writel(evt, reg, CASHEW_EVT_INT_ENABLE);

	reg = cashew_evt_readl(evt, CASHEW_EVT_INT_STATUS);
	while (reg & mask) {
		reg = cashew_evt_readl(evt, fifo_offset);
		cashew_evt_writel(evt, mask, CASHEW_EVT_INT_CLEAR);
		cashew_evt_writel(evt, 0, CASHEW_EVT_INT_CLEAR);
		reg = cashew_evt_readl(evt, CASHEW_EVT_INT_STATUS);
	}
}

static void cashew_evt_new_sr(struct cashew_evt *evt, int id, u32 mask)
{
	u32 reg;
	enum cashew_evt_state new_state;
	struct cashew_evt_measurement *sr = &evt->sample_rates[id];

	switch (sr->state) {
	case CASHEW_EVT_STATE_ACTIVE_FIRST:
		/* First sample rate measurement is always invalid */
		reg = cashew_evt_readl(evt, CASHEW_EVT_SAMPLE_FIFO(id));
		cashew_evt_writel(evt, mask, CASHEW_EVT_INT_CLEAR);
		cashew_evt_writel(evt, 0, CASHEW_EVT_INT_CLEAR);
		reg = cashew_evt_readl(evt, CASHEW_EVT_INT_STATUS);
		if (reg & mask)
			new_state = CASHEW_EVT_STATE_COMPLETE;
		else
			new_state = CASHEW_EVT_STATE_ACTIVE_SECOND;
		break;

	case CASHEW_EVT_STATE_ACTIVE_SECOND:
		new_state = CASHEW_EVT_STATE_COMPLETE;
		break;

	default:
		dev_err(evt->dev, "cashew_evt_new_sr bad state (%d)\n",
			(int)sr->state);
		return;
	}

	if (new_state == CASHEW_EVT_STATE_COMPLETE) {
		reg = cashew_evt_readl(evt, CASHEW_EVT_INT_ENABLE);
		reg &= ~mask;
		cashew_evt_writel(evt, reg, CASHEW_EVT_INT_ENABLE);
		cashew_evt_writel(evt, mask, CASHEW_EVT_INT_CLEAR);
		cashew_evt_writel(evt, 0, CASHEW_EVT_INT_CLEAR);
		if (sr->callback)
			sr->callback(sr->context);
	}

	sr->state = new_state;
}

static void cashew_evt_new_pd(struct cashew_evt *evt)
{
	u32 reg;
	enum cashew_evt_state new_state;
	u32 mask = CASHEW_EVT_INT_PHASE_FNE_MASK;
	struct cashew_evt_measurement *pd = &evt->phase_difference;

	switch (pd->state) {
	case CASHEW_EVT_STATE_ACTIVE_FIRST:
		/* First two phase measurements are always invalid */
		reg = cashew_evt_readl(evt, CASHEW_EVT_PHASE_FIFO);
		cashew_evt_writel(evt, mask, CASHEW_EVT_INT_CLEAR);
		cashew_evt_writel(evt, 0, CASHEW_EVT_INT_CLEAR);
		reg = cashew_evt_readl(evt, CASHEW_EVT_INT_STATUS);
		if (!(reg & mask)) {
			new_state = CASHEW_EVT_STATE_ACTIVE_SECOND;
			break;
		}
		/* Fall through */
	case CASHEW_EVT_STATE_ACTIVE_SECOND:
		/* First two phase measurements are always invalid */
		reg = cashew_evt_readl(evt, CASHEW_EVT_PHASE_FIFO);
		cashew_evt_writel(evt, mask, CASHEW_EVT_INT_CLEAR);
		cashew_evt_writel(evt, 0, CASHEW_EVT_INT_CLEAR);
		reg = cashew_evt_readl(evt, CASHEW_EVT_INT_STATUS);
		if (reg & mask)
			new_state = CASHEW_EVT_STATE_COMPLETE;
		else
			new_state = CASHEW_EVT_STATE_ACTIVE_THIRD;
		break;

	case CASHEW_EVT_STATE_ACTIVE_THIRD:
		new_state = CASHEW_EVT_STATE_COMPLETE;
		break;

	default:
		dev_err(evt->dev, "cashew_evt_new_pd bad state (%d)\n",
			(int)pd->state);
		return;
	}

	if (new_state == CASHEW_EVT_STATE_COMPLETE) {
		reg = cashew_evt_readl(evt, CASHEW_EVT_INT_ENABLE);
		reg &= ~mask;
		cashew_evt_writel(evt, reg, CASHEW_EVT_INT_ENABLE);
		cashew_evt_writel(evt, mask, CASHEW_EVT_INT_CLEAR);
		cashew_evt_writel(evt, 0, CASHEW_EVT_INT_CLEAR);
		if (pd->callback)
			pd->callback(pd->context);
	}

	pd->state = new_state;
}

static irqreturn_t cashew_evt_general_irq(int irq, void *dev_id)
{
	struct cashew_evt *evt = (struct cashew_evt *)dev_id;
	unsigned long flags;
	u32 mask, i, isr, ier;

	spin_lock_irqsave(&evt->lock, flags);

	while (1) {
		isr = cashew_evt_readl(evt, CASHEW_EVT_INT_STATUS);
		ier = cashew_evt_readl(evt, CASHEW_EVT_INT_ENABLE);
		isr &= ier;

		if (!isr)
			break;

		for (i = 0; i < CASHEW_EVT_MAX_SOURCES; i++) {
			if (i == 0)
				mask = CASHEW_EVT_INT_SAMPLE_0_FNE_MASK;
			else
				mask = CASHEW_EVT_INT_SAMPLE_1_FNE_MASK;

			if (isr & mask)
				cashew_evt_new_sr(evt, i, mask);
		}

		if (isr & CASHEW_EVT_INT_PHASE_FNE_MASK)
			cashew_evt_new_pd(evt);
	}

	spin_unlock_irqrestore(&evt->lock, flags);

	return IRQ_HANDLED;
}

int cashew_evt_get_sample_rate(struct cashew_evt *evt, int id, u32 *val,
			       u32 *sys_freq, void (*callback)(void *context),
			       void *context)
{
	unsigned long flags;
	u32 mask;
	int ret;

	if (id >= CASHEW_EVT_MAX_SOURCES)
		return -EINVAL;

	spin_lock_irqsave(&evt->lock, flags);

	switch (evt->sample_rates[id].state) {
	case CASHEW_EVT_STATE_IDLE:
		if (id == 0)
			mask = CASHEW_EVT_INT_SAMPLE_0_FNE_MASK;
		else
			mask = CASHEW_EVT_INT_SAMPLE_1_FNE_MASK;

		cashew_evt_clear_fifo(evt, CASHEW_EVT_SAMPLE_FIFO(id),
				      mask, true);

		ret = -EBUSY;
		evt->sample_rates[id].state = CASHEW_EVT_STATE_ACTIVE_FIRST;
		evt->sample_rates[id].callback = callback;
		evt->sample_rates[id].context = context;
		break;

	case CASHEW_EVT_STATE_COMPLETE:
		*val = cashew_evt_readl(evt, CASHEW_EVT_SAMPLE_FIFO(id));
		*sys_freq = evt->sys_rate;
		evt->sample_rates[id].state = CASHEW_EVT_STATE_IDLE;
		ret = 0;
		break;

	default:
		ret = -EBUSY;
		break;
	}

	spin_unlock_irqrestore(&evt->lock, flags);

	return ret;
}
EXPORT_SYMBOL_GPL(cashew_evt_get_sample_rate);

extern int cashew_evt_get_phase_diff(struct cashew_evt *evt, u32 *val,
				     u32 *sys_freq,
				     void (*callback)(void *context),
				     void *context)
{
	unsigned long flags;
	u32 mask;
	int ret = 0;

	spin_lock_irqsave(&evt->lock, flags);

	switch (evt->phase_difference.state) {
	case CASHEW_EVT_STATE_IDLE:
		mask = CASHEW_EVT_INT_PHASE_FNE_MASK;

		cashew_evt_clear_fifo(evt, CASHEW_EVT_PHASE_FIFO,
				      mask, true);

		ret = -EBUSY;
		evt->phase_difference.state = CASHEW_EVT_STATE_ACTIVE_FIRST;
		evt->phase_difference.callback = callback;
		evt->phase_difference.context = context;
		break;

	case CASHEW_EVT_STATE_COMPLETE:
		*val = cashew_evt_readl(evt, CASHEW_EVT_PHASE_FIFO);
		*sys_freq = evt->sys_rate;
		evt->phase_difference.state = CASHEW_EVT_STATE_IDLE;
		break;

	default:
		ret = -EBUSY;
		break;
	}

	spin_unlock_irqrestore(&evt->lock, flags);

	return ret;
}
EXPORT_SYMBOL_GPL(cashew_evt_get_phase_diff);

void cashew_evt_abort_measurements(struct cashew_evt *evt)
{
	unsigned long flags;
	u32 reg;

	spin_lock_irqsave(&evt->lock, flags);
	evt->sample_rates[0].state = CASHEW_EVT_STATE_IDLE;
	evt->sample_rates[1].state = CASHEW_EVT_STATE_IDLE;
	evt->phase_difference.state = CASHEW_EVT_STATE_IDLE;
	reg = cashew_evt_readl(evt, CASHEW_EVT_INT_ENABLE);
	reg &= ~CASHEW_EVT_INT_SAMPLE_0_FNE_MASK;
	reg &= ~CASHEW_EVT_INT_SAMPLE_1_FNE_MASK;
	reg &= ~CASHEW_EVT_INT_PHASE_FNE_MASK;
	cashew_evt_writel(evt, reg, CASHEW_EVT_INT_ENABLE);
	spin_unlock_irqrestore(&evt->lock, flags);
}
EXPORT_SYMBOL_GPL(cashew_evt_abort_measurements);

extern int cashew_evt_read_time_stamp(struct cashew_evt *evt, u32 *val,
				      u32 *event, int index)
{
	u32 offset;

	if (index < 0 || index >= CASHEW_EVT_SOURCE_MAX)
		return -EINVAL;

	/* Return the event id for the index requested. */
	*event = cashew_evt_readl(evt, CASHEW_EVT_ET_EVENTS(index));
	*event >>= CASHEW_EVT_ET_EVENTS_SHIFT(index);
	*event &= CASHEW_EVT_ET_EVENTS_MASK;

	offset = CASHEW_EVT_SOURCE_MAX * 4 + index + 4;

	*val = cashew_evt_readl(evt, CASHEW_EVT_SOURCE_START + offset);
	*val &= CASHEW_EVT_TIME_STAMP_MASK;

	return 0;
}
EXPORT_SYMBOL_GPL(cashew_evt_read_time_stamp);

extern int cashew_evt_set_event_source(struct cashew_evt *evt, int event,
				       int index)
{
	unsigned long flags;
	u32 reg;

	if (event < 0 || index < 0)
		return -EINVAL;

	if (event >= CASHEW_EVT_NUM_EVENTS) {
		dev_err(evt->dev, "Event id [%d] not supported\n", event);
		return -EINVAL;
	}
	if (index >= CASHEW_EVT_SOURCE_MAX) {
		dev_err(evt->dev, "Index id [%d] not supported\n", index);
		return -EINVAL;
	}

	spin_lock_irqsave(&evt->lock, flags);

	reg = cashew_evt_readl(evt, CASHEW_EVT_ET_EVENTS(index));
	reg &= ~(CASHEW_EVT_ET_EVENTS_MASK <<
		 CASHEW_EVT_ET_EVENTS_SHIFT(index));
	reg |= event << CASHEW_EVT_ET_EVENTS_SHIFT(index);
	cashew_evt_writel(evt, event, CASHEW_EVT_ET_EVENTS(index));

	spin_unlock_irqrestore(&evt->lock, flags);

	return 0;
}
EXPORT_SYMBOL_GPL(cashew_evt_set_event_source);

void cashew_evt_set_event_select(struct cashew_evt *evt)
{
	unsigned long flags;
	u32 i;

	spin_lock_irqsave(&evt->lock, flags);

	for (i = 0; i < CASHEW_EVT_SOURCE_MAX; i++)
		cashew_evt_writel(evt, i, CASHEW_EVT_SOURCE_START + (i * 4));

	spin_unlock_irqrestore(&evt->lock, flags);
}

void cashew_evt_clk_rate_change(struct cashew_evt *evt)
{
	u64 tmp;
	unsigned long flags;
	unsigned long rate;
	ktime_t quarter_rollover;
	u32 mult, shift, mask;

	dev_dbg(evt->dev, "cashew_evt_clk_rate_change()\n");

	mask = CASHEW_EVT_COUNTER_MASK;

	rate = clk_get_rate(evt->clk_ref_internal);

	tmp = ((u64)mask + 1) * NSEC_PER_SEC;
	do_div(tmp, rate);
	tmp >>= 2;
	quarter_rollover.tv64 = tmp;

	clocks_calc_mult_shift(&mult, &shift, rate,
			       NSEC_PER_SEC, DIV_ROUND_UP(mask, rate));

	spin_lock_irqsave(&evt->lock, flags);
	evt->quarter_rollover = quarter_rollover;
	evt->cc.mult = mult;
	evt->cc.shift = shift;
	spin_unlock_irqrestore(&evt->lock, flags);

	dev_dbg(evt->dev, "rate %ld cc mult %u shift %u\n", rate, evt->cc.mult,
		evt->cc.shift);
}

static int cashew_evt_driver_probe(struct platform_device *pdev)
{
	struct cashew_evt *evt;
	int ret, irq;
	struct device_node *np = pdev->dev.of_node;
	u32 clk_select, rate;
	struct resource iomem;
	struct device *dev = &pdev->dev;

	evt = devm_kzalloc(&pdev->dev, sizeof(*evt), GFP_KERNEL);
	if (!evt)
		return -ENOMEM;
	platform_set_drvdata(pdev, evt);

	evt->dev = dev;
	evt->np = np;

	spin_lock_init(&evt->lock);

	ret = of_address_to_resource(np, 0, &iomem);
	if (ret) {
		dev_err(dev, "Could not get IO memory\n");
		return ret;
	}

	evt->base = devm_ioremap_resource(dev, &iomem);
	if (IS_ERR(evt->base))
		return PTR_ERR(evt->base);

	irq = platform_get_irq(pdev, 0);
	if (irq < 0) {
		dev_err(&pdev->dev, "can't get general irq\n");
		return irq;
	}

	ret = devm_request_irq(&pdev->dev, irq, cashew_evt_general_irq,
			       0, pdev->name, evt);
	if (ret) {
		dev_err(&pdev->dev, "can't request irq %d\n", irq);
		return ret;
	}

	irq = platform_get_irq(pdev, 3);
	if (irq < 0) {
		dev_err(&pdev->dev, "can't get trigger 0 irq\n");
		return irq;
	}

	ret = devm_request_irq(&pdev->dev, irq, cashew_evt_trigger_0_irq,
			       0, pdev->name, evt);
	if (ret) {
		dev_err(&pdev->dev, "can't request irq %d\n", irq);
		return ret;
	}

	if (of_property_read_u32(np, "img,clk-select", &clk_select)) {
		dev_err(&pdev->dev, "No img,clk-select property\n");
		return -EINVAL;
	}

	if (clk_select > 1)
		return -EINVAL;

	if (of_property_read_u32(np, "img,clk-rate", &rate))
		rate = 0;

	evt->audio_pll = devm_clk_get(&pdev->dev, "pll");
	if (IS_ERR(evt->audio_pll))
		return PTR_ERR(evt->audio_pll);

	ret = clk_prepare_enable(evt->audio_pll);
	if (ret)
		return ret;

	evt->clk_ref_a = devm_clk_get(&pdev->dev, "ref0");
	if (IS_ERR(evt->clk_ref_a))
		goto err_pll;

	ret = clk_prepare_enable(evt->clk_ref_a);
	if (ret)
		goto err_pll;

	evt->clk_ref_b = devm_clk_get(&pdev->dev, "ref1");
	if (IS_ERR(evt->clk_ref_b)) {
		ret = PTR_ERR(evt->clk_ref_b);
		goto err_ref_a;
	}

	ret = clk_prepare_enable(evt->clk_ref_b);
	if (ret)
		goto err_ref_a;

	evt->clk_sys = devm_clk_get(&pdev->dev, "sys");
	if (IS_ERR(evt->clk_sys)) {
		ret = PTR_ERR(evt->clk_sys);
		goto err_ref_b;
	}

	ret = clk_prepare_enable(evt->clk_sys);
	if (ret)
		goto err_ref_b;

	evt->sys_rate = clk_get_rate(evt->clk_sys);

	/* Always use the refclock. */
	cashew_evt_writel(evt, CASHEW_EVT_COUNTER_NOT_SYS_CLK_MASK,
			  CASHEW_EVT_COUNTER);

	evt->ref_names[0] = __clk_get_name(evt->clk_ref_a);
	evt->ref_names[1] = __clk_get_name(evt->clk_ref_b);

	evt->clk_ref_internal = clk_register_mux(NULL, "event_timer_internal",
		evt->ref_names, 2, CLK_SET_RATE_PARENT |
		CLK_SET_RATE_NO_REPARENT,
		evt->base + CASHEW_EVT_CLKSRC_SELECT,
		CASHEW_EVT_CLKSRC_SELECT_SHIFT,
		CASHEW_EVT_CLKSRC_SELECT_WIDTH,
		0, NULL);

	if (IS_ERR(evt->clk_ref_internal)) {
		ret = PTR_ERR(evt->clk_ref_internal);
		goto err_sys;
	}

	ret = of_clk_add_provider(np, of_clk_src_simple_get,
				  evt->clk_ref_internal);
	if (ret)
		goto err_mux;

	if (clk_select == 0)
		ret = clk_set_parent(evt->clk_ref_internal, evt->clk_ref_a);
	else
		ret = clk_set_parent(evt->clk_ref_internal, evt->clk_ref_b);

	if (ret)
		goto err_clkp;

	if (rate) {
		ret = clk_set_rate(evt->clk_ref_internal, rate);
		if (ret)
			goto err_clkp;
	}

	evt->cc.mask = CASHEW_EVT_COUNTER_MASK;
	evt->cc.read = cashew_evt_cc_read;

	cashew_evt_set_event_select(evt);

	cashew_evt_writel(evt, CASHEW_EVT_TIMER_ENABLE_MASK,
			  CASHEW_EVT_TIMER_ENABLE);

	cashew_evt_start_count(evt);

	cashew_evt_clk_rate_change(evt);

	ret = cashew_evt_init(evt);
	if (ret)
		goto err_count;

	spin_lock(&cashew_evt_list_spinlock);
	list_add(&evt->list, &cashew_evt_list);
	spin_unlock(&cashew_evt_list_spinlock);

	return 0;

err_count:
	cashew_evt_stop_count(evt);
	cashew_evt_writel(evt, 0, CASHEW_EVT_TIMER_ENABLE);
err_clkp:
	of_clk_del_provider(np);
err_mux:
	clk_unregister(evt->clk_ref_internal);
err_sys:
	clk_disable_unprepare(evt->clk_sys);
err_ref_b:
	clk_disable_unprepare(evt->clk_ref_b);
err_ref_a:
	clk_disable_unprepare(evt->clk_ref_a);
err_pll:
	clk_disable_unprepare(evt->audio_pll);

	return ret;
}

#ifdef CONFIG_PM_SLEEP
static int cashew_evt_suspend(struct device *dev)
{
	struct cashew_evt *evt = dev_get_drvdata(dev);
	struct cashew_evt_ctx *ctx = &evt->suspend_ctx;
	int i;

	cashew_evt_stop(evt);

	ctx->sources	  = cashew_evt_readl(evt, CASHEW_EVT_SOURCES);
	ctx->event_ctl	  = cashew_evt_readl(evt, CASHEW_EVT_EVENT_CTL);
	ctx->int_enable	  = cashew_evt_readl(evt, CASHEW_EVT_INT_ENABLE);

	for (i = 0; i < CASHEW_EVT_NUM_ENABLES; i++)
		ctx->time[i] = cashew_evt_readl(evt, CASHEW_EVT_TIME_REG(i));

	for (i = 0; i < CASHEW_EVT_SOURCE_MAX; i++) {
		ctx->source[i] =
			cashew_evt_readl(evt, CASHEW_EVT_SOURCE_START + i * 4);

		ctx->et_event[i] =
			cashew_evt_readl(evt, CASHEW_EVT_ET_EVENTS(i));
	}

	ctx->counter	  = cashew_evt_readl(evt, CASHEW_EVT_COUNTER);
	ctx->timer_enable = cashew_evt_readl(evt, CASHEW_EVT_TIMER_ENABLE);

	clk_disable_unprepare(evt->clk_sys);
	clk_disable_unprepare(evt->clk_ref_b);
	clk_disable_unprepare(evt->clk_ref_a);
	clk_disable_unprepare(evt->audio_pll);

	return 0;
}

static int cashew_evt_resume(struct device *dev)
{
	struct cashew_evt *evt = dev_get_drvdata(dev);
	struct cashew_evt_ctx *ctx = &evt->suspend_ctx;
	int ret, i;

	ret = clk_prepare_enable(evt->audio_pll);
	if (ret)
		return ret;

	ret = clk_prepare_enable(evt->clk_ref_a);
	if (ret)
		goto err_audio_pll;

	ret = clk_prepare_enable(evt->clk_ref_b);
	if (ret)
		goto err_ref_a;

	ret = clk_prepare_enable(evt->clk_sys);
	if (ret)
		goto err_ref_b;

	cashew_evt_writel(evt, ctx->sources,	  CASHEW_EVT_SOURCES);
	cashew_evt_writel(evt, ctx->event_ctl,	  CASHEW_EVT_EVENT_CTL);
	cashew_evt_writel(evt, ctx->int_enable,	  CASHEW_EVT_INT_ENABLE);

	for (i = 0; i < CASHEW_EVT_NUM_ENABLES; i++)
		cashew_evt_writel(evt, ctx->time[i], CASHEW_EVT_TIME_REG(i));

	for (i = 0; i < CASHEW_EVT_SOURCE_MAX; i++) {
		cashew_evt_writel(evt, ctx->source[i],
				  CASHEW_EVT_SOURCE_START + i * 4);

		cashew_evt_writel(evt, ctx->et_event[i],
				  CASHEW_EVT_ET_EVENTS(i));
	}

	cashew_evt_writel(evt, ctx->counter,	  CASHEW_EVT_COUNTER);
	cashew_evt_writel(evt, ctx->timer_enable, CASHEW_EVT_TIMER_ENABLE);

	cashew_evt_restart(evt);

	return 0;

err_ref_b:
	clk_disable_unprepare(evt->clk_ref_b);
err_ref_a:
	clk_disable_unprepare(evt->clk_ref_a);
err_audio_pll:
	clk_disable_unprepare(evt->audio_pll);
	return ret;
}
#endif /* CONFIG_PM_SLEEP */

static const struct dev_pm_ops cashew_evt_pm_ops = {
	SET_SYSTEM_SLEEP_PM_OPS(cashew_evt_suspend, cashew_evt_resume)
};

static const struct of_device_id cashew_evt_of_match[] = {
	{ .compatible = "img,cashew-event-timer" },
	{ },
};
MODULE_DEVICE_TABLE(of, cashew_evt_of_match);

static int cashew_evt_driver_remove(struct platform_device *pdev)
{
	struct cashew_evt *evt = platform_get_drvdata(pdev);

	spin_lock(&cashew_evt_list_spinlock);
	list_del(&evt->list);
	spin_unlock(&cashew_evt_list_spinlock);
	cashew_evt_deinit(evt);
	cashew_evt_stop_count(evt);
	cashew_evt_writel(evt, 0, CASHEW_EVT_TIMER_ENABLE);
	of_clk_del_provider(evt->dev->of_node);
	clk_unregister(evt->clk_ref_internal);
	clk_disable_unprepare(evt->clk_sys);
	clk_disable_unprepare(evt->clk_ref_b);
	clk_disable_unprepare(evt->clk_ref_a);

	return 0;
}

static struct platform_driver cashew_evt_driver = {
	.driver = {
		.name = "cashew-event-timer",
		.of_match_table = cashew_evt_of_match,
		.pm = &cashew_evt_pm_ops,
	},
	.probe = cashew_evt_driver_probe,
	.remove = cashew_evt_driver_remove,
};
module_platform_driver(cashew_evt_driver);

MODULE_DESCRIPTION("Event Timer driver");
MODULE_AUTHOR("Sifan Naeem");
MODULE_LICENSE("GPL v2");
