/*
 * IMG parallel audio input controller driver
 *
 * Copyright (C) 2016 Imagination Technologies Ltd.
 *
 * Author: Sadegh Abbasi <Sadegh.Abbasi@imgtec.com>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 */

#include <linux/clk.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/of.h>
#include <linux/platform_device.h>
#include <linux/pm_runtime.h>
#include <linux/reset.h>

#include <sound/core.h>
#include <sound/dmaengine_pcm.h>
#include <sound/initval.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <sound/soc.h>

#define MAX_PAI_CHAN			8
#define FIFO_THRESH			16
#define IMG_PRL_IN_RX_FIFO		0x0
#define IMG_PRL_IN_CTL			0x4
#define IMG_PRL_IN_NO_ACTIVE_CH_MASK	0x1E
#define IMG_PRL_IN_CTL_ME_MASK		BIT(0)
#define IMG_PRL_IN_STATUS		0x18
#define IMG_PRL_IN_AUDIO_FMT_MASK	0x7
#define IMG_PRL_IN_DMA_CTRL		0x14
struct img_prl_in {
	void __iomem *base;
	struct snd_dmaengine_dai_dma_data dma_data;
	struct device *dev;
};

/* cashew only supports AUDIO_SAMPLE and HBR_AUDIO */
enum audio_formats {
	/*
	 * AUDIO_SAMPLE covers:
	 * IEC60958 L-PCM 2 and 8 channels,
	 * IEC61937 compressed non-linear PCM
	 */
	AUDIO_SAMPLE,
	ONE_BIT_AUDIO,
	DST_AUDIO,
	/* HBR_AUDIO covers Dolby True-HD and DTS-HD Master Audio */
	HBR_AUDIO,
	MAS_AUDIO,
	OBM_AUDIO,
	AUDIO_FORMAT_MAX
};

static inline void img_prl_in_writel(struct img_prl_in *prl,
				     u32 val, u32 reg)
{
	writel(val, prl->base + reg);
}

static inline u32 img_prl_in_readl(struct img_prl_in *prl, u32 reg)
{
	return readl(prl->base + reg);
}

static void img_prl_in_enable(struct img_prl_in *prl, bool enable)
{
	u32 ctl = img_prl_in_readl(prl, IMG_PRL_IN_CTL);

	if (enable)
		ctl |= IMG_PRL_IN_CTL_ME_MASK;
	else
		ctl &= ~IMG_PRL_IN_CTL_ME_MASK;
	img_prl_in_writel(prl, ctl, IMG_PRL_IN_CTL);
}

static int img_prl_in_trigger(struct snd_pcm_substream *substream, int cmd,
			      struct snd_soc_dai *dai)
{
	struct img_prl_in *prl = snd_soc_dai_get_drvdata(dai);

	switch (cmd) {
	case SNDRV_PCM_TRIGGER_START:
	case SNDRV_PCM_TRIGGER_RESUME:
	case SNDRV_PCM_TRIGGER_PAUSE_RELEASE:
		img_prl_in_enable(prl, true);
		break;
	case SNDRV_PCM_TRIGGER_STOP:
	case SNDRV_PCM_TRIGGER_SUSPEND:
	case SNDRV_PCM_TRIGGER_PAUSE_PUSH:
		img_prl_in_enable(prl, false);
		break;
	default:
		return -EINVAL;
	}

	return 0;
}

/*
 * Called when the hardware parameters (rate, no_of_chan, format, etc)
 * of a stream change. Must be called before the stream starts.
 */

static int img_prl_in_hw_params(struct snd_pcm_substream *substream,
				struct snd_pcm_hw_params *params,
				struct snd_soc_dai *dai)
{
	struct img_prl_in *prl = snd_soc_dai_get_drvdata(dai);
	unsigned int channels;
	snd_pcm_format_t format;
	u32 reg;

	channels = params_channels(params);
	pr_info("%s: channels = %d\n", __func__, channels);
	if (channels > MAX_PAI_CHAN)
		return -EINVAL;

	format = params_format(params);
	if (format != SNDRV_PCM_FORMAT_S24_LE &&
	    format != SNDRV_PCM_FORMAT_S16_LE)
		return -EINVAL;

	reg = img_prl_in_readl(prl, IMG_PRL_IN_CTL);
	reg = (reg & ~IMG_PRL_IN_NO_ACTIVE_CH_MASK) | ((channels - 1) << 1) ;
	img_prl_in_writel(prl, reg, IMG_PRL_IN_CTL);
	/* TODO: do we need to check the rate and other paramaters? */
	/* Change the default FIFO threshold. */
	img_prl_in_writel(prl, FIFO_THRESH, IMG_PRL_IN_DMA_CTRL);

	return 0;
}

static int img_prl_in_fmt_info(struct snd_kcontrol *kcontrol,
			       struct snd_ctl_elem_info *uinfo)
{
	uinfo->type = SNDRV_CTL_ELEM_TYPE_ENUMERATED;
	uinfo->count = 1;
	uinfo->value.enumerated.items = AUDIO_FORMAT_MAX;
	/* cashew only supports two formats */
	if (uinfo->value.enumerated.item != 0)
		return -EINVAL;

	return 0;
}

static int img_prl_in_fmt_get(struct snd_kcontrol *kcontrol,
			      struct snd_ctl_elem_value *ucontrol)
{
	struct snd_soc_dai *cpu_dai = snd_kcontrol_chip(kcontrol);
	struct img_prl_in *prl = snd_soc_dai_get_drvdata(cpu_dai);

	ucontrol->value.enumerated.item[0] =
		img_prl_in_readl(prl, IMG_PRL_IN_STATUS) &
		IMG_PRL_IN_AUDIO_FMT_MASK;

	return 0;
}

static struct snd_kcontrol_new img_prl_in_controls[] = {
	{
		.access = SNDRV_CTL_ELEM_ACCESS_READ,
		.iface = SNDRV_CTL_ELEM_IFACE_PCM,
		.name = "PAI audio format",
		.info = img_prl_in_fmt_info,
		.get = img_prl_in_fmt_get,
	},
};

static const struct snd_soc_dai_ops img_prl_in_dai_ops = {
	.trigger = img_prl_in_trigger,
	.hw_params = img_prl_in_hw_params,
};

static int img_prl_in_dai_probe(struct snd_soc_dai *dai)
{
	struct img_prl_in *prl = snd_soc_dai_get_drvdata(dai);

	snd_soc_dai_init_dma_data(dai, NULL, &prl->dma_data);
	if (!dai->id)
		snd_soc_add_dai_controls(dai, img_prl_in_controls, 1);

	return 0;
}

static struct snd_soc_dai_driver img_prl_in_dai = {
	.probe = img_prl_in_dai_probe,
	.capture = {
		.channels_min = 1,
		.channels_max = 8,
		 /*
		  * This is frame rate which can go up to 1536000 resulting in
		  * hdmi_audio_clk rate of 196.608 MHz. In Europa, although
		  * this clockcan go up to this rate but the recommended rate
		  * is 12.288 MHz. But how far can the PAI go?
		  */
		.rates = SNDRV_PCM_RATE_8000_192000,
		/* S16_LE is for HBR */
		.formats = SNDRV_PCM_FMTBIT_S24_LE | SNDRV_PCM_FMTBIT_S16_LE,
	},
	.ops = &img_prl_in_dai_ops
};

static const struct snd_soc_component_driver img_prl_in_component = {
	.name = "img-prl-in"
};

static int img_prl_in_probe(struct platform_device *pdev)
{
	struct img_prl_in *prl;
	struct resource *res;
	void __iomem *base;
	int ret;

	prl = devm_kzalloc(&pdev->dev, sizeof(*prl), GFP_KERNEL);
	if (!prl)
		return -ENOMEM;

	platform_set_drvdata(pdev, prl);

	prl->dev = &pdev->dev;

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	base = devm_ioremap_resource(&pdev->dev, res);
	if (IS_ERR(base))
		return PTR_ERR(base);

	prl->base = base;

	img_prl_in_enable(prl, false);

	prl->dma_data.addr = res->start + IMG_PRL_IN_RX_FIFO;
	prl->dma_data.addr_width = 4;
	prl->dma_data.maxburst = 16;

	ret = devm_snd_soc_register_component(&pdev->dev,
					      &img_prl_in_component,
					      &img_prl_in_dai, 1);
	if (ret)
		goto err;

	ret = devm_snd_dmaengine_pcm_register(&pdev->dev, NULL, 0);
	if (ret)
		goto err;

	dev_info(prl->dev, "img parallel-audio-in probe succeeded.");
	return 0;

err:
	dev_err(prl->dev, "img parallel-audio-in failed.");
	return ret;
}

static int img_prl_in_dev_remove(struct platform_device *pdev)
{
/*	struct img_prl_in *prl = platform_get_drvdata(pdev); */

	return 0;
}

static const struct of_device_id img_prl_in_of_match[] = {
	{ .compatible = "img,parallel-audio-in" },
	{}
};

MODULE_DEVICE_TABLE(of, img_prl_in_of_match);

static struct platform_driver img_prl_in_driver = {
	.driver = {
		.name = "img-parallel-audio-in",
		.of_match_table = img_prl_in_of_match,
	},
	.probe = img_prl_in_probe,
	.remove = img_prl_in_dev_remove
};
module_platform_driver(img_prl_in_driver);

MODULE_AUTHOR("Sadegh Abbasi <Sadegh.Abbasi@imgtec.com>");
MODULE_DESCRIPTION("IMG Parallel Audio Input Driver");
MODULE_LICENSE("GPL v2");
