/*
 * PDP Framebuffer
 *
 * Copyright (c) 2008,09,10 Imagination Technologies Ltd.
 * Parts Copyright (C) 2009 Nokia Corporation
 *	(custom ISR code derived from drivers/video/omap2/dss/dispc.c)
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 */

#ifndef __PDPFB_USE_OSAL__
	#include <linux/kernel.h>
	#include <linux/init.h>
	#include <linux/delay.h>
	#include <linux/mm.h>
	#include <linux/fb.h>
	#include <linux/module.h>
	#include <linux/platform_device.h>
	#include <linux/dma-mapping.h>
	#include <linux/interrupt.h>
	#include <linux/uaccess.h>
	#include <linux/clk.h>
	#include <linux/console.h>
	#include <linux/slab.h>
	#include <linux/debugfs.h>
	#include <linux/lcm.h>
	#include <asm/irq.h>
	#include <video/of_videomode.h>
	#include <video/videomode.h>
	#include <video/of_display_timing.h>
	#include <video/pdpfb2.h>
	#include "pdpfb2.h"
	#include "pdpfb2_str.h"
	#include "pdpfb2_regs.h"
	#include "pdpfb2_mmu_regs.h"
#else
	#include <system.h>
	#include "pdpfb2_linuxal.h"
	#define __KERNEL__
	#include "pdpfb2.h"
	#include "pdpfb2_internal.h"
	#include "pdpfb2_str.h"
	#include "pdpfb2_regs.h"
	#include "pdpfb2_mmu_regs.h"
#endif

#ifdef CONFIG_FB_PDP2_PDUMP
	#include "pdpfb2_regfields.h"
#endif

#ifndef __PDPFB_USE_OSAL__
/* Physical address and length of video memory */
static ulong videomem_base;
module_param(videomem_base, ulong, 0);
static ulong videomem_len;
module_param(videomem_len, ulong, 0);

/* Physical address and length of graphics framebuffers */
static ulong gfx1_videomem_base;
module_param(gfx1_videomem_base, ulong, 0);
static ulong gfx1_videomem_len = CONFIG_FB_PDP2_GFX1_VIDEOMEM;
module_param(gfx1_videomem_len, ulong, 0);

static ulong gfx2_videomem_base;
module_param(gfx2_videomem_base, ulong, 0);
static ulong gfx2_videomem_len = CONFIG_FB_PDP2_GFX2_VIDEOMEM;
module_param(gfx2_videomem_len, ulong, 0);

static ulong gfx3_videomem_base;
module_param(gfx3_videomem_base, ulong, 0);
static ulong gfx3_videomem_len = CONFIG_FB_PDP2_GFX3_VIDEOMEM;
module_param(gfx3_videomem_len, ulong, 0);

static ulong gfx4_videomem_base;
module_param(gfx4_videomem_base, ulong, 0);
static ulong gfx4_videomem_len = CONFIG_FB_PDP2_GFX4_VIDEOMEM;
module_param(gfx4_videomem_len, ulong, 0);

/* Physical address and length of video framebuffers */
static ulong vid1_videomem_base;
module_param(vid1_videomem_base, ulong, 0);
static ulong vid1_videomem_len = CONFIG_FB_PDP2_VID1_VIDEOMEM;
module_param(vid1_videomem_len, ulong, 0);

static ulong vid2_videomem_base;
module_param(vid2_videomem_base, ulong, 0);
static ulong vid2_videomem_len = CONFIG_FB_PDP2_VID2_VIDEOMEM;
module_param(vid2_videomem_len, ulong, 0);

static ulong vid3_videomem_base;
module_param(vid3_videomem_base, ulong, 0);
static ulong vid3_videomem_len = CONFIG_FB_PDP2_VID3_VIDEOMEM;
module_param(vid3_videomem_len, ulong, 0);

static ulong vid4_videomem_base;
module_param(vid4_videomem_base, ulong, 0);
static ulong vid4_videomem_len = CONFIG_FB_PDP2_VID4_VIDEOMEM;
module_param(vid4_videomem_len, ulong, 0);
#endif

#define PDPFB_IRQ_HBLNK0	0x00001	/* blanking */
#define PDPFB_IRQ_HBLNK1	0x00002	/* blanking */
#define PDPFB_IRQ_VEVENT0	0x00004	/* safe update */
#define PDPFB_IRQ_VEVENT1	0x00008	/* safe update */

/* Groups of registers that may need updating on VEVENT */
#define PDP_UPDATE_SYNC			0x00000001

struct pdpfb_update_sync {
	u32 hsync1, hsync2, hsync3;
	u32 hde_ctrl;
	u32 str1blend2;
};

struct pdpfb_updates {
	spinlock_t lock;
	u32 updates;					/* PDP_UPDATE_* */
	struct pdpfb_update_sync sync;	/* PDP_UPDATE_SYNC */
};

typedef void (*pdpfb_isr_t) (void *arg, u32 mask);

struct pdpfb_isr_data {
	pdpfb_isr_t isr;
	void *arg;
	u32 mask;
};

#define PDPFB_MAX_NR_ISRS		8

struct pdpfb_rgb_fmt {
	struct fb_bitfield r, g, b;
};

#define PDPFB_MEMPOOLF_KERNEL	0x1	/* Kernel allocated memory */

struct pdpfb_mem_pool {
	unsigned long phys_base;
	void *base;
	unsigned long size;
	unsigned int flags;
	unsigned long allocated;
};

static char *pdpfb_mem_pool_names[] = {
	[PDPFB_MEMPOOL_MEM]	= "videomem",
	[PDPFB_MEMPOOL_GFX1MEM]	= "gfx1_videomem",
	[PDPFB_MEMPOOL_GFX2MEM]	= "gfx2_videomem",
	[PDPFB_MEMPOOL_GFX3MEM]	= "gfx3_videomem",
	[PDPFB_MEMPOOL_GFX4MEM]	= "gfx4_videomem",
	[PDPFB_MEMPOOL_VID1MEM]	= "vid1_videomem",
	[PDPFB_MEMPOOL_VID2MEM]	= "vid2_videomem",
	[PDPFB_MEMPOOL_VID3MEM]	= "vid3_videomem",
	[PDPFB_MEMPOOL_VID4MEM]	= "vid4_videomem",
	[PDPFB_MEMPOOL_USER]	= "usermem",
};

static const int nonstd2pixfmt[] = {
	[0] = -1,
	[PDP_PIXFMT_420_PL8] = PDP_SURF_PIXFMT_420_PL8,
	[PDP_PIXFMT_420_PL8IVU] = PDP_SURF_PIXFMT_420_PL8IVU,
	[PDP_PIXFMT_420_PL8IUV] = PDP_SURF_PIXFMT_420_PL8IUV,
	[PDP_PIXFMT_420_PL10IVU] = PDP_SURF_PIXFMT_420_PL10IVU,
	[PDP_PIXFMT_420_PL10IUV] = PDP_SURF_PIXFMT_420_PL10IUV,
	[PDP_PIXFMT_422_PL10IVU] = PDP_SURF_PIXFMT_422_PL10IVU,
	[PDP_PIXFMT_422_PL10IUV] = PDP_SURF_PIXFMT_422_PL10IUV,
	[PDP_PIXFMT_422_UY0VY1_8888] = PDP_SURF_PIXFMT_422_UY0VY1_8888,
	[PDP_PIXFMT_422_VY0UY1_8888] = PDP_SURF_PIXFMT_422_VY0UY1_8888,
	[PDP_PIXFMT_422_Y0UY1V_8888] = PDP_SURF_PIXFMT_422_Y0UY1V_8888,
	[PDP_PIXFMT_422_Y0VY1U_8888] = PDP_SURF_PIXFMT_422_Y0VY1U_8888,
	[PDP_PIXFMT_AYUV8888] = PDP_SURF_PIXFMT_AYUV8888,
	[PDP_PIXFMT_YUV101010] = PDP_SURF_PIXFMT_YUV101010,
};

static const struct pdpfb_lumakey_coefs pdpfb_lumakey_presets[] = {
	[PDP_CSCPRESET_HDTV] = { .cr = 0x364,			/* (0.2126 * 1023) << 2 */
							 .cg = 0xB6C,			/* (0.7152 * 1023) << 2 */
							 .cb = 0x124,			/* (0.0722 * 1023) << 2 */
							 .output_offset = 0 },
	[PDP_CSCPRESET_SDTV] = { .cr = 0x4C4,			/* (0.299 * 1023) << 2 */
							 .cg = 0x960,			/* (0.587 * 1023) << 2 */
							 .cb = 0x1D0,			/* (0.114 * 1023) << 2 */
							 .output_offset = 0 },
};

static const struct pdpfb_csc_coefs pdpfb_csc_presets[] = {
	[PDP_CSCPRESET_HDTV] = {
		.ry = 298,	.rv = 459,	.ru = 0,
		.gy = 298,	.gv = -137,	.gu = -55,
		.by = 298,	.bv = 0,	.bu = 541,
	},
	[PDP_CSCPRESET_SDTV] = {
		.ry = 298,	.rv = 409,	.ru = 0,
		.gy = 298,	.gv = -208,	.gu = -100,
		.by = 298,	.bv = 0,	.bu = 517,
	},
	[PDP_CSCPRESET_LEGACYHDTV] = {
		.ry = 298,	.rv = 459,	.ru = 0,
		.gy = 298,	.gv = -139,	.gu = -66,
		.by = 298,	.bv = 0,	.bu = 532,
	},
	[PDP_CSCPRESET_LEGACYSDTV] = {
		.ry = 298,	.rv = 409,	.ru = 0,
		.gy = 298,	.gv = -207,	.gu = -97,
		.by = 298,	.bv = 0,	.bu = 519,
	},
};

static const u16 gamme_table[9][9][PDP_GAMMA_TABLE_ENTRIES] = {
	{	{ 0x000,0x020,0x040,0x060,0x080,0x0a0,0x0c0,0x0e0, /* 1.0 -> 1.0 */
		  0x100,0x120,0x140,0x160,0x180,0x1a0,0x1c0,0x1e0,
		  0x200,0x220,0x240,0x260,0x280,0x2a0,0x2c0,0x2e0,
		  0x300,0x320,0x340,0x360,0x380,0x3a0,0x3c0,0x3e0,0x3ff },
		{ 0x000,0x056,0x08d,0x0bc,0x0e7,0x10f,0x135,0x159, /* 1.0 -> 1.4 */
		  0x17c,0x19d,0x1bd,0x1dd,0x1fb,0x219,0x236,0x253,
		  0x26f,0x28b,0x2a6,0x2c0,0x2db,0x2f5,0x30e,0x328,
		  0x340,0x359,0x371,0x38a,0x3a1,0x3b9,0x3d0,0x3e8,0x3ff },
		{ 0x000,0x095,0x0db,0x112,0x142,0x16c,0x193,0x1b7, /* 1.0 -> 1.8 */
		  0x1d9,0x1f9,0x218,0x235,0x251,0x26c,0x286,0x29f,
		  0x2b8,0x2cf,0x2e7,0x2fd,0x313,0x329,0x33e,0x353,
		  0x367,0x37b,0x38f,0x3a2,0x3b5,0x3c8,0x3da,0x3ed,0x3ff },
		{ 0x000,0x0b4,0x0ff,0x139,0x169,0x194,0x1ba,0x1de, /* 1.0 -> 2.0 */
		  0x1ff,0x21e,0x23b,0x257,0x272,0x28c,0x2a4,0x2bc,
		  0x2d3,0x2e9,0x2ff,0x314,0x328,0x33c,0x350,0x363,
		  0x375,0x388,0x39a,0x3ab,0x3bc,0x3cd,0x3de,0x3ee,0x3ff },
		{ 0x000,0x0d3,0x122,0x15c,0x18d,0x1b7,0x1dd,0x200, /* 1.0 -> 2.2 */
		  0x220,0x23e,0x25a,0x275,0x28f,0x2a7,0x2be,0x2d4,
		  0x2ea,0x2ff,0x313,0x327,0x33a,0x34c,0x35e,0x370,
		  0x381,0x392,0x3a2,0x3b2,0x3c2,0x3d2,0x3e1,0x3f0,0x3ff },
		{ 0x000,0x087,0x0dd,0x11e,0x153,0x182,0x1ac,0x1d2, /* 1.0 -> BT709 */
		  0x1f5,0x216,0x234,0x252,0x26d,0x288,0x2a1,0x2ba,
		  0x2d1,0x2e8,0x2fe,0x313,0x328,0x33c,0x350,0x363,
		  0x376,0x388,0x39a,0x3ac,0x3bd,0x3ce,0x3de,0x3ef,0x3ff },
		{ 0x000,0x07c,0x0d4,0x115,0x14b,0x17b,0x1a5,0x1cb, /* 1.0 -> 240M */
		  0x1ef,0x210,0x22f,0x24d,0x269,0x284,0x29d,0x2b6,
		  0x2ce,0x2e5,0x2fb,0x311,0x326,0x33a,0x34e,0x361,
		  0x374,0x387,0x399,0x3ab,0x3bc,0x3cd,0x3de,0x3ee,0x3ff },
		{ 0x000,0x0c6,0x11b,0x15a,0x18d,0x1b9,0x1e1,0x204, /* 1.0 -> sRGB */
		  0x225,0x243,0x260,0x27b,0x294,0x2ad,0x2c4,0x2da,
		  0x2f0,0x304,0x318,0x32c,0x33f,0x351,0x362,0x374,
		  0x385,0x395,0x3a5,0x3b5,0x3c4,0x3d3,0x3e2,0x3f0,0x3ff },
		{ 0x000,0x128,0x17c,0x1b7,0x1e6,0x20f,0x232,0x252, /* 1.0 -> 2.8 */
		  0x26f,0x28a,0x2a3,0x2ba,0x2d0,0x2e5,0x2f9,0x30c,
		  0x31e,0x330,0x340,0x351,0x360,0x370,0x37e,0x38d,
		  0x39b,0x3a8,0x3b5,0x3c2,0x3cf,0x3db,0x3e7,0x3f3,0x3ff},
	},
	{	{ 0x000,0x007,0x015,0x025,0x037,0x04c,0x062,0x079, /* 1.4 -> 1.0 */
		  0x092,0x0ad,0x0c8,0x0e5,0x103,0x121,0x141,0x162,
		  0x183,0x1a5,0x1c9,0x1ed,0x211,0x237,0x25d,0x284,
		  0x2ab,0x2d4,0x2fc,0x326,0x350,0x37b,0x3a6,0x3d2,0x3ff },
		{ 0x000,0x020,0x040,0x060,0x080,0x0a0,0x0c0,0x0e0, /* 1.4 -> 1.4 */
		  0x100,0x120,0x140,0x160,0x180,0x1a0,0x1c0,0x1e0,
		  0x200,0x220,0x240,0x260,0x280,0x2a0,0x2c0,0x2e0,
		  0x300,0x320,0x340,0x360,0x380,0x3a0,0x3c0,0x3e0,0x3ff },
		{ 0x000,0x045,0x076,0x0a2,0x0ca,0x0f1,0x116,0x139, /* 1.4 -> 1.8 */
		  0x15c,0x17d,0x19d,0x1bd,0x1dd,0x1fb,0x219,0x237,
		  0x254,0x271,0x28d,0x2aa,0x2c5,0x2e1,0x2fc,0x317,
		  0x331,0x34c,0x366,0x380,0x39a,0x3b3,0x3cc,0x3e6,0x3ff },
		{ 0x000,0x05a,0x092,0x0c3,0x0ee,0x116,0x13c,0x161, /* 1.4 -> 2.0 */
		  0x183,0x1a4,0x1c5,0x1e4,0x202,0x220,0x23d,0x259,
		  0x275,0x291,0x2ab,0x2c6,0x2e0,0x2f9,0x312,0x32b,
		  0x344,0x35c,0x374,0x38c,0x3a3,0x3ba,0x3d1,0x3e8,0x3ff },
		{ 0x000,0x070,0x0af,0x0e2,0x110,0x139,0x160,0x184, /* 1.4 -> 2.2 */
		  0x1a7,0x1c8,0x1e7,0x206,0x224,0x240,0x25c,0x277,
		  0x292,0x2ac,0x2c5,0x2de,0x2f6,0x30e,0x325,0x33d,
		  0x353,0x36a,0x380,0x396,0x3ab,0x3c0,0x3d5,0x3ea,0x3ff},
		{ 0x000,0x023,0x05e,0x097,0x0ca,0x0f7,0x122,0x14a, /* 1.4 -> BT709 */
		  0x170,0x194,0x1b7,0x1d8,0x1f8,0x218,0x236,0x254,
		  0x271,0x28d,0x2a9,0x2c4,0x2de,0x2f8,0x312,0x32b,
		  0x344,0x35d,0x375,0x38c,0x3a4,0x3bb,0x3d2,0x3e8,0x3ff},
		{ 0x000,0x01f,0x054,0x08d,0x0c0,0x0ef,0x11a,0x142, /* 1.4 -> 240M */
		  0x168,0x18d,0x1b0,0x1d2,0x1f2,0x212,0x231,0x24f,
		  0x26c,0x289,0x2a5,0x2c0,0x2db,0x2f5,0x30f,0x329,
		  0x342,0x35b,0x373,0x38b,0x3a3,0x3ba,0x3d1,0x3e8,0x3ff},
		{ 0x000,0x056,0x09d,0x0d7,0x108,0x135,0x15e,0x184, /* 1.4 -> sRGB */
		  0x1a8,0x1ca,0x1eb,0x20a,0x228,0x245,0x262,0x27d,
		  0x298,0x2b1,0x2cb,0x2e4,0x2fc,0x313,0x32b,0x341,
		  0x358,0x36e,0x383,0x399,0x3ae,0x3c2,0x3d7,0x3eb,0x3ff },
		{ 0x000,0x0b4,0x0ff,0x139,0x169,0x194,0x1ba,0x1de, /* 1.4 -> 2.8 */
		  0x1ff,0x21e,0x23b,0x257,0x272,0x28c,0x2a4,0x2bc,
		  0x2d3,0x2e9,0x2ff,0x314,0x328,0x33c,0x350,0x363,
		  0x375,0x388,0x39a,0x3ab,0x3bc,0x3cd,0x3de,0x3ee,0x3ff},
	},
	{	{ 0x000,0x001,0x006,0x00e,0x018,0x024,0x032,0x042, /* 1.8 -> 1.0 */
		  0x054,0x068,0x07e,0x095,0x0af,0x0ca,0x0e7,0x105,
		  0x125,0x147,0x16b,0x190,0x1b6,0x1df,0x209,0x234,
		  0x261,0x28f,0x2bf,0x2f1,0x324,0x358,0x38e,0x3c6,0x3ff},
		{ 0x000,0x00b,0x01c,0x030,0x046,0x05e,0x076,0x090, /* 1.8 -> 1.4 */
		  0x0ac,0x0c8,0x0e5,0x103,0x121,0x141,0x161,0x182,
		  0x1a3,0x1c5,0x1e8,0x20b,0x22f,0x253,0x277,0x29d,
		  0x2c2,0x2e8,0x30f,0x336,0x35d,0x385,0x3ad,0x3d6,0x3ff},
		{ 0x000,0x020,0x040,0x060,0x080,0x0a0,0x0c0,0x0e0, /* 1.8 -> 1.8 */
		  0x100,0x120,0x140,0x160,0x180,0x1a0,0x1c0,0x1e0,
		  0x200,0x220,0x240,0x260,0x280,0x2a0,0x2c0,0x2e0,
		  0x300,0x320,0x340,0x360,0x380,0x3a0,0x3c0,0x3e0,0x3ff},
		{ 0x000,0x02d,0x054,0x079,0x09d,0x0c0,0x0e2,0x104, /* 1.8 -> 2.0 */
		  0x125,0x146,0x167,0x187,0x1a7,0x1c6,0x1e6,0x205,
		  0x224,0x242,0x261,0x27f,0x29e,0x2bc,0x2da,0x2f7,
		  0x315,0x333,0x350,0x36d,0x38b,0x3a8,0x3c5,0x3e2,0x3ff },
		{ 0x000,0x03c,0x069,0x093,0x0ba,0x0e0,0x104,0x127, /* 1.8 -> 2.2 */
		  0x149,0x16a,0x18a,0x1ab,0x1ca,0x1e9,0x208,0x226,
		  0x244,0x261,0x27e,0x29b,0x2b8,0x2d4,0x2f0,0x30c,
		  0x328,0x343,0x35f,0x37a,0x395,0x3af,0x3ca,0x3e4,0x3ff },
		{ 0x000,0x008,0x01f,0x040,0x06b,0x094,0x0bc,0x0e2, /* 1.8 -> BT709 */
		  0x108,0x12d,0x150,0x174,0x196,0x1b8,0x1da,0x1fb,
		  0x21b,0x23c,0x25c,0x27b,0x29b,0x2ba,0x2d8,0x2f7,
		  0x315,0x333,0x350,0x36e,0x38b,0x3a8,0x3c5,0x3e2,0x3ff },
		{ 0x000,0x007,0x01b,0x039,0x060,0x08a,0x0b2,0x0d9, /* 1.8 -> 240M */
		  0x0ff,0x124,0x149,0x16c,0x18f,0x1b2,0x1d4,0x1f5,
		  0x216,0x237,0x257,0x277,0x296,0x2b6,0x2d5,0x2f4,
		  0x312,0x330,0x34e,0x36c,0x38a,0x3a7,0x3c5,0x3e2,0x3ff },
		{ 0x000,0x019,0x04e,0x07e,0x0aa,0x0d3,0x0fb,0x120, /* 1.8 -> sRGB */
		  0x145,0x168,0x18a,0x1ac,0x1cc,0x1ec,0x20c,0x22b,
		  0x249,0x267,0x284,0x2a1,0x2be,0x2da,0x2f6,0x312,
		  0x32d,0x348,0x363,0x37d,0x398,0x3b2,0x3cc,0x3e5,0x3ff },
		{ 0x000,0x06e,0x0ac,0x0df,0x10c,0x136,0x15c,0x181, /* 1.8 -> 2.8 */
		  0x1a3,0x1c4,0x1e4,0x202,0x220,0x23d,0x259,0x274,
		  0x28f,0x2a9,0x2c2,0x2db,0x2f4,0x30c,0x324,0x33b,
		  0x352,0x368,0x37f,0x395,0x3aa,0x3c0,0x3d5,0x3ea,0x3ff },
	},
	{	{ 0x000,0x000,0x003,0x008,0x00f,0x018,0x023,0x030, /* 2.0 -> 1.0 */
		  0x03f,0x050,0x063,0x078,0x08f,0x0a8,0x0c3,0x0e0,
		  0x0ff,0x120,0x143,0x168,0x18f,0x1b8,0x1e3,0x210,
		  0x23f,0x270,0x2a3,0x2d8,0x30f,0x348,0x383,0x3c0,0x3ff },
		{ 0x000,0x007,0x013,0x022,0x034,0x048,0x05d,0x074, /* 2.0 -> 1.4 */
		  0x08d,0x0a7,0x0c2,0x0de,0x0fb,0x11a,0x13a,0x15a,
		  0x17c,0x19e,0x1c1,0x1e5,0x20a,0x230,0x256,0x27e,
		  0x2a6,0x2ce,0x2f8,0x322,0x34d,0x378,0x3a4,0x3d1,0x3ff },
		{ 0x000,0x015,0x02e,0x049,0x065,0x082,0x09f,0x0bd, /* 2.0 -> 1.8 */
		  0x0db,0x0f9,0x118,0x138,0x158,0x178,0x198,0x1b8,0x1d9,
		  0x1fa,0x21b,0x23d,0x25e,0x280,0x2a2,0x2c4,0x2e7,
		  0x309,0x32c,0x34f,0x371,0x395,0x3b8,0x3db,0x3ff },
		{ 0x000,0x020,0x040,0x060,0x080,0x0a0,0x0c0,0x0e0, /* 2.0 -> 2.0 */
		  0x100,0x120,0x140,0x160,0x180,0x1a0,0x1c0,0x1e0,
		  0x200,0x220,0x240,0x260,0x280,0x2a0,0x2c0,0x2e0,
		  0x300,0x320,0x340,0x360,0x380,0x3a0,0x3c0,0x3e0,0x3ff },
		{ 0x000,0x02b,0x052,0x076,0x09a,0x0bd,0x0df,0x100, /* 2.0 -> 2.2 */
		  0x122,0x142,0x163,0x183,0x1a3,0x1c3,0x1e2,0x201,
		  0x220,0x23f,0x25e,0x27c,0x29b,0x2b9,0x2d7,0x2f5,
		  0x313,0x331,0x34f,0x36c,0x38a,0x3a7,0x3c4,0x3e1,0x3ff },
		{ 0x000,0x004,0x011,0x028,0x047,0x06e,0x093,0x0b9, /* 2.0 -> BT709 */
		  0x0dd,0x101,0x125,0x148,0x16b,0x18e,0x1b0,0x1d3,
		  0x1f5,0x216,0x238,0x259,0x27b,0x29c,0x2bd,0x2dd,
		  0x2fe,0x31f,0x33f,0x35f,0x37f,0x39f,0x3bf,0x3df,0x3ff },
		{ 0x000,0x003,0x00f,0x023,0x03f,0x063,0x089,0x0af, /* 2.0 -> 240M */
		  0x0d4,0x0f8,0x11d,0x140,0x164,0x187,0x1aa,0x1cc,
		  0x1ef,0x211,0x233,0x255,0x276,0x298,0x2b9,0x2da,
		  0x2fb,0x31c,0x33d,0x35d,0x37e,0x39e,0x3be,0x3de,0x3ff },
		{ 0x000,0x00c,0x032,0x05d,0x086,0x0ad,0x0d3,0x0f7, /* 2.0 -> sRGB */
		  0x11b,0x13e,0x161,0x183,0x1a4,0x1c5,0x1e5,0x205,
		  0x225,0x244,0x263,0x282,0x2a1,0x2bf,0x2dd,0x2fb,
		  0x318,0x336,0x353,0x370,0x38d,0x3a9,0x3c6,0x3e2,0x3ff },
		{ 0x000,0x056,0x08d,0x0bc,0x0e7,0x10f,0x135,0x159, /* 2.0 -> 2.8 */
		  0x17c,0x19d,0x1bd,0x1dd,0x1fb,0x219,0x236,0x253,
		  0x26f,0x28b,0x2a6,0x2c0,0x2db,0x2f5,0x30e,0x328,
		  0x340,0x359,0x371,0x38a,0x3a1,0x3b9,0x3d0,0x3e8,0x3ff },
	},
	{	{ 0x000,0x000,0x002,0x005,0x00a,0x011,0x019,0x024, /* 2.2 -> 1.0 */
		  0x030,0x03e,0x04f,0x061,0x076,0x08d,0x0a5,0x0c1,
		  0x0de, 0x0fe,0x120,0x144,0x16b,0x194,0x1c0,0x1ee,
		  0x21f,0x252,0x287,0x2bf,0x2fa,0x337,0x377,0x3b9,0x3ff },
		{ 0x000,0x004,0x00d,0x018,0x026,0x037,0x049,0x05d, /* 2.2 -> 1.4 */
		  0x073,0x08b,0x0a4,0x0bf,0x0db,0x0f8,0x117,0x137,
		  0x158,0x17a,0x19e,0x1c2,0x1e8,0x20f,0x237,0x260,
		  0x28a,0x2b6,0x2e2,0x30f,0x33d,0x36c,0x39c,0x3cd,0x3ff },
		{ 0x000,0x00e,0x022,0x038,0x050,0x069,0x084,0x09f, /* 2.2 -> 1.8 */
		  0x0bb,0x0d9,0x0f6,0x115,0x134,0x154,0x174,0x195,
		  0x1b6,0x1d8,0x1fa,0x21c,0x23f,0x263,0x287,0x2ab,
		  0x2cf,0x2f4,0x319,0x33f,0x364,0x38b,0x3b1,0x3d8,0x3ff },
		{ 0x000,0x016,0x030,0x04b,0x067,0x084,0x0a2,0x0c0, /* 2.2 -> 2.0 */
		  0x0de,0x0fd,0x11c,0x13c,0x15b,0x17b,0x19c,0x1bc,
		  0x1dd,0x1fe,0x21f,0x240,0x262,0x283,0x2a5,0x2c7,
		  0x2e9,0x30b,0x32e,0x350,0x373,0x396,0x3b8,0x3db,0x3ff },
		{ 0x000,0x020,0x040,0x060,0x080,0x0a0,0x0c0,0x0e0, /* 2.2 -> 2.2 */
		  0x100,0x120,0x140,0x160,0x180,0x1a0,0x1c0,0x1e0,
		  0x200,0x220,0x240,0x260,0x280,0x2a0,0x2c0,0x2e0,
		  0x300,0x320,0x340,0x360,0x380,0x3a0,0x3c0,0x3e0,0x3ff },
		{ 0x000,0x002,0x00a,0x019,0x02f,0x04d,0x071,0x094, /* 2.2 -> BT709 */
		  0x0b7,0x0da,0x0fe,0x121,0x144,0x167,0x18a,0x1ad,
		  0x1d0,0x1f3,0x216,0x239,0x25c,0x27f,0x2a2,0x2c5,
		  0x2e8,0x30b,0x32e,0x350,0x373,0x396,0x3b9,0x3dc,0x3ff },
		{ 0x000,0x001,0x009,0x016,0x02a,0x044,0x066,0x08a, /* 2.2 -> 240M */
		  0x0ae,0x0d1,0x0f5,0x118,0x13c,0x160,0x183,0x1a6,
		  0x1ca,0x1ed,0x211,0x234,0x257,0x27b,0x29e,0x2c1,
		  0x2e5,0x308,0x32b,0x34e,0x372,0x395,0x3b8,0x3db,0x3ff },
		{ 0x000,0x006,0x01d,0x042,0x068,0x08c,0x0b0,0x0d3, /* 2.2 -> sRGB */
		  0x0f6,0x119,0x13b,0x15d,0x17e,0x1a0,0x1c1,0x1e2,
		  0x203,0x224,0x244,0x264,0x285,0x2a5,0x2c5,0x2e5,
		  0x304,0x324,0x343,0x363,0x382,0x3a1,0x3c1,0x3e0,0x3ff },
		{ 0x000,0x043,0x073,0x09f,0x0c7,0x0ed,0x112,0x135, /* 2.2 -> 2.8 */
		  0x158,0x179,0x19a,0x1ba,0x1d9,0x1f8,0x216,0x234,
		  0x251,0x26e,0x28a,0x2a7,0x2c3,0x2de,0x2fa,0x315,
		  0x330,0x34a,0x365,0x37f,0x399,0x3b2,0x3cc,0x3e5,0x3ff },
	},
	{	{ 0x000,0x008,0x00e,0x015,0x01d,0x027,0x033,0x040, /* BT709 -> 1.0 */
		  0x04f,0x060,0x073,0x087,0x09d,0x0b5,0x0cf,0x0eb,
		  0x109,0x129,0x14b,0x16e,0x194,0x1bc,0x1e6,0x212,
		  0x240,0x270,0x2a3,0x2d7,0x30e,0x347,0x382,0x3bf,0x3ff },
		{ 0x000,0x022,0x030,0x040,0x051,0x064,0x079,0x08e, /* BT709 -> 1.4 */
		  0x0a5,0x0bd,0x0d7,0x0f1,0x10d,0x129,0x147,0x166,
		  0x186,0x1a7,0x1c9,0x1eb,0x20f,0x234,0x259,0x27f,
		  0x2a7,0x2cf,0x2f8,0x321,0x34c,0x377,0x3a4,0x3d1,0x3ff },
		{ 0x000,0x049,0x05f,0x077,0x08f,0x0a8,0x0c2,0x0dd, /* BT709 -> 1.8 */
		  0x0f8,0x113,0x130,0x14c,0x16a,0x187,0x1a6,0x1c4,
		  0x1e3,0x202,0x222,0x242,0x263,0x283,0x2a4,0x2c6,
		  0x2e7,0x309,0x32c,0x34e,0x371,0x394,0x3b7,0x3db,0x3ff },
		{ 0x000,0x05f,0x079,0x093,0x0ae,0x0ca,0x0e5,0x101, /* BT709 -> 2.0 */
		  0x11d,0x13a,0x157,0x174,0x191,0x1af,0x1cd,0x1eb,
		  0x209,0x227,0x245,0x264,0x283,0x2a2,0x2c1,0x2e0,
		  0x2ff,0x31f,0x33f,0x35e,0x37e,0x39e,0x3be,0x3de,0x3ff },
		{ 0x000,0x076,0x093,0x0b0,0x0cd,0x0ea,0x107,0x124, /* BT709 -> 2.2 */
		  0x141,0x15e,0x17b,0x198,0x1b5,0x1d2,0x1ef,0x20c,
		  0x22a,0x247,0x264,0x281,0x29f,0x2bc,0x2d9,0x2f6,
		  0x314,0x331,0x34e,0x36c,0x389,0x3a6,0x3c4,0x3e1,0x3ff },
		{ 0x000,0x020,0x040,0x060,0x080,0x0a0,0x0c0,0x0e0, /* BT709 -> BT709 */
		  0x100,0x120,0x140,0x160,0x180,0x1a0,0x1c0,0x1e0,
		  0x200,0x220,0x240,0x260,0x280,0x2a0,0x2c0,0x2e0,
		  0x300,0x320,0x340,0x360,0x380,0x3a0,0x3c0,0x3e0,0x3ff },
		{ 0x000,0x023,0x039,0x055,0x075,0x096,0x0b6,0x0d6, /* BT709 -> 240M */
		  0x0f7,0x117,0x137,0x158,0x178,0x198,0x1b9,0x1d9,
		  0x1f9,0x21a,0x23a,0x25a,0x27b,0x29b,0x2bb,0x2dc,
		  0x2fc,0x31c,0x33d,0x35d,0x37d,0x39e,0x3be,0x3de,0x3ff },
		{ 0x000,0x05d,0x07e,0x09f,0x0bf,0x0df,0x0fe,0x11d, /* BT709 -> sRGB */
		  0x13c,0x15b,0x17a,0x198,0x1b7,0x1d5,0x1f3,0x211,
		  0x22f,0x24c,0x26a,0x287,0x2a5,0x2c2,0x2df,0x2fc,
		  0x319,0x336,0x353,0x370,0x38c,0x3a9,0x3c6,0x3e2,0x3ff },
		{ 0x000,0x0bc,0x0df,0x100,0x121,0x141,0x15f,0x17e, /* BT709 -> 2.8 */
		  0x19b,0x1b8,0x1d5,0x1f1,0x20c,0x228,0x243,0x25d,
		  0x277,0x291,0x2ab,0x2c5,0x2de,0x2f7,0x310,0x329,
		  0x341,0x359,0x371,0x389,0x3a1,0x3b9,0x3d0,0x3e7,0x3ff },
	},
	{	{ 0x000,0x00a,0x010,0x017,0x020,0x02b,0x037,0x044, /* 240M -> 1.0 */
		  0x054,0x065,0x078,0x08c,0x0a3,0x0bb,0x0d5,0x0f1,
		  0x10f,0x12e,0x150,0x174,0x199,0x1c1,0x1eb,0x216,
		  0x244,0x274,0x2a6,0x2da,0x310,0x349,0x383,0x3c0,0x3ff },
		{ 0x000,0x027,0x035,0x046,0x057,0x06a,0x07f,0x095,
		  0x0ac,0x0c4,0x0dd,0x0f8,0x113,0x130,0x14d,0x16c, /* 240M -> 1.4 */
		  0x18c,0x1ac,0x1ce,0x1f0,0x214,0x238,0x25d,0x283,
		  0x2aa,0x2d2,0x2fa,0x324,0x34e,0x379,0x3a5,0x3d1,0x3ff },
		{ 0x000,0x051,0x067,0x07f,0x097,0x0b0,0x0ca,0x0e4, /* 240M -> 1.8 */
		  0x0ff,0x11b,0x137,0x153,0x170,0x18e,0x1ac,0x1ca,
		  0x1e9,0x208,0x227,0x247,0x267,0x287,0x2a8,0x2c9,
		  0x2ea,0x30c,0x32e,0x350,0x372,0x395,0x3b8,0x3db,0x3ff },
		{ 0x000,0x068,0x082,0x09c,0x0b7,0x0d2,0x0ed,0x109, /* 240M -> 2.0 */
		  0x125,0x142,0x15e,0x17b,0x198,0x1b5,0x1d3,0x1f0,
		  0x20e,0x22c,0x24a,0x269,0x287,0x2a6,0x2c4,0x2e3,
		  0x302,0x321,0x341,0x360,0x37f,0x39f,0x3bf,0x3df,0x3ff },
		{ 0x000,0x080,0x09d,0x0b9,0x0d6,0x0f2,0x10f,0x12c, /* 240M -> 2.2 */
		  0x148,0x165,0x182,0x19f,0x1bc,0x1d8,0x1f5,0x212,
		  0x22f,0x24c,0x269,0x286,0x2a3,0x2bf,0x2dc,0x2f9,
		  0x316,0x333,0x350,0x36d,0x38a,0x3a7,0x3c4,0x3e1,0x3ff },
		{ 0x000,0x030,0x04a,0x06a,0x089,0x0a9,0x0c9,0x0e8, /* 240M -> 2.2 */
		  0x108,0x127,0x147,0x167,0x186,0x1a6,0x1c6,0x1e5,
		  0x205,0x224,0x244,0x264,0x283,0x2a3,0x2c2,0x2e2,
		  0x302,0x321,0x341,0x360,0x380,0x3a0,0x3bf,0x3df,0x3ff },
		{ 0x000,0x020,0x040,0x060,0x080,0x0a0,0x0c0,0x0e0, /* 240M -> 240M */
		  0x100,0x120,0x140,0x160,0x180,0x1a0,0x1c0,0x1e0,
		  0x200,0x220,0x240,0x260,0x280,0x2a0,0x2c0,0x2e0,
		  0x300,0x320,0x340,0x360,0x380,0x3a0,0x3c0,0x3e0,0x3ff },
		{ 0x000,0x069,0x089,0x0a9,0x0c9,0x0e8,0x107,0x126, /* 240M -> sRGB */
		  0x145,0x163,0x181,0x19f,0x1bd,0x1db,0x1f9,0x216,
		  0x234,0x251,0x26e,0x28c,0x2a9,0x2c5,0x2e2,0x2ff,
		  0x31c,0x338,0x355,0x371,0x38e,0x3aa,0x3c6,0x3e2,0x3ff },
		{ 0x000,0x0c8,0x0ea,0x10b,0x12b,0x14a,0x168,0x186, /* 240M -> 2.8 */
		  0x1a3,0x1c0,0x1dc,0x1f7,0x212,0x22d,0x248,0x262,
		  0x27c,0x296,0x2af,0x2c8,0x2e1,0x2fa,0x313,0x32b,
		  0x343,0x35b,0x373,0x38b,0x3a2,0x3b9,0x3d1,0x3e8,0x3ff },
	},
	{	{ 0x000,0x002,0x005,0x009,0x00e,0x015,0x01e,0x028, /* sRGB -> 1.0 */
		  0x034,0x041,0x051,0x063,0x076,0x08c,0x0a4,0x0be,
		  0x0da,0x0f9,0x11a,0x13e,0x164,0x18d,0x1b8,0x1e6,
		  0x216,0x249,0x27f,0x2b8,0x2f3,0x332,0x373,0x3b7,0x3ff },
		{ 0x000,0x00d,0x017,0x023,0x031,0x040,0x052,0x065, /* sRGB -> 1.4 */
		  0x079,0x090,0x0a7,0x0c0,0x0db,0x0f7,0x115,0x133,
		  0x154,0x175,0x198,0x1bc,0x1e1,0x208,0x230,0x259,
		  0x283,0x2ae,0x2db,0x309,0x338,0x368,0x399,0x3cb,0x3ff },
		{ 0x000,0x024,0x036,0x04b,0x060,0x077,0x090,0x0a9, /* sRGB -> 1.8 */
		  0x0c3,0x0de,0x0fa,0x117,0x135,0x153,0x172,0x192,
		  0x1b2, 0x1d3,0x1f4,0x216,0x239,0x25c,0x280,0x2a4,
		  0x2c9,0x2ee,0x314,0x33a,0x360,0x387,0x3af,0x3d6,0x3ff },
		{ 0x000,0x032,0x049,0x061,0x07a,0x094,0x0af,0x0ca, /* sRGB -> 2.0 */
		  0x0e6,0x103,0x120,0x13e,0x15c,0x17b,0x19a,0x1b9,
		  0x1d9, 0x1f9,0x219,0x23a,0x25b,0x27d,0x29f,0x2c1,
		  0x2e3,0x306,0x328,0x34b,0x36f,0x392,0x3b6,0x3da,0x3ff },
		{ 0x000,0x042,0x05d,0x078,0x094,0x0b0,0x0cd,0x0ea, /* sRGB -> 2.2 */
		  0x108,0x125,0x143,0x161,0x180,0x19e,0x1bd,0x1dc,
		  0x1fb,0x21a,0x23a,0x259,0x279,0x299,0x2b9,0x2d9,
		  0x2f9,0x319,0x33a,0x35a,0x37b,0x39c,0x3bd,0x3dd,0x3ff },
		{ 0x000,0x00b,0x017,0x029,0x042,0x060,0x080,0x0a0, /* sRGB -> BT709 */
		  0x0c1,0x0e1,0x102,0x123,0x145,0x166,0x188,0x1aa,
		  0x1cc,0x1ee,0x211,0x233,0x256,0x279,0x29c,0x2bf,
		  0x2e2,0x305,0x328,0x34c,0x36f,0x393,0x3b7,0x3db,0x3ff },
		{ 0x000,0x00a,0x015,0x025,0x03a,0x056,0x076,0x096, /* sRGB -> 240M */
		  0x0b7,0x0d8,0x0f9,0x11b,0x13d,0x15f,0x181,0x1a3,
		  0x1c6,0x1e8,0x20b,0x22e,0x251,0x274,0x298,0x2bb,
		  0x2de,0x302,0x326,0x34a,0x36e,0x392,0x3b6,0x3da,0x3ff },
		{ 0x000,0x020,0x040,0x060,0x080,0x0a0,0x0c0,0x0e0, /* sRGB -> sRGB */
		  0x100,0x120,0x140,0x160,0x180,0x1a0,0x1c0,0x1e0,
		  0x200,0x220,0x240,0x260,0x280,0x2a0,0x2c0,0x2e0,
		  0x300,0x320,0x340,0x360,0x380,0x3a0,0x3c0,0x3e0,0x3ff },
		{ 0x000,0x077,0x09b,0x0be,0x0e0,0x101,0x122,0x141, /* sRGB -> 2.8 */
		  0x161,0x17f,0x19e,0x1bc,0x1d9,0x1f7,0x214,0x231,
		  0x24d,0x26a,0x286,0x2a2,0x2be,0x2d9,0x2f5,0x310,
		  0x32b,0x346,0x361,0x37b,0x396,0x3b0,0x3ca,0x3e4,0x3ff },
	},
	{	{ 0x000,0x000,0x000,0x001,0x003,0x005,0x009,0x00e, /* 2.8 -> 1.0 */
		  0x015,0x01d,0x027,0x033,0x041,0x052,0x065,0x07a,
		  0x092,0x0ae,0x0cc,0x0ed,0x112,0x13a,0x166,0x195,
		  0x1c9,0x200,0x23b,0x27b,0x2bf,0x308,0x355,0x3a7,0x3ff },
		{ 0x000,0x000,0x003,0x008,0x00f,0x018,0x023,0x030, /* 2.8 -> 1.4 */
		  0x03f,0x050,0x063,0x078,0x08f,0x0a8,0x0c3,0x0e0,
		  0x0ff,0x120,0x143,0x168,0x18f,0x1b8,0x1e3,0x210,
		  0x23f,0x270,0x2a3,0x2d8,0x30f,0x348,0x383,0x3c0,0x3ff },
		{ 0x000,0x004,0x00d,0x019,0x028,0x038,0x04b,0x060, /* 2.8 -> 1.8 */
		  0x076,0x08e,0x0a7,0x0c2,0x0de,0x0fb,0x11a,0x13a,
		  0x15c,0x17e,0x1a2,0x1c6,0x1ec,0x213,0x23b,0x264,
		  0x28d,0x2b8,0x2e4,0x311,0x33f,0x36d,0x39d,0x3cd,0x3ff },
		{ 0x000,0x007,0x015,0x025,0x037,0x04c,0x062,0x079, /* 2.8 -> 2.0 */
		  0x092,0x0ad,0x0c8,0x0e5,0x103,0x121,0x141,0x162,
		  0x183,0x1a5,0x1c9,0x1ed,0x211,0x237,0x25d,0x284,
		  0x2ab,0x2d4,0x2fc,0x326,0x350,0x37b,0x3a6,0x3d2,0x3ff },
		{ 0x000,0x00c,0x01e,0x032,0x048,0x060,0x079,0x093, /* 2.8 -> 2.2 */
		  0x0af,0x0cb,0x0e8,0x106,0x125,0x145,0x165,0x186,
		  0x1a7,0x1c9,0x1eb,0x20e,0x232,0x256,0x27a,0x29f,
		  0x2c5,0x2eb,0x311,0x338,0x35f,0x386,0x3ae,0x3d6,0x3ff },
		{ 0x000,0x000,0x001,0x006,0x00d,0x019,0x02a,0x041, /* 2.8 -> BT709 */
		  0x05e,0x07e,0x09e,0x0bf,0x0e1,0x104,0x127,0x14b,
		  0x170,0x195,0x1bb,0x1e1,0x208,0x22f,0x257,0x280,
		  0x2a9,0x2d2,0x2fc,0x326,0x350,0x37b,0x3a7,0x3d2,0x3ff },
		{ 0x000,0x000,0x001,0x005,0x00c,0x016,0x025,0x03a, /* 2.8 -> 240M */
		  0x054,0x073,0x094,0x0b6,0x0d8,0x0fb,0x11f,0x143,
		  0x168,0x18e,0x1b4,0x1db,0x202,0x22a,0x253,0x27b,
		  0x2a5,0x2cf,0x2f9,0x323,0x34e,0x37a,0x3a6,0x3d2,0x3ff },
		{ 0x000,0x000,0x005,0x011,0x027,0x043,0x060,0x07e, /* 2.8 -> sRGB */
		  0x09d,0x0bd,0x0dd,0x0fe,0x11f,0x141,0x163,0x185,
		  0x1a8,0x1cb,0x1ef,0x213,0x237,0x25b,0x280,0x2a5,
		  0x2cb,0x2f0,0x316,0x33c,0x363,0x389,0x3b0,0x3d7,0x3ff },
		{ 0x000,0x020,0x040,0x060,0x080,0x0a0,0x0c0,0x0e0, /* 2.8 -> 2.8 */
		  0x100,0x120,0x140,0x160,0x180,0x1a0,0x1c0,0x1e0,
		  0x200,0x220,0x240,0x260,0x280,0x2a0,0x2c0,0x2e0,
		  0x300,0x320,0x340,0x360,0x380,0x3a0,0x3c0,0x3e0,0x3ff },
	},
};

static const struct {
	u8 int_field;		/* Interlaced field bit */
	u8 src_height_div;	/* Source height divisor */
	u8 dst_height_div;	/* Destination height divisor */
	u8 src_yoffset_div;	/* Source Y offset divisor */
	} scaler_modifier[][2][2] = {
		{	/* Progressive data */
			{	/* Non interlaced display */
				{	/* Not scaling */
					0,						/* Interlaced field bit			*/
					1,						/* Source height divisor		*/
					1,						/* Destination height divisor	*/
					1						/* Source Y offset divisor		*/
				},
				{	/* Scaling */
					0,						/* Interlaced field bit			*/
					1,						/* Source height divisor		*/
					1,						/* Destination height divisor	*/
					1						/* Source Y offset divisor		*/
				}
			},
			{	/* Interlaced display */
				{	/* Not scaling	*/
					1,						/* Interlaced field bit			*/
					2,						/* Source height divisor		*/
					2,						/* Destination height divisor	*/
					2						/* Source Y offset divisor		*/
				},
				{	/* Scaling */
					0,						/* Interlaced field bit			*/
					1,						/* Source height divisor		*/
					2,						/* Destination height divisor	*/
					1						/* Source Y offset divisor		*/
				}
			}
		},
		{	/* Interleaved fields */
			{	/* Non interlaced display */
				{	/* Not scaling */
					1,						/* Interlaced field bit			*/
					2,						/* Source height divisor		*/
					1,						/* Destination height divisor	*/
					1						/* Source Y offset divisor		*/
				},
				{	/* Scaling */
					1,						/* Interlaced field bit			*/
					2,						/* Source height divisor		*/
					1,						/* Destination height divisor	*/
					1						/* Source Y offset divisor		*/
				}
			},
			{	/* Interlaced display */
				{	/* Not scaling */
					1,						/* Interlaced field bit			*/
					2,						/* Source height divisor		*/
					2,						/* Destination height divisor	*/
					1						/* Source Y offset divisor		*/
				},
				{	/* Scaling */
					1,						/* Interlaced field bit			*/
					2,						/* Source height divisor		*/
					2,						/* Destination height divisor	*/
					1						/* Source Y offset divisor		*/
				}
			}
		},
		{	/* Single fields */
			{	/* Non interlaced display */
				{	/* Not scaling */
					0,						/* Interlaced field bit			*/
					1,						/* Source height divisor		*/
					1,						/* Destination height divisor	*/
					1						/* Source Y offset divisor		*/
				},
				{	/* Scaling */
					0,						/* Interlaced field bit			*/
					1,						/* Source height divisor		*/
					1,						/* Destination height divisor	*/
					1						/* Source Y offset divisor		*/
				}
			},
			{	/* Interlaced display */
				{	/* Not scaling */
					0,						/* Interlaced field bit			*/
					1,						/* Source height divisor		*/
					1,						/* Destination height divisor	*/
					2						/* Source Y offset divisor		*/
				},
				{	/* Scaling */
					0,						/* Interlaced field bit			*/
					1,						/* Source height divisor		*/
					1,						/* Destination height divisor	*/
					2						/* Source Y offset divisor		*/
				}																
			}
		}
};

#define SCALER_MODIFIER(priv,stream) \
	(scaler_modifier[(stream->info.var.vmode & FB_VMODE_INTERLACED) ? 1 : 0] \
	[(priv->mode.vmode & FB_VMODE_INTERLACED) ? 1 : 0][stream->vs_en])

static const struct {
	/* information for setting up pdp */
	u32 id;
	/* information for matching */
	int bpp, grayscale;
	struct fb_bitfield a, r, g, b;
	} pixfmts[] = {
		{
			.id = PDP_SURF_PIXFMT_RGB565,
			.bpp = 16,
			.r = { .offset = 11, .length = 5 },
			.g = { .offset = 5,	 .length = 6 },
			.b = { .offset = 0,	 .length = 5 },
		},
		{
			.id = PDP_SURF_PIXFMT_ARGB4444,
			.bpp = 16,
			.a = { .offset = 12, .length = 4 },
			.r = { .offset = 8,	 .length = 4 },
			.g = { .offset = 4,	 .length = 4 },
			.b = { .offset = 0,	 .length = 4 },
		},
		{
			.id = PDP_SURF_PIXFMT_ARGB1555,
			.bpp = 16,
			.a = { .offset = 15, .length = 1 },
			.r = { .offset = 10, .length = 5 },
			.g = { .offset = 5,	 .length = 5 },
			.b = { .offset = 0,	 .length = 5 },
		},
		{
			.id = PDP_SURF_PIXFMT_RGB888,
			.bpp = 24,
			.r = { .offset = 16, .length = 8 },
			.g = { .offset = 8,	 .length = 8 },
			.b = { .offset = 0,	 .length = 8 },
		},
		{
			.id = PDP_SURF_PIXFMT_ARGB8888,
			.bpp = 32,
			.a = { .offset = 24, .length = 8 },
			.r = { .offset = 16, .length = 8 },
			.g = { .offset = 8,	 .length = 8 },
			.b = { .offset = 0,	 .length = 8 },
		},
		{
			.id = PDP_SURF_PIXFMT_RGB101010,
			.bpp = 32,
			.r = { .offset = 20, .length = 10 },
			.g = { .offset = 10, .length = 10 },
			.b = { .offset = 0,	 .length = 10 },
		},
		{
			.id = PDP_SURF_PIXFMT_RGB121212,
			.bpp = 40,
			.r = { .offset = 24, .length = 12 },
			.g = { .offset = 12, .length = 12 },
			.b = { .offset = 0,	 .length = 12 },
		},
		{
			.id = PDP_SURF_PIXFMT_RGB161616,
			.bpp = 48,
			.r = { .offset = 32, .length = 16 },
			.g = { .offset = 16, .length = 16 },
			.b = { .offset = 0,	 .length = 16 },
		},
};

/* Bilinear filtering coefficients for 16 interpolation points in 0.8 format */
static const int bf_coeffs_i16[32] = {
		/* T = 0 */
		/* coeff[0][i] = 1 - (i / I) */
		/* coeff[0][0] is capped to fit within 0.8 format */
		255, 240, 224, 208, 192, 176, 160, 144,
		128, 112, 96, 80, 64, 48, 32, 16,

		/* T = 1 */
		/* coeff[1][i] = 1 - coeff[0][i] */
		1, 16, 32, 48, 64, 80, 96, 112,
		128, 144, 160, 176, 192, 208, 224, 240
};

enum {
	pixfmt_match_none = 0,
	pixfmt_match_bpp,
	pixfmt_match_alpha,
	pixfmt_match_bitsorgray,
	pixfmt_match_exact,
};

struct pdpfb_priv {
	void __iomem *base;
	void __iomem *mmu_base;
	struct pdpfb_mem_pool pools[PDPFB_MEMPOOL_NR_POOLS];
	unsigned int irq;
	struct clk *pdp_clk;
	struct clk *pixel_clk;
	int pdp_clk_en;
	int pixel_clk_en;
	struct device *device;
	struct pdpfb_cfg cfg;
	struct fb_videomode mode;
	struct pdpfb_updates upd;
	struct pdpfb_bgnd_col bgnd_col;
	struct pdpfb_csc_config csc_config;
	struct pdpfb_procamp procamp;
	struct pdpfb_gamma_config gamma_config;
	int alt3d_en;
	int alt3d_blend_sel;
	struct pdpfb_stream streams[PDPFB_MAX_NR_STREAMS];
	int stream_nr;
	u32 *pdir; /* The kernel virtual base address of the page directories */
	dma_addr_t pdir_bus; /* The bus base address of the page directories */
#ifdef CONFIG_FB_PDP2_PDUMP
	int pdump_pdir_mem_block;
	int pdump_next_mem_block;
	int pdump_suspended;
#ifdef __PDPFB_USE_OSAL__
	FILE * hOut2Txt;
#endif /* __PDPFB_USE_OSAL__ */
#endif /* CONFIG_FB_PDP2_PDUMP */

	spinlock_t irq_lock;
	int isr_disable;
	struct pdpfb_isr_data registered_isr[PDPFB_MAX_NR_ISRS];

	unsigned int vsync_count;
	unsigned int underrun_count[PDPFB_MAX_NR_STREAMS];
	unsigned int overrun_count[PDPFB_MAX_NR_STREAMS];
	struct dentry *stats_dir;
	struct timeval time_last_frame;
	unsigned int measured_fps;
	int is_sim;
};

/* Internal function prototypes */
static int pdpfb_configure_sync(struct pdpfb_priv *priv);
static int pdpfb_configure(struct pdpfb_priv *priv);
static void	pdpfb_set_mode(struct pdpfb_priv *priv,
						struct fb_var_screeninfo *var);
static int pdpfb_str_configure_addr(struct pdpfb_priv *priv,
									struct pdpfb_stream *stream);
static int pdpfb_str_change_mode(struct pdpfb_priv *priv,
								struct pdpfb_stream *stream);
static int pdpfb_str_check_geom(struct pdpfb_priv *priv,
						struct pdpfb_stream *stream, struct pdpfb_geom *geom);
static void pdpfb_str_configure(struct pdpfb_priv *priv,
							struct pdpfb_stream *stream);
static int pdpfb_register_isr(struct pdpfb_priv *priv, pdpfb_isr_t isr,
							void *arg, u32 mask);
static int pdpfb_unregister_isr(struct pdpfb_priv *priv, pdpfb_isr_t isr,
								void *arg, u32 mask);
static int pdpfb_init_mempool(struct pdpfb_priv *priv, int type,
							unsigned long phys, unsigned long len);

static void pdpfb_pdp_clk_enable(struct pdpfb_priv *priv)
{
	if (priv->pdp_clk != NULL) {
		if (!priv->pdp_clk_en) {
			clk_enable(priv->pdp_clk);
			priv->pdp_clk_en = 1;
		}
	}
}

static void pdpfb_pdp_clk_disable(struct pdpfb_priv *priv)
{
	if (priv->pdp_clk != NULL) {
		if (priv->pdp_clk_en) {
			clk_disable(priv->pdp_clk);
			priv->pdp_clk_en = 0;
		}
	}
}

static void pdpfb_pixel_clk_enable(struct pdpfb_priv *priv)
{
	if (priv->pixel_clk != NULL) {
		if (!priv->pixel_clk_en) {
			clk_prepare_enable(priv->pixel_clk);
			priv->pixel_clk_en = 1;
		}
	}
}

static void pdpfb_pixel_clk_disable(struct pdpfb_priv *priv)
{
	if (priv->pixel_clk != NULL) {
		if (priv->pixel_clk_en) {
			clk_disable_unprepare(priv->pixel_clk);
			priv->pixel_clk_en = 0;
		}
	}
}

/* Register access */

#define PDUMP_TAG "pdpfb pdump: "

#ifdef CONFIG_FB_PDP2_PDUMP_V1
#define PDUMP_WR_FMT "WRW :REG:%08x %08x -- %s\n"
#define PDUMP_RD_FMT "RDW :REG:%08x -- %s\n"
#endif

#ifdef CONFIG_FB_PDP2_PDUMP_V2
#define PDUMP_WR_FMT "WRW :REGMET_PDP:0x%08x 0x%08x -- %s\n"
#define PDUMP_RD_FMT "RDW :REGMET_PDP:0x%08x -- %s\n"
#define PDUMP_MMU_WR_FMT "WRW :REGPDP_BIF:0x%08x 0x%08x\n"
#define PDUMP_MMU_RD_FMT "RDW :REGPDP_BIF:0x%08x\n"
#endif

#ifdef __PDPFB_USE_OSAL__
#define PDUMP_LOG(...) if (!priv->pdump_suspended) \
	fprintf(priv->hOut2Txt, __VA_ARGS__)
#else
#define PDUMP_LOG(...) if (!priv->pdump_suspended) \
	printk(PDUMP_TAG __VA_ARGS__)
#endif

#ifdef CONFIG_FB_PDP2_PDUMP
static const PDPFB2_REGFIELDS;
static const char * pdpfb_get_reg_name(unsigned int reg_offs)
{
	int i;
	for (i = 0; i < ARRAY_SIZE(pdpfb2_regfields); i++) {
		if (pdpfb2_regfields[i].reg_offset == reg_offs)
			return pdpfb2_regfields[i].reg_name;
	}
	return NULL;
}
#endif

static void pdpfb_write(struct pdpfb_priv *priv,
						unsigned int reg_offs, unsigned int data)
{
#ifdef CONFIG_FB_PDP2_PDUMP
	if (reg_offs != PDP_INTCLR_OFFSET) {
		/* Only log the write if the value has changed */
		if (ioread32(priv->base + reg_offs) != data)
			PDUMP_LOG(PDUMP_WR_FMT, reg_offs, data,
						pdpfb_get_reg_name(reg_offs));
	}
#endif
	iowrite32(data, priv->base + reg_offs);
}

static unsigned int pdpfb_read(struct pdpfb_priv *priv, unsigned int reg_offs)
{
#ifdef __PDPFB_USE_OSAL__
	IMG_ASSERT (reg_offs != PDPFB_NO_REG);
#endif
	return ioread32(priv->base + reg_offs);
}

#ifdef CONFIG_FB_PDP2_MMU
static void pdpfb_mmu_write(struct pdpfb_priv *priv,
							unsigned int reg_offs, unsigned int data)
{
#ifdef CONFIG_FB_PDP2_PDUMP
	PDUMP_LOG(PDUMP_MMU_WR_FMT, reg_offs, data);
#endif
	iowrite32(data, priv->mmu_base + reg_offs);
}

static unsigned int pdpfb_mmu_read(struct pdpfb_priv *priv,
								unsigned int reg_offs)
{
	return ioread32(priv->mmu_base + reg_offs);
}
#endif /* ifdef CONFIG_FB_PDP2_MMU */

static void pdpfb_write_base_addr(struct pdpfb_priv *priv,
				struct pdpfb_stream *stream, unsigned int base_addr_reg_offs,
				unsigned int base_offset)
{
	const u32 base_addr = stream->videomem_pdp_base + base_offset;
	pdpfb_write(priv, base_addr_reg_offs, base_addr);
	#if defined(CONFIG_FB_PDP2_PDUMP) && !defined(CONFIG_FB_PDP2_MMU)
		/*
		 * If using the mmu, the virtual base address already written in the
		 * pdump will be correct, if not it must be overwritten here with the
		 * memory block reference.
		*/
		if (priv->pdump_next_mem_block != 0) {
			PDUMP_LOG("WRW :MEM:$1 :MEM:BLOCK_%d:0x%X -- Place image base address in $1\n",
				stream->pdump_image_mem_block, base_offset);
			PDUMP_LOG("SHR :MEM:$2 :MEM:$1 32 -- Place upper address bits in $2\n");
			PDUMP_LOG("SHL :MEM:$2 :MEM:$2 %d -- Position upper address bits in UPPER_ADDRESS_FIXED field\n",
				PDP_BIF_ADDRESS_CONTROL_UPPER_ADDRESS_FIXED_SHIFT);
			PDUMP_LOG("OR :MEM:$2 :MEM:$2 %d -- Set MMU_BYPASS field\n",
				PDP_BIF_ADDRESS_CONTROL_MMU_BYPASS_MASK);
			PDUMP_LOG("WRW :REGPDP_BIF:0x%08x :MEM:$2 -- MMU_ADDRESS_CONTROL\n",
				PDP_BIF_ADDRESS_CONTROL_OFFSET);
			PDUMP_LOG("AND :MEM:$1 :MEM:$1 0xffffffff -- Clear upper address bits\n");
			PDUMP_LOG("WRW :REGMET_PDP:0x%08x :MEM:$1 -- %s\n",
				base_addr_reg_offs, pdpfb_get_reg_name(base_addr_reg_offs));
		}
	#endif
}

static void pdpfb_flush_regs(struct pdpfb_priv *priv)
{
	u32 reg_upd_ctrl;
	reg_upd_ctrl = pdpfb_read(priv, PDP_REGISTER_UPDATE_CTRL_OFFSET);
	SET_FIELD(reg_upd_ctrl, PDP_REGISTER_UPDATE_CTRL_REGISTERS_VALID, 1);
	SET_FIELD(reg_upd_ctrl, PDP_REGISTER_UPDATE_CTRL_USE_VBLANK, 1);
	pdpfb_write(priv, PDP_REGISTER_UPDATE_CTRL_OFFSET, reg_upd_ctrl );
}

/* Scaler Code */

static inline int fix(int p, int a)
{
	return a << p;
}

static inline int fix_div(int p, int n, int d)
{
	return (n << p) / d;
}

static inline int fix_divl(int p, int n, int d)
{
	return (int)(((u64)n << p) / d);
}

static inline int fix_mul(int p, int a, int b)
{
	return (a * b) >> p;
}

static inline int fix_mull(int p, int a, int b)
{
	return (int)(((u64)a * (u64)b) >> p);
}

static inline int fix_sin(int p, int x)
{
	/* tailor expansion */
	int pow;
	int sum;

	/* in [-pi,pi] */
	int pi2 = fix_divl(p, 6283185, 1000000);
	int pi = pi2 / 2;
	x = x % pi2;
	if (x > pi)
		x -= pi2;
	else if (x < -pi)
		x += pi2;

	sum = x;

	/* x represents x^2 from now on */
	x = fix_mull(p, x, x);

	/* -x^3/3! */
	pow = fix_mull(p, sum, x);
	sum -= pow / 6;
	/* +x^5/5! */
	pow = fix_mull(p, pow, x);
	sum += pow / 120;
	/* -x^7/7! */
	pow = fix_mull(p, pow, x);
	sum -= pow / 5040;
	/* +x^9/9! */
	pow = fix_mull(p, pow, x);
	sum += pow / 362880;

	return sum;
}

static inline int fix_cos(int p, int x)
{
	/* tailor expansion */
	int pow;
	int sum = fix(p, 1);

	/* in [-pi,pi] */
	int pi2 = fix_divl(p, 6283185, 1000000);
	int pi = pi2 / 2;
	x = x % pi2;
	if (x > pi)
		x -= pi2;
	else if (x < -pi)
		x += pi2;

	/* x represents x^2 from now on */
	x = fix_mull(p, x, x);

	/* -x^2/2! */
	pow = x;
	sum -= pow / 2;
	/* +x^4/4! */
	pow = fix_mull(p, pow, x);
	sum += pow / 24;
	/* -x^6/6! */
	pow = fix_mull(p, pow, x);
	sum -= pow / 720;
	/* +x^8/8! */
	pow = fix_mull(p, pow, x);
	sum += pow / 40320;
	/* -x^10/10! */
	pow = fix_mull(p, pow, x);
	sum -= pow / 3628800;

	return sum;
}

static int pdpfb_str_scale_coeffs_stale(u32 new_res, u32 old_res,
										u32 new_size, u32 old_size)
{
	/* scale factor clamped to max of 1 */
	if (new_size >= new_res && old_size >= old_res)
		return 0;
	/* has scale factor changed? */
	return new_size * old_res != old_size * new_res;
}

static void pdpfb_str_calc_scale_coeffs(u32 res, u32 size,
										int T,	/* taps */
										int I,	/* interpolation points */
										int *coeffs, int count)
{
	int midpoint = T * I / 2;
#define SCALE_PRECISION 8
#define RESULT_PRECISION 6
#define DIV(a, b)	fix_div(SCALE_PRECISION, a, b)
#define DIVL(a, b)	fix_divl(SCALE_PRECISION, a, b)
#define MUL(a, b)	fix_mul(SCALE_PRECISION, a, b)
#define SIN(a)		fix_sin(SCALE_PRECISION, a)
#define COS(a)		fix_cos(SCALE_PRECISION, a)
#define FIX(a)		fix(SCALE_PRECISION, a)
#define RESULT(a)	((a) >> (SCALE_PRECISION - RESULT_PRECISION))
	int fpi = DIVL(3141593, 1000000);
	int fS = min(FIX(1), /* scale factor clamped to 1 */
		DIV(size, res));
	int fA = DIV(54, 100);
	int c, i;

	int fpiS = MUL(fpi, fS);
	int f2piSoT = fpiS * 2 / T;
	int f1mA = FIX(1) - fA;
	int fT = FIX(T);
	int ftotal = 0;

	/* Calculate coefficients */
	for (c = 0; c <= midpoint; ++c) {
		/* t + i/I */
		int ftpioI = FIX(c) / I;
		int fx = MUL(fpiS, fT / 2 - ftpioI);
		int fco = fA - MUL(f1mA, COS(MUL(f2piSoT, ftpioI)));
		if (fx > DIV(1, 10))
			fco = MUL(DIV(SIN(fx), fx), fco);
		ftotal += fco;
		coeffs[c] = fco;
	}

	/*
	 * Interpolation points I/2+1 to I-1 have same values as 1 to I/2-1,
	 * but in opposite order, so we only need to normalise the first half
	 * of the interpolation points taking mirroring into account.
	 */
	for (i = 0; i <= I / 2; ++i) {
		int sum, err, dir = 0;

		/*
		 * Normalise coefficients in each interpolation point and
		 * convert into result format.
		 */
		sum = 0;
		for (c = i; c <= midpoint; c += I)
			sum += coeffs[c];
		for (; c < midpoint * 2; c += I)
			sum += coeffs[midpoint * 2 - c];
		/* Careful not to modify a coefficient twice */
		for (c = i; c <= midpoint; c += I)
			coeffs[c] = RESULT(DIV(coeffs[c], sum));
		if (i & (I / 2 - 1))
			for (c = I - i; c <= midpoint; c += I)
				coeffs[c] = RESULT(DIV(coeffs[c], sum));

		/*
		 * Find fixed point error from normalisation.
		 */
		sum = 0;
		for (c = i; c <= midpoint; c += I)
			sum += coeffs[c];
		for (; c < midpoint * 2; c += I)
			sum += coeffs[midpoint * 2 - c];
		err = sum - (1 << RESULT_PRECISION);
		if (err > 0)
			dir = -1;
		else if (err < 0)
			dir = 1;
		else
			continue;

#ifdef DEBUG_SCALE_COEFFS
		printk(KERN_DEBUG "pdp: i=%d, err=%d\n", i, err);
#define SCALE_DEBUG(C, N) \
		printk(KERN_DEBUG "pdp:   adj*%d [%d] to 0x%x (err=%d)\n", \
				(N), (C), coeffs[(C)], err)
#else
#define SCALE_DEBUG(C, N) do {} while (0)
#endif

		/*
		 * Distribute the error over the tap coefficients, preferring
		 * to change central coefficients.
		 */
		if (i == 0) {
			/*
			 * Special case: contains midpoint. Work out from
			 * center until no error or an odd value (which can be
			 * fixed by adjusting midpoint again).
			 */
			while (err > 1 || err < -1) {
				coeffs[midpoint] += dir;
				err += dir;
				SCALE_DEBUG(midpoint, 1);
				for (c = midpoint - I;
					 (err > 1 || err < -1) && c >= I;
					 c -= I) {
					coeffs[c] += dir;
					err += dir * 2;
					SCALE_DEBUG(c, 2);
				}
			}
			if (err & 1) {
				coeffs[midpoint] += dir;
				err += dir;
				SCALE_DEBUG(midpoint, 1);
			}
		} else if (i == I / 2) {
			/*
			 * Special case: tap coeffs mirror. This also means err
			 * is always even with this interpolation point.
			 */
			int last = i + I * (T / 2 - 1);
#ifdef CHECK_SCALE_COEFFS
			WARN(err & 1, "err (=%d) should be even for i=%d\n",
					err, i);
#endif
			while (err) {
				for (c = last; err && c >= 0; c -= I) {
					coeffs[c] += dir;
					err += dir * 2;
					SCALE_DEBUG(c, 2);
				}
			}
		} else {
			/*
			 * No tap coefficient mirrors another. Work out from
			 * center adjusting values.
			 */
			int last = midpoint - i;
			int offset = i * 2 - I;
			while (err) {
				for (c = last; err && c >= 0; c -= I) {
					coeffs[c] += dir;
					err += dir;
					SCALE_DEBUG(c, 1);
					if (!err)
						break;
					coeffs[c+offset] += dir;
					err += dir;
					SCALE_DEBUG(c+offset, 1);
				}
			}
		}
#ifdef CHECK_SCALE_COEFFS
		sum = 0;
		for (c = i; c <= midpoint; c += I)
			sum += coeffs[c];
		for (; c < midpoint * 2; c += I)
			sum += coeffs[midpoint * 2 - c];
		err = sum - (1 << RESULT_PRECISION);
		WARN_ONCE(err, "Scale coefficients not normalised"
				   " (i=%d, err=%d)\n", i, err);
#endif
	}

	/* reflect about midpoint in spare coeffs */
	if (count > midpoint * 2)
		count = midpoint * 2;
	for (c = midpoint + 1; c < count; ++c)
		coeffs[c] = coeffs[2 * midpoint - c];
#undef SCALE_PRECISION
#undef RESULT_PRECISION
#undef DIV
#undef DIVL
#undef MUL
#undef SIN
#undef COS
#undef FIX
#undef RESULT
}

static void pdpfb_str_calc_pass_thru_coeffs(int T,	/* taps */
											int I,	/* interpolation points */
											int *coeffs, int count)
{
	const int midpoint = T * I / 2;
	int c;

	/* Zero coefficients */
	for (c = 0; c < count; ++c)
		coeffs[c] = 0;

	/* 
	 * Set all coeffs to 0x40 to select single tap, apart from mirror at centre
	 * of table (I/2) where coeffs of 0x20 are used to filter 2 taps.
	 * In practice, the scaler should not select phase I/2 as this pass through
	 * set of coefficients should be used for 1:1 scaling when phase = 0 and
	 * centre tap T/2 is selected.
	*/
	for (c = midpoint; c < midpoint + I / 2; ++c)
		coeffs[c] = 0x40;

	coeffs[midpoint + I / 2] = 0x20;
	coeffs[midpoint - I / 2] = 0x20;

	for (c = midpoint - I / 2 + 1; c < midpoint; ++c)
		coeffs[c] = 0x40;
}

static void pdpfb_calculate_vinitial(u32 stream_vmode, u32 display_vmode,
									u32 taps, u32 vpitch, u32 *vinitial0,
									u32 *vinitial1)
{
	/* pitch and vinitial values are in Q11 fixed point format */

	/* Set vinitial0 and vinitial1 to the centre tap position as the default */
	*vinitial0 = taps << 10; /* (taps / 2) in Q11 format */
	*vinitial1 = *vinitial0;

	if (stream_vmode & FB_VMODE_INTERLACED) {
		if (display_vmode & FB_VMODE_INTERLACED) {
			/* Interlaced data on an interlaced display */
			if (vpitch > (1 << 11))  {
				*vinitial1 += (vpitch - (1 << 11)) >> 1; /* ((vpitch-1) / 2) */
			} else if (vpitch < (1 << 11)) {
				*vinitial0 += ((1 << 11) - vpitch) >> 1; /* ((1-vpitch) / 2) */
			}
		} else {
			/* Interlaced data on a progressive display */
			*vinitial0 += 1 << 10;	/* 0.500 in Q11 format */
		}
	} else if (display_vmode & FB_VMODE_INTERLACED) {
		/* Progressive data on an interlaced display */
		*vinitial1 += vpitch >> 1;	/* (vpitch / 2) */
	}
}

static int pdpfb_str_configure_hscale(struct pdpfb_priv *priv,
									struct pdpfb_stream *stream)
{
	u32 reg;
	u32 xres = stream->info.var.xres;
	u32 geomw = stream->geom.w ? stream->geom.w : stream->info.var.xres;

	if (stream->caps.has_scaler) {
		u32 i;
		u32 scale_width;
		u32 hpitch, hinitial, hpostclip;

		stream->hs_en = (geomw != xres);

		/* Workaround for SBM 73585 */
		if (stream->hskip) {
			/* Force scaler on */
			stream->hs_en = 1;

			/* Re-write source width to allow for additional pixels to skip */
			reg = pdpfb_read(priv, stream->regs.size);
			if (stream->split_scale) {
				SET_FIELD(reg, PDP_GRPH1SIZE_GRPH1WIDTH,
					xres + stream->caps.hs_taps + stream->hskip - 1);
			} else {
				SET_FIELD(reg, PDP_GRPH1SIZE_GRPH1WIDTH,
					xres + stream->hskip - 1);
			}
			pdpfb_write(priv, stream->regs.size, reg);
		}

		reg = pdpfb_read(priv, stream->regs.scale_ctrl);
		SET_FIELD(reg, PDP_VID1SCALECTRL_VID1HSCALEBP, !stream->hs_en);
		pdpfb_write(priv, stream->regs.scale_ctrl, reg);

		if (!stream->hs_en) {
			reg = pdpfb_read(priv, stream->regs.scale_size);
			SET_FIELD(reg, PDP_VID1SCALESIZE_VID1SCALEWIDTH, geomw - 1);
			pdpfb_write(priv, stream->regs.scale_size, reg);
			return 0;
		}

		reg = pdpfb_read(priv, stream->regs.scale_ctrl);
		SET_FIELD(reg, PDP_VID1SCALECTRL_VID1HORDER, stream->bilinear_en ? 2 : 7);
		pdpfb_write(priv, stream->regs.scale_ctrl, reg);

		hpitch = (fix_div(12, xres, geomw) + 1) >> 1;
		hinitial = fix_div(11, stream->caps.hs_taps, 2);
		hpostclip = stream->clip_geom.w ? stream->clip_geom.x : 0;
		scale_width = stream->clip_geom.w ?	stream->clip_geom.w : geomw;

		/*
		   For the right image of a split scaling operation, add the fractional
		   part of (hpitch * scale_width) and the number of taps to hinitial.
		*/
		if (stream->split_scale == PDP_SPLIT_SCALE_RIGHT_PLANE)
			hinitial += ((hpitch * scale_width) & 0x3FF) + (stream->caps.hs_taps << 11);
		
		/*
		   For the left image of a split scaling operation, add 1 to the
		   scale width (SBM 73740)
		*/
		if (stream->split_scale == PDP_SPLIT_SCALE_LEFT_PLANE)
			scale_width += 1;

		/* 
		 * Workaround for SBM 73585 - up to 19 pixels are skipped using
		 * hinitial, the rest are skipped using post scale clipping.
		 */
		if (stream->hskip < 19) {
			hinitial += stream->hskip << 11;
		} else {
			/* Calculate hpostclip to add: 6.26 / x.11 = x.15 */
			u32 hpostclip_to_add = ((stream->hskip - 19) << 26) / hpitch;
			hpostclip_to_add >>= 15;
			hpostclip += hpostclip_to_add;
			if (hpostclip > MAX_FIELD_VALUE(PDP_VID1POSTSKIPCTRL_VID1HPOSTCLIP)) {
				hpostclip = MAX_FIELD_VALUE(PDP_VID1POSTSKIPCTRL_VID1HPOSTCLIP);
				dev_warn(priv->device, "insufficient post scale clip pixels available to achieve the required configuration");
			}
			hinitial += 19 << 11;
		}

		reg = pdpfb_read(priv, stream->regs.hsinit);
		SET_FIELD(reg, PDP_VID1HSINIT_VID1HINITIAL, hinitial);
		SET_FIELD(reg, PDP_VID1HSINIT_VID1HPITCH, hpitch);
		pdpfb_write(priv, stream->regs.hsinit, reg);

		reg = pdpfb_read(priv, stream->regs.postskip_ctrl);
		SET_FIELD(reg, PDP_VID1POSTSKIPCTRL_VID1HPOSTCLIP, hpostclip);
		pdpfb_write(priv, stream->regs.postskip_ctrl, reg);

		reg = pdpfb_read(priv, stream->regs.scale_size);
		SET_FIELD(reg, PDP_VID1SCALESIZE_VID1SCALEWIDTH,
					scale_width + hpostclip - 1);
		pdpfb_write(priv, stream->regs.scale_size, reg);

		if (stream->bilinear_en) {
			if (stream->caps.hs_interpolation_points != 16) {
				dev_err(priv->device, "unsupported number of interpolation points");
				return -EINVAL;
			}

			/* Number of coefficients = 2 * I */
			/* Number of coeff registers to write = I / 2 */
			for (i = 0; i < stream->caps.hs_interpolation_points / 2; ++i) {
				reg =  (0xFF & bf_coeffs_i16[i << 2])
					| ((0xFF & bf_coeffs_i16[(i << 2) + 1]) << 8)
					| ((0xFF & bf_coeffs_i16[(i << 2) + 2]) << 16)
					| ((0xFF & bf_coeffs_i16[(i << 2) + 3]) << 24);
				pdpfb_write(priv, stream->regs.first_hscoeff + (i << 2), reg);
			}
		}
		else {
			/*
			   Don't calculate coefficients unless they haven't been
			   calculated before, or something has changed.
			*/
			if (!stream->hs_oldres ||
				pdpfb_str_scale_coeffs_stale(stream->info.var.xres,
										stream->hs_oldres,
										stream->geom.w,
										stream->hs_oldsize)) {
				stream->hs_oldres = stream->info.var.xres;
				stream->hs_oldsize = stream->geom.w;

				if (geomw == xres) {
					pdpfb_str_calc_pass_thru_coeffs(stream->caps.hs_taps,
										stream->caps.hs_interpolation_points,
										stream->hscoeffs,
										ARRAY_SIZE(stream->hscoeffs));
				} else {
					pdpfb_str_calc_scale_coeffs(stream->info.var.xres,
										stream->geom.w,
										stream->caps.hs_taps,
										stream->caps.hs_interpolation_points,
										stream->hscoeffs,
										ARRAY_SIZE(stream->hscoeffs));
				}
			}

			/* Number of coefficients = (T/2)*I + 1 */
			/* Number of coeff registers to write = ((T/2)*I)/4 + 1 */
			/* T = 8, so registers to write = I + 1 */
			for (i = 0; i < stream->caps.hs_interpolation_points; ++i) {
				reg =  (0xFF & stream->hscoeffs[i << 2])
					| ((0xFF & stream->hscoeffs[(i << 2) + 1]) << 8)
					| ((0xFF & stream->hscoeffs[(i << 2) + 2]) << 16)
					| ((0xFF & stream->hscoeffs[(i << 2) + 3]) << 24);
				pdpfb_write(priv, stream->regs.first_hscoeff + (i << 2), reg);
			}
			pdpfb_write(priv, stream->regs.last_hscoeff,
						0xFF & stream->hscoeffs[i << 2]);
		}
	} else {
		/* No scaler */
		if (stream->caps.has_doubler) {
			reg = pdpfb_read(priv, stream->regs.blend2_r);
			SET_FIELD(reg, PDP_GRPH1BLND2_R_GRPH1PIXDBL, (geomw > xres));
			pdpfb_write(priv, stream->regs.blend2_r, reg);
		}
		if (stream->caps.has_decimator) {
			reg = pdpfb_read(priv, stream->regs.decimate_ctrl);
			SET_FIELD(reg, PDP_GRPH1DECIMATE_CTRL_GRPH1DECIMATE_PIXEL_HALVE,
						(geomw < xres));
			if (geomw < xres)
				SET_FIELD(reg, PDP_GRPH1DECIMATE_CTRL_GRPH1DECIMATE_EN, 1);
			pdpfb_write(priv, stream->regs.decimate_ctrl, reg);
		}
	}
	return 0;
}

#define MAX_LINE_DISCARD_COUNT \
	(MAX_FIELD_VALUE(PDP_GRPH1DECIMATE_CTRL_GRPH1DECIMATE_LINE_DISCARD_COUNT))

static void pdpfb_str_configure_decimator(struct pdpfb_priv *priv,
										struct pdpfb_stream *stream,
										u32 yres, u32 geomh)
{
	u32 read_or_skip_n = 0;
	u32 reg = pdpfb_read(priv, stream->regs.decimate_ctrl);

	if (geomh < yres) {
		if ((geomh * 2) > yres) {
			/* Pitch is between 1 and 2 */
			/* Use 'Read N, skip 1' */
			SET_FIELD(reg,
					PDP_GRPH1DECIMATE_CTRL_GRPH1DECIMATE_LINE_DISCARD_MODE, 1);
			/*
			* n = geomh / (yres - geomh)
			* Round n down to the nearest integer to give resulting
			* pitch less than or equal the target value (the rounding
			* down happens as a result of the truncation)
			*/
			read_or_skip_n = geomh / (yres - geomh);
		} else {
			/* Pitch is greater than or equal to 2 */
			/* Use 'Read 1, skip N' */
			SET_FIELD(reg,
					PDP_GRPH1DECIMATE_CTRL_GRPH1DECIMATE_LINE_DISCARD_MODE, 0);
			/*
			* n = (yres - geomh) / geomh
			* Round up to the nearest integer to give resulting pitch
			* less than or equal the target value by adding the
			* denominator to the numerator and subtracting 1
			*/
			read_or_skip_n = (yres - 1) / geomh;
		}
		/* Clip read_or_skip_n to the maximum */
		if (read_or_skip_n > MAX_LINE_DISCARD_COUNT)
			read_or_skip_n = MAX_LINE_DISCARD_COUNT;

		SET_FIELD(reg, PDP_GRPH1DECIMATE_CTRL_GRPH1DECIMATE_EN, 1);
	}
	SET_FIELD(reg, PDP_GRPH1DECIMATE_CTRL_GRPH1DECIMATE_LINE_DISCARD_COUNT,
				read_or_skip_n);
	pdpfb_write(priv, stream->regs.decimate_ctrl, reg);
}

static int pdpfb_str_configure_vscale(struct pdpfb_priv *priv,
									struct pdpfb_stream *stream)
{
	u32 reg;
	u32 yres = stream->info.var.yres
				/ SCALER_MODIFIER(priv,stream).src_height_div;
	u32 geomh = (stream->geom.h ? stream->geom.h : stream->info.var.yres)
				/ SCALER_MODIFIER(priv,stream).dst_height_div;

	if (stream->caps.has_scaler) {
		u32 i;
		u32 vpitch;
		u32 hs_before_vs;
		u32 vinitial0;
		u32 vinitial1;
		u32 scale_height;

		stream->vs_en = (geomh != yres);
		vpitch = fix_div(11, yres - 2, geomh); /* SBM 72977 */
		pdpfb_calculate_vinitial(stream->info.var.vmode, priv->mode.vmode,
									stream->caps.vs_taps, vpitch,
									&vinitial0, &vinitial1);

		if ((vinitial0 != vinitial1) && !stream->vs_en) {
			/* Force scaler on */
			stream->vs_en = 1;

			/*
			 * Forcing the scaler on has changed the scaler modifier values,
			 * so recalculate yres, geomh and vpitch
			 */
			yres = stream->info.var.yres
					/ SCALER_MODIFIER(priv,stream).src_height_div;
			geomh = (stream->geom.h ? stream->geom.h : stream->info.var.yres)
					/ SCALER_MODIFIER(priv,stream).dst_height_div;
			vpitch = fix_div(11, yres - 2, geomh); /* SBM 72977 */		
			pdpfb_calculate_vinitial(stream->info.var.vmode, priv->mode.vmode,
										stream->caps.vs_taps, vpitch,
										&vinitial0, &vinitial1);
		}

		reg = pdpfb_read(priv, stream->regs.scale_ctrl);
		SET_FIELD(reg, PDP_VID1SCALECTRL_VID1VSCALEBP, !stream->vs_en);
		pdpfb_write(priv, stream->regs.scale_ctrl, reg);

		scale_height = (stream->vs_en && stream->clip_geom.h) ?
						stream->clip_geom.h + stream->clip_geom.y : geomh;
		reg = pdpfb_read(priv, stream->regs.scale_size);
		SET_FIELD(reg, PDP_VID1SCALESIZE_VID1SCALEHEIGHT, scale_height - 1);
		pdpfb_write(priv, stream->regs.scale_size, reg);

		reg = pdpfb_read(priv, stream->regs.postskip_ctrl);
		SET_FIELD(reg, PDP_VID1POSTSKIPCTRL_VID1VPOSTCLIP,
					stream->clip_geom.h ? stream->clip_geom.y : 0);
		pdpfb_write(priv, stream->regs.postskip_ctrl, reg);

		if (!stream->vs_en)
			return 0;

		hs_before_vs = (yres > priv->cfg.linestore_len);
		reg = pdpfb_read(priv, stream->regs.scale_ctrl);
		SET_FIELD(reg, PDP_VID1SCALECTRL_VID1HSBEFOREVS, hs_before_vs);
		SET_FIELD(reg, PDP_VID1SCALECTRL_VID1VSURUNCTRL, 1);
		SET_FIELD(reg, PDP_VID1SCALECTRL_VID1VORDER, stream->bilinear_en ? 2 : 3);
		SET_FIELD(reg, PDP_VID1SCALECTRL_VID1VPITCH, vpitch);
		pdpfb_write(priv, stream->regs.scale_ctrl, reg);

		reg = pdpfb_read(priv, stream->regs.vsinit);
		SET_FIELD(reg, PDP_VID1VSINIT_VID1VINITIAL0, vinitial0);
		SET_FIELD(reg, PDP_VID1VSINIT_VID1VINITIAL1, vinitial1);
		pdpfb_write(priv, stream->regs.vsinit, reg);

		if (stream->bilinear_en) {
			if (stream->caps.vs_interpolation_points != 16) {
				dev_err(priv->device, "unsupported number of interpolation points");
				return -EINVAL;
			}

			/* Number of coefficients = 2 * I */
			/* Number of coeff registers to write = I / 2 */
			for (i = 0; i < stream->caps.vs_interpolation_points / 2; ++i) {
				reg =  (0xFF & bf_coeffs_i16[i << 2])
					| ((0xFF & bf_coeffs_i16[(i << 2) + 1]) << 8)
					| ((0xFF & bf_coeffs_i16[(i << 2) + 2]) << 16)
					| ((0xFF & bf_coeffs_i16[(i << 2) + 3]) << 24);
				pdpfb_write(priv, stream->regs.first_vscoeff + (i << 2), reg);
			}
		}
		else {
			/*
			   Don't calculate coefficients unless they haven't been
			   calculated before, or something has changed.
			*/
			if (!stream->vs_oldres ||
				pdpfb_str_scale_coeffs_stale(yres,
										stream->vs_oldres,
										geomh,
										stream->vs_oldsize)) {
				stream->vs_oldres = yres;
				stream->vs_oldsize = geomh;
				pdpfb_str_calc_scale_coeffs(yres,
											geomh,
											stream->caps.vs_taps,
											stream->caps.vs_interpolation_points,
											stream->vscoeffs,
											ARRAY_SIZE(stream->vscoeffs));
			}

			/* Number of coefficients = (T/2)*I + 1 */
			/* Number of coeff registers to write = ((T/2)*I)/4 + 1 */
			/* T = 4, so registers to write = I/2 + 1 */
			for (i = 0; i < (stream->caps.vs_interpolation_points / 2); ++i) {
				u32 val =  (0xFF & stream->vscoeffs[i << 2])
					| ((0xFF & stream->vscoeffs[(i << 2) + 1]) << 8)
					| ((0xFF & stream->vscoeffs[(i << 2) + 2]) << 16)
					| ((0xFF & stream->vscoeffs[(i << 2) + 3]) << 24);
				pdpfb_write(priv, stream->regs.first_vscoeff + (i << 2), val);
			}
			pdpfb_write(priv, stream->regs.last_vscoeff,
						0xFF & stream->vscoeffs[i << 2]);
		}
	} else {
		/* No scaler */
		if (stream->caps.has_doubler) {
			reg = pdpfb_read(priv, stream->regs.blend2_r);
			SET_FIELD(reg, PDP_GRPH1BLND2_R_GRPH1LINDBL, (geomh > yres));
			pdpfb_write(priv, stream->regs.blend2_r, reg);
		}
		if (stream->caps.has_decimator) 
			pdpfb_str_configure_decimator(priv, stream, yres, geomh);
	}
	return 0;
}

/* Procamp code */

#define DIVL(a, b)	fix_divl(PDP_PROCAMP_HSBC_FRACTIONAL_BITS, a, b)
#define MULL(a, b)	fix_mull(PDP_PROCAMP_HSBC_FRACTIONAL_BITS, a, b)
#define SIN(a)		fix_sin(PDP_PROCAMP_HSBC_FRACTIONAL_BITS, a)
#define COS(a)		fix_cos(PDP_PROCAMP_HSBC_FRACTIONAL_BITS, a)
#define FIX(a)		fix(PDP_PROCAMP_HSBC_FRACTIONAL_BITS, a)

static void pdpfb_fix_procamp_matrix_mul(const int matrix_a[3][3],
							const int matrix_b[3][3], int result_matrix[3][3])
{
	int i;
	int j;
	int k;
	int sum;
	for (i = 0; i < 3; i++) {
		for (j = 0; j < 3; j++) {
			sum = 0;
			for (k = 0; k < 3; k++)
				sum += MULL(matrix_a[i][k], matrix_b[k][j]);
			result_matrix[i][j] = sum;
		}
	}
}

static void pdpfb_calc_procamp_coefs(const struct pdpfb_procamp_hsbc *hsbc,
								const u32 output_range,
								struct pdpfb_procamp_coefs *coefs,
								struct pdpfb_procamp_offsets *input_offsets,
								struct pdpfb_procamp_offsets *output_offsets)
{
	const int one_degree_in_radians = DIVL(31415927, (180*10000000));
	/* In x.25 format */
	const int rgb_to_yuv[3][3] = {	{0x00839581,  0x01020C4A,  0x00322D0E},
									{-0x004BC6A8, -0x0094FDF4,	0x00E0C49C},
									{0x00E0C49C,  -0x00BC6A7F, -0x00245A1D}	};
	/* In x.25 format */
	const int yuv_to_rgb[3][3] = {	{0x0253F7CF,  0x00000000,  0x033126E9},
									{0x0253F7CF, -0x00C83127, -0x01A04189},
									{0x0253F7CF,  0x0409374C,  0x00000000}	};
	int inverted_scale_factor;
	int output_offset;
	int hue_in_radians;
	int cos_hue;
	int sin_hue;
	int contrast_times_saturation;
	int result_matrix[2][3][3];

	switch (output_range) {
	case 0:
		inverted_scale_factor = FIX(1);
		output_offset = 0;
		break;
	case 1:
		inverted_scale_factor = DIVL(876, 1023);
		output_offset = fix(PDP_PROCAMP_OUTPUT_OFFSET_FRACTIONAL_BITS, 64);
		break;
	case 2:
	default:
		inverted_scale_factor = DIVL(640, 1023);
		output_offset = fix(PDP_PROCAMP_OUTPUT_OFFSET_FRACTIONAL_BITS, 192);
		break;
	}

	/*
	 * Cosine is symmetrical about 0 degrees, hence cos(theta)=cos(-theta)
	 * Sine is not symmetrical, however sin(theta)=-1 * sin(-theta)
	 */
	hue_in_radians = hsbc->hue;
	if (hue_in_radians < 0)
		hue_in_radians = hue_in_radians * -1;

	hue_in_radians = MULL(hue_in_radians, one_degree_in_radians);
	cos_hue = COS(hue_in_radians);
	sin_hue = SIN(hue_in_radians);

	if (hsbc->hue < 0)
		sin_hue = -1 * sin_hue;

	contrast_times_saturation = MULL(hsbc->contrast, hsbc->saturation);

	/*
	 * Create the HSBC matrix:
	 * [ con,		0,						0					]
	 * [ 0,			(con*sat*cos(hue)),		(con*sat*sin(hue))	]
	 * [ 0,			-(con*sat*sin(hue)),	(con*sat*cos(hue))	]
	 */
	result_matrix[0][0][0] = hsbc->contrast;
	result_matrix[0][0][1] = 0;
	result_matrix[0][0][2] = 0;

	result_matrix[0][1][0] = 0;
	result_matrix[0][1][1] = MULL(contrast_times_saturation, cos_hue);
	result_matrix[0][1][2] = MULL(contrast_times_saturation, sin_hue);

	result_matrix[0][2][0] = 0;
	result_matrix[0][2][1] = -1 * result_matrix[0][1][2];
	result_matrix[0][2][2] = result_matrix[0][1][1];

	/* Scale the HSBC matrix, if required */
	if (inverted_scale_factor != FIX(1)) {
		result_matrix[0][0][0] = MULL(result_matrix[0][0][0],
										inverted_scale_factor);
		result_matrix[0][1][1] = MULL(result_matrix[0][1][1],
										inverted_scale_factor);
		result_matrix[0][1][2] = MULL(result_matrix[0][1][2],
										inverted_scale_factor);
		result_matrix[0][2][1] = MULL(result_matrix[0][2][1],
										inverted_scale_factor);
		result_matrix[0][2][2] = MULL(result_matrix[0][2][2],
										inverted_scale_factor);
	}

	/* Multiply the HSBC matrix by the RGB to YUV matrix */
	pdpfb_fix_procamp_matrix_mul((const int (*)[3])result_matrix[0],
								rgb_to_yuv, result_matrix[1]);

	/* Multiply the result by the YUV to RGB matrix to get final coefs */
	pdpfb_fix_procamp_matrix_mul(yuv_to_rgb,
								(const int (*)[3])result_matrix[1],
								result_matrix[0]);

	#define HSBC_TO_COEFF(hsbc) \
		((hsbc) >> (PDP_PROCAMP_HSBC_FRACTIONAL_BITS \
			- PDP_PROCAMP_COEFFICIENT_FRACTIONAL_BITS))
	coefs->c11 = HSBC_TO_COEFF(result_matrix[0][0][0]);
	coefs->c12 = HSBC_TO_COEFF(result_matrix[0][0][1]);
	coefs->c13 = HSBC_TO_COEFF(result_matrix[0][0][2]);
	coefs->c21 = HSBC_TO_COEFF(result_matrix[0][1][0]);
	coefs->c22 = HSBC_TO_COEFF(result_matrix[0][1][1]);
	coefs->c23 = HSBC_TO_COEFF(result_matrix[0][1][2]);
	coefs->c31 = HSBC_TO_COEFF(result_matrix[0][2][0]);
	coefs->c32 = HSBC_TO_COEFF(result_matrix[0][2][1]);
	coefs->c33 = HSBC_TO_COEFF(result_matrix[0][2][2]);
	#undef HSBC_TO_COEFF

	input_offsets->r = input_offsets->g = input_offsets->b = 0;
	
	/*
	 * Add the brightness to the output offset, shifted for the correct number
	 * of fractional bits and to convert from 8 to 10 integer bits
	 */
	output_offset += hsbc->brightness >> (PDP_PROCAMP_HSBC_FRACTIONAL_BITS -
						PDP_PROCAMP_OUTPUT_OFFSET_FRACTIONAL_BITS - 2);
	output_offsets->r = output_offsets->g = output_offsets->b = output_offset;
}
#undef DIVL
#undef MULL
#undef SIN
#undef COS
#undef FIX

#ifndef __PDPFB_USE_OSAL__
static void pdpfb_debugfs_create_file(struct pdpfb_priv *priv,
									const char *name, u32 *var)
{
	struct dentry *dentry_p;
	dentry_p = debugfs_create_u32(name, S_IRUGO, priv->stats_dir, var);
	if (!dentry_p || IS_ERR(dentry_p))
		dev_warn(priv->device, "could not create debugfs file %s\n", name);
}
#endif

static void pdpfb_debugfs_setup(struct pdpfb_priv *priv)
{
#ifndef __PDPFB_USE_OSAL__
	const char *name = dev_name(priv->device);
	priv->stats_dir = debugfs_create_dir(name, NULL);
	if (!priv->stats_dir || IS_ERR(priv->stats_dir))
		dev_warn(priv->device, "could not create debugfs dir %s\n", name);
	else {
		#define CREATE_FILE(priv, name, var) \
			pdpfb_debugfs_create_file((priv), (name), &(var))
		CREATE_FILE(priv, "vsync_count", priv->vsync_count);
		CREATE_FILE(priv, "measured_fps", priv->measured_fps);
		CREATE_FILE(priv, "graphics1_underrun_count", priv->underrun_count[0]);
		CREATE_FILE(priv, "graphics1_overrun_count", priv->overrun_count[0]);
		CREATE_FILE(priv, "graphics2_underrun_count", priv->underrun_count[1]);
		CREATE_FILE(priv, "graphics2_overrun_count", priv->overrun_count[1]);
		CREATE_FILE(priv, "graphics3_underrun_count", priv->underrun_count[2]);
		CREATE_FILE(priv, "graphics3_overrun_count", priv->overrun_count[2]);
		CREATE_FILE(priv, "graphics4_underrun_count", priv->underrun_count[3]);
		CREATE_FILE(priv, "graphics4_overrun_count", priv->overrun_count[3]);
		CREATE_FILE(priv, "video1_underrun_count", priv->underrun_count[4]);
		CREATE_FILE(priv, "video1_overrun_count", priv->overrun_count[4]);
		CREATE_FILE(priv, "video2_underrun_count", priv->underrun_count[5]);
		CREATE_FILE(priv, "video2_overrun_count", priv->overrun_count[5]);
		CREATE_FILE(priv, "video3_underrun_count", priv->underrun_count[6]);
		CREATE_FILE(priv, "video3_overrun_count", priv->overrun_count[6]);
		CREATE_FILE(priv, "video4_underrun_count", priv->underrun_count[7]);
		CREATE_FILE(priv, "video4_overrun_count", priv->overrun_count[7]);
		#undef CREATE_FILE
	}
#endif
}

static void pdpfb_ident(struct pdpfb_priv *priv)
{
	unsigned int coreid;
	unsigned int corerev;

	coreid = pdpfb_read(priv, PDP_CORE_ID_OFFSET);
	corerev = pdpfb_read(priv, PDP_CORE_REV_OFFSET);

	dev_info(priv->device,
			"PDP id: %#x revision: %d.%d.%d\n", coreid,
			(corerev >> 16) & 0xff,
			(corerev >> 8) & 0xff,
			corerev & 0xff);
}

/* Handler for all PDP interrupts */
PDPFB_STATIC irqreturn_t pdpfb_interrupt(int irq, void *dev_id)
{
	struct pdpfb_priv *priv = dev_id;
	u32 int_stat;
	int i;
	struct pdpfb_isr_data *isr_data;
	struct pdpfb_isr_data registered_isr[PDPFB_MAX_NR_ISRS];

	int_stat = pdpfb_read(priv, PDP_INTSTAT_OFFSET);
	pdpfb_write(priv, PDP_INTCLR_OFFSET, int_stat);

#ifndef __PDPFB_USE_OSAL__
	spin_lock(&priv->irq_lock);
#endif

	if (priv->isr_disable) {
#ifndef __PDPFB_USE_OSAL__
		spin_unlock(&priv->irq_lock);
#endif
		return IRQ_HANDLED;
	}

	/* make a copy and unlock, so that isrs can unregister themselves */
	memcpy(registered_isr, priv->registered_isr, sizeof(registered_isr));

#ifndef __PDPFB_USE_OSAL__
	spin_unlock(&priv->irq_lock);
#endif

	for (i = 0; i < PDPFB_MAX_NR_ISRS; i++) {
		isr_data = &registered_isr[i];

		if (!isr_data->isr)
			continue;

		if (isr_data->mask & int_stat)
			isr_data->isr(isr_data->arg, int_stat);
	}

	return IRQ_HANDLED;
}

/* priv->irq_lock has to be locked by the caller */
static void _pdpfb_set_irqs(struct pdpfb_priv *priv)
{
	u32 mask;
	u32 old_mask;
	int i;
	struct pdpfb_isr_data *isr_data;

	if (priv->isr_disable) {
		pdpfb_write(priv, PDP_INTENAB_OFFSET, 0);
		return;
	}

	mask = 0;
	for (i = 0; i < PDPFB_MAX_NR_ISRS; i++) {
		isr_data = &priv->registered_isr[i];

		if (isr_data->isr == NULL)
			continue;

		mask |= isr_data->mask;
	}

	if (priv->pdp_clk != NULL)
		clk_enable(priv->pdp_clk);

	old_mask = pdpfb_read(priv, PDP_INTENAB_OFFSET);
	/* clear the irqstatus for newly enabled irqs */
	pdpfb_write(priv, PDP_INTCLR_OFFSET, (mask ^ old_mask) & mask);

	pdpfb_write(priv, PDP_INTENAB_OFFSET, mask);

	if (priv->pdp_clk != NULL)
		clk_disable(priv->pdp_clk);
}

static void pdpfb_disable_isr(struct pdpfb_priv *priv)
{
	unsigned long flags;

	spin_lock_irqsave(&priv->irq_lock, flags);
	priv->isr_disable = 1;
	_pdpfb_set_irqs(priv);
	spin_unlock_irqrestore(&priv->irq_lock, flags);
}

static void pdpfb_enable_isr(struct pdpfb_priv *priv)
{
	unsigned long flags;

	spin_lock_irqsave(&priv->irq_lock, flags);
	priv->isr_disable = 0;
	_pdpfb_set_irqs(priv);
	spin_unlock_irqrestore(&priv->irq_lock, flags);
}

static int pdpfb_register_isr(struct pdpfb_priv *priv, pdpfb_isr_t isr,
							void *arg, u32 mask)
{
	int i;
	int ret;
	unsigned long flags;
	struct pdpfb_isr_data *isr_data;

	if (!priv)
		return -ENODEV;

	if (isr == NULL)
		return -EINVAL;

	spin_lock_irqsave(&priv->irq_lock, flags);

	/* check for duplicate entry */
	for (i = 0; i < PDPFB_MAX_NR_ISRS; i++) {
		isr_data = &priv->registered_isr[i];
		if (isr_data->isr == isr && isr_data->arg == arg &&
				isr_data->mask == mask) {
			ret = -EINVAL;
			goto err;
		}
	}

	isr_data = NULL;
	ret = -EBUSY;

	for (i = 0; i < PDPFB_MAX_NR_ISRS; i++) {
		isr_data = &priv->registered_isr[i];

		if (isr_data->isr != NULL)
			continue;

		isr_data->isr = isr;
		isr_data->arg = arg;
		isr_data->mask = mask;
		ret = 0;

		break;
	}

	_pdpfb_set_irqs(priv);

	spin_unlock_irqrestore(&priv->irq_lock, flags);

	return 0;
err:
	spin_unlock_irqrestore(&priv->irq_lock, flags);

	return ret;
}

static int pdpfb_unregister_isr(struct pdpfb_priv *priv, pdpfb_isr_t isr,
								void *arg, u32 mask)
{
	int i;
	unsigned long flags;
	int ret = -EINVAL;
	struct pdpfb_isr_data *isr_data;

	if (!priv)
		return -ENODEV;

	spin_lock_irqsave(&priv->irq_lock, flags);

	for (i = 0; i < PDPFB_MAX_NR_ISRS; i++) {
		isr_data = &priv->registered_isr[i];
		if (isr_data->isr != isr || isr_data->arg != arg ||
				isr_data->mask != mask)
			continue;

		/* found the correct isr */

		isr_data->isr = NULL;
		isr_data->arg = NULL;
		isr_data->mask = 0;

		ret = 0;
		break;
	}

	if (ret == 0)
		_pdpfb_set_irqs(priv);

	spin_unlock_irqrestore(&priv->irq_lock, flags);

	return ret;
}

static void pdpfb_irq_wait_handler(void *data, u32 mask)
{
	complete((struct completion *)data);
}

#if 0
static int pdpfb_wait_for_irq_timeout(struct pdpfb_priv *priv, u32 irqmask,
									unsigned long timeout)
{
	int r;
	DECLARE_COMPLETION_ONSTACK(completion);

	r = pdpfb_register_isr(priv, pdpfb_irq_wait_handler, &completion, irqmask);

	if (r)
		return r;

	timeout = wait_for_completion_timeout(&completion, timeout);

	pdpfb_unregister_isr(priv, pdpfb_irq_wait_handler, &completion, irqmask);

	if (timeout == 0)
		return -ETIMEDOUT;

	if (timeout == -ERESTARTSYS)
		return -ERESTARTSYS;

	return 0;
}
#endif

static int pdpfb_wait_for_irq_interruptible_timeout(struct pdpfb_priv *priv,
													u32 irqmask,
													unsigned long timeout)
{
	int r;
	DECLARE_COMPLETION_ONSTACK(completion);

	r = pdpfb_register_isr(priv, pdpfb_irq_wait_handler, &completion, irqmask);

	if (r)
		return r;

#if 0 /* This times out on the virtual platform */
	timeout = wait_for_completion_interruptible_timeout(&completion, timeout);

	pdpfb_unregister_isr(priv, pdpfb_irq_wait_handler, &completion, irqmask);

	if (timeout == 0)
		return -ETIMEDOUT;
#else
	timeout = wait_for_completion_interruptible(&completion);

	pdpfb_unregister_isr(priv, pdpfb_irq_wait_handler, &completion, irqmask);
#endif

	if (timeout == -ERESTARTSYS)
		return -ERESTARTSYS;

	return 0;
}

static int pdpfb_wait_vsync(struct pdpfb_priv *priv)
{
	unsigned long timeout = msecs_to_jiffies(500);
	u32 irq = PDPFB_IRQ_VEVENT0 | PDPFB_IRQ_VEVENT1;
	int ret;
	int linenum;

	if (priv->is_sim) {
		/* Write the magic value to INTCLR to trigger a render */
		pdpfb_write(priv, PDP_INTCLR_OFFSET, 0x11111111);
	}

	ret = pdpfb_wait_for_irq_interruptible_timeout(priv, irq, timeout);

	if (priv->is_sim) {
		/*
		   Poll for end of blanking (linenum = 0).
		   This is needed to avoid kicking again before end of blanking when
		   the kick would be ignored.
		*/
		do {
			linenum = GET_FIELD(pdpfb_read(priv, PDP_LINESTAT_OFFSET),
								PDP_LINESTAT_LINENO);
		} while (linenum != 0);
	}

	return ret;
}

#ifdef CONFIG_FB_PDP2_MMU
static void pdpfb_str_mmu_swap_dirs(struct pdpfb_priv *priv,
									struct pdpfb_stream *stream)
{
	int i;
	int j;
	u32 reg;
	struct device *dev = priv->device;
	const int old_dir = !stream->next_dir; /* The old directory */

	for (i = 0; i < PDPFB_PTRS_PER_PDIR; i++) {
		/* Unmap and release any user pages from the old directory */
		for (j = 0; j < PDPFB_PTRS_PER_PTBL; j++) {
			if (stream->pdir_user_page[old_dir][i][j] != NULL) {
				/* & ~0x1 masks off the valid bit */
				const u32 page_addr_phys =
					stream->pdir_kvirt[old_dir][i][j] & ~0x1;
				dma_unmap_page(dev, page_addr_phys, PDP_PAGE_SIZE,
								DMA_TO_DEVICE);
				put_page(stream->pdir_user_page[old_dir][i][j]);
				stream->pdir_user_page[old_dir][i][j] = NULL;
			} else {
				break;
			}
		}
		/* Unmap and free any page tables from the old directory */
		if (stream->pdir_kvirt[old_dir][i] != NULL) {
			dma_unmap_single(dev, stream->pdir_bus[old_dir][i],
								PDP_PAGE_SIZE, DMA_TO_DEVICE);
			kfree(stream->pdir_kvirt[old_dir][i]);
			stream->pdir_bus[old_dir][i] = 0;
			stream->pdir_kvirt[old_dir][i] = NULL;
			/* Invalidate the physical directory entry */
			(*stream->pdir)[old_dir][i] = 0;

#ifdef CONFIG_FB_PDP2_PDUMP
			PDUMP_LOG("FREE :MEM:BLOCK_%d\n",
						stream->pdump_pdir_mem_block[old_dir][i]);
			/* Invalidate the physical directory entry */
			PDUMP_LOG("WRW :MEM:BLOCK_%d:0x%X 0x0\n",
						priv->pdump_pdir_mem_block,
						(stream->pdir_offset[old_dir] + i) * 4);
#endif
		}
	}

	/* Swap the current and next directories */
	stream->next_dir = !stream->next_dir;

	/* Invalidate the cache */
	reg = pdpfb_mmu_read(priv, PDP_BIF_CONTROL1_OFFSET);
	SET_FIELD(reg, PDP_BIF_CONTROL1_MMU_INVALDC0, 1);
	pdpfb_mmu_write(priv, PDP_BIF_CONTROL1_OFFSET, reg );
	SET_FIELD(reg, PDP_BIF_CONTROL1_MMU_INVALDC0, 0);
	pdpfb_mmu_write(priv, PDP_BIF_CONTROL1_OFFSET, reg );
}
#endif

static void pdpfb_vsync_isr(void *arg, u32 mask)
{
	struct pdpfb_priv *priv = (struct pdpfb_priv *)arg;
	u32 reg;
	struct timeval current_time;
	int elapsed_usecs;
#ifdef CONFIG_FB_PDP2_MMU
	int i;
#endif

	/* Calculate current frame rate */
	do_gettimeofday(&current_time);
	elapsed_usecs = current_time.tv_usec - priv->time_last_frame.tv_usec;
	if (current_time.tv_sec > priv->time_last_frame.tv_sec)
		elapsed_usecs +=
			(current_time.tv_sec - priv->time_last_frame.tv_sec) * 1000000;
	priv->measured_fps = 1000000 / elapsed_usecs;
	priv->time_last_frame = current_time;

#ifdef CONFIG_FB_PDP2_MMU
	for (i = 0; i < priv->stream_nr; ++i) {
		struct pdpfb_stream *stream = &priv->streams[i];
		if (stream->next_dir_updated) {
			pdpfb_str_mmu_swap_dirs(priv, stream);
			stream->next_dir_updated = 0;
		}
	}
#endif

#ifdef CONFIG_FB_PDP2_PDUMP

#ifdef CONFIG_FB_PDP2_PDUMP_SINGLE_FRAME
	/* Don't log the first kick for single frame pdumps */
	if (priv->vsync_count == 0)
		priv->pdump_suspended = 1;
#endif /* CONFIG_FB_PDP2_PDUMP_SINGLE_FRAME */

	/*
	 * When pdumping, the kick / pols must be logged after the MMU directory is
	 * updated and the cache invalidated, and before registers_valid is cleared
	 */
	PDUMP_LOG(PDUMP_WR_FMT, PDP_INTCLR_OFFSET, 0x11111111, "Kick simulator");
	PDUMP_LOG("POL :REGMET_PDP:0x%08x 0x00000005 0x0007FFFF 0 5000000 190\n",
				PDP_INTSTAT_OFFSET);
	PDUMP_LOG("POL :REGMET_PDP:0x%08x 0x00000000 0xFFFFFFFF 0 5000000 490\n",
				PDP_LINESTAT_OFFSET);

#ifdef CONFIG_FB_PDP2_PDUMP_SINGLE_FRAME
	if (priv->vsync_count == 0) {
		priv->pdump_suspended = 0;
	} else if (priv->vsync_count == 1) {
		priv->pdump_suspended = 1;
	} else if (priv->vsync_count == 2) {
		/* Add signature polls for single frame pdumps */
		priv->pdump_suspended = 0;
		PDUMP_LOG("POL :REGMET_PDP:0x%08x 0x%08x 0xFFFFFFFF 0\n",
				PDP_SIGNAT_R_OFFSET, pdpfb_read(priv, PDP_SIGNAT_R_OFFSET));
		PDUMP_LOG("POL :REGMET_PDP:0x%08x 0x%08x 0xFFFFFFFF 0\n",
				PDP_SIGNAT_GB_OFFSET, pdpfb_read(priv, PDP_SIGNAT_GB_OFFSET));
		priv->pdump_suspended = 1;
	}
#endif /* CONFIG_FB_PDP2_PDUMP_SINGLE_FRAME */

#ifdef __PDPFB_USE_OSAL__
	fflush(priv->hOut2Txt);
#endif /* __PDPFB_USE_OSAL__ */

#endif /* CONFIG_FB_PDP2_PDUMP */

	reg = pdpfb_read(priv, PDP_REGISTER_UPDATE_CTRL_OFFSET);
	if (GET_FIELD(reg, PDP_REGISTER_UPDATE_CTRL_REGISTERS_VALID)) {
		/* Wait for regs to be updated before clearing registers valid flag */
		while (!GET_FIELD(pdpfb_read(priv, PDP_REGISTER_UPDATE_STATUS_OFFSET),
				PDP_REGISTER_UPDATE_STATUS_REGISTERS_UPDATED));
		SET_FIELD(reg, PDP_REGISTER_UPDATE_CTRL_REGISTERS_VALID, 0);
		pdpfb_write(priv, PDP_REGISTER_UPDATE_CTRL_OFFSET, reg);
	}

	++priv->vsync_count;

	if (mask & PDP_INTSTAT_INTS_GRPH1URUN_MASK)
		++priv->underrun_count[0];
	if (mask & PDP_INTSTAT_INTS_GRPH2URUN_MASK)
		++priv->underrun_count[1];
	if (mask & PDP_INTSTAT_INTS_GRPH3URUN_MASK)
		++priv->underrun_count[2];
	if (mask & PDP_INTSTAT_INTS_GRPH4URUN_MASK)
		++priv->underrun_count[3];
	if (mask & PDP_INTSTAT_INTS_VID1URUN_MASK)
		++priv->underrun_count[4];
	if (mask & PDP_INTSTAT_INTS_VID2URUN_MASK)
		++priv->underrun_count[5];
	if (mask & PDP_INTSTAT_INTS_VID3URUN_MASK)
		++priv->underrun_count[6];
	if (mask & PDP_INTSTAT_INTS_VID4URUN_MASK)
		++priv->underrun_count[7];

	if (mask & PDP_INTSTAT_INTS_GRPH1ORUN_MASK)
		++priv->overrun_count[0];
	if (mask & PDP_INTSTAT_INTS_GRPH2ORUN_MASK)
		++priv->overrun_count[1];
	if (mask & PDP_INTSTAT_INTS_GRPH3ORUN_MASK)
		++priv->overrun_count[2];
	if (mask & PDP_INTSTAT_INTS_GRPH4ORUN_MASK)
		++priv->overrun_count[3];
	if (mask & PDP_INTSTAT_INTS_VID1ORUN_MASK)
		++priv->overrun_count[4];
	if (mask & PDP_INTSTAT_INTS_VID2ORUN_MASK)
		++priv->overrun_count[5];
	if (mask & PDP_INTSTAT_INTS_VID3ORUN_MASK)
		++priv->overrun_count[6];
	if (mask & PDP_INTSTAT_INTS_VID4ORUN_MASK)
		++priv->overrun_count[7];
}

struct pdpfb_cfg *pdpfb_get_config(struct pdpfb_priv *priv)
{
	return &priv->cfg;
}

struct pdpfb_stream *pdpfb_get_stream(struct pdpfb_priv *priv,
									unsigned int stream_id)
{
	if (stream_id >= ARRAY_SIZE(priv->streams))
		return NULL;

	return &priv->streams[stream_id];
}

enum pdpfb_display_mode {
	PDP_FULLY_POWERED_DOWN,
	PDP_PARTIALLY_POWERED_DOWN,
	PDP_ENABLED
};

static void pdpfb_set_display_enabled(struct pdpfb_priv *priv,
									enum pdpfb_display_mode mode)
{
	u32 sync_ctrl;
	u32 op_mask;
	int powerdn = 0;

	if (mode == PDP_FULLY_POWERED_DOWN && priv->cfg.sync_cfg.force_vsyncs)
		mode = PDP_PARTIALLY_POWERED_DOWN;

	switch (mode) {
	case PDP_FULLY_POWERED_DOWN:
		sync_ctrl = pdpfb_read(priv, PDP_SYNCCTRL_OFFSET);
		SET_FIELD(sync_ctrl, PDP_SYNCCTRL_SYNCACTIVE, 0);
		SET_FIELD(sync_ctrl, PDP_SYNCCTRL_POWERDN, 1);
		pdpfb_write(priv, PDP_SYNCCTRL_OFFSET, sync_ctrl);

		pdpfb_disable_isr(priv);
		pdpfb_pixel_clk_disable(priv);

		break;
	case PDP_PARTIALLY_POWERED_DOWN:
		powerdn = 1;
	case PDP_ENABLED:
		op_mask = pdpfb_read(priv, PDP_OPMASK_R_OFFSET);
		SET_FIELD(op_mask, PDP_OPMASK_R_MASKR, 0x000);
		SET_FIELD(op_mask, PDP_OPMASK_R_MASKLEVEL, 0);
		pdpfb_write(priv, PDP_OPMASK_R_OFFSET, op_mask);

		op_mask = pdpfb_read(priv, PDP_OPMASK_GB_OFFSET);
		SET_FIELD(op_mask, PDP_OPMASK_GB_MASKG, 0x000);
		SET_FIELD(op_mask, PDP_OPMASK_GB_MASKB, 0x000);
		pdpfb_write(priv, PDP_OPMASK_GB_OFFSET, op_mask);

		sync_ctrl = pdpfb_read(priv, PDP_SYNCCTRL_OFFSET);
		SET_FIELD(sync_ctrl, PDP_SYNCCTRL_UPDCTRL, 0);
		SET_FIELD(sync_ctrl, PDP_SYNCCTRL_SYNCACTIVE, 1);
		SET_FIELD(sync_ctrl, PDP_SYNCCTRL_POWERDN, powerdn);
		pdpfb_write(priv, PDP_SYNCCTRL_OFFSET, sync_ctrl);

		pdpfb_pixel_clk_enable(priv);
		pdpfb_enable_isr(priv);

		break;
	default:
		printk(KERN_ERR
				"PDP: Unrecognised pdpfb_set_display_enabled parameter\n");
	}
}

static void pdpfb_update_margins(struct pdpfb_priv *priv,
								struct pdpfb_stream *stream)
{
	struct fb_info *info = &stream->info;
	unsigned long pix_freq = (priv->pixel_clk != NULL) ?
								clk_get_rate(priv->pixel_clk) : 0;
	u32 pix_clock;
	const int xres = info->var.xres;
	const int yres = info->var.yres;
	const int geomw = stream->geom.w ? stream->geom.w : xres;
	const int geomh = stream->geom.h ? stream->geom.h : yres;
	const int hpost_clip_en = stream->hs_en && stream->clip_geom.w;
	const int vpost_clip_en = stream->vs_en && stream->clip_geom.h;
	const int w = hpost_clip_en ? stream->clip_geom.w : geomw;
	const int h = vpost_clip_en ? stream->clip_geom.h : geomh;
	int xgap[2];
	int ygap[2];

	pix_clock = KHZ2PICOS(pix_freq / 1000);
	xgap[0] = stream->geom.x;
	ygap[0] = stream->geom.y;
	/* we can't really handle the case of a plane positioned offscreen */
	xgap[1] = max(0, (int)priv->mode.xres - w - xgap[0]);
	ygap[1] = max(0, (int)priv->mode.yres - h - ygap[0]);

	/* physical size of plane */
	info->var.width = priv->cfg.lcd_size.width * w / priv->mode.xres;
	info->var.height = priv->cfg.lcd_size.height * h / priv->mode.yres;

	/* measurements in units of plane pixels */
	info->var.pixclock = pix_clock * w / xres * h / yres;
	info->var.hsync_len = priv->mode.hsync_len * xres / w;
	info->var.vsync_len = priv->mode.vsync_len * yres / h;
	info->var.left_margin  = (priv->mode.left_margin  + xgap[0]) * xres / w;
	info->var.upper_margin = (priv->mode.upper_margin + ygap[0]) * yres / h;
	info->var.right_margin = (priv->mode.right_margin + xgap[1]) * xres / w;
	info->var.lower_margin = (priv->mode.lower_margin + ygap[1]) * yres / h;
}

/* update variables but not registers */
static void pdpfb_update_refresh_rate(struct pdpfb_priv *priv)
{
	const u32 ht = priv->mode.hsync_len + priv->mode.left_margin
					+ priv->mode.xres + priv->mode.right_margin;
	const u32 vt = priv->mode.vsync_len + priv->mode.upper_margin
					+ priv->mode.yres + priv->mode.lower_margin;
	u32 pix_freq;
	int i;

	if (priv->mode.pixclock)
		pix_freq = 1000 * PICOS2KHZ(priv->mode.pixclock);
	else
		pix_freq = ht * vt * priv->mode.refresh;

	if (priv->pixel_clk != NULL) {
		clk_set_rate(priv->pixel_clk, pix_freq);
		clk_prepare_enable(priv->pixel_clk);
		pix_freq = clk_get_rate(priv->pixel_clk);
		clk_disable_unprepare(priv->pixel_clk);
	}
	else {
		pix_freq = 0;
	}
	priv->mode.pixclock = KHZ2PICOS(pix_freq / 1000);

	for (i = 0; i < priv->stream_nr; ++i) {
		struct pdpfb_stream *stream = &priv->streams[i];
		if (!stream || !stream->probed)
			continue;
		pdpfb_update_margins(priv, stream);
	}
}

/*
 * Updates pixel clock if it's changed (pixclock is pS).
 * Returns 1 if changed.
 */
int pdpfb_update_pixclock(struct pdpfb_priv *priv, unsigned int pixclock)
{
	if (pixclock != priv->mode.pixclock) {
		priv->mode.pixclock = pixclock;
		pdpfb_update_refresh_rate(priv);
		return 1;
	}
	return 0;
}

static unsigned long pdpfb_get_line_len(int xres_virtual,
										int pixels_per_block,
										int bits_per_block,
										int linelen_align)
{
	const int linelen_align_in_bits = 8 << linelen_align;
	u_long length;

	/* round up to line length alignment */
	length = (xres_virtual * bits_per_block) / pixels_per_block;
	length = (length + (linelen_align_in_bits - 1)) & -linelen_align_in_bits;
	length >>= 3;
	return length;
}

static void pdpfb_soft_reset(struct pdpfb_priv *priv)
{
	int i;
	u32 sync_ctrl;
	u32 ctrl;

	/* disable all the streams */
	for (i = 0; i < priv->stream_nr; ++i) {
		struct pdpfb_stream *stream = &priv->streams[i];
		if (stream) {
			ctrl = pdpfb_read(priv, stream->regs.ctrl);
			SET_FIELD(ctrl, PDP_GRPH1CTRL_GRPH1STREN, 0);
			pdpfb_write(priv, stream->regs.ctrl, ctrl);
		}
	}

	sync_ctrl = pdpfb_read(priv, PDP_SYNCCTRL_OFFSET);

	/* wait for stream disable to take effect after two vsyncs */
	if (GET_FIELD(sync_ctrl, PDP_SYNCCTRL_SYNCACTIVE)) {
		pdpfb_wait_vsync(priv);
		pdpfb_wait_vsync(priv);
	}

	SET_FIELD(sync_ctrl, PDP_SYNCCTRL_PDP_RST, 1);
	pdpfb_write(priv, PDP_SYNCCTRL_OFFSET, sync_ctrl);
	SET_FIELD(sync_ctrl, PDP_SYNCCTRL_PDP_RST, 0);
	pdpfb_write(priv, PDP_SYNCCTRL_OFFSET, sync_ctrl);
}

static void pdpfb_configure_dithering(struct pdpfb_priv *priv)
{
	pdpfb_write(priv, PDP_DITHERMATRIX2_OFFSET, 0x0000006C);
	pdpfb_write(priv, PDP_DITHERMATRIX4_0_OFFSET, 0x7B48F3C0);
	pdpfb_write(priv, PDP_DITHERMATRIX4_1_OFFSET, 0x596AC1E2);
	pdpfb_write(priv, PDP_DITHERMATRIX8_0_OFFSET, 0x03F0CC00);
	pdpfb_write(priv, PDP_DITHERMATRIX8_1_OFFSET, 0x1083F3F3);
	pdpfb_write(priv, PDP_DITHERMATRIX8_2_OFFSET, 0x2F4E372C);
	pdpfb_write(priv, PDP_DITHERMATRIX8_3_OFFSET, 0x3413821F);
	pdpfb_write(priv, PDP_DITHERMATRIX8_4_OFFSET, 0x28DC7ECB);
	pdpfb_write(priv, PDP_DITHERMATRIX8_5_OFFSET, 0x1BAD4918);
	pdpfb_write(priv, PDP_DITHERMATRIX8_6_OFFSET, 0x0EC825E7);
	pdpfb_write(priv, PDP_DITHERMATRIX8_7_OFFSET, 0x3D37107E);
	pdpfb_write(priv, PDP_DITHERMATRIX8_8_OFFSET, 0x217AE4A2);
	pdpfb_write(priv, PDP_DITHERMATRIX8_9_OFFSET, 0x3A29DB51);
	pdpfb_write(priv, PDP_DITHERMATRIX8_10_OFFSET, 0x05E49D86);
	pdpfb_write(priv, PDP_DITHERMATRIX8_11_OFFSET, 0x1699AAB5);
	pdpfb_write(priv, PDP_DITHERMATRIX8_12_OFFSET, 0x00565669);
}

static void pdpfb_configure_display(struct pdpfb_priv *priv)
{
	u32 sync_ctrl;

	sync_ctrl = pdpfb_read(priv, PDP_SYNCCTRL_OFFSET);
	SET_FIELD(sync_ctrl, PDP_SYNCCTRL_CLKPOL,
				priv->cfg.sync_cfg.clock_inverted);
	SET_FIELD(sync_ctrl, PDP_SYNCCTRL_VSPOL,
				0 == (priv->mode.sync & FB_SYNC_VERT_HIGH_ACT));
	SET_FIELD(sync_ctrl, PDP_SYNCCTRL_HSPOL,
				0 == (priv->mode.sync & FB_SYNC_HOR_HIGH_ACT));
	SET_FIELD(sync_ctrl, PDP_SYNCCTRL_BLNKPOL,
				priv->cfg.sync_cfg.de_active_high);
	SET_FIELD(sync_ctrl, PDP_SYNCCTRL_VSDIS, priv->cfg.sync_cfg.vsync_dis);
	SET_FIELD(sync_ctrl, PDP_SYNCCTRL_HSDIS, priv->cfg.sync_cfg.hsync_dis);
	SET_FIELD(sync_ctrl, PDP_SYNCCTRL_BLNKDIS, priv->cfg.sync_cfg.blank_dis);
	SET_FIELD(sync_ctrl, PDP_SYNCCTRL_FIELD_EN,
				(priv->mode.vmode & FB_VMODE_INTERLACED) ? 1 : 0);
	pdpfb_write(priv, PDP_SYNCCTRL_OFFSET, sync_ctrl);
}

static void pdpfb_configure_streams(struct pdpfb_priv *priv)
{
	static const unsigned int porter_blnd_regs[] = {
		PDP_PORTER_BLND0_OFFSET,
		PDP_PORTER_BLND1_OFFSET,
		PDP_PORTER_BLND2_OFFSET,
		PDP_PORTER_BLND3_OFFSET,
		PDP_PORTER_BLND4_OFFSET,
		PDP_PORTER_BLND5_OFFSET,
		PDP_PORTER_BLND6_OFFSET,
		PDP_PORTER_BLND7_OFFSET
	};
	u32 reg;
	int i;

	reg = pdpfb_read(priv, PDP_BGNDCOL_AR_OFFSET);
	SET_FIELD(reg, PDP_BGNDCOL_AR_BGNDCOL_A, priv->bgnd_col.alpha);
	SET_FIELD(reg, PDP_BGNDCOL_AR_BGNDCOL_R, (priv->bgnd_col.rgb >> 20));
	pdpfb_write(priv, PDP_BGNDCOL_AR_OFFSET, reg);

	reg = pdpfb_read(priv, PDP_BGNDCOL_GB_OFFSET);
	SET_FIELD(reg, PDP_BGNDCOL_GB_BGNDCOL_G,
				((priv->bgnd_col.rgb >> 10) & 0x3FF));
	SET_FIELD(reg, PDP_BGNDCOL_GB_BGNDCOL_B, (priv->bgnd_col.rgb & 0x3FF));
	pdpfb_write(priv, PDP_BGNDCOL_GB_OFFSET, reg);

	for (i = 0; i < priv->stream_nr; ++i) {
		struct pdpfb_stream *stream = &priv->streams[i];
		if (!stream || !stream->probed)
			continue;
		if (stream->enable) {
			pdpfb_str_configure(priv, stream);
			pdpfb_str_configure_addr(priv, stream);

			/* stream position on screen */
			reg = pdpfb_read(priv, stream->regs.posn);
			SET_FIELD(reg, PDP_GRPH1POSN_GRPH1XSTART, stream->geom.x);
			SET_FIELD(reg, PDP_GRPH1POSN_GRPH1YSTART,
				stream->geom.y / SCALER_MODIFIER(priv,stream).dst_height_div);
			pdpfb_write(priv, stream->regs.posn, reg);
		}

		/* set up blending */
		reg = pdpfb_read(priv, stream->regs.ctrl);
		SET_FIELD(reg, PDP_GRPH1CTRL_GRPH1STREN, stream->enable);
		SET_FIELD(reg, PDP_GRPH1CTRL_GRPH1BLEND,
					stream->blend_config.blend_mode);
		SET_FIELD(reg, PDP_GRPH1CTRL_GRPH1BLENDPOS, stream->pos);
		pdpfb_write(priv, stream->regs.ctrl, reg);

		reg = pdpfb_read(priv, porter_blnd_regs[stream->pos]);
		SET_FIELD(reg, PDP_PORTER_BLND1_BLND1PORTERMODE,
					stream->blend_config.porter_mode);
		SET_FIELD(reg, PDP_PORTER_BLND1_BLND1BLENDTYPE,
					!stream->blend_config.blend_type);
		pdpfb_write(priv, porter_blnd_regs[stream->pos], reg);

		reg = pdpfb_read(priv, stream->regs.galpha);
		SET_FIELD(reg, PDP_GRPH1GALPHA_GRPH1GALPHA, stream->global_alpha);
		pdpfb_write(priv, stream->regs.galpha, reg);

		/* set up colour keying */
		reg = pdpfb_read(priv, stream->regs.ctrl);
		SET_FIELD(reg, PDP_GRPH1CTRL_GRPH1CKEYEN, stream->ckey_en);
		SET_FIELD(reg, PDP_GRPH1CTRL_GRPH1CKEYSRC, stream->ckey_src);
		pdpfb_write(priv, stream->regs.ctrl, reg);

		reg = pdpfb_read(priv, stream->regs.ckey_r);
		SET_FIELD(reg, PDP_GRPH1CKEY_R_GRPH1CKEY_R,
					((stream->ckey.ckey >> 20) & 0x3FF));
		pdpfb_write(priv, stream->regs.ckey_r, reg);

		reg = pdpfb_read(priv, stream->regs.ckey_gb);
		SET_FIELD(reg, PDP_GRPH1CKEY_GB_GRPH1CKEY_G,
					((stream->ckey.ckey >> 10) & 0x3FF));
		SET_FIELD(reg, PDP_GRPH1CKEY_GB_GRPH1CKEY_B,
					(stream->ckey.ckey & 0x3FF));
		pdpfb_write(priv, stream->regs.ckey_gb, reg);

		reg = pdpfb_read(priv, stream->regs.blend2_r);
		SET_FIELD(reg, PDP_GRPH1BLND2_R_GRPH1CKEYMASK_R,
					((stream->ckey.mask >> 20) & 0x3FF));
		pdpfb_write(priv, stream->regs.blend2_r, reg);

		reg = pdpfb_read(priv, stream->regs.blend2_gb);
		SET_FIELD(reg, PDP_GRPH1BLND2_GB_GRPH1CKEYMASK_G,
					((stream->ckey.mask >> 10) & 0x3FF));
		SET_FIELD(reg, PDP_GRPH1BLND2_GB_GRPH1CKEYMASK_B,
					(stream->ckey.mask & 0x3FF));
		pdpfb_write(priv, stream->regs.blend2_gb, reg);

		/* set interleaved control */
		reg = pdpfb_read(priv, stream->regs.interleave_ctrl);
		SET_FIELD(reg, PDP_GRPH1INTERLEAVE_CTRL_GRPH1INTFIELD,
					SCALER_MODIFIER(priv,stream).int_field);
		pdpfb_write(priv, stream->regs.interleave_ctrl, reg);
	}

	pdpfb_flush_regs(priv);
}

/* Set up sync registers to position the frame buffers on the screen */
static int pdpfb_configure_sync(struct pdpfb_priv *priv)
{
	const int border[2][2] = { {3, 3}, {3, 3} };

	const u16 hbps = priv->mode.hsync_len;
	const u16 hdes = hbps + priv->mode.left_margin;
	const u16 has = hdes;
	const u16 hlbs = has - border[0][0];
	const u16 hdef = hdes + priv->mode.xres;
	const u16 hrbs = hdef;
	const u16 hfps = hrbs + border[0][1];
	const u16 ht = hrbs + priv->mode.right_margin;

	const u16 vbps = priv->mode.vsync_len;
	const u16 vdes = vbps + priv->mode.upper_margin;
	const u16 vas = vdes;
	const u16 vtbs = vas - border[1][0];
	const u16 vdef = vdes + priv->mode.yres;
	const u16 vbbs = vdef;
	const u16 vfps = vbbs + border[1][1];
	const u16 vt = vbbs + priv->mode.lower_margin;

	const u16 vevent = 0;
	const u16 vfetch = vbps;

	/* write registers that shouldn't cause artifacts */

	pdpfb_write(priv, PDP_VSYNC1_OFFSET,
				PLACE_FIELD(PDP_VSYNC1_VBPS, vbps)
				| PLACE_FIELD(PDP_VSYNC1_VT, vt));

	pdpfb_write(priv, PDP_VSYNC2_OFFSET,
				PLACE_FIELD(PDP_VSYNC2_VAS, vas)
				| PLACE_FIELD(PDP_VSYNC2_VTBS, vtbs));

	pdpfb_write(priv, PDP_VSYNC3_OFFSET,
				PLACE_FIELD(PDP_VSYNC3_VFPS, vfps)
				| PLACE_FIELD(PDP_VSYNC3_VBBS, vbbs));

	pdpfb_write(priv, PDP_VEVENT_OFFSET,
				PLACE_FIELD(PDP_VEVENT_VEVENT, vevent)
				| PLACE_FIELD(PDP_VEVENT_VFETCH, vfetch));

	pdpfb_write(priv, PDP_VDECTRL_OFFSET,
				PLACE_FIELD(PDP_VDECTRL_VDES, vdes)
				| PLACE_FIELD(PDP_VDECTRL_VDEF, vdef));

	pdpfb_write(priv, PDP_HSYNC1_OFFSET,
				PLACE_FIELD(PDP_HSYNC1_HBPS, hbps)
				| PLACE_FIELD(PDP_HSYNC1_HT, ht));

	pdpfb_write(priv, PDP_HSYNC2_OFFSET,
				PLACE_FIELD(PDP_HSYNC2_HAS, has)
				| PLACE_FIELD(PDP_HSYNC2_HLBS, hlbs));

	pdpfb_write(priv, PDP_HSYNC3_OFFSET,
				PLACE_FIELD(PDP_HSYNC3_HFPS, hfps)
				| PLACE_FIELD(PDP_HSYNC3_HRBS, hrbs));

	pdpfb_write(priv, PDP_HDECTRL_OFFSET,
				PLACE_FIELD(PDP_HDECTRL_HDES, hdes)
				| PLACE_FIELD(PDP_HDECTRL_HDEF, hdef));

	return 0;
}

static int pdpfb_configure(struct pdpfb_priv *priv)
{
	pdpfb_soft_reset(priv);
	/* ensure pixel clock keeps running while powered down */
	pdpfb_set_display_enabled(priv, PDP_PARTIALLY_POWERED_DOWN);

	pdpfb_write(priv, PDP_INTCLR_OFFSET, 0xFFFFFFFF);
	pdpfb_configure_display(priv);
	pdpfb_update_refresh_rate(priv);

	pdpfb_configure_sync(priv);

	pdpfb_set_display_enabled(priv, PDP_ENABLED);

	pdpfb_configure_streams(priv);

	return 0;
}

static void pdpfb_set_mode(struct pdpfb_priv *priv,
						struct fb_var_screeninfo *var)
{
	if (var->activate & FB_ACTIVATE_FORCE) {
		fb_var_to_videomode(&priv->mode, var);
		pdpfb_configure(priv);
	} else {
		struct fb_videomode old_mode = priv->mode;
		fb_var_to_videomode(&priv->mode, var);
		if (!fb_mode_is_equal(&priv->mode, &old_mode))
			pdpfb_configure(priv);
		else
			pdpfb_configure_streams(priv);
	}
}

int pdpfb_blank(int blank, struct fb_info *info)
{
	struct pdpfb_priv *priv = dev_get_drvdata(info->device);

	switch (blank) {
	case FB_BLANK_UNBLANK:
		pdpfb_set_display_enabled(priv, PDP_ENABLED);
		break;
	case FB_BLANK_NORMAL:
		pdpfb_set_display_enabled(priv, PDP_PARTIALLY_POWERED_DOWN);
		break;
	case FB_BLANK_VSYNC_SUSPEND:
	case FB_BLANK_HSYNC_SUSPEND:
	case FB_BLANK_POWERDOWN:
		pdpfb_set_display_enabled(priv, PDP_FULLY_POWERED_DOWN);
		break;
	}
	return 0;
}

static void pdpfb_get_bgnd_col(struct pdpfb_priv *priv,
							struct pdpfb_bgnd_col *bgnd_col)
{
	*bgnd_col = priv->bgnd_col;
}

static void pdpfb_set_bgnd_col(struct pdpfb_priv *priv,
							struct pdpfb_bgnd_col *bgnd_col)
{
	bgnd_col->rgb &= 0x3FFFFFFF;
	bgnd_col->alpha &= 0x3FF;
	priv->bgnd_col = *bgnd_col;
	pdpfb_configure_streams(priv);
}

static void pdpfb_get_disp_mode(struct pdpfb_priv *priv,
								struct pdpfb_disp_mode *disp_mode)
{
	disp_mode->xres = priv->mode.xres;
	disp_mode->yres = priv->mode.yres
					* ((priv->mode.vmode & FB_VMODE_INTERLACED) ? 2 : 1);
	disp_mode->refresh = priv->mode.refresh;
}

static int pdpfb_set_disp_mode(struct pdpfb_priv *priv,
							struct pdpfb_disp_mode *disp_mode)
								
{
	struct fb_videomode vm;
	const struct fb_videomode *nearest_vm;
	struct fb_var_screeninfo var;

	vm.xres = disp_mode->xres;
	vm.yres = disp_mode->yres;
	vm.refresh = disp_mode->refresh;

	nearest_vm = fb_find_nearest_mode(&vm,
							&priv->streams[PDPFB_STREAM_GFX1].info.modelist);
	if (!nearest_vm)
		return -EINVAL;

	fb_videomode_to_var(&var, nearest_vm);
	pdpfb_set_mode(priv, &var);

	/* Update disp_mode based on the mode used */
	pdpfb_get_disp_mode(priv, disp_mode);

	return 0;
}

static void pdpfb_get_csc_config(struct pdpfb_priv *priv,
								struct pdpfb_csc_config *csc_config)
{
	*csc_config = priv->csc_config;
}

static int pdpfb_set_csc_config(struct pdpfb_priv *priv,
								struct pdpfb_csc_config *csc_config)
{
	if (csc_config->preset) {
		if (csc_config->preset < ARRAY_SIZE(pdpfb_csc_presets))
			csc_config->coefs = pdpfb_csc_presets[csc_config->preset];
		else
			return -EINVAL;
	} else {
		#define CSC_COEF_INVALID(x) (((x) >= (1 << 10)) || ((x) < (-1 << 10)))
		if (CSC_COEF_INVALID(csc_config->coefs.ry) ||
			CSC_COEF_INVALID(csc_config->coefs.rv) ||
			CSC_COEF_INVALID(csc_config->coefs.ru) ||
			CSC_COEF_INVALID(csc_config->coefs.gy) ||
			CSC_COEF_INVALID(csc_config->coefs.gv) ||
			CSC_COEF_INVALID(csc_config->coefs.gu) ||
			CSC_COEF_INVALID(csc_config->coefs.by) ||
			CSC_COEF_INVALID(csc_config->coefs.bv) ||
			CSC_COEF_INVALID(csc_config->coefs.bu)) {
			return -EINVAL;
		}
		#undef CSC_COEF_INVALID
	}

	priv->csc_config = *csc_config;

	pdpfb_write(priv, PDP_CSCCOEFF0_OFFSET,
				PLACE_FIELD(PDP_CSCCOEFF0_CSCCOEFFRU, csc_config->coefs.ru)
				| PLACE_FIELD(PDP_CSCCOEFF0_CSCCOEFFRY, csc_config->coefs.ry));
	pdpfb_write(priv, PDP_CSCCOEFF1_OFFSET,
				PLACE_FIELD(PDP_CSCCOEFF1_CSCCOEFFGY, csc_config->coefs.gy)
				| PLACE_FIELD(PDP_CSCCOEFF1_CSCCOEFFRV, csc_config->coefs.rv));
	pdpfb_write(priv, PDP_CSCCOEFF2_OFFSET,
				PLACE_FIELD(PDP_CSCCOEFF2_CSCCOEFFGV, csc_config->coefs.gv)
				| PLACE_FIELD(PDP_CSCCOEFF2_CSCCOEFFGU, csc_config->coefs.gu));
	pdpfb_write(priv, PDP_CSCCOEFF3_OFFSET,
				PLACE_FIELD(PDP_CSCCOEFF3_CSCCOEFFBU, csc_config->coefs.bu)
				| PLACE_FIELD(PDP_CSCCOEFF3_CSCCOEFFBY, csc_config->coefs.by));
	pdpfb_write(priv, PDP_CSCCOEFF4_OFFSET,
				PLACE_FIELD(PDP_CSCCOEFF4_CSCCOEFFBV, csc_config->coefs.bv));

	pdpfb_flush_regs(priv);

	return 0;
}

static void pdpfb_get_procamp(struct pdpfb_priv *priv,
							struct pdpfb_procamp *procamp)
{
	*procamp = priv->procamp;
}

#define PROCAMP_COEF_INVALID(x) \
	(((x) >= (1 << PDP_CR_PDP_PROCAMP_C11C12_CR_PROCAMP_C11_LENGTH)) \
	|| ((x) < (-1 << PDP_CR_PDP_PROCAMP_C11C12_CR_PROCAMP_C11_LENGTH)))

#define PROCAMP_INPUT_OFFSET_INVALID(x) \
	(((x) >= (1 << PDP_CR_PDP_PROCAMP_INOFFSET_R_CR_PROCAMP_INOFF_R_LENGTH)) \
	|| ((x) < (-1 << PDP_CR_PDP_PROCAMP_INOFFSET_R_CR_PROCAMP_INOFF_R_LENGTH)))

#define OUTOFF_LENGTH PDP_CR_PDP_PROCAMP_OUTOFFSET_R_CR_PROCAMP_OUTOFF_R_LENGTH
#define PROCAMP_OUTPUT_OFFSET_INVALID(x) \
	(((x) >= (1 << OUTOFF_LENGTH)) || ((x) < (-1 << OUTOFF_LENGTH)))

#define PROCAMP_HSBC_INVALID(hsbc, min, max) \
	(((hsbc) > ((max) << PDP_PROCAMP_HSBC_FRACTIONAL_BITS)) \
	|| ((hsbc) < ((min) << PDP_PROCAMP_HSBC_FRACTIONAL_BITS)))

static int pdpfb_set_procamp(struct pdpfb_priv *priv,
							struct pdpfb_procamp *procamp)
{
	u32 reg;

	if (procamp->enable) {
		if (procamp->output_range > PDP_PROCAMP_OUTPUT_RANGE_192_832)
			return -EINVAL;

		if (procamp->custom) {
			if (PROCAMP_COEF_INVALID(procamp->coefs.c11) ||
				PROCAMP_COEF_INVALID(procamp->coefs.c12) ||
				PROCAMP_COEF_INVALID(procamp->coefs.c13) ||
				PROCAMP_COEF_INVALID(procamp->coefs.c21) ||
				PROCAMP_COEF_INVALID(procamp->coefs.c22) ||
				PROCAMP_COEF_INVALID(procamp->coefs.c23) ||
				PROCAMP_COEF_INVALID(procamp->coefs.c31) ||
				PROCAMP_COEF_INVALID(procamp->coefs.c32) ||
				PROCAMP_COEF_INVALID(procamp->coefs.c33)) {
				return -EINVAL;
			}
			if (PROCAMP_INPUT_OFFSET_INVALID(procamp->input_offsets.r) ||
				PROCAMP_INPUT_OFFSET_INVALID(procamp->input_offsets.g) ||
				PROCAMP_INPUT_OFFSET_INVALID(procamp->input_offsets.b)) {
				return -EINVAL;
			}
			if (PROCAMP_OUTPUT_OFFSET_INVALID(procamp->output_offsets.r) ||
				PROCAMP_OUTPUT_OFFSET_INVALID(procamp->output_offsets.g) ||
				PROCAMP_OUTPUT_OFFSET_INVALID(procamp->output_offsets.b)) {
				return -EINVAL;
			}
			priv->procamp = *procamp;
		} else {
			if (PROCAMP_HSBC_INVALID(procamp->hsbc.hue, -30, 30) ||
				PROCAMP_HSBC_INVALID(procamp->hsbc.saturation, 0, 2) ||
				PROCAMP_HSBC_INVALID(procamp->hsbc.brightness, -50, 50) ||
				PROCAMP_HSBC_INVALID(procamp->hsbc.contrast, 0, 2)) {
				return -EINVAL;
			}
			/*
			 * Retain the procamp setup for use by pdpfb_get_procamp before the
			 * coefs are overwritten 
			 */
			priv->procamp = *procamp;
			pdpfb_calc_procamp_coefs(&procamp->hsbc,
									procamp->output_range,
									&procamp->coefs,
									&procamp->input_offsets,
									&procamp->output_offsets);
		}

		reg = PLACE_FIELD(PDP_CR_PDP_PROCAMP_C11C12_CR_PROCAMP_C11,
							procamp->coefs.c11);
		SET_FIELD(reg, PDP_CR_PDP_PROCAMP_C11C12_CR_PROCAMP_C12,
							procamp->coefs.c12);
		pdpfb_write(priv, PDP_CR_PDP_PROCAMP_C11C12_OFFSET, reg);

		reg = PLACE_FIELD(PDP_CR_PDP_PROCAMP_C13C21_CR_PROCAMP_C13,
							procamp->coefs.c13);
		SET_FIELD(reg, PDP_CR_PDP_PROCAMP_C13C21_CR_PROCAMP_C21,
							procamp->coefs.c21);
		pdpfb_write(priv, PDP_CR_PDP_PROCAMP_C13C21_OFFSET, reg);

		reg = PLACE_FIELD(PDP_CR_PDP_PROCAMP_C22C23_CR_PROCAMP_C22,
							procamp->coefs.c22);
		SET_FIELD(reg, PDP_CR_PDP_PROCAMP_C22C23_CR_PROCAMP_C23,
							procamp->coefs.c23);
		pdpfb_write(priv, PDP_CR_PDP_PROCAMP_C22C23_OFFSET, reg);

		reg = PLACE_FIELD(PDP_CR_PDP_PROCAMP_C31C32_CR_PROCAMP_C31,
							procamp->coefs.c31);
		SET_FIELD(reg, PDP_CR_PDP_PROCAMP_C31C32_CR_PROCAMP_C32,
							procamp->coefs.c32);
		pdpfb_write(priv, PDP_CR_PDP_PROCAMP_C31C32_OFFSET, reg);

		reg = PLACE_FIELD(PDP_CR_PDP_PROCAMP_C33_CR_PROCAMP_EN, 1);
		SET_FIELD(reg, PDP_CR_PDP_PROCAMP_C33_CR_PROCAMP_RANGE,
							procamp->output_range);
		SET_FIELD(reg, PDP_CR_PDP_PROCAMP_C33_CR_PROCAMP_C33,
							procamp->coefs.c33);
		pdpfb_write(priv, PDP_CR_PDP_PROCAMP_C33_OFFSET, reg);

		reg = PLACE_FIELD(PDP_CR_PDP_PROCAMP_INOFFSET_BG_CR_PROCAMP_INOFF_B,
							procamp->input_offsets.b);
		SET_FIELD(reg, PDP_CR_PDP_PROCAMP_INOFFSET_BG_CR_PROCAMP_INOFF_G,
							procamp->input_offsets.g);
		pdpfb_write(priv, PDP_CR_PDP_PROCAMP_INOFFSET_BG_OFFSET, reg);

		reg = PLACE_FIELD(PDP_CR_PDP_PROCAMP_INOFFSET_R_CR_PROCAMP_INOFF_R,
							procamp->input_offsets.r);
		pdpfb_write(priv, PDP_CR_PDP_PROCAMP_INOFFSET_R_OFFSET, reg);

		reg = PLACE_FIELD(PDP_CR_PDP_PROCAMP_OUTOFFSET_BG_CR_PROCAMP_OUTOFF_B,
							procamp->output_offsets.b);
		SET_FIELD(reg, PDP_CR_PDP_PROCAMP_OUTOFFSET_BG_CR_PROCAMP_OUTOFF_G,
							procamp->output_offsets.g);
		pdpfb_write(priv, PDP_CR_PDP_PROCAMP_OUTOFFSET_BG_OFFSET, reg);

		reg = PLACE_FIELD(PDP_CR_PDP_PROCAMP_OUTOFFSET_R_CR_PROCAMP_OUTOFF_R,
							procamp->output_offsets.r);
		pdpfb_write(priv, PDP_CR_PDP_PROCAMP_OUTOFFSET_R_OFFSET, reg);
	} else {
		pdpfb_write(priv, PDP_CR_PDP_PROCAMP_C33_OFFSET,
					PLACE_FIELD(PDP_CR_PDP_PROCAMP_C33_CR_PROCAMP_EN, 0));
	}

	pdpfb_flush_regs(priv);

	return 0;
}
#undef PROCAMP_COEF_INVALID
#undef PROCAMP_INPUT_OFFSET_INVALID
#undef OUTOFF_LENGTH
#undef PROCAMP_OUTPUT_OFFSET_INVALID
#undef PROCAMP_HSBC_INVALID

static void pdpfb_get_gamma_config(struct pdpfb_priv *priv,
								struct pdpfb_gamma_config *gamma_config)
{
	*gamma_config = priv->gamma_config;
}

static int pdpfb_set_gamma_config(struct pdpfb_priv *priv,
								struct pdpfb_gamma_config *gamma_config)
{
	int i;
	int gamma_entry;
	u32 reg;

	if (gamma_config->custom) {
		/* Check coefficients */
		for (i = 0; i < PDP_GAMMA_TABLE_ENTRIES; ++i) {
			if (gamma_config->coefs[i] > 0x3FFFFFFF)
				return -EINVAL;
		}
		priv->gamma_config = *gamma_config;
	} else {
		if ((gamma_config->input_transfer_function > PDP_GAMMA_2_8) ||
			(gamma_config->output_transfer_function > PDP_GAMMA_2_8)) {
				return -EINVAL;
			} else {
				const u32 input_func = gamma_config->input_transfer_function;
				const u32 output_func = gamma_config->output_transfer_function;
				/* 
				 * Retain the gamma config for use by pdpfb_get_gamma_config
				 * before the coefs are overwritten
				 */
				priv->gamma_config = *gamma_config;
				for (i = 0; i < PDP_GAMMA_TABLE_ENTRIES; ++i) {
					gamma_entry = gamme_table[input_func][output_func][i];
					gamma_config->coefs[i] = (gamma_entry << 20)
										+ (gamma_entry << 10) + gamma_entry;
				}
			}
	}

	for (i = 0; i < PDP_GAMMA_TABLE_ENTRIES; ++i) {
		reg = PLACE_FIELD(PDP_GAMMA0_R_GAMMA0_R,
							((gamma_config->coefs[i] >> 20) & 0x3FF));
		pdpfb_write(priv, PDP_GAMMA0_R_OFFSET + (i * 8), reg);
		reg = PLACE_FIELD(PDP_GAMMA0_GB_GAMMA0_G,
							((gamma_config->coefs[i] >> 10) & 0x3FF));
		SET_FIELD(reg, PDP_GAMMA0_GB_GAMMA0_B,
							(gamma_config->coefs[i] & 0x3FF));
		pdpfb_write(priv, PDP_GAMMA0_GB_OFFSET + (i * 8), reg);
	}

	pdpfb_flush_regs(priv);

	return 0;
}

static int pdpfb_get_3d_mode(struct pdpfb_priv *priv)
{
	if (!priv->alt3d_en) {
		return PDP_3D_MODE_DISABLE;
	}
	else {
		switch (priv->alt3d_blend_sel) {
		case PDP_3D_CTRL_BLENDSEL_BGND_WITH_POS0:
			return PDP_3D_MODE_BGND_WITH_POS0;
		case PDP_3D_CTRL_BLENDSEL_POS0_WITH_POS1:
			return PDP_3D_MODE_POS0_WITH_POS1;
		case PDP_3D_CTRL_BLENDSEL_POS1_WITH_POS2:
			return PDP_3D_MODE_POS1_WITH_POS2;
		case PDP_3D_CTRL_BLENDSEL_POS2_WITH_POS3:
			return PDP_3D_MODE_POS2_WITH_POS3;
		case PDP_3D_CTRL_BLENDSEL_POS3_WITH_POS4:
			return PDP_3D_MODE_POS3_WITH_POS4;
		default:
			return -EINVAL;
		}
	}
}

static int pdpfb_set_3d_mode(struct pdpfb_priv *priv, int val)
{
	u32 alt3d_ctrl;

	if (val == PDP_3D_MODE_DISABLE) {
		priv->alt3d_en = 0;
	}
	else {
		switch (val) {
		case PDP_3D_MODE_BGND_WITH_POS0:
			priv->alt3d_blend_sel = PDP_3D_CTRL_BLENDSEL_BGND_WITH_POS0;
			break;
		case PDP_3D_MODE_POS0_WITH_POS1:
			priv->alt3d_blend_sel = PDP_3D_CTRL_BLENDSEL_POS0_WITH_POS1;
			break;
		case PDP_3D_MODE_POS1_WITH_POS2:
			priv->alt3d_blend_sel = PDP_3D_CTRL_BLENDSEL_POS1_WITH_POS2;
			break;
		case PDP_3D_MODE_POS2_WITH_POS3:
			priv->alt3d_blend_sel = PDP_3D_CTRL_BLENDSEL_POS2_WITH_POS3;
			break;
		case PDP_3D_MODE_POS3_WITH_POS4:
			priv->alt3d_blend_sel = PDP_3D_CTRL_BLENDSEL_POS3_WITH_POS4;
			break;
		default:
			return -EINVAL;
		}
		priv->alt3d_en = 1;
	}

	alt3d_ctrl = pdpfb_read(priv, PDP_ALTERNATE_3D_CTRL_OFFSET);
	SET_FIELD(alt3d_ctrl, PDP_ALTERNATE_3D_CTRL_ALT3D_BLENDSEL,
				priv->alt3d_blend_sel);
	SET_FIELD(alt3d_ctrl, PDP_ALTERNATE_3D_CTRL_ALT3D_ON, priv->alt3d_en);
	pdpfb_write(priv, PDP_ALTERNATE_3D_CTRL_OFFSET, alt3d_ctrl);

	pdpfb_flush_regs(priv);

	return 0;
}

static int pdpfb_get_output_depth(struct pdpfb_priv *priv)
{
	int eight_bit_output;

	eight_bit_output = GET_FIELD(pdpfb_read(priv, PDP_OUTPUT_OFFSET),
									PDP_OUTPUT_EIGHT_BIT_OUTPUT);

	return eight_bit_output ? 8 : 10;

}

static int pdpfb_set_output_depth(struct pdpfb_priv *priv, int val)
{
	u32 reg;

	if (val != 8 && val != 10)
		return -EINVAL;

	reg = pdpfb_read(priv, PDP_OUTPUT_OFFSET);
	SET_FIELD(reg, PDP_OUTPUT_EIGHT_BIT_OUTPUT, val == 8);
	pdpfb_write(priv, PDP_OUTPUT_OFFSET, reg);

	pdpfb_flush_regs(priv);

	return 0;
}

/* Stream Functions */

static void pdpfb_str_get_vblank(struct pdpfb_priv *priv,
								 struct pdpfb_stream *stream,
								 struct fb_vblank *vblank)
{
	u32 line_no;
	u32 geomy;
	u32 geomh;

	line_no = GET_FIELD(pdpfb_read(priv, PDP_LINESTAT_OFFSET),
						PDP_LINESTAT_LINENO);

	geomy = priv->mode.vsync_len + priv->mode.upper_margin + stream->geom.y;
	geomh = stream->geom.h;
	if (!geomh)
		geomh = stream->info.var.yres;

	memset(vblank, 0, sizeof(*vblank));

	vblank->flags = FB_VBLANK_HAVE_VBLANK
					| FB_VBLANK_HAVE_COUNT
					| FB_VBLANK_HAVE_VCOUNT
					| FB_VBLANK_HAVE_VSYNC;
	if (line_no < geomy || line_no >= geomy + geomh)
		vblank->flags |= FB_VBLANK_VBLANKING;
	if (line_no < priv->mode.vsync_len)
		vblank->flags |= FB_VBLANK_VSYNCING;
	vblank->count = priv->vsync_count;
	vblank->vcount = line_no * stream->info.var.yres / geomh;
}

static int pdpfb_str_get_en(struct pdpfb_priv *priv,
							struct pdpfb_stream *stream)
{
	return stream->enable;
}

static void pdpfb_str_set_en(struct pdpfb_priv *priv,
							struct pdpfb_stream *stream, int val)
{
	stream->enable = (0 != val);
	pdpfb_configure_streams(priv);
}

static int pdpfb_str_get_plane_pos(struct pdpfb_priv *priv,
								struct pdpfb_stream *stream)
{
	return stream->pos;
}

/* returns new position, or negative on error */
static int pdpfb_str_set_plane_pos(struct pdpfb_priv *priv,
								struct pdpfb_stream *stream, int pos)
{
	int i;
	const int old_pos = stream->pos;
	if (unlikely(pos < 0 || pos >= priv->stream_nr))
		return -EINVAL;
	if (pos == old_pos)
		return pos;

	if (pos > old_pos) {
		for (i = 0; i < priv->stream_nr; i++) {
			if (priv->streams[i].pos > old_pos && priv->streams[i].pos <= pos)
				priv->streams[i].pos--;
		}
	} else if (pos < old_pos) {
		for (i = 0; i < priv->stream_nr; i++) {
			if (priv->streams[i].pos < old_pos && priv->streams[i].pos >= pos)
				priv->streams[i].pos++;
		}
	}

	stream->pos = pos;
	pdpfb_configure_streams(priv);
	return pos;
}

static void pdpfb_str_get_geom(struct pdpfb_priv *priv,
							struct pdpfb_stream *stream,
							struct pdpfb_geom *geom)
{
	*geom = stream->geom;
}

static int pdpfb_str_set_geom(struct pdpfb_priv *priv,
							struct pdpfb_stream *stream,
							struct pdpfb_geom *geom)
{
	int error;

	if (geom->x > MAX_FIELD_VALUE(PDP_GRPH1POSN_GRPH1XSTART) ||
		geom->y > MAX_FIELD_VALUE(PDP_GRPH1POSN_GRPH1XSTART))
		return -EINVAL;

	error = pdpfb_str_check_geom(priv, stream, geom);
	if (error)
		return -EINVAL;

	stream->geom = *geom;

	pdpfb_configure_streams(priv);
	pdpfb_update_margins(priv, stream);

	return 0;
}

static int pdpfb_str_get_ckey_mode(struct pdpfb_priv *priv,
								struct pdpfb_stream *stream)
{
	if (!stream->ckey_en)
		return PDP_CKEYMODE_DISABLE;
	else if (stream->ckey_src == PDP_CTRL_CKEYSRC_PREV)
		return PDP_CKEYMODE_PREVIOUS;
	else if (stream->ckey_src == PDP_CTRL_CKEYSRC_CUR)
		return PDP_CKEYMODE_CURRENT;
	else
		return -EINVAL;
}

static int pdpfb_str_set_ckey_mode(struct pdpfb_priv *priv,
								struct pdpfb_stream *stream, int val)
{
	switch (val) {
	case PDP_CKEYMODE_DISABLE:
		stream->ckey_en = 0;
		break;
	case PDP_CKEYMODE_PREVIOUS:
		stream->ckey_en = 1;
		stream->ckey_src = PDP_CTRL_CKEYSRC_PREV;
		break;
	case PDP_CKEYMODE_CURRENT:
		stream->ckey_en = 1;
		stream->ckey_src = PDP_CTRL_CKEYSRC_CUR;
		break;
	default:
		return -EINVAL;
	}
	pdpfb_configure_streams(priv);
	return 0;
}

static void pdpfb_str_get_ckey(struct pdpfb_priv *priv,
							struct pdpfb_stream *stream,
							struct pdpfb_ckey *ckey)
{
	*ckey = stream->ckey;
}

static void pdpfb_str_set_ckey(struct pdpfb_priv *priv,
							struct pdpfb_stream *stream,
							struct pdpfb_ckey *ckey)
{
	ckey->mask = ckey->mask & 0x3FFFFFFF;
	ckey->ckey = ckey->ckey & ckey->mask;
	stream->ckey = *ckey;
	if (stream->ckey_en)
		pdpfb_configure_streams(priv);
}

static void pdpfb_str_get_blend_config(struct pdpfb_priv *priv,
									struct pdpfb_stream *stream,
									struct pdpfb_blend_config *blend_config)
{
	*blend_config = stream->blend_config;
}

static int pdpfb_str_set_blend_config(struct pdpfb_priv *priv,
									struct pdpfb_stream *stream,
									struct pdpfb_blend_config *blend_config)
{
	if (blend_config->blend_mode > PDP_BLENDMODE_PIXEL)
		return -EINVAL;
	if (blend_config->porter_mode > PDP_PORTERMODE_FOREGROUND)
		return -EINVAL;
	if (blend_config->blend_type > PDP_BLENDTYPE_CLIP_TO_SELF)
		return -EINVAL;
	stream->blend_config = *blend_config;
	pdpfb_configure_streams(priv);
	return 0;
}

static int pdpfb_str_get_galpha(struct pdpfb_priv *priv,
								struct pdpfb_stream *stream)
{
	return stream->global_alpha;
}

static int pdpfb_str_set_galpha(struct pdpfb_priv *priv,
								struct pdpfb_stream *stream, int val)
{
	val &= 0x3FF;
	stream->global_alpha = val;
	if (stream->blend_config.blend_mode != PDP_BLENDMODE_NOALPHA)
		pdpfb_configure_streams(priv);
	return val;
}

int pdpfb_str_pan_display(struct pdpfb_priv *priv, struct pdpfb_stream *stream,
						struct fb_var_screeninfo *var, struct fb_info *info)
{
	if (var->xoffset + info->var.xres > info->var.xres_virtual ||
		var->yoffset + info->var.yres > info->var.yres_virtual)
		return -EINVAL;

	info->var.xoffset = var->xoffset;
	info->var.yoffset = var->yoffset;

	pdpfb_str_configure_addr(priv, stream);

	return 0;
}

static int pdpfb_str_configure_addr(struct pdpfb_priv *priv,
									struct pdpfb_stream *stream)
{
	struct fb_info *info = &stream->info;

	u32 xoffset = info->var.xoffset;
	u32 yoffset = info->var.yoffset
					/ SCALER_MODIFIER(priv,stream).src_yoffset_div;

	u32 ybuf_offset = 0;	/* Y base addr offset in bytes */
	u32 ubuf_offset = 0;	/* UV base addr offset in bytes */
	u32 vbuf_offset = 0;	/* V base addr offset in bytes */
	u32 xoffset_base = 0;	/* X offset applied to base address in pixels */

	if (stream->split_scale == PDP_SPLIT_SCALE_RIGHT_PLANE) 
		xoffset -= stream->caps.hs_taps;

	/* Use base address manipulation to implement x/y offsets */
	if (info->fix.type == FB_TYPE_PLANES) {
		switch (info->var.nonstd) {
		case PDP_PIXFMT_420_PL8:
			/* 64 pixel blocks of 64 bytes (Y) and 32 bytes (U and V) */
			xoffset_base = xoffset & ~0x3F;
			ybuf_offset = yoffset * stream->planar.y_line_length
							+ xoffset_base;
			ubuf_offset = (yoffset / 2) * stream->planar.u_line_length
							+ xoffset_base / 2;
			vbuf_offset = (yoffset / 2) * stream->planar.v_line_length
							+ xoffset_base / 2;
			break;
		case PDP_PIXFMT_420_PL8IVU:
		case PDP_PIXFMT_420_PL8IUV:
			/* 32 pixel blocks of 32 bytes (Y and UV) */
			xoffset_base = xoffset & ~0x1F;
			ybuf_offset = yoffset * stream->planar.y_line_length
							+ xoffset_base;
			ubuf_offset = (yoffset / 2) * stream->planar.u_line_length
							+ xoffset_base;
			vbuf_offset = ubuf_offset;
			break;
		case PDP_PIXFMT_420_PL10IVU:
		case PDP_PIXFMT_420_PL10IUV:
			/* 24 pixel blocks of 32 bytes (Y and UV) */
			xoffset_base = (xoffset / 24) * 24;
			ybuf_offset = yoffset * stream->planar.y_line_length
							+ xoffset_base * 4 / 3;
			ubuf_offset = (yoffset / 2) * stream->planar.u_line_length
							+ xoffset_base * 4 / 3;
			vbuf_offset = ubuf_offset;
			break;
		case PDP_PIXFMT_422_PL10IVU:
		case PDP_PIXFMT_422_PL10IUV:
			/* 24 pixel blocks of 32 bytes (Y and UV) */
			xoffset_base = (xoffset / 24) * 24;
			ybuf_offset = yoffset * stream->planar.y_line_length
							+ xoffset_base * 4 / 3;
			ubuf_offset = yoffset * stream->planar.u_line_length
							+ xoffset_base * 4 / 3;
			vbuf_offset = ubuf_offset;
			break;
		}
	} else {
		/* Non-planar */
		u32 ybuf_xoffset;		/* Y base addr xoffset in bytes */
		u32 ybuf_xoffset_align;	/* Aligned Y base addr xoffset in bytes */

		/* Find LCM of required adr alignment and pixel boundary in bytes */
		u32 align = lcm(1 << PDP_YADDR_ALIGN, info->var.bits_per_pixel / 8);

		ybuf_xoffset = xoffset * info->var.bits_per_pixel / 8;
		ybuf_xoffset_align = (ybuf_xoffset / align) * align;
		ybuf_offset = yoffset * info->fix.line_length + ybuf_xoffset_align;
		xoffset_base = ybuf_xoffset_align * 8 / info->var.bits_per_pixel;
	}

#if 1
	/* Workaround for SBM 73585 */
	stream->hskip = xoffset - xoffset_base;
#else
	if (PDPFB_REG_EXISTS(stream->regs.skip_ctrl)) {
		/* If supported, use HSkip to skip any remaining pixels */
		u32 skip_ctrl;
		u32 hskip = xoffset - xoffset_base;

		skip_ctrl = pdpfb_read(priv, stream->regs.skip_ctrl);
		SET_FIELD(skip_ctrl, PDP_VID1SKIPCTRL_VID1HSKIP, hskip);
		pdpfb_write(priv, stream->regs.skip_ctrl, skip_ctrl);
	}
#endif

	if (info->fix.type == FB_TYPE_PLANES) {
		u32 uctrl;
		u32 uv_half_stride;

		if (stream->planar.u_line_length == stream->planar.y_line_length)
			uv_half_stride = PDP_UADDR_UV_STRIDE_EQUAL_TO_Y_STRIDE;
		else if (stream->planar.u_line_length > stream->planar.y_line_length)
			uv_half_stride = PDP_UADDR_UV_STRIDE_EQUAL_TO_DOUBLE_Y_STRIDE;
		else
			uv_half_stride = PDP_UADDR_UV_STRIDE_EQUAL_TO_HALF_Y_STRIDE;

		uctrl = pdpfb_read(priv, stream->regs.uctrl);
		SET_FIELD(uctrl, PDP_VID1UCTRL_VID1UVHALFSTR, uv_half_stride);
		pdpfb_write(priv, stream->regs.uctrl, uctrl);

		pdpfb_write_base_addr(priv, stream, stream->regs.base_addr,
								stream->planar.y_offset + ybuf_offset);
		pdpfb_write_base_addr(priv, stream, stream->regs.u_base_addr,
								stream->planar.u_offset + ubuf_offset);
		pdpfb_write_base_addr(priv, stream, stream->regs.v_base_addr,
								stream->planar.v_offset + vbuf_offset);
	} else {
		/* Non-planar */
		pdpfb_write_base_addr(priv, stream, stream->regs.base_addr,
								ybuf_offset);
	}

	return 0;
}

int pdpfb_str_check_var(struct pdpfb_priv *priv, struct pdpfb_stream *stream,
						struct fb_var_screeninfo *var, struct fb_info *info)
{
#ifndef __PDPFB_USE_OSAL__
	const struct fb_videomode *mode;
#endif

	if (PDPFB_IS_PLANAR(var->nonstd) && !PDPFB_REG_EXISTS(stream->regs.uctrl))
		return -EINVAL;  /* This stream doesn't support planar formats */

	if (!var->xres)
		var->xres = 1;
	if (!var->yres)
		var->yres = 1;

	if (var->xres > MAX_FIELD_VALUE(PDP_GRPH1SIZE_GRPH1WIDTH) + 1)
		var->xres = MAX_FIELD_VALUE(PDP_GRPH1SIZE_GRPH1WIDTH) + 1;
	if (var->yres > MAX_FIELD_VALUE(PDP_GRPH1SIZE_GRPH1HEIGHT) + 1)
		var->yres = MAX_FIELD_VALUE(PDP_GRPH1SIZE_GRPH1HEIGHT) + 1;

	/* xres must be even, round up */
	var->xres = (var->xres + 1) & ~0x1;

	/* Mode match if mode master */
	if (stream->mode_master) {
	#ifndef __PDPFB_USE_OSAL__
		mode = fb_find_best_mode(var, &info->modelist);
		if (mode) {
			unsigned int xres_virtual = var->xres_virtual;
			unsigned int yres_virtual = var->yres_virtual;
			unsigned int xoffset = var->xoffset;
			unsigned int yoffset = var->yoffset;
			fb_videomode_to_var(var, mode);
			var->xres_virtual = xres_virtual;
			var->yres_virtual = yres_virtual;
			var->xoffset = xoffset;
			var->yoffset = yoffset;
		} else {
			return -EINVAL;
		}
	#else
		/* Not implemented */
		IMG_ASSERT(IMG_FALSE);
	#endif
	}

	if (var->xres > var->xres_virtual)
		var->xres_virtual = var->xres;
	if (var->yres > var->yres_virtual)
		var->yres_virtual = var->yres;

	if (var->xres_virtual < var->xoffset + var->xres)
		var->xres_virtual = var->xoffset + var->xres;
	if (var->yres_virtual < var->yoffset + var->yres)
		var->yres_virtual = var->yoffset + var->yres;

	pdpfb_str_match_var(priv, stream, var);

	/* Memory limit */
	if (pdpfb_str_required_mem(var) > stream->videomem_len)
		return -ENOMEM;

	return 0;
}

unsigned long pdpfb_str_required_mem(struct fb_var_screeninfo *var)
{
#ifndef __PDPFB_USE_OSAL__
	const int xres_virt = var->xres_virtual;
	const int align = PDP_YSTRIDE_ALIGN;

	if (PDPFB_IS_PLANAR(var->nonstd)) {
		switch (var->nonstd) {
		case PDP_PIXFMT_420_PL10IVU:
		case PDP_PIXFMT_420_PL10IUV:
			return ((pdpfb_get_line_len(xres_virt, 3, 32, align) * 3) / 2)
					* var->yres_virtual;
		case PDP_PIXFMT_422_PL10IVU:
		case PDP_PIXFMT_422_PL10IUV:
			return (pdpfb_get_line_len(xres_virt, 3, 32, align) * 2)
					* var->yres_virtual;
		case PDP_PIXFMT_420_PL8:
			return ((pdpfb_get_line_len(xres_virt, 1, 8, (align + 1)) * 3) / 2)
					* var->yres_virtual;
		case PDP_PIXFMT_420_PL8IVU:
		case PDP_PIXFMT_420_PL8IUV:
		default:
			return ((pdpfb_get_line_len(xres_virt, 1, 8, align) * 3) / 2)
					* var->yres_virtual;
		}
	} else {
		return pdpfb_get_line_len(xres_virt, 1, var->bits_per_pixel, align)
				* var->yres_virtual;
	}
#else
	return 0;
#endif
}

void pdpfb_str_match_var(struct pdpfb_priv *priv, struct pdpfb_stream *stream,
						struct fb_var_screeninfo *var)
{
	if (var->nonstd != 0) {
		/* Non-standard pixel formats */
		switch (var->nonstd) {
		case PDP_PIXFMT_420_PL8:
		case PDP_PIXFMT_420_PL8IVU:
		case PDP_PIXFMT_420_PL8IUV:
			var->bits_per_pixel = 12;
			break;
		case PDP_PIXFMT_420_PL10IVU:
		case PDP_PIXFMT_420_PL10IUV:
			var->bits_per_pixel = 16;
			break;
		case PDP_PIXFMT_422_PL10IVU:
		case PDP_PIXFMT_422_PL10IUV:
			var->bits_per_pixel = 22;
			break;
		case PDP_PIXFMT_422_UY0VY1_8888:
		case PDP_PIXFMT_422_VY0UY1_8888:
		case PDP_PIXFMT_422_Y0UY1V_8888:
		case PDP_PIXFMT_422_Y0VY1U_8888:
			var->bits_per_pixel = 16;
			break;
		case PDP_PIXFMT_AYUV8888:
		case PDP_PIXFMT_YUV101010:
			var->bits_per_pixel = 32;
			break;
		default:
			/* Unknown non-standard format, set to default */
			var->nonstd = PDP_PIXFMT_422_UY0VY1_8888;
			var->bits_per_pixel = 16;
			break;
		}

		if ((var->nonstd <= 0) || (var->nonstd >= ARRAY_SIZE(nonstd2pixfmt)))
			dev_err(priv->device, "invalid non-standard pixel format");
		else
			stream->pixfmt = nonstd2pixfmt[var->nonstd];

		var->red.offset = 0;
		var->red.length = 0;
		var->green.offset = 0;
		var->green.length = 0;
		var->blue.offset = 0;
		var->blue.length = 0;
		var->transp.offset = 0;
		var->transp.length = 0;
		var->red.msb_right = 0;
		var->green.msb_right = 0;
		var->blue.msb_right = 0;
		var->transp.msb_right = 0;

		if (PDPFB_IS_PLANAR(var->nonstd)) {
			if (stream->planar_override &&
				stream->planar_ov_virtres[0] == var->xres_virtual &&
				stream->planar_ov_virtres[1] == var->yres_virtual &&
				stream->planar_ov_pixfmt == var->nonstd) {
				/* if overridden planar setup is still ok, leave it */
			}
			else {
				const int xres_virt = var->xres_virtual;
				const int align = PDP_YSTRIDE_ALIGN;

				stream->planar_override = 0;
				switch (var->nonstd) {
				case PDP_PIXFMT_420_PL8:
					stream->planar.y_line_length =
							pdpfb_get_line_len(xres_virt, 1, 8, (align + 1));
					stream->planar.u_line_length =
							stream->planar.y_line_length / 2;
					stream->planar.v_line_length =
							stream->planar.u_line_length;
					stream->planar.y_offset = 0;
					stream->planar.u_offset = stream->planar.y_offset
							+ stream->planar.y_line_length * var->yres_virtual;
					stream->planar.v_offset = stream->planar.u_offset
						+ stream->planar.u_line_length * var->yres_virtual / 2;
					break;

				case PDP_PIXFMT_420_PL8IVU:
				case PDP_PIXFMT_420_PL8IUV:
					stream->planar.y_line_length =
							pdpfb_get_line_len(xres_virt, 1, 8, align);
					stream->planar.u_line_length =
							stream->planar.y_line_length;
					stream->planar.v_line_length =
							stream->planar.y_line_length;
					stream->planar.y_offset = 0;
					stream->planar.u_offset = stream->planar.y_offset
							+ stream->planar.y_line_length * var->yres_virtual;
					stream->planar.v_offset = stream->planar.u_offset;
					break;
				case PDP_PIXFMT_420_PL10IVU:
				case PDP_PIXFMT_420_PL10IUV:
				case PDP_PIXFMT_422_PL10IVU:
				case PDP_PIXFMT_422_PL10IUV:
					stream->planar.y_line_length =
							pdpfb_get_line_len(xres_virt, 3, 32, align);
					stream->planar.u_line_length =
							stream->planar.y_line_length;
					stream->planar.v_line_length =
							stream->planar.y_line_length;
					stream->planar.y_offset = 0;
					stream->planar.u_offset = stream->planar.y_offset
							+ stream->planar.y_line_length * var->yres_virtual;
					stream->planar.v_offset = stream->planar.u_offset;
					break;
				}
			}
		}
	} else {
		/* Standard pixel formats */
		int i;
		int best_pixfmt = -1;
		int best_rating = pixfmt_match_none;

		if (var->bits_per_pixel <= 8)
			var->bits_per_pixel = 8;
		else if (var->bits_per_pixel <= 16)
			var->bits_per_pixel = 16;
		else if (var->bits_per_pixel <= 24)
			var->bits_per_pixel = 24;
		else if (var->bits_per_pixel <= 32)
			var->bits_per_pixel = 32;
		else if (var->bits_per_pixel <= 40)
			var->bits_per_pixel = 40;
		else
			var->bits_per_pixel = 48;
			

	#define PIXFMT_MATCH(rating, cond) \
			if (!(cond)) \
				continue; \
			if (best_rating < (rating)) { \
				best_rating = (rating); \
				best_pixfmt = i; \
				if ((rating) == pixfmt_match_exact) \
					break; \
			}

		for (i = 0; i < ARRAY_SIZE(pixfmts); ++i) {
			int match_bits, match_gray;

			PIXFMT_MATCH(pixfmt_match_bpp,
					var->bits_per_pixel == pixfmts[i].bpp);
			PIXFMT_MATCH(pixfmt_match_alpha,
					(var->transp.length != 0) == (pixfmts[i].a.length != 0));

			match_bits = (var->red.offset == pixfmts[i].r.offset &&
							var->red.length == pixfmts[i].r.length &&
							var->green.offset == pixfmts[i].g.offset &&
							var->green.length == pixfmts[i].g.length &&
							var->blue.offset == pixfmts[i].b.offset &&
							var->blue.length == pixfmts[i].b.length &&
							var->transp.offset == pixfmts[i].a.offset &&
							var->transp.length == pixfmts[i].a.length);
			match_gray = (var->grayscale == pixfmts[i].grayscale);

			PIXFMT_MATCH(pixfmt_match_bitsorgray, match_bits || match_gray);
			PIXFMT_MATCH(pixfmt_match_exact, match_bits && match_gray);
		}

	#undef PIXFMT_MATCH

		var->bits_per_pixel = pixfmts[best_pixfmt].bpp;
		var->red = pixfmts[best_pixfmt].r;
		var->green = pixfmts[best_pixfmt].g;
		var->blue = pixfmts[best_pixfmt].b;
		var->transp = pixfmts[best_pixfmt].a;
		var->grayscale = pixfmts[best_pixfmt].grayscale;

		stream->pixfmt = pixfmts[best_pixfmt].id;
	}
}

static int pdpfb_str_check_planar(struct pdpfb_stream *stream,
								struct fb_info *info, struct pdpfb_planar *pl)
{
	int extra;

	/* check alignment */
	if (unlikely(pl->y_offset & PDP_YADDR_ALIGNMASK)) {
		pl->y_offset &= ~PDP_YADDR_ALIGNMASK;
		return -EINVAL;
	}
	if (unlikely(pl->u_offset & PDP_UADDR_ALIGNMASK)) {
		pl->u_offset &= ~PDP_UADDR_ALIGNMASK;
		return -EINVAL;
	}
	if (unlikely(pl->v_offset & PDP_VADDR_ALIGNMASK)) {
		pl->v_offset &= ~PDP_VADDR_ALIGNMASK;
		return -EINVAL;
	}
	if (unlikely(pl->y_line_length & PDP_YSTRIDE_ALIGNMASK)) {
		pl->y_line_length &= ~PDP_YSTRIDE_ALIGNMASK;
		return -EINVAL;
	}
	/* u/v strides must be the same */
	if (unlikely(pl->u_line_length != pl->v_line_length)) {
		pl->v_line_length = pl->u_line_length;
		return -EINVAL;
	}
	/* strides must be in range */
	if (unlikely(pl->y_line_length > PDP_YSTRIDE_MAX)) {
		pl->y_line_length = PDP_YSTRIDE_MAX;
		return -EINVAL;
	}
	/* u/v strides must be half, double or equal to y stride */
	if (unlikely((pl->u_line_length != pl->y_line_length) &&
					(pl->u_line_length * 2 != pl->y_line_length) &&
					(pl->u_line_length != pl->y_line_length * 2))) {
		if (pl->u_line_length < pl->y_line_length)
			pl->u_line_length = pl->y_line_length / 2;
		else if (pl->u_line_length > pl->y_line_length)
			pl->u_line_length = pl->y_line_length * 2;
		else
			pl->u_line_length = pl->y_line_length;
		pl->v_line_length = pl->u_line_length;
		return -EINVAL;
	}

	/* offsets must be in range (don't want to overflow later) */
	if (unlikely(pl->y_offset > PDP_YADDR_MAX ||
					pl->y_offset >= stream->videomem_len)) {
		pl->y_offset = 0;
		return -ENOMEM;
	}
	if (unlikely(pl->u_offset > PDP_UADDR_MAX ||
					pl->u_offset >= stream->videomem_len)) {
		pl->u_offset = 0;
		return -ENOMEM;
	}
	if (unlikely(pl->v_offset > PDP_VADDR_MAX ||
					pl->v_offset >= stream->videomem_len)) {
		pl->v_offset = 0;
		return -ENOMEM;
	}
	/* does y plane fit in memory? */
	extra = info->var.xres_virtual - pl->y_line_length;
	if (extra < 0)
		extra = 0;
	if (pl->y_offset + pl->y_line_length * info->var.yres_virtual + extra
													> stream->videomem_len) {
		pl->y_offset = 0;
		return -ENOMEM;
	}
	/* does u plane fit in memory? */
	extra = info->var.xres_virtual/2 - pl->u_line_length;
	if (extra < 0)
		extra = 0;
	if (pl->u_offset + pl->u_line_length * info->var.yres_virtual / 2 + extra
													> stream->videomem_len) {
		pl->u_offset = 0;
		return -ENOMEM;
	}
	/* does v plane fit in memory? (same extra as u plane) */
	if (pl->v_offset + pl->v_line_length * info->var.yres_virtual / 2 + extra
													> stream->videomem_len) {
		pl->v_offset = 0;
		return -ENOMEM;
	}

	return 0;
}

static int pdpfb_str_change_mode(struct pdpfb_priv *priv,
								struct pdpfb_stream *stream)
{
	struct fb_info *info = &stream->info;
	u32 stride;

	/* handle planar YUV */
	switch (info->var.nonstd) {
	case PDP_PIXFMT_420_PL8:
	case PDP_PIXFMT_420_PL8IVU:
	case PDP_PIXFMT_420_PL8IUV:
	case PDP_PIXFMT_420_PL10IVU:
	case PDP_PIXFMT_420_PL10IUV:
	case PDP_PIXFMT_422_PL10IVU:
	case PDP_PIXFMT_422_PL10IUV:
		info->fix.type = FB_TYPE_PLANES;
		break;

	case PDP_PIXFMT_422_UY0VY1_8888:
	case PDP_PIXFMT_422_VY0UY1_8888:
	case PDP_PIXFMT_422_Y0UY1V_8888:
	case PDP_PIXFMT_422_Y0VY1U_8888:
	case PDP_PIXFMT_AYUV8888:
	case PDP_PIXFMT_YUV101010:
	default:
		info->fix.type = FB_TYPE_PACKED_PIXELS;
		break;
	}

	if (info->fix.type == FB_TYPE_PLANES)
		info->fix.line_length = stream->planar.y_line_length;
	else
		info->fix.line_length = pdpfb_get_line_len(info->var.xres_virtual, 1,
													info->var.bits_per_pixel,
													PDP_YSTRIDE_ALIGN);

	stride = pdpfb_read(priv, stream->regs.stride);
	SET_FIELD(stride, PDP_GRPH1STRIDE_GRPH1STRIDE,
				(info->fix.line_length >> PDP_YSTRIDE_ALIGN) - 1);
	pdpfb_write(priv, stream->regs.stride, stride);

	if (info->var.nonstd == 0) {
		/* Set fix.visual for standard pixel formats */
		if (info->var.bits_per_pixel == 8)
			info->fix.visual = FB_VISUAL_PSEUDOCOLOR;
		else
			info->fix.visual = FB_VISUAL_TRUECOLOR;
	}

	return 0;
}

static void pdpfb_str_get_csc(struct pdpfb_stream *stream,
							struct pdpfb_csc *csc)
{
	*csc = stream->csc;
}

static void pdpfb_str_set_csc(struct pdpfb_stream *stream,
							struct pdpfb_csc *csc)
{
	csc->enable = (csc->enable != 0);
	csc->cosited = (csc->cosited != 0);
	stream->csc = *csc;
}

static int pdpfb_str_get_gamma_en(struct pdpfb_stream *stream)
{
	return stream->gamma_en;
}

static void pdpfb_str_set_gamma_en(struct pdpfb_priv *priv,
								struct pdpfb_stream *stream, int val)
{
	u32 surf;
	stream->gamma_en = (0 != val);
	surf = pdpfb_read(priv, stream->regs.surf);
	SET_FIELD(surf, PDP_GRPH1SURF_GRPH1USEGAMMA, stream->gamma_en);
	pdpfb_write(priv, stream->regs.surf, surf);

	pdpfb_flush_regs(priv);
}

static int pdpfb_str_get_dithering_en(struct pdpfb_stream *stream)
{
	return stream->dithering_en;
}

static void pdpfb_str_set_dithering_en(struct pdpfb_priv *priv,
									struct pdpfb_stream *stream, int val)
{
	u32 ctrl;
	stream->dithering_en = (0 != val);
	ctrl = pdpfb_read(priv, stream->regs.ctrl);
	SET_FIELD(ctrl, PDP_GRPH1CTRL_GRPH1DITHEREN, stream->dithering_en);
	pdpfb_write(priv, stream->regs.ctrl, ctrl);

	pdpfb_flush_regs(priv);
}

static int pdpfb_str_get_bilinear_en(struct pdpfb_stream *stream)
{
	return stream->bilinear_en;
}

static void pdpfb_str_set_bilinear_en(struct pdpfb_priv *priv,
									struct pdpfb_stream *stream, int val)
{
	int en = (val != 0);

	if (stream->bilinear_en != en) {
		stream->bilinear_en = en;

		pdpfb_str_configure_hscale(priv, stream);
		pdpfb_str_configure_vscale(priv, stream);
		pdpfb_flush_regs(priv);
	}
}

void pdpfb_str_get_planar(struct pdpfb_stream *stream, struct pdpfb_planar *pl)
{
	*pl = stream->planar;
}

int pdpfb_str_set_planar(struct pdpfb_priv *priv, struct pdpfb_stream *stream,
						struct fb_info *info, struct pdpfb_planar *pl)
{
	int error;

	/* pixel format must be planar */
	if (unlikely(info->fix.type != FB_TYPE_PLANES))
		return -EINVAL;

	error = pdpfb_str_check_planar(stream, info, pl);
	if (error < 0)
		return error;

	/* override the planar setup */
	stream->planar = *pl;
	stream->planar_override = 1;
	stream->planar_ov_virtres[0] = info->var.xres_virtual;
	stream->planar_ov_virtres[1] = info->var.yres_virtual;
	stream->planar_ov_pixfmt = info->var.nonstd;

	/* start using the new setup */
	pdpfb_str_change_mode(priv, stream);
	pdpfb_str_configure_addr(priv, stream);
	return 0;
}

void pdpfb_str_get_lumakey(struct pdpfb_stream *stream,
						struct pdpfb_lumakey *lumakey)
{
	*lumakey = stream->lumakey;
}

int pdpfb_str_set_lumakey(struct pdpfb_priv *priv, struct pdpfb_stream *stream,
						struct pdpfb_lumakey *lumakey)
{
	struct pdpfb_lumakey_coefs const *coefs;
	u32 reg;
	const int max_luma_key_min =
		PDP_VID1LUMAKEY_LUMA_MAX_MIN_VID1LUMAKEYYMIN_LSBMASK;
	const int max_luma_key_max =
		PDP_VID1LUMAKEY_LUMA_MAX_MIN_VID1LUMAKEYYMAX_LSBMASK;
	const int max_alpha_opaque =
		PDP_VID1LUMAKEY_ALPHA_TRANS_OPAQUE_VID1LUMAKEYALPHA_OPAQUE_LSBMASK;
	const int max_alpha_trans =
		PDP_VID1LUMAKEY_ALPHA_TRANS_OPAQUE_VID1LUMAKEYALPHA_TRANS_LSBMASK;

	if (lumakey->enable) {
		if (lumakey->preset) {
			if (lumakey->preset < ARRAY_SIZE(pdpfb_lumakey_presets))
				coefs = &pdpfb_lumakey_presets[lumakey->preset];
			else
				return -EINVAL;
		}
		else {
			#define LUMA_KEY_COEF_INVALID(x) \
				((x) > PDP_VID1LUMAKEY_C_RG_VID1LUMAKEYC_R_LSBMASK)
			if (LUMA_KEY_COEF_INVALID(lumakey->coefs.cr) ||
				LUMA_KEY_COEF_INVALID(lumakey->coefs.cg) ||
				LUMA_KEY_COEF_INVALID(lumakey->coefs.cb)) {
				return -EINVAL;
			}
			#undef LUMA_KEY_COEF_INVALID

			if (lumakey->coefs.output_offset >
				PDP_VID1LUMAKEY_C_B_VID1LUMAKEYOUTOFF_LSBMASK)
				return -EINVAL;
			coefs = &lumakey->coefs;
		}

		if ((lumakey->luma_key_min > max_luma_key_min) ||
			(lumakey->luma_key_max > max_luma_key_max) ||
			(lumakey->alpha_opaque > max_alpha_opaque) ||
			(lumakey->alpha_trans > max_alpha_trans))
			return -EINVAL;

		reg = PLACE_FIELD(PDP_VID1LUMAKEY_LUMA_MAX_MIN_VID1LUMAKEYYMIN,
					lumakey->luma_key_min);
		SET_FIELD(reg, PDP_VID1LUMAKEY_LUMA_MAX_MIN_VID1LUMAKEYYMAX,
					lumakey->luma_key_max);
		pdpfb_write(priv, stream->regs.lumakey_luma_max_min, reg);
			
		reg = PLACE_FIELD(
					PDP_VID1LUMAKEY_ALPHA_TRANS_OPAQUE_VID1LUMAKEYALPHA_OPAQUE,
					lumakey->alpha_opaque);
		SET_FIELD(reg,
					PDP_VID1LUMAKEY_ALPHA_TRANS_OPAQUE_VID1LUMAKEYALPHA_TRANS,
					lumakey->alpha_trans);
		pdpfb_write(priv, stream->regs.lumakey_alpha_trans_opaque, reg);

		reg = PLACE_FIELD(PDP_VID1LUMAKEY_C_RG_VID1LUMAKEYC_G, coefs->cg);
		SET_FIELD(reg, PDP_VID1LUMAKEY_C_RG_VID1LUMAKEYC_R, coefs->cr);
		pdpfb_write(priv, stream->regs.lumakey_c_rg, reg);			
			
		reg = PLACE_FIELD(PDP_VID1LUMAKEY_C_B_VID1LUMAKEYC_B, coefs->cb);
		SET_FIELD(reg, PDP_VID1LUMAKEY_C_B_VID1LUMAKEYOUTOFF, 
					coefs->output_offset);
		SET_FIELD(reg, PDP_VID1LUMAKEY_C_B_VID1LUMAKEYEN, 1);
		SET_FIELD(reg, PDP_VID1LUMAKEY_C_B_VID1LUMAKEYALPHAMULT,
					lumakey->luma_key_alpha_mult ? 1 : 0);
		pdpfb_write(priv, stream->regs.lumakey_c_b, reg);

	} else {
		pdpfb_write(priv, stream->regs.lumakey_c_b,
					PLACE_FIELD(PDP_VID1LUMAKEY_C_B_VID1LUMAKEYEN, 0));
	}

	stream->lumakey = *lumakey;
	pdpfb_flush_regs(priv);

	return 0;
}

void pdpfb_str_get_clip_geom(struct pdpfb_stream *stream,
							struct pdpfb_geom *clip_geom)
{
	*clip_geom = stream->clip_geom;
}

int pdpfb_str_set_clip_geom(struct pdpfb_priv *priv,
					struct pdpfb_stream *stream, struct pdpfb_geom *clip_geom)
{
	const u32 geomw = stream->geom.w ? stream->geom.w : stream->info.var.xres;
	const u32 geomh = stream->geom.h ? stream->geom.h : stream->info.var.yres;

	if (clip_geom->w + clip_geom->x > geomw ||
		clip_geom->h + clip_geom->y > geomh)
		return -EINVAL;

	if (clip_geom->x > MAX_FIELD_VALUE(PDP_VID1POSTSKIPCTRL_VID1HPOSTCLIP) ||
		clip_geom->y > MAX_FIELD_VALUE(PDP_VID1POSTSKIPCTRL_VID1VPOSTCLIP))
		return -EINVAL;

	stream->clip_geom = *clip_geom;

	pdpfb_configure_streams(priv);
	pdpfb_update_margins(priv, stream);

	return 0;
}

int pdpfb_str_get_split_scale(struct pdpfb_stream *stream)
{
	return stream->split_scale;
}

int pdpfb_str_set_split_scale(struct pdpfb_priv *priv,
					struct pdpfb_stream *stream, int split_scale)
{
	if (!stream->caps.has_scaler)
		return -EINVAL;	

	if (split_scale < 0 || split_scale > PDP_SPLIT_SCALE_RIGHT_PLANE)
		return -EINVAL;	

	stream->split_scale = split_scale;
	pdpfb_configure_streams(priv);

	return 0;
}

int pdpfb_str_set_par(struct pdpfb_priv *priv, struct pdpfb_stream *stream,
					struct fb_info *info)
{
	pdpfb_str_check_geom(priv, stream, &stream->geom);

	/* pixel format maps from nonstd, so may need updating */
	pdpfb_str_match_var(priv, stream, &info->var);

	if (stream->mode_master) {
		pdpfb_set_mode(priv, &info->var);
	} else {
		pdpfb_str_configure(priv, stream);
		pdpfb_str_configure_addr(priv, stream);
		pdpfb_update_margins(priv, stream);
	}

	pdpfb_flush_regs(priv);

	/* Calculate info->fix.xpanstep */
	if (PDPFB_REG_EXISTS(stream->regs.skip_ctrl)) {
		/* If supported, HSkip is used to implement single pixel X panning */
		info->fix.xpanstep = 1;
	} else {
		/* If not supported, only base address manipulation is used */
		if (info->fix.type == FB_TYPE_PLANES) {
			/* Planar pixel formats */
			switch (info->var.nonstd) {
			case PDP_PIXFMT_420_PL8:
				/* 64 pixel blocks of 64 bytes (Y) and 32 bytes (U and V) */
				info->fix.xpanstep = 64;
				break;
			case PDP_PIXFMT_420_PL8IVU:
			case PDP_PIXFMT_420_PL8IUV:
				/* 32 pixel blocks of 32 bytes (Y and UV) */
				info->fix.xpanstep = 32;
				break;
			case PDP_PIXFMT_420_PL10IVU:
			case PDP_PIXFMT_420_PL10IUV:
			case PDP_PIXFMT_422_PL10IVU:
			case PDP_PIXFMT_422_PL10IUV:
				/* 24 pixel blocks of 32 bytes (Y and UV) */
				info->fix.xpanstep = 24;
				break;
			}
		} else {
			/* Non-planar formats */
			/* Find LCM of required adr alignment and pixel boundary in bits */
			u32 align = lcm(8 << PDP_YADDR_ALIGN, info->var.bits_per_pixel);
			info->fix.xpanstep = align / info->var.bits_per_pixel;
		}
	}

	return 0;
}

static int pdpfb_str_check_geom(struct pdpfb_priv *priv,
						struct pdpfb_stream *stream, struct pdpfb_geom *geom)
{
	struct fb_info *info = &stream->info;

	if (stream->caps.has_scaler) {
		/* geom->w must be even because it's in YUV */
		if (geom->w) {
			if (geom->w == info->var.xres) {
				geom->w = 0;
			} else if (info->var.xres > stream->caps.max_scaler_input_width) {
				/* The scaler cannot be enabled if input width exceeds the max */
				geom->w = 0;
			} else if (geom->w > info->var.xres * 8) {
				/* Max horizontal scale factor is x8  */
				geom->w = info->var.xres * 8;
			} else if ((u32)(geom->w * 8) < info->var.xres) {
				/* Min horizontal scale factor is x1/8 */
				geom->w = (info->var.xres + 7) / 8;
				geom->w = (geom->w + 1) & ~0x1; /* round up to even */
			} else {
				geom->w &= ~0x1;
			}

			if (geom->w && geom->w < info->var.xres && geom->y == 0) {
				/*
				 * When scaling down horizontally with screen y position = 0,
				 * set screen y position to 1 to avoid a h/w issue (SBM 73206)
				*/
				geom->y = 1;
			}
		}
		/* geom->h must be even because it's in YUV */
		if (geom->h) {
			if (geom->h == info->var.yres) {
				geom->h = 0;
			} else if (geom->h > info->var.yres * 8) {
				/* Max vertical scale factor is x8	*/
				geom->h = info->var.yres * 8;
			} else if ((u32)(geom->h * 31) < info->var.yres) {
				/* Min vertical scale factor is x1/31  */
				geom->h = (info->var.yres + 30) / 31;
				geom->h = (geom->h + 1) & ~0x1; /* round up to even */
			} else {
				geom->h &= ~0x1;
			}
		}
	}
	else {
		/* No scaling capability */
		if (geom->w) {
			if (((u32)(geom->w * 2) >= info->var.xres * 3)
						&& stream->caps.has_doubler)
				geom->w = info->var.xres * 2;
			else if (((u32)(geom->w * 3) <= info->var.xres * 2)
						&& stream->caps.has_decimator)
				geom->w = info->var.xres / 2;
			else
				geom->w = 0;
		}
		if (geom->h) {
			const int max_n = MAX_FIELD_VALUE(
					PDP_GRPH1DECIMATE_CTRL_GRPH1DECIMATE_LINE_DISCARD_COUNT);
			/* Max geomh achievable with decimation using 'Read N, skip 1' */
			const int max_geomh = ((info->var.yres + 1) * max_n) / (max_n + 1);
			/* geomh threshold below which decimation will be enabled */
			const int geomh_threshold = (info->var.yres - max_geomh) / 2
										+ max_geomh;

			if (((u32)(geom->h * 2) >= info->var.yres * 3)
					&& stream->caps.has_doubler) {
				geom->h = info->var.yres * 2;
			}
			else if (((u32)geom->h <= geomh_threshold)
						&& stream->caps.has_decimator) {
				u32 read_or_skip_n = 0;

				if ((u32)(geom->h * 2) > info->var.yres) {
					/*
					 * Pitch is between 1 and 2 - use 'Read N, skip 1'
					 * n = geomh / (yres - geomh)
					 * Round n to the nearest integer to achieve the closest
					 * possible resulting geomh value by adding half the
					 * denominator to the numerator
					 * n = (geomh + (yres - geomh) / 2) / (yres - geomh)
					 *   = (yres + geomh) / (2 * (yres - geomh)
					 */
					read_or_skip_n = (info->var.yres + geom->h)
										/ (2 * (info->var.yres - geom->h));

					/* Clip read_or_skip_n to the maximum */
					if (read_or_skip_n > max_n)
						read_or_skip_n = max_n;

					/*
					 * Calculate resulting geomh
					 * geomh = ((yres + 1) * n) / (n + 1)
					 */
					geom->h = ((info->var.yres + 1) * read_or_skip_n)
								/ (read_or_skip_n + 1);
				} else {
					/* 
					 * Pitch is >= 2 - use 'Read 1, skip N'
					 * n = (yres - geomh) / geomh
					 * Round n to the nearest integer to achieve the closest
					 * possible resulting geomh value by adding half the
					 * denominator to the numerator
					 * n = (yres - geomh + (geomh / 2)) / geomh 
					 *   = ((2 * yres) - geomh) / (2 * geomh)
					 */
					read_or_skip_n = ((2 * info->var.yres) - geom->h)
										/ (2 * geom->h);

					/* Clip read_or_skip_n to the maximum */
					if (read_or_skip_n > max_n)
						read_or_skip_n = max_n;

					/*
					 * Calculate resulting geomh
					 * geomh = (yres + n) / (n + 1)
					 */
					geom->h = (info->var.yres + read_or_skip_n)
								/ (read_or_skip_n + 1);
				}
			}
			else {
				geom->h = 0;
			}
		}
	}

	return 0;
}

static void pdpfb_str_configure(struct pdpfb_priv *priv,
							struct pdpfb_stream *stream)
{
	struct fb_info *info = &stream->info;
	u32 reg;

	reg = pdpfb_read(priv, stream->regs.surf);
	if (stream->caps.has_csc)
		SET_FIELD(reg, PDP_GRPH1SURF_GRPH1USECSC, stream->csc.enable);
	if (stream->caps.has_cosited_chroma_control)
		SET_FIELD(reg, PDP_VID1SURF_VID1COSITED, stream->csc.cosited);
	SET_FIELD(reg, PDP_VID1SURF_VID1PIXFMT, stream->pixfmt);
	pdpfb_write(priv, stream->regs.surf, reg);

#define BURSTLENGTH 8
	reg = pdpfb_read(priv, stream->regs.memthresh);
	SET_FIELD(reg, PDP_GRPH1_MEM_THRESH_GRPH1_THRESHOLD, 256 - BURSTLENGTH);
	SET_FIELD(reg, PDP_GRPH1_MEM_THRESH_GRPH1_YTHRESHOLD, 256 - BURSTLENGTH);
	SET_FIELD(reg, PDP_GRPH1_MEM_THRESH_GRPH1_UVTHRESHOLD, 128 - BURSTLENGTH);
	pdpfb_write(priv, stream->regs.memthresh, reg);

	reg = pdpfb_read(priv, stream->regs.memctrl);
	SET_FIELD(reg, PDP_GRPH1_MEMCTRL_GRPH1_BURSTLEN, BURSTLENGTH);
	SET_FIELD(reg, PDP_GRPH1_MEMCTRL_GRPH1_LOCAL_GLOBAL_MEMCTRL, 1);
	pdpfb_write(priv, stream->regs.memctrl, reg);
#undef BURSTLENGTH

	reg = pdpfb_read(priv, stream->regs.size);
	if (stream->split_scale) {
		SET_FIELD(reg, PDP_GRPH1SIZE_GRPH1WIDTH,
					info->var.xres + stream->caps.hs_taps - 1);
	} else {
		SET_FIELD(reg, PDP_GRPH1SIZE_GRPH1WIDTH, info->var.xres - 1);
	}
	SET_FIELD(reg, PDP_GRPH1SIZE_GRPH1HEIGHT,
		(info->var.yres / SCALER_MODIFIER(priv,stream).src_height_div) - 1);
	pdpfb_write(priv, stream->regs.size, reg);

	pdpfb_str_change_mode(priv, stream);

	pdpfb_str_configure_hscale(priv, stream);
	pdpfb_str_configure_vscale(priv, stream);
}

static unsigned int pdpfb_get_signature(struct pdpfb_priv *priv)
{
	u32 signat_r, signat_gb, signature;
	signat_r = pdpfb_read(priv, PDP_SIGNAT_R_OFFSET);
	signat_gb = pdpfb_read(priv, PDP_SIGNAT_GB_OFFSET);
	signature = GET_FIELD(signat_r, PDP_SIGNAT_R_SIGNATURE_R) << 20
				| GET_FIELD(signat_gb, PDP_SIGNAT_GB_SIGNATURE_G) << 10
				| GET_FIELD(signat_gb, PDP_SIGNAT_GB_SIGNATURE_B);
	return signature;
}

enum pdpfb_addr_type {
	PDP_ADDR_PHYS,
	PDP_ADDR_USER_VIRT
};

static int pdpfb_update_pdp_base(struct pdpfb_priv *priv,
								struct pdpfb_stream *stream,
								u32 addr, u32 len,
								enum pdpfb_addr_type addr_type)
{
#if defined(CONFIG_FB_PDP2_MMU) || defined(CONFIG_FB_PDP2_PDUMP)
	u32 addr_to_map;
	u32 len_to_map;
#endif

#ifdef CONFIG_FB_PDP2_MMU
	int error;
	int ret;
	int dir_entry;
	int page_entry;
	u32 cur_offset;
	u32 *page_table;
	u32 cpu_page_offset;
	u32 page_addr_phys;
	dma_addr_t ptbl_bus;
#endif

#ifdef CONFIG_FB_PDP2_MMU
	if (unlikely(len > PDP_MAX_IMAGE_BYTES))
		return -EINVAL;

	if (addr & ~PDP_PAGE_MASK) {
		/*
		 * If usermem->addr is not on a page boundary, start mapping from
		 * the start of the page
		*/
		addr_to_map = addr - (addr & ~PDP_PAGE_MASK);
		len_to_map = len + (addr & ~PDP_PAGE_MASK);
	}
	else {
		addr_to_map = addr;
		len_to_map = len;
	}
#elif defined(CONFIG_FB_PDP2_PDUMP)
	addr_to_map = addr;
	len_to_map = len;
#endif

#ifdef CONFIG_FB_PDP2_PDUMP
	if (!priv->pdump_suspended) {
		PDUMP_LOG("MALLOC :MEM:BLOCK_%d %d %d\n",
				priv->pdump_next_mem_block, len_to_map, PDP_PAGE_SIZE);
		stream->pdump_image_mem_block = priv->pdump_next_mem_block++;
		if (strlen(stream->usermem_file.name)) {
			PDUMP_LOG("LDB :MEM:BLOCK_%d:0x%x 0x%x %d %s\n",
				stream->pdump_image_mem_block, (addr - addr_to_map),
				len, stream->usermem_file.offset,
				stream->usermem_file.name);
		} else {
			PDUMP_LOG("-- LDB :MEM:BLOCK_%d -- NO FILE NAME / OFFSET\n",
				stream->pdump_image_mem_block);
		}
	}
#endif

#ifdef CONFIG_FB_PDP2_MMU
	for (dir_entry = 0, cur_offset = 0; cur_offset < len_to_map; dir_entry++) {
		/* Allocate one page table at a time */
		page_table = (u32*)kmalloc(PDP_PAGE_SIZE, GFP_KERNEL);
		if (unlikely(!page_table)) {
			error = -ENOMEM;
			goto err;
		}
#ifdef CONFIG_FB_PDP2_PDUMP
		PDUMP_LOG("MALLOC :MEM:BLOCK_%d %d %d\n",
				priv->pdump_next_mem_block, PDP_PAGE_SIZE, PDP_PAGE_SIZE);
		stream->pdump_pdir_mem_block[stream->next_dir][dir_entry]
				= priv->pdump_next_mem_block++;
#endif

		/* Fill in the page table */
		for (page_entry = 0;
				cur_offset < len_to_map && page_entry < PDPFB_PTRS_PER_PTBL;
				cur_offset += PDP_PAGE_SIZE, page_entry++) {
			if (addr_type == PDP_ADDR_USER_VIRT) {
				struct page **pdir_user_page = &stream->pdir_user_page
									[stream->next_dir][dir_entry][page_entry];
				down_read(&current->mm->mmap_sem);
				ret = get_user_pages(current, current->mm,
									(unsigned long)(addr_to_map + cur_offset),
									1, 0, 0, pdir_user_page, NULL);
				up_read(&current->mm->mmap_sem);
				if (unlikely(ret <= 0)) {
					error = -EFAULT;
					goto err;
				}
#if (PAGE_SIZE != PDP_PAGE_SIZE)
				cpu_page_offset = (addr_to_map + cur_offset) & ~PAGE_MASK;
#else
				cpu_page_offset = 0;
#endif
				page_addr_phys = dma_map_page(priv->device,	*pdir_user_page,
												cpu_page_offset, PDP_PAGE_SIZE,
												DMA_TO_DEVICE);
				if (unlikely(!page_addr_phys)) {
					put_page(*pdir_user_page);
					*pdir_user_page = NULL;
					error = -EFAULT;
					goto err;
				}
			}
			else {
				page_addr_phys = addr_to_map + cur_offset;
			}

			if (unlikely(page_addr_phys & ~PDP_PAGE_MASK)) {
				error = -EFAULT;
				goto err;
			}
			/* | 0x1 sets the valid bit */
			page_table[page_entry] = page_addr_phys | 0x1;

#ifdef CONFIG_FB_PDP2_PDUMP
			PDUMP_LOG("WRW :MEM:$1 :MEM:BLOCK_%d:0x%X\n",
					stream->pdump_image_mem_block, cur_offset);
			PDUMP_LOG("OR :MEM:$1 :MEM:$1 0x1\n");
			PDUMP_LOG("WRW :MEM:BLOCK_%d:0x%X :MEM:$1\n",
					stream->pdump_pdir_mem_block[stream->next_dir][dir_entry],
					page_entry * 4);
#endif
		}

		/* Mark any unused page table entries as invalid */
		for (; page_entry < PDPFB_PTRS_PER_PTBL; page_entry++) {
			page_table[page_entry] = 0;
#ifdef CONFIG_FB_PDP2_PDUMP
			PDUMP_LOG("WRW :MEM:BLOCK_%d:0x%X 0x0\n",
					stream->pdump_pdir_mem_block[stream->next_dir][dir_entry],
					page_entry * 4);
#endif
		}

		/* Page table complete, now fill in the directory entry */
		ptbl_bus = dma_map_single(priv->device, page_table, PDP_PAGE_SIZE,
									DMA_TO_DEVICE);
		if (unlikely(!ptbl_bus)) {
			error = -EFAULT;
			goto err;
		}
		if (unlikely(ptbl_bus & ~PDP_PAGE_MASK)) {
			dma_unmap_single(priv->device, ptbl_bus, PDP_PAGE_SIZE, DMA_TO_DEVICE);
			error = -EFAULT;
			goto err;
		}
		/* | 0x1 sets the valid bit */
		(*stream->pdir)[stream->next_dir][dir_entry] = ptbl_bus | 0x1;
		/* The bus address is retained for unmapping */
		stream->pdir_bus[stream->next_dir][dir_entry] = ptbl_bus;
		/* The virtual address is retained for freeing */
		stream->pdir_kvirt[stream->next_dir][dir_entry] = page_table;

#ifdef CONFIG_FB_PDP2_PDUMP
		PDUMP_LOG("WRW :MEM:$1 :MEM:BLOCK_%d:0x0\n",
					stream->pdump_pdir_mem_block[stream->next_dir][dir_entry]);
		PDUMP_LOG("OR :MEM:$1 :MEM:$1 0x1\n");
		PDUMP_LOG("WRW :MEM:BLOCK_%d:0x%X :MEM:$1\n",
					priv->pdump_pdir_mem_block,
					(stream->pdir_offset[stream->next_dir] + dir_entry) * 4);
#endif
	}

	/*
	 * The PDP virtual base address is:
	 * <dir_offset> << 22 | <page_table_offset> << 12 | <page_offset>
	 * The page table offset is zero, as the first page mapped is the one
	 * containing the buffer base 
	*/
	stream->videomem_pdp_base =	stream->pdir_offset[stream->next_dir] << 22
								| (addr & ~PDP_PAGE_MASK);
#else /* !CONFIG_FB_PDP2_MMU */
	if (addr_type != PDP_ADDR_PHYS)
		return -EINVAL;

	stream->videomem_pdp_base = addr;
#endif /* !CONFIG_FB_PDP2_MMU */

	/*
	 * Program the base address registers now, which will be latched on vfetch
	 * and used for the next frame
	*/
	pdpfb_str_configure_addr(priv, stream);
	/*
	 * Flag the directory as updated after programming the base address regs,
	 * in case we get a VSYNC interrupt between the two, to prevent the ISR
	 * from invalidating the old page tables while they're still being used
	*/
	stream->next_dir_updated = 1;

	return 0;
#ifdef CONFIG_FB_PDP2_MMU
err:
	/*
	 * Process the page table that was being completed when the error occured
	 * (if allocated)
	*/
	if (page_table) {
		/* Unmap and release any user pages */
		for (page_entry = 0; page_entry < PDPFB_PTRS_PER_PTBL; page_entry++) {
			struct page **pdir_user_page = &stream->pdir_user_page
									[stream->next_dir][dir_entry][page_entry];
			if (*pdir_user_page != NULL) {
				dma_unmap_page(priv->device, (page_table[page_entry] & ~0x1),
								PDP_PAGE_SIZE, DMA_TO_DEVICE);
				put_page(*pdir_user_page);
				*pdir_user_page = NULL;
			} else {
				break;
			}
		}
		kfree(page_table);
	}
	/* Process any page tables already entered into the directory */
	for (; dir_entry > 0; dir_entry--) {
		/* Unmap and release any user pages */
		for (page_entry = 0; page_entry < PDPFB_PTRS_PER_PTBL; page_entry++) {
			struct page **pdir_user_page = &stream->pdir_user_page
								[stream->next_dir][dir_entry - 1][page_entry];
			if (*pdir_user_page != NULL) {
				/* & ~0x1 masks off the valid bit */
				const u32 page_addr_phys = stream->pdir_kvirt
						[stream->next_dir][dir_entry - 1][page_entry] & ~0x1;
				dma_unmap_page(priv->device, page_addr_phys,
								PDP_PAGE_SIZE, DMA_TO_DEVICE);
				put_page(*pdir_user_page);
				*pdir_user_page = NULL;
			} else {
				break;
			}
		}
		/* Unmap and free the page table */
		if (stream->pdir_kvirt[stream->next_dir][dir_entry- 1]) {
			dma_unmap_single(priv->device,
							stream->pdir_bus[stream->next_dir][dir_entry - 1],
							PDP_PAGE_SIZE, DMA_TO_DEVICE);
			kfree(stream->pdir_kvirt[stream->next_dir][dir_entry - 1]);
			stream->pdir_bus[stream->next_dir][dir_entry - 1] = 0;
			stream->pdir_kvirt[stream->next_dir][dir_entry - 1] = NULL;
			/* Also invalidate the physical directory entry */
			(*stream->pdir)[stream->next_dir][dir_entry - 1] = 0;
		}
	}
	return error;
#endif /* CONFIG_FB_PDP2_MMU */
}

static int pdpfb_set_usermem(struct pdpfb_priv *priv,
					struct pdpfb_stream *stream, struct pdpfb_usermem *usermem)
{
	int i;
	int ret;
	int error;
	enum pdpfb_addr_type addr_type;

	/* Usermem addresses must be correctly aligned */
	if (usermem->addr & PDP_YADDR_ALIGNMASK)
		return -EINVAL;

	if (usermem->flags & PDP_USERMEM_ALLPLANES) {
		/* Virtual user memory for all planes is not supported */
		if (usermem->flags & PDP_USERMEM_VIRTUAL)
			return -EINVAL;

		pdpfb_init_mempool(priv, PDPFB_MEMPOOL_USER, usermem->addr,
							usermem->len);

		ret = 0;
		for (i = priv->stream_nr - 1; i >= 0; i--) {
			struct pdpfb_stream *stream = &priv->streams[i];
			if (!stream || !stream->probed)
				continue;
			pdpfb_str_videomem_free(priv, stream);
		}
		for (i = 0; i < priv->stream_nr; i++) {
			struct pdpfb_stream *stream = &priv->streams[i];
			if (!stream || !stream->probed)
				continue;
			error = pdpfb_str_videomem_alloc(priv, stream);
			if (error && !ret)
				ret = error;
		}

		if (!ret)
			return 0;
		return -ENOMEM;
	}

	addr_type = usermem->flags & PDP_USERMEM_VIRTUAL ? 
					PDP_ADDR_USER_VIRT : PDP_ADDR_PHYS;
	error = pdpfb_update_pdp_base(priv, stream, usermem->addr, usermem->len,
									addr_type);
	if (error)
		return error;

	stream->videomem_cpu_virt = NULL;
	stream->videomem_len = usermem->len;

	/* Usermem will not be accessible from framebuffer file */
	stream->info.screen_base = 
							(char __force __iomem *)stream->videomem_cpu_virt;
	stream->info.fix.smem_start = usermem->addr;
	stream->info.fix.smem_len = usermem->len;
	stream->mem_pool = PDPFB_MEMPOOL_USERPLANE;

	return 0;
}

int pdpfb_str_ioctl(struct pdpfb_priv *priv, struct pdpfb_stream *stream,
					unsigned int cmd, unsigned long arg)
{
	void __user *argp = (void __user *)arg;
	struct fb_vblank vblank;
	struct pdpfb_bgnd_col bgnd_col;
	struct pdpfb_disp_mode disp_mode;
	struct pdpfb_geom geom;
	struct pdpfb_csc_config csc_config;
	struct pdpfb_procamp procamp;
	struct pdpfb_gamma_config gamma_config;
	int plane_pos;
	struct pdpfb_ckey ckey;
	struct pdpfb_blend_config blend_config;
	struct pdpfb_csc csc;
	int error;
	u32 val;
#ifdef CONFIG_FB_PDP2_USERMEM
	struct pdpfb_usermem usermem;
#endif /* CONFIG_FB_PDP2_USERMEM */

	switch (cmd) {
	/* get blanking info */
	case FBIOGET_VBLANK:
		pdpfb_str_get_vblank(priv, stream, &vblank);
		if (copy_to_user(argp, &vblank, sizeof(vblank)))
			return -EFAULT;
		return 0;
	/* wait for vsync */
	case FBIO_WAITFORVSYNC:
		if (get_user(val, (u32 __user *)arg))
			return -EFAULT;
		if (val != 0)
			return -ENODEV;
		return pdpfb_wait_vsync(priv);
	/* get background colour */
	case PDPIO_GETBGND:
		pdpfb_get_bgnd_col(priv, &bgnd_col);
		if (copy_to_user(argp, &bgnd_col, sizeof(bgnd_col)))
			return -EFAULT;
		return 0;
	/* set background colour */
	case PDPIO_SETBGND:
		if (copy_from_user(&bgnd_col, argp, sizeof(bgnd_col)))
			return -EFAULT;
		pdpfb_set_bgnd_col(priv, &bgnd_col);
		return 0;
	/* get display mode */
	case PDPIO_GETDISPMODE:
		pdpfb_get_disp_mode(priv, &disp_mode);
		if (copy_to_user(argp, &disp_mode, sizeof(disp_mode)))
			return -EFAULT;
		return 0;
	/* set display mode */
	case PDPIO_SETDISPMODE:
		if (copy_from_user(&disp_mode, argp, sizeof(disp_mode)))
			return -EFAULT;
		error = pdpfb_set_disp_mode(priv, &disp_mode);
		if (error)
			return error;
		if (copy_to_user(argp, &disp_mode, sizeof(disp_mode)))
			return -EFAULT;
		return 0;
	/* get colour space conversion settings */
	case PDPIO_GETCSCCONFIG:
		pdpfb_get_csc_config(priv, &csc_config);
		if (copy_to_user(argp, &csc_config, sizeof(csc_config)))
			return -EFAULT;
		return 0;
	/* set colour space conversion settings */
	case PDPIO_SETCSCCONFIG:
		if (copy_from_user(&csc_config, argp, sizeof(csc_config)))
			return -EFAULT;
		error = pdpfb_set_csc_config(priv, &csc_config);
		if (error)
			return error;
		if (copy_to_user(argp, &csc_config, sizeof(csc_config)))
			return -EFAULT;
		return 0;
	/* get procamp settings */
	case PDPIO_GETPROCAMP:
		pdpfb_get_procamp(priv, &procamp);
		if (copy_to_user(argp, &procamp, sizeof(procamp)))
			return -EFAULT;
		return 0;
	/* set procamp settings */
	case PDPIO_SETPROCAMP:
		if (copy_from_user(&procamp, argp, sizeof(procamp)))
			return -EFAULT;
		error = pdpfb_set_procamp(priv, &procamp);
		if (error)
			return error;
		return 0;
	/* get gamma conversion settings */
	case PDPIO_GETGAMMACONFIG:
		pdpfb_get_gamma_config(priv, &gamma_config);
		if (copy_to_user(argp, &gamma_config, sizeof(gamma_config)))
			return -EFAULT;
		return 0;
	/* set gamma conversion settings */
	case PDPIO_SETGAMMACONFIG:
		if (copy_from_user(&gamma_config, argp, sizeof(gamma_config)))
			return -EFAULT;
		error = pdpfb_set_gamma_config(priv, &gamma_config);
		if (error)
			return error;
		return 0;
	/* get 3D mode */
	case PDPIO_GET3DMODE:
		val = pdpfb_get_3d_mode(priv);
		if (unlikely(val < 0))
			return -EINVAL;
		if (copy_to_user(argp, &val, sizeof(val)))
			return -EFAULT;
		return 0;
	/* set 3D mode */
	case PDPIO_SET3DMODE:
		if (copy_from_user(&val, argp, sizeof(val)))
			return -EFAULT;
		error = pdpfb_set_3d_mode(priv, val);
		if (error)
			return error;
		return 0;
	/* get output colour depth */
	case PDPIO_GETOUTPUTDEPTH:
		val = pdpfb_get_output_depth(priv);
		if (unlikely(val < 0))
			return -EINVAL;
		if (copy_to_user(argp, &val, sizeof(val)))
			return -EFAULT;
		return 0;
	/* set output colour depth */
	case PDPIO_SETOUTPUTDEPTH:
		if (copy_from_user(&val, argp, sizeof(val)))
			return -EFAULT;
		error = pdpfb_set_output_depth(priv, val);
		if (error)
			return error;
		return 0;
	/* get stream enable */
	case PDPIO_GETEN:
		val = pdpfb_str_get_en(priv, stream);
		if (copy_to_user(argp, &val, sizeof(val)))
			return -EFAULT;
		return 0;
	/* set stream enable */
	case PDPIO_SETEN:
		if (copy_from_user(&val, argp, sizeof(val)))
			return -EFAULT;
		pdpfb_str_set_en(priv, stream, val);
		return 0;
	/* get stream plane position */
	case PDPIO_GETPLANEPOS:
		plane_pos = pdpfb_str_get_plane_pos(priv, stream);
		if (unlikely(plane_pos < 0))
			return -EINVAL;
		if (copy_to_user(argp, &plane_pos, sizeof(plane_pos)))
			return -EFAULT;
		return 0;
	/* set stream plane position */
	case PDPIO_SETPLANEPOS:
		if (copy_from_user(&plane_pos, argp, sizeof(plane_pos)))
			return -EFAULT;

		plane_pos = pdpfb_str_set_plane_pos(priv, stream, plane_pos);
		if (plane_pos < 0)
			return plane_pos;

		if (copy_to_user(argp, &plane_pos, sizeof(plane_pos)))
			return -EFAULT;
		return 0;
	/* get stream geometry */
	case PDPIO_GETGEOM:
		pdpfb_str_get_geom(priv, stream, &geom);
		if (copy_to_user(argp, &geom, sizeof(geom)))
			return -EFAULT;
		return 0;
	/* set stream geometry */
	case PDPIO_SETGEOM:
		if (copy_from_user(&geom, argp, sizeof(geom)))
			return -EFAULT;
		error = pdpfb_str_set_geom(priv, stream, &geom);
		if (error)
			return error;
		if (copy_to_user(argp, &geom, sizeof(geom)))
			return -EFAULT;
		return 0;
	/* get stream colour key mode */
	case PDPIO_GETCKEYMODE:
		val = pdpfb_str_get_ckey_mode(priv, stream);
		if (unlikely(val < 0))
			return -EINVAL;
		if (copy_to_user(argp, &val, sizeof(val)))
			return -EFAULT;
		return 0;
	/* set stream colour key mode */
	case PDPIO_SETCKEYMODE:
		if (copy_from_user(&val, argp, sizeof(val)))
			return -EFAULT;
		error = pdpfb_str_set_ckey_mode(priv, stream, val);
		if (error)
			return error;
		return 0;
	/* get stream colour key */
	case PDPIO_GETCKEY:
		pdpfb_str_get_ckey(priv, stream, &ckey);
		if (copy_to_user(argp, &ckey, sizeof(ckey)))
			return -EFAULT;
		return 0;
	/* set stream colour key */
	case PDPIO_SETCKEY:
		if (copy_from_user(&ckey, argp, sizeof(ckey)))
			return -EFAULT;
		pdpfb_str_set_ckey(priv, stream, &ckey);
		if (copy_to_user(argp, &ckey, sizeof(ckey)))
			return -EFAULT;
		return 0;
	/* get stream blending mode */
	case PDPIO_GETBLENDCONFIG:
		pdpfb_str_get_blend_config(priv, stream, &blend_config);
		if (copy_to_user(argp, &blend_config, sizeof(blend_config)))
			return -EFAULT;
		return 0;
	/* set stream blending mode */
	case PDPIO_SETBLENDCONFIG:
		if (copy_from_user(&blend_config, argp, sizeof(blend_config)))
			return -EFAULT;
		error = pdpfb_str_set_blend_config(priv, stream, &blend_config);
		if (error)
			return error;
		return 0;
	/* gset stream global alpha */
	case PDPIO_GETGALPHA:
		val = pdpfb_str_get_galpha(priv, stream);
		if (copy_to_user(argp, &val, sizeof(val)))
			return -EFAULT;
		return 0;
	/* set stream global alpha */
	case PDPIO_SETGALPHA:
		if (copy_from_user(&val, argp, sizeof(val)))
			return -EFAULT;
		val = pdpfb_str_set_galpha(priv, stream, val);
		if (copy_to_user(argp, &val, sizeof(val)))
			return -EFAULT;
		return 0;
	/* get colour space conversion settings */
	case PDPIO_GETCSC:
		pdpfb_str_get_csc(stream, &csc);
		if (copy_to_user(argp, &csc, sizeof(csc)))
			return -EFAULT;
		return 0;
	/* set colour space conversion settings */
	case PDPIO_SETCSC:
		if (copy_from_user(&csc, argp, sizeof(csc)))
			return -EFAULT;
		pdpfb_str_set_csc(stream, &csc);
		if (copy_to_user(argp, &csc, sizeof(csc)))
			return -EFAULT;
		return 0;
	/* get gamma enable */
	case PDPIO_GETGAMMAEN:
		val = pdpfb_str_get_gamma_en(stream);
		if (copy_to_user(argp, &val, sizeof(val)))
			return -EFAULT;
		return 0;
	/* set gamma enable */
	case PDPIO_SETGAMMAEN:
		if (copy_from_user(&val, argp, sizeof(val)))
			return -EFAULT;
		pdpfb_str_set_gamma_en(priv, stream, val);
		return 0;
	/* get dithering enable */
	case PDPIO_GETDITHERINGEN:
		val = pdpfb_str_get_dithering_en(stream);
		if (copy_to_user(argp, &val, sizeof(val)))
			return -EFAULT;
		return 0;
	/* set dithering enable */
	case PDPIO_SETDITHERINGEN:
		if (copy_from_user(&val, argp, sizeof(val)))
			return -EFAULT;
		pdpfb_str_set_dithering_en(priv, stream, val);
		return 0;
	/* get bilinear filtering enable */
	case PDPIO_GETBILINEAREN:
		val = pdpfb_str_get_bilinear_en(stream);
		if (copy_to_user(argp, &val, sizeof(val)))
			return -EFAULT;
		return 0;
	/* set bilinear filtering enable */
	case PDPIO_SETBILINEAREN:
		if (copy_from_user(&val, argp, sizeof(val)))
			return -EFAULT;
		pdpfb_str_set_bilinear_en(priv, stream, val);
		return 0;
	/* set user provided memory */
	case PDPIO_SETUSERMEM:
#ifdef CONFIG_FB_PDP2_USERMEM
		if (copy_from_user(&usermem, argp, sizeof(usermem)))
				return -EFAULT;
		error = pdpfb_set_usermem(priv, stream, &usermem);
		if (error)
			return error;
		else
			return 0;
#else /* !CONFIG_FB_PDP2_USERMEM */
		return -EINVAL;
#endif /* !CONFIG_FB_PDP2_USERMEM */
	/* get signature */
	case PDPIO_GETSIGNATURE:
		val = pdpfb_get_signature(priv);
		if (copy_to_user(argp, &val, sizeof(val)))
			return -EFAULT;
		return 0;
	case PDPIO_SETSIMULATOR:
		if (copy_from_user(&val, argp, sizeof(val)))
			return -EFAULT;
		priv->is_sim = (val != 0);
		return 0;
	case PDPIO_SETUSERMEMFILE:
#if defined(CONFIG_FB_PDP2_USERMEM) && defined(CONFIG_FB_PDP2_PDUMP)
		if (copy_from_user(&stream->usermem_file, argp,
							sizeof(stream->usermem_file)))
			return -EFAULT;
		else
			return 0;
#else /* !CONFIG_FB_PDP2_USERMEM || !FB_PDP2_PDUMP */
		return -EINVAL;
#endif /* !CONFIG_FB_PDP2_USERMEM || !FB_PDP2_PDUMP */
	default:
		return -ENOTTY;
	}
	return 0;
}

void pdpfb_str_videomem_free(struct pdpfb_priv *priv,
							struct pdpfb_stream *stream)
{
	struct fb_info *info = &stream->info;
	struct pdpfb_mem_pool *pool;

	if (stream->mem_pool == PDPFB_MEMPOOL_KERNEL) {
#ifndef __PDPFB_USE_OSAL__
		kfree(stream->videomem_cpu_virt);
#else
		/* Not implemented */
		IMG_ASSERT(IMG_FALSE);
#endif
	} else if (stream->mem_pool < PDPFB_MEMPOOL_NR_POOLS) {
		pool = &priv->pools[stream->mem_pool];

		if ((unsigned long)pool->base + pool->allocated ==
			(unsigned long)stream->videomem_cpu_virt + stream->videomem_len) {
			pool->allocated -= stream->videomem_len;
		} else {
			dev_warn(priv->device, "free losing memory in pool '%s'",
						pdpfb_mem_pool_names[stream->mem_pool]);
		}
	}

	stream->videomem_cpu_virt = NULL;
	stream->videomem_pdp_base = 0;
	info->screen_base = NULL;
	info->fix.smem_start = 0;
	info->fix.smem_len = 0;
	stream->mem_pool = PDPFB_MEMPOOL_NONE;
}

#ifndef __PDPFB_USE_OSAL__
static int pdpfb_pool_alloc(struct pdpfb_priv *priv,
							struct pdpfb_stream *stream,
							struct pdpfb_mem_pool *pool,
							unsigned long len)
{
	struct fb_info *info = &stream->info;
	unsigned long offset;
	unsigned long len_aligned = (len + PDP_PAGE_SIZE - 1) & PDP_PAGE_MASK;
	int error;

	offset = pool->allocated;
	if (offset + len > pool->size)
		return -ENOMEM;
	pool->allocated += len_aligned;

	error = pdpfb_update_pdp_base(priv, stream, pool->phys_base + offset,
									len_aligned, PDP_ADDR_PHYS);
	if (error)
		return error;

	stream->videomem_cpu_virt = (void *)((unsigned long)pool->base + offset);
	stream->videomem_len = len_aligned;

	info->screen_base = (char __force __iomem *)stream->videomem_cpu_virt;
	info->fix.smem_start = pool->phys_base + offset;
	info->fix.smem_len = len;

	return 0;
}
#endif

int pdpfb_str_videomem_alloc(struct pdpfb_priv *priv,
							struct pdpfb_stream *stream)
{
#ifndef __PDPFB_USE_OSAL__
	struct fb_info *info = &stream->info;
	void *videomem = NULL;
	unsigned long len;
	int error;

	/* Set default videomem length to the full pool size */
	len = priv->pools[stream->mem_pool].size;

	/* Lengths specified in kernel config or module params override default */
	if (stream->mem_pool == PDPFB_MEMPOOL_GFX1MEM && gfx1_videomem_len)
		len = gfx1_videomem_len;
	else if (stream->mem_pool == PDPFB_MEMPOOL_GFX2MEM && gfx2_videomem_len)
		len = gfx2_videomem_len;
	else if (stream->mem_pool == PDPFB_MEMPOOL_GFX3MEM && gfx3_videomem_len)
		len = gfx3_videomem_len;
	else if (stream->mem_pool == PDPFB_MEMPOOL_GFX4MEM && gfx4_videomem_len)
		len = gfx4_videomem_len;
	else if (stream->mem_pool == PDPFB_MEMPOOL_VID1MEM && vid1_videomem_len)
		len = vid1_videomem_len;
	else if (stream->mem_pool == PDPFB_MEMPOOL_VID2MEM && vid2_videomem_len)
		len = vid2_videomem_len;
	else if (stream->mem_pool == PDPFB_MEMPOOL_VID3MEM && vid3_videomem_len)
		len = vid3_videomem_len;
	else if (vid4_videomem_len)
		len = vid4_videomem_len;
	if (!len)
		return 0;

	/* Try user memory pool */
	if (priv->pools[PDPFB_MEMPOOL_USER].base) {
		if (!pdpfb_pool_alloc(priv, stream, &priv->pools[PDPFB_MEMPOOL_USER],
								len)) {
			stream->mem_pool = PDPFB_MEMPOOL_USER;
			return 0;
		}
	}
	/* Try default memory pool (depends on stream type) */
	if (stream->mem_pool < PDPFB_MEMPOOL_NR_POOLS &&
		priv->pools[stream->mem_pool].base) {
		if (pdpfb_pool_alloc(priv, stream, &priv->pools[stream->mem_pool],
								priv->pools[stream->mem_pool].size))
			goto err_nomem;
		return 0;
	}
	/* Try combined memory pool */
	if (priv->pools[PDPFB_MEMPOOL_MEM].base) {
		if (pdpfb_pool_alloc(priv, stream, &priv->pools[PDPFB_MEMPOOL_MEM],
								len))
			goto err_nomem;
		stream->mem_pool = PDPFB_MEMPOOL_MEM;
		return 0;
	}
	/* Finally try kernel allocated memory */
	videomem = kzalloc(len, GFP_KERNEL);
	if (!videomem)
		goto err_nomem;
	stream->mem_pool = PDPFB_MEMPOOL_KERNEL;

	error = pdpfb_update_pdp_base(priv, stream, __pa(videomem), len,
									PDP_ADDR_PHYS);
	if (error)
		return error;

	stream->videomem_cpu_virt = videomem;
	stream->videomem_len = len;

	info->screen_base = (char __force __iomem *)stream->videomem_cpu_virt;
	info->fix.smem_start = __pa(videomem);
	info->fix.smem_len = len;
	return 0;

err_nomem:
	dev_err(priv->device, "unable to allocate video memory\n");
	return -ENOMEM;
#else
	return 0;
#endif
}

static int pdpfb_probe_streams(struct pdpfb_priv *priv,
							struct platform_device *pdev)
{
	int i;
	int error;

	for (i = 0; i < priv->stream_nr; ++i) {
		struct pdpfb_stream *stream = &priv->streams[i];
		if (!stream)
			continue;
		stream->probed = 1;
		error = stream->ops.probe(priv, pdev);
		if (error)
			stream->probed = 0;
		if (error == -ENOMEM)
			dev_err(priv->device,
					"unable to initialise fb %d, "
					"not enough video memory\n",
					i);
	}

	return 0;
}

static void pdpfb_remove_streams(struct pdpfb_priv *priv,
								struct platform_device *pdev)
{
	int i;

	for (i = 0; i < priv->stream_nr; ++i) {
		struct pdpfb_stream *stream = &priv->streams[i];
		if (stream && stream->probed)
			stream->ops.remove(priv, pdev);
	}
}

static int pdpfb_stop(struct pdpfb_priv *priv)
{
	pdpfb_set_display_enabled(priv, PDP_FULLY_POWERED_DOWN);

	return 0;
}

static inline int pdpfb_init_mem(struct pdpfb_priv *priv)
{
	return 0;
}

#ifndef __PDPFB_USE_OSAL__
static void pdpfb_free_pools(struct pdpfb_priv *priv)
{
	int i;
	for (i = 0; i < PDPFB_MEMPOOL_NR_POOLS; ++i) {
		if (!priv->pools[i].base)
			continue;
		if (priv->pools[i].flags & PDPFB_MEMPOOLF_KERNEL)
			kfree(priv->pools[i].base);
		else
			iounmap((void __force __iomem *)priv->pools[i].base);
	}
}
#endif

static int pdpfb_init_mempool(struct pdpfb_priv *priv, int type,
							unsigned long phys, unsigned long len)
{
	struct pdpfb_mem_pool *pool = &priv->pools[type];
	/* First come first served */
	if (unlikely(pool->base))
		return -EBUSY;
	pool->base = (void __force *)ioremap_cachable(phys, len);
	if (unlikely(!pool->base)) {
		dev_err(priv->device, "cannot ioremap %s @0x%08lx:0x%lx\n",
				pdpfb_mem_pool_names[type], phys, len);
		return -ENOMEM;
	}
	dev_info(priv->device, "%s @0x%08lx:0x%lx (ioremapped to 0x%p)\n",
				pdpfb_mem_pool_names[type],
				phys, len, pool->base);
	priv->pools[type].phys_base = phys;
	priv->pools[type].size = len;
	priv->pools[type].flags = 0;
	/* Clear the memory */
	memset(priv->pools[type].base, 0, priv->pools[type].size);
	return 0;
}

#ifndef __PDPFB_USE_OSAL__
static int pdpfb_probe_params(struct platform_device *pdev,
							struct pdpfb_priv *priv)
{
	if (videomem_base > 0 && videomem_len > 0)
		pdpfb_init_mempool(priv, PDPFB_MEMPOOL_MEM, videomem_base,
							videomem_len);
	if (gfx1_videomem_base > 0 && gfx1_videomem_len > 0)
		pdpfb_init_mempool(priv, PDPFB_MEMPOOL_GFX1MEM, gfx1_videomem_base,
							gfx1_videomem_len);
	if (gfx2_videomem_base > 0 && gfx2_videomem_len > 0)
		pdpfb_init_mempool(priv, PDPFB_MEMPOOL_GFX2MEM, gfx2_videomem_base,
							gfx2_videomem_len);
	if (gfx3_videomem_base > 0 && gfx3_videomem_len > 0)
		pdpfb_init_mempool(priv, PDPFB_MEMPOOL_GFX3MEM, gfx3_videomem_base,
							gfx3_videomem_len);
	if (gfx4_videomem_base > 0 && gfx4_videomem_len > 0)
		pdpfb_init_mempool(priv, PDPFB_MEMPOOL_GFX4MEM, gfx4_videomem_base,
							gfx4_videomem_len);
	if (vid1_videomem_base > 0 && vid1_videomem_len > 0)
		pdpfb_init_mempool(priv, PDPFB_MEMPOOL_VID1MEM, vid1_videomem_base,
							vid1_videomem_len);
	if (vid2_videomem_base > 0 && vid2_videomem_len > 0)
		pdpfb_init_mempool(priv, PDPFB_MEMPOOL_VID2MEM, vid2_videomem_base,
							vid2_videomem_len);
	if (vid3_videomem_base > 0 && vid3_videomem_len > 0)
		pdpfb_init_mempool(priv, PDPFB_MEMPOOL_VID3MEM, vid3_videomem_base,
							vid3_videomem_len);
	if (vid4_videomem_base > 0 && vid4_videomem_len > 0)
		pdpfb_init_mempool(priv, PDPFB_MEMPOOL_VID4MEM, vid4_videomem_base,
							vid4_videomem_len);
	return 0;
}
#endif

static int pdpfb_probe_mems(struct platform_device *pdev,
							struct pdpfb_priv *priv)
{
	struct resource *res_mem;
	int i;
	int error;

	res_mem = platform_get_resource_byname(pdev, IORESOURCE_MEM, "pdp_reg");
	if (!res_mem) {
		dev_err(&pdev->dev, "no pdp register ioresource\n");
		return -ENOMEM;
	}
	priv->base = ioremap_nocache(res_mem->start,
									res_mem->end - res_mem->start);
	if (!priv->base)
		return -ENOMEM;

#ifdef CONFIG_FB_PDP2_MMU
	res_mem = platform_get_resource_byname(pdev, IORESOURCE_MEM,
											"pdp_mmu_reg");
	if (!res_mem) {
		dev_err(&pdev->dev, "no pdp mmu register ioresource\n");
		return -ENOMEM;
	}
	priv->mmu_base = ioremap_nocache(res_mem->start,
										res_mem->end - res_mem->start);
	if (!priv->mmu_base)
		return -ENOMEM;
#endif

	/* Video memory pools (optional) */
	for (i = 0; i < PDPFB_MEMPOOL_NR_POOLS; i++) {
		res_mem = platform_get_resource_byname(pdev, IORESOURCE_MEM,
												pdpfb_mem_pool_names[i]);												
		if (res_mem) {
			error = pdpfb_init_mempool(priv, i, res_mem->start,
										res_mem->end - res_mem->start);
			if (error)
				return error;
		}
	}

	return 0;
}

static int pdpfb_get_properties(struct device_node *node,
								struct pdpfb_cfg *config)
{
	int error;
	struct videomode vm;

	/* Display timings */
	error = of_get_videomode(node, &vm, OF_USE_NATIVE_MODE);
	if (error)
		return error;

	if (vm.flags & DISPLAY_FLAGS_DE_HIGH)
		config->sync_cfg.de_active_high = 1;
	if (vm.flags & DISPLAY_FLAGS_PIXDATA_POSEDGE)
		config->sync_cfg.clock_inverted = 1;

	error = fb_videomode_from_videomode(&vm, &config->lcd_cfg);
	if (error)
		return error;

	/* Required properties */
	if (of_property_read_u32(node, "img,pdp-bpp", &config->bpp))
		return -EINVAL;
	if (of_property_read_u32(node, "img,pdp-linestore_len",
								&config->linestore_len))
		return -EINVAL;

	/* Optional properties */
	of_property_read_u32(node, "img,pdp-dynamic_mode",
							&config->lcd_size.dynamic_mode);
	of_property_read_u32(node, "img,pdp-width_in_mm",
							&config->lcd_size.width);
	of_property_read_u32(node, "img,pdp-height_in_mm",
							&config->lcd_size.height);
	of_property_read_u32(node, "img,pdp-force_vsyncs",
							&config->sync_cfg.force_vsyncs);
	of_property_read_u32(node, "img,pdp-hsync_dis",
							&config->sync_cfg.hsync_dis);
	of_property_read_u32(node, "img,pdp-vsync_dis",
							&config->sync_cfg.vsync_dis);
	of_property_read_u32(node, "img,pdp-blank_dis",
							&config->sync_cfg.blank_dis);

	return 0;
}

static int pdpfb_add_videomodes(struct device_node *node,
								struct list_head *modelist)
{
	int i;
	int error;
	struct display_timings *timings;
	struct videomode vm;
	struct fb_videomode fb_vm;

	timings = of_get_display_timings(node);
	if (!timings)
		return -EINVAL;

	for (i = 0; i < timings->num_timings; i++) {
		error = videomode_from_timings(timings, &vm, i);
		if (error)
			return error;
		error = fb_videomode_from_videomode(&vm, &fb_vm);
		if (error)
			return error;
		fb_add_videomode(&fb_vm, modelist);
	}

	return 0;
}

PDPFB_STATIC int pdpfb_probe(struct platform_device *pdev)
{
	struct pdpfb_priv *priv;
#ifndef __PDPFB_USE_OSAL__
	struct resource *res_irq;
#endif
	int error;

	if (!pdev->dev.of_node) {
		dev_err(&pdev->dev, "no device tree info\n");
		error = -EINVAL;
		goto err0;
	}

#ifndef __PDPFB_USE_OSAL__
	res_irq = platform_get_resource(pdev, IORESOURCE_IRQ, 0);
	if (res_irq == NULL) {
		dev_err(&pdev->dev, "cannot find IRQ resource\n");
		error = -ENOENT;
		goto err0;
	}
#endif

	priv = kzalloc(sizeof(*priv), GFP_KERNEL);
	if (!priv) {
		dev_err(&pdev->dev, "cannot allocate device data\n");
		error = -ENOMEM;
		goto err0;
	}

	priv->device = &pdev->dev;

	error = pdpfb_get_properties(pdev->dev.of_node, &priv->cfg);
	if (error) {
		dev_err(&pdev->dev, "error reading device tree properties\n");
		goto err1;
	}

#ifdef __PDPFB_USE_OSAL__
	pdpfb_initialise_osal(pdev->dev.of_node, priv);
#endif /* __PDPFB_USE_OSAL__ */

#if defined(CONFIG_FB_PDP2_PDUMP) && defined(__PDPFB_USE_OSAL__)
	priv->hOut2Txt = fopen ( "out2.txt", "w" );
#endif

	spin_lock_init(&priv->upd.lock);
	spin_lock_init(&priv->irq_lock);

	platform_set_drvdata(pdev, priv);
	priv->mode = priv->cfg.lcd_cfg;

#ifndef __PDPFB_USE_OSAL__
	error = pdpfb_probe_params(pdev, priv);
	if (error)
		goto err1;
#endif

	error = pdpfb_probe_mems(pdev, priv);

#ifndef __PDPFB_USE_OSAL__
	if (error)
		goto err1;
#endif

#ifndef __PDPFB_USE_OSAL__
	priv->pdp_clk = clk_get(priv->device, "core");
	if (IS_ERR(priv->pdp_clk)) {
		dev_warn(&pdev->dev, "could not get core clock resource\n");
		priv->pdp_clk = NULL;
	}
	else {
		clk_prepare(priv->pdp_clk);
	}

	priv->pixel_clk = clk_get(priv->device, "pixel");
	if (IS_ERR(priv->pixel_clk)) {
		dev_warn(&pdev->dev, "could not get pixel clock resource\n");
		priv->pixel_clk = NULL;
	}
#endif

	pdpfb_pdp_clk_enable(priv);

	error = pdpfb_init_mem(priv);
	if (error)
		goto err3;

#ifndef __PDPFB_USE_OSAL__
	priv->irq = res_irq->start;
	error = request_irq(priv->irq, pdpfb_interrupt, 0, "pdp", priv);
	if (error) {
		dev_err(&pdev->dev, "cannot register IRQ %d (%d)\n", priv->irq, error);
		error = -EIO;
		goto err3;
	}
#endif

	pdpfb_configure(priv);
	pdpfb_configure_dithering(priv);
	pdpfb_register_isr(priv, pdpfb_vsync_isr, priv,
						PDPFB_IRQ_VEVENT0 | PDPFB_IRQ_VEVENT1);

	pdpfb_gfx1_init_stream(&priv->streams[PDPFB_STREAM_GFX1]);
#ifdef CONFIG_FB_PDP2_GFX1_MODEMASTER
	priv->streams[PDPFB_STREAM_GFX1].mode_master
						= priv->cfg.lcd_size.dynamic_mode;
#endif
	pdpfb_gfx2_init_stream(&priv->streams[PDPFB_STREAM_GFX2]);
	pdpfb_gfx3_init_stream(&priv->streams[PDPFB_STREAM_GFX3]);
	pdpfb_gfx4_init_stream(&priv->streams[PDPFB_STREAM_GFX4]);
	pdpfb_vid1_init_stream(&priv->streams[PDPFB_STREAM_VID1]);
	pdpfb_vid2_init_stream(&priv->streams[PDPFB_STREAM_VID2]);
	pdpfb_vid3_init_stream(&priv->streams[PDPFB_STREAM_VID3]);
	pdpfb_vid4_init_stream(&priv->streams[PDPFB_STREAM_VID4]);
	priv->stream_nr = 8;

#ifdef CONFIG_FB_PDP2_MMU
	{
		int i, dir_offset;
		u32 reg;

#if (PAGE_SIZE % PDP_PAGE_SIZE != 0)
#error PAGE_SIZE must be a multiple of PDP_PAGE_SIZE
#endif

		priv->pdir = dma_alloc_coherent(priv->device, PDPFB_PAGE_DIR_BYTES,
										&priv->pdir_bus, GFP_KERNEL);
		if (!priv->pdir) {
			dev_err(&pdev->dev, "cannot allocate mmu page directory\n");
			error = -ENOMEM;
			goto err4;
		}
		if (priv->pdir_bus & ~PDP_PAGE_MASK) {
			dev_err(&pdev->dev,
					"mmu page directory bus address is not page aligned\n");
			error = -EFAULT;
			goto err5;
		}

		/* Initialise stream MMU page directory pointers */
		for (i = 0; i < priv->stream_nr; i++) {
			dir_offset = i * 2 * PDPFB_PTRS_PER_PDIR;
			priv->streams[i].pdir =
				(u32(*)[2][PDPFB_PTRS_PER_PDIR])(priv->pdir + dir_offset);
			priv->streams[i].pdir_offset[0] = dir_offset;
			priv->streams[i].pdir_offset[1] = dir_offset + PDPFB_PTRS_PER_PDIR;
		}

		/* Set MMU directory base address */
		pdpfb_mmu_write(priv, PDP_BIF_DIR_BASE_ADDR_OFFSET, priv->pdir_bus);

#ifdef CONFIG_FB_PDP2_PDUMP
		PDUMP_LOG("MALLOC :MEM:BLOCK_%d %d %d\n", priv->pdump_next_mem_block,
					PDP_PAGE_SIZE, PDP_PAGE_SIZE);
		priv->pdump_pdir_mem_block = priv->pdump_next_mem_block++;
		PDUMP_LOG("WRW :REGPDP_BIF:0x%08x :MEM:BLOCK_%d:0x0\n",
					PDP_BIF_DIR_BASE_ADDR_OFFSET,
					priv->pdump_pdir_mem_block);
#endif

		/* Enable MMU access for all requesters */
		reg = pdpfb_mmu_read(priv, PDP_BIF_ADDRESS_CONTROL_OFFSET);
		SET_FIELD(reg, PDP_BIF_ADDRESS_CONTROL_MMU_BYPASS, 0);
		pdpfb_mmu_write(priv, PDP_BIF_ADDRESS_CONTROL_OFFSET, reg);
	}
#endif

	if (pdpfb_probe_streams(priv, pdev))
		goto err5;

	if (priv->cfg.lcd_size.dynamic_mode) {
		error = pdpfb_add_videomodes(pdev->dev.of_node,
							&priv->streams[PDPFB_STREAM_GFX1].info.modelist);
		if (error) {
			dev_err(&pdev->dev, "error adding videomodes\n");
		}
	}

	priv->csc_config.preset = PDP_CSCPRESET_HDTV;
	priv->csc_config.coefs = pdpfb_csc_presets[PDP_CSCPRESET_HDTV];

	pdpfb_configure_streams(priv);

	_pdpfb_set_irqs(priv);

	pdpfb_debugfs_setup(priv);
	pdpfb_ident(priv);

	return 0;
err5:
#ifdef CONFIG_FB_PDP2_MMU
	dma_free_coherent(priv->device, PDPFB_PAGE_DIR_BYTES, priv->pdir,
					priv->pdir_bus);
err4:
#endif
	pdpfb_unregister_isr(priv, pdpfb_vsync_isr, priv,
						PDPFB_IRQ_VEVENT0 | PDPFB_IRQ_VEVENT1);
#ifndef __PDPFB_USE_OSAL__
	free_irq(priv->irq, priv);
#endif
err3:
#ifndef __PDPFB_USE_OSAL__
	pdpfb_pdp_clk_disable(priv);
	if (priv->pixel_clk != NULL)
		clk_put(priv->pixel_clk);
#endif
err1:
	if (priv->base)
		iounmap(priv->base);
#ifndef __PDPFB_USE_OSAL__
	pdpfb_free_pools(priv);

	kfree(priv);
#endif
err0:
	return error;
}


PDPFB_STATIC int pdpfb_remove(struct platform_device *pdev)
{
	struct pdpfb_priv *priv = platform_get_drvdata(pdev);

#ifndef __PDPFB_USE_OSAL__
	if (priv->stats_dir)
		debugfs_remove_recursive(priv->stats_dir);
#endif

	pdpfb_stop(priv);
	pdpfb_pdp_clk_disable(priv);

#ifdef CONFIG_FB_PDP2_MMU
	dma_free_coherent(priv->device, PDPFB_PAGE_DIR_BYTES, priv->pdir,
						priv->pdir_bus);
#endif

	pdpfb_unregister_isr(priv, pdpfb_vsync_isr, priv,
						PDPFB_IRQ_VEVENT0 | PDPFB_IRQ_VEVENT1);
	pdpfb_remove_streams(priv, pdev);

#ifndef __PDPFB_USE_OSAL__
	free_irq(priv->irq, priv);
	if (priv->pixel_clk != NULL)
		clk_put(priv->pixel_clk);

	if (priv->pdp_clk != NULL) {
		clk_unprepare(priv->pdp_clk);
		clk_put(priv->pdp_clk);
	}
	if (priv->base)
		iounmap(priv->base);
	pdpfb_free_pools(priv);

	kfree(priv);
#endif
	return 0;
}

PDPFB_STATIC void pdpfb_shutdown(struct platform_device *pdev)
{
	struct pdpfb_priv *priv = platform_get_drvdata(pdev);

	pdpfb_stop(priv);
}

#ifndef __PDPFB_USE_OSAL__
#ifdef CONFIG_PM
static int pdpfb_suspend(struct platform_device *pdev, pm_message_t state)
{
	struct pdpfb_priv *priv = platform_get_drvdata(pdev);
	int i;

	console_lock();
	for (i = 0; i < priv->stream_nr; ++i) {
		struct pdpfb_stream *stream = &priv->streams[i];
		if (!stream || !stream->probed)
			continue;
		fb_set_suspend(&stream->info, 1);
	}
	pdpfb_stop(priv);
	pdpfb_pdp_clk_disable(priv);
	console_unlock();

	return 0;
}

static int pdpfb_resume(struct platform_device *pdev)
{
	struct pdpfb_priv *priv = platform_get_drvdata(pdev);
	int i;

	console_lock();
	pdpfb_pdp_clk_enable(priv);
	pdpfb_configure(priv);
	for (i = 0; i < priv->stream_nr; ++i) {
		struct pdpfb_stream *stream = &priv->streams[i];
		if (!stream || !stream->probed)
			continue;
		fb_set_suspend(&stream->info, 0);
	}
	console_unlock();

	return 0;
}
#else
#define pdpfb_suspend NULL
#define pdpfb_resume NULL
#endif	/* CONFIG_PM */

static struct of_device_id pdpfb_match[] = {
	{ .compatible = "img,pdp-cashew" },
	{},
};
MODULE_DEVICE_TABLE(of, pdpfb_match);

static struct platform_driver pdpfb_driver = {
	.driver		= {
		.name		= "pdpfb",
		.owner		= THIS_MODULE,
		.of_match_table = pdpfb_match,
	},
	.probe		= pdpfb_probe,
	.remove		= pdpfb_remove,
	.shutdown	= pdpfb_shutdown,
	.suspend	= pdpfb_suspend,
	.resume		= pdpfb_resume,
};

static int pdpfb_init(void)
{
	return platform_driver_register(&pdpfb_driver);
}

static void pdpfb_exit(void)
{
	platform_driver_unregister(&pdpfb_driver);
}

module_init(pdpfb_init);
module_exit(pdpfb_exit);

MODULE_DESCRIPTION("PDP Framebuffer driver");
MODULE_AUTHOR("Will Newton <will.newton@imgtec.com>");
MODULE_AUTHOR("James Hogan <james.hogan@imgtec.com>");
MODULE_AUTHOR("Ed Blake <ed.blake@imgtec.com>");
MODULE_LICENSE("GPL v2");
#endif
